/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 8365
#define LARGE_STATE_COUNT 366
#define SYMBOL_COUNT 610
#define ALIAS_COUNT 3
#define TOKEN_COUNT 204
#define EXTERNAL_TOKEN_COUNT 167
#define FIELD_COUNT 14
#define MAX_ALIAS_SEQUENCE_LENGTH 9
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 61
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  anon_sym_SEMI = 1,
  anon_sym_COMMA = 2,
  anon_sym_DOT = 3,
  anon_sym_LPAREN = 4,
  anon_sym_RPAREN = 5,
  anon_sym_COLON = 6,
  anon_sym_EQ_GT = 7,
  anon_sym_QMARK = 8,
  anon_sym_LT_EQ = 9,
  sym_variable_assignment = 10,
  sym_condition_conversion = 11,
  anon_sym_EQ = 12,
  anon_sym_SLASH_EQ = 13,
  anon_sym_LT = 14,
  anon_sym_GT = 15,
  anon_sym_GT_EQ = 16,
  anon_sym_QMARK_EQ = 17,
  anon_sym_QMARK_SLASH_EQ = 18,
  anon_sym_QMARK_LT = 19,
  anon_sym_QMARK_LT_EQ = 20,
  anon_sym_QMARK_GT = 21,
  anon_sym_QMARK_GT_EQ = 22,
  anon_sym_PLUS = 23,
  anon_sym_DASH = 24,
  anon_sym_AMP = 25,
  anon_sym_STAR = 26,
  anon_sym_SLASH = 27,
  sym_exponentiate = 28,
  anon_sym_LT_LT = 29,
  anon_sym_GT_GT = 30,
  anon_sym_AT = 31,
  anon_sym_CARET = 32,
  anon_sym_SQUOTE = 33,
  anon_sym_LBRACK = 34,
  anon_sym_RBRACK = 35,
  anon_sym_PIPE = 36,
  sym_identifier = 37,
  sym_ABS = 38,
  sym_ACCESS = 39,
  sym_AFTER = 40,
  sym_ALIAS = 41,
  sym_ALL = 42,
  sym_AND = 43,
  sym_ARCHITECTURE = 44,
  sym_ARRAY = 45,
  sym_ASSERT = 46,
  sym_ASSUME = 47,
  sym_ATTRIBUTE = 48,
  sym_BEGIN = 49,
  sym_BLOCK = 50,
  sym_BODY = 51,
  sym_BUFFER = 52,
  sym_BUS = 53,
  sym_CASE = 54,
  sym_COMPONENT = 55,
  sym_CONFIGURATION = 56,
  sym_CONSTANT = 57,
  sym_CONTEXT = 58,
  sym_COVER = 59,
  sym_DEFAULT = 60,
  sym_DISCONNECT = 61,
  sym_DOWNTO = 62,
  sym_ELSE = 63,
  sym_ELSIF = 64,
  sym_END = 65,
  sym_ENTITY = 66,
  sym_EXIT = 67,
  sym_FAIRNESS = 68,
  sym_FILE = 69,
  sym_FOR = 70,
  sym_FORCE = 71,
  sym_FUNCTION = 72,
  sym_GENERATE = 73,
  sym_GENERIC = 74,
  sym_GROUP = 75,
  sym_GUARDED = 76,
  sym_IF = 77,
  sym_IMPURE = 78,
  sym_IN = 79,
  sym_INERTIAL = 80,
  sym_INOUT = 81,
  sym_IS = 82,
  sym_LABEL = 83,
  sym_LIBRARY = 84,
  sym_LINKAGE = 85,
  sym_LITERAL = 86,
  sym_LOOP = 87,
  sym_MAP = 88,
  sym_MOD = 89,
  sym_NAND = 90,
  sym_NEW = 91,
  sym_NEXT = 92,
  sym_NOR = 93,
  sym_NOT = 94,
  sym_NULL = 95,
  sym_OF = 96,
  sym_ON = 97,
  sym_OPEN = 98,
  sym_OR = 99,
  sym_OTHERS = 100,
  sym_OUT = 101,
  sym_PACKAGE = 102,
  sym_PARAMETER = 103,
  sym_PORT = 104,
  sym_POSTPONED = 105,
  sym_PROCEDURE = 106,
  sym_PROCESS = 107,
  sym_PROPERTY = 108,
  sym_PROTECTED = 109,
  sym_PRIVATE = 110,
  sym_PURE = 111,
  sym_RANGE = 112,
  sym_RECORD = 113,
  sym_REGISTER = 114,
  sym_REJECT = 115,
  sym_RELEASE = 116,
  sym_REM = 117,
  sym_REPORT = 118,
  sym_RESTRICT = 119,
  sym_RETURN = 120,
  sym_ROL = 121,
  sym_ROR = 122,
  sym_SELECT = 123,
  sym_SEQUENCE = 124,
  sym_SEVERITY = 125,
  sym_SIGNAL = 126,
  sym_SHARED = 127,
  sym_SLA = 128,
  sym_SLL = 129,
  sym_SRA = 130,
  sym_SRL = 131,
  sym_STRONG = 132,
  sym_SUBTYPE = 133,
  sym_THEN = 134,
  sym_TO = 135,
  sym_TRANSPORT = 136,
  sym_TYPE = 137,
  sym_UNAFFECTED = 138,
  sym_UNITS = 139,
  sym_UNTIL = 140,
  sym_USE = 141,
  sym_VARIABLE = 142,
  sym_VIEW = 143,
  sym_VMODE = 144,
  sym_VPKG = 145,
  sym_VPROP = 146,
  sym_VUNIT = 147,
  sym_WAIT = 148,
  sym_WHEN = 149,
  sym_WHILE = 150,
  sym_WITH = 151,
  sym_XNOR = 152,
  sym_XOR = 153,
  sym_reserved_end_marker = 154,
  sym_directive_body = 155,
  sym_directive_constant_builtin = 156,
  sym_directive_error = 157,
  sym_directive_protect = 158,
  sym_directive_warning = 159,
  sym__directive_newline = 160,
  sym__grave_accent = 161,
  sym_box = 162,
  sym_delimiter_end_marker = 163,
  sym_decimal_integer = 164,
  sym_decimal_float = 165,
  sym_based_base = 166,
  sym_based_integer = 167,
  sym_based_float = 168,
  sym_character_literal = 169,
  sym_string_literal = 170,
  sym_string_literal_std_logic = 171,
  sym_bit_string_length = 172,
  sym_bit_string_base = 173,
  sym_bit_string_value = 174,
  sym_operator_symbol = 175,
  sym__line_comment_start = 176,
  sym__block_comment_start = 177,
  sym__block_comment_end = 178,
  sym_comment_content = 179,
  sym_token_end_marker = 180,
  sym_attribute_function = 181,
  sym_attribute_impure_function = 182,
  sym_attribute_mode_view = 183,
  sym_attribute_pure_function = 184,
  sym_attribute_range = 185,
  sym_attribute_signal = 186,
  sym_attribute_subtype = 187,
  sym_attribute_type = 188,
  sym_attribute_value = 189,
  sym_library_attribute = 190,
  sym_library_constant = 191,
  sym_library_constant_boolean = 192,
  sym_library_constant_character = 193,
  sym_library_constant_debug = 194,
  sym_library_constant_env = 195,
  sym_library_constant_standard = 196,
  sym_library_constant_std_logic = 197,
  sym_library_constant_unit = 198,
  sym_library_function = 199,
  sym_library_namespace = 200,
  sym_library_type = 201,
  sym__end_of_file = 202,
  sym_error_sentinel = 203,
  sym_design_file = 204,
  sym_design_unit = 205,
  sym__context_item = 206,
  sym_library_clause = 207,
  sym_use_clause = 208,
  sym_context_reference = 209,
  sym_logical_name_list = 210,
  sym__logical_name = 211,
  sym_selected_name_list = 212,
  sym_selected_name = 213,
  sym__library_unit = 214,
  sym_entity_declaration = 215,
  sym_entity_head = 216,
  sym_port_clause = 217,
  sym_entity_body = 218,
  sym_end_entity = 219,
  sym_configuration_declaration = 220,
  sym_configuration_head = 221,
  sym_end_configuration = 222,
  sym_package_declaration = 223,
  sym_package_declaration_body = 224,
  sym_end_package = 225,
  sym_package_instantiation_declaration = 226,
  sym_interface_package_declaration = 227,
  sym_context_declaration = 228,
  sym_context_declaration_body = 229,
  sym_end_context = 230,
  sym_architecture_definition = 231,
  sym_architecture_head = 232,
  sym_end_architecture = 233,
  sym_package_definition = 234,
  sym_package_definition_body = 235,
  sym_end_package_body = 236,
  sym__entity_declarative_item = 237,
  sym__package_declarative_item = 238,
  sym__package_body_declarative_item = 239,
  sym__protected_type_declarative_item = 240,
  sym__protected_type_body_declarative_item = 241,
  sym__process_declarative_item = 242,
  sym__subprogram_declarative_item = 243,
  sym__block_declarative_item = 244,
  sym_subprogram_declaration = 245,
  sym_subprogram_definition = 246,
  sym_subprogram_head = 247,
  sym_subprogram_end = 248,
  sym_subprogram_instantiation_declaration = 249,
  sym_type_declaration = 250,
  sym_subtype_declaration = 251,
  sym_mode_view_declaration = 252,
  sym_mode_view_body = 253,
  sym_end_view = 254,
  sym_constant_declaration = 255,
  sym_signal_declaration = 256,
  sym_signal_kind = 257,
  sym_variable_declaration = 258,
  sym_initialiser = 259,
  sym_file_declaration = 260,
  sym_interface_file_declaration = 261,
  sym_alias_declaration = 262,
  sym_component_declaration = 263,
  sym_component_body = 264,
  sym_end_component = 265,
  sym_attribute_declaration = 266,
  sym_attribute_specification = 267,
  sym_disconnection_specification = 268,
  sym_group_template_declaration = 269,
  sym_group_declaration = 270,
  sym_private_variable_declaration = 271,
  sym__interface_type_indication = 272,
  sym_unspecified_type_indication = 273,
  sym_interface_type_declaration = 274,
  sym__incomplete_type_definition = 275,
  sym_private_incomplete_type_definition = 276,
  sym_scalar_incomplete_type_definition = 277,
  sym_discrete_incomplete_type_definition = 278,
  sym_integer_incomplete_type_definition = 279,
  sym_physical_incomplete_type_definition = 280,
  sym_floating_incomplete_type_definition = 281,
  sym_array_type_definition = 282,
  sym_array_index_incomplete_type_list = 283,
  sym__array_index_incomplete_type = 284,
  sym_index_constraint = 285,
  sym_index_subtype_definition = 286,
  sym_access_type_definition = 287,
  sym_access_incomplete_type_definition = 288,
  sym_file_type_definition = 289,
  sym_file_incomplete_type_definition = 290,
  sym_subtype_indication = 291,
  sym_resolution_indication = 292,
  sym__element_resolution = 293,
  sym_record_resolution = 294,
  sym_record_element_resolution = 295,
  sym__incomplete_subtype_indication = 296,
  sym__incomplete_type_mark = 297,
  sym_concurrent_block = 298,
  sym_sequential_block = 299,
  sym_generate_body = 300,
  sym_generate_head = 301,
  sym_case_generate_alternative = 302,
  sym_case_generate_body = 303,
  sym_generate_block = 304,
  sym_generate_block_end = 305,
  sym__entity_statement = 306,
  sym__concurrent_statement = 307,
  sym__sequential_statement = 308,
  sym_block_statement = 309,
  sym_guard_condition = 310,
  sym_block_head = 311,
  sym_end_block = 312,
  sym_sequential_block_statement = 313,
  sym_sequential_block_head = 314,
  sym_component_instantiation_statement = 315,
  sym_instantiated_unit = 316,
  sym_process_statement = 317,
  sym_sensitivity_specification = 318,
  sym_process_head = 319,
  sym_end_process = 320,
  sym_case_generate_statement = 321,
  sym_case_generate_block = 322,
  sym_for_generate_statement = 323,
  sym_if_generate_statement = 324,
  sym_if_generate = 325,
  sym__elsif_else_generate = 326,
  sym_elsif_generate = 327,
  sym_else_generate = 328,
  sym_end_generate = 329,
  sym_assertion_statement = 330,
  sym_concurrent_assertion_statement = 331,
  sym_assertion = 332,
  sym_report_expression = 333,
  sym_severity_expression = 334,
  sym_case_statement = 335,
  sym_case_expression = 336,
  sym_case_body = 337,
  sym_end_case = 338,
  sym_exit_statement = 339,
  sym_when_expression = 340,
  sym_if_statement_block = 341,
  sym_if_statement = 342,
  sym__elsif_else_statement = 343,
  sym_elsif_statement = 344,
  sym_else_statement = 345,
  sym_if_statement_body = 346,
  sym_end_if = 347,
  sym_loop_statement = 348,
  sym__iteration_scheme = 349,
  sym_loop_body = 350,
  sym_end_loop = 351,
  sym_next_statement = 352,
  sym_null_statement = 353,
  sym_procedure_call_statement = 354,
  sym_concurrent_procedure_call_statement = 355,
  sym_report_statement = 356,
  sym_return_statement = 357,
  sym_signal_assignment = 358,
  sym_simple_waveform_assignment = 359,
  sym_simple_variable_assignment = 360,
  sym_concurrent_simple_signal_assignment = 361,
  sym_simple_force_assignment = 362,
  sym_simple_release_assignment = 363,
  sym_conditional_signal_assignment = 364,
  sym_concurrent_conditional_signal_assignment = 365,
  sym_selected_waveform_assignment = 366,
  sym_concurrent_selected_signal_assignment = 367,
  sym_selected_force_assignment = 368,
  sym_selected_variable_assignment = 369,
  sym_with_expression = 370,
  sym_select_target = 371,
  sym_wait_statement = 372,
  sym_conditional_expression = 373,
  sym_else_expression = 374,
  sym__expression = 375,
  sym_condition_expression = 376,
  sym_logical_expression = 377,
  sym_relational_expression = 378,
  sym_shift_expression = 379,
  sym_simple_expression = 380,
  sym_unary_operator = 381,
  sym_logical_operator = 382,
  sym_relational_operator = 383,
  sym_shift_operator = 384,
  sym_sign = 385,
  sym_adding_operator = 386,
  sym_multiplying_operator = 387,
  sym__primary = 388,
  sym_name = 389,
  sym__direct_name = 390,
  sym__label = 391,
  sym__attribute = 392,
  sym__unit = 393,
  sym__identifier = 394,
  sym__name_selector = 395,
  sym__external_name = 396,
  sym_external_constant_name = 397,
  sym_external_signal_name = 398,
  sym_external_variable_name = 399,
  sym__external_pathname = 400,
  sym_package_pathname = 401,
  sym_absolute_pathname = 402,
  sym_relative_pathname = 403,
  sym_partial_pathname = 404,
  sym_pathname_element = 405,
  sym_function_call = 406,
  sym_parenthesis_group = 407,
  sym_association_or_range_list = 408,
  sym__association_or_range = 409,
  sym_association_element = 410,
  sym__actual_part = 411,
  sym_attribute = 412,
  sym__attribute_designator = 413,
  sym_signature = 414,
  sym_selection = 415,
  sym__suffix = 416,
  sym__literal = 417,
  sym_bit_string_literal = 418,
  sym__abstract_literal = 419,
  sym_based_literal = 420,
  sym_allocator = 421,
  sym_parenthesis_expression = 422,
  sym_element_association_list = 423,
  sym_element_association = 424,
  sym__element = 425,
  sym_choices = 426,
  sym__range = 427,
  sym_simple_range = 428,
  sym__direction = 429,
  sym__tool_directive = 430,
  sym__directive = 431,
  sym_user_directive = 432,
  sym_protect_directive = 433,
  sym_warning_directive = 434,
  sym_error_directive = 435,
  sym_if_conditional_analysis = 436,
  sym_elsif_conditional_analysis = 437,
  sym_else_conditional_analysis = 438,
  sym_end_conditional_analysis = 439,
  sym_conditional_analysis_expression = 440,
  sym_conditional_analysis_relation = 441,
  sym__conditional_analysis_operator = 442,
  sym__conditional_analysis_identifier = 443,
  sym__configuration_declarative_item = 444,
  sym__configuration_item = 445,
  sym_block_configuration = 446,
  sym_component_configuration = 447,
  sym_end_for = 448,
  sym_generic_interface_list = 449,
  sym__generic_interface_declaration = 450,
  sym_generic_interface_declaration = 451,
  sym_interface_list = 452,
  sym__interface_declaration = 453,
  sym_interface_declaration = 454,
  sym_interface_constant_declaration = 455,
  sym_interface_signal_declaration = 456,
  sym_interface_variable_declaration = 457,
  sym_interface_subprogram_declaration = 458,
  sym__interface_subprogram_specification = 459,
  sym_procedure_specification = 460,
  sym_interface_procedure_specification = 461,
  sym_function_specification = 462,
  sym_interface_function_specification = 463,
  sym__interface_package_generic_map_aspect = 464,
  sym_generic_map_any = 465,
  sym_generic_map_default = 466,
  sym_generic_map_aspect = 467,
  sym_port_map_aspect = 468,
  sym_association_list = 469,
  sym__interface_subprogram_default = 470,
  sym__designator = 471,
  sym_generic_identifier_list = 472,
  sym_constant_identifier_list = 473,
  sym_identifier_list = 474,
  sym__mode_indication = 475,
  sym_simple_mode_indication = 476,
  sym__mode_view_indication = 477,
  sym_record_mode_view_indication = 478,
  sym_array_mode_view_indication = 479,
  sym_mode_view_element_definition = 480,
  sym_record_element_list = 481,
  sym__element_mode_indication = 482,
  sym__element_mode_view_indication = 483,
  sym_element_record_mode_view_indication = 484,
  sym_element_array_mode_view_indication = 485,
  sym_mode = 486,
  sym_range_constraint = 487,
  sym_group_constituent_list = 488,
  sym_entity_class_entry_list = 489,
  sym_entity_class_entry = 490,
  sym_guarded_signal_specification = 491,
  sym_signal_list = 492,
  sym_entity_specification = 493,
  sym__entity_class = 494,
  sym_entity_name_list = 495,
  sym_entity_designator = 496,
  sym__entity_tag = 497,
  sym__alias_designator = 498,
  sym_file_open_information = 499,
  sym_file_logical_name = 500,
  sym_package_header = 501,
  sym_generic_clause = 502,
  sym_subprogram_header = 503,
  sym__subprogram_specification = 504,
  sym_parameter_list_specification = 505,
  sym__type_definition = 506,
  sym_protected_type_instantiation_definition = 507,
  sym_protected_type_declaration = 508,
  sym_protected_type_declaration_end = 509,
  sym_protected_type_header = 510,
  sym_protected_type_body = 511,
  sym_protected_type_body_end = 512,
  sym_while_loop = 513,
  sym_for_loop = 514,
  sym_parameter_specification = 515,
  sym_case_statement_alternative = 516,
  sym_selected_waveforms = 517,
  sym_waveform = 518,
  sym_waveform_element = 519,
  sym_force_mode = 520,
  sym_selected_expressions = 521,
  sym_when_element = 522,
  sym_conditional_waveforms = 523,
  sym_else_waveform = 524,
  sym_delay_mechanism = 525,
  sym__target = 526,
  sym_aggregate = 527,
  sym_conditional_or_unaffected_expression = 528,
  sym_else_expression_or_unaffected = 529,
  sym__expression_or_unaffected = 530,
  sym_label_declaration = 531,
  sym_sensitivity_clause = 532,
  sym_sensitivity_list = 533,
  sym_condition_clause = 534,
  sym_timeout_clause = 535,
  sym_physical_type_definition = 536,
  sym_primary_unit_declaration = 537,
  sym_secondary_unit_declaration = 538,
  sym_end_units = 539,
  sym_enumeration_type_definition = 540,
  sym_enumeration_literal = 541,
  sym_record_type_definition = 542,
  sym_end_record = 543,
  sym_element_declaration = 544,
  sym__process_sensitivity_list = 545,
  sym_configuration_specification = 546,
  sym_verification_unit_binding_indication = 547,
  sym_verification_unit_list = 548,
  sym_component_specification = 549,
  sym_instantiation_list = 550,
  sym_binding_indication = 551,
  sym_entity_aspect = 552,
  sym_line_comment = 553,
  sym_block_comment = 554,
  aux_sym_design_file_repeat1 = 555,
  aux_sym_design_unit_repeat1 = 556,
  aux_sym_design_unit_repeat2 = 557,
  aux_sym_design_unit_repeat3 = 558,
  aux_sym_logical_name_list_repeat1 = 559,
  aux_sym_selected_name_list_repeat1 = 560,
  aux_sym_selected_name_repeat1 = 561,
  aux_sym_entity_head_repeat1 = 562,
  aux_sym_entity_body_repeat1 = 563,
  aux_sym_configuration_head_repeat1 = 564,
  aux_sym_configuration_head_repeat2 = 565,
  aux_sym_package_declaration_body_repeat1 = 566,
  aux_sym_package_definition_body_repeat1 = 567,
  aux_sym_subprogram_head_repeat1 = 568,
  aux_sym_mode_view_body_repeat1 = 569,
  aux_sym_array_index_incomplete_type_list_repeat1 = 570,
  aux_sym_index_constraint_repeat1 = 571,
  aux_sym_record_resolution_repeat1 = 572,
  aux_sym_sequential_block_repeat1 = 573,
  aux_sym_sequential_block_head_repeat1 = 574,
  aux_sym_case_generate_block_repeat1 = 575,
  aux_sym_case_body_repeat1 = 576,
  aux_sym_conditional_expression_repeat1 = 577,
  aux_sym_name_repeat1 = 578,
  aux_sym_package_pathname_repeat1 = 579,
  aux_sym_relative_pathname_repeat1 = 580,
  aux_sym_partial_pathname_repeat1 = 581,
  aux_sym_association_or_range_list_repeat1 = 582,
  aux_sym_signature_repeat1 = 583,
  aux_sym_element_association_list_repeat1 = 584,
  aux_sym_user_directive_repeat1 = 585,
  aux_sym_conditional_analysis_expression_repeat1 = 586,
  aux_sym_block_configuration_repeat1 = 587,
  aux_sym_block_configuration_repeat2 = 588,
  aux_sym_generic_interface_list_repeat1 = 589,
  aux_sym_interface_list_repeat1 = 590,
  aux_sym_association_list_repeat1 = 591,
  aux_sym_generic_identifier_list_repeat1 = 592,
  aux_sym_constant_identifier_list_repeat1 = 593,
  aux_sym_identifier_list_repeat1 = 594,
  aux_sym_group_constituent_list_repeat1 = 595,
  aux_sym_entity_class_entry_list_repeat1 = 596,
  aux_sym_entity_name_list_repeat1 = 597,
  aux_sym_protected_type_declaration_repeat1 = 598,
  aux_sym_protected_type_body_repeat1 = 599,
  aux_sym_selected_waveforms_repeat1 = 600,
  aux_sym_waveform_repeat1 = 601,
  aux_sym_selected_expressions_repeat1 = 602,
  aux_sym_conditional_waveforms_repeat1 = 603,
  aux_sym_aggregate_repeat1 = 604,
  aux_sym_conditional_or_unaffected_expression_repeat1 = 605,
  aux_sym_physical_type_definition_repeat1 = 606,
  aux_sym_enumeration_type_definition_repeat1 = 607,
  aux_sym_record_type_definition_repeat1 = 608,
  aux_sym_instantiation_list_repeat1 = 609,
  alias_sym_attribute_identifier = 610,
  alias_sym_label = 611,
  alias_sym_unit = 612,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [anon_sym_SEMI] = ";",
  [anon_sym_COMMA] = ",",
  [anon_sym_DOT] = ".",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_COLON] = ":",
  [anon_sym_EQ_GT] = "=>",
  [anon_sym_QMARK] = "\?",
  [anon_sym_LT_EQ] = "<=",
  [sym_variable_assignment] = "variable_assignment",
  [sym_condition_conversion] = "condition_conversion",
  [anon_sym_EQ] = "=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_LT] = "<",
  [anon_sym_GT] = ">",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_QMARK_EQ] = "\?=",
  [anon_sym_QMARK_SLASH_EQ] = "\?/=",
  [anon_sym_QMARK_LT] = "\?<",
  [anon_sym_QMARK_LT_EQ] = "\?<=",
  [anon_sym_QMARK_GT] = "\?>",
  [anon_sym_QMARK_GT_EQ] = "\?>=",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_AMP] = "&",
  [anon_sym_STAR] = "*",
  [anon_sym_SLASH] = "/",
  [sym_exponentiate] = "exponentiate",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_AT] = "@",
  [anon_sym_CARET] = "^",
  [anon_sym_SQUOTE] = "'",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_PIPE] = "|",
  [sym_identifier] = "identifier",
  [sym_ABS] = "abs",
  [sym_ACCESS] = "access",
  [sym_AFTER] = "after",
  [sym_ALIAS] = "alias",
  [sym_ALL] = "ALL",
  [sym_AND] = "and",
  [sym_ARCHITECTURE] = "architecture",
  [sym_ARRAY] = "array",
  [sym_ASSERT] = "assert",
  [sym_ASSUME] = "ASSUME",
  [sym_ATTRIBUTE] = "attribute",
  [sym_BEGIN] = "begin",
  [sym_BLOCK] = "block",
  [sym_BODY] = "body",
  [sym_BUFFER] = "buffer",
  [sym_BUS] = "bus",
  [sym_CASE] = "case",
  [sym_COMPONENT] = "component",
  [sym_CONFIGURATION] = "configuration",
  [sym_CONSTANT] = "constant",
  [sym_CONTEXT] = "context",
  [sym_COVER] = "COVER",
  [sym_DEFAULT] = "DEFAULT",
  [sym_DISCONNECT] = "disconnect",
  [sym_DOWNTO] = "downto",
  [sym_ELSE] = "else",
  [sym_ELSIF] = "elsif",
  [sym_END] = "end",
  [sym_ENTITY] = "entity",
  [sym_EXIT] = "exit",
  [sym_FAIRNESS] = "FAIRNESS",
  [sym_FILE] = "file",
  [sym_FOR] = "for",
  [sym_FORCE] = "force",
  [sym_FUNCTION] = "function",
  [sym_GENERATE] = "generate",
  [sym_GENERIC] = "generic",
  [sym_GROUP] = "group",
  [sym_GUARDED] = "guarded",
  [sym_IF] = "if",
  [sym_IMPURE] = "impure",
  [sym_IN] = "in",
  [sym_INERTIAL] = "inertial",
  [sym_INOUT] = "inout",
  [sym_IS] = "is",
  [sym_LABEL] = "label",
  [sym_LIBRARY] = "library",
  [sym_LINKAGE] = "linkage",
  [sym_LITERAL] = "literal",
  [sym_LOOP] = "loop",
  [sym_MAP] = "map",
  [sym_MOD] = "mod",
  [sym_NAND] = "nand",
  [sym_NEW] = "new",
  [sym_NEXT] = "next",
  [sym_NOR] = "nor",
  [sym_NOT] = "not",
  [sym_NULL] = "null",
  [sym_OF] = "of",
  [sym_ON] = "on",
  [sym_OPEN] = "OPEN",
  [sym_OR] = "or",
  [sym_OTHERS] = "OTHERS",
  [sym_OUT] = "out",
  [sym_PACKAGE] = "package",
  [sym_PARAMETER] = "parameter",
  [sym_PORT] = "port",
  [sym_POSTPONED] = "postponed",
  [sym_PROCEDURE] = "procedure",
  [sym_PROCESS] = "process",
  [sym_PROPERTY] = "property",
  [sym_PROTECTED] = "protected",
  [sym_PRIVATE] = "private",
  [sym_PURE] = "pure",
  [sym_RANGE] = "range",
  [sym_RECORD] = "record",
  [sym_REGISTER] = "register",
  [sym_REJECT] = "reject",
  [sym_RELEASE] = "release",
  [sym_REM] = "rem",
  [sym_REPORT] = "report",
  [sym_RESTRICT] = "RESTRICT",
  [sym_RETURN] = "return",
  [sym_ROL] = "rol",
  [sym_ROR] = "ror",
  [sym_SELECT] = "select",
  [sym_SEQUENCE] = "sequence",
  [sym_SEVERITY] = "severity",
  [sym_SIGNAL] = "signal",
  [sym_SHARED] = "shared",
  [sym_SLA] = "sla",
  [sym_SLL] = "sll",
  [sym_SRA] = "sra",
  [sym_SRL] = "srl",
  [sym_STRONG] = "STRONG",
  [sym_SUBTYPE] = "subtype",
  [sym_THEN] = "then",
  [sym_TO] = "to",
  [sym_TRANSPORT] = "transport",
  [sym_TYPE] = "type",
  [sym_UNAFFECTED] = "unaffected",
  [sym_UNITS] = "units",
  [sym_UNTIL] = "until",
  [sym_USE] = "use",
  [sym_VARIABLE] = "variable",
  [sym_VIEW] = "view",
  [sym_VMODE] = "VMODE",
  [sym_VPKG] = "VPKG",
  [sym_VPROP] = "VPROP",
  [sym_VUNIT] = "vunit",
  [sym_WAIT] = "wait",
  [sym_WHEN] = "when",
  [sym_WHILE] = "while",
  [sym_WITH] = "with",
  [sym_XNOR] = "xnor",
  [sym_XOR] = "xor",
  [sym_reserved_end_marker] = "reserved_end_marker",
  [sym_directive_body] = "directive_body",
  [sym_directive_constant_builtin] = "directive_constant_builtin",
  [sym_directive_error] = "directive_error",
  [sym_directive_protect] = "directive_protect",
  [sym_directive_warning] = "directive_warning",
  [sym__directive_newline] = "_directive_newline",
  [sym__grave_accent] = "_grave_accent",
  [sym_box] = "<>",
  [sym_delimiter_end_marker] = "delimiter_end_marker",
  [sym_decimal_integer] = "decimal_integer",
  [sym_decimal_float] = "decimal_float",
  [sym_based_base] = "based_base",
  [sym_based_integer] = "based_integer",
  [sym_based_float] = "based_float",
  [sym_character_literal] = "character_literal",
  [sym_string_literal] = "string_literal",
  [sym_string_literal_std_logic] = "string_literal_std_logic",
  [sym_bit_string_length] = "bit_string_length",
  [sym_bit_string_base] = "bit_string_base",
  [sym_bit_string_value] = "bit_string_value",
  [sym_operator_symbol] = "operator_symbol",
  [sym__line_comment_start] = "_line_comment_start",
  [sym__block_comment_start] = "_block_comment_start",
  [sym__block_comment_end] = "_block_comment_end",
  [sym_comment_content] = "comment_content",
  [sym_token_end_marker] = "token_end_marker",
  [sym_attribute_function] = "attribute_function",
  [sym_attribute_impure_function] = "attribute_impure_function",
  [sym_attribute_mode_view] = "attribute_mode_view",
  [sym_attribute_pure_function] = "attribute_pure_function",
  [sym_attribute_range] = "attribute_range",
  [sym_attribute_signal] = "attribute_signal",
  [sym_attribute_subtype] = "attribute_subtype",
  [sym_attribute_type] = "attribute_type",
  [sym_attribute_value] = "attribute_value",
  [sym_library_attribute] = "library_attribute",
  [sym_library_constant] = "library_constant",
  [sym_library_constant_boolean] = "library_constant_boolean",
  [sym_library_constant_character] = "library_constant_character",
  [sym_library_constant_debug] = "library_constant_debug",
  [sym_library_constant_env] = "library_constant_env",
  [sym_library_constant_standard] = "library_constant_standard",
  [sym_library_constant_std_logic] = "library_constant_std_logic",
  [sym_library_constant_unit] = "library_constant_unit",
  [sym_library_function] = "library_function",
  [sym_library_namespace] = "library_namespace",
  [sym_library_type] = "library_type",
  [sym__end_of_file] = "_end_of_file",
  [sym_error_sentinel] = "error_sentinel",
  [sym_design_file] = "design_file",
  [sym_design_unit] = "design_unit",
  [sym__context_item] = "_context_item",
  [sym_library_clause] = "library_clause",
  [sym_use_clause] = "use_clause",
  [sym_context_reference] = "context_reference",
  [sym_logical_name_list] = "logical_name_list",
  [sym__logical_name] = "_logical_name",
  [sym_selected_name_list] = "selected_name_list",
  [sym_selected_name] = "selected_name",
  [sym__library_unit] = "_library_unit",
  [sym_entity_declaration] = "entity_declaration",
  [sym_entity_head] = "entity_head",
  [sym_port_clause] = "port_clause",
  [sym_entity_body] = "entity_body",
  [sym_end_entity] = "end_entity",
  [sym_configuration_declaration] = "configuration_declaration",
  [sym_configuration_head] = "configuration_head",
  [sym_end_configuration] = "end_configuration",
  [sym_package_declaration] = "package_declaration",
  [sym_package_declaration_body] = "package_declaration_body",
  [sym_end_package] = "end_package",
  [sym_package_instantiation_declaration] = "package_instantiation_declaration",
  [sym_interface_package_declaration] = "interface_package_declaration",
  [sym_context_declaration] = "context_declaration",
  [sym_context_declaration_body] = "context_declaration_body",
  [sym_end_context] = "end_context",
  [sym_architecture_definition] = "architecture_definition",
  [sym_architecture_head] = "architecture_head",
  [sym_end_architecture] = "end_architecture",
  [sym_package_definition] = "package_definition",
  [sym_package_definition_body] = "package_definition_body",
  [sym_end_package_body] = "end_package_body",
  [sym__entity_declarative_item] = "_entity_declarative_item",
  [sym__package_declarative_item] = "_package_declarative_item",
  [sym__package_body_declarative_item] = "_package_body_declarative_item",
  [sym__protected_type_declarative_item] = "_protected_type_declarative_item",
  [sym__protected_type_body_declarative_item] = "_protected_type_body_declarative_item",
  [sym__process_declarative_item] = "_process_declarative_item",
  [sym__subprogram_declarative_item] = "_subprogram_declarative_item",
  [sym__block_declarative_item] = "_block_declarative_item",
  [sym_subprogram_declaration] = "subprogram_declaration",
  [sym_subprogram_definition] = "subprogram_definition",
  [sym_subprogram_head] = "subprogram_head",
  [sym_subprogram_end] = "subprogram_end",
  [sym_subprogram_instantiation_declaration] = "subprogram_instantiation_declaration",
  [sym_type_declaration] = "type_declaration",
  [sym_subtype_declaration] = "subtype_declaration",
  [sym_mode_view_declaration] = "mode_view_declaration",
  [sym_mode_view_body] = "mode_view_body",
  [sym_end_view] = "end_view",
  [sym_constant_declaration] = "constant_declaration",
  [sym_signal_declaration] = "signal_declaration",
  [sym_signal_kind] = "signal_kind",
  [sym_variable_declaration] = "variable_declaration",
  [sym_initialiser] = "initialiser",
  [sym_file_declaration] = "file_declaration",
  [sym_interface_file_declaration] = "interface_file_declaration",
  [sym_alias_declaration] = "alias_declaration",
  [sym_component_declaration] = "component_declaration",
  [sym_component_body] = "component_body",
  [sym_end_component] = "end_component",
  [sym_attribute_declaration] = "attribute_declaration",
  [sym_attribute_specification] = "attribute_specification",
  [sym_disconnection_specification] = "disconnection_specification",
  [sym_group_template_declaration] = "group_template_declaration",
  [sym_group_declaration] = "group_declaration",
  [sym_private_variable_declaration] = "private_variable_declaration",
  [sym__interface_type_indication] = "_interface_type_indication",
  [sym_unspecified_type_indication] = "unspecified_type_indication",
  [sym_interface_type_declaration] = "interface_type_declaration",
  [sym__incomplete_type_definition] = "_incomplete_type_definition",
  [sym_private_incomplete_type_definition] = "private_incomplete_type_definition",
  [sym_scalar_incomplete_type_definition] = "scalar_incomplete_type_definition",
  [sym_discrete_incomplete_type_definition] = "discrete_incomplete_type_definition",
  [sym_integer_incomplete_type_definition] = "integer_incomplete_type_definition",
  [sym_physical_incomplete_type_definition] = "physical_incomplete_type_definition",
  [sym_floating_incomplete_type_definition] = "floating_incomplete_type_definition",
  [sym_array_type_definition] = "array_type_definition",
  [sym_array_index_incomplete_type_list] = "array_index_incomplete_type_list",
  [sym__array_index_incomplete_type] = "_array_index_incomplete_type",
  [sym_index_constraint] = "index_constraint",
  [sym_index_subtype_definition] = "index_subtype_definition",
  [sym_access_type_definition] = "access_type_definition",
  [sym_access_incomplete_type_definition] = "access_incomplete_type_definition",
  [sym_file_type_definition] = "file_type_definition",
  [sym_file_incomplete_type_definition] = "file_incomplete_type_definition",
  [sym_subtype_indication] = "subtype_indication",
  [sym_resolution_indication] = "resolution_indication",
  [sym__element_resolution] = "_element_resolution",
  [sym_record_resolution] = "record_resolution",
  [sym_record_element_resolution] = "record_element_resolution",
  [sym__incomplete_subtype_indication] = "_incomplete_subtype_indication",
  [sym__incomplete_type_mark] = "_incomplete_type_mark",
  [sym_concurrent_block] = "concurrent_block",
  [sym_sequential_block] = "sequential_block",
  [sym_generate_body] = "generate_body",
  [sym_generate_head] = "generate_head",
  [sym_case_generate_alternative] = "case_generate_alternative",
  [sym_case_generate_body] = "case_generate_body",
  [sym_generate_block] = "generate_block",
  [sym_generate_block_end] = "generate_block_end",
  [sym__entity_statement] = "_entity_statement",
  [sym__concurrent_statement] = "_concurrent_statement",
  [sym__sequential_statement] = "_sequential_statement",
  [sym_block_statement] = "block_statement",
  [sym_guard_condition] = "guard_condition",
  [sym_block_head] = "block_head",
  [sym_end_block] = "end_block",
  [sym_sequential_block_statement] = "sequential_block_statement",
  [sym_sequential_block_head] = "sequential_block_head",
  [sym_component_instantiation_statement] = "component_instantiation_statement",
  [sym_instantiated_unit] = "instantiated_unit",
  [sym_process_statement] = "process_statement",
  [sym_sensitivity_specification] = "sensitivity_specification",
  [sym_process_head] = "process_head",
  [sym_end_process] = "end_process",
  [sym_case_generate_statement] = "case_generate_statement",
  [sym_case_generate_block] = "case_generate_block",
  [sym_for_generate_statement] = "for_generate_statement",
  [sym_if_generate_statement] = "if_generate_statement",
  [sym_if_generate] = "if_generate",
  [sym__elsif_else_generate] = "_elsif_else_generate",
  [sym_elsif_generate] = "elsif_generate",
  [sym_else_generate] = "else_generate",
  [sym_end_generate] = "end_generate",
  [sym_assertion_statement] = "assertion_statement",
  [sym_concurrent_assertion_statement] = "concurrent_assertion_statement",
  [sym_assertion] = "assertion",
  [sym_report_expression] = "report_expression",
  [sym_severity_expression] = "severity_expression",
  [sym_case_statement] = "case_statement",
  [sym_case_expression] = "case_expression",
  [sym_case_body] = "case_body",
  [sym_end_case] = "end_case",
  [sym_exit_statement] = "exit_statement",
  [sym_when_expression] = "when_expression",
  [sym_if_statement_block] = "if_statement_block",
  [sym_if_statement] = "if_statement",
  [sym__elsif_else_statement] = "_elsif_else_statement",
  [sym_elsif_statement] = "elsif_statement",
  [sym_else_statement] = "else_statement",
  [sym_if_statement_body] = "if_statement_body",
  [sym_end_if] = "end_if",
  [sym_loop_statement] = "loop_statement",
  [sym__iteration_scheme] = "_iteration_scheme",
  [sym_loop_body] = "loop_body",
  [sym_end_loop] = "end_loop",
  [sym_next_statement] = "next_statement",
  [sym_null_statement] = "null_statement",
  [sym_procedure_call_statement] = "procedure_call_statement",
  [sym_concurrent_procedure_call_statement] = "concurrent_procedure_call_statement",
  [sym_report_statement] = "report_statement",
  [sym_return_statement] = "return_statement",
  [sym_signal_assignment] = "signal_assignment",
  [sym_simple_waveform_assignment] = "simple_waveform_assignment",
  [sym_simple_variable_assignment] = "simple_variable_assignment",
  [sym_concurrent_simple_signal_assignment] = "concurrent_simple_signal_assignment",
  [sym_simple_force_assignment] = "simple_force_assignment",
  [sym_simple_release_assignment] = "simple_release_assignment",
  [sym_conditional_signal_assignment] = "conditional_signal_assignment",
  [sym_concurrent_conditional_signal_assignment] = "concurrent_conditional_signal_assignment",
  [sym_selected_waveform_assignment] = "selected_waveform_assignment",
  [sym_concurrent_selected_signal_assignment] = "concurrent_selected_signal_assignment",
  [sym_selected_force_assignment] = "selected_force_assignment",
  [sym_selected_variable_assignment] = "selected_variable_assignment",
  [sym_with_expression] = "with_expression",
  [sym_select_target] = "select_target",
  [sym_wait_statement] = "wait_statement",
  [sym_conditional_expression] = "conditional_expression",
  [sym_else_expression] = "else_expression",
  [sym__expression] = "_expression",
  [sym_condition_expression] = "condition_expression",
  [sym_logical_expression] = "logical_expression",
  [sym_relational_expression] = "relational_expression",
  [sym_shift_expression] = "shift_expression",
  [sym_simple_expression] = "simple_expression",
  [sym_unary_operator] = "unary_operator",
  [sym_logical_operator] = "logical_operator",
  [sym_relational_operator] = "relational_operator",
  [sym_shift_operator] = "shift_operator",
  [sym_sign] = "sign",
  [sym_adding_operator] = "adding_operator",
  [sym_multiplying_operator] = "multiplying_operator",
  [sym__primary] = "_primary",
  [sym_name] = "name",
  [sym__direct_name] = "_direct_name",
  [sym__label] = "_label",
  [sym__attribute] = "_attribute",
  [sym__unit] = "_unit",
  [sym__identifier] = "_identifier",
  [sym__name_selector] = "_name_selector",
  [sym__external_name] = "_external_name",
  [sym_external_constant_name] = "external_constant_name",
  [sym_external_signal_name] = "external_signal_name",
  [sym_external_variable_name] = "external_variable_name",
  [sym__external_pathname] = "_external_pathname",
  [sym_package_pathname] = "package_pathname",
  [sym_absolute_pathname] = "absolute_pathname",
  [sym_relative_pathname] = "relative_pathname",
  [sym_partial_pathname] = "partial_pathname",
  [sym_pathname_element] = "pathname_element",
  [sym_function_call] = "function_call",
  [sym_parenthesis_group] = "parenthesis_group",
  [sym_association_or_range_list] = "association_or_range_list",
  [sym__association_or_range] = "_association_or_range",
  [sym_association_element] = "association_element",
  [sym__actual_part] = "_actual_part",
  [sym_attribute] = "attribute",
  [sym__attribute_designator] = "_attribute_designator",
  [sym_signature] = "signature",
  [sym_selection] = "selection",
  [sym__suffix] = "_suffix",
  [sym__literal] = "_literal",
  [sym_bit_string_literal] = "bit_string_literal",
  [sym__abstract_literal] = "_abstract_literal",
  [sym_based_literal] = "based_literal",
  [sym_allocator] = "allocator",
  [sym_parenthesis_expression] = "parenthesis_expression",
  [sym_element_association_list] = "element_association_list",
  [sym_element_association] = "element_association",
  [sym__element] = "_element",
  [sym_choices] = "choices",
  [sym__range] = "_range",
  [sym_simple_range] = "simple_range",
  [sym__direction] = "_direction",
  [sym__tool_directive] = "_tool_directive",
  [sym__directive] = "_directive",
  [sym_user_directive] = "user_directive",
  [sym_protect_directive] = "protect_directive",
  [sym_warning_directive] = "warning_directive",
  [sym_error_directive] = "error_directive",
  [sym_if_conditional_analysis] = "if_conditional_analysis",
  [sym_elsif_conditional_analysis] = "elsif_conditional_analysis",
  [sym_else_conditional_analysis] = "else_conditional_analysis",
  [sym_end_conditional_analysis] = "end_conditional_analysis",
  [sym_conditional_analysis_expression] = "conditional_analysis_expression",
  [sym_conditional_analysis_relation] = "conditional_analysis_relation",
  [sym__conditional_analysis_operator] = "_conditional_analysis_operator",
  [sym__conditional_analysis_identifier] = "_conditional_analysis_identifier",
  [sym__configuration_declarative_item] = "_configuration_declarative_item",
  [sym__configuration_item] = "_configuration_item",
  [sym_block_configuration] = "block_configuration",
  [sym_component_configuration] = "component_configuration",
  [sym_end_for] = "end_for",
  [sym_generic_interface_list] = "interface_list",
  [sym__generic_interface_declaration] = "_generic_interface_declaration",
  [sym_generic_interface_declaration] = "interface_declaration",
  [sym_interface_list] = "interface_list",
  [sym__interface_declaration] = "_interface_declaration",
  [sym_interface_declaration] = "interface_declaration",
  [sym_interface_constant_declaration] = "interface_constant_declaration",
  [sym_interface_signal_declaration] = "interface_signal_declaration",
  [sym_interface_variable_declaration] = "interface_variable_declaration",
  [sym_interface_subprogram_declaration] = "interface_subprogram_declaration",
  [sym__interface_subprogram_specification] = "_interface_subprogram_specification",
  [sym_procedure_specification] = "procedure_specification",
  [sym_interface_procedure_specification] = "interface_procedure_specification",
  [sym_function_specification] = "function_specification",
  [sym_interface_function_specification] = "interface_function_specification",
  [sym__interface_package_generic_map_aspect] = "_interface_package_generic_map_aspect",
  [sym_generic_map_any] = "generic_map_any",
  [sym_generic_map_default] = "generic_map_default",
  [sym_generic_map_aspect] = "generic_map_aspect",
  [sym_port_map_aspect] = "port_map_aspect",
  [sym_association_list] = "association_list",
  [sym__interface_subprogram_default] = "_interface_subprogram_default",
  [sym__designator] = "_designator",
  [sym_generic_identifier_list] = "identifier_list",
  [sym_constant_identifier_list] = "identifier_list",
  [sym_identifier_list] = "identifier_list",
  [sym__mode_indication] = "_mode_indication",
  [sym_simple_mode_indication] = "simple_mode_indication",
  [sym__mode_view_indication] = "_mode_view_indication",
  [sym_record_mode_view_indication] = "record_mode_view_indication",
  [sym_array_mode_view_indication] = "array_mode_view_indication",
  [sym_mode_view_element_definition] = "mode_view_element_definition",
  [sym_record_element_list] = "record_element_list",
  [sym__element_mode_indication] = "_element_mode_indication",
  [sym__element_mode_view_indication] = "_element_mode_view_indication",
  [sym_element_record_mode_view_indication] = "element_record_mode_view_indication",
  [sym_element_array_mode_view_indication] = "element_array_mode_view_indication",
  [sym_mode] = "mode",
  [sym_range_constraint] = "range_constraint",
  [sym_group_constituent_list] = "group_constituent_list",
  [sym_entity_class_entry_list] = "entity_class_entry_list",
  [sym_entity_class_entry] = "entity_class_entry",
  [sym_guarded_signal_specification] = "guarded_signal_specification",
  [sym_signal_list] = "signal_list",
  [sym_entity_specification] = "entity_specification",
  [sym__entity_class] = "_entity_class",
  [sym_entity_name_list] = "entity_name_list",
  [sym_entity_designator] = "entity_designator",
  [sym__entity_tag] = "_entity_tag",
  [sym__alias_designator] = "_alias_designator",
  [sym_file_open_information] = "file_open_information",
  [sym_file_logical_name] = "file_logical_name",
  [sym_package_header] = "package_header",
  [sym_generic_clause] = "generic_clause",
  [sym_subprogram_header] = "subprogram_header",
  [sym__subprogram_specification] = "_subprogram_specification",
  [sym_parameter_list_specification] = "parameter_list_specification",
  [sym__type_definition] = "_type_definition",
  [sym_protected_type_instantiation_definition] = "protected_type_instantiation_definition",
  [sym_protected_type_declaration] = "protected_type_declaration",
  [sym_protected_type_declaration_end] = "protected_type_declaration_end",
  [sym_protected_type_header] = "protected_type_header",
  [sym_protected_type_body] = "protected_type_body",
  [sym_protected_type_body_end] = "protected_type_body_end",
  [sym_while_loop] = "while_loop",
  [sym_for_loop] = "for_loop",
  [sym_parameter_specification] = "parameter_specification",
  [sym_case_statement_alternative] = "case_statement_alternative",
  [sym_selected_waveforms] = "selected_waveforms",
  [sym_waveform] = "waveform",
  [sym_waveform_element] = "waveform_element",
  [sym_force_mode] = "force_mode",
  [sym_selected_expressions] = "selected_expressions",
  [sym_when_element] = "when_element",
  [sym_conditional_waveforms] = "conditional_waveforms",
  [sym_else_waveform] = "else_waveform",
  [sym_delay_mechanism] = "delay_mechanism",
  [sym__target] = "_target",
  [sym_aggregate] = "aggregate",
  [sym_conditional_or_unaffected_expression] = "conditional_or_unaffected_expression",
  [sym_else_expression_or_unaffected] = "else_expression_or_unaffected",
  [sym__expression_or_unaffected] = "_expression_or_unaffected",
  [sym_label_declaration] = "label_declaration",
  [sym_sensitivity_clause] = "sensitivity_clause",
  [sym_sensitivity_list] = "sensitivity_list",
  [sym_condition_clause] = "condition_clause",
  [sym_timeout_clause] = "timeout_clause",
  [sym_physical_type_definition] = "physical_type_definition",
  [sym_primary_unit_declaration] = "primary_unit_declaration",
  [sym_secondary_unit_declaration] = "secondary_unit_declaration",
  [sym_end_units] = "end_units",
  [sym_enumeration_type_definition] = "enumeration_type_definition",
  [sym_enumeration_literal] = "enumeration_literal",
  [sym_record_type_definition] = "record_type_definition",
  [sym_end_record] = "end_record",
  [sym_element_declaration] = "element_declaration",
  [sym__process_sensitivity_list] = "_process_sensitivity_list",
  [sym_configuration_specification] = "configuration_specification",
  [sym_verification_unit_binding_indication] = "verification_unit_binding_indication",
  [sym_verification_unit_list] = "verification_unit_list",
  [sym_component_specification] = "component_specification",
  [sym_instantiation_list] = "instantiation_list",
  [sym_binding_indication] = "binding_indication",
  [sym_entity_aspect] = "entity_aspect",
  [sym_line_comment] = "line_comment",
  [sym_block_comment] = "block_comment",
  [aux_sym_design_file_repeat1] = "design_file_repeat1",
  [aux_sym_design_unit_repeat1] = "design_unit_repeat1",
  [aux_sym_design_unit_repeat2] = "design_unit_repeat2",
  [aux_sym_design_unit_repeat3] = "design_unit_repeat3",
  [aux_sym_logical_name_list_repeat1] = "logical_name_list_repeat1",
  [aux_sym_selected_name_list_repeat1] = "selected_name_list_repeat1",
  [aux_sym_selected_name_repeat1] = "selected_name_repeat1",
  [aux_sym_entity_head_repeat1] = "entity_head_repeat1",
  [aux_sym_entity_body_repeat1] = "entity_body_repeat1",
  [aux_sym_configuration_head_repeat1] = "configuration_head_repeat1",
  [aux_sym_configuration_head_repeat2] = "configuration_head_repeat2",
  [aux_sym_package_declaration_body_repeat1] = "package_declaration_body_repeat1",
  [aux_sym_package_definition_body_repeat1] = "package_definition_body_repeat1",
  [aux_sym_subprogram_head_repeat1] = "subprogram_head_repeat1",
  [aux_sym_mode_view_body_repeat1] = "mode_view_body_repeat1",
  [aux_sym_array_index_incomplete_type_list_repeat1] = "array_index_incomplete_type_list_repeat1",
  [aux_sym_index_constraint_repeat1] = "index_constraint_repeat1",
  [aux_sym_record_resolution_repeat1] = "record_resolution_repeat1",
  [aux_sym_sequential_block_repeat1] = "sequential_block_repeat1",
  [aux_sym_sequential_block_head_repeat1] = "sequential_block_head_repeat1",
  [aux_sym_case_generate_block_repeat1] = "case_generate_block_repeat1",
  [aux_sym_case_body_repeat1] = "case_body_repeat1",
  [aux_sym_conditional_expression_repeat1] = "conditional_expression_repeat1",
  [aux_sym_name_repeat1] = "name_repeat1",
  [aux_sym_package_pathname_repeat1] = "package_pathname_repeat1",
  [aux_sym_relative_pathname_repeat1] = "relative_pathname_repeat1",
  [aux_sym_partial_pathname_repeat1] = "partial_pathname_repeat1",
  [aux_sym_association_or_range_list_repeat1] = "association_or_range_list_repeat1",
  [aux_sym_signature_repeat1] = "signature_repeat1",
  [aux_sym_element_association_list_repeat1] = "element_association_list_repeat1",
  [aux_sym_user_directive_repeat1] = "user_directive_repeat1",
  [aux_sym_conditional_analysis_expression_repeat1] = "conditional_analysis_expression_repeat1",
  [aux_sym_block_configuration_repeat1] = "block_configuration_repeat1",
  [aux_sym_block_configuration_repeat2] = "block_configuration_repeat2",
  [aux_sym_generic_interface_list_repeat1] = "generic_interface_list_repeat1",
  [aux_sym_interface_list_repeat1] = "interface_list_repeat1",
  [aux_sym_association_list_repeat1] = "association_list_repeat1",
  [aux_sym_generic_identifier_list_repeat1] = "generic_identifier_list_repeat1",
  [aux_sym_constant_identifier_list_repeat1] = "constant_identifier_list_repeat1",
  [aux_sym_identifier_list_repeat1] = "identifier_list_repeat1",
  [aux_sym_group_constituent_list_repeat1] = "group_constituent_list_repeat1",
  [aux_sym_entity_class_entry_list_repeat1] = "entity_class_entry_list_repeat1",
  [aux_sym_entity_name_list_repeat1] = "entity_name_list_repeat1",
  [aux_sym_protected_type_declaration_repeat1] = "protected_type_declaration_repeat1",
  [aux_sym_protected_type_body_repeat1] = "protected_type_body_repeat1",
  [aux_sym_selected_waveforms_repeat1] = "selected_waveforms_repeat1",
  [aux_sym_waveform_repeat1] = "waveform_repeat1",
  [aux_sym_selected_expressions_repeat1] = "selected_expressions_repeat1",
  [aux_sym_conditional_waveforms_repeat1] = "conditional_waveforms_repeat1",
  [aux_sym_aggregate_repeat1] = "aggregate_repeat1",
  [aux_sym_conditional_or_unaffected_expression_repeat1] = "conditional_or_unaffected_expression_repeat1",
  [aux_sym_physical_type_definition_repeat1] = "physical_type_definition_repeat1",
  [aux_sym_enumeration_type_definition_repeat1] = "enumeration_type_definition_repeat1",
  [aux_sym_record_type_definition_repeat1] = "record_type_definition_repeat1",
  [aux_sym_instantiation_list_repeat1] = "instantiation_list_repeat1",
  [alias_sym_attribute_identifier] = "attribute_identifier",
  [alias_sym_label] = "label",
  [alias_sym_unit] = "unit",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_EQ_GT] = anon_sym_EQ_GT,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [sym_variable_assignment] = sym_variable_assignment,
  [sym_condition_conversion] = sym_condition_conversion,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_QMARK_EQ] = anon_sym_QMARK_EQ,
  [anon_sym_QMARK_SLASH_EQ] = anon_sym_QMARK_SLASH_EQ,
  [anon_sym_QMARK_LT] = anon_sym_QMARK_LT,
  [anon_sym_QMARK_LT_EQ] = anon_sym_QMARK_LT_EQ,
  [anon_sym_QMARK_GT] = anon_sym_QMARK_GT,
  [anon_sym_QMARK_GT_EQ] = anon_sym_QMARK_GT_EQ,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [sym_exponentiate] = sym_exponentiate,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [sym_identifier] = sym_identifier,
  [sym_ABS] = sym_ABS,
  [sym_ACCESS] = sym_ACCESS,
  [sym_AFTER] = sym_AFTER,
  [sym_ALIAS] = sym_ALIAS,
  [sym_ALL] = sym_ALL,
  [sym_AND] = sym_AND,
  [sym_ARCHITECTURE] = sym_ARCHITECTURE,
  [sym_ARRAY] = sym_ARRAY,
  [sym_ASSERT] = sym_ASSERT,
  [sym_ASSUME] = sym_ASSUME,
  [sym_ATTRIBUTE] = sym_ATTRIBUTE,
  [sym_BEGIN] = sym_BEGIN,
  [sym_BLOCK] = sym_BLOCK,
  [sym_BODY] = sym_BODY,
  [sym_BUFFER] = sym_BUFFER,
  [sym_BUS] = sym_BUS,
  [sym_CASE] = sym_CASE,
  [sym_COMPONENT] = sym_COMPONENT,
  [sym_CONFIGURATION] = sym_CONFIGURATION,
  [sym_CONSTANT] = sym_CONSTANT,
  [sym_CONTEXT] = sym_CONTEXT,
  [sym_COVER] = sym_COVER,
  [sym_DEFAULT] = sym_DEFAULT,
  [sym_DISCONNECT] = sym_DISCONNECT,
  [sym_DOWNTO] = sym_DOWNTO,
  [sym_ELSE] = sym_ELSE,
  [sym_ELSIF] = sym_ELSIF,
  [sym_END] = sym_END,
  [sym_ENTITY] = sym_ENTITY,
  [sym_EXIT] = sym_EXIT,
  [sym_FAIRNESS] = sym_FAIRNESS,
  [sym_FILE] = sym_FILE,
  [sym_FOR] = sym_FOR,
  [sym_FORCE] = sym_FORCE,
  [sym_FUNCTION] = sym_FUNCTION,
  [sym_GENERATE] = sym_GENERATE,
  [sym_GENERIC] = sym_GENERIC,
  [sym_GROUP] = sym_GROUP,
  [sym_GUARDED] = sym_GUARDED,
  [sym_IF] = sym_IF,
  [sym_IMPURE] = sym_IMPURE,
  [sym_IN] = sym_IN,
  [sym_INERTIAL] = sym_INERTIAL,
  [sym_INOUT] = sym_INOUT,
  [sym_IS] = sym_IS,
  [sym_LABEL] = sym_LABEL,
  [sym_LIBRARY] = sym_LIBRARY,
  [sym_LINKAGE] = sym_LINKAGE,
  [sym_LITERAL] = sym_LITERAL,
  [sym_LOOP] = sym_LOOP,
  [sym_MAP] = sym_MAP,
  [sym_MOD] = sym_MOD,
  [sym_NAND] = sym_NAND,
  [sym_NEW] = sym_NEW,
  [sym_NEXT] = sym_NEXT,
  [sym_NOR] = sym_NOR,
  [sym_NOT] = sym_NOT,
  [sym_NULL] = sym_NULL,
  [sym_OF] = sym_OF,
  [sym_ON] = sym_ON,
  [sym_OPEN] = sym_OPEN,
  [sym_OR] = sym_OR,
  [sym_OTHERS] = sym_OTHERS,
  [sym_OUT] = sym_OUT,
  [sym_PACKAGE] = sym_PACKAGE,
  [sym_PARAMETER] = sym_PARAMETER,
  [sym_PORT] = sym_PORT,
  [sym_POSTPONED] = sym_POSTPONED,
  [sym_PROCEDURE] = sym_PROCEDURE,
  [sym_PROCESS] = sym_PROCESS,
  [sym_PROPERTY] = sym_PROPERTY,
  [sym_PROTECTED] = sym_PROTECTED,
  [sym_PRIVATE] = sym_PRIVATE,
  [sym_PURE] = sym_PURE,
  [sym_RANGE] = sym_RANGE,
  [sym_RECORD] = sym_RECORD,
  [sym_REGISTER] = sym_REGISTER,
  [sym_REJECT] = sym_REJECT,
  [sym_RELEASE] = sym_RELEASE,
  [sym_REM] = sym_REM,
  [sym_REPORT] = sym_REPORT,
  [sym_RESTRICT] = sym_RESTRICT,
  [sym_RETURN] = sym_RETURN,
  [sym_ROL] = sym_ROL,
  [sym_ROR] = sym_ROR,
  [sym_SELECT] = sym_SELECT,
  [sym_SEQUENCE] = sym_SEQUENCE,
  [sym_SEVERITY] = sym_SEVERITY,
  [sym_SIGNAL] = sym_SIGNAL,
  [sym_SHARED] = sym_SHARED,
  [sym_SLA] = sym_SLA,
  [sym_SLL] = sym_SLL,
  [sym_SRA] = sym_SRA,
  [sym_SRL] = sym_SRL,
  [sym_STRONG] = sym_STRONG,
  [sym_SUBTYPE] = sym_SUBTYPE,
  [sym_THEN] = sym_THEN,
  [sym_TO] = sym_TO,
  [sym_TRANSPORT] = sym_TRANSPORT,
  [sym_TYPE] = sym_TYPE,
  [sym_UNAFFECTED] = sym_UNAFFECTED,
  [sym_UNITS] = sym_UNITS,
  [sym_UNTIL] = sym_UNTIL,
  [sym_USE] = sym_USE,
  [sym_VARIABLE] = sym_VARIABLE,
  [sym_VIEW] = sym_VIEW,
  [sym_VMODE] = sym_VMODE,
  [sym_VPKG] = sym_VPKG,
  [sym_VPROP] = sym_VPROP,
  [sym_VUNIT] = sym_VUNIT,
  [sym_WAIT] = sym_WAIT,
  [sym_WHEN] = sym_WHEN,
  [sym_WHILE] = sym_WHILE,
  [sym_WITH] = sym_WITH,
  [sym_XNOR] = sym_XNOR,
  [sym_XOR] = sym_XOR,
  [sym_reserved_end_marker] = sym_reserved_end_marker,
  [sym_directive_body] = sym_directive_body,
  [sym_directive_constant_builtin] = sym_directive_constant_builtin,
  [sym_directive_error] = sym_directive_error,
  [sym_directive_protect] = sym_directive_protect,
  [sym_directive_warning] = sym_directive_warning,
  [sym__directive_newline] = sym__directive_newline,
  [sym__grave_accent] = sym__grave_accent,
  [sym_box] = sym_box,
  [sym_delimiter_end_marker] = sym_delimiter_end_marker,
  [sym_decimal_integer] = sym_decimal_integer,
  [sym_decimal_float] = sym_decimal_float,
  [sym_based_base] = sym_based_base,
  [sym_based_integer] = sym_based_integer,
  [sym_based_float] = sym_based_float,
  [sym_character_literal] = sym_character_literal,
  [sym_string_literal] = sym_string_literal,
  [sym_string_literal_std_logic] = sym_string_literal_std_logic,
  [sym_bit_string_length] = sym_bit_string_length,
  [sym_bit_string_base] = sym_bit_string_base,
  [sym_bit_string_value] = sym_bit_string_value,
  [sym_operator_symbol] = sym_operator_symbol,
  [sym__line_comment_start] = sym__line_comment_start,
  [sym__block_comment_start] = sym__block_comment_start,
  [sym__block_comment_end] = sym__block_comment_end,
  [sym_comment_content] = sym_comment_content,
  [sym_token_end_marker] = sym_token_end_marker,
  [sym_attribute_function] = sym_attribute_function,
  [sym_attribute_impure_function] = sym_attribute_impure_function,
  [sym_attribute_mode_view] = sym_attribute_mode_view,
  [sym_attribute_pure_function] = sym_attribute_pure_function,
  [sym_attribute_range] = sym_attribute_range,
  [sym_attribute_signal] = sym_attribute_signal,
  [sym_attribute_subtype] = sym_attribute_subtype,
  [sym_attribute_type] = sym_attribute_type,
  [sym_attribute_value] = sym_attribute_value,
  [sym_library_attribute] = sym_library_attribute,
  [sym_library_constant] = sym_library_constant,
  [sym_library_constant_boolean] = sym_library_constant_boolean,
  [sym_library_constant_character] = sym_library_constant_character,
  [sym_library_constant_debug] = sym_library_constant_debug,
  [sym_library_constant_env] = sym_library_constant_env,
  [sym_library_constant_standard] = sym_library_constant_standard,
  [sym_library_constant_std_logic] = sym_library_constant_std_logic,
  [sym_library_constant_unit] = sym_library_constant_unit,
  [sym_library_function] = sym_library_function,
  [sym_library_namespace] = sym_library_namespace,
  [sym_library_type] = sym_library_type,
  [sym__end_of_file] = sym__end_of_file,
  [sym_error_sentinel] = sym_error_sentinel,
  [sym_design_file] = sym_design_file,
  [sym_design_unit] = sym_design_unit,
  [sym__context_item] = sym__context_item,
  [sym_library_clause] = sym_library_clause,
  [sym_use_clause] = sym_use_clause,
  [sym_context_reference] = sym_context_reference,
  [sym_logical_name_list] = sym_logical_name_list,
  [sym__logical_name] = sym__logical_name,
  [sym_selected_name_list] = sym_selected_name_list,
  [sym_selected_name] = sym_selected_name,
  [sym__library_unit] = sym__library_unit,
  [sym_entity_declaration] = sym_entity_declaration,
  [sym_entity_head] = sym_entity_head,
  [sym_port_clause] = sym_port_clause,
  [sym_entity_body] = sym_entity_body,
  [sym_end_entity] = sym_end_entity,
  [sym_configuration_declaration] = sym_configuration_declaration,
  [sym_configuration_head] = sym_configuration_head,
  [sym_end_configuration] = sym_end_configuration,
  [sym_package_declaration] = sym_package_declaration,
  [sym_package_declaration_body] = sym_package_declaration_body,
  [sym_end_package] = sym_end_package,
  [sym_package_instantiation_declaration] = sym_package_instantiation_declaration,
  [sym_interface_package_declaration] = sym_interface_package_declaration,
  [sym_context_declaration] = sym_context_declaration,
  [sym_context_declaration_body] = sym_context_declaration_body,
  [sym_end_context] = sym_end_context,
  [sym_architecture_definition] = sym_architecture_definition,
  [sym_architecture_head] = sym_architecture_head,
  [sym_end_architecture] = sym_end_architecture,
  [sym_package_definition] = sym_package_definition,
  [sym_package_definition_body] = sym_package_definition_body,
  [sym_end_package_body] = sym_end_package_body,
  [sym__entity_declarative_item] = sym__entity_declarative_item,
  [sym__package_declarative_item] = sym__package_declarative_item,
  [sym__package_body_declarative_item] = sym__package_body_declarative_item,
  [sym__protected_type_declarative_item] = sym__protected_type_declarative_item,
  [sym__protected_type_body_declarative_item] = sym__protected_type_body_declarative_item,
  [sym__process_declarative_item] = sym__process_declarative_item,
  [sym__subprogram_declarative_item] = sym__subprogram_declarative_item,
  [sym__block_declarative_item] = sym__block_declarative_item,
  [sym_subprogram_declaration] = sym_subprogram_declaration,
  [sym_subprogram_definition] = sym_subprogram_definition,
  [sym_subprogram_head] = sym_subprogram_head,
  [sym_subprogram_end] = sym_subprogram_end,
  [sym_subprogram_instantiation_declaration] = sym_subprogram_instantiation_declaration,
  [sym_type_declaration] = sym_type_declaration,
  [sym_subtype_declaration] = sym_subtype_declaration,
  [sym_mode_view_declaration] = sym_mode_view_declaration,
  [sym_mode_view_body] = sym_mode_view_body,
  [sym_end_view] = sym_end_view,
  [sym_constant_declaration] = sym_constant_declaration,
  [sym_signal_declaration] = sym_signal_declaration,
  [sym_signal_kind] = sym_signal_kind,
  [sym_variable_declaration] = sym_variable_declaration,
  [sym_initialiser] = sym_initialiser,
  [sym_file_declaration] = sym_file_declaration,
  [sym_interface_file_declaration] = sym_interface_file_declaration,
  [sym_alias_declaration] = sym_alias_declaration,
  [sym_component_declaration] = sym_component_declaration,
  [sym_component_body] = sym_component_body,
  [sym_end_component] = sym_end_component,
  [sym_attribute_declaration] = sym_attribute_declaration,
  [sym_attribute_specification] = sym_attribute_specification,
  [sym_disconnection_specification] = sym_disconnection_specification,
  [sym_group_template_declaration] = sym_group_template_declaration,
  [sym_group_declaration] = sym_group_declaration,
  [sym_private_variable_declaration] = sym_private_variable_declaration,
  [sym__interface_type_indication] = sym__interface_type_indication,
  [sym_unspecified_type_indication] = sym_unspecified_type_indication,
  [sym_interface_type_declaration] = sym_interface_type_declaration,
  [sym__incomplete_type_definition] = sym__incomplete_type_definition,
  [sym_private_incomplete_type_definition] = sym_private_incomplete_type_definition,
  [sym_scalar_incomplete_type_definition] = sym_scalar_incomplete_type_definition,
  [sym_discrete_incomplete_type_definition] = sym_discrete_incomplete_type_definition,
  [sym_integer_incomplete_type_definition] = sym_integer_incomplete_type_definition,
  [sym_physical_incomplete_type_definition] = sym_physical_incomplete_type_definition,
  [sym_floating_incomplete_type_definition] = sym_floating_incomplete_type_definition,
  [sym_array_type_definition] = sym_array_type_definition,
  [sym_array_index_incomplete_type_list] = sym_array_index_incomplete_type_list,
  [sym__array_index_incomplete_type] = sym__array_index_incomplete_type,
  [sym_index_constraint] = sym_index_constraint,
  [sym_index_subtype_definition] = sym_index_subtype_definition,
  [sym_access_type_definition] = sym_access_type_definition,
  [sym_access_incomplete_type_definition] = sym_access_incomplete_type_definition,
  [sym_file_type_definition] = sym_file_type_definition,
  [sym_file_incomplete_type_definition] = sym_file_incomplete_type_definition,
  [sym_subtype_indication] = sym_subtype_indication,
  [sym_resolution_indication] = sym_resolution_indication,
  [sym__element_resolution] = sym__element_resolution,
  [sym_record_resolution] = sym_record_resolution,
  [sym_record_element_resolution] = sym_record_element_resolution,
  [sym__incomplete_subtype_indication] = sym__incomplete_subtype_indication,
  [sym__incomplete_type_mark] = sym__incomplete_type_mark,
  [sym_concurrent_block] = sym_concurrent_block,
  [sym_sequential_block] = sym_sequential_block,
  [sym_generate_body] = sym_generate_body,
  [sym_generate_head] = sym_generate_head,
  [sym_case_generate_alternative] = sym_case_generate_alternative,
  [sym_case_generate_body] = sym_case_generate_body,
  [sym_generate_block] = sym_generate_block,
  [sym_generate_block_end] = sym_generate_block_end,
  [sym__entity_statement] = sym__entity_statement,
  [sym__concurrent_statement] = sym__concurrent_statement,
  [sym__sequential_statement] = sym__sequential_statement,
  [sym_block_statement] = sym_block_statement,
  [sym_guard_condition] = sym_guard_condition,
  [sym_block_head] = sym_block_head,
  [sym_end_block] = sym_end_block,
  [sym_sequential_block_statement] = sym_sequential_block_statement,
  [sym_sequential_block_head] = sym_sequential_block_head,
  [sym_component_instantiation_statement] = sym_component_instantiation_statement,
  [sym_instantiated_unit] = sym_instantiated_unit,
  [sym_process_statement] = sym_process_statement,
  [sym_sensitivity_specification] = sym_sensitivity_specification,
  [sym_process_head] = sym_process_head,
  [sym_end_process] = sym_end_process,
  [sym_case_generate_statement] = sym_case_generate_statement,
  [sym_case_generate_block] = sym_case_generate_block,
  [sym_for_generate_statement] = sym_for_generate_statement,
  [sym_if_generate_statement] = sym_if_generate_statement,
  [sym_if_generate] = sym_if_generate,
  [sym__elsif_else_generate] = sym__elsif_else_generate,
  [sym_elsif_generate] = sym_elsif_generate,
  [sym_else_generate] = sym_else_generate,
  [sym_end_generate] = sym_end_generate,
  [sym_assertion_statement] = sym_assertion_statement,
  [sym_concurrent_assertion_statement] = sym_concurrent_assertion_statement,
  [sym_assertion] = sym_assertion,
  [sym_report_expression] = sym_report_expression,
  [sym_severity_expression] = sym_severity_expression,
  [sym_case_statement] = sym_case_statement,
  [sym_case_expression] = sym_case_expression,
  [sym_case_body] = sym_case_body,
  [sym_end_case] = sym_end_case,
  [sym_exit_statement] = sym_exit_statement,
  [sym_when_expression] = sym_when_expression,
  [sym_if_statement_block] = sym_if_statement_block,
  [sym_if_statement] = sym_if_statement,
  [sym__elsif_else_statement] = sym__elsif_else_statement,
  [sym_elsif_statement] = sym_elsif_statement,
  [sym_else_statement] = sym_else_statement,
  [sym_if_statement_body] = sym_if_statement_body,
  [sym_end_if] = sym_end_if,
  [sym_loop_statement] = sym_loop_statement,
  [sym__iteration_scheme] = sym__iteration_scheme,
  [sym_loop_body] = sym_loop_body,
  [sym_end_loop] = sym_end_loop,
  [sym_next_statement] = sym_next_statement,
  [sym_null_statement] = sym_null_statement,
  [sym_procedure_call_statement] = sym_procedure_call_statement,
  [sym_concurrent_procedure_call_statement] = sym_concurrent_procedure_call_statement,
  [sym_report_statement] = sym_report_statement,
  [sym_return_statement] = sym_return_statement,
  [sym_signal_assignment] = sym_signal_assignment,
  [sym_simple_waveform_assignment] = sym_simple_waveform_assignment,
  [sym_simple_variable_assignment] = sym_simple_variable_assignment,
  [sym_concurrent_simple_signal_assignment] = sym_concurrent_simple_signal_assignment,
  [sym_simple_force_assignment] = sym_simple_force_assignment,
  [sym_simple_release_assignment] = sym_simple_release_assignment,
  [sym_conditional_signal_assignment] = sym_conditional_signal_assignment,
  [sym_concurrent_conditional_signal_assignment] = sym_concurrent_conditional_signal_assignment,
  [sym_selected_waveform_assignment] = sym_selected_waveform_assignment,
  [sym_concurrent_selected_signal_assignment] = sym_concurrent_selected_signal_assignment,
  [sym_selected_force_assignment] = sym_selected_force_assignment,
  [sym_selected_variable_assignment] = sym_selected_variable_assignment,
  [sym_with_expression] = sym_with_expression,
  [sym_select_target] = sym_select_target,
  [sym_wait_statement] = sym_wait_statement,
  [sym_conditional_expression] = sym_conditional_expression,
  [sym_else_expression] = sym_else_expression,
  [sym__expression] = sym__expression,
  [sym_condition_expression] = sym_condition_expression,
  [sym_logical_expression] = sym_logical_expression,
  [sym_relational_expression] = sym_relational_expression,
  [sym_shift_expression] = sym_shift_expression,
  [sym_simple_expression] = sym_simple_expression,
  [sym_unary_operator] = sym_unary_operator,
  [sym_logical_operator] = sym_logical_operator,
  [sym_relational_operator] = sym_relational_operator,
  [sym_shift_operator] = sym_shift_operator,
  [sym_sign] = sym_sign,
  [sym_adding_operator] = sym_adding_operator,
  [sym_multiplying_operator] = sym_multiplying_operator,
  [sym__primary] = sym__primary,
  [sym_name] = sym_name,
  [sym__direct_name] = sym__direct_name,
  [sym__label] = sym__label,
  [sym__attribute] = sym__attribute,
  [sym__unit] = sym__unit,
  [sym__identifier] = sym__identifier,
  [sym__name_selector] = sym__name_selector,
  [sym__external_name] = sym__external_name,
  [sym_external_constant_name] = sym_external_constant_name,
  [sym_external_signal_name] = sym_external_signal_name,
  [sym_external_variable_name] = sym_external_variable_name,
  [sym__external_pathname] = sym__external_pathname,
  [sym_package_pathname] = sym_package_pathname,
  [sym_absolute_pathname] = sym_absolute_pathname,
  [sym_relative_pathname] = sym_relative_pathname,
  [sym_partial_pathname] = sym_partial_pathname,
  [sym_pathname_element] = sym_pathname_element,
  [sym_function_call] = sym_function_call,
  [sym_parenthesis_group] = sym_parenthesis_group,
  [sym_association_or_range_list] = sym_association_or_range_list,
  [sym__association_or_range] = sym__association_or_range,
  [sym_association_element] = sym_association_element,
  [sym__actual_part] = sym__actual_part,
  [sym_attribute] = sym_attribute,
  [sym__attribute_designator] = sym__attribute_designator,
  [sym_signature] = sym_signature,
  [sym_selection] = sym_selection,
  [sym__suffix] = sym__suffix,
  [sym__literal] = sym__literal,
  [sym_bit_string_literal] = sym_bit_string_literal,
  [sym__abstract_literal] = sym__abstract_literal,
  [sym_based_literal] = sym_based_literal,
  [sym_allocator] = sym_allocator,
  [sym_parenthesis_expression] = sym_parenthesis_expression,
  [sym_element_association_list] = sym_element_association_list,
  [sym_element_association] = sym_element_association,
  [sym__element] = sym__element,
  [sym_choices] = sym_choices,
  [sym__range] = sym__range,
  [sym_simple_range] = sym_simple_range,
  [sym__direction] = sym__direction,
  [sym__tool_directive] = sym__tool_directive,
  [sym__directive] = sym__directive,
  [sym_user_directive] = sym_user_directive,
  [sym_protect_directive] = sym_protect_directive,
  [sym_warning_directive] = sym_warning_directive,
  [sym_error_directive] = sym_error_directive,
  [sym_if_conditional_analysis] = sym_if_conditional_analysis,
  [sym_elsif_conditional_analysis] = sym_elsif_conditional_analysis,
  [sym_else_conditional_analysis] = sym_else_conditional_analysis,
  [sym_end_conditional_analysis] = sym_end_conditional_analysis,
  [sym_conditional_analysis_expression] = sym_conditional_analysis_expression,
  [sym_conditional_analysis_relation] = sym_conditional_analysis_relation,
  [sym__conditional_analysis_operator] = sym__conditional_analysis_operator,
  [sym__conditional_analysis_identifier] = sym__conditional_analysis_identifier,
  [sym__configuration_declarative_item] = sym__configuration_declarative_item,
  [sym__configuration_item] = sym__configuration_item,
  [sym_block_configuration] = sym_block_configuration,
  [sym_component_configuration] = sym_component_configuration,
  [sym_end_for] = sym_end_for,
  [sym_generic_interface_list] = sym_interface_list,
  [sym__generic_interface_declaration] = sym__generic_interface_declaration,
  [sym_generic_interface_declaration] = sym_interface_declaration,
  [sym_interface_list] = sym_interface_list,
  [sym__interface_declaration] = sym__interface_declaration,
  [sym_interface_declaration] = sym_interface_declaration,
  [sym_interface_constant_declaration] = sym_interface_constant_declaration,
  [sym_interface_signal_declaration] = sym_interface_signal_declaration,
  [sym_interface_variable_declaration] = sym_interface_variable_declaration,
  [sym_interface_subprogram_declaration] = sym_interface_subprogram_declaration,
  [sym__interface_subprogram_specification] = sym__interface_subprogram_specification,
  [sym_procedure_specification] = sym_procedure_specification,
  [sym_interface_procedure_specification] = sym_interface_procedure_specification,
  [sym_function_specification] = sym_function_specification,
  [sym_interface_function_specification] = sym_interface_function_specification,
  [sym__interface_package_generic_map_aspect] = sym__interface_package_generic_map_aspect,
  [sym_generic_map_any] = sym_generic_map_any,
  [sym_generic_map_default] = sym_generic_map_default,
  [sym_generic_map_aspect] = sym_generic_map_aspect,
  [sym_port_map_aspect] = sym_port_map_aspect,
  [sym_association_list] = sym_association_list,
  [sym__interface_subprogram_default] = sym__interface_subprogram_default,
  [sym__designator] = sym__designator,
  [sym_generic_identifier_list] = sym_identifier_list,
  [sym_constant_identifier_list] = sym_identifier_list,
  [sym_identifier_list] = sym_identifier_list,
  [sym__mode_indication] = sym__mode_indication,
  [sym_simple_mode_indication] = sym_simple_mode_indication,
  [sym__mode_view_indication] = sym__mode_view_indication,
  [sym_record_mode_view_indication] = sym_record_mode_view_indication,
  [sym_array_mode_view_indication] = sym_array_mode_view_indication,
  [sym_mode_view_element_definition] = sym_mode_view_element_definition,
  [sym_record_element_list] = sym_record_element_list,
  [sym__element_mode_indication] = sym__element_mode_indication,
  [sym__element_mode_view_indication] = sym__element_mode_view_indication,
  [sym_element_record_mode_view_indication] = sym_element_record_mode_view_indication,
  [sym_element_array_mode_view_indication] = sym_element_array_mode_view_indication,
  [sym_mode] = sym_mode,
  [sym_range_constraint] = sym_range_constraint,
  [sym_group_constituent_list] = sym_group_constituent_list,
  [sym_entity_class_entry_list] = sym_entity_class_entry_list,
  [sym_entity_class_entry] = sym_entity_class_entry,
  [sym_guarded_signal_specification] = sym_guarded_signal_specification,
  [sym_signal_list] = sym_signal_list,
  [sym_entity_specification] = sym_entity_specification,
  [sym__entity_class] = sym__entity_class,
  [sym_entity_name_list] = sym_entity_name_list,
  [sym_entity_designator] = sym_entity_designator,
  [sym__entity_tag] = sym__entity_tag,
  [sym__alias_designator] = sym__alias_designator,
  [sym_file_open_information] = sym_file_open_information,
  [sym_file_logical_name] = sym_file_logical_name,
  [sym_package_header] = sym_package_header,
  [sym_generic_clause] = sym_generic_clause,
  [sym_subprogram_header] = sym_subprogram_header,
  [sym__subprogram_specification] = sym__subprogram_specification,
  [sym_parameter_list_specification] = sym_parameter_list_specification,
  [sym__type_definition] = sym__type_definition,
  [sym_protected_type_instantiation_definition] = sym_protected_type_instantiation_definition,
  [sym_protected_type_declaration] = sym_protected_type_declaration,
  [sym_protected_type_declaration_end] = sym_protected_type_declaration_end,
  [sym_protected_type_header] = sym_protected_type_header,
  [sym_protected_type_body] = sym_protected_type_body,
  [sym_protected_type_body_end] = sym_protected_type_body_end,
  [sym_while_loop] = sym_while_loop,
  [sym_for_loop] = sym_for_loop,
  [sym_parameter_specification] = sym_parameter_specification,
  [sym_case_statement_alternative] = sym_case_statement_alternative,
  [sym_selected_waveforms] = sym_selected_waveforms,
  [sym_waveform] = sym_waveform,
  [sym_waveform_element] = sym_waveform_element,
  [sym_force_mode] = sym_force_mode,
  [sym_selected_expressions] = sym_selected_expressions,
  [sym_when_element] = sym_when_element,
  [sym_conditional_waveforms] = sym_conditional_waveforms,
  [sym_else_waveform] = sym_else_waveform,
  [sym_delay_mechanism] = sym_delay_mechanism,
  [sym__target] = sym__target,
  [sym_aggregate] = sym_aggregate,
  [sym_conditional_or_unaffected_expression] = sym_conditional_or_unaffected_expression,
  [sym_else_expression_or_unaffected] = sym_else_expression_or_unaffected,
  [sym__expression_or_unaffected] = sym__expression_or_unaffected,
  [sym_label_declaration] = sym_label_declaration,
  [sym_sensitivity_clause] = sym_sensitivity_clause,
  [sym_sensitivity_list] = sym_sensitivity_list,
  [sym_condition_clause] = sym_condition_clause,
  [sym_timeout_clause] = sym_timeout_clause,
  [sym_physical_type_definition] = sym_physical_type_definition,
  [sym_primary_unit_declaration] = sym_primary_unit_declaration,
  [sym_secondary_unit_declaration] = sym_secondary_unit_declaration,
  [sym_end_units] = sym_end_units,
  [sym_enumeration_type_definition] = sym_enumeration_type_definition,
  [sym_enumeration_literal] = sym_enumeration_literal,
  [sym_record_type_definition] = sym_record_type_definition,
  [sym_end_record] = sym_end_record,
  [sym_element_declaration] = sym_element_declaration,
  [sym__process_sensitivity_list] = sym__process_sensitivity_list,
  [sym_configuration_specification] = sym_configuration_specification,
  [sym_verification_unit_binding_indication] = sym_verification_unit_binding_indication,
  [sym_verification_unit_list] = sym_verification_unit_list,
  [sym_component_specification] = sym_component_specification,
  [sym_instantiation_list] = sym_instantiation_list,
  [sym_binding_indication] = sym_binding_indication,
  [sym_entity_aspect] = sym_entity_aspect,
  [sym_line_comment] = sym_line_comment,
  [sym_block_comment] = sym_block_comment,
  [aux_sym_design_file_repeat1] = aux_sym_design_file_repeat1,
  [aux_sym_design_unit_repeat1] = aux_sym_design_unit_repeat1,
  [aux_sym_design_unit_repeat2] = aux_sym_design_unit_repeat2,
  [aux_sym_design_unit_repeat3] = aux_sym_design_unit_repeat3,
  [aux_sym_logical_name_list_repeat1] = aux_sym_logical_name_list_repeat1,
  [aux_sym_selected_name_list_repeat1] = aux_sym_selected_name_list_repeat1,
  [aux_sym_selected_name_repeat1] = aux_sym_selected_name_repeat1,
  [aux_sym_entity_head_repeat1] = aux_sym_entity_head_repeat1,
  [aux_sym_entity_body_repeat1] = aux_sym_entity_body_repeat1,
  [aux_sym_configuration_head_repeat1] = aux_sym_configuration_head_repeat1,
  [aux_sym_configuration_head_repeat2] = aux_sym_configuration_head_repeat2,
  [aux_sym_package_declaration_body_repeat1] = aux_sym_package_declaration_body_repeat1,
  [aux_sym_package_definition_body_repeat1] = aux_sym_package_definition_body_repeat1,
  [aux_sym_subprogram_head_repeat1] = aux_sym_subprogram_head_repeat1,
  [aux_sym_mode_view_body_repeat1] = aux_sym_mode_view_body_repeat1,
  [aux_sym_array_index_incomplete_type_list_repeat1] = aux_sym_array_index_incomplete_type_list_repeat1,
  [aux_sym_index_constraint_repeat1] = aux_sym_index_constraint_repeat1,
  [aux_sym_record_resolution_repeat1] = aux_sym_record_resolution_repeat1,
  [aux_sym_sequential_block_repeat1] = aux_sym_sequential_block_repeat1,
  [aux_sym_sequential_block_head_repeat1] = aux_sym_sequential_block_head_repeat1,
  [aux_sym_case_generate_block_repeat1] = aux_sym_case_generate_block_repeat1,
  [aux_sym_case_body_repeat1] = aux_sym_case_body_repeat1,
  [aux_sym_conditional_expression_repeat1] = aux_sym_conditional_expression_repeat1,
  [aux_sym_name_repeat1] = aux_sym_name_repeat1,
  [aux_sym_package_pathname_repeat1] = aux_sym_package_pathname_repeat1,
  [aux_sym_relative_pathname_repeat1] = aux_sym_relative_pathname_repeat1,
  [aux_sym_partial_pathname_repeat1] = aux_sym_partial_pathname_repeat1,
  [aux_sym_association_or_range_list_repeat1] = aux_sym_association_or_range_list_repeat1,
  [aux_sym_signature_repeat1] = aux_sym_signature_repeat1,
  [aux_sym_element_association_list_repeat1] = aux_sym_element_association_list_repeat1,
  [aux_sym_user_directive_repeat1] = aux_sym_user_directive_repeat1,
  [aux_sym_conditional_analysis_expression_repeat1] = aux_sym_conditional_analysis_expression_repeat1,
  [aux_sym_block_configuration_repeat1] = aux_sym_block_configuration_repeat1,
  [aux_sym_block_configuration_repeat2] = aux_sym_block_configuration_repeat2,
  [aux_sym_generic_interface_list_repeat1] = aux_sym_generic_interface_list_repeat1,
  [aux_sym_interface_list_repeat1] = aux_sym_interface_list_repeat1,
  [aux_sym_association_list_repeat1] = aux_sym_association_list_repeat1,
  [aux_sym_generic_identifier_list_repeat1] = aux_sym_generic_identifier_list_repeat1,
  [aux_sym_constant_identifier_list_repeat1] = aux_sym_constant_identifier_list_repeat1,
  [aux_sym_identifier_list_repeat1] = aux_sym_identifier_list_repeat1,
  [aux_sym_group_constituent_list_repeat1] = aux_sym_group_constituent_list_repeat1,
  [aux_sym_entity_class_entry_list_repeat1] = aux_sym_entity_class_entry_list_repeat1,
  [aux_sym_entity_name_list_repeat1] = aux_sym_entity_name_list_repeat1,
  [aux_sym_protected_type_declaration_repeat1] = aux_sym_protected_type_declaration_repeat1,
  [aux_sym_protected_type_body_repeat1] = aux_sym_protected_type_body_repeat1,
  [aux_sym_selected_waveforms_repeat1] = aux_sym_selected_waveforms_repeat1,
  [aux_sym_waveform_repeat1] = aux_sym_waveform_repeat1,
  [aux_sym_selected_expressions_repeat1] = aux_sym_selected_expressions_repeat1,
  [aux_sym_conditional_waveforms_repeat1] = aux_sym_conditional_waveforms_repeat1,
  [aux_sym_aggregate_repeat1] = aux_sym_aggregate_repeat1,
  [aux_sym_conditional_or_unaffected_expression_repeat1] = aux_sym_conditional_or_unaffected_expression_repeat1,
  [aux_sym_physical_type_definition_repeat1] = aux_sym_physical_type_definition_repeat1,
  [aux_sym_enumeration_type_definition_repeat1] = aux_sym_enumeration_type_definition_repeat1,
  [aux_sym_record_type_definition_repeat1] = aux_sym_record_type_definition_repeat1,
  [aux_sym_instantiation_list_repeat1] = aux_sym_instantiation_list_repeat1,
  [alias_sym_attribute_identifier] = alias_sym_attribute_identifier,
  [alias_sym_label] = alias_sym_label,
  [alias_sym_unit] = alias_sym_unit,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [sym_variable_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_condition_conversion] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [sym_exponentiate] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_ABS] = {
    .visible = true,
    .named = false,
  },
  [sym_ACCESS] = {
    .visible = true,
    .named = false,
  },
  [sym_AFTER] = {
    .visible = true,
    .named = false,
  },
  [sym_ALIAS] = {
    .visible = true,
    .named = false,
  },
  [sym_ALL] = {
    .visible = true,
    .named = true,
  },
  [sym_AND] = {
    .visible = true,
    .named = false,
  },
  [sym_ARCHITECTURE] = {
    .visible = true,
    .named = false,
  },
  [sym_ARRAY] = {
    .visible = true,
    .named = false,
  },
  [sym_ASSERT] = {
    .visible = true,
    .named = false,
  },
  [sym_ASSUME] = {
    .visible = true,
    .named = true,
  },
  [sym_ATTRIBUTE] = {
    .visible = true,
    .named = false,
  },
  [sym_BEGIN] = {
    .visible = true,
    .named = false,
  },
  [sym_BLOCK] = {
    .visible = true,
    .named = false,
  },
  [sym_BODY] = {
    .visible = true,
    .named = false,
  },
  [sym_BUFFER] = {
    .visible = true,
    .named = false,
  },
  [sym_BUS] = {
    .visible = true,
    .named = false,
  },
  [sym_CASE] = {
    .visible = true,
    .named = false,
  },
  [sym_COMPONENT] = {
    .visible = true,
    .named = false,
  },
  [sym_CONFIGURATION] = {
    .visible = true,
    .named = false,
  },
  [sym_CONSTANT] = {
    .visible = true,
    .named = false,
  },
  [sym_CONTEXT] = {
    .visible = true,
    .named = false,
  },
  [sym_COVER] = {
    .visible = true,
    .named = true,
  },
  [sym_DEFAULT] = {
    .visible = true,
    .named = true,
  },
  [sym_DISCONNECT] = {
    .visible = true,
    .named = false,
  },
  [sym_DOWNTO] = {
    .visible = true,
    .named = false,
  },
  [sym_ELSE] = {
    .visible = true,
    .named = false,
  },
  [sym_ELSIF] = {
    .visible = true,
    .named = false,
  },
  [sym_END] = {
    .visible = true,
    .named = false,
  },
  [sym_ENTITY] = {
    .visible = true,
    .named = false,
  },
  [sym_EXIT] = {
    .visible = true,
    .named = false,
  },
  [sym_FAIRNESS] = {
    .visible = true,
    .named = true,
  },
  [sym_FILE] = {
    .visible = true,
    .named = false,
  },
  [sym_FOR] = {
    .visible = true,
    .named = false,
  },
  [sym_FORCE] = {
    .visible = true,
    .named = false,
  },
  [sym_FUNCTION] = {
    .visible = true,
    .named = false,
  },
  [sym_GENERATE] = {
    .visible = true,
    .named = false,
  },
  [sym_GENERIC] = {
    .visible = true,
    .named = false,
  },
  [sym_GROUP] = {
    .visible = true,
    .named = false,
  },
  [sym_GUARDED] = {
    .visible = true,
    .named = false,
  },
  [sym_IF] = {
    .visible = true,
    .named = false,
  },
  [sym_IMPURE] = {
    .visible = true,
    .named = false,
  },
  [sym_IN] = {
    .visible = true,
    .named = false,
  },
  [sym_INERTIAL] = {
    .visible = true,
    .named = false,
  },
  [sym_INOUT] = {
    .visible = true,
    .named = false,
  },
  [sym_IS] = {
    .visible = true,
    .named = false,
  },
  [sym_LABEL] = {
    .visible = true,
    .named = false,
  },
  [sym_LIBRARY] = {
    .visible = true,
    .named = false,
  },
  [sym_LINKAGE] = {
    .visible = true,
    .named = false,
  },
  [sym_LITERAL] = {
    .visible = true,
    .named = false,
  },
  [sym_LOOP] = {
    .visible = true,
    .named = false,
  },
  [sym_MAP] = {
    .visible = true,
    .named = false,
  },
  [sym_MOD] = {
    .visible = true,
    .named = false,
  },
  [sym_NAND] = {
    .visible = true,
    .named = false,
  },
  [sym_NEW] = {
    .visible = true,
    .named = false,
  },
  [sym_NEXT] = {
    .visible = true,
    .named = false,
  },
  [sym_NOR] = {
    .visible = true,
    .named = false,
  },
  [sym_NOT] = {
    .visible = true,
    .named = false,
  },
  [sym_NULL] = {
    .visible = true,
    .named = false,
  },
  [sym_OF] = {
    .visible = true,
    .named = false,
  },
  [sym_ON] = {
    .visible = true,
    .named = false,
  },
  [sym_OPEN] = {
    .visible = true,
    .named = true,
  },
  [sym_OR] = {
    .visible = true,
    .named = false,
  },
  [sym_OTHERS] = {
    .visible = true,
    .named = true,
  },
  [sym_OUT] = {
    .visible = true,
    .named = false,
  },
  [sym_PACKAGE] = {
    .visible = true,
    .named = false,
  },
  [sym_PARAMETER] = {
    .visible = true,
    .named = false,
  },
  [sym_PORT] = {
    .visible = true,
    .named = false,
  },
  [sym_POSTPONED] = {
    .visible = true,
    .named = false,
  },
  [sym_PROCEDURE] = {
    .visible = true,
    .named = false,
  },
  [sym_PROCESS] = {
    .visible = true,
    .named = false,
  },
  [sym_PROPERTY] = {
    .visible = true,
    .named = false,
  },
  [sym_PROTECTED] = {
    .visible = true,
    .named = false,
  },
  [sym_PRIVATE] = {
    .visible = true,
    .named = false,
  },
  [sym_PURE] = {
    .visible = true,
    .named = false,
  },
  [sym_RANGE] = {
    .visible = true,
    .named = false,
  },
  [sym_RECORD] = {
    .visible = true,
    .named = false,
  },
  [sym_REGISTER] = {
    .visible = true,
    .named = false,
  },
  [sym_REJECT] = {
    .visible = true,
    .named = false,
  },
  [sym_RELEASE] = {
    .visible = true,
    .named = false,
  },
  [sym_REM] = {
    .visible = true,
    .named = false,
  },
  [sym_REPORT] = {
    .visible = true,
    .named = false,
  },
  [sym_RESTRICT] = {
    .visible = true,
    .named = true,
  },
  [sym_RETURN] = {
    .visible = true,
    .named = false,
  },
  [sym_ROL] = {
    .visible = true,
    .named = false,
  },
  [sym_ROR] = {
    .visible = true,
    .named = false,
  },
  [sym_SELECT] = {
    .visible = true,
    .named = false,
  },
  [sym_SEQUENCE] = {
    .visible = true,
    .named = false,
  },
  [sym_SEVERITY] = {
    .visible = true,
    .named = false,
  },
  [sym_SIGNAL] = {
    .visible = true,
    .named = false,
  },
  [sym_SHARED] = {
    .visible = true,
    .named = false,
  },
  [sym_SLA] = {
    .visible = true,
    .named = false,
  },
  [sym_SLL] = {
    .visible = true,
    .named = false,
  },
  [sym_SRA] = {
    .visible = true,
    .named = false,
  },
  [sym_SRL] = {
    .visible = true,
    .named = false,
  },
  [sym_STRONG] = {
    .visible = true,
    .named = true,
  },
  [sym_SUBTYPE] = {
    .visible = true,
    .named = false,
  },
  [sym_THEN] = {
    .visible = true,
    .named = false,
  },
  [sym_TO] = {
    .visible = true,
    .named = false,
  },
  [sym_TRANSPORT] = {
    .visible = true,
    .named = false,
  },
  [sym_TYPE] = {
    .visible = true,
    .named = false,
  },
  [sym_UNAFFECTED] = {
    .visible = true,
    .named = false,
  },
  [sym_UNITS] = {
    .visible = true,
    .named = false,
  },
  [sym_UNTIL] = {
    .visible = true,
    .named = false,
  },
  [sym_USE] = {
    .visible = true,
    .named = false,
  },
  [sym_VARIABLE] = {
    .visible = true,
    .named = false,
  },
  [sym_VIEW] = {
    .visible = true,
    .named = false,
  },
  [sym_VMODE] = {
    .visible = true,
    .named = true,
  },
  [sym_VPKG] = {
    .visible = true,
    .named = true,
  },
  [sym_VPROP] = {
    .visible = true,
    .named = true,
  },
  [sym_VUNIT] = {
    .visible = true,
    .named = false,
  },
  [sym_WAIT] = {
    .visible = true,
    .named = false,
  },
  [sym_WHEN] = {
    .visible = true,
    .named = false,
  },
  [sym_WHILE] = {
    .visible = true,
    .named = false,
  },
  [sym_WITH] = {
    .visible = true,
    .named = false,
  },
  [sym_XNOR] = {
    .visible = true,
    .named = false,
  },
  [sym_XOR] = {
    .visible = true,
    .named = false,
  },
  [sym_reserved_end_marker] = {
    .visible = true,
    .named = true,
  },
  [sym_directive_body] = {
    .visible = true,
    .named = true,
  },
  [sym_directive_constant_builtin] = {
    .visible = true,
    .named = true,
  },
  [sym_directive_error] = {
    .visible = true,
    .named = true,
  },
  [sym_directive_protect] = {
    .visible = true,
    .named = true,
  },
  [sym_directive_warning] = {
    .visible = true,
    .named = true,
  },
  [sym__directive_newline] = {
    .visible = false,
    .named = true,
  },
  [sym__grave_accent] = {
    .visible = false,
    .named = true,
  },
  [sym_box] = {
    .visible = true,
    .named = false,
  },
  [sym_delimiter_end_marker] = {
    .visible = true,
    .named = true,
  },
  [sym_decimal_integer] = {
    .visible = true,
    .named = true,
  },
  [sym_decimal_float] = {
    .visible = true,
    .named = true,
  },
  [sym_based_base] = {
    .visible = true,
    .named = true,
  },
  [sym_based_integer] = {
    .visible = true,
    .named = true,
  },
  [sym_based_float] = {
    .visible = true,
    .named = true,
  },
  [sym_character_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_string_literal_std_logic] = {
    .visible = true,
    .named = true,
  },
  [sym_bit_string_length] = {
    .visible = true,
    .named = true,
  },
  [sym_bit_string_base] = {
    .visible = true,
    .named = true,
  },
  [sym_bit_string_value] = {
    .visible = true,
    .named = true,
  },
  [sym_operator_symbol] = {
    .visible = true,
    .named = true,
  },
  [sym__line_comment_start] = {
    .visible = false,
    .named = true,
  },
  [sym__block_comment_start] = {
    .visible = false,
    .named = true,
  },
  [sym__block_comment_end] = {
    .visible = false,
    .named = true,
  },
  [sym_comment_content] = {
    .visible = true,
    .named = true,
  },
  [sym_token_end_marker] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_function] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_impure_function] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_mode_view] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_pure_function] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_range] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_signal] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_subtype] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_type] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_value] = {
    .visible = true,
    .named = true,
  },
  [sym_library_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_library_constant] = {
    .visible = true,
    .named = true,
  },
  [sym_library_constant_boolean] = {
    .visible = true,
    .named = true,
  },
  [sym_library_constant_character] = {
    .visible = true,
    .named = true,
  },
  [sym_library_constant_debug] = {
    .visible = true,
    .named = true,
  },
  [sym_library_constant_env] = {
    .visible = true,
    .named = true,
  },
  [sym_library_constant_standard] = {
    .visible = true,
    .named = true,
  },
  [sym_library_constant_std_logic] = {
    .visible = true,
    .named = true,
  },
  [sym_library_constant_unit] = {
    .visible = true,
    .named = true,
  },
  [sym_library_function] = {
    .visible = true,
    .named = true,
  },
  [sym_library_namespace] = {
    .visible = true,
    .named = true,
  },
  [sym_library_type] = {
    .visible = true,
    .named = true,
  },
  [sym__end_of_file] = {
    .visible = false,
    .named = true,
  },
  [sym_error_sentinel] = {
    .visible = true,
    .named = true,
  },
  [sym_design_file] = {
    .visible = true,
    .named = true,
  },
  [sym_design_unit] = {
    .visible = true,
    .named = true,
  },
  [sym__context_item] = {
    .visible = false,
    .named = true,
  },
  [sym_library_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_use_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_context_reference] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_name_list] = {
    .visible = true,
    .named = true,
  },
  [sym__logical_name] = {
    .visible = false,
    .named = true,
  },
  [sym_selected_name_list] = {
    .visible = true,
    .named = true,
  },
  [sym_selected_name] = {
    .visible = true,
    .named = true,
  },
  [sym__library_unit] = {
    .visible = false,
    .named = true,
  },
  [sym_entity_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_entity_head] = {
    .visible = true,
    .named = true,
  },
  [sym_port_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_entity_body] = {
    .visible = true,
    .named = true,
  },
  [sym_end_entity] = {
    .visible = true,
    .named = true,
  },
  [sym_configuration_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_configuration_head] = {
    .visible = true,
    .named = true,
  },
  [sym_end_configuration] = {
    .visible = true,
    .named = true,
  },
  [sym_package_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_package_declaration_body] = {
    .visible = true,
    .named = true,
  },
  [sym_end_package] = {
    .visible = true,
    .named = true,
  },
  [sym_package_instantiation_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_package_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_context_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_context_declaration_body] = {
    .visible = true,
    .named = true,
  },
  [sym_end_context] = {
    .visible = true,
    .named = true,
  },
  [sym_architecture_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_architecture_head] = {
    .visible = true,
    .named = true,
  },
  [sym_end_architecture] = {
    .visible = true,
    .named = true,
  },
  [sym_package_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_package_definition_body] = {
    .visible = true,
    .named = true,
  },
  [sym_end_package_body] = {
    .visible = true,
    .named = true,
  },
  [sym__entity_declarative_item] = {
    .visible = false,
    .named = true,
  },
  [sym__package_declarative_item] = {
    .visible = false,
    .named = true,
  },
  [sym__package_body_declarative_item] = {
    .visible = false,
    .named = true,
  },
  [sym__protected_type_declarative_item] = {
    .visible = false,
    .named = true,
  },
  [sym__protected_type_body_declarative_item] = {
    .visible = false,
    .named = true,
  },
  [sym__process_declarative_item] = {
    .visible = false,
    .named = true,
  },
  [sym__subprogram_declarative_item] = {
    .visible = false,
    .named = true,
  },
  [sym__block_declarative_item] = {
    .visible = false,
    .named = true,
  },
  [sym_subprogram_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_subprogram_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_subprogram_head] = {
    .visible = true,
    .named = true,
  },
  [sym_subprogram_end] = {
    .visible = true,
    .named = true,
  },
  [sym_subprogram_instantiation_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_type_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_subtype_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_mode_view_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_mode_view_body] = {
    .visible = true,
    .named = true,
  },
  [sym_end_view] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_signal_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_signal_kind] = {
    .visible = true,
    .named = true,
  },
  [sym_variable_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_initialiser] = {
    .visible = true,
    .named = true,
  },
  [sym_file_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_file_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_alias_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_component_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_component_body] = {
    .visible = true,
    .named = true,
  },
  [sym_end_component] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_disconnection_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_group_template_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_group_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_private_variable_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__interface_type_indication] = {
    .visible = false,
    .named = true,
  },
  [sym_unspecified_type_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_type_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__incomplete_type_definition] = {
    .visible = false,
    .named = true,
  },
  [sym_private_incomplete_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_scalar_incomplete_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_discrete_incomplete_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_integer_incomplete_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_physical_incomplete_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_floating_incomplete_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_array_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_array_index_incomplete_type_list] = {
    .visible = true,
    .named = true,
  },
  [sym__array_index_incomplete_type] = {
    .visible = false,
    .named = true,
  },
  [sym_index_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_index_subtype_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_access_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_access_incomplete_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_file_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_file_incomplete_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_subtype_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_resolution_indication] = {
    .visible = true,
    .named = true,
  },
  [sym__element_resolution] = {
    .visible = false,
    .named = true,
  },
  [sym_record_resolution] = {
    .visible = true,
    .named = true,
  },
  [sym_record_element_resolution] = {
    .visible = true,
    .named = true,
  },
  [sym__incomplete_subtype_indication] = {
    .visible = false,
    .named = true,
  },
  [sym__incomplete_type_mark] = {
    .visible = false,
    .named = true,
  },
  [sym_concurrent_block] = {
    .visible = true,
    .named = true,
  },
  [sym_sequential_block] = {
    .visible = true,
    .named = true,
  },
  [sym_generate_body] = {
    .visible = true,
    .named = true,
  },
  [sym_generate_head] = {
    .visible = true,
    .named = true,
  },
  [sym_case_generate_alternative] = {
    .visible = true,
    .named = true,
  },
  [sym_case_generate_body] = {
    .visible = true,
    .named = true,
  },
  [sym_generate_block] = {
    .visible = true,
    .named = true,
  },
  [sym_generate_block_end] = {
    .visible = true,
    .named = true,
  },
  [sym__entity_statement] = {
    .visible = false,
    .named = true,
  },
  [sym__concurrent_statement] = {
    .visible = false,
    .named = true,
  },
  [sym__sequential_statement] = {
    .visible = false,
    .named = true,
  },
  [sym_block_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_guard_condition] = {
    .visible = true,
    .named = true,
  },
  [sym_block_head] = {
    .visible = true,
    .named = true,
  },
  [sym_end_block] = {
    .visible = true,
    .named = true,
  },
  [sym_sequential_block_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_sequential_block_head] = {
    .visible = true,
    .named = true,
  },
  [sym_component_instantiation_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_instantiated_unit] = {
    .visible = true,
    .named = true,
  },
  [sym_process_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_sensitivity_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_process_head] = {
    .visible = true,
    .named = true,
  },
  [sym_end_process] = {
    .visible = true,
    .named = true,
  },
  [sym_case_generate_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_case_generate_block] = {
    .visible = true,
    .named = true,
  },
  [sym_for_generate_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_if_generate_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_if_generate] = {
    .visible = true,
    .named = true,
  },
  [sym__elsif_else_generate] = {
    .visible = false,
    .named = true,
  },
  [sym_elsif_generate] = {
    .visible = true,
    .named = true,
  },
  [sym_else_generate] = {
    .visible = true,
    .named = true,
  },
  [sym_end_generate] = {
    .visible = true,
    .named = true,
  },
  [sym_assertion_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_concurrent_assertion_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_assertion] = {
    .visible = true,
    .named = true,
  },
  [sym_report_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_severity_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_case_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_case_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_case_body] = {
    .visible = true,
    .named = true,
  },
  [sym_end_case] = {
    .visible = true,
    .named = true,
  },
  [sym_exit_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_when_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_if_statement_block] = {
    .visible = true,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__elsif_else_statement] = {
    .visible = false,
    .named = true,
  },
  [sym_elsif_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_else_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_if_statement_body] = {
    .visible = true,
    .named = true,
  },
  [sym_end_if] = {
    .visible = true,
    .named = true,
  },
  [sym_loop_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__iteration_scheme] = {
    .visible = false,
    .named = true,
  },
  [sym_loop_body] = {
    .visible = true,
    .named = true,
  },
  [sym_end_loop] = {
    .visible = true,
    .named = true,
  },
  [sym_next_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_null_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_procedure_call_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_concurrent_procedure_call_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_report_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_return_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_signal_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_waveform_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_variable_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_concurrent_simple_signal_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_force_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_release_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_signal_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_concurrent_conditional_signal_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_selected_waveform_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_concurrent_selected_signal_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_selected_force_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_selected_variable_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym_with_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_select_target] = {
    .visible = true,
    .named = true,
  },
  [sym_wait_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_else_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
  },
  [sym_condition_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_relational_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_shift_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_simple_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_logical_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_relational_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_shift_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_sign] = {
    .visible = true,
    .named = true,
  },
  [sym_adding_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_multiplying_operator] = {
    .visible = true,
    .named = true,
  },
  [sym__primary] = {
    .visible = false,
    .named = true,
  },
  [sym_name] = {
    .visible = true,
    .named = true,
  },
  [sym__direct_name] = {
    .visible = false,
    .named = true,
  },
  [sym__label] = {
    .visible = false,
    .named = true,
  },
  [sym__attribute] = {
    .visible = false,
    .named = true,
  },
  [sym__unit] = {
    .visible = false,
    .named = true,
  },
  [sym__identifier] = {
    .visible = false,
    .named = true,
  },
  [sym__name_selector] = {
    .visible = false,
    .named = true,
  },
  [sym__external_name] = {
    .visible = false,
    .named = true,
  },
  [sym_external_constant_name] = {
    .visible = true,
    .named = true,
  },
  [sym_external_signal_name] = {
    .visible = true,
    .named = true,
  },
  [sym_external_variable_name] = {
    .visible = true,
    .named = true,
  },
  [sym__external_pathname] = {
    .visible = false,
    .named = true,
  },
  [sym_package_pathname] = {
    .visible = true,
    .named = true,
  },
  [sym_absolute_pathname] = {
    .visible = true,
    .named = true,
  },
  [sym_relative_pathname] = {
    .visible = true,
    .named = true,
  },
  [sym_partial_pathname] = {
    .visible = true,
    .named = true,
  },
  [sym_pathname_element] = {
    .visible = true,
    .named = true,
  },
  [sym_function_call] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesis_group] = {
    .visible = true,
    .named = true,
  },
  [sym_association_or_range_list] = {
    .visible = true,
    .named = true,
  },
  [sym__association_or_range] = {
    .visible = false,
    .named = true,
  },
  [sym_association_element] = {
    .visible = true,
    .named = true,
  },
  [sym__actual_part] = {
    .visible = false,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym__attribute_designator] = {
    .visible = false,
    .named = true,
  },
  [sym_signature] = {
    .visible = true,
    .named = true,
  },
  [sym_selection] = {
    .visible = true,
    .named = true,
  },
  [sym__suffix] = {
    .visible = false,
    .named = true,
  },
  [sym__literal] = {
    .visible = false,
    .named = true,
  },
  [sym_bit_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym__abstract_literal] = {
    .visible = false,
    .named = true,
  },
  [sym_based_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_allocator] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesis_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_element_association_list] = {
    .visible = true,
    .named = true,
  },
  [sym_element_association] = {
    .visible = true,
    .named = true,
  },
  [sym__element] = {
    .visible = false,
    .named = true,
  },
  [sym_choices] = {
    .visible = true,
    .named = true,
  },
  [sym__range] = {
    .visible = false,
    .named = true,
  },
  [sym_simple_range] = {
    .visible = true,
    .named = true,
  },
  [sym__direction] = {
    .visible = false,
    .named = true,
  },
  [sym__tool_directive] = {
    .visible = false,
    .named = true,
  },
  [sym__directive] = {
    .visible = false,
    .named = true,
  },
  [sym_user_directive] = {
    .visible = true,
    .named = true,
  },
  [sym_protect_directive] = {
    .visible = true,
    .named = true,
  },
  [sym_warning_directive] = {
    .visible = true,
    .named = true,
  },
  [sym_error_directive] = {
    .visible = true,
    .named = true,
  },
  [sym_if_conditional_analysis] = {
    .visible = true,
    .named = true,
  },
  [sym_elsif_conditional_analysis] = {
    .visible = true,
    .named = true,
  },
  [sym_else_conditional_analysis] = {
    .visible = true,
    .named = true,
  },
  [sym_end_conditional_analysis] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_analysis_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_analysis_relation] = {
    .visible = true,
    .named = true,
  },
  [sym__conditional_analysis_operator] = {
    .visible = false,
    .named = true,
  },
  [sym__conditional_analysis_identifier] = {
    .visible = false,
    .named = true,
  },
  [sym__configuration_declarative_item] = {
    .visible = false,
    .named = true,
  },
  [sym__configuration_item] = {
    .visible = false,
    .named = true,
  },
  [sym_block_configuration] = {
    .visible = true,
    .named = true,
  },
  [sym_component_configuration] = {
    .visible = true,
    .named = true,
  },
  [sym_end_for] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_interface_list] = {
    .visible = true,
    .named = true,
  },
  [sym__generic_interface_declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_generic_interface_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_list] = {
    .visible = true,
    .named = true,
  },
  [sym__interface_declaration] = {
    .visible = false,
    .named = true,
  },
  [sym_interface_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_constant_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_signal_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_variable_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_subprogram_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__interface_subprogram_specification] = {
    .visible = false,
    .named = true,
  },
  [sym_procedure_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_procedure_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_function_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_interface_function_specification] = {
    .visible = true,
    .named = true,
  },
  [sym__interface_package_generic_map_aspect] = {
    .visible = false,
    .named = true,
  },
  [sym_generic_map_any] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_map_default] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_map_aspect] = {
    .visible = true,
    .named = true,
  },
  [sym_port_map_aspect] = {
    .visible = true,
    .named = true,
  },
  [sym_association_list] = {
    .visible = true,
    .named = true,
  },
  [sym__interface_subprogram_default] = {
    .visible = false,
    .named = true,
  },
  [sym__designator] = {
    .visible = false,
    .named = true,
  },
  [sym_generic_identifier_list] = {
    .visible = true,
    .named = true,
  },
  [sym_constant_identifier_list] = {
    .visible = true,
    .named = true,
  },
  [sym_identifier_list] = {
    .visible = true,
    .named = true,
  },
  [sym__mode_indication] = {
    .visible = false,
    .named = true,
  },
  [sym_simple_mode_indication] = {
    .visible = true,
    .named = true,
  },
  [sym__mode_view_indication] = {
    .visible = false,
    .named = true,
  },
  [sym_record_mode_view_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_array_mode_view_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_mode_view_element_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_record_element_list] = {
    .visible = true,
    .named = true,
  },
  [sym__element_mode_indication] = {
    .visible = false,
    .named = true,
  },
  [sym__element_mode_view_indication] = {
    .visible = false,
    .named = true,
  },
  [sym_element_record_mode_view_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_element_array_mode_view_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_mode] = {
    .visible = true,
    .named = true,
  },
  [sym_range_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_group_constituent_list] = {
    .visible = true,
    .named = true,
  },
  [sym_entity_class_entry_list] = {
    .visible = true,
    .named = true,
  },
  [sym_entity_class_entry] = {
    .visible = true,
    .named = true,
  },
  [sym_guarded_signal_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_signal_list] = {
    .visible = true,
    .named = true,
  },
  [sym_entity_specification] = {
    .visible = true,
    .named = true,
  },
  [sym__entity_class] = {
    .visible = false,
    .named = true,
  },
  [sym_entity_name_list] = {
    .visible = true,
    .named = true,
  },
  [sym_entity_designator] = {
    .visible = true,
    .named = true,
  },
  [sym__entity_tag] = {
    .visible = false,
    .named = true,
  },
  [sym__alias_designator] = {
    .visible = false,
    .named = true,
  },
  [sym_file_open_information] = {
    .visible = true,
    .named = true,
  },
  [sym_file_logical_name] = {
    .visible = true,
    .named = true,
  },
  [sym_package_header] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_subprogram_header] = {
    .visible = true,
    .named = true,
  },
  [sym__subprogram_specification] = {
    .visible = false,
    .named = true,
  },
  [sym_parameter_list_specification] = {
    .visible = true,
    .named = true,
  },
  [sym__type_definition] = {
    .visible = false,
    .named = true,
  },
  [sym_protected_type_instantiation_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_protected_type_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_protected_type_declaration_end] = {
    .visible = true,
    .named = true,
  },
  [sym_protected_type_header] = {
    .visible = true,
    .named = true,
  },
  [sym_protected_type_body] = {
    .visible = true,
    .named = true,
  },
  [sym_protected_type_body_end] = {
    .visible = true,
    .named = true,
  },
  [sym_while_loop] = {
    .visible = true,
    .named = true,
  },
  [sym_for_loop] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_case_statement_alternative] = {
    .visible = true,
    .named = true,
  },
  [sym_selected_waveforms] = {
    .visible = true,
    .named = true,
  },
  [sym_waveform] = {
    .visible = true,
    .named = true,
  },
  [sym_waveform_element] = {
    .visible = true,
    .named = true,
  },
  [sym_force_mode] = {
    .visible = true,
    .named = true,
  },
  [sym_selected_expressions] = {
    .visible = true,
    .named = true,
  },
  [sym_when_element] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_waveforms] = {
    .visible = true,
    .named = true,
  },
  [sym_else_waveform] = {
    .visible = true,
    .named = true,
  },
  [sym_delay_mechanism] = {
    .visible = true,
    .named = true,
  },
  [sym__target] = {
    .visible = false,
    .named = true,
  },
  [sym_aggregate] = {
    .visible = true,
    .named = true,
  },
  [sym_conditional_or_unaffected_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_else_expression_or_unaffected] = {
    .visible = true,
    .named = true,
  },
  [sym__expression_or_unaffected] = {
    .visible = false,
    .named = true,
  },
  [sym_label_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_sensitivity_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_sensitivity_list] = {
    .visible = true,
    .named = true,
  },
  [sym_condition_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_timeout_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_physical_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_primary_unit_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_secondary_unit_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_end_units] = {
    .visible = true,
    .named = true,
  },
  [sym_enumeration_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_enumeration_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_record_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_end_record] = {
    .visible = true,
    .named = true,
  },
  [sym_element_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__process_sensitivity_list] = {
    .visible = false,
    .named = true,
  },
  [sym_configuration_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_verification_unit_binding_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_verification_unit_list] = {
    .visible = true,
    .named = true,
  },
  [sym_component_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_instantiation_list] = {
    .visible = true,
    .named = true,
  },
  [sym_binding_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_entity_aspect] = {
    .visible = true,
    .named = true,
  },
  [sym_line_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_block_comment] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_design_file_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_design_unit_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_design_unit_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_design_unit_repeat3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_logical_name_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_selected_name_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_selected_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_entity_head_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_entity_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_configuration_head_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_configuration_head_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_package_declaration_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_package_definition_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_subprogram_head_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_mode_view_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_array_index_incomplete_type_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_index_constraint_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_resolution_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_sequential_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_sequential_block_head_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_case_generate_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_case_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_conditional_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_name_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_package_pathname_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_relative_pathname_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_partial_pathname_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_association_or_range_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_signature_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_element_association_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_user_directive_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_conditional_analysis_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_block_configuration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_block_configuration_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_generic_interface_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_interface_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_association_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_generic_identifier_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_constant_identifier_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_identifier_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_group_constituent_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_entity_class_entry_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_entity_name_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_protected_type_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_protected_type_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_selected_waveforms_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_waveform_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_selected_expressions_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_conditional_waveforms_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_aggregate_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_conditional_or_unaffected_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_physical_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enumeration_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_instantiation_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_attribute_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_label] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_unit] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_architecture = 1,
  field_attribute = 2,
  field_component = 3,
  field_configuration = 4,
  field_constant = 5,
  field_entity = 6,
  field_function = 7,
  field_generic = 8,
  field_library = 9,
  field_package = 10,
  field_procedure = 11,
  field_type = 12,
  field_unit = 13,
  field_view = 14,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_architecture] = "architecture",
  [field_attribute] = "attribute",
  [field_component] = "component",
  [field_configuration] = "configuration",
  [field_constant] = "constant",
  [field_entity] = "entity",
  [field_function] = "function",
  [field_generic] = "generic",
  [field_library] = "library",
  [field_package] = "package",
  [field_procedure] = "procedure",
  [field_type] = "type",
  [field_unit] = "unit",
  [field_view] = "view",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [2] = {.index = 0, .length = 1},
  [4] = {.index = 1, .length = 1},
  [5] = {.index = 2, .length = 1},
  [6] = {.index = 3, .length = 1},
  [7] = {.index = 4, .length = 1},
  [8] = {.index = 5, .length = 1},
  [10] = {.index = 6, .length = 1},
  [11] = {.index = 7, .length = 2},
  [12] = {.index = 9, .length = 2},
  [13] = {.index = 11, .length = 1},
  [14] = {.index = 12, .length = 1},
  [15] = {.index = 13, .length = 1},
  [16] = {.index = 14, .length = 1},
  [17] = {.index = 15, .length = 1},
  [18] = {.index = 16, .length = 1},
  [19] = {.index = 17, .length = 1},
  [20] = {.index = 18, .length = 2},
  [21] = {.index = 20, .length = 2},
  [22] = {.index = 22, .length = 1},
  [23] = {.index = 23, .length = 2},
  [24] = {.index = 25, .length = 1},
  [25] = {.index = 26, .length = 2},
  [26] = {.index = 28, .length = 1},
  [27] = {.index = 29, .length = 1},
  [28] = {.index = 30, .length = 2},
  [29] = {.index = 32, .length = 2},
  [30] = {.index = 34, .length = 1},
  [31] = {.index = 35, .length = 2},
  [32] = {.index = 37, .length = 2},
  [33] = {.index = 39, .length = 1},
  [34] = {.index = 40, .length = 2},
  [35] = {.index = 42, .length = 2},
  [36] = {.index = 44, .length = 1},
  [37] = {.index = 45, .length = 2},
  [38] = {.index = 47, .length = 2},
  [39] = {.index = 49, .length = 1},
  [40] = {.index = 50, .length = 2},
  [41] = {.index = 52, .length = 1},
  [42] = {.index = 53, .length = 2},
  [43] = {.index = 55, .length = 1},
  [44] = {.index = 56, .length = 2},
  [45] = {.index = 58, .length = 1},
  [46] = {.index = 59, .length = 2},
  [47] = {.index = 61, .length = 2},
  [48] = {.index = 63, .length = 3},
  [49] = {.index = 66, .length = 1},
  [50] = {.index = 67, .length = 2},
  [51] = {.index = 69, .length = 2},
  [52] = {.index = 71, .length = 2},
  [53] = {.index = 73, .length = 2},
  [54] = {.index = 75, .length = 1},
  [55] = {.index = 76, .length = 1},
  [56] = {.index = 77, .length = 3},
  [57] = {.index = 80, .length = 1},
  [58] = {.index = 81, .length = 1},
  [59] = {.index = 82, .length = 1},
  [60] = {.index = 83, .length = 2},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_unit, 0, .inherited = true},
  [1] =
    {field_attribute, 0},
  [2] =
    {field_constant, 0},
  [3] =
    {field_library, 0},
  [4] =
    {field_library, 0, .inherited = true},
  [5] =
    {field_procedure, 1},
  [6] =
    {field_unit, 1},
  [7] =
    {field_constant, 0},
    {field_constant, 1, .inherited = true},
  [9] =
    {field_library, 0, .inherited = true},
    {field_library, 1, .inherited = true},
  [11] =
    {field_type, 1},
  [12] =
    {field_attribute, 1, .inherited = true},
  [13] =
    {field_component, 1},
  [14] =
    {field_configuration, 1},
  [15] =
    {field_entity, 1},
  [16] =
    {field_type, 0},
  [17] =
    {field_constant, 1},
  [18] =
    {field_constant, 0, .inherited = true},
    {field_constant, 1, .inherited = true},
  [20] =
    {field_library, 0, .inherited = true},
    {field_package, 2},
  [22] =
    {field_type, 2},
  [23] =
    {field_function, 1},
    {field_type, 3},
  [25] =
    {field_library, 1, .inherited = true},
  [26] =
    {field_attribute, 1},
    {field_type, 3},
  [28] =
    {field_component, 2},
  [29] =
    {field_generic, 0},
  [30] =
    {field_function, 1},
    {field_type, 4},
  [32] =
    {field_function, 2},
    {field_type, 4},
  [34] =
    {field_package, 1},
  [35] =
    {field_type, 1},
    {field_type, 2, .inherited = true},
  [37] =
    {field_type, 0, .inherited = true},
    {field_type, 1, .inherited = true},
  [39] =
    {field_function, 1},
  [40] =
    {field_entity, 3},
    {field_library, 1},
  [42] =
    {field_generic, 0},
    {field_generic, 1, .inherited = true},
  [44] =
    {field_entity, 2},
  [45] =
    {field_function, 1},
    {field_type, 5},
  [47] =
    {field_function, 2},
    {field_type, 5},
  [49] =
    {field_package, 2},
  [50] =
    {field_type, 1},
    {field_type, 3},
  [52] =
    {field_function, 2},
  [53] =
    {field_architecture, 3},
    {field_entity, 1},
  [55] =
    {field_generic, 1},
  [56] =
    {field_generic, 0, .inherited = true},
    {field_generic, 1, .inherited = true},
  [58] =
    {field_view, 1},
  [59] =
    {field_function, 1},
    {field_type, 6},
  [61] =
    {field_function, 2},
    {field_type, 6},
  [63] =
    {field_type, 1},
    {field_type, 2, .inherited = true},
    {field_type, 4},
  [66] =
    {field_architecture, 1},
  [67] =
    {field_architecture, 1},
    {field_entity, 3},
  [69] =
    {field_configuration, 1},
    {field_entity, 3},
  [71] =
    {field_function, 1},
    {field_type, 7},
  [73] =
    {field_function, 2},
    {field_type, 7},
  [75] =
    {field_package, 3},
  [76] =
    {field_view, 2},
  [77] =
    {field_architecture, 5},
    {field_entity, 3},
    {field_library, 1},
  [80] =
    {field_architecture, 2},
  [81] =
    {field_configuration, 2},
  [82] =
    {field_type, 2, .inherited = true},
  [83] =
    {field_function, 2},
    {field_type, 8},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = alias_sym_label,
  },
  [3] = {
    [0] = alias_sym_attribute_identifier,
  },
  [9] = {
    [0] = alias_sym_unit,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 4,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 9,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 12,
  [17] = 17,
  [18] = 18,
  [19] = 19,
  [20] = 20,
  [21] = 21,
  [22] = 22,
  [23] = 23,
  [24] = 24,
  [25] = 25,
  [26] = 26,
  [27] = 27,
  [28] = 11,
  [29] = 29,
  [30] = 17,
  [31] = 17,
  [32] = 17,
  [33] = 17,
  [34] = 17,
  [35] = 17,
  [36] = 17,
  [37] = 17,
  [38] = 17,
  [39] = 17,
  [40] = 17,
  [41] = 17,
  [42] = 17,
  [43] = 17,
  [44] = 17,
  [45] = 17,
  [46] = 17,
  [47] = 17,
  [48] = 17,
  [49] = 17,
  [50] = 17,
  [51] = 17,
  [52] = 17,
  [53] = 17,
  [54] = 17,
  [55] = 17,
  [56] = 17,
  [57] = 17,
  [58] = 17,
  [59] = 17,
  [60] = 17,
  [61] = 12,
  [62] = 17,
  [63] = 17,
  [64] = 17,
  [65] = 17,
  [66] = 17,
  [67] = 17,
  [68] = 17,
  [69] = 17,
  [70] = 17,
  [71] = 17,
  [72] = 17,
  [73] = 17,
  [74] = 17,
  [75] = 17,
  [76] = 17,
  [77] = 17,
  [78] = 17,
  [79] = 17,
  [80] = 25,
  [81] = 27,
  [82] = 82,
  [83] = 18,
  [84] = 19,
  [85] = 20,
  [86] = 22,
  [87] = 25,
  [88] = 27,
  [89] = 82,
  [90] = 18,
  [91] = 19,
  [92] = 20,
  [93] = 22,
  [94] = 94,
  [95] = 25,
  [96] = 27,
  [97] = 20,
  [98] = 82,
  [99] = 17,
  [100] = 100,
  [101] = 101,
  [102] = 102,
  [103] = 100,
  [104] = 102,
  [105] = 100,
  [106] = 102,
  [107] = 101,
  [108] = 108,
  [109] = 109,
  [110] = 108,
  [111] = 111,
  [112] = 101,
  [113] = 108,
  [114] = 114,
  [115] = 102,
  [116] = 109,
  [117] = 100,
  [118] = 109,
  [119] = 101,
  [120] = 108,
  [121] = 109,
  [122] = 114,
  [123] = 114,
  [124] = 124,
  [125] = 114,
  [126] = 126,
  [127] = 126,
  [128] = 128,
  [129] = 129,
  [130] = 130,
  [131] = 131,
  [132] = 126,
  [133] = 130,
  [134] = 128,
  [135] = 129,
  [136] = 130,
  [137] = 126,
  [138] = 128,
  [139] = 128,
  [140] = 126,
  [141] = 130,
  [142] = 129,
  [143] = 129,
  [144] = 126,
  [145] = 126,
  [146] = 126,
  [147] = 126,
  [148] = 126,
  [149] = 126,
  [150] = 126,
  [151] = 126,
  [152] = 126,
  [153] = 126,
  [154] = 126,
  [155] = 126,
  [156] = 126,
  [157] = 126,
  [158] = 126,
  [159] = 126,
  [160] = 126,
  [161] = 126,
  [162] = 126,
  [163] = 126,
  [164] = 126,
  [165] = 126,
  [166] = 126,
  [167] = 126,
  [168] = 126,
  [169] = 126,
  [170] = 126,
  [171] = 126,
  [172] = 126,
  [173] = 126,
  [174] = 126,
  [175] = 126,
  [176] = 126,
  [177] = 126,
  [178] = 126,
  [179] = 126,
  [180] = 126,
  [181] = 126,
  [182] = 126,
  [183] = 126,
  [184] = 126,
  [185] = 126,
  [186] = 126,
  [187] = 126,
  [188] = 126,
  [189] = 126,
  [190] = 126,
  [191] = 126,
  [192] = 126,
  [193] = 126,
  [194] = 126,
  [195] = 126,
  [196] = 126,
  [197] = 126,
  [198] = 126,
  [199] = 126,
  [200] = 126,
  [201] = 126,
  [202] = 126,
  [203] = 126,
  [204] = 126,
  [205] = 126,
  [206] = 126,
  [207] = 126,
  [208] = 126,
  [209] = 126,
  [210] = 126,
  [211] = 126,
  [212] = 126,
  [213] = 126,
  [214] = 126,
  [215] = 126,
  [216] = 126,
  [217] = 126,
  [218] = 126,
  [219] = 126,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 223,
  [224] = 224,
  [225] = 225,
  [226] = 226,
  [227] = 227,
  [228] = 226,
  [229] = 229,
  [230] = 227,
  [231] = 231,
  [232] = 231,
  [233] = 231,
  [234] = 227,
  [235] = 229,
  [236] = 229,
  [237] = 226,
  [238] = 238,
  [239] = 239,
  [240] = 240,
  [241] = 241,
  [242] = 242,
  [243] = 240,
  [244] = 241,
  [245] = 245,
  [246] = 238,
  [247] = 247,
  [248] = 247,
  [249] = 249,
  [250] = 241,
  [251] = 238,
  [252] = 252,
  [253] = 238,
  [254] = 238,
  [255] = 238,
  [256] = 238,
  [257] = 238,
  [258] = 238,
  [259] = 238,
  [260] = 238,
  [261] = 238,
  [262] = 238,
  [263] = 238,
  [264] = 238,
  [265] = 238,
  [266] = 238,
  [267] = 238,
  [268] = 238,
  [269] = 245,
  [270] = 238,
  [271] = 238,
  [272] = 238,
  [273] = 238,
  [274] = 238,
  [275] = 238,
  [276] = 247,
  [277] = 238,
  [278] = 238,
  [279] = 238,
  [280] = 238,
  [281] = 238,
  [282] = 238,
  [283] = 238,
  [284] = 238,
  [285] = 238,
  [286] = 238,
  [287] = 249,
  [288] = 238,
  [289] = 289,
  [290] = 238,
  [291] = 239,
  [292] = 238,
  [293] = 238,
  [294] = 238,
  [295] = 295,
  [296] = 238,
  [297] = 249,
  [298] = 238,
  [299] = 238,
  [300] = 245,
  [301] = 238,
  [302] = 238,
  [303] = 238,
  [304] = 242,
  [305] = 238,
  [306] = 240,
  [307] = 238,
  [308] = 238,
  [309] = 241,
  [310] = 238,
  [311] = 238,
  [312] = 238,
  [313] = 238,
  [314] = 238,
  [315] = 289,
  [316] = 238,
  [317] = 239,
  [318] = 245,
  [319] = 238,
  [320] = 247,
  [321] = 321,
  [322] = 322,
  [323] = 249,
  [324] = 289,
  [325] = 242,
  [326] = 326,
  [327] = 327,
  [328] = 328,
  [329] = 329,
  [330] = 330,
  [331] = 329,
  [332] = 328,
  [333] = 327,
  [334] = 334,
  [335] = 329,
  [336] = 330,
  [337] = 328,
  [338] = 326,
  [339] = 339,
  [340] = 326,
  [341] = 330,
  [342] = 329,
  [343] = 343,
  [344] = 328,
  [345] = 345,
  [346] = 334,
  [347] = 326,
  [348] = 327,
  [349] = 330,
  [350] = 334,
  [351] = 351,
  [352] = 352,
  [353] = 353,
  [354] = 354,
  [355] = 353,
  [356] = 356,
  [357] = 352,
  [358] = 356,
  [359] = 359,
  [360] = 359,
  [361] = 354,
  [362] = 356,
  [363] = 353,
  [364] = 354,
  [365] = 359,
  [366] = 366,
  [367] = 367,
  [368] = 368,
  [369] = 369,
  [370] = 370,
  [371] = 371,
  [372] = 372,
  [373] = 369,
  [374] = 370,
  [375] = 370,
  [376] = 367,
  [377] = 370,
  [378] = 378,
  [379] = 367,
  [380] = 369,
  [381] = 381,
  [382] = 382,
  [383] = 378,
  [384] = 378,
  [385] = 382,
  [386] = 370,
  [387] = 370,
  [388] = 368,
  [389] = 368,
  [390] = 370,
  [391] = 366,
  [392] = 392,
  [393] = 393,
  [394] = 393,
  [395] = 395,
  [396] = 396,
  [397] = 397,
  [398] = 398,
  [399] = 393,
  [400] = 400,
  [401] = 401,
  [402] = 402,
  [403] = 393,
  [404] = 404,
  [405] = 405,
  [406] = 393,
  [407] = 393,
  [408] = 393,
  [409] = 401,
  [410] = 393,
  [411] = 401,
  [412] = 401,
  [413] = 413,
  [414] = 414,
  [415] = 415,
  [416] = 416,
  [417] = 417,
  [418] = 413,
  [419] = 414,
  [420] = 415,
  [421] = 416,
  [422] = 417,
  [423] = 413,
  [424] = 414,
  [425] = 415,
  [426] = 416,
  [427] = 417,
  [428] = 413,
  [429] = 414,
  [430] = 415,
  [431] = 416,
  [432] = 417,
  [433] = 413,
  [434] = 414,
  [435] = 415,
  [436] = 416,
  [437] = 417,
  [438] = 438,
  [439] = 414,
  [440] = 415,
  [441] = 416,
  [442] = 417,
  [443] = 413,
  [444] = 414,
  [445] = 415,
  [446] = 416,
  [447] = 417,
  [448] = 413,
  [449] = 414,
  [450] = 415,
  [451] = 416,
  [452] = 417,
  [453] = 413,
  [454] = 414,
  [455] = 415,
  [456] = 416,
  [457] = 417,
  [458] = 413,
  [459] = 414,
  [460] = 415,
  [461] = 416,
  [462] = 417,
  [463] = 413,
  [464] = 414,
  [465] = 415,
  [466] = 416,
  [467] = 417,
  [468] = 413,
  [469] = 414,
  [470] = 415,
  [471] = 416,
  [472] = 417,
  [473] = 413,
  [474] = 414,
  [475] = 415,
  [476] = 416,
  [477] = 417,
  [478] = 413,
  [479] = 414,
  [480] = 415,
  [481] = 416,
  [482] = 417,
  [483] = 413,
  [484] = 414,
  [485] = 415,
  [486] = 416,
  [487] = 417,
  [488] = 413,
  [489] = 414,
  [490] = 415,
  [491] = 416,
  [492] = 417,
  [493] = 413,
  [494] = 414,
  [495] = 438,
  [496] = 415,
  [497] = 416,
  [498] = 417,
  [499] = 413,
  [500] = 414,
  [501] = 415,
  [502] = 416,
  [503] = 417,
  [504] = 413,
  [505] = 414,
  [506] = 415,
  [507] = 416,
  [508] = 417,
  [509] = 413,
  [510] = 414,
  [511] = 415,
  [512] = 416,
  [513] = 417,
  [514] = 413,
  [515] = 414,
  [516] = 415,
  [517] = 416,
  [518] = 417,
  [519] = 413,
  [520] = 414,
  [521] = 415,
  [522] = 416,
  [523] = 417,
  [524] = 413,
  [525] = 414,
  [526] = 415,
  [527] = 416,
  [528] = 417,
  [529] = 529,
  [530] = 530,
  [531] = 531,
  [532] = 532,
  [533] = 533,
  [534] = 534,
  [535] = 535,
  [536] = 536,
  [537] = 537,
  [538] = 531,
  [539] = 534,
  [540] = 535,
  [541] = 536,
  [542] = 537,
  [543] = 543,
  [544] = 531,
  [545] = 534,
  [546] = 535,
  [547] = 536,
  [548] = 537,
  [549] = 531,
  [550] = 534,
  [551] = 535,
  [552] = 536,
  [553] = 537,
  [554] = 531,
  [555] = 534,
  [556] = 535,
  [557] = 536,
  [558] = 537,
  [559] = 531,
  [560] = 534,
  [561] = 535,
  [562] = 536,
  [563] = 537,
  [564] = 531,
  [565] = 534,
  [566] = 535,
  [567] = 536,
  [568] = 537,
  [569] = 531,
  [570] = 534,
  [571] = 535,
  [572] = 536,
  [573] = 537,
  [574] = 531,
  [575] = 534,
  [576] = 535,
  [577] = 536,
  [578] = 537,
  [579] = 534,
  [580] = 535,
  [581] = 536,
  [582] = 537,
  [583] = 534,
  [584] = 535,
  [585] = 536,
  [586] = 537,
  [587] = 534,
  [588] = 535,
  [589] = 536,
  [590] = 537,
  [591] = 591,
  [592] = 534,
  [593] = 535,
  [594] = 536,
  [595] = 537,
  [596] = 534,
  [597] = 535,
  [598] = 536,
  [599] = 537,
  [600] = 534,
  [601] = 535,
  [602] = 536,
  [603] = 537,
  [604] = 534,
  [605] = 535,
  [606] = 536,
  [607] = 537,
  [608] = 534,
  [609] = 535,
  [610] = 536,
  [611] = 537,
  [612] = 612,
  [613] = 534,
  [614] = 535,
  [615] = 536,
  [616] = 537,
  [617] = 534,
  [618] = 535,
  [619] = 536,
  [620] = 537,
  [621] = 534,
  [622] = 535,
  [623] = 536,
  [624] = 537,
  [625] = 534,
  [626] = 535,
  [627] = 536,
  [628] = 537,
  [629] = 534,
  [630] = 535,
  [631] = 536,
  [632] = 537,
  [633] = 633,
  [634] = 534,
  [635] = 535,
  [636] = 536,
  [637] = 537,
  [638] = 534,
  [639] = 535,
  [640] = 536,
  [641] = 537,
  [642] = 534,
  [643] = 535,
  [644] = 536,
  [645] = 537,
  [646] = 530,
  [647] = 534,
  [648] = 535,
  [649] = 536,
  [650] = 537,
  [651] = 531,
  [652] = 652,
  [653] = 534,
  [654] = 438,
  [655] = 655,
  [656] = 656,
  [657] = 657,
  [658] = 658,
  [659] = 533,
  [660] = 660,
  [661] = 414,
  [662] = 662,
  [663] = 663,
  [664] = 664,
  [665] = 665,
  [666] = 415,
  [667] = 667,
  [668] = 668,
  [669] = 535,
  [670] = 536,
  [671] = 533,
  [672] = 534,
  [673] = 413,
  [674] = 414,
  [675] = 415,
  [676] = 535,
  [677] = 536,
  [678] = 537,
  [679] = 416,
  [680] = 417,
  [681] = 537,
  [682] = 416,
  [683] = 533,
  [684] = 417,
  [685] = 534,
  [686] = 413,
  [687] = 414,
  [688] = 415,
  [689] = 535,
  [690] = 536,
  [691] = 537,
  [692] = 416,
  [693] = 417,
  [694] = 534,
  [695] = 413,
  [696] = 414,
  [697] = 415,
  [698] = 535,
  [699] = 536,
  [700] = 537,
  [701] = 416,
  [702] = 417,
  [703] = 438,
  [704] = 413,
  [705] = 414,
  [706] = 415,
  [707] = 416,
  [708] = 417,
  [709] = 413,
  [710] = 414,
  [711] = 415,
  [712] = 416,
  [713] = 417,
  [714] = 413,
  [715] = 414,
  [716] = 415,
  [717] = 416,
  [718] = 417,
  [719] = 413,
  [720] = 413,
  [721] = 721,
  [722] = 722,
  [723] = 723,
  [724] = 722,
  [725] = 725,
  [726] = 722,
  [727] = 722,
  [728] = 725,
  [729] = 723,
  [730] = 721,
  [731] = 725,
  [732] = 723,
  [733] = 721,
  [734] = 721,
  [735] = 723,
  [736] = 725,
  [737] = 737,
  [738] = 725,
  [739] = 725,
  [740] = 740,
  [741] = 723,
  [742] = 725,
  [743] = 721,
  [744] = 744,
  [745] = 745,
  [746] = 721,
  [747] = 723,
  [748] = 721,
  [749] = 723,
  [750] = 721,
  [751] = 723,
  [752] = 721,
  [753] = 725,
  [754] = 725,
  [755] = 723,
  [756] = 723,
  [757] = 757,
  [758] = 725,
  [759] = 725,
  [760] = 725,
  [761] = 761,
  [762] = 762,
  [763] = 721,
  [764] = 723,
  [765] = 765,
  [766] = 725,
  [767] = 721,
  [768] = 723,
  [769] = 721,
  [770] = 723,
  [771] = 723,
  [772] = 721,
  [773] = 773,
  [774] = 725,
  [775] = 775,
  [776] = 776,
  [777] = 721,
  [778] = 723,
  [779] = 721,
  [780] = 721,
  [781] = 723,
  [782] = 725,
  [783] = 725,
  [784] = 784,
  [785] = 725,
  [786] = 721,
  [787] = 725,
  [788] = 721,
  [789] = 723,
  [790] = 723,
  [791] = 725,
  [792] = 723,
  [793] = 721,
  [794] = 723,
  [795] = 721,
  [796] = 796,
  [797] = 725,
  [798] = 725,
  [799] = 725,
  [800] = 725,
  [801] = 801,
  [802] = 802,
  [803] = 721,
  [804] = 804,
  [805] = 805,
  [806] = 723,
  [807] = 807,
  [808] = 721,
  [809] = 723,
  [810] = 810,
  [811] = 811,
  [812] = 721,
  [813] = 723,
  [814] = 814,
  [815] = 815,
  [816] = 816,
  [817] = 817,
  [818] = 818,
  [819] = 819,
  [820] = 721,
  [821] = 821,
  [822] = 721,
  [823] = 723,
  [824] = 21,
  [825] = 825,
  [826] = 723,
  [827] = 816,
  [828] = 725,
  [829] = 829,
  [830] = 721,
  [831] = 816,
  [832] = 723,
  [833] = 816,
  [834] = 816,
  [835] = 816,
  [836] = 836,
  [837] = 721,
  [838] = 838,
  [839] = 839,
  [840] = 840,
  [841] = 725,
  [842] = 721,
  [843] = 723,
  [844] = 844,
  [845] = 723,
  [846] = 846,
  [847] = 847,
  [848] = 848,
  [849] = 723,
  [850] = 805,
  [851] = 725,
  [852] = 852,
  [853] = 721,
  [854] = 854,
  [855] = 855,
  [856] = 856,
  [857] = 723,
  [858] = 858,
  [859] = 859,
  [860] = 860,
  [861] = 861,
  [862] = 725,
  [863] = 863,
  [864] = 864,
  [865] = 865,
  [866] = 866,
  [867] = 867,
  [868] = 721,
  [869] = 869,
  [870] = 723,
  [871] = 871,
  [872] = 725,
  [873] = 721,
  [874] = 874,
  [875] = 875,
  [876] = 725,
  [877] = 877,
  [878] = 878,
  [879] = 879,
  [880] = 880,
  [881] = 881,
  [882] = 725,
  [883] = 721,
  [884] = 723,
  [885] = 885,
  [886] = 886,
  [887] = 858,
  [888] = 805,
  [889] = 805,
  [890] = 859,
  [891] = 852,
  [892] = 854,
  [893] = 893,
  [894] = 839,
  [895] = 895,
  [896] = 881,
  [897] = 877,
  [898] = 898,
  [899] = 886,
  [900] = 860,
  [901] = 901,
  [902] = 902,
  [903] = 838,
  [904] = 863,
  [905] = 878,
  [906] = 864,
  [907] = 844,
  [908] = 879,
  [909] = 909,
  [910] = 875,
  [911] = 866,
  [912] = 855,
  [913] = 856,
  [914] = 880,
  [915] = 840,
  [916] = 916,
  [917] = 867,
  [918] = 918,
  [919] = 919,
  [920] = 920,
  [921] = 871,
  [922] = 885,
  [923] = 869,
  [924] = 847,
  [925] = 874,
  [926] = 848,
  [927] = 927,
  [928] = 865,
  [929] = 856,
  [930] = 852,
  [931] = 854,
  [932] = 855,
  [933] = 874,
  [934] = 839,
  [935] = 856,
  [936] = 858,
  [937] = 875,
  [938] = 859,
  [939] = 838,
  [940] = 840,
  [941] = 941,
  [942] = 886,
  [943] = 860,
  [944] = 944,
  [945] = 945,
  [946] = 863,
  [947] = 877,
  [948] = 878,
  [949] = 879,
  [950] = 864,
  [951] = 865,
  [952] = 866,
  [953] = 867,
  [954] = 880,
  [955] = 881,
  [956] = 805,
  [957] = 886,
  [958] = 958,
  [959] = 869,
  [960] = 960,
  [961] = 961,
  [962] = 847,
  [963] = 848,
  [964] = 839,
  [965] = 846,
  [966] = 838,
  [967] = 840,
  [968] = 847,
  [969] = 848,
  [970] = 852,
  [971] = 854,
  [972] = 855,
  [973] = 874,
  [974] = 858,
  [975] = 875,
  [976] = 859,
  [977] = 860,
  [978] = 863,
  [979] = 877,
  [980] = 878,
  [981] = 879,
  [982] = 864,
  [983] = 865,
  [984] = 866,
  [985] = 867,
  [986] = 880,
  [987] = 881,
  [988] = 869,
  [989] = 871,
  [990] = 885,
  [991] = 871,
  [992] = 805,
  [993] = 885,
  [994] = 844,
  [995] = 805,
  [996] = 844,
  [997] = 846,
  [998] = 918,
  [999] = 867,
  [1000] = 840,
  [1001] = 847,
  [1002] = 848,
  [1003] = 886,
  [1004] = 852,
  [1005] = 854,
  [1006] = 855,
  [1007] = 856,
  [1008] = 858,
  [1009] = 859,
  [1010] = 920,
  [1011] = 860,
  [1012] = 863,
  [1013] = 864,
  [1014] = 865,
  [1015] = 866,
  [1016] = 867,
  [1017] = 869,
  [1018] = 805,
  [1019] = 1019,
  [1020] = 874,
  [1021] = 875,
  [1022] = 895,
  [1023] = 877,
  [1024] = 886,
  [1025] = 1025,
  [1026] = 1026,
  [1027] = 875,
  [1028] = 877,
  [1029] = 878,
  [1030] = 879,
  [1031] = 880,
  [1032] = 881,
  [1033] = 871,
  [1034] = 885,
  [1035] = 844,
  [1036] = 878,
  [1037] = 879,
  [1038] = 880,
  [1039] = 898,
  [1040] = 919,
  [1041] = 1041,
  [1042] = 886,
  [1043] = 1043,
  [1044] = 902,
  [1045] = 839,
  [1046] = 839,
  [1047] = 881,
  [1048] = 893,
  [1049] = 805,
  [1050] = 1050,
  [1051] = 916,
  [1052] = 920,
  [1053] = 918,
  [1054] = 909,
  [1055] = 901,
  [1056] = 846,
  [1057] = 927,
  [1058] = 895,
  [1059] = 909,
  [1060] = 846,
  [1061] = 869,
  [1062] = 874,
  [1063] = 871,
  [1064] = 901,
  [1065] = 898,
  [1066] = 919,
  [1067] = 927,
  [1068] = 885,
  [1069] = 844,
  [1070] = 893,
  [1071] = 846,
  [1072] = 1072,
  [1073] = 839,
  [1074] = 885,
  [1075] = 1075,
  [1076] = 916,
  [1077] = 838,
  [1078] = 840,
  [1079] = 847,
  [1080] = 848,
  [1081] = 838,
  [1082] = 840,
  [1083] = 852,
  [1084] = 854,
  [1085] = 855,
  [1086] = 874,
  [1087] = 856,
  [1088] = 858,
  [1089] = 875,
  [1090] = 859,
  [1091] = 847,
  [1092] = 860,
  [1093] = 863,
  [1094] = 877,
  [1095] = 878,
  [1096] = 879,
  [1097] = 864,
  [1098] = 865,
  [1099] = 866,
  [1100] = 867,
  [1101] = 880,
  [1102] = 1102,
  [1103] = 881,
  [1104] = 848,
  [1105] = 869,
  [1106] = 852,
  [1107] = 854,
  [1108] = 871,
  [1109] = 855,
  [1110] = 902,
  [1111] = 856,
  [1112] = 858,
  [1113] = 859,
  [1114] = 860,
  [1115] = 863,
  [1116] = 838,
  [1117] = 864,
  [1118] = 865,
  [1119] = 866,
  [1120] = 844,
  [1121] = 805,
  [1122] = 840,
  [1123] = 898,
  [1124] = 919,
  [1125] = 856,
  [1126] = 860,
  [1127] = 893,
  [1128] = 920,
  [1129] = 863,
  [1130] = 864,
  [1131] = 865,
  [1132] = 866,
  [1133] = 867,
  [1134] = 919,
  [1135] = 916,
  [1136] = 920,
  [1137] = 927,
  [1138] = 846,
  [1139] = 927,
  [1140] = 846,
  [1141] = 869,
  [1142] = 875,
  [1143] = 885,
  [1144] = 877,
  [1145] = 878,
  [1146] = 918,
  [1147] = 1147,
  [1148] = 879,
  [1149] = 880,
  [1150] = 909,
  [1151] = 901,
  [1152] = 881,
  [1153] = 1153,
  [1154] = 902,
  [1155] = 844,
  [1156] = 858,
  [1157] = 859,
  [1158] = 893,
  [1159] = 847,
  [1160] = 916,
  [1161] = 805,
  [1162] = 895,
  [1163] = 874,
  [1164] = 918,
  [1165] = 916,
  [1166] = 871,
  [1167] = 875,
  [1168] = 920,
  [1169] = 885,
  [1170] = 844,
  [1171] = 877,
  [1172] = 805,
  [1173] = 878,
  [1174] = 879,
  [1175] = 860,
  [1176] = 863,
  [1177] = 886,
  [1178] = 880,
  [1179] = 881,
  [1180] = 901,
  [1181] = 859,
  [1182] = 1182,
  [1183] = 1183,
  [1184] = 838,
  [1185] = 865,
  [1186] = 848,
  [1187] = 866,
  [1188] = 886,
  [1189] = 1189,
  [1190] = 839,
  [1191] = 867,
  [1192] = 838,
  [1193] = 840,
  [1194] = 869,
  [1195] = 805,
  [1196] = 839,
  [1197] = 874,
  [1198] = 895,
  [1199] = 847,
  [1200] = 805,
  [1201] = 1201,
  [1202] = 848,
  [1203] = 871,
  [1204] = 893,
  [1205] = 918,
  [1206] = 852,
  [1207] = 852,
  [1208] = 1208,
  [1209] = 854,
  [1210] = 854,
  [1211] = 898,
  [1212] = 1182,
  [1213] = 1208,
  [1214] = 898,
  [1215] = 919,
  [1216] = 927,
  [1217] = 1183,
  [1218] = 895,
  [1219] = 1182,
  [1220] = 1208,
  [1221] = 909,
  [1222] = 855,
  [1223] = 909,
  [1224] = 1183,
  [1225] = 1182,
  [1226] = 1208,
  [1227] = 855,
  [1228] = 901,
  [1229] = 902,
  [1230] = 1183,
  [1231] = 1182,
  [1232] = 1208,
  [1233] = 856,
  [1234] = 1183,
  [1235] = 858,
  [1236] = 902,
  [1237] = 864,
  [1238] = 886,
  [1239] = 848,
  [1240] = 838,
  [1241] = 840,
  [1242] = 864,
  [1243] = 847,
  [1244] = 848,
  [1245] = 852,
  [1246] = 854,
  [1247] = 855,
  [1248] = 856,
  [1249] = 858,
  [1250] = 859,
  [1251] = 860,
  [1252] = 863,
  [1253] = 864,
  [1254] = 865,
  [1255] = 866,
  [1256] = 867,
  [1257] = 902,
  [1258] = 869,
  [1259] = 885,
  [1260] = 893,
  [1261] = 895,
  [1262] = 874,
  [1263] = 916,
  [1264] = 920,
  [1265] = 839,
  [1266] = 918,
  [1267] = 898,
  [1268] = 919,
  [1269] = 927,
  [1270] = 909,
  [1271] = 901,
  [1272] = 844,
  [1273] = 865,
  [1274] = 875,
  [1275] = 877,
  [1276] = 878,
  [1277] = 879,
  [1278] = 880,
  [1279] = 881,
  [1280] = 838,
  [1281] = 840,
  [1282] = 847,
  [1283] = 848,
  [1284] = 852,
  [1285] = 854,
  [1286] = 855,
  [1287] = 856,
  [1288] = 858,
  [1289] = 846,
  [1290] = 859,
  [1291] = 860,
  [1292] = 863,
  [1293] = 864,
  [1294] = 865,
  [1295] = 866,
  [1296] = 867,
  [1297] = 866,
  [1298] = 869,
  [1299] = 874,
  [1300] = 871,
  [1301] = 885,
  [1302] = 844,
  [1303] = 875,
  [1304] = 867,
  [1305] = 877,
  [1306] = 878,
  [1307] = 879,
  [1308] = 859,
  [1309] = 880,
  [1310] = 881,
  [1311] = 895,
  [1312] = 880,
  [1313] = 871,
  [1314] = 885,
  [1315] = 844,
  [1316] = 898,
  [1317] = 919,
  [1318] = 927,
  [1319] = 886,
  [1320] = 839,
  [1321] = 839,
  [1322] = 886,
  [1323] = 852,
  [1324] = 839,
  [1325] = 839,
  [1326] = 902,
  [1327] = 1327,
  [1328] = 881,
  [1329] = 854,
  [1330] = 893,
  [1331] = 846,
  [1332] = 916,
  [1333] = 920,
  [1334] = 847,
  [1335] = 1041,
  [1336] = 838,
  [1337] = 840,
  [1338] = 918,
  [1339] = 847,
  [1340] = 848,
  [1341] = 852,
  [1342] = 854,
  [1343] = 855,
  [1344] = 856,
  [1345] = 858,
  [1346] = 859,
  [1347] = 860,
  [1348] = 863,
  [1349] = 864,
  [1350] = 865,
  [1351] = 866,
  [1352] = 867,
  [1353] = 838,
  [1354] = 840,
  [1355] = 869,
  [1356] = 847,
  [1357] = 848,
  [1358] = 852,
  [1359] = 854,
  [1360] = 855,
  [1361] = 874,
  [1362] = 856,
  [1363] = 858,
  [1364] = 859,
  [1365] = 909,
  [1366] = 901,
  [1367] = 860,
  [1368] = 863,
  [1369] = 864,
  [1370] = 865,
  [1371] = 866,
  [1372] = 867,
  [1373] = 869,
  [1374] = 805,
  [1375] = 874,
  [1376] = 875,
  [1377] = 874,
  [1378] = 877,
  [1379] = 878,
  [1380] = 879,
  [1381] = 880,
  [1382] = 881,
  [1383] = 805,
  [1384] = 869,
  [1385] = 875,
  [1386] = 1386,
  [1387] = 877,
  [1388] = 878,
  [1389] = 879,
  [1390] = 880,
  [1391] = 881,
  [1392] = 1043,
  [1393] = 871,
  [1394] = 885,
  [1395] = 844,
  [1396] = 871,
  [1397] = 885,
  [1398] = 844,
  [1399] = 805,
  [1400] = 805,
  [1401] = 1401,
  [1402] = 111,
  [1403] = 856,
  [1404] = 858,
  [1405] = 875,
  [1406] = 860,
  [1407] = 846,
  [1408] = 846,
  [1409] = 871,
  [1410] = 863,
  [1411] = 838,
  [1412] = 840,
  [1413] = 846,
  [1414] = 886,
  [1415] = 877,
  [1416] = 878,
  [1417] = 886,
  [1418] = 805,
  [1419] = 879,
  [1420] = 855,
  [1421] = 881,
  [1422] = 886,
  [1423] = 839,
  [1424] = 875,
  [1425] = 839,
  [1426] = 839,
  [1427] = 885,
  [1428] = 839,
  [1429] = 898,
  [1430] = 839,
  [1431] = 844,
  [1432] = 919,
  [1433] = 877,
  [1434] = 805,
  [1435] = 927,
  [1436] = 895,
  [1437] = 878,
  [1438] = 886,
  [1439] = 879,
  [1440] = 898,
  [1441] = 919,
  [1442] = 927,
  [1443] = 1443,
  [1444] = 1444,
  [1445] = 902,
  [1446] = 1446,
  [1447] = 893,
  [1448] = 895,
  [1449] = 805,
  [1450] = 1450,
  [1451] = 916,
  [1452] = 920,
  [1453] = 918,
  [1454] = 1454,
  [1455] = 846,
  [1456] = 909,
  [1457] = 901,
  [1458] = 1458,
  [1459] = 898,
  [1460] = 880,
  [1461] = 919,
  [1462] = 856,
  [1463] = 858,
  [1464] = 918,
  [1465] = 859,
  [1466] = 927,
  [1467] = 902,
  [1468] = 1468,
  [1469] = 916,
  [1470] = 1470,
  [1471] = 860,
  [1472] = 863,
  [1473] = 1473,
  [1474] = 864,
  [1475] = 838,
  [1476] = 840,
  [1477] = 893,
  [1478] = 847,
  [1479] = 848,
  [1480] = 852,
  [1481] = 854,
  [1482] = 855,
  [1483] = 1483,
  [1484] = 856,
  [1485] = 858,
  [1486] = 865,
  [1487] = 859,
  [1488] = 866,
  [1489] = 860,
  [1490] = 863,
  [1491] = 867,
  [1492] = 920,
  [1493] = 864,
  [1494] = 865,
  [1495] = 866,
  [1496] = 867,
  [1497] = 846,
  [1498] = 916,
  [1499] = 838,
  [1500] = 1500,
  [1501] = 840,
  [1502] = 869,
  [1503] = 847,
  [1504] = 848,
  [1505] = 920,
  [1506] = 852,
  [1507] = 854,
  [1508] = 855,
  [1509] = 874,
  [1510] = 856,
  [1511] = 858,
  [1512] = 918,
  [1513] = 859,
  [1514] = 860,
  [1515] = 863,
  [1516] = 839,
  [1517] = 1517,
  [1518] = 909,
  [1519] = 1519,
  [1520] = 864,
  [1521] = 865,
  [1522] = 866,
  [1523] = 867,
  [1524] = 895,
  [1525] = 909,
  [1526] = 846,
  [1527] = 869,
  [1528] = 895,
  [1529] = 1500,
  [1530] = 874,
  [1531] = 846,
  [1532] = 875,
  [1533] = 898,
  [1534] = 919,
  [1535] = 901,
  [1536] = 927,
  [1537] = 877,
  [1538] = 878,
  [1539] = 879,
  [1540] = 880,
  [1541] = 881,
  [1542] = 875,
  [1543] = 877,
  [1544] = 878,
  [1545] = 879,
  [1546] = 901,
  [1547] = 1500,
  [1548] = 880,
  [1549] = 1549,
  [1550] = 838,
  [1551] = 840,
  [1552] = 881,
  [1553] = 838,
  [1554] = 847,
  [1555] = 848,
  [1556] = 1050,
  [1557] = 854,
  [1558] = 847,
  [1559] = 848,
  [1560] = 886,
  [1561] = 852,
  [1562] = 854,
  [1563] = 855,
  [1564] = 852,
  [1565] = 854,
  [1566] = 855,
  [1567] = 874,
  [1568] = 856,
  [1569] = 858,
  [1570] = 875,
  [1571] = 859,
  [1572] = 856,
  [1573] = 858,
  [1574] = 805,
  [1575] = 871,
  [1576] = 860,
  [1577] = 863,
  [1578] = 877,
  [1579] = 859,
  [1580] = 869,
  [1581] = 878,
  [1582] = 879,
  [1583] = 864,
  [1584] = 871,
  [1585] = 860,
  [1586] = 865,
  [1587] = 866,
  [1588] = 867,
  [1589] = 885,
  [1590] = 880,
  [1591] = 844,
  [1592] = 881,
  [1593] = 863,
  [1594] = 864,
  [1595] = 865,
  [1596] = 866,
  [1597] = 869,
  [1598] = 867,
  [1599] = 885,
  [1600] = 844,
  [1601] = 869,
  [1602] = 874,
  [1603] = 1401,
  [1604] = 871,
  [1605] = 886,
  [1606] = 885,
  [1607] = 844,
  [1608] = 111,
  [1609] = 846,
  [1610] = 846,
  [1611] = 874,
  [1612] = 1612,
  [1613] = 1050,
  [1614] = 875,
  [1615] = 902,
  [1616] = 893,
  [1617] = 877,
  [1618] = 878,
  [1619] = 879,
  [1620] = 1620,
  [1621] = 1621,
  [1622] = 1622,
  [1623] = 880,
  [1624] = 916,
  [1625] = 881,
  [1626] = 920,
  [1627] = 902,
  [1628] = 805,
  [1629] = 893,
  [1630] = 895,
  [1631] = 916,
  [1632] = 920,
  [1633] = 918,
  [1634] = 898,
  [1635] = 919,
  [1636] = 927,
  [1637] = 909,
  [1638] = 901,
  [1639] = 918,
  [1640] = 909,
  [1641] = 901,
  [1642] = 902,
  [1643] = 805,
  [1644] = 1644,
  [1645] = 871,
  [1646] = 885,
  [1647] = 1473,
  [1648] = 844,
  [1649] = 893,
  [1650] = 1443,
  [1651] = 1454,
  [1652] = 855,
  [1653] = 1549,
  [1654] = 1450,
  [1655] = 1473,
  [1656] = 1443,
  [1657] = 1454,
  [1658] = 1549,
  [1659] = 1450,
  [1660] = 1473,
  [1661] = 838,
  [1662] = 1443,
  [1663] = 840,
  [1664] = 1450,
  [1665] = 847,
  [1666] = 1473,
  [1667] = 848,
  [1668] = 1443,
  [1669] = 886,
  [1670] = 1450,
  [1671] = 871,
  [1672] = 852,
  [1673] = 840,
  [1674] = 871,
  [1675] = 902,
  [1676] = 886,
  [1677] = 901,
  [1678] = 846,
  [1679] = 847,
  [1680] = 902,
  [1681] = 918,
  [1682] = 1682,
  [1683] = 902,
  [1684] = 1684,
  [1685] = 869,
  [1686] = 111,
  [1687] = 840,
  [1688] = 909,
  [1689] = 848,
  [1690] = 847,
  [1691] = 848,
  [1692] = 901,
  [1693] = 920,
  [1694] = 871,
  [1695] = 838,
  [1696] = 840,
  [1697] = 874,
  [1698] = 895,
  [1699] = 880,
  [1700] = 852,
  [1701] = 916,
  [1702] = 920,
  [1703] = 854,
  [1704] = 885,
  [1705] = 855,
  [1706] = 844,
  [1707] = 1707,
  [1708] = 856,
  [1709] = 902,
  [1710] = 893,
  [1711] = 895,
  [1712] = 918,
  [1713] = 863,
  [1714] = 898,
  [1715] = 885,
  [1716] = 1716,
  [1717] = 839,
  [1718] = 898,
  [1719] = 919,
  [1720] = 916,
  [1721] = 920,
  [1722] = 918,
  [1723] = 898,
  [1724] = 877,
  [1725] = 895,
  [1726] = 878,
  [1727] = 1401,
  [1728] = 875,
  [1729] = 858,
  [1730] = 902,
  [1731] = 805,
  [1732] = 852,
  [1733] = 927,
  [1734] = 859,
  [1735] = 879,
  [1736] = 839,
  [1737] = 864,
  [1738] = 865,
  [1739] = 871,
  [1740] = 844,
  [1741] = 852,
  [1742] = 854,
  [1743] = 855,
  [1744] = 856,
  [1745] = 858,
  [1746] = 859,
  [1747] = 805,
  [1748] = 885,
  [1749] = 844,
  [1750] = 898,
  [1751] = 916,
  [1752] = 839,
  [1753] = 919,
  [1754] = 877,
  [1755] = 927,
  [1756] = 1756,
  [1757] = 866,
  [1758] = 867,
  [1759] = 860,
  [1760] = 863,
  [1761] = 864,
  [1762] = 878,
  [1763] = 893,
  [1764] = 865,
  [1765] = 866,
  [1766] = 879,
  [1767] = 867,
  [1768] = 880,
  [1769] = 846,
  [1770] = 881,
  [1771] = 919,
  [1772] = 846,
  [1773] = 869,
  [1774] = 927,
  [1775] = 909,
  [1776] = 1776,
  [1777] = 895,
  [1778] = 901,
  [1779] = 880,
  [1780] = 881,
  [1781] = 886,
  [1782] = 1050,
  [1783] = 893,
  [1784] = 881,
  [1785] = 854,
  [1786] = 1786,
  [1787] = 871,
  [1788] = 874,
  [1789] = 893,
  [1790] = 871,
  [1791] = 875,
  [1792] = 855,
  [1793] = 874,
  [1794] = 875,
  [1795] = 856,
  [1796] = 838,
  [1797] = 916,
  [1798] = 840,
  [1799] = 886,
  [1800] = 898,
  [1801] = 847,
  [1802] = 844,
  [1803] = 909,
  [1804] = 920,
  [1805] = 919,
  [1806] = 805,
  [1807] = 852,
  [1808] = 927,
  [1809] = 918,
  [1810] = 854,
  [1811] = 855,
  [1812] = 927,
  [1813] = 898,
  [1814] = 919,
  [1815] = 927,
  [1816] = 909,
  [1817] = 1817,
  [1818] = 901,
  [1819] = 918,
  [1820] = 847,
  [1821] = 856,
  [1822] = 858,
  [1823] = 893,
  [1824] = 859,
  [1825] = 838,
  [1826] = 860,
  [1827] = 863,
  [1828] = 877,
  [1829] = 878,
  [1830] = 879,
  [1831] = 864,
  [1832] = 865,
  [1833] = 866,
  [1834] = 867,
  [1835] = 848,
  [1836] = 805,
  [1837] = 1837,
  [1838] = 838,
  [1839] = 909,
  [1840] = 840,
  [1841] = 869,
  [1842] = 847,
  [1843] = 848,
  [1844] = 1050,
  [1845] = 909,
  [1846] = 885,
  [1847] = 852,
  [1848] = 854,
  [1849] = 855,
  [1850] = 874,
  [1851] = 856,
  [1852] = 858,
  [1853] = 880,
  [1854] = 859,
  [1855] = 881,
  [1856] = 1856,
  [1857] = 860,
  [1858] = 863,
  [1859] = 869,
  [1860] = 846,
  [1861] = 916,
  [1862] = 846,
  [1863] = 864,
  [1864] = 865,
  [1865] = 866,
  [1866] = 920,
  [1867] = 867,
  [1868] = 916,
  [1869] = 860,
  [1870] = 920,
  [1871] = 1871,
  [1872] = 863,
  [1873] = 893,
  [1874] = 895,
  [1875] = 901,
  [1876] = 838,
  [1877] = 869,
  [1878] = 858,
  [1879] = 874,
  [1880] = 840,
  [1881] = 875,
  [1882] = 805,
  [1883] = 919,
  [1884] = 877,
  [1885] = 878,
  [1886] = 901,
  [1887] = 902,
  [1888] = 879,
  [1889] = 895,
  [1890] = 1890,
  [1891] = 875,
  [1892] = 880,
  [1893] = 885,
  [1894] = 859,
  [1895] = 881,
  [1896] = 805,
  [1897] = 860,
  [1898] = 839,
  [1899] = 877,
  [1900] = 886,
  [1901] = 918,
  [1902] = 878,
  [1903] = 879,
  [1904] = 886,
  [1905] = 864,
  [1906] = 865,
  [1907] = 844,
  [1908] = 866,
  [1909] = 839,
  [1910] = 1189,
  [1911] = 867,
  [1912] = 848,
  [1913] = 847,
  [1914] = 1684,
  [1915] = 1915,
  [1916] = 1916,
  [1917] = 1917,
  [1918] = 1918,
  [1919] = 1919,
  [1920] = 1920,
  [1921] = 1921,
  [1922] = 1922,
  [1923] = 1923,
  [1924] = 1924,
  [1925] = 1925,
  [1926] = 1926,
  [1927] = 902,
  [1928] = 1928,
  [1929] = 1929,
  [1930] = 893,
  [1931] = 895,
  [1932] = 1932,
  [1933] = 916,
  [1934] = 920,
  [1935] = 918,
  [1936] = 898,
  [1937] = 919,
  [1938] = 927,
  [1939] = 909,
  [1940] = 901,
  [1941] = 1941,
  [1942] = 1942,
  [1943] = 1612,
  [1944] = 1644,
  [1945] = 886,
  [1946] = 1620,
  [1947] = 1621,
  [1948] = 1622,
  [1949] = 1468,
  [1950] = 1470,
  [1951] = 1951,
  [1952] = 1952,
  [1953] = 1953,
  [1954] = 1446,
  [1955] = 1458,
  [1956] = 1483,
  [1957] = 1957,
  [1958] = 1958,
  [1959] = 1959,
  [1960] = 1960,
  [1961] = 1961,
  [1962] = 1962,
  [1963] = 1963,
  [1964] = 1444,
  [1965] = 1965,
  [1966] = 1966,
  [1967] = 1967,
  [1968] = 1968,
  [1969] = 1969,
  [1970] = 1970,
  [1971] = 1971,
  [1972] = 1972,
  [1973] = 1973,
  [1974] = 1974,
  [1975] = 838,
  [1976] = 840,
  [1977] = 1977,
  [1978] = 846,
  [1979] = 1979,
  [1980] = 847,
  [1981] = 848,
  [1982] = 1982,
  [1983] = 1983,
  [1984] = 1984,
  [1985] = 1985,
  [1986] = 852,
  [1987] = 854,
  [1988] = 855,
  [1989] = 1837,
  [1990] = 874,
  [1991] = 856,
  [1992] = 858,
  [1993] = 875,
  [1994] = 859,
  [1995] = 839,
  [1996] = 860,
  [1997] = 863,
  [1998] = 877,
  [1999] = 878,
  [2000] = 879,
  [2001] = 864,
  [2002] = 865,
  [2003] = 866,
  [2004] = 867,
  [2005] = 880,
  [2006] = 881,
  [2007] = 886,
  [2008] = 846,
  [2009] = 869,
  [2010] = 2010,
  [2011] = 2011,
  [2012] = 2012,
  [2013] = 839,
  [2014] = 846,
  [2015] = 1837,
  [2016] = 871,
  [2017] = 885,
  [2018] = 844,
  [2019] = 902,
  [2020] = 846,
  [2021] = 1716,
  [2022] = 2022,
  [2023] = 893,
  [2024] = 895,
  [2025] = 2025,
  [2026] = 2026,
  [2027] = 916,
  [2028] = 920,
  [2029] = 2029,
  [2030] = 2030,
  [2031] = 2031,
  [2032] = 839,
  [2033] = 2033,
  [2034] = 918,
  [2035] = 898,
  [2036] = 919,
  [2037] = 927,
  [2038] = 909,
  [2039] = 901,
  [2040] = 2040,
  [2041] = 2041,
  [2042] = 839,
  [2043] = 2043,
  [2044] = 2044,
  [2045] = 2045,
  [2046] = 2046,
  [2047] = 2047,
  [2048] = 1050,
  [2049] = 2049,
  [2050] = 2050,
  [2051] = 2051,
  [2052] = 2052,
  [2053] = 2053,
  [2054] = 846,
  [2055] = 895,
  [2056] = 839,
  [2057] = 898,
  [2058] = 919,
  [2059] = 927,
  [2060] = 2060,
  [2061] = 902,
  [2062] = 893,
  [2063] = 895,
  [2064] = 916,
  [2065] = 920,
  [2066] = 918,
  [2067] = 898,
  [2068] = 919,
  [2069] = 927,
  [2070] = 909,
  [2071] = 901,
  [2072] = 886,
  [2073] = 2073,
  [2074] = 2074,
  [2075] = 2075,
  [2076] = 895,
  [2077] = 898,
  [2078] = 919,
  [2079] = 927,
  [2080] = 839,
  [2081] = 2081,
  [2082] = 838,
  [2083] = 840,
  [2084] = 847,
  [2085] = 848,
  [2086] = 2086,
  [2087] = 838,
  [2088] = 840,
  [2089] = 1612,
  [2090] = 852,
  [2091] = 854,
  [2092] = 855,
  [2093] = 874,
  [2094] = 856,
  [2095] = 858,
  [2096] = 859,
  [2097] = 1620,
  [2098] = 1621,
  [2099] = 1622,
  [2100] = 860,
  [2101] = 863,
  [2102] = 864,
  [2103] = 865,
  [2104] = 866,
  [2105] = 867,
  [2106] = 838,
  [2107] = 840,
  [2108] = 847,
  [2109] = 848,
  [2110] = 2110,
  [2111] = 869,
  [2112] = 847,
  [2113] = 848,
  [2114] = 1684,
  [2115] = 1468,
  [2116] = 2116,
  [2117] = 2117,
  [2118] = 852,
  [2119] = 854,
  [2120] = 855,
  [2121] = 874,
  [2122] = 856,
  [2123] = 858,
  [2124] = 875,
  [2125] = 859,
  [2126] = 860,
  [2127] = 863,
  [2128] = 877,
  [2129] = 878,
  [2130] = 879,
  [2131] = 864,
  [2132] = 865,
  [2133] = 866,
  [2134] = 867,
  [2135] = 880,
  [2136] = 881,
  [2137] = 838,
  [2138] = 1446,
  [2139] = 840,
  [2140] = 1458,
  [2141] = 1483,
  [2142] = 852,
  [2143] = 869,
  [2144] = 847,
  [2145] = 848,
  [2146] = 854,
  [2147] = 855,
  [2148] = 874,
  [2149] = 856,
  [2150] = 858,
  [2151] = 875,
  [2152] = 859,
  [2153] = 852,
  [2154] = 854,
  [2155] = 855,
  [2156] = 874,
  [2157] = 856,
  [2158] = 858,
  [2159] = 875,
  [2160] = 859,
  [2161] = 1444,
  [2162] = 839,
  [2163] = 2163,
  [2164] = 2164,
  [2165] = 2165,
  [2166] = 2166,
  [2167] = 871,
  [2168] = 860,
  [2169] = 863,
  [2170] = 877,
  [2171] = 878,
  [2172] = 879,
  [2173] = 864,
  [2174] = 865,
  [2175] = 866,
  [2176] = 867,
  [2177] = 885,
  [2178] = 880,
  [2179] = 844,
  [2180] = 881,
  [2181] = 2181,
  [2182] = 869,
  [2183] = 2183,
  [2184] = 1050,
  [2185] = 860,
  [2186] = 875,
  [2187] = 863,
  [2188] = 877,
  [2189] = 871,
  [2190] = 877,
  [2191] = 878,
  [2192] = 879,
  [2193] = 885,
  [2194] = 880,
  [2195] = 844,
  [2196] = 881,
  [2197] = 878,
  [2198] = 879,
  [2199] = 864,
  [2200] = 865,
  [2201] = 866,
  [2202] = 867,
  [2203] = 880,
  [2204] = 871,
  [2205] = 885,
  [2206] = 844,
  [2207] = 902,
  [2208] = 881,
  [2209] = 893,
  [2210] = 916,
  [2211] = 920,
  [2212] = 2212,
  [2213] = 2213,
  [2214] = 918,
  [2215] = 2215,
  [2216] = 1644,
  [2217] = 909,
  [2218] = 901,
  [2219] = 869,
  [2220] = 2220,
  [2221] = 2221,
  [2222] = 838,
  [2223] = 840,
  [2224] = 902,
  [2225] = 2225,
  [2226] = 848,
  [2227] = 893,
  [2228] = 2228,
  [2229] = 2229,
  [2230] = 2230,
  [2231] = 916,
  [2232] = 920,
  [2233] = 852,
  [2234] = 854,
  [2235] = 855,
  [2236] = 856,
  [2237] = 858,
  [2238] = 859,
  [2239] = 918,
  [2240] = 909,
  [2241] = 901,
  [2242] = 2242,
  [2243] = 860,
  [2244] = 863,
  [2245] = 864,
  [2246] = 865,
  [2247] = 866,
  [2248] = 867,
  [2249] = 2249,
  [2250] = 2250,
  [2251] = 2251,
  [2252] = 2252,
  [2253] = 869,
  [2254] = 871,
  [2255] = 874,
  [2256] = 875,
  [2257] = 2257,
  [2258] = 2258,
  [2259] = 2259,
  [2260] = 2260,
  [2261] = 2261,
  [2262] = 2262,
  [2263] = 2263,
  [2264] = 2264,
  [2265] = 877,
  [2266] = 878,
  [2267] = 879,
  [2268] = 880,
  [2269] = 881,
  [2270] = 2270,
  [2271] = 886,
  [2272] = 2272,
  [2273] = 839,
  [2274] = 2274,
  [2275] = 2275,
  [2276] = 871,
  [2277] = 885,
  [2278] = 844,
  [2279] = 886,
  [2280] = 885,
  [2281] = 844,
  [2282] = 886,
  [2283] = 846,
  [2284] = 2284,
  [2285] = 2285,
  [2286] = 2286,
  [2287] = 2287,
  [2288] = 1470,
  [2289] = 1446,
  [2290] = 1620,
  [2291] = 1951,
  [2292] = 898,
  [2293] = 919,
  [2294] = 927,
  [2295] = 2295,
  [2296] = 902,
  [2297] = 2297,
  [2298] = 893,
  [2299] = 916,
  [2300] = 920,
  [2301] = 1621,
  [2302] = 909,
  [2303] = 901,
  [2304] = 902,
  [2305] = 2305,
  [2306] = 1468,
  [2307] = 893,
  [2308] = 895,
  [2309] = 2309,
  [2310] = 916,
  [2311] = 920,
  [2312] = 2312,
  [2313] = 1644,
  [2314] = 2314,
  [2315] = 1837,
  [2316] = 918,
  [2317] = 1444,
  [2318] = 839,
  [2319] = 898,
  [2320] = 918,
  [2321] = 898,
  [2322] = 2322,
  [2323] = 919,
  [2324] = 927,
  [2325] = 909,
  [2326] = 2326,
  [2327] = 901,
  [2328] = 919,
  [2329] = 927,
  [2330] = 2326,
  [2331] = 909,
  [2332] = 1050,
  [2333] = 2326,
  [2334] = 901,
  [2335] = 2335,
  [2336] = 2336,
  [2337] = 902,
  [2338] = 2338,
  [2339] = 2339,
  [2340] = 1470,
  [2341] = 893,
  [2342] = 895,
  [2343] = 916,
  [2344] = 920,
  [2345] = 1837,
  [2346] = 918,
  [2347] = 898,
  [2348] = 919,
  [2349] = 927,
  [2350] = 909,
  [2351] = 901,
  [2352] = 1837,
  [2353] = 2353,
  [2354] = 902,
  [2355] = 893,
  [2356] = 895,
  [2357] = 902,
  [2358] = 2358,
  [2359] = 916,
  [2360] = 920,
  [2361] = 918,
  [2362] = 898,
  [2363] = 919,
  [2364] = 927,
  [2365] = 909,
  [2366] = 1612,
  [2367] = 1620,
  [2368] = 1621,
  [2369] = 1622,
  [2370] = 901,
  [2371] = 1468,
  [2372] = 1470,
  [2373] = 1446,
  [2374] = 1458,
  [2375] = 1483,
  [2376] = 1444,
  [2377] = 2377,
  [2378] = 920,
  [2379] = 2379,
  [2380] = 1612,
  [2381] = 2326,
  [2382] = 2382,
  [2383] = 1620,
  [2384] = 1621,
  [2385] = 1622,
  [2386] = 2386,
  [2387] = 1644,
  [2388] = 2388,
  [2389] = 1468,
  [2390] = 1470,
  [2391] = 2391,
  [2392] = 2392,
  [2393] = 893,
  [2394] = 895,
  [2395] = 1446,
  [2396] = 1458,
  [2397] = 1483,
  [2398] = 2398,
  [2399] = 1622,
  [2400] = 2400,
  [2401] = 2401,
  [2402] = 916,
  [2403] = 920,
  [2404] = 1444,
  [2405] = 839,
  [2406] = 2406,
  [2407] = 1050,
  [2408] = 1458,
  [2409] = 1483,
  [2410] = 2410,
  [2411] = 2411,
  [2412] = 916,
  [2413] = 2413,
  [2414] = 918,
  [2415] = 898,
  [2416] = 2416,
  [2417] = 1644,
  [2418] = 1050,
  [2419] = 895,
  [2420] = 919,
  [2421] = 839,
  [2422] = 902,
  [2423] = 927,
  [2424] = 909,
  [2425] = 901,
  [2426] = 2426,
  [2427] = 893,
  [2428] = 895,
  [2429] = 1612,
  [2430] = 2295,
  [2431] = 2295,
  [2432] = 2295,
  [2433] = 918,
  [2434] = 2434,
  [2435] = 2434,
  [2436] = 1612,
  [2437] = 2434,
  [2438] = 1620,
  [2439] = 1621,
  [2440] = 1622,
  [2441] = 2434,
  [2442] = 1468,
  [2443] = 1837,
  [2444] = 1470,
  [2445] = 1446,
  [2446] = 1458,
  [2447] = 1483,
  [2448] = 1444,
  [2449] = 2434,
  [2450] = 2434,
  [2451] = 839,
  [2452] = 1612,
  [2453] = 1644,
  [2454] = 1620,
  [2455] = 1621,
  [2456] = 1622,
  [2457] = 1468,
  [2458] = 1470,
  [2459] = 1446,
  [2460] = 1458,
  [2461] = 1483,
  [2462] = 1444,
  [2463] = 1837,
  [2464] = 2434,
  [2465] = 2434,
  [2466] = 2434,
  [2467] = 1644,
  [2468] = 2434,
  [2469] = 1817,
  [2470] = 839,
  [2471] = 1468,
  [2472] = 1444,
  [2473] = 1458,
  [2474] = 1483,
  [2475] = 839,
  [2476] = 839,
  [2477] = 1612,
  [2478] = 839,
  [2479] = 1446,
  [2480] = 1622,
  [2481] = 1644,
  [2482] = 1620,
  [2483] = 1483,
  [2484] = 1621,
  [2485] = 1622,
  [2486] = 1837,
  [2487] = 839,
  [2488] = 1468,
  [2489] = 1837,
  [2490] = 1470,
  [2491] = 1644,
  [2492] = 1470,
  [2493] = 1446,
  [2494] = 1458,
  [2495] = 1444,
  [2496] = 1612,
  [2497] = 1620,
  [2498] = 1958,
  [2499] = 1468,
  [2500] = 1621,
  [2501] = 1622,
  [2502] = 1468,
  [2503] = 1470,
  [2504] = 2353,
  [2505] = 1446,
  [2506] = 1612,
  [2507] = 1621,
  [2508] = 1620,
  [2509] = 1621,
  [2510] = 1622,
  [2511] = 1468,
  [2512] = 1470,
  [2513] = 1612,
  [2514] = 1446,
  [2515] = 1458,
  [2516] = 1620,
  [2517] = 1621,
  [2518] = 1622,
  [2519] = 1483,
  [2520] = 1837,
  [2521] = 1644,
  [2522] = 1444,
  [2523] = 1470,
  [2524] = 1446,
  [2525] = 1458,
  [2526] = 1483,
  [2527] = 1444,
  [2528] = 1458,
  [2529] = 1483,
  [2530] = 1612,
  [2531] = 1444,
  [2532] = 839,
  [2533] = 1644,
  [2534] = 1958,
  [2535] = 1620,
  [2536] = 1644,
  [2537] = 1468,
  [2538] = 2538,
  [2539] = 1444,
  [2540] = 1444,
  [2541] = 1620,
  [2542] = 2542,
  [2543] = 1621,
  [2544] = 1644,
  [2545] = 839,
  [2546] = 1622,
  [2547] = 1644,
  [2548] = 1468,
  [2549] = 1470,
  [2550] = 1458,
  [2551] = 1446,
  [2552] = 1458,
  [2553] = 1483,
  [2554] = 1612,
  [2555] = 1444,
  [2556] = 1483,
  [2557] = 1620,
  [2558] = 1621,
  [2559] = 1612,
  [2560] = 1644,
  [2561] = 1620,
  [2562] = 1621,
  [2563] = 1622,
  [2564] = 1622,
  [2565] = 1644,
  [2566] = 1468,
  [2567] = 1620,
  [2568] = 1470,
  [2569] = 839,
  [2570] = 1612,
  [2571] = 1644,
  [2572] = 1612,
  [2573] = 1621,
  [2574] = 1622,
  [2575] = 1958,
  [2576] = 1468,
  [2577] = 1470,
  [2578] = 839,
  [2579] = 2579,
  [2580] = 1446,
  [2581] = 1458,
  [2582] = 1483,
  [2583] = 1444,
  [2584] = 1621,
  [2585] = 1468,
  [2586] = 1446,
  [2587] = 1470,
  [2588] = 839,
  [2589] = 1458,
  [2590] = 1470,
  [2591] = 1483,
  [2592] = 1644,
  [2593] = 2593,
  [2594] = 839,
  [2595] = 1446,
  [2596] = 1458,
  [2597] = 1483,
  [2598] = 1612,
  [2599] = 1620,
  [2600] = 1621,
  [2601] = 1622,
  [2602] = 1468,
  [2603] = 1470,
  [2604] = 1446,
  [2605] = 1458,
  [2606] = 1483,
  [2607] = 1444,
  [2608] = 839,
  [2609] = 1958,
  [2610] = 1612,
  [2611] = 1444,
  [2612] = 1622,
  [2613] = 2613,
  [2614] = 1446,
  [2615] = 1620,
  [2616] = 1622,
  [2617] = 2617,
  [2618] = 1483,
  [2619] = 2619,
  [2620] = 2593,
  [2621] = 1644,
  [2622] = 1620,
  [2623] = 1444,
  [2624] = 1444,
  [2625] = 1621,
  [2626] = 2626,
  [2627] = 1622,
  [2628] = 1468,
  [2629] = 2629,
  [2630] = 2579,
  [2631] = 1612,
  [2632] = 1468,
  [2633] = 839,
  [2634] = 1444,
  [2635] = 1468,
  [2636] = 2626,
  [2637] = 2637,
  [2638] = 2638,
  [2639] = 1446,
  [2640] = 1612,
  [2641] = 1483,
  [2642] = 1444,
  [2643] = 1621,
  [2644] = 839,
  [2645] = 2645,
  [2646] = 1470,
  [2647] = 1470,
  [2648] = 1612,
  [2649] = 1620,
  [2650] = 1644,
  [2651] = 1612,
  [2652] = 1622,
  [2653] = 1620,
  [2654] = 1621,
  [2655] = 839,
  [2656] = 1458,
  [2657] = 1468,
  [2658] = 2626,
  [2659] = 1620,
  [2660] = 1470,
  [2661] = 1621,
  [2662] = 1621,
  [2663] = 1622,
  [2664] = 1458,
  [2665] = 1458,
  [2666] = 1468,
  [2667] = 1622,
  [2668] = 1483,
  [2669] = 1470,
  [2670] = 1483,
  [2671] = 839,
  [2672] = 1446,
  [2673] = 1644,
  [2674] = 1446,
  [2675] = 1958,
  [2676] = 2676,
  [2677] = 1620,
  [2678] = 2678,
  [2679] = 839,
  [2680] = 1458,
  [2681] = 1644,
  [2682] = 1458,
  [2683] = 1612,
  [2684] = 1483,
  [2685] = 1958,
  [2686] = 1446,
  [2687] = 1444,
  [2688] = 1644,
  [2689] = 2626,
  [2690] = 1446,
  [2691] = 2691,
  [2692] = 1470,
  [2693] = 1444,
  [2694] = 1620,
  [2695] = 1444,
  [2696] = 1470,
  [2697] = 1621,
  [2698] = 1622,
  [2699] = 1458,
  [2700] = 1620,
  [2701] = 1621,
  [2702] = 1446,
  [2703] = 1458,
  [2704] = 1483,
  [2705] = 1622,
  [2706] = 1444,
  [2707] = 1958,
  [2708] = 1483,
  [2709] = 1612,
  [2710] = 1644,
  [2711] = 1468,
  [2712] = 1470,
  [2713] = 1620,
  [2714] = 2714,
  [2715] = 839,
  [2716] = 1621,
  [2717] = 1622,
  [2718] = 1468,
  [2719] = 1446,
  [2720] = 1470,
  [2721] = 1468,
  [2722] = 2722,
  [2723] = 1446,
  [2724] = 2724,
  [2725] = 1446,
  [2726] = 2726,
  [2727] = 1468,
  [2728] = 1458,
  [2729] = 1458,
  [2730] = 1483,
  [2731] = 1483,
  [2732] = 1444,
  [2733] = 1470,
  [2734] = 1644,
  [2735] = 1612,
  [2736] = 1444,
  [2737] = 1612,
  [2738] = 1644,
  [2739] = 1612,
  [2740] = 1620,
  [2741] = 1621,
  [2742] = 1622,
  [2743] = 1468,
  [2744] = 839,
  [2745] = 1470,
  [2746] = 1620,
  [2747] = 1446,
  [2748] = 1458,
  [2749] = 1483,
  [2750] = 1644,
  [2751] = 1958,
  [2752] = 1621,
  [2753] = 1622,
  [2754] = 2754,
  [2755] = 1446,
  [2756] = 1612,
  [2757] = 839,
  [2758] = 1612,
  [2759] = 1458,
  [2760] = 1483,
  [2761] = 1644,
  [2762] = 1620,
  [2763] = 1621,
  [2764] = 1622,
  [2765] = 1444,
  [2766] = 839,
  [2767] = 839,
  [2768] = 1644,
  [2769] = 1468,
  [2770] = 1470,
  [2771] = 839,
  [2772] = 1958,
  [2773] = 2773,
  [2774] = 2774,
  [2775] = 2775,
  [2776] = 2774,
  [2777] = 2774,
  [2778] = 2778,
  [2779] = 2779,
  [2780] = 2779,
  [2781] = 2773,
  [2782] = 2775,
  [2783] = 2774,
  [2784] = 2774,
  [2785] = 2775,
  [2786] = 2779,
  [2787] = 2773,
  [2788] = 2778,
  [2789] = 2778,
  [2790] = 2790,
  [2791] = 2791,
  [2792] = 2792,
  [2793] = 2793,
  [2794] = 721,
  [2795] = 2795,
  [2796] = 2796,
  [2797] = 725,
  [2798] = 723,
  [2799] = 721,
  [2800] = 2800,
  [2801] = 723,
  [2802] = 725,
  [2803] = 2803,
  [2804] = 723,
  [2805] = 2805,
  [2806] = 721,
  [2807] = 725,
  [2808] = 2808,
  [2809] = 2809,
  [2810] = 2810,
  [2811] = 2811,
  [2812] = 2812,
  [2813] = 2813,
  [2814] = 2814,
  [2815] = 2815,
  [2816] = 2816,
  [2817] = 2817,
  [2818] = 2818,
  [2819] = 2819,
  [2820] = 2820,
  [2821] = 2821,
  [2822] = 2822,
  [2823] = 2823,
  [2824] = 2824,
  [2825] = 2825,
  [2826] = 2826,
  [2827] = 2827,
  [2828] = 2828,
  [2829] = 2829,
  [2830] = 2830,
  [2831] = 2831,
  [2832] = 2832,
  [2833] = 2833,
  [2834] = 2834,
  [2835] = 2835,
  [2836] = 2836,
  [2837] = 725,
  [2838] = 2838,
  [2839] = 2839,
  [2840] = 2840,
  [2841] = 2841,
  [2842] = 2842,
  [2843] = 2843,
  [2844] = 2844,
  [2845] = 2845,
  [2846] = 2846,
  [2847] = 2847,
  [2848] = 2848,
  [2849] = 721,
  [2850] = 2850,
  [2851] = 2851,
  [2852] = 2852,
  [2853] = 2853,
  [2854] = 2854,
  [2855] = 2855,
  [2856] = 2856,
  [2857] = 2857,
  [2858] = 2858,
  [2859] = 723,
  [2860] = 2860,
  [2861] = 2861,
  [2862] = 2862,
  [2863] = 2863,
  [2864] = 2864,
  [2865] = 2865,
  [2866] = 2866,
  [2867] = 2867,
  [2868] = 2868,
  [2869] = 2869,
  [2870] = 2854,
  [2871] = 2833,
  [2872] = 2815,
  [2873] = 2845,
  [2874] = 721,
  [2875] = 2836,
  [2876] = 2850,
  [2877] = 2825,
  [2878] = 2868,
  [2879] = 2866,
  [2880] = 2867,
  [2881] = 2856,
  [2882] = 2818,
  [2883] = 2820,
  [2884] = 2834,
  [2885] = 2832,
  [2886] = 2838,
  [2887] = 2887,
  [2888] = 2840,
  [2889] = 2821,
  [2890] = 2853,
  [2891] = 2839,
  [2892] = 2864,
  [2893] = 2822,
  [2894] = 2823,
  [2895] = 2851,
  [2896] = 2852,
  [2897] = 2824,
  [2898] = 2844,
  [2899] = 2865,
  [2900] = 723,
  [2901] = 2817,
  [2902] = 2848,
  [2903] = 2819,
  [2904] = 2858,
  [2905] = 2814,
  [2906] = 2861,
  [2907] = 2869,
  [2908] = 2863,
  [2909] = 2827,
  [2910] = 2828,
  [2911] = 2809,
  [2912] = 2813,
  [2913] = 2860,
  [2914] = 2862,
  [2915] = 2831,
  [2916] = 2835,
  [2917] = 2810,
  [2918] = 2811,
  [2919] = 2841,
  [2920] = 2846,
  [2921] = 2829,
  [2922] = 2843,
  [2923] = 2842,
  [2924] = 2855,
  [2925] = 2830,
  [2926] = 2857,
  [2927] = 2812,
  [2928] = 2808,
  [2929] = 2816,
  [2930] = 2826,
  [2931] = 2847,
  [2932] = 2865,
  [2933] = 2825,
  [2934] = 2854,
  [2935] = 723,
  [2936] = 2838,
  [2937] = 2831,
  [2938] = 2835,
  [2939] = 2852,
  [2940] = 2836,
  [2941] = 2811,
  [2942] = 2822,
  [2943] = 2866,
  [2944] = 2816,
  [2945] = 2826,
  [2946] = 2832,
  [2947] = 2813,
  [2948] = 2850,
  [2949] = 2857,
  [2950] = 2843,
  [2951] = 2823,
  [2952] = 2808,
  [2953] = 2812,
  [2954] = 2846,
  [2955] = 2847,
  [2956] = 2848,
  [2957] = 2814,
  [2958] = 2817,
  [2959] = 2959,
  [2960] = 2821,
  [2961] = 2863,
  [2962] = 725,
  [2963] = 1327,
  [2964] = 2869,
  [2965] = 2842,
  [2966] = 2868,
  [2967] = 2833,
  [2968] = 2841,
  [2969] = 2829,
  [2970] = 2855,
  [2971] = 2867,
  [2972] = 2809,
  [2973] = 2818,
  [2974] = 2861,
  [2975] = 2851,
  [2976] = 2864,
  [2977] = 2856,
  [2978] = 2834,
  [2979] = 2839,
  [2980] = 2853,
  [2981] = 2858,
  [2982] = 2845,
  [2983] = 2860,
  [2984] = 2820,
  [2985] = 2824,
  [2986] = 2827,
  [2987] = 2844,
  [2988] = 2815,
  [2989] = 2830,
  [2990] = 721,
  [2991] = 2819,
  [2992] = 2810,
  [2993] = 2828,
  [2994] = 2862,
  [2995] = 2840,
  [2996] = 2996,
  [2997] = 2997,
  [2998] = 2998,
  [2999] = 2999,
  [3000] = 2997,
  [3001] = 2997,
  [3002] = 2996,
  [3003] = 3003,
  [3004] = 2996,
  [3005] = 2999,
  [3006] = 2999,
  [3007] = 2996,
  [3008] = 2997,
  [3009] = 2997,
  [3010] = 2579,
  [3011] = 2999,
  [3012] = 2997,
  [3013] = 3013,
  [3014] = 1871,
  [3015] = 3015,
  [3016] = 1682,
  [3017] = 3017,
  [3018] = 1786,
  [3019] = 1707,
  [3020] = 1856,
  [3021] = 723,
  [3022] = 3022,
  [3023] = 721,
  [3024] = 3024,
  [3025] = 725,
  [3026] = 3026,
  [3027] = 3026,
  [3028] = 3028,
  [3029] = 3026,
  [3030] = 2287,
  [3031] = 3031,
  [3032] = 1959,
  [3033] = 1961,
  [3034] = 3034,
  [3035] = 2041,
  [3036] = 2043,
  [3037] = 2045,
  [3038] = 2051,
  [3039] = 2074,
  [3040] = 2164,
  [3041] = 2181,
  [3042] = 2183,
  [3043] = 2213,
  [3044] = 3003,
  [3045] = 2242,
  [3046] = 2250,
  [3047] = 2252,
  [3048] = 2257,
  [3049] = 2258,
  [3050] = 2264,
  [3051] = 3051,
  [3052] = 2286,
  [3053] = 1918,
  [3054] = 1919,
  [3055] = 2998,
  [3056] = 1921,
  [3057] = 1922,
  [3058] = 1923,
  [3059] = 805,
  [3060] = 1929,
  [3061] = 1953,
  [3062] = 1969,
  [3063] = 1970,
  [3064] = 1972,
  [3065] = 1977,
  [3066] = 3066,
  [3067] = 3067,
  [3068] = 3003,
  [3069] = 2998,
  [3070] = 2029,
  [3071] = 2040,
  [3072] = 3034,
  [3073] = 3066,
  [3074] = 3074,
  [3075] = 3075,
  [3076] = 2215,
  [3077] = 3077,
  [3078] = 2044,
  [3079] = 3079,
  [3080] = 3028,
  [3081] = 3026,
  [3082] = 2047,
  [3083] = 2053,
  [3084] = 3079,
  [3085] = 3028,
  [3086] = 3026,
  [3087] = 2163,
  [3088] = 3079,
  [3089] = 3028,
  [3090] = 3026,
  [3091] = 3079,
  [3092] = 3028,
  [3093] = 3026,
  [3094] = 3079,
  [3095] = 3028,
  [3096] = 3026,
  [3097] = 3079,
  [3098] = 3028,
  [3099] = 3026,
  [3100] = 3079,
  [3101] = 3028,
  [3102] = 3026,
  [3103] = 3079,
  [3104] = 3028,
  [3105] = 3026,
  [3106] = 3079,
  [3107] = 3028,
  [3108] = 3026,
  [3109] = 3079,
  [3110] = 3028,
  [3111] = 3026,
  [3112] = 3112,
  [3113] = 3079,
  [3114] = 3028,
  [3115] = 3026,
  [3116] = 2285,
  [3117] = 3079,
  [3118] = 3028,
  [3119] = 3026,
  [3120] = 3079,
  [3121] = 3028,
  [3122] = 3026,
  [3123] = 3079,
  [3124] = 3028,
  [3125] = 3026,
  [3126] = 3079,
  [3127] = 3028,
  [3128] = 3026,
  [3129] = 3079,
  [3130] = 3028,
  [3131] = 3079,
  [3132] = 3028,
  [3133] = 3026,
  [3134] = 3079,
  [3135] = 3028,
  [3136] = 3026,
  [3137] = 3079,
  [3138] = 3028,
  [3139] = 3026,
  [3140] = 3079,
  [3141] = 3028,
  [3142] = 3026,
  [3143] = 3079,
  [3144] = 3028,
  [3145] = 3026,
  [3146] = 3079,
  [3147] = 3028,
  [3148] = 3026,
  [3149] = 1920,
  [3150] = 3079,
  [3151] = 3028,
  [3152] = 3026,
  [3153] = 3079,
  [3154] = 3028,
  [3155] = 3026,
  [3156] = 3079,
  [3157] = 3028,
  [3158] = 3026,
  [3159] = 3079,
  [3160] = 3028,
  [3161] = 3026,
  [3162] = 3079,
  [3163] = 3028,
  [3164] = 3026,
  [3165] = 3079,
  [3166] = 3028,
  [3167] = 3026,
  [3168] = 3079,
  [3169] = 3028,
  [3170] = 3026,
  [3171] = 3079,
  [3172] = 3028,
  [3173] = 3026,
  [3174] = 3079,
  [3175] = 3028,
  [3176] = 3026,
  [3177] = 3079,
  [3178] = 3028,
  [3179] = 3026,
  [3180] = 3079,
  [3181] = 3028,
  [3182] = 3026,
  [3183] = 3079,
  [3184] = 3028,
  [3185] = 3026,
  [3186] = 3079,
  [3187] = 3028,
  [3188] = 3026,
  [3189] = 3079,
  [3190] = 3028,
  [3191] = 3026,
  [3192] = 3079,
  [3193] = 3028,
  [3194] = 3026,
  [3195] = 3079,
  [3196] = 3028,
  [3197] = 3026,
  [3198] = 3079,
  [3199] = 3028,
  [3200] = 3026,
  [3201] = 3079,
  [3202] = 3028,
  [3203] = 3026,
  [3204] = 3079,
  [3205] = 3028,
  [3206] = 3026,
  [3207] = 3079,
  [3208] = 3028,
  [3209] = 3026,
  [3210] = 3079,
  [3211] = 3028,
  [3212] = 3026,
  [3213] = 1952,
  [3214] = 3079,
  [3215] = 3028,
  [3216] = 3026,
  [3217] = 1957,
  [3218] = 3079,
  [3219] = 3028,
  [3220] = 3026,
  [3221] = 3079,
  [3222] = 3028,
  [3223] = 3026,
  [3224] = 1968,
  [3225] = 3079,
  [3226] = 3028,
  [3227] = 3026,
  [3228] = 3079,
  [3229] = 3028,
  [3230] = 3026,
  [3231] = 3079,
  [3232] = 3028,
  [3233] = 3026,
  [3234] = 3079,
  [3235] = 3028,
  [3236] = 3079,
  [3237] = 1952,
  [3238] = 1972,
  [3239] = 1977,
  [3240] = 2287,
  [3241] = 3241,
  [3242] = 1959,
  [3243] = 1961,
  [3244] = 3244,
  [3245] = 2041,
  [3246] = 2043,
  [3247] = 2045,
  [3248] = 2051,
  [3249] = 2074,
  [3250] = 2164,
  [3251] = 2181,
  [3252] = 2183,
  [3253] = 2242,
  [3254] = 2250,
  [3255] = 2252,
  [3256] = 2258,
  [3257] = 2264,
  [3258] = 2286,
  [3259] = 1918,
  [3260] = 1919,
  [3261] = 1921,
  [3262] = 886,
  [3263] = 1922,
  [3264] = 1923,
  [3265] = 1929,
  [3266] = 1953,
  [3267] = 1969,
  [3268] = 1970,
  [3269] = 1972,
  [3270] = 805,
  [3271] = 1977,
  [3272] = 1961,
  [3273] = 3273,
  [3274] = 3274,
  [3275] = 3275,
  [3276] = 2215,
  [3277] = 2285,
  [3278] = 838,
  [3279] = 805,
  [3280] = 840,
  [3281] = 721,
  [3282] = 3017,
  [3283] = 839,
  [3284] = 2044,
  [3285] = 847,
  [3286] = 848,
  [3287] = 723,
  [3288] = 2047,
  [3289] = 852,
  [3290] = 854,
  [3291] = 855,
  [3292] = 874,
  [3293] = 856,
  [3294] = 858,
  [3295] = 859,
  [3296] = 1969,
  [3297] = 1970,
  [3298] = 1957,
  [3299] = 2163,
  [3300] = 860,
  [3301] = 863,
  [3302] = 864,
  [3303] = 865,
  [3304] = 866,
  [3305] = 867,
  [3306] = 725,
  [3307] = 869,
  [3308] = 2579,
  [3309] = 3309,
  [3310] = 875,
  [3311] = 877,
  [3312] = 878,
  [3313] = 879,
  [3314] = 880,
  [3315] = 881,
  [3316] = 2287,
  [3317] = 3275,
  [3318] = 1959,
  [3319] = 871,
  [3320] = 885,
  [3321] = 3321,
  [3322] = 844,
  [3323] = 2041,
  [3324] = 2043,
  [3325] = 2045,
  [3326] = 2047,
  [3327] = 2051,
  [3328] = 2074,
  [3329] = 2163,
  [3330] = 2164,
  [3331] = 2181,
  [3332] = 2183,
  [3333] = 2213,
  [3334] = 2242,
  [3335] = 2250,
  [3336] = 2252,
  [3337] = 2257,
  [3338] = 2258,
  [3339] = 2264,
  [3340] = 2285,
  [3341] = 2286,
  [3342] = 1918,
  [3343] = 1919,
  [3344] = 3024,
  [3345] = 1921,
  [3346] = 1922,
  [3347] = 1923,
  [3348] = 1929,
  [3349] = 1952,
  [3350] = 1953,
  [3351] = 1957,
  [3352] = 3352,
  [3353] = 886,
  [3354] = 877,
  [3355] = 878,
  [3356] = 879,
  [3357] = 880,
  [3358] = 881,
  [3359] = 871,
  [3360] = 885,
  [3361] = 844,
  [3362] = 3362,
  [3363] = 3363,
  [3364] = 838,
  [3365] = 3362,
  [3366] = 3366,
  [3367] = 840,
  [3368] = 847,
  [3369] = 848,
  [3370] = 3362,
  [3371] = 852,
  [3372] = 854,
  [3373] = 3362,
  [3374] = 855,
  [3375] = 856,
  [3376] = 3362,
  [3377] = 858,
  [3378] = 859,
  [3379] = 860,
  [3380] = 863,
  [3381] = 721,
  [3382] = 864,
  [3383] = 865,
  [3384] = 866,
  [3385] = 867,
  [3386] = 721,
  [3387] = 869,
  [3388] = 838,
  [3389] = 723,
  [3390] = 874,
  [3391] = 840,
  [3392] = 839,
  [3393] = 3363,
  [3394] = 721,
  [3395] = 847,
  [3396] = 848,
  [3397] = 723,
  [3398] = 875,
  [3399] = 877,
  [3400] = 878,
  [3401] = 879,
  [3402] = 880,
  [3403] = 881,
  [3404] = 852,
  [3405] = 3405,
  [3406] = 854,
  [3407] = 855,
  [3408] = 874,
  [3409] = 856,
  [3410] = 858,
  [3411] = 871,
  [3412] = 885,
  [3413] = 844,
  [3414] = 859,
  [3415] = 805,
  [3416] = 725,
  [3417] = 860,
  [3418] = 863,
  [3419] = 839,
  [3420] = 725,
  [3421] = 864,
  [3422] = 865,
  [3423] = 866,
  [3424] = 867,
  [3425] = 886,
  [3426] = 869,
  [3427] = 875,
  [3428] = 723,
  [3429] = 3362,
  [3430] = 3363,
  [3431] = 3363,
  [3432] = 725,
  [3433] = 3433,
  [3434] = 3434,
  [3435] = 3435,
  [3436] = 3436,
  [3437] = 3433,
  [3438] = 3438,
  [3439] = 3439,
  [3440] = 3440,
  [3441] = 3434,
  [3442] = 3442,
  [3443] = 3435,
  [3444] = 725,
  [3445] = 3445,
  [3446] = 3435,
  [3447] = 3436,
  [3448] = 3438,
  [3449] = 3439,
  [3450] = 886,
  [3451] = 3440,
  [3452] = 3433,
  [3453] = 3442,
  [3454] = 721,
  [3455] = 3440,
  [3456] = 3456,
  [3457] = 3433,
  [3458] = 3433,
  [3459] = 3435,
  [3460] = 3436,
  [3461] = 3438,
  [3462] = 3439,
  [3463] = 3440,
  [3464] = 3442,
  [3465] = 3445,
  [3466] = 3433,
  [3467] = 3435,
  [3468] = 3436,
  [3469] = 3439,
  [3470] = 3440,
  [3471] = 3442,
  [3472] = 3472,
  [3473] = 3473,
  [3474] = 3436,
  [3475] = 3435,
  [3476] = 3436,
  [3477] = 3439,
  [3478] = 3433,
  [3479] = 3440,
  [3480] = 3442,
  [3481] = 721,
  [3482] = 3433,
  [3483] = 3483,
  [3484] = 3440,
  [3485] = 3442,
  [3486] = 723,
  [3487] = 839,
  [3488] = 3488,
  [3489] = 3433,
  [3490] = 723,
  [3491] = 3433,
  [3492] = 3433,
  [3493] = 721,
  [3494] = 3433,
  [3495] = 3433,
  [3496] = 3433,
  [3497] = 3433,
  [3498] = 3433,
  [3499] = 3433,
  [3500] = 3500,
  [3501] = 3433,
  [3502] = 3433,
  [3503] = 3433,
  [3504] = 3433,
  [3505] = 3473,
  [3506] = 3433,
  [3507] = 3433,
  [3508] = 3508,
  [3509] = 3433,
  [3510] = 3433,
  [3511] = 723,
  [3512] = 3433,
  [3513] = 3433,
  [3514] = 3514,
  [3515] = 3515,
  [3516] = 3433,
  [3517] = 3433,
  [3518] = 838,
  [3519] = 3433,
  [3520] = 840,
  [3521] = 3433,
  [3522] = 847,
  [3523] = 848,
  [3524] = 3433,
  [3525] = 3433,
  [3526] = 852,
  [3527] = 854,
  [3528] = 855,
  [3529] = 874,
  [3530] = 856,
  [3531] = 858,
  [3532] = 859,
  [3533] = 3433,
  [3534] = 860,
  [3535] = 863,
  [3536] = 864,
  [3537] = 865,
  [3538] = 866,
  [3539] = 867,
  [3540] = 3433,
  [3541] = 3541,
  [3542] = 869,
  [3543] = 3543,
  [3544] = 875,
  [3545] = 3433,
  [3546] = 877,
  [3547] = 878,
  [3548] = 879,
  [3549] = 880,
  [3550] = 881,
  [3551] = 3433,
  [3552] = 3433,
  [3553] = 3433,
  [3554] = 3433,
  [3555] = 3433,
  [3556] = 3433,
  [3557] = 3433,
  [3558] = 3433,
  [3559] = 3433,
  [3560] = 3442,
  [3561] = 3433,
  [3562] = 3562,
  [3563] = 871,
  [3564] = 3433,
  [3565] = 3433,
  [3566] = 885,
  [3567] = 844,
  [3568] = 3433,
  [3569] = 3433,
  [3570] = 725,
  [3571] = 3514,
  [3572] = 3572,
  [3573] = 3573,
  [3574] = 3438,
  [3575] = 3439,
  [3576] = 725,
  [3577] = 3434,
  [3578] = 3472,
  [3579] = 3434,
  [3580] = 3472,
  [3581] = 3434,
  [3582] = 3472,
  [3583] = 3434,
  [3584] = 3434,
  [3585] = 3433,
  [3586] = 3586,
  [3587] = 725,
  [3588] = 3588,
  [3589] = 3586,
  [3590] = 3586,
  [3591] = 3586,
  [3592] = 3586,
  [3593] = 725,
  [3594] = 3586,
  [3595] = 725,
  [3596] = 3586,
  [3597] = 3586,
  [3598] = 3586,
  [3599] = 3586,
  [3600] = 3586,
  [3601] = 3586,
  [3602] = 3586,
  [3603] = 721,
  [3604] = 3586,
  [3605] = 3586,
  [3606] = 3586,
  [3607] = 3586,
  [3608] = 3586,
  [3609] = 721,
  [3610] = 3610,
  [3611] = 3586,
  [3612] = 3586,
  [3613] = 3586,
  [3614] = 3586,
  [3615] = 723,
  [3616] = 3586,
  [3617] = 3617,
  [3618] = 3586,
  [3619] = 3586,
  [3620] = 3586,
  [3621] = 723,
  [3622] = 3586,
  [3623] = 723,
  [3624] = 3586,
  [3625] = 721,
  [3626] = 3586,
  [3627] = 723,
  [3628] = 3586,
  [3629] = 3586,
  [3630] = 3586,
  [3631] = 3586,
  [3632] = 3586,
  [3633] = 3586,
  [3634] = 3586,
  [3635] = 3586,
  [3636] = 721,
  [3637] = 3637,
  [3638] = 3586,
  [3639] = 725,
  [3640] = 723,
  [3641] = 721,
  [3642] = 721,
  [3643] = 3586,
  [3644] = 3586,
  [3645] = 3586,
  [3646] = 723,
  [3647] = 805,
  [3648] = 3586,
  [3649] = 3586,
  [3650] = 3586,
  [3651] = 3617,
  [3652] = 3586,
  [3653] = 3586,
  [3654] = 3586,
  [3655] = 3617,
  [3656] = 3586,
  [3657] = 3586,
  [3658] = 3586,
  [3659] = 3617,
  [3660] = 725,
  [3661] = 3586,
  [3662] = 725,
  [3663] = 3617,
  [3664] = 3586,
  [3665] = 3617,
  [3666] = 3586,
  [3667] = 723,
  [3668] = 721,
  [3669] = 2242,
  [3670] = 880,
  [3671] = 2030,
  [3672] = 2031,
  [3673] = 2033,
  [3674] = 871,
  [3675] = 2259,
  [3676] = 881,
  [3677] = 875,
  [3678] = 885,
  [3679] = 1924,
  [3680] = 2081,
  [3681] = 844,
  [3682] = 2086,
  [3683] = 1983,
  [3684] = 839,
  [3685] = 2212,
  [3686] = 3686,
  [3687] = 3687,
  [3688] = 2221,
  [3689] = 2228,
  [3690] = 2229,
  [3691] = 2230,
  [3692] = 3692,
  [3693] = 2287,
  [3694] = 869,
  [3695] = 2213,
  [3696] = 2261,
  [3697] = 2262,
  [3698] = 2257,
  [3699] = 2263,
  [3700] = 3700,
  [3701] = 859,
  [3702] = 1915,
  [3703] = 858,
  [3704] = 3704,
  [3705] = 2287,
  [3706] = 1959,
  [3707] = 1959,
  [3708] = 1925,
  [3709] = 1926,
  [3710] = 1961,
  [3711] = 1928,
  [3712] = 3712,
  [3713] = 2041,
  [3714] = 2043,
  [3715] = 2045,
  [3716] = 2051,
  [3717] = 2074,
  [3718] = 2164,
  [3719] = 2181,
  [3720] = 2183,
  [3721] = 2250,
  [3722] = 2252,
  [3723] = 2258,
  [3724] = 2264,
  [3725] = 2052,
  [3726] = 2286,
  [3727] = 860,
  [3728] = 863,
  [3729] = 1918,
  [3730] = 1919,
  [3731] = 1921,
  [3732] = 1922,
  [3733] = 1960,
  [3734] = 1923,
  [3735] = 1929,
  [3736] = 1953,
  [3737] = 1962,
  [3738] = 1969,
  [3739] = 1970,
  [3740] = 1972,
  [3741] = 1977,
  [3742] = 3742,
  [3743] = 1971,
  [3744] = 3744,
  [3745] = 877,
  [3746] = 1984,
  [3747] = 1985,
  [3748] = 2010,
  [3749] = 1961,
  [3750] = 2041,
  [3751] = 2043,
  [3752] = 2045,
  [3753] = 2051,
  [3754] = 2074,
  [3755] = 878,
  [3756] = 3756,
  [3757] = 2225,
  [3758] = 2260,
  [3759] = 2164,
  [3760] = 2181,
  [3761] = 2183,
  [3762] = 2213,
  [3763] = 2242,
  [3764] = 2250,
  [3765] = 2252,
  [3766] = 725,
  [3767] = 2257,
  [3768] = 2258,
  [3769] = 2264,
  [3770] = 3770,
  [3771] = 2073,
  [3772] = 3772,
  [3773] = 879,
  [3774] = 864,
  [3775] = 865,
  [3776] = 3776,
  [3777] = 2272,
  [3778] = 2274,
  [3779] = 3779,
  [3780] = 2286,
  [3781] = 866,
  [3782] = 856,
  [3783] = 1918,
  [3784] = 1919,
  [3785] = 3785,
  [3786] = 3786,
  [3787] = 1921,
  [3788] = 1922,
  [3789] = 3789,
  [3790] = 1932,
  [3791] = 1923,
  [3792] = 3792,
  [3793] = 3793,
  [3794] = 1929,
  [3795] = 2025,
  [3796] = 2026,
  [3797] = 1953,
  [3798] = 867,
  [3799] = 1969,
  [3800] = 838,
  [3801] = 721,
  [3802] = 840,
  [3803] = 721,
  [3804] = 1970,
  [3805] = 1972,
  [3806] = 886,
  [3807] = 723,
  [3808] = 847,
  [3809] = 848,
  [3810] = 723,
  [3811] = 3770,
  [3812] = 1977,
  [3813] = 3770,
  [3814] = 2060,
  [3815] = 852,
  [3816] = 854,
  [3817] = 855,
  [3818] = 3770,
  [3819] = 3819,
  [3820] = 3770,
  [3821] = 3770,
  [3822] = 3770,
  [3823] = 3712,
  [3824] = 3742,
  [3825] = 3744,
  [3826] = 2075,
  [3827] = 3712,
  [3828] = 3742,
  [3829] = 3744,
  [3830] = 3712,
  [3831] = 3742,
  [3832] = 3744,
  [3833] = 3712,
  [3834] = 3742,
  [3835] = 3744,
  [3836] = 3712,
  [3837] = 3742,
  [3838] = 3744,
  [3839] = 2011,
  [3840] = 3712,
  [3841] = 3742,
  [3842] = 3744,
  [3843] = 2012,
  [3844] = 3712,
  [3845] = 3742,
  [3846] = 3744,
  [3847] = 3712,
  [3848] = 3742,
  [3849] = 3744,
  [3850] = 3712,
  [3851] = 3742,
  [3852] = 3744,
  [3853] = 3712,
  [3854] = 3742,
  [3855] = 3744,
  [3856] = 3712,
  [3857] = 3742,
  [3858] = 3744,
  [3859] = 3712,
  [3860] = 3742,
  [3861] = 3744,
  [3862] = 3712,
  [3863] = 3742,
  [3864] = 3744,
  [3865] = 3712,
  [3866] = 3742,
  [3867] = 3744,
  [3868] = 3712,
  [3869] = 3742,
  [3870] = 3744,
  [3871] = 3712,
  [3872] = 3742,
  [3873] = 3744,
  [3874] = 3712,
  [3875] = 3742,
  [3876] = 3744,
  [3877] = 3712,
  [3878] = 3742,
  [3879] = 3744,
  [3880] = 3712,
  [3881] = 3742,
  [3882] = 3744,
  [3883] = 3712,
  [3884] = 3742,
  [3885] = 3744,
  [3886] = 3712,
  [3887] = 3742,
  [3888] = 3744,
  [3889] = 3712,
  [3890] = 3742,
  [3891] = 3744,
  [3892] = 3712,
  [3893] = 3742,
  [3894] = 3744,
  [3895] = 3712,
  [3896] = 3742,
  [3897] = 3744,
  [3898] = 3712,
  [3899] = 3742,
  [3900] = 3744,
  [3901] = 3712,
  [3902] = 3742,
  [3903] = 3744,
  [3904] = 3712,
  [3905] = 3742,
  [3906] = 3744,
  [3907] = 3712,
  [3908] = 3742,
  [3909] = 3744,
  [3910] = 3712,
  [3911] = 3742,
  [3912] = 3744,
  [3913] = 3712,
  [3914] = 3742,
  [3915] = 3744,
  [3916] = 3712,
  [3917] = 3742,
  [3918] = 3744,
  [3919] = 3712,
  [3920] = 3742,
  [3921] = 3744,
  [3922] = 3712,
  [3923] = 3742,
  [3924] = 3744,
  [3925] = 3712,
  [3926] = 3742,
  [3927] = 3744,
  [3928] = 3712,
  [3929] = 3742,
  [3930] = 3744,
  [3931] = 3712,
  [3932] = 3742,
  [3933] = 3744,
  [3934] = 3712,
  [3935] = 3742,
  [3936] = 3744,
  [3937] = 3712,
  [3938] = 3742,
  [3939] = 3744,
  [3940] = 3712,
  [3941] = 3742,
  [3942] = 3744,
  [3943] = 3712,
  [3944] = 3742,
  [3945] = 3744,
  [3946] = 3712,
  [3947] = 3742,
  [3948] = 3744,
  [3949] = 3712,
  [3950] = 3742,
  [3951] = 3744,
  [3952] = 3712,
  [3953] = 3742,
  [3954] = 3744,
  [3955] = 3712,
  [3956] = 3742,
  [3957] = 3744,
  [3958] = 3712,
  [3959] = 3742,
  [3960] = 3744,
  [3961] = 3712,
  [3962] = 3742,
  [3963] = 3744,
  [3964] = 3712,
  [3965] = 3742,
  [3966] = 3744,
  [3967] = 3712,
  [3968] = 3742,
  [3969] = 3744,
  [3970] = 3712,
  [3971] = 3742,
  [3972] = 3744,
  [3973] = 3712,
  [3974] = 3742,
  [3975] = 3744,
  [3976] = 874,
  [3977] = 3704,
  [3978] = 3978,
  [3979] = 1926,
  [3980] = 3980,
  [3981] = 1928,
  [3982] = 3982,
  [3983] = 3983,
  [3984] = 3980,
  [3985] = 3982,
  [3986] = 3983,
  [3987] = 3980,
  [3988] = 1932,
  [3989] = 3982,
  [3990] = 3983,
  [3991] = 3980,
  [3992] = 2263,
  [3993] = 3982,
  [3994] = 3983,
  [3995] = 3980,
  [3996] = 3982,
  [3997] = 3983,
  [3998] = 3980,
  [3999] = 2073,
  [4000] = 4000,
  [4001] = 3982,
  [4002] = 3983,
  [4003] = 3980,
  [4004] = 3982,
  [4005] = 3983,
  [4006] = 4006,
  [4007] = 3980,
  [4008] = 4008,
  [4009] = 3982,
  [4010] = 3983,
  [4011] = 2025,
  [4012] = 3980,
  [4013] = 4013,
  [4014] = 2026,
  [4015] = 3978,
  [4016] = 3982,
  [4017] = 3983,
  [4018] = 3980,
  [4019] = 4019,
  [4020] = 4020,
  [4021] = 4006,
  [4022] = 3982,
  [4023] = 3983,
  [4024] = 3980,
  [4025] = 2075,
  [4026] = 1960,
  [4027] = 3982,
  [4028] = 3983,
  [4029] = 3980,
  [4030] = 3982,
  [4031] = 3983,
  [4032] = 3980,
  [4033] = 3982,
  [4034] = 3983,
  [4035] = 3980,
  [4036] = 3982,
  [4037] = 3983,
  [4038] = 3980,
  [4039] = 3982,
  [4040] = 3983,
  [4041] = 3980,
  [4042] = 3982,
  [4043] = 3983,
  [4044] = 3980,
  [4045] = 4045,
  [4046] = 3982,
  [4047] = 3983,
  [4048] = 3980,
  [4049] = 3982,
  [4050] = 3983,
  [4051] = 3980,
  [4052] = 4006,
  [4053] = 3982,
  [4054] = 3983,
  [4055] = 3980,
  [4056] = 3982,
  [4057] = 3983,
  [4058] = 3980,
  [4059] = 4006,
  [4060] = 3982,
  [4061] = 3983,
  [4062] = 3980,
  [4063] = 3982,
  [4064] = 3983,
  [4065] = 3980,
  [4066] = 3982,
  [4067] = 3983,
  [4068] = 3980,
  [4069] = 3982,
  [4070] = 3983,
  [4071] = 3980,
  [4072] = 2011,
  [4073] = 3982,
  [4074] = 3983,
  [4075] = 3980,
  [4076] = 3982,
  [4077] = 3983,
  [4078] = 3980,
  [4079] = 2012,
  [4080] = 3982,
  [4081] = 3983,
  [4082] = 3980,
  [4083] = 4083,
  [4084] = 4006,
  [4085] = 3982,
  [4086] = 3983,
  [4087] = 1962,
  [4088] = 3980,
  [4089] = 2030,
  [4090] = 2031,
  [4091] = 2033,
  [4092] = 3982,
  [4093] = 3983,
  [4094] = 3980,
  [4095] = 3982,
  [4096] = 3983,
  [4097] = 3980,
  [4098] = 3982,
  [4099] = 3983,
  [4100] = 1971,
  [4101] = 3980,
  [4102] = 3982,
  [4103] = 3983,
  [4104] = 3980,
  [4105] = 3982,
  [4106] = 3983,
  [4107] = 3980,
  [4108] = 4108,
  [4109] = 3982,
  [4110] = 3983,
  [4111] = 3980,
  [4112] = 3982,
  [4113] = 3983,
  [4114] = 3980,
  [4115] = 3982,
  [4116] = 3983,
  [4117] = 3980,
  [4118] = 4006,
  [4119] = 3982,
  [4120] = 3983,
  [4121] = 4121,
  [4122] = 4122,
  [4123] = 4006,
  [4124] = 4124,
  [4125] = 3980,
  [4126] = 4126,
  [4127] = 2228,
  [4128] = 4128,
  [4129] = 1984,
  [4130] = 2259,
  [4131] = 4006,
  [4132] = 1985,
  [4133] = 4006,
  [4134] = 4134,
  [4135] = 4124,
  [4136] = 4136,
  [4137] = 4006,
  [4138] = 4138,
  [4139] = 4139,
  [4140] = 1924,
  [4141] = 4141,
  [4142] = 4006,
  [4143] = 4124,
  [4144] = 4006,
  [4145] = 4006,
  [4146] = 1983,
  [4147] = 4008,
  [4148] = 4006,
  [4149] = 4006,
  [4150] = 2010,
  [4151] = 886,
  [4152] = 4006,
  [4153] = 4153,
  [4154] = 4006,
  [4155] = 4006,
  [4156] = 2081,
  [4157] = 2086,
  [4158] = 721,
  [4159] = 4083,
  [4160] = 4160,
  [4161] = 4161,
  [4162] = 4162,
  [4163] = 4124,
  [4164] = 4164,
  [4165] = 2212,
  [4166] = 3983,
  [4167] = 4006,
  [4168] = 4168,
  [4169] = 2052,
  [4170] = 2060,
  [4171] = 725,
  [4172] = 4172,
  [4173] = 4134,
  [4174] = 4174,
  [4175] = 4126,
  [4176] = 3980,
  [4177] = 4124,
  [4178] = 4168,
  [4179] = 4020,
  [4180] = 4174,
  [4181] = 4162,
  [4182] = 4006,
  [4183] = 3982,
  [4184] = 4136,
  [4185] = 3983,
  [4186] = 4186,
  [4187] = 4187,
  [4188] = 4188,
  [4189] = 4189,
  [4190] = 4134,
  [4191] = 4174,
  [4192] = 4192,
  [4193] = 3982,
  [4194] = 2225,
  [4195] = 721,
  [4196] = 4196,
  [4197] = 2260,
  [4198] = 4006,
  [4199] = 3980,
  [4200] = 4168,
  [4201] = 4020,
  [4202] = 1915,
  [4203] = 4162,
  [4204] = 4204,
  [4205] = 3982,
  [4206] = 4136,
  [4207] = 4207,
  [4208] = 4164,
  [4209] = 4006,
  [4210] = 4160,
  [4211] = 3983,
  [4212] = 4006,
  [4213] = 4213,
  [4214] = 2229,
  [4215] = 4134,
  [4216] = 4174,
  [4217] = 4188,
  [4218] = 4006,
  [4219] = 2272,
  [4220] = 3980,
  [4221] = 4221,
  [4222] = 4168,
  [4223] = 4020,
  [4224] = 4224,
  [4225] = 4225,
  [4226] = 4162,
  [4227] = 723,
  [4228] = 2274,
  [4229] = 2230,
  [4230] = 3982,
  [4231] = 4136,
  [4232] = 3983,
  [4233] = 4006,
  [4234] = 4134,
  [4235] = 4174,
  [4236] = 4124,
  [4237] = 4189,
  [4238] = 4006,
  [4239] = 3980,
  [4240] = 4168,
  [4241] = 4020,
  [4242] = 4162,
  [4243] = 2221,
  [4244] = 3982,
  [4245] = 4136,
  [4246] = 3983,
  [4247] = 3983,
  [4248] = 4134,
  [4249] = 4174,
  [4250] = 4006,
  [4251] = 721,
  [4252] = 4252,
  [4253] = 3980,
  [4254] = 4168,
  [4255] = 4020,
  [4256] = 4162,
  [4257] = 4006,
  [4258] = 3982,
  [4259] = 4136,
  [4260] = 3983,
  [4261] = 4261,
  [4262] = 4134,
  [4263] = 4121,
  [4264] = 3980,
  [4265] = 4168,
  [4266] = 4162,
  [4267] = 2261,
  [4268] = 4006,
  [4269] = 3982,
  [4270] = 4136,
  [4271] = 3983,
  [4272] = 805,
  [4273] = 3980,
  [4274] = 4192,
  [4275] = 3982,
  [4276] = 3983,
  [4277] = 4122,
  [4278] = 3980,
  [4279] = 3982,
  [4280] = 3983,
  [4281] = 3980,
  [4282] = 723,
  [4283] = 3982,
  [4284] = 3983,
  [4285] = 3980,
  [4286] = 3982,
  [4287] = 3983,
  [4288] = 3980,
  [4289] = 3982,
  [4290] = 3983,
  [4291] = 2262,
  [4292] = 3980,
  [4293] = 4221,
  [4294] = 3982,
  [4295] = 3983,
  [4296] = 3980,
  [4297] = 1925,
  [4298] = 4006,
  [4299] = 3982,
  [4300] = 3983,
  [4301] = 4224,
  [4302] = 3980,
  [4303] = 3982,
  [4304] = 3983,
  [4305] = 4225,
  [4306] = 4187,
  [4307] = 3980,
  [4308] = 4187,
  [4309] = 3982,
  [4310] = 4187,
  [4311] = 4187,
  [4312] = 4187,
  [4313] = 4187,
  [4314] = 723,
  [4315] = 854,
  [4316] = 2031,
  [4317] = 2033,
  [4318] = 863,
  [4319] = 2075,
  [4320] = 839,
  [4321] = 2225,
  [4322] = 2081,
  [4323] = 2086,
  [4324] = 2260,
  [4325] = 2212,
  [4326] = 2221,
  [4327] = 2228,
  [4328] = 2229,
  [4329] = 877,
  [4330] = 2230,
  [4331] = 847,
  [4332] = 2259,
  [4333] = 2261,
  [4334] = 2262,
  [4335] = 2263,
  [4336] = 721,
  [4337] = 848,
  [4338] = 874,
  [4339] = 852,
  [4340] = 1915,
  [4341] = 856,
  [4342] = 855,
  [4343] = 858,
  [4344] = 4252,
  [4345] = 723,
  [4346] = 875,
  [4347] = 1924,
  [4348] = 1925,
  [4349] = 1926,
  [4350] = 1928,
  [4351] = 840,
  [4352] = 859,
  [4353] = 2030,
  [4354] = 2060,
  [4355] = 865,
  [4356] = 866,
  [4357] = 867,
  [4358] = 2272,
  [4359] = 2274,
  [4360] = 1960,
  [4361] = 1962,
  [4362] = 838,
  [4363] = 871,
  [4364] = 1971,
  [4365] = 4252,
  [4366] = 878,
  [4367] = 2073,
  [4368] = 879,
  [4369] = 1983,
  [4370] = 1984,
  [4371] = 1985,
  [4372] = 864,
  [4373] = 2010,
  [4374] = 860,
  [4375] = 885,
  [4376] = 844,
  [4377] = 881,
  [4378] = 880,
  [4379] = 886,
  [4380] = 869,
  [4381] = 1932,
  [4382] = 2011,
  [4383] = 2012,
  [4384] = 2025,
  [4385] = 2026,
  [4386] = 2052,
  [4387] = 4387,
  [4388] = 4388,
  [4389] = 4252,
  [4390] = 886,
  [4391] = 4252,
  [4392] = 4387,
  [4393] = 4387,
  [4394] = 4387,
  [4395] = 4387,
  [4396] = 4387,
  [4397] = 4387,
  [4398] = 4387,
  [4399] = 805,
  [4400] = 852,
  [4401] = 880,
  [4402] = 881,
  [4403] = 860,
  [4404] = 2075,
  [4405] = 863,
  [4406] = 856,
  [4407] = 864,
  [4408] = 865,
  [4409] = 866,
  [4410] = 867,
  [4411] = 859,
  [4412] = 847,
  [4413] = 871,
  [4414] = 1932,
  [4415] = 885,
  [4416] = 2225,
  [4417] = 2259,
  [4418] = 1924,
  [4419] = 844,
  [4420] = 2025,
  [4421] = 875,
  [4422] = 858,
  [4423] = 877,
  [4424] = 2260,
  [4425] = 848,
  [4426] = 878,
  [4427] = 2272,
  [4428] = 2274,
  [4429] = 886,
  [4430] = 2026,
  [4431] = 805,
  [4432] = 805,
  [4433] = 839,
  [4434] = 838,
  [4435] = 840,
  [4436] = 869,
  [4437] = 4437,
  [4438] = 4438,
  [4439] = 1983,
  [4440] = 879,
  [4441] = 805,
  [4442] = 854,
  [4443] = 855,
  [4444] = 874,
  [4445] = 869,
  [4446] = 885,
  [4447] = 847,
  [4448] = 4448,
  [4449] = 844,
  [4450] = 848,
  [4451] = 877,
  [4452] = 838,
  [4453] = 878,
  [4454] = 840,
  [4455] = 847,
  [4456] = 848,
  [4457] = 839,
  [4458] = 877,
  [4459] = 885,
  [4460] = 852,
  [4461] = 805,
  [4462] = 1189,
  [4463] = 854,
  [4464] = 855,
  [4465] = 856,
  [4466] = 4466,
  [4467] = 858,
  [4468] = 859,
  [4469] = 844,
  [4470] = 879,
  [4471] = 860,
  [4472] = 839,
  [4473] = 863,
  [4474] = 864,
  [4475] = 865,
  [4476] = 866,
  [4477] = 867,
  [4478] = 852,
  [4479] = 880,
  [4480] = 880,
  [4481] = 805,
  [4482] = 874,
  [4483] = 854,
  [4484] = 855,
  [4485] = 871,
  [4486] = 4486,
  [4487] = 4486,
  [4488] = 856,
  [4489] = 875,
  [4490] = 839,
  [4491] = 805,
  [4492] = 881,
  [4493] = 881,
  [4494] = 858,
  [4495] = 4495,
  [4496] = 875,
  [4497] = 871,
  [4498] = 885,
  [4499] = 838,
  [4500] = 844,
  [4501] = 4501,
  [4502] = 874,
  [4503] = 886,
  [4504] = 878,
  [4505] = 4505,
  [4506] = 4252,
  [4507] = 2579,
  [4508] = 840,
  [4509] = 879,
  [4510] = 859,
  [4511] = 838,
  [4512] = 840,
  [4513] = 875,
  [4514] = 4514,
  [4515] = 847,
  [4516] = 877,
  [4517] = 848,
  [4518] = 852,
  [4519] = 854,
  [4520] = 855,
  [4521] = 856,
  [4522] = 858,
  [4523] = 860,
  [4524] = 859,
  [4525] = 4525,
  [4526] = 860,
  [4527] = 863,
  [4528] = 863,
  [4529] = 864,
  [4530] = 878,
  [4531] = 864,
  [4532] = 865,
  [4533] = 865,
  [4534] = 866,
  [4535] = 867,
  [4536] = 866,
  [4537] = 867,
  [4538] = 879,
  [4539] = 880,
  [4540] = 869,
  [4541] = 874,
  [4542] = 881,
  [4543] = 4525,
  [4544] = 869,
  [4545] = 886,
  [4546] = 886,
  [4547] = 871,
  [4548] = 886,
  [4549] = 4549,
  [4550] = 844,
  [4551] = 2998,
  [4552] = 1189,
  [4553] = 852,
  [4554] = 854,
  [4555] = 855,
  [4556] = 874,
  [4557] = 856,
  [4558] = 858,
  [4559] = 875,
  [4560] = 859,
  [4561] = 805,
  [4562] = 840,
  [4563] = 839,
  [4564] = 856,
  [4565] = 852,
  [4566] = 860,
  [4567] = 863,
  [4568] = 877,
  [4569] = 878,
  [4570] = 879,
  [4571] = 864,
  [4572] = 865,
  [4573] = 866,
  [4574] = 858,
  [4575] = 859,
  [4576] = 4576,
  [4577] = 867,
  [4578] = 3003,
  [4579] = 875,
  [4580] = 854,
  [4581] = 878,
  [4582] = 860,
  [4583] = 880,
  [4584] = 863,
  [4585] = 855,
  [4586] = 4586,
  [4587] = 805,
  [4588] = 847,
  [4589] = 864,
  [4590] = 865,
  [4591] = 848,
  [4592] = 866,
  [4593] = 867,
  [4594] = 856,
  [4595] = 858,
  [4596] = 885,
  [4597] = 4597,
  [4598] = 4597,
  [4599] = 879,
  [4600] = 874,
  [4601] = 839,
  [4602] = 871,
  [4603] = 869,
  [4604] = 838,
  [4605] = 859,
  [4606] = 871,
  [4607] = 880,
  [4608] = 840,
  [4609] = 869,
  [4610] = 886,
  [4611] = 871,
  [4612] = 844,
  [4613] = 860,
  [4614] = 874,
  [4615] = 863,
  [4616] = 4616,
  [4617] = 885,
  [4618] = 805,
  [4619] = 864,
  [4620] = 865,
  [4621] = 866,
  [4622] = 867,
  [4623] = 838,
  [4624] = 877,
  [4625] = 878,
  [4626] = 879,
  [4627] = 881,
  [4628] = 838,
  [4629] = 847,
  [4630] = 840,
  [4631] = 852,
  [4632] = 4632,
  [4633] = 848,
  [4634] = 4634,
  [4635] = 880,
  [4636] = 854,
  [4637] = 869,
  [4638] = 881,
  [4639] = 805,
  [4640] = 847,
  [4641] = 848,
  [4642] = 4634,
  [4643] = 855,
  [4644] = 875,
  [4645] = 4586,
  [4646] = 805,
  [4647] = 885,
  [4648] = 844,
  [4649] = 4649,
  [4650] = 881,
  [4651] = 4651,
  [4652] = 839,
  [4653] = 877,
  [4654] = 886,
  [4655] = 805,
  [4656] = 856,
  [4657] = 855,
  [4658] = 4658,
  [4659] = 856,
  [4660] = 4660,
  [4661] = 858,
  [4662] = 2041,
  [4663] = 859,
  [4664] = 866,
  [4665] = 867,
  [4666] = 1019,
  [4667] = 880,
  [4668] = 860,
  [4669] = 863,
  [4670] = 886,
  [4671] = 4671,
  [4672] = 4672,
  [4673] = 2183,
  [4674] = 864,
  [4675] = 865,
  [4676] = 866,
  [4677] = 4677,
  [4678] = 878,
  [4679] = 871,
  [4680] = 867,
  [4681] = 879,
  [4682] = 864,
  [4683] = 2252,
  [4684] = 2264,
  [4685] = 2286,
  [4686] = 865,
  [4687] = 838,
  [4688] = 866,
  [4689] = 867,
  [4690] = 869,
  [4691] = 840,
  [4692] = 874,
  [4693] = 885,
  [4694] = 880,
  [4695] = 1918,
  [4696] = 1919,
  [4697] = 1921,
  [4698] = 4698,
  [4699] = 4699,
  [4700] = 874,
  [4701] = 881,
  [4702] = 1929,
  [4703] = 1953,
  [4704] = 4704,
  [4705] = 885,
  [4706] = 847,
  [4707] = 844,
  [4708] = 848,
  [4709] = 881,
  [4710] = 1969,
  [4711] = 1972,
  [4712] = 844,
  [4713] = 1977,
  [4714] = 854,
  [4715] = 838,
  [4716] = 4660,
  [4717] = 838,
  [4718] = 856,
  [4719] = 838,
  [4720] = 840,
  [4721] = 852,
  [4722] = 854,
  [4723] = 847,
  [4724] = 839,
  [4725] = 840,
  [4726] = 848,
  [4727] = 855,
  [4728] = 852,
  [4729] = 854,
  [4730] = 855,
  [4731] = 4731,
  [4732] = 874,
  [4733] = 856,
  [4734] = 858,
  [4735] = 4735,
  [4736] = 875,
  [4737] = 859,
  [4738] = 858,
  [4739] = 4739,
  [4740] = 856,
  [4741] = 858,
  [4742] = 860,
  [4743] = 875,
  [4744] = 4744,
  [4745] = 863,
  [4746] = 877,
  [4747] = 878,
  [4748] = 859,
  [4749] = 879,
  [4750] = 864,
  [4751] = 865,
  [4752] = 866,
  [4753] = 855,
  [4754] = 881,
  [4755] = 867,
  [4756] = 880,
  [4757] = 4757,
  [4758] = 869,
  [4759] = 859,
  [4760] = 881,
  [4761] = 4735,
  [4762] = 864,
  [4763] = 886,
  [4764] = 869,
  [4765] = 869,
  [4766] = 874,
  [4767] = 886,
  [4768] = 4768,
  [4769] = 856,
  [4770] = 4770,
  [4771] = 865,
  [4772] = 4772,
  [4773] = 874,
  [4774] = 4774,
  [4775] = 852,
  [4776] = 854,
  [4777] = 855,
  [4778] = 874,
  [4779] = 865,
  [4780] = 875,
  [4781] = 859,
  [4782] = 886,
  [4783] = 858,
  [4784] = 871,
  [4785] = 1961,
  [4786] = 4735,
  [4787] = 860,
  [4788] = 863,
  [4789] = 886,
  [4790] = 4790,
  [4791] = 864,
  [4792] = 805,
  [4793] = 875,
  [4794] = 859,
  [4795] = 865,
  [4796] = 866,
  [4797] = 847,
  [4798] = 866,
  [4799] = 867,
  [4800] = 880,
  [4801] = 4801,
  [4802] = 4735,
  [4803] = 885,
  [4804] = 844,
  [4805] = 848,
  [4806] = 4806,
  [4807] = 860,
  [4808] = 893,
  [4809] = 879,
  [4810] = 863,
  [4811] = 4811,
  [4812] = 877,
  [4813] = 871,
  [4814] = 881,
  [4815] = 875,
  [4816] = 852,
  [4817] = 839,
  [4818] = 867,
  [4819] = 852,
  [4820] = 4774,
  [4821] = 871,
  [4822] = 839,
  [4823] = 4823,
  [4824] = 4735,
  [4825] = 878,
  [4826] = 4826,
  [4827] = 847,
  [4828] = 4672,
  [4829] = 878,
  [4830] = 871,
  [4831] = 839,
  [4832] = 877,
  [4833] = 840,
  [4834] = 878,
  [4835] = 879,
  [4836] = 886,
  [4837] = 875,
  [4838] = 885,
  [4839] = 880,
  [4840] = 844,
  [4841] = 881,
  [4842] = 4842,
  [4843] = 879,
  [4844] = 869,
  [4845] = 4735,
  [4846] = 4846,
  [4847] = 848,
  [4848] = 854,
  [4849] = 839,
  [4850] = 869,
  [4851] = 871,
  [4852] = 4852,
  [4853] = 864,
  [4854] = 839,
  [4855] = 885,
  [4856] = 886,
  [4857] = 4735,
  [4858] = 4811,
  [4859] = 855,
  [4860] = 844,
  [4861] = 838,
  [4862] = 838,
  [4863] = 840,
  [4864] = 4735,
  [4865] = 886,
  [4866] = 840,
  [4867] = 4867,
  [4868] = 4735,
  [4869] = 847,
  [4870] = 885,
  [4871] = 848,
  [4872] = 877,
  [4873] = 844,
  [4874] = 2287,
  [4875] = 847,
  [4876] = 860,
  [4877] = 863,
  [4878] = 848,
  [4879] = 902,
  [4880] = 880,
  [4881] = 877,
  [4882] = 878,
  [4883] = 4735,
  [4884] = 879,
  [4885] = 852,
  [4886] = 854,
  [4887] = 860,
  [4888] = 863,
  [4889] = 877,
  [4890] = 4772,
  [4891] = 858,
  [4892] = 864,
  [4893] = 871,
  [4894] = 858,
  [4895] = 4895,
  [4896] = 4896,
  [4897] = 886,
  [4898] = 859,
  [4899] = 4895,
  [4900] = 4900,
  [4901] = 885,
  [4902] = 844,
  [4903] = 4895,
  [4904] = 4904,
  [4905] = 4905,
  [4906] = 4906,
  [4907] = 875,
  [4908] = 877,
  [4909] = 4671,
  [4910] = 4910,
  [4911] = 4911,
  [4912] = 4896,
  [4913] = 4757,
  [4914] = 4914,
  [4915] = 4896,
  [4916] = 4916,
  [4917] = 4677,
  [4918] = 4895,
  [4919] = 860,
  [4920] = 847,
  [4921] = 863,
  [4922] = 4911,
  [4923] = 4900,
  [4924] = 4924,
  [4925] = 4904,
  [4926] = 848,
  [4927] = 856,
  [4928] = 4826,
  [4929] = 865,
  [4930] = 866,
  [4931] = 4931,
  [4932] = 880,
  [4933] = 867,
  [4934] = 4924,
  [4935] = 4935,
  [4936] = 878,
  [4937] = 879,
  [4938] = 4906,
  [4939] = 4904,
  [4940] = 881,
  [4941] = 4906,
  [4942] = 874,
  [4943] = 4699,
  [4944] = 4895,
  [4945] = 852,
  [4946] = 854,
  [4947] = 838,
  [4948] = 4911,
  [4949] = 869,
  [4950] = 855,
  [4951] = 805,
  [4952] = 840,
  [4953] = 4895,
  [4954] = 4924,
  [4955] = 839,
  [4956] = 4895,
  [4957] = 4867,
  [4958] = 4958,
  [4959] = 4959,
  [4960] = 4960,
  [4961] = 4961,
  [4962] = 4962,
  [4963] = 885,
  [4964] = 4964,
  [4965] = 844,
  [4966] = 4966,
  [4967] = 881,
  [4968] = 4968,
  [4969] = 4961,
  [4970] = 4970,
  [4971] = 4971,
  [4972] = 879,
  [4973] = 4958,
  [4974] = 4974,
  [4975] = 869,
  [4976] = 4976,
  [4977] = 860,
  [4978] = 4978,
  [4979] = 863,
  [4980] = 4980,
  [4981] = 856,
  [4982] = 4982,
  [4983] = 858,
  [4984] = 878,
  [4985] = 4985,
  [4986] = 4986,
  [4987] = 4987,
  [4988] = 855,
  [4989] = 859,
  [4990] = 4990,
  [4991] = 4970,
  [4992] = 864,
  [4993] = 4993,
  [4994] = 4994,
  [4995] = 4993,
  [4996] = 4990,
  [4997] = 4997,
  [4998] = 874,
  [4999] = 4990,
  [5000] = 4994,
  [5001] = 4990,
  [5002] = 5002,
  [5003] = 4990,
  [5004] = 4990,
  [5005] = 4997,
  [5006] = 5006,
  [5007] = 5007,
  [5008] = 5008,
  [5009] = 5009,
  [5010] = 875,
  [5011] = 4961,
  [5012] = 852,
  [5013] = 5013,
  [5014] = 5002,
  [5015] = 865,
  [5016] = 854,
  [5017] = 866,
  [5018] = 4997,
  [5019] = 5019,
  [5020] = 877,
  [5021] = 4970,
  [5022] = 5022,
  [5023] = 5023,
  [5024] = 5024,
  [5025] = 867,
  [5026] = 2637,
  [5027] = 4982,
  [5028] = 5028,
  [5029] = 838,
  [5030] = 840,
  [5031] = 4990,
  [5032] = 4960,
  [5033] = 4993,
  [5034] = 4966,
  [5035] = 880,
  [5036] = 839,
  [5037] = 5037,
  [5038] = 871,
  [5039] = 4962,
  [5040] = 5040,
  [5041] = 886,
  [5042] = 847,
  [5043] = 848,
  [5044] = 4997,
  [5045] = 4997,
  [5046] = 4997,
  [5047] = 4997,
  [5048] = 4985,
  [5049] = 4994,
  [5050] = 5040,
  [5051] = 5051,
  [5052] = 5052,
  [5053] = 5053,
  [5054] = 5054,
  [5055] = 5055,
  [5056] = 5056,
  [5057] = 5057,
  [5058] = 5058,
  [5059] = 5059,
  [5060] = 5055,
  [5061] = 5061,
  [5062] = 5052,
  [5063] = 5059,
  [5064] = 5054,
  [5065] = 886,
  [5066] = 5066,
  [5067] = 5067,
  [5068] = 5059,
  [5069] = 5055,
  [5070] = 5052,
  [5071] = 5054,
  [5072] = 5072,
  [5073] = 5052,
  [5074] = 5052,
  [5075] = 5054,
  [5076] = 5076,
  [5077] = 5054,
  [5078] = 5057,
  [5079] = 5079,
  [5080] = 5080,
  [5081] = 5081,
  [5082] = 5058,
  [5083] = 5083,
  [5084] = 5084,
  [5085] = 5085,
  [5086] = 5086,
  [5087] = 5087,
  [5088] = 5088,
  [5089] = 5089,
  [5090] = 5081,
  [5091] = 5091,
  [5092] = 5092,
  [5093] = 5093,
  [5094] = 5094,
  [5095] = 5095,
  [5096] = 5096,
  [5097] = 5097,
  [5098] = 5098,
  [5099] = 5059,
  [5100] = 5100,
  [5101] = 5101,
  [5102] = 5061,
  [5103] = 5100,
  [5104] = 5104,
  [5105] = 5055,
  [5106] = 5061,
  [5107] = 5100,
  [5108] = 5108,
  [5109] = 5094,
  [5110] = 5052,
  [5111] = 5111,
  [5112] = 5007,
  [5113] = 5054,
  [5114] = 5114,
  [5115] = 5061,
  [5116] = 5094,
  [5117] = 5117,
  [5118] = 5118,
  [5119] = 5119,
  [5120] = 5081,
  [5121] = 5121,
  [5122] = 5081,
  [5123] = 5123,
  [5124] = 5124,
  [5125] = 5125,
  [5126] = 5126,
  [5127] = 5059,
  [5128] = 5128,
  [5129] = 5129,
  [5130] = 5089,
  [5131] = 5131,
  [5132] = 5055,
  [5133] = 5133,
  [5134] = 5061,
  [5135] = 5100,
  [5136] = 5052,
  [5137] = 5137,
  [5138] = 5094,
  [5139] = 5054,
  [5140] = 5007,
  [5141] = 5141,
  [5142] = 5142,
  [5143] = 5089,
  [5144] = 5059,
  [5145] = 5145,
  [5146] = 5055,
  [5147] = 5147,
  [5148] = 5148,
  [5149] = 5149,
  [5150] = 5149,
  [5151] = 5151,
  [5152] = 5152,
  [5153] = 5153,
  [5154] = 5154,
  [5155] = 5121,
  [5156] = 5156,
  [5157] = 5157,
  [5158] = 5158,
  [5159] = 5148,
  [5160] = 5160,
  [5161] = 5149,
  [5162] = 5162,
  [5163] = 5149,
  [5164] = 5160,
  [5165] = 5149,
  [5166] = 5166,
  [5167] = 5160,
  [5168] = 5151,
  [5169] = 5169,
  [5170] = 5170,
  [5171] = 5171,
  [5172] = 5172,
  [5173] = 5007,
  [5174] = 5174,
  [5175] = 5175,
  [5176] = 5176,
  [5177] = 5149,
  [5178] = 5178,
  [5179] = 5179,
  [5180] = 5180,
  [5181] = 5181,
  [5182] = 5182,
  [5183] = 5172,
  [5184] = 5151,
  [5185] = 5151,
  [5186] = 5186,
  [5187] = 5149,
  [5188] = 5188,
  [5189] = 5189,
  [5190] = 5149,
  [5191] = 5148,
  [5192] = 5192,
  [5193] = 5172,
  [5194] = 5194,
  [5195] = 5195,
  [5196] = 5149,
  [5197] = 5197,
  [5198] = 5197,
  [5199] = 5169,
  [5200] = 5137,
  [5201] = 5149,
  [5202] = 5151,
  [5203] = 5149,
  [5204] = 5204,
  [5205] = 5149,
  [5206] = 5169,
  [5207] = 5178,
  [5208] = 5172,
  [5209] = 5172,
  [5210] = 5149,
  [5211] = 5211,
  [5212] = 5178,
  [5213] = 5149,
  [5214] = 5149,
  [5215] = 5215,
  [5216] = 5153,
  [5217] = 5175,
  [5218] = 5007,
  [5219] = 5137,
  [5220] = 5151,
  [5221] = 886,
  [5222] = 5222,
  [5223] = 5149,
  [5224] = 5149,
  [5225] = 5225,
  [5226] = 5153,
  [5227] = 5149,
  [5228] = 5228,
  [5229] = 5229,
  [5230] = 5230,
  [5231] = 5231,
  [5232] = 5149,
  [5233] = 5233,
  [5234] = 5234,
  [5235] = 5228,
  [5236] = 5188,
  [5237] = 5237,
  [5238] = 5238,
  [5239] = 5197,
  [5240] = 5169,
  [5241] = 5149,
  [5242] = 5148,
  [5243] = 5243,
  [5244] = 5195,
  [5245] = 5245,
  [5246] = 5121,
  [5247] = 5231,
  [5248] = 5195,
  [5249] = 5249,
  [5250] = 5149,
  [5251] = 5251,
  [5252] = 5169,
  [5253] = 5160,
  [5254] = 5254,
  [5255] = 5149,
  [5256] = 5237,
  [5257] = 5178,
  [5258] = 5149,
  [5259] = 5149,
  [5260] = 5160,
  [5261] = 5149,
  [5262] = 5262,
  [5263] = 5263,
  [5264] = 5149,
  [5265] = 5149,
  [5266] = 886,
  [5267] = 5237,
  [5268] = 5154,
  [5269] = 5149,
  [5270] = 5160,
  [5271] = 5271,
  [5272] = 5154,
  [5273] = 5149,
  [5274] = 5149,
  [5275] = 5169,
  [5276] = 5276,
  [5277] = 5149,
  [5278] = 5148,
  [5279] = 5154,
  [5280] = 5195,
  [5281] = 5149,
  [5282] = 5231,
  [5283] = 5149,
  [5284] = 5149,
  [5285] = 5149,
  [5286] = 5195,
  [5287] = 5287,
  [5288] = 5288,
  [5289] = 5149,
  [5290] = 5151,
  [5291] = 5169,
  [5292] = 5171,
  [5293] = 5149,
  [5294] = 5294,
  [5295] = 5149,
  [5296] = 5149,
  [5297] = 5149,
  [5298] = 5298,
  [5299] = 5299,
  [5300] = 5160,
  [5301] = 5149,
  [5302] = 5302,
  [5303] = 5171,
  [5304] = 5149,
  [5305] = 5149,
  [5306] = 5172,
  [5307] = 5181,
  [5308] = 5149,
  [5309] = 5309,
  [5310] = 5149,
  [5311] = 5172,
  [5312] = 5149,
  [5313] = 5149,
  [5314] = 5154,
  [5315] = 5149,
  [5316] = 5149,
  [5317] = 5148,
  [5318] = 5149,
  [5319] = 5319,
  [5320] = 5154,
  [5321] = 5157,
  [5322] = 5211,
  [5323] = 5323,
  [5324] = 5157,
  [5325] = 5211,
  [5326] = 5157,
  [5327] = 5211,
  [5328] = 5157,
  [5329] = 5157,
  [5330] = 5149,
  [5331] = 5331,
  [5332] = 5332,
  [5333] = 5333,
  [5334] = 5334,
  [5335] = 5332,
  [5336] = 5333,
  [5337] = 5337,
  [5338] = 5334,
  [5339] = 5332,
  [5340] = 5340,
  [5341] = 5333,
  [5342] = 5342,
  [5343] = 5334,
  [5344] = 5332,
  [5345] = 5334,
  [5346] = 5346,
  [5347] = 5332,
  [5348] = 5334,
  [5349] = 5349,
  [5350] = 5233,
  [5351] = 5332,
  [5352] = 5334,
  [5353] = 5332,
  [5354] = 5334,
  [5355] = 871,
  [5356] = 5332,
  [5357] = 5334,
  [5358] = 885,
  [5359] = 5359,
  [5360] = 844,
  [5361] = 5332,
  [5362] = 5334,
  [5363] = 5363,
  [5364] = 5364,
  [5365] = 5332,
  [5366] = 5334,
  [5367] = 5367,
  [5368] = 5332,
  [5369] = 5334,
  [5370] = 1961,
  [5371] = 5334,
  [5372] = 5332,
  [5373] = 5334,
  [5374] = 5332,
  [5375] = 5334,
  [5376] = 5376,
  [5377] = 5377,
  [5378] = 5332,
  [5379] = 5334,
  [5380] = 5332,
  [5381] = 5332,
  [5382] = 5334,
  [5383] = 5186,
  [5384] = 5332,
  [5385] = 5334,
  [5386] = 5386,
  [5387] = 5332,
  [5388] = 5334,
  [5389] = 5389,
  [5390] = 5332,
  [5391] = 5334,
  [5392] = 5249,
  [5393] = 5332,
  [5394] = 5334,
  [5395] = 5332,
  [5396] = 5334,
  [5397] = 5397,
  [5398] = 5398,
  [5399] = 5332,
  [5400] = 5334,
  [5401] = 5332,
  [5402] = 5334,
  [5403] = 5332,
  [5404] = 5334,
  [5405] = 886,
  [5406] = 5332,
  [5407] = 5334,
  [5408] = 5332,
  [5409] = 5334,
  [5410] = 5332,
  [5411] = 5334,
  [5412] = 5332,
  [5413] = 5334,
  [5414] = 5332,
  [5415] = 5334,
  [5416] = 5416,
  [5417] = 5417,
  [5418] = 5332,
  [5419] = 5334,
  [5420] = 5420,
  [5421] = 5332,
  [5422] = 5334,
  [5423] = 5423,
  [5424] = 5332,
  [5425] = 5334,
  [5426] = 5426,
  [5427] = 5332,
  [5428] = 5334,
  [5429] = 5332,
  [5430] = 5334,
  [5431] = 5332,
  [5432] = 5334,
  [5433] = 5332,
  [5434] = 5334,
  [5435] = 5332,
  [5436] = 5334,
  [5437] = 5332,
  [5438] = 5334,
  [5439] = 5332,
  [5440] = 5334,
  [5441] = 5441,
  [5442] = 5442,
  [5443] = 5332,
  [5444] = 5334,
  [5445] = 5332,
  [5446] = 5334,
  [5447] = 5152,
  [5448] = 5332,
  [5449] = 5334,
  [5450] = 5332,
  [5451] = 5334,
  [5452] = 5332,
  [5453] = 5334,
  [5454] = 5332,
  [5455] = 5455,
  [5456] = 5332,
  [5457] = 5334,
  [5458] = 5332,
  [5459] = 5334,
  [5460] = 5332,
  [5461] = 5334,
  [5462] = 5332,
  [5463] = 5334,
  [5464] = 5332,
  [5465] = 5334,
  [5466] = 5332,
  [5467] = 5334,
  [5468] = 5468,
  [5469] = 5469,
  [5470] = 5470,
  [5471] = 5471,
  [5472] = 5472,
  [5473] = 5473,
  [5474] = 5474,
  [5475] = 5475,
  [5476] = 5476,
  [5477] = 5186,
  [5478] = 2270,
  [5479] = 5249,
  [5480] = 1917,
  [5481] = 5481,
  [5482] = 5482,
  [5483] = 5483,
  [5484] = 2165,
  [5485] = 2251,
  [5486] = 5271,
  [5487] = 5487,
  [5488] = 5488,
  [5489] = 5152,
  [5490] = 5490,
  [5491] = 5491,
  [5492] = 5492,
  [5493] = 5493,
  [5494] = 5493,
  [5495] = 1970,
  [5496] = 5473,
  [5497] = 5022,
  [5498] = 5333,
  [5499] = 1919,
  [5500] = 5088,
  [5501] = 5501,
  [5502] = 5502,
  [5503] = 5503,
  [5504] = 5504,
  [5505] = 5505,
  [5506] = 5506,
  [5507] = 1961,
  [5508] = 5508,
  [5509] = 5509,
  [5510] = 5510,
  [5511] = 5511,
  [5512] = 5512,
  [5513] = 5022,
  [5514] = 5233,
  [5515] = 5515,
  [5516] = 5359,
  [5517] = 5363,
  [5518] = 5364,
  [5519] = 5519,
  [5520] = 5520,
  [5521] = 5521,
  [5522] = 5389,
  [5523] = 5398,
  [5524] = 5524,
  [5525] = 5416,
  [5526] = 5417,
  [5527] = 5420,
  [5528] = 5468,
  [5529] = 5469,
  [5530] = 5470,
  [5531] = 5471,
  [5532] = 5472,
  [5533] = 5476,
  [5534] = 5534,
  [5535] = 5237,
  [5536] = 5536,
  [5537] = 5537,
  [5538] = 5538,
  [5539] = 5502,
  [5540] = 5332,
  [5541] = 5333,
  [5542] = 5511,
  [5543] = 5334,
  [5544] = 5492,
  [5545] = 5545,
  [5546] = 5386,
  [5547] = 5386,
  [5548] = 5545,
  [5549] = 5386,
  [5550] = 5332,
  [5551] = 5333,
  [5552] = 5334,
  [5553] = 5492,
  [5554] = 5545,
  [5555] = 5334,
  [5556] = 5556,
  [5557] = 5557,
  [5558] = 5558,
  [5559] = 5559,
  [5560] = 5557,
  [5561] = 5561,
  [5562] = 5562,
  [5563] = 5562,
  [5564] = 5559,
  [5565] = 5557,
  [5566] = 5559,
  [5567] = 886,
  [5568] = 5557,
  [5569] = 5237,
  [5570] = 5559,
  [5571] = 5557,
  [5572] = 5572,
  [5573] = 5573,
  [5574] = 5574,
  [5575] = 1871,
  [5576] = 5559,
  [5577] = 5557,
  [5578] = 5578,
  [5579] = 5515,
  [5580] = 5559,
  [5581] = 5557,
  [5582] = 5582,
  [5583] = 5559,
  [5584] = 5557,
  [5585] = 5559,
  [5586] = 5557,
  [5587] = 5587,
  [5588] = 5588,
  [5589] = 5589,
  [5590] = 5559,
  [5591] = 5557,
  [5592] = 5592,
  [5593] = 5559,
  [5594] = 5559,
  [5595] = 5557,
  [5596] = 5596,
  [5597] = 5559,
  [5598] = 5559,
  [5599] = 5557,
  [5600] = 5600,
  [5601] = 5557,
  [5602] = 5602,
  [5603] = 5559,
  [5604] = 5557,
  [5605] = 5605,
  [5606] = 5606,
  [5607] = 5607,
  [5608] = 5608,
  [5609] = 5559,
  [5610] = 5557,
  [5611] = 5186,
  [5612] = 5612,
  [5613] = 5557,
  [5614] = 5559,
  [5615] = 5557,
  [5616] = 5559,
  [5617] = 5557,
  [5618] = 5618,
  [5619] = 5559,
  [5620] = 5557,
  [5621] = 5621,
  [5622] = 5622,
  [5623] = 5559,
  [5624] = 5557,
  [5625] = 5249,
  [5626] = 5626,
  [5627] = 5588,
  [5628] = 5559,
  [5629] = 5557,
  [5630] = 5626,
  [5631] = 5559,
  [5632] = 5557,
  [5633] = 5559,
  [5634] = 5557,
  [5635] = 5635,
  [5636] = 5636,
  [5637] = 5562,
  [5638] = 5559,
  [5639] = 5557,
  [5640] = 5640,
  [5641] = 5559,
  [5642] = 5559,
  [5643] = 5557,
  [5644] = 5557,
  [5645] = 5645,
  [5646] = 5559,
  [5647] = 5557,
  [5648] = 5648,
  [5649] = 5649,
  [5650] = 5650,
  [5651] = 5559,
  [5652] = 5557,
  [5653] = 5588,
  [5654] = 5654,
  [5655] = 5559,
  [5656] = 5557,
  [5657] = 5559,
  [5658] = 5557,
  [5659] = 5559,
  [5660] = 5557,
  [5661] = 5559,
  [5662] = 5557,
  [5663] = 5559,
  [5664] = 5557,
  [5665] = 5559,
  [5666] = 5557,
  [5667] = 5589,
  [5668] = 5559,
  [5669] = 5596,
  [5670] = 5559,
  [5671] = 5671,
  [5672] = 5672,
  [5673] = 5557,
  [5674] = 5602,
  [5675] = 5022,
  [5676] = 5612,
  [5677] = 5677,
  [5678] = 5557,
  [5679] = 5679,
  [5680] = 5680,
  [5681] = 5559,
  [5682] = 5682,
  [5683] = 5683,
  [5684] = 5557,
  [5685] = 5152,
  [5686] = 5557,
  [5687] = 5687,
  [5688] = 5589,
  [5689] = 5559,
  [5690] = 5690,
  [5691] = 5691,
  [5692] = 5692,
  [5693] = 5559,
  [5694] = 5694,
  [5695] = 5695,
  [5696] = 5696,
  [5697] = 5697,
  [5698] = 5698,
  [5699] = 5699,
  [5700] = 5588,
  [5701] = 5701,
  [5702] = 5702,
  [5703] = 5640,
  [5704] = 5704,
  [5705] = 5705,
  [5706] = 5557,
  [5707] = 5152,
  [5708] = 5708,
  [5709] = 5709,
  [5710] = 5589,
  [5711] = 5711,
  [5712] = 5712,
  [5713] = 5559,
  [5714] = 5714,
  [5715] = 5596,
  [5716] = 5716,
  [5717] = 5717,
  [5718] = 5559,
  [5719] = 5562,
  [5720] = 5557,
  [5721] = 5721,
  [5722] = 5722,
  [5723] = 5723,
  [5724] = 5602,
  [5725] = 5725,
  [5726] = 5726,
  [5727] = 5727,
  [5728] = 5728,
  [5729] = 5612,
  [5730] = 5730,
  [5731] = 5557,
  [5732] = 5732,
  [5733] = 5733,
  [5734] = 5559,
  [5735] = 5557,
  [5736] = 5650,
  [5737] = 5737,
  [5738] = 5738,
  [5739] = 5739,
  [5740] = 5740,
  [5741] = 5672,
  [5742] = 5742,
  [5743] = 5743,
  [5744] = 5744,
  [5745] = 5745,
  [5746] = 5746,
  [5747] = 5747,
  [5748] = 5748,
  [5749] = 5589,
  [5750] = 5750,
  [5751] = 871,
  [5752] = 5739,
  [5753] = 5559,
  [5754] = 5559,
  [5755] = 5755,
  [5756] = 5756,
  [5757] = 5022,
  [5758] = 5557,
  [5759] = 5574,
  [5760] = 5760,
  [5761] = 5761,
  [5762] = 5762,
  [5763] = 885,
  [5764] = 5764,
  [5765] = 5765,
  [5766] = 5766,
  [5767] = 5767,
  [5768] = 5768,
  [5769] = 5769,
  [5770] = 5596,
  [5771] = 5559,
  [5772] = 5772,
  [5773] = 5773,
  [5774] = 5774,
  [5775] = 5775,
  [5776] = 5557,
  [5777] = 5777,
  [5778] = 5778,
  [5779] = 5779,
  [5780] = 5780,
  [5781] = 5781,
  [5782] = 5782,
  [5783] = 5559,
  [5784] = 5559,
  [5785] = 5557,
  [5786] = 844,
  [5787] = 5787,
  [5788] = 1961,
  [5789] = 5789,
  [5790] = 5790,
  [5791] = 5791,
  [5792] = 5557,
  [5793] = 5793,
  [5794] = 5559,
  [5795] = 871,
  [5796] = 5557,
  [5797] = 5721,
  [5798] = 5798,
  [5799] = 5799,
  [5800] = 885,
  [5801] = 5801,
  [5802] = 5802,
  [5803] = 5803,
  [5804] = 844,
  [5805] = 5602,
  [5806] = 5559,
  [5807] = 5807,
  [5808] = 5808,
  [5809] = 5578,
  [5810] = 5810,
  [5811] = 5506,
  [5812] = 5812,
  [5813] = 5596,
  [5814] = 5557,
  [5815] = 5815,
  [5816] = 5816,
  [5817] = 5612,
  [5818] = 5557,
  [5819] = 871,
  [5820] = 5559,
  [5821] = 885,
  [5822] = 5557,
  [5823] = 844,
  [5824] = 5824,
  [5825] = 5825,
  [5826] = 5826,
  [5827] = 5827,
  [5828] = 5007,
  [5829] = 5829,
  [5830] = 5830,
  [5831] = 5831,
  [5832] = 5832,
  [5833] = 5833,
  [5834] = 1707,
  [5835] = 5835,
  [5836] = 5781,
  [5837] = 5837,
  [5838] = 5602,
  [5839] = 5799,
  [5840] = 5840,
  [5841] = 5559,
  [5842] = 5559,
  [5843] = 1856,
  [5844] = 5708,
  [5845] = 5845,
  [5846] = 5309,
  [5847] = 5737,
  [5848] = 5848,
  [5849] = 5557,
  [5850] = 5829,
  [5851] = 5833,
  [5852] = 5852,
  [5853] = 5574,
  [5854] = 5626,
  [5855] = 5855,
  [5856] = 5856,
  [5857] = 5562,
  [5858] = 5557,
  [5859] = 5859,
  [5860] = 5578,
  [5861] = 5827,
  [5862] = 5832,
  [5863] = 5186,
  [5864] = 5837,
  [5865] = 5865,
  [5866] = 5855,
  [5867] = 5867,
  [5868] = 1682,
  [5869] = 5559,
  [5870] = 5867,
  [5871] = 5680,
  [5872] = 5588,
  [5873] = 5640,
  [5874] = 5874,
  [5875] = 5716,
  [5876] = 5557,
  [5877] = 5725,
  [5878] = 5727,
  [5879] = 5728,
  [5880] = 5880,
  [5881] = 5650,
  [5882] = 5882,
  [5883] = 5589,
  [5884] = 5559,
  [5885] = 5835,
  [5886] = 5845,
  [5887] = 5874,
  [5888] = 5596,
  [5889] = 5889,
  [5890] = 5732,
  [5891] = 5781,
  [5892] = 5602,
  [5893] = 5835,
  [5894] = 5845,
  [5895] = 5874,
  [5896] = 5896,
  [5897] = 5856,
  [5898] = 5732,
  [5899] = 5865,
  [5900] = 5515,
  [5901] = 5708,
  [5902] = 5612,
  [5903] = 5254,
  [5904] = 5557,
  [5905] = 5687,
  [5906] = 5799,
  [5907] = 5612,
  [5908] = 5908,
  [5909] = 5708,
  [5910] = 5910,
  [5911] = 5775,
  [5912] = 5737,
  [5913] = 5829,
  [5914] = 5833,
  [5915] = 5559,
  [5916] = 5574,
  [5917] = 5626,
  [5918] = 5249,
  [5919] = 5562,
  [5920] = 5557,
  [5921] = 5578,
  [5922] = 5827,
  [5923] = 5832,
  [5924] = 1786,
  [5925] = 5837,
  [5926] = 5506,
  [5927] = 5927,
  [5928] = 5855,
  [5929] = 5867,
  [5930] = 5930,
  [5931] = 5931,
  [5932] = 5680,
  [5933] = 5588,
  [5934] = 5640,
  [5935] = 5935,
  [5936] = 5716,
  [5937] = 5559,
  [5938] = 5725,
  [5939] = 5727,
  [5940] = 5728,
  [5941] = 5650,
  [5942] = 5589,
  [5943] = 5559,
  [5944] = 5944,
  [5945] = 5596,
  [5946] = 5781,
  [5947] = 5602,
  [5948] = 5856,
  [5949] = 5865,
  [5950] = 5950,
  [5951] = 5612,
  [5952] = 5557,
  [5953] = 5687,
  [5954] = 5799,
  [5955] = 5672,
  [5956] = 5672,
  [5957] = 5672,
  [5958] = 5672,
  [5959] = 5672,
  [5960] = 5672,
  [5961] = 5708,
  [5962] = 5672,
  [5963] = 5672,
  [5964] = 5672,
  [5965] = 5672,
  [5966] = 5672,
  [5967] = 5672,
  [5968] = 5672,
  [5969] = 5672,
  [5970] = 5672,
  [5971] = 5672,
  [5972] = 5672,
  [5973] = 5672,
  [5974] = 5672,
  [5975] = 5672,
  [5976] = 5672,
  [5977] = 5672,
  [5978] = 5672,
  [5979] = 5672,
  [5980] = 5672,
  [5981] = 5672,
  [5982] = 5672,
  [5983] = 5672,
  [5984] = 5672,
  [5985] = 5672,
  [5986] = 5672,
  [5987] = 5672,
  [5988] = 5672,
  [5989] = 5672,
  [5990] = 5672,
  [5991] = 5672,
  [5992] = 5672,
  [5993] = 5672,
  [5994] = 5672,
  [5995] = 5672,
  [5996] = 5672,
  [5997] = 5672,
  [5998] = 5672,
  [5999] = 5672,
  [6000] = 5672,
  [6001] = 5672,
  [6002] = 5672,
  [6003] = 5672,
  [6004] = 5672,
  [6005] = 6005,
  [6006] = 6006,
  [6007] = 6007,
  [6008] = 6008,
  [6009] = 6009,
  [6010] = 6010,
  [6011] = 6011,
  [6012] = 6009,
  [6013] = 5777,
  [6014] = 6014,
  [6015] = 875,
  [6016] = 6014,
  [6017] = 6017,
  [6018] = 6018,
  [6019] = 6019,
  [6020] = 6020,
  [6021] = 6021,
  [6022] = 6022,
  [6023] = 6023,
  [6024] = 6024,
  [6025] = 6025,
  [6026] = 6026,
  [6027] = 6027,
  [6028] = 6028,
  [6029] = 6029,
  [6030] = 6030,
  [6031] = 877,
  [6032] = 6032,
  [6033] = 5882,
  [6034] = 6009,
  [6035] = 6035,
  [6036] = 6028,
  [6037] = 6037,
  [6038] = 5022,
  [6039] = 6028,
  [6040] = 6018,
  [6041] = 6041,
  [6042] = 6042,
  [6043] = 6043,
  [6044] = 6044,
  [6045] = 6045,
  [6046] = 6046,
  [6047] = 6014,
  [6048] = 6048,
  [6049] = 6049,
  [6050] = 6050,
  [6051] = 6051,
  [6052] = 6052,
  [6053] = 6053,
  [6054] = 6054,
  [6055] = 6055,
  [6056] = 6056,
  [6057] = 6057,
  [6058] = 6058,
  [6059] = 6056,
  [6060] = 6060,
  [6061] = 6061,
  [6062] = 6028,
  [6063] = 6028,
  [6064] = 6064,
  [6065] = 6065,
  [6066] = 6061,
  [6067] = 6067,
  [6068] = 6068,
  [6069] = 6069,
  [6070] = 6070,
  [6071] = 6071,
  [6072] = 6072,
  [6073] = 6032,
  [6074] = 6074,
  [6075] = 6009,
  [6076] = 6028,
  [6077] = 6077,
  [6078] = 6007,
  [6079] = 6014,
  [6080] = 6080,
  [6081] = 878,
  [6082] = 6082,
  [6083] = 6028,
  [6084] = 6084,
  [6085] = 6085,
  [6086] = 6086,
  [6087] = 6032,
  [6088] = 6017,
  [6089] = 6019,
  [6090] = 6032,
  [6091] = 6009,
  [6092] = 6009,
  [6093] = 6026,
  [6094] = 6027,
  [6095] = 879,
  [6096] = 6096,
  [6097] = 6014,
  [6098] = 6098,
  [6099] = 2998,
  [6100] = 6014,
  [6101] = 6037,
  [6102] = 6028,
  [6103] = 6018,
  [6104] = 6041,
  [6105] = 6042,
  [6106] = 6106,
  [6107] = 6044,
  [6108] = 6108,
  [6109] = 6046,
  [6110] = 6110,
  [6111] = 6032,
  [6112] = 6009,
  [6113] = 6032,
  [6114] = 6114,
  [6115] = 6115,
  [6116] = 6064,
  [6117] = 6057,
  [6118] = 6065,
  [6119] = 6014,
  [6120] = 6056,
  [6121] = 6009,
  [6122] = 6068,
  [6123] = 6028,
  [6124] = 6070,
  [6125] = 6071,
  [6126] = 6126,
  [6127] = 6007,
  [6128] = 6028,
  [6129] = 6129,
  [6130] = 6032,
  [6131] = 6131,
  [6132] = 6017,
  [6133] = 6014,
  [6134] = 6134,
  [6135] = 6026,
  [6136] = 6027,
  [6137] = 6032,
  [6138] = 6138,
  [6139] = 6037,
  [6140] = 6009,
  [6141] = 6032,
  [6142] = 6044,
  [6143] = 6009,
  [6144] = 6046,
  [6145] = 6014,
  [6146] = 6146,
  [6147] = 6014,
  [6148] = 6148,
  [6149] = 5152,
  [6150] = 6056,
  [6151] = 6007,
  [6152] = 6152,
  [6153] = 6153,
  [6154] = 6009,
  [6155] = 6014,
  [6156] = 6156,
  [6157] = 6026,
  [6158] = 6032,
  [6159] = 6068,
  [6160] = 6160,
  [6161] = 6009,
  [6162] = 6162,
  [6163] = 6163,
  [6164] = 6032,
  [6165] = 6009,
  [6166] = 6166,
  [6167] = 6167,
  [6168] = 6028,
  [6169] = 6169,
  [6170] = 6014,
  [6171] = 6171,
  [6172] = 6172,
  [6173] = 6032,
  [6174] = 6032,
  [6175] = 6009,
  [6176] = 6009,
  [6177] = 6177,
  [6178] = 4910,
  [6179] = 886,
  [6180] = 6180,
  [6181] = 6014,
  [6182] = 878,
  [6183] = 6014,
  [6184] = 6184,
  [6185] = 6028,
  [6186] = 5572,
  [6187] = 6187,
  [6188] = 6188,
  [6189] = 6189,
  [6190] = 6070,
  [6191] = 6191,
  [6192] = 6192,
  [6193] = 6071,
  [6194] = 6032,
  [6195] = 6009,
  [6196] = 6196,
  [6197] = 6197,
  [6198] = 6014,
  [6199] = 6199,
  [6200] = 6026,
  [6201] = 6027,
  [6202] = 880,
  [6203] = 5504,
  [6204] = 879,
  [6205] = 6028,
  [6206] = 6206,
  [6207] = 6207,
  [6208] = 6208,
  [6209] = 6032,
  [6210] = 6009,
  [6211] = 6032,
  [6212] = 6212,
  [6213] = 5572,
  [6214] = 6032,
  [6215] = 6014,
  [6216] = 6216,
  [6217] = 6009,
  [6218] = 6009,
  [6219] = 6219,
  [6220] = 6220,
  [6221] = 6032,
  [6222] = 6222,
  [6223] = 6009,
  [6224] = 3003,
  [6225] = 6044,
  [6226] = 6226,
  [6227] = 6014,
  [6228] = 6228,
  [6229] = 6014,
  [6230] = 6014,
  [6231] = 6231,
  [6232] = 6014,
  [6233] = 6233,
  [6234] = 6234,
  [6235] = 6235,
  [6236] = 6236,
  [6237] = 6237,
  [6238] = 6238,
  [6239] = 881,
  [6240] = 6032,
  [6241] = 6009,
  [6242] = 6028,
  [6243] = 6243,
  [6244] = 6014,
  [6245] = 6245,
  [6246] = 6032,
  [6247] = 6009,
  [6248] = 6032,
  [6249] = 6046,
  [6250] = 6250,
  [6251] = 6009,
  [6252] = 6014,
  [6253] = 6253,
  [6254] = 6014,
  [6255] = 6014,
  [6256] = 6256,
  [6257] = 6032,
  [6258] = 6258,
  [6259] = 6009,
  [6260] = 6028,
  [6261] = 6057,
  [6262] = 6014,
  [6263] = 6263,
  [6264] = 6264,
  [6265] = 6028,
  [6266] = 6049,
  [6267] = 6267,
  [6268] = 6268,
  [6269] = 6269,
  [6270] = 6270,
  [6271] = 6058,
  [6272] = 6032,
  [6273] = 6009,
  [6274] = 6274,
  [6275] = 6275,
  [6276] = 6276,
  [6277] = 6277,
  [6278] = 6032,
  [6279] = 6014,
  [6280] = 6280,
  [6281] = 6281,
  [6282] = 6009,
  [6283] = 6283,
  [6284] = 6284,
  [6285] = 6285,
  [6286] = 6286,
  [6287] = 5022,
  [6288] = 6288,
  [6289] = 6032,
  [6290] = 6290,
  [6291] = 6032,
  [6292] = 6009,
  [6293] = 6293,
  [6294] = 6294,
  [6295] = 6295,
  [6296] = 6029,
  [6297] = 6297,
  [6298] = 6030,
  [6299] = 6299,
  [6300] = 6014,
  [6301] = 6032,
  [6302] = 6014,
  [6303] = 6019,
  [6304] = 6009,
  [6305] = 6009,
  [6306] = 6306,
  [6307] = 6307,
  [6308] = 6308,
  [6309] = 6309,
  [6310] = 6050,
  [6311] = 6311,
  [6312] = 6028,
  [6313] = 6032,
  [6314] = 6009,
  [6315] = 6006,
  [6316] = 6316,
  [6317] = 6317,
  [6318] = 6037,
  [6319] = 6028,
  [6320] = 6014,
  [6321] = 6245,
  [6322] = 6058,
  [6323] = 6032,
  [6324] = 6324,
  [6325] = 6009,
  [6326] = 6014,
  [6327] = 6032,
  [6328] = 871,
  [6329] = 6009,
  [6330] = 6014,
  [6331] = 6032,
  [6332] = 6009,
  [6333] = 6014,
  [6334] = 6014,
  [6335] = 6226,
  [6336] = 885,
  [6337] = 6337,
  [6338] = 6338,
  [6339] = 6339,
  [6340] = 6074,
  [6341] = 6032,
  [6342] = 6342,
  [6343] = 6009,
  [6344] = 6344,
  [6345] = 6028,
  [6346] = 6346,
  [6347] = 6347,
  [6348] = 6348,
  [6349] = 6349,
  [6350] = 6053,
  [6351] = 6351,
  [6352] = 6226,
  [6353] = 6032,
  [6354] = 6009,
  [6355] = 6355,
  [6356] = 6238,
  [6357] = 6014,
  [6358] = 6032,
  [6359] = 6014,
  [6360] = 6009,
  [6361] = 6028,
  [6362] = 6362,
  [6363] = 6032,
  [6364] = 6032,
  [6365] = 6009,
  [6366] = 6014,
  [6367] = 6009,
  [6368] = 6014,
  [6369] = 6369,
  [6370] = 6032,
  [6371] = 6009,
  [6372] = 6372,
  [6373] = 6080,
  [6374] = 6374,
  [6375] = 6172,
  [6376] = 6309,
  [6377] = 6377,
  [6378] = 6115,
  [6379] = 6231,
  [6380] = 6308,
  [6381] = 6189,
  [6382] = 6216,
  [6383] = 6228,
  [6384] = 6294,
  [6385] = 6297,
  [6386] = 6386,
  [6387] = 6020,
  [6388] = 6362,
  [6389] = 6250,
  [6390] = 6275,
  [6391] = 6032,
  [6392] = 6342,
  [6393] = 6351,
  [6394] = 6009,
  [6395] = 6028,
  [6396] = 6051,
  [6397] = 6268,
  [6398] = 6028,
  [6399] = 6399,
  [6400] = 6108,
  [6401] = 6014,
  [6402] = 6199,
  [6403] = 6080,
  [6404] = 6374,
  [6405] = 6172,
  [6406] = 6309,
  [6407] = 6399,
  [6408] = 6377,
  [6409] = 6115,
  [6410] = 6231,
  [6411] = 6308,
  [6412] = 6189,
  [6413] = 6032,
  [6414] = 6216,
  [6415] = 6228,
  [6416] = 6294,
  [6417] = 6297,
  [6418] = 6009,
  [6419] = 6020,
  [6420] = 6362,
  [6421] = 6250,
  [6422] = 6275,
  [6423] = 6032,
  [6424] = 6342,
  [6425] = 6351,
  [6426] = 6009,
  [6427] = 6051,
  [6428] = 6268,
  [6429] = 6007,
  [6430] = 6399,
  [6431] = 6014,
  [6432] = 6108,
  [6433] = 6014,
  [6434] = 6199,
  [6435] = 6080,
  [6436] = 6374,
  [6437] = 6172,
  [6438] = 6309,
  [6439] = 6377,
  [6440] = 6115,
  [6441] = 6231,
  [6442] = 6014,
  [6443] = 6308,
  [6444] = 6189,
  [6445] = 6216,
  [6446] = 6228,
  [6447] = 6020,
  [6448] = 6032,
  [6449] = 6362,
  [6450] = 6250,
  [6451] = 6051,
  [6452] = 6268,
  [6453] = 6032,
  [6454] = 6399,
  [6455] = 6009,
  [6456] = 6199,
  [6457] = 6009,
  [6458] = 6374,
  [6459] = 6172,
  [6460] = 6377,
  [6461] = 6231,
  [6462] = 6189,
  [6463] = 6463,
  [6464] = 6216,
  [6465] = 6250,
  [6466] = 6199,
  [6467] = 6374,
  [6468] = 6172,
  [6469] = 6374,
  [6470] = 6470,
  [6471] = 6243,
  [6472] = 6014,
  [6473] = 6473,
  [6474] = 6299,
  [6475] = 6338,
  [6476] = 6355,
  [6477] = 6477,
  [6478] = 6478,
  [6479] = 6041,
  [6480] = 6470,
  [6481] = 6317,
  [6482] = 6243,
  [6483] = 6032,
  [6484] = 6028,
  [6485] = 6299,
  [6486] = 6245,
  [6487] = 6058,
  [6488] = 6338,
  [6489] = 6355,
  [6490] = 6009,
  [6491] = 6470,
  [6492] = 6028,
  [6493] = 6470,
  [6494] = 6032,
  [6495] = 6243,
  [6496] = 6496,
  [6497] = 6338,
  [6498] = 6014,
  [6499] = 6009,
  [6500] = 6032,
  [6501] = 6009,
  [6502] = 6502,
  [6503] = 6028,
  [6504] = 6226,
  [6505] = 6032,
  [6506] = 6009,
  [6507] = 6014,
  [6508] = 844,
  [6509] = 6014,
  [6510] = 5512,
  [6511] = 880,
  [6512] = 6032,
  [6513] = 6032,
  [6514] = 6009,
  [6515] = 6014,
  [6516] = 6028,
  [6517] = 6517,
  [6518] = 6518,
  [6519] = 6032,
  [6520] = 6374,
  [6521] = 6014,
  [6522] = 6522,
  [6523] = 6032,
  [6524] = 6009,
  [6525] = 6032,
  [6526] = 6014,
  [6527] = 5802,
  [6528] = 6032,
  [6529] = 6014,
  [6530] = 877,
  [6531] = 6032,
  [6532] = 6532,
  [6533] = 881,
  [6534] = 6534,
  [6535] = 6535,
  [6536] = 6536,
  [6537] = 6028,
  [6538] = 6538,
  [6539] = 6017,
  [6540] = 6317,
  [6541] = 6541,
  [6542] = 6542,
  [6543] = 6543,
  [6544] = 6017,
  [6545] = 6019,
  [6546] = 6042,
  [6547] = 6369,
  [6548] = 6032,
  [6549] = 6369,
  [6550] = 875,
  [6551] = 6022,
  [6552] = 6369,
  [6553] = 6009,
  [6554] = 6369,
  [6555] = 6317,
  [6556] = 6028,
  [6557] = 6369,
  [6558] = 6026,
  [6559] = 6027,
  [6560] = 6029,
  [6561] = 6030,
  [6562] = 6028,
  [6563] = 6037,
  [6564] = 6014,
  [6565] = 6018,
  [6566] = 6041,
  [6567] = 6042,
  [6568] = 6568,
  [6569] = 6569,
  [6570] = 6570,
  [6571] = 6058,
  [6572] = 6044,
  [6573] = 6573,
  [6574] = 6046,
  [6575] = 6575,
  [6576] = 6049,
  [6577] = 6050,
  [6578] = 6006,
  [6579] = 6579,
  [6580] = 6053,
  [6581] = 6022,
  [6582] = 6582,
  [6583] = 6028,
  [6584] = 6057,
  [6585] = 6056,
  [6586] = 6586,
  [6587] = 6061,
  [6588] = 6032,
  [6589] = 6589,
  [6590] = 6590,
  [6591] = 6064,
  [6592] = 6065,
  [6593] = 6068,
  [6594] = 6070,
  [6595] = 6071,
  [6596] = 6596,
  [6597] = 6377,
  [6598] = 6598,
  [6599] = 6599,
  [6600] = 6600,
  [6601] = 6074,
  [6602] = 6032,
  [6603] = 6603,
  [6604] = 6028,
  [6605] = 6245,
  [6606] = 6058,
  [6607] = 6014,
  [6608] = 6608,
  [6609] = 6609,
  [6610] = 6610,
  [6611] = 6611,
  [6612] = 6612,
  [6613] = 6613,
  [6614] = 6611,
  [6615] = 5778,
  [6616] = 6616,
  [6617] = 6617,
  [6618] = 6618,
  [6619] = 6613,
  [6620] = 6620,
  [6621] = 6621,
  [6622] = 6622,
  [6623] = 6623,
  [6624] = 6624,
  [6625] = 6625,
  [6626] = 6613,
  [6627] = 6609,
  [6628] = 6616,
  [6629] = 6617,
  [6630] = 6611,
  [6631] = 6618,
  [6632] = 6613,
  [6633] = 6622,
  [6634] = 6620,
  [6635] = 6616,
  [6636] = 6617,
  [6637] = 6618,
  [6638] = 6638,
  [6639] = 6620,
  [6640] = 6622,
  [6641] = 6622,
  [6642] = 6642,
  [6643] = 6609,
  [6644] = 6644,
  [6645] = 6645,
  [6646] = 6613,
  [6647] = 6647,
  [6648] = 6648,
  [6649] = 6620,
  [6650] = 6650,
  [6651] = 6622,
  [6652] = 6613,
  [6653] = 6609,
  [6654] = 6654,
  [6655] = 6655,
  [6656] = 6613,
  [6657] = 6657,
  [6658] = 6658,
  [6659] = 6620,
  [6660] = 886,
  [6661] = 6622,
  [6662] = 6609,
  [6663] = 6663,
  [6664] = 6609,
  [6665] = 6610,
  [6666] = 6609,
  [6667] = 6616,
  [6668] = 6611,
  [6669] = 6617,
  [6670] = 6609,
  [6671] = 6613,
  [6672] = 6618,
  [6673] = 6654,
  [6674] = 6609,
  [6675] = 6609,
  [6676] = 6609,
  [6677] = 6609,
  [6678] = 6609,
  [6679] = 6609,
  [6680] = 6609,
  [6681] = 6609,
  [6682] = 6609,
  [6683] = 6609,
  [6684] = 6609,
  [6685] = 6609,
  [6686] = 6609,
  [6687] = 6609,
  [6688] = 6609,
  [6689] = 6609,
  [6690] = 6609,
  [6691] = 6609,
  [6692] = 6609,
  [6693] = 6609,
  [6694] = 6609,
  [6695] = 6609,
  [6696] = 6609,
  [6697] = 6609,
  [6698] = 6609,
  [6699] = 6616,
  [6700] = 6617,
  [6701] = 6618,
  [6702] = 6620,
  [6703] = 6703,
  [6704] = 6620,
  [6705] = 6705,
  [6706] = 6706,
  [6707] = 6622,
  [6708] = 6708,
  [6709] = 6709,
  [6710] = 6710,
  [6711] = 6711,
  [6712] = 6712,
  [6713] = 6713,
  [6714] = 6714,
  [6715] = 6715,
  [6716] = 6716,
  [6717] = 6717,
  [6718] = 6718,
  [6719] = 6719,
  [6720] = 6654,
  [6721] = 6721,
  [6722] = 6722,
  [6723] = 6723,
  [6724] = 6724,
  [6725] = 6725,
  [6726] = 6726,
  [6727] = 6727,
  [6728] = 6728,
  [6729] = 6729,
  [6730] = 6663,
  [6731] = 6609,
  [6732] = 6732,
  [6733] = 5237,
  [6734] = 6734,
  [6735] = 6735,
  [6736] = 6736,
  [6737] = 6737,
  [6738] = 6738,
  [6739] = 6739,
  [6740] = 6740,
  [6741] = 2579,
  [6742] = 6611,
  [6743] = 6647,
  [6744] = 6744,
  [6745] = 6613,
  [6746] = 6622,
  [6747] = 6747,
  [6748] = 6616,
  [6749] = 6617,
  [6750] = 6750,
  [6751] = 6618,
  [6752] = 6752,
  [6753] = 6620,
  [6754] = 6754,
  [6755] = 6622,
  [6756] = 6756,
  [6757] = 6757,
  [6758] = 6758,
  [6759] = 6759,
  [6760] = 6760,
  [6761] = 6759,
  [6762] = 6762,
  [6763] = 6763,
  [6764] = 6764,
  [6765] = 6765,
  [6766] = 6766,
  [6767] = 6267,
  [6768] = 6768,
  [6769] = 6654,
  [6770] = 6770,
  [6771] = 6771,
  [6772] = 6772,
  [6773] = 6773,
  [6774] = 6774,
  [6775] = 6775,
  [6776] = 6776,
  [6777] = 6777,
  [6778] = 6778,
  [6779] = 6779,
  [6780] = 6780,
  [6781] = 6781,
  [6782] = 6782,
  [6783] = 6783,
  [6784] = 6784,
  [6785] = 6663,
  [6786] = 6609,
  [6787] = 6787,
  [6788] = 6663,
  [6789] = 6654,
  [6790] = 6611,
  [6791] = 6740,
  [6792] = 5746,
  [6793] = 6270,
  [6794] = 6794,
  [6795] = 6613,
  [6796] = 6796,
  [6797] = 6797,
  [6798] = 6609,
  [6799] = 6799,
  [6800] = 6616,
  [6801] = 6617,
  [6802] = 6802,
  [6803] = 6803,
  [6804] = 6618,
  [6805] = 6805,
  [6806] = 6806,
  [6807] = 6620,
  [6808] = 6808,
  [6809] = 6622,
  [6810] = 6810,
  [6811] = 6811,
  [6812] = 6663,
  [6813] = 6813,
  [6814] = 6814,
  [6815] = 6815,
  [6816] = 6609,
  [6817] = 6817,
  [6818] = 6818,
  [6819] = 6611,
  [6820] = 6820,
  [6821] = 6821,
  [6822] = 6762,
  [6823] = 6654,
  [6824] = 6824,
  [6825] = 6736,
  [6826] = 6826,
  [6827] = 6610,
  [6828] = 6828,
  [6829] = 6829,
  [6830] = 6830,
  [6831] = 6831,
  [6832] = 6706,
  [6833] = 6069,
  [6834] = 6834,
  [6835] = 6663,
  [6836] = 6712,
  [6837] = 6713,
  [6838] = 6838,
  [6839] = 6609,
  [6840] = 6611,
  [6841] = 6841,
  [6842] = 6611,
  [6843] = 6843,
  [6844] = 6613,
  [6845] = 6613,
  [6846] = 6610,
  [6847] = 6725,
  [6848] = 6726,
  [6849] = 6616,
  [6850] = 6617,
  [6851] = 6618,
  [6852] = 6852,
  [6853] = 6620,
  [6854] = 6780,
  [6855] = 6616,
  [6856] = 6622,
  [6857] = 6857,
  [6858] = 6617,
  [6859] = 6859,
  [6860] = 6618,
  [6861] = 6739,
  [6862] = 6862,
  [6863] = 6620,
  [6864] = 6654,
  [6865] = 6865,
  [6866] = 6866,
  [6867] = 6867,
  [6868] = 6868,
  [6869] = 6869,
  [6870] = 6706,
  [6871] = 6663,
  [6872] = 6609,
  [6873] = 6710,
  [6874] = 6874,
  [6875] = 6875,
  [6876] = 6875,
  [6877] = 6611,
  [6878] = 6622,
  [6879] = 6613,
  [6880] = 6611,
  [6881] = 6881,
  [6882] = 6882,
  [6883] = 6616,
  [6884] = 6617,
  [6885] = 6885,
  [6886] = 5022,
  [6887] = 6654,
  [6888] = 6618,
  [6889] = 6889,
  [6890] = 6890,
  [6891] = 6862,
  [6892] = 6620,
  [6893] = 6893,
  [6894] = 6894,
  [6895] = 6895,
  [6896] = 6777,
  [6897] = 6778,
  [6898] = 6780,
  [6899] = 6622,
  [6900] = 6900,
  [6901] = 3017,
  [6902] = 6902,
  [6903] = 6889,
  [6904] = 6904,
  [6905] = 6654,
  [6906] = 6770,
  [6907] = 6907,
  [6908] = 6740,
  [6909] = 6909,
  [6910] = 6654,
  [6911] = 6911,
  [6912] = 6781,
  [6913] = 6913,
  [6914] = 6914,
  [6915] = 6663,
  [6916] = 6916,
  [6917] = 6917,
  [6918] = 6803,
  [6919] = 6609,
  [6920] = 6893,
  [6921] = 6921,
  [6922] = 6922,
  [6923] = 6663,
  [6924] = 6611,
  [6925] = 6233,
  [6926] = 6609,
  [6927] = 6613,
  [6928] = 6862,
  [6929] = 6613,
  [6930] = 6616,
  [6931] = 6617,
  [6932] = 6932,
  [6933] = 6618,
  [6934] = 6900,
  [6935] = 6935,
  [6936] = 6620,
  [6937] = 874,
  [6938] = 6611,
  [6939] = 6622,
  [6940] = 6940,
  [6941] = 6706,
  [6942] = 6942,
  [6943] = 6799,
  [6944] = 6712,
  [6945] = 6713,
  [6946] = 6946,
  [6947] = 6947,
  [6948] = 6948,
  [6949] = 6949,
  [6950] = 6950,
  [6951] = 6613,
  [6952] = 6952,
  [6953] = 6654,
  [6954] = 6724,
  [6955] = 5556,
  [6956] = 6808,
  [6957] = 6616,
  [6958] = 6663,
  [6959] = 6608,
  [6960] = 6609,
  [6961] = 6617,
  [6962] = 6739,
  [6963] = 6618,
  [6964] = 6611,
  [6965] = 6965,
  [6966] = 6818,
  [6967] = 6613,
  [6968] = 6620,
  [6969] = 6969,
  [6970] = 6736,
  [6971] = 6616,
  [6972] = 6617,
  [6973] = 6618,
  [6974] = 6622,
  [6975] = 6620,
  [6976] = 6177,
  [6977] = 6977,
  [6978] = 6622,
  [6979] = 6805,
  [6980] = 6980,
  [6981] = 6981,
  [6982] = 6264,
  [6983] = 6983,
  [6984] = 6984,
  [6985] = 6985,
  [6986] = 6986,
  [6987] = 6987,
  [6988] = 6988,
  [6989] = 6875,
  [6990] = 6777,
  [6991] = 6780,
  [6992] = 6654,
  [6993] = 6993,
  [6994] = 6994,
  [6995] = 6995,
  [6996] = 6996,
  [6997] = 6997,
  [6998] = 6998,
  [6999] = 6663,
  [7000] = 7000,
  [7001] = 6609,
  [7002] = 6803,
  [7003] = 7003,
  [7004] = 6654,
  [7005] = 6740,
  [7006] = 6611,
  [7007] = 6996,
  [7008] = 6621,
  [7009] = 7009,
  [7010] = 6706,
  [7011] = 6613,
  [7012] = 6889,
  [7013] = 6712,
  [7014] = 6713,
  [7015] = 6712,
  [7016] = 6794,
  [7017] = 6911,
  [7018] = 6616,
  [7019] = 6617,
  [7020] = 6618,
  [7021] = 7021,
  [7022] = 7022,
  [7023] = 6620,
  [7024] = 7024,
  [7025] = 6739,
  [7026] = 6663,
  [7027] = 6622,
  [7028] = 6609,
  [7029] = 7029,
  [7030] = 7030,
  [7031] = 7031,
  [7032] = 6657,
  [7033] = 6657,
  [7034] = 7034,
  [7035] = 6654,
  [7036] = 7036,
  [7037] = 5802,
  [7038] = 7038,
  [7039] = 7039,
  [7040] = 7040,
  [7041] = 6777,
  [7042] = 6780,
  [7043] = 5803,
  [7044] = 6713,
  [7045] = 6803,
  [7046] = 6611,
  [7047] = 6725,
  [7048] = 6663,
  [7049] = 6706,
  [7050] = 6609,
  [7051] = 6712,
  [7052] = 6713,
  [7053] = 6616,
  [7054] = 6023,
  [7055] = 6611,
  [7056] = 6613,
  [7057] = 7057,
  [7058] = 5694,
  [7059] = 6035,
  [7060] = 6726,
  [7061] = 6739,
  [7062] = 6657,
  [7063] = 6613,
  [7064] = 7064,
  [7065] = 6617,
  [7066] = 7066,
  [7067] = 7067,
  [7068] = 7068,
  [7069] = 6796,
  [7070] = 6616,
  [7071] = 5501,
  [7072] = 6618,
  [7073] = 6780,
  [7074] = 6617,
  [7075] = 6803,
  [7076] = 6616,
  [7077] = 6706,
  [7078] = 6618,
  [7079] = 6712,
  [7080] = 6713,
  [7081] = 7081,
  [7082] = 6617,
  [7083] = 6618,
  [7084] = 6739,
  [7085] = 6616,
  [7086] = 6620,
  [7087] = 5152,
  [7088] = 6617,
  [7089] = 6622,
  [7090] = 7090,
  [7091] = 7091,
  [7092] = 7092,
  [7093] = 7093,
  [7094] = 6620,
  [7095] = 6618,
  [7096] = 6803,
  [7097] = 7097,
  [7098] = 6706,
  [7099] = 6882,
  [7100] = 6712,
  [7101] = 6713,
  [7102] = 7102,
  [7103] = 6739,
  [7104] = 6620,
  [7105] = 6654,
  [7106] = 6620,
  [7107] = 6622,
  [7108] = 6747,
  [7109] = 7109,
  [7110] = 7110,
  [7111] = 7111,
  [7112] = 6706,
  [7113] = 7113,
  [7114] = 6712,
  [7115] = 6713,
  [7116] = 7116,
  [7117] = 6739,
  [7118] = 6663,
  [7119] = 7119,
  [7120] = 6609,
  [7121] = 6705,
  [7122] = 6710,
  [7123] = 6611,
  [7124] = 7124,
  [7125] = 6706,
  [7126] = 6613,
  [7127] = 6712,
  [7128] = 6713,
  [7129] = 6616,
  [7130] = 6739,
  [7131] = 6263,
  [7132] = 4616,
  [7133] = 7133,
  [7134] = 7134,
  [7135] = 6719,
  [7136] = 6293,
  [7137] = 6706,
  [7138] = 7138,
  [7139] = 6712,
  [7140] = 6713,
  [7141] = 6616,
  [7142] = 6739,
  [7143] = 6617,
  [7144] = 869,
  [7145] = 5730,
  [7146] = 7146,
  [7147] = 6618,
  [7148] = 6654,
  [7149] = 6706,
  [7150] = 7150,
  [7151] = 6712,
  [7152] = 6713,
  [7153] = 6620,
  [7154] = 6739,
  [7155] = 7155,
  [7156] = 6622,
  [7157] = 7157,
  [7158] = 7158,
  [7159] = 7102,
  [7160] = 7160,
  [7161] = 6706,
  [7162] = 6654,
  [7163] = 6712,
  [7164] = 6713,
  [7165] = 2593,
  [7166] = 6739,
  [7167] = 7167,
  [7168] = 7168,
  [7169] = 6648,
  [7170] = 5022,
  [7171] = 7171,
  [7172] = 7172,
  [7173] = 6706,
  [7174] = 7174,
  [7175] = 6712,
  [7176] = 6713,
  [7177] = 7177,
  [7178] = 6739,
  [7179] = 7179,
  [7180] = 6663,
  [7181] = 7181,
  [7182] = 6609,
  [7183] = 5761,
  [7184] = 7184,
  [7185] = 6706,
  [7186] = 7186,
  [7187] = 6712,
  [7188] = 6713,
  [7189] = 7189,
  [7190] = 6739,
  [7191] = 6663,
  [7192] = 6611,
  [7193] = 6609,
  [7194] = 6613,
  [7195] = 6617,
  [7196] = 6728,
  [7197] = 6706,
  [7198] = 6616,
  [7199] = 6712,
  [7200] = 6713,
  [7201] = 6617,
  [7202] = 6739,
  [7203] = 6618,
  [7204] = 6611,
  [7205] = 6705,
  [7206] = 5249,
  [7207] = 6620,
  [7208] = 6734,
  [7209] = 6706,
  [7210] = 6622,
  [7211] = 6712,
  [7212] = 6713,
  [7213] = 5186,
  [7214] = 6739,
  [7215] = 6613,
  [7216] = 6055,
  [7217] = 6654,
  [7218] = 6710,
  [7219] = 6114,
  [7220] = 7220,
  [7221] = 6706,
  [7222] = 7222,
  [7223] = 6712,
  [7224] = 6713,
  [7225] = 6618,
  [7226] = 6739,
  [7227] = 7227,
  [7228] = 6890,
  [7229] = 6663,
  [7230] = 6609,
  [7231] = 7124,
  [7232] = 7232,
  [7233] = 6706,
  [7234] = 6616,
  [7235] = 6712,
  [7236] = 6713,
  [7237] = 6617,
  [7238] = 6739,
  [7239] = 3024,
  [7240] = 7240,
  [7241] = 6638,
  [7242] = 7242,
  [7243] = 6611,
  [7244] = 6618,
  [7245] = 6706,
  [7246] = 7246,
  [7247] = 6712,
  [7248] = 6713,
  [7249] = 6613,
  [7250] = 6739,
  [7251] = 6622,
  [7252] = 6620,
  [7253] = 6616,
  [7254] = 6617,
  [7255] = 6618,
  [7256] = 7256,
  [7257] = 6706,
  [7258] = 6747,
  [7259] = 6712,
  [7260] = 6713,
  [7261] = 6620,
  [7262] = 6739,
  [7263] = 6622,
  [7264] = 6622,
  [7265] = 6752,
  [7266] = 6754,
  [7267] = 7267,
  [7268] = 5022,
  [7269] = 6706,
  [7270] = 7270,
  [7271] = 6712,
  [7272] = 6713,
  [7273] = 7273,
  [7274] = 6739,
  [7275] = 7275,
  [7276] = 7276,
  [7277] = 6882,
  [7278] = 6609,
  [7279] = 6759,
  [7280] = 7280,
  [7281] = 6706,
  [7282] = 6890,
  [7283] = 6712,
  [7284] = 6713,
  [7285] = 6654,
  [7286] = 6739,
  [7287] = 6188,
  [7288] = 6611,
  [7289] = 6206,
  [7290] = 7290,
  [7291] = 7030,
  [7292] = 6613,
  [7293] = 6706,
  [7294] = 6620,
  [7295] = 6712,
  [7296] = 6713,
  [7297] = 7146,
  [7298] = 6739,
  [7299] = 7299,
  [7300] = 6616,
  [7301] = 7301,
  [7302] = 7302,
  [7303] = 7303,
  [7304] = 6617,
  [7305] = 6706,
  [7306] = 7306,
  [7307] = 6712,
  [7308] = 6713,
  [7309] = 6618,
  [7310] = 6739,
  [7311] = 7311,
  [7312] = 6719,
  [7313] = 5768,
  [7314] = 6620,
  [7315] = 6663,
  [7316] = 6622,
  [7317] = 6706,
  [7318] = 6609,
  [7319] = 6712,
  [7320] = 6713,
  [7321] = 7321,
  [7322] = 6739,
  [7323] = 6543,
  [7324] = 7324,
  [7325] = 6611,
  [7326] = 7326,
  [7327] = 6609,
  [7328] = 6648,
  [7329] = 6706,
  [7330] = 6613,
  [7331] = 6712,
  [7332] = 6713,
  [7333] = 6611,
  [7334] = 6739,
  [7335] = 6705,
  [7336] = 7336,
  [7337] = 6613,
  [7338] = 6710,
  [7339] = 6616,
  [7340] = 6616,
  [7341] = 6706,
  [7342] = 7342,
  [7343] = 6712,
  [7344] = 6713,
  [7345] = 6617,
  [7346] = 6739,
  [7347] = 6618,
  [7348] = 6617,
  [7349] = 7349,
  [7350] = 886,
  [7351] = 6620,
  [7352] = 6618,
  [7353] = 6706,
  [7354] = 6622,
  [7355] = 6712,
  [7356] = 6713,
  [7357] = 6654,
  [7358] = 6739,
  [7359] = 7359,
  [7360] = 7360,
  [7361] = 7361,
  [7362] = 7362,
  [7363] = 7363,
  [7364] = 6609,
  [7365] = 6706,
  [7366] = 6620,
  [7367] = 6712,
  [7368] = 6713,
  [7369] = 7124,
  [7370] = 6739,
  [7371] = 6611,
  [7372] = 7372,
  [7373] = 7373,
  [7374] = 6622,
  [7375] = 7155,
  [7376] = 7376,
  [7377] = 6613,
  [7378] = 6712,
  [7379] = 6713,
  [7380] = 7380,
  [7381] = 6739,
  [7382] = 6740,
  [7383] = 6803,
  [7384] = 6616,
  [7385] = 6617,
  [7386] = 6618,
  [7387] = 6817,
  [7388] = 6620,
  [7389] = 6712,
  [7390] = 6713,
  [7391] = 7391,
  [7392] = 6739,
  [7393] = 7393,
  [7394] = 7394,
  [7395] = 6841,
  [7396] = 7160,
  [7397] = 6622,
  [7398] = 7398,
  [7399] = 6719,
  [7400] = 6712,
  [7401] = 6713,
  [7402] = 7186,
  [7403] = 6739,
  [7404] = 7404,
  [7405] = 7405,
  [7406] = 6654,
  [7407] = 7407,
  [7408] = 7408,
  [7409] = 6609,
  [7410] = 6663,
  [7411] = 6712,
  [7412] = 6713,
  [7413] = 7413,
  [7414] = 6739,
  [7415] = 7415,
  [7416] = 6611,
  [7417] = 6728,
  [7418] = 7418,
  [7419] = 7419,
  [7420] = 6609,
  [7421] = 6613,
  [7422] = 6712,
  [7423] = 6713,
  [7424] = 7424,
  [7425] = 6739,
  [7426] = 7426,
  [7427] = 7427,
  [7428] = 7428,
  [7429] = 6663,
  [7430] = 6609,
  [7431] = 7431,
  [7432] = 6616,
  [7433] = 6712,
  [7434] = 6713,
  [7435] = 6617,
  [7436] = 6739,
  [7437] = 6618,
  [7438] = 7438,
  [7439] = 7439,
  [7440] = 7155,
  [7441] = 6620,
  [7442] = 7442,
  [7443] = 7160,
  [7444] = 6712,
  [7445] = 6713,
  [7446] = 5730,
  [7447] = 6739,
  [7448] = 6638,
  [7449] = 6622,
  [7450] = 6648,
  [7451] = 6611,
  [7452] = 7452,
  [7453] = 7186,
  [7454] = 6654,
  [7455] = 6712,
  [7456] = 6713,
  [7457] = 7306,
  [7458] = 6739,
  [7459] = 6613,
  [7460] = 7460,
  [7461] = 7415,
  [7462] = 6711,
  [7463] = 7463,
  [7464] = 6609,
  [7465] = 6740,
  [7466] = 6712,
  [7467] = 6713,
  [7468] = 6611,
  [7469] = 6739,
  [7470] = 7470,
  [7471] = 7471,
  [7472] = 7472,
  [7473] = 7404,
  [7474] = 6799,
  [7475] = 7475,
  [7476] = 6663,
  [7477] = 6712,
  [7478] = 6713,
  [7479] = 6609,
  [7480] = 6739,
  [7481] = 6829,
  [7482] = 6611,
  [7483] = 7009,
  [7484] = 7484,
  [7485] = 6616,
  [7486] = 6613,
  [7487] = 7109,
  [7488] = 6712,
  [7489] = 6713,
  [7490] = 6617,
  [7491] = 6739,
  [7492] = 7492,
  [7493] = 7493,
  [7494] = 7326,
  [7495] = 7495,
  [7496] = 7496,
  [7497] = 7373,
  [7498] = 6618,
  [7499] = 6712,
  [7500] = 6713,
  [7501] = 6616,
  [7502] = 6739,
  [7503] = 6799,
  [7504] = 7504,
  [7505] = 7505,
  [7506] = 6617,
  [7507] = 6618,
  [7508] = 6611,
  [7509] = 6610,
  [7510] = 6712,
  [7511] = 6713,
  [7512] = 6620,
  [7513] = 6739,
  [7514] = 6911,
  [7515] = 6841,
  [7516] = 6620,
  [7517] = 6622,
  [7518] = 7518,
  [7519] = 6612,
  [7520] = 6655,
  [7521] = 6712,
  [7522] = 6713,
  [7523] = 7303,
  [7524] = 6739,
  [7525] = 6735,
  [7526] = 7526,
  [7527] = 6774,
  [7528] = 6610,
  [7529] = 7529,
  [7530] = 6862,
  [7531] = 6932,
  [7532] = 6712,
  [7533] = 6713,
  [7534] = 6808,
  [7535] = 6739,
  [7536] = 7536,
  [7537] = 6609,
  [7538] = 7538,
  [7539] = 7064,
  [7540] = 7068,
  [7541] = 7541,
  [7542] = 6622,
  [7543] = 6712,
  [7544] = 6713,
  [7545] = 7150,
  [7546] = 6739,
  [7547] = 7174,
  [7548] = 6862,
  [7549] = 7189,
  [7550] = 7256,
  [7551] = 6882,
  [7552] = 6890,
  [7553] = 6611,
  [7554] = 6712,
  [7555] = 6713,
  [7556] = 7303,
  [7557] = 6739,
  [7558] = 7361,
  [7559] = 7413,
  [7560] = 7376,
  [7561] = 6613,
  [7562] = 7394,
  [7563] = 6663,
  [7564] = 6882,
  [7565] = 6712,
  [7566] = 6713,
  [7567] = 7418,
  [7568] = 6739,
  [7569] = 7428,
  [7570] = 7438,
  [7571] = 7452,
  [7572] = 6616,
  [7573] = 6617,
  [7574] = 7492,
  [7575] = 7575,
  [7576] = 6712,
  [7577] = 6713,
  [7578] = 6618,
  [7579] = 6739,
  [7580] = 7505,
  [7581] = 6895,
  [7582] = 6620,
  [7583] = 7583,
  [7584] = 6895,
  [7585] = 7585,
  [7586] = 6913,
  [7587] = 6712,
  [7588] = 6713,
  [7589] = 6622,
  [7590] = 6739,
  [7591] = 5746,
  [7592] = 6621,
  [7593] = 6613,
  [7594] = 6654,
  [7595] = 6913,
  [7596] = 6724,
  [7597] = 6647,
  [7598] = 6712,
  [7599] = 6713,
  [7600] = 6728,
  [7601] = 6609,
  [7602] = 6770,
  [7603] = 6781,
  [7604] = 6950,
  [7605] = 6805,
  [7606] = 6986,
  [7607] = 7092,
  [7608] = 6808,
  [7609] = 6712,
  [7610] = 7610,
  [7611] = 6818,
  [7612] = 6736,
  [7613] = 7613,
  [7614] = 6757,
  [7615] = 6712,
  [7616] = 6775,
  [7617] = 7617,
  [7618] = 6797,
  [7619] = 6868,
  [7620] = 6875,
  [7621] = 6712,
  [7622] = 6609,
  [7623] = 6889,
  [7624] = 6712,
  [7625] = 6893,
  [7626] = 7626,
  [7627] = 6712,
  [7628] = 6900,
  [7629] = 6613,
  [7630] = 6907,
  [7631] = 6917,
  [7632] = 6922,
  [7633] = 6611,
  [7634] = 6935,
  [7635] = 6947,
  [7636] = 6952,
  [7637] = 6608,
  [7638] = 6663,
  [7639] = 6613,
  [7640] = 6965,
  [7641] = 6609,
  [7642] = 6734,
  [7643] = 6980,
  [7644] = 6616,
  [7645] = 6984,
  [7646] = 6987,
  [7647] = 6617,
  [7648] = 6996,
  [7649] = 6618,
  [7650] = 6657,
  [7651] = 6895,
  [7652] = 6616,
  [7653] = 6617,
  [7654] = 6618,
  [7655] = 7097,
  [7656] = 6620,
  [7657] = 6611,
  [7658] = 6622,
  [7659] = 7184,
  [7660] = 7276,
  [7661] = 7360,
  [7662] = 7496,
  [7663] = 6777,
  [7664] = 6620,
  [7665] = 6818,
  [7666] = 7146,
  [7667] = 7667,
  [7668] = 7157,
  [7669] = 7227,
  [7670] = 7670,
  [7671] = 7671,
  [7672] = 6609,
  [7673] = 7673,
  [7674] = 6613,
  [7675] = 7675,
  [7676] = 6623,
  [7677] = 6721,
  [7678] = 7057,
  [7679] = 6736,
  [7680] = 6705,
  [7681] = 6710,
  [7682] = 6611,
  [7683] = 7683,
  [7684] = 7124,
  [7685] = 6613,
  [7686] = 6616,
  [7687] = 6617,
  [7688] = 7493,
  [7689] = 7324,
  [7690] = 6843,
  [7691] = 6914,
  [7692] = 7240,
  [7693] = 6616,
  [7694] = 7246,
  [7695] = 6617,
  [7696] = 6838,
  [7697] = 6902,
  [7698] = 7081,
  [7699] = 7299,
  [7700] = 6618,
  [7701] = 6618,
  [7702] = 6620,
  [7703] = 7302,
  [7704] = 6609,
  [7705] = 6622,
  [7706] = 7610,
  [7707] = 6778,
  [7708] = 6744,
  [7709] = 7709,
  [7710] = 6609,
  [7711] = 6719,
  [7712] = 6620,
  [7713] = 7155,
  [7714] = 7714,
  [7715] = 6611,
  [7716] = 6622,
  [7717] = 7717,
  [7718] = 7160,
  [7719] = 7186,
  [7720] = 6613,
  [7721] = 6752,
  [7722] = 7184,
  [7723] = 7276,
  [7724] = 7360,
  [7725] = 7496,
  [7726] = 6622,
  [7727] = 7404,
  [7728] = 6616,
  [7729] = 6617,
  [7730] = 6618,
  [7731] = 7157,
  [7732] = 7227,
  [7733] = 7670,
  [7734] = 6613,
  [7735] = 6620,
  [7736] = 7413,
  [7737] = 6657,
  [7738] = 6623,
  [7739] = 6721,
  [7740] = 7057,
  [7741] = 6622,
  [7742] = 7742,
  [7743] = 6728,
  [7744] = 6949,
  [7745] = 6663,
  [7746] = 7419,
  [7747] = 7405,
  [7748] = 6609,
  [7749] = 6654,
  [7750] = 7493,
  [7751] = 7324,
  [7752] = 6843,
  [7753] = 6914,
  [7754] = 7439,
  [7755] = 6616,
  [7756] = 7472,
  [7757] = 6611,
  [7758] = 6838,
  [7759] = 6902,
  [7760] = 7475,
  [7761] = 6611,
  [7762] = 6613,
  [7763] = 6617,
  [7764] = 7495,
  [7765] = 6618,
  [7766] = 6616,
  [7767] = 6617,
  [7768] = 6618,
  [7769] = 6663,
  [7770] = 6620,
  [7771] = 7504,
  [7772] = 6609,
  [7773] = 6622,
  [7774] = 6740,
  [7775] = 6818,
  [7776] = 7585,
  [7777] = 7529,
  [7778] = 6620,
  [7779] = 6609,
  [7780] = 7184,
  [7781] = 7276,
  [7782] = 7360,
  [7783] = 7496,
  [7784] = 6611,
  [7785] = 7538,
  [7786] = 6705,
  [7787] = 7541,
  [7788] = 6611,
  [7789] = 7157,
  [7790] = 7227,
  [7791] = 7670,
  [7792] = 6613,
  [7793] = 6613,
  [7794] = 6710,
  [7795] = 6747,
  [7796] = 6623,
  [7797] = 7057,
  [7798] = 6616,
  [7799] = 7575,
  [7800] = 7583,
  [7801] = 6617,
  [7802] = 6618,
  [7803] = 7493,
  [7804] = 7324,
  [7805] = 6843,
  [7806] = 6914,
  [7807] = 7617,
  [7808] = 6616,
  [7809] = 6620,
  [7810] = 6617,
  [7811] = 6838,
  [7812] = 6902,
  [7813] = 7626,
  [7814] = 6622,
  [7815] = 6618,
  [7816] = 7670,
  [7817] = 6719,
  [7818] = 6734,
  [7819] = 6620,
  [7820] = 6609,
  [7821] = 7155,
  [7822] = 6622,
  [7823] = 6611,
  [7824] = 7184,
  [7825] = 7276,
  [7826] = 7496,
  [7827] = 7827,
  [7828] = 7626,
  [7829] = 7683,
  [7830] = 6613,
  [7831] = 7227,
  [7832] = 7670,
  [7833] = 7160,
  [7834] = 6623,
  [7835] = 7057,
  [7836] = 6752,
  [7837] = 7837,
  [7838] = 6616,
  [7839] = 6617,
  [7840] = 7840,
  [7841] = 7493,
  [7842] = 7324,
  [7843] = 6843,
  [7844] = 6914,
  [7845] = 6622,
  [7846] = 6838,
  [7847] = 6902,
  [7848] = 7405,
  [7849] = 6618,
  [7850] = 7585,
  [7851] = 7851,
  [7852] = 6724,
  [7853] = 6620,
  [7854] = 7184,
  [7855] = 7276,
  [7856] = 7496,
  [7857] = 6754,
  [7858] = 7227,
  [7859] = 7670,
  [7860] = 7057,
  [7861] = 7021,
  [7862] = 7493,
  [7863] = 7324,
  [7864] = 6843,
  [7865] = 6914,
  [7866] = 6838,
  [7867] = 6902,
  [7868] = 6747,
  [7869] = 6734,
  [7870] = 6622,
  [7871] = 7670,
  [7872] = 7057,
  [7873] = 7840,
  [7874] = 7493,
  [7875] = 7324,
  [7876] = 6843,
  [7877] = 6914,
  [7878] = 6838,
  [7879] = 6902,
  [7880] = 7275,
  [7881] = 7324,
  [7882] = 6843,
  [7883] = 6914,
  [7884] = 6902,
  [7885] = 7324,
  [7886] = 6843,
  [7887] = 6914,
  [7888] = 7324,
  [7889] = 6843,
  [7890] = 6914,
  [7891] = 7324,
  [7892] = 6843,
  [7893] = 6914,
  [7894] = 7324,
  [7895] = 6843,
  [7896] = 6914,
  [7897] = 7324,
  [7898] = 6843,
  [7899] = 6914,
  [7900] = 7324,
  [7901] = 6843,
  [7902] = 6914,
  [7903] = 7324,
  [7904] = 6843,
  [7905] = 6914,
  [7906] = 7324,
  [7907] = 6843,
  [7908] = 6914,
  [7909] = 7324,
  [7910] = 6843,
  [7911] = 6914,
  [7912] = 7324,
  [7913] = 6843,
  [7914] = 6914,
  [7915] = 7324,
  [7916] = 6843,
  [7917] = 6914,
  [7918] = 7324,
  [7919] = 6843,
  [7920] = 6914,
  [7921] = 7324,
  [7922] = 6843,
  [7923] = 6914,
  [7924] = 7324,
  [7925] = 6843,
  [7926] = 6914,
  [7927] = 7324,
  [7928] = 6843,
  [7929] = 6914,
  [7930] = 7324,
  [7931] = 6843,
  [7932] = 6914,
  [7933] = 7324,
  [7934] = 6843,
  [7935] = 6914,
  [7936] = 7324,
  [7937] = 6843,
  [7938] = 6914,
  [7939] = 7324,
  [7940] = 6843,
  [7941] = 6914,
  [7942] = 7324,
  [7943] = 6843,
  [7944] = 6914,
  [7945] = 7324,
  [7946] = 6843,
  [7947] = 6914,
  [7948] = 7324,
  [7949] = 6843,
  [7950] = 6914,
  [7951] = 7324,
  [7952] = 6843,
  [7953] = 6914,
  [7954] = 7324,
  [7955] = 6843,
  [7956] = 6914,
  [7957] = 7324,
  [7958] = 6843,
  [7959] = 6914,
  [7960] = 7324,
  [7961] = 6843,
  [7962] = 6914,
  [7963] = 7324,
  [7964] = 6843,
  [7965] = 6914,
  [7966] = 7324,
  [7967] = 6843,
  [7968] = 6914,
  [7969] = 7324,
  [7970] = 6843,
  [7971] = 6914,
  [7972] = 7324,
  [7973] = 6843,
  [7974] = 6914,
  [7975] = 7324,
  [7976] = 6843,
  [7977] = 6914,
  [7978] = 7324,
  [7979] = 6843,
  [7980] = 6914,
  [7981] = 7324,
  [7982] = 6843,
  [7983] = 6914,
  [7984] = 7324,
  [7985] = 6843,
  [7986] = 6914,
  [7987] = 7324,
  [7988] = 6843,
  [7989] = 6914,
  [7990] = 7324,
  [7991] = 6843,
  [7992] = 6914,
  [7993] = 7324,
  [7994] = 6843,
  [7995] = 6914,
  [7996] = 7324,
  [7997] = 6843,
  [7998] = 6914,
  [7999] = 7324,
  [8000] = 6843,
  [8001] = 6914,
  [8002] = 7324,
  [8003] = 6843,
  [8004] = 6914,
  [8005] = 7324,
  [8006] = 6843,
  [8007] = 6914,
  [8008] = 7324,
  [8009] = 6843,
  [8010] = 6914,
  [8011] = 7324,
  [8012] = 6843,
  [8013] = 6914,
  [8014] = 7311,
  [8015] = 7380,
  [8016] = 6752,
  [8017] = 7827,
  [8018] = 6754,
  [8019] = 8019,
  [8020] = 8020,
  [8021] = 8021,
  [8022] = 6645,
  [8023] = 6728,
  [8024] = 6650,
  [8025] = 6708,
  [8026] = 6841,
  [8027] = 6722,
  [8028] = 8028,
  [8029] = 6727,
  [8030] = 6759,
  [8031] = 7030,
  [8032] = 8032,
  [8033] = 7113,
  [8034] = 6609,
  [8035] = 7036,
  [8036] = 7111,
  [8037] = 6654,
  [8038] = 7181,
  [8039] = 6734,
  [8040] = 7359,
  [8041] = 7613,
  [8042] = 7709,
  [8043] = 7742,
  [8044] = 6611,
  [8045] = 8045,
  [8046] = 8046,
  [8047] = 6703,
  [8048] = 6718,
  [8049] = 6768,
  [8050] = 8050,
  [8051] = 8051,
  [8052] = 7717,
  [8053] = 6613,
  [8054] = 7827,
  [8055] = 6638,
  [8056] = 8056,
  [8057] = 6616,
  [8058] = 6648,
  [8059] = 6617,
  [8060] = 8060,
  [8061] = 6618,
  [8062] = 6654,
  [8063] = 6711,
  [8064] = 7113,
  [8065] = 6663,
  [8066] = 6620,
  [8067] = 6609,
  [8068] = 6622,
  [8069] = 7181,
  [8070] = 6754,
  [8071] = 6740,
  [8072] = 8072,
  [8073] = 7675,
  [8074] = 8074,
  [8075] = 6622,
  [8076] = 8076,
  [8077] = 8077,
  [8078] = 6663,
  [8079] = 6609,
  [8080] = 6609,
  [8081] = 6829,
  [8082] = 6611,
  [8083] = 6747,
  [8084] = 7009,
  [8085] = 8085,
  [8086] = 6709,
  [8087] = 6895,
  [8088] = 7326,
  [8089] = 8089,
  [8090] = 7113,
  [8091] = 6611,
  [8092] = 7373,
  [8093] = 6613,
  [8094] = 7181,
  [8095] = 6613,
  [8096] = 6752,
  [8097] = 6799,
  [8098] = 8098,
  [8099] = 6754,
  [8100] = 6616,
  [8101] = 6611,
  [8102] = 6617,
  [8103] = 6618,
  [8104] = 6911,
  [8105] = 6841,
  [8106] = 6616,
  [8107] = 6620,
  [8108] = 7518,
  [8109] = 6612,
  [8110] = 7113,
  [8111] = 7181,
  [8112] = 6655,
  [8113] = 6610,
  [8114] = 6617,
  [8115] = 6862,
  [8116] = 6932,
  [8117] = 6622,
  [8118] = 8118,
  [8119] = 8119,
  [8120] = 8120,
  [8121] = 7064,
  [8122] = 8122,
  [8123] = 8123,
  [8124] = 7113,
  [8125] = 7181,
  [8126] = 7150,
  [8127] = 6882,
  [8128] = 6890,
  [8129] = 8129,
  [8130] = 7303,
  [8131] = 7361,
  [8132] = 6618,
  [8133] = 7376,
  [8134] = 7181,
  [8135] = 6759,
  [8136] = 7394,
  [8137] = 6609,
  [8138] = 6620,
  [8139] = 7418,
  [8140] = 7428,
  [8141] = 7438,
  [8142] = 7452,
  [8143] = 8143,
  [8144] = 6611,
  [8145] = 7492,
  [8146] = 8146,
  [8147] = 5768,
  [8148] = 7505,
  [8149] = 6622,
  [8150] = 6613,
  [8151] = 6759,
  [8152] = 6895,
  [8153] = 8153,
  [8154] = 6913,
  [8155] = 8155,
  [8156] = 8156,
  [8157] = 6621,
  [8158] = 6613,
  [8159] = 6616,
  [8160] = 6617,
  [8161] = 6724,
  [8162] = 6647,
  [8163] = 6770,
  [8164] = 6781,
  [8165] = 6805,
  [8166] = 6808,
  [8167] = 7610,
  [8168] = 6818,
  [8169] = 6736,
  [8170] = 6618,
  [8171] = 6757,
  [8172] = 6775,
  [8173] = 8173,
  [8174] = 6797,
  [8175] = 6875,
  [8176] = 6609,
  [8177] = 6889,
  [8178] = 6893,
  [8179] = 6620,
  [8180] = 6900,
  [8181] = 8181,
  [8182] = 6907,
  [8183] = 6917,
  [8184] = 6922,
  [8185] = 6622,
  [8186] = 6935,
  [8187] = 6947,
  [8188] = 6952,
  [8189] = 8189,
  [8190] = 8190,
  [8191] = 8191,
  [8192] = 6965,
  [8193] = 7030,
  [8194] = 8194,
  [8195] = 6980,
  [8196] = 8196,
  [8197] = 6984,
  [8198] = 6987,
  [8199] = 6609,
  [8200] = 6996,
  [8201] = 6654,
  [8202] = 6657,
  [8203] = 8203,
  [8204] = 6616,
  [8205] = 6617,
  [8206] = 6618,
  [8207] = 6611,
  [8208] = 6620,
  [8209] = 7146,
  [8210] = 6654,
  [8211] = 6705,
  [8212] = 6710,
  [8213] = 7124,
  [8214] = 7240,
  [8215] = 7246,
  [8216] = 7299,
  [8217] = 6613,
  [8218] = 6719,
  [8219] = 7155,
  [8220] = 7160,
  [8221] = 7186,
  [8222] = 7404,
  [8223] = 6740,
  [8224] = 7413,
  [8225] = 8225,
  [8226] = 6728,
  [8227] = 7419,
  [8228] = 7439,
  [8229] = 7472,
  [8230] = 7475,
  [8231] = 7495,
  [8232] = 7504,
  [8233] = 6616,
  [8234] = 7529,
  [8235] = 7538,
  [8236] = 7541,
  [8237] = 6617,
  [8238] = 7575,
  [8239] = 7583,
  [8240] = 7617,
  [8241] = 6618,
  [8242] = 7626,
  [8243] = 6734,
  [8244] = 6663,
  [8245] = 7683,
  [8246] = 6620,
  [8247] = 6622,
  [8248] = 7405,
  [8249] = 7585,
  [8250] = 7851,
  [8251] = 6747,
  [8252] = 6752,
  [8253] = 7827,
  [8254] = 6754,
  [8255] = 8019,
  [8256] = 8020,
  [8257] = 8021,
  [8258] = 6645,
  [8259] = 6650,
  [8260] = 6708,
  [8261] = 6722,
  [8262] = 6609,
  [8263] = 6727,
  [8264] = 6759,
  [8265] = 7030,
  [8266] = 6622,
  [8267] = 7036,
  [8268] = 7613,
  [8269] = 7029,
  [8270] = 6913,
  [8271] = 7232,
  [8272] = 8272,
  [8273] = 8194,
  [8274] = 7709,
  [8275] = 7742,
  [8276] = 8045,
  [8277] = 7029,
  [8278] = 8278,
  [8279] = 7232,
  [8280] = 6611,
  [8281] = 8194,
  [8282] = 6718,
  [8283] = 6913,
  [8284] = 6609,
  [8285] = 7029,
  [8286] = 8286,
  [8287] = 7232,
  [8288] = 6613,
  [8289] = 8194,
  [8290] = 8290,
  [8291] = 8291,
  [8292] = 6638,
  [8293] = 7029,
  [8294] = 6611,
  [8295] = 7232,
  [8296] = 5777,
  [8297] = 6613,
  [8298] = 7029,
  [8299] = 6663,
  [8300] = 7232,
  [8301] = 6616,
  [8302] = 6648,
  [8303] = 7029,
  [8304] = 6616,
  [8305] = 7232,
  [8306] = 6617,
  [8307] = 6618,
  [8308] = 7232,
  [8309] = 6617,
  [8310] = 6620,
  [8311] = 6618,
  [8312] = 6622,
  [8313] = 6609,
  [8314] = 8314,
  [8315] = 7518,
  [8316] = 8316,
  [8317] = 6609,
  [8318] = 6620,
  [8319] = 7851,
  [8320] = 6611,
  [8321] = 6622,
  [8322] = 6613,
  [8323] = 6711,
  [8324] = 7036,
  [8325] = 6616,
  [8326] = 6617,
  [8327] = 6618,
  [8328] = 8019,
  [8329] = 6620,
  [8330] = 6739,
  [8331] = 6622,
  [8332] = 8045,
  [8333] = 6612,
  [8334] = 8334,
  [8335] = 8335,
  [8336] = 6609,
  [8337] = 8020,
  [8338] = 8045,
  [8339] = 6611,
  [8340] = 6654,
  [8341] = 6613,
  [8342] = 8021,
  [8343] = 6611,
  [8344] = 6616,
  [8345] = 6617,
  [8346] = 6618,
  [8347] = 7518,
  [8348] = 6620,
  [8349] = 8349,
  [8350] = 6622,
  [8351] = 6663,
  [8352] = 8352,
  [8353] = 6258,
  [8354] = 6654,
  [8355] = 6711,
  [8356] = 8046,
  [8357] = 6655,
  [8358] = 6740,
  [8359] = 7683,
  [8360] = 8360,
  [8361] = 6739,
  [8362] = 8362,
  [8363] = 8363,
  [8364] = 8364,
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(9);
      ADVANCE_MAP(
        '&', 36,
        '\'', 44,
        '(', 13,
        ')', 14,
        '*', 37,
        '+', 34,
        ',', 11,
        '-', 35,
        '.', 12,
        '/', 38,
        ':', 15,
        ';', 10,
        '<', 25,
        '=', 23,
        '>', 26,
        '?', 18,
        '@', 42,
        '[', 45,
        ']', 46,
        '^', 43,
        '|', 47,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(0);
      END_STATE();
    case 1:
      ADVANCE_MAP(
        '&', 36,
        '\'', 44,
        '(', 13,
        ')', 14,
        '*', 37,
        '+', 34,
        ',', 11,
        '-', 35,
        '.', 12,
        '/', 38,
        ':', 15,
        ';', 10,
        '<', 25,
        '=', 23,
        '>', 26,
        '?', 4,
        '[', 45,
        '|', 47,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(1);
      END_STATE();
    case 2:
      ADVANCE_MAP(
        '\'', 44,
        '(', 13,
        ')', 14,
        '+', 34,
        ',', 11,
        '-', 35,
        '.', 12,
        ':', 6,
        ';', 10,
        '<', 5,
        '>', 8,
        '?', 19,
        '[', 45,
        ']', 46,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(2);
      END_STATE();
    case 3:
      if (lookahead == '(') ADVANCE(13);
      if (lookahead == ';') ADVANCE(10);
      if (lookahead == '<') ADVANCE(5);
      if (lookahead == '?') ADVANCE(17);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(3);
      END_STATE();
    case 4:
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == '<') ADVANCE(30);
      if (lookahead == '=') ADVANCE(28);
      if (lookahead == '>') ADVANCE(32);
      END_STATE();
    case 5:
      if (lookahead == '<') ADVANCE(40);
      END_STATE();
    case 6:
      if (lookahead == '=') ADVANCE(21);
      END_STATE();
    case 7:
      if (lookahead == '=') ADVANCE(29);
      END_STATE();
    case 8:
      if (lookahead == '>') ADVANCE(41);
      END_STATE();
    case 9:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 10:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 11:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 12:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 13:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 14:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 15:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '=') ADVANCE(21);
      END_STATE();
    case 16:
      ACCEPT_TOKEN(anon_sym_EQ_GT);
      END_STATE();
    case 17:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 18:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '/') ADVANCE(7);
      if (lookahead == '=') ADVANCE(28);
      if (lookahead == '>') ADVANCE(32);
      if (lookahead == '?') ADVANCE(22);
      END_STATE();
    case 19:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '?') ADVANCE(22);
      END_STATE();
    case 20:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 21:
      ACCEPT_TOKEN(sym_variable_assignment);
      END_STATE();
    case 22:
      ACCEPT_TOKEN(sym_condition_conversion);
      END_STATE();
    case 23:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '>') ADVANCE(16);
      END_STATE();
    case 24:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 25:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '=') ADVANCE(20);
      END_STATE();
    case 26:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(27);
      if (lookahead == '>') ADVANCE(41);
      END_STATE();
    case 27:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 28:
      ACCEPT_TOKEN(anon_sym_QMARK_EQ);
      END_STATE();
    case 29:
      ACCEPT_TOKEN(anon_sym_QMARK_SLASH_EQ);
      END_STATE();
    case 30:
      ACCEPT_TOKEN(anon_sym_QMARK_LT);
      if (lookahead == '=') ADVANCE(31);
      END_STATE();
    case 31:
      ACCEPT_TOKEN(anon_sym_QMARK_LT_EQ);
      END_STATE();
    case 32:
      ACCEPT_TOKEN(anon_sym_QMARK_GT);
      if (lookahead == '=') ADVANCE(33);
      END_STATE();
    case 33:
      ACCEPT_TOKEN(anon_sym_QMARK_GT_EQ);
      END_STATE();
    case 34:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 35:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 36:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 37:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(39);
      END_STATE();
    case 38:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '=') ADVANCE(24);
      END_STATE();
    case 39:
      ACCEPT_TOKEN(sym_exponentiate);
      END_STATE();
    case 40:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 41:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      END_STATE();
    case 42:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 43:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 44:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 45:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 46:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 47:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 2, .external_lex_state = 2},
  [2] = {.lex_state = 2, .external_lex_state = 2},
  [3] = {.lex_state = 2, .external_lex_state = 2},
  [4] = {.lex_state = 2, .external_lex_state = 3},
  [5] = {.lex_state = 2, .external_lex_state = 4},
  [6] = {.lex_state = 2, .external_lex_state = 5},
  [7] = {.lex_state = 2, .external_lex_state = 6},
  [8] = {.lex_state = 2, .external_lex_state = 6},
  [9] = {.lex_state = 2, .external_lex_state = 7},
  [10] = {.lex_state = 2, .external_lex_state = 7},
  [11] = {.lex_state = 2, .external_lex_state = 6},
  [12] = {.lex_state = 2, .external_lex_state = 6},
  [13] = {.lex_state = 2, .external_lex_state = 8},
  [14] = {.lex_state = 2, .external_lex_state = 9},
  [15] = {.lex_state = 2, .external_lex_state = 8},
  [16] = {.lex_state = 2, .external_lex_state = 8},
  [17] = {.lex_state = 2, .external_lex_state = 10},
  [18] = {.lex_state = 2, .external_lex_state = 11},
  [19] = {.lex_state = 2, .external_lex_state = 12},
  [20] = {.lex_state = 2, .external_lex_state = 13},
  [21] = {.lex_state = 2, .external_lex_state = 2},
  [22] = {.lex_state = 2, .external_lex_state = 11},
  [23] = {.lex_state = 2, .external_lex_state = 9},
  [24] = {.lex_state = 2, .external_lex_state = 9},
  [25] = {.lex_state = 2, .external_lex_state = 13},
  [26] = {.lex_state = 2, .external_lex_state = 9},
  [27] = {.lex_state = 2, .external_lex_state = 13},
  [28] = {.lex_state = 2, .external_lex_state = 9},
  [29] = {.lex_state = 2, .external_lex_state = 9},
  [30] = {.lex_state = 2, .external_lex_state = 10},
  [31] = {.lex_state = 2, .external_lex_state = 10},
  [32] = {.lex_state = 2, .external_lex_state = 10},
  [33] = {.lex_state = 2, .external_lex_state = 10},
  [34] = {.lex_state = 2, .external_lex_state = 10},
  [35] = {.lex_state = 2, .external_lex_state = 10},
  [36] = {.lex_state = 2, .external_lex_state = 10},
  [37] = {.lex_state = 2, .external_lex_state = 10},
  [38] = {.lex_state = 2, .external_lex_state = 10},
  [39] = {.lex_state = 2, .external_lex_state = 10},
  [40] = {.lex_state = 2, .external_lex_state = 10},
  [41] = {.lex_state = 2, .external_lex_state = 10},
  [42] = {.lex_state = 2, .external_lex_state = 10},
  [43] = {.lex_state = 2, .external_lex_state = 10},
  [44] = {.lex_state = 2, .external_lex_state = 10},
  [45] = {.lex_state = 2, .external_lex_state = 10},
  [46] = {.lex_state = 2, .external_lex_state = 10},
  [47] = {.lex_state = 2, .external_lex_state = 10},
  [48] = {.lex_state = 2, .external_lex_state = 10},
  [49] = {.lex_state = 2, .external_lex_state = 10},
  [50] = {.lex_state = 2, .external_lex_state = 10},
  [51] = {.lex_state = 2, .external_lex_state = 10},
  [52] = {.lex_state = 2, .external_lex_state = 10},
  [53] = {.lex_state = 2, .external_lex_state = 10},
  [54] = {.lex_state = 2, .external_lex_state = 10},
  [55] = {.lex_state = 2, .external_lex_state = 10},
  [56] = {.lex_state = 2, .external_lex_state = 10},
  [57] = {.lex_state = 2, .external_lex_state = 10},
  [58] = {.lex_state = 2, .external_lex_state = 10},
  [59] = {.lex_state = 2, .external_lex_state = 10},
  [60] = {.lex_state = 2, .external_lex_state = 10},
  [61] = {.lex_state = 2, .external_lex_state = 9},
  [62] = {.lex_state = 2, .external_lex_state = 10},
  [63] = {.lex_state = 2, .external_lex_state = 10},
  [64] = {.lex_state = 2, .external_lex_state = 10},
  [65] = {.lex_state = 2, .external_lex_state = 10},
  [66] = {.lex_state = 2, .external_lex_state = 10},
  [67] = {.lex_state = 2, .external_lex_state = 10},
  [68] = {.lex_state = 2, .external_lex_state = 10},
  [69] = {.lex_state = 2, .external_lex_state = 10},
  [70] = {.lex_state = 2, .external_lex_state = 10},
  [71] = {.lex_state = 2, .external_lex_state = 10},
  [72] = {.lex_state = 2, .external_lex_state = 10},
  [73] = {.lex_state = 2, .external_lex_state = 10},
  [74] = {.lex_state = 2, .external_lex_state = 10},
  [75] = {.lex_state = 2, .external_lex_state = 10},
  [76] = {.lex_state = 2, .external_lex_state = 10},
  [77] = {.lex_state = 2, .external_lex_state = 10},
  [78] = {.lex_state = 2, .external_lex_state = 10},
  [79] = {.lex_state = 2, .external_lex_state = 10},
  [80] = {.lex_state = 2, .external_lex_state = 13},
  [81] = {.lex_state = 2, .external_lex_state = 13},
  [82] = {.lex_state = 2, .external_lex_state = 12},
  [83] = {.lex_state = 2, .external_lex_state = 11},
  [84] = {.lex_state = 2, .external_lex_state = 12},
  [85] = {.lex_state = 2, .external_lex_state = 13},
  [86] = {.lex_state = 2, .external_lex_state = 11},
  [87] = {.lex_state = 2, .external_lex_state = 13},
  [88] = {.lex_state = 2, .external_lex_state = 13},
  [89] = {.lex_state = 2, .external_lex_state = 12},
  [90] = {.lex_state = 2, .external_lex_state = 11},
  [91] = {.lex_state = 2, .external_lex_state = 12},
  [92] = {.lex_state = 2, .external_lex_state = 13},
  [93] = {.lex_state = 2, .external_lex_state = 11},
  [94] = {.lex_state = 2, .external_lex_state = 2},
  [95] = {.lex_state = 2, .external_lex_state = 13},
  [96] = {.lex_state = 2, .external_lex_state = 13},
  [97] = {.lex_state = 2, .external_lex_state = 13},
  [98] = {.lex_state = 2, .external_lex_state = 12},
  [99] = {.lex_state = 2, .external_lex_state = 10},
  [100] = {.lex_state = 2, .external_lex_state = 13},
  [101] = {.lex_state = 2, .external_lex_state = 14},
  [102] = {.lex_state = 2, .external_lex_state = 13},
  [103] = {.lex_state = 2, .external_lex_state = 13},
  [104] = {.lex_state = 2, .external_lex_state = 13},
  [105] = {.lex_state = 2, .external_lex_state = 13},
  [106] = {.lex_state = 2, .external_lex_state = 13},
  [107] = {.lex_state = 2, .external_lex_state = 14},
  [108] = {.lex_state = 2, .external_lex_state = 13},
  [109] = {.lex_state = 2, .external_lex_state = 14},
  [110] = {.lex_state = 2, .external_lex_state = 13},
  [111] = {.lex_state = 2, .external_lex_state = 2},
  [112] = {.lex_state = 2, .external_lex_state = 14},
  [113] = {.lex_state = 2, .external_lex_state = 13},
  [114] = {.lex_state = 2, .external_lex_state = 14},
  [115] = {.lex_state = 2, .external_lex_state = 13},
  [116] = {.lex_state = 2, .external_lex_state = 14},
  [117] = {.lex_state = 2, .external_lex_state = 13},
  [118] = {.lex_state = 2, .external_lex_state = 14},
  [119] = {.lex_state = 2, .external_lex_state = 14},
  [120] = {.lex_state = 2, .external_lex_state = 13},
  [121] = {.lex_state = 2, .external_lex_state = 14},
  [122] = {.lex_state = 2, .external_lex_state = 14},
  [123] = {.lex_state = 2, .external_lex_state = 14},
  [124] = {.lex_state = 2, .external_lex_state = 2},
  [125] = {.lex_state = 2, .external_lex_state = 14},
  [126] = {.lex_state = 2, .external_lex_state = 15},
  [127] = {.lex_state = 2, .external_lex_state = 15},
  [128] = {.lex_state = 2, .external_lex_state = 14},
  [129] = {.lex_state = 2, .external_lex_state = 14},
  [130] = {.lex_state = 2, .external_lex_state = 14},
  [131] = {.lex_state = 2, .external_lex_state = 10},
  [132] = {.lex_state = 2, .external_lex_state = 15},
  [133] = {.lex_state = 2, .external_lex_state = 14},
  [134] = {.lex_state = 2, .external_lex_state = 14},
  [135] = {.lex_state = 2, .external_lex_state = 14},
  [136] = {.lex_state = 2, .external_lex_state = 14},
  [137] = {.lex_state = 2, .external_lex_state = 15},
  [138] = {.lex_state = 2, .external_lex_state = 14},
  [139] = {.lex_state = 2, .external_lex_state = 14},
  [140] = {.lex_state = 2, .external_lex_state = 15},
  [141] = {.lex_state = 2, .external_lex_state = 14},
  [142] = {.lex_state = 2, .external_lex_state = 14},
  [143] = {.lex_state = 2, .external_lex_state = 14},
  [144] = {.lex_state = 2, .external_lex_state = 15},
  [145] = {.lex_state = 2, .external_lex_state = 15},
  [146] = {.lex_state = 2, .external_lex_state = 15},
  [147] = {.lex_state = 2, .external_lex_state = 15},
  [148] = {.lex_state = 2, .external_lex_state = 15},
  [149] = {.lex_state = 2, .external_lex_state = 15},
  [150] = {.lex_state = 2, .external_lex_state = 15},
  [151] = {.lex_state = 2, .external_lex_state = 15},
  [152] = {.lex_state = 2, .external_lex_state = 15},
  [153] = {.lex_state = 2, .external_lex_state = 15},
  [154] = {.lex_state = 2, .external_lex_state = 15},
  [155] = {.lex_state = 2, .external_lex_state = 15},
  [156] = {.lex_state = 2, .external_lex_state = 15},
  [157] = {.lex_state = 2, .external_lex_state = 15},
  [158] = {.lex_state = 2, .external_lex_state = 15},
  [159] = {.lex_state = 2, .external_lex_state = 15},
  [160] = {.lex_state = 2, .external_lex_state = 15},
  [161] = {.lex_state = 2, .external_lex_state = 15},
  [162] = {.lex_state = 2, .external_lex_state = 15},
  [163] = {.lex_state = 2, .external_lex_state = 15},
  [164] = {.lex_state = 2, .external_lex_state = 15},
  [165] = {.lex_state = 2, .external_lex_state = 15},
  [166] = {.lex_state = 2, .external_lex_state = 15},
  [167] = {.lex_state = 2, .external_lex_state = 15},
  [168] = {.lex_state = 2, .external_lex_state = 15},
  [169] = {.lex_state = 2, .external_lex_state = 15},
  [170] = {.lex_state = 2, .external_lex_state = 15},
  [171] = {.lex_state = 2, .external_lex_state = 15},
  [172] = {.lex_state = 2, .external_lex_state = 15},
  [173] = {.lex_state = 2, .external_lex_state = 15},
  [174] = {.lex_state = 2, .external_lex_state = 15},
  [175] = {.lex_state = 2, .external_lex_state = 15},
  [176] = {.lex_state = 2, .external_lex_state = 15},
  [177] = {.lex_state = 2, .external_lex_state = 15},
  [178] = {.lex_state = 2, .external_lex_state = 15},
  [179] = {.lex_state = 2, .external_lex_state = 15},
  [180] = {.lex_state = 2, .external_lex_state = 15},
  [181] = {.lex_state = 2, .external_lex_state = 15},
  [182] = {.lex_state = 2, .external_lex_state = 15},
  [183] = {.lex_state = 2, .external_lex_state = 15},
  [184] = {.lex_state = 2, .external_lex_state = 15},
  [185] = {.lex_state = 2, .external_lex_state = 15},
  [186] = {.lex_state = 2, .external_lex_state = 15},
  [187] = {.lex_state = 2, .external_lex_state = 15},
  [188] = {.lex_state = 2, .external_lex_state = 15},
  [189] = {.lex_state = 2, .external_lex_state = 15},
  [190] = {.lex_state = 2, .external_lex_state = 15},
  [191] = {.lex_state = 2, .external_lex_state = 15},
  [192] = {.lex_state = 2, .external_lex_state = 15},
  [193] = {.lex_state = 2, .external_lex_state = 15},
  [194] = {.lex_state = 2, .external_lex_state = 15},
  [195] = {.lex_state = 2, .external_lex_state = 15},
  [196] = {.lex_state = 2, .external_lex_state = 15},
  [197] = {.lex_state = 2, .external_lex_state = 15},
  [198] = {.lex_state = 2, .external_lex_state = 15},
  [199] = {.lex_state = 2, .external_lex_state = 15},
  [200] = {.lex_state = 2, .external_lex_state = 15},
  [201] = {.lex_state = 2, .external_lex_state = 15},
  [202] = {.lex_state = 2, .external_lex_state = 15},
  [203] = {.lex_state = 2, .external_lex_state = 15},
  [204] = {.lex_state = 2, .external_lex_state = 15},
  [205] = {.lex_state = 2, .external_lex_state = 15},
  [206] = {.lex_state = 2, .external_lex_state = 15},
  [207] = {.lex_state = 2, .external_lex_state = 15},
  [208] = {.lex_state = 2, .external_lex_state = 15},
  [209] = {.lex_state = 2, .external_lex_state = 15},
  [210] = {.lex_state = 2, .external_lex_state = 15},
  [211] = {.lex_state = 2, .external_lex_state = 15},
  [212] = {.lex_state = 2, .external_lex_state = 15},
  [213] = {.lex_state = 2, .external_lex_state = 15},
  [214] = {.lex_state = 2, .external_lex_state = 15},
  [215] = {.lex_state = 2, .external_lex_state = 15},
  [216] = {.lex_state = 2, .external_lex_state = 15},
  [217] = {.lex_state = 2, .external_lex_state = 15},
  [218] = {.lex_state = 2, .external_lex_state = 15},
  [219] = {.lex_state = 2, .external_lex_state = 15},
  [220] = {.lex_state = 2, .external_lex_state = 16},
  [221] = {.lex_state = 2, .external_lex_state = 15},
  [222] = {.lex_state = 2, .external_lex_state = 15},
  [223] = {.lex_state = 2, .external_lex_state = 15},
  [224] = {.lex_state = 2, .external_lex_state = 15},
  [225] = {.lex_state = 2, .external_lex_state = 17},
  [226] = {.lex_state = 2, .external_lex_state = 18},
  [227] = {.lex_state = 2, .external_lex_state = 18},
  [228] = {.lex_state = 2, .external_lex_state = 18},
  [229] = {.lex_state = 2, .external_lex_state = 19},
  [230] = {.lex_state = 2, .external_lex_state = 18},
  [231] = {.lex_state = 2, .external_lex_state = 19},
  [232] = {.lex_state = 2, .external_lex_state = 19},
  [233] = {.lex_state = 2, .external_lex_state = 19},
  [234] = {.lex_state = 2, .external_lex_state = 18},
  [235] = {.lex_state = 2, .external_lex_state = 19},
  [236] = {.lex_state = 2, .external_lex_state = 19},
  [237] = {.lex_state = 2, .external_lex_state = 18},
  [238] = {.lex_state = 2, .external_lex_state = 10},
  [239] = {.lex_state = 2, .external_lex_state = 20},
  [240] = {.lex_state = 2, .external_lex_state = 20},
  [241] = {.lex_state = 2, .external_lex_state = 20},
  [242] = {.lex_state = 2, .external_lex_state = 21},
  [243] = {.lex_state = 2, .external_lex_state = 20},
  [244] = {.lex_state = 2, .external_lex_state = 20},
  [245] = {.lex_state = 2, .external_lex_state = 20},
  [246] = {.lex_state = 2, .external_lex_state = 10},
  [247] = {.lex_state = 2, .external_lex_state = 20},
  [248] = {.lex_state = 2, .external_lex_state = 20},
  [249] = {.lex_state = 2, .external_lex_state = 20},
  [250] = {.lex_state = 2, .external_lex_state = 20},
  [251] = {.lex_state = 2, .external_lex_state = 10},
  [252] = {.lex_state = 2, .external_lex_state = 15},
  [253] = {.lex_state = 2, .external_lex_state = 10},
  [254] = {.lex_state = 2, .external_lex_state = 10},
  [255] = {.lex_state = 2, .external_lex_state = 10},
  [256] = {.lex_state = 2, .external_lex_state = 10},
  [257] = {.lex_state = 2, .external_lex_state = 10},
  [258] = {.lex_state = 2, .external_lex_state = 10},
  [259] = {.lex_state = 2, .external_lex_state = 10},
  [260] = {.lex_state = 2, .external_lex_state = 10},
  [261] = {.lex_state = 2, .external_lex_state = 10},
  [262] = {.lex_state = 2, .external_lex_state = 10},
  [263] = {.lex_state = 2, .external_lex_state = 10},
  [264] = {.lex_state = 2, .external_lex_state = 10},
  [265] = {.lex_state = 2, .external_lex_state = 10},
  [266] = {.lex_state = 2, .external_lex_state = 10},
  [267] = {.lex_state = 2, .external_lex_state = 10},
  [268] = {.lex_state = 2, .external_lex_state = 10},
  [269] = {.lex_state = 2, .external_lex_state = 20},
  [270] = {.lex_state = 2, .external_lex_state = 10},
  [271] = {.lex_state = 2, .external_lex_state = 10},
  [272] = {.lex_state = 2, .external_lex_state = 10},
  [273] = {.lex_state = 2, .external_lex_state = 10},
  [274] = {.lex_state = 2, .external_lex_state = 10},
  [275] = {.lex_state = 2, .external_lex_state = 10},
  [276] = {.lex_state = 2, .external_lex_state = 20},
  [277] = {.lex_state = 2, .external_lex_state = 10},
  [278] = {.lex_state = 2, .external_lex_state = 10},
  [279] = {.lex_state = 2, .external_lex_state = 10},
  [280] = {.lex_state = 2, .external_lex_state = 10},
  [281] = {.lex_state = 2, .external_lex_state = 10},
  [282] = {.lex_state = 2, .external_lex_state = 10},
  [283] = {.lex_state = 2, .external_lex_state = 10},
  [284] = {.lex_state = 2, .external_lex_state = 10},
  [285] = {.lex_state = 2, .external_lex_state = 10},
  [286] = {.lex_state = 2, .external_lex_state = 10},
  [287] = {.lex_state = 2, .external_lex_state = 20},
  [288] = {.lex_state = 2, .external_lex_state = 10},
  [289] = {.lex_state = 2, .external_lex_state = 21},
  [290] = {.lex_state = 2, .external_lex_state = 10},
  [291] = {.lex_state = 2, .external_lex_state = 20},
  [292] = {.lex_state = 2, .external_lex_state = 10},
  [293] = {.lex_state = 2, .external_lex_state = 10},
  [294] = {.lex_state = 2, .external_lex_state = 10},
  [295] = {.lex_state = 2, .external_lex_state = 15},
  [296] = {.lex_state = 2, .external_lex_state = 10},
  [297] = {.lex_state = 2, .external_lex_state = 20},
  [298] = {.lex_state = 2, .external_lex_state = 10},
  [299] = {.lex_state = 2, .external_lex_state = 10},
  [300] = {.lex_state = 2, .external_lex_state = 20},
  [301] = {.lex_state = 2, .external_lex_state = 10},
  [302] = {.lex_state = 2, .external_lex_state = 10},
  [303] = {.lex_state = 2, .external_lex_state = 10},
  [304] = {.lex_state = 2, .external_lex_state = 21},
  [305] = {.lex_state = 2, .external_lex_state = 10},
  [306] = {.lex_state = 2, .external_lex_state = 20},
  [307] = {.lex_state = 2, .external_lex_state = 10},
  [308] = {.lex_state = 2, .external_lex_state = 10},
  [309] = {.lex_state = 2, .external_lex_state = 20},
  [310] = {.lex_state = 2, .external_lex_state = 10},
  [311] = {.lex_state = 2, .external_lex_state = 10},
  [312] = {.lex_state = 2, .external_lex_state = 10},
  [313] = {.lex_state = 2, .external_lex_state = 10},
  [314] = {.lex_state = 2, .external_lex_state = 10},
  [315] = {.lex_state = 2, .external_lex_state = 21},
  [316] = {.lex_state = 2, .external_lex_state = 10},
  [317] = {.lex_state = 2, .external_lex_state = 20},
  [318] = {.lex_state = 2, .external_lex_state = 20},
  [319] = {.lex_state = 2, .external_lex_state = 10},
  [320] = {.lex_state = 2, .external_lex_state = 20},
  [321] = {.lex_state = 2, .external_lex_state = 10},
  [322] = {.lex_state = 2, .external_lex_state = 15},
  [323] = {.lex_state = 2, .external_lex_state = 20},
  [324] = {.lex_state = 2, .external_lex_state = 21},
  [325] = {.lex_state = 2, .external_lex_state = 21},
  [326] = {.lex_state = 2, .external_lex_state = 20},
  [327] = {.lex_state = 2, .external_lex_state = 20},
  [328] = {.lex_state = 2, .external_lex_state = 20},
  [329] = {.lex_state = 2, .external_lex_state = 20},
  [330] = {.lex_state = 2, .external_lex_state = 20},
  [331] = {.lex_state = 2, .external_lex_state = 20},
  [332] = {.lex_state = 2, .external_lex_state = 20},
  [333] = {.lex_state = 2, .external_lex_state = 20},
  [334] = {.lex_state = 2, .external_lex_state = 20},
  [335] = {.lex_state = 2, .external_lex_state = 20},
  [336] = {.lex_state = 2, .external_lex_state = 20},
  [337] = {.lex_state = 2, .external_lex_state = 20},
  [338] = {.lex_state = 2, .external_lex_state = 20},
  [339] = {.lex_state = 2, .external_lex_state = 20},
  [340] = {.lex_state = 2, .external_lex_state = 20},
  [341] = {.lex_state = 2, .external_lex_state = 20},
  [342] = {.lex_state = 2, .external_lex_state = 20},
  [343] = {.lex_state = 2, .external_lex_state = 20},
  [344] = {.lex_state = 2, .external_lex_state = 20},
  [345] = {.lex_state = 2, .external_lex_state = 10},
  [346] = {.lex_state = 2, .external_lex_state = 20},
  [347] = {.lex_state = 2, .external_lex_state = 20},
  [348] = {.lex_state = 2, .external_lex_state = 20},
  [349] = {.lex_state = 2, .external_lex_state = 20},
  [350] = {.lex_state = 2, .external_lex_state = 20},
  [351] = {.lex_state = 2, .external_lex_state = 20},
  [352] = {.lex_state = 2, .external_lex_state = 22},
  [353] = {.lex_state = 2, .external_lex_state = 20},
  [354] = {.lex_state = 2, .external_lex_state = 20},
  [355] = {.lex_state = 2, .external_lex_state = 20},
  [356] = {.lex_state = 2, .external_lex_state = 20},
  [357] = {.lex_state = 2, .external_lex_state = 22},
  [358] = {.lex_state = 2, .external_lex_state = 20},
  [359] = {.lex_state = 2, .external_lex_state = 20},
  [360] = {.lex_state = 2, .external_lex_state = 20},
  [361] = {.lex_state = 2, .external_lex_state = 20},
  [362] = {.lex_state = 2, .external_lex_state = 20},
  [363] = {.lex_state = 2, .external_lex_state = 20},
  [364] = {.lex_state = 2, .external_lex_state = 20},
  [365] = {.lex_state = 2, .external_lex_state = 20},
  [366] = {.lex_state = 2, .external_lex_state = 22},
  [367] = {.lex_state = 2, .external_lex_state = 22},
  [368] = {.lex_state = 2, .external_lex_state = 22},
  [369] = {.lex_state = 2, .external_lex_state = 22},
  [370] = {.lex_state = 2, .external_lex_state = 22},
  [371] = {.lex_state = 2, .external_lex_state = 22},
  [372] = {.lex_state = 2, .external_lex_state = 20},
  [373] = {.lex_state = 2, .external_lex_state = 22},
  [374] = {.lex_state = 2, .external_lex_state = 22},
  [375] = {.lex_state = 2, .external_lex_state = 22},
  [376] = {.lex_state = 2, .external_lex_state = 22},
  [377] = {.lex_state = 2, .external_lex_state = 22},
  [378] = {.lex_state = 2, .external_lex_state = 22},
  [379] = {.lex_state = 2, .external_lex_state = 22},
  [380] = {.lex_state = 2, .external_lex_state = 22},
  [381] = {.lex_state = 2, .external_lex_state = 22},
  [382] = {.lex_state = 2, .external_lex_state = 22},
  [383] = {.lex_state = 2, .external_lex_state = 22},
  [384] = {.lex_state = 2, .external_lex_state = 22},
  [385] = {.lex_state = 2, .external_lex_state = 22},
  [386] = {.lex_state = 2, .external_lex_state = 22},
  [387] = {.lex_state = 2, .external_lex_state = 22},
  [388] = {.lex_state = 2, .external_lex_state = 22},
  [389] = {.lex_state = 2, .external_lex_state = 22},
  [390] = {.lex_state = 2, .external_lex_state = 22},
  [391] = {.lex_state = 2, .external_lex_state = 22},
  [392] = {.lex_state = 2, .external_lex_state = 22},
  [393] = {.lex_state = 2, .external_lex_state = 22},
  [394] = {.lex_state = 2, .external_lex_state = 22},
  [395] = {.lex_state = 2, .external_lex_state = 22},
  [396] = {.lex_state = 2, .external_lex_state = 22},
  [397] = {.lex_state = 2, .external_lex_state = 22},
  [398] = {.lex_state = 2, .external_lex_state = 22},
  [399] = {.lex_state = 2, .external_lex_state = 22},
  [400] = {.lex_state = 2, .external_lex_state = 22},
  [401] = {.lex_state = 0, .external_lex_state = 23},
  [402] = {.lex_state = 2, .external_lex_state = 22},
  [403] = {.lex_state = 2, .external_lex_state = 22},
  [404] = {.lex_state = 2, .external_lex_state = 22},
  [405] = {.lex_state = 2, .external_lex_state = 22},
  [406] = {.lex_state = 2, .external_lex_state = 22},
  [407] = {.lex_state = 2, .external_lex_state = 22},
  [408] = {.lex_state = 2, .external_lex_state = 22},
  [409] = {.lex_state = 0, .external_lex_state = 23},
  [410] = {.lex_state = 2, .external_lex_state = 22},
  [411] = {.lex_state = 0, .external_lex_state = 23},
  [412] = {.lex_state = 0, .external_lex_state = 23},
  [413] = {.lex_state = 2, .external_lex_state = 22},
  [414] = {.lex_state = 2, .external_lex_state = 22},
  [415] = {.lex_state = 2, .external_lex_state = 22},
  [416] = {.lex_state = 2, .external_lex_state = 22},
  [417] = {.lex_state = 2, .external_lex_state = 22},
  [418] = {.lex_state = 2, .external_lex_state = 22},
  [419] = {.lex_state = 2, .external_lex_state = 22},
  [420] = {.lex_state = 2, .external_lex_state = 22},
  [421] = {.lex_state = 2, .external_lex_state = 22},
  [422] = {.lex_state = 2, .external_lex_state = 22},
  [423] = {.lex_state = 2, .external_lex_state = 22},
  [424] = {.lex_state = 2, .external_lex_state = 22},
  [425] = {.lex_state = 2, .external_lex_state = 22},
  [426] = {.lex_state = 2, .external_lex_state = 22},
  [427] = {.lex_state = 2, .external_lex_state = 22},
  [428] = {.lex_state = 2, .external_lex_state = 22},
  [429] = {.lex_state = 2, .external_lex_state = 22},
  [430] = {.lex_state = 2, .external_lex_state = 22},
  [431] = {.lex_state = 2, .external_lex_state = 22},
  [432] = {.lex_state = 2, .external_lex_state = 22},
  [433] = {.lex_state = 2, .external_lex_state = 22},
  [434] = {.lex_state = 2, .external_lex_state = 22},
  [435] = {.lex_state = 2, .external_lex_state = 22},
  [436] = {.lex_state = 2, .external_lex_state = 22},
  [437] = {.lex_state = 2, .external_lex_state = 22},
  [438] = {.lex_state = 2, .external_lex_state = 22},
  [439] = {.lex_state = 2, .external_lex_state = 22},
  [440] = {.lex_state = 2, .external_lex_state = 22},
  [441] = {.lex_state = 2, .external_lex_state = 22},
  [442] = {.lex_state = 2, .external_lex_state = 22},
  [443] = {.lex_state = 2, .external_lex_state = 22},
  [444] = {.lex_state = 2, .external_lex_state = 22},
  [445] = {.lex_state = 2, .external_lex_state = 22},
  [446] = {.lex_state = 2, .external_lex_state = 22},
  [447] = {.lex_state = 2, .external_lex_state = 22},
  [448] = {.lex_state = 2, .external_lex_state = 22},
  [449] = {.lex_state = 2, .external_lex_state = 22},
  [450] = {.lex_state = 2, .external_lex_state = 22},
  [451] = {.lex_state = 2, .external_lex_state = 22},
  [452] = {.lex_state = 2, .external_lex_state = 22},
  [453] = {.lex_state = 2, .external_lex_state = 22},
  [454] = {.lex_state = 2, .external_lex_state = 22},
  [455] = {.lex_state = 2, .external_lex_state = 22},
  [456] = {.lex_state = 2, .external_lex_state = 22},
  [457] = {.lex_state = 2, .external_lex_state = 22},
  [458] = {.lex_state = 2, .external_lex_state = 22},
  [459] = {.lex_state = 2, .external_lex_state = 22},
  [460] = {.lex_state = 2, .external_lex_state = 22},
  [461] = {.lex_state = 2, .external_lex_state = 22},
  [462] = {.lex_state = 2, .external_lex_state = 22},
  [463] = {.lex_state = 2, .external_lex_state = 22},
  [464] = {.lex_state = 2, .external_lex_state = 22},
  [465] = {.lex_state = 2, .external_lex_state = 22},
  [466] = {.lex_state = 2, .external_lex_state = 22},
  [467] = {.lex_state = 2, .external_lex_state = 22},
  [468] = {.lex_state = 2, .external_lex_state = 22},
  [469] = {.lex_state = 2, .external_lex_state = 22},
  [470] = {.lex_state = 2, .external_lex_state = 22},
  [471] = {.lex_state = 2, .external_lex_state = 22},
  [472] = {.lex_state = 2, .external_lex_state = 22},
  [473] = {.lex_state = 2, .external_lex_state = 22},
  [474] = {.lex_state = 2, .external_lex_state = 22},
  [475] = {.lex_state = 2, .external_lex_state = 22},
  [476] = {.lex_state = 2, .external_lex_state = 22},
  [477] = {.lex_state = 2, .external_lex_state = 22},
  [478] = {.lex_state = 2, .external_lex_state = 22},
  [479] = {.lex_state = 2, .external_lex_state = 22},
  [480] = {.lex_state = 2, .external_lex_state = 22},
  [481] = {.lex_state = 2, .external_lex_state = 22},
  [482] = {.lex_state = 2, .external_lex_state = 22},
  [483] = {.lex_state = 2, .external_lex_state = 22},
  [484] = {.lex_state = 2, .external_lex_state = 22},
  [485] = {.lex_state = 2, .external_lex_state = 22},
  [486] = {.lex_state = 2, .external_lex_state = 22},
  [487] = {.lex_state = 2, .external_lex_state = 22},
  [488] = {.lex_state = 2, .external_lex_state = 22},
  [489] = {.lex_state = 2, .external_lex_state = 22},
  [490] = {.lex_state = 2, .external_lex_state = 22},
  [491] = {.lex_state = 2, .external_lex_state = 22},
  [492] = {.lex_state = 2, .external_lex_state = 22},
  [493] = {.lex_state = 2, .external_lex_state = 22},
  [494] = {.lex_state = 2, .external_lex_state = 22},
  [495] = {.lex_state = 2, .external_lex_state = 22},
  [496] = {.lex_state = 2, .external_lex_state = 22},
  [497] = {.lex_state = 2, .external_lex_state = 22},
  [498] = {.lex_state = 2, .external_lex_state = 22},
  [499] = {.lex_state = 2, .external_lex_state = 22},
  [500] = {.lex_state = 2, .external_lex_state = 22},
  [501] = {.lex_state = 2, .external_lex_state = 22},
  [502] = {.lex_state = 2, .external_lex_state = 22},
  [503] = {.lex_state = 2, .external_lex_state = 22},
  [504] = {.lex_state = 2, .external_lex_state = 22},
  [505] = {.lex_state = 2, .external_lex_state = 22},
  [506] = {.lex_state = 2, .external_lex_state = 22},
  [507] = {.lex_state = 2, .external_lex_state = 22},
  [508] = {.lex_state = 2, .external_lex_state = 22},
  [509] = {.lex_state = 2, .external_lex_state = 22},
  [510] = {.lex_state = 2, .external_lex_state = 22},
  [511] = {.lex_state = 2, .external_lex_state = 22},
  [512] = {.lex_state = 2, .external_lex_state = 22},
  [513] = {.lex_state = 2, .external_lex_state = 22},
  [514] = {.lex_state = 2, .external_lex_state = 22},
  [515] = {.lex_state = 2, .external_lex_state = 22},
  [516] = {.lex_state = 2, .external_lex_state = 22},
  [517] = {.lex_state = 2, .external_lex_state = 22},
  [518] = {.lex_state = 2, .external_lex_state = 22},
  [519] = {.lex_state = 2, .external_lex_state = 22},
  [520] = {.lex_state = 2, .external_lex_state = 22},
  [521] = {.lex_state = 2, .external_lex_state = 22},
  [522] = {.lex_state = 2, .external_lex_state = 22},
  [523] = {.lex_state = 2, .external_lex_state = 22},
  [524] = {.lex_state = 2, .external_lex_state = 22},
  [525] = {.lex_state = 2, .external_lex_state = 22},
  [526] = {.lex_state = 2, .external_lex_state = 22},
  [527] = {.lex_state = 2, .external_lex_state = 22},
  [528] = {.lex_state = 2, .external_lex_state = 22},
  [529] = {.lex_state = 2, .external_lex_state = 22},
  [530] = {.lex_state = 2, .external_lex_state = 22},
  [531] = {.lex_state = 2, .external_lex_state = 22},
  [532] = {.lex_state = 2, .external_lex_state = 22},
  [533] = {.lex_state = 2, .external_lex_state = 22},
  [534] = {.lex_state = 2, .external_lex_state = 22},
  [535] = {.lex_state = 2, .external_lex_state = 22},
  [536] = {.lex_state = 2, .external_lex_state = 22},
  [537] = {.lex_state = 2, .external_lex_state = 22},
  [538] = {.lex_state = 2, .external_lex_state = 22},
  [539] = {.lex_state = 2, .external_lex_state = 22},
  [540] = {.lex_state = 2, .external_lex_state = 22},
  [541] = {.lex_state = 2, .external_lex_state = 22},
  [542] = {.lex_state = 2, .external_lex_state = 22},
  [543] = {.lex_state = 2, .external_lex_state = 22},
  [544] = {.lex_state = 2, .external_lex_state = 22},
  [545] = {.lex_state = 2, .external_lex_state = 22},
  [546] = {.lex_state = 2, .external_lex_state = 22},
  [547] = {.lex_state = 2, .external_lex_state = 22},
  [548] = {.lex_state = 2, .external_lex_state = 22},
  [549] = {.lex_state = 2, .external_lex_state = 22},
  [550] = {.lex_state = 2, .external_lex_state = 22},
  [551] = {.lex_state = 2, .external_lex_state = 22},
  [552] = {.lex_state = 2, .external_lex_state = 22},
  [553] = {.lex_state = 2, .external_lex_state = 22},
  [554] = {.lex_state = 2, .external_lex_state = 22},
  [555] = {.lex_state = 2, .external_lex_state = 22},
  [556] = {.lex_state = 2, .external_lex_state = 22},
  [557] = {.lex_state = 2, .external_lex_state = 22},
  [558] = {.lex_state = 2, .external_lex_state = 22},
  [559] = {.lex_state = 2, .external_lex_state = 22},
  [560] = {.lex_state = 2, .external_lex_state = 22},
  [561] = {.lex_state = 2, .external_lex_state = 22},
  [562] = {.lex_state = 2, .external_lex_state = 22},
  [563] = {.lex_state = 2, .external_lex_state = 22},
  [564] = {.lex_state = 2, .external_lex_state = 22},
  [565] = {.lex_state = 2, .external_lex_state = 22},
  [566] = {.lex_state = 2, .external_lex_state = 22},
  [567] = {.lex_state = 2, .external_lex_state = 22},
  [568] = {.lex_state = 2, .external_lex_state = 22},
  [569] = {.lex_state = 2, .external_lex_state = 22},
  [570] = {.lex_state = 2, .external_lex_state = 22},
  [571] = {.lex_state = 2, .external_lex_state = 22},
  [572] = {.lex_state = 2, .external_lex_state = 22},
  [573] = {.lex_state = 2, .external_lex_state = 22},
  [574] = {.lex_state = 2, .external_lex_state = 22},
  [575] = {.lex_state = 2, .external_lex_state = 22},
  [576] = {.lex_state = 2, .external_lex_state = 22},
  [577] = {.lex_state = 2, .external_lex_state = 22},
  [578] = {.lex_state = 2, .external_lex_state = 22},
  [579] = {.lex_state = 2, .external_lex_state = 22},
  [580] = {.lex_state = 2, .external_lex_state = 22},
  [581] = {.lex_state = 2, .external_lex_state = 22},
  [582] = {.lex_state = 2, .external_lex_state = 22},
  [583] = {.lex_state = 2, .external_lex_state = 22},
  [584] = {.lex_state = 2, .external_lex_state = 22},
  [585] = {.lex_state = 2, .external_lex_state = 22},
  [586] = {.lex_state = 2, .external_lex_state = 22},
  [587] = {.lex_state = 2, .external_lex_state = 22},
  [588] = {.lex_state = 2, .external_lex_state = 22},
  [589] = {.lex_state = 2, .external_lex_state = 22},
  [590] = {.lex_state = 2, .external_lex_state = 22},
  [591] = {.lex_state = 2, .external_lex_state = 22},
  [592] = {.lex_state = 2, .external_lex_state = 22},
  [593] = {.lex_state = 2, .external_lex_state = 22},
  [594] = {.lex_state = 2, .external_lex_state = 22},
  [595] = {.lex_state = 2, .external_lex_state = 22},
  [596] = {.lex_state = 2, .external_lex_state = 22},
  [597] = {.lex_state = 2, .external_lex_state = 22},
  [598] = {.lex_state = 2, .external_lex_state = 22},
  [599] = {.lex_state = 2, .external_lex_state = 22},
  [600] = {.lex_state = 2, .external_lex_state = 22},
  [601] = {.lex_state = 2, .external_lex_state = 22},
  [602] = {.lex_state = 2, .external_lex_state = 22},
  [603] = {.lex_state = 2, .external_lex_state = 22},
  [604] = {.lex_state = 2, .external_lex_state = 22},
  [605] = {.lex_state = 2, .external_lex_state = 22},
  [606] = {.lex_state = 2, .external_lex_state = 22},
  [607] = {.lex_state = 2, .external_lex_state = 22},
  [608] = {.lex_state = 2, .external_lex_state = 22},
  [609] = {.lex_state = 2, .external_lex_state = 22},
  [610] = {.lex_state = 2, .external_lex_state = 22},
  [611] = {.lex_state = 2, .external_lex_state = 22},
  [612] = {.lex_state = 2, .external_lex_state = 22},
  [613] = {.lex_state = 2, .external_lex_state = 22},
  [614] = {.lex_state = 2, .external_lex_state = 22},
  [615] = {.lex_state = 2, .external_lex_state = 22},
  [616] = {.lex_state = 2, .external_lex_state = 22},
  [617] = {.lex_state = 2, .external_lex_state = 22},
  [618] = {.lex_state = 2, .external_lex_state = 22},
  [619] = {.lex_state = 2, .external_lex_state = 22},
  [620] = {.lex_state = 2, .external_lex_state = 22},
  [621] = {.lex_state = 2, .external_lex_state = 22},
  [622] = {.lex_state = 2, .external_lex_state = 22},
  [623] = {.lex_state = 2, .external_lex_state = 22},
  [624] = {.lex_state = 2, .external_lex_state = 22},
  [625] = {.lex_state = 2, .external_lex_state = 22},
  [626] = {.lex_state = 2, .external_lex_state = 22},
  [627] = {.lex_state = 2, .external_lex_state = 22},
  [628] = {.lex_state = 2, .external_lex_state = 22},
  [629] = {.lex_state = 2, .external_lex_state = 22},
  [630] = {.lex_state = 2, .external_lex_state = 22},
  [631] = {.lex_state = 2, .external_lex_state = 22},
  [632] = {.lex_state = 2, .external_lex_state = 22},
  [633] = {.lex_state = 2, .external_lex_state = 22},
  [634] = {.lex_state = 2, .external_lex_state = 22},
  [635] = {.lex_state = 2, .external_lex_state = 22},
  [636] = {.lex_state = 2, .external_lex_state = 22},
  [637] = {.lex_state = 2, .external_lex_state = 22},
  [638] = {.lex_state = 2, .external_lex_state = 22},
  [639] = {.lex_state = 2, .external_lex_state = 22},
  [640] = {.lex_state = 2, .external_lex_state = 22},
  [641] = {.lex_state = 2, .external_lex_state = 22},
  [642] = {.lex_state = 2, .external_lex_state = 22},
  [643] = {.lex_state = 2, .external_lex_state = 22},
  [644] = {.lex_state = 2, .external_lex_state = 22},
  [645] = {.lex_state = 2, .external_lex_state = 22},
  [646] = {.lex_state = 2, .external_lex_state = 22},
  [647] = {.lex_state = 2, .external_lex_state = 22},
  [648] = {.lex_state = 2, .external_lex_state = 22},
  [649] = {.lex_state = 2, .external_lex_state = 22},
  [650] = {.lex_state = 2, .external_lex_state = 22},
  [651] = {.lex_state = 2, .external_lex_state = 22},
  [652] = {.lex_state = 2, .external_lex_state = 22},
  [653] = {.lex_state = 2, .external_lex_state = 22},
  [654] = {.lex_state = 2, .external_lex_state = 22},
  [655] = {.lex_state = 2, .external_lex_state = 22},
  [656] = {.lex_state = 2, .external_lex_state = 22},
  [657] = {.lex_state = 2, .external_lex_state = 22},
  [658] = {.lex_state = 2, .external_lex_state = 22},
  [659] = {.lex_state = 2, .external_lex_state = 22},
  [660] = {.lex_state = 2, .external_lex_state = 22},
  [661] = {.lex_state = 2, .external_lex_state = 22},
  [662] = {.lex_state = 2, .external_lex_state = 22},
  [663] = {.lex_state = 2, .external_lex_state = 22},
  [664] = {.lex_state = 2, .external_lex_state = 22},
  [665] = {.lex_state = 2, .external_lex_state = 22},
  [666] = {.lex_state = 2, .external_lex_state = 22},
  [667] = {.lex_state = 2, .external_lex_state = 22},
  [668] = {.lex_state = 2, .external_lex_state = 22},
  [669] = {.lex_state = 2, .external_lex_state = 22},
  [670] = {.lex_state = 2, .external_lex_state = 22},
  [671] = {.lex_state = 2, .external_lex_state = 22},
  [672] = {.lex_state = 2, .external_lex_state = 22},
  [673] = {.lex_state = 2, .external_lex_state = 22},
  [674] = {.lex_state = 2, .external_lex_state = 22},
  [675] = {.lex_state = 2, .external_lex_state = 22},
  [676] = {.lex_state = 2, .external_lex_state = 22},
  [677] = {.lex_state = 2, .external_lex_state = 22},
  [678] = {.lex_state = 2, .external_lex_state = 22},
  [679] = {.lex_state = 2, .external_lex_state = 22},
  [680] = {.lex_state = 2, .external_lex_state = 22},
  [681] = {.lex_state = 2, .external_lex_state = 22},
  [682] = {.lex_state = 2, .external_lex_state = 22},
  [683] = {.lex_state = 2, .external_lex_state = 22},
  [684] = {.lex_state = 2, .external_lex_state = 22},
  [685] = {.lex_state = 2, .external_lex_state = 22},
  [686] = {.lex_state = 2, .external_lex_state = 22},
  [687] = {.lex_state = 2, .external_lex_state = 22},
  [688] = {.lex_state = 2, .external_lex_state = 22},
  [689] = {.lex_state = 2, .external_lex_state = 22},
  [690] = {.lex_state = 2, .external_lex_state = 22},
  [691] = {.lex_state = 2, .external_lex_state = 22},
  [692] = {.lex_state = 2, .external_lex_state = 22},
  [693] = {.lex_state = 2, .external_lex_state = 22},
  [694] = {.lex_state = 2, .external_lex_state = 22},
  [695] = {.lex_state = 2, .external_lex_state = 22},
  [696] = {.lex_state = 2, .external_lex_state = 22},
  [697] = {.lex_state = 2, .external_lex_state = 22},
  [698] = {.lex_state = 2, .external_lex_state = 22},
  [699] = {.lex_state = 2, .external_lex_state = 22},
  [700] = {.lex_state = 2, .external_lex_state = 22},
  [701] = {.lex_state = 2, .external_lex_state = 22},
  [702] = {.lex_state = 2, .external_lex_state = 22},
  [703] = {.lex_state = 2, .external_lex_state = 22},
  [704] = {.lex_state = 2, .external_lex_state = 22},
  [705] = {.lex_state = 2, .external_lex_state = 22},
  [706] = {.lex_state = 2, .external_lex_state = 22},
  [707] = {.lex_state = 2, .external_lex_state = 22},
  [708] = {.lex_state = 2, .external_lex_state = 22},
  [709] = {.lex_state = 2, .external_lex_state = 22},
  [710] = {.lex_state = 2, .external_lex_state = 22},
  [711] = {.lex_state = 2, .external_lex_state = 22},
  [712] = {.lex_state = 2, .external_lex_state = 22},
  [713] = {.lex_state = 2, .external_lex_state = 22},
  [714] = {.lex_state = 2, .external_lex_state = 22},
  [715] = {.lex_state = 2, .external_lex_state = 22},
  [716] = {.lex_state = 2, .external_lex_state = 22},
  [717] = {.lex_state = 2, .external_lex_state = 22},
  [718] = {.lex_state = 2, .external_lex_state = 22},
  [719] = {.lex_state = 2, .external_lex_state = 22},
  [720] = {.lex_state = 2, .external_lex_state = 22},
  [721] = {.lex_state = 1, .external_lex_state = 24},
  [722] = {.lex_state = 0, .external_lex_state = 23},
  [723] = {.lex_state = 1, .external_lex_state = 24},
  [724] = {.lex_state = 0, .external_lex_state = 23},
  [725] = {.lex_state = 1, .external_lex_state = 24},
  [726] = {.lex_state = 0, .external_lex_state = 23},
  [727] = {.lex_state = 0, .external_lex_state = 23},
  [728] = {.lex_state = 1, .external_lex_state = 25},
  [729] = {.lex_state = 1, .external_lex_state = 25},
  [730] = {.lex_state = 1, .external_lex_state = 25},
  [731] = {.lex_state = 1, .external_lex_state = 26},
  [732] = {.lex_state = 1, .external_lex_state = 26},
  [733] = {.lex_state = 1, .external_lex_state = 26},
  [734] = {.lex_state = 1, .external_lex_state = 27},
  [735] = {.lex_state = 1, .external_lex_state = 27},
  [736] = {.lex_state = 1, .external_lex_state = 27},
  [737] = {.lex_state = 0, .external_lex_state = 28},
  [738] = {.lex_state = 1, .external_lex_state = 29},
  [739] = {.lex_state = 1, .external_lex_state = 30},
  [740] = {.lex_state = 2, .external_lex_state = 2},
  [741] = {.lex_state = 1, .external_lex_state = 30},
  [742] = {.lex_state = 1, .external_lex_state = 31},
  [743] = {.lex_state = 1, .external_lex_state = 30},
  [744] = {.lex_state = 0, .external_lex_state = 28},
  [745] = {.lex_state = 0, .external_lex_state = 28},
  [746] = {.lex_state = 1, .external_lex_state = 31},
  [747] = {.lex_state = 1, .external_lex_state = 31},
  [748] = {.lex_state = 1, .external_lex_state = 29},
  [749] = {.lex_state = 1, .external_lex_state = 29},
  [750] = {.lex_state = 1, .external_lex_state = 32},
  [751] = {.lex_state = 1, .external_lex_state = 32},
  [752] = {.lex_state = 1, .external_lex_state = 33},
  [753] = {.lex_state = 1, .external_lex_state = 32},
  [754] = {.lex_state = 1, .external_lex_state = 33},
  [755] = {.lex_state = 1, .external_lex_state = 33},
  [756] = {.lex_state = 1, .external_lex_state = 24},
  [757] = {.lex_state = 0, .external_lex_state = 34},
  [758] = {.lex_state = 1, .external_lex_state = 35},
  [759] = {.lex_state = 1, .external_lex_state = 36},
  [760] = {.lex_state = 1, .external_lex_state = 37},
  [761] = {.lex_state = 0, .external_lex_state = 34},
  [762] = {.lex_state = 0, .external_lex_state = 34},
  [763] = {.lex_state = 1, .external_lex_state = 24},
  [764] = {.lex_state = 1, .external_lex_state = 37},
  [765] = {.lex_state = 0, .external_lex_state = 34},
  [766] = {.lex_state = 1, .external_lex_state = 38},
  [767] = {.lex_state = 1, .external_lex_state = 36},
  [768] = {.lex_state = 1, .external_lex_state = 36},
  [769] = {.lex_state = 1, .external_lex_state = 38},
  [770] = {.lex_state = 1, .external_lex_state = 38},
  [771] = {.lex_state = 1, .external_lex_state = 39},
  [772] = {.lex_state = 1, .external_lex_state = 35},
  [773] = {.lex_state = 0, .external_lex_state = 34},
  [774] = {.lex_state = 1, .external_lex_state = 39},
  [775] = {.lex_state = 0, .external_lex_state = 34},
  [776] = {.lex_state = 0, .external_lex_state = 34},
  [777] = {.lex_state = 1, .external_lex_state = 39},
  [778] = {.lex_state = 1, .external_lex_state = 35},
  [779] = {.lex_state = 1, .external_lex_state = 37},
  [780] = {.lex_state = 1, .external_lex_state = 40},
  [781] = {.lex_state = 1, .external_lex_state = 41},
  [782] = {.lex_state = 1, .external_lex_state = 40},
  [783] = {.lex_state = 1, .external_lex_state = 42},
  [784] = {.lex_state = 0, .external_lex_state = 43},
  [785] = {.lex_state = 1, .external_lex_state = 44},
  [786] = {.lex_state = 1, .external_lex_state = 42},
  [787] = {.lex_state = 1, .external_lex_state = 45},
  [788] = {.lex_state = 1, .external_lex_state = 45},
  [789] = {.lex_state = 1, .external_lex_state = 40},
  [790] = {.lex_state = 1, .external_lex_state = 45},
  [791] = {.lex_state = 1, .external_lex_state = 41},
  [792] = {.lex_state = 1, .external_lex_state = 42},
  [793] = {.lex_state = 1, .external_lex_state = 44},
  [794] = {.lex_state = 1, .external_lex_state = 44},
  [795] = {.lex_state = 1, .external_lex_state = 41},
  [796] = {.lex_state = 0, .external_lex_state = 46},
  [797] = {.lex_state = 1, .external_lex_state = 47},
  [798] = {.lex_state = 1, .external_lex_state = 48},
  [799] = {.lex_state = 1, .external_lex_state = 49},
  [800] = {.lex_state = 1, .external_lex_state = 50},
  [801] = {.lex_state = 0, .external_lex_state = 46},
  [802] = {.lex_state = 0, .external_lex_state = 46},
  [803] = {.lex_state = 1, .external_lex_state = 47},
  [804] = {.lex_state = 0, .external_lex_state = 46},
  [805] = {.lex_state = 1, .external_lex_state = 24},
  [806] = {.lex_state = 1, .external_lex_state = 47},
  [807] = {.lex_state = 0, .external_lex_state = 46},
  [808] = {.lex_state = 1, .external_lex_state = 48},
  [809] = {.lex_state = 1, .external_lex_state = 48},
  [810] = {.lex_state = 0, .external_lex_state = 46},
  [811] = {.lex_state = 0, .external_lex_state = 46},
  [812] = {.lex_state = 1, .external_lex_state = 50},
  [813] = {.lex_state = 1, .external_lex_state = 50},
  [814] = {.lex_state = 0, .external_lex_state = 46},
  [815] = {.lex_state = 0, .external_lex_state = 46},
  [816] = {.lex_state = 0, .external_lex_state = 51},
  [817] = {.lex_state = 0, .external_lex_state = 46},
  [818] = {.lex_state = 0, .external_lex_state = 46},
  [819] = {.lex_state = 0, .external_lex_state = 46},
  [820] = {.lex_state = 1, .external_lex_state = 52},
  [821] = {.lex_state = 0, .external_lex_state = 46},
  [822] = {.lex_state = 1, .external_lex_state = 26},
  [823] = {.lex_state = 1, .external_lex_state = 26},
  [824] = {.lex_state = 0, .external_lex_state = 46},
  [825] = {.lex_state = 0, .external_lex_state = 46},
  [826] = {.lex_state = 1, .external_lex_state = 52},
  [827] = {.lex_state = 0, .external_lex_state = 51},
  [828] = {.lex_state = 1, .external_lex_state = 52},
  [829] = {.lex_state = 0, .external_lex_state = 46},
  [830] = {.lex_state = 1, .external_lex_state = 49},
  [831] = {.lex_state = 0, .external_lex_state = 51},
  [832] = {.lex_state = 1, .external_lex_state = 49},
  [833] = {.lex_state = 0, .external_lex_state = 51},
  [834] = {.lex_state = 0, .external_lex_state = 51},
  [835] = {.lex_state = 0, .external_lex_state = 51},
  [836] = {.lex_state = 0, .external_lex_state = 46},
  [837] = {.lex_state = 1, .external_lex_state = 53},
  [838] = {.lex_state = 1, .external_lex_state = 24},
  [839] = {.lex_state = 1, .external_lex_state = 24},
  [840] = {.lex_state = 1, .external_lex_state = 24},
  [841] = {.lex_state = 1, .external_lex_state = 53},
  [842] = {.lex_state = 1, .external_lex_state = 54},
  [843] = {.lex_state = 1, .external_lex_state = 55},
  [844] = {.lex_state = 1, .external_lex_state = 24},
  [845] = {.lex_state = 1, .external_lex_state = 54},
  [846] = {.lex_state = 1, .external_lex_state = 56},
  [847] = {.lex_state = 1, .external_lex_state = 24},
  [848] = {.lex_state = 1, .external_lex_state = 24},
  [849] = {.lex_state = 1, .external_lex_state = 57},
  [850] = {.lex_state = 1, .external_lex_state = 25},
  [851] = {.lex_state = 1, .external_lex_state = 57},
  [852] = {.lex_state = 1, .external_lex_state = 24},
  [853] = {.lex_state = 1, .external_lex_state = 58},
  [854] = {.lex_state = 1, .external_lex_state = 24},
  [855] = {.lex_state = 1, .external_lex_state = 24},
  [856] = {.lex_state = 1, .external_lex_state = 24},
  [857] = {.lex_state = 1, .external_lex_state = 58},
  [858] = {.lex_state = 1, .external_lex_state = 24},
  [859] = {.lex_state = 1, .external_lex_state = 24},
  [860] = {.lex_state = 1, .external_lex_state = 24},
  [861] = {.lex_state = 0, .external_lex_state = 59},
  [862] = {.lex_state = 1, .external_lex_state = 58},
  [863] = {.lex_state = 1, .external_lex_state = 24},
  [864] = {.lex_state = 1, .external_lex_state = 24},
  [865] = {.lex_state = 1, .external_lex_state = 24},
  [866] = {.lex_state = 1, .external_lex_state = 24},
  [867] = {.lex_state = 1, .external_lex_state = 24},
  [868] = {.lex_state = 1, .external_lex_state = 60},
  [869] = {.lex_state = 1, .external_lex_state = 24},
  [870] = {.lex_state = 1, .external_lex_state = 60},
  [871] = {.lex_state = 1, .external_lex_state = 24},
  [872] = {.lex_state = 1, .external_lex_state = 55},
  [873] = {.lex_state = 1, .external_lex_state = 57},
  [874] = {.lex_state = 1, .external_lex_state = 24},
  [875] = {.lex_state = 1, .external_lex_state = 24},
  [876] = {.lex_state = 1, .external_lex_state = 54},
  [877] = {.lex_state = 1, .external_lex_state = 24},
  [878] = {.lex_state = 1, .external_lex_state = 24},
  [879] = {.lex_state = 1, .external_lex_state = 24},
  [880] = {.lex_state = 1, .external_lex_state = 24},
  [881] = {.lex_state = 1, .external_lex_state = 24},
  [882] = {.lex_state = 1, .external_lex_state = 60},
  [883] = {.lex_state = 1, .external_lex_state = 55},
  [884] = {.lex_state = 1, .external_lex_state = 53},
  [885] = {.lex_state = 1, .external_lex_state = 24},
  [886] = {.lex_state = 1, .external_lex_state = 24},
  [887] = {.lex_state = 1, .external_lex_state = 25},
  [888] = {.lex_state = 1, .external_lex_state = 26},
  [889] = {.lex_state = 1, .external_lex_state = 27},
  [890] = {.lex_state = 1, .external_lex_state = 25},
  [891] = {.lex_state = 1, .external_lex_state = 25},
  [892] = {.lex_state = 1, .external_lex_state = 25},
  [893] = {.lex_state = 1, .external_lex_state = 56},
  [894] = {.lex_state = 1, .external_lex_state = 25},
  [895] = {.lex_state = 1, .external_lex_state = 61},
  [896] = {.lex_state = 1, .external_lex_state = 25},
  [897] = {.lex_state = 1, .external_lex_state = 25},
  [898] = {.lex_state = 1, .external_lex_state = 61},
  [899] = {.lex_state = 1, .external_lex_state = 25},
  [900] = {.lex_state = 1, .external_lex_state = 25},
  [901] = {.lex_state = 1, .external_lex_state = 61},
  [902] = {.lex_state = 1, .external_lex_state = 56},
  [903] = {.lex_state = 1, .external_lex_state = 25},
  [904] = {.lex_state = 1, .external_lex_state = 25},
  [905] = {.lex_state = 1, .external_lex_state = 25},
  [906] = {.lex_state = 1, .external_lex_state = 25},
  [907] = {.lex_state = 1, .external_lex_state = 25},
  [908] = {.lex_state = 1, .external_lex_state = 25},
  [909] = {.lex_state = 1, .external_lex_state = 61},
  [910] = {.lex_state = 1, .external_lex_state = 25},
  [911] = {.lex_state = 1, .external_lex_state = 25},
  [912] = {.lex_state = 1, .external_lex_state = 25},
  [913] = {.lex_state = 1, .external_lex_state = 25},
  [914] = {.lex_state = 1, .external_lex_state = 25},
  [915] = {.lex_state = 1, .external_lex_state = 25},
  [916] = {.lex_state = 1, .external_lex_state = 61},
  [917] = {.lex_state = 1, .external_lex_state = 25},
  [918] = {.lex_state = 1, .external_lex_state = 61},
  [919] = {.lex_state = 1, .external_lex_state = 61},
  [920] = {.lex_state = 1, .external_lex_state = 61},
  [921] = {.lex_state = 1, .external_lex_state = 25},
  [922] = {.lex_state = 1, .external_lex_state = 25},
  [923] = {.lex_state = 1, .external_lex_state = 25},
  [924] = {.lex_state = 1, .external_lex_state = 25},
  [925] = {.lex_state = 1, .external_lex_state = 25},
  [926] = {.lex_state = 1, .external_lex_state = 25},
  [927] = {.lex_state = 1, .external_lex_state = 61},
  [928] = {.lex_state = 1, .external_lex_state = 25},
  [929] = {.lex_state = 1, .external_lex_state = 27},
  [930] = {.lex_state = 1, .external_lex_state = 26},
  [931] = {.lex_state = 1, .external_lex_state = 26},
  [932] = {.lex_state = 1, .external_lex_state = 26},
  [933] = {.lex_state = 1, .external_lex_state = 26},
  [934] = {.lex_state = 1, .external_lex_state = 27},
  [935] = {.lex_state = 1, .external_lex_state = 26},
  [936] = {.lex_state = 1, .external_lex_state = 26},
  [937] = {.lex_state = 1, .external_lex_state = 26},
  [938] = {.lex_state = 1, .external_lex_state = 26},
  [939] = {.lex_state = 1, .external_lex_state = 26},
  [940] = {.lex_state = 1, .external_lex_state = 26},
  [941] = {.lex_state = 0, .external_lex_state = 62},
  [942] = {.lex_state = 1, .external_lex_state = 27},
  [943] = {.lex_state = 1, .external_lex_state = 26},
  [944] = {.lex_state = 0, .external_lex_state = 62},
  [945] = {.lex_state = 0, .external_lex_state = 62},
  [946] = {.lex_state = 1, .external_lex_state = 26},
  [947] = {.lex_state = 1, .external_lex_state = 26},
  [948] = {.lex_state = 1, .external_lex_state = 26},
  [949] = {.lex_state = 1, .external_lex_state = 26},
  [950] = {.lex_state = 1, .external_lex_state = 26},
  [951] = {.lex_state = 1, .external_lex_state = 26},
  [952] = {.lex_state = 1, .external_lex_state = 26},
  [953] = {.lex_state = 1, .external_lex_state = 26},
  [954] = {.lex_state = 1, .external_lex_state = 26},
  [955] = {.lex_state = 1, .external_lex_state = 26},
  [956] = {.lex_state = 1, .external_lex_state = 31},
  [957] = {.lex_state = 1, .external_lex_state = 26},
  [958] = {.lex_state = 0, .external_lex_state = 62},
  [959] = {.lex_state = 1, .external_lex_state = 26},
  [960] = {.lex_state = 0, .external_lex_state = 62},
  [961] = {.lex_state = 0, .external_lex_state = 62},
  [962] = {.lex_state = 1, .external_lex_state = 26},
  [963] = {.lex_state = 1, .external_lex_state = 26},
  [964] = {.lex_state = 1, .external_lex_state = 26},
  [965] = {.lex_state = 1, .external_lex_state = 63},
  [966] = {.lex_state = 1, .external_lex_state = 27},
  [967] = {.lex_state = 1, .external_lex_state = 27},
  [968] = {.lex_state = 1, .external_lex_state = 27},
  [969] = {.lex_state = 1, .external_lex_state = 27},
  [970] = {.lex_state = 1, .external_lex_state = 27},
  [971] = {.lex_state = 1, .external_lex_state = 27},
  [972] = {.lex_state = 1, .external_lex_state = 27},
  [973] = {.lex_state = 1, .external_lex_state = 27},
  [974] = {.lex_state = 1, .external_lex_state = 27},
  [975] = {.lex_state = 1, .external_lex_state = 27},
  [976] = {.lex_state = 1, .external_lex_state = 27},
  [977] = {.lex_state = 1, .external_lex_state = 27},
  [978] = {.lex_state = 1, .external_lex_state = 27},
  [979] = {.lex_state = 1, .external_lex_state = 27},
  [980] = {.lex_state = 1, .external_lex_state = 27},
  [981] = {.lex_state = 1, .external_lex_state = 27},
  [982] = {.lex_state = 1, .external_lex_state = 27},
  [983] = {.lex_state = 1, .external_lex_state = 27},
  [984] = {.lex_state = 1, .external_lex_state = 27},
  [985] = {.lex_state = 1, .external_lex_state = 27},
  [986] = {.lex_state = 1, .external_lex_state = 27},
  [987] = {.lex_state = 1, .external_lex_state = 27},
  [988] = {.lex_state = 1, .external_lex_state = 27},
  [989] = {.lex_state = 1, .external_lex_state = 26},
  [990] = {.lex_state = 1, .external_lex_state = 26},
  [991] = {.lex_state = 1, .external_lex_state = 27},
  [992] = {.lex_state = 1, .external_lex_state = 29},
  [993] = {.lex_state = 1, .external_lex_state = 27},
  [994] = {.lex_state = 1, .external_lex_state = 27},
  [995] = {.lex_state = 1, .external_lex_state = 30},
  [996] = {.lex_state = 1, .external_lex_state = 26},
  [997] = {.lex_state = 1, .external_lex_state = 64},
  [998] = {.lex_state = 1, .external_lex_state = 65},
  [999] = {.lex_state = 1, .external_lex_state = 30},
  [1000] = {.lex_state = 1, .external_lex_state = 31},
  [1001] = {.lex_state = 1, .external_lex_state = 31},
  [1002] = {.lex_state = 1, .external_lex_state = 31},
  [1003] = {.lex_state = 1, .external_lex_state = 29},
  [1004] = {.lex_state = 1, .external_lex_state = 31},
  [1005] = {.lex_state = 1, .external_lex_state = 31},
  [1006] = {.lex_state = 1, .external_lex_state = 31},
  [1007] = {.lex_state = 1, .external_lex_state = 31},
  [1008] = {.lex_state = 1, .external_lex_state = 31},
  [1009] = {.lex_state = 1, .external_lex_state = 31},
  [1010] = {.lex_state = 1, .external_lex_state = 65},
  [1011] = {.lex_state = 1, .external_lex_state = 31},
  [1012] = {.lex_state = 1, .external_lex_state = 31},
  [1013] = {.lex_state = 1, .external_lex_state = 31},
  [1014] = {.lex_state = 1, .external_lex_state = 31},
  [1015] = {.lex_state = 1, .external_lex_state = 31},
  [1016] = {.lex_state = 1, .external_lex_state = 31},
  [1017] = {.lex_state = 1, .external_lex_state = 31},
  [1018] = {.lex_state = 1, .external_lex_state = 33},
  [1019] = {.lex_state = 2, .external_lex_state = 2},
  [1020] = {.lex_state = 1, .external_lex_state = 31},
  [1021] = {.lex_state = 1, .external_lex_state = 30},
  [1022] = {.lex_state = 1, .external_lex_state = 65},
  [1023] = {.lex_state = 1, .external_lex_state = 30},
  [1024] = {.lex_state = 1, .external_lex_state = 30},
  [1025] = {.lex_state = 0, .external_lex_state = 66},
  [1026] = {.lex_state = 0, .external_lex_state = 66},
  [1027] = {.lex_state = 1, .external_lex_state = 31},
  [1028] = {.lex_state = 1, .external_lex_state = 31},
  [1029] = {.lex_state = 1, .external_lex_state = 31},
  [1030] = {.lex_state = 1, .external_lex_state = 31},
  [1031] = {.lex_state = 1, .external_lex_state = 31},
  [1032] = {.lex_state = 1, .external_lex_state = 31},
  [1033] = {.lex_state = 1, .external_lex_state = 31},
  [1034] = {.lex_state = 1, .external_lex_state = 31},
  [1035] = {.lex_state = 1, .external_lex_state = 31},
  [1036] = {.lex_state = 1, .external_lex_state = 30},
  [1037] = {.lex_state = 1, .external_lex_state = 30},
  [1038] = {.lex_state = 1, .external_lex_state = 30},
  [1039] = {.lex_state = 1, .external_lex_state = 65},
  [1040] = {.lex_state = 1, .external_lex_state = 65},
  [1041] = {.lex_state = 2, .external_lex_state = 67},
  [1042] = {.lex_state = 1, .external_lex_state = 31},
  [1043] = {.lex_state = 2, .external_lex_state = 67},
  [1044] = {.lex_state = 1, .external_lex_state = 64},
  [1045] = {.lex_state = 1, .external_lex_state = 30},
  [1046] = {.lex_state = 1, .external_lex_state = 31},
  [1047] = {.lex_state = 1, .external_lex_state = 30},
  [1048] = {.lex_state = 1, .external_lex_state = 64},
  [1049] = {.lex_state = 1, .external_lex_state = 32},
  [1050] = {.lex_state = 1, .external_lex_state = 65},
  [1051] = {.lex_state = 1, .external_lex_state = 68},
  [1052] = {.lex_state = 1, .external_lex_state = 68},
  [1053] = {.lex_state = 1, .external_lex_state = 68},
  [1054] = {.lex_state = 1, .external_lex_state = 68},
  [1055] = {.lex_state = 1, .external_lex_state = 68},
  [1056] = {.lex_state = 1, .external_lex_state = 69},
  [1057] = {.lex_state = 1, .external_lex_state = 65},
  [1058] = {.lex_state = 1, .external_lex_state = 68},
  [1059] = {.lex_state = 1, .external_lex_state = 65},
  [1060] = {.lex_state = 1, .external_lex_state = 70},
  [1061] = {.lex_state = 1, .external_lex_state = 30},
  [1062] = {.lex_state = 1, .external_lex_state = 30},
  [1063] = {.lex_state = 1, .external_lex_state = 30},
  [1064] = {.lex_state = 1, .external_lex_state = 65},
  [1065] = {.lex_state = 1, .external_lex_state = 68},
  [1066] = {.lex_state = 1, .external_lex_state = 68},
  [1067] = {.lex_state = 1, .external_lex_state = 68},
  [1068] = {.lex_state = 1, .external_lex_state = 30},
  [1069] = {.lex_state = 1, .external_lex_state = 30},
  [1070] = {.lex_state = 1, .external_lex_state = 63},
  [1071] = {.lex_state = 1, .external_lex_state = 71},
  [1072] = {.lex_state = 0, .external_lex_state = 66},
  [1073] = {.lex_state = 1, .external_lex_state = 29},
  [1074] = {.lex_state = 1, .external_lex_state = 29},
  [1075] = {.lex_state = 0, .external_lex_state = 66},
  [1076] = {.lex_state = 1, .external_lex_state = 65},
  [1077] = {.lex_state = 1, .external_lex_state = 29},
  [1078] = {.lex_state = 1, .external_lex_state = 29},
  [1079] = {.lex_state = 1, .external_lex_state = 29},
  [1080] = {.lex_state = 1, .external_lex_state = 29},
  [1081] = {.lex_state = 1, .external_lex_state = 30},
  [1082] = {.lex_state = 1, .external_lex_state = 30},
  [1083] = {.lex_state = 1, .external_lex_state = 29},
  [1084] = {.lex_state = 1, .external_lex_state = 29},
  [1085] = {.lex_state = 1, .external_lex_state = 29},
  [1086] = {.lex_state = 1, .external_lex_state = 29},
  [1087] = {.lex_state = 1, .external_lex_state = 29},
  [1088] = {.lex_state = 1, .external_lex_state = 29},
  [1089] = {.lex_state = 1, .external_lex_state = 29},
  [1090] = {.lex_state = 1, .external_lex_state = 29},
  [1091] = {.lex_state = 1, .external_lex_state = 30},
  [1092] = {.lex_state = 1, .external_lex_state = 29},
  [1093] = {.lex_state = 1, .external_lex_state = 29},
  [1094] = {.lex_state = 1, .external_lex_state = 29},
  [1095] = {.lex_state = 1, .external_lex_state = 29},
  [1096] = {.lex_state = 1, .external_lex_state = 29},
  [1097] = {.lex_state = 1, .external_lex_state = 29},
  [1098] = {.lex_state = 1, .external_lex_state = 29},
  [1099] = {.lex_state = 1, .external_lex_state = 29},
  [1100] = {.lex_state = 1, .external_lex_state = 29},
  [1101] = {.lex_state = 1, .external_lex_state = 29},
  [1102] = {.lex_state = 1, .external_lex_state = 72},
  [1103] = {.lex_state = 1, .external_lex_state = 29},
  [1104] = {.lex_state = 1, .external_lex_state = 30},
  [1105] = {.lex_state = 1, .external_lex_state = 29},
  [1106] = {.lex_state = 1, .external_lex_state = 30},
  [1107] = {.lex_state = 1, .external_lex_state = 30},
  [1108] = {.lex_state = 1, .external_lex_state = 29},
  [1109] = {.lex_state = 1, .external_lex_state = 30},
  [1110] = {.lex_state = 1, .external_lex_state = 63},
  [1111] = {.lex_state = 1, .external_lex_state = 30},
  [1112] = {.lex_state = 1, .external_lex_state = 30},
  [1113] = {.lex_state = 1, .external_lex_state = 30},
  [1114] = {.lex_state = 1, .external_lex_state = 30},
  [1115] = {.lex_state = 1, .external_lex_state = 30},
  [1116] = {.lex_state = 1, .external_lex_state = 31},
  [1117] = {.lex_state = 1, .external_lex_state = 30},
  [1118] = {.lex_state = 1, .external_lex_state = 30},
  [1119] = {.lex_state = 1, .external_lex_state = 30},
  [1120] = {.lex_state = 1, .external_lex_state = 29},
  [1121] = {.lex_state = 1, .external_lex_state = 39},
  [1122] = {.lex_state = 1, .external_lex_state = 33},
  [1123] = {.lex_state = 1, .external_lex_state = 73},
  [1124] = {.lex_state = 1, .external_lex_state = 73},
  [1125] = {.lex_state = 1, .external_lex_state = 33},
  [1126] = {.lex_state = 1, .external_lex_state = 32},
  [1127] = {.lex_state = 1, .external_lex_state = 69},
  [1128] = {.lex_state = 1, .external_lex_state = 73},
  [1129] = {.lex_state = 1, .external_lex_state = 32},
  [1130] = {.lex_state = 1, .external_lex_state = 32},
  [1131] = {.lex_state = 1, .external_lex_state = 32},
  [1132] = {.lex_state = 1, .external_lex_state = 32},
  [1133] = {.lex_state = 1, .external_lex_state = 32},
  [1134] = {.lex_state = 1, .external_lex_state = 74},
  [1135] = {.lex_state = 1, .external_lex_state = 72},
  [1136] = {.lex_state = 1, .external_lex_state = 72},
  [1137] = {.lex_state = 1, .external_lex_state = 73},
  [1138] = {.lex_state = 1, .external_lex_state = 75},
  [1139] = {.lex_state = 1, .external_lex_state = 74},
  [1140] = {.lex_state = 1, .external_lex_state = 76},
  [1141] = {.lex_state = 1, .external_lex_state = 32},
  [1142] = {.lex_state = 1, .external_lex_state = 33},
  [1143] = {.lex_state = 1, .external_lex_state = 32},
  [1144] = {.lex_state = 1, .external_lex_state = 33},
  [1145] = {.lex_state = 1, .external_lex_state = 33},
  [1146] = {.lex_state = 1, .external_lex_state = 72},
  [1147] = {.lex_state = 2, .external_lex_state = 77},
  [1148] = {.lex_state = 1, .external_lex_state = 33},
  [1149] = {.lex_state = 1, .external_lex_state = 33},
  [1150] = {.lex_state = 1, .external_lex_state = 72},
  [1151] = {.lex_state = 1, .external_lex_state = 72},
  [1152] = {.lex_state = 1, .external_lex_state = 33},
  [1153] = {.lex_state = 2, .external_lex_state = 77},
  [1154] = {.lex_state = 1, .external_lex_state = 71},
  [1155] = {.lex_state = 1, .external_lex_state = 32},
  [1156] = {.lex_state = 1, .external_lex_state = 33},
  [1157] = {.lex_state = 1, .external_lex_state = 33},
  [1158] = {.lex_state = 1, .external_lex_state = 71},
  [1159] = {.lex_state = 1, .external_lex_state = 33},
  [1160] = {.lex_state = 1, .external_lex_state = 73},
  [1161] = {.lex_state = 1, .external_lex_state = 35},
  [1162] = {.lex_state = 1, .external_lex_state = 73},
  [1163] = {.lex_state = 1, .external_lex_state = 32},
  [1164] = {.lex_state = 1, .external_lex_state = 73},
  [1165] = {.lex_state = 1, .external_lex_state = 74},
  [1166] = {.lex_state = 1, .external_lex_state = 33},
  [1167] = {.lex_state = 1, .external_lex_state = 32},
  [1168] = {.lex_state = 1, .external_lex_state = 74},
  [1169] = {.lex_state = 1, .external_lex_state = 33},
  [1170] = {.lex_state = 1, .external_lex_state = 33},
  [1171] = {.lex_state = 1, .external_lex_state = 32},
  [1172] = {.lex_state = 1, .external_lex_state = 37},
  [1173] = {.lex_state = 1, .external_lex_state = 32},
  [1174] = {.lex_state = 1, .external_lex_state = 32},
  [1175] = {.lex_state = 1, .external_lex_state = 33},
  [1176] = {.lex_state = 1, .external_lex_state = 33},
  [1177] = {.lex_state = 1, .external_lex_state = 32},
  [1178] = {.lex_state = 1, .external_lex_state = 32},
  [1179] = {.lex_state = 1, .external_lex_state = 32},
  [1180] = {.lex_state = 1, .external_lex_state = 74},
  [1181] = {.lex_state = 1, .external_lex_state = 32},
  [1182] = {.lex_state = 0, .external_lex_state = 78},
  [1183] = {.lex_state = 0, .external_lex_state = 78},
  [1184] = {.lex_state = 1, .external_lex_state = 32},
  [1185] = {.lex_state = 1, .external_lex_state = 33},
  [1186] = {.lex_state = 1, .external_lex_state = 33},
  [1187] = {.lex_state = 1, .external_lex_state = 33},
  [1188] = {.lex_state = 1, .external_lex_state = 33},
  [1189] = {.lex_state = 1, .external_lex_state = 24},
  [1190] = {.lex_state = 1, .external_lex_state = 32},
  [1191] = {.lex_state = 1, .external_lex_state = 33},
  [1192] = {.lex_state = 1, .external_lex_state = 33},
  [1193] = {.lex_state = 1, .external_lex_state = 32},
  [1194] = {.lex_state = 1, .external_lex_state = 33},
  [1195] = {.lex_state = 1, .external_lex_state = 38},
  [1196] = {.lex_state = 1, .external_lex_state = 33},
  [1197] = {.lex_state = 1, .external_lex_state = 33},
  [1198] = {.lex_state = 1, .external_lex_state = 72},
  [1199] = {.lex_state = 1, .external_lex_state = 32},
  [1200] = {.lex_state = 1, .external_lex_state = 36},
  [1201] = {.lex_state = 1, .external_lex_state = 72},
  [1202] = {.lex_state = 1, .external_lex_state = 32},
  [1203] = {.lex_state = 1, .external_lex_state = 32},
  [1204] = {.lex_state = 1, .external_lex_state = 70},
  [1205] = {.lex_state = 1, .external_lex_state = 74},
  [1206] = {.lex_state = 1, .external_lex_state = 32},
  [1207] = {.lex_state = 1, .external_lex_state = 33},
  [1208] = {.lex_state = 0, .external_lex_state = 78},
  [1209] = {.lex_state = 1, .external_lex_state = 33},
  [1210] = {.lex_state = 1, .external_lex_state = 32},
  [1211] = {.lex_state = 1, .external_lex_state = 74},
  [1212] = {.lex_state = 0, .external_lex_state = 78},
  [1213] = {.lex_state = 0, .external_lex_state = 78},
  [1214] = {.lex_state = 1, .external_lex_state = 72},
  [1215] = {.lex_state = 1, .external_lex_state = 72},
  [1216] = {.lex_state = 1, .external_lex_state = 72},
  [1217] = {.lex_state = 0, .external_lex_state = 78},
  [1218] = {.lex_state = 1, .external_lex_state = 74},
  [1219] = {.lex_state = 0, .external_lex_state = 78},
  [1220] = {.lex_state = 0, .external_lex_state = 78},
  [1221] = {.lex_state = 1, .external_lex_state = 74},
  [1222] = {.lex_state = 1, .external_lex_state = 33},
  [1223] = {.lex_state = 1, .external_lex_state = 73},
  [1224] = {.lex_state = 0, .external_lex_state = 78},
  [1225] = {.lex_state = 0, .external_lex_state = 78},
  [1226] = {.lex_state = 0, .external_lex_state = 78},
  [1227] = {.lex_state = 1, .external_lex_state = 32},
  [1228] = {.lex_state = 1, .external_lex_state = 73},
  [1229] = {.lex_state = 1, .external_lex_state = 69},
  [1230] = {.lex_state = 0, .external_lex_state = 78},
  [1231] = {.lex_state = 0, .external_lex_state = 78},
  [1232] = {.lex_state = 0, .external_lex_state = 78},
  [1233] = {.lex_state = 1, .external_lex_state = 32},
  [1234] = {.lex_state = 0, .external_lex_state = 78},
  [1235] = {.lex_state = 1, .external_lex_state = 32},
  [1236] = {.lex_state = 1, .external_lex_state = 70},
  [1237] = {.lex_state = 1, .external_lex_state = 33},
  [1238] = {.lex_state = 1, .external_lex_state = 39},
  [1239] = {.lex_state = 1, .external_lex_state = 39},
  [1240] = {.lex_state = 1, .external_lex_state = 36},
  [1241] = {.lex_state = 1, .external_lex_state = 36},
  [1242] = {.lex_state = 1, .external_lex_state = 39},
  [1243] = {.lex_state = 1, .external_lex_state = 36},
  [1244] = {.lex_state = 1, .external_lex_state = 36},
  [1245] = {.lex_state = 1, .external_lex_state = 36},
  [1246] = {.lex_state = 1, .external_lex_state = 36},
  [1247] = {.lex_state = 1, .external_lex_state = 36},
  [1248] = {.lex_state = 1, .external_lex_state = 36},
  [1249] = {.lex_state = 1, .external_lex_state = 36},
  [1250] = {.lex_state = 1, .external_lex_state = 36},
  [1251] = {.lex_state = 1, .external_lex_state = 36},
  [1252] = {.lex_state = 1, .external_lex_state = 36},
  [1253] = {.lex_state = 1, .external_lex_state = 36},
  [1254] = {.lex_state = 1, .external_lex_state = 36},
  [1255] = {.lex_state = 1, .external_lex_state = 36},
  [1256] = {.lex_state = 1, .external_lex_state = 36},
  [1257] = {.lex_state = 1, .external_lex_state = 75},
  [1258] = {.lex_state = 1, .external_lex_state = 36},
  [1259] = {.lex_state = 1, .external_lex_state = 39},
  [1260] = {.lex_state = 1, .external_lex_state = 75},
  [1261] = {.lex_state = 1, .external_lex_state = 79},
  [1262] = {.lex_state = 1, .external_lex_state = 36},
  [1263] = {.lex_state = 1, .external_lex_state = 80},
  [1264] = {.lex_state = 1, .external_lex_state = 80},
  [1265] = {.lex_state = 1, .external_lex_state = 39},
  [1266] = {.lex_state = 1, .external_lex_state = 80},
  [1267] = {.lex_state = 1, .external_lex_state = 79},
  [1268] = {.lex_state = 1, .external_lex_state = 79},
  [1269] = {.lex_state = 1, .external_lex_state = 79},
  [1270] = {.lex_state = 1, .external_lex_state = 80},
  [1271] = {.lex_state = 1, .external_lex_state = 80},
  [1272] = {.lex_state = 1, .external_lex_state = 39},
  [1273] = {.lex_state = 1, .external_lex_state = 39},
  [1274] = {.lex_state = 1, .external_lex_state = 36},
  [1275] = {.lex_state = 1, .external_lex_state = 36},
  [1276] = {.lex_state = 1, .external_lex_state = 36},
  [1277] = {.lex_state = 1, .external_lex_state = 36},
  [1278] = {.lex_state = 1, .external_lex_state = 36},
  [1279] = {.lex_state = 1, .external_lex_state = 36},
  [1280] = {.lex_state = 1, .external_lex_state = 38},
  [1281] = {.lex_state = 1, .external_lex_state = 38},
  [1282] = {.lex_state = 1, .external_lex_state = 38},
  [1283] = {.lex_state = 1, .external_lex_state = 38},
  [1284] = {.lex_state = 1, .external_lex_state = 38},
  [1285] = {.lex_state = 1, .external_lex_state = 38},
  [1286] = {.lex_state = 1, .external_lex_state = 38},
  [1287] = {.lex_state = 1, .external_lex_state = 38},
  [1288] = {.lex_state = 1, .external_lex_state = 38},
  [1289] = {.lex_state = 1, .external_lex_state = 81},
  [1290] = {.lex_state = 1, .external_lex_state = 38},
  [1291] = {.lex_state = 1, .external_lex_state = 38},
  [1292] = {.lex_state = 1, .external_lex_state = 38},
  [1293] = {.lex_state = 1, .external_lex_state = 38},
  [1294] = {.lex_state = 1, .external_lex_state = 38},
  [1295] = {.lex_state = 1, .external_lex_state = 38},
  [1296] = {.lex_state = 1, .external_lex_state = 38},
  [1297] = {.lex_state = 1, .external_lex_state = 39},
  [1298] = {.lex_state = 1, .external_lex_state = 38},
  [1299] = {.lex_state = 1, .external_lex_state = 38},
  [1300] = {.lex_state = 1, .external_lex_state = 36},
  [1301] = {.lex_state = 1, .external_lex_state = 36},
  [1302] = {.lex_state = 1, .external_lex_state = 36},
  [1303] = {.lex_state = 1, .external_lex_state = 38},
  [1304] = {.lex_state = 1, .external_lex_state = 39},
  [1305] = {.lex_state = 1, .external_lex_state = 38},
  [1306] = {.lex_state = 1, .external_lex_state = 38},
  [1307] = {.lex_state = 1, .external_lex_state = 38},
  [1308] = {.lex_state = 1, .external_lex_state = 39},
  [1309] = {.lex_state = 1, .external_lex_state = 38},
  [1310] = {.lex_state = 1, .external_lex_state = 38},
  [1311] = {.lex_state = 1, .external_lex_state = 80},
  [1312] = {.lex_state = 1, .external_lex_state = 39},
  [1313] = {.lex_state = 1, .external_lex_state = 38},
  [1314] = {.lex_state = 1, .external_lex_state = 38},
  [1315] = {.lex_state = 1, .external_lex_state = 38},
  [1316] = {.lex_state = 1, .external_lex_state = 80},
  [1317] = {.lex_state = 1, .external_lex_state = 80},
  [1318] = {.lex_state = 1, .external_lex_state = 80},
  [1319] = {.lex_state = 1, .external_lex_state = 36},
  [1320] = {.lex_state = 1, .external_lex_state = 35},
  [1321] = {.lex_state = 1, .external_lex_state = 37},
  [1322] = {.lex_state = 1, .external_lex_state = 38},
  [1323] = {.lex_state = 1, .external_lex_state = 39},
  [1324] = {.lex_state = 1, .external_lex_state = 36},
  [1325] = {.lex_state = 1, .external_lex_state = 38},
  [1326] = {.lex_state = 1, .external_lex_state = 76},
  [1327] = {.lex_state = 2, .external_lex_state = 82},
  [1328] = {.lex_state = 1, .external_lex_state = 39},
  [1329] = {.lex_state = 1, .external_lex_state = 39},
  [1330] = {.lex_state = 1, .external_lex_state = 76},
  [1331] = {.lex_state = 1, .external_lex_state = 83},
  [1332] = {.lex_state = 1, .external_lex_state = 79},
  [1333] = {.lex_state = 1, .external_lex_state = 79},
  [1334] = {.lex_state = 1, .external_lex_state = 39},
  [1335] = {.lex_state = 2, .external_lex_state = 84},
  [1336] = {.lex_state = 1, .external_lex_state = 35},
  [1337] = {.lex_state = 1, .external_lex_state = 35},
  [1338] = {.lex_state = 1, .external_lex_state = 79},
  [1339] = {.lex_state = 1, .external_lex_state = 35},
  [1340] = {.lex_state = 1, .external_lex_state = 35},
  [1341] = {.lex_state = 1, .external_lex_state = 35},
  [1342] = {.lex_state = 1, .external_lex_state = 35},
  [1343] = {.lex_state = 1, .external_lex_state = 35},
  [1344] = {.lex_state = 1, .external_lex_state = 35},
  [1345] = {.lex_state = 1, .external_lex_state = 35},
  [1346] = {.lex_state = 1, .external_lex_state = 35},
  [1347] = {.lex_state = 1, .external_lex_state = 35},
  [1348] = {.lex_state = 1, .external_lex_state = 35},
  [1349] = {.lex_state = 1, .external_lex_state = 35},
  [1350] = {.lex_state = 1, .external_lex_state = 35},
  [1351] = {.lex_state = 1, .external_lex_state = 35},
  [1352] = {.lex_state = 1, .external_lex_state = 35},
  [1353] = {.lex_state = 1, .external_lex_state = 37},
  [1354] = {.lex_state = 1, .external_lex_state = 37},
  [1355] = {.lex_state = 1, .external_lex_state = 35},
  [1356] = {.lex_state = 1, .external_lex_state = 37},
  [1357] = {.lex_state = 1, .external_lex_state = 37},
  [1358] = {.lex_state = 1, .external_lex_state = 37},
  [1359] = {.lex_state = 1, .external_lex_state = 37},
  [1360] = {.lex_state = 1, .external_lex_state = 37},
  [1361] = {.lex_state = 1, .external_lex_state = 35},
  [1362] = {.lex_state = 1, .external_lex_state = 37},
  [1363] = {.lex_state = 1, .external_lex_state = 37},
  [1364] = {.lex_state = 1, .external_lex_state = 37},
  [1365] = {.lex_state = 1, .external_lex_state = 79},
  [1366] = {.lex_state = 1, .external_lex_state = 79},
  [1367] = {.lex_state = 1, .external_lex_state = 37},
  [1368] = {.lex_state = 1, .external_lex_state = 37},
  [1369] = {.lex_state = 1, .external_lex_state = 37},
  [1370] = {.lex_state = 1, .external_lex_state = 37},
  [1371] = {.lex_state = 1, .external_lex_state = 37},
  [1372] = {.lex_state = 1, .external_lex_state = 37},
  [1373] = {.lex_state = 1, .external_lex_state = 37},
  [1374] = {.lex_state = 1, .external_lex_state = 42},
  [1375] = {.lex_state = 1, .external_lex_state = 37},
  [1376] = {.lex_state = 1, .external_lex_state = 35},
  [1377] = {.lex_state = 1, .external_lex_state = 39},
  [1378] = {.lex_state = 1, .external_lex_state = 35},
  [1379] = {.lex_state = 1, .external_lex_state = 35},
  [1380] = {.lex_state = 1, .external_lex_state = 35},
  [1381] = {.lex_state = 1, .external_lex_state = 35},
  [1382] = {.lex_state = 1, .external_lex_state = 35},
  [1383] = {.lex_state = 1, .external_lex_state = 44},
  [1384] = {.lex_state = 1, .external_lex_state = 39},
  [1385] = {.lex_state = 1, .external_lex_state = 37},
  [1386] = {.lex_state = 1, .external_lex_state = 85},
  [1387] = {.lex_state = 1, .external_lex_state = 37},
  [1388] = {.lex_state = 1, .external_lex_state = 37},
  [1389] = {.lex_state = 1, .external_lex_state = 37},
  [1390] = {.lex_state = 1, .external_lex_state = 37},
  [1391] = {.lex_state = 1, .external_lex_state = 37},
  [1392] = {.lex_state = 2, .external_lex_state = 84},
  [1393] = {.lex_state = 1, .external_lex_state = 35},
  [1394] = {.lex_state = 1, .external_lex_state = 35},
  [1395] = {.lex_state = 1, .external_lex_state = 35},
  [1396] = {.lex_state = 1, .external_lex_state = 37},
  [1397] = {.lex_state = 1, .external_lex_state = 37},
  [1398] = {.lex_state = 1, .external_lex_state = 37},
  [1399] = {.lex_state = 1, .external_lex_state = 40},
  [1400] = {.lex_state = 1, .external_lex_state = 45},
  [1401] = {.lex_state = 2, .external_lex_state = 67},
  [1402] = {.lex_state = 2, .external_lex_state = 67},
  [1403] = {.lex_state = 1, .external_lex_state = 39},
  [1404] = {.lex_state = 1, .external_lex_state = 39},
  [1405] = {.lex_state = 1, .external_lex_state = 39},
  [1406] = {.lex_state = 1, .external_lex_state = 39},
  [1407] = {.lex_state = 1, .external_lex_state = 86},
  [1408] = {.lex_state = 1, .external_lex_state = 87},
  [1409] = {.lex_state = 1, .external_lex_state = 39},
  [1410] = {.lex_state = 1, .external_lex_state = 39},
  [1411] = {.lex_state = 1, .external_lex_state = 39},
  [1412] = {.lex_state = 1, .external_lex_state = 39},
  [1413] = {.lex_state = 1, .external_lex_state = 88},
  [1414] = {.lex_state = 1, .external_lex_state = 35},
  [1415] = {.lex_state = 1, .external_lex_state = 39},
  [1416] = {.lex_state = 1, .external_lex_state = 39},
  [1417] = {.lex_state = 1, .external_lex_state = 37},
  [1418] = {.lex_state = 1, .external_lex_state = 41},
  [1419] = {.lex_state = 1, .external_lex_state = 39},
  [1420] = {.lex_state = 1, .external_lex_state = 39},
  [1421] = {.lex_state = 1, .external_lex_state = 41},
  [1422] = {.lex_state = 1, .external_lex_state = 44},
  [1423] = {.lex_state = 1, .external_lex_state = 61},
  [1424] = {.lex_state = 1, .external_lex_state = 41},
  [1425] = {.lex_state = 1, .external_lex_state = 42},
  [1426] = {.lex_state = 1, .external_lex_state = 44},
  [1427] = {.lex_state = 1, .external_lex_state = 41},
  [1428] = {.lex_state = 1, .external_lex_state = 45},
  [1429] = {.lex_state = 1, .external_lex_state = 89},
  [1430] = {.lex_state = 1, .external_lex_state = 41},
  [1431] = {.lex_state = 1, .external_lex_state = 41},
  [1432] = {.lex_state = 1, .external_lex_state = 89},
  [1433] = {.lex_state = 1, .external_lex_state = 41},
  [1434] = {.lex_state = 1, .external_lex_state = 50},
  [1435] = {.lex_state = 1, .external_lex_state = 89},
  [1436] = {.lex_state = 1, .external_lex_state = 90},
  [1437] = {.lex_state = 1, .external_lex_state = 41},
  [1438] = {.lex_state = 1, .external_lex_state = 45},
  [1439] = {.lex_state = 1, .external_lex_state = 41},
  [1440] = {.lex_state = 1, .external_lex_state = 90},
  [1441] = {.lex_state = 1, .external_lex_state = 90},
  [1442] = {.lex_state = 1, .external_lex_state = 90},
  [1443] = {.lex_state = 0, .external_lex_state = 78},
  [1444] = {.lex_state = 1, .external_lex_state = 61},
  [1445] = {.lex_state = 1, .external_lex_state = 83},
  [1446] = {.lex_state = 1, .external_lex_state = 61},
  [1447] = {.lex_state = 1, .external_lex_state = 83},
  [1448] = {.lex_state = 1, .external_lex_state = 91},
  [1449] = {.lex_state = 1, .external_lex_state = 52},
  [1450] = {.lex_state = 0, .external_lex_state = 78},
  [1451] = {.lex_state = 1, .external_lex_state = 90},
  [1452] = {.lex_state = 1, .external_lex_state = 90},
  [1453] = {.lex_state = 1, .external_lex_state = 90},
  [1454] = {.lex_state = 0, .external_lex_state = 78},
  [1455] = {.lex_state = 1, .external_lex_state = 92},
  [1456] = {.lex_state = 1, .external_lex_state = 90},
  [1457] = {.lex_state = 1, .external_lex_state = 90},
  [1458] = {.lex_state = 1, .external_lex_state = 61},
  [1459] = {.lex_state = 1, .external_lex_state = 91},
  [1460] = {.lex_state = 1, .external_lex_state = 41},
  [1461] = {.lex_state = 1, .external_lex_state = 91},
  [1462] = {.lex_state = 1, .external_lex_state = 41},
  [1463] = {.lex_state = 1, .external_lex_state = 41},
  [1464] = {.lex_state = 1, .external_lex_state = 93},
  [1465] = {.lex_state = 1, .external_lex_state = 41},
  [1466] = {.lex_state = 1, .external_lex_state = 91},
  [1467] = {.lex_state = 1, .external_lex_state = 81},
  [1468] = {.lex_state = 1, .external_lex_state = 61},
  [1469] = {.lex_state = 1, .external_lex_state = 93},
  [1470] = {.lex_state = 1, .external_lex_state = 61},
  [1471] = {.lex_state = 1, .external_lex_state = 41},
  [1472] = {.lex_state = 1, .external_lex_state = 41},
  [1473] = {.lex_state = 0, .external_lex_state = 78},
  [1474] = {.lex_state = 1, .external_lex_state = 41},
  [1475] = {.lex_state = 1, .external_lex_state = 42},
  [1476] = {.lex_state = 1, .external_lex_state = 42},
  [1477] = {.lex_state = 1, .external_lex_state = 81},
  [1478] = {.lex_state = 1, .external_lex_state = 42},
  [1479] = {.lex_state = 1, .external_lex_state = 42},
  [1480] = {.lex_state = 1, .external_lex_state = 42},
  [1481] = {.lex_state = 1, .external_lex_state = 42},
  [1482] = {.lex_state = 1, .external_lex_state = 42},
  [1483] = {.lex_state = 1, .external_lex_state = 61},
  [1484] = {.lex_state = 1, .external_lex_state = 42},
  [1485] = {.lex_state = 1, .external_lex_state = 42},
  [1486] = {.lex_state = 1, .external_lex_state = 41},
  [1487] = {.lex_state = 1, .external_lex_state = 42},
  [1488] = {.lex_state = 1, .external_lex_state = 41},
  [1489] = {.lex_state = 1, .external_lex_state = 42},
  [1490] = {.lex_state = 1, .external_lex_state = 42},
  [1491] = {.lex_state = 1, .external_lex_state = 41},
  [1492] = {.lex_state = 1, .external_lex_state = 93},
  [1493] = {.lex_state = 1, .external_lex_state = 42},
  [1494] = {.lex_state = 1, .external_lex_state = 42},
  [1495] = {.lex_state = 1, .external_lex_state = 42},
  [1496] = {.lex_state = 1, .external_lex_state = 42},
  [1497] = {.lex_state = 1, .external_lex_state = 94},
  [1498] = {.lex_state = 1, .external_lex_state = 89},
  [1499] = {.lex_state = 1, .external_lex_state = 44},
  [1500] = {.lex_state = 2, .external_lex_state = 95},
  [1501] = {.lex_state = 1, .external_lex_state = 44},
  [1502] = {.lex_state = 1, .external_lex_state = 42},
  [1503] = {.lex_state = 1, .external_lex_state = 44},
  [1504] = {.lex_state = 1, .external_lex_state = 44},
  [1505] = {.lex_state = 1, .external_lex_state = 89},
  [1506] = {.lex_state = 1, .external_lex_state = 44},
  [1507] = {.lex_state = 1, .external_lex_state = 44},
  [1508] = {.lex_state = 1, .external_lex_state = 44},
  [1509] = {.lex_state = 1, .external_lex_state = 42},
  [1510] = {.lex_state = 1, .external_lex_state = 44},
  [1511] = {.lex_state = 1, .external_lex_state = 44},
  [1512] = {.lex_state = 1, .external_lex_state = 89},
  [1513] = {.lex_state = 1, .external_lex_state = 44},
  [1514] = {.lex_state = 1, .external_lex_state = 44},
  [1515] = {.lex_state = 1, .external_lex_state = 44},
  [1516] = {.lex_state = 1, .external_lex_state = 40},
  [1517] = {.lex_state = 1, .external_lex_state = 72},
  [1518] = {.lex_state = 1, .external_lex_state = 93},
  [1519] = {.lex_state = 1, .external_lex_state = 96},
  [1520] = {.lex_state = 1, .external_lex_state = 44},
  [1521] = {.lex_state = 1, .external_lex_state = 44},
  [1522] = {.lex_state = 1, .external_lex_state = 44},
  [1523] = {.lex_state = 1, .external_lex_state = 44},
  [1524] = {.lex_state = 1, .external_lex_state = 89},
  [1525] = {.lex_state = 1, .external_lex_state = 89},
  [1526] = {.lex_state = 1, .external_lex_state = 97},
  [1527] = {.lex_state = 1, .external_lex_state = 44},
  [1528] = {.lex_state = 1, .external_lex_state = 93},
  [1529] = {.lex_state = 2, .external_lex_state = 95},
  [1530] = {.lex_state = 1, .external_lex_state = 44},
  [1531] = {.lex_state = 1, .external_lex_state = 98},
  [1532] = {.lex_state = 1, .external_lex_state = 42},
  [1533] = {.lex_state = 1, .external_lex_state = 93},
  [1534] = {.lex_state = 1, .external_lex_state = 93},
  [1535] = {.lex_state = 1, .external_lex_state = 89},
  [1536] = {.lex_state = 1, .external_lex_state = 93},
  [1537] = {.lex_state = 1, .external_lex_state = 42},
  [1538] = {.lex_state = 1, .external_lex_state = 42},
  [1539] = {.lex_state = 1, .external_lex_state = 42},
  [1540] = {.lex_state = 1, .external_lex_state = 42},
  [1541] = {.lex_state = 1, .external_lex_state = 42},
  [1542] = {.lex_state = 1, .external_lex_state = 44},
  [1543] = {.lex_state = 1, .external_lex_state = 44},
  [1544] = {.lex_state = 1, .external_lex_state = 44},
  [1545] = {.lex_state = 1, .external_lex_state = 44},
  [1546] = {.lex_state = 1, .external_lex_state = 93},
  [1547] = {.lex_state = 2, .external_lex_state = 95},
  [1548] = {.lex_state = 1, .external_lex_state = 44},
  [1549] = {.lex_state = 0, .external_lex_state = 78},
  [1550] = {.lex_state = 1, .external_lex_state = 40},
  [1551] = {.lex_state = 1, .external_lex_state = 40},
  [1552] = {.lex_state = 1, .external_lex_state = 44},
  [1553] = {.lex_state = 1, .external_lex_state = 45},
  [1554] = {.lex_state = 1, .external_lex_state = 40},
  [1555] = {.lex_state = 1, .external_lex_state = 40},
  [1556] = {.lex_state = 1, .external_lex_state = 90},
  [1557] = {.lex_state = 1, .external_lex_state = 41},
  [1558] = {.lex_state = 1, .external_lex_state = 45},
  [1559] = {.lex_state = 1, .external_lex_state = 45},
  [1560] = {.lex_state = 1, .external_lex_state = 42},
  [1561] = {.lex_state = 1, .external_lex_state = 45},
  [1562] = {.lex_state = 1, .external_lex_state = 45},
  [1563] = {.lex_state = 1, .external_lex_state = 45},
  [1564] = {.lex_state = 1, .external_lex_state = 40},
  [1565] = {.lex_state = 1, .external_lex_state = 40},
  [1566] = {.lex_state = 1, .external_lex_state = 40},
  [1567] = {.lex_state = 1, .external_lex_state = 40},
  [1568] = {.lex_state = 1, .external_lex_state = 40},
  [1569] = {.lex_state = 1, .external_lex_state = 40},
  [1570] = {.lex_state = 1, .external_lex_state = 40},
  [1571] = {.lex_state = 1, .external_lex_state = 40},
  [1572] = {.lex_state = 1, .external_lex_state = 45},
  [1573] = {.lex_state = 1, .external_lex_state = 45},
  [1574] = {.lex_state = 1, .external_lex_state = 49},
  [1575] = {.lex_state = 1, .external_lex_state = 40},
  [1576] = {.lex_state = 1, .external_lex_state = 40},
  [1577] = {.lex_state = 1, .external_lex_state = 40},
  [1578] = {.lex_state = 1, .external_lex_state = 40},
  [1579] = {.lex_state = 1, .external_lex_state = 45},
  [1580] = {.lex_state = 1, .external_lex_state = 41},
  [1581] = {.lex_state = 1, .external_lex_state = 40},
  [1582] = {.lex_state = 1, .external_lex_state = 40},
  [1583] = {.lex_state = 1, .external_lex_state = 40},
  [1584] = {.lex_state = 1, .external_lex_state = 42},
  [1585] = {.lex_state = 1, .external_lex_state = 45},
  [1586] = {.lex_state = 1, .external_lex_state = 40},
  [1587] = {.lex_state = 1, .external_lex_state = 40},
  [1588] = {.lex_state = 1, .external_lex_state = 40},
  [1589] = {.lex_state = 1, .external_lex_state = 40},
  [1590] = {.lex_state = 1, .external_lex_state = 40},
  [1591] = {.lex_state = 1, .external_lex_state = 40},
  [1592] = {.lex_state = 1, .external_lex_state = 40},
  [1593] = {.lex_state = 1, .external_lex_state = 45},
  [1594] = {.lex_state = 1, .external_lex_state = 45},
  [1595] = {.lex_state = 1, .external_lex_state = 45},
  [1596] = {.lex_state = 1, .external_lex_state = 45},
  [1597] = {.lex_state = 1, .external_lex_state = 40},
  [1598] = {.lex_state = 1, .external_lex_state = 45},
  [1599] = {.lex_state = 1, .external_lex_state = 42},
  [1600] = {.lex_state = 1, .external_lex_state = 42},
  [1601] = {.lex_state = 1, .external_lex_state = 45},
  [1602] = {.lex_state = 1, .external_lex_state = 45},
  [1603] = {.lex_state = 2, .external_lex_state = 77},
  [1604] = {.lex_state = 1, .external_lex_state = 44},
  [1605] = {.lex_state = 1, .external_lex_state = 40},
  [1606] = {.lex_state = 1, .external_lex_state = 44},
  [1607] = {.lex_state = 1, .external_lex_state = 44},
  [1608] = {.lex_state = 2, .external_lex_state = 77},
  [1609] = {.lex_state = 1, .external_lex_state = 99},
  [1610] = {.lex_state = 1, .external_lex_state = 100},
  [1611] = {.lex_state = 1, .external_lex_state = 41},
  [1612] = {.lex_state = 1, .external_lex_state = 61},
  [1613] = {.lex_state = 1, .external_lex_state = 89},
  [1614] = {.lex_state = 1, .external_lex_state = 45},
  [1615] = {.lex_state = 1, .external_lex_state = 86},
  [1616] = {.lex_state = 1, .external_lex_state = 86},
  [1617] = {.lex_state = 1, .external_lex_state = 45},
  [1618] = {.lex_state = 1, .external_lex_state = 45},
  [1619] = {.lex_state = 1, .external_lex_state = 45},
  [1620] = {.lex_state = 1, .external_lex_state = 61},
  [1621] = {.lex_state = 1, .external_lex_state = 61},
  [1622] = {.lex_state = 1, .external_lex_state = 61},
  [1623] = {.lex_state = 1, .external_lex_state = 45},
  [1624] = {.lex_state = 1, .external_lex_state = 91},
  [1625] = {.lex_state = 1, .external_lex_state = 45},
  [1626] = {.lex_state = 1, .external_lex_state = 91},
  [1627] = {.lex_state = 1, .external_lex_state = 87},
  [1628] = {.lex_state = 1, .external_lex_state = 47},
  [1629] = {.lex_state = 1, .external_lex_state = 87},
  [1630] = {.lex_state = 1, .external_lex_state = 96},
  [1631] = {.lex_state = 1, .external_lex_state = 96},
  [1632] = {.lex_state = 1, .external_lex_state = 96},
  [1633] = {.lex_state = 1, .external_lex_state = 96},
  [1634] = {.lex_state = 1, .external_lex_state = 96},
  [1635] = {.lex_state = 1, .external_lex_state = 96},
  [1636] = {.lex_state = 1, .external_lex_state = 96},
  [1637] = {.lex_state = 1, .external_lex_state = 96},
  [1638] = {.lex_state = 1, .external_lex_state = 96},
  [1639] = {.lex_state = 1, .external_lex_state = 91},
  [1640] = {.lex_state = 1, .external_lex_state = 91},
  [1641] = {.lex_state = 1, .external_lex_state = 91},
  [1642] = {.lex_state = 1, .external_lex_state = 88},
  [1643] = {.lex_state = 1, .external_lex_state = 48},
  [1644] = {.lex_state = 1, .external_lex_state = 61},
  [1645] = {.lex_state = 1, .external_lex_state = 45},
  [1646] = {.lex_state = 1, .external_lex_state = 45},
  [1647] = {.lex_state = 0, .external_lex_state = 78},
  [1648] = {.lex_state = 1, .external_lex_state = 45},
  [1649] = {.lex_state = 1, .external_lex_state = 88},
  [1650] = {.lex_state = 0, .external_lex_state = 78},
  [1651] = {.lex_state = 0, .external_lex_state = 78},
  [1652] = {.lex_state = 1, .external_lex_state = 41},
  [1653] = {.lex_state = 0, .external_lex_state = 78},
  [1654] = {.lex_state = 0, .external_lex_state = 78},
  [1655] = {.lex_state = 0, .external_lex_state = 78},
  [1656] = {.lex_state = 0, .external_lex_state = 78},
  [1657] = {.lex_state = 0, .external_lex_state = 78},
  [1658] = {.lex_state = 0, .external_lex_state = 78},
  [1659] = {.lex_state = 0, .external_lex_state = 78},
  [1660] = {.lex_state = 0, .external_lex_state = 78},
  [1661] = {.lex_state = 1, .external_lex_state = 41},
  [1662] = {.lex_state = 0, .external_lex_state = 78},
  [1663] = {.lex_state = 1, .external_lex_state = 41},
  [1664] = {.lex_state = 0, .external_lex_state = 78},
  [1665] = {.lex_state = 1, .external_lex_state = 41},
  [1666] = {.lex_state = 0, .external_lex_state = 78},
  [1667] = {.lex_state = 1, .external_lex_state = 41},
  [1668] = {.lex_state = 0, .external_lex_state = 78},
  [1669] = {.lex_state = 1, .external_lex_state = 41},
  [1670] = {.lex_state = 0, .external_lex_state = 78},
  [1671] = {.lex_state = 1, .external_lex_state = 41},
  [1672] = {.lex_state = 1, .external_lex_state = 41},
  [1673] = {.lex_state = 1, .external_lex_state = 45},
  [1674] = {.lex_state = 1, .external_lex_state = 47},
  [1675] = {.lex_state = 1, .external_lex_state = 99},
  [1676] = {.lex_state = 1, .external_lex_state = 49},
  [1677] = {.lex_state = 1, .external_lex_state = 101},
  [1678] = {.lex_state = 1, .external_lex_state = 102},
  [1679] = {.lex_state = 1, .external_lex_state = 49},
  [1680] = {.lex_state = 1, .external_lex_state = 94},
  [1681] = {.lex_state = 1, .external_lex_state = 103},
  [1682] = {.lex_state = 2, .external_lex_state = 82},
  [1683] = {.lex_state = 1, .external_lex_state = 92},
  [1684] = {.lex_state = 2, .external_lex_state = 12},
  [1685] = {.lex_state = 1, .external_lex_state = 50},
  [1686] = {.lex_state = 2, .external_lex_state = 84},
  [1687] = {.lex_state = 1, .external_lex_state = 50},
  [1688] = {.lex_state = 1, .external_lex_state = 103},
  [1689] = {.lex_state = 1, .external_lex_state = 49},
  [1690] = {.lex_state = 1, .external_lex_state = 52},
  [1691] = {.lex_state = 1, .external_lex_state = 52},
  [1692] = {.lex_state = 1, .external_lex_state = 103},
  [1693] = {.lex_state = 1, .external_lex_state = 85},
  [1694] = {.lex_state = 1, .external_lex_state = 50},
  [1695] = {.lex_state = 1, .external_lex_state = 52},
  [1696] = {.lex_state = 1, .external_lex_state = 52},
  [1697] = {.lex_state = 1, .external_lex_state = 50},
  [1698] = {.lex_state = 1, .external_lex_state = 104},
  [1699] = {.lex_state = 1, .external_lex_state = 48},
  [1700] = {.lex_state = 1, .external_lex_state = 50},
  [1701] = {.lex_state = 1, .external_lex_state = 105},
  [1702] = {.lex_state = 1, .external_lex_state = 105},
  [1703] = {.lex_state = 1, .external_lex_state = 50},
  [1704] = {.lex_state = 1, .external_lex_state = 50},
  [1705] = {.lex_state = 1, .external_lex_state = 50},
  [1706] = {.lex_state = 1, .external_lex_state = 50},
  [1707] = {.lex_state = 2, .external_lex_state = 82},
  [1708] = {.lex_state = 1, .external_lex_state = 50},
  [1709] = {.lex_state = 1, .external_lex_state = 97},
  [1710] = {.lex_state = 1, .external_lex_state = 92},
  [1711] = {.lex_state = 1, .external_lex_state = 106},
  [1712] = {.lex_state = 1, .external_lex_state = 105},
  [1713] = {.lex_state = 1, .external_lex_state = 52},
  [1714] = {.lex_state = 1, .external_lex_state = 104},
  [1715] = {.lex_state = 1, .external_lex_state = 48},
  [1716] = {.lex_state = 1, .external_lex_state = 68},
  [1717] = {.lex_state = 1, .external_lex_state = 47},
  [1718] = {.lex_state = 1, .external_lex_state = 103},
  [1719] = {.lex_state = 1, .external_lex_state = 104},
  [1720] = {.lex_state = 1, .external_lex_state = 104},
  [1721] = {.lex_state = 1, .external_lex_state = 104},
  [1722] = {.lex_state = 1, .external_lex_state = 104},
  [1723] = {.lex_state = 1, .external_lex_state = 106},
  [1724] = {.lex_state = 1, .external_lex_state = 52},
  [1725] = {.lex_state = 1, .external_lex_state = 101},
  [1726] = {.lex_state = 1, .external_lex_state = 52},
  [1727] = {.lex_state = 2, .external_lex_state = 84},
  [1728] = {.lex_state = 1, .external_lex_state = 50},
  [1729] = {.lex_state = 1, .external_lex_state = 50},
  [1730] = {.lex_state = 1, .external_lex_state = 98},
  [1731] = {.lex_state = 1, .external_lex_state = 58},
  [1732] = {.lex_state = 1, .external_lex_state = 52},
  [1733] = {.lex_state = 1, .external_lex_state = 103},
  [1734] = {.lex_state = 1, .external_lex_state = 50},
  [1735] = {.lex_state = 1, .external_lex_state = 52},
  [1736] = {.lex_state = 1, .external_lex_state = 52},
  [1737] = {.lex_state = 1, .external_lex_state = 52},
  [1738] = {.lex_state = 1, .external_lex_state = 52},
  [1739] = {.lex_state = 1, .external_lex_state = 52},
  [1740] = {.lex_state = 1, .external_lex_state = 48},
  [1741] = {.lex_state = 1, .external_lex_state = 49},
  [1742] = {.lex_state = 1, .external_lex_state = 49},
  [1743] = {.lex_state = 1, .external_lex_state = 49},
  [1744] = {.lex_state = 1, .external_lex_state = 49},
  [1745] = {.lex_state = 1, .external_lex_state = 49},
  [1746] = {.lex_state = 1, .external_lex_state = 49},
  [1747] = {.lex_state = 1, .external_lex_state = 55},
  [1748] = {.lex_state = 1, .external_lex_state = 52},
  [1749] = {.lex_state = 1, .external_lex_state = 52},
  [1750] = {.lex_state = 1, .external_lex_state = 101},
  [1751] = {.lex_state = 1, .external_lex_state = 85},
  [1752] = {.lex_state = 1, .external_lex_state = 48},
  [1753] = {.lex_state = 1, .external_lex_state = 101},
  [1754] = {.lex_state = 1, .external_lex_state = 50},
  [1755] = {.lex_state = 1, .external_lex_state = 101},
  [1756] = {.lex_state = 1, .external_lex_state = 72},
  [1757] = {.lex_state = 1, .external_lex_state = 52},
  [1758] = {.lex_state = 1, .external_lex_state = 52},
  [1759] = {.lex_state = 1, .external_lex_state = 49},
  [1760] = {.lex_state = 1, .external_lex_state = 49},
  [1761] = {.lex_state = 1, .external_lex_state = 49},
  [1762] = {.lex_state = 1, .external_lex_state = 50},
  [1763] = {.lex_state = 1, .external_lex_state = 94},
  [1764] = {.lex_state = 1, .external_lex_state = 49},
  [1765] = {.lex_state = 1, .external_lex_state = 49},
  [1766] = {.lex_state = 1, .external_lex_state = 50},
  [1767] = {.lex_state = 1, .external_lex_state = 49},
  [1768] = {.lex_state = 1, .external_lex_state = 52},
  [1769] = {.lex_state = 1, .external_lex_state = 107},
  [1770] = {.lex_state = 1, .external_lex_state = 52},
  [1771] = {.lex_state = 1, .external_lex_state = 106},
  [1772] = {.lex_state = 1, .external_lex_state = 108},
  [1773] = {.lex_state = 1, .external_lex_state = 49},
  [1774] = {.lex_state = 1, .external_lex_state = 106},
  [1775] = {.lex_state = 1, .external_lex_state = 104},
  [1776] = {.lex_state = 2, .external_lex_state = 84},
  [1777] = {.lex_state = 1, .external_lex_state = 105},
  [1778] = {.lex_state = 1, .external_lex_state = 104},
  [1779] = {.lex_state = 1, .external_lex_state = 50},
  [1780] = {.lex_state = 1, .external_lex_state = 48},
  [1781] = {.lex_state = 1, .external_lex_state = 48},
  [1782] = {.lex_state = 1, .external_lex_state = 101},
  [1783] = {.lex_state = 1, .external_lex_state = 100},
  [1784] = {.lex_state = 1, .external_lex_state = 50},
  [1785] = {.lex_state = 1, .external_lex_state = 52},
  [1786] = {.lex_state = 2, .external_lex_state = 82},
  [1787] = {.lex_state = 1, .external_lex_state = 48},
  [1788] = {.lex_state = 1, .external_lex_state = 49},
  [1789] = {.lex_state = 1, .external_lex_state = 97},
  [1790] = {.lex_state = 1, .external_lex_state = 49},
  [1791] = {.lex_state = 1, .external_lex_state = 49},
  [1792] = {.lex_state = 1, .external_lex_state = 52},
  [1793] = {.lex_state = 1, .external_lex_state = 52},
  [1794] = {.lex_state = 1, .external_lex_state = 48},
  [1795] = {.lex_state = 1, .external_lex_state = 52},
  [1796] = {.lex_state = 1, .external_lex_state = 47},
  [1797] = {.lex_state = 1, .external_lex_state = 106},
  [1798] = {.lex_state = 1, .external_lex_state = 47},
  [1799] = {.lex_state = 1, .external_lex_state = 50},
  [1800] = {.lex_state = 1, .external_lex_state = 105},
  [1801] = {.lex_state = 1, .external_lex_state = 47},
  [1802] = {.lex_state = 1, .external_lex_state = 47},
  [1803] = {.lex_state = 1, .external_lex_state = 101},
  [1804] = {.lex_state = 1, .external_lex_state = 106},
  [1805] = {.lex_state = 1, .external_lex_state = 105},
  [1806] = {.lex_state = 1, .external_lex_state = 57},
  [1807] = {.lex_state = 1, .external_lex_state = 47},
  [1808] = {.lex_state = 1, .external_lex_state = 104},
  [1809] = {.lex_state = 1, .external_lex_state = 85},
  [1810] = {.lex_state = 1, .external_lex_state = 47},
  [1811] = {.lex_state = 1, .external_lex_state = 47},
  [1812] = {.lex_state = 1, .external_lex_state = 105},
  [1813] = {.lex_state = 1, .external_lex_state = 85},
  [1814] = {.lex_state = 1, .external_lex_state = 85},
  [1815] = {.lex_state = 1, .external_lex_state = 85},
  [1816] = {.lex_state = 1, .external_lex_state = 85},
  [1817] = {.lex_state = 1, .external_lex_state = 109},
  [1818] = {.lex_state = 1, .external_lex_state = 85},
  [1819] = {.lex_state = 1, .external_lex_state = 106},
  [1820] = {.lex_state = 1, .external_lex_state = 50},
  [1821] = {.lex_state = 1, .external_lex_state = 47},
  [1822] = {.lex_state = 1, .external_lex_state = 47},
  [1823] = {.lex_state = 1, .external_lex_state = 99},
  [1824] = {.lex_state = 1, .external_lex_state = 47},
  [1825] = {.lex_state = 1, .external_lex_state = 50},
  [1826] = {.lex_state = 1, .external_lex_state = 47},
  [1827] = {.lex_state = 1, .external_lex_state = 47},
  [1828] = {.lex_state = 1, .external_lex_state = 49},
  [1829] = {.lex_state = 1, .external_lex_state = 49},
  [1830] = {.lex_state = 1, .external_lex_state = 49},
  [1831] = {.lex_state = 1, .external_lex_state = 47},
  [1832] = {.lex_state = 1, .external_lex_state = 47},
  [1833] = {.lex_state = 1, .external_lex_state = 47},
  [1834] = {.lex_state = 1, .external_lex_state = 47},
  [1835] = {.lex_state = 1, .external_lex_state = 50},
  [1836] = {.lex_state = 1, .external_lex_state = 53},
  [1837] = {.lex_state = 1, .external_lex_state = 61},
  [1838] = {.lex_state = 1, .external_lex_state = 48},
  [1839] = {.lex_state = 1, .external_lex_state = 105},
  [1840] = {.lex_state = 1, .external_lex_state = 48},
  [1841] = {.lex_state = 1, .external_lex_state = 47},
  [1842] = {.lex_state = 1, .external_lex_state = 48},
  [1843] = {.lex_state = 1, .external_lex_state = 48},
  [1844] = {.lex_state = 1, .external_lex_state = 104},
  [1845] = {.lex_state = 1, .external_lex_state = 106},
  [1846] = {.lex_state = 1, .external_lex_state = 49},
  [1847] = {.lex_state = 1, .external_lex_state = 48},
  [1848] = {.lex_state = 1, .external_lex_state = 48},
  [1849] = {.lex_state = 1, .external_lex_state = 48},
  [1850] = {.lex_state = 1, .external_lex_state = 47},
  [1851] = {.lex_state = 1, .external_lex_state = 48},
  [1852] = {.lex_state = 1, .external_lex_state = 48},
  [1853] = {.lex_state = 1, .external_lex_state = 49},
  [1854] = {.lex_state = 1, .external_lex_state = 48},
  [1855] = {.lex_state = 1, .external_lex_state = 49},
  [1856] = {.lex_state = 2, .external_lex_state = 82},
  [1857] = {.lex_state = 1, .external_lex_state = 48},
  [1858] = {.lex_state = 1, .external_lex_state = 48},
  [1859] = {.lex_state = 1, .external_lex_state = 52},
  [1860] = {.lex_state = 1, .external_lex_state = 110},
  [1861] = {.lex_state = 1, .external_lex_state = 103},
  [1862] = {.lex_state = 1, .external_lex_state = 111},
  [1863] = {.lex_state = 1, .external_lex_state = 48},
  [1864] = {.lex_state = 1, .external_lex_state = 48},
  [1865] = {.lex_state = 1, .external_lex_state = 48},
  [1866] = {.lex_state = 1, .external_lex_state = 103},
  [1867] = {.lex_state = 1, .external_lex_state = 48},
  [1868] = {.lex_state = 1, .external_lex_state = 101},
  [1869] = {.lex_state = 1, .external_lex_state = 50},
  [1870] = {.lex_state = 1, .external_lex_state = 101},
  [1871] = {.lex_state = 2, .external_lex_state = 82},
  [1872] = {.lex_state = 1, .external_lex_state = 50},
  [1873] = {.lex_state = 1, .external_lex_state = 98},
  [1874] = {.lex_state = 1, .external_lex_state = 85},
  [1875] = {.lex_state = 1, .external_lex_state = 106},
  [1876] = {.lex_state = 1, .external_lex_state = 49},
  [1877] = {.lex_state = 1, .external_lex_state = 48},
  [1878] = {.lex_state = 1, .external_lex_state = 52},
  [1879] = {.lex_state = 1, .external_lex_state = 48},
  [1880] = {.lex_state = 1, .external_lex_state = 49},
  [1881] = {.lex_state = 1, .external_lex_state = 47},
  [1882] = {.lex_state = 1, .external_lex_state = 54},
  [1883] = {.lex_state = 1, .external_lex_state = 103},
  [1884] = {.lex_state = 1, .external_lex_state = 47},
  [1885] = {.lex_state = 1, .external_lex_state = 47},
  [1886] = {.lex_state = 1, .external_lex_state = 105},
  [1887] = {.lex_state = 1, .external_lex_state = 100},
  [1888] = {.lex_state = 1, .external_lex_state = 47},
  [1889] = {.lex_state = 1, .external_lex_state = 103},
  [1890] = {.lex_state = 2, .external_lex_state = 84},
  [1891] = {.lex_state = 1, .external_lex_state = 52},
  [1892] = {.lex_state = 1, .external_lex_state = 47},
  [1893] = {.lex_state = 1, .external_lex_state = 47},
  [1894] = {.lex_state = 1, .external_lex_state = 52},
  [1895] = {.lex_state = 1, .external_lex_state = 47},
  [1896] = {.lex_state = 1, .external_lex_state = 60},
  [1897] = {.lex_state = 1, .external_lex_state = 52},
  [1898] = {.lex_state = 1, .external_lex_state = 49},
  [1899] = {.lex_state = 1, .external_lex_state = 48},
  [1900] = {.lex_state = 1, .external_lex_state = 47},
  [1901] = {.lex_state = 1, .external_lex_state = 101},
  [1902] = {.lex_state = 1, .external_lex_state = 48},
  [1903] = {.lex_state = 1, .external_lex_state = 48},
  [1904] = {.lex_state = 1, .external_lex_state = 52},
  [1905] = {.lex_state = 1, .external_lex_state = 50},
  [1906] = {.lex_state = 1, .external_lex_state = 50},
  [1907] = {.lex_state = 1, .external_lex_state = 49},
  [1908] = {.lex_state = 1, .external_lex_state = 50},
  [1909] = {.lex_state = 1, .external_lex_state = 50},
  [1910] = {.lex_state = 1, .external_lex_state = 60},
  [1911] = {.lex_state = 1, .external_lex_state = 50},
  [1912] = {.lex_state = 1, .external_lex_state = 47},
  [1913] = {.lex_state = 1, .external_lex_state = 54},
  [1914] = {.lex_state = 2, .external_lex_state = 13},
  [1915] = {.lex_state = 2, .external_lex_state = 2},
  [1916] = {.lex_state = 1, .external_lex_state = 112},
  [1917] = {.lex_state = 2, .external_lex_state = 2},
  [1918] = {.lex_state = 2, .external_lex_state = 2},
  [1919] = {.lex_state = 2, .external_lex_state = 2},
  [1920] = {.lex_state = 2, .external_lex_state = 2},
  [1921] = {.lex_state = 2, .external_lex_state = 2},
  [1922] = {.lex_state = 2, .external_lex_state = 2},
  [1923] = {.lex_state = 2, .external_lex_state = 2},
  [1924] = {.lex_state = 2, .external_lex_state = 2},
  [1925] = {.lex_state = 2, .external_lex_state = 2},
  [1926] = {.lex_state = 2, .external_lex_state = 2},
  [1927] = {.lex_state = 1, .external_lex_state = 110},
  [1928] = {.lex_state = 2, .external_lex_state = 2},
  [1929] = {.lex_state = 2, .external_lex_state = 2},
  [1930] = {.lex_state = 1, .external_lex_state = 110},
  [1931] = {.lex_state = 1, .external_lex_state = 113},
  [1932] = {.lex_state = 2, .external_lex_state = 2},
  [1933] = {.lex_state = 1, .external_lex_state = 113},
  [1934] = {.lex_state = 1, .external_lex_state = 113},
  [1935] = {.lex_state = 1, .external_lex_state = 113},
  [1936] = {.lex_state = 1, .external_lex_state = 113},
  [1937] = {.lex_state = 1, .external_lex_state = 113},
  [1938] = {.lex_state = 1, .external_lex_state = 113},
  [1939] = {.lex_state = 1, .external_lex_state = 113},
  [1940] = {.lex_state = 1, .external_lex_state = 113},
  [1941] = {.lex_state = 1, .external_lex_state = 72},
  [1942] = {.lex_state = 2, .external_lex_state = 2},
  [1943] = {.lex_state = 1, .external_lex_state = 65},
  [1944] = {.lex_state = 1, .external_lex_state = 65},
  [1945] = {.lex_state = 1, .external_lex_state = 54},
  [1946] = {.lex_state = 1, .external_lex_state = 65},
  [1947] = {.lex_state = 1, .external_lex_state = 65},
  [1948] = {.lex_state = 1, .external_lex_state = 65},
  [1949] = {.lex_state = 1, .external_lex_state = 65},
  [1950] = {.lex_state = 1, .external_lex_state = 65},
  [1951] = {.lex_state = 1, .external_lex_state = 114},
  [1952] = {.lex_state = 2, .external_lex_state = 2},
  [1953] = {.lex_state = 2, .external_lex_state = 2},
  [1954] = {.lex_state = 1, .external_lex_state = 65},
  [1955] = {.lex_state = 1, .external_lex_state = 65},
  [1956] = {.lex_state = 1, .external_lex_state = 65},
  [1957] = {.lex_state = 2, .external_lex_state = 2},
  [1958] = {.lex_state = 1, .external_lex_state = 65},
  [1959] = {.lex_state = 2, .external_lex_state = 2},
  [1960] = {.lex_state = 2, .external_lex_state = 2},
  [1961] = {.lex_state = 2, .external_lex_state = 2},
  [1962] = {.lex_state = 2, .external_lex_state = 2},
  [1963] = {.lex_state = 1, .external_lex_state = 112},
  [1964] = {.lex_state = 1, .external_lex_state = 65},
  [1965] = {.lex_state = 2, .external_lex_state = 2},
  [1966] = {.lex_state = 2, .external_lex_state = 2},
  [1967] = {.lex_state = 2, .external_lex_state = 2},
  [1968] = {.lex_state = 2, .external_lex_state = 2},
  [1969] = {.lex_state = 2, .external_lex_state = 2},
  [1970] = {.lex_state = 2, .external_lex_state = 2},
  [1971] = {.lex_state = 2, .external_lex_state = 2},
  [1972] = {.lex_state = 2, .external_lex_state = 2},
  [1973] = {.lex_state = 1, .external_lex_state = 72},
  [1974] = {.lex_state = 2, .external_lex_state = 2},
  [1975] = {.lex_state = 1, .external_lex_state = 60},
  [1976] = {.lex_state = 1, .external_lex_state = 60},
  [1977] = {.lex_state = 2, .external_lex_state = 2},
  [1978] = {.lex_state = 1, .external_lex_state = 115},
  [1979] = {.lex_state = 2, .external_lex_state = 2},
  [1980] = {.lex_state = 1, .external_lex_state = 60},
  [1981] = {.lex_state = 1, .external_lex_state = 60},
  [1982] = {.lex_state = 2, .external_lex_state = 2},
  [1983] = {.lex_state = 2, .external_lex_state = 2},
  [1984] = {.lex_state = 2, .external_lex_state = 2},
  [1985] = {.lex_state = 2, .external_lex_state = 2},
  [1986] = {.lex_state = 1, .external_lex_state = 60},
  [1987] = {.lex_state = 1, .external_lex_state = 60},
  [1988] = {.lex_state = 1, .external_lex_state = 60},
  [1989] = {.lex_state = 1, .external_lex_state = 73},
  [1990] = {.lex_state = 1, .external_lex_state = 60},
  [1991] = {.lex_state = 1, .external_lex_state = 60},
  [1992] = {.lex_state = 1, .external_lex_state = 60},
  [1993] = {.lex_state = 1, .external_lex_state = 60},
  [1994] = {.lex_state = 1, .external_lex_state = 60},
  [1995] = {.lex_state = 1, .external_lex_state = 65},
  [1996] = {.lex_state = 1, .external_lex_state = 60},
  [1997] = {.lex_state = 1, .external_lex_state = 60},
  [1998] = {.lex_state = 1, .external_lex_state = 60},
  [1999] = {.lex_state = 1, .external_lex_state = 60},
  [2000] = {.lex_state = 1, .external_lex_state = 60},
  [2001] = {.lex_state = 1, .external_lex_state = 60},
  [2002] = {.lex_state = 1, .external_lex_state = 60},
  [2003] = {.lex_state = 1, .external_lex_state = 60},
  [2004] = {.lex_state = 1, .external_lex_state = 60},
  [2005] = {.lex_state = 1, .external_lex_state = 60},
  [2006] = {.lex_state = 1, .external_lex_state = 60},
  [2007] = {.lex_state = 1, .external_lex_state = 58},
  [2008] = {.lex_state = 1, .external_lex_state = 116},
  [2009] = {.lex_state = 1, .external_lex_state = 60},
  [2010] = {.lex_state = 2, .external_lex_state = 2},
  [2011] = {.lex_state = 2, .external_lex_state = 2},
  [2012] = {.lex_state = 2, .external_lex_state = 2},
  [2013] = {.lex_state = 1, .external_lex_state = 57},
  [2014] = {.lex_state = 1, .external_lex_state = 117},
  [2015] = {.lex_state = 1, .external_lex_state = 74},
  [2016] = {.lex_state = 1, .external_lex_state = 60},
  [2017] = {.lex_state = 1, .external_lex_state = 60},
  [2018] = {.lex_state = 1, .external_lex_state = 60},
  [2019] = {.lex_state = 1, .external_lex_state = 102},
  [2020] = {.lex_state = 1, .external_lex_state = 118},
  [2021] = {.lex_state = 1, .external_lex_state = 61},
  [2022] = {.lex_state = 1, .external_lex_state = 112},
  [2023] = {.lex_state = 1, .external_lex_state = 102},
  [2024] = {.lex_state = 1, .external_lex_state = 114},
  [2025] = {.lex_state = 2, .external_lex_state = 2},
  [2026] = {.lex_state = 2, .external_lex_state = 2},
  [2027] = {.lex_state = 1, .external_lex_state = 114},
  [2028] = {.lex_state = 1, .external_lex_state = 114},
  [2029] = {.lex_state = 2, .external_lex_state = 2},
  [2030] = {.lex_state = 2, .external_lex_state = 2},
  [2031] = {.lex_state = 2, .external_lex_state = 2},
  [2032] = {.lex_state = 1, .external_lex_state = 55},
  [2033] = {.lex_state = 2, .external_lex_state = 2},
  [2034] = {.lex_state = 1, .external_lex_state = 114},
  [2035] = {.lex_state = 1, .external_lex_state = 114},
  [2036] = {.lex_state = 1, .external_lex_state = 114},
  [2037] = {.lex_state = 1, .external_lex_state = 114},
  [2038] = {.lex_state = 1, .external_lex_state = 114},
  [2039] = {.lex_state = 1, .external_lex_state = 114},
  [2040] = {.lex_state = 2, .external_lex_state = 2},
  [2041] = {.lex_state = 2, .external_lex_state = 2},
  [2042] = {.lex_state = 1, .external_lex_state = 53},
  [2043] = {.lex_state = 2, .external_lex_state = 2},
  [2044] = {.lex_state = 2, .external_lex_state = 2},
  [2045] = {.lex_state = 2, .external_lex_state = 2},
  [2046] = {.lex_state = 2, .external_lex_state = 2},
  [2047] = {.lex_state = 2, .external_lex_state = 2},
  [2048] = {.lex_state = 1, .external_lex_state = 119},
  [2049] = {.lex_state = 2, .external_lex_state = 2},
  [2050] = {.lex_state = 1, .external_lex_state = 112},
  [2051] = {.lex_state = 2, .external_lex_state = 2},
  [2052] = {.lex_state = 2, .external_lex_state = 2},
  [2053] = {.lex_state = 2, .external_lex_state = 2},
  [2054] = {.lex_state = 1, .external_lex_state = 120},
  [2055] = {.lex_state = 1, .external_lex_state = 121},
  [2056] = {.lex_state = 1, .external_lex_state = 58},
  [2057] = {.lex_state = 1, .external_lex_state = 121},
  [2058] = {.lex_state = 1, .external_lex_state = 121},
  [2059] = {.lex_state = 1, .external_lex_state = 121},
  [2060] = {.lex_state = 2, .external_lex_state = 2},
  [2061] = {.lex_state = 1, .external_lex_state = 107},
  [2062] = {.lex_state = 1, .external_lex_state = 107},
  [2063] = {.lex_state = 1, .external_lex_state = 119},
  [2064] = {.lex_state = 1, .external_lex_state = 121},
  [2065] = {.lex_state = 1, .external_lex_state = 121},
  [2066] = {.lex_state = 1, .external_lex_state = 121},
  [2067] = {.lex_state = 1, .external_lex_state = 119},
  [2068] = {.lex_state = 1, .external_lex_state = 119},
  [2069] = {.lex_state = 1, .external_lex_state = 119},
  [2070] = {.lex_state = 1, .external_lex_state = 121},
  [2071] = {.lex_state = 1, .external_lex_state = 121},
  [2072] = {.lex_state = 1, .external_lex_state = 60},
  [2073] = {.lex_state = 2, .external_lex_state = 2},
  [2074] = {.lex_state = 2, .external_lex_state = 2},
  [2075] = {.lex_state = 2, .external_lex_state = 2},
  [2076] = {.lex_state = 1, .external_lex_state = 122},
  [2077] = {.lex_state = 1, .external_lex_state = 122},
  [2078] = {.lex_state = 1, .external_lex_state = 122},
  [2079] = {.lex_state = 1, .external_lex_state = 122},
  [2080] = {.lex_state = 1, .external_lex_state = 54},
  [2081] = {.lex_state = 2, .external_lex_state = 2},
  [2082] = {.lex_state = 1, .external_lex_state = 57},
  [2083] = {.lex_state = 1, .external_lex_state = 57},
  [2084] = {.lex_state = 1, .external_lex_state = 57},
  [2085] = {.lex_state = 1, .external_lex_state = 57},
  [2086] = {.lex_state = 2, .external_lex_state = 2},
  [2087] = {.lex_state = 1, .external_lex_state = 58},
  [2088] = {.lex_state = 1, .external_lex_state = 58},
  [2089] = {.lex_state = 1, .external_lex_state = 68},
  [2090] = {.lex_state = 1, .external_lex_state = 57},
  [2091] = {.lex_state = 1, .external_lex_state = 57},
  [2092] = {.lex_state = 1, .external_lex_state = 57},
  [2093] = {.lex_state = 1, .external_lex_state = 57},
  [2094] = {.lex_state = 1, .external_lex_state = 57},
  [2095] = {.lex_state = 1, .external_lex_state = 57},
  [2096] = {.lex_state = 1, .external_lex_state = 57},
  [2097] = {.lex_state = 1, .external_lex_state = 68},
  [2098] = {.lex_state = 1, .external_lex_state = 68},
  [2099] = {.lex_state = 1, .external_lex_state = 68},
  [2100] = {.lex_state = 1, .external_lex_state = 57},
  [2101] = {.lex_state = 1, .external_lex_state = 57},
  [2102] = {.lex_state = 1, .external_lex_state = 57},
  [2103] = {.lex_state = 1, .external_lex_state = 57},
  [2104] = {.lex_state = 1, .external_lex_state = 57},
  [2105] = {.lex_state = 1, .external_lex_state = 57},
  [2106] = {.lex_state = 1, .external_lex_state = 55},
  [2107] = {.lex_state = 1, .external_lex_state = 55},
  [2108] = {.lex_state = 1, .external_lex_state = 58},
  [2109] = {.lex_state = 1, .external_lex_state = 58},
  [2110] = {.lex_state = 1, .external_lex_state = 121},
  [2111] = {.lex_state = 1, .external_lex_state = 57},
  [2112] = {.lex_state = 1, .external_lex_state = 55},
  [2113] = {.lex_state = 1, .external_lex_state = 55},
  [2114] = {.lex_state = 2, .external_lex_state = 11},
  [2115] = {.lex_state = 1, .external_lex_state = 68},
  [2116] = {.lex_state = 1, .external_lex_state = 113},
  [2117] = {.lex_state = 0, .external_lex_state = 123},
  [2118] = {.lex_state = 1, .external_lex_state = 55},
  [2119] = {.lex_state = 1, .external_lex_state = 55},
  [2120] = {.lex_state = 1, .external_lex_state = 55},
  [2121] = {.lex_state = 1, .external_lex_state = 55},
  [2122] = {.lex_state = 1, .external_lex_state = 55},
  [2123] = {.lex_state = 1, .external_lex_state = 55},
  [2124] = {.lex_state = 1, .external_lex_state = 57},
  [2125] = {.lex_state = 1, .external_lex_state = 55},
  [2126] = {.lex_state = 1, .external_lex_state = 55},
  [2127] = {.lex_state = 1, .external_lex_state = 55},
  [2128] = {.lex_state = 1, .external_lex_state = 57},
  [2129] = {.lex_state = 1, .external_lex_state = 57},
  [2130] = {.lex_state = 1, .external_lex_state = 57},
  [2131] = {.lex_state = 1, .external_lex_state = 55},
  [2132] = {.lex_state = 1, .external_lex_state = 55},
  [2133] = {.lex_state = 1, .external_lex_state = 55},
  [2134] = {.lex_state = 1, .external_lex_state = 55},
  [2135] = {.lex_state = 1, .external_lex_state = 57},
  [2136] = {.lex_state = 1, .external_lex_state = 57},
  [2137] = {.lex_state = 1, .external_lex_state = 53},
  [2138] = {.lex_state = 1, .external_lex_state = 68},
  [2139] = {.lex_state = 1, .external_lex_state = 53},
  [2140] = {.lex_state = 1, .external_lex_state = 68},
  [2141] = {.lex_state = 1, .external_lex_state = 68},
  [2142] = {.lex_state = 1, .external_lex_state = 58},
  [2143] = {.lex_state = 1, .external_lex_state = 55},
  [2144] = {.lex_state = 1, .external_lex_state = 53},
  [2145] = {.lex_state = 1, .external_lex_state = 53},
  [2146] = {.lex_state = 1, .external_lex_state = 58},
  [2147] = {.lex_state = 1, .external_lex_state = 58},
  [2148] = {.lex_state = 1, .external_lex_state = 58},
  [2149] = {.lex_state = 1, .external_lex_state = 58},
  [2150] = {.lex_state = 1, .external_lex_state = 58},
  [2151] = {.lex_state = 1, .external_lex_state = 58},
  [2152] = {.lex_state = 1, .external_lex_state = 58},
  [2153] = {.lex_state = 1, .external_lex_state = 53},
  [2154] = {.lex_state = 1, .external_lex_state = 53},
  [2155] = {.lex_state = 1, .external_lex_state = 53},
  [2156] = {.lex_state = 1, .external_lex_state = 53},
  [2157] = {.lex_state = 1, .external_lex_state = 53},
  [2158] = {.lex_state = 1, .external_lex_state = 53},
  [2159] = {.lex_state = 1, .external_lex_state = 55},
  [2160] = {.lex_state = 1, .external_lex_state = 53},
  [2161] = {.lex_state = 1, .external_lex_state = 68},
  [2162] = {.lex_state = 1, .external_lex_state = 68},
  [2163] = {.lex_state = 2, .external_lex_state = 2},
  [2164] = {.lex_state = 2, .external_lex_state = 2},
  [2165] = {.lex_state = 2, .external_lex_state = 2},
  [2166] = {.lex_state = 0, .external_lex_state = 123},
  [2167] = {.lex_state = 1, .external_lex_state = 57},
  [2168] = {.lex_state = 1, .external_lex_state = 53},
  [2169] = {.lex_state = 1, .external_lex_state = 53},
  [2170] = {.lex_state = 1, .external_lex_state = 55},
  [2171] = {.lex_state = 1, .external_lex_state = 55},
  [2172] = {.lex_state = 1, .external_lex_state = 55},
  [2173] = {.lex_state = 1, .external_lex_state = 53},
  [2174] = {.lex_state = 1, .external_lex_state = 53},
  [2175] = {.lex_state = 1, .external_lex_state = 53},
  [2176] = {.lex_state = 1, .external_lex_state = 53},
  [2177] = {.lex_state = 1, .external_lex_state = 57},
  [2178] = {.lex_state = 1, .external_lex_state = 55},
  [2179] = {.lex_state = 1, .external_lex_state = 57},
  [2180] = {.lex_state = 1, .external_lex_state = 55},
  [2181] = {.lex_state = 2, .external_lex_state = 2},
  [2182] = {.lex_state = 1, .external_lex_state = 53},
  [2183] = {.lex_state = 2, .external_lex_state = 2},
  [2184] = {.lex_state = 1, .external_lex_state = 122},
  [2185] = {.lex_state = 1, .external_lex_state = 58},
  [2186] = {.lex_state = 1, .external_lex_state = 53},
  [2187] = {.lex_state = 1, .external_lex_state = 58},
  [2188] = {.lex_state = 1, .external_lex_state = 58},
  [2189] = {.lex_state = 1, .external_lex_state = 55},
  [2190] = {.lex_state = 1, .external_lex_state = 53},
  [2191] = {.lex_state = 1, .external_lex_state = 53},
  [2192] = {.lex_state = 1, .external_lex_state = 53},
  [2193] = {.lex_state = 1, .external_lex_state = 55},
  [2194] = {.lex_state = 1, .external_lex_state = 53},
  [2195] = {.lex_state = 1, .external_lex_state = 55},
  [2196] = {.lex_state = 1, .external_lex_state = 53},
  [2197] = {.lex_state = 1, .external_lex_state = 58},
  [2198] = {.lex_state = 1, .external_lex_state = 58},
  [2199] = {.lex_state = 1, .external_lex_state = 58},
  [2200] = {.lex_state = 1, .external_lex_state = 58},
  [2201] = {.lex_state = 1, .external_lex_state = 58},
  [2202] = {.lex_state = 1, .external_lex_state = 58},
  [2203] = {.lex_state = 1, .external_lex_state = 58},
  [2204] = {.lex_state = 1, .external_lex_state = 53},
  [2205] = {.lex_state = 1, .external_lex_state = 53},
  [2206] = {.lex_state = 1, .external_lex_state = 53},
  [2207] = {.lex_state = 1, .external_lex_state = 108},
  [2208] = {.lex_state = 1, .external_lex_state = 58},
  [2209] = {.lex_state = 1, .external_lex_state = 108},
  [2210] = {.lex_state = 1, .external_lex_state = 119},
  [2211] = {.lex_state = 1, .external_lex_state = 119},
  [2212] = {.lex_state = 2, .external_lex_state = 2},
  [2213] = {.lex_state = 2, .external_lex_state = 2},
  [2214] = {.lex_state = 1, .external_lex_state = 119},
  [2215] = {.lex_state = 2, .external_lex_state = 2},
  [2216] = {.lex_state = 1, .external_lex_state = 68},
  [2217] = {.lex_state = 1, .external_lex_state = 119},
  [2218] = {.lex_state = 1, .external_lex_state = 119},
  [2219] = {.lex_state = 1, .external_lex_state = 58},
  [2220] = {.lex_state = 2, .external_lex_state = 2},
  [2221] = {.lex_state = 2, .external_lex_state = 2},
  [2222] = {.lex_state = 1, .external_lex_state = 54},
  [2223] = {.lex_state = 1, .external_lex_state = 54},
  [2224] = {.lex_state = 1, .external_lex_state = 111},
  [2225] = {.lex_state = 2, .external_lex_state = 2},
  [2226] = {.lex_state = 1, .external_lex_state = 54},
  [2227] = {.lex_state = 1, .external_lex_state = 111},
  [2228] = {.lex_state = 2, .external_lex_state = 2},
  [2229] = {.lex_state = 2, .external_lex_state = 2},
  [2230] = {.lex_state = 2, .external_lex_state = 2},
  [2231] = {.lex_state = 1, .external_lex_state = 122},
  [2232] = {.lex_state = 1, .external_lex_state = 122},
  [2233] = {.lex_state = 1, .external_lex_state = 54},
  [2234] = {.lex_state = 1, .external_lex_state = 54},
  [2235] = {.lex_state = 1, .external_lex_state = 54},
  [2236] = {.lex_state = 1, .external_lex_state = 54},
  [2237] = {.lex_state = 1, .external_lex_state = 54},
  [2238] = {.lex_state = 1, .external_lex_state = 54},
  [2239] = {.lex_state = 1, .external_lex_state = 122},
  [2240] = {.lex_state = 1, .external_lex_state = 122},
  [2241] = {.lex_state = 1, .external_lex_state = 122},
  [2242] = {.lex_state = 2, .external_lex_state = 2},
  [2243] = {.lex_state = 1, .external_lex_state = 54},
  [2244] = {.lex_state = 1, .external_lex_state = 54},
  [2245] = {.lex_state = 1, .external_lex_state = 54},
  [2246] = {.lex_state = 1, .external_lex_state = 54},
  [2247] = {.lex_state = 1, .external_lex_state = 54},
  [2248] = {.lex_state = 1, .external_lex_state = 54},
  [2249] = {.lex_state = 2, .external_lex_state = 2},
  [2250] = {.lex_state = 2, .external_lex_state = 2},
  [2251] = {.lex_state = 2, .external_lex_state = 2},
  [2252] = {.lex_state = 2, .external_lex_state = 2},
  [2253] = {.lex_state = 1, .external_lex_state = 54},
  [2254] = {.lex_state = 1, .external_lex_state = 58},
  [2255] = {.lex_state = 1, .external_lex_state = 54},
  [2256] = {.lex_state = 1, .external_lex_state = 54},
  [2257] = {.lex_state = 2, .external_lex_state = 2},
  [2258] = {.lex_state = 2, .external_lex_state = 2},
  [2259] = {.lex_state = 2, .external_lex_state = 2},
  [2260] = {.lex_state = 2, .external_lex_state = 2},
  [2261] = {.lex_state = 2, .external_lex_state = 2},
  [2262] = {.lex_state = 2, .external_lex_state = 2},
  [2263] = {.lex_state = 2, .external_lex_state = 2},
  [2264] = {.lex_state = 2, .external_lex_state = 2},
  [2265] = {.lex_state = 1, .external_lex_state = 54},
  [2266] = {.lex_state = 1, .external_lex_state = 54},
  [2267] = {.lex_state = 1, .external_lex_state = 54},
  [2268] = {.lex_state = 1, .external_lex_state = 54},
  [2269] = {.lex_state = 1, .external_lex_state = 54},
  [2270] = {.lex_state = 2, .external_lex_state = 2},
  [2271] = {.lex_state = 1, .external_lex_state = 57},
  [2272] = {.lex_state = 2, .external_lex_state = 2},
  [2273] = {.lex_state = 1, .external_lex_state = 60},
  [2274] = {.lex_state = 2, .external_lex_state = 2},
  [2275] = {.lex_state = 1, .external_lex_state = 72},
  [2276] = {.lex_state = 1, .external_lex_state = 54},
  [2277] = {.lex_state = 1, .external_lex_state = 54},
  [2278] = {.lex_state = 1, .external_lex_state = 54},
  [2279] = {.lex_state = 1, .external_lex_state = 55},
  [2280] = {.lex_state = 1, .external_lex_state = 58},
  [2281] = {.lex_state = 1, .external_lex_state = 58},
  [2282] = {.lex_state = 1, .external_lex_state = 53},
  [2283] = {.lex_state = 1, .external_lex_state = 124},
  [2284] = {.lex_state = 1, .external_lex_state = 72},
  [2285] = {.lex_state = 2, .external_lex_state = 2},
  [2286] = {.lex_state = 2, .external_lex_state = 2},
  [2287] = {.lex_state = 2, .external_lex_state = 2},
  [2288] = {.lex_state = 1, .external_lex_state = 68},
  [2289] = {.lex_state = 1, .external_lex_state = 74},
  [2290] = {.lex_state = 1, .external_lex_state = 74},
  [2291] = {.lex_state = 1, .external_lex_state = 65},
  [2292] = {.lex_state = 1, .external_lex_state = 125},
  [2293] = {.lex_state = 1, .external_lex_state = 125},
  [2294] = {.lex_state = 1, .external_lex_state = 125},
  [2295] = {.lex_state = 1, .external_lex_state = 112},
  [2296] = {.lex_state = 1, .external_lex_state = 120},
  [2297] = {.lex_state = 0, .external_lex_state = 123},
  [2298] = {.lex_state = 1, .external_lex_state = 120},
  [2299] = {.lex_state = 1, .external_lex_state = 125},
  [2300] = {.lex_state = 1, .external_lex_state = 125},
  [2301] = {.lex_state = 1, .external_lex_state = 74},
  [2302] = {.lex_state = 1, .external_lex_state = 125},
  [2303] = {.lex_state = 1, .external_lex_state = 125},
  [2304] = {.lex_state = 1, .external_lex_state = 115},
  [2305] = {.lex_state = 1, .external_lex_state = 126},
  [2306] = {.lex_state = 1, .external_lex_state = 74},
  [2307] = {.lex_state = 1, .external_lex_state = 115},
  [2308] = {.lex_state = 1, .external_lex_state = 127},
  [2309] = {.lex_state = 0, .external_lex_state = 128},
  [2310] = {.lex_state = 1, .external_lex_state = 127},
  [2311] = {.lex_state = 1, .external_lex_state = 127},
  [2312] = {.lex_state = 1, .external_lex_state = 127},
  [2313] = {.lex_state = 1, .external_lex_state = 72},
  [2314] = {.lex_state = 0, .external_lex_state = 128},
  [2315] = {.lex_state = 1, .external_lex_state = 79},
  [2316] = {.lex_state = 1, .external_lex_state = 112},
  [2317] = {.lex_state = 1, .external_lex_state = 74},
  [2318] = {.lex_state = 1, .external_lex_state = 74},
  [2319] = {.lex_state = 1, .external_lex_state = 112},
  [2320] = {.lex_state = 1, .external_lex_state = 127},
  [2321] = {.lex_state = 1, .external_lex_state = 127},
  [2322] = {.lex_state = 1, .external_lex_state = 65},
  [2323] = {.lex_state = 1, .external_lex_state = 127},
  [2324] = {.lex_state = 1, .external_lex_state = 127},
  [2325] = {.lex_state = 1, .external_lex_state = 127},
  [2326] = {.lex_state = 1, .external_lex_state = 65},
  [2327] = {.lex_state = 1, .external_lex_state = 127},
  [2328] = {.lex_state = 1, .external_lex_state = 112},
  [2329] = {.lex_state = 1, .external_lex_state = 112},
  [2330] = {.lex_state = 1, .external_lex_state = 65},
  [2331] = {.lex_state = 1, .external_lex_state = 112},
  [2332] = {.lex_state = 1, .external_lex_state = 129},
  [2333] = {.lex_state = 1, .external_lex_state = 65},
  [2334] = {.lex_state = 1, .external_lex_state = 112},
  [2335] = {.lex_state = 1, .external_lex_state = 65},
  [2336] = {.lex_state = 0, .external_lex_state = 128},
  [2337] = {.lex_state = 1, .external_lex_state = 116},
  [2338] = {.lex_state = 0, .external_lex_state = 128},
  [2339] = {.lex_state = 1, .external_lex_state = 125},
  [2340] = {.lex_state = 1, .external_lex_state = 74},
  [2341] = {.lex_state = 1, .external_lex_state = 116},
  [2342] = {.lex_state = 1, .external_lex_state = 126},
  [2343] = {.lex_state = 1, .external_lex_state = 126},
  [2344] = {.lex_state = 1, .external_lex_state = 126},
  [2345] = {.lex_state = 1, .external_lex_state = 80},
  [2346] = {.lex_state = 1, .external_lex_state = 126},
  [2347] = {.lex_state = 1, .external_lex_state = 126},
  [2348] = {.lex_state = 1, .external_lex_state = 126},
  [2349] = {.lex_state = 1, .external_lex_state = 126},
  [2350] = {.lex_state = 1, .external_lex_state = 126},
  [2351] = {.lex_state = 1, .external_lex_state = 126},
  [2352] = {.lex_state = 1, .external_lex_state = 68},
  [2353] = {.lex_state = 1, .external_lex_state = 68},
  [2354] = {.lex_state = 1, .external_lex_state = 117},
  [2355] = {.lex_state = 1, .external_lex_state = 117},
  [2356] = {.lex_state = 1, .external_lex_state = 129},
  [2357] = {.lex_state = 1, .external_lex_state = 118},
  [2358] = {.lex_state = 0, .external_lex_state = 128},
  [2359] = {.lex_state = 1, .external_lex_state = 129},
  [2360] = {.lex_state = 1, .external_lex_state = 129},
  [2361] = {.lex_state = 1, .external_lex_state = 129},
  [2362] = {.lex_state = 1, .external_lex_state = 129},
  [2363] = {.lex_state = 1, .external_lex_state = 129},
  [2364] = {.lex_state = 1, .external_lex_state = 129},
  [2365] = {.lex_state = 1, .external_lex_state = 129},
  [2366] = {.lex_state = 1, .external_lex_state = 73},
  [2367] = {.lex_state = 1, .external_lex_state = 73},
  [2368] = {.lex_state = 1, .external_lex_state = 73},
  [2369] = {.lex_state = 1, .external_lex_state = 73},
  [2370] = {.lex_state = 1, .external_lex_state = 129},
  [2371] = {.lex_state = 1, .external_lex_state = 73},
  [2372] = {.lex_state = 1, .external_lex_state = 73},
  [2373] = {.lex_state = 1, .external_lex_state = 73},
  [2374] = {.lex_state = 1, .external_lex_state = 73},
  [2375] = {.lex_state = 1, .external_lex_state = 73},
  [2376] = {.lex_state = 1, .external_lex_state = 73},
  [2377] = {.lex_state = 1, .external_lex_state = 65},
  [2378] = {.lex_state = 1, .external_lex_state = 112},
  [2379] = {.lex_state = 1, .external_lex_state = 65},
  [2380] = {.lex_state = 1, .external_lex_state = 72},
  [2381] = {.lex_state = 1, .external_lex_state = 65},
  [2382] = {.lex_state = 1, .external_lex_state = 130},
  [2383] = {.lex_state = 1, .external_lex_state = 72},
  [2384] = {.lex_state = 1, .external_lex_state = 72},
  [2385] = {.lex_state = 1, .external_lex_state = 72},
  [2386] = {.lex_state = 0, .external_lex_state = 128},
  [2387] = {.lex_state = 1, .external_lex_state = 74},
  [2388] = {.lex_state = 1, .external_lex_state = 129},
  [2389] = {.lex_state = 1, .external_lex_state = 72},
  [2390] = {.lex_state = 1, .external_lex_state = 72},
  [2391] = {.lex_state = 0, .external_lex_state = 123},
  [2392] = {.lex_state = 0, .external_lex_state = 123},
  [2393] = {.lex_state = 1, .external_lex_state = 118},
  [2394] = {.lex_state = 1, .external_lex_state = 130},
  [2395] = {.lex_state = 1, .external_lex_state = 72},
  [2396] = {.lex_state = 1, .external_lex_state = 72},
  [2397] = {.lex_state = 1, .external_lex_state = 72},
  [2398] = {.lex_state = 0, .external_lex_state = 128},
  [2399] = {.lex_state = 1, .external_lex_state = 74},
  [2400] = {.lex_state = 1, .external_lex_state = 65},
  [2401] = {.lex_state = 1, .external_lex_state = 127},
  [2402] = {.lex_state = 1, .external_lex_state = 130},
  [2403] = {.lex_state = 1, .external_lex_state = 130},
  [2404] = {.lex_state = 1, .external_lex_state = 72},
  [2405] = {.lex_state = 1, .external_lex_state = 72},
  [2406] = {.lex_state = 0, .external_lex_state = 123},
  [2407] = {.lex_state = 1, .external_lex_state = 112},
  [2408] = {.lex_state = 1, .external_lex_state = 74},
  [2409] = {.lex_state = 1, .external_lex_state = 74},
  [2410] = {.lex_state = 0, .external_lex_state = 128},
  [2411] = {.lex_state = 0, .external_lex_state = 128},
  [2412] = {.lex_state = 1, .external_lex_state = 112},
  [2413] = {.lex_state = 1, .external_lex_state = 127},
  [2414] = {.lex_state = 1, .external_lex_state = 130},
  [2415] = {.lex_state = 1, .external_lex_state = 130},
  [2416] = {.lex_state = 1, .external_lex_state = 126},
  [2417] = {.lex_state = 1, .external_lex_state = 73},
  [2418] = {.lex_state = 1, .external_lex_state = 127},
  [2419] = {.lex_state = 1, .external_lex_state = 125},
  [2420] = {.lex_state = 1, .external_lex_state = 130},
  [2421] = {.lex_state = 1, .external_lex_state = 73},
  [2422] = {.lex_state = 1, .external_lex_state = 124},
  [2423] = {.lex_state = 1, .external_lex_state = 130},
  [2424] = {.lex_state = 1, .external_lex_state = 130},
  [2425] = {.lex_state = 1, .external_lex_state = 130},
  [2426] = {.lex_state = 0, .external_lex_state = 128},
  [2427] = {.lex_state = 1, .external_lex_state = 124},
  [2428] = {.lex_state = 1, .external_lex_state = 112},
  [2429] = {.lex_state = 1, .external_lex_state = 74},
  [2430] = {.lex_state = 1, .external_lex_state = 112},
  [2431] = {.lex_state = 1, .external_lex_state = 112},
  [2432] = {.lex_state = 1, .external_lex_state = 112},
  [2433] = {.lex_state = 1, .external_lex_state = 125},
  [2434] = {.lex_state = 1, .external_lex_state = 65},
  [2435] = {.lex_state = 1, .external_lex_state = 65},
  [2436] = {.lex_state = 1, .external_lex_state = 79},
  [2437] = {.lex_state = 1, .external_lex_state = 65},
  [2438] = {.lex_state = 1, .external_lex_state = 79},
  [2439] = {.lex_state = 1, .external_lex_state = 79},
  [2440] = {.lex_state = 1, .external_lex_state = 79},
  [2441] = {.lex_state = 1, .external_lex_state = 65},
  [2442] = {.lex_state = 1, .external_lex_state = 79},
  [2443] = {.lex_state = 1, .external_lex_state = 93},
  [2444] = {.lex_state = 1, .external_lex_state = 79},
  [2445] = {.lex_state = 1, .external_lex_state = 79},
  [2446] = {.lex_state = 1, .external_lex_state = 79},
  [2447] = {.lex_state = 1, .external_lex_state = 79},
  [2448] = {.lex_state = 1, .external_lex_state = 79},
  [2449] = {.lex_state = 1, .external_lex_state = 65},
  [2450] = {.lex_state = 1, .external_lex_state = 65},
  [2451] = {.lex_state = 1, .external_lex_state = 80},
  [2452] = {.lex_state = 1, .external_lex_state = 80},
  [2453] = {.lex_state = 1, .external_lex_state = 79},
  [2454] = {.lex_state = 1, .external_lex_state = 80},
  [2455] = {.lex_state = 1, .external_lex_state = 80},
  [2456] = {.lex_state = 1, .external_lex_state = 80},
  [2457] = {.lex_state = 1, .external_lex_state = 80},
  [2458] = {.lex_state = 1, .external_lex_state = 80},
  [2459] = {.lex_state = 1, .external_lex_state = 80},
  [2460] = {.lex_state = 1, .external_lex_state = 80},
  [2461] = {.lex_state = 1, .external_lex_state = 80},
  [2462] = {.lex_state = 1, .external_lex_state = 80},
  [2463] = {.lex_state = 1, .external_lex_state = 91},
  [2464] = {.lex_state = 1, .external_lex_state = 65},
  [2465] = {.lex_state = 1, .external_lex_state = 65},
  [2466] = {.lex_state = 1, .external_lex_state = 65},
  [2467] = {.lex_state = 1, .external_lex_state = 80},
  [2468] = {.lex_state = 1, .external_lex_state = 65},
  [2469] = {.lex_state = 1, .external_lex_state = 131},
  [2470] = {.lex_state = 1, .external_lex_state = 79},
  [2471] = {.lex_state = 1, .external_lex_state = 89},
  [2472] = {.lex_state = 1, .external_lex_state = 90},
  [2473] = {.lex_state = 1, .external_lex_state = 90},
  [2474] = {.lex_state = 1, .external_lex_state = 90},
  [2475] = {.lex_state = 1, .external_lex_state = 93},
  [2476] = {.lex_state = 1, .external_lex_state = 96},
  [2477] = {.lex_state = 1, .external_lex_state = 90},
  [2478] = {.lex_state = 1, .external_lex_state = 89},
  [2479] = {.lex_state = 1, .external_lex_state = 90},
  [2480] = {.lex_state = 1, .external_lex_state = 91},
  [2481] = {.lex_state = 1, .external_lex_state = 91},
  [2482] = {.lex_state = 1, .external_lex_state = 90},
  [2483] = {.lex_state = 1, .external_lex_state = 91},
  [2484] = {.lex_state = 1, .external_lex_state = 90},
  [2485] = {.lex_state = 1, .external_lex_state = 90},
  [2486] = {.lex_state = 1, .external_lex_state = 106},
  [2487] = {.lex_state = 1, .external_lex_state = 90},
  [2488] = {.lex_state = 1, .external_lex_state = 90},
  [2489] = {.lex_state = 1, .external_lex_state = 105},
  [2490] = {.lex_state = 1, .external_lex_state = 90},
  [2491] = {.lex_state = 1, .external_lex_state = 89},
  [2492] = {.lex_state = 1, .external_lex_state = 91},
  [2493] = {.lex_state = 1, .external_lex_state = 91},
  [2494] = {.lex_state = 1, .external_lex_state = 91},
  [2495] = {.lex_state = 1, .external_lex_state = 91},
  [2496] = {.lex_state = 1, .external_lex_state = 93},
  [2497] = {.lex_state = 1, .external_lex_state = 93},
  [2498] = {.lex_state = 1, .external_lex_state = 90},
  [2499] = {.lex_state = 1, .external_lex_state = 91},
  [2500] = {.lex_state = 1, .external_lex_state = 93},
  [2501] = {.lex_state = 1, .external_lex_state = 93},
  [2502] = {.lex_state = 1, .external_lex_state = 93},
  [2503] = {.lex_state = 1, .external_lex_state = 93},
  [2504] = {.lex_state = 1, .external_lex_state = 72},
  [2505] = {.lex_state = 1, .external_lex_state = 93},
  [2506] = {.lex_state = 1, .external_lex_state = 96},
  [2507] = {.lex_state = 1, .external_lex_state = 91},
  [2508] = {.lex_state = 1, .external_lex_state = 96},
  [2509] = {.lex_state = 1, .external_lex_state = 96},
  [2510] = {.lex_state = 1, .external_lex_state = 96},
  [2511] = {.lex_state = 1, .external_lex_state = 96},
  [2512] = {.lex_state = 1, .external_lex_state = 96},
  [2513] = {.lex_state = 1, .external_lex_state = 89},
  [2514] = {.lex_state = 1, .external_lex_state = 96},
  [2515] = {.lex_state = 1, .external_lex_state = 96},
  [2516] = {.lex_state = 1, .external_lex_state = 89},
  [2517] = {.lex_state = 1, .external_lex_state = 89},
  [2518] = {.lex_state = 1, .external_lex_state = 89},
  [2519] = {.lex_state = 1, .external_lex_state = 96},
  [2520] = {.lex_state = 1, .external_lex_state = 103},
  [2521] = {.lex_state = 1, .external_lex_state = 93},
  [2522] = {.lex_state = 1, .external_lex_state = 96},
  [2523] = {.lex_state = 1, .external_lex_state = 89},
  [2524] = {.lex_state = 1, .external_lex_state = 89},
  [2525] = {.lex_state = 1, .external_lex_state = 89},
  [2526] = {.lex_state = 1, .external_lex_state = 89},
  [2527] = {.lex_state = 1, .external_lex_state = 89},
  [2528] = {.lex_state = 1, .external_lex_state = 93},
  [2529] = {.lex_state = 1, .external_lex_state = 93},
  [2530] = {.lex_state = 1, .external_lex_state = 91},
  [2531] = {.lex_state = 1, .external_lex_state = 93},
  [2532] = {.lex_state = 1, .external_lex_state = 91},
  [2533] = {.lex_state = 1, .external_lex_state = 90},
  [2534] = {.lex_state = 1, .external_lex_state = 89},
  [2535] = {.lex_state = 1, .external_lex_state = 91},
  [2536] = {.lex_state = 1, .external_lex_state = 96},
  [2537] = {.lex_state = 1, .external_lex_state = 85},
  [2538] = {.lex_state = 2, .external_lex_state = 20},
  [2539] = {.lex_state = 1, .external_lex_state = 105},
  [2540] = {.lex_state = 1, .external_lex_state = 85},
  [2541] = {.lex_state = 1, .external_lex_state = 106},
  [2542] = {.lex_state = 2, .external_lex_state = 20},
  [2543] = {.lex_state = 1, .external_lex_state = 106},
  [2544] = {.lex_state = 1, .external_lex_state = 103},
  [2545] = {.lex_state = 1, .external_lex_state = 101},
  [2546] = {.lex_state = 1, .external_lex_state = 106},
  [2547] = {.lex_state = 1, .external_lex_state = 85},
  [2548] = {.lex_state = 1, .external_lex_state = 106},
  [2549] = {.lex_state = 1, .external_lex_state = 106},
  [2550] = {.lex_state = 1, .external_lex_state = 85},
  [2551] = {.lex_state = 1, .external_lex_state = 106},
  [2552] = {.lex_state = 1, .external_lex_state = 106},
  [2553] = {.lex_state = 1, .external_lex_state = 106},
  [2554] = {.lex_state = 1, .external_lex_state = 103},
  [2555] = {.lex_state = 1, .external_lex_state = 106},
  [2556] = {.lex_state = 1, .external_lex_state = 85},
  [2557] = {.lex_state = 1, .external_lex_state = 103},
  [2558] = {.lex_state = 1, .external_lex_state = 103},
  [2559] = {.lex_state = 1, .external_lex_state = 105},
  [2560] = {.lex_state = 1, .external_lex_state = 104},
  [2561] = {.lex_state = 1, .external_lex_state = 105},
  [2562] = {.lex_state = 1, .external_lex_state = 105},
  [2563] = {.lex_state = 1, .external_lex_state = 105},
  [2564] = {.lex_state = 1, .external_lex_state = 103},
  [2565] = {.lex_state = 1, .external_lex_state = 101},
  [2566] = {.lex_state = 1, .external_lex_state = 105},
  [2567] = {.lex_state = 1, .external_lex_state = 85},
  [2568] = {.lex_state = 1, .external_lex_state = 105},
  [2569] = {.lex_state = 1, .external_lex_state = 106},
  [2570] = {.lex_state = 1, .external_lex_state = 104},
  [2571] = {.lex_state = 1, .external_lex_state = 106},
  [2572] = {.lex_state = 1, .external_lex_state = 106},
  [2573] = {.lex_state = 1, .external_lex_state = 104},
  [2574] = {.lex_state = 1, .external_lex_state = 104},
  [2575] = {.lex_state = 1, .external_lex_state = 101},
  [2576] = {.lex_state = 1, .external_lex_state = 104},
  [2577] = {.lex_state = 1, .external_lex_state = 104},
  [2578] = {.lex_state = 1, .external_lex_state = 104},
  [2579] = {.lex_state = 2, .external_lex_state = 15},
  [2580] = {.lex_state = 1, .external_lex_state = 104},
  [2581] = {.lex_state = 1, .external_lex_state = 104},
  [2582] = {.lex_state = 1, .external_lex_state = 104},
  [2583] = {.lex_state = 1, .external_lex_state = 104},
  [2584] = {.lex_state = 1, .external_lex_state = 85},
  [2585] = {.lex_state = 1, .external_lex_state = 103},
  [2586] = {.lex_state = 1, .external_lex_state = 105},
  [2587] = {.lex_state = 1, .external_lex_state = 103},
  [2588] = {.lex_state = 1, .external_lex_state = 85},
  [2589] = {.lex_state = 1, .external_lex_state = 105},
  [2590] = {.lex_state = 1, .external_lex_state = 85},
  [2591] = {.lex_state = 1, .external_lex_state = 105},
  [2592] = {.lex_state = 1, .external_lex_state = 105},
  [2593] = {.lex_state = 2, .external_lex_state = 20},
  [2594] = {.lex_state = 1, .external_lex_state = 105},
  [2595] = {.lex_state = 1, .external_lex_state = 103},
  [2596] = {.lex_state = 1, .external_lex_state = 103},
  [2597] = {.lex_state = 1, .external_lex_state = 103},
  [2598] = {.lex_state = 1, .external_lex_state = 101},
  [2599] = {.lex_state = 1, .external_lex_state = 101},
  [2600] = {.lex_state = 1, .external_lex_state = 101},
  [2601] = {.lex_state = 1, .external_lex_state = 101},
  [2602] = {.lex_state = 1, .external_lex_state = 101},
  [2603] = {.lex_state = 1, .external_lex_state = 101},
  [2604] = {.lex_state = 1, .external_lex_state = 101},
  [2605] = {.lex_state = 1, .external_lex_state = 101},
  [2606] = {.lex_state = 1, .external_lex_state = 101},
  [2607] = {.lex_state = 1, .external_lex_state = 101},
  [2608] = {.lex_state = 1, .external_lex_state = 103},
  [2609] = {.lex_state = 1, .external_lex_state = 104},
  [2610] = {.lex_state = 1, .external_lex_state = 85},
  [2611] = {.lex_state = 1, .external_lex_state = 103},
  [2612] = {.lex_state = 1, .external_lex_state = 85},
  [2613] = {.lex_state = 2, .external_lex_state = 20},
  [2614] = {.lex_state = 1, .external_lex_state = 85},
  [2615] = {.lex_state = 1, .external_lex_state = 104},
  [2616] = {.lex_state = 1, .external_lex_state = 119},
  [2617] = {.lex_state = 2, .external_lex_state = 22},
  [2618] = {.lex_state = 1, .external_lex_state = 122},
  [2619] = {.lex_state = 2, .external_lex_state = 22},
  [2620] = {.lex_state = 2, .external_lex_state = 22},
  [2621] = {.lex_state = 1, .external_lex_state = 119},
  [2622] = {.lex_state = 1, .external_lex_state = 121},
  [2623] = {.lex_state = 1, .external_lex_state = 114},
  [2624] = {.lex_state = 1, .external_lex_state = 122},
  [2625] = {.lex_state = 1, .external_lex_state = 121},
  [2626] = {.lex_state = 2, .external_lex_state = 132},
  [2627] = {.lex_state = 1, .external_lex_state = 121},
  [2628] = {.lex_state = 1, .external_lex_state = 121},
  [2629] = {.lex_state = 2, .external_lex_state = 22},
  [2630] = {.lex_state = 2, .external_lex_state = 22},
  [2631] = {.lex_state = 1, .external_lex_state = 122},
  [2632] = {.lex_state = 1, .external_lex_state = 113},
  [2633] = {.lex_state = 1, .external_lex_state = 121},
  [2634] = {.lex_state = 1, .external_lex_state = 113},
  [2635] = {.lex_state = 1, .external_lex_state = 114},
  [2636] = {.lex_state = 2, .external_lex_state = 132},
  [2637] = {.lex_state = 2, .external_lex_state = 22},
  [2638] = {.lex_state = 2, .external_lex_state = 22},
  [2639] = {.lex_state = 1, .external_lex_state = 114},
  [2640] = {.lex_state = 1, .external_lex_state = 113},
  [2641] = {.lex_state = 1, .external_lex_state = 121},
  [2642] = {.lex_state = 1, .external_lex_state = 121},
  [2643] = {.lex_state = 1, .external_lex_state = 113},
  [2644] = {.lex_state = 1, .external_lex_state = 114},
  [2645] = {.lex_state = 2, .external_lex_state = 22},
  [2646] = {.lex_state = 1, .external_lex_state = 114},
  [2647] = {.lex_state = 1, .external_lex_state = 121},
  [2648] = {.lex_state = 1, .external_lex_state = 114},
  [2649] = {.lex_state = 1, .external_lex_state = 122},
  [2650] = {.lex_state = 1, .external_lex_state = 114},
  [2651] = {.lex_state = 1, .external_lex_state = 119},
  [2652] = {.lex_state = 1, .external_lex_state = 113},
  [2653] = {.lex_state = 1, .external_lex_state = 119},
  [2654] = {.lex_state = 1, .external_lex_state = 119},
  [2655] = {.lex_state = 1, .external_lex_state = 122},
  [2656] = {.lex_state = 1, .external_lex_state = 122},
  [2657] = {.lex_state = 1, .external_lex_state = 119},
  [2658] = {.lex_state = 2, .external_lex_state = 132},
  [2659] = {.lex_state = 1, .external_lex_state = 114},
  [2660] = {.lex_state = 1, .external_lex_state = 119},
  [2661] = {.lex_state = 1, .external_lex_state = 122},
  [2662] = {.lex_state = 1, .external_lex_state = 114},
  [2663] = {.lex_state = 1, .external_lex_state = 122},
  [2664] = {.lex_state = 1, .external_lex_state = 113},
  [2665] = {.lex_state = 1, .external_lex_state = 114},
  [2666] = {.lex_state = 1, .external_lex_state = 122},
  [2667] = {.lex_state = 1, .external_lex_state = 114},
  [2668] = {.lex_state = 1, .external_lex_state = 114},
  [2669] = {.lex_state = 1, .external_lex_state = 122},
  [2670] = {.lex_state = 1, .external_lex_state = 113},
  [2671] = {.lex_state = 1, .external_lex_state = 119},
  [2672] = {.lex_state = 1, .external_lex_state = 121},
  [2673] = {.lex_state = 1, .external_lex_state = 121},
  [2674] = {.lex_state = 1, .external_lex_state = 119},
  [2675] = {.lex_state = 1, .external_lex_state = 119},
  [2676] = {.lex_state = 2, .external_lex_state = 22},
  [2677] = {.lex_state = 1, .external_lex_state = 113},
  [2678] = {.lex_state = 2, .external_lex_state = 22},
  [2679] = {.lex_state = 1, .external_lex_state = 113},
  [2680] = {.lex_state = 1, .external_lex_state = 119},
  [2681] = {.lex_state = 1, .external_lex_state = 122},
  [2682] = {.lex_state = 1, .external_lex_state = 121},
  [2683] = {.lex_state = 1, .external_lex_state = 121},
  [2684] = {.lex_state = 1, .external_lex_state = 119},
  [2685] = {.lex_state = 1, .external_lex_state = 122},
  [2686] = {.lex_state = 1, .external_lex_state = 122},
  [2687] = {.lex_state = 1, .external_lex_state = 119},
  [2688] = {.lex_state = 1, .external_lex_state = 113},
  [2689] = {.lex_state = 2, .external_lex_state = 132},
  [2690] = {.lex_state = 1, .external_lex_state = 113},
  [2691] = {.lex_state = 2, .external_lex_state = 22},
  [2692] = {.lex_state = 1, .external_lex_state = 113},
  [2693] = {.lex_state = 1, .external_lex_state = 130},
  [2694] = {.lex_state = 1, .external_lex_state = 127},
  [2695] = {.lex_state = 1, .external_lex_state = 127},
  [2696] = {.lex_state = 1, .external_lex_state = 125},
  [2697] = {.lex_state = 1, .external_lex_state = 127},
  [2698] = {.lex_state = 1, .external_lex_state = 127},
  [2699] = {.lex_state = 1, .external_lex_state = 112},
  [2700] = {.lex_state = 1, .external_lex_state = 130},
  [2701] = {.lex_state = 1, .external_lex_state = 130},
  [2702] = {.lex_state = 1, .external_lex_state = 125},
  [2703] = {.lex_state = 1, .external_lex_state = 125},
  [2704] = {.lex_state = 1, .external_lex_state = 125},
  [2705] = {.lex_state = 1, .external_lex_state = 130},
  [2706] = {.lex_state = 1, .external_lex_state = 125},
  [2707] = {.lex_state = 1, .external_lex_state = 129},
  [2708] = {.lex_state = 1, .external_lex_state = 112},
  [2709] = {.lex_state = 1, .external_lex_state = 126},
  [2710] = {.lex_state = 1, .external_lex_state = 126},
  [2711] = {.lex_state = 1, .external_lex_state = 130},
  [2712] = {.lex_state = 1, .external_lex_state = 130},
  [2713] = {.lex_state = 1, .external_lex_state = 126},
  [2714] = {.lex_state = 2, .external_lex_state = 133},
  [2715] = {.lex_state = 1, .external_lex_state = 126},
  [2716] = {.lex_state = 1, .external_lex_state = 126},
  [2717] = {.lex_state = 1, .external_lex_state = 126},
  [2718] = {.lex_state = 1, .external_lex_state = 126},
  [2719] = {.lex_state = 1, .external_lex_state = 112},
  [2720] = {.lex_state = 1, .external_lex_state = 126},
  [2721] = {.lex_state = 1, .external_lex_state = 127},
  [2722] = {.lex_state = 2, .external_lex_state = 133},
  [2723] = {.lex_state = 1, .external_lex_state = 130},
  [2724] = {.lex_state = 2, .external_lex_state = 133},
  [2725] = {.lex_state = 1, .external_lex_state = 126},
  [2726] = {.lex_state = 2, .external_lex_state = 133},
  [2727] = {.lex_state = 1, .external_lex_state = 125},
  [2728] = {.lex_state = 1, .external_lex_state = 130},
  [2729] = {.lex_state = 1, .external_lex_state = 126},
  [2730] = {.lex_state = 1, .external_lex_state = 126},
  [2731] = {.lex_state = 1, .external_lex_state = 130},
  [2732] = {.lex_state = 1, .external_lex_state = 126},
  [2733] = {.lex_state = 1, .external_lex_state = 127},
  [2734] = {.lex_state = 1, .external_lex_state = 125},
  [2735] = {.lex_state = 1, .external_lex_state = 125},
  [2736] = {.lex_state = 1, .external_lex_state = 112},
  [2737] = {.lex_state = 1, .external_lex_state = 129},
  [2738] = {.lex_state = 1, .external_lex_state = 129},
  [2739] = {.lex_state = 1, .external_lex_state = 130},
  [2740] = {.lex_state = 1, .external_lex_state = 129},
  [2741] = {.lex_state = 1, .external_lex_state = 129},
  [2742] = {.lex_state = 1, .external_lex_state = 129},
  [2743] = {.lex_state = 1, .external_lex_state = 129},
  [2744] = {.lex_state = 1, .external_lex_state = 130},
  [2745] = {.lex_state = 1, .external_lex_state = 129},
  [2746] = {.lex_state = 1, .external_lex_state = 125},
  [2747] = {.lex_state = 1, .external_lex_state = 127},
  [2748] = {.lex_state = 1, .external_lex_state = 127},
  [2749] = {.lex_state = 1, .external_lex_state = 127},
  [2750] = {.lex_state = 1, .external_lex_state = 130},
  [2751] = {.lex_state = 1, .external_lex_state = 127},
  [2752] = {.lex_state = 1, .external_lex_state = 125},
  [2753] = {.lex_state = 1, .external_lex_state = 125},
  [2754] = {.lex_state = 2, .external_lex_state = 133},
  [2755] = {.lex_state = 1, .external_lex_state = 129},
  [2756] = {.lex_state = 1, .external_lex_state = 112},
  [2757] = {.lex_state = 1, .external_lex_state = 125},
  [2758] = {.lex_state = 1, .external_lex_state = 127},
  [2759] = {.lex_state = 1, .external_lex_state = 129},
  [2760] = {.lex_state = 1, .external_lex_state = 129},
  [2761] = {.lex_state = 1, .external_lex_state = 112},
  [2762] = {.lex_state = 1, .external_lex_state = 112},
  [2763] = {.lex_state = 1, .external_lex_state = 112},
  [2764] = {.lex_state = 1, .external_lex_state = 112},
  [2765] = {.lex_state = 1, .external_lex_state = 129},
  [2766] = {.lex_state = 1, .external_lex_state = 127},
  [2767] = {.lex_state = 1, .external_lex_state = 129},
  [2768] = {.lex_state = 1, .external_lex_state = 127},
  [2769] = {.lex_state = 1, .external_lex_state = 112},
  [2770] = {.lex_state = 1, .external_lex_state = 112},
  [2771] = {.lex_state = 1, .external_lex_state = 112},
  [2772] = {.lex_state = 1, .external_lex_state = 112},
  [2773] = {.lex_state = 0, .external_lex_state = 134},
  [2774] = {.lex_state = 0, .external_lex_state = 134},
  [2775] = {.lex_state = 0, .external_lex_state = 134},
  [2776] = {.lex_state = 0, .external_lex_state = 134},
  [2777] = {.lex_state = 0, .external_lex_state = 134},
  [2778] = {.lex_state = 0, .external_lex_state = 134},
  [2779] = {.lex_state = 0, .external_lex_state = 134},
  [2780] = {.lex_state = 0, .external_lex_state = 134},
  [2781] = {.lex_state = 0, .external_lex_state = 134},
  [2782] = {.lex_state = 0, .external_lex_state = 134},
  [2783] = {.lex_state = 0, .external_lex_state = 134},
  [2784] = {.lex_state = 0, .external_lex_state = 134},
  [2785] = {.lex_state = 0, .external_lex_state = 134},
  [2786] = {.lex_state = 0, .external_lex_state = 134},
  [2787] = {.lex_state = 0, .external_lex_state = 134},
  [2788] = {.lex_state = 0, .external_lex_state = 134},
  [2789] = {.lex_state = 0, .external_lex_state = 134},
  [2790] = {.lex_state = 0, .external_lex_state = 134},
  [2791] = {.lex_state = 0, .external_lex_state = 134},
  [2792] = {.lex_state = 0, .external_lex_state = 134},
  [2793] = {.lex_state = 0, .external_lex_state = 134},
  [2794] = {.lex_state = 2, .external_lex_state = 135},
  [2795] = {.lex_state = 0, .external_lex_state = 134},
  [2796] = {.lex_state = 0, .external_lex_state = 134},
  [2797] = {.lex_state = 2, .external_lex_state = 135},
  [2798] = {.lex_state = 2, .external_lex_state = 135},
  [2799] = {.lex_state = 2, .external_lex_state = 136},
  [2800] = {.lex_state = 2, .external_lex_state = 137},
  [2801] = {.lex_state = 2, .external_lex_state = 138},
  [2802] = {.lex_state = 2, .external_lex_state = 138},
  [2803] = {.lex_state = 2, .external_lex_state = 137},
  [2804] = {.lex_state = 2, .external_lex_state = 136},
  [2805] = {.lex_state = 2, .external_lex_state = 137},
  [2806] = {.lex_state = 2, .external_lex_state = 138},
  [2807] = {.lex_state = 2, .external_lex_state = 136},
  [2808] = {.lex_state = 2, .external_lex_state = 6},
  [2809] = {.lex_state = 2, .external_lex_state = 6},
  [2810] = {.lex_state = 2, .external_lex_state = 6},
  [2811] = {.lex_state = 2, .external_lex_state = 6},
  [2812] = {.lex_state = 2, .external_lex_state = 6},
  [2813] = {.lex_state = 2, .external_lex_state = 6},
  [2814] = {.lex_state = 2, .external_lex_state = 6},
  [2815] = {.lex_state = 2, .external_lex_state = 6},
  [2816] = {.lex_state = 2, .external_lex_state = 6},
  [2817] = {.lex_state = 2, .external_lex_state = 6},
  [2818] = {.lex_state = 2, .external_lex_state = 6},
  [2819] = {.lex_state = 2, .external_lex_state = 6},
  [2820] = {.lex_state = 2, .external_lex_state = 6},
  [2821] = {.lex_state = 2, .external_lex_state = 6},
  [2822] = {.lex_state = 2, .external_lex_state = 6},
  [2823] = {.lex_state = 2, .external_lex_state = 6},
  [2824] = {.lex_state = 2, .external_lex_state = 6},
  [2825] = {.lex_state = 2, .external_lex_state = 6},
  [2826] = {.lex_state = 2, .external_lex_state = 6},
  [2827] = {.lex_state = 2, .external_lex_state = 6},
  [2828] = {.lex_state = 2, .external_lex_state = 6},
  [2829] = {.lex_state = 2, .external_lex_state = 6},
  [2830] = {.lex_state = 2, .external_lex_state = 6},
  [2831] = {.lex_state = 2, .external_lex_state = 6},
  [2832] = {.lex_state = 2, .external_lex_state = 6},
  [2833] = {.lex_state = 2, .external_lex_state = 6},
  [2834] = {.lex_state = 2, .external_lex_state = 6},
  [2835] = {.lex_state = 2, .external_lex_state = 6},
  [2836] = {.lex_state = 2, .external_lex_state = 6},
  [2837] = {.lex_state = 2, .external_lex_state = 139},
  [2838] = {.lex_state = 2, .external_lex_state = 6},
  [2839] = {.lex_state = 2, .external_lex_state = 6},
  [2840] = {.lex_state = 2, .external_lex_state = 6},
  [2841] = {.lex_state = 2, .external_lex_state = 6},
  [2842] = {.lex_state = 2, .external_lex_state = 6},
  [2843] = {.lex_state = 2, .external_lex_state = 6},
  [2844] = {.lex_state = 2, .external_lex_state = 6},
  [2845] = {.lex_state = 2, .external_lex_state = 6},
  [2846] = {.lex_state = 2, .external_lex_state = 6},
  [2847] = {.lex_state = 2, .external_lex_state = 6},
  [2848] = {.lex_state = 2, .external_lex_state = 6},
  [2849] = {.lex_state = 2, .external_lex_state = 139},
  [2850] = {.lex_state = 2, .external_lex_state = 6},
  [2851] = {.lex_state = 2, .external_lex_state = 6},
  [2852] = {.lex_state = 2, .external_lex_state = 6},
  [2853] = {.lex_state = 2, .external_lex_state = 6},
  [2854] = {.lex_state = 2, .external_lex_state = 6},
  [2855] = {.lex_state = 2, .external_lex_state = 6},
  [2856] = {.lex_state = 2, .external_lex_state = 6},
  [2857] = {.lex_state = 2, .external_lex_state = 6},
  [2858] = {.lex_state = 2, .external_lex_state = 6},
  [2859] = {.lex_state = 2, .external_lex_state = 139},
  [2860] = {.lex_state = 2, .external_lex_state = 6},
  [2861] = {.lex_state = 2, .external_lex_state = 6},
  [2862] = {.lex_state = 2, .external_lex_state = 6},
  [2863] = {.lex_state = 2, .external_lex_state = 6},
  [2864] = {.lex_state = 2, .external_lex_state = 6},
  [2865] = {.lex_state = 2, .external_lex_state = 6},
  [2866] = {.lex_state = 2, .external_lex_state = 6},
  [2867] = {.lex_state = 2, .external_lex_state = 6},
  [2868] = {.lex_state = 2, .external_lex_state = 6},
  [2869] = {.lex_state = 2, .external_lex_state = 6},
  [2870] = {.lex_state = 2, .external_lex_state = 8},
  [2871] = {.lex_state = 2, .external_lex_state = 8},
  [2872] = {.lex_state = 2, .external_lex_state = 8},
  [2873] = {.lex_state = 2, .external_lex_state = 8},
  [2874] = {.lex_state = 2, .external_lex_state = 135},
  [2875] = {.lex_state = 2, .external_lex_state = 8},
  [2876] = {.lex_state = 2, .external_lex_state = 8},
  [2877] = {.lex_state = 2, .external_lex_state = 8},
  [2878] = {.lex_state = 2, .external_lex_state = 8},
  [2879] = {.lex_state = 2, .external_lex_state = 8},
  [2880] = {.lex_state = 2, .external_lex_state = 8},
  [2881] = {.lex_state = 2, .external_lex_state = 8},
  [2882] = {.lex_state = 2, .external_lex_state = 8},
  [2883] = {.lex_state = 2, .external_lex_state = 8},
  [2884] = {.lex_state = 2, .external_lex_state = 8},
  [2885] = {.lex_state = 2, .external_lex_state = 8},
  [2886] = {.lex_state = 2, .external_lex_state = 8},
  [2887] = {.lex_state = 0, .external_lex_state = 140},
  [2888] = {.lex_state = 2, .external_lex_state = 8},
  [2889] = {.lex_state = 2, .external_lex_state = 8},
  [2890] = {.lex_state = 2, .external_lex_state = 8},
  [2891] = {.lex_state = 2, .external_lex_state = 8},
  [2892] = {.lex_state = 2, .external_lex_state = 8},
  [2893] = {.lex_state = 2, .external_lex_state = 8},
  [2894] = {.lex_state = 2, .external_lex_state = 8},
  [2895] = {.lex_state = 2, .external_lex_state = 8},
  [2896] = {.lex_state = 2, .external_lex_state = 8},
  [2897] = {.lex_state = 2, .external_lex_state = 8},
  [2898] = {.lex_state = 2, .external_lex_state = 8},
  [2899] = {.lex_state = 2, .external_lex_state = 8},
  [2900] = {.lex_state = 2, .external_lex_state = 135},
  [2901] = {.lex_state = 2, .external_lex_state = 8},
  [2902] = {.lex_state = 2, .external_lex_state = 8},
  [2903] = {.lex_state = 2, .external_lex_state = 8},
  [2904] = {.lex_state = 2, .external_lex_state = 8},
  [2905] = {.lex_state = 2, .external_lex_state = 8},
  [2906] = {.lex_state = 2, .external_lex_state = 8},
  [2907] = {.lex_state = 2, .external_lex_state = 8},
  [2908] = {.lex_state = 2, .external_lex_state = 8},
  [2909] = {.lex_state = 2, .external_lex_state = 8},
  [2910] = {.lex_state = 2, .external_lex_state = 8},
  [2911] = {.lex_state = 2, .external_lex_state = 8},
  [2912] = {.lex_state = 2, .external_lex_state = 8},
  [2913] = {.lex_state = 2, .external_lex_state = 8},
  [2914] = {.lex_state = 2, .external_lex_state = 8},
  [2915] = {.lex_state = 2, .external_lex_state = 8},
  [2916] = {.lex_state = 2, .external_lex_state = 8},
  [2917] = {.lex_state = 2, .external_lex_state = 8},
  [2918] = {.lex_state = 2, .external_lex_state = 8},
  [2919] = {.lex_state = 2, .external_lex_state = 8},
  [2920] = {.lex_state = 2, .external_lex_state = 8},
  [2921] = {.lex_state = 2, .external_lex_state = 8},
  [2922] = {.lex_state = 2, .external_lex_state = 8},
  [2923] = {.lex_state = 2, .external_lex_state = 8},
  [2924] = {.lex_state = 2, .external_lex_state = 8},
  [2925] = {.lex_state = 2, .external_lex_state = 8},
  [2926] = {.lex_state = 2, .external_lex_state = 8},
  [2927] = {.lex_state = 2, .external_lex_state = 8},
  [2928] = {.lex_state = 2, .external_lex_state = 8},
  [2929] = {.lex_state = 2, .external_lex_state = 8},
  [2930] = {.lex_state = 2, .external_lex_state = 8},
  [2931] = {.lex_state = 2, .external_lex_state = 8},
  [2932] = {.lex_state = 2, .external_lex_state = 9},
  [2933] = {.lex_state = 2, .external_lex_state = 9},
  [2934] = {.lex_state = 2, .external_lex_state = 9},
  [2935] = {.lex_state = 2, .external_lex_state = 141},
  [2936] = {.lex_state = 2, .external_lex_state = 9},
  [2937] = {.lex_state = 2, .external_lex_state = 9},
  [2938] = {.lex_state = 2, .external_lex_state = 9},
  [2939] = {.lex_state = 2, .external_lex_state = 9},
  [2940] = {.lex_state = 2, .external_lex_state = 9},
  [2941] = {.lex_state = 2, .external_lex_state = 9},
  [2942] = {.lex_state = 2, .external_lex_state = 9},
  [2943] = {.lex_state = 2, .external_lex_state = 9},
  [2944] = {.lex_state = 2, .external_lex_state = 9},
  [2945] = {.lex_state = 2, .external_lex_state = 9},
  [2946] = {.lex_state = 2, .external_lex_state = 9},
  [2947] = {.lex_state = 2, .external_lex_state = 9},
  [2948] = {.lex_state = 2, .external_lex_state = 9},
  [2949] = {.lex_state = 2, .external_lex_state = 9},
  [2950] = {.lex_state = 2, .external_lex_state = 9},
  [2951] = {.lex_state = 2, .external_lex_state = 9},
  [2952] = {.lex_state = 2, .external_lex_state = 9},
  [2953] = {.lex_state = 2, .external_lex_state = 9},
  [2954] = {.lex_state = 2, .external_lex_state = 9},
  [2955] = {.lex_state = 2, .external_lex_state = 9},
  [2956] = {.lex_state = 2, .external_lex_state = 9},
  [2957] = {.lex_state = 2, .external_lex_state = 9},
  [2958] = {.lex_state = 2, .external_lex_state = 9},
  [2959] = {.lex_state = 0, .external_lex_state = 23},
  [2960] = {.lex_state = 2, .external_lex_state = 9},
  [2961] = {.lex_state = 2, .external_lex_state = 9},
  [2962] = {.lex_state = 2, .external_lex_state = 141},
  [2963] = {.lex_state = 0, .external_lex_state = 142},
  [2964] = {.lex_state = 2, .external_lex_state = 9},
  [2965] = {.lex_state = 2, .external_lex_state = 9},
  [2966] = {.lex_state = 2, .external_lex_state = 9},
  [2967] = {.lex_state = 2, .external_lex_state = 9},
  [2968] = {.lex_state = 2, .external_lex_state = 9},
  [2969] = {.lex_state = 2, .external_lex_state = 9},
  [2970] = {.lex_state = 2, .external_lex_state = 9},
  [2971] = {.lex_state = 2, .external_lex_state = 9},
  [2972] = {.lex_state = 2, .external_lex_state = 9},
  [2973] = {.lex_state = 2, .external_lex_state = 9},
  [2974] = {.lex_state = 2, .external_lex_state = 9},
  [2975] = {.lex_state = 2, .external_lex_state = 9},
  [2976] = {.lex_state = 2, .external_lex_state = 9},
  [2977] = {.lex_state = 2, .external_lex_state = 9},
  [2978] = {.lex_state = 2, .external_lex_state = 9},
  [2979] = {.lex_state = 2, .external_lex_state = 9},
  [2980] = {.lex_state = 2, .external_lex_state = 9},
  [2981] = {.lex_state = 2, .external_lex_state = 9},
  [2982] = {.lex_state = 2, .external_lex_state = 9},
  [2983] = {.lex_state = 2, .external_lex_state = 9},
  [2984] = {.lex_state = 2, .external_lex_state = 9},
  [2985] = {.lex_state = 2, .external_lex_state = 9},
  [2986] = {.lex_state = 2, .external_lex_state = 9},
  [2987] = {.lex_state = 2, .external_lex_state = 9},
  [2988] = {.lex_state = 2, .external_lex_state = 9},
  [2989] = {.lex_state = 2, .external_lex_state = 9},
  [2990] = {.lex_state = 2, .external_lex_state = 141},
  [2991] = {.lex_state = 2, .external_lex_state = 9},
  [2992] = {.lex_state = 2, .external_lex_state = 9},
  [2993] = {.lex_state = 2, .external_lex_state = 9},
  [2994] = {.lex_state = 2, .external_lex_state = 9},
  [2995] = {.lex_state = 2, .external_lex_state = 9},
  [2996] = {.lex_state = 2, .external_lex_state = 143},
  [2997] = {.lex_state = 0, .external_lex_state = 144},
  [2998] = {.lex_state = 0, .external_lex_state = 28},
  [2999] = {.lex_state = 2, .external_lex_state = 143},
  [3000] = {.lex_state = 0, .external_lex_state = 144},
  [3001] = {.lex_state = 0, .external_lex_state = 144},
  [3002] = {.lex_state = 2, .external_lex_state = 143},
  [3003] = {.lex_state = 0, .external_lex_state = 28},
  [3004] = {.lex_state = 2, .external_lex_state = 143},
  [3005] = {.lex_state = 2, .external_lex_state = 143},
  [3006] = {.lex_state = 2, .external_lex_state = 143},
  [3007] = {.lex_state = 2, .external_lex_state = 143},
  [3008] = {.lex_state = 0, .external_lex_state = 144},
  [3009] = {.lex_state = 0, .external_lex_state = 144},
  [3010] = {.lex_state = 2, .external_lex_state = 95},
  [3011] = {.lex_state = 2, .external_lex_state = 143},
  [3012] = {.lex_state = 0, .external_lex_state = 144},
  [3013] = {.lex_state = 2, .external_lex_state = 145},
  [3014] = {.lex_state = 0, .external_lex_state = 142},
  [3015] = {.lex_state = 0, .external_lex_state = 146},
  [3016] = {.lex_state = 0, .external_lex_state = 142},
  [3017] = {.lex_state = 0, .external_lex_state = 34},
  [3018] = {.lex_state = 0, .external_lex_state = 142},
  [3019] = {.lex_state = 0, .external_lex_state = 142},
  [3020] = {.lex_state = 0, .external_lex_state = 142},
  [3021] = {.lex_state = 0, .external_lex_state = 147},
  [3022] = {.lex_state = 0, .external_lex_state = 144},
  [3023] = {.lex_state = 0, .external_lex_state = 147},
  [3024] = {.lex_state = 0, .external_lex_state = 34},
  [3025] = {.lex_state = 0, .external_lex_state = 147},
  [3026] = {.lex_state = 2, .external_lex_state = 148},
  [3027] = {.lex_state = 2, .external_lex_state = 148},
  [3028] = {.lex_state = 2, .external_lex_state = 148},
  [3029] = {.lex_state = 2, .external_lex_state = 148},
  [3030] = {.lex_state = 0, .external_lex_state = 46},
  [3031] = {.lex_state = 0, .external_lex_state = 149},
  [3032] = {.lex_state = 0, .external_lex_state = 46},
  [3033] = {.lex_state = 0, .external_lex_state = 46},
  [3034] = {.lex_state = 2, .external_lex_state = 148},
  [3035] = {.lex_state = 0, .external_lex_state = 46},
  [3036] = {.lex_state = 0, .external_lex_state = 46},
  [3037] = {.lex_state = 0, .external_lex_state = 46},
  [3038] = {.lex_state = 0, .external_lex_state = 46},
  [3039] = {.lex_state = 0, .external_lex_state = 46},
  [3040] = {.lex_state = 0, .external_lex_state = 46},
  [3041] = {.lex_state = 0, .external_lex_state = 46},
  [3042] = {.lex_state = 0, .external_lex_state = 46},
  [3043] = {.lex_state = 0, .external_lex_state = 46},
  [3044] = {.lex_state = 0, .external_lex_state = 59},
  [3045] = {.lex_state = 0, .external_lex_state = 46},
  [3046] = {.lex_state = 0, .external_lex_state = 46},
  [3047] = {.lex_state = 0, .external_lex_state = 46},
  [3048] = {.lex_state = 0, .external_lex_state = 46},
  [3049] = {.lex_state = 0, .external_lex_state = 46},
  [3050] = {.lex_state = 0, .external_lex_state = 46},
  [3051] = {.lex_state = 0, .external_lex_state = 144},
  [3052] = {.lex_state = 0, .external_lex_state = 46},
  [3053] = {.lex_state = 0, .external_lex_state = 46},
  [3054] = {.lex_state = 0, .external_lex_state = 46},
  [3055] = {.lex_state = 0, .external_lex_state = 59},
  [3056] = {.lex_state = 0, .external_lex_state = 46},
  [3057] = {.lex_state = 0, .external_lex_state = 46},
  [3058] = {.lex_state = 0, .external_lex_state = 46},
  [3059] = {.lex_state = 2, .external_lex_state = 135},
  [3060] = {.lex_state = 0, .external_lex_state = 46},
  [3061] = {.lex_state = 0, .external_lex_state = 46},
  [3062] = {.lex_state = 0, .external_lex_state = 46},
  [3063] = {.lex_state = 0, .external_lex_state = 46},
  [3064] = {.lex_state = 0, .external_lex_state = 46},
  [3065] = {.lex_state = 0, .external_lex_state = 46},
  [3066] = {.lex_state = 2, .external_lex_state = 148},
  [3067] = {.lex_state = 2, .external_lex_state = 150},
  [3068] = {.lex_state = 0, .external_lex_state = 146},
  [3069] = {.lex_state = 0, .external_lex_state = 146},
  [3070] = {.lex_state = 0, .external_lex_state = 46},
  [3071] = {.lex_state = 0, .external_lex_state = 46},
  [3072] = {.lex_state = 2, .external_lex_state = 148},
  [3073] = {.lex_state = 2, .external_lex_state = 148},
  [3074] = {.lex_state = 0, .external_lex_state = 149},
  [3075] = {.lex_state = 0, .external_lex_state = 149},
  [3076] = {.lex_state = 0, .external_lex_state = 46},
  [3077] = {.lex_state = 0, .external_lex_state = 151},
  [3078] = {.lex_state = 0, .external_lex_state = 46},
  [3079] = {.lex_state = 2, .external_lex_state = 148},
  [3080] = {.lex_state = 2, .external_lex_state = 148},
  [3081] = {.lex_state = 2, .external_lex_state = 148},
  [3082] = {.lex_state = 0, .external_lex_state = 46},
  [3083] = {.lex_state = 0, .external_lex_state = 46},
  [3084] = {.lex_state = 2, .external_lex_state = 148},
  [3085] = {.lex_state = 2, .external_lex_state = 148},
  [3086] = {.lex_state = 2, .external_lex_state = 148},
  [3087] = {.lex_state = 0, .external_lex_state = 46},
  [3088] = {.lex_state = 2, .external_lex_state = 148},
  [3089] = {.lex_state = 2, .external_lex_state = 148},
  [3090] = {.lex_state = 2, .external_lex_state = 148},
  [3091] = {.lex_state = 2, .external_lex_state = 148},
  [3092] = {.lex_state = 2, .external_lex_state = 148},
  [3093] = {.lex_state = 2, .external_lex_state = 148},
  [3094] = {.lex_state = 2, .external_lex_state = 148},
  [3095] = {.lex_state = 2, .external_lex_state = 148},
  [3096] = {.lex_state = 2, .external_lex_state = 148},
  [3097] = {.lex_state = 2, .external_lex_state = 148},
  [3098] = {.lex_state = 2, .external_lex_state = 148},
  [3099] = {.lex_state = 2, .external_lex_state = 148},
  [3100] = {.lex_state = 2, .external_lex_state = 148},
  [3101] = {.lex_state = 2, .external_lex_state = 148},
  [3102] = {.lex_state = 2, .external_lex_state = 148},
  [3103] = {.lex_state = 2, .external_lex_state = 148},
  [3104] = {.lex_state = 2, .external_lex_state = 148},
  [3105] = {.lex_state = 2, .external_lex_state = 148},
  [3106] = {.lex_state = 2, .external_lex_state = 148},
  [3107] = {.lex_state = 2, .external_lex_state = 148},
  [3108] = {.lex_state = 2, .external_lex_state = 148},
  [3109] = {.lex_state = 2, .external_lex_state = 148},
  [3110] = {.lex_state = 2, .external_lex_state = 148},
  [3111] = {.lex_state = 2, .external_lex_state = 148},
  [3112] = {.lex_state = 2, .external_lex_state = 148},
  [3113] = {.lex_state = 2, .external_lex_state = 148},
  [3114] = {.lex_state = 2, .external_lex_state = 148},
  [3115] = {.lex_state = 2, .external_lex_state = 148},
  [3116] = {.lex_state = 0, .external_lex_state = 46},
  [3117] = {.lex_state = 2, .external_lex_state = 148},
  [3118] = {.lex_state = 2, .external_lex_state = 148},
  [3119] = {.lex_state = 2, .external_lex_state = 148},
  [3120] = {.lex_state = 2, .external_lex_state = 148},
  [3121] = {.lex_state = 2, .external_lex_state = 148},
  [3122] = {.lex_state = 2, .external_lex_state = 148},
  [3123] = {.lex_state = 2, .external_lex_state = 148},
  [3124] = {.lex_state = 2, .external_lex_state = 148},
  [3125] = {.lex_state = 2, .external_lex_state = 148},
  [3126] = {.lex_state = 2, .external_lex_state = 148},
  [3127] = {.lex_state = 2, .external_lex_state = 148},
  [3128] = {.lex_state = 2, .external_lex_state = 148},
  [3129] = {.lex_state = 2, .external_lex_state = 148},
  [3130] = {.lex_state = 2, .external_lex_state = 148},
  [3131] = {.lex_state = 2, .external_lex_state = 148},
  [3132] = {.lex_state = 2, .external_lex_state = 148},
  [3133] = {.lex_state = 2, .external_lex_state = 148},
  [3134] = {.lex_state = 2, .external_lex_state = 148},
  [3135] = {.lex_state = 2, .external_lex_state = 148},
  [3136] = {.lex_state = 2, .external_lex_state = 148},
  [3137] = {.lex_state = 2, .external_lex_state = 148},
  [3138] = {.lex_state = 2, .external_lex_state = 148},
  [3139] = {.lex_state = 2, .external_lex_state = 148},
  [3140] = {.lex_state = 2, .external_lex_state = 148},
  [3141] = {.lex_state = 2, .external_lex_state = 148},
  [3142] = {.lex_state = 2, .external_lex_state = 148},
  [3143] = {.lex_state = 2, .external_lex_state = 148},
  [3144] = {.lex_state = 2, .external_lex_state = 148},
  [3145] = {.lex_state = 2, .external_lex_state = 148},
  [3146] = {.lex_state = 2, .external_lex_state = 148},
  [3147] = {.lex_state = 2, .external_lex_state = 148},
  [3148] = {.lex_state = 2, .external_lex_state = 148},
  [3149] = {.lex_state = 0, .external_lex_state = 46},
  [3150] = {.lex_state = 2, .external_lex_state = 148},
  [3151] = {.lex_state = 2, .external_lex_state = 148},
  [3152] = {.lex_state = 2, .external_lex_state = 148},
  [3153] = {.lex_state = 2, .external_lex_state = 148},
  [3154] = {.lex_state = 2, .external_lex_state = 148},
  [3155] = {.lex_state = 2, .external_lex_state = 148},
  [3156] = {.lex_state = 2, .external_lex_state = 148},
  [3157] = {.lex_state = 2, .external_lex_state = 148},
  [3158] = {.lex_state = 2, .external_lex_state = 148},
  [3159] = {.lex_state = 2, .external_lex_state = 148},
  [3160] = {.lex_state = 2, .external_lex_state = 148},
  [3161] = {.lex_state = 2, .external_lex_state = 148},
  [3162] = {.lex_state = 2, .external_lex_state = 148},
  [3163] = {.lex_state = 2, .external_lex_state = 148},
  [3164] = {.lex_state = 2, .external_lex_state = 148},
  [3165] = {.lex_state = 2, .external_lex_state = 148},
  [3166] = {.lex_state = 2, .external_lex_state = 148},
  [3167] = {.lex_state = 2, .external_lex_state = 148},
  [3168] = {.lex_state = 2, .external_lex_state = 148},
  [3169] = {.lex_state = 2, .external_lex_state = 148},
  [3170] = {.lex_state = 2, .external_lex_state = 148},
  [3171] = {.lex_state = 2, .external_lex_state = 148},
  [3172] = {.lex_state = 2, .external_lex_state = 148},
  [3173] = {.lex_state = 2, .external_lex_state = 148},
  [3174] = {.lex_state = 2, .external_lex_state = 148},
  [3175] = {.lex_state = 2, .external_lex_state = 148},
  [3176] = {.lex_state = 2, .external_lex_state = 148},
  [3177] = {.lex_state = 2, .external_lex_state = 148},
  [3178] = {.lex_state = 2, .external_lex_state = 148},
  [3179] = {.lex_state = 2, .external_lex_state = 148},
  [3180] = {.lex_state = 2, .external_lex_state = 148},
  [3181] = {.lex_state = 2, .external_lex_state = 148},
  [3182] = {.lex_state = 2, .external_lex_state = 148},
  [3183] = {.lex_state = 2, .external_lex_state = 148},
  [3184] = {.lex_state = 2, .external_lex_state = 148},
  [3185] = {.lex_state = 2, .external_lex_state = 148},
  [3186] = {.lex_state = 2, .external_lex_state = 148},
  [3187] = {.lex_state = 2, .external_lex_state = 148},
  [3188] = {.lex_state = 2, .external_lex_state = 148},
  [3189] = {.lex_state = 2, .external_lex_state = 148},
  [3190] = {.lex_state = 2, .external_lex_state = 148},
  [3191] = {.lex_state = 2, .external_lex_state = 148},
  [3192] = {.lex_state = 2, .external_lex_state = 148},
  [3193] = {.lex_state = 2, .external_lex_state = 148},
  [3194] = {.lex_state = 2, .external_lex_state = 148},
  [3195] = {.lex_state = 2, .external_lex_state = 148},
  [3196] = {.lex_state = 2, .external_lex_state = 148},
  [3197] = {.lex_state = 2, .external_lex_state = 148},
  [3198] = {.lex_state = 2, .external_lex_state = 148},
  [3199] = {.lex_state = 2, .external_lex_state = 148},
  [3200] = {.lex_state = 2, .external_lex_state = 148},
  [3201] = {.lex_state = 2, .external_lex_state = 148},
  [3202] = {.lex_state = 2, .external_lex_state = 148},
  [3203] = {.lex_state = 2, .external_lex_state = 148},
  [3204] = {.lex_state = 2, .external_lex_state = 148},
  [3205] = {.lex_state = 2, .external_lex_state = 148},
  [3206] = {.lex_state = 2, .external_lex_state = 148},
  [3207] = {.lex_state = 2, .external_lex_state = 148},
  [3208] = {.lex_state = 2, .external_lex_state = 148},
  [3209] = {.lex_state = 2, .external_lex_state = 148},
  [3210] = {.lex_state = 2, .external_lex_state = 148},
  [3211] = {.lex_state = 2, .external_lex_state = 148},
  [3212] = {.lex_state = 2, .external_lex_state = 148},
  [3213] = {.lex_state = 0, .external_lex_state = 46},
  [3214] = {.lex_state = 2, .external_lex_state = 148},
  [3215] = {.lex_state = 2, .external_lex_state = 148},
  [3216] = {.lex_state = 2, .external_lex_state = 148},
  [3217] = {.lex_state = 0, .external_lex_state = 46},
  [3218] = {.lex_state = 2, .external_lex_state = 148},
  [3219] = {.lex_state = 2, .external_lex_state = 148},
  [3220] = {.lex_state = 2, .external_lex_state = 148},
  [3221] = {.lex_state = 2, .external_lex_state = 148},
  [3222] = {.lex_state = 2, .external_lex_state = 148},
  [3223] = {.lex_state = 2, .external_lex_state = 148},
  [3224] = {.lex_state = 0, .external_lex_state = 46},
  [3225] = {.lex_state = 2, .external_lex_state = 148},
  [3226] = {.lex_state = 2, .external_lex_state = 148},
  [3227] = {.lex_state = 2, .external_lex_state = 148},
  [3228] = {.lex_state = 2, .external_lex_state = 148},
  [3229] = {.lex_state = 2, .external_lex_state = 148},
  [3230] = {.lex_state = 2, .external_lex_state = 148},
  [3231] = {.lex_state = 2, .external_lex_state = 148},
  [3232] = {.lex_state = 2, .external_lex_state = 148},
  [3233] = {.lex_state = 2, .external_lex_state = 148},
  [3234] = {.lex_state = 2, .external_lex_state = 148},
  [3235] = {.lex_state = 2, .external_lex_state = 148},
  [3236] = {.lex_state = 2, .external_lex_state = 148},
  [3237] = {.lex_state = 0, .external_lex_state = 62},
  [3238] = {.lex_state = 0, .external_lex_state = 62},
  [3239] = {.lex_state = 0, .external_lex_state = 62},
  [3240] = {.lex_state = 0, .external_lex_state = 66},
  [3241] = {.lex_state = 0, .external_lex_state = 62},
  [3242] = {.lex_state = 0, .external_lex_state = 66},
  [3243] = {.lex_state = 0, .external_lex_state = 62},
  [3244] = {.lex_state = 0, .external_lex_state = 62},
  [3245] = {.lex_state = 0, .external_lex_state = 66},
  [3246] = {.lex_state = 0, .external_lex_state = 66},
  [3247] = {.lex_state = 0, .external_lex_state = 66},
  [3248] = {.lex_state = 0, .external_lex_state = 66},
  [3249] = {.lex_state = 0, .external_lex_state = 66},
  [3250] = {.lex_state = 0, .external_lex_state = 66},
  [3251] = {.lex_state = 0, .external_lex_state = 66},
  [3252] = {.lex_state = 0, .external_lex_state = 66},
  [3253] = {.lex_state = 0, .external_lex_state = 66},
  [3254] = {.lex_state = 0, .external_lex_state = 66},
  [3255] = {.lex_state = 0, .external_lex_state = 66},
  [3256] = {.lex_state = 0, .external_lex_state = 66},
  [3257] = {.lex_state = 0, .external_lex_state = 66},
  [3258] = {.lex_state = 0, .external_lex_state = 66},
  [3259] = {.lex_state = 0, .external_lex_state = 66},
  [3260] = {.lex_state = 0, .external_lex_state = 66},
  [3261] = {.lex_state = 0, .external_lex_state = 66},
  [3262] = {.lex_state = 2, .external_lex_state = 135},
  [3263] = {.lex_state = 0, .external_lex_state = 66},
  [3264] = {.lex_state = 0, .external_lex_state = 66},
  [3265] = {.lex_state = 0, .external_lex_state = 66},
  [3266] = {.lex_state = 0, .external_lex_state = 66},
  [3267] = {.lex_state = 0, .external_lex_state = 66},
  [3268] = {.lex_state = 0, .external_lex_state = 66},
  [3269] = {.lex_state = 0, .external_lex_state = 66},
  [3270] = {.lex_state = 2, .external_lex_state = 136},
  [3271] = {.lex_state = 0, .external_lex_state = 66},
  [3272] = {.lex_state = 0, .external_lex_state = 66},
  [3273] = {.lex_state = 0, .external_lex_state = 66},
  [3274] = {.lex_state = 0, .external_lex_state = 66},
  [3275] = {.lex_state = 2, .external_lex_state = 152},
  [3276] = {.lex_state = 0, .external_lex_state = 66},
  [3277] = {.lex_state = 0, .external_lex_state = 62},
  [3278] = {.lex_state = 2, .external_lex_state = 135},
  [3279] = {.lex_state = 2, .external_lex_state = 138},
  [3280] = {.lex_state = 2, .external_lex_state = 135},
  [3281] = {.lex_state = 0, .external_lex_state = 153},
  [3282] = {.lex_state = 0, .external_lex_state = 62},
  [3283] = {.lex_state = 2, .external_lex_state = 135},
  [3284] = {.lex_state = 0, .external_lex_state = 66},
  [3285] = {.lex_state = 2, .external_lex_state = 135},
  [3286] = {.lex_state = 2, .external_lex_state = 135},
  [3287] = {.lex_state = 0, .external_lex_state = 153},
  [3288] = {.lex_state = 0, .external_lex_state = 62},
  [3289] = {.lex_state = 2, .external_lex_state = 135},
  [3290] = {.lex_state = 2, .external_lex_state = 135},
  [3291] = {.lex_state = 2, .external_lex_state = 135},
  [3292] = {.lex_state = 2, .external_lex_state = 135},
  [3293] = {.lex_state = 2, .external_lex_state = 135},
  [3294] = {.lex_state = 2, .external_lex_state = 135},
  [3295] = {.lex_state = 2, .external_lex_state = 135},
  [3296] = {.lex_state = 0, .external_lex_state = 62},
  [3297] = {.lex_state = 0, .external_lex_state = 62},
  [3298] = {.lex_state = 0, .external_lex_state = 62},
  [3299] = {.lex_state = 0, .external_lex_state = 62},
  [3300] = {.lex_state = 2, .external_lex_state = 135},
  [3301] = {.lex_state = 2, .external_lex_state = 135},
  [3302] = {.lex_state = 2, .external_lex_state = 135},
  [3303] = {.lex_state = 2, .external_lex_state = 135},
  [3304] = {.lex_state = 2, .external_lex_state = 135},
  [3305] = {.lex_state = 2, .external_lex_state = 135},
  [3306] = {.lex_state = 0, .external_lex_state = 153},
  [3307] = {.lex_state = 2, .external_lex_state = 135},
  [3308] = {.lex_state = 2, .external_lex_state = 132},
  [3309] = {.lex_state = 2, .external_lex_state = 154},
  [3310] = {.lex_state = 2, .external_lex_state = 135},
  [3311] = {.lex_state = 2, .external_lex_state = 135},
  [3312] = {.lex_state = 2, .external_lex_state = 135},
  [3313] = {.lex_state = 2, .external_lex_state = 135},
  [3314] = {.lex_state = 2, .external_lex_state = 135},
  [3315] = {.lex_state = 2, .external_lex_state = 135},
  [3316] = {.lex_state = 0, .external_lex_state = 62},
  [3317] = {.lex_state = 2, .external_lex_state = 152},
  [3318] = {.lex_state = 0, .external_lex_state = 62},
  [3319] = {.lex_state = 2, .external_lex_state = 135},
  [3320] = {.lex_state = 2, .external_lex_state = 135},
  [3321] = {.lex_state = 0, .external_lex_state = 66},
  [3322] = {.lex_state = 2, .external_lex_state = 135},
  [3323] = {.lex_state = 0, .external_lex_state = 62},
  [3324] = {.lex_state = 0, .external_lex_state = 62},
  [3325] = {.lex_state = 0, .external_lex_state = 62},
  [3326] = {.lex_state = 0, .external_lex_state = 66},
  [3327] = {.lex_state = 0, .external_lex_state = 62},
  [3328] = {.lex_state = 0, .external_lex_state = 62},
  [3329] = {.lex_state = 0, .external_lex_state = 66},
  [3330] = {.lex_state = 0, .external_lex_state = 62},
  [3331] = {.lex_state = 0, .external_lex_state = 62},
  [3332] = {.lex_state = 0, .external_lex_state = 62},
  [3333] = {.lex_state = 0, .external_lex_state = 62},
  [3334] = {.lex_state = 0, .external_lex_state = 62},
  [3335] = {.lex_state = 0, .external_lex_state = 62},
  [3336] = {.lex_state = 0, .external_lex_state = 62},
  [3337] = {.lex_state = 0, .external_lex_state = 62},
  [3338] = {.lex_state = 0, .external_lex_state = 62},
  [3339] = {.lex_state = 0, .external_lex_state = 62},
  [3340] = {.lex_state = 0, .external_lex_state = 66},
  [3341] = {.lex_state = 0, .external_lex_state = 62},
  [3342] = {.lex_state = 0, .external_lex_state = 62},
  [3343] = {.lex_state = 0, .external_lex_state = 62},
  [3344] = {.lex_state = 0, .external_lex_state = 62},
  [3345] = {.lex_state = 0, .external_lex_state = 62},
  [3346] = {.lex_state = 0, .external_lex_state = 62},
  [3347] = {.lex_state = 0, .external_lex_state = 62},
  [3348] = {.lex_state = 0, .external_lex_state = 62},
  [3349] = {.lex_state = 0, .external_lex_state = 66},
  [3350] = {.lex_state = 0, .external_lex_state = 62},
  [3351] = {.lex_state = 0, .external_lex_state = 66},
  [3352] = {.lex_state = 0, .external_lex_state = 149},
  [3353] = {.lex_state = 2, .external_lex_state = 136},
  [3354] = {.lex_state = 2, .external_lex_state = 136},
  [3355] = {.lex_state = 2, .external_lex_state = 136},
  [3356] = {.lex_state = 2, .external_lex_state = 136},
  [3357] = {.lex_state = 2, .external_lex_state = 136},
  [3358] = {.lex_state = 2, .external_lex_state = 136},
  [3359] = {.lex_state = 2, .external_lex_state = 136},
  [3360] = {.lex_state = 2, .external_lex_state = 136},
  [3361] = {.lex_state = 2, .external_lex_state = 136},
  [3362] = {.lex_state = 0, .external_lex_state = 155},
  [3363] = {.lex_state = 2, .external_lex_state = 154},
  [3364] = {.lex_state = 2, .external_lex_state = 138},
  [3365] = {.lex_state = 0, .external_lex_state = 155},
  [3366] = {.lex_state = 3, .external_lex_state = 152},
  [3367] = {.lex_state = 2, .external_lex_state = 138},
  [3368] = {.lex_state = 2, .external_lex_state = 138},
  [3369] = {.lex_state = 2, .external_lex_state = 138},
  [3370] = {.lex_state = 0, .external_lex_state = 155},
  [3371] = {.lex_state = 2, .external_lex_state = 138},
  [3372] = {.lex_state = 2, .external_lex_state = 138},
  [3373] = {.lex_state = 0, .external_lex_state = 155},
  [3374] = {.lex_state = 2, .external_lex_state = 138},
  [3375] = {.lex_state = 2, .external_lex_state = 138},
  [3376] = {.lex_state = 0, .external_lex_state = 155},
  [3377] = {.lex_state = 2, .external_lex_state = 138},
  [3378] = {.lex_state = 2, .external_lex_state = 138},
  [3379] = {.lex_state = 2, .external_lex_state = 138},
  [3380] = {.lex_state = 2, .external_lex_state = 138},
  [3381] = {.lex_state = 0, .external_lex_state = 156},
  [3382] = {.lex_state = 2, .external_lex_state = 138},
  [3383] = {.lex_state = 2, .external_lex_state = 138},
  [3384] = {.lex_state = 2, .external_lex_state = 138},
  [3385] = {.lex_state = 2, .external_lex_state = 138},
  [3386] = {.lex_state = 0, .external_lex_state = 157},
  [3387] = {.lex_state = 2, .external_lex_state = 138},
  [3388] = {.lex_state = 2, .external_lex_state = 136},
  [3389] = {.lex_state = 0, .external_lex_state = 157},
  [3390] = {.lex_state = 2, .external_lex_state = 138},
  [3391] = {.lex_state = 2, .external_lex_state = 136},
  [3392] = {.lex_state = 2, .external_lex_state = 138},
  [3393] = {.lex_state = 2, .external_lex_state = 154},
  [3394] = {.lex_state = 0, .external_lex_state = 158},
  [3395] = {.lex_state = 2, .external_lex_state = 136},
  [3396] = {.lex_state = 2, .external_lex_state = 136},
  [3397] = {.lex_state = 0, .external_lex_state = 158},
  [3398] = {.lex_state = 2, .external_lex_state = 138},
  [3399] = {.lex_state = 2, .external_lex_state = 138},
  [3400] = {.lex_state = 2, .external_lex_state = 138},
  [3401] = {.lex_state = 2, .external_lex_state = 138},
  [3402] = {.lex_state = 2, .external_lex_state = 138},
  [3403] = {.lex_state = 2, .external_lex_state = 138},
  [3404] = {.lex_state = 2, .external_lex_state = 136},
  [3405] = {.lex_state = 2, .external_lex_state = 159},
  [3406] = {.lex_state = 2, .external_lex_state = 136},
  [3407] = {.lex_state = 2, .external_lex_state = 136},
  [3408] = {.lex_state = 2, .external_lex_state = 136},
  [3409] = {.lex_state = 2, .external_lex_state = 136},
  [3410] = {.lex_state = 2, .external_lex_state = 136},
  [3411] = {.lex_state = 2, .external_lex_state = 138},
  [3412] = {.lex_state = 2, .external_lex_state = 138},
  [3413] = {.lex_state = 2, .external_lex_state = 138},
  [3414] = {.lex_state = 2, .external_lex_state = 136},
  [3415] = {.lex_state = 2, .external_lex_state = 139},
  [3416] = {.lex_state = 0, .external_lex_state = 156},
  [3417] = {.lex_state = 2, .external_lex_state = 136},
  [3418] = {.lex_state = 2, .external_lex_state = 136},
  [3419] = {.lex_state = 2, .external_lex_state = 136},
  [3420] = {.lex_state = 0, .external_lex_state = 158},
  [3421] = {.lex_state = 2, .external_lex_state = 136},
  [3422] = {.lex_state = 2, .external_lex_state = 136},
  [3423] = {.lex_state = 2, .external_lex_state = 136},
  [3424] = {.lex_state = 2, .external_lex_state = 136},
  [3425] = {.lex_state = 2, .external_lex_state = 138},
  [3426] = {.lex_state = 2, .external_lex_state = 136},
  [3427] = {.lex_state = 2, .external_lex_state = 136},
  [3428] = {.lex_state = 0, .external_lex_state = 156},
  [3429] = {.lex_state = 0, .external_lex_state = 155},
  [3430] = {.lex_state = 2, .external_lex_state = 154},
  [3431] = {.lex_state = 2, .external_lex_state = 154},
  [3432] = {.lex_state = 0, .external_lex_state = 157},
  [3433] = {.lex_state = 0, .external_lex_state = 160},
  [3434] = {.lex_state = 2, .external_lex_state = 152},
  [3435] = {.lex_state = 2, .external_lex_state = 152},
  [3436] = {.lex_state = 2, .external_lex_state = 152},
  [3437] = {.lex_state = 0, .external_lex_state = 160},
  [3438] = {.lex_state = 2, .external_lex_state = 152},
  [3439] = {.lex_state = 2, .external_lex_state = 152},
  [3440] = {.lex_state = 2, .external_lex_state = 152},
  [3441] = {.lex_state = 2, .external_lex_state = 152},
  [3442] = {.lex_state = 2, .external_lex_state = 152},
  [3443] = {.lex_state = 2, .external_lex_state = 152},
  [3444] = {.lex_state = 0, .external_lex_state = 161},
  [3445] = {.lex_state = 2, .external_lex_state = 152},
  [3446] = {.lex_state = 2, .external_lex_state = 152},
  [3447] = {.lex_state = 2, .external_lex_state = 152},
  [3448] = {.lex_state = 2, .external_lex_state = 152},
  [3449] = {.lex_state = 2, .external_lex_state = 152},
  [3450] = {.lex_state = 2, .external_lex_state = 139},
  [3451] = {.lex_state = 2, .external_lex_state = 152},
  [3452] = {.lex_state = 0, .external_lex_state = 160},
  [3453] = {.lex_state = 2, .external_lex_state = 152},
  [3454] = {.lex_state = 0, .external_lex_state = 161},
  [3455] = {.lex_state = 2, .external_lex_state = 152},
  [3456] = {.lex_state = 2, .external_lex_state = 162},
  [3457] = {.lex_state = 0, .external_lex_state = 160},
  [3458] = {.lex_state = 0, .external_lex_state = 160},
  [3459] = {.lex_state = 2, .external_lex_state = 152},
  [3460] = {.lex_state = 2, .external_lex_state = 152},
  [3461] = {.lex_state = 2, .external_lex_state = 152},
  [3462] = {.lex_state = 2, .external_lex_state = 152},
  [3463] = {.lex_state = 2, .external_lex_state = 152},
  [3464] = {.lex_state = 2, .external_lex_state = 152},
  [3465] = {.lex_state = 2, .external_lex_state = 152},
  [3466] = {.lex_state = 0, .external_lex_state = 160},
  [3467] = {.lex_state = 2, .external_lex_state = 152},
  [3468] = {.lex_state = 2, .external_lex_state = 152},
  [3469] = {.lex_state = 2, .external_lex_state = 152},
  [3470] = {.lex_state = 2, .external_lex_state = 152},
  [3471] = {.lex_state = 2, .external_lex_state = 152},
  [3472] = {.lex_state = 2, .external_lex_state = 152},
  [3473] = {.lex_state = 0, .external_lex_state = 163},
  [3474] = {.lex_state = 2, .external_lex_state = 152},
  [3475] = {.lex_state = 2, .external_lex_state = 152},
  [3476] = {.lex_state = 2, .external_lex_state = 152},
  [3477] = {.lex_state = 2, .external_lex_state = 152},
  [3478] = {.lex_state = 0, .external_lex_state = 160},
  [3479] = {.lex_state = 2, .external_lex_state = 152},
  [3480] = {.lex_state = 2, .external_lex_state = 152},
  [3481] = {.lex_state = 0, .external_lex_state = 164},
  [3482] = {.lex_state = 0, .external_lex_state = 160},
  [3483] = {.lex_state = 2, .external_lex_state = 152},
  [3484] = {.lex_state = 2, .external_lex_state = 152},
  [3485] = {.lex_state = 2, .external_lex_state = 152},
  [3486] = {.lex_state = 0, .external_lex_state = 164},
  [3487] = {.lex_state = 2, .external_lex_state = 139},
  [3488] = {.lex_state = 0, .external_lex_state = 163},
  [3489] = {.lex_state = 0, .external_lex_state = 160},
  [3490] = {.lex_state = 0, .external_lex_state = 161},
  [3491] = {.lex_state = 0, .external_lex_state = 160},
  [3492] = {.lex_state = 0, .external_lex_state = 160},
  [3493] = {.lex_state = 0, .external_lex_state = 165},
  [3494] = {.lex_state = 0, .external_lex_state = 160},
  [3495] = {.lex_state = 0, .external_lex_state = 160},
  [3496] = {.lex_state = 0, .external_lex_state = 160},
  [3497] = {.lex_state = 0, .external_lex_state = 160},
  [3498] = {.lex_state = 0, .external_lex_state = 160},
  [3499] = {.lex_state = 0, .external_lex_state = 160},
  [3500] = {.lex_state = 2, .external_lex_state = 152},
  [3501] = {.lex_state = 0, .external_lex_state = 160},
  [3502] = {.lex_state = 0, .external_lex_state = 160},
  [3503] = {.lex_state = 0, .external_lex_state = 160},
  [3504] = {.lex_state = 0, .external_lex_state = 160},
  [3505] = {.lex_state = 0, .external_lex_state = 163},
  [3506] = {.lex_state = 0, .external_lex_state = 160},
  [3507] = {.lex_state = 0, .external_lex_state = 160},
  [3508] = {.lex_state = 2, .external_lex_state = 162},
  [3509] = {.lex_state = 0, .external_lex_state = 160},
  [3510] = {.lex_state = 0, .external_lex_state = 160},
  [3511] = {.lex_state = 0, .external_lex_state = 165},
  [3512] = {.lex_state = 0, .external_lex_state = 160},
  [3513] = {.lex_state = 0, .external_lex_state = 160},
  [3514] = {.lex_state = 2, .external_lex_state = 148},
  [3515] = {.lex_state = 2, .external_lex_state = 152},
  [3516] = {.lex_state = 0, .external_lex_state = 160},
  [3517] = {.lex_state = 0, .external_lex_state = 160},
  [3518] = {.lex_state = 2, .external_lex_state = 139},
  [3519] = {.lex_state = 0, .external_lex_state = 160},
  [3520] = {.lex_state = 2, .external_lex_state = 139},
  [3521] = {.lex_state = 0, .external_lex_state = 160},
  [3522] = {.lex_state = 2, .external_lex_state = 139},
  [3523] = {.lex_state = 2, .external_lex_state = 139},
  [3524] = {.lex_state = 0, .external_lex_state = 160},
  [3525] = {.lex_state = 0, .external_lex_state = 160},
  [3526] = {.lex_state = 2, .external_lex_state = 139},
  [3527] = {.lex_state = 2, .external_lex_state = 139},
  [3528] = {.lex_state = 2, .external_lex_state = 139},
  [3529] = {.lex_state = 2, .external_lex_state = 139},
  [3530] = {.lex_state = 2, .external_lex_state = 139},
  [3531] = {.lex_state = 2, .external_lex_state = 139},
  [3532] = {.lex_state = 2, .external_lex_state = 139},
  [3533] = {.lex_state = 0, .external_lex_state = 160},
  [3534] = {.lex_state = 2, .external_lex_state = 139},
  [3535] = {.lex_state = 2, .external_lex_state = 139},
  [3536] = {.lex_state = 2, .external_lex_state = 139},
  [3537] = {.lex_state = 2, .external_lex_state = 139},
  [3538] = {.lex_state = 2, .external_lex_state = 139},
  [3539] = {.lex_state = 2, .external_lex_state = 139},
  [3540] = {.lex_state = 0, .external_lex_state = 160},
  [3541] = {.lex_state = 2, .external_lex_state = 152},
  [3542] = {.lex_state = 2, .external_lex_state = 139},
  [3543] = {.lex_state = 2, .external_lex_state = 166},
  [3544] = {.lex_state = 2, .external_lex_state = 139},
  [3545] = {.lex_state = 0, .external_lex_state = 160},
  [3546] = {.lex_state = 2, .external_lex_state = 139},
  [3547] = {.lex_state = 2, .external_lex_state = 139},
  [3548] = {.lex_state = 2, .external_lex_state = 139},
  [3549] = {.lex_state = 2, .external_lex_state = 139},
  [3550] = {.lex_state = 2, .external_lex_state = 139},
  [3551] = {.lex_state = 0, .external_lex_state = 160},
  [3552] = {.lex_state = 0, .external_lex_state = 160},
  [3553] = {.lex_state = 0, .external_lex_state = 160},
  [3554] = {.lex_state = 0, .external_lex_state = 160},
  [3555] = {.lex_state = 0, .external_lex_state = 160},
  [3556] = {.lex_state = 0, .external_lex_state = 160},
  [3557] = {.lex_state = 0, .external_lex_state = 160},
  [3558] = {.lex_state = 0, .external_lex_state = 160},
  [3559] = {.lex_state = 0, .external_lex_state = 160},
  [3560] = {.lex_state = 2, .external_lex_state = 152},
  [3561] = {.lex_state = 0, .external_lex_state = 160},
  [3562] = {.lex_state = 2, .external_lex_state = 152},
  [3563] = {.lex_state = 2, .external_lex_state = 139},
  [3564] = {.lex_state = 0, .external_lex_state = 160},
  [3565] = {.lex_state = 0, .external_lex_state = 160},
  [3566] = {.lex_state = 2, .external_lex_state = 139},
  [3567] = {.lex_state = 2, .external_lex_state = 139},
  [3568] = {.lex_state = 0, .external_lex_state = 160},
  [3569] = {.lex_state = 0, .external_lex_state = 160},
  [3570] = {.lex_state = 0, .external_lex_state = 165},
  [3571] = {.lex_state = 2, .external_lex_state = 148},
  [3572] = {.lex_state = 2, .external_lex_state = 152},
  [3573] = {.lex_state = 2, .external_lex_state = 152},
  [3574] = {.lex_state = 2, .external_lex_state = 152},
  [3575] = {.lex_state = 2, .external_lex_state = 152},
  [3576] = {.lex_state = 0, .external_lex_state = 164},
  [3577] = {.lex_state = 2, .external_lex_state = 152},
  [3578] = {.lex_state = 2, .external_lex_state = 152},
  [3579] = {.lex_state = 2, .external_lex_state = 152},
  [3580] = {.lex_state = 2, .external_lex_state = 152},
  [3581] = {.lex_state = 2, .external_lex_state = 152},
  [3582] = {.lex_state = 2, .external_lex_state = 152},
  [3583] = {.lex_state = 2, .external_lex_state = 152},
  [3584] = {.lex_state = 2, .external_lex_state = 152},
  [3585] = {.lex_state = 0, .external_lex_state = 160},
  [3586] = {.lex_state = 2, .external_lex_state = 167},
  [3587] = {.lex_state = 0, .external_lex_state = 168},
  [3588] = {.lex_state = 2, .external_lex_state = 152},
  [3589] = {.lex_state = 2, .external_lex_state = 167},
  [3590] = {.lex_state = 2, .external_lex_state = 167},
  [3591] = {.lex_state = 2, .external_lex_state = 167},
  [3592] = {.lex_state = 2, .external_lex_state = 167},
  [3593] = {.lex_state = 0, .external_lex_state = 169},
  [3594] = {.lex_state = 2, .external_lex_state = 167},
  [3595] = {.lex_state = 0, .external_lex_state = 170},
  [3596] = {.lex_state = 2, .external_lex_state = 167},
  [3597] = {.lex_state = 2, .external_lex_state = 167},
  [3598] = {.lex_state = 2, .external_lex_state = 167},
  [3599] = {.lex_state = 2, .external_lex_state = 167},
  [3600] = {.lex_state = 2, .external_lex_state = 167},
  [3601] = {.lex_state = 2, .external_lex_state = 167},
  [3602] = {.lex_state = 2, .external_lex_state = 167},
  [3603] = {.lex_state = 0, .external_lex_state = 170},
  [3604] = {.lex_state = 2, .external_lex_state = 167},
  [3605] = {.lex_state = 2, .external_lex_state = 167},
  [3606] = {.lex_state = 2, .external_lex_state = 167},
  [3607] = {.lex_state = 2, .external_lex_state = 167},
  [3608] = {.lex_state = 2, .external_lex_state = 167},
  [3609] = {.lex_state = 0, .external_lex_state = 171},
  [3610] = {.lex_state = 0, .external_lex_state = 78},
  [3611] = {.lex_state = 2, .external_lex_state = 167},
  [3612] = {.lex_state = 2, .external_lex_state = 167},
  [3613] = {.lex_state = 2, .external_lex_state = 167},
  [3614] = {.lex_state = 2, .external_lex_state = 167},
  [3615] = {.lex_state = 0, .external_lex_state = 170},
  [3616] = {.lex_state = 2, .external_lex_state = 167},
  [3617] = {.lex_state = 0, .external_lex_state = 160},
  [3618] = {.lex_state = 2, .external_lex_state = 167},
  [3619] = {.lex_state = 2, .external_lex_state = 167},
  [3620] = {.lex_state = 2, .external_lex_state = 167},
  [3621] = {.lex_state = 0, .external_lex_state = 168},
  [3622] = {.lex_state = 2, .external_lex_state = 167},
  [3623] = {.lex_state = 0, .external_lex_state = 172},
  [3624] = {.lex_state = 2, .external_lex_state = 167},
  [3625] = {.lex_state = 0, .external_lex_state = 169},
  [3626] = {.lex_state = 2, .external_lex_state = 167},
  [3627] = {.lex_state = 0, .external_lex_state = 169},
  [3628] = {.lex_state = 2, .external_lex_state = 167},
  [3629] = {.lex_state = 2, .external_lex_state = 167},
  [3630] = {.lex_state = 2, .external_lex_state = 167},
  [3631] = {.lex_state = 2, .external_lex_state = 167},
  [3632] = {.lex_state = 2, .external_lex_state = 167},
  [3633] = {.lex_state = 2, .external_lex_state = 167},
  [3634] = {.lex_state = 2, .external_lex_state = 167},
  [3635] = {.lex_state = 2, .external_lex_state = 167},
  [3636] = {.lex_state = 0, .external_lex_state = 168},
  [3637] = {.lex_state = 2, .external_lex_state = 173},
  [3638] = {.lex_state = 2, .external_lex_state = 167},
  [3639] = {.lex_state = 0, .external_lex_state = 174},
  [3640] = {.lex_state = 0, .external_lex_state = 174},
  [3641] = {.lex_state = 0, .external_lex_state = 172},
  [3642] = {.lex_state = 0, .external_lex_state = 153},
  [3643] = {.lex_state = 2, .external_lex_state = 167},
  [3644] = {.lex_state = 2, .external_lex_state = 167},
  [3645] = {.lex_state = 2, .external_lex_state = 167},
  [3646] = {.lex_state = 0, .external_lex_state = 153},
  [3647] = {.lex_state = 2, .external_lex_state = 141},
  [3648] = {.lex_state = 2, .external_lex_state = 167},
  [3649] = {.lex_state = 2, .external_lex_state = 167},
  [3650] = {.lex_state = 2, .external_lex_state = 167},
  [3651] = {.lex_state = 0, .external_lex_state = 160},
  [3652] = {.lex_state = 2, .external_lex_state = 167},
  [3653] = {.lex_state = 2, .external_lex_state = 167},
  [3654] = {.lex_state = 2, .external_lex_state = 167},
  [3655] = {.lex_state = 0, .external_lex_state = 160},
  [3656] = {.lex_state = 2, .external_lex_state = 167},
  [3657] = {.lex_state = 2, .external_lex_state = 167},
  [3658] = {.lex_state = 2, .external_lex_state = 167},
  [3659] = {.lex_state = 0, .external_lex_state = 160},
  [3660] = {.lex_state = 0, .external_lex_state = 171},
  [3661] = {.lex_state = 2, .external_lex_state = 167},
  [3662] = {.lex_state = 0, .external_lex_state = 172},
  [3663] = {.lex_state = 0, .external_lex_state = 160},
  [3664] = {.lex_state = 2, .external_lex_state = 167},
  [3665] = {.lex_state = 0, .external_lex_state = 160},
  [3666] = {.lex_state = 2, .external_lex_state = 167},
  [3667] = {.lex_state = 0, .external_lex_state = 171},
  [3668] = {.lex_state = 0, .external_lex_state = 174},
  [3669] = {.lex_state = 0, .external_lex_state = 123},
  [3670] = {.lex_state = 2, .external_lex_state = 141},
  [3671] = {.lex_state = 2, .external_lex_state = 67},
  [3672] = {.lex_state = 2, .external_lex_state = 67},
  [3673] = {.lex_state = 2, .external_lex_state = 67},
  [3674] = {.lex_state = 2, .external_lex_state = 141},
  [3675] = {.lex_state = 2, .external_lex_state = 67},
  [3676] = {.lex_state = 2, .external_lex_state = 141},
  [3677] = {.lex_state = 2, .external_lex_state = 141},
  [3678] = {.lex_state = 2, .external_lex_state = 141},
  [3679] = {.lex_state = 2, .external_lex_state = 67},
  [3680] = {.lex_state = 2, .external_lex_state = 67},
  [3681] = {.lex_state = 2, .external_lex_state = 141},
  [3682] = {.lex_state = 2, .external_lex_state = 67},
  [3683] = {.lex_state = 2, .external_lex_state = 67},
  [3684] = {.lex_state = 2, .external_lex_state = 141},
  [3685] = {.lex_state = 2, .external_lex_state = 67},
  [3686] = {.lex_state = 0, .external_lex_state = 128},
  [3687] = {.lex_state = 0, .external_lex_state = 123},
  [3688] = {.lex_state = 2, .external_lex_state = 67},
  [3689] = {.lex_state = 2, .external_lex_state = 67},
  [3690] = {.lex_state = 2, .external_lex_state = 67},
  [3691] = {.lex_state = 2, .external_lex_state = 67},
  [3692] = {.lex_state = 0, .external_lex_state = 123},
  [3693] = {.lex_state = 0, .external_lex_state = 128},
  [3694] = {.lex_state = 2, .external_lex_state = 141},
  [3695] = {.lex_state = 0, .external_lex_state = 123},
  [3696] = {.lex_state = 2, .external_lex_state = 67},
  [3697] = {.lex_state = 2, .external_lex_state = 67},
  [3698] = {.lex_state = 0, .external_lex_state = 123},
  [3699] = {.lex_state = 2, .external_lex_state = 67},
  [3700] = {.lex_state = 0, .external_lex_state = 128},
  [3701] = {.lex_state = 2, .external_lex_state = 141},
  [3702] = {.lex_state = 2, .external_lex_state = 67},
  [3703] = {.lex_state = 2, .external_lex_state = 141},
  [3704] = {.lex_state = 0, .external_lex_state = 160},
  [3705] = {.lex_state = 0, .external_lex_state = 123},
  [3706] = {.lex_state = 0, .external_lex_state = 128},
  [3707] = {.lex_state = 0, .external_lex_state = 123},
  [3708] = {.lex_state = 2, .external_lex_state = 67},
  [3709] = {.lex_state = 2, .external_lex_state = 67},
  [3710] = {.lex_state = 0, .external_lex_state = 128},
  [3711] = {.lex_state = 2, .external_lex_state = 67},
  [3712] = {.lex_state = 0, .external_lex_state = 175},
  [3713] = {.lex_state = 0, .external_lex_state = 123},
  [3714] = {.lex_state = 0, .external_lex_state = 123},
  [3715] = {.lex_state = 0, .external_lex_state = 123},
  [3716] = {.lex_state = 0, .external_lex_state = 123},
  [3717] = {.lex_state = 0, .external_lex_state = 123},
  [3718] = {.lex_state = 0, .external_lex_state = 123},
  [3719] = {.lex_state = 0, .external_lex_state = 123},
  [3720] = {.lex_state = 0, .external_lex_state = 123},
  [3721] = {.lex_state = 0, .external_lex_state = 123},
  [3722] = {.lex_state = 0, .external_lex_state = 123},
  [3723] = {.lex_state = 0, .external_lex_state = 123},
  [3724] = {.lex_state = 0, .external_lex_state = 123},
  [3725] = {.lex_state = 2, .external_lex_state = 67},
  [3726] = {.lex_state = 0, .external_lex_state = 123},
  [3727] = {.lex_state = 2, .external_lex_state = 141},
  [3728] = {.lex_state = 2, .external_lex_state = 141},
  [3729] = {.lex_state = 0, .external_lex_state = 123},
  [3730] = {.lex_state = 0, .external_lex_state = 123},
  [3731] = {.lex_state = 0, .external_lex_state = 123},
  [3732] = {.lex_state = 0, .external_lex_state = 123},
  [3733] = {.lex_state = 2, .external_lex_state = 67},
  [3734] = {.lex_state = 0, .external_lex_state = 123},
  [3735] = {.lex_state = 0, .external_lex_state = 123},
  [3736] = {.lex_state = 0, .external_lex_state = 123},
  [3737] = {.lex_state = 2, .external_lex_state = 67},
  [3738] = {.lex_state = 0, .external_lex_state = 123},
  [3739] = {.lex_state = 0, .external_lex_state = 123},
  [3740] = {.lex_state = 0, .external_lex_state = 123},
  [3741] = {.lex_state = 0, .external_lex_state = 123},
  [3742] = {.lex_state = 0, .external_lex_state = 175},
  [3743] = {.lex_state = 2, .external_lex_state = 67},
  [3744] = {.lex_state = 0, .external_lex_state = 175},
  [3745] = {.lex_state = 2, .external_lex_state = 141},
  [3746] = {.lex_state = 2, .external_lex_state = 67},
  [3747] = {.lex_state = 2, .external_lex_state = 67},
  [3748] = {.lex_state = 2, .external_lex_state = 67},
  [3749] = {.lex_state = 0, .external_lex_state = 123},
  [3750] = {.lex_state = 0, .external_lex_state = 128},
  [3751] = {.lex_state = 0, .external_lex_state = 128},
  [3752] = {.lex_state = 0, .external_lex_state = 128},
  [3753] = {.lex_state = 0, .external_lex_state = 128},
  [3754] = {.lex_state = 0, .external_lex_state = 128},
  [3755] = {.lex_state = 2, .external_lex_state = 141},
  [3756] = {.lex_state = 2, .external_lex_state = 152},
  [3757] = {.lex_state = 2, .external_lex_state = 67},
  [3758] = {.lex_state = 2, .external_lex_state = 67},
  [3759] = {.lex_state = 0, .external_lex_state = 128},
  [3760] = {.lex_state = 0, .external_lex_state = 128},
  [3761] = {.lex_state = 0, .external_lex_state = 128},
  [3762] = {.lex_state = 0, .external_lex_state = 128},
  [3763] = {.lex_state = 0, .external_lex_state = 128},
  [3764] = {.lex_state = 0, .external_lex_state = 128},
  [3765] = {.lex_state = 0, .external_lex_state = 128},
  [3766] = {.lex_state = 0, .external_lex_state = 176},
  [3767] = {.lex_state = 0, .external_lex_state = 128},
  [3768] = {.lex_state = 0, .external_lex_state = 128},
  [3769] = {.lex_state = 0, .external_lex_state = 128},
  [3770] = {.lex_state = 2, .external_lex_state = 152},
  [3771] = {.lex_state = 2, .external_lex_state = 67},
  [3772] = {.lex_state = 2, .external_lex_state = 152},
  [3773] = {.lex_state = 2, .external_lex_state = 141},
  [3774] = {.lex_state = 2, .external_lex_state = 141},
  [3775] = {.lex_state = 2, .external_lex_state = 141},
  [3776] = {.lex_state = 2, .external_lex_state = 152},
  [3777] = {.lex_state = 2, .external_lex_state = 67},
  [3778] = {.lex_state = 2, .external_lex_state = 67},
  [3779] = {.lex_state = 2, .external_lex_state = 152},
  [3780] = {.lex_state = 0, .external_lex_state = 128},
  [3781] = {.lex_state = 2, .external_lex_state = 141},
  [3782] = {.lex_state = 2, .external_lex_state = 141},
  [3783] = {.lex_state = 0, .external_lex_state = 128},
  [3784] = {.lex_state = 0, .external_lex_state = 128},
  [3785] = {.lex_state = 0, .external_lex_state = 123},
  [3786] = {.lex_state = 0, .external_lex_state = 123},
  [3787] = {.lex_state = 0, .external_lex_state = 128},
  [3788] = {.lex_state = 0, .external_lex_state = 128},
  [3789] = {.lex_state = 2, .external_lex_state = 177},
  [3790] = {.lex_state = 2, .external_lex_state = 67},
  [3791] = {.lex_state = 0, .external_lex_state = 128},
  [3792] = {.lex_state = 0, .external_lex_state = 128},
  [3793] = {.lex_state = 0, .external_lex_state = 128},
  [3794] = {.lex_state = 0, .external_lex_state = 128},
  [3795] = {.lex_state = 2, .external_lex_state = 67},
  [3796] = {.lex_state = 2, .external_lex_state = 67},
  [3797] = {.lex_state = 0, .external_lex_state = 128},
  [3798] = {.lex_state = 2, .external_lex_state = 141},
  [3799] = {.lex_state = 0, .external_lex_state = 128},
  [3800] = {.lex_state = 2, .external_lex_state = 141},
  [3801] = {.lex_state = 0, .external_lex_state = 168},
  [3802] = {.lex_state = 2, .external_lex_state = 141},
  [3803] = {.lex_state = 0, .external_lex_state = 176},
  [3804] = {.lex_state = 0, .external_lex_state = 128},
  [3805] = {.lex_state = 0, .external_lex_state = 128},
  [3806] = {.lex_state = 2, .external_lex_state = 141},
  [3807] = {.lex_state = 0, .external_lex_state = 176},
  [3808] = {.lex_state = 2, .external_lex_state = 141},
  [3809] = {.lex_state = 2, .external_lex_state = 141},
  [3810] = {.lex_state = 0, .external_lex_state = 168},
  [3811] = {.lex_state = 2, .external_lex_state = 152},
  [3812] = {.lex_state = 0, .external_lex_state = 128},
  [3813] = {.lex_state = 2, .external_lex_state = 152},
  [3814] = {.lex_state = 2, .external_lex_state = 67},
  [3815] = {.lex_state = 2, .external_lex_state = 141},
  [3816] = {.lex_state = 2, .external_lex_state = 141},
  [3817] = {.lex_state = 2, .external_lex_state = 141},
  [3818] = {.lex_state = 2, .external_lex_state = 152},
  [3819] = {.lex_state = 2, .external_lex_state = 178},
  [3820] = {.lex_state = 2, .external_lex_state = 152},
  [3821] = {.lex_state = 2, .external_lex_state = 152},
  [3822] = {.lex_state = 2, .external_lex_state = 152},
  [3823] = {.lex_state = 0, .external_lex_state = 175},
  [3824] = {.lex_state = 0, .external_lex_state = 175},
  [3825] = {.lex_state = 0, .external_lex_state = 175},
  [3826] = {.lex_state = 2, .external_lex_state = 67},
  [3827] = {.lex_state = 0, .external_lex_state = 175},
  [3828] = {.lex_state = 0, .external_lex_state = 175},
  [3829] = {.lex_state = 0, .external_lex_state = 175},
  [3830] = {.lex_state = 0, .external_lex_state = 175},
  [3831] = {.lex_state = 0, .external_lex_state = 175},
  [3832] = {.lex_state = 0, .external_lex_state = 175},
  [3833] = {.lex_state = 0, .external_lex_state = 175},
  [3834] = {.lex_state = 0, .external_lex_state = 175},
  [3835] = {.lex_state = 0, .external_lex_state = 175},
  [3836] = {.lex_state = 0, .external_lex_state = 175},
  [3837] = {.lex_state = 0, .external_lex_state = 175},
  [3838] = {.lex_state = 0, .external_lex_state = 175},
  [3839] = {.lex_state = 2, .external_lex_state = 67},
  [3840] = {.lex_state = 0, .external_lex_state = 175},
  [3841] = {.lex_state = 0, .external_lex_state = 175},
  [3842] = {.lex_state = 0, .external_lex_state = 175},
  [3843] = {.lex_state = 2, .external_lex_state = 67},
  [3844] = {.lex_state = 0, .external_lex_state = 175},
  [3845] = {.lex_state = 0, .external_lex_state = 175},
  [3846] = {.lex_state = 0, .external_lex_state = 175},
  [3847] = {.lex_state = 0, .external_lex_state = 175},
  [3848] = {.lex_state = 0, .external_lex_state = 175},
  [3849] = {.lex_state = 0, .external_lex_state = 175},
  [3850] = {.lex_state = 0, .external_lex_state = 175},
  [3851] = {.lex_state = 0, .external_lex_state = 175},
  [3852] = {.lex_state = 0, .external_lex_state = 175},
  [3853] = {.lex_state = 0, .external_lex_state = 175},
  [3854] = {.lex_state = 0, .external_lex_state = 175},
  [3855] = {.lex_state = 0, .external_lex_state = 175},
  [3856] = {.lex_state = 0, .external_lex_state = 175},
  [3857] = {.lex_state = 0, .external_lex_state = 175},
  [3858] = {.lex_state = 0, .external_lex_state = 175},
  [3859] = {.lex_state = 0, .external_lex_state = 175},
  [3860] = {.lex_state = 0, .external_lex_state = 175},
  [3861] = {.lex_state = 0, .external_lex_state = 175},
  [3862] = {.lex_state = 0, .external_lex_state = 175},
  [3863] = {.lex_state = 0, .external_lex_state = 175},
  [3864] = {.lex_state = 0, .external_lex_state = 175},
  [3865] = {.lex_state = 0, .external_lex_state = 175},
  [3866] = {.lex_state = 0, .external_lex_state = 175},
  [3867] = {.lex_state = 0, .external_lex_state = 175},
  [3868] = {.lex_state = 0, .external_lex_state = 175},
  [3869] = {.lex_state = 0, .external_lex_state = 175},
  [3870] = {.lex_state = 0, .external_lex_state = 175},
  [3871] = {.lex_state = 0, .external_lex_state = 175},
  [3872] = {.lex_state = 0, .external_lex_state = 175},
  [3873] = {.lex_state = 0, .external_lex_state = 175},
  [3874] = {.lex_state = 0, .external_lex_state = 175},
  [3875] = {.lex_state = 0, .external_lex_state = 175},
  [3876] = {.lex_state = 0, .external_lex_state = 175},
  [3877] = {.lex_state = 0, .external_lex_state = 175},
  [3878] = {.lex_state = 0, .external_lex_state = 175},
  [3879] = {.lex_state = 0, .external_lex_state = 175},
  [3880] = {.lex_state = 0, .external_lex_state = 175},
  [3881] = {.lex_state = 0, .external_lex_state = 175},
  [3882] = {.lex_state = 0, .external_lex_state = 175},
  [3883] = {.lex_state = 0, .external_lex_state = 175},
  [3884] = {.lex_state = 0, .external_lex_state = 175},
  [3885] = {.lex_state = 0, .external_lex_state = 175},
  [3886] = {.lex_state = 0, .external_lex_state = 175},
  [3887] = {.lex_state = 0, .external_lex_state = 175},
  [3888] = {.lex_state = 0, .external_lex_state = 175},
  [3889] = {.lex_state = 0, .external_lex_state = 175},
  [3890] = {.lex_state = 0, .external_lex_state = 175},
  [3891] = {.lex_state = 0, .external_lex_state = 175},
  [3892] = {.lex_state = 0, .external_lex_state = 175},
  [3893] = {.lex_state = 0, .external_lex_state = 175},
  [3894] = {.lex_state = 0, .external_lex_state = 175},
  [3895] = {.lex_state = 0, .external_lex_state = 175},
  [3896] = {.lex_state = 0, .external_lex_state = 175},
  [3897] = {.lex_state = 0, .external_lex_state = 175},
  [3898] = {.lex_state = 0, .external_lex_state = 175},
  [3899] = {.lex_state = 0, .external_lex_state = 175},
  [3900] = {.lex_state = 0, .external_lex_state = 175},
  [3901] = {.lex_state = 0, .external_lex_state = 175},
  [3902] = {.lex_state = 0, .external_lex_state = 175},
  [3903] = {.lex_state = 0, .external_lex_state = 175},
  [3904] = {.lex_state = 0, .external_lex_state = 175},
  [3905] = {.lex_state = 0, .external_lex_state = 175},
  [3906] = {.lex_state = 0, .external_lex_state = 175},
  [3907] = {.lex_state = 0, .external_lex_state = 175},
  [3908] = {.lex_state = 0, .external_lex_state = 175},
  [3909] = {.lex_state = 0, .external_lex_state = 175},
  [3910] = {.lex_state = 0, .external_lex_state = 175},
  [3911] = {.lex_state = 0, .external_lex_state = 175},
  [3912] = {.lex_state = 0, .external_lex_state = 175},
  [3913] = {.lex_state = 0, .external_lex_state = 175},
  [3914] = {.lex_state = 0, .external_lex_state = 175},
  [3915] = {.lex_state = 0, .external_lex_state = 175},
  [3916] = {.lex_state = 0, .external_lex_state = 175},
  [3917] = {.lex_state = 0, .external_lex_state = 175},
  [3918] = {.lex_state = 0, .external_lex_state = 175},
  [3919] = {.lex_state = 0, .external_lex_state = 175},
  [3920] = {.lex_state = 0, .external_lex_state = 175},
  [3921] = {.lex_state = 0, .external_lex_state = 175},
  [3922] = {.lex_state = 0, .external_lex_state = 175},
  [3923] = {.lex_state = 0, .external_lex_state = 175},
  [3924] = {.lex_state = 0, .external_lex_state = 175},
  [3925] = {.lex_state = 0, .external_lex_state = 175},
  [3926] = {.lex_state = 0, .external_lex_state = 175},
  [3927] = {.lex_state = 0, .external_lex_state = 175},
  [3928] = {.lex_state = 0, .external_lex_state = 175},
  [3929] = {.lex_state = 0, .external_lex_state = 175},
  [3930] = {.lex_state = 0, .external_lex_state = 175},
  [3931] = {.lex_state = 0, .external_lex_state = 175},
  [3932] = {.lex_state = 0, .external_lex_state = 175},
  [3933] = {.lex_state = 0, .external_lex_state = 175},
  [3934] = {.lex_state = 0, .external_lex_state = 175},
  [3935] = {.lex_state = 0, .external_lex_state = 175},
  [3936] = {.lex_state = 0, .external_lex_state = 175},
  [3937] = {.lex_state = 0, .external_lex_state = 175},
  [3938] = {.lex_state = 0, .external_lex_state = 175},
  [3939] = {.lex_state = 0, .external_lex_state = 175},
  [3940] = {.lex_state = 0, .external_lex_state = 175},
  [3941] = {.lex_state = 0, .external_lex_state = 175},
  [3942] = {.lex_state = 0, .external_lex_state = 175},
  [3943] = {.lex_state = 0, .external_lex_state = 175},
  [3944] = {.lex_state = 0, .external_lex_state = 175},
  [3945] = {.lex_state = 0, .external_lex_state = 175},
  [3946] = {.lex_state = 0, .external_lex_state = 175},
  [3947] = {.lex_state = 0, .external_lex_state = 175},
  [3948] = {.lex_state = 0, .external_lex_state = 175},
  [3949] = {.lex_state = 0, .external_lex_state = 175},
  [3950] = {.lex_state = 0, .external_lex_state = 175},
  [3951] = {.lex_state = 0, .external_lex_state = 175},
  [3952] = {.lex_state = 0, .external_lex_state = 175},
  [3953] = {.lex_state = 0, .external_lex_state = 175},
  [3954] = {.lex_state = 0, .external_lex_state = 175},
  [3955] = {.lex_state = 0, .external_lex_state = 175},
  [3956] = {.lex_state = 0, .external_lex_state = 175},
  [3957] = {.lex_state = 0, .external_lex_state = 175},
  [3958] = {.lex_state = 0, .external_lex_state = 175},
  [3959] = {.lex_state = 0, .external_lex_state = 175},
  [3960] = {.lex_state = 0, .external_lex_state = 175},
  [3961] = {.lex_state = 0, .external_lex_state = 175},
  [3962] = {.lex_state = 0, .external_lex_state = 175},
  [3963] = {.lex_state = 0, .external_lex_state = 175},
  [3964] = {.lex_state = 0, .external_lex_state = 175},
  [3965] = {.lex_state = 0, .external_lex_state = 175},
  [3966] = {.lex_state = 0, .external_lex_state = 175},
  [3967] = {.lex_state = 0, .external_lex_state = 175},
  [3968] = {.lex_state = 0, .external_lex_state = 175},
  [3969] = {.lex_state = 0, .external_lex_state = 175},
  [3970] = {.lex_state = 0, .external_lex_state = 175},
  [3971] = {.lex_state = 0, .external_lex_state = 175},
  [3972] = {.lex_state = 0, .external_lex_state = 175},
  [3973] = {.lex_state = 0, .external_lex_state = 175},
  [3974] = {.lex_state = 0, .external_lex_state = 175},
  [3975] = {.lex_state = 0, .external_lex_state = 175},
  [3976] = {.lex_state = 2, .external_lex_state = 141},
  [3977] = {.lex_state = 0, .external_lex_state = 160},
  [3978] = {.lex_state = 2, .external_lex_state = 152},
  [3979] = {.lex_state = 2, .external_lex_state = 77},
  [3980] = {.lex_state = 2, .external_lex_state = 152},
  [3981] = {.lex_state = 2, .external_lex_state = 77},
  [3982] = {.lex_state = 2, .external_lex_state = 152},
  [3983] = {.lex_state = 2, .external_lex_state = 152},
  [3984] = {.lex_state = 2, .external_lex_state = 152},
  [3985] = {.lex_state = 2, .external_lex_state = 152},
  [3986] = {.lex_state = 2, .external_lex_state = 152},
  [3987] = {.lex_state = 2, .external_lex_state = 152},
  [3988] = {.lex_state = 2, .external_lex_state = 77},
  [3989] = {.lex_state = 2, .external_lex_state = 152},
  [3990] = {.lex_state = 2, .external_lex_state = 152},
  [3991] = {.lex_state = 2, .external_lex_state = 152},
  [3992] = {.lex_state = 2, .external_lex_state = 77},
  [3993] = {.lex_state = 2, .external_lex_state = 152},
  [3994] = {.lex_state = 2, .external_lex_state = 152},
  [3995] = {.lex_state = 2, .external_lex_state = 152},
  [3996] = {.lex_state = 2, .external_lex_state = 152},
  [3997] = {.lex_state = 2, .external_lex_state = 152},
  [3998] = {.lex_state = 2, .external_lex_state = 152},
  [3999] = {.lex_state = 2, .external_lex_state = 77},
  [4000] = {.lex_state = 2, .external_lex_state = 152},
  [4001] = {.lex_state = 2, .external_lex_state = 152},
  [4002] = {.lex_state = 2, .external_lex_state = 152},
  [4003] = {.lex_state = 2, .external_lex_state = 152},
  [4004] = {.lex_state = 2, .external_lex_state = 152},
  [4005] = {.lex_state = 2, .external_lex_state = 152},
  [4006] = {.lex_state = 2, .external_lex_state = 152},
  [4007] = {.lex_state = 2, .external_lex_state = 152},
  [4008] = {.lex_state = 2, .external_lex_state = 152},
  [4009] = {.lex_state = 2, .external_lex_state = 152},
  [4010] = {.lex_state = 2, .external_lex_state = 152},
  [4011] = {.lex_state = 2, .external_lex_state = 77},
  [4012] = {.lex_state = 2, .external_lex_state = 152},
  [4013] = {.lex_state = 2, .external_lex_state = 152},
  [4014] = {.lex_state = 2, .external_lex_state = 77},
  [4015] = {.lex_state = 2, .external_lex_state = 152},
  [4016] = {.lex_state = 2, .external_lex_state = 152},
  [4017] = {.lex_state = 2, .external_lex_state = 152},
  [4018] = {.lex_state = 2, .external_lex_state = 152},
  [4019] = {.lex_state = 2, .external_lex_state = 152},
  [4020] = {.lex_state = 2, .external_lex_state = 152},
  [4021] = {.lex_state = 2, .external_lex_state = 152},
  [4022] = {.lex_state = 2, .external_lex_state = 152},
  [4023] = {.lex_state = 2, .external_lex_state = 152},
  [4024] = {.lex_state = 2, .external_lex_state = 152},
  [4025] = {.lex_state = 2, .external_lex_state = 77},
  [4026] = {.lex_state = 2, .external_lex_state = 77},
  [4027] = {.lex_state = 2, .external_lex_state = 152},
  [4028] = {.lex_state = 2, .external_lex_state = 152},
  [4029] = {.lex_state = 2, .external_lex_state = 152},
  [4030] = {.lex_state = 2, .external_lex_state = 152},
  [4031] = {.lex_state = 2, .external_lex_state = 152},
  [4032] = {.lex_state = 2, .external_lex_state = 152},
  [4033] = {.lex_state = 2, .external_lex_state = 152},
  [4034] = {.lex_state = 2, .external_lex_state = 152},
  [4035] = {.lex_state = 2, .external_lex_state = 152},
  [4036] = {.lex_state = 2, .external_lex_state = 152},
  [4037] = {.lex_state = 2, .external_lex_state = 152},
  [4038] = {.lex_state = 2, .external_lex_state = 152},
  [4039] = {.lex_state = 2, .external_lex_state = 152},
  [4040] = {.lex_state = 2, .external_lex_state = 152},
  [4041] = {.lex_state = 2, .external_lex_state = 152},
  [4042] = {.lex_state = 2, .external_lex_state = 152},
  [4043] = {.lex_state = 2, .external_lex_state = 152},
  [4044] = {.lex_state = 2, .external_lex_state = 152},
  [4045] = {.lex_state = 2, .external_lex_state = 152},
  [4046] = {.lex_state = 2, .external_lex_state = 152},
  [4047] = {.lex_state = 2, .external_lex_state = 152},
  [4048] = {.lex_state = 2, .external_lex_state = 152},
  [4049] = {.lex_state = 2, .external_lex_state = 152},
  [4050] = {.lex_state = 2, .external_lex_state = 152},
  [4051] = {.lex_state = 2, .external_lex_state = 152},
  [4052] = {.lex_state = 2, .external_lex_state = 152},
  [4053] = {.lex_state = 2, .external_lex_state = 152},
  [4054] = {.lex_state = 2, .external_lex_state = 152},
  [4055] = {.lex_state = 2, .external_lex_state = 152},
  [4056] = {.lex_state = 2, .external_lex_state = 152},
  [4057] = {.lex_state = 2, .external_lex_state = 152},
  [4058] = {.lex_state = 2, .external_lex_state = 152},
  [4059] = {.lex_state = 2, .external_lex_state = 152},
  [4060] = {.lex_state = 2, .external_lex_state = 152},
  [4061] = {.lex_state = 2, .external_lex_state = 152},
  [4062] = {.lex_state = 2, .external_lex_state = 152},
  [4063] = {.lex_state = 2, .external_lex_state = 152},
  [4064] = {.lex_state = 2, .external_lex_state = 152},
  [4065] = {.lex_state = 2, .external_lex_state = 152},
  [4066] = {.lex_state = 2, .external_lex_state = 152},
  [4067] = {.lex_state = 2, .external_lex_state = 152},
  [4068] = {.lex_state = 2, .external_lex_state = 152},
  [4069] = {.lex_state = 2, .external_lex_state = 152},
  [4070] = {.lex_state = 2, .external_lex_state = 152},
  [4071] = {.lex_state = 2, .external_lex_state = 152},
  [4072] = {.lex_state = 2, .external_lex_state = 77},
  [4073] = {.lex_state = 2, .external_lex_state = 152},
  [4074] = {.lex_state = 2, .external_lex_state = 152},
  [4075] = {.lex_state = 2, .external_lex_state = 152},
  [4076] = {.lex_state = 2, .external_lex_state = 152},
  [4077] = {.lex_state = 2, .external_lex_state = 152},
  [4078] = {.lex_state = 2, .external_lex_state = 152},
  [4079] = {.lex_state = 2, .external_lex_state = 77},
  [4080] = {.lex_state = 2, .external_lex_state = 152},
  [4081] = {.lex_state = 2, .external_lex_state = 152},
  [4082] = {.lex_state = 2, .external_lex_state = 152},
  [4083] = {.lex_state = 2, .external_lex_state = 152},
  [4084] = {.lex_state = 2, .external_lex_state = 152},
  [4085] = {.lex_state = 2, .external_lex_state = 152},
  [4086] = {.lex_state = 2, .external_lex_state = 152},
  [4087] = {.lex_state = 2, .external_lex_state = 77},
  [4088] = {.lex_state = 2, .external_lex_state = 152},
  [4089] = {.lex_state = 2, .external_lex_state = 77},
  [4090] = {.lex_state = 2, .external_lex_state = 77},
  [4091] = {.lex_state = 2, .external_lex_state = 77},
  [4092] = {.lex_state = 2, .external_lex_state = 152},
  [4093] = {.lex_state = 2, .external_lex_state = 152},
  [4094] = {.lex_state = 2, .external_lex_state = 152},
  [4095] = {.lex_state = 2, .external_lex_state = 152},
  [4096] = {.lex_state = 2, .external_lex_state = 152},
  [4097] = {.lex_state = 2, .external_lex_state = 152},
  [4098] = {.lex_state = 2, .external_lex_state = 152},
  [4099] = {.lex_state = 2, .external_lex_state = 152},
  [4100] = {.lex_state = 2, .external_lex_state = 77},
  [4101] = {.lex_state = 2, .external_lex_state = 152},
  [4102] = {.lex_state = 2, .external_lex_state = 152},
  [4103] = {.lex_state = 2, .external_lex_state = 152},
  [4104] = {.lex_state = 2, .external_lex_state = 152},
  [4105] = {.lex_state = 2, .external_lex_state = 152},
  [4106] = {.lex_state = 2, .external_lex_state = 152},
  [4107] = {.lex_state = 2, .external_lex_state = 152},
  [4108] = {.lex_state = 2, .external_lex_state = 152},
  [4109] = {.lex_state = 2, .external_lex_state = 152},
  [4110] = {.lex_state = 2, .external_lex_state = 152},
  [4111] = {.lex_state = 2, .external_lex_state = 152},
  [4112] = {.lex_state = 2, .external_lex_state = 152},
  [4113] = {.lex_state = 2, .external_lex_state = 152},
  [4114] = {.lex_state = 2, .external_lex_state = 152},
  [4115] = {.lex_state = 2, .external_lex_state = 152},
  [4116] = {.lex_state = 2, .external_lex_state = 152},
  [4117] = {.lex_state = 2, .external_lex_state = 152},
  [4118] = {.lex_state = 2, .external_lex_state = 152},
  [4119] = {.lex_state = 2, .external_lex_state = 152},
  [4120] = {.lex_state = 2, .external_lex_state = 152},
  [4121] = {.lex_state = 2, .external_lex_state = 152},
  [4122] = {.lex_state = 2, .external_lex_state = 152},
  [4123] = {.lex_state = 2, .external_lex_state = 152},
  [4124] = {.lex_state = 2, .external_lex_state = 152},
  [4125] = {.lex_state = 2, .external_lex_state = 152},
  [4126] = {.lex_state = 2, .external_lex_state = 152},
  [4127] = {.lex_state = 2, .external_lex_state = 77},
  [4128] = {.lex_state = 2, .external_lex_state = 152},
  [4129] = {.lex_state = 2, .external_lex_state = 77},
  [4130] = {.lex_state = 2, .external_lex_state = 77},
  [4131] = {.lex_state = 2, .external_lex_state = 152},
  [4132] = {.lex_state = 2, .external_lex_state = 77},
  [4133] = {.lex_state = 2, .external_lex_state = 152},
  [4134] = {.lex_state = 2, .external_lex_state = 152},
  [4135] = {.lex_state = 2, .external_lex_state = 152},
  [4136] = {.lex_state = 2, .external_lex_state = 152},
  [4137] = {.lex_state = 2, .external_lex_state = 152},
  [4138] = {.lex_state = 2, .external_lex_state = 152},
  [4139] = {.lex_state = 2, .external_lex_state = 152},
  [4140] = {.lex_state = 2, .external_lex_state = 77},
  [4141] = {.lex_state = 2, .external_lex_state = 152},
  [4142] = {.lex_state = 2, .external_lex_state = 152},
  [4143] = {.lex_state = 2, .external_lex_state = 152},
  [4144] = {.lex_state = 2, .external_lex_state = 152},
  [4145] = {.lex_state = 2, .external_lex_state = 152},
  [4146] = {.lex_state = 2, .external_lex_state = 77},
  [4147] = {.lex_state = 2, .external_lex_state = 152},
  [4148] = {.lex_state = 2, .external_lex_state = 152},
  [4149] = {.lex_state = 2, .external_lex_state = 152},
  [4150] = {.lex_state = 2, .external_lex_state = 77},
  [4151] = {.lex_state = 2, .external_lex_state = 145},
  [4152] = {.lex_state = 2, .external_lex_state = 152},
  [4153] = {.lex_state = 2, .external_lex_state = 152},
  [4154] = {.lex_state = 2, .external_lex_state = 152},
  [4155] = {.lex_state = 2, .external_lex_state = 152},
  [4156] = {.lex_state = 2, .external_lex_state = 77},
  [4157] = {.lex_state = 2, .external_lex_state = 77},
  [4158] = {.lex_state = 0, .external_lex_state = 147},
  [4159] = {.lex_state = 2, .external_lex_state = 152},
  [4160] = {.lex_state = 2, .external_lex_state = 152},
  [4161] = {.lex_state = 2, .external_lex_state = 152},
  [4162] = {.lex_state = 2, .external_lex_state = 152},
  [4163] = {.lex_state = 2, .external_lex_state = 152},
  [4164] = {.lex_state = 2, .external_lex_state = 152},
  [4165] = {.lex_state = 2, .external_lex_state = 77},
  [4166] = {.lex_state = 2, .external_lex_state = 152},
  [4167] = {.lex_state = 2, .external_lex_state = 152},
  [4168] = {.lex_state = 2, .external_lex_state = 152},
  [4169] = {.lex_state = 2, .external_lex_state = 77},
  [4170] = {.lex_state = 2, .external_lex_state = 77},
  [4171] = {.lex_state = 0, .external_lex_state = 179},
  [4172] = {.lex_state = 2, .external_lex_state = 152},
  [4173] = {.lex_state = 2, .external_lex_state = 152},
  [4174] = {.lex_state = 2, .external_lex_state = 152},
  [4175] = {.lex_state = 2, .external_lex_state = 152},
  [4176] = {.lex_state = 2, .external_lex_state = 152},
  [4177] = {.lex_state = 2, .external_lex_state = 152},
  [4178] = {.lex_state = 2, .external_lex_state = 152},
  [4179] = {.lex_state = 2, .external_lex_state = 152},
  [4180] = {.lex_state = 2, .external_lex_state = 152},
  [4181] = {.lex_state = 2, .external_lex_state = 152},
  [4182] = {.lex_state = 2, .external_lex_state = 152},
  [4183] = {.lex_state = 2, .external_lex_state = 152},
  [4184] = {.lex_state = 2, .external_lex_state = 152},
  [4185] = {.lex_state = 2, .external_lex_state = 152},
  [4186] = {.lex_state = 2, .external_lex_state = 152},
  [4187] = {.lex_state = 2, .external_lex_state = 152},
  [4188] = {.lex_state = 2, .external_lex_state = 152},
  [4189] = {.lex_state = 2, .external_lex_state = 152},
  [4190] = {.lex_state = 2, .external_lex_state = 152},
  [4191] = {.lex_state = 2, .external_lex_state = 152},
  [4192] = {.lex_state = 2, .external_lex_state = 152},
  [4193] = {.lex_state = 2, .external_lex_state = 152},
  [4194] = {.lex_state = 2, .external_lex_state = 77},
  [4195] = {.lex_state = 0, .external_lex_state = 179},
  [4196] = {.lex_state = 2, .external_lex_state = 152},
  [4197] = {.lex_state = 2, .external_lex_state = 77},
  [4198] = {.lex_state = 2, .external_lex_state = 152},
  [4199] = {.lex_state = 2, .external_lex_state = 152},
  [4200] = {.lex_state = 2, .external_lex_state = 152},
  [4201] = {.lex_state = 2, .external_lex_state = 152},
  [4202] = {.lex_state = 2, .external_lex_state = 77},
  [4203] = {.lex_state = 2, .external_lex_state = 152},
  [4204] = {.lex_state = 2, .external_lex_state = 152},
  [4205] = {.lex_state = 2, .external_lex_state = 152},
  [4206] = {.lex_state = 2, .external_lex_state = 152},
  [4207] = {.lex_state = 2, .external_lex_state = 152},
  [4208] = {.lex_state = 2, .external_lex_state = 152},
  [4209] = {.lex_state = 2, .external_lex_state = 152},
  [4210] = {.lex_state = 2, .external_lex_state = 152},
  [4211] = {.lex_state = 2, .external_lex_state = 152},
  [4212] = {.lex_state = 2, .external_lex_state = 152},
  [4213] = {.lex_state = 2, .external_lex_state = 152},
  [4214] = {.lex_state = 2, .external_lex_state = 77},
  [4215] = {.lex_state = 2, .external_lex_state = 152},
  [4216] = {.lex_state = 2, .external_lex_state = 152},
  [4217] = {.lex_state = 2, .external_lex_state = 152},
  [4218] = {.lex_state = 2, .external_lex_state = 152},
  [4219] = {.lex_state = 2, .external_lex_state = 77},
  [4220] = {.lex_state = 2, .external_lex_state = 152},
  [4221] = {.lex_state = 2, .external_lex_state = 152},
  [4222] = {.lex_state = 2, .external_lex_state = 152},
  [4223] = {.lex_state = 2, .external_lex_state = 152},
  [4224] = {.lex_state = 2, .external_lex_state = 152},
  [4225] = {.lex_state = 2, .external_lex_state = 152},
  [4226] = {.lex_state = 2, .external_lex_state = 152},
  [4227] = {.lex_state = 0, .external_lex_state = 179},
  [4228] = {.lex_state = 2, .external_lex_state = 77},
  [4229] = {.lex_state = 2, .external_lex_state = 77},
  [4230] = {.lex_state = 2, .external_lex_state = 152},
  [4231] = {.lex_state = 2, .external_lex_state = 152},
  [4232] = {.lex_state = 2, .external_lex_state = 152},
  [4233] = {.lex_state = 2, .external_lex_state = 152},
  [4234] = {.lex_state = 2, .external_lex_state = 152},
  [4235] = {.lex_state = 2, .external_lex_state = 152},
  [4236] = {.lex_state = 2, .external_lex_state = 152},
  [4237] = {.lex_state = 2, .external_lex_state = 152},
  [4238] = {.lex_state = 2, .external_lex_state = 152},
  [4239] = {.lex_state = 2, .external_lex_state = 152},
  [4240] = {.lex_state = 2, .external_lex_state = 152},
  [4241] = {.lex_state = 2, .external_lex_state = 152},
  [4242] = {.lex_state = 2, .external_lex_state = 152},
  [4243] = {.lex_state = 2, .external_lex_state = 77},
  [4244] = {.lex_state = 2, .external_lex_state = 152},
  [4245] = {.lex_state = 2, .external_lex_state = 152},
  [4246] = {.lex_state = 2, .external_lex_state = 152},
  [4247] = {.lex_state = 2, .external_lex_state = 152},
  [4248] = {.lex_state = 2, .external_lex_state = 152},
  [4249] = {.lex_state = 2, .external_lex_state = 152},
  [4250] = {.lex_state = 2, .external_lex_state = 152},
  [4251] = {.lex_state = 0, .external_lex_state = 147},
  [4252] = {.lex_state = 2, .external_lex_state = 180},
  [4253] = {.lex_state = 2, .external_lex_state = 152},
  [4254] = {.lex_state = 2, .external_lex_state = 152},
  [4255] = {.lex_state = 2, .external_lex_state = 152},
  [4256] = {.lex_state = 2, .external_lex_state = 152},
  [4257] = {.lex_state = 2, .external_lex_state = 152},
  [4258] = {.lex_state = 2, .external_lex_state = 152},
  [4259] = {.lex_state = 2, .external_lex_state = 152},
  [4260] = {.lex_state = 2, .external_lex_state = 152},
  [4261] = {.lex_state = 2, .external_lex_state = 152},
  [4262] = {.lex_state = 2, .external_lex_state = 152},
  [4263] = {.lex_state = 2, .external_lex_state = 152},
  [4264] = {.lex_state = 2, .external_lex_state = 152},
  [4265] = {.lex_state = 2, .external_lex_state = 152},
  [4266] = {.lex_state = 2, .external_lex_state = 152},
  [4267] = {.lex_state = 2, .external_lex_state = 77},
  [4268] = {.lex_state = 2, .external_lex_state = 152},
  [4269] = {.lex_state = 2, .external_lex_state = 152},
  [4270] = {.lex_state = 2, .external_lex_state = 152},
  [4271] = {.lex_state = 2, .external_lex_state = 152},
  [4272] = {.lex_state = 0, .external_lex_state = 147},
  [4273] = {.lex_state = 2, .external_lex_state = 152},
  [4274] = {.lex_state = 2, .external_lex_state = 152},
  [4275] = {.lex_state = 2, .external_lex_state = 152},
  [4276] = {.lex_state = 2, .external_lex_state = 152},
  [4277] = {.lex_state = 2, .external_lex_state = 152},
  [4278] = {.lex_state = 2, .external_lex_state = 152},
  [4279] = {.lex_state = 2, .external_lex_state = 152},
  [4280] = {.lex_state = 2, .external_lex_state = 152},
  [4281] = {.lex_state = 2, .external_lex_state = 152},
  [4282] = {.lex_state = 0, .external_lex_state = 147},
  [4283] = {.lex_state = 2, .external_lex_state = 152},
  [4284] = {.lex_state = 2, .external_lex_state = 152},
  [4285] = {.lex_state = 2, .external_lex_state = 152},
  [4286] = {.lex_state = 2, .external_lex_state = 152},
  [4287] = {.lex_state = 2, .external_lex_state = 152},
  [4288] = {.lex_state = 2, .external_lex_state = 152},
  [4289] = {.lex_state = 2, .external_lex_state = 152},
  [4290] = {.lex_state = 2, .external_lex_state = 152},
  [4291] = {.lex_state = 2, .external_lex_state = 77},
  [4292] = {.lex_state = 2, .external_lex_state = 152},
  [4293] = {.lex_state = 2, .external_lex_state = 152},
  [4294] = {.lex_state = 2, .external_lex_state = 152},
  [4295] = {.lex_state = 2, .external_lex_state = 152},
  [4296] = {.lex_state = 2, .external_lex_state = 152},
  [4297] = {.lex_state = 2, .external_lex_state = 77},
  [4298] = {.lex_state = 2, .external_lex_state = 152},
  [4299] = {.lex_state = 2, .external_lex_state = 152},
  [4300] = {.lex_state = 2, .external_lex_state = 152},
  [4301] = {.lex_state = 2, .external_lex_state = 152},
  [4302] = {.lex_state = 2, .external_lex_state = 152},
  [4303] = {.lex_state = 2, .external_lex_state = 152},
  [4304] = {.lex_state = 2, .external_lex_state = 152},
  [4305] = {.lex_state = 2, .external_lex_state = 152},
  [4306] = {.lex_state = 2, .external_lex_state = 152},
  [4307] = {.lex_state = 2, .external_lex_state = 152},
  [4308] = {.lex_state = 2, .external_lex_state = 152},
  [4309] = {.lex_state = 2, .external_lex_state = 152},
  [4310] = {.lex_state = 2, .external_lex_state = 152},
  [4311] = {.lex_state = 2, .external_lex_state = 152},
  [4312] = {.lex_state = 2, .external_lex_state = 152},
  [4313] = {.lex_state = 2, .external_lex_state = 152},
  [4314] = {.lex_state = 0, .external_lex_state = 147},
  [4315] = {.lex_state = 0, .external_lex_state = 147},
  [4316] = {.lex_state = 2, .external_lex_state = 84},
  [4317] = {.lex_state = 2, .external_lex_state = 84},
  [4318] = {.lex_state = 0, .external_lex_state = 147},
  [4319] = {.lex_state = 2, .external_lex_state = 84},
  [4320] = {.lex_state = 0, .external_lex_state = 147},
  [4321] = {.lex_state = 2, .external_lex_state = 84},
  [4322] = {.lex_state = 2, .external_lex_state = 84},
  [4323] = {.lex_state = 2, .external_lex_state = 84},
  [4324] = {.lex_state = 2, .external_lex_state = 84},
  [4325] = {.lex_state = 2, .external_lex_state = 84},
  [4326] = {.lex_state = 2, .external_lex_state = 84},
  [4327] = {.lex_state = 2, .external_lex_state = 84},
  [4328] = {.lex_state = 2, .external_lex_state = 84},
  [4329] = {.lex_state = 0, .external_lex_state = 147},
  [4330] = {.lex_state = 2, .external_lex_state = 84},
  [4331] = {.lex_state = 0, .external_lex_state = 147},
  [4332] = {.lex_state = 2, .external_lex_state = 84},
  [4333] = {.lex_state = 2, .external_lex_state = 84},
  [4334] = {.lex_state = 2, .external_lex_state = 84},
  [4335] = {.lex_state = 2, .external_lex_state = 84},
  [4336] = {.lex_state = 0, .external_lex_state = 147},
  [4337] = {.lex_state = 0, .external_lex_state = 147},
  [4338] = {.lex_state = 0, .external_lex_state = 147},
  [4339] = {.lex_state = 0, .external_lex_state = 147},
  [4340] = {.lex_state = 2, .external_lex_state = 84},
  [4341] = {.lex_state = 0, .external_lex_state = 147},
  [4342] = {.lex_state = 0, .external_lex_state = 147},
  [4343] = {.lex_state = 0, .external_lex_state = 147},
  [4344] = {.lex_state = 2, .external_lex_state = 181},
  [4345] = {.lex_state = 0, .external_lex_state = 147},
  [4346] = {.lex_state = 0, .external_lex_state = 147},
  [4347] = {.lex_state = 2, .external_lex_state = 84},
  [4348] = {.lex_state = 2, .external_lex_state = 84},
  [4349] = {.lex_state = 2, .external_lex_state = 84},
  [4350] = {.lex_state = 2, .external_lex_state = 84},
  [4351] = {.lex_state = 0, .external_lex_state = 147},
  [4352] = {.lex_state = 0, .external_lex_state = 147},
  [4353] = {.lex_state = 2, .external_lex_state = 84},
  [4354] = {.lex_state = 2, .external_lex_state = 84},
  [4355] = {.lex_state = 0, .external_lex_state = 147},
  [4356] = {.lex_state = 0, .external_lex_state = 147},
  [4357] = {.lex_state = 0, .external_lex_state = 147},
  [4358] = {.lex_state = 2, .external_lex_state = 84},
  [4359] = {.lex_state = 2, .external_lex_state = 84},
  [4360] = {.lex_state = 2, .external_lex_state = 84},
  [4361] = {.lex_state = 2, .external_lex_state = 84},
  [4362] = {.lex_state = 0, .external_lex_state = 147},
  [4363] = {.lex_state = 0, .external_lex_state = 147},
  [4364] = {.lex_state = 2, .external_lex_state = 84},
  [4365] = {.lex_state = 2, .external_lex_state = 182},
  [4366] = {.lex_state = 0, .external_lex_state = 147},
  [4367] = {.lex_state = 2, .external_lex_state = 84},
  [4368] = {.lex_state = 0, .external_lex_state = 147},
  [4369] = {.lex_state = 2, .external_lex_state = 84},
  [4370] = {.lex_state = 2, .external_lex_state = 84},
  [4371] = {.lex_state = 2, .external_lex_state = 84},
  [4372] = {.lex_state = 0, .external_lex_state = 147},
  [4373] = {.lex_state = 2, .external_lex_state = 84},
  [4374] = {.lex_state = 0, .external_lex_state = 147},
  [4375] = {.lex_state = 0, .external_lex_state = 147},
  [4376] = {.lex_state = 0, .external_lex_state = 147},
  [4377] = {.lex_state = 0, .external_lex_state = 147},
  [4378] = {.lex_state = 0, .external_lex_state = 147},
  [4379] = {.lex_state = 0, .external_lex_state = 147},
  [4380] = {.lex_state = 0, .external_lex_state = 147},
  [4381] = {.lex_state = 2, .external_lex_state = 84},
  [4382] = {.lex_state = 2, .external_lex_state = 84},
  [4383] = {.lex_state = 2, .external_lex_state = 84},
  [4384] = {.lex_state = 2, .external_lex_state = 84},
  [4385] = {.lex_state = 2, .external_lex_state = 84},
  [4386] = {.lex_state = 2, .external_lex_state = 84},
  [4387] = {.lex_state = 0, .external_lex_state = 183},
  [4388] = {.lex_state = 0, .external_lex_state = 184},
  [4389] = {.lex_state = 2, .external_lex_state = 185},
  [4390] = {.lex_state = 0, .external_lex_state = 186},
  [4391] = {.lex_state = 2, .external_lex_state = 187},
  [4392] = {.lex_state = 0, .external_lex_state = 183},
  [4393] = {.lex_state = 0, .external_lex_state = 183},
  [4394] = {.lex_state = 0, .external_lex_state = 183},
  [4395] = {.lex_state = 0, .external_lex_state = 183},
  [4396] = {.lex_state = 0, .external_lex_state = 183},
  [4397] = {.lex_state = 0, .external_lex_state = 183},
  [4398] = {.lex_state = 0, .external_lex_state = 183},
  [4399] = {.lex_state = 0, .external_lex_state = 153},
  [4400] = {.lex_state = 0, .external_lex_state = 153},
  [4401] = {.lex_state = 0, .external_lex_state = 153},
  [4402] = {.lex_state = 0, .external_lex_state = 153},
  [4403] = {.lex_state = 0, .external_lex_state = 153},
  [4404] = {.lex_state = 2, .external_lex_state = 137},
  [4405] = {.lex_state = 0, .external_lex_state = 153},
  [4406] = {.lex_state = 0, .external_lex_state = 153},
  [4407] = {.lex_state = 0, .external_lex_state = 153},
  [4408] = {.lex_state = 0, .external_lex_state = 153},
  [4409] = {.lex_state = 0, .external_lex_state = 153},
  [4410] = {.lex_state = 0, .external_lex_state = 153},
  [4411] = {.lex_state = 0, .external_lex_state = 153},
  [4412] = {.lex_state = 0, .external_lex_state = 153},
  [4413] = {.lex_state = 0, .external_lex_state = 153},
  [4414] = {.lex_state = 2, .external_lex_state = 137},
  [4415] = {.lex_state = 0, .external_lex_state = 153},
  [4416] = {.lex_state = 2, .external_lex_state = 137},
  [4417] = {.lex_state = 2, .external_lex_state = 137},
  [4418] = {.lex_state = 2, .external_lex_state = 137},
  [4419] = {.lex_state = 0, .external_lex_state = 153},
  [4420] = {.lex_state = 2, .external_lex_state = 137},
  [4421] = {.lex_state = 0, .external_lex_state = 153},
  [4422] = {.lex_state = 0, .external_lex_state = 153},
  [4423] = {.lex_state = 0, .external_lex_state = 153},
  [4424] = {.lex_state = 2, .external_lex_state = 137},
  [4425] = {.lex_state = 0, .external_lex_state = 153},
  [4426] = {.lex_state = 0, .external_lex_state = 153},
  [4427] = {.lex_state = 2, .external_lex_state = 137},
  [4428] = {.lex_state = 2, .external_lex_state = 137},
  [4429] = {.lex_state = 0, .external_lex_state = 153},
  [4430] = {.lex_state = 2, .external_lex_state = 137},
  [4431] = {.lex_state = 0, .external_lex_state = 157},
  [4432] = {.lex_state = 0, .external_lex_state = 156},
  [4433] = {.lex_state = 0, .external_lex_state = 153},
  [4434] = {.lex_state = 0, .external_lex_state = 153},
  [4435] = {.lex_state = 0, .external_lex_state = 153},
  [4436] = {.lex_state = 0, .external_lex_state = 153},
  [4437] = {.lex_state = 2, .external_lex_state = 137},
  [4438] = {.lex_state = 2, .external_lex_state = 137},
  [4439] = {.lex_state = 2, .external_lex_state = 137},
  [4440] = {.lex_state = 0, .external_lex_state = 153},
  [4441] = {.lex_state = 0, .external_lex_state = 158},
  [4442] = {.lex_state = 0, .external_lex_state = 153},
  [4443] = {.lex_state = 0, .external_lex_state = 153},
  [4444] = {.lex_state = 0, .external_lex_state = 153},
  [4445] = {.lex_state = 0, .external_lex_state = 158},
  [4446] = {.lex_state = 0, .external_lex_state = 156},
  [4447] = {.lex_state = 0, .external_lex_state = 156},
  [4448] = {.lex_state = 0, .external_lex_state = 188},
  [4449] = {.lex_state = 0, .external_lex_state = 156},
  [4450] = {.lex_state = 0, .external_lex_state = 156},
  [4451] = {.lex_state = 0, .external_lex_state = 156},
  [4452] = {.lex_state = 0, .external_lex_state = 158},
  [4453] = {.lex_state = 0, .external_lex_state = 156},
  [4454] = {.lex_state = 0, .external_lex_state = 158},
  [4455] = {.lex_state = 0, .external_lex_state = 158},
  [4456] = {.lex_state = 0, .external_lex_state = 158},
  [4457] = {.lex_state = 0, .external_lex_state = 158},
  [4458] = {.lex_state = 0, .external_lex_state = 158},
  [4459] = {.lex_state = 0, .external_lex_state = 158},
  [4460] = {.lex_state = 0, .external_lex_state = 158},
  [4461] = {.lex_state = 0, .external_lex_state = 165},
  [4462] = {.lex_state = 0, .external_lex_state = 172},
  [4463] = {.lex_state = 0, .external_lex_state = 158},
  [4464] = {.lex_state = 0, .external_lex_state = 158},
  [4465] = {.lex_state = 0, .external_lex_state = 158},
  [4466] = {.lex_state = 0, .external_lex_state = 189},
  [4467] = {.lex_state = 0, .external_lex_state = 158},
  [4468] = {.lex_state = 0, .external_lex_state = 158},
  [4469] = {.lex_state = 0, .external_lex_state = 158},
  [4470] = {.lex_state = 0, .external_lex_state = 156},
  [4471] = {.lex_state = 0, .external_lex_state = 158},
  [4472] = {.lex_state = 0, .external_lex_state = 157},
  [4473] = {.lex_state = 0, .external_lex_state = 158},
  [4474] = {.lex_state = 0, .external_lex_state = 158},
  [4475] = {.lex_state = 0, .external_lex_state = 158},
  [4476] = {.lex_state = 0, .external_lex_state = 158},
  [4477] = {.lex_state = 0, .external_lex_state = 158},
  [4478] = {.lex_state = 0, .external_lex_state = 156},
  [4479] = {.lex_state = 0, .external_lex_state = 158},
  [4480] = {.lex_state = 0, .external_lex_state = 156},
  [4481] = {.lex_state = 0, .external_lex_state = 161},
  [4482] = {.lex_state = 0, .external_lex_state = 158},
  [4483] = {.lex_state = 0, .external_lex_state = 156},
  [4484] = {.lex_state = 0, .external_lex_state = 156},
  [4485] = {.lex_state = 0, .external_lex_state = 156},
  [4486] = {.lex_state = 0, .external_lex_state = 188},
  [4487] = {.lex_state = 0, .external_lex_state = 188},
  [4488] = {.lex_state = 0, .external_lex_state = 156},
  [4489] = {.lex_state = 0, .external_lex_state = 157},
  [4490] = {.lex_state = 0, .external_lex_state = 156},
  [4491] = {.lex_state = 0, .external_lex_state = 164},
  [4492] = {.lex_state = 0, .external_lex_state = 158},
  [4493] = {.lex_state = 0, .external_lex_state = 156},
  [4494] = {.lex_state = 0, .external_lex_state = 156},
  [4495] = {.lex_state = 0, .external_lex_state = 188},
  [4496] = {.lex_state = 0, .external_lex_state = 156},
  [4497] = {.lex_state = 0, .external_lex_state = 158},
  [4498] = {.lex_state = 0, .external_lex_state = 157},
  [4499] = {.lex_state = 0, .external_lex_state = 156},
  [4500] = {.lex_state = 0, .external_lex_state = 157},
  [4501] = {.lex_state = 0, .external_lex_state = 190},
  [4502] = {.lex_state = 0, .external_lex_state = 156},
  [4503] = {.lex_state = 0, .external_lex_state = 158},
  [4504] = {.lex_state = 0, .external_lex_state = 158},
  [4505] = {.lex_state = 0, .external_lex_state = 191},
  [4506] = {.lex_state = 2, .external_lex_state = 192},
  [4507] = {.lex_state = 2, .external_lex_state = 159},
  [4508] = {.lex_state = 0, .external_lex_state = 156},
  [4509] = {.lex_state = 0, .external_lex_state = 158},
  [4510] = {.lex_state = 0, .external_lex_state = 156},
  [4511] = {.lex_state = 0, .external_lex_state = 157},
  [4512] = {.lex_state = 0, .external_lex_state = 157},
  [4513] = {.lex_state = 0, .external_lex_state = 158},
  [4514] = {.lex_state = 0, .external_lex_state = 190},
  [4515] = {.lex_state = 0, .external_lex_state = 157},
  [4516] = {.lex_state = 0, .external_lex_state = 157},
  [4517] = {.lex_state = 0, .external_lex_state = 157},
  [4518] = {.lex_state = 0, .external_lex_state = 157},
  [4519] = {.lex_state = 0, .external_lex_state = 157},
  [4520] = {.lex_state = 0, .external_lex_state = 157},
  [4521] = {.lex_state = 0, .external_lex_state = 157},
  [4522] = {.lex_state = 0, .external_lex_state = 157},
  [4523] = {.lex_state = 0, .external_lex_state = 156},
  [4524] = {.lex_state = 0, .external_lex_state = 157},
  [4525] = {.lex_state = 0, .external_lex_state = 188},
  [4526] = {.lex_state = 0, .external_lex_state = 157},
  [4527] = {.lex_state = 0, .external_lex_state = 156},
  [4528] = {.lex_state = 0, .external_lex_state = 157},
  [4529] = {.lex_state = 0, .external_lex_state = 157},
  [4530] = {.lex_state = 0, .external_lex_state = 157},
  [4531] = {.lex_state = 0, .external_lex_state = 156},
  [4532] = {.lex_state = 0, .external_lex_state = 157},
  [4533] = {.lex_state = 0, .external_lex_state = 156},
  [4534] = {.lex_state = 0, .external_lex_state = 156},
  [4535] = {.lex_state = 0, .external_lex_state = 156},
  [4536] = {.lex_state = 0, .external_lex_state = 157},
  [4537] = {.lex_state = 0, .external_lex_state = 157},
  [4538] = {.lex_state = 0, .external_lex_state = 157},
  [4539] = {.lex_state = 0, .external_lex_state = 157},
  [4540] = {.lex_state = 0, .external_lex_state = 157},
  [4541] = {.lex_state = 0, .external_lex_state = 157},
  [4542] = {.lex_state = 0, .external_lex_state = 157},
  [4543] = {.lex_state = 0, .external_lex_state = 188},
  [4544] = {.lex_state = 0, .external_lex_state = 156},
  [4545] = {.lex_state = 0, .external_lex_state = 156},
  [4546] = {.lex_state = 0, .external_lex_state = 157},
  [4547] = {.lex_state = 0, .external_lex_state = 157},
  [4548] = {.lex_state = 0, .external_lex_state = 164},
  [4549] = {.lex_state = 0, .external_lex_state = 193},
  [4550] = {.lex_state = 0, .external_lex_state = 165},
  [4551] = {.lex_state = 0, .external_lex_state = 189},
  [4552] = {.lex_state = 0, .external_lex_state = 147},
  [4553] = {.lex_state = 0, .external_lex_state = 164},
  [4554] = {.lex_state = 0, .external_lex_state = 164},
  [4555] = {.lex_state = 0, .external_lex_state = 164},
  [4556] = {.lex_state = 0, .external_lex_state = 165},
  [4557] = {.lex_state = 0, .external_lex_state = 164},
  [4558] = {.lex_state = 0, .external_lex_state = 164},
  [4559] = {.lex_state = 0, .external_lex_state = 165},
  [4560] = {.lex_state = 0, .external_lex_state = 164},
  [4561] = {.lex_state = 0, .external_lex_state = 168},
  [4562] = {.lex_state = 0, .external_lex_state = 161},
  [4563] = {.lex_state = 0, .external_lex_state = 164},
  [4564] = {.lex_state = 0, .external_lex_state = 161},
  [4565] = {.lex_state = 0, .external_lex_state = 165},
  [4566] = {.lex_state = 0, .external_lex_state = 164},
  [4567] = {.lex_state = 0, .external_lex_state = 164},
  [4568] = {.lex_state = 0, .external_lex_state = 165},
  [4569] = {.lex_state = 0, .external_lex_state = 165},
  [4570] = {.lex_state = 0, .external_lex_state = 165},
  [4571] = {.lex_state = 0, .external_lex_state = 164},
  [4572] = {.lex_state = 0, .external_lex_state = 164},
  [4573] = {.lex_state = 0, .external_lex_state = 164},
  [4574] = {.lex_state = 0, .external_lex_state = 161},
  [4575] = {.lex_state = 0, .external_lex_state = 161},
  [4576] = {.lex_state = 0, .external_lex_state = 194},
  [4577] = {.lex_state = 0, .external_lex_state = 164},
  [4578] = {.lex_state = 0, .external_lex_state = 189},
  [4579] = {.lex_state = 0, .external_lex_state = 164},
  [4580] = {.lex_state = 0, .external_lex_state = 165},
  [4581] = {.lex_state = 0, .external_lex_state = 161},
  [4582] = {.lex_state = 0, .external_lex_state = 161},
  [4583] = {.lex_state = 0, .external_lex_state = 165},
  [4584] = {.lex_state = 0, .external_lex_state = 161},
  [4585] = {.lex_state = 0, .external_lex_state = 165},
  [4586] = {.lex_state = 0, .external_lex_state = 188},
  [4587] = {.lex_state = 0, .external_lex_state = 174},
  [4588] = {.lex_state = 0, .external_lex_state = 161},
  [4589] = {.lex_state = 0, .external_lex_state = 161},
  [4590] = {.lex_state = 0, .external_lex_state = 161},
  [4591] = {.lex_state = 0, .external_lex_state = 161},
  [4592] = {.lex_state = 0, .external_lex_state = 161},
  [4593] = {.lex_state = 0, .external_lex_state = 161},
  [4594] = {.lex_state = 0, .external_lex_state = 165},
  [4595] = {.lex_state = 0, .external_lex_state = 165},
  [4596] = {.lex_state = 0, .external_lex_state = 164},
  [4597] = {.lex_state = 0, .external_lex_state = 195},
  [4598] = {.lex_state = 0, .external_lex_state = 195},
  [4599] = {.lex_state = 0, .external_lex_state = 161},
  [4600] = {.lex_state = 0, .external_lex_state = 164},
  [4601] = {.lex_state = 0, .external_lex_state = 161},
  [4602] = {.lex_state = 0, .external_lex_state = 164},
  [4603] = {.lex_state = 0, .external_lex_state = 161},
  [4604] = {.lex_state = 0, .external_lex_state = 165},
  [4605] = {.lex_state = 0, .external_lex_state = 165},
  [4606] = {.lex_state = 0, .external_lex_state = 165},
  [4607] = {.lex_state = 0, .external_lex_state = 161},
  [4608] = {.lex_state = 0, .external_lex_state = 165},
  [4609] = {.lex_state = 0, .external_lex_state = 164},
  [4610] = {.lex_state = 0, .external_lex_state = 165},
  [4611] = {.lex_state = 0, .external_lex_state = 161},
  [4612] = {.lex_state = 0, .external_lex_state = 164},
  [4613] = {.lex_state = 0, .external_lex_state = 165},
  [4614] = {.lex_state = 0, .external_lex_state = 161},
  [4615] = {.lex_state = 0, .external_lex_state = 165},
  [4616] = {.lex_state = 2, .external_lex_state = 148},
  [4617] = {.lex_state = 0, .external_lex_state = 165},
  [4618] = {.lex_state = 0, .external_lex_state = 172},
  [4619] = {.lex_state = 0, .external_lex_state = 165},
  [4620] = {.lex_state = 0, .external_lex_state = 165},
  [4621] = {.lex_state = 0, .external_lex_state = 165},
  [4622] = {.lex_state = 0, .external_lex_state = 165},
  [4623] = {.lex_state = 0, .external_lex_state = 161},
  [4624] = {.lex_state = 0, .external_lex_state = 164},
  [4625] = {.lex_state = 0, .external_lex_state = 164},
  [4626] = {.lex_state = 0, .external_lex_state = 164},
  [4627] = {.lex_state = 0, .external_lex_state = 161},
  [4628] = {.lex_state = 0, .external_lex_state = 164},
  [4629] = {.lex_state = 0, .external_lex_state = 165},
  [4630] = {.lex_state = 0, .external_lex_state = 164},
  [4631] = {.lex_state = 0, .external_lex_state = 161},
  [4632] = {.lex_state = 0, .external_lex_state = 175},
  [4633] = {.lex_state = 0, .external_lex_state = 165},
  [4634] = {.lex_state = 0, .external_lex_state = 195},
  [4635] = {.lex_state = 0, .external_lex_state = 164},
  [4636] = {.lex_state = 0, .external_lex_state = 161},
  [4637] = {.lex_state = 0, .external_lex_state = 165},
  [4638] = {.lex_state = 0, .external_lex_state = 164},
  [4639] = {.lex_state = 0, .external_lex_state = 171},
  [4640] = {.lex_state = 0, .external_lex_state = 164},
  [4641] = {.lex_state = 0, .external_lex_state = 164},
  [4642] = {.lex_state = 0, .external_lex_state = 195},
  [4643] = {.lex_state = 0, .external_lex_state = 161},
  [4644] = {.lex_state = 0, .external_lex_state = 161},
  [4645] = {.lex_state = 0, .external_lex_state = 188},
  [4646] = {.lex_state = 0, .external_lex_state = 169},
  [4647] = {.lex_state = 0, .external_lex_state = 161},
  [4648] = {.lex_state = 0, .external_lex_state = 161},
  [4649] = {.lex_state = 0, .external_lex_state = 193},
  [4650] = {.lex_state = 0, .external_lex_state = 165},
  [4651] = {.lex_state = 0, .external_lex_state = 196},
  [4652] = {.lex_state = 0, .external_lex_state = 165},
  [4653] = {.lex_state = 0, .external_lex_state = 161},
  [4654] = {.lex_state = 0, .external_lex_state = 161},
  [4655] = {.lex_state = 0, .external_lex_state = 170},
  [4656] = {.lex_state = 0, .external_lex_state = 174},
  [4657] = {.lex_state = 0, .external_lex_state = 172},
  [4658] = {.lex_state = 0, .external_lex_state = 190},
  [4659] = {.lex_state = 0, .external_lex_state = 172},
  [4660] = {.lex_state = 0, .external_lex_state = 197},
  [4661] = {.lex_state = 0, .external_lex_state = 172},
  [4662] = {.lex_state = 0, .external_lex_state = 149},
  [4663] = {.lex_state = 0, .external_lex_state = 172},
  [4664] = {.lex_state = 0, .external_lex_state = 168},
  [4665] = {.lex_state = 0, .external_lex_state = 168},
  [4666] = {.lex_state = 0, .external_lex_state = 198},
  [4667] = {.lex_state = 0, .external_lex_state = 168},
  [4668] = {.lex_state = 0, .external_lex_state = 172},
  [4669] = {.lex_state = 0, .external_lex_state = 172},
  [4670] = {.lex_state = 0, .external_lex_state = 172},
  [4671] = {.lex_state = 0, .external_lex_state = 199},
  [4672] = {.lex_state = 0, .external_lex_state = 197},
  [4673] = {.lex_state = 0, .external_lex_state = 149},
  [4674] = {.lex_state = 0, .external_lex_state = 172},
  [4675] = {.lex_state = 0, .external_lex_state = 172},
  [4676] = {.lex_state = 0, .external_lex_state = 172},
  [4677] = {.lex_state = 0, .external_lex_state = 199},
  [4678] = {.lex_state = 0, .external_lex_state = 171},
  [4679] = {.lex_state = 0, .external_lex_state = 174},
  [4680] = {.lex_state = 0, .external_lex_state = 172},
  [4681] = {.lex_state = 0, .external_lex_state = 171},
  [4682] = {.lex_state = 0, .external_lex_state = 171},
  [4683] = {.lex_state = 0, .external_lex_state = 149},
  [4684] = {.lex_state = 0, .external_lex_state = 149},
  [4685] = {.lex_state = 0, .external_lex_state = 149},
  [4686] = {.lex_state = 0, .external_lex_state = 171},
  [4687] = {.lex_state = 0, .external_lex_state = 170},
  [4688] = {.lex_state = 0, .external_lex_state = 171},
  [4689] = {.lex_state = 0, .external_lex_state = 171},
  [4690] = {.lex_state = 0, .external_lex_state = 172},
  [4691] = {.lex_state = 0, .external_lex_state = 170},
  [4692] = {.lex_state = 0, .external_lex_state = 168},
  [4693] = {.lex_state = 0, .external_lex_state = 168},
  [4694] = {.lex_state = 0, .external_lex_state = 171},
  [4695] = {.lex_state = 0, .external_lex_state = 149},
  [4696] = {.lex_state = 0, .external_lex_state = 149},
  [4697] = {.lex_state = 0, .external_lex_state = 149},
  [4698] = {.lex_state = 0, .external_lex_state = 149},
  [4699] = {.lex_state = 0, .external_lex_state = 199},
  [4700] = {.lex_state = 0, .external_lex_state = 172},
  [4701] = {.lex_state = 0, .external_lex_state = 168},
  [4702] = {.lex_state = 0, .external_lex_state = 149},
  [4703] = {.lex_state = 0, .external_lex_state = 149},
  [4704] = {.lex_state = 0, .external_lex_state = 200},
  [4705] = {.lex_state = 0, .external_lex_state = 174},
  [4706] = {.lex_state = 0, .external_lex_state = 170},
  [4707] = {.lex_state = 0, .external_lex_state = 168},
  [4708] = {.lex_state = 0, .external_lex_state = 170},
  [4709] = {.lex_state = 0, .external_lex_state = 171},
  [4710] = {.lex_state = 0, .external_lex_state = 149},
  [4711] = {.lex_state = 0, .external_lex_state = 149},
  [4712] = {.lex_state = 0, .external_lex_state = 174},
  [4713] = {.lex_state = 0, .external_lex_state = 149},
  [4714] = {.lex_state = 0, .external_lex_state = 171},
  [4715] = {.lex_state = 0, .external_lex_state = 174},
  [4716] = {.lex_state = 0, .external_lex_state = 201},
  [4717] = {.lex_state = 0, .external_lex_state = 168},
  [4718] = {.lex_state = 0, .external_lex_state = 168},
  [4719] = {.lex_state = 0, .external_lex_state = 169},
  [4720] = {.lex_state = 0, .external_lex_state = 169},
  [4721] = {.lex_state = 0, .external_lex_state = 170},
  [4722] = {.lex_state = 0, .external_lex_state = 170},
  [4723] = {.lex_state = 0, .external_lex_state = 169},
  [4724] = {.lex_state = 0, .external_lex_state = 171},
  [4725] = {.lex_state = 0, .external_lex_state = 168},
  [4726] = {.lex_state = 0, .external_lex_state = 169},
  [4727] = {.lex_state = 0, .external_lex_state = 170},
  [4728] = {.lex_state = 0, .external_lex_state = 169},
  [4729] = {.lex_state = 0, .external_lex_state = 169},
  [4730] = {.lex_state = 0, .external_lex_state = 169},
  [4731] = {.lex_state = 0, .external_lex_state = 193},
  [4732] = {.lex_state = 0, .external_lex_state = 170},
  [4733] = {.lex_state = 0, .external_lex_state = 169},
  [4734] = {.lex_state = 0, .external_lex_state = 169},
  [4735] = {.lex_state = 0, .external_lex_state = 202},
  [4736] = {.lex_state = 0, .external_lex_state = 172},
  [4737] = {.lex_state = 0, .external_lex_state = 169},
  [4738] = {.lex_state = 0, .external_lex_state = 168},
  [4739] = {.lex_state = 0, .external_lex_state = 193},
  [4740] = {.lex_state = 0, .external_lex_state = 170},
  [4741] = {.lex_state = 0, .external_lex_state = 170},
  [4742] = {.lex_state = 0, .external_lex_state = 169},
  [4743] = {.lex_state = 0, .external_lex_state = 168},
  [4744] = {.lex_state = 0, .external_lex_state = 193},
  [4745] = {.lex_state = 0, .external_lex_state = 169},
  [4746] = {.lex_state = 0, .external_lex_state = 172},
  [4747] = {.lex_state = 0, .external_lex_state = 172},
  [4748] = {.lex_state = 0, .external_lex_state = 168},
  [4749] = {.lex_state = 0, .external_lex_state = 172},
  [4750] = {.lex_state = 0, .external_lex_state = 169},
  [4751] = {.lex_state = 0, .external_lex_state = 169},
  [4752] = {.lex_state = 0, .external_lex_state = 169},
  [4753] = {.lex_state = 0, .external_lex_state = 171},
  [4754] = {.lex_state = 0, .external_lex_state = 174},
  [4755] = {.lex_state = 0, .external_lex_state = 169},
  [4756] = {.lex_state = 0, .external_lex_state = 172},
  [4757] = {.lex_state = 0, .external_lex_state = 199},
  [4758] = {.lex_state = 0, .external_lex_state = 168},
  [4759] = {.lex_state = 0, .external_lex_state = 170},
  [4760] = {.lex_state = 0, .external_lex_state = 172},
  [4761] = {.lex_state = 0, .external_lex_state = 202},
  [4762] = {.lex_state = 0, .external_lex_state = 174},
  [4763] = {.lex_state = 0, .external_lex_state = 169},
  [4764] = {.lex_state = 0, .external_lex_state = 171},
  [4765] = {.lex_state = 0, .external_lex_state = 169},
  [4766] = {.lex_state = 0, .external_lex_state = 171},
  [4767] = {.lex_state = 2, .external_lex_state = 152},
  [4768] = {.lex_state = 0, .external_lex_state = 149},
  [4769] = {.lex_state = 0, .external_lex_state = 171},
  [4770] = {.lex_state = 0, .external_lex_state = 149},
  [4771] = {.lex_state = 0, .external_lex_state = 174},
  [4772] = {.lex_state = 0, .external_lex_state = 203},
  [4773] = {.lex_state = 0, .external_lex_state = 169},
  [4774] = {.lex_state = 0, .external_lex_state = 197},
  [4775] = {.lex_state = 0, .external_lex_state = 174},
  [4776] = {.lex_state = 0, .external_lex_state = 174},
  [4777] = {.lex_state = 0, .external_lex_state = 174},
  [4778] = {.lex_state = 0, .external_lex_state = 174},
  [4779] = {.lex_state = 0, .external_lex_state = 168},
  [4780] = {.lex_state = 0, .external_lex_state = 174},
  [4781] = {.lex_state = 0, .external_lex_state = 174},
  [4782] = {.lex_state = 0, .external_lex_state = 199},
  [4783] = {.lex_state = 0, .external_lex_state = 171},
  [4784] = {.lex_state = 0, .external_lex_state = 169},
  [4785] = {.lex_state = 0, .external_lex_state = 149},
  [4786] = {.lex_state = 0, .external_lex_state = 202},
  [4787] = {.lex_state = 0, .external_lex_state = 170},
  [4788] = {.lex_state = 0, .external_lex_state = 170},
  [4789] = {.lex_state = 0, .external_lex_state = 168},
  [4790] = {.lex_state = 0, .external_lex_state = 202},
  [4791] = {.lex_state = 0, .external_lex_state = 170},
  [4792] = {.lex_state = 0, .external_lex_state = 176},
  [4793] = {.lex_state = 0, .external_lex_state = 171},
  [4794] = {.lex_state = 0, .external_lex_state = 171},
  [4795] = {.lex_state = 0, .external_lex_state = 170},
  [4796] = {.lex_state = 0, .external_lex_state = 170},
  [4797] = {.lex_state = 0, .external_lex_state = 168},
  [4798] = {.lex_state = 0, .external_lex_state = 174},
  [4799] = {.lex_state = 0, .external_lex_state = 170},
  [4800] = {.lex_state = 0, .external_lex_state = 170},
  [4801] = {.lex_state = 0, .external_lex_state = 193},
  [4802] = {.lex_state = 0, .external_lex_state = 202},
  [4803] = {.lex_state = 0, .external_lex_state = 171},
  [4804] = {.lex_state = 0, .external_lex_state = 171},
  [4805] = {.lex_state = 0, .external_lex_state = 168},
  [4806] = {.lex_state = 0, .external_lex_state = 198},
  [4807] = {.lex_state = 0, .external_lex_state = 168},
  [4808] = {.lex_state = 2, .external_lex_state = 177},
  [4809] = {.lex_state = 0, .external_lex_state = 170},
  [4810] = {.lex_state = 0, .external_lex_state = 168},
  [4811] = {.lex_state = 0, .external_lex_state = 202},
  [4812] = {.lex_state = 0, .external_lex_state = 168},
  [4813] = {.lex_state = 0, .external_lex_state = 171},
  [4814] = {.lex_state = 0, .external_lex_state = 170},
  [4815] = {.lex_state = 0, .external_lex_state = 169},
  [4816] = {.lex_state = 0, .external_lex_state = 168},
  [4817] = {.lex_state = 0, .external_lex_state = 172},
  [4818] = {.lex_state = 0, .external_lex_state = 174},
  [4819] = {.lex_state = 0, .external_lex_state = 171},
  [4820] = {.lex_state = 0, .external_lex_state = 201},
  [4821] = {.lex_state = 0, .external_lex_state = 172},
  [4822] = {.lex_state = 0, .external_lex_state = 169},
  [4823] = {.lex_state = 0, .external_lex_state = 149},
  [4824] = {.lex_state = 0, .external_lex_state = 202},
  [4825] = {.lex_state = 0, .external_lex_state = 170},
  [4826] = {.lex_state = 0, .external_lex_state = 199},
  [4827] = {.lex_state = 0, .external_lex_state = 174},
  [4828] = {.lex_state = 0, .external_lex_state = 201},
  [4829] = {.lex_state = 0, .external_lex_state = 168},
  [4830] = {.lex_state = 0, .external_lex_state = 168},
  [4831] = {.lex_state = 0, .external_lex_state = 168},
  [4832] = {.lex_state = 0, .external_lex_state = 169},
  [4833] = {.lex_state = 0, .external_lex_state = 174},
  [4834] = {.lex_state = 0, .external_lex_state = 169},
  [4835] = {.lex_state = 0, .external_lex_state = 169},
  [4836] = {.lex_state = 0, .external_lex_state = 170},
  [4837] = {.lex_state = 0, .external_lex_state = 170},
  [4838] = {.lex_state = 0, .external_lex_state = 172},
  [4839] = {.lex_state = 0, .external_lex_state = 169},
  [4840] = {.lex_state = 0, .external_lex_state = 172},
  [4841] = {.lex_state = 0, .external_lex_state = 169},
  [4842] = {.lex_state = 0, .external_lex_state = 193},
  [4843] = {.lex_state = 0, .external_lex_state = 168},
  [4844] = {.lex_state = 0, .external_lex_state = 170},
  [4845] = {.lex_state = 0, .external_lex_state = 202},
  [4846] = {.lex_state = 0, .external_lex_state = 193},
  [4847] = {.lex_state = 0, .external_lex_state = 174},
  [4848] = {.lex_state = 0, .external_lex_state = 168},
  [4849] = {.lex_state = 0, .external_lex_state = 170},
  [4850] = {.lex_state = 0, .external_lex_state = 174},
  [4851] = {.lex_state = 0, .external_lex_state = 170},
  [4852] = {.lex_state = 0, .external_lex_state = 198},
  [4853] = {.lex_state = 0, .external_lex_state = 168},
  [4854] = {.lex_state = 0, .external_lex_state = 174},
  [4855] = {.lex_state = 0, .external_lex_state = 169},
  [4856] = {.lex_state = 0, .external_lex_state = 174},
  [4857] = {.lex_state = 0, .external_lex_state = 202},
  [4858] = {.lex_state = 0, .external_lex_state = 202},
  [4859] = {.lex_state = 0, .external_lex_state = 168},
  [4860] = {.lex_state = 0, .external_lex_state = 169},
  [4861] = {.lex_state = 0, .external_lex_state = 172},
  [4862] = {.lex_state = 0, .external_lex_state = 171},
  [4863] = {.lex_state = 0, .external_lex_state = 171},
  [4864] = {.lex_state = 0, .external_lex_state = 202},
  [4865] = {.lex_state = 0, .external_lex_state = 171},
  [4866] = {.lex_state = 0, .external_lex_state = 172},
  [4867] = {.lex_state = 0, .external_lex_state = 199},
  [4868] = {.lex_state = 0, .external_lex_state = 202},
  [4869] = {.lex_state = 0, .external_lex_state = 171},
  [4870] = {.lex_state = 0, .external_lex_state = 170},
  [4871] = {.lex_state = 0, .external_lex_state = 171},
  [4872] = {.lex_state = 0, .external_lex_state = 170},
  [4873] = {.lex_state = 0, .external_lex_state = 170},
  [4874] = {.lex_state = 0, .external_lex_state = 149},
  [4875] = {.lex_state = 0, .external_lex_state = 172},
  [4876] = {.lex_state = 0, .external_lex_state = 174},
  [4877] = {.lex_state = 0, .external_lex_state = 174},
  [4878] = {.lex_state = 0, .external_lex_state = 172},
  [4879] = {.lex_state = 2, .external_lex_state = 177},
  [4880] = {.lex_state = 0, .external_lex_state = 174},
  [4881] = {.lex_state = 0, .external_lex_state = 174},
  [4882] = {.lex_state = 0, .external_lex_state = 174},
  [4883] = {.lex_state = 0, .external_lex_state = 202},
  [4884] = {.lex_state = 0, .external_lex_state = 174},
  [4885] = {.lex_state = 0, .external_lex_state = 172},
  [4886] = {.lex_state = 0, .external_lex_state = 172},
  [4887] = {.lex_state = 0, .external_lex_state = 171},
  [4888] = {.lex_state = 0, .external_lex_state = 171},
  [4889] = {.lex_state = 0, .external_lex_state = 171},
  [4890] = {.lex_state = 0, .external_lex_state = 203},
  [4891] = {.lex_state = 0, .external_lex_state = 174},
  [4892] = {.lex_state = 0, .external_lex_state = 176},
  [4893] = {.lex_state = 0, .external_lex_state = 176},
  [4894] = {.lex_state = 0, .external_lex_state = 176},
  [4895] = {.lex_state = 0, .external_lex_state = 200},
  [4896] = {.lex_state = 0, .external_lex_state = 204},
  [4897] = {.lex_state = 0, .external_lex_state = 176},
  [4898] = {.lex_state = 0, .external_lex_state = 176},
  [4899] = {.lex_state = 0, .external_lex_state = 200},
  [4900] = {.lex_state = 0, .external_lex_state = 202},
  [4901] = {.lex_state = 0, .external_lex_state = 176},
  [4902] = {.lex_state = 0, .external_lex_state = 176},
  [4903] = {.lex_state = 0, .external_lex_state = 200},
  [4904] = {.lex_state = 0, .external_lex_state = 204},
  [4905] = {.lex_state = 0, .external_lex_state = 175},
  [4906] = {.lex_state = 0, .external_lex_state = 205},
  [4907] = {.lex_state = 0, .external_lex_state = 176},
  [4908] = {.lex_state = 0, .external_lex_state = 176},
  [4909] = {.lex_state = 0, .external_lex_state = 169},
  [4910] = {.lex_state = 2, .external_lex_state = 152},
  [4911] = {.lex_state = 0, .external_lex_state = 204},
  [4912] = {.lex_state = 0, .external_lex_state = 204},
  [4913] = {.lex_state = 0, .external_lex_state = 169},
  [4914] = {.lex_state = 0, .external_lex_state = 193},
  [4915] = {.lex_state = 0, .external_lex_state = 204},
  [4916] = {.lex_state = 0, .external_lex_state = 202},
  [4917] = {.lex_state = 0, .external_lex_state = 169},
  [4918] = {.lex_state = 0, .external_lex_state = 200},
  [4919] = {.lex_state = 0, .external_lex_state = 176},
  [4920] = {.lex_state = 0, .external_lex_state = 176},
  [4921] = {.lex_state = 0, .external_lex_state = 176},
  [4922] = {.lex_state = 0, .external_lex_state = 204},
  [4923] = {.lex_state = 0, .external_lex_state = 202},
  [4924] = {.lex_state = 0, .external_lex_state = 204},
  [4925] = {.lex_state = 0, .external_lex_state = 204},
  [4926] = {.lex_state = 0, .external_lex_state = 176},
  [4927] = {.lex_state = 0, .external_lex_state = 176},
  [4928] = {.lex_state = 0, .external_lex_state = 169},
  [4929] = {.lex_state = 0, .external_lex_state = 176},
  [4930] = {.lex_state = 0, .external_lex_state = 176},
  [4931] = {.lex_state = 0, .external_lex_state = 206},
  [4932] = {.lex_state = 0, .external_lex_state = 176},
  [4933] = {.lex_state = 0, .external_lex_state = 176},
  [4934] = {.lex_state = 0, .external_lex_state = 204},
  [4935] = {.lex_state = 0, .external_lex_state = 207},
  [4936] = {.lex_state = 0, .external_lex_state = 176},
  [4937] = {.lex_state = 0, .external_lex_state = 176},
  [4938] = {.lex_state = 0, .external_lex_state = 205},
  [4939] = {.lex_state = 0, .external_lex_state = 204},
  [4940] = {.lex_state = 0, .external_lex_state = 176},
  [4941] = {.lex_state = 0, .external_lex_state = 205},
  [4942] = {.lex_state = 0, .external_lex_state = 176},
  [4943] = {.lex_state = 0, .external_lex_state = 169},
  [4944] = {.lex_state = 0, .external_lex_state = 200},
  [4945] = {.lex_state = 0, .external_lex_state = 176},
  [4946] = {.lex_state = 0, .external_lex_state = 176},
  [4947] = {.lex_state = 0, .external_lex_state = 176},
  [4948] = {.lex_state = 0, .external_lex_state = 204},
  [4949] = {.lex_state = 0, .external_lex_state = 176},
  [4950] = {.lex_state = 0, .external_lex_state = 176},
  [4951] = {.lex_state = 0, .external_lex_state = 179},
  [4952] = {.lex_state = 0, .external_lex_state = 176},
  [4953] = {.lex_state = 0, .external_lex_state = 200},
  [4954] = {.lex_state = 0, .external_lex_state = 204},
  [4955] = {.lex_state = 0, .external_lex_state = 176},
  [4956] = {.lex_state = 0, .external_lex_state = 200},
  [4957] = {.lex_state = 0, .external_lex_state = 169},
  [4958] = {.lex_state = 0, .external_lex_state = 186},
  [4959] = {.lex_state = 0, .external_lex_state = 208},
  [4960] = {.lex_state = 0, .external_lex_state = 207},
  [4961] = {.lex_state = 0, .external_lex_state = 209},
  [4962] = {.lex_state = 0, .external_lex_state = 201},
  [4963] = {.lex_state = 0, .external_lex_state = 179},
  [4964] = {.lex_state = 0, .external_lex_state = 207},
  [4965] = {.lex_state = 0, .external_lex_state = 179},
  [4966] = {.lex_state = 0, .external_lex_state = 210},
  [4967] = {.lex_state = 0, .external_lex_state = 179},
  [4968] = {.lex_state = 0, .external_lex_state = 211},
  [4969] = {.lex_state = 0, .external_lex_state = 209},
  [4970] = {.lex_state = 0, .external_lex_state = 209},
  [4971] = {.lex_state = 0, .external_lex_state = 207},
  [4972] = {.lex_state = 0, .external_lex_state = 179},
  [4973] = {.lex_state = 0, .external_lex_state = 186},
  [4974] = {.lex_state = 0, .external_lex_state = 212},
  [4975] = {.lex_state = 0, .external_lex_state = 179},
  [4976] = {.lex_state = 3, .external_lex_state = 175},
  [4977] = {.lex_state = 0, .external_lex_state = 179},
  [4978] = {.lex_state = 0, .external_lex_state = 207},
  [4979] = {.lex_state = 0, .external_lex_state = 179},
  [4980] = {.lex_state = 0, .external_lex_state = 208},
  [4981] = {.lex_state = 0, .external_lex_state = 179},
  [4982] = {.lex_state = 0, .external_lex_state = 201},
  [4983] = {.lex_state = 0, .external_lex_state = 179},
  [4984] = {.lex_state = 0, .external_lex_state = 179},
  [4985] = {.lex_state = 0, .external_lex_state = 201},
  [4986] = {.lex_state = 0, .external_lex_state = 175},
  [4987] = {.lex_state = 0, .external_lex_state = 195},
  [4988] = {.lex_state = 0, .external_lex_state = 179},
  [4989] = {.lex_state = 0, .external_lex_state = 179},
  [4990] = {.lex_state = 0, .external_lex_state = 213},
  [4991] = {.lex_state = 0, .external_lex_state = 209},
  [4992] = {.lex_state = 0, .external_lex_state = 179},
  [4993] = {.lex_state = 0, .external_lex_state = 207},
  [4994] = {.lex_state = 0, .external_lex_state = 174},
  [4995] = {.lex_state = 0, .external_lex_state = 207},
  [4996] = {.lex_state = 0, .external_lex_state = 213},
  [4997] = {.lex_state = 0, .external_lex_state = 207},
  [4998] = {.lex_state = 0, .external_lex_state = 179},
  [4999] = {.lex_state = 0, .external_lex_state = 213},
  [5000] = {.lex_state = 0, .external_lex_state = 174},
  [5001] = {.lex_state = 0, .external_lex_state = 213},
  [5002] = {.lex_state = 0, .external_lex_state = 214},
  [5003] = {.lex_state = 0, .external_lex_state = 213},
  [5004] = {.lex_state = 0, .external_lex_state = 213},
  [5005] = {.lex_state = 0, .external_lex_state = 207},
  [5006] = {.lex_state = 0, .external_lex_state = 215},
  [5007] = {.lex_state = 0, .external_lex_state = 216},
  [5008] = {.lex_state = 0, .external_lex_state = 217},
  [5009] = {.lex_state = 0, .external_lex_state = 218},
  [5010] = {.lex_state = 0, .external_lex_state = 179},
  [5011] = {.lex_state = 0, .external_lex_state = 209},
  [5012] = {.lex_state = 0, .external_lex_state = 179},
  [5013] = {.lex_state = 0, .external_lex_state = 202},
  [5014] = {.lex_state = 0, .external_lex_state = 214},
  [5015] = {.lex_state = 0, .external_lex_state = 179},
  [5016] = {.lex_state = 0, .external_lex_state = 179},
  [5017] = {.lex_state = 0, .external_lex_state = 179},
  [5018] = {.lex_state = 0, .external_lex_state = 207},
  [5019] = {.lex_state = 0, .external_lex_state = 219},
  [5020] = {.lex_state = 0, .external_lex_state = 179},
  [5021] = {.lex_state = 0, .external_lex_state = 209},
  [5022] = {.lex_state = 0, .external_lex_state = 186},
  [5023] = {.lex_state = 0, .external_lex_state = 195},
  [5024] = {.lex_state = 0, .external_lex_state = 219},
  [5025] = {.lex_state = 0, .external_lex_state = 179},
  [5026] = {.lex_state = 0, .external_lex_state = 188},
  [5027] = {.lex_state = 0, .external_lex_state = 197},
  [5028] = {.lex_state = 0, .external_lex_state = 195},
  [5029] = {.lex_state = 0, .external_lex_state = 179},
  [5030] = {.lex_state = 0, .external_lex_state = 179},
  [5031] = {.lex_state = 0, .external_lex_state = 213},
  [5032] = {.lex_state = 0, .external_lex_state = 207},
  [5033] = {.lex_state = 0, .external_lex_state = 207},
  [5034] = {.lex_state = 0, .external_lex_state = 210},
  [5035] = {.lex_state = 0, .external_lex_state = 179},
  [5036] = {.lex_state = 0, .external_lex_state = 179},
  [5037] = {.lex_state = 0, .external_lex_state = 175},
  [5038] = {.lex_state = 0, .external_lex_state = 179},
  [5039] = {.lex_state = 0, .external_lex_state = 197},
  [5040] = {.lex_state = 0, .external_lex_state = 210},
  [5041] = {.lex_state = 0, .external_lex_state = 179},
  [5042] = {.lex_state = 0, .external_lex_state = 179},
  [5043] = {.lex_state = 0, .external_lex_state = 179},
  [5044] = {.lex_state = 0, .external_lex_state = 207},
  [5045] = {.lex_state = 0, .external_lex_state = 207},
  [5046] = {.lex_state = 0, .external_lex_state = 207},
  [5047] = {.lex_state = 0, .external_lex_state = 207},
  [5048] = {.lex_state = 0, .external_lex_state = 197},
  [5049] = {.lex_state = 0, .external_lex_state = 174},
  [5050] = {.lex_state = 0, .external_lex_state = 210},
  [5051] = {.lex_state = 0, .external_lex_state = 175},
  [5052] = {.lex_state = 0, .external_lex_state = 175},
  [5053] = {.lex_state = 0, .external_lex_state = 175},
  [5054] = {.lex_state = 0, .external_lex_state = 175},
  [5055] = {.lex_state = 0, .external_lex_state = 175},
  [5056] = {.lex_state = 0, .external_lex_state = 220},
  [5057] = {.lex_state = 0, .external_lex_state = 175},
  [5058] = {.lex_state = 0, .external_lex_state = 171},
  [5059] = {.lex_state = 0, .external_lex_state = 175},
  [5060] = {.lex_state = 0, .external_lex_state = 175},
  [5061] = {.lex_state = 0, .external_lex_state = 221},
  [5062] = {.lex_state = 0, .external_lex_state = 175},
  [5063] = {.lex_state = 0, .external_lex_state = 175},
  [5064] = {.lex_state = 0, .external_lex_state = 175},
  [5065] = {.lex_state = 0, .external_lex_state = 222},
  [5066] = {.lex_state = 0, .external_lex_state = 210},
  [5067] = {.lex_state = 0, .external_lex_state = 186},
  [5068] = {.lex_state = 0, .external_lex_state = 175},
  [5069] = {.lex_state = 0, .external_lex_state = 175},
  [5070] = {.lex_state = 0, .external_lex_state = 175},
  [5071] = {.lex_state = 0, .external_lex_state = 175},
  [5072] = {.lex_state = 0, .external_lex_state = 175},
  [5073] = {.lex_state = 0, .external_lex_state = 175},
  [5074] = {.lex_state = 0, .external_lex_state = 175},
  [5075] = {.lex_state = 0, .external_lex_state = 175},
  [5076] = {.lex_state = 0, .external_lex_state = 175},
  [5077] = {.lex_state = 0, .external_lex_state = 175},
  [5078] = {.lex_state = 0, .external_lex_state = 175},
  [5079] = {.lex_state = 0, .external_lex_state = 175},
  [5080] = {.lex_state = 0, .external_lex_state = 175},
  [5081] = {.lex_state = 0, .external_lex_state = 223},
  [5082] = {.lex_state = 0, .external_lex_state = 171},
  [5083] = {.lex_state = 0, .external_lex_state = 175},
  [5084] = {.lex_state = 0, .external_lex_state = 175},
  [5085] = {.lex_state = 0, .external_lex_state = 224},
  [5086] = {.lex_state = 0, .external_lex_state = 225},
  [5087] = {.lex_state = 0, .external_lex_state = 175},
  [5088] = {.lex_state = 0, .external_lex_state = 184},
  [5089] = {.lex_state = 0, .external_lex_state = 175},
  [5090] = {.lex_state = 0, .external_lex_state = 223},
  [5091] = {.lex_state = 0, .external_lex_state = 220},
  [5092] = {.lex_state = 0, .external_lex_state = 220},
  [5093] = {.lex_state = 0, .external_lex_state = 175},
  [5094] = {.lex_state = 0, .external_lex_state = 226},
  [5095] = {.lex_state = 0, .external_lex_state = 175},
  [5096] = {.lex_state = 0, .external_lex_state = 175},
  [5097] = {.lex_state = 0, .external_lex_state = 175},
  [5098] = {.lex_state = 0, .external_lex_state = 175},
  [5099] = {.lex_state = 0, .external_lex_state = 175},
  [5100] = {.lex_state = 0, .external_lex_state = 175},
  [5101] = {.lex_state = 0, .external_lex_state = 175},
  [5102] = {.lex_state = 0, .external_lex_state = 221},
  [5103] = {.lex_state = 0, .external_lex_state = 175},
  [5104] = {.lex_state = 0, .external_lex_state = 175},
  [5105] = {.lex_state = 0, .external_lex_state = 175},
  [5106] = {.lex_state = 0, .external_lex_state = 221},
  [5107] = {.lex_state = 0, .external_lex_state = 175},
  [5108] = {.lex_state = 0, .external_lex_state = 175},
  [5109] = {.lex_state = 0, .external_lex_state = 226},
  [5110] = {.lex_state = 0, .external_lex_state = 175},
  [5111] = {.lex_state = 0, .external_lex_state = 220},
  [5112] = {.lex_state = 0, .external_lex_state = 227},
  [5113] = {.lex_state = 0, .external_lex_state = 175},
  [5114] = {.lex_state = 0, .external_lex_state = 175},
  [5115] = {.lex_state = 0, .external_lex_state = 221},
  [5116] = {.lex_state = 0, .external_lex_state = 226},
  [5117] = {.lex_state = 0, .external_lex_state = 175},
  [5118] = {.lex_state = 0, .external_lex_state = 175},
  [5119] = {.lex_state = 0, .external_lex_state = 175},
  [5120] = {.lex_state = 0, .external_lex_state = 223},
  [5121] = {.lex_state = 0, .external_lex_state = 174},
  [5122] = {.lex_state = 0, .external_lex_state = 223},
  [5123] = {.lex_state = 0, .external_lex_state = 228},
  [5124] = {.lex_state = 0, .external_lex_state = 175},
  [5125] = {.lex_state = 0, .external_lex_state = 175},
  [5126] = {.lex_state = 0, .external_lex_state = 175},
  [5127] = {.lex_state = 0, .external_lex_state = 175},
  [5128] = {.lex_state = 0, .external_lex_state = 175},
  [5129] = {.lex_state = 0, .external_lex_state = 220},
  [5130] = {.lex_state = 0, .external_lex_state = 175},
  [5131] = {.lex_state = 0, .external_lex_state = 175},
  [5132] = {.lex_state = 0, .external_lex_state = 175},
  [5133] = {.lex_state = 0, .external_lex_state = 175},
  [5134] = {.lex_state = 0, .external_lex_state = 221},
  [5135] = {.lex_state = 0, .external_lex_state = 175},
  [5136] = {.lex_state = 0, .external_lex_state = 175},
  [5137] = {.lex_state = 0, .external_lex_state = 174},
  [5138] = {.lex_state = 0, .external_lex_state = 226},
  [5139] = {.lex_state = 0, .external_lex_state = 175},
  [5140] = {.lex_state = 0, .external_lex_state = 229},
  [5141] = {.lex_state = 0, .external_lex_state = 175},
  [5142] = {.lex_state = 0, .external_lex_state = 175},
  [5143] = {.lex_state = 0, .external_lex_state = 175},
  [5144] = {.lex_state = 0, .external_lex_state = 175},
  [5145] = {.lex_state = 0, .external_lex_state = 224},
  [5146] = {.lex_state = 0, .external_lex_state = 175},
  [5147] = {.lex_state = 0, .external_lex_state = 220},
  [5148] = {.lex_state = 0, .external_lex_state = 175},
  [5149] = {.lex_state = 0, .external_lex_state = 230},
  [5150] = {.lex_state = 0, .external_lex_state = 230},
  [5151] = {.lex_state = 0, .external_lex_state = 231},
  [5152] = {.lex_state = 0, .external_lex_state = 186},
  [5153] = {.lex_state = 0, .external_lex_state = 223},
  [5154] = {.lex_state = 0, .external_lex_state = 175},
  [5155] = {.lex_state = 0, .external_lex_state = 171},
  [5156] = {.lex_state = 0, .external_lex_state = 175},
  [5157] = {.lex_state = 0, .external_lex_state = 175},
  [5158] = {.lex_state = 0, .external_lex_state = 175},
  [5159] = {.lex_state = 0, .external_lex_state = 175},
  [5160] = {.lex_state = 0, .external_lex_state = 231},
  [5161] = {.lex_state = 0, .external_lex_state = 230},
  [5162] = {.lex_state = 0, .external_lex_state = 175},
  [5163] = {.lex_state = 0, .external_lex_state = 230},
  [5164] = {.lex_state = 0, .external_lex_state = 231},
  [5165] = {.lex_state = 0, .external_lex_state = 230},
  [5166] = {.lex_state = 0, .external_lex_state = 231},
  [5167] = {.lex_state = 0, .external_lex_state = 231},
  [5168] = {.lex_state = 0, .external_lex_state = 231},
  [5169] = {.lex_state = 0, .external_lex_state = 231},
  [5170] = {.lex_state = 0, .external_lex_state = 193},
  [5171] = {.lex_state = 0, .external_lex_state = 232},
  [5172] = {.lex_state = 0, .external_lex_state = 231},
  [5173] = {.lex_state = 0, .external_lex_state = 233},
  [5174] = {.lex_state = 0, .external_lex_state = 175},
  [5175] = {.lex_state = 0, .external_lex_state = 232},
  [5176] = {.lex_state = 0, .external_lex_state = 234},
  [5177] = {.lex_state = 0, .external_lex_state = 230},
  [5178] = {.lex_state = 0, .external_lex_state = 223},
  [5179] = {.lex_state = 0, .external_lex_state = 175},
  [5180] = {.lex_state = 0, .external_lex_state = 175},
  [5181] = {.lex_state = 0, .external_lex_state = 235},
  [5182] = {.lex_state = 0, .external_lex_state = 186},
  [5183] = {.lex_state = 0, .external_lex_state = 231},
  [5184] = {.lex_state = 0, .external_lex_state = 231},
  [5185] = {.lex_state = 0, .external_lex_state = 231},
  [5186] = {.lex_state = 0, .external_lex_state = 186},
  [5187] = {.lex_state = 0, .external_lex_state = 230},
  [5188] = {.lex_state = 0, .external_lex_state = 235},
  [5189] = {.lex_state = 0, .external_lex_state = 175},
  [5190] = {.lex_state = 0, .external_lex_state = 230},
  [5191] = {.lex_state = 0, .external_lex_state = 175},
  [5192] = {.lex_state = 0, .external_lex_state = 236},
  [5193] = {.lex_state = 0, .external_lex_state = 231},
  [5194] = {.lex_state = 0, .external_lex_state = 175},
  [5195] = {.lex_state = 0, .external_lex_state = 175},
  [5196] = {.lex_state = 0, .external_lex_state = 230},
  [5197] = {.lex_state = 0, .external_lex_state = 175},
  [5198] = {.lex_state = 0, .external_lex_state = 175},
  [5199] = {.lex_state = 0, .external_lex_state = 231},
  [5200] = {.lex_state = 0, .external_lex_state = 171},
  [5201] = {.lex_state = 0, .external_lex_state = 230},
  [5202] = {.lex_state = 0, .external_lex_state = 231},
  [5203] = {.lex_state = 0, .external_lex_state = 230},
  [5204] = {.lex_state = 0, .external_lex_state = 237},
  [5205] = {.lex_state = 0, .external_lex_state = 230},
  [5206] = {.lex_state = 0, .external_lex_state = 231},
  [5207] = {.lex_state = 0, .external_lex_state = 223},
  [5208] = {.lex_state = 0, .external_lex_state = 231},
  [5209] = {.lex_state = 0, .external_lex_state = 231},
  [5210] = {.lex_state = 0, .external_lex_state = 230},
  [5211] = {.lex_state = 0, .external_lex_state = 175},
  [5212] = {.lex_state = 0, .external_lex_state = 223},
  [5213] = {.lex_state = 0, .external_lex_state = 230},
  [5214] = {.lex_state = 0, .external_lex_state = 230},
  [5215] = {.lex_state = 0, .external_lex_state = 175},
  [5216] = {.lex_state = 0, .external_lex_state = 223},
  [5217] = {.lex_state = 0, .external_lex_state = 186},
  [5218] = {.lex_state = 0, .external_lex_state = 238},
  [5219] = {.lex_state = 0, .external_lex_state = 147},
  [5220] = {.lex_state = 0, .external_lex_state = 231},
  [5221] = {.lex_state = 0, .external_lex_state = 228},
  [5222] = {.lex_state = 0, .external_lex_state = 175},
  [5223] = {.lex_state = 0, .external_lex_state = 230},
  [5224] = {.lex_state = 0, .external_lex_state = 230},
  [5225] = {.lex_state = 0, .external_lex_state = 193},
  [5226] = {.lex_state = 0, .external_lex_state = 223},
  [5227] = {.lex_state = 0, .external_lex_state = 230},
  [5228] = {.lex_state = 0, .external_lex_state = 186},
  [5229] = {.lex_state = 0, .external_lex_state = 236},
  [5230] = {.lex_state = 0, .external_lex_state = 193},
  [5231] = {.lex_state = 0, .external_lex_state = 232},
  [5232] = {.lex_state = 0, .external_lex_state = 230},
  [5233] = {.lex_state = 0, .external_lex_state = 195},
  [5234] = {.lex_state = 0, .external_lex_state = 239},
  [5235] = {.lex_state = 0, .external_lex_state = 239},
  [5236] = {.lex_state = 0, .external_lex_state = 235},
  [5237] = {.lex_state = 0, .external_lex_state = 228},
  [5238] = {.lex_state = 0, .external_lex_state = 222},
  [5239] = {.lex_state = 0, .external_lex_state = 175},
  [5240] = {.lex_state = 0, .external_lex_state = 231},
  [5241] = {.lex_state = 0, .external_lex_state = 230},
  [5242] = {.lex_state = 0, .external_lex_state = 175},
  [5243] = {.lex_state = 0, .external_lex_state = 195},
  [5244] = {.lex_state = 0, .external_lex_state = 175},
  [5245] = {.lex_state = 0, .external_lex_state = 175},
  [5246] = {.lex_state = 0, .external_lex_state = 147},
  [5247] = {.lex_state = 0, .external_lex_state = 232},
  [5248] = {.lex_state = 0, .external_lex_state = 175},
  [5249] = {.lex_state = 0, .external_lex_state = 186},
  [5250] = {.lex_state = 0, .external_lex_state = 230},
  [5251] = {.lex_state = 0, .external_lex_state = 175},
  [5252] = {.lex_state = 0, .external_lex_state = 231},
  [5253] = {.lex_state = 0, .external_lex_state = 231},
  [5254] = {.lex_state = 0, .external_lex_state = 232},
  [5255] = {.lex_state = 0, .external_lex_state = 230},
  [5256] = {.lex_state = 0, .external_lex_state = 174},
  [5257] = {.lex_state = 0, .external_lex_state = 223},
  [5258] = {.lex_state = 0, .external_lex_state = 230},
  [5259] = {.lex_state = 0, .external_lex_state = 230},
  [5260] = {.lex_state = 0, .external_lex_state = 231},
  [5261] = {.lex_state = 0, .external_lex_state = 230},
  [5262] = {.lex_state = 0, .external_lex_state = 193},
  [5263] = {.lex_state = 0, .external_lex_state = 175},
  [5264] = {.lex_state = 0, .external_lex_state = 230},
  [5265] = {.lex_state = 0, .external_lex_state = 230},
  [5266] = {.lex_state = 0, .external_lex_state = 213},
  [5267] = {.lex_state = 0, .external_lex_state = 213},
  [5268] = {.lex_state = 0, .external_lex_state = 175},
  [5269] = {.lex_state = 0, .external_lex_state = 230},
  [5270] = {.lex_state = 0, .external_lex_state = 231},
  [5271] = {.lex_state = 0, .external_lex_state = 228},
  [5272] = {.lex_state = 0, .external_lex_state = 175},
  [5273] = {.lex_state = 0, .external_lex_state = 230},
  [5274] = {.lex_state = 0, .external_lex_state = 230},
  [5275] = {.lex_state = 0, .external_lex_state = 231},
  [5276] = {.lex_state = 0, .external_lex_state = 175},
  [5277] = {.lex_state = 0, .external_lex_state = 230},
  [5278] = {.lex_state = 0, .external_lex_state = 175},
  [5279] = {.lex_state = 0, .external_lex_state = 175},
  [5280] = {.lex_state = 0, .external_lex_state = 175},
  [5281] = {.lex_state = 0, .external_lex_state = 230},
  [5282] = {.lex_state = 0, .external_lex_state = 232},
  [5283] = {.lex_state = 0, .external_lex_state = 230},
  [5284] = {.lex_state = 0, .external_lex_state = 230},
  [5285] = {.lex_state = 0, .external_lex_state = 230},
  [5286] = {.lex_state = 0, .external_lex_state = 175},
  [5287] = {.lex_state = 0, .external_lex_state = 193},
  [5288] = {.lex_state = 0, .external_lex_state = 239},
  [5289] = {.lex_state = 0, .external_lex_state = 230},
  [5290] = {.lex_state = 0, .external_lex_state = 231},
  [5291] = {.lex_state = 0, .external_lex_state = 231},
  [5292] = {.lex_state = 0, .external_lex_state = 232},
  [5293] = {.lex_state = 0, .external_lex_state = 230},
  [5294] = {.lex_state = 0, .external_lex_state = 175},
  [5295] = {.lex_state = 0, .external_lex_state = 230},
  [5296] = {.lex_state = 0, .external_lex_state = 230},
  [5297] = {.lex_state = 0, .external_lex_state = 230},
  [5298] = {.lex_state = 0, .external_lex_state = 193},
  [5299] = {.lex_state = 0, .external_lex_state = 175},
  [5300] = {.lex_state = 0, .external_lex_state = 231},
  [5301] = {.lex_state = 0, .external_lex_state = 230},
  [5302] = {.lex_state = 0, .external_lex_state = 175},
  [5303] = {.lex_state = 0, .external_lex_state = 232},
  [5304] = {.lex_state = 0, .external_lex_state = 230},
  [5305] = {.lex_state = 0, .external_lex_state = 230},
  [5306] = {.lex_state = 0, .external_lex_state = 231},
  [5307] = {.lex_state = 0, .external_lex_state = 235},
  [5308] = {.lex_state = 0, .external_lex_state = 230},
  [5309] = {.lex_state = 0, .external_lex_state = 232},
  [5310] = {.lex_state = 0, .external_lex_state = 230},
  [5311] = {.lex_state = 0, .external_lex_state = 231},
  [5312] = {.lex_state = 0, .external_lex_state = 230},
  [5313] = {.lex_state = 0, .external_lex_state = 230},
  [5314] = {.lex_state = 0, .external_lex_state = 175},
  [5315] = {.lex_state = 0, .external_lex_state = 230},
  [5316] = {.lex_state = 0, .external_lex_state = 230},
  [5317] = {.lex_state = 0, .external_lex_state = 175},
  [5318] = {.lex_state = 0, .external_lex_state = 230},
  [5319] = {.lex_state = 0, .external_lex_state = 193},
  [5320] = {.lex_state = 0, .external_lex_state = 175},
  [5321] = {.lex_state = 0, .external_lex_state = 175},
  [5322] = {.lex_state = 0, .external_lex_state = 175},
  [5323] = {.lex_state = 0, .external_lex_state = 193},
  [5324] = {.lex_state = 0, .external_lex_state = 175},
  [5325] = {.lex_state = 0, .external_lex_state = 175},
  [5326] = {.lex_state = 0, .external_lex_state = 175},
  [5327] = {.lex_state = 0, .external_lex_state = 175},
  [5328] = {.lex_state = 0, .external_lex_state = 175},
  [5329] = {.lex_state = 0, .external_lex_state = 175},
  [5330] = {.lex_state = 0, .external_lex_state = 230},
  [5331] = {.lex_state = 0, .external_lex_state = 224},
  [5332] = {.lex_state = 0, .external_lex_state = 240},
  [5333] = {.lex_state = 0, .external_lex_state = 241},
  [5334] = {.lex_state = 0, .external_lex_state = 240},
  [5335] = {.lex_state = 0, .external_lex_state = 240},
  [5336] = {.lex_state = 0, .external_lex_state = 241},
  [5337] = {.lex_state = 0, .external_lex_state = 242},
  [5338] = {.lex_state = 0, .external_lex_state = 240},
  [5339] = {.lex_state = 0, .external_lex_state = 240},
  [5340] = {.lex_state = 0, .external_lex_state = 186},
  [5341] = {.lex_state = 0, .external_lex_state = 241},
  [5342] = {.lex_state = 0, .external_lex_state = 236},
  [5343] = {.lex_state = 0, .external_lex_state = 240},
  [5344] = {.lex_state = 0, .external_lex_state = 240},
  [5345] = {.lex_state = 0, .external_lex_state = 240},
  [5346] = {.lex_state = 0, .external_lex_state = 175},
  [5347] = {.lex_state = 0, .external_lex_state = 240},
  [5348] = {.lex_state = 0, .external_lex_state = 240},
  [5349] = {.lex_state = 0, .external_lex_state = 243},
  [5350] = {.lex_state = 0, .external_lex_state = 243},
  [5351] = {.lex_state = 0, .external_lex_state = 240},
  [5352] = {.lex_state = 0, .external_lex_state = 240},
  [5353] = {.lex_state = 0, .external_lex_state = 240},
  [5354] = {.lex_state = 0, .external_lex_state = 240},
  [5355] = {.lex_state = 0, .external_lex_state = 228},
  [5356] = {.lex_state = 0, .external_lex_state = 240},
  [5357] = {.lex_state = 0, .external_lex_state = 240},
  [5358] = {.lex_state = 0, .external_lex_state = 228},
  [5359] = {.lex_state = 0, .external_lex_state = 186},
  [5360] = {.lex_state = 0, .external_lex_state = 228},
  [5361] = {.lex_state = 0, .external_lex_state = 240},
  [5362] = {.lex_state = 0, .external_lex_state = 240},
  [5363] = {.lex_state = 0, .external_lex_state = 186},
  [5364] = {.lex_state = 0, .external_lex_state = 186},
  [5365] = {.lex_state = 0, .external_lex_state = 240},
  [5366] = {.lex_state = 0, .external_lex_state = 240},
  [5367] = {.lex_state = 0, .external_lex_state = 186},
  [5368] = {.lex_state = 0, .external_lex_state = 240},
  [5369] = {.lex_state = 0, .external_lex_state = 240},
  [5370] = {.lex_state = 0, .external_lex_state = 198},
  [5371] = {.lex_state = 0, .external_lex_state = 240},
  [5372] = {.lex_state = 0, .external_lex_state = 240},
  [5373] = {.lex_state = 0, .external_lex_state = 240},
  [5374] = {.lex_state = 0, .external_lex_state = 240},
  [5375] = {.lex_state = 0, .external_lex_state = 240},
  [5376] = {.lex_state = 0, .external_lex_state = 242},
  [5377] = {.lex_state = 0, .external_lex_state = 224},
  [5378] = {.lex_state = 0, .external_lex_state = 240},
  [5379] = {.lex_state = 0, .external_lex_state = 240},
  [5380] = {.lex_state = 0, .external_lex_state = 240},
  [5381] = {.lex_state = 0, .external_lex_state = 240},
  [5382] = {.lex_state = 0, .external_lex_state = 240},
  [5383] = {.lex_state = 0, .external_lex_state = 226},
  [5384] = {.lex_state = 0, .external_lex_state = 240},
  [5385] = {.lex_state = 0, .external_lex_state = 240},
  [5386] = {.lex_state = 0, .external_lex_state = 224},
  [5387] = {.lex_state = 0, .external_lex_state = 240},
  [5388] = {.lex_state = 0, .external_lex_state = 240},
  [5389] = {.lex_state = 0, .external_lex_state = 186},
  [5390] = {.lex_state = 0, .external_lex_state = 240},
  [5391] = {.lex_state = 0, .external_lex_state = 240},
  [5392] = {.lex_state = 0, .external_lex_state = 226},
  [5393] = {.lex_state = 0, .external_lex_state = 240},
  [5394] = {.lex_state = 0, .external_lex_state = 240},
  [5395] = {.lex_state = 0, .external_lex_state = 240},
  [5396] = {.lex_state = 0, .external_lex_state = 240},
  [5397] = {.lex_state = 0, .external_lex_state = 224},
  [5398] = {.lex_state = 0, .external_lex_state = 186},
  [5399] = {.lex_state = 0, .external_lex_state = 240},
  [5400] = {.lex_state = 0, .external_lex_state = 240},
  [5401] = {.lex_state = 0, .external_lex_state = 240},
  [5402] = {.lex_state = 0, .external_lex_state = 240},
  [5403] = {.lex_state = 0, .external_lex_state = 240},
  [5404] = {.lex_state = 0, .external_lex_state = 240},
  [5405] = {.lex_state = 0, .external_lex_state = 205},
  [5406] = {.lex_state = 0, .external_lex_state = 240},
  [5407] = {.lex_state = 0, .external_lex_state = 240},
  [5408] = {.lex_state = 0, .external_lex_state = 240},
  [5409] = {.lex_state = 0, .external_lex_state = 240},
  [5410] = {.lex_state = 0, .external_lex_state = 240},
  [5411] = {.lex_state = 0, .external_lex_state = 240},
  [5412] = {.lex_state = 0, .external_lex_state = 240},
  [5413] = {.lex_state = 0, .external_lex_state = 240},
  [5414] = {.lex_state = 0, .external_lex_state = 240},
  [5415] = {.lex_state = 0, .external_lex_state = 240},
  [5416] = {.lex_state = 0, .external_lex_state = 186},
  [5417] = {.lex_state = 0, .external_lex_state = 186},
  [5418] = {.lex_state = 0, .external_lex_state = 240},
  [5419] = {.lex_state = 0, .external_lex_state = 240},
  [5420] = {.lex_state = 0, .external_lex_state = 186},
  [5421] = {.lex_state = 0, .external_lex_state = 240},
  [5422] = {.lex_state = 0, .external_lex_state = 240},
  [5423] = {.lex_state = 0, .external_lex_state = 186},
  [5424] = {.lex_state = 0, .external_lex_state = 240},
  [5425] = {.lex_state = 0, .external_lex_state = 240},
  [5426] = {.lex_state = 0, .external_lex_state = 223},
  [5427] = {.lex_state = 0, .external_lex_state = 240},
  [5428] = {.lex_state = 0, .external_lex_state = 240},
  [5429] = {.lex_state = 0, .external_lex_state = 240},
  [5430] = {.lex_state = 0, .external_lex_state = 240},
  [5431] = {.lex_state = 0, .external_lex_state = 240},
  [5432] = {.lex_state = 0, .external_lex_state = 240},
  [5433] = {.lex_state = 0, .external_lex_state = 240},
  [5434] = {.lex_state = 0, .external_lex_state = 240},
  [5435] = {.lex_state = 0, .external_lex_state = 240},
  [5436] = {.lex_state = 0, .external_lex_state = 240},
  [5437] = {.lex_state = 0, .external_lex_state = 240},
  [5438] = {.lex_state = 0, .external_lex_state = 240},
  [5439] = {.lex_state = 0, .external_lex_state = 240},
  [5440] = {.lex_state = 0, .external_lex_state = 240},
  [5441] = {.lex_state = 0, .external_lex_state = 186},
  [5442] = {.lex_state = 0, .external_lex_state = 186},
  [5443] = {.lex_state = 0, .external_lex_state = 240},
  [5444] = {.lex_state = 0, .external_lex_state = 240},
  [5445] = {.lex_state = 0, .external_lex_state = 240},
  [5446] = {.lex_state = 0, .external_lex_state = 240},
  [5447] = {.lex_state = 0, .external_lex_state = 226},
  [5448] = {.lex_state = 0, .external_lex_state = 240},
  [5449] = {.lex_state = 0, .external_lex_state = 240},
  [5450] = {.lex_state = 0, .external_lex_state = 240},
  [5451] = {.lex_state = 0, .external_lex_state = 240},
  [5452] = {.lex_state = 0, .external_lex_state = 240},
  [5453] = {.lex_state = 0, .external_lex_state = 240},
  [5454] = {.lex_state = 0, .external_lex_state = 240},
  [5455] = {.lex_state = 0, .external_lex_state = 224},
  [5456] = {.lex_state = 0, .external_lex_state = 240},
  [5457] = {.lex_state = 0, .external_lex_state = 240},
  [5458] = {.lex_state = 0, .external_lex_state = 240},
  [5459] = {.lex_state = 0, .external_lex_state = 240},
  [5460] = {.lex_state = 0, .external_lex_state = 240},
  [5461] = {.lex_state = 0, .external_lex_state = 240},
  [5462] = {.lex_state = 0, .external_lex_state = 240},
  [5463] = {.lex_state = 0, .external_lex_state = 240},
  [5464] = {.lex_state = 0, .external_lex_state = 240},
  [5465] = {.lex_state = 0, .external_lex_state = 240},
  [5466] = {.lex_state = 0, .external_lex_state = 240},
  [5467] = {.lex_state = 0, .external_lex_state = 240},
  [5468] = {.lex_state = 0, .external_lex_state = 186},
  [5469] = {.lex_state = 0, .external_lex_state = 186},
  [5470] = {.lex_state = 0, .external_lex_state = 186},
  [5471] = {.lex_state = 0, .external_lex_state = 186},
  [5472] = {.lex_state = 0, .external_lex_state = 186},
  [5473] = {.lex_state = 0, .external_lex_state = 244},
  [5474] = {.lex_state = 0, .external_lex_state = 175},
  [5475] = {.lex_state = 0, .external_lex_state = 223},
  [5476] = {.lex_state = 0, .external_lex_state = 186},
  [5477] = {.lex_state = 0, .external_lex_state = 239},
  [5478] = {.lex_state = 0, .external_lex_state = 198},
  [5479] = {.lex_state = 0, .external_lex_state = 239},
  [5480] = {.lex_state = 0, .external_lex_state = 198},
  [5481] = {.lex_state = 0, .external_lex_state = 245},
  [5482] = {.lex_state = 0, .external_lex_state = 186},
  [5483] = {.lex_state = 0, .external_lex_state = 186},
  [5484] = {.lex_state = 0, .external_lex_state = 198},
  [5485] = {.lex_state = 0, .external_lex_state = 198},
  [5486] = {.lex_state = 0, .external_lex_state = 246},
  [5487] = {.lex_state = 0, .external_lex_state = 224},
  [5488] = {.lex_state = 0, .external_lex_state = 247},
  [5489] = {.lex_state = 0, .external_lex_state = 239},
  [5490] = {.lex_state = 0, .external_lex_state = 186},
  [5491] = {.lex_state = 0, .external_lex_state = 186},
  [5492] = {.lex_state = 0, .external_lex_state = 248},
  [5493] = {.lex_state = 0, .external_lex_state = 232},
  [5494] = {.lex_state = 0, .external_lex_state = 186},
  [5495] = {.lex_state = 0, .external_lex_state = 190},
  [5496] = {.lex_state = 0, .external_lex_state = 244},
  [5497] = {.lex_state = 0, .external_lex_state = 226},
  [5498] = {.lex_state = 0, .external_lex_state = 241},
  [5499] = {.lex_state = 0, .external_lex_state = 190},
  [5500] = {.lex_state = 0, .external_lex_state = 214},
  [5501] = {.lex_state = 0, .external_lex_state = 239},
  [5502] = {.lex_state = 0, .external_lex_state = 245},
  [5503] = {.lex_state = 0, .external_lex_state = 240},
  [5504] = {.lex_state = 0, .external_lex_state = 220},
  [5505] = {.lex_state = 0, .external_lex_state = 242},
  [5506] = {.lex_state = 0, .external_lex_state = 228},
  [5507] = {.lex_state = 0, .external_lex_state = 190},
  [5508] = {.lex_state = 0, .external_lex_state = 224},
  [5509] = {.lex_state = 0, .external_lex_state = 236},
  [5510] = {.lex_state = 0, .external_lex_state = 223},
  [5511] = {.lex_state = 0, .external_lex_state = 245},
  [5512] = {.lex_state = 0, .external_lex_state = 220},
  [5513] = {.lex_state = 0, .external_lex_state = 239},
  [5514] = {.lex_state = 0, .external_lex_state = 235},
  [5515] = {.lex_state = 0, .external_lex_state = 228},
  [5516] = {.lex_state = 0, .external_lex_state = 239},
  [5517] = {.lex_state = 0, .external_lex_state = 239},
  [5518] = {.lex_state = 0, .external_lex_state = 239},
  [5519] = {.lex_state = 0, .external_lex_state = 190},
  [5520] = {.lex_state = 0, .external_lex_state = 190},
  [5521] = {.lex_state = 0, .external_lex_state = 243},
  [5522] = {.lex_state = 0, .external_lex_state = 239},
  [5523] = {.lex_state = 0, .external_lex_state = 239},
  [5524] = {.lex_state = 0, .external_lex_state = 186},
  [5525] = {.lex_state = 0, .external_lex_state = 239},
  [5526] = {.lex_state = 0, .external_lex_state = 239},
  [5527] = {.lex_state = 0, .external_lex_state = 239},
  [5528] = {.lex_state = 0, .external_lex_state = 239},
  [5529] = {.lex_state = 0, .external_lex_state = 239},
  [5530] = {.lex_state = 0, .external_lex_state = 239},
  [5531] = {.lex_state = 0, .external_lex_state = 239},
  [5532] = {.lex_state = 0, .external_lex_state = 239},
  [5533] = {.lex_state = 0, .external_lex_state = 239},
  [5534] = {.lex_state = 0, .external_lex_state = 245},
  [5535] = {.lex_state = 0, .external_lex_state = 171},
  [5536] = {.lex_state = 0, .external_lex_state = 224},
  [5537] = {.lex_state = 0, .external_lex_state = 224},
  [5538] = {.lex_state = 0, .external_lex_state = 186},
  [5539] = {.lex_state = 0, .external_lex_state = 245},
  [5540] = {.lex_state = 0, .external_lex_state = 240},
  [5541] = {.lex_state = 0, .external_lex_state = 241},
  [5542] = {.lex_state = 0, .external_lex_state = 245},
  [5543] = {.lex_state = 0, .external_lex_state = 240},
  [5544] = {.lex_state = 0, .external_lex_state = 248},
  [5545] = {.lex_state = 0, .external_lex_state = 248},
  [5546] = {.lex_state = 0, .external_lex_state = 224},
  [5547] = {.lex_state = 0, .external_lex_state = 224},
  [5548] = {.lex_state = 0, .external_lex_state = 248},
  [5549] = {.lex_state = 0, .external_lex_state = 224},
  [5550] = {.lex_state = 0, .external_lex_state = 240},
  [5551] = {.lex_state = 0, .external_lex_state = 241},
  [5552] = {.lex_state = 0, .external_lex_state = 240},
  [5553] = {.lex_state = 0, .external_lex_state = 248},
  [5554] = {.lex_state = 0, .external_lex_state = 248},
  [5555] = {.lex_state = 0, .external_lex_state = 240},
  [5556] = {.lex_state = 0, .external_lex_state = 220},
  [5557] = {.lex_state = 0, .external_lex_state = 186},
  [5558] = {.lex_state = 0, .external_lex_state = 222},
  [5559] = {.lex_state = 0, .external_lex_state = 186},
  [5560] = {.lex_state = 0, .external_lex_state = 186},
  [5561] = {.lex_state = 0, .external_lex_state = 224},
  [5562] = {.lex_state = 0, .external_lex_state = 186},
  [5563] = {.lex_state = 0, .external_lex_state = 186},
  [5564] = {.lex_state = 0, .external_lex_state = 186},
  [5565] = {.lex_state = 0, .external_lex_state = 186},
  [5566] = {.lex_state = 0, .external_lex_state = 186},
  [5567] = {.lex_state = 0, .external_lex_state = 245},
  [5568] = {.lex_state = 0, .external_lex_state = 186},
  [5569] = {.lex_state = 0, .external_lex_state = 245},
  [5570] = {.lex_state = 0, .external_lex_state = 186},
  [5571] = {.lex_state = 0, .external_lex_state = 186},
  [5572] = {.lex_state = 0, .external_lex_state = 195},
  [5573] = {.lex_state = 0, .external_lex_state = 186},
  [5574] = {.lex_state = 0, .external_lex_state = 224},
  [5575] = {.lex_state = 0, .external_lex_state = 195},
  [5576] = {.lex_state = 0, .external_lex_state = 186},
  [5577] = {.lex_state = 0, .external_lex_state = 186},
  [5578] = {.lex_state = 0, .external_lex_state = 224},
  [5579] = {.lex_state = 0, .external_lex_state = 246},
  [5580] = {.lex_state = 0, .external_lex_state = 186},
  [5581] = {.lex_state = 0, .external_lex_state = 186},
  [5582] = {.lex_state = 0, .external_lex_state = 186},
  [5583] = {.lex_state = 0, .external_lex_state = 186},
  [5584] = {.lex_state = 0, .external_lex_state = 186},
  [5585] = {.lex_state = 0, .external_lex_state = 186},
  [5586] = {.lex_state = 0, .external_lex_state = 186},
  [5587] = {.lex_state = 0, .external_lex_state = 186},
  [5588] = {.lex_state = 0, .external_lex_state = 231},
  [5589] = {.lex_state = 0, .external_lex_state = 186},
  [5590] = {.lex_state = 0, .external_lex_state = 186},
  [5591] = {.lex_state = 0, .external_lex_state = 186},
  [5592] = {.lex_state = 0, .external_lex_state = 249},
  [5593] = {.lex_state = 0, .external_lex_state = 186},
  [5594] = {.lex_state = 0, .external_lex_state = 186},
  [5595] = {.lex_state = 0, .external_lex_state = 186},
  [5596] = {.lex_state = 0, .external_lex_state = 231},
  [5597] = {.lex_state = 0, .external_lex_state = 186},
  [5598] = {.lex_state = 0, .external_lex_state = 186},
  [5599] = {.lex_state = 0, .external_lex_state = 186},
  [5600] = {.lex_state = 0, .external_lex_state = 186},
  [5601] = {.lex_state = 0, .external_lex_state = 186},
  [5602] = {.lex_state = 0, .external_lex_state = 231},
  [5603] = {.lex_state = 0, .external_lex_state = 186},
  [5604] = {.lex_state = 0, .external_lex_state = 186},
  [5605] = {.lex_state = 0, .external_lex_state = 186},
  [5606] = {.lex_state = 0, .external_lex_state = 186},
  [5607] = {.lex_state = 0, .external_lex_state = 186},
  [5608] = {.lex_state = 0, .external_lex_state = 186},
  [5609] = {.lex_state = 0, .external_lex_state = 186},
  [5610] = {.lex_state = 0, .external_lex_state = 186},
  [5611] = {.lex_state = 0, .external_lex_state = 231},
  [5612] = {.lex_state = 0, .external_lex_state = 186},
  [5613] = {.lex_state = 0, .external_lex_state = 186},
  [5614] = {.lex_state = 0, .external_lex_state = 186},
  [5615] = {.lex_state = 0, .external_lex_state = 186},
  [5616] = {.lex_state = 0, .external_lex_state = 186},
  [5617] = {.lex_state = 0, .external_lex_state = 186},
  [5618] = {.lex_state = 0, .external_lex_state = 250},
  [5619] = {.lex_state = 0, .external_lex_state = 186},
  [5620] = {.lex_state = 0, .external_lex_state = 186},
  [5621] = {.lex_state = 0, .external_lex_state = 186},
  [5622] = {.lex_state = 0, .external_lex_state = 186},
  [5623] = {.lex_state = 0, .external_lex_state = 186},
  [5624] = {.lex_state = 0, .external_lex_state = 186},
  [5625] = {.lex_state = 0, .external_lex_state = 231},
  [5626] = {.lex_state = 0, .external_lex_state = 251},
  [5627] = {.lex_state = 0, .external_lex_state = 231},
  [5628] = {.lex_state = 0, .external_lex_state = 186},
  [5629] = {.lex_state = 0, .external_lex_state = 186},
  [5630] = {.lex_state = 0, .external_lex_state = 251},
  [5631] = {.lex_state = 0, .external_lex_state = 186},
  [5632] = {.lex_state = 0, .external_lex_state = 186},
  [5633] = {.lex_state = 0, .external_lex_state = 186},
  [5634] = {.lex_state = 0, .external_lex_state = 186},
  [5635] = {.lex_state = 0, .external_lex_state = 223},
  [5636] = {.lex_state = 0, .external_lex_state = 186},
  [5637] = {.lex_state = 0, .external_lex_state = 186},
  [5638] = {.lex_state = 0, .external_lex_state = 186},
  [5639] = {.lex_state = 0, .external_lex_state = 186},
  [5640] = {.lex_state = 0, .external_lex_state = 224},
  [5641] = {.lex_state = 0, .external_lex_state = 186},
  [5642] = {.lex_state = 0, .external_lex_state = 186},
  [5643] = {.lex_state = 0, .external_lex_state = 186},
  [5644] = {.lex_state = 0, .external_lex_state = 186},
  [5645] = {.lex_state = 0, .external_lex_state = 186},
  [5646] = {.lex_state = 0, .external_lex_state = 186},
  [5647] = {.lex_state = 0, .external_lex_state = 186},
  [5648] = {.lex_state = 0, .external_lex_state = 222},
  [5649] = {.lex_state = 0, .external_lex_state = 186},
  [5650] = {.lex_state = 0, .external_lex_state = 186},
  [5651] = {.lex_state = 0, .external_lex_state = 186},
  [5652] = {.lex_state = 0, .external_lex_state = 186},
  [5653] = {.lex_state = 0, .external_lex_state = 231},
  [5654] = {.lex_state = 0, .external_lex_state = 223},
  [5655] = {.lex_state = 0, .external_lex_state = 186},
  [5656] = {.lex_state = 0, .external_lex_state = 186},
  [5657] = {.lex_state = 0, .external_lex_state = 186},
  [5658] = {.lex_state = 0, .external_lex_state = 186},
  [5659] = {.lex_state = 0, .external_lex_state = 186},
  [5660] = {.lex_state = 0, .external_lex_state = 186},
  [5661] = {.lex_state = 0, .external_lex_state = 186},
  [5662] = {.lex_state = 0, .external_lex_state = 186},
  [5663] = {.lex_state = 0, .external_lex_state = 186},
  [5664] = {.lex_state = 0, .external_lex_state = 186},
  [5665] = {.lex_state = 0, .external_lex_state = 186},
  [5666] = {.lex_state = 0, .external_lex_state = 186},
  [5667] = {.lex_state = 0, .external_lex_state = 186},
  [5668] = {.lex_state = 0, .external_lex_state = 186},
  [5669] = {.lex_state = 0, .external_lex_state = 231},
  [5670] = {.lex_state = 0, .external_lex_state = 186},
  [5671] = {.lex_state = 0, .external_lex_state = 249},
  [5672] = {.lex_state = 0, .external_lex_state = 252},
  [5673] = {.lex_state = 0, .external_lex_state = 186},
  [5674] = {.lex_state = 0, .external_lex_state = 231},
  [5675] = {.lex_state = 0, .external_lex_state = 241},
  [5676] = {.lex_state = 0, .external_lex_state = 186},
  [5677] = {.lex_state = 0, .external_lex_state = 223},
  [5678] = {.lex_state = 0, .external_lex_state = 186},
  [5679] = {.lex_state = 0, .external_lex_state = 236},
  [5680] = {.lex_state = 0, .external_lex_state = 251},
  [5681] = {.lex_state = 0, .external_lex_state = 186},
  [5682] = {.lex_state = 0, .external_lex_state = 236},
  [5683] = {.lex_state = 0, .external_lex_state = 250},
  [5684] = {.lex_state = 0, .external_lex_state = 186},
  [5685] = {.lex_state = 0, .external_lex_state = 231},
  [5686] = {.lex_state = 0, .external_lex_state = 186},
  [5687] = {.lex_state = 0, .external_lex_state = 224},
  [5688] = {.lex_state = 0, .external_lex_state = 186},
  [5689] = {.lex_state = 0, .external_lex_state = 186},
  [5690] = {.lex_state = 0, .external_lex_state = 222},
  [5691] = {.lex_state = 0, .external_lex_state = 186},
  [5692] = {.lex_state = 0, .external_lex_state = 253},
  [5693] = {.lex_state = 0, .external_lex_state = 186},
  [5694] = {.lex_state = 0, .external_lex_state = 254},
  [5695] = {.lex_state = 0, .external_lex_state = 186},
  [5696] = {.lex_state = 0, .external_lex_state = 254},
  [5697] = {.lex_state = 0, .external_lex_state = 186},
  [5698] = {.lex_state = 0, .external_lex_state = 186},
  [5699] = {.lex_state = 0, .external_lex_state = 186},
  [5700] = {.lex_state = 0, .external_lex_state = 231},
  [5701] = {.lex_state = 0, .external_lex_state = 186},
  [5702] = {.lex_state = 0, .external_lex_state = 186},
  [5703] = {.lex_state = 0, .external_lex_state = 224},
  [5704] = {.lex_state = 0, .external_lex_state = 186},
  [5705] = {.lex_state = 0, .external_lex_state = 186},
  [5706] = {.lex_state = 0, .external_lex_state = 186},
  [5707] = {.lex_state = 0, .external_lex_state = 241},
  [5708] = {.lex_state = 0, .external_lex_state = 222},
  [5709] = {.lex_state = 0, .external_lex_state = 186},
  [5710] = {.lex_state = 0, .external_lex_state = 186},
  [5711] = {.lex_state = 0, .external_lex_state = 253},
  [5712] = {.lex_state = 0, .external_lex_state = 242},
  [5713] = {.lex_state = 0, .external_lex_state = 186},
  [5714] = {.lex_state = 0, .external_lex_state = 232},
  [5715] = {.lex_state = 0, .external_lex_state = 231},
  [5716] = {.lex_state = 0, .external_lex_state = 255},
  [5717] = {.lex_state = 0, .external_lex_state = 186},
  [5718] = {.lex_state = 0, .external_lex_state = 186},
  [5719] = {.lex_state = 0, .external_lex_state = 186},
  [5720] = {.lex_state = 0, .external_lex_state = 186},
  [5721] = {.lex_state = 0, .external_lex_state = 249},
  [5722] = {.lex_state = 0, .external_lex_state = 186},
  [5723] = {.lex_state = 0, .external_lex_state = 186},
  [5724] = {.lex_state = 0, .external_lex_state = 231},
  [5725] = {.lex_state = 0, .external_lex_state = 224},
  [5726] = {.lex_state = 0, .external_lex_state = 256},
  [5727] = {.lex_state = 0, .external_lex_state = 224},
  [5728] = {.lex_state = 0, .external_lex_state = 224},
  [5729] = {.lex_state = 0, .external_lex_state = 186},
  [5730] = {.lex_state = 0, .external_lex_state = 222},
  [5731] = {.lex_state = 0, .external_lex_state = 186},
  [5732] = {.lex_state = 0, .external_lex_state = 186},
  [5733] = {.lex_state = 0, .external_lex_state = 186},
  [5734] = {.lex_state = 0, .external_lex_state = 186},
  [5735] = {.lex_state = 0, .external_lex_state = 186},
  [5736] = {.lex_state = 0, .external_lex_state = 186},
  [5737] = {.lex_state = 0, .external_lex_state = 251},
  [5738] = {.lex_state = 0, .external_lex_state = 231},
  [5739] = {.lex_state = 0, .external_lex_state = 249},
  [5740] = {.lex_state = 0, .external_lex_state = 231},
  [5741] = {.lex_state = 0, .external_lex_state = 252},
  [5742] = {.lex_state = 0, .external_lex_state = 257},
  [5743] = {.lex_state = 0, .external_lex_state = 186},
  [5744] = {.lex_state = 0, .external_lex_state = 186},
  [5745] = {.lex_state = 0, .external_lex_state = 186},
  [5746] = {.lex_state = 0, .external_lex_state = 222},
  [5747] = {.lex_state = 0, .external_lex_state = 186},
  [5748] = {.lex_state = 0, .external_lex_state = 223},
  [5749] = {.lex_state = 0, .external_lex_state = 186},
  [5750] = {.lex_state = 0, .external_lex_state = 186},
  [5751] = {.lex_state = 0, .external_lex_state = 245},
  [5752] = {.lex_state = 0, .external_lex_state = 186},
  [5753] = {.lex_state = 0, .external_lex_state = 186},
  [5754] = {.lex_state = 0, .external_lex_state = 186},
  [5755] = {.lex_state = 0, .external_lex_state = 186},
  [5756] = {.lex_state = 0, .external_lex_state = 240},
  [5757] = {.lex_state = 0, .external_lex_state = 231},
  [5758] = {.lex_state = 0, .external_lex_state = 186},
  [5759] = {.lex_state = 0, .external_lex_state = 224},
  [5760] = {.lex_state = 0, .external_lex_state = 186},
  [5761] = {.lex_state = 0, .external_lex_state = 220},
  [5762] = {.lex_state = 0, .external_lex_state = 224},
  [5763] = {.lex_state = 0, .external_lex_state = 186},
  [5764] = {.lex_state = 0, .external_lex_state = 250},
  [5765] = {.lex_state = 0, .external_lex_state = 186},
  [5766] = {.lex_state = 0, .external_lex_state = 224},
  [5767] = {.lex_state = 0, .external_lex_state = 186},
  [5768] = {.lex_state = 0, .external_lex_state = 222},
  [5769] = {.lex_state = 0, .external_lex_state = 186},
  [5770] = {.lex_state = 0, .external_lex_state = 231},
  [5771] = {.lex_state = 0, .external_lex_state = 186},
  [5772] = {.lex_state = 0, .external_lex_state = 186},
  [5773] = {.lex_state = 0, .external_lex_state = 222},
  [5774] = {.lex_state = 0, .external_lex_state = 222},
  [5775] = {.lex_state = 0, .external_lex_state = 186},
  [5776] = {.lex_state = 0, .external_lex_state = 186},
  [5777] = {.lex_state = 0, .external_lex_state = 220},
  [5778] = {.lex_state = 0, .external_lex_state = 220},
  [5779] = {.lex_state = 0, .external_lex_state = 186},
  [5780] = {.lex_state = 0, .external_lex_state = 186},
  [5781] = {.lex_state = 0, .external_lex_state = 224},
  [5782] = {.lex_state = 0, .external_lex_state = 186},
  [5783] = {.lex_state = 0, .external_lex_state = 186},
  [5784] = {.lex_state = 0, .external_lex_state = 186},
  [5785] = {.lex_state = 0, .external_lex_state = 186},
  [5786] = {.lex_state = 0, .external_lex_state = 186},
  [5787] = {.lex_state = 0, .external_lex_state = 186},
  [5788] = {.lex_state = 0, .external_lex_state = 195},
  [5789] = {.lex_state = 0, .external_lex_state = 186},
  [5790] = {.lex_state = 0, .external_lex_state = 186},
  [5791] = {.lex_state = 0, .external_lex_state = 186},
  [5792] = {.lex_state = 0, .external_lex_state = 186},
  [5793] = {.lex_state = 0, .external_lex_state = 186},
  [5794] = {.lex_state = 0, .external_lex_state = 186},
  [5795] = {.lex_state = 0, .external_lex_state = 186},
  [5796] = {.lex_state = 0, .external_lex_state = 186},
  [5797] = {.lex_state = 0, .external_lex_state = 186},
  [5798] = {.lex_state = 0, .external_lex_state = 186},
  [5799] = {.lex_state = 0, .external_lex_state = 224},
  [5800] = {.lex_state = 0, .external_lex_state = 245},
  [5801] = {.lex_state = 0, .external_lex_state = 186},
  [5802] = {.lex_state = 0, .external_lex_state = 220},
  [5803] = {.lex_state = 0, .external_lex_state = 220},
  [5804] = {.lex_state = 0, .external_lex_state = 245},
  [5805] = {.lex_state = 0, .external_lex_state = 231},
  [5806] = {.lex_state = 0, .external_lex_state = 186},
  [5807] = {.lex_state = 0, .external_lex_state = 186},
  [5808] = {.lex_state = 0, .external_lex_state = 223},
  [5809] = {.lex_state = 0, .external_lex_state = 224},
  [5810] = {.lex_state = 0, .external_lex_state = 186},
  [5811] = {.lex_state = 0, .external_lex_state = 245},
  [5812] = {.lex_state = 0, .external_lex_state = 186},
  [5813] = {.lex_state = 0, .external_lex_state = 231},
  [5814] = {.lex_state = 0, .external_lex_state = 186},
  [5815] = {.lex_state = 0, .external_lex_state = 186},
  [5816] = {.lex_state = 0, .external_lex_state = 195},
  [5817] = {.lex_state = 0, .external_lex_state = 186},
  [5818] = {.lex_state = 0, .external_lex_state = 186},
  [5819] = {.lex_state = 0, .external_lex_state = 246},
  [5820] = {.lex_state = 0, .external_lex_state = 186},
  [5821] = {.lex_state = 0, .external_lex_state = 246},
  [5822] = {.lex_state = 0, .external_lex_state = 186},
  [5823] = {.lex_state = 0, .external_lex_state = 246},
  [5824] = {.lex_state = 0, .external_lex_state = 222},
  [5825] = {.lex_state = 0, .external_lex_state = 186},
  [5826] = {.lex_state = 0, .external_lex_state = 186},
  [5827] = {.lex_state = 0, .external_lex_state = 224},
  [5828] = {.lex_state = 0, .external_lex_state = 258},
  [5829] = {.lex_state = 0, .external_lex_state = 256},
  [5830] = {.lex_state = 0, .external_lex_state = 186},
  [5831] = {.lex_state = 0, .external_lex_state = 186},
  [5832] = {.lex_state = 0, .external_lex_state = 224},
  [5833] = {.lex_state = 0, .external_lex_state = 186},
  [5834] = {.lex_state = 0, .external_lex_state = 195},
  [5835] = {.lex_state = 0, .external_lex_state = 186},
  [5836] = {.lex_state = 0, .external_lex_state = 224},
  [5837] = {.lex_state = 0, .external_lex_state = 255},
  [5838] = {.lex_state = 0, .external_lex_state = 231},
  [5839] = {.lex_state = 0, .external_lex_state = 224},
  [5840] = {.lex_state = 0, .external_lex_state = 242},
  [5841] = {.lex_state = 0, .external_lex_state = 186},
  [5842] = {.lex_state = 0, .external_lex_state = 186},
  [5843] = {.lex_state = 0, .external_lex_state = 195},
  [5844] = {.lex_state = 0, .external_lex_state = 222},
  [5845] = {.lex_state = 0, .external_lex_state = 186},
  [5846] = {.lex_state = 0, .external_lex_state = 186},
  [5847] = {.lex_state = 0, .external_lex_state = 251},
  [5848] = {.lex_state = 0, .external_lex_state = 186},
  [5849] = {.lex_state = 0, .external_lex_state = 186},
  [5850] = {.lex_state = 0, .external_lex_state = 256},
  [5851] = {.lex_state = 0, .external_lex_state = 186},
  [5852] = {.lex_state = 0, .external_lex_state = 223},
  [5853] = {.lex_state = 0, .external_lex_state = 224},
  [5854] = {.lex_state = 0, .external_lex_state = 251},
  [5855] = {.lex_state = 0, .external_lex_state = 256},
  [5856] = {.lex_state = 0, .external_lex_state = 224},
  [5857] = {.lex_state = 0, .external_lex_state = 186},
  [5858] = {.lex_state = 0, .external_lex_state = 186},
  [5859] = {.lex_state = 0, .external_lex_state = 186},
  [5860] = {.lex_state = 0, .external_lex_state = 224},
  [5861] = {.lex_state = 0, .external_lex_state = 224},
  [5862] = {.lex_state = 0, .external_lex_state = 224},
  [5863] = {.lex_state = 0, .external_lex_state = 241},
  [5864] = {.lex_state = 0, .external_lex_state = 255},
  [5865] = {.lex_state = 0, .external_lex_state = 255},
  [5866] = {.lex_state = 0, .external_lex_state = 256},
  [5867] = {.lex_state = 0, .external_lex_state = 186},
  [5868] = {.lex_state = 0, .external_lex_state = 195},
  [5869] = {.lex_state = 0, .external_lex_state = 186},
  [5870] = {.lex_state = 0, .external_lex_state = 186},
  [5871] = {.lex_state = 0, .external_lex_state = 251},
  [5872] = {.lex_state = 0, .external_lex_state = 231},
  [5873] = {.lex_state = 0, .external_lex_state = 224},
  [5874] = {.lex_state = 0, .external_lex_state = 186},
  [5875] = {.lex_state = 0, .external_lex_state = 255},
  [5876] = {.lex_state = 0, .external_lex_state = 186},
  [5877] = {.lex_state = 0, .external_lex_state = 224},
  [5878] = {.lex_state = 0, .external_lex_state = 224},
  [5879] = {.lex_state = 0, .external_lex_state = 224},
  [5880] = {.lex_state = 0, .external_lex_state = 222},
  [5881] = {.lex_state = 0, .external_lex_state = 186},
  [5882] = {.lex_state = 0, .external_lex_state = 186},
  [5883] = {.lex_state = 0, .external_lex_state = 186},
  [5884] = {.lex_state = 0, .external_lex_state = 186},
  [5885] = {.lex_state = 0, .external_lex_state = 186},
  [5886] = {.lex_state = 0, .external_lex_state = 186},
  [5887] = {.lex_state = 0, .external_lex_state = 186},
  [5888] = {.lex_state = 0, .external_lex_state = 231},
  [5889] = {.lex_state = 0, .external_lex_state = 222},
  [5890] = {.lex_state = 0, .external_lex_state = 186},
  [5891] = {.lex_state = 0, .external_lex_state = 224},
  [5892] = {.lex_state = 0, .external_lex_state = 231},
  [5893] = {.lex_state = 0, .external_lex_state = 186},
  [5894] = {.lex_state = 0, .external_lex_state = 186},
  [5895] = {.lex_state = 0, .external_lex_state = 186},
  [5896] = {.lex_state = 0, .external_lex_state = 186},
  [5897] = {.lex_state = 0, .external_lex_state = 224},
  [5898] = {.lex_state = 0, .external_lex_state = 186},
  [5899] = {.lex_state = 0, .external_lex_state = 255},
  [5900] = {.lex_state = 0, .external_lex_state = 245},
  [5901] = {.lex_state = 0, .external_lex_state = 222},
  [5902] = {.lex_state = 0, .external_lex_state = 186},
  [5903] = {.lex_state = 0, .external_lex_state = 186},
  [5904] = {.lex_state = 0, .external_lex_state = 186},
  [5905] = {.lex_state = 0, .external_lex_state = 224},
  [5906] = {.lex_state = 0, .external_lex_state = 224},
  [5907] = {.lex_state = 0, .external_lex_state = 186},
  [5908] = {.lex_state = 0, .external_lex_state = 186},
  [5909] = {.lex_state = 0, .external_lex_state = 222},
  [5910] = {.lex_state = 0, .external_lex_state = 186},
  [5911] = {.lex_state = 0, .external_lex_state = 249},
  [5912] = {.lex_state = 0, .external_lex_state = 251},
  [5913] = {.lex_state = 0, .external_lex_state = 256},
  [5914] = {.lex_state = 0, .external_lex_state = 186},
  [5915] = {.lex_state = 0, .external_lex_state = 186},
  [5916] = {.lex_state = 0, .external_lex_state = 224},
  [5917] = {.lex_state = 0, .external_lex_state = 251},
  [5918] = {.lex_state = 0, .external_lex_state = 241},
  [5919] = {.lex_state = 0, .external_lex_state = 186},
  [5920] = {.lex_state = 0, .external_lex_state = 186},
  [5921] = {.lex_state = 0, .external_lex_state = 224},
  [5922] = {.lex_state = 0, .external_lex_state = 224},
  [5923] = {.lex_state = 0, .external_lex_state = 224},
  [5924] = {.lex_state = 0, .external_lex_state = 195},
  [5925] = {.lex_state = 0, .external_lex_state = 255},
  [5926] = {.lex_state = 0, .external_lex_state = 246},
  [5927] = {.lex_state = 0, .external_lex_state = 250},
  [5928] = {.lex_state = 0, .external_lex_state = 256},
  [5929] = {.lex_state = 0, .external_lex_state = 186},
  [5930] = {.lex_state = 0, .external_lex_state = 186},
  [5931] = {.lex_state = 0, .external_lex_state = 250},
  [5932] = {.lex_state = 0, .external_lex_state = 251},
  [5933] = {.lex_state = 0, .external_lex_state = 231},
  [5934] = {.lex_state = 0, .external_lex_state = 224},
  [5935] = {.lex_state = 0, .external_lex_state = 186},
  [5936] = {.lex_state = 0, .external_lex_state = 255},
  [5937] = {.lex_state = 0, .external_lex_state = 186},
  [5938] = {.lex_state = 0, .external_lex_state = 224},
  [5939] = {.lex_state = 0, .external_lex_state = 224},
  [5940] = {.lex_state = 0, .external_lex_state = 224},
  [5941] = {.lex_state = 0, .external_lex_state = 186},
  [5942] = {.lex_state = 0, .external_lex_state = 186},
  [5943] = {.lex_state = 0, .external_lex_state = 186},
  [5944] = {.lex_state = 0, .external_lex_state = 186},
  [5945] = {.lex_state = 0, .external_lex_state = 231},
  [5946] = {.lex_state = 0, .external_lex_state = 224},
  [5947] = {.lex_state = 0, .external_lex_state = 231},
  [5948] = {.lex_state = 0, .external_lex_state = 224},
  [5949] = {.lex_state = 0, .external_lex_state = 255},
  [5950] = {.lex_state = 0, .external_lex_state = 223},
  [5951] = {.lex_state = 0, .external_lex_state = 186},
  [5952] = {.lex_state = 0, .external_lex_state = 186},
  [5953] = {.lex_state = 0, .external_lex_state = 224},
  [5954] = {.lex_state = 0, .external_lex_state = 224},
  [5955] = {.lex_state = 0, .external_lex_state = 252},
  [5956] = {.lex_state = 0, .external_lex_state = 252},
  [5957] = {.lex_state = 0, .external_lex_state = 252},
  [5958] = {.lex_state = 0, .external_lex_state = 252},
  [5959] = {.lex_state = 0, .external_lex_state = 252},
  [5960] = {.lex_state = 0, .external_lex_state = 252},
  [5961] = {.lex_state = 0, .external_lex_state = 222},
  [5962] = {.lex_state = 0, .external_lex_state = 252},
  [5963] = {.lex_state = 0, .external_lex_state = 252},
  [5964] = {.lex_state = 0, .external_lex_state = 252},
  [5965] = {.lex_state = 0, .external_lex_state = 252},
  [5966] = {.lex_state = 0, .external_lex_state = 252},
  [5967] = {.lex_state = 0, .external_lex_state = 252},
  [5968] = {.lex_state = 0, .external_lex_state = 252},
  [5969] = {.lex_state = 0, .external_lex_state = 252},
  [5970] = {.lex_state = 0, .external_lex_state = 252},
  [5971] = {.lex_state = 0, .external_lex_state = 252},
  [5972] = {.lex_state = 0, .external_lex_state = 252},
  [5973] = {.lex_state = 0, .external_lex_state = 252},
  [5974] = {.lex_state = 0, .external_lex_state = 252},
  [5975] = {.lex_state = 0, .external_lex_state = 252},
  [5976] = {.lex_state = 0, .external_lex_state = 252},
  [5977] = {.lex_state = 0, .external_lex_state = 252},
  [5978] = {.lex_state = 0, .external_lex_state = 252},
  [5979] = {.lex_state = 0, .external_lex_state = 252},
  [5980] = {.lex_state = 0, .external_lex_state = 252},
  [5981] = {.lex_state = 0, .external_lex_state = 252},
  [5982] = {.lex_state = 0, .external_lex_state = 252},
  [5983] = {.lex_state = 0, .external_lex_state = 252},
  [5984] = {.lex_state = 0, .external_lex_state = 252},
  [5985] = {.lex_state = 0, .external_lex_state = 252},
  [5986] = {.lex_state = 0, .external_lex_state = 252},
  [5987] = {.lex_state = 0, .external_lex_state = 252},
  [5988] = {.lex_state = 0, .external_lex_state = 252},
  [5989] = {.lex_state = 0, .external_lex_state = 252},
  [5990] = {.lex_state = 0, .external_lex_state = 252},
  [5991] = {.lex_state = 0, .external_lex_state = 252},
  [5992] = {.lex_state = 0, .external_lex_state = 252},
  [5993] = {.lex_state = 0, .external_lex_state = 252},
  [5994] = {.lex_state = 0, .external_lex_state = 252},
  [5995] = {.lex_state = 0, .external_lex_state = 252},
  [5996] = {.lex_state = 0, .external_lex_state = 252},
  [5997] = {.lex_state = 0, .external_lex_state = 252},
  [5998] = {.lex_state = 0, .external_lex_state = 252},
  [5999] = {.lex_state = 0, .external_lex_state = 252},
  [6000] = {.lex_state = 0, .external_lex_state = 252},
  [6001] = {.lex_state = 0, .external_lex_state = 252},
  [6002] = {.lex_state = 0, .external_lex_state = 252},
  [6003] = {.lex_state = 0, .external_lex_state = 252},
  [6004] = {.lex_state = 0, .external_lex_state = 252},
  [6005] = {.lex_state = 0, .external_lex_state = 186},
  [6006] = {.lex_state = 0, .external_lex_state = 259},
  [6007] = {.lex_state = 0, .external_lex_state = 259},
  [6008] = {.lex_state = 0, .external_lex_state = 186},
  [6009] = {.lex_state = 0, .external_lex_state = 186},
  [6010] = {.lex_state = 0, .external_lex_state = 186},
  [6011] = {.lex_state = 0, .external_lex_state = 210},
  [6012] = {.lex_state = 0, .external_lex_state = 186},
  [6013] = {.lex_state = 0, .external_lex_state = 236},
  [6014] = {.lex_state = 0, .external_lex_state = 186},
  [6015] = {.lex_state = 0, .external_lex_state = 186},
  [6016] = {.lex_state = 0, .external_lex_state = 186},
  [6017] = {.lex_state = 0, .external_lex_state = 259},
  [6018] = {.lex_state = 0, .external_lex_state = 259},
  [6019] = {.lex_state = 0, .external_lex_state = 259},
  [6020] = {.lex_state = 0, .external_lex_state = 260},
  [6021] = {.lex_state = 0, .external_lex_state = 186},
  [6022] = {.lex_state = 0, .external_lex_state = 259},
  [6023] = {.lex_state = 0, .external_lex_state = 222},
  [6024] = {.lex_state = 0, .external_lex_state = 222},
  [6025] = {.lex_state = 0, .external_lex_state = 186},
  [6026] = {.lex_state = 0, .external_lex_state = 259},
  [6027] = {.lex_state = 0, .external_lex_state = 259},
  [6028] = {.lex_state = 0, .external_lex_state = 261},
  [6029] = {.lex_state = 0, .external_lex_state = 259},
  [6030] = {.lex_state = 0, .external_lex_state = 259},
  [6031] = {.lex_state = 0, .external_lex_state = 186},
  [6032] = {.lex_state = 0, .external_lex_state = 186},
  [6033] = {.lex_state = 0, .external_lex_state = 224},
  [6034] = {.lex_state = 0, .external_lex_state = 186},
  [6035] = {.lex_state = 0, .external_lex_state = 222},
  [6036] = {.lex_state = 0, .external_lex_state = 261},
  [6037] = {.lex_state = 0, .external_lex_state = 259},
  [6038] = {.lex_state = 0, .external_lex_state = 249},
  [6039] = {.lex_state = 0, .external_lex_state = 261},
  [6040] = {.lex_state = 0, .external_lex_state = 259},
  [6041] = {.lex_state = 0, .external_lex_state = 251},
  [6042] = {.lex_state = 0, .external_lex_state = 259},
  [6043] = {.lex_state = 0, .external_lex_state = 262},
  [6044] = {.lex_state = 0, .external_lex_state = 259},
  [6045] = {.lex_state = 0, .external_lex_state = 186},
  [6046] = {.lex_state = 0, .external_lex_state = 259},
  [6047] = {.lex_state = 0, .external_lex_state = 186},
  [6048] = {.lex_state = 0, .external_lex_state = 186},
  [6049] = {.lex_state = 0, .external_lex_state = 259},
  [6050] = {.lex_state = 0, .external_lex_state = 259},
  [6051] = {.lex_state = 0, .external_lex_state = 222},
  [6052] = {.lex_state = 0, .external_lex_state = 186},
  [6053] = {.lex_state = 0, .external_lex_state = 259},
  [6054] = {.lex_state = 0, .external_lex_state = 259},
  [6055] = {.lex_state = 0, .external_lex_state = 222},
  [6056] = {.lex_state = 0, .external_lex_state = 259},
  [6057] = {.lex_state = 0, .external_lex_state = 259},
  [6058] = {.lex_state = 0, .external_lex_state = 222},
  [6059] = {.lex_state = 0, .external_lex_state = 259},
  [6060] = {.lex_state = 0, .external_lex_state = 186},
  [6061] = {.lex_state = 0, .external_lex_state = 259},
  [6062] = {.lex_state = 0, .external_lex_state = 261},
  [6063] = {.lex_state = 0, .external_lex_state = 261},
  [6064] = {.lex_state = 0, .external_lex_state = 259},
  [6065] = {.lex_state = 0, .external_lex_state = 259},
  [6066] = {.lex_state = 0, .external_lex_state = 259},
  [6067] = {.lex_state = 0, .external_lex_state = 210},
  [6068] = {.lex_state = 0, .external_lex_state = 259},
  [6069] = {.lex_state = 0, .external_lex_state = 222},
  [6070] = {.lex_state = 0, .external_lex_state = 259},
  [6071] = {.lex_state = 0, .external_lex_state = 259},
  [6072] = {.lex_state = 0, .external_lex_state = 210},
  [6073] = {.lex_state = 0, .external_lex_state = 186},
  [6074] = {.lex_state = 0, .external_lex_state = 259},
  [6075] = {.lex_state = 0, .external_lex_state = 186},
  [6076] = {.lex_state = 0, .external_lex_state = 261},
  [6077] = {.lex_state = 0, .external_lex_state = 186},
  [6078] = {.lex_state = 0, .external_lex_state = 259},
  [6079] = {.lex_state = 0, .external_lex_state = 186},
  [6080] = {.lex_state = 0, .external_lex_state = 186},
  [6081] = {.lex_state = 0, .external_lex_state = 186},
  [6082] = {.lex_state = 0, .external_lex_state = 186},
  [6083] = {.lex_state = 0, .external_lex_state = 261},
  [6084] = {.lex_state = 0, .external_lex_state = 222},
  [6085] = {.lex_state = 0, .external_lex_state = 186},
  [6086] = {.lex_state = 0, .external_lex_state = 236},
  [6087] = {.lex_state = 0, .external_lex_state = 186},
  [6088] = {.lex_state = 0, .external_lex_state = 259},
  [6089] = {.lex_state = 0, .external_lex_state = 259},
  [6090] = {.lex_state = 0, .external_lex_state = 186},
  [6091] = {.lex_state = 0, .external_lex_state = 186},
  [6092] = {.lex_state = 0, .external_lex_state = 186},
  [6093] = {.lex_state = 0, .external_lex_state = 259},
  [6094] = {.lex_state = 0, .external_lex_state = 259},
  [6095] = {.lex_state = 0, .external_lex_state = 186},
  [6096] = {.lex_state = 0, .external_lex_state = 186},
  [6097] = {.lex_state = 0, .external_lex_state = 186},
  [6098] = {.lex_state = 0, .external_lex_state = 186},
  [6099] = {.lex_state = 0, .external_lex_state = 253},
  [6100] = {.lex_state = 0, .external_lex_state = 186},
  [6101] = {.lex_state = 0, .external_lex_state = 259},
  [6102] = {.lex_state = 0, .external_lex_state = 261},
  [6103] = {.lex_state = 0, .external_lex_state = 259},
  [6104] = {.lex_state = 0, .external_lex_state = 251},
  [6105] = {.lex_state = 0, .external_lex_state = 259},
  [6106] = {.lex_state = 0, .external_lex_state = 186},
  [6107] = {.lex_state = 0, .external_lex_state = 259},
  [6108] = {.lex_state = 0, .external_lex_state = 260},
  [6109] = {.lex_state = 0, .external_lex_state = 259},
  [6110] = {.lex_state = 0, .external_lex_state = 260},
  [6111] = {.lex_state = 0, .external_lex_state = 186},
  [6112] = {.lex_state = 0, .external_lex_state = 186},
  [6113] = {.lex_state = 0, .external_lex_state = 186},
  [6114] = {.lex_state = 0, .external_lex_state = 222},
  [6115] = {.lex_state = 0, .external_lex_state = 186},
  [6116] = {.lex_state = 0, .external_lex_state = 259},
  [6117] = {.lex_state = 0, .external_lex_state = 259},
  [6118] = {.lex_state = 0, .external_lex_state = 259},
  [6119] = {.lex_state = 0, .external_lex_state = 186},
  [6120] = {.lex_state = 0, .external_lex_state = 259},
  [6121] = {.lex_state = 0, .external_lex_state = 186},
  [6122] = {.lex_state = 0, .external_lex_state = 259},
  [6123] = {.lex_state = 0, .external_lex_state = 261},
  [6124] = {.lex_state = 0, .external_lex_state = 259},
  [6125] = {.lex_state = 0, .external_lex_state = 259},
  [6126] = {.lex_state = 0, .external_lex_state = 186},
  [6127] = {.lex_state = 0, .external_lex_state = 259},
  [6128] = {.lex_state = 0, .external_lex_state = 261},
  [6129] = {.lex_state = 0, .external_lex_state = 224},
  [6130] = {.lex_state = 0, .external_lex_state = 186},
  [6131] = {.lex_state = 0, .external_lex_state = 186},
  [6132] = {.lex_state = 0, .external_lex_state = 259},
  [6133] = {.lex_state = 0, .external_lex_state = 186},
  [6134] = {.lex_state = 0, .external_lex_state = 259},
  [6135] = {.lex_state = 0, .external_lex_state = 259},
  [6136] = {.lex_state = 0, .external_lex_state = 259},
  [6137] = {.lex_state = 0, .external_lex_state = 186},
  [6138] = {.lex_state = 0, .external_lex_state = 186},
  [6139] = {.lex_state = 0, .external_lex_state = 259},
  [6140] = {.lex_state = 0, .external_lex_state = 186},
  [6141] = {.lex_state = 0, .external_lex_state = 186},
  [6142] = {.lex_state = 0, .external_lex_state = 259},
  [6143] = {.lex_state = 0, .external_lex_state = 186},
  [6144] = {.lex_state = 0, .external_lex_state = 259},
  [6145] = {.lex_state = 0, .external_lex_state = 186},
  [6146] = {.lex_state = 0, .external_lex_state = 186},
  [6147] = {.lex_state = 0, .external_lex_state = 186},
  [6148] = {.lex_state = 0, .external_lex_state = 186},
  [6149] = {.lex_state = 0, .external_lex_state = 263},
  [6150] = {.lex_state = 0, .external_lex_state = 259},
  [6151] = {.lex_state = 0, .external_lex_state = 259},
  [6152] = {.lex_state = 0, .external_lex_state = 186},
  [6153] = {.lex_state = 0, .external_lex_state = 186},
  [6154] = {.lex_state = 0, .external_lex_state = 186},
  [6155] = {.lex_state = 0, .external_lex_state = 186},
  [6156] = {.lex_state = 0, .external_lex_state = 242},
  [6157] = {.lex_state = 0, .external_lex_state = 259},
  [6158] = {.lex_state = 0, .external_lex_state = 186},
  [6159] = {.lex_state = 0, .external_lex_state = 259},
  [6160] = {.lex_state = 0, .external_lex_state = 186},
  [6161] = {.lex_state = 0, .external_lex_state = 186},
  [6162] = {.lex_state = 0, .external_lex_state = 186},
  [6163] = {.lex_state = 0, .external_lex_state = 224},
  [6164] = {.lex_state = 0, .external_lex_state = 186},
  [6165] = {.lex_state = 0, .external_lex_state = 186},
  [6166] = {.lex_state = 0, .external_lex_state = 186},
  [6167] = {.lex_state = 0, .external_lex_state = 186},
  [6168] = {.lex_state = 0, .external_lex_state = 261},
  [6169] = {.lex_state = 0, .external_lex_state = 186},
  [6170] = {.lex_state = 0, .external_lex_state = 186},
  [6171] = {.lex_state = 0, .external_lex_state = 186},
  [6172] = {.lex_state = 0, .external_lex_state = 222},
  [6173] = {.lex_state = 0, .external_lex_state = 186},
  [6174] = {.lex_state = 0, .external_lex_state = 186},
  [6175] = {.lex_state = 0, .external_lex_state = 186},
  [6176] = {.lex_state = 0, .external_lex_state = 186},
  [6177] = {.lex_state = 0, .external_lex_state = 222},
  [6178] = {.lex_state = 0, .external_lex_state = 186},
  [6179] = {.lex_state = 0, .external_lex_state = 250},
  [6180] = {.lex_state = 0, .external_lex_state = 186},
  [6181] = {.lex_state = 0, .external_lex_state = 186},
  [6182] = {.lex_state = 0, .external_lex_state = 231},
  [6183] = {.lex_state = 0, .external_lex_state = 186},
  [6184] = {.lex_state = 0, .external_lex_state = 186},
  [6185] = {.lex_state = 0, .external_lex_state = 261},
  [6186] = {.lex_state = 0, .external_lex_state = 243},
  [6187] = {.lex_state = 0, .external_lex_state = 249},
  [6188] = {.lex_state = 0, .external_lex_state = 222},
  [6189] = {.lex_state = 0, .external_lex_state = 222},
  [6190] = {.lex_state = 0, .external_lex_state = 259},
  [6191] = {.lex_state = 0, .external_lex_state = 186},
  [6192] = {.lex_state = 0, .external_lex_state = 236},
  [6193] = {.lex_state = 0, .external_lex_state = 259},
  [6194] = {.lex_state = 0, .external_lex_state = 186},
  [6195] = {.lex_state = 0, .external_lex_state = 186},
  [6196] = {.lex_state = 0, .external_lex_state = 224},
  [6197] = {.lex_state = 0, .external_lex_state = 242},
  [6198] = {.lex_state = 0, .external_lex_state = 186},
  [6199] = {.lex_state = 0, .external_lex_state = 260},
  [6200] = {.lex_state = 0, .external_lex_state = 259},
  [6201] = {.lex_state = 0, .external_lex_state = 259},
  [6202] = {.lex_state = 0, .external_lex_state = 186},
  [6203] = {.lex_state = 0, .external_lex_state = 259},
  [6204] = {.lex_state = 0, .external_lex_state = 231},
  [6205] = {.lex_state = 0, .external_lex_state = 261},
  [6206] = {.lex_state = 0, .external_lex_state = 222},
  [6207] = {.lex_state = 0, .external_lex_state = 186},
  [6208] = {.lex_state = 0, .external_lex_state = 186},
  [6209] = {.lex_state = 0, .external_lex_state = 186},
  [6210] = {.lex_state = 0, .external_lex_state = 186},
  [6211] = {.lex_state = 0, .external_lex_state = 186},
  [6212] = {.lex_state = 0, .external_lex_state = 186},
  [6213] = {.lex_state = 0, .external_lex_state = 235},
  [6214] = {.lex_state = 0, .external_lex_state = 186},
  [6215] = {.lex_state = 0, .external_lex_state = 186},
  [6216] = {.lex_state = 0, .external_lex_state = 260},
  [6217] = {.lex_state = 0, .external_lex_state = 186},
  [6218] = {.lex_state = 0, .external_lex_state = 186},
  [6219] = {.lex_state = 0, .external_lex_state = 186},
  [6220] = {.lex_state = 0, .external_lex_state = 259},
  [6221] = {.lex_state = 0, .external_lex_state = 186},
  [6222] = {.lex_state = 0, .external_lex_state = 186},
  [6223] = {.lex_state = 0, .external_lex_state = 186},
  [6224] = {.lex_state = 0, .external_lex_state = 253},
  [6225] = {.lex_state = 0, .external_lex_state = 259},
  [6226] = {.lex_state = 0, .external_lex_state = 186},
  [6227] = {.lex_state = 0, .external_lex_state = 186},
  [6228] = {.lex_state = 0, .external_lex_state = 186},
  [6229] = {.lex_state = 0, .external_lex_state = 186},
  [6230] = {.lex_state = 0, .external_lex_state = 186},
  [6231] = {.lex_state = 0, .external_lex_state = 260},
  [6232] = {.lex_state = 0, .external_lex_state = 186},
  [6233] = {.lex_state = 0, .external_lex_state = 222},
  [6234] = {.lex_state = 0, .external_lex_state = 259},
  [6235] = {.lex_state = 0, .external_lex_state = 186},
  [6236] = {.lex_state = 0, .external_lex_state = 186},
  [6237] = {.lex_state = 0, .external_lex_state = 186},
  [6238] = {.lex_state = 0, .external_lex_state = 186},
  [6239] = {.lex_state = 0, .external_lex_state = 186},
  [6240] = {.lex_state = 0, .external_lex_state = 186},
  [6241] = {.lex_state = 0, .external_lex_state = 186},
  [6242] = {.lex_state = 0, .external_lex_state = 261},
  [6243] = {.lex_state = 0, .external_lex_state = 264},
  [6244] = {.lex_state = 0, .external_lex_state = 186},
  [6245] = {.lex_state = 0, .external_lex_state = 186},
  [6246] = {.lex_state = 0, .external_lex_state = 186},
  [6247] = {.lex_state = 0, .external_lex_state = 186},
  [6248] = {.lex_state = 0, .external_lex_state = 186},
  [6249] = {.lex_state = 0, .external_lex_state = 259},
  [6250] = {.lex_state = 0, .external_lex_state = 260},
  [6251] = {.lex_state = 0, .external_lex_state = 186},
  [6252] = {.lex_state = 0, .external_lex_state = 186},
  [6253] = {.lex_state = 0, .external_lex_state = 186},
  [6254] = {.lex_state = 0, .external_lex_state = 186},
  [6255] = {.lex_state = 0, .external_lex_state = 186},
  [6256] = {.lex_state = 0, .external_lex_state = 186},
  [6257] = {.lex_state = 0, .external_lex_state = 186},
  [6258] = {.lex_state = 0, .external_lex_state = 210},
  [6259] = {.lex_state = 0, .external_lex_state = 186},
  [6260] = {.lex_state = 0, .external_lex_state = 261},
  [6261] = {.lex_state = 0, .external_lex_state = 259},
  [6262] = {.lex_state = 0, .external_lex_state = 186},
  [6263] = {.lex_state = 0, .external_lex_state = 222},
  [6264] = {.lex_state = 0, .external_lex_state = 222},
  [6265] = {.lex_state = 0, .external_lex_state = 261},
  [6266] = {.lex_state = 0, .external_lex_state = 259},
  [6267] = {.lex_state = 0, .external_lex_state = 210},
  [6268] = {.lex_state = 0, .external_lex_state = 260},
  [6269] = {.lex_state = 0, .external_lex_state = 255},
  [6270] = {.lex_state = 0, .external_lex_state = 210},
  [6271] = {.lex_state = 0, .external_lex_state = 222},
  [6272] = {.lex_state = 0, .external_lex_state = 186},
  [6273] = {.lex_state = 0, .external_lex_state = 186},
  [6274] = {.lex_state = 0, .external_lex_state = 186},
  [6275] = {.lex_state = 0, .external_lex_state = 260},
  [6276] = {.lex_state = 0, .external_lex_state = 265},
  [6277] = {.lex_state = 0, .external_lex_state = 186},
  [6278] = {.lex_state = 0, .external_lex_state = 186},
  [6279] = {.lex_state = 0, .external_lex_state = 186},
  [6280] = {.lex_state = 0, .external_lex_state = 186},
  [6281] = {.lex_state = 0, .external_lex_state = 186},
  [6282] = {.lex_state = 0, .external_lex_state = 186},
  [6283] = {.lex_state = 0, .external_lex_state = 186},
  [6284] = {.lex_state = 0, .external_lex_state = 186},
  [6285] = {.lex_state = 0, .external_lex_state = 186},
  [6286] = {.lex_state = 0, .external_lex_state = 224},
  [6287] = {.lex_state = 0, .external_lex_state = 263},
  [6288] = {.lex_state = 0, .external_lex_state = 186},
  [6289] = {.lex_state = 0, .external_lex_state = 186},
  [6290] = {.lex_state = 0, .external_lex_state = 186},
  [6291] = {.lex_state = 0, .external_lex_state = 186},
  [6292] = {.lex_state = 0, .external_lex_state = 186},
  [6293] = {.lex_state = 0, .external_lex_state = 222},
  [6294] = {.lex_state = 0, .external_lex_state = 222},
  [6295] = {.lex_state = 0, .external_lex_state = 224},
  [6296] = {.lex_state = 0, .external_lex_state = 259},
  [6297] = {.lex_state = 0, .external_lex_state = 266},
  [6298] = {.lex_state = 0, .external_lex_state = 259},
  [6299] = {.lex_state = 0, .external_lex_state = 264},
  [6300] = {.lex_state = 0, .external_lex_state = 186},
  [6301] = {.lex_state = 0, .external_lex_state = 186},
  [6302] = {.lex_state = 0, .external_lex_state = 186},
  [6303] = {.lex_state = 0, .external_lex_state = 259},
  [6304] = {.lex_state = 0, .external_lex_state = 186},
  [6305] = {.lex_state = 0, .external_lex_state = 186},
  [6306] = {.lex_state = 0, .external_lex_state = 186},
  [6307] = {.lex_state = 0, .external_lex_state = 222},
  [6308] = {.lex_state = 0, .external_lex_state = 267},
  [6309] = {.lex_state = 0, .external_lex_state = 186},
  [6310] = {.lex_state = 0, .external_lex_state = 259},
  [6311] = {.lex_state = 0, .external_lex_state = 224},
  [6312] = {.lex_state = 0, .external_lex_state = 261},
  [6313] = {.lex_state = 0, .external_lex_state = 186},
  [6314] = {.lex_state = 0, .external_lex_state = 186},
  [6315] = {.lex_state = 0, .external_lex_state = 259},
  [6316] = {.lex_state = 0, .external_lex_state = 224},
  [6317] = {.lex_state = 0, .external_lex_state = 186},
  [6318] = {.lex_state = 0, .external_lex_state = 259},
  [6319] = {.lex_state = 0, .external_lex_state = 261},
  [6320] = {.lex_state = 0, .external_lex_state = 186},
  [6321] = {.lex_state = 0, .external_lex_state = 186},
  [6322] = {.lex_state = 0, .external_lex_state = 222},
  [6323] = {.lex_state = 0, .external_lex_state = 186},
  [6324] = {.lex_state = 0, .external_lex_state = 268},
  [6325] = {.lex_state = 0, .external_lex_state = 186},
  [6326] = {.lex_state = 0, .external_lex_state = 186},
  [6327] = {.lex_state = 0, .external_lex_state = 186},
  [6328] = {.lex_state = 0, .external_lex_state = 251},
  [6329] = {.lex_state = 0, .external_lex_state = 186},
  [6330] = {.lex_state = 0, .external_lex_state = 186},
  [6331] = {.lex_state = 0, .external_lex_state = 186},
  [6332] = {.lex_state = 0, .external_lex_state = 186},
  [6333] = {.lex_state = 0, .external_lex_state = 186},
  [6334] = {.lex_state = 0, .external_lex_state = 186},
  [6335] = {.lex_state = 0, .external_lex_state = 186},
  [6336] = {.lex_state = 0, .external_lex_state = 251},
  [6337] = {.lex_state = 0, .external_lex_state = 186},
  [6338] = {.lex_state = 0, .external_lex_state = 264},
  [6339] = {.lex_state = 0, .external_lex_state = 186},
  [6340] = {.lex_state = 0, .external_lex_state = 259},
  [6341] = {.lex_state = 0, .external_lex_state = 186},
  [6342] = {.lex_state = 0, .external_lex_state = 222},
  [6343] = {.lex_state = 0, .external_lex_state = 186},
  [6344] = {.lex_state = 0, .external_lex_state = 222},
  [6345] = {.lex_state = 0, .external_lex_state = 261},
  [6346] = {.lex_state = 0, .external_lex_state = 250},
  [6347] = {.lex_state = 0, .external_lex_state = 186},
  [6348] = {.lex_state = 0, .external_lex_state = 267},
  [6349] = {.lex_state = 0, .external_lex_state = 186},
  [6350] = {.lex_state = 0, .external_lex_state = 259},
  [6351] = {.lex_state = 0, .external_lex_state = 266},
  [6352] = {.lex_state = 0, .external_lex_state = 186},
  [6353] = {.lex_state = 0, .external_lex_state = 186},
  [6354] = {.lex_state = 0, .external_lex_state = 186},
  [6355] = {.lex_state = 0, .external_lex_state = 264},
  [6356] = {.lex_state = 0, .external_lex_state = 249},
  [6357] = {.lex_state = 0, .external_lex_state = 186},
  [6358] = {.lex_state = 0, .external_lex_state = 186},
  [6359] = {.lex_state = 0, .external_lex_state = 186},
  [6360] = {.lex_state = 0, .external_lex_state = 186},
  [6361] = {.lex_state = 0, .external_lex_state = 261},
  [6362] = {.lex_state = 0, .external_lex_state = 186},
  [6363] = {.lex_state = 0, .external_lex_state = 186},
  [6364] = {.lex_state = 0, .external_lex_state = 186},
  [6365] = {.lex_state = 0, .external_lex_state = 186},
  [6366] = {.lex_state = 0, .external_lex_state = 186},
  [6367] = {.lex_state = 0, .external_lex_state = 186},
  [6368] = {.lex_state = 0, .external_lex_state = 186},
  [6369] = {.lex_state = 0, .external_lex_state = 222},
  [6370] = {.lex_state = 0, .external_lex_state = 186},
  [6371] = {.lex_state = 0, .external_lex_state = 186},
  [6372] = {.lex_state = 0, .external_lex_state = 236},
  [6373] = {.lex_state = 0, .external_lex_state = 186},
  [6374] = {.lex_state = 0, .external_lex_state = 222},
  [6375] = {.lex_state = 0, .external_lex_state = 222},
  [6376] = {.lex_state = 0, .external_lex_state = 186},
  [6377] = {.lex_state = 0, .external_lex_state = 260},
  [6378] = {.lex_state = 0, .external_lex_state = 186},
  [6379] = {.lex_state = 0, .external_lex_state = 260},
  [6380] = {.lex_state = 0, .external_lex_state = 267},
  [6381] = {.lex_state = 0, .external_lex_state = 222},
  [6382] = {.lex_state = 0, .external_lex_state = 260},
  [6383] = {.lex_state = 0, .external_lex_state = 186},
  [6384] = {.lex_state = 0, .external_lex_state = 222},
  [6385] = {.lex_state = 0, .external_lex_state = 266},
  [6386] = {.lex_state = 0, .external_lex_state = 236},
  [6387] = {.lex_state = 0, .external_lex_state = 260},
  [6388] = {.lex_state = 0, .external_lex_state = 186},
  [6389] = {.lex_state = 0, .external_lex_state = 260},
  [6390] = {.lex_state = 0, .external_lex_state = 260},
  [6391] = {.lex_state = 0, .external_lex_state = 186},
  [6392] = {.lex_state = 0, .external_lex_state = 222},
  [6393] = {.lex_state = 0, .external_lex_state = 266},
  [6394] = {.lex_state = 0, .external_lex_state = 186},
  [6395] = {.lex_state = 0, .external_lex_state = 261},
  [6396] = {.lex_state = 0, .external_lex_state = 222},
  [6397] = {.lex_state = 0, .external_lex_state = 260},
  [6398] = {.lex_state = 0, .external_lex_state = 261},
  [6399] = {.lex_state = 0, .external_lex_state = 186},
  [6400] = {.lex_state = 0, .external_lex_state = 260},
  [6401] = {.lex_state = 0, .external_lex_state = 186},
  [6402] = {.lex_state = 0, .external_lex_state = 260},
  [6403] = {.lex_state = 0, .external_lex_state = 186},
  [6404] = {.lex_state = 0, .external_lex_state = 222},
  [6405] = {.lex_state = 0, .external_lex_state = 222},
  [6406] = {.lex_state = 0, .external_lex_state = 186},
  [6407] = {.lex_state = 0, .external_lex_state = 186},
  [6408] = {.lex_state = 0, .external_lex_state = 260},
  [6409] = {.lex_state = 0, .external_lex_state = 186},
  [6410] = {.lex_state = 0, .external_lex_state = 260},
  [6411] = {.lex_state = 0, .external_lex_state = 267},
  [6412] = {.lex_state = 0, .external_lex_state = 222},
  [6413] = {.lex_state = 0, .external_lex_state = 186},
  [6414] = {.lex_state = 0, .external_lex_state = 260},
  [6415] = {.lex_state = 0, .external_lex_state = 186},
  [6416] = {.lex_state = 0, .external_lex_state = 222},
  [6417] = {.lex_state = 0, .external_lex_state = 266},
  [6418] = {.lex_state = 0, .external_lex_state = 186},
  [6419] = {.lex_state = 0, .external_lex_state = 260},
  [6420] = {.lex_state = 0, .external_lex_state = 186},
  [6421] = {.lex_state = 0, .external_lex_state = 260},
  [6422] = {.lex_state = 0, .external_lex_state = 260},
  [6423] = {.lex_state = 0, .external_lex_state = 186},
  [6424] = {.lex_state = 0, .external_lex_state = 222},
  [6425] = {.lex_state = 0, .external_lex_state = 266},
  [6426] = {.lex_state = 0, .external_lex_state = 186},
  [6427] = {.lex_state = 0, .external_lex_state = 222},
  [6428] = {.lex_state = 0, .external_lex_state = 260},
  [6429] = {.lex_state = 0, .external_lex_state = 259},
  [6430] = {.lex_state = 0, .external_lex_state = 186},
  [6431] = {.lex_state = 0, .external_lex_state = 186},
  [6432] = {.lex_state = 0, .external_lex_state = 260},
  [6433] = {.lex_state = 0, .external_lex_state = 186},
  [6434] = {.lex_state = 0, .external_lex_state = 260},
  [6435] = {.lex_state = 0, .external_lex_state = 186},
  [6436] = {.lex_state = 0, .external_lex_state = 222},
  [6437] = {.lex_state = 0, .external_lex_state = 222},
  [6438] = {.lex_state = 0, .external_lex_state = 186},
  [6439] = {.lex_state = 0, .external_lex_state = 260},
  [6440] = {.lex_state = 0, .external_lex_state = 186},
  [6441] = {.lex_state = 0, .external_lex_state = 260},
  [6442] = {.lex_state = 0, .external_lex_state = 186},
  [6443] = {.lex_state = 0, .external_lex_state = 267},
  [6444] = {.lex_state = 0, .external_lex_state = 222},
  [6445] = {.lex_state = 0, .external_lex_state = 260},
  [6446] = {.lex_state = 0, .external_lex_state = 186},
  [6447] = {.lex_state = 0, .external_lex_state = 260},
  [6448] = {.lex_state = 0, .external_lex_state = 186},
  [6449] = {.lex_state = 0, .external_lex_state = 186},
  [6450] = {.lex_state = 0, .external_lex_state = 260},
  [6451] = {.lex_state = 0, .external_lex_state = 222},
  [6452] = {.lex_state = 0, .external_lex_state = 260},
  [6453] = {.lex_state = 0, .external_lex_state = 186},
  [6454] = {.lex_state = 0, .external_lex_state = 186},
  [6455] = {.lex_state = 0, .external_lex_state = 186},
  [6456] = {.lex_state = 0, .external_lex_state = 260},
  [6457] = {.lex_state = 0, .external_lex_state = 186},
  [6458] = {.lex_state = 0, .external_lex_state = 222},
  [6459] = {.lex_state = 0, .external_lex_state = 222},
  [6460] = {.lex_state = 0, .external_lex_state = 260},
  [6461] = {.lex_state = 0, .external_lex_state = 260},
  [6462] = {.lex_state = 0, .external_lex_state = 222},
  [6463] = {.lex_state = 0, .external_lex_state = 186},
  [6464] = {.lex_state = 0, .external_lex_state = 260},
  [6465] = {.lex_state = 0, .external_lex_state = 260},
  [6466] = {.lex_state = 0, .external_lex_state = 260},
  [6467] = {.lex_state = 0, .external_lex_state = 222},
  [6468] = {.lex_state = 0, .external_lex_state = 222},
  [6469] = {.lex_state = 0, .external_lex_state = 222},
  [6470] = {.lex_state = 0, .external_lex_state = 264},
  [6471] = {.lex_state = 0, .external_lex_state = 264},
  [6472] = {.lex_state = 0, .external_lex_state = 186},
  [6473] = {.lex_state = 0, .external_lex_state = 186},
  [6474] = {.lex_state = 0, .external_lex_state = 264},
  [6475] = {.lex_state = 0, .external_lex_state = 264},
  [6476] = {.lex_state = 0, .external_lex_state = 264},
  [6477] = {.lex_state = 0, .external_lex_state = 224},
  [6478] = {.lex_state = 0, .external_lex_state = 186},
  [6479] = {.lex_state = 0, .external_lex_state = 251},
  [6480] = {.lex_state = 0, .external_lex_state = 264},
  [6481] = {.lex_state = 0, .external_lex_state = 186},
  [6482] = {.lex_state = 0, .external_lex_state = 264},
  [6483] = {.lex_state = 0, .external_lex_state = 186},
  [6484] = {.lex_state = 0, .external_lex_state = 261},
  [6485] = {.lex_state = 0, .external_lex_state = 264},
  [6486] = {.lex_state = 0, .external_lex_state = 186},
  [6487] = {.lex_state = 0, .external_lex_state = 222},
  [6488] = {.lex_state = 0, .external_lex_state = 264},
  [6489] = {.lex_state = 0, .external_lex_state = 264},
  [6490] = {.lex_state = 0, .external_lex_state = 186},
  [6491] = {.lex_state = 0, .external_lex_state = 264},
  [6492] = {.lex_state = 0, .external_lex_state = 261},
  [6493] = {.lex_state = 0, .external_lex_state = 264},
  [6494] = {.lex_state = 0, .external_lex_state = 186},
  [6495] = {.lex_state = 0, .external_lex_state = 264},
  [6496] = {.lex_state = 0, .external_lex_state = 242},
  [6497] = {.lex_state = 0, .external_lex_state = 264},
  [6498] = {.lex_state = 0, .external_lex_state = 186},
  [6499] = {.lex_state = 0, .external_lex_state = 186},
  [6500] = {.lex_state = 0, .external_lex_state = 186},
  [6501] = {.lex_state = 0, .external_lex_state = 186},
  [6502] = {.lex_state = 0, .external_lex_state = 186},
  [6503] = {.lex_state = 0, .external_lex_state = 261},
  [6504] = {.lex_state = 0, .external_lex_state = 186},
  [6505] = {.lex_state = 0, .external_lex_state = 186},
  [6506] = {.lex_state = 0, .external_lex_state = 186},
  [6507] = {.lex_state = 0, .external_lex_state = 186},
  [6508] = {.lex_state = 0, .external_lex_state = 251},
  [6509] = {.lex_state = 0, .external_lex_state = 186},
  [6510] = {.lex_state = 0, .external_lex_state = 259},
  [6511] = {.lex_state = 0, .external_lex_state = 231},
  [6512] = {.lex_state = 0, .external_lex_state = 186},
  [6513] = {.lex_state = 0, .external_lex_state = 186},
  [6514] = {.lex_state = 0, .external_lex_state = 186},
  [6515] = {.lex_state = 0, .external_lex_state = 186},
  [6516] = {.lex_state = 0, .external_lex_state = 261},
  [6517] = {.lex_state = 0, .external_lex_state = 186},
  [6518] = {.lex_state = 0, .external_lex_state = 186},
  [6519] = {.lex_state = 0, .external_lex_state = 186},
  [6520] = {.lex_state = 0, .external_lex_state = 222},
  [6521] = {.lex_state = 0, .external_lex_state = 186},
  [6522] = {.lex_state = 0, .external_lex_state = 186},
  [6523] = {.lex_state = 0, .external_lex_state = 186},
  [6524] = {.lex_state = 0, .external_lex_state = 186},
  [6525] = {.lex_state = 0, .external_lex_state = 186},
  [6526] = {.lex_state = 0, .external_lex_state = 186},
  [6527] = {.lex_state = 0, .external_lex_state = 236},
  [6528] = {.lex_state = 0, .external_lex_state = 186},
  [6529] = {.lex_state = 0, .external_lex_state = 186},
  [6530] = {.lex_state = 0, .external_lex_state = 231},
  [6531] = {.lex_state = 0, .external_lex_state = 186},
  [6532] = {.lex_state = 0, .external_lex_state = 263},
  [6533] = {.lex_state = 0, .external_lex_state = 231},
  [6534] = {.lex_state = 0, .external_lex_state = 186},
  [6535] = {.lex_state = 0, .external_lex_state = 186},
  [6536] = {.lex_state = 0, .external_lex_state = 186},
  [6537] = {.lex_state = 0, .external_lex_state = 261},
  [6538] = {.lex_state = 0, .external_lex_state = 224},
  [6539] = {.lex_state = 0, .external_lex_state = 259},
  [6540] = {.lex_state = 0, .external_lex_state = 186},
  [6541] = {.lex_state = 0, .external_lex_state = 186},
  [6542] = {.lex_state = 0, .external_lex_state = 186},
  [6543] = {.lex_state = 0, .external_lex_state = 222},
  [6544] = {.lex_state = 0, .external_lex_state = 259},
  [6545] = {.lex_state = 0, .external_lex_state = 259},
  [6546] = {.lex_state = 0, .external_lex_state = 259},
  [6547] = {.lex_state = 0, .external_lex_state = 222},
  [6548] = {.lex_state = 0, .external_lex_state = 186},
  [6549] = {.lex_state = 0, .external_lex_state = 222},
  [6550] = {.lex_state = 0, .external_lex_state = 231},
  [6551] = {.lex_state = 0, .external_lex_state = 259},
  [6552] = {.lex_state = 0, .external_lex_state = 222},
  [6553] = {.lex_state = 0, .external_lex_state = 186},
  [6554] = {.lex_state = 0, .external_lex_state = 222},
  [6555] = {.lex_state = 0, .external_lex_state = 186},
  [6556] = {.lex_state = 0, .external_lex_state = 261},
  [6557] = {.lex_state = 0, .external_lex_state = 222},
  [6558] = {.lex_state = 0, .external_lex_state = 259},
  [6559] = {.lex_state = 0, .external_lex_state = 259},
  [6560] = {.lex_state = 0, .external_lex_state = 259},
  [6561] = {.lex_state = 0, .external_lex_state = 259},
  [6562] = {.lex_state = 0, .external_lex_state = 261},
  [6563] = {.lex_state = 0, .external_lex_state = 259},
  [6564] = {.lex_state = 0, .external_lex_state = 186},
  [6565] = {.lex_state = 0, .external_lex_state = 259},
  [6566] = {.lex_state = 0, .external_lex_state = 251},
  [6567] = {.lex_state = 0, .external_lex_state = 259},
  [6568] = {.lex_state = 0, .external_lex_state = 186},
  [6569] = {.lex_state = 0, .external_lex_state = 186},
  [6570] = {.lex_state = 0, .external_lex_state = 186},
  [6571] = {.lex_state = 0, .external_lex_state = 222},
  [6572] = {.lex_state = 0, .external_lex_state = 259},
  [6573] = {.lex_state = 0, .external_lex_state = 236},
  [6574] = {.lex_state = 0, .external_lex_state = 259},
  [6575] = {.lex_state = 0, .external_lex_state = 236},
  [6576] = {.lex_state = 0, .external_lex_state = 259},
  [6577] = {.lex_state = 0, .external_lex_state = 259},
  [6578] = {.lex_state = 0, .external_lex_state = 259},
  [6579] = {.lex_state = 0, .external_lex_state = 259},
  [6580] = {.lex_state = 0, .external_lex_state = 259},
  [6581] = {.lex_state = 0, .external_lex_state = 259},
  [6582] = {.lex_state = 0, .external_lex_state = 186},
  [6583] = {.lex_state = 0, .external_lex_state = 261},
  [6584] = {.lex_state = 0, .external_lex_state = 259},
  [6585] = {.lex_state = 0, .external_lex_state = 259},
  [6586] = {.lex_state = 0, .external_lex_state = 236},
  [6587] = {.lex_state = 0, .external_lex_state = 259},
  [6588] = {.lex_state = 0, .external_lex_state = 186},
  [6589] = {.lex_state = 0, .external_lex_state = 210},
  [6590] = {.lex_state = 0, .external_lex_state = 259},
  [6591] = {.lex_state = 0, .external_lex_state = 259},
  [6592] = {.lex_state = 0, .external_lex_state = 259},
  [6593] = {.lex_state = 0, .external_lex_state = 259},
  [6594] = {.lex_state = 0, .external_lex_state = 259},
  [6595] = {.lex_state = 0, .external_lex_state = 259},
  [6596] = {.lex_state = 0, .external_lex_state = 186},
  [6597] = {.lex_state = 0, .external_lex_state = 260},
  [6598] = {.lex_state = 0, .external_lex_state = 236},
  [6599] = {.lex_state = 0, .external_lex_state = 210},
  [6600] = {.lex_state = 0, .external_lex_state = 259},
  [6601] = {.lex_state = 0, .external_lex_state = 259},
  [6602] = {.lex_state = 0, .external_lex_state = 186},
  [6603] = {.lex_state = 0, .external_lex_state = 210},
  [6604] = {.lex_state = 0, .external_lex_state = 261},
  [6605] = {.lex_state = 0, .external_lex_state = 186},
  [6606] = {.lex_state = 0, .external_lex_state = 222},
  [6607] = {.lex_state = 0, .external_lex_state = 186},
  [6608] = {.lex_state = 0, .external_lex_state = 186},
  [6609] = {.lex_state = 0, .external_lex_state = 186},
  [6610] = {.lex_state = 0, .external_lex_state = 186},
  [6611] = {.lex_state = 0, .external_lex_state = 186},
  [6612] = {.lex_state = 0, .external_lex_state = 186},
  [6613] = {.lex_state = 0, .external_lex_state = 186},
  [6614] = {.lex_state = 0, .external_lex_state = 186},
  [6615] = {.lex_state = 0, .external_lex_state = 259},
  [6616] = {.lex_state = 0, .external_lex_state = 186},
  [6617] = {.lex_state = 0, .external_lex_state = 186},
  [6618] = {.lex_state = 0, .external_lex_state = 186},
  [6619] = {.lex_state = 0, .external_lex_state = 186},
  [6620] = {.lex_state = 0, .external_lex_state = 186},
  [6621] = {.lex_state = 0, .external_lex_state = 186},
  [6622] = {.lex_state = 0, .external_lex_state = 186},
  [6623] = {.lex_state = 0, .external_lex_state = 186},
  [6624] = {.lex_state = 0, .external_lex_state = 186},
  [6625] = {.lex_state = 0, .external_lex_state = 186},
  [6626] = {.lex_state = 0, .external_lex_state = 186},
  [6627] = {.lex_state = 0, .external_lex_state = 186},
  [6628] = {.lex_state = 0, .external_lex_state = 186},
  [6629] = {.lex_state = 0, .external_lex_state = 186},
  [6630] = {.lex_state = 0, .external_lex_state = 186},
  [6631] = {.lex_state = 0, .external_lex_state = 186},
  [6632] = {.lex_state = 0, .external_lex_state = 186},
  [6633] = {.lex_state = 0, .external_lex_state = 186},
  [6634] = {.lex_state = 0, .external_lex_state = 186},
  [6635] = {.lex_state = 0, .external_lex_state = 186},
  [6636] = {.lex_state = 0, .external_lex_state = 186},
  [6637] = {.lex_state = 0, .external_lex_state = 186},
  [6638] = {.lex_state = 0, .external_lex_state = 186},
  [6639] = {.lex_state = 0, .external_lex_state = 186},
  [6640] = {.lex_state = 0, .external_lex_state = 186},
  [6641] = {.lex_state = 0, .external_lex_state = 186},
  [6642] = {.lex_state = 0, .external_lex_state = 269},
  [6643] = {.lex_state = 0, .external_lex_state = 186},
  [6644] = {.lex_state = 0, .external_lex_state = 270},
  [6645] = {.lex_state = 0, .external_lex_state = 186},
  [6646] = {.lex_state = 0, .external_lex_state = 186},
  [6647] = {.lex_state = 0, .external_lex_state = 186},
  [6648] = {.lex_state = 0, .external_lex_state = 186},
  [6649] = {.lex_state = 0, .external_lex_state = 186},
  [6650] = {.lex_state = 0, .external_lex_state = 186},
  [6651] = {.lex_state = 0, .external_lex_state = 186},
  [6652] = {.lex_state = 0, .external_lex_state = 186},
  [6653] = {.lex_state = 0, .external_lex_state = 186},
  [6654] = {.lex_state = 0, .external_lex_state = 271},
  [6655] = {.lex_state = 0, .external_lex_state = 186},
  [6656] = {.lex_state = 0, .external_lex_state = 186},
  [6657] = {.lex_state = 0, .external_lex_state = 186},
  [6658] = {.lex_state = 0, .external_lex_state = 186},
  [6659] = {.lex_state = 0, .external_lex_state = 186},
  [6660] = {.lex_state = 0, .external_lex_state = 272},
  [6661] = {.lex_state = 0, .external_lex_state = 186},
  [6662] = {.lex_state = 0, .external_lex_state = 186},
  [6663] = {.lex_state = 0, .external_lex_state = 271},
  [6664] = {.lex_state = 0, .external_lex_state = 186},
  [6665] = {.lex_state = 0, .external_lex_state = 186},
  [6666] = {.lex_state = 0, .external_lex_state = 186},
  [6667] = {.lex_state = 0, .external_lex_state = 186},
  [6668] = {.lex_state = 0, .external_lex_state = 186},
  [6669] = {.lex_state = 0, .external_lex_state = 186},
  [6670] = {.lex_state = 0, .external_lex_state = 186},
  [6671] = {.lex_state = 0, .external_lex_state = 186},
  [6672] = {.lex_state = 0, .external_lex_state = 186},
  [6673] = {.lex_state = 0, .external_lex_state = 271},
  [6674] = {.lex_state = 0, .external_lex_state = 186},
  [6675] = {.lex_state = 0, .external_lex_state = 186},
  [6676] = {.lex_state = 0, .external_lex_state = 186},
  [6677] = {.lex_state = 0, .external_lex_state = 186},
  [6678] = {.lex_state = 0, .external_lex_state = 186},
  [6679] = {.lex_state = 0, .external_lex_state = 186},
  [6680] = {.lex_state = 0, .external_lex_state = 186},
  [6681] = {.lex_state = 0, .external_lex_state = 186},
  [6682] = {.lex_state = 0, .external_lex_state = 186},
  [6683] = {.lex_state = 0, .external_lex_state = 186},
  [6684] = {.lex_state = 0, .external_lex_state = 186},
  [6685] = {.lex_state = 0, .external_lex_state = 186},
  [6686] = {.lex_state = 0, .external_lex_state = 186},
  [6687] = {.lex_state = 0, .external_lex_state = 186},
  [6688] = {.lex_state = 0, .external_lex_state = 186},
  [6689] = {.lex_state = 0, .external_lex_state = 186},
  [6690] = {.lex_state = 0, .external_lex_state = 186},
  [6691] = {.lex_state = 0, .external_lex_state = 186},
  [6692] = {.lex_state = 0, .external_lex_state = 186},
  [6693] = {.lex_state = 0, .external_lex_state = 186},
  [6694] = {.lex_state = 0, .external_lex_state = 186},
  [6695] = {.lex_state = 0, .external_lex_state = 186},
  [6696] = {.lex_state = 0, .external_lex_state = 186},
  [6697] = {.lex_state = 0, .external_lex_state = 186},
  [6698] = {.lex_state = 0, .external_lex_state = 186},
  [6699] = {.lex_state = 0, .external_lex_state = 186},
  [6700] = {.lex_state = 0, .external_lex_state = 186},
  [6701] = {.lex_state = 0, .external_lex_state = 186},
  [6702] = {.lex_state = 0, .external_lex_state = 186},
  [6703] = {.lex_state = 0, .external_lex_state = 254},
  [6704] = {.lex_state = 0, .external_lex_state = 186},
  [6705] = {.lex_state = 0, .external_lex_state = 186},
  [6706] = {.lex_state = 0, .external_lex_state = 273},
  [6707] = {.lex_state = 0, .external_lex_state = 186},
  [6708] = {.lex_state = 0, .external_lex_state = 186},
  [6709] = {.lex_state = 0, .external_lex_state = 186},
  [6710] = {.lex_state = 0, .external_lex_state = 186},
  [6711] = {.lex_state = 0, .external_lex_state = 186},
  [6712] = {.lex_state = 0, .external_lex_state = 269},
  [6713] = {.lex_state = 0, .external_lex_state = 269},
  [6714] = {.lex_state = 0, .external_lex_state = 259},
  [6715] = {.lex_state = 0, .external_lex_state = 186},
  [6716] = {.lex_state = 0, .external_lex_state = 274},
  [6717] = {.lex_state = 0, .external_lex_state = 186},
  [6718] = {.lex_state = 0, .external_lex_state = 186},
  [6719] = {.lex_state = 0, .external_lex_state = 186},
  [6720] = {.lex_state = 0, .external_lex_state = 271},
  [6721] = {.lex_state = 0, .external_lex_state = 186},
  [6722] = {.lex_state = 0, .external_lex_state = 186},
  [6723] = {.lex_state = 0, .external_lex_state = 186},
  [6724] = {.lex_state = 0, .external_lex_state = 186},
  [6725] = {.lex_state = 0, .external_lex_state = 186},
  [6726] = {.lex_state = 0, .external_lex_state = 186},
  [6727] = {.lex_state = 0, .external_lex_state = 186},
  [6728] = {.lex_state = 0, .external_lex_state = 186},
  [6729] = {.lex_state = 0, .external_lex_state = 186},
  [6730] = {.lex_state = 0, .external_lex_state = 271},
  [6731] = {.lex_state = 0, .external_lex_state = 186},
  [6732] = {.lex_state = 0, .external_lex_state = 186},
  [6733] = {.lex_state = 0, .external_lex_state = 186},
  [6734] = {.lex_state = 0, .external_lex_state = 186},
  [6735] = {.lex_state = 0, .external_lex_state = 186},
  [6736] = {.lex_state = 0, .external_lex_state = 186},
  [6737] = {.lex_state = 0, .external_lex_state = 186},
  [6738] = {.lex_state = 0, .external_lex_state = 186},
  [6739] = {.lex_state = 0, .external_lex_state = 269},
  [6740] = {.lex_state = 0, .external_lex_state = 186},
  [6741] = {.lex_state = 0, .external_lex_state = 267},
  [6742] = {.lex_state = 0, .external_lex_state = 186},
  [6743] = {.lex_state = 0, .external_lex_state = 186},
  [6744] = {.lex_state = 0, .external_lex_state = 272},
  [6745] = {.lex_state = 0, .external_lex_state = 186},
  [6746] = {.lex_state = 0, .external_lex_state = 186},
  [6747] = {.lex_state = 0, .external_lex_state = 186},
  [6748] = {.lex_state = 0, .external_lex_state = 186},
  [6749] = {.lex_state = 0, .external_lex_state = 186},
  [6750] = {.lex_state = 0, .external_lex_state = 274},
  [6751] = {.lex_state = 0, .external_lex_state = 186},
  [6752] = {.lex_state = 0, .external_lex_state = 186},
  [6753] = {.lex_state = 0, .external_lex_state = 186},
  [6754] = {.lex_state = 0, .external_lex_state = 186},
  [6755] = {.lex_state = 0, .external_lex_state = 186},
  [6756] = {.lex_state = 0, .external_lex_state = 259},
  [6757] = {.lex_state = 0, .external_lex_state = 186},
  [6758] = {.lex_state = 0, .external_lex_state = 186},
  [6759] = {.lex_state = 0, .external_lex_state = 186},
  [6760] = {.lex_state = 0, .external_lex_state = 275},
  [6761] = {.lex_state = 0, .external_lex_state = 186},
  [6762] = {.lex_state = 0, .external_lex_state = 249},
  [6763] = {.lex_state = 0, .external_lex_state = 186},
  [6764] = {.lex_state = 0, .external_lex_state = 186},
  [6765] = {.lex_state = 0, .external_lex_state = 259},
  [6766] = {.lex_state = 0, .external_lex_state = 186},
  [6767] = {.lex_state = 0, .external_lex_state = 259},
  [6768] = {.lex_state = 0, .external_lex_state = 186},
  [6769] = {.lex_state = 0, .external_lex_state = 271},
  [6770] = {.lex_state = 0, .external_lex_state = 186},
  [6771] = {.lex_state = 0, .external_lex_state = 276},
  [6772] = {.lex_state = 0, .external_lex_state = 259},
  [6773] = {.lex_state = 0, .external_lex_state = 254},
  [6774] = {.lex_state = 0, .external_lex_state = 276},
  [6775] = {.lex_state = 0, .external_lex_state = 186},
  [6776] = {.lex_state = 0, .external_lex_state = 186},
  [6777] = {.lex_state = 0, .external_lex_state = 186},
  [6778] = {.lex_state = 0, .external_lex_state = 186},
  [6779] = {.lex_state = 0, .external_lex_state = 186},
  [6780] = {.lex_state = 0, .external_lex_state = 186},
  [6781] = {.lex_state = 0, .external_lex_state = 186},
  [6782] = {.lex_state = 0, .external_lex_state = 186},
  [6783] = {.lex_state = 0, .external_lex_state = 274},
  [6784] = {.lex_state = 0, .external_lex_state = 186},
  [6785] = {.lex_state = 0, .external_lex_state = 271},
  [6786] = {.lex_state = 0, .external_lex_state = 186},
  [6787] = {.lex_state = 0, .external_lex_state = 186},
  [6788] = {.lex_state = 0, .external_lex_state = 271},
  [6789] = {.lex_state = 0, .external_lex_state = 271},
  [6790] = {.lex_state = 0, .external_lex_state = 186},
  [6791] = {.lex_state = 0, .external_lex_state = 186},
  [6792] = {.lex_state = 0, .external_lex_state = 240},
  [6793] = {.lex_state = 0, .external_lex_state = 259},
  [6794] = {.lex_state = 0, .external_lex_state = 249},
  [6795] = {.lex_state = 0, .external_lex_state = 186},
  [6796] = {.lex_state = 0, .external_lex_state = 222},
  [6797] = {.lex_state = 0, .external_lex_state = 186},
  [6798] = {.lex_state = 0, .external_lex_state = 186},
  [6799] = {.lex_state = 0, .external_lex_state = 186},
  [6800] = {.lex_state = 0, .external_lex_state = 186},
  [6801] = {.lex_state = 0, .external_lex_state = 186},
  [6802] = {.lex_state = 0, .external_lex_state = 186},
  [6803] = {.lex_state = 0, .external_lex_state = 186},
  [6804] = {.lex_state = 0, .external_lex_state = 186},
  [6805] = {.lex_state = 0, .external_lex_state = 186},
  [6806] = {.lex_state = 0, .external_lex_state = 186},
  [6807] = {.lex_state = 0, .external_lex_state = 186},
  [6808] = {.lex_state = 0, .external_lex_state = 186},
  [6809] = {.lex_state = 0, .external_lex_state = 186},
  [6810] = {.lex_state = 0, .external_lex_state = 186},
  [6811] = {.lex_state = 0, .external_lex_state = 186},
  [6812] = {.lex_state = 0, .external_lex_state = 271},
  [6813] = {.lex_state = 0, .external_lex_state = 277},
  [6814] = {.lex_state = 0, .external_lex_state = 186},
  [6815] = {.lex_state = 0, .external_lex_state = 186},
  [6816] = {.lex_state = 0, .external_lex_state = 186},
  [6817] = {.lex_state = 0, .external_lex_state = 254},
  [6818] = {.lex_state = 0, .external_lex_state = 186},
  [6819] = {.lex_state = 0, .external_lex_state = 186},
  [6820] = {.lex_state = 0, .external_lex_state = 186},
  [6821] = {.lex_state = 0, .external_lex_state = 186},
  [6822] = {.lex_state = 0, .external_lex_state = 249},
  [6823] = {.lex_state = 0, .external_lex_state = 271},
  [6824] = {.lex_state = 0, .external_lex_state = 186},
  [6825] = {.lex_state = 0, .external_lex_state = 186},
  [6826] = {.lex_state = 0, .external_lex_state = 186},
  [6827] = {.lex_state = 0, .external_lex_state = 186},
  [6828] = {.lex_state = 0, .external_lex_state = 222},
  [6829] = {.lex_state = 0, .external_lex_state = 186},
  [6830] = {.lex_state = 0, .external_lex_state = 186},
  [6831] = {.lex_state = 0, .external_lex_state = 186},
  [6832] = {.lex_state = 0, .external_lex_state = 273},
  [6833] = {.lex_state = 0, .external_lex_state = 186},
  [6834] = {.lex_state = 0, .external_lex_state = 186},
  [6835] = {.lex_state = 0, .external_lex_state = 271},
  [6836] = {.lex_state = 0, .external_lex_state = 269},
  [6837] = {.lex_state = 0, .external_lex_state = 269},
  [6838] = {.lex_state = 0, .external_lex_state = 222},
  [6839] = {.lex_state = 0, .external_lex_state = 186},
  [6840] = {.lex_state = 0, .external_lex_state = 186},
  [6841] = {.lex_state = 0, .external_lex_state = 186},
  [6842] = {.lex_state = 0, .external_lex_state = 186},
  [6843] = {.lex_state = 0, .external_lex_state = 186},
  [6844] = {.lex_state = 0, .external_lex_state = 186},
  [6845] = {.lex_state = 0, .external_lex_state = 186},
  [6846] = {.lex_state = 0, .external_lex_state = 186},
  [6847] = {.lex_state = 0, .external_lex_state = 186},
  [6848] = {.lex_state = 0, .external_lex_state = 186},
  [6849] = {.lex_state = 0, .external_lex_state = 186},
  [6850] = {.lex_state = 0, .external_lex_state = 186},
  [6851] = {.lex_state = 0, .external_lex_state = 186},
  [6852] = {.lex_state = 0, .external_lex_state = 186},
  [6853] = {.lex_state = 0, .external_lex_state = 186},
  [6854] = {.lex_state = 0, .external_lex_state = 186},
  [6855] = {.lex_state = 0, .external_lex_state = 186},
  [6856] = {.lex_state = 0, .external_lex_state = 186},
  [6857] = {.lex_state = 0, .external_lex_state = 186},
  [6858] = {.lex_state = 0, .external_lex_state = 186},
  [6859] = {.lex_state = 0, .external_lex_state = 186},
  [6860] = {.lex_state = 0, .external_lex_state = 186},
  [6861] = {.lex_state = 0, .external_lex_state = 269},
  [6862] = {.lex_state = 0, .external_lex_state = 186},
  [6863] = {.lex_state = 0, .external_lex_state = 186},
  [6864] = {.lex_state = 0, .external_lex_state = 271},
  [6865] = {.lex_state = 0, .external_lex_state = 263},
  [6866] = {.lex_state = 0, .external_lex_state = 186},
  [6867] = {.lex_state = 0, .external_lex_state = 186},
  [6868] = {.lex_state = 0, .external_lex_state = 240},
  [6869] = {.lex_state = 0, .external_lex_state = 186},
  [6870] = {.lex_state = 0, .external_lex_state = 273},
  [6871] = {.lex_state = 0, .external_lex_state = 271},
  [6872] = {.lex_state = 0, .external_lex_state = 186},
  [6873] = {.lex_state = 0, .external_lex_state = 186},
  [6874] = {.lex_state = 0, .external_lex_state = 186},
  [6875] = {.lex_state = 0, .external_lex_state = 186},
  [6876] = {.lex_state = 0, .external_lex_state = 186},
  [6877] = {.lex_state = 0, .external_lex_state = 186},
  [6878] = {.lex_state = 0, .external_lex_state = 186},
  [6879] = {.lex_state = 0, .external_lex_state = 186},
  [6880] = {.lex_state = 0, .external_lex_state = 186},
  [6881] = {.lex_state = 0, .external_lex_state = 186},
  [6882] = {.lex_state = 0, .external_lex_state = 186},
  [6883] = {.lex_state = 0, .external_lex_state = 186},
  [6884] = {.lex_state = 0, .external_lex_state = 186},
  [6885] = {.lex_state = 0, .external_lex_state = 186},
  [6886] = {.lex_state = 0, .external_lex_state = 266},
  [6887] = {.lex_state = 0, .external_lex_state = 271},
  [6888] = {.lex_state = 0, .external_lex_state = 186},
  [6889] = {.lex_state = 0, .external_lex_state = 186},
  [6890] = {.lex_state = 0, .external_lex_state = 186},
  [6891] = {.lex_state = 0, .external_lex_state = 186},
  [6892] = {.lex_state = 0, .external_lex_state = 186},
  [6893] = {.lex_state = 0, .external_lex_state = 186},
  [6894] = {.lex_state = 0, .external_lex_state = 186},
  [6895] = {.lex_state = 0, .external_lex_state = 186},
  [6896] = {.lex_state = 0, .external_lex_state = 186},
  [6897] = {.lex_state = 0, .external_lex_state = 186},
  [6898] = {.lex_state = 0, .external_lex_state = 186},
  [6899] = {.lex_state = 0, .external_lex_state = 186},
  [6900] = {.lex_state = 0, .external_lex_state = 186},
  [6901] = {.lex_state = 0, .external_lex_state = 259},
  [6902] = {.lex_state = 0, .external_lex_state = 222},
  [6903] = {.lex_state = 0, .external_lex_state = 186},
  [6904] = {.lex_state = 0, .external_lex_state = 186},
  [6905] = {.lex_state = 0, .external_lex_state = 271},
  [6906] = {.lex_state = 0, .external_lex_state = 186},
  [6907] = {.lex_state = 0, .external_lex_state = 186},
  [6908] = {.lex_state = 0, .external_lex_state = 186},
  [6909] = {.lex_state = 0, .external_lex_state = 186},
  [6910] = {.lex_state = 0, .external_lex_state = 271},
  [6911] = {.lex_state = 0, .external_lex_state = 186},
  [6912] = {.lex_state = 0, .external_lex_state = 186},
  [6913] = {.lex_state = 0, .external_lex_state = 186},
  [6914] = {.lex_state = 0, .external_lex_state = 186},
  [6915] = {.lex_state = 0, .external_lex_state = 271},
  [6916] = {.lex_state = 0, .external_lex_state = 186},
  [6917] = {.lex_state = 0, .external_lex_state = 186},
  [6918] = {.lex_state = 0, .external_lex_state = 186},
  [6919] = {.lex_state = 0, .external_lex_state = 186},
  [6920] = {.lex_state = 0, .external_lex_state = 186},
  [6921] = {.lex_state = 0, .external_lex_state = 259},
  [6922] = {.lex_state = 0, .external_lex_state = 186},
  [6923] = {.lex_state = 0, .external_lex_state = 271},
  [6924] = {.lex_state = 0, .external_lex_state = 186},
  [6925] = {.lex_state = 0, .external_lex_state = 186},
  [6926] = {.lex_state = 0, .external_lex_state = 186},
  [6927] = {.lex_state = 0, .external_lex_state = 186},
  [6928] = {.lex_state = 0, .external_lex_state = 186},
  [6929] = {.lex_state = 0, .external_lex_state = 186},
  [6930] = {.lex_state = 0, .external_lex_state = 186},
  [6931] = {.lex_state = 0, .external_lex_state = 186},
  [6932] = {.lex_state = 0, .external_lex_state = 186},
  [6933] = {.lex_state = 0, .external_lex_state = 186},
  [6934] = {.lex_state = 0, .external_lex_state = 186},
  [6935] = {.lex_state = 0, .external_lex_state = 186},
  [6936] = {.lex_state = 0, .external_lex_state = 186},
  [6937] = {.lex_state = 0, .external_lex_state = 249},
  [6938] = {.lex_state = 0, .external_lex_state = 186},
  [6939] = {.lex_state = 0, .external_lex_state = 186},
  [6940] = {.lex_state = 0, .external_lex_state = 186},
  [6941] = {.lex_state = 0, .external_lex_state = 273},
  [6942] = {.lex_state = 0, .external_lex_state = 186},
  [6943] = {.lex_state = 0, .external_lex_state = 186},
  [6944] = {.lex_state = 0, .external_lex_state = 269},
  [6945] = {.lex_state = 0, .external_lex_state = 269},
  [6946] = {.lex_state = 0, .external_lex_state = 278},
  [6947] = {.lex_state = 0, .external_lex_state = 186},
  [6948] = {.lex_state = 0, .external_lex_state = 249},
  [6949] = {.lex_state = 0, .external_lex_state = 266},
  [6950] = {.lex_state = 0, .external_lex_state = 260},
  [6951] = {.lex_state = 0, .external_lex_state = 186},
  [6952] = {.lex_state = 0, .external_lex_state = 186},
  [6953] = {.lex_state = 0, .external_lex_state = 271},
  [6954] = {.lex_state = 0, .external_lex_state = 186},
  [6955] = {.lex_state = 0, .external_lex_state = 259},
  [6956] = {.lex_state = 0, .external_lex_state = 186},
  [6957] = {.lex_state = 0, .external_lex_state = 186},
  [6958] = {.lex_state = 0, .external_lex_state = 271},
  [6959] = {.lex_state = 0, .external_lex_state = 186},
  [6960] = {.lex_state = 0, .external_lex_state = 186},
  [6961] = {.lex_state = 0, .external_lex_state = 186},
  [6962] = {.lex_state = 0, .external_lex_state = 269},
  [6963] = {.lex_state = 0, .external_lex_state = 186},
  [6964] = {.lex_state = 0, .external_lex_state = 186},
  [6965] = {.lex_state = 0, .external_lex_state = 186},
  [6966] = {.lex_state = 0, .external_lex_state = 186},
  [6967] = {.lex_state = 0, .external_lex_state = 186},
  [6968] = {.lex_state = 0, .external_lex_state = 186},
  [6969] = {.lex_state = 0, .external_lex_state = 186},
  [6970] = {.lex_state = 0, .external_lex_state = 186},
  [6971] = {.lex_state = 0, .external_lex_state = 186},
  [6972] = {.lex_state = 0, .external_lex_state = 186},
  [6973] = {.lex_state = 0, .external_lex_state = 186},
  [6974] = {.lex_state = 0, .external_lex_state = 186},
  [6975] = {.lex_state = 0, .external_lex_state = 186},
  [6976] = {.lex_state = 0, .external_lex_state = 186},
  [6977] = {.lex_state = 0, .external_lex_state = 274},
  [6978] = {.lex_state = 0, .external_lex_state = 186},
  [6979] = {.lex_state = 0, .external_lex_state = 186},
  [6980] = {.lex_state = 0, .external_lex_state = 186},
  [6981] = {.lex_state = 0, .external_lex_state = 186},
  [6982] = {.lex_state = 0, .external_lex_state = 186},
  [6983] = {.lex_state = 0, .external_lex_state = 186},
  [6984] = {.lex_state = 0, .external_lex_state = 186},
  [6985] = {.lex_state = 0, .external_lex_state = 186},
  [6986] = {.lex_state = 0, .external_lex_state = 186},
  [6987] = {.lex_state = 0, .external_lex_state = 186},
  [6988] = {.lex_state = 0, .external_lex_state = 186},
  [6989] = {.lex_state = 0, .external_lex_state = 186},
  [6990] = {.lex_state = 0, .external_lex_state = 186},
  [6991] = {.lex_state = 0, .external_lex_state = 186},
  [6992] = {.lex_state = 0, .external_lex_state = 271},
  [6993] = {.lex_state = 0, .external_lex_state = 279},
  [6994] = {.lex_state = 0, .external_lex_state = 186},
  [6995] = {.lex_state = 0, .external_lex_state = 269},
  [6996] = {.lex_state = 0, .external_lex_state = 186},
  [6997] = {.lex_state = 0, .external_lex_state = 186},
  [6998] = {.lex_state = 0, .external_lex_state = 259},
  [6999] = {.lex_state = 0, .external_lex_state = 271},
  [7000] = {.lex_state = 0, .external_lex_state = 186},
  [7001] = {.lex_state = 0, .external_lex_state = 186},
  [7002] = {.lex_state = 0, .external_lex_state = 186},
  [7003] = {.lex_state = 0, .external_lex_state = 280},
  [7004] = {.lex_state = 0, .external_lex_state = 271},
  [7005] = {.lex_state = 0, .external_lex_state = 186},
  [7006] = {.lex_state = 0, .external_lex_state = 186},
  [7007] = {.lex_state = 0, .external_lex_state = 186},
  [7008] = {.lex_state = 0, .external_lex_state = 186},
  [7009] = {.lex_state = 0, .external_lex_state = 186},
  [7010] = {.lex_state = 0, .external_lex_state = 273},
  [7011] = {.lex_state = 0, .external_lex_state = 186},
  [7012] = {.lex_state = 0, .external_lex_state = 186},
  [7013] = {.lex_state = 0, .external_lex_state = 269},
  [7014] = {.lex_state = 0, .external_lex_state = 269},
  [7015] = {.lex_state = 0, .external_lex_state = 269},
  [7016] = {.lex_state = 0, .external_lex_state = 249},
  [7017] = {.lex_state = 0, .external_lex_state = 186},
  [7018] = {.lex_state = 0, .external_lex_state = 186},
  [7019] = {.lex_state = 0, .external_lex_state = 186},
  [7020] = {.lex_state = 0, .external_lex_state = 186},
  [7021] = {.lex_state = 0, .external_lex_state = 186},
  [7022] = {.lex_state = 0, .external_lex_state = 186},
  [7023] = {.lex_state = 0, .external_lex_state = 186},
  [7024] = {.lex_state = 0, .external_lex_state = 281},
  [7025] = {.lex_state = 0, .external_lex_state = 269},
  [7026] = {.lex_state = 0, .external_lex_state = 271},
  [7027] = {.lex_state = 0, .external_lex_state = 186},
  [7028] = {.lex_state = 0, .external_lex_state = 186},
  [7029] = {.lex_state = 0, .external_lex_state = 282},
  [7030] = {.lex_state = 0, .external_lex_state = 186},
  [7031] = {.lex_state = 0, .external_lex_state = 186},
  [7032] = {.lex_state = 0, .external_lex_state = 186},
  [7033] = {.lex_state = 0, .external_lex_state = 186},
  [7034] = {.lex_state = 0, .external_lex_state = 186},
  [7035] = {.lex_state = 0, .external_lex_state = 271},
  [7036] = {.lex_state = 0, .external_lex_state = 186},
  [7037] = {.lex_state = 0, .external_lex_state = 259},
  [7038] = {.lex_state = 0, .external_lex_state = 283},
  [7039] = {.lex_state = 0, .external_lex_state = 186},
  [7040] = {.lex_state = 0, .external_lex_state = 186},
  [7041] = {.lex_state = 0, .external_lex_state = 186},
  [7042] = {.lex_state = 0, .external_lex_state = 186},
  [7043] = {.lex_state = 0, .external_lex_state = 259},
  [7044] = {.lex_state = 0, .external_lex_state = 269},
  [7045] = {.lex_state = 0, .external_lex_state = 186},
  [7046] = {.lex_state = 0, .external_lex_state = 186},
  [7047] = {.lex_state = 0, .external_lex_state = 186},
  [7048] = {.lex_state = 0, .external_lex_state = 271},
  [7049] = {.lex_state = 0, .external_lex_state = 273},
  [7050] = {.lex_state = 0, .external_lex_state = 186},
  [7051] = {.lex_state = 0, .external_lex_state = 269},
  [7052] = {.lex_state = 0, .external_lex_state = 269},
  [7053] = {.lex_state = 0, .external_lex_state = 186},
  [7054] = {.lex_state = 0, .external_lex_state = 186},
  [7055] = {.lex_state = 0, .external_lex_state = 186},
  [7056] = {.lex_state = 0, .external_lex_state = 186},
  [7057] = {.lex_state = 0, .external_lex_state = 284},
  [7058] = {.lex_state = 0, .external_lex_state = 222},
  [7059] = {.lex_state = 0, .external_lex_state = 186},
  [7060] = {.lex_state = 0, .external_lex_state = 186},
  [7061] = {.lex_state = 0, .external_lex_state = 269},
  [7062] = {.lex_state = 0, .external_lex_state = 186},
  [7063] = {.lex_state = 0, .external_lex_state = 186},
  [7064] = {.lex_state = 0, .external_lex_state = 186},
  [7065] = {.lex_state = 0, .external_lex_state = 186},
  [7066] = {.lex_state = 0, .external_lex_state = 255},
  [7067] = {.lex_state = 0, .external_lex_state = 285},
  [7068] = {.lex_state = 0, .external_lex_state = 186},
  [7069] = {.lex_state = 0, .external_lex_state = 222},
  [7070] = {.lex_state = 0, .external_lex_state = 186},
  [7071] = {.lex_state = 0, .external_lex_state = 186},
  [7072] = {.lex_state = 0, .external_lex_state = 186},
  [7073] = {.lex_state = 0, .external_lex_state = 186},
  [7074] = {.lex_state = 0, .external_lex_state = 186},
  [7075] = {.lex_state = 0, .external_lex_state = 186},
  [7076] = {.lex_state = 0, .external_lex_state = 186},
  [7077] = {.lex_state = 0, .external_lex_state = 273},
  [7078] = {.lex_state = 0, .external_lex_state = 186},
  [7079] = {.lex_state = 0, .external_lex_state = 269},
  [7080] = {.lex_state = 0, .external_lex_state = 269},
  [7081] = {.lex_state = 0, .external_lex_state = 186},
  [7082] = {.lex_state = 0, .external_lex_state = 186},
  [7083] = {.lex_state = 0, .external_lex_state = 186},
  [7084] = {.lex_state = 0, .external_lex_state = 269},
  [7085] = {.lex_state = 0, .external_lex_state = 186},
  [7086] = {.lex_state = 0, .external_lex_state = 186},
  [7087] = {.lex_state = 0, .external_lex_state = 222},
  [7088] = {.lex_state = 0, .external_lex_state = 186},
  [7089] = {.lex_state = 0, .external_lex_state = 186},
  [7090] = {.lex_state = 0, .external_lex_state = 186},
  [7091] = {.lex_state = 0, .external_lex_state = 186},
  [7092] = {.lex_state = 0, .external_lex_state = 186},
  [7093] = {.lex_state = 0, .external_lex_state = 280},
  [7094] = {.lex_state = 0, .external_lex_state = 186},
  [7095] = {.lex_state = 0, .external_lex_state = 186},
  [7096] = {.lex_state = 0, .external_lex_state = 186},
  [7097] = {.lex_state = 0, .external_lex_state = 286},
  [7098] = {.lex_state = 0, .external_lex_state = 273},
  [7099] = {.lex_state = 0, .external_lex_state = 186},
  [7100] = {.lex_state = 0, .external_lex_state = 269},
  [7101] = {.lex_state = 0, .external_lex_state = 269},
  [7102] = {.lex_state = 0, .external_lex_state = 266},
  [7103] = {.lex_state = 0, .external_lex_state = 269},
  [7104] = {.lex_state = 0, .external_lex_state = 186},
  [7105] = {.lex_state = 0, .external_lex_state = 271},
  [7106] = {.lex_state = 0, .external_lex_state = 186},
  [7107] = {.lex_state = 0, .external_lex_state = 186},
  [7108] = {.lex_state = 0, .external_lex_state = 186},
  [7109] = {.lex_state = 0, .external_lex_state = 240},
  [7110] = {.lex_state = 0, .external_lex_state = 186},
  [7111] = {.lex_state = 0, .external_lex_state = 186},
  [7112] = {.lex_state = 0, .external_lex_state = 273},
  [7113] = {.lex_state = 0, .external_lex_state = 186},
  [7114] = {.lex_state = 0, .external_lex_state = 269},
  [7115] = {.lex_state = 0, .external_lex_state = 269},
  [7116] = {.lex_state = 0, .external_lex_state = 186},
  [7117] = {.lex_state = 0, .external_lex_state = 269},
  [7118] = {.lex_state = 0, .external_lex_state = 271},
  [7119] = {.lex_state = 0, .external_lex_state = 286},
  [7120] = {.lex_state = 0, .external_lex_state = 186},
  [7121] = {.lex_state = 0, .external_lex_state = 186},
  [7122] = {.lex_state = 0, .external_lex_state = 186},
  [7123] = {.lex_state = 0, .external_lex_state = 186},
  [7124] = {.lex_state = 0, .external_lex_state = 186},
  [7125] = {.lex_state = 0, .external_lex_state = 273},
  [7126] = {.lex_state = 0, .external_lex_state = 186},
  [7127] = {.lex_state = 0, .external_lex_state = 269},
  [7128] = {.lex_state = 0, .external_lex_state = 269},
  [7129] = {.lex_state = 0, .external_lex_state = 186},
  [7130] = {.lex_state = 0, .external_lex_state = 269},
  [7131] = {.lex_state = 0, .external_lex_state = 186},
  [7132] = {.lex_state = 0, .external_lex_state = 186},
  [7133] = {.lex_state = 0, .external_lex_state = 186},
  [7134] = {.lex_state = 0, .external_lex_state = 274},
  [7135] = {.lex_state = 0, .external_lex_state = 186},
  [7136] = {.lex_state = 0, .external_lex_state = 186},
  [7137] = {.lex_state = 0, .external_lex_state = 273},
  [7138] = {.lex_state = 0, .external_lex_state = 186},
  [7139] = {.lex_state = 0, .external_lex_state = 269},
  [7140] = {.lex_state = 0, .external_lex_state = 269},
  [7141] = {.lex_state = 0, .external_lex_state = 186},
  [7142] = {.lex_state = 0, .external_lex_state = 269},
  [7143] = {.lex_state = 0, .external_lex_state = 186},
  [7144] = {.lex_state = 0, .external_lex_state = 249},
  [7145] = {.lex_state = 0, .external_lex_state = 240},
  [7146] = {.lex_state = 0, .external_lex_state = 186},
  [7147] = {.lex_state = 0, .external_lex_state = 186},
  [7148] = {.lex_state = 0, .external_lex_state = 271},
  [7149] = {.lex_state = 0, .external_lex_state = 273},
  [7150] = {.lex_state = 0, .external_lex_state = 186},
  [7151] = {.lex_state = 0, .external_lex_state = 269},
  [7152] = {.lex_state = 0, .external_lex_state = 269},
  [7153] = {.lex_state = 0, .external_lex_state = 186},
  [7154] = {.lex_state = 0, .external_lex_state = 269},
  [7155] = {.lex_state = 0, .external_lex_state = 186},
  [7156] = {.lex_state = 0, .external_lex_state = 186},
  [7157] = {.lex_state = 0, .external_lex_state = 186},
  [7158] = {.lex_state = 0, .external_lex_state = 186},
  [7159] = {.lex_state = 0, .external_lex_state = 267},
  [7160] = {.lex_state = 0, .external_lex_state = 186},
  [7161] = {.lex_state = 0, .external_lex_state = 273},
  [7162] = {.lex_state = 0, .external_lex_state = 271},
  [7163] = {.lex_state = 0, .external_lex_state = 269},
  [7164] = {.lex_state = 0, .external_lex_state = 269},
  [7165] = {.lex_state = 0, .external_lex_state = 186},
  [7166] = {.lex_state = 0, .external_lex_state = 269},
  [7167] = {.lex_state = 0, .external_lex_state = 259},
  [7168] = {.lex_state = 0, .external_lex_state = 259},
  [7169] = {.lex_state = 0, .external_lex_state = 186},
  [7170] = {.lex_state = 0, .external_lex_state = 222},
  [7171] = {.lex_state = 0, .external_lex_state = 186},
  [7172] = {.lex_state = 0, .external_lex_state = 240},
  [7173] = {.lex_state = 0, .external_lex_state = 273},
  [7174] = {.lex_state = 0, .external_lex_state = 240},
  [7175] = {.lex_state = 0, .external_lex_state = 269},
  [7176] = {.lex_state = 0, .external_lex_state = 269},
  [7177] = {.lex_state = 0, .external_lex_state = 186},
  [7178] = {.lex_state = 0, .external_lex_state = 269},
  [7179] = {.lex_state = 0, .external_lex_state = 186},
  [7180] = {.lex_state = 0, .external_lex_state = 271},
  [7181] = {.lex_state = 0, .external_lex_state = 186},
  [7182] = {.lex_state = 0, .external_lex_state = 186},
  [7183] = {.lex_state = 0, .external_lex_state = 259},
  [7184] = {.lex_state = 0, .external_lex_state = 186},
  [7185] = {.lex_state = 0, .external_lex_state = 273},
  [7186] = {.lex_state = 0, .external_lex_state = 186},
  [7187] = {.lex_state = 0, .external_lex_state = 269},
  [7188] = {.lex_state = 0, .external_lex_state = 269},
  [7189] = {.lex_state = 0, .external_lex_state = 240},
  [7190] = {.lex_state = 0, .external_lex_state = 269},
  [7191] = {.lex_state = 0, .external_lex_state = 271},
  [7192] = {.lex_state = 0, .external_lex_state = 186},
  [7193] = {.lex_state = 0, .external_lex_state = 186},
  [7194] = {.lex_state = 0, .external_lex_state = 186},
  [7195] = {.lex_state = 0, .external_lex_state = 186},
  [7196] = {.lex_state = 0, .external_lex_state = 186},
  [7197] = {.lex_state = 0, .external_lex_state = 273},
  [7198] = {.lex_state = 0, .external_lex_state = 186},
  [7199] = {.lex_state = 0, .external_lex_state = 269},
  [7200] = {.lex_state = 0, .external_lex_state = 269},
  [7201] = {.lex_state = 0, .external_lex_state = 186},
  [7202] = {.lex_state = 0, .external_lex_state = 269},
  [7203] = {.lex_state = 0, .external_lex_state = 186},
  [7204] = {.lex_state = 0, .external_lex_state = 186},
  [7205] = {.lex_state = 0, .external_lex_state = 186},
  [7206] = {.lex_state = 0, .external_lex_state = 222},
  [7207] = {.lex_state = 0, .external_lex_state = 186},
  [7208] = {.lex_state = 0, .external_lex_state = 186},
  [7209] = {.lex_state = 0, .external_lex_state = 273},
  [7210] = {.lex_state = 0, .external_lex_state = 186},
  [7211] = {.lex_state = 0, .external_lex_state = 269},
  [7212] = {.lex_state = 0, .external_lex_state = 269},
  [7213] = {.lex_state = 0, .external_lex_state = 222},
  [7214] = {.lex_state = 0, .external_lex_state = 269},
  [7215] = {.lex_state = 0, .external_lex_state = 186},
  [7216] = {.lex_state = 0, .external_lex_state = 186},
  [7217] = {.lex_state = 0, .external_lex_state = 271},
  [7218] = {.lex_state = 0, .external_lex_state = 186},
  [7219] = {.lex_state = 0, .external_lex_state = 186},
  [7220] = {.lex_state = 0, .external_lex_state = 222},
  [7221] = {.lex_state = 0, .external_lex_state = 273},
  [7222] = {.lex_state = 0, .external_lex_state = 186},
  [7223] = {.lex_state = 0, .external_lex_state = 269},
  [7224] = {.lex_state = 0, .external_lex_state = 269},
  [7225] = {.lex_state = 0, .external_lex_state = 186},
  [7226] = {.lex_state = 0, .external_lex_state = 269},
  [7227] = {.lex_state = 0, .external_lex_state = 222},
  [7228] = {.lex_state = 0, .external_lex_state = 186},
  [7229] = {.lex_state = 0, .external_lex_state = 271},
  [7230] = {.lex_state = 0, .external_lex_state = 186},
  [7231] = {.lex_state = 0, .external_lex_state = 186},
  [7232] = {.lex_state = 0, .external_lex_state = 249},
  [7233] = {.lex_state = 0, .external_lex_state = 273},
  [7234] = {.lex_state = 0, .external_lex_state = 186},
  [7235] = {.lex_state = 0, .external_lex_state = 269},
  [7236] = {.lex_state = 0, .external_lex_state = 269},
  [7237] = {.lex_state = 0, .external_lex_state = 186},
  [7238] = {.lex_state = 0, .external_lex_state = 269},
  [7239] = {.lex_state = 0, .external_lex_state = 259},
  [7240] = {.lex_state = 0, .external_lex_state = 186},
  [7241] = {.lex_state = 0, .external_lex_state = 186},
  [7242] = {.lex_state = 0, .external_lex_state = 249},
  [7243] = {.lex_state = 0, .external_lex_state = 186},
  [7244] = {.lex_state = 0, .external_lex_state = 186},
  [7245] = {.lex_state = 0, .external_lex_state = 273},
  [7246] = {.lex_state = 0, .external_lex_state = 186},
  [7247] = {.lex_state = 0, .external_lex_state = 269},
  [7248] = {.lex_state = 0, .external_lex_state = 269},
  [7249] = {.lex_state = 0, .external_lex_state = 186},
  [7250] = {.lex_state = 0, .external_lex_state = 269},
  [7251] = {.lex_state = 0, .external_lex_state = 186},
  [7252] = {.lex_state = 0, .external_lex_state = 186},
  [7253] = {.lex_state = 0, .external_lex_state = 186},
  [7254] = {.lex_state = 0, .external_lex_state = 186},
  [7255] = {.lex_state = 0, .external_lex_state = 186},
  [7256] = {.lex_state = 0, .external_lex_state = 186},
  [7257] = {.lex_state = 0, .external_lex_state = 273},
  [7258] = {.lex_state = 0, .external_lex_state = 186},
  [7259] = {.lex_state = 0, .external_lex_state = 269},
  [7260] = {.lex_state = 0, .external_lex_state = 269},
  [7261] = {.lex_state = 0, .external_lex_state = 186},
  [7262] = {.lex_state = 0, .external_lex_state = 269},
  [7263] = {.lex_state = 0, .external_lex_state = 186},
  [7264] = {.lex_state = 0, .external_lex_state = 186},
  [7265] = {.lex_state = 0, .external_lex_state = 186},
  [7266] = {.lex_state = 0, .external_lex_state = 186},
  [7267] = {.lex_state = 0, .external_lex_state = 186},
  [7268] = {.lex_state = 0, .external_lex_state = 267},
  [7269] = {.lex_state = 0, .external_lex_state = 273},
  [7270] = {.lex_state = 0, .external_lex_state = 186},
  [7271] = {.lex_state = 0, .external_lex_state = 269},
  [7272] = {.lex_state = 0, .external_lex_state = 269},
  [7273] = {.lex_state = 0, .external_lex_state = 276},
  [7274] = {.lex_state = 0, .external_lex_state = 269},
  [7275] = {.lex_state = 0, .external_lex_state = 249},
  [7276] = {.lex_state = 0, .external_lex_state = 186},
  [7277] = {.lex_state = 0, .external_lex_state = 186},
  [7278] = {.lex_state = 0, .external_lex_state = 186},
  [7279] = {.lex_state = 0, .external_lex_state = 186},
  [7280] = {.lex_state = 0, .external_lex_state = 186},
  [7281] = {.lex_state = 0, .external_lex_state = 273},
  [7282] = {.lex_state = 0, .external_lex_state = 186},
  [7283] = {.lex_state = 0, .external_lex_state = 269},
  [7284] = {.lex_state = 0, .external_lex_state = 269},
  [7285] = {.lex_state = 0, .external_lex_state = 271},
  [7286] = {.lex_state = 0, .external_lex_state = 269},
  [7287] = {.lex_state = 0, .external_lex_state = 186},
  [7288] = {.lex_state = 0, .external_lex_state = 186},
  [7289] = {.lex_state = 0, .external_lex_state = 186},
  [7290] = {.lex_state = 0, .external_lex_state = 186},
  [7291] = {.lex_state = 0, .external_lex_state = 186},
  [7292] = {.lex_state = 0, .external_lex_state = 186},
  [7293] = {.lex_state = 0, .external_lex_state = 273},
  [7294] = {.lex_state = 0, .external_lex_state = 186},
  [7295] = {.lex_state = 0, .external_lex_state = 269},
  [7296] = {.lex_state = 0, .external_lex_state = 269},
  [7297] = {.lex_state = 0, .external_lex_state = 186},
  [7298] = {.lex_state = 0, .external_lex_state = 269},
  [7299] = {.lex_state = 0, .external_lex_state = 186},
  [7300] = {.lex_state = 0, .external_lex_state = 186},
  [7301] = {.lex_state = 0, .external_lex_state = 186},
  [7302] = {.lex_state = 0, .external_lex_state = 186},
  [7303] = {.lex_state = 0, .external_lex_state = 186},
  [7304] = {.lex_state = 0, .external_lex_state = 186},
  [7305] = {.lex_state = 0, .external_lex_state = 273},
  [7306] = {.lex_state = 0, .external_lex_state = 186},
  [7307] = {.lex_state = 0, .external_lex_state = 269},
  [7308] = {.lex_state = 0, .external_lex_state = 269},
  [7309] = {.lex_state = 0, .external_lex_state = 186},
  [7310] = {.lex_state = 0, .external_lex_state = 269},
  [7311] = {.lex_state = 0, .external_lex_state = 254},
  [7312] = {.lex_state = 0, .external_lex_state = 186},
  [7313] = {.lex_state = 0, .external_lex_state = 240},
  [7314] = {.lex_state = 0, .external_lex_state = 186},
  [7315] = {.lex_state = 0, .external_lex_state = 271},
  [7316] = {.lex_state = 0, .external_lex_state = 186},
  [7317] = {.lex_state = 0, .external_lex_state = 273},
  [7318] = {.lex_state = 0, .external_lex_state = 186},
  [7319] = {.lex_state = 0, .external_lex_state = 269},
  [7320] = {.lex_state = 0, .external_lex_state = 269},
  [7321] = {.lex_state = 0, .external_lex_state = 186},
  [7322] = {.lex_state = 0, .external_lex_state = 269},
  [7323] = {.lex_state = 0, .external_lex_state = 186},
  [7324] = {.lex_state = 0, .external_lex_state = 186},
  [7325] = {.lex_state = 0, .external_lex_state = 186},
  [7326] = {.lex_state = 0, .external_lex_state = 186},
  [7327] = {.lex_state = 0, .external_lex_state = 186},
  [7328] = {.lex_state = 0, .external_lex_state = 186},
  [7329] = {.lex_state = 0, .external_lex_state = 273},
  [7330] = {.lex_state = 0, .external_lex_state = 186},
  [7331] = {.lex_state = 0, .external_lex_state = 269},
  [7332] = {.lex_state = 0, .external_lex_state = 269},
  [7333] = {.lex_state = 0, .external_lex_state = 186},
  [7334] = {.lex_state = 0, .external_lex_state = 269},
  [7335] = {.lex_state = 0, .external_lex_state = 186},
  [7336] = {.lex_state = 0, .external_lex_state = 240},
  [7337] = {.lex_state = 0, .external_lex_state = 186},
  [7338] = {.lex_state = 0, .external_lex_state = 186},
  [7339] = {.lex_state = 0, .external_lex_state = 186},
  [7340] = {.lex_state = 0, .external_lex_state = 186},
  [7341] = {.lex_state = 0, .external_lex_state = 273},
  [7342] = {.lex_state = 0, .external_lex_state = 284},
  [7343] = {.lex_state = 0, .external_lex_state = 269},
  [7344] = {.lex_state = 0, .external_lex_state = 269},
  [7345] = {.lex_state = 0, .external_lex_state = 186},
  [7346] = {.lex_state = 0, .external_lex_state = 269},
  [7347] = {.lex_state = 0, .external_lex_state = 186},
  [7348] = {.lex_state = 0, .external_lex_state = 186},
  [7349] = {.lex_state = 0, .external_lex_state = 186},
  [7350] = {.lex_state = 0, .external_lex_state = 249},
  [7351] = {.lex_state = 0, .external_lex_state = 186},
  [7352] = {.lex_state = 0, .external_lex_state = 186},
  [7353] = {.lex_state = 0, .external_lex_state = 273},
  [7354] = {.lex_state = 0, .external_lex_state = 186},
  [7355] = {.lex_state = 0, .external_lex_state = 269},
  [7356] = {.lex_state = 0, .external_lex_state = 269},
  [7357] = {.lex_state = 0, .external_lex_state = 271},
  [7358] = {.lex_state = 0, .external_lex_state = 269},
  [7359] = {.lex_state = 0, .external_lex_state = 186},
  [7360] = {.lex_state = 0, .external_lex_state = 279},
  [7361] = {.lex_state = 0, .external_lex_state = 186},
  [7362] = {.lex_state = 0, .external_lex_state = 186},
  [7363] = {.lex_state = 0, .external_lex_state = 186},
  [7364] = {.lex_state = 0, .external_lex_state = 186},
  [7365] = {.lex_state = 0, .external_lex_state = 273},
  [7366] = {.lex_state = 0, .external_lex_state = 186},
  [7367] = {.lex_state = 0, .external_lex_state = 269},
  [7368] = {.lex_state = 0, .external_lex_state = 269},
  [7369] = {.lex_state = 0, .external_lex_state = 186},
  [7370] = {.lex_state = 0, .external_lex_state = 269},
  [7371] = {.lex_state = 0, .external_lex_state = 186},
  [7372] = {.lex_state = 0, .external_lex_state = 186},
  [7373] = {.lex_state = 0, .external_lex_state = 186},
  [7374] = {.lex_state = 0, .external_lex_state = 186},
  [7375] = {.lex_state = 0, .external_lex_state = 186},
  [7376] = {.lex_state = 0, .external_lex_state = 186},
  [7377] = {.lex_state = 0, .external_lex_state = 186},
  [7378] = {.lex_state = 0, .external_lex_state = 269},
  [7379] = {.lex_state = 0, .external_lex_state = 269},
  [7380] = {.lex_state = 0, .external_lex_state = 254},
  [7381] = {.lex_state = 0, .external_lex_state = 269},
  [7382] = {.lex_state = 0, .external_lex_state = 186},
  [7383] = {.lex_state = 0, .external_lex_state = 186},
  [7384] = {.lex_state = 0, .external_lex_state = 186},
  [7385] = {.lex_state = 0, .external_lex_state = 186},
  [7386] = {.lex_state = 0, .external_lex_state = 186},
  [7387] = {.lex_state = 0, .external_lex_state = 254},
  [7388] = {.lex_state = 0, .external_lex_state = 186},
  [7389] = {.lex_state = 0, .external_lex_state = 269},
  [7390] = {.lex_state = 0, .external_lex_state = 269},
  [7391] = {.lex_state = 0, .external_lex_state = 277},
  [7392] = {.lex_state = 0, .external_lex_state = 269},
  [7393] = {.lex_state = 0, .external_lex_state = 186},
  [7394] = {.lex_state = 0, .external_lex_state = 186},
  [7395] = {.lex_state = 0, .external_lex_state = 186},
  [7396] = {.lex_state = 0, .external_lex_state = 186},
  [7397] = {.lex_state = 0, .external_lex_state = 186},
  [7398] = {.lex_state = 0, .external_lex_state = 186},
  [7399] = {.lex_state = 0, .external_lex_state = 186},
  [7400] = {.lex_state = 0, .external_lex_state = 269},
  [7401] = {.lex_state = 0, .external_lex_state = 269},
  [7402] = {.lex_state = 0, .external_lex_state = 186},
  [7403] = {.lex_state = 0, .external_lex_state = 269},
  [7404] = {.lex_state = 0, .external_lex_state = 186},
  [7405] = {.lex_state = 0, .external_lex_state = 186},
  [7406] = {.lex_state = 0, .external_lex_state = 271},
  [7407] = {.lex_state = 0, .external_lex_state = 266},
  [7408] = {.lex_state = 0, .external_lex_state = 186},
  [7409] = {.lex_state = 0, .external_lex_state = 186},
  [7410] = {.lex_state = 0, .external_lex_state = 271},
  [7411] = {.lex_state = 0, .external_lex_state = 269},
  [7412] = {.lex_state = 0, .external_lex_state = 269},
  [7413] = {.lex_state = 0, .external_lex_state = 186},
  [7414] = {.lex_state = 0, .external_lex_state = 269},
  [7415] = {.lex_state = 0, .external_lex_state = 186},
  [7416] = {.lex_state = 0, .external_lex_state = 186},
  [7417] = {.lex_state = 0, .external_lex_state = 186},
  [7418] = {.lex_state = 0, .external_lex_state = 186},
  [7419] = {.lex_state = 0, .external_lex_state = 186},
  [7420] = {.lex_state = 0, .external_lex_state = 186},
  [7421] = {.lex_state = 0, .external_lex_state = 186},
  [7422] = {.lex_state = 0, .external_lex_state = 269},
  [7423] = {.lex_state = 0, .external_lex_state = 269},
  [7424] = {.lex_state = 0, .external_lex_state = 186},
  [7425] = {.lex_state = 0, .external_lex_state = 269},
  [7426] = {.lex_state = 0, .external_lex_state = 259},
  [7427] = {.lex_state = 0, .external_lex_state = 259},
  [7428] = {.lex_state = 0, .external_lex_state = 186},
  [7429] = {.lex_state = 0, .external_lex_state = 271},
  [7430] = {.lex_state = 0, .external_lex_state = 186},
  [7431] = {.lex_state = 0, .external_lex_state = 287},
  [7432] = {.lex_state = 0, .external_lex_state = 186},
  [7433] = {.lex_state = 0, .external_lex_state = 269},
  [7434] = {.lex_state = 0, .external_lex_state = 269},
  [7435] = {.lex_state = 0, .external_lex_state = 186},
  [7436] = {.lex_state = 0, .external_lex_state = 269},
  [7437] = {.lex_state = 0, .external_lex_state = 186},
  [7438] = {.lex_state = 0, .external_lex_state = 186},
  [7439] = {.lex_state = 0, .external_lex_state = 186},
  [7440] = {.lex_state = 0, .external_lex_state = 186},
  [7441] = {.lex_state = 0, .external_lex_state = 186},
  [7442] = {.lex_state = 0, .external_lex_state = 266},
  [7443] = {.lex_state = 0, .external_lex_state = 186},
  [7444] = {.lex_state = 0, .external_lex_state = 269},
  [7445] = {.lex_state = 0, .external_lex_state = 269},
  [7446] = {.lex_state = 0, .external_lex_state = 186},
  [7447] = {.lex_state = 0, .external_lex_state = 269},
  [7448] = {.lex_state = 0, .external_lex_state = 186},
  [7449] = {.lex_state = 0, .external_lex_state = 186},
  [7450] = {.lex_state = 0, .external_lex_state = 186},
  [7451] = {.lex_state = 0, .external_lex_state = 186},
  [7452] = {.lex_state = 0, .external_lex_state = 186},
  [7453] = {.lex_state = 0, .external_lex_state = 186},
  [7454] = {.lex_state = 0, .external_lex_state = 271},
  [7455] = {.lex_state = 0, .external_lex_state = 269},
  [7456] = {.lex_state = 0, .external_lex_state = 269},
  [7457] = {.lex_state = 0, .external_lex_state = 186},
  [7458] = {.lex_state = 0, .external_lex_state = 269},
  [7459] = {.lex_state = 0, .external_lex_state = 186},
  [7460] = {.lex_state = 0, .external_lex_state = 186},
  [7461] = {.lex_state = 0, .external_lex_state = 186},
  [7462] = {.lex_state = 0, .external_lex_state = 186},
  [7463] = {.lex_state = 0, .external_lex_state = 186},
  [7464] = {.lex_state = 0, .external_lex_state = 186},
  [7465] = {.lex_state = 0, .external_lex_state = 186},
  [7466] = {.lex_state = 0, .external_lex_state = 269},
  [7467] = {.lex_state = 0, .external_lex_state = 269},
  [7468] = {.lex_state = 0, .external_lex_state = 186},
  [7469] = {.lex_state = 0, .external_lex_state = 269},
  [7470] = {.lex_state = 0, .external_lex_state = 186},
  [7471] = {.lex_state = 0, .external_lex_state = 186},
  [7472] = {.lex_state = 0, .external_lex_state = 186},
  [7473] = {.lex_state = 0, .external_lex_state = 186},
  [7474] = {.lex_state = 0, .external_lex_state = 186},
  [7475] = {.lex_state = 0, .external_lex_state = 186},
  [7476] = {.lex_state = 0, .external_lex_state = 271},
  [7477] = {.lex_state = 0, .external_lex_state = 269},
  [7478] = {.lex_state = 0, .external_lex_state = 269},
  [7479] = {.lex_state = 0, .external_lex_state = 186},
  [7480] = {.lex_state = 0, .external_lex_state = 269},
  [7481] = {.lex_state = 0, .external_lex_state = 186},
  [7482] = {.lex_state = 0, .external_lex_state = 186},
  [7483] = {.lex_state = 0, .external_lex_state = 186},
  [7484] = {.lex_state = 0, .external_lex_state = 186},
  [7485] = {.lex_state = 0, .external_lex_state = 186},
  [7486] = {.lex_state = 0, .external_lex_state = 186},
  [7487] = {.lex_state = 0, .external_lex_state = 240},
  [7488] = {.lex_state = 0, .external_lex_state = 269},
  [7489] = {.lex_state = 0, .external_lex_state = 269},
  [7490] = {.lex_state = 0, .external_lex_state = 186},
  [7491] = {.lex_state = 0, .external_lex_state = 269},
  [7492] = {.lex_state = 0, .external_lex_state = 186},
  [7493] = {.lex_state = 0, .external_lex_state = 186},
  [7494] = {.lex_state = 0, .external_lex_state = 186},
  [7495] = {.lex_state = 0, .external_lex_state = 186},
  [7496] = {.lex_state = 0, .external_lex_state = 186},
  [7497] = {.lex_state = 0, .external_lex_state = 186},
  [7498] = {.lex_state = 0, .external_lex_state = 186},
  [7499] = {.lex_state = 0, .external_lex_state = 269},
  [7500] = {.lex_state = 0, .external_lex_state = 269},
  [7501] = {.lex_state = 0, .external_lex_state = 186},
  [7502] = {.lex_state = 0, .external_lex_state = 269},
  [7503] = {.lex_state = 0, .external_lex_state = 186},
  [7504] = {.lex_state = 0, .external_lex_state = 186},
  [7505] = {.lex_state = 0, .external_lex_state = 186},
  [7506] = {.lex_state = 0, .external_lex_state = 186},
  [7507] = {.lex_state = 0, .external_lex_state = 186},
  [7508] = {.lex_state = 0, .external_lex_state = 186},
  [7509] = {.lex_state = 0, .external_lex_state = 186},
  [7510] = {.lex_state = 0, .external_lex_state = 269},
  [7511] = {.lex_state = 0, .external_lex_state = 269},
  [7512] = {.lex_state = 0, .external_lex_state = 186},
  [7513] = {.lex_state = 0, .external_lex_state = 269},
  [7514] = {.lex_state = 0, .external_lex_state = 186},
  [7515] = {.lex_state = 0, .external_lex_state = 186},
  [7516] = {.lex_state = 0, .external_lex_state = 186},
  [7517] = {.lex_state = 0, .external_lex_state = 186},
  [7518] = {.lex_state = 0, .external_lex_state = 186},
  [7519] = {.lex_state = 0, .external_lex_state = 186},
  [7520] = {.lex_state = 0, .external_lex_state = 186},
  [7521] = {.lex_state = 0, .external_lex_state = 269},
  [7522] = {.lex_state = 0, .external_lex_state = 269},
  [7523] = {.lex_state = 0, .external_lex_state = 186},
  [7524] = {.lex_state = 0, .external_lex_state = 269},
  [7525] = {.lex_state = 0, .external_lex_state = 186},
  [7526] = {.lex_state = 0, .external_lex_state = 186},
  [7527] = {.lex_state = 0, .external_lex_state = 276},
  [7528] = {.lex_state = 0, .external_lex_state = 186},
  [7529] = {.lex_state = 0, .external_lex_state = 186},
  [7530] = {.lex_state = 0, .external_lex_state = 186},
  [7531] = {.lex_state = 0, .external_lex_state = 186},
  [7532] = {.lex_state = 0, .external_lex_state = 269},
  [7533] = {.lex_state = 0, .external_lex_state = 269},
  [7534] = {.lex_state = 0, .external_lex_state = 186},
  [7535] = {.lex_state = 0, .external_lex_state = 269},
  [7536] = {.lex_state = 0, .external_lex_state = 186},
  [7537] = {.lex_state = 0, .external_lex_state = 186},
  [7538] = {.lex_state = 0, .external_lex_state = 186},
  [7539] = {.lex_state = 0, .external_lex_state = 186},
  [7540] = {.lex_state = 0, .external_lex_state = 186},
  [7541] = {.lex_state = 0, .external_lex_state = 186},
  [7542] = {.lex_state = 0, .external_lex_state = 186},
  [7543] = {.lex_state = 0, .external_lex_state = 269},
  [7544] = {.lex_state = 0, .external_lex_state = 269},
  [7545] = {.lex_state = 0, .external_lex_state = 186},
  [7546] = {.lex_state = 0, .external_lex_state = 269},
  [7547] = {.lex_state = 0, .external_lex_state = 240},
  [7548] = {.lex_state = 0, .external_lex_state = 186},
  [7549] = {.lex_state = 0, .external_lex_state = 240},
  [7550] = {.lex_state = 0, .external_lex_state = 186},
  [7551] = {.lex_state = 0, .external_lex_state = 186},
  [7552] = {.lex_state = 0, .external_lex_state = 186},
  [7553] = {.lex_state = 0, .external_lex_state = 186},
  [7554] = {.lex_state = 0, .external_lex_state = 269},
  [7555] = {.lex_state = 0, .external_lex_state = 269},
  [7556] = {.lex_state = 0, .external_lex_state = 186},
  [7557] = {.lex_state = 0, .external_lex_state = 269},
  [7558] = {.lex_state = 0, .external_lex_state = 186},
  [7559] = {.lex_state = 0, .external_lex_state = 186},
  [7560] = {.lex_state = 0, .external_lex_state = 186},
  [7561] = {.lex_state = 0, .external_lex_state = 186},
  [7562] = {.lex_state = 0, .external_lex_state = 186},
  [7563] = {.lex_state = 0, .external_lex_state = 271},
  [7564] = {.lex_state = 0, .external_lex_state = 186},
  [7565] = {.lex_state = 0, .external_lex_state = 269},
  [7566] = {.lex_state = 0, .external_lex_state = 269},
  [7567] = {.lex_state = 0, .external_lex_state = 186},
  [7568] = {.lex_state = 0, .external_lex_state = 269},
  [7569] = {.lex_state = 0, .external_lex_state = 186},
  [7570] = {.lex_state = 0, .external_lex_state = 186},
  [7571] = {.lex_state = 0, .external_lex_state = 186},
  [7572] = {.lex_state = 0, .external_lex_state = 186},
  [7573] = {.lex_state = 0, .external_lex_state = 186},
  [7574] = {.lex_state = 0, .external_lex_state = 186},
  [7575] = {.lex_state = 0, .external_lex_state = 186},
  [7576] = {.lex_state = 0, .external_lex_state = 269},
  [7577] = {.lex_state = 0, .external_lex_state = 269},
  [7578] = {.lex_state = 0, .external_lex_state = 186},
  [7579] = {.lex_state = 0, .external_lex_state = 269},
  [7580] = {.lex_state = 0, .external_lex_state = 186},
  [7581] = {.lex_state = 0, .external_lex_state = 186},
  [7582] = {.lex_state = 0, .external_lex_state = 186},
  [7583] = {.lex_state = 0, .external_lex_state = 186},
  [7584] = {.lex_state = 0, .external_lex_state = 186},
  [7585] = {.lex_state = 0, .external_lex_state = 186},
  [7586] = {.lex_state = 0, .external_lex_state = 186},
  [7587] = {.lex_state = 0, .external_lex_state = 269},
  [7588] = {.lex_state = 0, .external_lex_state = 269},
  [7589] = {.lex_state = 0, .external_lex_state = 186},
  [7590] = {.lex_state = 0, .external_lex_state = 269},
  [7591] = {.lex_state = 0, .external_lex_state = 186},
  [7592] = {.lex_state = 0, .external_lex_state = 186},
  [7593] = {.lex_state = 0, .external_lex_state = 186},
  [7594] = {.lex_state = 0, .external_lex_state = 271},
  [7595] = {.lex_state = 0, .external_lex_state = 186},
  [7596] = {.lex_state = 0, .external_lex_state = 186},
  [7597] = {.lex_state = 0, .external_lex_state = 186},
  [7598] = {.lex_state = 0, .external_lex_state = 269},
  [7599] = {.lex_state = 0, .external_lex_state = 269},
  [7600] = {.lex_state = 0, .external_lex_state = 186},
  [7601] = {.lex_state = 0, .external_lex_state = 186},
  [7602] = {.lex_state = 0, .external_lex_state = 186},
  [7603] = {.lex_state = 0, .external_lex_state = 186},
  [7604] = {.lex_state = 0, .external_lex_state = 260},
  [7605] = {.lex_state = 0, .external_lex_state = 186},
  [7606] = {.lex_state = 0, .external_lex_state = 186},
  [7607] = {.lex_state = 0, .external_lex_state = 186},
  [7608] = {.lex_state = 0, .external_lex_state = 186},
  [7609] = {.lex_state = 0, .external_lex_state = 269},
  [7610] = {.lex_state = 0, .external_lex_state = 186},
  [7611] = {.lex_state = 0, .external_lex_state = 186},
  [7612] = {.lex_state = 0, .external_lex_state = 186},
  [7613] = {.lex_state = 0, .external_lex_state = 186},
  [7614] = {.lex_state = 0, .external_lex_state = 186},
  [7615] = {.lex_state = 0, .external_lex_state = 269},
  [7616] = {.lex_state = 0, .external_lex_state = 186},
  [7617] = {.lex_state = 0, .external_lex_state = 186},
  [7618] = {.lex_state = 0, .external_lex_state = 186},
  [7619] = {.lex_state = 0, .external_lex_state = 240},
  [7620] = {.lex_state = 0, .external_lex_state = 186},
  [7621] = {.lex_state = 0, .external_lex_state = 269},
  [7622] = {.lex_state = 0, .external_lex_state = 186},
  [7623] = {.lex_state = 0, .external_lex_state = 186},
  [7624] = {.lex_state = 0, .external_lex_state = 269},
  [7625] = {.lex_state = 0, .external_lex_state = 186},
  [7626] = {.lex_state = 0, .external_lex_state = 186},
  [7627] = {.lex_state = 0, .external_lex_state = 269},
  [7628] = {.lex_state = 0, .external_lex_state = 186},
  [7629] = {.lex_state = 0, .external_lex_state = 186},
  [7630] = {.lex_state = 0, .external_lex_state = 186},
  [7631] = {.lex_state = 0, .external_lex_state = 186},
  [7632] = {.lex_state = 0, .external_lex_state = 186},
  [7633] = {.lex_state = 0, .external_lex_state = 186},
  [7634] = {.lex_state = 0, .external_lex_state = 186},
  [7635] = {.lex_state = 0, .external_lex_state = 186},
  [7636] = {.lex_state = 0, .external_lex_state = 186},
  [7637] = {.lex_state = 0, .external_lex_state = 186},
  [7638] = {.lex_state = 0, .external_lex_state = 271},
  [7639] = {.lex_state = 0, .external_lex_state = 186},
  [7640] = {.lex_state = 0, .external_lex_state = 186},
  [7641] = {.lex_state = 0, .external_lex_state = 186},
  [7642] = {.lex_state = 0, .external_lex_state = 186},
  [7643] = {.lex_state = 0, .external_lex_state = 186},
  [7644] = {.lex_state = 0, .external_lex_state = 186},
  [7645] = {.lex_state = 0, .external_lex_state = 186},
  [7646] = {.lex_state = 0, .external_lex_state = 186},
  [7647] = {.lex_state = 0, .external_lex_state = 186},
  [7648] = {.lex_state = 0, .external_lex_state = 186},
  [7649] = {.lex_state = 0, .external_lex_state = 186},
  [7650] = {.lex_state = 0, .external_lex_state = 186},
  [7651] = {.lex_state = 0, .external_lex_state = 186},
  [7652] = {.lex_state = 0, .external_lex_state = 186},
  [7653] = {.lex_state = 0, .external_lex_state = 186},
  [7654] = {.lex_state = 0, .external_lex_state = 186},
  [7655] = {.lex_state = 0, .external_lex_state = 286},
  [7656] = {.lex_state = 0, .external_lex_state = 186},
  [7657] = {.lex_state = 0, .external_lex_state = 186},
  [7658] = {.lex_state = 0, .external_lex_state = 186},
  [7659] = {.lex_state = 0, .external_lex_state = 186},
  [7660] = {.lex_state = 0, .external_lex_state = 186},
  [7661] = {.lex_state = 0, .external_lex_state = 279},
  [7662] = {.lex_state = 0, .external_lex_state = 186},
  [7663] = {.lex_state = 0, .external_lex_state = 186},
  [7664] = {.lex_state = 0, .external_lex_state = 186},
  [7665] = {.lex_state = 0, .external_lex_state = 186},
  [7666] = {.lex_state = 0, .external_lex_state = 186},
  [7667] = {.lex_state = 0, .external_lex_state = 186},
  [7668] = {.lex_state = 0, .external_lex_state = 186},
  [7669] = {.lex_state = 0, .external_lex_state = 222},
  [7670] = {.lex_state = 0, .external_lex_state = 186},
  [7671] = {.lex_state = 0, .external_lex_state = 259},
  [7672] = {.lex_state = 0, .external_lex_state = 186},
  [7673] = {.lex_state = 0, .external_lex_state = 186},
  [7674] = {.lex_state = 0, .external_lex_state = 186},
  [7675] = {.lex_state = 0, .external_lex_state = 186},
  [7676] = {.lex_state = 0, .external_lex_state = 186},
  [7677] = {.lex_state = 0, .external_lex_state = 186},
  [7678] = {.lex_state = 0, .external_lex_state = 284},
  [7679] = {.lex_state = 0, .external_lex_state = 186},
  [7680] = {.lex_state = 0, .external_lex_state = 186},
  [7681] = {.lex_state = 0, .external_lex_state = 186},
  [7682] = {.lex_state = 0, .external_lex_state = 186},
  [7683] = {.lex_state = 0, .external_lex_state = 186},
  [7684] = {.lex_state = 0, .external_lex_state = 186},
  [7685] = {.lex_state = 0, .external_lex_state = 186},
  [7686] = {.lex_state = 0, .external_lex_state = 186},
  [7687] = {.lex_state = 0, .external_lex_state = 186},
  [7688] = {.lex_state = 0, .external_lex_state = 186},
  [7689] = {.lex_state = 0, .external_lex_state = 186},
  [7690] = {.lex_state = 0, .external_lex_state = 186},
  [7691] = {.lex_state = 0, .external_lex_state = 186},
  [7692] = {.lex_state = 0, .external_lex_state = 186},
  [7693] = {.lex_state = 0, .external_lex_state = 186},
  [7694] = {.lex_state = 0, .external_lex_state = 186},
  [7695] = {.lex_state = 0, .external_lex_state = 186},
  [7696] = {.lex_state = 0, .external_lex_state = 222},
  [7697] = {.lex_state = 0, .external_lex_state = 222},
  [7698] = {.lex_state = 0, .external_lex_state = 186},
  [7699] = {.lex_state = 0, .external_lex_state = 186},
  [7700] = {.lex_state = 0, .external_lex_state = 186},
  [7701] = {.lex_state = 0, .external_lex_state = 186},
  [7702] = {.lex_state = 0, .external_lex_state = 186},
  [7703] = {.lex_state = 0, .external_lex_state = 186},
  [7704] = {.lex_state = 0, .external_lex_state = 186},
  [7705] = {.lex_state = 0, .external_lex_state = 186},
  [7706] = {.lex_state = 0, .external_lex_state = 186},
  [7707] = {.lex_state = 0, .external_lex_state = 186},
  [7708] = {.lex_state = 0, .external_lex_state = 272},
  [7709] = {.lex_state = 0, .external_lex_state = 186},
  [7710] = {.lex_state = 0, .external_lex_state = 186},
  [7711] = {.lex_state = 0, .external_lex_state = 186},
  [7712] = {.lex_state = 0, .external_lex_state = 186},
  [7713] = {.lex_state = 0, .external_lex_state = 186},
  [7714] = {.lex_state = 0, .external_lex_state = 186},
  [7715] = {.lex_state = 0, .external_lex_state = 186},
  [7716] = {.lex_state = 0, .external_lex_state = 186},
  [7717] = {.lex_state = 0, .external_lex_state = 186},
  [7718] = {.lex_state = 0, .external_lex_state = 186},
  [7719] = {.lex_state = 0, .external_lex_state = 186},
  [7720] = {.lex_state = 0, .external_lex_state = 186},
  [7721] = {.lex_state = 0, .external_lex_state = 186},
  [7722] = {.lex_state = 0, .external_lex_state = 186},
  [7723] = {.lex_state = 0, .external_lex_state = 186},
  [7724] = {.lex_state = 0, .external_lex_state = 279},
  [7725] = {.lex_state = 0, .external_lex_state = 186},
  [7726] = {.lex_state = 0, .external_lex_state = 186},
  [7727] = {.lex_state = 0, .external_lex_state = 186},
  [7728] = {.lex_state = 0, .external_lex_state = 186},
  [7729] = {.lex_state = 0, .external_lex_state = 186},
  [7730] = {.lex_state = 0, .external_lex_state = 186},
  [7731] = {.lex_state = 0, .external_lex_state = 186},
  [7732] = {.lex_state = 0, .external_lex_state = 222},
  [7733] = {.lex_state = 0, .external_lex_state = 186},
  [7734] = {.lex_state = 0, .external_lex_state = 186},
  [7735] = {.lex_state = 0, .external_lex_state = 186},
  [7736] = {.lex_state = 0, .external_lex_state = 186},
  [7737] = {.lex_state = 0, .external_lex_state = 186},
  [7738] = {.lex_state = 0, .external_lex_state = 186},
  [7739] = {.lex_state = 0, .external_lex_state = 186},
  [7740] = {.lex_state = 0, .external_lex_state = 284},
  [7741] = {.lex_state = 0, .external_lex_state = 186},
  [7742] = {.lex_state = 0, .external_lex_state = 186},
  [7743] = {.lex_state = 0, .external_lex_state = 186},
  [7744] = {.lex_state = 0, .external_lex_state = 267},
  [7745] = {.lex_state = 0, .external_lex_state = 271},
  [7746] = {.lex_state = 0, .external_lex_state = 186},
  [7747] = {.lex_state = 0, .external_lex_state = 186},
  [7748] = {.lex_state = 0, .external_lex_state = 186},
  [7749] = {.lex_state = 0, .external_lex_state = 271},
  [7750] = {.lex_state = 0, .external_lex_state = 186},
  [7751] = {.lex_state = 0, .external_lex_state = 186},
  [7752] = {.lex_state = 0, .external_lex_state = 186},
  [7753] = {.lex_state = 0, .external_lex_state = 186},
  [7754] = {.lex_state = 0, .external_lex_state = 186},
  [7755] = {.lex_state = 0, .external_lex_state = 186},
  [7756] = {.lex_state = 0, .external_lex_state = 186},
  [7757] = {.lex_state = 0, .external_lex_state = 186},
  [7758] = {.lex_state = 0, .external_lex_state = 222},
  [7759] = {.lex_state = 0, .external_lex_state = 222},
  [7760] = {.lex_state = 0, .external_lex_state = 186},
  [7761] = {.lex_state = 0, .external_lex_state = 186},
  [7762] = {.lex_state = 0, .external_lex_state = 186},
  [7763] = {.lex_state = 0, .external_lex_state = 186},
  [7764] = {.lex_state = 0, .external_lex_state = 186},
  [7765] = {.lex_state = 0, .external_lex_state = 186},
  [7766] = {.lex_state = 0, .external_lex_state = 186},
  [7767] = {.lex_state = 0, .external_lex_state = 186},
  [7768] = {.lex_state = 0, .external_lex_state = 186},
  [7769] = {.lex_state = 0, .external_lex_state = 271},
  [7770] = {.lex_state = 0, .external_lex_state = 186},
  [7771] = {.lex_state = 0, .external_lex_state = 186},
  [7772] = {.lex_state = 0, .external_lex_state = 186},
  [7773] = {.lex_state = 0, .external_lex_state = 186},
  [7774] = {.lex_state = 0, .external_lex_state = 186},
  [7775] = {.lex_state = 0, .external_lex_state = 186},
  [7776] = {.lex_state = 0, .external_lex_state = 186},
  [7777] = {.lex_state = 0, .external_lex_state = 186},
  [7778] = {.lex_state = 0, .external_lex_state = 186},
  [7779] = {.lex_state = 0, .external_lex_state = 186},
  [7780] = {.lex_state = 0, .external_lex_state = 186},
  [7781] = {.lex_state = 0, .external_lex_state = 186},
  [7782] = {.lex_state = 0, .external_lex_state = 279},
  [7783] = {.lex_state = 0, .external_lex_state = 186},
  [7784] = {.lex_state = 0, .external_lex_state = 186},
  [7785] = {.lex_state = 0, .external_lex_state = 186},
  [7786] = {.lex_state = 0, .external_lex_state = 186},
  [7787] = {.lex_state = 0, .external_lex_state = 186},
  [7788] = {.lex_state = 0, .external_lex_state = 186},
  [7789] = {.lex_state = 0, .external_lex_state = 186},
  [7790] = {.lex_state = 0, .external_lex_state = 222},
  [7791] = {.lex_state = 0, .external_lex_state = 186},
  [7792] = {.lex_state = 0, .external_lex_state = 186},
  [7793] = {.lex_state = 0, .external_lex_state = 186},
  [7794] = {.lex_state = 0, .external_lex_state = 186},
  [7795] = {.lex_state = 0, .external_lex_state = 186},
  [7796] = {.lex_state = 0, .external_lex_state = 186},
  [7797] = {.lex_state = 0, .external_lex_state = 284},
  [7798] = {.lex_state = 0, .external_lex_state = 186},
  [7799] = {.lex_state = 0, .external_lex_state = 186},
  [7800] = {.lex_state = 0, .external_lex_state = 186},
  [7801] = {.lex_state = 0, .external_lex_state = 186},
  [7802] = {.lex_state = 0, .external_lex_state = 186},
  [7803] = {.lex_state = 0, .external_lex_state = 186},
  [7804] = {.lex_state = 0, .external_lex_state = 186},
  [7805] = {.lex_state = 0, .external_lex_state = 186},
  [7806] = {.lex_state = 0, .external_lex_state = 186},
  [7807] = {.lex_state = 0, .external_lex_state = 186},
  [7808] = {.lex_state = 0, .external_lex_state = 186},
  [7809] = {.lex_state = 0, .external_lex_state = 186},
  [7810] = {.lex_state = 0, .external_lex_state = 186},
  [7811] = {.lex_state = 0, .external_lex_state = 222},
  [7812] = {.lex_state = 0, .external_lex_state = 222},
  [7813] = {.lex_state = 0, .external_lex_state = 186},
  [7814] = {.lex_state = 0, .external_lex_state = 186},
  [7815] = {.lex_state = 0, .external_lex_state = 186},
  [7816] = {.lex_state = 0, .external_lex_state = 186},
  [7817] = {.lex_state = 0, .external_lex_state = 186},
  [7818] = {.lex_state = 0, .external_lex_state = 186},
  [7819] = {.lex_state = 0, .external_lex_state = 186},
  [7820] = {.lex_state = 0, .external_lex_state = 186},
  [7821] = {.lex_state = 0, .external_lex_state = 186},
  [7822] = {.lex_state = 0, .external_lex_state = 186},
  [7823] = {.lex_state = 0, .external_lex_state = 186},
  [7824] = {.lex_state = 0, .external_lex_state = 186},
  [7825] = {.lex_state = 0, .external_lex_state = 186},
  [7826] = {.lex_state = 0, .external_lex_state = 186},
  [7827] = {.lex_state = 0, .external_lex_state = 186},
  [7828] = {.lex_state = 0, .external_lex_state = 186},
  [7829] = {.lex_state = 0, .external_lex_state = 186},
  [7830] = {.lex_state = 0, .external_lex_state = 186},
  [7831] = {.lex_state = 0, .external_lex_state = 222},
  [7832] = {.lex_state = 0, .external_lex_state = 186},
  [7833] = {.lex_state = 0, .external_lex_state = 186},
  [7834] = {.lex_state = 0, .external_lex_state = 186},
  [7835] = {.lex_state = 0, .external_lex_state = 284},
  [7836] = {.lex_state = 0, .external_lex_state = 186},
  [7837] = {.lex_state = 0, .external_lex_state = 186},
  [7838] = {.lex_state = 0, .external_lex_state = 186},
  [7839] = {.lex_state = 0, .external_lex_state = 186},
  [7840] = {.lex_state = 0, .external_lex_state = 284},
  [7841] = {.lex_state = 0, .external_lex_state = 186},
  [7842] = {.lex_state = 0, .external_lex_state = 186},
  [7843] = {.lex_state = 0, .external_lex_state = 186},
  [7844] = {.lex_state = 0, .external_lex_state = 186},
  [7845] = {.lex_state = 0, .external_lex_state = 186},
  [7846] = {.lex_state = 0, .external_lex_state = 222},
  [7847] = {.lex_state = 0, .external_lex_state = 222},
  [7848] = {.lex_state = 0, .external_lex_state = 186},
  [7849] = {.lex_state = 0, .external_lex_state = 186},
  [7850] = {.lex_state = 0, .external_lex_state = 186},
  [7851] = {.lex_state = 0, .external_lex_state = 186},
  [7852] = {.lex_state = 0, .external_lex_state = 186},
  [7853] = {.lex_state = 0, .external_lex_state = 186},
  [7854] = {.lex_state = 0, .external_lex_state = 186},
  [7855] = {.lex_state = 0, .external_lex_state = 186},
  [7856] = {.lex_state = 0, .external_lex_state = 186},
  [7857] = {.lex_state = 0, .external_lex_state = 186},
  [7858] = {.lex_state = 0, .external_lex_state = 222},
  [7859] = {.lex_state = 0, .external_lex_state = 186},
  [7860] = {.lex_state = 0, .external_lex_state = 284},
  [7861] = {.lex_state = 0, .external_lex_state = 186},
  [7862] = {.lex_state = 0, .external_lex_state = 186},
  [7863] = {.lex_state = 0, .external_lex_state = 186},
  [7864] = {.lex_state = 0, .external_lex_state = 186},
  [7865] = {.lex_state = 0, .external_lex_state = 186},
  [7866] = {.lex_state = 0, .external_lex_state = 222},
  [7867] = {.lex_state = 0, .external_lex_state = 222},
  [7868] = {.lex_state = 0, .external_lex_state = 186},
  [7869] = {.lex_state = 0, .external_lex_state = 186},
  [7870] = {.lex_state = 0, .external_lex_state = 186},
  [7871] = {.lex_state = 0, .external_lex_state = 186},
  [7872] = {.lex_state = 0, .external_lex_state = 284},
  [7873] = {.lex_state = 0, .external_lex_state = 284},
  [7874] = {.lex_state = 0, .external_lex_state = 186},
  [7875] = {.lex_state = 0, .external_lex_state = 186},
  [7876] = {.lex_state = 0, .external_lex_state = 186},
  [7877] = {.lex_state = 0, .external_lex_state = 186},
  [7878] = {.lex_state = 0, .external_lex_state = 222},
  [7879] = {.lex_state = 0, .external_lex_state = 222},
  [7880] = {.lex_state = 0, .external_lex_state = 249},
  [7881] = {.lex_state = 0, .external_lex_state = 186},
  [7882] = {.lex_state = 0, .external_lex_state = 186},
  [7883] = {.lex_state = 0, .external_lex_state = 186},
  [7884] = {.lex_state = 0, .external_lex_state = 222},
  [7885] = {.lex_state = 0, .external_lex_state = 186},
  [7886] = {.lex_state = 0, .external_lex_state = 186},
  [7887] = {.lex_state = 0, .external_lex_state = 186},
  [7888] = {.lex_state = 0, .external_lex_state = 186},
  [7889] = {.lex_state = 0, .external_lex_state = 186},
  [7890] = {.lex_state = 0, .external_lex_state = 186},
  [7891] = {.lex_state = 0, .external_lex_state = 186},
  [7892] = {.lex_state = 0, .external_lex_state = 186},
  [7893] = {.lex_state = 0, .external_lex_state = 186},
  [7894] = {.lex_state = 0, .external_lex_state = 186},
  [7895] = {.lex_state = 0, .external_lex_state = 186},
  [7896] = {.lex_state = 0, .external_lex_state = 186},
  [7897] = {.lex_state = 0, .external_lex_state = 186},
  [7898] = {.lex_state = 0, .external_lex_state = 186},
  [7899] = {.lex_state = 0, .external_lex_state = 186},
  [7900] = {.lex_state = 0, .external_lex_state = 186},
  [7901] = {.lex_state = 0, .external_lex_state = 186},
  [7902] = {.lex_state = 0, .external_lex_state = 186},
  [7903] = {.lex_state = 0, .external_lex_state = 186},
  [7904] = {.lex_state = 0, .external_lex_state = 186},
  [7905] = {.lex_state = 0, .external_lex_state = 186},
  [7906] = {.lex_state = 0, .external_lex_state = 186},
  [7907] = {.lex_state = 0, .external_lex_state = 186},
  [7908] = {.lex_state = 0, .external_lex_state = 186},
  [7909] = {.lex_state = 0, .external_lex_state = 186},
  [7910] = {.lex_state = 0, .external_lex_state = 186},
  [7911] = {.lex_state = 0, .external_lex_state = 186},
  [7912] = {.lex_state = 0, .external_lex_state = 186},
  [7913] = {.lex_state = 0, .external_lex_state = 186},
  [7914] = {.lex_state = 0, .external_lex_state = 186},
  [7915] = {.lex_state = 0, .external_lex_state = 186},
  [7916] = {.lex_state = 0, .external_lex_state = 186},
  [7917] = {.lex_state = 0, .external_lex_state = 186},
  [7918] = {.lex_state = 0, .external_lex_state = 186},
  [7919] = {.lex_state = 0, .external_lex_state = 186},
  [7920] = {.lex_state = 0, .external_lex_state = 186},
  [7921] = {.lex_state = 0, .external_lex_state = 186},
  [7922] = {.lex_state = 0, .external_lex_state = 186},
  [7923] = {.lex_state = 0, .external_lex_state = 186},
  [7924] = {.lex_state = 0, .external_lex_state = 186},
  [7925] = {.lex_state = 0, .external_lex_state = 186},
  [7926] = {.lex_state = 0, .external_lex_state = 186},
  [7927] = {.lex_state = 0, .external_lex_state = 186},
  [7928] = {.lex_state = 0, .external_lex_state = 186},
  [7929] = {.lex_state = 0, .external_lex_state = 186},
  [7930] = {.lex_state = 0, .external_lex_state = 186},
  [7931] = {.lex_state = 0, .external_lex_state = 186},
  [7932] = {.lex_state = 0, .external_lex_state = 186},
  [7933] = {.lex_state = 0, .external_lex_state = 186},
  [7934] = {.lex_state = 0, .external_lex_state = 186},
  [7935] = {.lex_state = 0, .external_lex_state = 186},
  [7936] = {.lex_state = 0, .external_lex_state = 186},
  [7937] = {.lex_state = 0, .external_lex_state = 186},
  [7938] = {.lex_state = 0, .external_lex_state = 186},
  [7939] = {.lex_state = 0, .external_lex_state = 186},
  [7940] = {.lex_state = 0, .external_lex_state = 186},
  [7941] = {.lex_state = 0, .external_lex_state = 186},
  [7942] = {.lex_state = 0, .external_lex_state = 186},
  [7943] = {.lex_state = 0, .external_lex_state = 186},
  [7944] = {.lex_state = 0, .external_lex_state = 186},
  [7945] = {.lex_state = 0, .external_lex_state = 186},
  [7946] = {.lex_state = 0, .external_lex_state = 186},
  [7947] = {.lex_state = 0, .external_lex_state = 186},
  [7948] = {.lex_state = 0, .external_lex_state = 186},
  [7949] = {.lex_state = 0, .external_lex_state = 186},
  [7950] = {.lex_state = 0, .external_lex_state = 186},
  [7951] = {.lex_state = 0, .external_lex_state = 186},
  [7952] = {.lex_state = 0, .external_lex_state = 186},
  [7953] = {.lex_state = 0, .external_lex_state = 186},
  [7954] = {.lex_state = 0, .external_lex_state = 186},
  [7955] = {.lex_state = 0, .external_lex_state = 186},
  [7956] = {.lex_state = 0, .external_lex_state = 186},
  [7957] = {.lex_state = 0, .external_lex_state = 186},
  [7958] = {.lex_state = 0, .external_lex_state = 186},
  [7959] = {.lex_state = 0, .external_lex_state = 186},
  [7960] = {.lex_state = 0, .external_lex_state = 186},
  [7961] = {.lex_state = 0, .external_lex_state = 186},
  [7962] = {.lex_state = 0, .external_lex_state = 186},
  [7963] = {.lex_state = 0, .external_lex_state = 186},
  [7964] = {.lex_state = 0, .external_lex_state = 186},
  [7965] = {.lex_state = 0, .external_lex_state = 186},
  [7966] = {.lex_state = 0, .external_lex_state = 186},
  [7967] = {.lex_state = 0, .external_lex_state = 186},
  [7968] = {.lex_state = 0, .external_lex_state = 186},
  [7969] = {.lex_state = 0, .external_lex_state = 186},
  [7970] = {.lex_state = 0, .external_lex_state = 186},
  [7971] = {.lex_state = 0, .external_lex_state = 186},
  [7972] = {.lex_state = 0, .external_lex_state = 186},
  [7973] = {.lex_state = 0, .external_lex_state = 186},
  [7974] = {.lex_state = 0, .external_lex_state = 186},
  [7975] = {.lex_state = 0, .external_lex_state = 186},
  [7976] = {.lex_state = 0, .external_lex_state = 186},
  [7977] = {.lex_state = 0, .external_lex_state = 186},
  [7978] = {.lex_state = 0, .external_lex_state = 186},
  [7979] = {.lex_state = 0, .external_lex_state = 186},
  [7980] = {.lex_state = 0, .external_lex_state = 186},
  [7981] = {.lex_state = 0, .external_lex_state = 186},
  [7982] = {.lex_state = 0, .external_lex_state = 186},
  [7983] = {.lex_state = 0, .external_lex_state = 186},
  [7984] = {.lex_state = 0, .external_lex_state = 186},
  [7985] = {.lex_state = 0, .external_lex_state = 186},
  [7986] = {.lex_state = 0, .external_lex_state = 186},
  [7987] = {.lex_state = 0, .external_lex_state = 186},
  [7988] = {.lex_state = 0, .external_lex_state = 186},
  [7989] = {.lex_state = 0, .external_lex_state = 186},
  [7990] = {.lex_state = 0, .external_lex_state = 186},
  [7991] = {.lex_state = 0, .external_lex_state = 186},
  [7992] = {.lex_state = 0, .external_lex_state = 186},
  [7993] = {.lex_state = 0, .external_lex_state = 186},
  [7994] = {.lex_state = 0, .external_lex_state = 186},
  [7995] = {.lex_state = 0, .external_lex_state = 186},
  [7996] = {.lex_state = 0, .external_lex_state = 186},
  [7997] = {.lex_state = 0, .external_lex_state = 186},
  [7998] = {.lex_state = 0, .external_lex_state = 186},
  [7999] = {.lex_state = 0, .external_lex_state = 186},
  [8000] = {.lex_state = 0, .external_lex_state = 186},
  [8001] = {.lex_state = 0, .external_lex_state = 186},
  [8002] = {.lex_state = 0, .external_lex_state = 186},
  [8003] = {.lex_state = 0, .external_lex_state = 186},
  [8004] = {.lex_state = 0, .external_lex_state = 186},
  [8005] = {.lex_state = 0, .external_lex_state = 186},
  [8006] = {.lex_state = 0, .external_lex_state = 186},
  [8007] = {.lex_state = 0, .external_lex_state = 186},
  [8008] = {.lex_state = 0, .external_lex_state = 186},
  [8009] = {.lex_state = 0, .external_lex_state = 186},
  [8010] = {.lex_state = 0, .external_lex_state = 186},
  [8011] = {.lex_state = 0, .external_lex_state = 186},
  [8012] = {.lex_state = 0, .external_lex_state = 186},
  [8013] = {.lex_state = 0, .external_lex_state = 186},
  [8014] = {.lex_state = 0, .external_lex_state = 254},
  [8015] = {.lex_state = 0, .external_lex_state = 254},
  [8016] = {.lex_state = 0, .external_lex_state = 186},
  [8017] = {.lex_state = 0, .external_lex_state = 186},
  [8018] = {.lex_state = 0, .external_lex_state = 186},
  [8019] = {.lex_state = 0, .external_lex_state = 186},
  [8020] = {.lex_state = 0, .external_lex_state = 186},
  [8021] = {.lex_state = 0, .external_lex_state = 186},
  [8022] = {.lex_state = 0, .external_lex_state = 186},
  [8023] = {.lex_state = 0, .external_lex_state = 186},
  [8024] = {.lex_state = 0, .external_lex_state = 186},
  [8025] = {.lex_state = 0, .external_lex_state = 186},
  [8026] = {.lex_state = 0, .external_lex_state = 186},
  [8027] = {.lex_state = 0, .external_lex_state = 186},
  [8028] = {.lex_state = 0, .external_lex_state = 186},
  [8029] = {.lex_state = 0, .external_lex_state = 186},
  [8030] = {.lex_state = 0, .external_lex_state = 186},
  [8031] = {.lex_state = 0, .external_lex_state = 186},
  [8032] = {.lex_state = 0, .external_lex_state = 186},
  [8033] = {.lex_state = 0, .external_lex_state = 186},
  [8034] = {.lex_state = 0, .external_lex_state = 186},
  [8035] = {.lex_state = 0, .external_lex_state = 186},
  [8036] = {.lex_state = 0, .external_lex_state = 186},
  [8037] = {.lex_state = 0, .external_lex_state = 271},
  [8038] = {.lex_state = 0, .external_lex_state = 186},
  [8039] = {.lex_state = 0, .external_lex_state = 186},
  [8040] = {.lex_state = 0, .external_lex_state = 186},
  [8041] = {.lex_state = 0, .external_lex_state = 186},
  [8042] = {.lex_state = 0, .external_lex_state = 186},
  [8043] = {.lex_state = 0, .external_lex_state = 186},
  [8044] = {.lex_state = 0, .external_lex_state = 186},
  [8045] = {.lex_state = 0, .external_lex_state = 186},
  [8046] = {.lex_state = 0, .external_lex_state = 186},
  [8047] = {.lex_state = 0, .external_lex_state = 254},
  [8048] = {.lex_state = 0, .external_lex_state = 186},
  [8049] = {.lex_state = 0, .external_lex_state = 186},
  [8050] = {.lex_state = 0, .external_lex_state = 186},
  [8051] = {.lex_state = 0, .external_lex_state = 259},
  [8052] = {.lex_state = 0, .external_lex_state = 186},
  [8053] = {.lex_state = 0, .external_lex_state = 186},
  [8054] = {.lex_state = 0, .external_lex_state = 186},
  [8055] = {.lex_state = 0, .external_lex_state = 186},
  [8056] = {.lex_state = 0, .external_lex_state = 259},
  [8057] = {.lex_state = 0, .external_lex_state = 186},
  [8058] = {.lex_state = 0, .external_lex_state = 186},
  [8059] = {.lex_state = 0, .external_lex_state = 186},
  [8060] = {.lex_state = 0, .external_lex_state = 288},
  [8061] = {.lex_state = 0, .external_lex_state = 186},
  [8062] = {.lex_state = 0, .external_lex_state = 271},
  [8063] = {.lex_state = 0, .external_lex_state = 186},
  [8064] = {.lex_state = 0, .external_lex_state = 186},
  [8065] = {.lex_state = 0, .external_lex_state = 271},
  [8066] = {.lex_state = 0, .external_lex_state = 186},
  [8067] = {.lex_state = 0, .external_lex_state = 186},
  [8068] = {.lex_state = 0, .external_lex_state = 186},
  [8069] = {.lex_state = 0, .external_lex_state = 186},
  [8070] = {.lex_state = 0, .external_lex_state = 186},
  [8071] = {.lex_state = 0, .external_lex_state = 186},
  [8072] = {.lex_state = 0, .external_lex_state = 186},
  [8073] = {.lex_state = 0, .external_lex_state = 186},
  [8074] = {.lex_state = 0, .external_lex_state = 186},
  [8075] = {.lex_state = 0, .external_lex_state = 186},
  [8076] = {.lex_state = 0, .external_lex_state = 186},
  [8077] = {.lex_state = 0, .external_lex_state = 186},
  [8078] = {.lex_state = 0, .external_lex_state = 271},
  [8079] = {.lex_state = 0, .external_lex_state = 186},
  [8080] = {.lex_state = 0, .external_lex_state = 186},
  [8081] = {.lex_state = 0, .external_lex_state = 186},
  [8082] = {.lex_state = 0, .external_lex_state = 186},
  [8083] = {.lex_state = 0, .external_lex_state = 186},
  [8084] = {.lex_state = 0, .external_lex_state = 186},
  [8085] = {.lex_state = 0, .external_lex_state = 281},
  [8086] = {.lex_state = 0, .external_lex_state = 186},
  [8087] = {.lex_state = 0, .external_lex_state = 186},
  [8088] = {.lex_state = 0, .external_lex_state = 186},
  [8089] = {.lex_state = 0, .external_lex_state = 186},
  [8090] = {.lex_state = 0, .external_lex_state = 186},
  [8091] = {.lex_state = 0, .external_lex_state = 186},
  [8092] = {.lex_state = 0, .external_lex_state = 186},
  [8093] = {.lex_state = 0, .external_lex_state = 186},
  [8094] = {.lex_state = 0, .external_lex_state = 186},
  [8095] = {.lex_state = 0, .external_lex_state = 186},
  [8096] = {.lex_state = 0, .external_lex_state = 186},
  [8097] = {.lex_state = 0, .external_lex_state = 186},
  [8098] = {.lex_state = 0, .external_lex_state = 289},
  [8099] = {.lex_state = 0, .external_lex_state = 186},
  [8100] = {.lex_state = 0, .external_lex_state = 186},
  [8101] = {.lex_state = 0, .external_lex_state = 186},
  [8102] = {.lex_state = 0, .external_lex_state = 186},
  [8103] = {.lex_state = 0, .external_lex_state = 186},
  [8104] = {.lex_state = 0, .external_lex_state = 186},
  [8105] = {.lex_state = 0, .external_lex_state = 186},
  [8106] = {.lex_state = 0, .external_lex_state = 186},
  [8107] = {.lex_state = 0, .external_lex_state = 186},
  [8108] = {.lex_state = 0, .external_lex_state = 186},
  [8109] = {.lex_state = 0, .external_lex_state = 186},
  [8110] = {.lex_state = 0, .external_lex_state = 186},
  [8111] = {.lex_state = 0, .external_lex_state = 186},
  [8112] = {.lex_state = 0, .external_lex_state = 186},
  [8113] = {.lex_state = 0, .external_lex_state = 186},
  [8114] = {.lex_state = 0, .external_lex_state = 186},
  [8115] = {.lex_state = 0, .external_lex_state = 186},
  [8116] = {.lex_state = 0, .external_lex_state = 186},
  [8117] = {.lex_state = 0, .external_lex_state = 186},
  [8118] = {.lex_state = 0, .external_lex_state = 249},
  [8119] = {.lex_state = 0, .external_lex_state = 186},
  [8120] = {.lex_state = 0, .external_lex_state = 186},
  [8121] = {.lex_state = 0, .external_lex_state = 186},
  [8122] = {.lex_state = 0, .external_lex_state = 186},
  [8123] = {.lex_state = 0, .external_lex_state = 276},
  [8124] = {.lex_state = 0, .external_lex_state = 186},
  [8125] = {.lex_state = 0, .external_lex_state = 186},
  [8126] = {.lex_state = 0, .external_lex_state = 186},
  [8127] = {.lex_state = 0, .external_lex_state = 186},
  [8128] = {.lex_state = 0, .external_lex_state = 186},
  [8129] = {.lex_state = 0, .external_lex_state = 186},
  [8130] = {.lex_state = 0, .external_lex_state = 186},
  [8131] = {.lex_state = 0, .external_lex_state = 186},
  [8132] = {.lex_state = 0, .external_lex_state = 186},
  [8133] = {.lex_state = 0, .external_lex_state = 186},
  [8134] = {.lex_state = 0, .external_lex_state = 186},
  [8135] = {.lex_state = 0, .external_lex_state = 186},
  [8136] = {.lex_state = 0, .external_lex_state = 186},
  [8137] = {.lex_state = 0, .external_lex_state = 186},
  [8138] = {.lex_state = 0, .external_lex_state = 186},
  [8139] = {.lex_state = 0, .external_lex_state = 186},
  [8140] = {.lex_state = 0, .external_lex_state = 186},
  [8141] = {.lex_state = 0, .external_lex_state = 186},
  [8142] = {.lex_state = 0, .external_lex_state = 186},
  [8143] = {.lex_state = 0, .external_lex_state = 259},
  [8144] = {.lex_state = 0, .external_lex_state = 186},
  [8145] = {.lex_state = 0, .external_lex_state = 186},
  [8146] = {.lex_state = 0, .external_lex_state = 274},
  [8147] = {.lex_state = 0, .external_lex_state = 186},
  [8148] = {.lex_state = 0, .external_lex_state = 186},
  [8149] = {.lex_state = 0, .external_lex_state = 186},
  [8150] = {.lex_state = 0, .external_lex_state = 186},
  [8151] = {.lex_state = 0, .external_lex_state = 186},
  [8152] = {.lex_state = 0, .external_lex_state = 186},
  [8153] = {.lex_state = 0, .external_lex_state = 186},
  [8154] = {.lex_state = 0, .external_lex_state = 186},
  [8155] = {.lex_state = 0, .external_lex_state = 186},
  [8156] = {.lex_state = 0, .external_lex_state = 186},
  [8157] = {.lex_state = 0, .external_lex_state = 186},
  [8158] = {.lex_state = 0, .external_lex_state = 186},
  [8159] = {.lex_state = 0, .external_lex_state = 186},
  [8160] = {.lex_state = 0, .external_lex_state = 186},
  [8161] = {.lex_state = 0, .external_lex_state = 186},
  [8162] = {.lex_state = 0, .external_lex_state = 186},
  [8163] = {.lex_state = 0, .external_lex_state = 186},
  [8164] = {.lex_state = 0, .external_lex_state = 186},
  [8165] = {.lex_state = 0, .external_lex_state = 186},
  [8166] = {.lex_state = 0, .external_lex_state = 186},
  [8167] = {.lex_state = 0, .external_lex_state = 186},
  [8168] = {.lex_state = 0, .external_lex_state = 186},
  [8169] = {.lex_state = 0, .external_lex_state = 186},
  [8170] = {.lex_state = 0, .external_lex_state = 186},
  [8171] = {.lex_state = 0, .external_lex_state = 186},
  [8172] = {.lex_state = 0, .external_lex_state = 186},
  [8173] = {.lex_state = 0, .external_lex_state = 186},
  [8174] = {.lex_state = 0, .external_lex_state = 186},
  [8175] = {.lex_state = 0, .external_lex_state = 186},
  [8176] = {.lex_state = 0, .external_lex_state = 186},
  [8177] = {.lex_state = 0, .external_lex_state = 186},
  [8178] = {.lex_state = 0, .external_lex_state = 186},
  [8179] = {.lex_state = 0, .external_lex_state = 186},
  [8180] = {.lex_state = 0, .external_lex_state = 186},
  [8181] = {.lex_state = 0, .external_lex_state = 186},
  [8182] = {.lex_state = 0, .external_lex_state = 186},
  [8183] = {.lex_state = 0, .external_lex_state = 186},
  [8184] = {.lex_state = 0, .external_lex_state = 186},
  [8185] = {.lex_state = 0, .external_lex_state = 186},
  [8186] = {.lex_state = 0, .external_lex_state = 186},
  [8187] = {.lex_state = 0, .external_lex_state = 186},
  [8188] = {.lex_state = 0, .external_lex_state = 186},
  [8189] = {.lex_state = 0, .external_lex_state = 186},
  [8190] = {.lex_state = 0, .external_lex_state = 259},
  [8191] = {.lex_state = 0, .external_lex_state = 186},
  [8192] = {.lex_state = 0, .external_lex_state = 186},
  [8193] = {.lex_state = 0, .external_lex_state = 186},
  [8194] = {.lex_state = 0, .external_lex_state = 249},
  [8195] = {.lex_state = 0, .external_lex_state = 186},
  [8196] = {.lex_state = 0, .external_lex_state = 186},
  [8197] = {.lex_state = 0, .external_lex_state = 186},
  [8198] = {.lex_state = 0, .external_lex_state = 186},
  [8199] = {.lex_state = 0, .external_lex_state = 186},
  [8200] = {.lex_state = 0, .external_lex_state = 186},
  [8201] = {.lex_state = 0, .external_lex_state = 271},
  [8202] = {.lex_state = 0, .external_lex_state = 186},
  [8203] = {.lex_state = 0, .external_lex_state = 186},
  [8204] = {.lex_state = 0, .external_lex_state = 186},
  [8205] = {.lex_state = 0, .external_lex_state = 186},
  [8206] = {.lex_state = 0, .external_lex_state = 186},
  [8207] = {.lex_state = 0, .external_lex_state = 186},
  [8208] = {.lex_state = 0, .external_lex_state = 186},
  [8209] = {.lex_state = 0, .external_lex_state = 186},
  [8210] = {.lex_state = 0, .external_lex_state = 271},
  [8211] = {.lex_state = 0, .external_lex_state = 186},
  [8212] = {.lex_state = 0, .external_lex_state = 186},
  [8213] = {.lex_state = 0, .external_lex_state = 186},
  [8214] = {.lex_state = 0, .external_lex_state = 186},
  [8215] = {.lex_state = 0, .external_lex_state = 186},
  [8216] = {.lex_state = 0, .external_lex_state = 186},
  [8217] = {.lex_state = 0, .external_lex_state = 186},
  [8218] = {.lex_state = 0, .external_lex_state = 186},
  [8219] = {.lex_state = 0, .external_lex_state = 186},
  [8220] = {.lex_state = 0, .external_lex_state = 186},
  [8221] = {.lex_state = 0, .external_lex_state = 186},
  [8222] = {.lex_state = 0, .external_lex_state = 186},
  [8223] = {.lex_state = 0, .external_lex_state = 186},
  [8224] = {.lex_state = 0, .external_lex_state = 186},
  [8225] = {.lex_state = 0, .external_lex_state = 274},
  [8226] = {.lex_state = 0, .external_lex_state = 186},
  [8227] = {.lex_state = 0, .external_lex_state = 186},
  [8228] = {.lex_state = 0, .external_lex_state = 186},
  [8229] = {.lex_state = 0, .external_lex_state = 186},
  [8230] = {.lex_state = 0, .external_lex_state = 186},
  [8231] = {.lex_state = 0, .external_lex_state = 186},
  [8232] = {.lex_state = 0, .external_lex_state = 186},
  [8233] = {.lex_state = 0, .external_lex_state = 186},
  [8234] = {.lex_state = 0, .external_lex_state = 186},
  [8235] = {.lex_state = 0, .external_lex_state = 186},
  [8236] = {.lex_state = 0, .external_lex_state = 186},
  [8237] = {.lex_state = 0, .external_lex_state = 186},
  [8238] = {.lex_state = 0, .external_lex_state = 186},
  [8239] = {.lex_state = 0, .external_lex_state = 186},
  [8240] = {.lex_state = 0, .external_lex_state = 186},
  [8241] = {.lex_state = 0, .external_lex_state = 186},
  [8242] = {.lex_state = 0, .external_lex_state = 186},
  [8243] = {.lex_state = 0, .external_lex_state = 186},
  [8244] = {.lex_state = 0, .external_lex_state = 271},
  [8245] = {.lex_state = 0, .external_lex_state = 186},
  [8246] = {.lex_state = 0, .external_lex_state = 186},
  [8247] = {.lex_state = 0, .external_lex_state = 186},
  [8248] = {.lex_state = 0, .external_lex_state = 186},
  [8249] = {.lex_state = 0, .external_lex_state = 186},
  [8250] = {.lex_state = 0, .external_lex_state = 186},
  [8251] = {.lex_state = 0, .external_lex_state = 186},
  [8252] = {.lex_state = 0, .external_lex_state = 186},
  [8253] = {.lex_state = 0, .external_lex_state = 186},
  [8254] = {.lex_state = 0, .external_lex_state = 186},
  [8255] = {.lex_state = 0, .external_lex_state = 186},
  [8256] = {.lex_state = 0, .external_lex_state = 186},
  [8257] = {.lex_state = 0, .external_lex_state = 186},
  [8258] = {.lex_state = 0, .external_lex_state = 186},
  [8259] = {.lex_state = 0, .external_lex_state = 186},
  [8260] = {.lex_state = 0, .external_lex_state = 186},
  [8261] = {.lex_state = 0, .external_lex_state = 186},
  [8262] = {.lex_state = 0, .external_lex_state = 186},
  [8263] = {.lex_state = 0, .external_lex_state = 186},
  [8264] = {.lex_state = 0, .external_lex_state = 186},
  [8265] = {.lex_state = 0, .external_lex_state = 186},
  [8266] = {.lex_state = 0, .external_lex_state = 186},
  [8267] = {.lex_state = 0, .external_lex_state = 186},
  [8268] = {.lex_state = 0, .external_lex_state = 186},
  [8269] = {.lex_state = 0, .external_lex_state = 282},
  [8270] = {.lex_state = 0, .external_lex_state = 186},
  [8271] = {.lex_state = 0, .external_lex_state = 249},
  [8272] = {.lex_state = 0, .external_lex_state = 260},
  [8273] = {.lex_state = 0, .external_lex_state = 249},
  [8274] = {.lex_state = 0, .external_lex_state = 186},
  [8275] = {.lex_state = 0, .external_lex_state = 186},
  [8276] = {.lex_state = 0, .external_lex_state = 186},
  [8277] = {.lex_state = 0, .external_lex_state = 282},
  [8278] = {.lex_state = 0, .external_lex_state = 186},
  [8279] = {.lex_state = 0, .external_lex_state = 249},
  [8280] = {.lex_state = 0, .external_lex_state = 186},
  [8281] = {.lex_state = 0, .external_lex_state = 249},
  [8282] = {.lex_state = 0, .external_lex_state = 186},
  [8283] = {.lex_state = 0, .external_lex_state = 186},
  [8284] = {.lex_state = 0, .external_lex_state = 186},
  [8285] = {.lex_state = 0, .external_lex_state = 282},
  [8286] = {.lex_state = 0, .external_lex_state = 274},
  [8287] = {.lex_state = 0, .external_lex_state = 249},
  [8288] = {.lex_state = 0, .external_lex_state = 186},
  [8289] = {.lex_state = 0, .external_lex_state = 249},
  [8290] = {.lex_state = 0, .external_lex_state = 186},
  [8291] = {.lex_state = 0, .external_lex_state = 289},
  [8292] = {.lex_state = 0, .external_lex_state = 186},
  [8293] = {.lex_state = 0, .external_lex_state = 282},
  [8294] = {.lex_state = 0, .external_lex_state = 186},
  [8295] = {.lex_state = 0, .external_lex_state = 249},
  [8296] = {.lex_state = 0, .external_lex_state = 259},
  [8297] = {.lex_state = 0, .external_lex_state = 186},
  [8298] = {.lex_state = 0, .external_lex_state = 282},
  [8299] = {.lex_state = 0, .external_lex_state = 271},
  [8300] = {.lex_state = 0, .external_lex_state = 249},
  [8301] = {.lex_state = 0, .external_lex_state = 186},
  [8302] = {.lex_state = 0, .external_lex_state = 186},
  [8303] = {.lex_state = 0, .external_lex_state = 282},
  [8304] = {.lex_state = 0, .external_lex_state = 186},
  [8305] = {.lex_state = 0, .external_lex_state = 249},
  [8306] = {.lex_state = 0, .external_lex_state = 186},
  [8307] = {.lex_state = 0, .external_lex_state = 186},
  [8308] = {.lex_state = 0, .external_lex_state = 249},
  [8309] = {.lex_state = 0, .external_lex_state = 186},
  [8310] = {.lex_state = 0, .external_lex_state = 186},
  [8311] = {.lex_state = 0, .external_lex_state = 186},
  [8312] = {.lex_state = 0, .external_lex_state = 186},
  [8313] = {.lex_state = 0, .external_lex_state = 186},
  [8314] = {.lex_state = 0, .external_lex_state = 259},
  [8315] = {.lex_state = 0, .external_lex_state = 186},
  [8316] = {.lex_state = 0, .external_lex_state = 267},
  [8317] = {.lex_state = 0, .external_lex_state = 186},
  [8318] = {.lex_state = 0, .external_lex_state = 186},
  [8319] = {.lex_state = 0, .external_lex_state = 186},
  [8320] = {.lex_state = 0, .external_lex_state = 186},
  [8321] = {.lex_state = 0, .external_lex_state = 186},
  [8322] = {.lex_state = 0, .external_lex_state = 186},
  [8323] = {.lex_state = 0, .external_lex_state = 186},
  [8324] = {.lex_state = 0, .external_lex_state = 186},
  [8325] = {.lex_state = 0, .external_lex_state = 186},
  [8326] = {.lex_state = 0, .external_lex_state = 186},
  [8327] = {.lex_state = 0, .external_lex_state = 186},
  [8328] = {.lex_state = 0, .external_lex_state = 186},
  [8329] = {.lex_state = 0, .external_lex_state = 186},
  [8330] = {.lex_state = 0, .external_lex_state = 269},
  [8331] = {.lex_state = 0, .external_lex_state = 186},
  [8332] = {.lex_state = 0, .external_lex_state = 186},
  [8333] = {.lex_state = 0, .external_lex_state = 186},
  [8334] = {.lex_state = 0, .external_lex_state = 186},
  [8335] = {.lex_state = 0, .external_lex_state = 186},
  [8336] = {.lex_state = 0, .external_lex_state = 186},
  [8337] = {.lex_state = 0, .external_lex_state = 186},
  [8338] = {.lex_state = 0, .external_lex_state = 186},
  [8339] = {.lex_state = 0, .external_lex_state = 186},
  [8340] = {.lex_state = 0, .external_lex_state = 271},
  [8341] = {.lex_state = 0, .external_lex_state = 186},
  [8342] = {.lex_state = 0, .external_lex_state = 186},
  [8343] = {.lex_state = 0, .external_lex_state = 186},
  [8344] = {.lex_state = 0, .external_lex_state = 186},
  [8345] = {.lex_state = 0, .external_lex_state = 186},
  [8346] = {.lex_state = 0, .external_lex_state = 186},
  [8347] = {.lex_state = 0, .external_lex_state = 186},
  [8348] = {.lex_state = 0, .external_lex_state = 186},
  [8349] = {.lex_state = 0, .external_lex_state = 186},
  [8350] = {.lex_state = 0, .external_lex_state = 186},
  [8351] = {.lex_state = 0, .external_lex_state = 271},
  [8352] = {.lex_state = 0, .external_lex_state = 290},
  [8353] = {.lex_state = 0, .external_lex_state = 259},
  [8354] = {.lex_state = 0, .external_lex_state = 271},
  [8355] = {.lex_state = 0, .external_lex_state = 186},
  [8356] = {.lex_state = 0, .external_lex_state = 186},
  [8357] = {.lex_state = 0, .external_lex_state = 186},
  [8358] = {.lex_state = 0, .external_lex_state = 186},
  [8359] = {.lex_state = 0, .external_lex_state = 186},
  [8360] = {.lex_state = 0, .external_lex_state = 186},
  [8361] = {.lex_state = 0, .external_lex_state = 269},
  [8362] = {(TSStateId)(-1),},
  [8363] = {(TSStateId)(-1),},
  [8364] = {(TSStateId)(-1),},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [sym__tool_directive] = STATE(0),
    [sym_line_comment] = STATE(0),
    [sym_block_comment] = STATE(0),
    [ts_builtin_sym_end] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_EQ_GT] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [sym_variable_assignment] = ACTIONS(1),
    [sym_condition_conversion] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_QMARK_EQ] = ACTIONS(1),
    [anon_sym_QMARK_SLASH_EQ] = ACTIONS(1),
    [anon_sym_QMARK_GT] = ACTIONS(1),
    [anon_sym_QMARK_GT_EQ] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [sym_exponentiate] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [sym_ABS] = ACTIONS(1),
    [sym_ACCESS] = ACTIONS(1),
    [sym_AFTER] = ACTIONS(1),
    [sym_ALIAS] = ACTIONS(1),
    [sym_ALL] = ACTIONS(1),
    [sym_AND] = ACTIONS(1),
    [sym_ARCHITECTURE] = ACTIONS(1),
    [sym_ARRAY] = ACTIONS(1),
    [sym_ASSERT] = ACTIONS(1),
    [sym_ASSUME] = ACTIONS(1),
    [sym_ATTRIBUTE] = ACTIONS(1),
    [sym_BEGIN] = ACTIONS(1),
    [sym_BLOCK] = ACTIONS(1),
    [sym_BODY] = ACTIONS(1),
    [sym_BUFFER] = ACTIONS(1),
    [sym_BUS] = ACTIONS(1),
    [sym_CASE] = ACTIONS(1),
    [sym_COMPONENT] = ACTIONS(1),
    [sym_CONFIGURATION] = ACTIONS(1),
    [sym_CONSTANT] = ACTIONS(1),
    [sym_CONTEXT] = ACTIONS(1),
    [sym_COVER] = ACTIONS(1),
    [sym_DEFAULT] = ACTIONS(1),
    [sym_DISCONNECT] = ACTIONS(1),
    [sym_DOWNTO] = ACTIONS(1),
    [sym_ELSE] = ACTIONS(1),
    [sym_ELSIF] = ACTIONS(1),
    [sym_END] = ACTIONS(1),
    [sym_ENTITY] = ACTIONS(1),
    [sym_EXIT] = ACTIONS(1),
    [sym_FAIRNESS] = ACTIONS(1),
    [sym_FILE] = ACTIONS(1),
    [sym_FOR] = ACTIONS(1),
    [sym_FORCE] = ACTIONS(1),
    [sym_FUNCTION] = ACTIONS(1),
    [sym_GENERATE] = ACTIONS(1),
    [sym_GENERIC] = ACTIONS(1),
    [sym_GROUP] = ACTIONS(1),
    [sym_GUARDED] = ACTIONS(1),
    [sym_IF] = ACTIONS(1),
    [sym_IMPURE] = ACTIONS(1),
    [sym_IN] = ACTIONS(1),
    [sym_INERTIAL] = ACTIONS(1),
    [sym_INOUT] = ACTIONS(1),
    [sym_IS] = ACTIONS(1),
    [sym_LABEL] = ACTIONS(1),
    [sym_LIBRARY] = ACTIONS(1),
    [sym_LINKAGE] = ACTIONS(1),
    [sym_LITERAL] = ACTIONS(1),
    [sym_LOOP] = ACTIONS(1),
    [sym_MAP] = ACTIONS(1),
    [sym_MOD] = ACTIONS(1),
    [sym_NAND] = ACTIONS(1),
    [sym_NEW] = ACTIONS(1),
    [sym_NEXT] = ACTIONS(1),
    [sym_NOR] = ACTIONS(1),
    [sym_NOT] = ACTIONS(1),
    [sym_NULL] = ACTIONS(1),
    [sym_OF] = ACTIONS(1),
    [sym_ON] = ACTIONS(1),
    [sym_OPEN] = ACTIONS(1),
    [sym_OR] = ACTIONS(1),
    [sym_OTHERS] = ACTIONS(1),
    [sym_OUT] = ACTIONS(1),
    [sym_PACKAGE] = ACTIONS(1),
    [sym_PARAMETER] = ACTIONS(1),
    [sym_PORT] = ACTIONS(1),
    [sym_POSTPONED] = ACTIONS(1),
    [sym_PROCEDURE] = ACTIONS(1),
    [sym_PROCESS] = ACTIONS(1),
    [sym_PROPERTY] = ACTIONS(1),
    [sym_PROTECTED] = ACTIONS(1),
    [sym_PRIVATE] = ACTIONS(1),
    [sym_PURE] = ACTIONS(1),
    [sym_RANGE] = ACTIONS(1),
    [sym_RECORD] = ACTIONS(1),
    [sym_REGISTER] = ACTIONS(1),
    [sym_REJECT] = ACTIONS(1),
    [sym_RELEASE] = ACTIONS(1),
    [sym_REM] = ACTIONS(1),
    [sym_REPORT] = ACTIONS(1),
    [sym_RESTRICT] = ACTIONS(1),
    [sym_RETURN] = ACTIONS(1),
    [sym_ROL] = ACTIONS(1),
    [sym_ROR] = ACTIONS(1),
    [sym_SELECT] = ACTIONS(1),
    [sym_SEQUENCE] = ACTIONS(1),
    [sym_SEVERITY] = ACTIONS(1),
    [sym_SIGNAL] = ACTIONS(1),
    [sym_SHARED] = ACTIONS(1),
    [sym_SLA] = ACTIONS(1),
    [sym_SLL] = ACTIONS(1),
    [sym_SRA] = ACTIONS(1),
    [sym_SRL] = ACTIONS(1),
    [sym_STRONG] = ACTIONS(1),
    [sym_SUBTYPE] = ACTIONS(1),
    [sym_THEN] = ACTIONS(1),
    [sym_TO] = ACTIONS(1),
    [sym_TRANSPORT] = ACTIONS(1),
    [sym_TYPE] = ACTIONS(1),
    [sym_UNAFFECTED] = ACTIONS(1),
    [sym_UNITS] = ACTIONS(1),
    [sym_UNTIL] = ACTIONS(1),
    [sym_USE] = ACTIONS(1),
    [sym_VARIABLE] = ACTIONS(1),
    [sym_VIEW] = ACTIONS(1),
    [sym_VMODE] = ACTIONS(1),
    [sym_VPKG] = ACTIONS(1),
    [sym_VPROP] = ACTIONS(1),
    [sym_VUNIT] = ACTIONS(1),
    [sym_WAIT] = ACTIONS(1),
    [sym_WHEN] = ACTIONS(1),
    [sym_WHILE] = ACTIONS(1),
    [sym_WITH] = ACTIONS(1),
    [sym_XNOR] = ACTIONS(1),
    [sym_XOR] = ACTIONS(1),
    [sym_reserved_end_marker] = ACTIONS(1),
    [sym_directive_body] = ACTIONS(1),
    [sym_directive_constant_builtin] = ACTIONS(1),
    [sym_directive_error] = ACTIONS(1),
    [sym_directive_protect] = ACTIONS(1),
    [sym_directive_warning] = ACTIONS(1),
    [sym__directive_newline] = ACTIONS(1),
    [sym__grave_accent] = ACTIONS(3),
    [sym_box] = ACTIONS(1),
    [sym_delimiter_end_marker] = ACTIONS(1),
    [sym_decimal_integer] = ACTIONS(1),
    [sym_decimal_float] = ACTIONS(1),
    [sym_based_base] = ACTIONS(1),
    [sym_based_integer] = ACTIONS(1),
    [sym_based_float] = ACTIONS(1),
    [sym_character_literal] = ACTIONS(1),
    [sym_string_literal] = ACTIONS(1),
    [sym_string_literal_std_logic] = ACTIONS(1),
    [sym_bit_string_length] = ACTIONS(1),
    [sym_bit_string_base] = ACTIONS(1),
    [sym_bit_string_value] = ACTIONS(1),
    [sym_operator_symbol] = ACTIONS(1),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym__block_comment_end] = ACTIONS(1),
    [sym_comment_content] = ACTIONS(1),
    [sym_token_end_marker] = ACTIONS(1),
    [sym_attribute_function] = ACTIONS(1),
    [sym_attribute_impure_function] = ACTIONS(1),
    [sym_attribute_mode_view] = ACTIONS(1),
    [sym_attribute_pure_function] = ACTIONS(1),
    [sym_attribute_range] = ACTIONS(1),
    [sym_attribute_signal] = ACTIONS(1),
    [sym_attribute_subtype] = ACTIONS(1),
    [sym_attribute_type] = ACTIONS(1),
    [sym_attribute_value] = ACTIONS(1),
    [sym_library_attribute] = ACTIONS(1),
    [sym_library_constant] = ACTIONS(1),
    [sym_library_constant_boolean] = ACTIONS(1),
    [sym_library_constant_character] = ACTIONS(1),
    [sym_library_constant_debug] = ACTIONS(1),
    [sym_library_constant_env] = ACTIONS(1),
    [sym_library_constant_standard] = ACTIONS(1),
    [sym_library_constant_std_logic] = ACTIONS(1),
    [sym_library_constant_unit] = ACTIONS(1),
    [sym_library_function] = ACTIONS(1),
    [sym_library_namespace] = ACTIONS(1),
    [sym_library_type] = ACTIONS(1),
    [sym__end_of_file] = ACTIONS(1),
    [sym_error_sentinel] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_design_file] = STATE(6981),
    [sym_design_unit] = STATE(2220),
    [sym__context_item] = STATE(2270),
    [sym_library_clause] = STATE(1917),
    [sym_use_clause] = STATE(1942),
    [sym_context_reference] = STATE(1917),
    [sym__library_unit] = STATE(1965),
    [sym_entity_declaration] = STATE(1979),
    [sym_configuration_declaration] = STATE(1979),
    [sym_package_declaration] = STATE(1982),
    [sym_package_instantiation_declaration] = STATE(1982),
    [sym_context_declaration] = STATE(1979),
    [sym_architecture_definition] = STATE(1979),
    [sym_package_definition] = STATE(1982),
    [sym__block_declarative_item] = STATE(2029),
    [sym_subprogram_declaration] = STATE(2040),
    [sym_subprogram_definition] = STATE(2040),
    [sym_subprogram_instantiation_declaration] = STATE(2040),
    [sym_type_declaration] = STATE(2040),
    [sym_subtype_declaration] = STATE(2040),
    [sym_mode_view_declaration] = STATE(2040),
    [sym_constant_declaration] = STATE(2040),
    [sym_signal_declaration] = STATE(2040),
    [sym_variable_declaration] = STATE(2040),
    [sym_file_declaration] = STATE(2040),
    [sym_alias_declaration] = STATE(2040),
    [sym_component_declaration] = STATE(2040),
    [sym_attribute_declaration] = STATE(2040),
    [sym_attribute_specification] = STATE(2040),
    [sym_disconnection_specification] = STATE(2040),
    [sym_group_template_declaration] = STATE(2040),
    [sym_group_declaration] = STATE(2040),
    [sym__concurrent_statement] = STATE(2052),
    [sym_block_statement] = STATE(2060),
    [sym_component_instantiation_statement] = STATE(2060),
    [sym_process_statement] = STATE(2060),
    [sym_case_generate_statement] = STATE(2060),
    [sym_for_generate_statement] = STATE(2060),
    [sym_if_generate_statement] = STATE(2060),
    [sym_concurrent_assertion_statement] = STATE(2060),
    [sym_assertion] = STATE(7241),
    [sym_concurrent_procedure_call_statement] = STATE(2060),
    [sym_concurrent_simple_signal_assignment] = STATE(2060),
    [sym_concurrent_conditional_signal_assignment] = STATE(2060),
    [sym_concurrent_selected_signal_assignment] = STATE(2060),
    [sym_with_expression] = STATE(6491),
    [sym_name] = STATE(6540),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7169),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(1),
    [sym_procedure_specification] = STATE(6069),
    [sym_function_specification] = STATE(6069),
    [sym__subprogram_specification] = STATE(5708),
    [sym__target] = STATE(6080),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(2689),
    [sym_configuration_specification] = STATE(2040),
    [sym_line_comment] = STATE(1),
    [sym_block_comment] = STATE(1),
    [aux_sym_design_file_repeat1] = STATE(3),
    [aux_sym_design_unit_repeat1] = STATE(740),
    [aux_sym_design_unit_repeat2] = STATE(94),
    [aux_sym_design_unit_repeat3] = STATE(124),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ALIAS] = ACTIONS(15),
    [sym_ARCHITECTURE] = ACTIONS(17),
    [sym_ASSERT] = ACTIONS(19),
    [sym_ATTRIBUTE] = ACTIONS(21),
    [sym_COMPONENT] = ACTIONS(23),
    [sym_CONFIGURATION] = ACTIONS(25),
    [sym_CONSTANT] = ACTIONS(27),
    [sym_CONTEXT] = ACTIONS(29),
    [sym_DISCONNECT] = ACTIONS(31),
    [sym_ENTITY] = ACTIONS(33),
    [sym_FILE] = ACTIONS(35),
    [sym_FOR] = ACTIONS(37),
    [sym_FUNCTION] = ACTIONS(39),
    [sym_GROUP] = ACTIONS(41),
    [sym_IMPURE] = ACTIONS(43),
    [sym_LIBRARY] = ACTIONS(45),
    [sym_PACKAGE] = ACTIONS(47),
    [sym_POSTPONED] = ACTIONS(49),
    [sym_PROCEDURE] = ACTIONS(51),
    [sym_PROCESS] = ACTIONS(53),
    [sym_PURE] = ACTIONS(43),
    [sym_SIGNAL] = ACTIONS(55),
    [sym_SHARED] = ACTIONS(57),
    [sym_SUBTYPE] = ACTIONS(59),
    [sym_TYPE] = ACTIONS(61),
    [sym_USE] = ACTIONS(63),
    [sym_VARIABLE] = ACTIONS(65),
    [sym_VIEW] = ACTIONS(67),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
    [sym__end_of_file] = ACTIONS(73),
  },
  [STATE(2)] = {
    [sym_design_unit] = STATE(2220),
    [sym__context_item] = STATE(2270),
    [sym_library_clause] = STATE(1917),
    [sym_use_clause] = STATE(1942),
    [sym_context_reference] = STATE(1917),
    [sym__library_unit] = STATE(1965),
    [sym_entity_declaration] = STATE(1979),
    [sym_configuration_declaration] = STATE(1979),
    [sym_package_declaration] = STATE(1982),
    [sym_package_instantiation_declaration] = STATE(1982),
    [sym_context_declaration] = STATE(1979),
    [sym_architecture_definition] = STATE(1979),
    [sym_package_definition] = STATE(1982),
    [sym__block_declarative_item] = STATE(2029),
    [sym_subprogram_declaration] = STATE(2040),
    [sym_subprogram_definition] = STATE(2040),
    [sym_subprogram_instantiation_declaration] = STATE(2040),
    [sym_type_declaration] = STATE(2040),
    [sym_subtype_declaration] = STATE(2040),
    [sym_mode_view_declaration] = STATE(2040),
    [sym_constant_declaration] = STATE(2040),
    [sym_signal_declaration] = STATE(2040),
    [sym_variable_declaration] = STATE(2040),
    [sym_file_declaration] = STATE(2040),
    [sym_alias_declaration] = STATE(2040),
    [sym_component_declaration] = STATE(2040),
    [sym_attribute_declaration] = STATE(2040),
    [sym_attribute_specification] = STATE(2040),
    [sym_disconnection_specification] = STATE(2040),
    [sym_group_template_declaration] = STATE(2040),
    [sym_group_declaration] = STATE(2040),
    [sym__concurrent_statement] = STATE(2052),
    [sym_block_statement] = STATE(2060),
    [sym_component_instantiation_statement] = STATE(2060),
    [sym_process_statement] = STATE(2060),
    [sym_case_generate_statement] = STATE(2060),
    [sym_for_generate_statement] = STATE(2060),
    [sym_if_generate_statement] = STATE(2060),
    [sym_concurrent_assertion_statement] = STATE(2060),
    [sym_assertion] = STATE(7241),
    [sym_concurrent_procedure_call_statement] = STATE(2060),
    [sym_concurrent_simple_signal_assignment] = STATE(2060),
    [sym_concurrent_conditional_signal_assignment] = STATE(2060),
    [sym_concurrent_selected_signal_assignment] = STATE(2060),
    [sym_with_expression] = STATE(6491),
    [sym_name] = STATE(6540),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7169),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(2),
    [sym_procedure_specification] = STATE(6069),
    [sym_function_specification] = STATE(6069),
    [sym__subprogram_specification] = STATE(5708),
    [sym__target] = STATE(6080),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(2689),
    [sym_configuration_specification] = STATE(2040),
    [sym_line_comment] = STATE(2),
    [sym_block_comment] = STATE(2),
    [aux_sym_design_file_repeat1] = STATE(2),
    [aux_sym_design_unit_repeat1] = STATE(740),
    [aux_sym_design_unit_repeat2] = STATE(94),
    [aux_sym_design_unit_repeat3] = STATE(124),
    [anon_sym_LPAREN] = ACTIONS(75),
    [anon_sym_LT_LT] = ACTIONS(78),
    [sym_identifier] = ACTIONS(81),
    [sym_ALIAS] = ACTIONS(84),
    [sym_ARCHITECTURE] = ACTIONS(87),
    [sym_ASSERT] = ACTIONS(90),
    [sym_ATTRIBUTE] = ACTIONS(93),
    [sym_COMPONENT] = ACTIONS(96),
    [sym_CONFIGURATION] = ACTIONS(99),
    [sym_CONSTANT] = ACTIONS(102),
    [sym_CONTEXT] = ACTIONS(105),
    [sym_DISCONNECT] = ACTIONS(108),
    [sym_ENTITY] = ACTIONS(111),
    [sym_FILE] = ACTIONS(114),
    [sym_FOR] = ACTIONS(117),
    [sym_FUNCTION] = ACTIONS(120),
    [sym_GROUP] = ACTIONS(123),
    [sym_IMPURE] = ACTIONS(126),
    [sym_LIBRARY] = ACTIONS(129),
    [sym_PACKAGE] = ACTIONS(132),
    [sym_POSTPONED] = ACTIONS(135),
    [sym_PROCEDURE] = ACTIONS(138),
    [sym_PROCESS] = ACTIONS(141),
    [sym_PURE] = ACTIONS(126),
    [sym_SIGNAL] = ACTIONS(144),
    [sym_SHARED] = ACTIONS(147),
    [sym_SUBTYPE] = ACTIONS(150),
    [sym_TYPE] = ACTIONS(153),
    [sym_USE] = ACTIONS(156),
    [sym_VARIABLE] = ACTIONS(159),
    [sym_VIEW] = ACTIONS(162),
    [sym_WITH] = ACTIONS(165),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(168),
    [sym_operator_symbol] = ACTIONS(168),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(81),
    [sym_library_constant_std_logic] = ACTIONS(168),
    [sym_library_function] = ACTIONS(81),
    [sym_library_type] = ACTIONS(81),
    [sym__end_of_file] = ACTIONS(171),
  },
  [STATE(3)] = {
    [sym_design_unit] = STATE(2220),
    [sym__context_item] = STATE(2270),
    [sym_library_clause] = STATE(1917),
    [sym_use_clause] = STATE(1942),
    [sym_context_reference] = STATE(1917),
    [sym__library_unit] = STATE(1965),
    [sym_entity_declaration] = STATE(1979),
    [sym_configuration_declaration] = STATE(1979),
    [sym_package_declaration] = STATE(1982),
    [sym_package_instantiation_declaration] = STATE(1982),
    [sym_context_declaration] = STATE(1979),
    [sym_architecture_definition] = STATE(1979),
    [sym_package_definition] = STATE(1982),
    [sym__block_declarative_item] = STATE(2029),
    [sym_subprogram_declaration] = STATE(2040),
    [sym_subprogram_definition] = STATE(2040),
    [sym_subprogram_instantiation_declaration] = STATE(2040),
    [sym_type_declaration] = STATE(2040),
    [sym_subtype_declaration] = STATE(2040),
    [sym_mode_view_declaration] = STATE(2040),
    [sym_constant_declaration] = STATE(2040),
    [sym_signal_declaration] = STATE(2040),
    [sym_variable_declaration] = STATE(2040),
    [sym_file_declaration] = STATE(2040),
    [sym_alias_declaration] = STATE(2040),
    [sym_component_declaration] = STATE(2040),
    [sym_attribute_declaration] = STATE(2040),
    [sym_attribute_specification] = STATE(2040),
    [sym_disconnection_specification] = STATE(2040),
    [sym_group_template_declaration] = STATE(2040),
    [sym_group_declaration] = STATE(2040),
    [sym__concurrent_statement] = STATE(2052),
    [sym_block_statement] = STATE(2060),
    [sym_component_instantiation_statement] = STATE(2060),
    [sym_process_statement] = STATE(2060),
    [sym_case_generate_statement] = STATE(2060),
    [sym_for_generate_statement] = STATE(2060),
    [sym_if_generate_statement] = STATE(2060),
    [sym_concurrent_assertion_statement] = STATE(2060),
    [sym_assertion] = STATE(7241),
    [sym_concurrent_procedure_call_statement] = STATE(2060),
    [sym_concurrent_simple_signal_assignment] = STATE(2060),
    [sym_concurrent_conditional_signal_assignment] = STATE(2060),
    [sym_concurrent_selected_signal_assignment] = STATE(2060),
    [sym_with_expression] = STATE(6491),
    [sym_name] = STATE(6540),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7169),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(3),
    [sym_procedure_specification] = STATE(6069),
    [sym_function_specification] = STATE(6069),
    [sym__subprogram_specification] = STATE(5708),
    [sym__target] = STATE(6080),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(2689),
    [sym_configuration_specification] = STATE(2040),
    [sym_line_comment] = STATE(3),
    [sym_block_comment] = STATE(3),
    [aux_sym_design_file_repeat1] = STATE(2),
    [aux_sym_design_unit_repeat1] = STATE(740),
    [aux_sym_design_unit_repeat2] = STATE(94),
    [aux_sym_design_unit_repeat3] = STATE(124),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ALIAS] = ACTIONS(15),
    [sym_ARCHITECTURE] = ACTIONS(17),
    [sym_ASSERT] = ACTIONS(19),
    [sym_ATTRIBUTE] = ACTIONS(21),
    [sym_COMPONENT] = ACTIONS(23),
    [sym_CONFIGURATION] = ACTIONS(25),
    [sym_CONSTANT] = ACTIONS(27),
    [sym_CONTEXT] = ACTIONS(29),
    [sym_DISCONNECT] = ACTIONS(31),
    [sym_ENTITY] = ACTIONS(33),
    [sym_FILE] = ACTIONS(35),
    [sym_FOR] = ACTIONS(37),
    [sym_FUNCTION] = ACTIONS(39),
    [sym_GROUP] = ACTIONS(41),
    [sym_IMPURE] = ACTIONS(43),
    [sym_LIBRARY] = ACTIONS(45),
    [sym_PACKAGE] = ACTIONS(47),
    [sym_POSTPONED] = ACTIONS(49),
    [sym_PROCEDURE] = ACTIONS(51),
    [sym_PROCESS] = ACTIONS(53),
    [sym_PURE] = ACTIONS(43),
    [sym_SIGNAL] = ACTIONS(55),
    [sym_SHARED] = ACTIONS(57),
    [sym_SUBTYPE] = ACTIONS(59),
    [sym_TYPE] = ACTIONS(61),
    [sym_USE] = ACTIONS(63),
    [sym_VARIABLE] = ACTIONS(65),
    [sym_VIEW] = ACTIONS(67),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
    [sym__end_of_file] = ACTIONS(173),
  },
  [STATE(4)] = {
    [sym_use_clause] = STATE(3071),
    [sym_package_declaration] = STATE(3071),
    [sym_package_instantiation_declaration] = STATE(3071),
    [sym_package_definition] = STATE(3071),
    [sym__block_declarative_item] = STATE(3070),
    [sym_subprogram_declaration] = STATE(3071),
    [sym_subprogram_definition] = STATE(3071),
    [sym_subprogram_instantiation_declaration] = STATE(3071),
    [sym_type_declaration] = STATE(3071),
    [sym_subtype_declaration] = STATE(3071),
    [sym_mode_view_declaration] = STATE(3071),
    [sym_constant_declaration] = STATE(3071),
    [sym_signal_declaration] = STATE(3071),
    [sym_variable_declaration] = STATE(3071),
    [sym_file_declaration] = STATE(3071),
    [sym_alias_declaration] = STATE(3071),
    [sym_component_declaration] = STATE(3071),
    [sym_attribute_declaration] = STATE(3071),
    [sym_attribute_specification] = STATE(3071),
    [sym_disconnection_specification] = STATE(3071),
    [sym_group_template_declaration] = STATE(3071),
    [sym_group_declaration] = STATE(3071),
    [sym_generate_head] = STATE(7604),
    [sym_generate_block] = STATE(5556),
    [sym__concurrent_statement] = STATE(3725),
    [sym_block_statement] = STATE(3814),
    [sym_component_instantiation_statement] = STATE(3814),
    [sym_process_statement] = STATE(3814),
    [sym_case_generate_statement] = STATE(3814),
    [sym_for_generate_statement] = STATE(3814),
    [sym_if_generate_statement] = STATE(3814),
    [sym_concurrent_assertion_statement] = STATE(3814),
    [sym_assertion] = STATE(8292),
    [sym_concurrent_procedure_call_statement] = STATE(3814),
    [sym_concurrent_simple_signal_assignment] = STATE(3814),
    [sym_concurrent_conditional_signal_assignment] = STATE(3814),
    [sym_concurrent_selected_signal_assignment] = STATE(3814),
    [sym_with_expression] = STATE(6480),
    [sym_name] = STATE(6481),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7169),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(4),
    [sym_procedure_specification] = STATE(6069),
    [sym_function_specification] = STATE(6069),
    [sym__subprogram_specification] = STATE(5909),
    [sym__target] = STATE(6403),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(2658),
    [sym_configuration_specification] = STATE(3071),
    [sym_line_comment] = STATE(4),
    [sym_block_comment] = STATE(4),
    [aux_sym_design_unit_repeat2] = STATE(796),
    [aux_sym_design_unit_repeat3] = STATE(1401),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ALIAS] = ACTIONS(175),
    [sym_ASSERT] = ACTIONS(19),
    [sym_ATTRIBUTE] = ACTIONS(177),
    [sym_BEGIN] = ACTIONS(179),
    [sym_COMPONENT] = ACTIONS(181),
    [sym_CONSTANT] = ACTIONS(183),
    [sym_DISCONNECT] = ACTIONS(185),
    [sym_ELSE] = ACTIONS(187),
    [sym_ELSIF] = ACTIONS(187),
    [sym_END] = ACTIONS(187),
    [sym_FILE] = ACTIONS(189),
    [sym_FOR] = ACTIONS(191),
    [sym_FUNCTION] = ACTIONS(193),
    [sym_GROUP] = ACTIONS(195),
    [sym_IMPURE] = ACTIONS(43),
    [sym_PACKAGE] = ACTIONS(197),
    [sym_POSTPONED] = ACTIONS(199),
    [sym_PROCEDURE] = ACTIONS(51),
    [sym_PROCESS] = ACTIONS(201),
    [sym_PURE] = ACTIONS(43),
    [sym_SIGNAL] = ACTIONS(203),
    [sym_SHARED] = ACTIONS(205),
    [sym_SUBTYPE] = ACTIONS(207),
    [sym_TYPE] = ACTIONS(209),
    [sym_USE] = ACTIONS(211),
    [sym_VARIABLE] = ACTIONS(213),
    [sym_VIEW] = ACTIONS(215),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(5)] = {
    [sym_use_clause] = STATE(3071),
    [sym_package_declaration] = STATE(3071),
    [sym_package_instantiation_declaration] = STATE(3071),
    [sym_package_definition] = STATE(3071),
    [sym__block_declarative_item] = STATE(3070),
    [sym_subprogram_declaration] = STATE(3071),
    [sym_subprogram_definition] = STATE(3071),
    [sym_subprogram_instantiation_declaration] = STATE(3071),
    [sym_type_declaration] = STATE(3071),
    [sym_subtype_declaration] = STATE(3071),
    [sym_mode_view_declaration] = STATE(3071),
    [sym_constant_declaration] = STATE(3071),
    [sym_signal_declaration] = STATE(3071),
    [sym_variable_declaration] = STATE(3071),
    [sym_file_declaration] = STATE(3071),
    [sym_alias_declaration] = STATE(3071),
    [sym_component_declaration] = STATE(3071),
    [sym_attribute_declaration] = STATE(3071),
    [sym_attribute_specification] = STATE(3071),
    [sym_disconnection_specification] = STATE(3071),
    [sym_group_template_declaration] = STATE(3071),
    [sym_group_declaration] = STATE(3071),
    [sym_generate_head] = STATE(8272),
    [sym_generate_block] = STATE(6372),
    [sym__concurrent_statement] = STATE(4169),
    [sym_block_statement] = STATE(4170),
    [sym_component_instantiation_statement] = STATE(4170),
    [sym_process_statement] = STATE(4170),
    [sym_case_generate_statement] = STATE(4170),
    [sym_for_generate_statement] = STATE(4170),
    [sym_if_generate_statement] = STATE(4170),
    [sym_concurrent_assertion_statement] = STATE(4170),
    [sym_assertion] = STATE(6638),
    [sym_concurrent_procedure_call_statement] = STATE(4170),
    [sym_concurrent_simple_signal_assignment] = STATE(4170),
    [sym_concurrent_conditional_signal_assignment] = STATE(4170),
    [sym_concurrent_selected_signal_assignment] = STATE(4170),
    [sym_with_expression] = STATE(6493),
    [sym_name] = STATE(6555),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7169),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(5),
    [sym_procedure_specification] = STATE(6069),
    [sym_function_specification] = STATE(6069),
    [sym__subprogram_specification] = STATE(5909),
    [sym__target] = STATE(6435),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(2636),
    [sym_configuration_specification] = STATE(3071),
    [sym_line_comment] = STATE(5),
    [sym_block_comment] = STATE(5),
    [aux_sym_design_unit_repeat2] = STATE(796),
    [aux_sym_design_unit_repeat3] = STATE(1603),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ALIAS] = ACTIONS(175),
    [sym_ASSERT] = ACTIONS(19),
    [sym_ATTRIBUTE] = ACTIONS(177),
    [sym_BEGIN] = ACTIONS(217),
    [sym_COMPONENT] = ACTIONS(181),
    [sym_CONSTANT] = ACTIONS(183),
    [sym_DISCONNECT] = ACTIONS(185),
    [sym_END] = ACTIONS(219),
    [sym_FILE] = ACTIONS(189),
    [sym_FOR] = ACTIONS(191),
    [sym_FUNCTION] = ACTIONS(193),
    [sym_GROUP] = ACTIONS(195),
    [sym_IMPURE] = ACTIONS(43),
    [sym_PACKAGE] = ACTIONS(197),
    [sym_POSTPONED] = ACTIONS(221),
    [sym_PROCEDURE] = ACTIONS(51),
    [sym_PROCESS] = ACTIONS(223),
    [sym_PURE] = ACTIONS(43),
    [sym_SIGNAL] = ACTIONS(203),
    [sym_SHARED] = ACTIONS(205),
    [sym_SUBTYPE] = ACTIONS(207),
    [sym_TYPE] = ACTIONS(209),
    [sym_USE] = ACTIONS(211),
    [sym_VARIABLE] = ACTIONS(213),
    [sym_VIEW] = ACTIONS(215),
    [sym_WHEN] = ACTIONS(219),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(6)] = {
    [sym_use_clause] = STATE(3071),
    [sym_package_declaration] = STATE(3071),
    [sym_package_instantiation_declaration] = STATE(3071),
    [sym_package_definition] = STATE(3071),
    [sym__block_declarative_item] = STATE(3070),
    [sym_subprogram_declaration] = STATE(3071),
    [sym_subprogram_definition] = STATE(3071),
    [sym_subprogram_instantiation_declaration] = STATE(3071),
    [sym_type_declaration] = STATE(3071),
    [sym_subtype_declaration] = STATE(3071),
    [sym_mode_view_declaration] = STATE(3071),
    [sym_constant_declaration] = STATE(3071),
    [sym_signal_declaration] = STATE(3071),
    [sym_variable_declaration] = STATE(3071),
    [sym_file_declaration] = STATE(3071),
    [sym_alias_declaration] = STATE(3071),
    [sym_component_declaration] = STATE(3071),
    [sym_attribute_declaration] = STATE(3071),
    [sym_attribute_specification] = STATE(3071),
    [sym_disconnection_specification] = STATE(3071),
    [sym_group_template_declaration] = STATE(3071),
    [sym_group_declaration] = STATE(3071),
    [sym_generate_head] = STATE(6950),
    [sym_generate_block] = STATE(6955),
    [sym__concurrent_statement] = STATE(4386),
    [sym_block_statement] = STATE(4354),
    [sym_component_instantiation_statement] = STATE(4354),
    [sym_process_statement] = STATE(4354),
    [sym_case_generate_statement] = STATE(4354),
    [sym_for_generate_statement] = STATE(4354),
    [sym_if_generate_statement] = STATE(4354),
    [sym_concurrent_assertion_statement] = STATE(4354),
    [sym_assertion] = STATE(7448),
    [sym_concurrent_procedure_call_statement] = STATE(4354),
    [sym_concurrent_simple_signal_assignment] = STATE(4354),
    [sym_concurrent_conditional_signal_assignment] = STATE(4354),
    [sym_concurrent_selected_signal_assignment] = STATE(4354),
    [sym_with_expression] = STATE(6470),
    [sym_name] = STATE(6317),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7169),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(6),
    [sym_procedure_specification] = STATE(6069),
    [sym_function_specification] = STATE(6069),
    [sym__subprogram_specification] = STATE(5909),
    [sym__target] = STATE(6373),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(2626),
    [sym_configuration_specification] = STATE(3071),
    [sym_line_comment] = STATE(6),
    [sym_block_comment] = STATE(6),
    [aux_sym_design_unit_repeat2] = STATE(796),
    [aux_sym_design_unit_repeat3] = STATE(1727),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ALIAS] = ACTIONS(175),
    [sym_ASSERT] = ACTIONS(19),
    [sym_ATTRIBUTE] = ACTIONS(177),
    [sym_BEGIN] = ACTIONS(225),
    [sym_COMPONENT] = ACTIONS(181),
    [sym_CONSTANT] = ACTIONS(183),
    [sym_DISCONNECT] = ACTIONS(185),
    [sym_END] = ACTIONS(187),
    [sym_FILE] = ACTIONS(189),
    [sym_FOR] = ACTIONS(191),
    [sym_FUNCTION] = ACTIONS(193),
    [sym_GROUP] = ACTIONS(195),
    [sym_IMPURE] = ACTIONS(43),
    [sym_PACKAGE] = ACTIONS(197),
    [sym_POSTPONED] = ACTIONS(227),
    [sym_PROCEDURE] = ACTIONS(51),
    [sym_PROCESS] = ACTIONS(229),
    [sym_PURE] = ACTIONS(43),
    [sym_SIGNAL] = ACTIONS(203),
    [sym_SHARED] = ACTIONS(205),
    [sym_SUBTYPE] = ACTIONS(207),
    [sym_TYPE] = ACTIONS(209),
    [sym_USE] = ACTIONS(211),
    [sym_VARIABLE] = ACTIONS(213),
    [sym_VIEW] = ACTIONS(215),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(7)] = {
    [sym__sequential_statement] = STATE(2824),
    [sym_sequential_block_statement] = STATE(2825),
    [sym_assertion_statement] = STATE(2825),
    [sym_assertion] = STATE(7497),
    [sym_report_expression] = STATE(5850),
    [sym_case_statement] = STATE(2825),
    [sym_case_expression] = STATE(6384),
    [sym_exit_statement] = STATE(2825),
    [sym_if_statement_block] = STATE(2825),
    [sym_if_statement] = STATE(6560),
    [sym__elsif_else_statement] = STATE(6714),
    [sym_elsif_statement] = STATE(7427),
    [sym_else_statement] = STATE(7427),
    [sym_if_statement_body] = STATE(5092),
    [sym_loop_statement] = STATE(2825),
    [sym__iteration_scheme] = STATE(6385),
    [sym_loop_body] = STATE(6561),
    [sym_next_statement] = STATE(2825),
    [sym_null_statement] = STATE(2825),
    [sym_procedure_call_statement] = STATE(2825),
    [sym_report_statement] = STATE(2825),
    [sym_return_statement] = STATE(2825),
    [sym_simple_waveform_assignment] = STATE(2825),
    [sym_simple_variable_assignment] = STATE(2825),
    [sym_simple_force_assignment] = STATE(2825),
    [sym_simple_release_assignment] = STATE(2825),
    [sym_conditional_signal_assignment] = STATE(2825),
    [sym_selected_waveform_assignment] = STATE(2825),
    [sym_selected_force_assignment] = STATE(2825),
    [sym_selected_variable_assignment] = STATE(2825),
    [sym_with_expression] = STATE(6474),
    [sym_wait_statement] = STATE(2825),
    [sym_name] = STATE(5851),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(7),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5885),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1529),
    [sym_line_comment] = STATE(7),
    [sym_block_comment] = STATE(7),
    [aux_sym_sequential_block_repeat1] = STATE(11),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(231),
    [sym_CASE] = ACTIONS(233),
    [sym_ELSE] = ACTIONS(235),
    [sym_ELSIF] = ACTIONS(237),
    [sym_END] = ACTIONS(239),
    [sym_EXIT] = ACTIONS(241),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(249),
    [sym_NULL] = ACTIONS(251),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(255),
    [sym_WAIT] = ACTIONS(257),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(8)] = {
    [sym__sequential_statement] = STATE(2824),
    [sym_sequential_block_statement] = STATE(2825),
    [sym_assertion_statement] = STATE(2825),
    [sym_assertion] = STATE(7497),
    [sym_report_expression] = STATE(5850),
    [sym_case_statement] = STATE(2825),
    [sym_case_expression] = STATE(6384),
    [sym_exit_statement] = STATE(2825),
    [sym_if_statement_block] = STATE(2825),
    [sym_if_statement] = STATE(6560),
    [sym__elsif_else_statement] = STATE(7426),
    [sym_elsif_statement] = STATE(7427),
    [sym_else_statement] = STATE(7427),
    [sym_if_statement_body] = STATE(5091),
    [sym_loop_statement] = STATE(2825),
    [sym__iteration_scheme] = STATE(6385),
    [sym_loop_body] = STATE(6561),
    [sym_next_statement] = STATE(2825),
    [sym_null_statement] = STATE(2825),
    [sym_procedure_call_statement] = STATE(2825),
    [sym_report_statement] = STATE(2825),
    [sym_return_statement] = STATE(2825),
    [sym_simple_waveform_assignment] = STATE(2825),
    [sym_simple_variable_assignment] = STATE(2825),
    [sym_simple_force_assignment] = STATE(2825),
    [sym_simple_release_assignment] = STATE(2825),
    [sym_conditional_signal_assignment] = STATE(2825),
    [sym_selected_waveform_assignment] = STATE(2825),
    [sym_selected_force_assignment] = STATE(2825),
    [sym_selected_variable_assignment] = STATE(2825),
    [sym_with_expression] = STATE(6474),
    [sym_wait_statement] = STATE(2825),
    [sym_name] = STATE(5851),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(8),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5885),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1529),
    [sym_line_comment] = STATE(8),
    [sym_block_comment] = STATE(8),
    [aux_sym_sequential_block_repeat1] = STATE(11),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(231),
    [sym_CASE] = ACTIONS(233),
    [sym_ELSE] = ACTIONS(235),
    [sym_ELSIF] = ACTIONS(237),
    [sym_END] = ACTIONS(261),
    [sym_EXIT] = ACTIONS(241),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(249),
    [sym_NULL] = ACTIONS(251),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(255),
    [sym_WAIT] = ACTIONS(257),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(9)] = {
    [sym_unspecified_type_indication] = STATE(6478),
    [sym_array_index_incomplete_type_list] = STATE(7717),
    [sym__array_index_incomplete_type] = STATE(5908),
    [sym_index_constraint] = STATE(6478),
    [sym_index_subtype_definition] = STATE(6478),
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(1817),
    [sym__direct_name] = STATE(730),
    [sym__identifier] = STATE(903),
    [sym__external_name] = STATE(903),
    [sym_external_constant_name] = STATE(915),
    [sym_external_signal_name] = STATE(915),
    [sym_external_variable_name] = STATE(915),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6786),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5182),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(9),
    [sym_line_comment] = STATE(9),
    [sym_block_comment] = STATE(9),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(269),
    [sym_identifier] = ACTIONS(271),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_TYPE] = ACTIONS(281),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(287),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(287),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(271),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(287),
    [sym_library_function] = ACTIONS(271),
    [sym_library_type] = ACTIONS(271),
  },
  [STATE(10)] = {
    [sym_unspecified_type_indication] = STATE(6478),
    [sym_array_index_incomplete_type_list] = STATE(8052),
    [sym__array_index_incomplete_type] = STATE(5908),
    [sym_index_constraint] = STATE(6478),
    [sym_index_subtype_definition] = STATE(6478),
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(1817),
    [sym__direct_name] = STATE(730),
    [sym__identifier] = STATE(903),
    [sym__external_name] = STATE(903),
    [sym_external_constant_name] = STATE(915),
    [sym_external_signal_name] = STATE(915),
    [sym_external_variable_name] = STATE(915),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6786),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5182),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(10),
    [sym_line_comment] = STATE(10),
    [sym_block_comment] = STATE(10),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(269),
    [sym_identifier] = ACTIONS(271),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_TYPE] = ACTIONS(281),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(287),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(287),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(271),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(287),
    [sym_library_function] = ACTIONS(271),
    [sym_library_type] = ACTIONS(271),
  },
  [STATE(11)] = {
    [sym__sequential_statement] = STATE(2824),
    [sym_sequential_block_statement] = STATE(2825),
    [sym_assertion_statement] = STATE(2825),
    [sym_assertion] = STATE(7497),
    [sym_report_expression] = STATE(5850),
    [sym_case_statement] = STATE(2825),
    [sym_case_expression] = STATE(6384),
    [sym_exit_statement] = STATE(2825),
    [sym_if_statement_block] = STATE(2825),
    [sym_if_statement] = STATE(6560),
    [sym_loop_statement] = STATE(2825),
    [sym__iteration_scheme] = STATE(6385),
    [sym_loop_body] = STATE(6561),
    [sym_next_statement] = STATE(2825),
    [sym_null_statement] = STATE(2825),
    [sym_procedure_call_statement] = STATE(2825),
    [sym_report_statement] = STATE(2825),
    [sym_return_statement] = STATE(2825),
    [sym_simple_waveform_assignment] = STATE(2825),
    [sym_simple_variable_assignment] = STATE(2825),
    [sym_simple_force_assignment] = STATE(2825),
    [sym_simple_release_assignment] = STATE(2825),
    [sym_conditional_signal_assignment] = STATE(2825),
    [sym_selected_waveform_assignment] = STATE(2825),
    [sym_selected_force_assignment] = STATE(2825),
    [sym_selected_variable_assignment] = STATE(2825),
    [sym_with_expression] = STATE(6474),
    [sym_wait_statement] = STATE(2825),
    [sym_name] = STATE(5851),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(11),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5885),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1529),
    [sym_line_comment] = STATE(11),
    [sym_block_comment] = STATE(11),
    [aux_sym_sequential_block_repeat1] = STATE(12),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(231),
    [sym_CASE] = ACTIONS(233),
    [sym_ELSE] = ACTIONS(293),
    [sym_ELSIF] = ACTIONS(293),
    [sym_END] = ACTIONS(293),
    [sym_EXIT] = ACTIONS(241),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(249),
    [sym_NULL] = ACTIONS(251),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(255),
    [sym_WAIT] = ACTIONS(257),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(12)] = {
    [sym__sequential_statement] = STATE(2824),
    [sym_sequential_block_statement] = STATE(2825),
    [sym_assertion_statement] = STATE(2825),
    [sym_assertion] = STATE(7497),
    [sym_report_expression] = STATE(5850),
    [sym_case_statement] = STATE(2825),
    [sym_case_expression] = STATE(6384),
    [sym_exit_statement] = STATE(2825),
    [sym_if_statement_block] = STATE(2825),
    [sym_if_statement] = STATE(6560),
    [sym_loop_statement] = STATE(2825),
    [sym__iteration_scheme] = STATE(6385),
    [sym_loop_body] = STATE(6561),
    [sym_next_statement] = STATE(2825),
    [sym_null_statement] = STATE(2825),
    [sym_procedure_call_statement] = STATE(2825),
    [sym_report_statement] = STATE(2825),
    [sym_return_statement] = STATE(2825),
    [sym_simple_waveform_assignment] = STATE(2825),
    [sym_simple_variable_assignment] = STATE(2825),
    [sym_simple_force_assignment] = STATE(2825),
    [sym_simple_release_assignment] = STATE(2825),
    [sym_conditional_signal_assignment] = STATE(2825),
    [sym_selected_waveform_assignment] = STATE(2825),
    [sym_selected_force_assignment] = STATE(2825),
    [sym_selected_variable_assignment] = STATE(2825),
    [sym_with_expression] = STATE(6474),
    [sym_wait_statement] = STATE(2825),
    [sym_name] = STATE(5851),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(12),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5885),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1529),
    [sym_line_comment] = STATE(12),
    [sym_block_comment] = STATE(12),
    [aux_sym_sequential_block_repeat1] = STATE(12),
    [anon_sym_LPAREN] = ACTIONS(295),
    [anon_sym_LT_LT] = ACTIONS(298),
    [sym_identifier] = ACTIONS(301),
    [sym_ASSERT] = ACTIONS(304),
    [sym_BLOCK] = ACTIONS(307),
    [sym_CASE] = ACTIONS(310),
    [sym_ELSE] = ACTIONS(313),
    [sym_ELSIF] = ACTIONS(313),
    [sym_END] = ACTIONS(313),
    [sym_EXIT] = ACTIONS(315),
    [sym_FOR] = ACTIONS(318),
    [sym_IF] = ACTIONS(321),
    [sym_LOOP] = ACTIONS(324),
    [sym_NEXT] = ACTIONS(327),
    [sym_NULL] = ACTIONS(330),
    [sym_REPORT] = ACTIONS(333),
    [sym_RETURN] = ACTIONS(336),
    [sym_WAIT] = ACTIONS(339),
    [sym_WHILE] = ACTIONS(342),
    [sym_WITH] = ACTIONS(345),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(348),
    [sym_operator_symbol] = ACTIONS(348),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(301),
    [sym_library_constant_std_logic] = ACTIONS(348),
    [sym_library_function] = ACTIONS(301),
    [sym_library_type] = ACTIONS(301),
  },
  [STATE(13)] = {
    [sym__sequential_statement] = STATE(2897),
    [sym_sequential_block_statement] = STATE(2877),
    [sym_assertion_statement] = STATE(2877),
    [sym_assertion] = STATE(8092),
    [sym_report_expression] = STATE(5913),
    [sym_case_statement] = STATE(2877),
    [sym_case_expression] = STATE(6416),
    [sym_exit_statement] = STATE(2877),
    [sym_if_statement_block] = STATE(2877),
    [sym_if_statement] = STATE(6029),
    [sym_loop_statement] = STATE(2877),
    [sym__iteration_scheme] = STATE(6417),
    [sym_loop_body] = STATE(6030),
    [sym_next_statement] = STATE(2877),
    [sym_null_statement] = STATE(2877),
    [sym_procedure_call_statement] = STATE(2877),
    [sym_report_statement] = STATE(2877),
    [sym_return_statement] = STATE(2877),
    [sym_simple_waveform_assignment] = STATE(2877),
    [sym_simple_variable_assignment] = STATE(2877),
    [sym_simple_force_assignment] = STATE(2877),
    [sym_simple_release_assignment] = STATE(2877),
    [sym_conditional_signal_assignment] = STATE(2877),
    [sym_selected_waveform_assignment] = STATE(2877),
    [sym_selected_force_assignment] = STATE(2877),
    [sym_selected_variable_assignment] = STATE(2877),
    [sym_with_expression] = STATE(6485),
    [sym_wait_statement] = STATE(2877),
    [sym_name] = STATE(5914),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(13),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5893),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1547),
    [sym_line_comment] = STATE(13),
    [sym_block_comment] = STATE(13),
    [aux_sym_sequential_block_repeat1] = STATE(16),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(351),
    [sym_CASE] = ACTIONS(233),
    [sym_END] = ACTIONS(353),
    [sym_EXIT] = ACTIONS(355),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(357),
    [sym_NULL] = ACTIONS(359),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(361),
    [sym_WAIT] = ACTIONS(363),
    [sym_WHEN] = ACTIONS(353),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(14)] = {
    [sym__sequential_statement] = STATE(2985),
    [sym_sequential_block_statement] = STATE(2933),
    [sym_assertion_statement] = STATE(2933),
    [sym_assertion] = STATE(7373),
    [sym_report_expression] = STATE(5829),
    [sym_case_statement] = STATE(2933),
    [sym_case_expression] = STATE(6294),
    [sym_exit_statement] = STATE(2933),
    [sym_if_statement_block] = STATE(2933),
    [sym_if_statement] = STATE(6296),
    [sym_if_statement_body] = STATE(8051),
    [sym_loop_statement] = STATE(2933),
    [sym__iteration_scheme] = STATE(6297),
    [sym_loop_body] = STATE(6298),
    [sym_next_statement] = STATE(2933),
    [sym_null_statement] = STATE(2933),
    [sym_procedure_call_statement] = STATE(2933),
    [sym_report_statement] = STATE(2933),
    [sym_return_statement] = STATE(2933),
    [sym_simple_waveform_assignment] = STATE(2933),
    [sym_simple_variable_assignment] = STATE(2933),
    [sym_simple_force_assignment] = STATE(2933),
    [sym_simple_release_assignment] = STATE(2933),
    [sym_conditional_signal_assignment] = STATE(2933),
    [sym_selected_waveform_assignment] = STATE(2933),
    [sym_selected_force_assignment] = STATE(2933),
    [sym_selected_variable_assignment] = STATE(2933),
    [sym_with_expression] = STATE(6299),
    [sym_wait_statement] = STATE(2933),
    [sym_name] = STATE(5833),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(14),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5835),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1500),
    [sym_line_comment] = STATE(14),
    [sym_block_comment] = STATE(14),
    [aux_sym_sequential_block_repeat1] = STATE(28),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(365),
    [sym_CASE] = ACTIONS(233),
    [sym_END] = ACTIONS(367),
    [sym_EXIT] = ACTIONS(369),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(371),
    [sym_NULL] = ACTIONS(373),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(375),
    [sym_WAIT] = ACTIONS(377),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(15)] = {
    [sym__sequential_statement] = STATE(2897),
    [sym_sequential_block_statement] = STATE(2877),
    [sym_assertion_statement] = STATE(2877),
    [sym_assertion] = STATE(8092),
    [sym_report_expression] = STATE(5913),
    [sym_case_statement] = STATE(2877),
    [sym_case_expression] = STATE(6416),
    [sym_exit_statement] = STATE(2877),
    [sym_if_statement_block] = STATE(2877),
    [sym_if_statement] = STATE(6029),
    [sym_loop_statement] = STATE(2877),
    [sym__iteration_scheme] = STATE(6417),
    [sym_loop_body] = STATE(6030),
    [sym_next_statement] = STATE(2877),
    [sym_null_statement] = STATE(2877),
    [sym_procedure_call_statement] = STATE(2877),
    [sym_report_statement] = STATE(2877),
    [sym_return_statement] = STATE(2877),
    [sym_simple_waveform_assignment] = STATE(2877),
    [sym_simple_variable_assignment] = STATE(2877),
    [sym_simple_force_assignment] = STATE(2877),
    [sym_simple_release_assignment] = STATE(2877),
    [sym_conditional_signal_assignment] = STATE(2877),
    [sym_selected_waveform_assignment] = STATE(2877),
    [sym_selected_force_assignment] = STATE(2877),
    [sym_selected_variable_assignment] = STATE(2877),
    [sym_with_expression] = STATE(6485),
    [sym_wait_statement] = STATE(2877),
    [sym_name] = STATE(5914),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(15),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5893),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1547),
    [sym_line_comment] = STATE(15),
    [sym_block_comment] = STATE(15),
    [aux_sym_sequential_block_repeat1] = STATE(13),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(351),
    [sym_CASE] = ACTIONS(233),
    [sym_END] = ACTIONS(379),
    [sym_EXIT] = ACTIONS(355),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(357),
    [sym_NULL] = ACTIONS(359),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(361),
    [sym_WAIT] = ACTIONS(363),
    [sym_WHEN] = ACTIONS(379),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(16)] = {
    [sym__sequential_statement] = STATE(2897),
    [sym_sequential_block_statement] = STATE(2877),
    [sym_assertion_statement] = STATE(2877),
    [sym_assertion] = STATE(8092),
    [sym_report_expression] = STATE(5913),
    [sym_case_statement] = STATE(2877),
    [sym_case_expression] = STATE(6416),
    [sym_exit_statement] = STATE(2877),
    [sym_if_statement_block] = STATE(2877),
    [sym_if_statement] = STATE(6029),
    [sym_loop_statement] = STATE(2877),
    [sym__iteration_scheme] = STATE(6417),
    [sym_loop_body] = STATE(6030),
    [sym_next_statement] = STATE(2877),
    [sym_null_statement] = STATE(2877),
    [sym_procedure_call_statement] = STATE(2877),
    [sym_report_statement] = STATE(2877),
    [sym_return_statement] = STATE(2877),
    [sym_simple_waveform_assignment] = STATE(2877),
    [sym_simple_variable_assignment] = STATE(2877),
    [sym_simple_force_assignment] = STATE(2877),
    [sym_simple_release_assignment] = STATE(2877),
    [sym_conditional_signal_assignment] = STATE(2877),
    [sym_selected_waveform_assignment] = STATE(2877),
    [sym_selected_force_assignment] = STATE(2877),
    [sym_selected_variable_assignment] = STATE(2877),
    [sym_with_expression] = STATE(6485),
    [sym_wait_statement] = STATE(2877),
    [sym_name] = STATE(5914),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(16),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5893),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1547),
    [sym_line_comment] = STATE(16),
    [sym_block_comment] = STATE(16),
    [aux_sym_sequential_block_repeat1] = STATE(16),
    [anon_sym_LPAREN] = ACTIONS(295),
    [anon_sym_LT_LT] = ACTIONS(298),
    [sym_identifier] = ACTIONS(301),
    [sym_ASSERT] = ACTIONS(304),
    [sym_BLOCK] = ACTIONS(381),
    [sym_CASE] = ACTIONS(310),
    [sym_END] = ACTIONS(313),
    [sym_EXIT] = ACTIONS(384),
    [sym_FOR] = ACTIONS(318),
    [sym_IF] = ACTIONS(321),
    [sym_LOOP] = ACTIONS(324),
    [sym_NEXT] = ACTIONS(387),
    [sym_NULL] = ACTIONS(390),
    [sym_REPORT] = ACTIONS(333),
    [sym_RETURN] = ACTIONS(393),
    [sym_WAIT] = ACTIONS(396),
    [sym_WHEN] = ACTIONS(313),
    [sym_WHILE] = ACTIONS(342),
    [sym_WITH] = ACTIONS(345),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(348),
    [sym_operator_symbol] = ACTIONS(348),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(301),
    [sym_library_constant_std_logic] = ACTIONS(348),
    [sym_library_function] = ACTIONS(301),
    [sym_library_type] = ACTIONS(301),
  },
  [STATE(17)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7055),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(17),
    [sym_line_comment] = STATE(17),
    [sym_block_comment] = STATE(17),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(399),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(18)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(18),
    [sym_selected_waveforms] = STATE(7475),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(317),
    [sym_line_comment] = STATE(18),
    [sym_block_comment] = STATE(18),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(419),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(19)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(19),
    [sym_waveform] = STATE(5856),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7529),
    [sym_delay_mechanism] = STATE(346),
    [sym_line_comment] = STATE(19),
    [sym_block_comment] = STATE(19),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(441),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_RELEASE] = ACTIONS(443),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(20)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(20),
    [sym_selected_waveforms] = STATE(7404),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(297),
    [sym_line_comment] = STATE(20),
    [sym_block_comment] = STATE(20),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(445),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(21)] = {
    [sym_use_clause] = STATE(2040),
    [sym_package_declaration] = STATE(2040),
    [sym_package_instantiation_declaration] = STATE(2040),
    [sym_package_definition] = STATE(2040),
    [sym__block_declarative_item] = STATE(2029),
    [sym_subprogram_declaration] = STATE(2040),
    [sym_subprogram_definition] = STATE(2040),
    [sym_subprogram_instantiation_declaration] = STATE(2040),
    [sym_type_declaration] = STATE(2040),
    [sym_subtype_declaration] = STATE(2040),
    [sym_mode_view_declaration] = STATE(2040),
    [sym_constant_declaration] = STATE(2040),
    [sym_signal_declaration] = STATE(2040),
    [sym_variable_declaration] = STATE(2040),
    [sym_file_declaration] = STATE(2040),
    [sym_alias_declaration] = STATE(2040),
    [sym_component_declaration] = STATE(2040),
    [sym_attribute_declaration] = STATE(2040),
    [sym_attribute_specification] = STATE(2040),
    [sym_disconnection_specification] = STATE(2040),
    [sym_group_template_declaration] = STATE(2040),
    [sym_group_declaration] = STATE(2040),
    [sym__tool_directive] = STATE(21),
    [sym_procedure_specification] = STATE(6069),
    [sym_function_specification] = STATE(6069),
    [sym__subprogram_specification] = STATE(5708),
    [sym_configuration_specification] = STATE(2040),
    [sym_line_comment] = STATE(21),
    [sym_block_comment] = STATE(21),
    [aux_sym_design_unit_repeat2] = STATE(21),
    [anon_sym_LPAREN] = ACTIONS(447),
    [anon_sym_LT_LT] = ACTIONS(447),
    [sym_identifier] = ACTIONS(447),
    [sym_ALIAS] = ACTIONS(449),
    [sym_ARCHITECTURE] = ACTIONS(447),
    [sym_ASSERT] = ACTIONS(447),
    [sym_ATTRIBUTE] = ACTIONS(452),
    [sym_COMPONENT] = ACTIONS(455),
    [sym_CONFIGURATION] = ACTIONS(447),
    [sym_CONSTANT] = ACTIONS(458),
    [sym_CONTEXT] = ACTIONS(447),
    [sym_DISCONNECT] = ACTIONS(461),
    [sym_ENTITY] = ACTIONS(447),
    [sym_FILE] = ACTIONS(464),
    [sym_FOR] = ACTIONS(467),
    [sym_FUNCTION] = ACTIONS(470),
    [sym_GROUP] = ACTIONS(473),
    [sym_IMPURE] = ACTIONS(476),
    [sym_LIBRARY] = ACTIONS(447),
    [sym_PACKAGE] = ACTIONS(479),
    [sym_POSTPONED] = ACTIONS(447),
    [sym_PROCEDURE] = ACTIONS(482),
    [sym_PROCESS] = ACTIONS(447),
    [sym_PURE] = ACTIONS(476),
    [sym_SIGNAL] = ACTIONS(485),
    [sym_SHARED] = ACTIONS(488),
    [sym_SUBTYPE] = ACTIONS(491),
    [sym_TYPE] = ACTIONS(494),
    [sym_USE] = ACTIONS(497),
    [sym_VARIABLE] = ACTIONS(500),
    [sym_VIEW] = ACTIONS(503),
    [sym_WITH] = ACTIONS(447),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(447),
    [sym_operator_symbol] = ACTIONS(447),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(447),
    [sym_library_constant_std_logic] = ACTIONS(447),
    [sym_library_function] = ACTIONS(447),
    [sym_library_type] = ACTIONS(447),
    [sym__end_of_file] = ACTIONS(447),
  },
  [STATE(22)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(22),
    [sym_selected_waveforms] = STATE(8337),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(243),
    [sym_line_comment] = STATE(22),
    [sym_block_comment] = STATE(22),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(506),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(23)] = {
    [sym__sequential_statement] = STATE(2985),
    [sym_sequential_block_statement] = STATE(2933),
    [sym_assertion_statement] = STATE(2933),
    [sym_assertion] = STATE(7373),
    [sym_report_expression] = STATE(5829),
    [sym_case_statement] = STATE(2933),
    [sym_case_expression] = STATE(6294),
    [sym_exit_statement] = STATE(2933),
    [sym_if_statement_block] = STATE(2933),
    [sym_if_statement] = STATE(6296),
    [sym_loop_statement] = STATE(2933),
    [sym__iteration_scheme] = STATE(6297),
    [sym_loop_body] = STATE(6298),
    [sym_next_statement] = STATE(2933),
    [sym_null_statement] = STATE(2933),
    [sym_procedure_call_statement] = STATE(2933),
    [sym_report_statement] = STATE(2933),
    [sym_return_statement] = STATE(2933),
    [sym_simple_waveform_assignment] = STATE(2933),
    [sym_simple_variable_assignment] = STATE(2933),
    [sym_simple_force_assignment] = STATE(2933),
    [sym_simple_release_assignment] = STATE(2933),
    [sym_conditional_signal_assignment] = STATE(2933),
    [sym_selected_waveform_assignment] = STATE(2933),
    [sym_selected_force_assignment] = STATE(2933),
    [sym_selected_variable_assignment] = STATE(2933),
    [sym_with_expression] = STATE(6299),
    [sym_wait_statement] = STATE(2933),
    [sym_name] = STATE(5833),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(23),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5835),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1500),
    [sym_line_comment] = STATE(23),
    [sym_block_comment] = STATE(23),
    [aux_sym_sequential_block_repeat1] = STATE(29),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(365),
    [sym_CASE] = ACTIONS(233),
    [sym_END] = ACTIONS(508),
    [sym_EXIT] = ACTIONS(369),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(371),
    [sym_NULL] = ACTIONS(373),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(375),
    [sym_WAIT] = ACTIONS(377),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(24)] = {
    [sym__sequential_statement] = STATE(2985),
    [sym_sequential_block_statement] = STATE(2933),
    [sym_assertion_statement] = STATE(2933),
    [sym_assertion] = STATE(7373),
    [sym_report_expression] = STATE(5829),
    [sym_case_statement] = STATE(2933),
    [sym_case_expression] = STATE(6294),
    [sym_exit_statement] = STATE(2933),
    [sym_if_statement_block] = STATE(2933),
    [sym_if_statement] = STATE(6296),
    [sym_loop_statement] = STATE(2933),
    [sym__iteration_scheme] = STATE(6297),
    [sym_loop_body] = STATE(6298),
    [sym_next_statement] = STATE(2933),
    [sym_null_statement] = STATE(2933),
    [sym_procedure_call_statement] = STATE(2933),
    [sym_report_statement] = STATE(2933),
    [sym_return_statement] = STATE(2933),
    [sym_simple_waveform_assignment] = STATE(2933),
    [sym_simple_variable_assignment] = STATE(2933),
    [sym_simple_force_assignment] = STATE(2933),
    [sym_simple_release_assignment] = STATE(2933),
    [sym_conditional_signal_assignment] = STATE(2933),
    [sym_selected_waveform_assignment] = STATE(2933),
    [sym_selected_force_assignment] = STATE(2933),
    [sym_selected_variable_assignment] = STATE(2933),
    [sym_with_expression] = STATE(6299),
    [sym_wait_statement] = STATE(2933),
    [sym_name] = STATE(5833),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(24),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5835),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1500),
    [sym_line_comment] = STATE(24),
    [sym_block_comment] = STATE(24),
    [aux_sym_sequential_block_repeat1] = STATE(61),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(365),
    [sym_CASE] = ACTIONS(233),
    [sym_END] = ACTIONS(510),
    [sym_EXIT] = ACTIONS(369),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(371),
    [sym_NULL] = ACTIONS(373),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(375),
    [sym_WAIT] = ACTIONS(377),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(25)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(25),
    [sym_selected_waveforms] = STATE(7008),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(300),
    [sym_line_comment] = STATE(25),
    [sym_block_comment] = STATE(25),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(512),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(26)] = {
    [sym__sequential_statement] = STATE(2985),
    [sym_sequential_block_statement] = STATE(2933),
    [sym_assertion_statement] = STATE(2933),
    [sym_assertion] = STATE(7373),
    [sym_report_expression] = STATE(5829),
    [sym_case_statement] = STATE(2933),
    [sym_case_expression] = STATE(6294),
    [sym_exit_statement] = STATE(2933),
    [sym_if_statement_block] = STATE(2933),
    [sym_if_statement] = STATE(6296),
    [sym_loop_statement] = STATE(2933),
    [sym__iteration_scheme] = STATE(6297),
    [sym_loop_body] = STATE(6298),
    [sym_next_statement] = STATE(2933),
    [sym_null_statement] = STATE(2933),
    [sym_procedure_call_statement] = STATE(2933),
    [sym_report_statement] = STATE(2933),
    [sym_return_statement] = STATE(2933),
    [sym_simple_waveform_assignment] = STATE(2933),
    [sym_simple_variable_assignment] = STATE(2933),
    [sym_simple_force_assignment] = STATE(2933),
    [sym_simple_release_assignment] = STATE(2933),
    [sym_conditional_signal_assignment] = STATE(2933),
    [sym_selected_waveform_assignment] = STATE(2933),
    [sym_selected_force_assignment] = STATE(2933),
    [sym_selected_variable_assignment] = STATE(2933),
    [sym_with_expression] = STATE(6299),
    [sym_wait_statement] = STATE(2933),
    [sym_name] = STATE(5833),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(26),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5835),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1500),
    [sym_line_comment] = STATE(26),
    [sym_block_comment] = STATE(26),
    [aux_sym_sequential_block_repeat1] = STATE(24),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(365),
    [sym_CASE] = ACTIONS(233),
    [sym_END] = ACTIONS(514),
    [sym_EXIT] = ACTIONS(369),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(371),
    [sym_NULL] = ACTIONS(373),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(375),
    [sym_WAIT] = ACTIONS(377),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(27)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(27),
    [sym_selected_waveforms] = STATE(6893),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(248),
    [sym_line_comment] = STATE(27),
    [sym_block_comment] = STATE(27),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(516),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(28)] = {
    [sym__sequential_statement] = STATE(2985),
    [sym_sequential_block_statement] = STATE(2933),
    [sym_assertion_statement] = STATE(2933),
    [sym_assertion] = STATE(7373),
    [sym_report_expression] = STATE(5829),
    [sym_case_statement] = STATE(2933),
    [sym_case_expression] = STATE(6294),
    [sym_exit_statement] = STATE(2933),
    [sym_if_statement_block] = STATE(2933),
    [sym_if_statement] = STATE(6296),
    [sym_loop_statement] = STATE(2933),
    [sym__iteration_scheme] = STATE(6297),
    [sym_loop_body] = STATE(6298),
    [sym_next_statement] = STATE(2933),
    [sym_null_statement] = STATE(2933),
    [sym_procedure_call_statement] = STATE(2933),
    [sym_report_statement] = STATE(2933),
    [sym_return_statement] = STATE(2933),
    [sym_simple_waveform_assignment] = STATE(2933),
    [sym_simple_variable_assignment] = STATE(2933),
    [sym_simple_force_assignment] = STATE(2933),
    [sym_simple_release_assignment] = STATE(2933),
    [sym_conditional_signal_assignment] = STATE(2933),
    [sym_selected_waveform_assignment] = STATE(2933),
    [sym_selected_force_assignment] = STATE(2933),
    [sym_selected_variable_assignment] = STATE(2933),
    [sym_with_expression] = STATE(6299),
    [sym_wait_statement] = STATE(2933),
    [sym_name] = STATE(5833),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(28),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5835),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1500),
    [sym_line_comment] = STATE(28),
    [sym_block_comment] = STATE(28),
    [aux_sym_sequential_block_repeat1] = STATE(61),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(365),
    [sym_CASE] = ACTIONS(233),
    [sym_END] = ACTIONS(293),
    [sym_EXIT] = ACTIONS(369),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(371),
    [sym_NULL] = ACTIONS(373),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(375),
    [sym_WAIT] = ACTIONS(377),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(29)] = {
    [sym__sequential_statement] = STATE(2985),
    [sym_sequential_block_statement] = STATE(2933),
    [sym_assertion_statement] = STATE(2933),
    [sym_assertion] = STATE(7373),
    [sym_report_expression] = STATE(5829),
    [sym_case_statement] = STATE(2933),
    [sym_case_expression] = STATE(6294),
    [sym_exit_statement] = STATE(2933),
    [sym_if_statement_block] = STATE(2933),
    [sym_if_statement] = STATE(6296),
    [sym_loop_statement] = STATE(2933),
    [sym__iteration_scheme] = STATE(6297),
    [sym_loop_body] = STATE(6298),
    [sym_next_statement] = STATE(2933),
    [sym_null_statement] = STATE(2933),
    [sym_procedure_call_statement] = STATE(2933),
    [sym_report_statement] = STATE(2933),
    [sym_return_statement] = STATE(2933),
    [sym_simple_waveform_assignment] = STATE(2933),
    [sym_simple_variable_assignment] = STATE(2933),
    [sym_simple_force_assignment] = STATE(2933),
    [sym_simple_release_assignment] = STATE(2933),
    [sym_conditional_signal_assignment] = STATE(2933),
    [sym_selected_waveform_assignment] = STATE(2933),
    [sym_selected_force_assignment] = STATE(2933),
    [sym_selected_variable_assignment] = STATE(2933),
    [sym_with_expression] = STATE(6299),
    [sym_wait_statement] = STATE(2933),
    [sym_name] = STATE(5833),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(29),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5835),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1500),
    [sym_line_comment] = STATE(29),
    [sym_block_comment] = STATE(29),
    [aux_sym_sequential_block_repeat1] = STATE(61),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ASSERT] = ACTIONS(19),
    [sym_BLOCK] = ACTIONS(365),
    [sym_CASE] = ACTIONS(233),
    [sym_END] = ACTIONS(518),
    [sym_EXIT] = ACTIONS(369),
    [sym_FOR] = ACTIONS(243),
    [sym_IF] = ACTIONS(245),
    [sym_LOOP] = ACTIONS(247),
    [sym_NEXT] = ACTIONS(371),
    [sym_NULL] = ACTIONS(373),
    [sym_REPORT] = ACTIONS(253),
    [sym_RETURN] = ACTIONS(375),
    [sym_WAIT] = ACTIONS(377),
    [sym_WHILE] = ACTIONS(259),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
  },
  [STATE(30)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7508),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(30),
    [sym_line_comment] = STATE(30),
    [sym_block_comment] = STATE(30),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(520),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(31)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8101),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(31),
    [sym_line_comment] = STATE(31),
    [sym_block_comment] = STATE(31),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(522),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(32)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6880),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(32),
    [sym_line_comment] = STATE(32),
    [sym_block_comment] = STATE(32),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(524),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(33)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6819),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(33),
    [sym_line_comment] = STATE(33),
    [sym_block_comment] = STATE(33),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(526),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(34)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7468),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(34),
    [sym_line_comment] = STATE(34),
    [sym_block_comment] = STATE(34),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(528),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(35)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8343),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(35),
    [sym_line_comment] = STATE(35),
    [sym_block_comment] = STATE(35),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(530),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(36)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6840),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(36),
    [sym_line_comment] = STATE(36),
    [sym_block_comment] = STATE(36),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(532),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(37)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6938),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(37),
    [sym_line_comment] = STATE(37),
    [sym_block_comment] = STATE(37),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(534),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(38)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7046),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(38),
    [sym_line_comment] = STATE(38),
    [sym_block_comment] = STATE(38),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(536),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(39)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7204),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(39),
    [sym_line_comment] = STATE(39),
    [sym_block_comment] = STATE(39),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(538),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(40)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7325),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(40),
    [sym_line_comment] = STATE(40),
    [sym_block_comment] = STATE(40),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(540),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(41)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7451),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(41),
    [sym_line_comment] = STATE(41),
    [sym_block_comment] = STATE(41),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(542),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(42)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7657),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(42),
    [sym_line_comment] = STATE(42),
    [sym_block_comment] = STATE(42),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(544),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(43)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7784),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(43),
    [sym_line_comment] = STATE(43),
    [sym_block_comment] = STATE(43),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(546),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(44)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8082),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(44),
    [sym_line_comment] = STATE(44),
    [sym_block_comment] = STATE(44),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(548),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(45)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8280),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(45),
    [sym_line_comment] = STATE(45),
    [sym_block_comment] = STATE(45),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(550),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(46)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6614),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(46),
    [sym_line_comment] = STATE(46),
    [sym_block_comment] = STATE(46),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(552),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(47)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6668),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(47),
    [sym_line_comment] = STATE(47),
    [sym_block_comment] = STATE(47),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(554),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(48)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6742),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(48),
    [sym_line_comment] = STATE(48),
    [sym_block_comment] = STATE(48),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(556),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(49)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6790),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(49),
    [sym_line_comment] = STATE(49),
    [sym_block_comment] = STATE(49),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(558),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(50)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6842),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(50),
    [sym_line_comment] = STATE(50),
    [sym_block_comment] = STATE(50),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(560),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(51)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6877),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(51),
    [sym_line_comment] = STATE(51),
    [sym_block_comment] = STATE(51),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(562),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(52)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6924),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(52),
    [sym_line_comment] = STATE(52),
    [sym_block_comment] = STATE(52),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(564),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(53)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6964),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(53),
    [sym_line_comment] = STATE(53),
    [sym_block_comment] = STATE(53),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(566),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(54)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7006),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(54),
    [sym_line_comment] = STATE(54),
    [sym_block_comment] = STATE(54),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(568),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(55)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7761),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(55),
    [sym_line_comment] = STATE(55),
    [sym_block_comment] = STATE(55),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(570),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(56)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7123),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(56),
    [sym_line_comment] = STATE(56),
    [sym_block_comment] = STATE(56),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(572),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(57)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7192),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(57),
    [sym_line_comment] = STATE(57),
    [sym_block_comment] = STATE(57),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(574),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(58)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7243),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(58),
    [sym_line_comment] = STATE(58),
    [sym_block_comment] = STATE(58),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(576),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(59)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7288),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(59),
    [sym_line_comment] = STATE(59),
    [sym_block_comment] = STATE(59),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(578),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(60)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7333),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(60),
    [sym_line_comment] = STATE(60),
    [sym_block_comment] = STATE(60),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(580),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(61)] = {
    [sym__sequential_statement] = STATE(2985),
    [sym_sequential_block_statement] = STATE(2933),
    [sym_assertion_statement] = STATE(2933),
    [sym_assertion] = STATE(7373),
    [sym_report_expression] = STATE(5829),
    [sym_case_statement] = STATE(2933),
    [sym_case_expression] = STATE(6294),
    [sym_exit_statement] = STATE(2933),
    [sym_if_statement_block] = STATE(2933),
    [sym_if_statement] = STATE(6296),
    [sym_loop_statement] = STATE(2933),
    [sym__iteration_scheme] = STATE(6297),
    [sym_loop_body] = STATE(6298),
    [sym_next_statement] = STATE(2933),
    [sym_null_statement] = STATE(2933),
    [sym_procedure_call_statement] = STATE(2933),
    [sym_report_statement] = STATE(2933),
    [sym_return_statement] = STATE(2933),
    [sym_simple_waveform_assignment] = STATE(2933),
    [sym_simple_variable_assignment] = STATE(2933),
    [sym_simple_force_assignment] = STATE(2933),
    [sym_simple_release_assignment] = STATE(2933),
    [sym_conditional_signal_assignment] = STATE(2933),
    [sym_selected_waveform_assignment] = STATE(2933),
    [sym_selected_force_assignment] = STATE(2933),
    [sym_selected_variable_assignment] = STATE(2933),
    [sym_with_expression] = STATE(6299),
    [sym_wait_statement] = STATE(2933),
    [sym_name] = STATE(5833),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7450),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(61),
    [sym_while_loop] = STATE(7407),
    [sym_for_loop] = STATE(7407),
    [sym__target] = STATE(5835),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(1500),
    [sym_line_comment] = STATE(61),
    [sym_block_comment] = STATE(61),
    [aux_sym_sequential_block_repeat1] = STATE(61),
    [anon_sym_LPAREN] = ACTIONS(295),
    [anon_sym_LT_LT] = ACTIONS(298),
    [sym_identifier] = ACTIONS(301),
    [sym_ASSERT] = ACTIONS(304),
    [sym_BLOCK] = ACTIONS(582),
    [sym_CASE] = ACTIONS(310),
    [sym_END] = ACTIONS(313),
    [sym_EXIT] = ACTIONS(585),
    [sym_FOR] = ACTIONS(318),
    [sym_IF] = ACTIONS(321),
    [sym_LOOP] = ACTIONS(324),
    [sym_NEXT] = ACTIONS(588),
    [sym_NULL] = ACTIONS(591),
    [sym_REPORT] = ACTIONS(333),
    [sym_RETURN] = ACTIONS(594),
    [sym_WAIT] = ACTIONS(597),
    [sym_WHILE] = ACTIONS(342),
    [sym_WITH] = ACTIONS(345),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(348),
    [sym_operator_symbol] = ACTIONS(348),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(301),
    [sym_library_constant_std_logic] = ACTIONS(348),
    [sym_library_function] = ACTIONS(301),
    [sym_library_type] = ACTIONS(301),
  },
  [STATE(62)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7416),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(62),
    [sym_line_comment] = STATE(62),
    [sym_block_comment] = STATE(62),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(600),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(63)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7482),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(63),
    [sym_line_comment] = STATE(63),
    [sym_block_comment] = STATE(63),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(602),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(64)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7553),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(64),
    [sym_line_comment] = STATE(64),
    [sym_block_comment] = STATE(64),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(604),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(65)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7633),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(65),
    [sym_line_comment] = STATE(65),
    [sym_block_comment] = STATE(65),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(606),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(66)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7682),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(66),
    [sym_line_comment] = STATE(66),
    [sym_block_comment] = STATE(66),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(608),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(67)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7715),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(67),
    [sym_line_comment] = STATE(67),
    [sym_block_comment] = STATE(67),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(610),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(68)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7757),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(68),
    [sym_line_comment] = STATE(68),
    [sym_block_comment] = STATE(68),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(612),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(69)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7788),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(69),
    [sym_line_comment] = STATE(69),
    [sym_block_comment] = STATE(69),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(614),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(70)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7823),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(70),
    [sym_line_comment] = STATE(70),
    [sym_block_comment] = STATE(70),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(616),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(71)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8044),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(71),
    [sym_line_comment] = STATE(71),
    [sym_block_comment] = STATE(71),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(618),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(72)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8091),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(72),
    [sym_line_comment] = STATE(72),
    [sym_block_comment] = STATE(72),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(620),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(73)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8144),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(73),
    [sym_line_comment] = STATE(73),
    [sym_block_comment] = STATE(73),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(622),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(74)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8207),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(74),
    [sym_line_comment] = STATE(74),
    [sym_block_comment] = STATE(74),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(624),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(75)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8294),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(75),
    [sym_line_comment] = STATE(75),
    [sym_block_comment] = STATE(75),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(626),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(76)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8320),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(76),
    [sym_line_comment] = STATE(76),
    [sym_block_comment] = STATE(76),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(628),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(77)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(8339),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(77),
    [sym_line_comment] = STATE(77),
    [sym_block_comment] = STATE(77),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(630),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(78)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6611),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(78),
    [sym_line_comment] = STATE(78),
    [sym_block_comment] = STATE(78),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(632),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(79)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(6630),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(79),
    [sym_line_comment] = STATE(79),
    [sym_block_comment] = STATE(79),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(634),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(80)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(80),
    [sym_selected_waveforms] = STATE(7592),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(269),
    [sym_line_comment] = STATE(80),
    [sym_block_comment] = STATE(80),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(636),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(81)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(81),
    [sym_selected_waveforms] = STATE(7625),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(276),
    [sym_line_comment] = STATE(81),
    [sym_block_comment] = STATE(81),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(638),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(82)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(82),
    [sym_waveform] = STATE(5877),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7640),
    [sym_delay_mechanism] = STATE(327),
    [sym_line_comment] = STATE(82),
    [sym_block_comment] = STATE(82),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(640),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_RELEASE] = ACTIONS(642),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(83)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(83),
    [sym_selected_waveforms] = STATE(7760),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(291),
    [sym_line_comment] = STATE(83),
    [sym_block_comment] = STATE(83),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(644),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(84)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(84),
    [sym_waveform] = STATE(5897),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7777),
    [sym_delay_mechanism] = STATE(334),
    [sym_line_comment] = STATE(84),
    [sym_block_comment] = STATE(84),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(646),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_RELEASE] = ACTIONS(648),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(85)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(85),
    [sym_selected_waveforms] = STATE(7727),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(287),
    [sym_line_comment] = STATE(85),
    [sym_block_comment] = STATE(85),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(650),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(86)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(86),
    [sym_selected_waveforms] = STATE(8020),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(306),
    [sym_line_comment] = STATE(86),
    [sym_block_comment] = STATE(86),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(652),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(87)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(87),
    [sym_selected_waveforms] = STATE(8157),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(318),
    [sym_line_comment] = STATE(87),
    [sym_block_comment] = STATE(87),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(654),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(88)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(88),
    [sym_selected_waveforms] = STATE(8178),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(320),
    [sym_line_comment] = STATE(88),
    [sym_block_comment] = STATE(88),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(656),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(89)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(89),
    [sym_waveform] = STATE(5938),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8192),
    [sym_delay_mechanism] = STATE(348),
    [sym_line_comment] = STATE(89),
    [sym_block_comment] = STATE(89),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(658),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_RELEASE] = ACTIONS(660),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(90)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(90),
    [sym_selected_waveforms] = STATE(8230),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(239),
    [sym_line_comment] = STATE(90),
    [sym_block_comment] = STATE(90),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(662),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(91)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(91),
    [sym_waveform] = STATE(5948),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8234),
    [sym_delay_mechanism] = STATE(350),
    [sym_line_comment] = STATE(91),
    [sym_block_comment] = STATE(91),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(664),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_RELEASE] = ACTIONS(666),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(92)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(92),
    [sym_selected_waveforms] = STATE(8222),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(323),
    [sym_line_comment] = STATE(92),
    [sym_block_comment] = STATE(92),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(668),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(93)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(93),
    [sym_selected_waveforms] = STATE(8256),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(240),
    [sym_line_comment] = STATE(93),
    [sym_block_comment] = STATE(93),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(670),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(94)] = {
    [sym_use_clause] = STATE(2040),
    [sym_package_declaration] = STATE(2040),
    [sym_package_instantiation_declaration] = STATE(2040),
    [sym_package_definition] = STATE(2040),
    [sym__block_declarative_item] = STATE(2029),
    [sym_subprogram_declaration] = STATE(2040),
    [sym_subprogram_definition] = STATE(2040),
    [sym_subprogram_instantiation_declaration] = STATE(2040),
    [sym_type_declaration] = STATE(2040),
    [sym_subtype_declaration] = STATE(2040),
    [sym_mode_view_declaration] = STATE(2040),
    [sym_constant_declaration] = STATE(2040),
    [sym_signal_declaration] = STATE(2040),
    [sym_variable_declaration] = STATE(2040),
    [sym_file_declaration] = STATE(2040),
    [sym_alias_declaration] = STATE(2040),
    [sym_component_declaration] = STATE(2040),
    [sym_attribute_declaration] = STATE(2040),
    [sym_attribute_specification] = STATE(2040),
    [sym_disconnection_specification] = STATE(2040),
    [sym_group_template_declaration] = STATE(2040),
    [sym_group_declaration] = STATE(2040),
    [sym__tool_directive] = STATE(94),
    [sym_procedure_specification] = STATE(6069),
    [sym_function_specification] = STATE(6069),
    [sym__subprogram_specification] = STATE(5708),
    [sym_configuration_specification] = STATE(2040),
    [sym_line_comment] = STATE(94),
    [sym_block_comment] = STATE(94),
    [aux_sym_design_unit_repeat2] = STATE(21),
    [anon_sym_LPAREN] = ACTIONS(672),
    [anon_sym_LT_LT] = ACTIONS(672),
    [sym_identifier] = ACTIONS(672),
    [sym_ALIAS] = ACTIONS(15),
    [sym_ARCHITECTURE] = ACTIONS(672),
    [sym_ASSERT] = ACTIONS(672),
    [sym_ATTRIBUTE] = ACTIONS(21),
    [sym_COMPONENT] = ACTIONS(23),
    [sym_CONFIGURATION] = ACTIONS(672),
    [sym_CONSTANT] = ACTIONS(27),
    [sym_CONTEXT] = ACTIONS(672),
    [sym_DISCONNECT] = ACTIONS(31),
    [sym_ENTITY] = ACTIONS(672),
    [sym_FILE] = ACTIONS(35),
    [sym_FOR] = ACTIONS(37),
    [sym_FUNCTION] = ACTIONS(39),
    [sym_GROUP] = ACTIONS(41),
    [sym_IMPURE] = ACTIONS(43),
    [sym_LIBRARY] = ACTIONS(672),
    [sym_PACKAGE] = ACTIONS(47),
    [sym_POSTPONED] = ACTIONS(672),
    [sym_PROCEDURE] = ACTIONS(51),
    [sym_PROCESS] = ACTIONS(672),
    [sym_PURE] = ACTIONS(43),
    [sym_SIGNAL] = ACTIONS(55),
    [sym_SHARED] = ACTIONS(57),
    [sym_SUBTYPE] = ACTIONS(59),
    [sym_TYPE] = ACTIONS(61),
    [sym_USE] = ACTIONS(63),
    [sym_VARIABLE] = ACTIONS(65),
    [sym_VIEW] = ACTIONS(67),
    [sym_WITH] = ACTIONS(672),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(672),
    [sym_operator_symbol] = ACTIONS(672),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(672),
    [sym_library_constant_std_logic] = ACTIONS(672),
    [sym_library_function] = ACTIONS(672),
    [sym_library_type] = ACTIONS(672),
    [sym__end_of_file] = ACTIONS(672),
  },
  [STATE(95)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(95),
    [sym_selected_waveforms] = STATE(6621),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(245),
    [sym_line_comment] = STATE(95),
    [sym_block_comment] = STATE(95),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(674),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(96)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(96),
    [sym_selected_waveforms] = STATE(6920),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(247),
    [sym_line_comment] = STATE(96),
    [sym_block_comment] = STATE(96),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(676),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(97)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(97),
    [sym_selected_waveforms] = STATE(7473),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(249),
    [sym_line_comment] = STATE(97),
    [sym_block_comment] = STATE(97),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(678),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(98)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(98),
    [sym_waveform] = STATE(5725),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(6965),
    [sym_delay_mechanism] = STATE(333),
    [sym_line_comment] = STATE(98),
    [sym_block_comment] = STATE(98),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_FORCE] = ACTIONS(680),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_RELEASE] = ACTIONS(682),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(99)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym_association_or_range_list] = STATE(7371),
    [sym__association_or_range] = STATE(5636),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(99),
    [sym_line_comment] = STATE(99),
    [sym_block_comment] = STATE(99),
    [anon_sym_LPAREN] = ACTIONS(263),
    [anon_sym_RPAREN] = ACTIONS(684),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(100)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(100),
    [sym_waveform] = STATE(5578),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7523),
    [sym_delay_mechanism] = STATE(332),
    [sym_line_comment] = STATE(100),
    [sym_block_comment] = STATE(100),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(686),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(101)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(101),
    [sym_selected_waveforms] = STATE(6893),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(248),
    [sym_line_comment] = STATE(101),
    [sym_block_comment] = STATE(101),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(102)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(102),
    [sym_waveform] = STATE(5916),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8104),
    [sym_delay_mechanism] = STATE(342),
    [sym_line_comment] = STATE(102),
    [sym_block_comment] = STATE(102),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(688),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(103)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(103),
    [sym_waveform] = STATE(5921),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8130),
    [sym_delay_mechanism] = STATE(344),
    [sym_line_comment] = STATE(103),
    [sym_block_comment] = STATE(103),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(690),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(104)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(104),
    [sym_waveform] = STATE(5853),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7514),
    [sym_delay_mechanism] = STATE(335),
    [sym_line_comment] = STATE(104),
    [sym_block_comment] = STATE(104),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(692),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(105)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(105),
    [sym_waveform] = STATE(5860),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7556),
    [sym_delay_mechanism] = STATE(337),
    [sym_line_comment] = STATE(105),
    [sym_block_comment] = STATE(105),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(694),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(106)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(106),
    [sym_waveform] = STATE(5759),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(6911),
    [sym_delay_mechanism] = STATE(331),
    [sym_line_comment] = STATE(106),
    [sym_block_comment] = STATE(106),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(696),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(107)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(107),
    [sym_selected_waveforms] = STATE(8178),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(320),
    [sym_line_comment] = STATE(107),
    [sym_block_comment] = STATE(107),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(108)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(108),
    [sym_waveform] = STATE(5934),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8180),
    [sym_delay_mechanism] = STATE(347),
    [sym_line_comment] = STATE(108),
    [sym_block_comment] = STATE(108),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(698),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(109)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(109),
    [sym_selected_waveforms] = STATE(8222),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(323),
    [sym_line_comment] = STATE(109),
    [sym_block_comment] = STATE(109),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(110)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(110),
    [sym_waveform] = STATE(5703),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(6900),
    [sym_delay_mechanism] = STATE(340),
    [sym_line_comment] = STATE(110),
    [sym_block_comment] = STATE(110),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(700),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(111)] = {
    [sym__concurrent_statement] = STATE(2052),
    [sym_block_statement] = STATE(2060),
    [sym_component_instantiation_statement] = STATE(2060),
    [sym_process_statement] = STATE(2060),
    [sym_case_generate_statement] = STATE(2060),
    [sym_for_generate_statement] = STATE(2060),
    [sym_if_generate_statement] = STATE(2060),
    [sym_concurrent_assertion_statement] = STATE(2060),
    [sym_assertion] = STATE(7241),
    [sym_concurrent_procedure_call_statement] = STATE(2060),
    [sym_concurrent_simple_signal_assignment] = STATE(2060),
    [sym_concurrent_conditional_signal_assignment] = STATE(2060),
    [sym_concurrent_selected_signal_assignment] = STATE(2060),
    [sym_with_expression] = STATE(6491),
    [sym_name] = STATE(6540),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7169),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(111),
    [sym__target] = STATE(6080),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(2689),
    [sym_line_comment] = STATE(111),
    [sym_block_comment] = STATE(111),
    [aux_sym_design_unit_repeat3] = STATE(111),
    [anon_sym_LPAREN] = ACTIONS(702),
    [anon_sym_LT_LT] = ACTIONS(705),
    [sym_identifier] = ACTIONS(708),
    [sym_ALIAS] = ACTIONS(711),
    [sym_ARCHITECTURE] = ACTIONS(711),
    [sym_ASSERT] = ACTIONS(713),
    [sym_ATTRIBUTE] = ACTIONS(711),
    [sym_COMPONENT] = ACTIONS(711),
    [sym_CONFIGURATION] = ACTIONS(711),
    [sym_CONSTANT] = ACTIONS(711),
    [sym_CONTEXT] = ACTIONS(711),
    [sym_DISCONNECT] = ACTIONS(711),
    [sym_ENTITY] = ACTIONS(711),
    [sym_FILE] = ACTIONS(711),
    [sym_FOR] = ACTIONS(711),
    [sym_FUNCTION] = ACTIONS(711),
    [sym_GROUP] = ACTIONS(711),
    [sym_IMPURE] = ACTIONS(711),
    [sym_LIBRARY] = ACTIONS(711),
    [sym_PACKAGE] = ACTIONS(711),
    [sym_POSTPONED] = ACTIONS(716),
    [sym_PROCEDURE] = ACTIONS(711),
    [sym_PROCESS] = ACTIONS(719),
    [sym_PURE] = ACTIONS(711),
    [sym_SIGNAL] = ACTIONS(711),
    [sym_SHARED] = ACTIONS(711),
    [sym_SUBTYPE] = ACTIONS(711),
    [sym_TYPE] = ACTIONS(711),
    [sym_USE] = ACTIONS(711),
    [sym_VARIABLE] = ACTIONS(711),
    [sym_VIEW] = ACTIONS(711),
    [sym_WITH] = ACTIONS(722),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(725),
    [sym_operator_symbol] = ACTIONS(725),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(708),
    [sym_library_constant_std_logic] = ACTIONS(725),
    [sym_library_function] = ACTIONS(708),
    [sym_library_type] = ACTIONS(708),
    [sym__end_of_file] = ACTIONS(711),
  },
  [STATE(112)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(112),
    [sym_selected_waveforms] = STATE(7625),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(276),
    [sym_line_comment] = STATE(112),
    [sym_block_comment] = STATE(112),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(113)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(113),
    [sym_waveform] = STATE(5873),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7628),
    [sym_delay_mechanism] = STATE(326),
    [sym_line_comment] = STATE(113),
    [sym_block_comment] = STATE(113),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(728),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(114)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(114),
    [sym_selected_waveforms] = STATE(8253),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(241),
    [sym_line_comment] = STATE(114),
    [sym_block_comment] = STATE(114),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(115)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(115),
    [sym_waveform] = STATE(5574),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7017),
    [sym_delay_mechanism] = STATE(329),
    [sym_line_comment] = STATE(115),
    [sym_block_comment] = STATE(115),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(730),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(116)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(116),
    [sym_selected_waveforms] = STATE(7727),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(287),
    [sym_line_comment] = STATE(116),
    [sym_block_comment] = STATE(116),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(117)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(117),
    [sym_waveform] = STATE(5809),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7303),
    [sym_delay_mechanism] = STATE(328),
    [sym_line_comment] = STATE(117),
    [sym_block_comment] = STATE(117),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(732),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(118)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(118),
    [sym_selected_waveforms] = STATE(7404),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(297),
    [sym_line_comment] = STATE(118),
    [sym_block_comment] = STATE(118),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(119)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(119),
    [sym_selected_waveforms] = STATE(6920),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(247),
    [sym_line_comment] = STATE(119),
    [sym_block_comment] = STATE(119),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(120)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(120),
    [sym_waveform] = STATE(5640),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(6934),
    [sym_delay_mechanism] = STATE(338),
    [sym_line_comment] = STATE(120),
    [sym_block_comment] = STATE(120),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_GUARDED] = ACTIONS(734),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(121)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(121),
    [sym_selected_waveforms] = STATE(7473),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(249),
    [sym_line_comment] = STATE(121),
    [sym_block_comment] = STATE(121),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(122)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(122),
    [sym_selected_waveforms] = STATE(7827),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(244),
    [sym_line_comment] = STATE(122),
    [sym_block_comment] = STATE(122),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(123)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(123),
    [sym_selected_waveforms] = STATE(8054),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(250),
    [sym_line_comment] = STATE(123),
    [sym_block_comment] = STATE(123),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(124)] = {
    [sym__concurrent_statement] = STATE(2052),
    [sym_block_statement] = STATE(2060),
    [sym_component_instantiation_statement] = STATE(2060),
    [sym_process_statement] = STATE(2060),
    [sym_case_generate_statement] = STATE(2060),
    [sym_for_generate_statement] = STATE(2060),
    [sym_if_generate_statement] = STATE(2060),
    [sym_concurrent_assertion_statement] = STATE(2060),
    [sym_assertion] = STATE(7241),
    [sym_concurrent_procedure_call_statement] = STATE(2060),
    [sym_concurrent_simple_signal_assignment] = STATE(2060),
    [sym_concurrent_conditional_signal_assignment] = STATE(2060),
    [sym_concurrent_selected_signal_assignment] = STATE(2060),
    [sym_with_expression] = STATE(6491),
    [sym_name] = STATE(6540),
    [sym__direct_name] = STATE(3023),
    [sym__label] = STATE(7169),
    [sym__identifier] = STATE(4362),
    [sym__external_name] = STATE(4362),
    [sym_external_constant_name] = STATE(4351),
    [sym_external_signal_name] = STATE(4351),
    [sym_external_variable_name] = STATE(4351),
    [sym__tool_directive] = STATE(124),
    [sym__target] = STATE(6080),
    [sym_aggregate] = STATE(6098),
    [sym_label_declaration] = STATE(2689),
    [sym_line_comment] = STATE(124),
    [sym_block_comment] = STATE(124),
    [aux_sym_design_unit_repeat3] = STATE(111),
    [anon_sym_LPAREN] = ACTIONS(9),
    [anon_sym_LT_LT] = ACTIONS(11),
    [sym_identifier] = ACTIONS(13),
    [sym_ALIAS] = ACTIONS(672),
    [sym_ARCHITECTURE] = ACTIONS(672),
    [sym_ASSERT] = ACTIONS(19),
    [sym_ATTRIBUTE] = ACTIONS(672),
    [sym_COMPONENT] = ACTIONS(672),
    [sym_CONFIGURATION] = ACTIONS(672),
    [sym_CONSTANT] = ACTIONS(672),
    [sym_CONTEXT] = ACTIONS(672),
    [sym_DISCONNECT] = ACTIONS(672),
    [sym_ENTITY] = ACTIONS(672),
    [sym_FILE] = ACTIONS(672),
    [sym_FOR] = ACTIONS(672),
    [sym_FUNCTION] = ACTIONS(672),
    [sym_GROUP] = ACTIONS(672),
    [sym_IMPURE] = ACTIONS(672),
    [sym_LIBRARY] = ACTIONS(672),
    [sym_PACKAGE] = ACTIONS(672),
    [sym_POSTPONED] = ACTIONS(49),
    [sym_PROCEDURE] = ACTIONS(672),
    [sym_PROCESS] = ACTIONS(53),
    [sym_PURE] = ACTIONS(672),
    [sym_SIGNAL] = ACTIONS(672),
    [sym_SHARED] = ACTIONS(672),
    [sym_SUBTYPE] = ACTIONS(672),
    [sym_TYPE] = ACTIONS(672),
    [sym_USE] = ACTIONS(672),
    [sym_VARIABLE] = ACTIONS(672),
    [sym_VIEW] = ACTIONS(672),
    [sym_WITH] = ACTIONS(69),
    [sym__grave_accent] = ACTIONS(3),
    [sym_character_literal] = ACTIONS(71),
    [sym_operator_symbol] = ACTIONS(71),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(13),
    [sym_library_constant_std_logic] = ACTIONS(71),
    [sym_library_function] = ACTIONS(13),
    [sym_library_type] = ACTIONS(13),
    [sym__end_of_file] = ACTIONS(672),
  },
  [STATE(125)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(125),
    [sym_selected_waveforms] = STATE(8017),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_delay_mechanism] = STATE(309),
    [sym_line_comment] = STATE(125),
    [sym_block_comment] = STATE(125),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(126)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6676),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(126),
    [sym_line_comment] = STATE(126),
    [sym_block_comment] = STATE(126),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(127)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6698),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(127),
    [sym_line_comment] = STATE(127),
    [sym_block_comment] = STATE(127),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(128)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(128),
    [sym_waveform] = STATE(5921),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8130),
    [sym_delay_mechanism] = STATE(344),
    [sym_line_comment] = STATE(128),
    [sym_block_comment] = STATE(128),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(129)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(129),
    [sym_waveform] = STATE(5836),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7413),
    [sym_delay_mechanism] = STATE(349),
    [sym_line_comment] = STATE(129),
    [sym_block_comment] = STATE(129),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(130)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(130),
    [sym_waveform] = STATE(5934),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8180),
    [sym_delay_mechanism] = STATE(347),
    [sym_line_comment] = STATE(130),
    [sym_block_comment] = STATE(130),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(131)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2352),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2353),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__association_or_range] = STATE(6169),
    [sym_association_element] = STATE(6166),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym__range] = STATE(6166),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(131),
    [sym_line_comment] = STATE(131),
    [sym_block_comment] = STATE(131),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(132)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7479),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(132),
    [sym_line_comment] = STATE(132),
    [sym_block_comment] = STATE(132),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(133)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(133),
    [sym_waveform] = STATE(5703),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(6900),
    [sym_delay_mechanism] = STATE(340),
    [sym_line_comment] = STATE(133),
    [sym_block_comment] = STATE(133),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(134)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(134),
    [sym_waveform] = STATE(5860),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7556),
    [sym_delay_mechanism] = STATE(337),
    [sym_line_comment] = STATE(134),
    [sym_block_comment] = STATE(134),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(135)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(135),
    [sym_waveform] = STATE(5946),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8224),
    [sym_delay_mechanism] = STATE(330),
    [sym_line_comment] = STATE(135),
    [sym_block_comment] = STATE(135),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(136)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(136),
    [sym_waveform] = STATE(5873),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7628),
    [sym_delay_mechanism] = STATE(326),
    [sym_line_comment] = STATE(136),
    [sym_block_comment] = STATE(136),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(137)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7704),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(137),
    [sym_line_comment] = STATE(137),
    [sym_block_comment] = STATE(137),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(138)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(138),
    [sym_waveform] = STATE(5809),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7303),
    [sym_delay_mechanism] = STATE(328),
    [sym_line_comment] = STATE(138),
    [sym_block_comment] = STATE(138),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(139)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(139),
    [sym_waveform] = STATE(5578),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7523),
    [sym_delay_mechanism] = STATE(332),
    [sym_line_comment] = STATE(139),
    [sym_block_comment] = STATE(139),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(140)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8176),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(140),
    [sym_line_comment] = STATE(140),
    [sym_block_comment] = STATE(140),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(141)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(141),
    [sym_waveform] = STATE(5640),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(6934),
    [sym_delay_mechanism] = STATE(338),
    [sym_line_comment] = STATE(141),
    [sym_block_comment] = STATE(141),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(142)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(142),
    [sym_waveform] = STATE(5891),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7736),
    [sym_delay_mechanism] = STATE(336),
    [sym_line_comment] = STATE(142),
    [sym_block_comment] = STATE(142),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(143)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(143),
    [sym_waveform] = STATE(5781),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7559),
    [sym_delay_mechanism] = STATE(341),
    [sym_line_comment] = STATE(143),
    [sym_block_comment] = STATE(143),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(421),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_REJECT] = ACTIONS(427),
    [sym_TRANSPORT] = ACTIONS(421),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(144)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6798),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(144),
    [sym_line_comment] = STATE(144),
    [sym_block_comment] = STATE(144),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(145)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7420),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(145),
    [sym_line_comment] = STATE(145),
    [sym_block_comment] = STATE(145),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(146)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8313),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(146),
    [sym_line_comment] = STATE(146),
    [sym_block_comment] = STATE(146),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(147)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6816),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(147),
    [sym_line_comment] = STATE(147),
    [sym_block_comment] = STATE(147),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(148)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6926),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(148),
    [sym_line_comment] = STATE(148),
    [sym_block_comment] = STATE(148),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(149)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7028),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(149),
    [sym_line_comment] = STATE(149),
    [sym_block_comment] = STATE(149),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(150)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7193),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(150),
    [sym_line_comment] = STATE(150),
    [sym_block_comment] = STATE(150),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(151)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7318),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(151),
    [sym_line_comment] = STATE(151),
    [sym_block_comment] = STATE(151),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(152)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7430),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(152),
    [sym_line_comment] = STATE(152),
    [sym_block_comment] = STATE(152),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(153)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7641),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(153),
    [sym_line_comment] = STATE(153),
    [sym_block_comment] = STATE(153),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(154)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7772),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(154),
    [sym_line_comment] = STATE(154),
    [sym_block_comment] = STATE(154),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(155)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8067),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(155),
    [sym_line_comment] = STATE(155),
    [sym_block_comment] = STATE(155),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(156)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8262),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(156),
    [sym_line_comment] = STATE(156),
    [sym_block_comment] = STATE(156),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(157)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6609),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(157),
    [sym_line_comment] = STATE(157),
    [sym_block_comment] = STATE(157),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(158)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6664),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(158),
    [sym_line_comment] = STATE(158),
    [sym_block_comment] = STATE(158),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(159)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6731),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(159),
    [sym_line_comment] = STATE(159),
    [sym_block_comment] = STATE(159),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(160)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6839),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(160),
    [sym_line_comment] = STATE(160),
    [sym_block_comment] = STATE(160),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(161)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6872),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(161),
    [sym_line_comment] = STATE(161),
    [sym_block_comment] = STATE(161),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(162)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6919),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(162),
    [sym_line_comment] = STATE(162),
    [sym_block_comment] = STATE(162),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(163)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6960),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(163),
    [sym_line_comment] = STATE(163),
    [sym_block_comment] = STATE(163),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(164)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7001),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(164),
    [sym_line_comment] = STATE(164),
    [sym_block_comment] = STATE(164),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(165)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7050),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(165),
    [sym_line_comment] = STATE(165),
    [sym_block_comment] = STATE(165),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(166)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7120),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(166),
    [sym_line_comment] = STATE(166),
    [sym_block_comment] = STATE(166),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(167)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8079),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(167),
    [sym_line_comment] = STATE(167),
    [sym_block_comment] = STATE(167),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(168)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7230),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(168),
    [sym_line_comment] = STATE(168),
    [sym_block_comment] = STATE(168),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(169)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7278),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(169),
    [sym_line_comment] = STATE(169),
    [sym_block_comment] = STATE(169),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(170)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7327),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(170),
    [sym_line_comment] = STATE(170),
    [sym_block_comment] = STATE(170),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(171)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7364),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(171),
    [sym_line_comment] = STATE(171),
    [sym_block_comment] = STATE(171),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(172)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7409),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(172),
    [sym_line_comment] = STATE(172),
    [sym_block_comment] = STATE(172),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(173)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7464),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(173),
    [sym_line_comment] = STATE(173),
    [sym_block_comment] = STATE(173),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(174)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7537),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(174),
    [sym_line_comment] = STATE(174),
    [sym_block_comment] = STATE(174),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(175)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7622),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(175),
    [sym_line_comment] = STATE(175),
    [sym_block_comment] = STATE(175),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(176)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7672),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(176),
    [sym_line_comment] = STATE(176),
    [sym_block_comment] = STATE(176),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(177)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7710),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(177),
    [sym_line_comment] = STATE(177),
    [sym_block_comment] = STATE(177),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(178)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7748),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(178),
    [sym_line_comment] = STATE(178),
    [sym_block_comment] = STATE(178),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(179)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7779),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(179),
    [sym_line_comment] = STATE(179),
    [sym_block_comment] = STATE(179),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(180)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7820),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(180),
    [sym_line_comment] = STATE(180),
    [sym_block_comment] = STATE(180),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(181)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8034),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(181),
    [sym_line_comment] = STATE(181),
    [sym_block_comment] = STATE(181),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(182)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8080),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(182),
    [sym_line_comment] = STATE(182),
    [sym_block_comment] = STATE(182),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(183)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6786),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(183),
    [sym_line_comment] = STATE(183),
    [sym_block_comment] = STATE(183),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(184)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8137),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(184),
    [sym_line_comment] = STATE(184),
    [sym_block_comment] = STATE(184),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(185)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8199),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(185),
    [sym_line_comment] = STATE(185),
    [sym_block_comment] = STATE(185),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(186)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8284),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(186),
    [sym_line_comment] = STATE(186),
    [sym_block_comment] = STATE(186),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(187)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8317),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(187),
    [sym_line_comment] = STATE(187),
    [sym_block_comment] = STATE(187),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(188)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(8336),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(188),
    [sym_line_comment] = STATE(188),
    [sym_block_comment] = STATE(188),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(189)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7601),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(189),
    [sym_line_comment] = STATE(189),
    [sym_block_comment] = STATE(189),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(190)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6627),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(190),
    [sym_line_comment] = STATE(190),
    [sym_block_comment] = STATE(190),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(191)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6643),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(191),
    [sym_line_comment] = STATE(191),
    [sym_block_comment] = STATE(191),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(192)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6653),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(192),
    [sym_line_comment] = STATE(192),
    [sym_block_comment] = STATE(192),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(193)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6662),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(193),
    [sym_line_comment] = STATE(193),
    [sym_block_comment] = STATE(193),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(194)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6666),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(194),
    [sym_line_comment] = STATE(194),
    [sym_block_comment] = STATE(194),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(195)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6670),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(195),
    [sym_line_comment] = STATE(195),
    [sym_block_comment] = STATE(195),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(196)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6674),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(196),
    [sym_line_comment] = STATE(196),
    [sym_block_comment] = STATE(196),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(197)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6675),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(197),
    [sym_line_comment] = STATE(197),
    [sym_block_comment] = STATE(197),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(198)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6677),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(198),
    [sym_line_comment] = STATE(198),
    [sym_block_comment] = STATE(198),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(199)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6678),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(199),
    [sym_line_comment] = STATE(199),
    [sym_block_comment] = STATE(199),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(200)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6679),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(200),
    [sym_line_comment] = STATE(200),
    [sym_block_comment] = STATE(200),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(201)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6680),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(201),
    [sym_line_comment] = STATE(201),
    [sym_block_comment] = STATE(201),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(202)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6681),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(202),
    [sym_line_comment] = STATE(202),
    [sym_block_comment] = STATE(202),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(203)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6682),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(203),
    [sym_line_comment] = STATE(203),
    [sym_block_comment] = STATE(203),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(204)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6683),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(204),
    [sym_line_comment] = STATE(204),
    [sym_block_comment] = STATE(204),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(205)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6684),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(205),
    [sym_line_comment] = STATE(205),
    [sym_block_comment] = STATE(205),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(206)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6685),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(206),
    [sym_line_comment] = STATE(206),
    [sym_block_comment] = STATE(206),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(207)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6686),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(207),
    [sym_line_comment] = STATE(207),
    [sym_block_comment] = STATE(207),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(208)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6687),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(208),
    [sym_line_comment] = STATE(208),
    [sym_block_comment] = STATE(208),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(209)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6688),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(209),
    [sym_line_comment] = STATE(209),
    [sym_block_comment] = STATE(209),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(210)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6689),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(210),
    [sym_line_comment] = STATE(210),
    [sym_block_comment] = STATE(210),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(211)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6690),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(211),
    [sym_line_comment] = STATE(211),
    [sym_block_comment] = STATE(211),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(212)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6691),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(212),
    [sym_line_comment] = STATE(212),
    [sym_block_comment] = STATE(212),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(213)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6692),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(213),
    [sym_line_comment] = STATE(213),
    [sym_block_comment] = STATE(213),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(214)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6693),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(214),
    [sym_line_comment] = STATE(214),
    [sym_block_comment] = STATE(214),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(215)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6694),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(215),
    [sym_line_comment] = STATE(215),
    [sym_block_comment] = STATE(215),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(216)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6695),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(216),
    [sym_line_comment] = STATE(216),
    [sym_block_comment] = STATE(216),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(217)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6696),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(217),
    [sym_line_comment] = STATE(217),
    [sym_block_comment] = STATE(217),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(218)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(6697),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(218),
    [sym_line_comment] = STATE(218),
    [sym_block_comment] = STATE(218),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(219)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association_list] = STATE(7182),
    [sym_element_association] = STATE(5649),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(219),
    [sym_line_comment] = STATE(219),
    [sym_block_comment] = STATE(219),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(220)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5689),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(220),
    [sym_line_comment] = STATE(220),
    [sym_block_comment] = STATE(220),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_DEFAULT] = ACTIONS(742),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_box] = ACTIONS(748),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(221)] = {
    [sym__expression] = STATE(1050),
    [sym_condition_expression] = STATE(1944),
    [sym_logical_expression] = STATE(1944),
    [sym_relational_expression] = STATE(1944),
    [sym_shift_expression] = STATE(1944),
    [sym_simple_expression] = STATE(2021),
    [sym_unary_operator] = STATE(720),
    [sym_sign] = STATE(439),
    [sym__primary] = STATE(1620),
    [sym_name] = STATE(1621),
    [sym__direct_name] = STATE(721),
    [sym__label] = STATE(6648),
    [sym__identifier] = STATE(838),
    [sym__external_name] = STATE(838),
    [sym_external_constant_name] = STATE(840),
    [sym_external_signal_name] = STATE(840),
    [sym_external_variable_name] = STATE(840),
    [sym__literal] = STATE(1622),
    [sym_bit_string_literal] = STATE(1612),
    [sym__abstract_literal] = STATE(846),
    [sym_based_literal] = STATE(902),
    [sym_allocator] = STATE(1621),
    [sym_parenthesis_expression] = STATE(1621),
    [sym__element] = STATE(5622),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(221),
    [sym_label_declaration] = STATE(322),
    [sym_line_comment] = STATE(221),
    [sym_block_comment] = STATE(221),
    [anon_sym_LPAREN] = ACTIONS(760),
    [sym_condition_conversion] = ACTIONS(762),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(764),
    [sym_identifier] = ACTIONS(766),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(768),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(770),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(772),
    [sym_decimal_float] = ACTIONS(772),
    [sym_based_base] = ACTIONS(774),
    [sym_character_literal] = ACTIONS(776),
    [sym_string_literal] = ACTIONS(770),
    [sym_string_literal_std_logic] = ACTIONS(770),
    [sym_bit_string_length] = ACTIONS(778),
    [sym_bit_string_base] = ACTIONS(780),
    [sym_operator_symbol] = ACTIONS(776),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(766),
    [sym_library_constant_boolean] = ACTIONS(770),
    [sym_library_constant_character] = ACTIONS(770),
    [sym_library_constant_debug] = ACTIONS(770),
    [sym_library_constant_env] = ACTIONS(770),
    [sym_library_constant_standard] = ACTIONS(770),
    [sym_library_constant_std_logic] = ACTIONS(776),
    [sym_library_function] = ACTIONS(766),
    [sym_library_type] = ACTIONS(766),
  },
  [STATE(222)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association] = STATE(6131),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(222),
    [sym_line_comment] = STATE(222),
    [sym_block_comment] = STATE(222),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(223)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association] = STATE(6347),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(223),
    [sym_line_comment] = STATE(223),
    [sym_block_comment] = STATE(223),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(224)] = {
    [sym_conditional_expression] = STATE(6025),
    [sym__expression] = STATE(1102),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(1716),
    [sym_unary_operator] = STATE(686),
    [sym_sign] = STATE(687),
    [sym__primary] = STATE(2097),
    [sym_name] = STATE(2098),
    [sym__direct_name] = STATE(734),
    [sym__identifier] = STATE(966),
    [sym__external_name] = STATE(966),
    [sym_external_constant_name] = STATE(967),
    [sym_external_signal_name] = STATE(967),
    [sym_external_variable_name] = STATE(967),
    [sym__literal] = STATE(2099),
    [sym_bit_string_literal] = STATE(2089),
    [sym__abstract_literal] = STATE(997),
    [sym_based_literal] = STATE(1044),
    [sym_allocator] = STATE(2098),
    [sym_parenthesis_expression] = STATE(2098),
    [sym_element_association] = STATE(5699),
    [sym__element] = STATE(6052),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(224),
    [sym_line_comment] = STATE(224),
    [sym_block_comment] = STATE(224),
    [anon_sym_LPAREN] = ACTIONS(263),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(401),
    [sym_identifier] = ACTIONS(403),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(275),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(277),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(283),
    [sym_decimal_float] = ACTIONS(283),
    [sym_based_base] = ACTIONS(285),
    [sym_character_literal] = ACTIONS(409),
    [sym_string_literal] = ACTIONS(277),
    [sym_string_literal_std_logic] = ACTIONS(277),
    [sym_bit_string_length] = ACTIONS(289),
    [sym_bit_string_base] = ACTIONS(291),
    [sym_operator_symbol] = ACTIONS(409),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(403),
    [sym_library_constant_boolean] = ACTIONS(277),
    [sym_library_constant_character] = ACTIONS(277),
    [sym_library_constant_debug] = ACTIONS(277),
    [sym_library_constant_env] = ACTIONS(277),
    [sym_library_constant_standard] = ACTIONS(277),
    [sym_library_constant_std_logic] = ACTIONS(409),
    [sym_library_function] = ACTIONS(403),
    [sym_library_type] = ACTIONS(403),
  },
  [STATE(225)] = {
    [sym_unspecified_type_indication] = STATE(6478),
    [sym__array_index_incomplete_type] = STATE(6288),
    [sym_index_constraint] = STATE(6478),
    [sym_index_subtype_definition] = STATE(6478),
    [sym__expression] = STATE(1050),
    [sym_condition_expression] = STATE(1944),
    [sym_logical_expression] = STATE(1944),
    [sym_relational_expression] = STATE(1944),
    [sym_shift_expression] = STATE(1944),
    [sym_simple_expression] = STATE(1837),
    [sym_unary_operator] = STATE(720),
    [sym_sign] = STATE(439),
    [sym__primary] = STATE(1620),
    [sym_name] = STATE(2469),
    [sym__direct_name] = STATE(752),
    [sym__identifier] = STATE(1192),
    [sym__external_name] = STATE(1192),
    [sym_external_constant_name] = STATE(1122),
    [sym_external_signal_name] = STATE(1122),
    [sym_external_variable_name] = STATE(1122),
    [sym__literal] = STATE(1622),
    [sym_bit_string_literal] = STATE(1612),
    [sym__abstract_literal] = STATE(846),
    [sym_based_literal] = STATE(902),
    [sym_allocator] = STATE(1621),
    [sym_parenthesis_expression] = STATE(1621),
    [sym__range] = STATE(5752),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(225),
    [sym_line_comment] = STATE(225),
    [sym_block_comment] = STATE(225),
    [anon_sym_LPAREN] = ACTIONS(760),
    [sym_condition_conversion] = ACTIONS(762),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(782),
    [sym_identifier] = ACTIONS(784),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(768),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(770),
    [sym_OR] = ACTIONS(273),
    [sym_TYPE] = ACTIONS(281),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(772),
    [sym_decimal_float] = ACTIONS(772),
    [sym_based_base] = ACTIONS(774),
    [sym_character_literal] = ACTIONS(786),
    [sym_string_literal] = ACTIONS(770),
    [sym_string_literal_std_logic] = ACTIONS(770),
    [sym_bit_string_length] = ACTIONS(778),
    [sym_bit_string_base] = ACTIONS(780),
    [sym_operator_symbol] = ACTIONS(786),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(784),
    [sym_library_constant_boolean] = ACTIONS(770),
    [sym_library_constant_character] = ACTIONS(770),
    [sym_library_constant_debug] = ACTIONS(770),
    [sym_library_constant_env] = ACTIONS(770),
    [sym_library_constant_standard] = ACTIONS(770),
    [sym_library_constant_std_logic] = ACTIONS(786),
    [sym_library_function] = ACTIONS(784),
    [sym_library_type] = ACTIONS(784),
  },
  [STATE(226)] = {
    [sym_when_expression] = STATE(8195),
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(226),
    [sym_conditional_or_unaffected_expression] = STATE(8195),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(226),
    [sym_block_comment] = STATE(226),
    [anon_sym_SEMI] = ACTIONS(788),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_WHEN] = ACTIONS(792),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(227)] = {
    [sym_when_expression] = STATE(8136),
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(227),
    [sym_conditional_or_unaffected_expression] = STATE(8136),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(227),
    [sym_block_comment] = STATE(227),
    [anon_sym_SEMI] = ACTIONS(794),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_WHEN] = ACTIONS(792),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(228)] = {
    [sym_when_expression] = STATE(7643),
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(228),
    [sym_conditional_or_unaffected_expression] = STATE(7643),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(228),
    [sym_block_comment] = STATE(228),
    [anon_sym_SEMI] = ACTIONS(796),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_WHEN] = ACTIONS(792),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(229)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(229),
    [sym_force_mode] = STATE(365),
    [sym_conditional_or_unaffected_expression] = STATE(8342),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(229),
    [sym_block_comment] = STATE(229),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(798),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(798),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(230)] = {
    [sym_when_expression] = STATE(7394),
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(230),
    [sym_conditional_or_unaffected_expression] = STATE(7394),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(230),
    [sym_block_comment] = STATE(230),
    [anon_sym_SEMI] = ACTIONS(800),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_WHEN] = ACTIONS(792),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(231)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(231),
    [sym_force_mode] = STATE(354),
    [sym_conditional_or_unaffected_expression] = STATE(8231),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(231),
    [sym_block_comment] = STATE(231),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(798),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(798),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(232)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(232),
    [sym_force_mode] = STATE(364),
    [sym_conditional_or_unaffected_expression] = STATE(7764),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(232),
    [sym_block_comment] = STATE(232),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(798),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(798),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(233)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(233),
    [sym_force_mode] = STATE(361),
    [sym_conditional_or_unaffected_expression] = STATE(7495),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(233),
    [sym_block_comment] = STATE(233),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(798),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(798),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(234)] = {
    [sym_when_expression] = STATE(7562),
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(234),
    [sym_conditional_or_unaffected_expression] = STATE(7562),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(234),
    [sym_block_comment] = STATE(234),
    [anon_sym_SEMI] = ACTIONS(802),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_WHEN] = ACTIONS(792),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(235)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(235),
    [sym_force_mode] = STATE(360),
    [sym_conditional_or_unaffected_expression] = STATE(8257),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(235),
    [sym_block_comment] = STATE(235),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(798),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(798),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(236)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(236),
    [sym_force_mode] = STATE(359),
    [sym_conditional_or_unaffected_expression] = STATE(8021),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(236),
    [sym_block_comment] = STATE(236),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(798),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(798),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(237)] = {
    [sym_when_expression] = STATE(6980),
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(237),
    [sym_conditional_or_unaffected_expression] = STATE(6980),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(237),
    [sym_block_comment] = STATE(237),
    [anon_sym_SEMI] = ACTIONS(804),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_WHEN] = ACTIONS(792),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(238)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5663),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(238),
    [sym_line_comment] = STATE(238),
    [sym_block_comment] = STATE(238),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(239)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(239),
    [sym_selected_waveforms] = STATE(8256),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(239),
    [sym_block_comment] = STATE(239),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(240)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(240),
    [sym_selected_waveforms] = STATE(8274),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(240),
    [sym_block_comment] = STATE(240),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(241)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(241),
    [sym_selected_waveforms] = STATE(8276),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(241),
    [sym_block_comment] = STATE(241),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(242)] = {
    [sym__expression] = STATE(2275),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(242),
    [sym_force_mode] = STATE(373),
    [sym_selected_expressions] = STATE(7613),
    [sym_line_comment] = STATE(242),
    [sym_block_comment] = STATE(242),
    [aux_sym_selected_expressions_repeat1] = STATE(405),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(806),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(806),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(243)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(243),
    [sym_selected_waveforms] = STATE(7709),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(243),
    [sym_block_comment] = STATE(243),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(244)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(244),
    [sym_selected_waveforms] = STATE(8332),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(244),
    [sym_block_comment] = STATE(244),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(245)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(245),
    [sym_selected_waveforms] = STATE(6920),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(245),
    [sym_block_comment] = STATE(245),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(246)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5689),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(246),
    [sym_line_comment] = STATE(246),
    [sym_block_comment] = STATE(246),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(247)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(247),
    [sym_selected_waveforms] = STATE(7473),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(247),
    [sym_block_comment] = STATE(247),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(248)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(248),
    [sym_selected_waveforms] = STATE(7404),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(248),
    [sym_block_comment] = STATE(248),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(249)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(249),
    [sym_selected_waveforms] = STATE(8054),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(249),
    [sym_block_comment] = STATE(249),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(250)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(250),
    [sym_selected_waveforms] = STATE(8338),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(250),
    [sym_block_comment] = STATE(250),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(251)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5593),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(251),
    [sym_line_comment] = STATE(251),
    [sym_block_comment] = STATE(251),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(252)] = {
    [sym__expression] = STATE(1050),
    [sym_condition_expression] = STATE(1944),
    [sym_logical_expression] = STATE(1944),
    [sym_relational_expression] = STATE(1944),
    [sym_shift_expression] = STATE(1944),
    [sym_simple_expression] = STATE(2021),
    [sym_unary_operator] = STATE(720),
    [sym_sign] = STATE(439),
    [sym__primary] = STATE(1620),
    [sym_name] = STATE(1621),
    [sym__direct_name] = STATE(721),
    [sym__identifier] = STATE(838),
    [sym__external_name] = STATE(838),
    [sym_external_constant_name] = STATE(840),
    [sym_external_signal_name] = STATE(840),
    [sym_external_variable_name] = STATE(840),
    [sym__literal] = STATE(1622),
    [sym_bit_string_literal] = STATE(1612),
    [sym__abstract_literal] = STATE(846),
    [sym_based_literal] = STATE(902),
    [sym_allocator] = STATE(1621),
    [sym_parenthesis_expression] = STATE(1621),
    [sym__element] = STATE(5538),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(252),
    [sym_line_comment] = STATE(252),
    [sym_block_comment] = STATE(252),
    [anon_sym_LPAREN] = ACTIONS(760),
    [sym_condition_conversion] = ACTIONS(762),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(764),
    [sym_identifier] = ACTIONS(808),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(768),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(770),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(772),
    [sym_decimal_float] = ACTIONS(772),
    [sym_based_base] = ACTIONS(774),
    [sym_character_literal] = ACTIONS(776),
    [sym_string_literal] = ACTIONS(770),
    [sym_string_literal_std_logic] = ACTIONS(770),
    [sym_bit_string_length] = ACTIONS(778),
    [sym_bit_string_base] = ACTIONS(780),
    [sym_operator_symbol] = ACTIONS(776),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(808),
    [sym_library_constant_boolean] = ACTIONS(770),
    [sym_library_constant_character] = ACTIONS(770),
    [sym_library_constant_debug] = ACTIONS(770),
    [sym_library_constant_env] = ACTIONS(770),
    [sym_library_constant_standard] = ACTIONS(770),
    [sym_library_constant_std_logic] = ACTIONS(776),
    [sym_library_function] = ACTIONS(808),
    [sym_library_type] = ACTIONS(808),
  },
  [STATE(253)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5668),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(253),
    [sym_line_comment] = STATE(253),
    [sym_block_comment] = STATE(253),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(254)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5713),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(254),
    [sym_line_comment] = STATE(254),
    [sym_block_comment] = STATE(254),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(255)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5783),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(255),
    [sym_line_comment] = STATE(255),
    [sym_block_comment] = STATE(255),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(256)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5841),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(256),
    [sym_line_comment] = STATE(256),
    [sym_block_comment] = STATE(256),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(257)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5937),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(257),
    [sym_line_comment] = STATE(257),
    [sym_block_comment] = STATE(257),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(258)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5566),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(258),
    [sym_line_comment] = STATE(258),
    [sym_block_comment] = STATE(258),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(259)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5583),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(259),
    [sym_line_comment] = STATE(259),
    [sym_block_comment] = STATE(259),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(260)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5597),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(260),
    [sym_line_comment] = STATE(260),
    [sym_block_comment] = STATE(260),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(261)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5616),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(261),
    [sym_line_comment] = STATE(261),
    [sym_block_comment] = STATE(261),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(262)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5631),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(262),
    [sym_line_comment] = STATE(262),
    [sym_block_comment] = STATE(262),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(263)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5641),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(263),
    [sym_line_comment] = STATE(263),
    [sym_block_comment] = STATE(263),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(264)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5657),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(264),
    [sym_line_comment] = STATE(264),
    [sym_block_comment] = STATE(264),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(265)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5670),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(265),
    [sym_line_comment] = STATE(265),
    [sym_block_comment] = STATE(265),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(266)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5681),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(266),
    [sym_line_comment] = STATE(266),
    [sym_block_comment] = STATE(266),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(267)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5693),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(267),
    [sym_line_comment] = STATE(267),
    [sym_block_comment] = STATE(267),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(268)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5718),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(268),
    [sym_line_comment] = STATE(268),
    [sym_block_comment] = STATE(268),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(269)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(269),
    [sym_selected_waveforms] = STATE(7625),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(269),
    [sym_block_comment] = STATE(269),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(270)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5734),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(270),
    [sym_line_comment] = STATE(270),
    [sym_block_comment] = STATE(270),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(271)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5884),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(271),
    [sym_line_comment] = STATE(271),
    [sym_block_comment] = STATE(271),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(272)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5754),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(272),
    [sym_line_comment] = STATE(272),
    [sym_block_comment] = STATE(272),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(273)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5771),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(273),
    [sym_line_comment] = STATE(273),
    [sym_block_comment] = STATE(273),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(274)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5784),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(274),
    [sym_line_comment] = STATE(274),
    [sym_block_comment] = STATE(274),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(275)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5794),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(275),
    [sym_line_comment] = STATE(275),
    [sym_block_comment] = STATE(275),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(276)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(276),
    [sym_selected_waveforms] = STATE(7727),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(276),
    [sym_block_comment] = STATE(276),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(277)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5806),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(277),
    [sym_line_comment] = STATE(277),
    [sym_block_comment] = STATE(277),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(278)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5820),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(278),
    [sym_line_comment] = STATE(278),
    [sym_block_comment] = STATE(278),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(279)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5842),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(279),
    [sym_line_comment] = STATE(279),
    [sym_block_comment] = STATE(279),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(280)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5869),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(280),
    [sym_line_comment] = STATE(280),
    [sym_block_comment] = STATE(280),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(281)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5915),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(281),
    [sym_line_comment] = STATE(281),
    [sym_block_comment] = STATE(281),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(282)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5559),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(282),
    [sym_line_comment] = STATE(282),
    [sym_block_comment] = STATE(282),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(283)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5564),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(283),
    [sym_line_comment] = STATE(283),
    [sym_block_comment] = STATE(283),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(284)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5570),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(284),
    [sym_line_comment] = STATE(284),
    [sym_block_comment] = STATE(284),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(285)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5576),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(285),
    [sym_line_comment] = STATE(285),
    [sym_block_comment] = STATE(285),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(286)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5580),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(286),
    [sym_line_comment] = STATE(286),
    [sym_block_comment] = STATE(286),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(287)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(287),
    [sym_selected_waveforms] = STATE(8017),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(287),
    [sym_block_comment] = STATE(287),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(288)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5585),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(288),
    [sym_line_comment] = STATE(288),
    [sym_block_comment] = STATE(288),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(289)] = {
    [sym__expression] = STATE(2275),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(289),
    [sym_force_mode] = STATE(388),
    [sym_selected_expressions] = STATE(8019),
    [sym_line_comment] = STATE(289),
    [sym_block_comment] = STATE(289),
    [aux_sym_selected_expressions_repeat1] = STATE(405),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(806),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(806),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(290)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5590),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(290),
    [sym_line_comment] = STATE(290),
    [sym_block_comment] = STATE(290),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(291)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(291),
    [sym_selected_waveforms] = STATE(8020),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(291),
    [sym_block_comment] = STATE(291),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(292)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5594),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(292),
    [sym_line_comment] = STATE(292),
    [sym_block_comment] = STATE(292),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(293)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5598),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(293),
    [sym_line_comment] = STATE(293),
    [sym_block_comment] = STATE(293),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(294)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5603),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(294),
    [sym_line_comment] = STATE(294),
    [sym_block_comment] = STATE(294),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(295)] = {
    [sym__expression] = STATE(1050),
    [sym_condition_expression] = STATE(1944),
    [sym_logical_expression] = STATE(1944),
    [sym_relational_expression] = STATE(1944),
    [sym_shift_expression] = STATE(1944),
    [sym_simple_expression] = STATE(2021),
    [sym_unary_operator] = STATE(720),
    [sym_sign] = STATE(439),
    [sym__primary] = STATE(1620),
    [sym_name] = STATE(1621),
    [sym__direct_name] = STATE(721),
    [sym__identifier] = STATE(838),
    [sym__external_name] = STATE(838),
    [sym_external_constant_name] = STATE(840),
    [sym_external_signal_name] = STATE(840),
    [sym_external_variable_name] = STATE(840),
    [sym__literal] = STATE(1622),
    [sym_bit_string_literal] = STATE(1612),
    [sym__abstract_literal] = STATE(846),
    [sym_based_literal] = STATE(902),
    [sym_allocator] = STATE(1621),
    [sym_parenthesis_expression] = STATE(1621),
    [sym__element] = STATE(5340),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(295),
    [sym_line_comment] = STATE(295),
    [sym_block_comment] = STATE(295),
    [anon_sym_LPAREN] = ACTIONS(760),
    [sym_condition_conversion] = ACTIONS(762),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(764),
    [sym_identifier] = ACTIONS(808),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(768),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(770),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(772),
    [sym_decimal_float] = ACTIONS(772),
    [sym_based_base] = ACTIONS(774),
    [sym_character_literal] = ACTIONS(776),
    [sym_string_literal] = ACTIONS(770),
    [sym_string_literal_std_logic] = ACTIONS(770),
    [sym_bit_string_length] = ACTIONS(778),
    [sym_bit_string_base] = ACTIONS(780),
    [sym_operator_symbol] = ACTIONS(776),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(808),
    [sym_library_constant_boolean] = ACTIONS(770),
    [sym_library_constant_character] = ACTIONS(770),
    [sym_library_constant_debug] = ACTIONS(770),
    [sym_library_constant_env] = ACTIONS(770),
    [sym_library_constant_standard] = ACTIONS(770),
    [sym_library_constant_std_logic] = ACTIONS(776),
    [sym_library_function] = ACTIONS(808),
    [sym_library_type] = ACTIONS(808),
  },
  [STATE(296)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5609),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(296),
    [sym_line_comment] = STATE(296),
    [sym_block_comment] = STATE(296),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(297)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(297),
    [sym_selected_waveforms] = STATE(7827),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(297),
    [sym_block_comment] = STATE(297),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(298)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5614),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(298),
    [sym_line_comment] = STATE(298),
    [sym_block_comment] = STATE(298),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(299)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5619),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(299),
    [sym_line_comment] = STATE(299),
    [sym_block_comment] = STATE(299),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(300)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(300),
    [sym_selected_waveforms] = STATE(6893),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(300),
    [sym_block_comment] = STATE(300),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(301)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5623),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(301),
    [sym_line_comment] = STATE(301),
    [sym_block_comment] = STATE(301),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(302)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5628),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(302),
    [sym_line_comment] = STATE(302),
    [sym_block_comment] = STATE(302),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(303)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5633),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(303),
    [sym_line_comment] = STATE(303),
    [sym_block_comment] = STATE(303),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(304)] = {
    [sym__expression] = STATE(2275),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(304),
    [sym_force_mode] = STATE(369),
    [sym_selected_expressions] = STATE(8041),
    [sym_line_comment] = STATE(304),
    [sym_block_comment] = STATE(304),
    [aux_sym_selected_expressions_repeat1] = STATE(405),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(806),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(806),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(305)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5638),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(305),
    [sym_line_comment] = STATE(305),
    [sym_block_comment] = STATE(305),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(306)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(306),
    [sym_selected_waveforms] = STATE(8042),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(306),
    [sym_block_comment] = STATE(306),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(307)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5642),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(307),
    [sym_line_comment] = STATE(307),
    [sym_block_comment] = STATE(307),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(308)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5646),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(308),
    [sym_line_comment] = STATE(308),
    [sym_block_comment] = STATE(308),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(309)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(309),
    [sym_selected_waveforms] = STATE(8045),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(309),
    [sym_block_comment] = STATE(309),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(310)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5651),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(310),
    [sym_line_comment] = STATE(310),
    [sym_block_comment] = STATE(310),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(311)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5655),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(311),
    [sym_line_comment] = STATE(311),
    [sym_block_comment] = STATE(311),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(312)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5659),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(312),
    [sym_line_comment] = STATE(312),
    [sym_block_comment] = STATE(312),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(313)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5661),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(313),
    [sym_line_comment] = STATE(313),
    [sym_block_comment] = STATE(313),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(314)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5753),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(314),
    [sym_line_comment] = STATE(314),
    [sym_block_comment] = STATE(314),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(315)] = {
    [sym__expression] = STATE(2275),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(315),
    [sym_force_mode] = STATE(389),
    [sym_selected_expressions] = STATE(8328),
    [sym_line_comment] = STATE(315),
    [sym_block_comment] = STATE(315),
    [aux_sym_selected_expressions_repeat1] = STATE(405),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(806),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(806),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(316)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5665),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(316),
    [sym_line_comment] = STATE(316),
    [sym_block_comment] = STATE(316),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(317)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(317),
    [sym_selected_waveforms] = STATE(8337),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(317),
    [sym_block_comment] = STATE(317),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(318)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(318),
    [sym_selected_waveforms] = STATE(8178),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(318),
    [sym_block_comment] = STATE(318),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(319)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(5943),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(319),
    [sym_line_comment] = STATE(319),
    [sym_block_comment] = STATE(319),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(320)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(320),
    [sym_selected_waveforms] = STATE(8222),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(320),
    [sym_block_comment] = STATE(320),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(321)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2504),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym_association_element] = STATE(6534),
    [sym__actual_part] = STATE(6171),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(321),
    [sym_line_comment] = STATE(321),
    [sym_block_comment] = STATE(321),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(322)] = {
    [sym__expression] = STATE(1050),
    [sym_condition_expression] = STATE(1944),
    [sym_logical_expression] = STATE(1944),
    [sym_relational_expression] = STATE(1944),
    [sym_shift_expression] = STATE(1944),
    [sym_simple_expression] = STATE(2021),
    [sym_unary_operator] = STATE(720),
    [sym_sign] = STATE(439),
    [sym__primary] = STATE(1620),
    [sym_name] = STATE(1621),
    [sym__direct_name] = STATE(721),
    [sym__identifier] = STATE(838),
    [sym__external_name] = STATE(838),
    [sym_external_constant_name] = STATE(840),
    [sym_external_signal_name] = STATE(840),
    [sym_external_variable_name] = STATE(840),
    [sym__literal] = STATE(1622),
    [sym_bit_string_literal] = STATE(1612),
    [sym__abstract_literal] = STATE(846),
    [sym_based_literal] = STATE(902),
    [sym_allocator] = STATE(1621),
    [sym_parenthesis_expression] = STATE(1621),
    [sym__element] = STATE(5787),
    [sym_choices] = STATE(5482),
    [sym__range] = STATE(5482),
    [sym_simple_range] = STATE(5022),
    [sym__tool_directive] = STATE(322),
    [sym_line_comment] = STATE(322),
    [sym_block_comment] = STATE(322),
    [anon_sym_LPAREN] = ACTIONS(760),
    [sym_condition_conversion] = ACTIONS(762),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(764),
    [sym_identifier] = ACTIONS(808),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(768),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(770),
    [sym_OR] = ACTIONS(273),
    [sym_OTHERS] = ACTIONS(279),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(772),
    [sym_decimal_float] = ACTIONS(772),
    [sym_based_base] = ACTIONS(774),
    [sym_character_literal] = ACTIONS(776),
    [sym_string_literal] = ACTIONS(770),
    [sym_string_literal_std_logic] = ACTIONS(770),
    [sym_bit_string_length] = ACTIONS(778),
    [sym_bit_string_base] = ACTIONS(780),
    [sym_operator_symbol] = ACTIONS(776),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(808),
    [sym_library_constant_boolean] = ACTIONS(770),
    [sym_library_constant_character] = ACTIONS(770),
    [sym_library_constant_debug] = ACTIONS(770),
    [sym_library_constant_env] = ACTIONS(770),
    [sym_library_constant_standard] = ACTIONS(770),
    [sym_library_constant_std_logic] = ACTIONS(776),
    [sym_library_function] = ACTIONS(808),
    [sym_library_type] = ACTIONS(808),
  },
  [STATE(323)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(323),
    [sym_selected_waveforms] = STATE(8253),
    [sym_waveform] = STATE(6286),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(323),
    [sym_block_comment] = STATE(323),
    [aux_sym_selected_waveforms_repeat1] = STATE(339),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(324)] = {
    [sym__expression] = STATE(2275),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(324),
    [sym_force_mode] = STATE(368),
    [sym_selected_expressions] = STATE(8255),
    [sym_line_comment] = STATE(324),
    [sym_block_comment] = STATE(324),
    [aux_sym_selected_expressions_repeat1] = STATE(405),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(806),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(806),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(325)] = {
    [sym__expression] = STATE(2275),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(325),
    [sym_force_mode] = STATE(380),
    [sym_selected_expressions] = STATE(8268),
    [sym_line_comment] = STATE(325),
    [sym_block_comment] = STATE(325),
    [aux_sym_selected_expressions_repeat1] = STATE(405),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_IN] = ACTIONS(806),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_OUT] = ACTIONS(806),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(326)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(326),
    [sym_waveform] = STATE(5891),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7736),
    [sym_line_comment] = STATE(326),
    [sym_block_comment] = STATE(326),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(327)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(327),
    [sym_waveform] = STATE(5897),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7777),
    [sym_line_comment] = STATE(327),
    [sym_block_comment] = STATE(327),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(328)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(328),
    [sym_waveform] = STATE(5703),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(6900),
    [sym_line_comment] = STATE(328),
    [sym_block_comment] = STATE(328),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(329)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(329),
    [sym_waveform] = STATE(5578),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7523),
    [sym_line_comment] = STATE(329),
    [sym_block_comment] = STATE(329),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(330)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(330),
    [sym_waveform] = STATE(5954),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8267),
    [sym_line_comment] = STATE(330),
    [sym_block_comment] = STATE(330),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(331)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(331),
    [sym_waveform] = STATE(5809),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7303),
    [sym_line_comment] = STATE(331),
    [sym_block_comment] = STATE(331),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(332)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(332),
    [sym_waveform] = STATE(5640),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(6934),
    [sym_line_comment] = STATE(332),
    [sym_block_comment] = STATE(332),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(333)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(333),
    [sym_waveform] = STATE(5856),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7529),
    [sym_line_comment] = STATE(333),
    [sym_block_comment] = STATE(333),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(334)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(334),
    [sym_waveform] = STATE(5905),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8029),
    [sym_line_comment] = STATE(334),
    [sym_block_comment] = STATE(334),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(335)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(335),
    [sym_waveform] = STATE(5860),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7556),
    [sym_line_comment] = STATE(335),
    [sym_block_comment] = STATE(335),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(336)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(336),
    [sym_waveform] = STATE(5906),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8035),
    [sym_line_comment] = STATE(336),
    [sym_block_comment] = STATE(336),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(337)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(337),
    [sym_waveform] = STATE(5873),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7628),
    [sym_line_comment] = STATE(337),
    [sym_block_comment] = STATE(337),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(338)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(338),
    [sym_waveform] = STATE(5781),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7559),
    [sym_line_comment] = STATE(338),
    [sym_block_comment] = STATE(338),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(339)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(339),
    [sym_waveform] = STATE(6163),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(339),
    [sym_block_comment] = STATE(339),
    [aux_sym_selected_waveforms_repeat1] = STATE(343),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(340)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(340),
    [sym_waveform] = STATE(5836),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7413),
    [sym_line_comment] = STATE(340),
    [sym_block_comment] = STATE(340),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(341)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(341),
    [sym_waveform] = STATE(5839),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8324),
    [sym_line_comment] = STATE(341),
    [sym_block_comment] = STATE(341),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(342)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(342),
    [sym_waveform] = STATE(5921),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8130),
    [sym_line_comment] = STATE(342),
    [sym_block_comment] = STATE(342),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(343)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(343),
    [sym_waveform] = STATE(6316),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(343),
    [sym_block_comment] = STATE(343),
    [aux_sym_selected_waveforms_repeat1] = STATE(343),
    [anon_sym_LPAREN] = ACTIONS(810),
    [sym_condition_conversion] = ACTIONS(813),
    [anon_sym_PLUS] = ACTIONS(816),
    [anon_sym_DASH] = ACTIONS(816),
    [anon_sym_LT_LT] = ACTIONS(819),
    [sym_identifier] = ACTIONS(822),
    [sym_ABS] = ACTIONS(825),
    [sym_AND] = ACTIONS(825),
    [sym_NAND] = ACTIONS(825),
    [sym_NEW] = ACTIONS(828),
    [sym_NOR] = ACTIONS(825),
    [sym_NOT] = ACTIONS(825),
    [sym_NULL] = ACTIONS(831),
    [sym_OR] = ACTIONS(825),
    [sym_UNAFFECTED] = ACTIONS(834),
    [sym_XNOR] = ACTIONS(825),
    [sym_XOR] = ACTIONS(825),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(837),
    [sym_decimal_float] = ACTIONS(837),
    [sym_based_base] = ACTIONS(840),
    [sym_character_literal] = ACTIONS(843),
    [sym_string_literal] = ACTIONS(831),
    [sym_string_literal_std_logic] = ACTIONS(831),
    [sym_bit_string_length] = ACTIONS(846),
    [sym_bit_string_base] = ACTIONS(849),
    [sym_operator_symbol] = ACTIONS(843),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(822),
    [sym_library_constant_boolean] = ACTIONS(831),
    [sym_library_constant_character] = ACTIONS(831),
    [sym_library_constant_debug] = ACTIONS(831),
    [sym_library_constant_env] = ACTIONS(831),
    [sym_library_constant_standard] = ACTIONS(831),
    [sym_library_constant_std_logic] = ACTIONS(843),
    [sym_library_function] = ACTIONS(822),
    [sym_library_type] = ACTIONS(822),
  },
  [STATE(344)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(344),
    [sym_waveform] = STATE(5934),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8180),
    [sym_line_comment] = STATE(344),
    [sym_block_comment] = STATE(344),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(345)] = {
    [sym_conditional_expression] = STATE(6048),
    [sym__expression] = STATE(1201),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__actual_part] = STATE(6167),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(345),
    [sym_line_comment] = STATE(345),
    [sym_block_comment] = STATE(345),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_INERTIAL] = ACTIONS(405),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OPEN] = ACTIONS(407),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(346)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(346),
    [sym_waveform] = STATE(5687),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(6727),
    [sym_line_comment] = STATE(346),
    [sym_block_comment] = STATE(346),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(347)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(347),
    [sym_waveform] = STATE(5946),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8224),
    [sym_line_comment] = STATE(347),
    [sym_block_comment] = STATE(347),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(348)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(348),
    [sym_waveform] = STATE(5948),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8234),
    [sym_line_comment] = STATE(348),
    [sym_block_comment] = STATE(348),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(349)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(349),
    [sym_waveform] = STATE(5799),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(7036),
    [sym_line_comment] = STATE(349),
    [sym_block_comment] = STATE(349),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(350)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(350),
    [sym_waveform] = STATE(5953),
    [sym_waveform_element] = STATE(5397),
    [sym_conditional_waveforms] = STATE(8263),
    [sym_line_comment] = STATE(350),
    [sym_block_comment] = STATE(350),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(351)] = {
    [sym__expression] = STATE(1519),
    [sym_condition_expression] = STATE(2536),
    [sym_logical_expression] = STATE(2536),
    [sym_relational_expression] = STATE(2536),
    [sym_shift_expression] = STATE(2536),
    [sym_simple_expression] = STATE(2536),
    [sym_unary_operator] = STATE(704),
    [sym_sign] = STATE(705),
    [sym__primary] = STATE(2508),
    [sym_name] = STATE(2509),
    [sym__direct_name] = STATE(777),
    [sym__identifier] = STATE(1411),
    [sym__external_name] = STATE(1411),
    [sym_external_constant_name] = STATE(1412),
    [sym_external_signal_name] = STATE(1412),
    [sym_external_variable_name] = STATE(1412),
    [sym__literal] = STATE(2510),
    [sym_bit_string_literal] = STATE(2506),
    [sym__abstract_literal] = STATE(1408),
    [sym_based_literal] = STATE(1627),
    [sym_allocator] = STATE(2509),
    [sym_parenthesis_expression] = STATE(2509),
    [sym__tool_directive] = STATE(351),
    [sym_waveform] = STATE(6129),
    [sym_waveform_element] = STATE(5397),
    [sym_line_comment] = STATE(351),
    [sym_block_comment] = STATE(351),
    [anon_sym_LPAREN] = ACTIONS(411),
    [sym_condition_conversion] = ACTIONS(413),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(415),
    [sym_identifier] = ACTIONS(417),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(423),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(425),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(429),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(431),
    [sym_decimal_float] = ACTIONS(431),
    [sym_based_base] = ACTIONS(433),
    [sym_character_literal] = ACTIONS(435),
    [sym_string_literal] = ACTIONS(425),
    [sym_string_literal_std_logic] = ACTIONS(425),
    [sym_bit_string_length] = ACTIONS(437),
    [sym_bit_string_base] = ACTIONS(439),
    [sym_operator_symbol] = ACTIONS(435),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(417),
    [sym_library_constant_boolean] = ACTIONS(425),
    [sym_library_constant_character] = ACTIONS(425),
    [sym_library_constant_debug] = ACTIONS(425),
    [sym_library_constant_env] = ACTIONS(425),
    [sym_library_constant_standard] = ACTIONS(425),
    [sym_library_constant_std_logic] = ACTIONS(435),
    [sym_library_function] = ACTIONS(417),
    [sym_library_type] = ACTIONS(417),
  },
  [STATE(352)] = {
    [sym_index_constraint] = STATE(6794),
    [sym__expression] = STATE(2048),
    [sym_condition_expression] = STATE(2621),
    [sym_logical_expression] = STATE(2621),
    [sym_relational_expression] = STATE(2621),
    [sym_shift_expression] = STATE(2621),
    [sym_simple_expression] = STATE(2463),
    [sym_unary_operator] = STATE(453),
    [sym_sign] = STATE(454),
    [sym__primary] = STATE(2535),
    [sym_name] = STATE(2507),
    [sym__direct_name] = STATE(772),
    [sym__identifier] = STATE(1336),
    [sym__external_name] = STATE(1336),
    [sym_external_constant_name] = STATE(1337),
    [sym_external_signal_name] = STATE(1337),
    [sym_external_variable_name] = STATE(1337),
    [sym__literal] = STATE(2480),
    [sym_bit_string_literal] = STATE(2530),
    [sym__abstract_literal] = STATE(1407),
    [sym_based_literal] = STATE(1615),
    [sym_allocator] = STATE(2507),
    [sym_parenthesis_expression] = STATE(2507),
    [sym__range] = STATE(5739),
    [sym_simple_range] = STATE(6038),
    [sym__tool_directive] = STATE(352),
    [sym_line_comment] = STATE(352),
    [sym_block_comment] = STATE(352),
    [anon_sym_LPAREN] = ACTIONS(852),
    [sym_condition_conversion] = ACTIONS(854),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(856),
    [sym_identifier] = ACTIONS(858),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(860),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(862),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(864),
    [sym_decimal_float] = ACTIONS(864),
    [sym_based_base] = ACTIONS(866),
    [sym_character_literal] = ACTIONS(868),
    [sym_string_literal] = ACTIONS(862),
    [sym_string_literal_std_logic] = ACTIONS(862),
    [sym_bit_string_length] = ACTIONS(870),
    [sym_bit_string_base] = ACTIONS(872),
    [sym_operator_symbol] = ACTIONS(868),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(858),
    [sym_library_constant_boolean] = ACTIONS(862),
    [sym_library_constant_character] = ACTIONS(862),
    [sym_library_constant_debug] = ACTIONS(862),
    [sym_library_constant_env] = ACTIONS(862),
    [sym_library_constant_standard] = ACTIONS(862),
    [sym_library_constant_std_logic] = ACTIONS(868),
    [sym_library_function] = ACTIONS(858),
    [sym_library_type] = ACTIONS(858),
  },
  [STATE(353)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(353),
    [sym_conditional_or_unaffected_expression] = STATE(7617),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(353),
    [sym_block_comment] = STATE(353),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(354)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(354),
    [sym_conditional_or_unaffected_expression] = STATE(8257),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(354),
    [sym_block_comment] = STATE(354),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(355)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(355),
    [sym_conditional_or_unaffected_expression] = STATE(7807),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(355),
    [sym_block_comment] = STATE(355),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(356)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(356),
    [sym_conditional_or_unaffected_expression] = STATE(6959),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(356),
    [sym_block_comment] = STATE(356),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(357)] = {
    [sym_index_constraint] = STATE(7016),
    [sym__expression] = STATE(2048),
    [sym_condition_expression] = STATE(2621),
    [sym_logical_expression] = STATE(2621),
    [sym_relational_expression] = STATE(2621),
    [sym_shift_expression] = STATE(2621),
    [sym_simple_expression] = STATE(2463),
    [sym_unary_operator] = STATE(453),
    [sym_sign] = STATE(454),
    [sym__primary] = STATE(2535),
    [sym_name] = STATE(2507),
    [sym__direct_name] = STATE(772),
    [sym__identifier] = STATE(1336),
    [sym__external_name] = STATE(1336),
    [sym_external_constant_name] = STATE(1337),
    [sym_external_signal_name] = STATE(1337),
    [sym_external_variable_name] = STATE(1337),
    [sym__literal] = STATE(2480),
    [sym_bit_string_literal] = STATE(2530),
    [sym__abstract_literal] = STATE(1407),
    [sym_based_literal] = STATE(1615),
    [sym_allocator] = STATE(2507),
    [sym_parenthesis_expression] = STATE(2507),
    [sym__range] = STATE(5739),
    [sym_simple_range] = STATE(6038),
    [sym__tool_directive] = STATE(357),
    [sym_line_comment] = STATE(357),
    [sym_block_comment] = STATE(357),
    [anon_sym_LPAREN] = ACTIONS(874),
    [sym_condition_conversion] = ACTIONS(854),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(856),
    [sym_identifier] = ACTIONS(858),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(860),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(862),
    [sym_OR] = ACTIONS(273),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(864),
    [sym_decimal_float] = ACTIONS(864),
    [sym_based_base] = ACTIONS(866),
    [sym_character_literal] = ACTIONS(868),
    [sym_string_literal] = ACTIONS(862),
    [sym_string_literal_std_logic] = ACTIONS(862),
    [sym_bit_string_length] = ACTIONS(870),
    [sym_bit_string_base] = ACTIONS(872),
    [sym_operator_symbol] = ACTIONS(868),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(858),
    [sym_library_constant_boolean] = ACTIONS(862),
    [sym_library_constant_character] = ACTIONS(862),
    [sym_library_constant_debug] = ACTIONS(862),
    [sym_library_constant_env] = ACTIONS(862),
    [sym_library_constant_standard] = ACTIONS(862),
    [sym_library_constant_std_logic] = ACTIONS(868),
    [sym_library_function] = ACTIONS(858),
    [sym_library_type] = ACTIONS(858),
  },
  [STATE(358)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(358),
    [sym_conditional_or_unaffected_expression] = STATE(7637),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(358),
    [sym_block_comment] = STATE(358),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(359)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(359),
    [sym_conditional_or_unaffected_expression] = STATE(8043),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(359),
    [sym_block_comment] = STATE(359),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(360)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(360),
    [sym_conditional_or_unaffected_expression] = STATE(8275),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(360),
    [sym_block_comment] = STATE(360),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(361)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(361),
    [sym_conditional_or_unaffected_expression] = STATE(8342),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(361),
    [sym_block_comment] = STATE(361),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(362)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(362),
    [sym_conditional_or_unaffected_expression] = STATE(6608),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(362),
    [sym_block_comment] = STATE(362),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(363)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(363),
    [sym_conditional_or_unaffected_expression] = STATE(8240),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(363),
    [sym_block_comment] = STATE(363),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(364)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(364),
    [sym_conditional_or_unaffected_expression] = STATE(8021),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(364),
    [sym_block_comment] = STATE(364),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
  [STATE(365)] = {
    [sym__expression] = STATE(2284),
    [sym_condition_expression] = STATE(2313),
    [sym_logical_expression] = STATE(2313),
    [sym_relational_expression] = STATE(2313),
    [sym_shift_expression] = STATE(2313),
    [sym_simple_expression] = STATE(2313),
    [sym_unary_operator] = STATE(695),
    [sym_sign] = STATE(696),
    [sym__primary] = STATE(2383),
    [sym_name] = STATE(2384),
    [sym__direct_name] = STATE(748),
    [sym__identifier] = STATE(1077),
    [sym__external_name] = STATE(1077),
    [sym_external_constant_name] = STATE(1078),
    [sym_external_signal_name] = STATE(1078),
    [sym_external_variable_name] = STATE(1078),
    [sym__literal] = STATE(2385),
    [sym_bit_string_literal] = STATE(2380),
    [sym__abstract_literal] = STATE(1056),
    [sym_based_literal] = STATE(1229),
    [sym_allocator] = STATE(2384),
    [sym_parenthesis_expression] = STATE(2384),
    [sym__tool_directive] = STATE(365),
    [sym_conditional_or_unaffected_expression] = STATE(7742),
    [sym__expression_or_unaffected] = STATE(5331),
    [sym_line_comment] = STATE(365),
    [sym_block_comment] = STATE(365),
    [anon_sym_LPAREN] = ACTIONS(736),
    [sym_condition_conversion] = ACTIONS(265),
    [anon_sym_PLUS] = ACTIONS(267),
    [anon_sym_DASH] = ACTIONS(267),
    [anon_sym_LT_LT] = ACTIONS(738),
    [sym_identifier] = ACTIONS(740),
    [sym_ABS] = ACTIONS(273),
    [sym_AND] = ACTIONS(273),
    [sym_NAND] = ACTIONS(273),
    [sym_NEW] = ACTIONS(744),
    [sym_NOR] = ACTIONS(273),
    [sym_NOT] = ACTIONS(273),
    [sym_NULL] = ACTIONS(746),
    [sym_OR] = ACTIONS(273),
    [sym_UNAFFECTED] = ACTIONS(790),
    [sym_XNOR] = ACTIONS(273),
    [sym_XOR] = ACTIONS(273),
    [sym__grave_accent] = ACTIONS(3),
    [sym_decimal_integer] = ACTIONS(750),
    [sym_decimal_float] = ACTIONS(750),
    [sym_based_base] = ACTIONS(752),
    [sym_character_literal] = ACTIONS(754),
    [sym_string_literal] = ACTIONS(746),
    [sym_string_literal_std_logic] = ACTIONS(746),
    [sym_bit_string_length] = ACTIONS(756),
    [sym_bit_string_base] = ACTIONS(758),
    [sym_operator_symbol] = ACTIONS(754),
    [sym__line_comment_start] = ACTIONS(5),
    [sym__block_comment_start] = ACTIONS(7),
    [sym_library_constant] = ACTIONS(740),
    [sym_library_constant_boolean] = ACTIONS(746),
    [sym_library_constant_character] = ACTIONS(746),
    [sym_library_constant_debug] = ACTIONS(746),
    [sym_library_constant_env] = ACTIONS(746),
    [sym_library_constant_standard] = ACTIONS(746),
    [sym_library_constant_std_logic] = ACTIONS(754),
    [sym_library_function] = ACTIONS(740),
    [sym_library_type] = ACTIONS(740),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(876), 1,
      anon_sym_LPAREN,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    ACTIONS(884), 1,
      sym_NEW,
    ACTIONS(890), 1,
      sym_based_base,
    ACTIONS(894), 1,
      sym_bit_string_length,
    ACTIONS(896), 1,
      sym_bit_string_base,
    STATE(463), 1,
      sym_unary_operator,
    STATE(464), 1,
      sym_sign,
    STATE(786), 1,
      sym__direct_name,
    STATE(1609), 1,
      sym__abstract_literal,
    STATE(1675), 1,
      sym_based_literal,
    STATE(2407), 1,
      sym__expression,
    STATE(2489), 1,
      sym_simple_expression,
    STATE(2559), 1,
      sym_bit_string_literal,
    STATE(2561), 1,
      sym__primary,
    STATE(2563), 1,
      sym__literal,
    STATE(7159), 1,
      sym__range,
    STATE(7268), 1,
      sym_simple_range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(888), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(366), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2562), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(886), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [131] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(8025), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(367), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [260] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(8268), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(368), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [389] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(8048), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(369), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [518] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      anon_sym_LPAREN,
    ACTIONS(900), 1,
      sym_condition_conversion,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    ACTIONS(906), 1,
      sym_NEW,
    ACTIONS(912), 1,
      sym_based_base,
    ACTIONS(916), 1,
      sym_bit_string_length,
    ACTIONS(918), 1,
      sym_bit_string_base,
    STATE(483), 1,
      sym_unary_operator,
    STATE(484), 1,
      sym_sign,
    STATE(743), 1,
      sym__direct_name,
    STATE(1060), 1,
      sym__abstract_literal,
    STATE(1236), 1,
      sym_based_literal,
    STATE(1613), 1,
      sym__expression,
    STATE(1989), 1,
      sym_simple_expression,
    STATE(2366), 1,
      sym_bit_string_literal,
    STATE(2367), 1,
      sym__primary,
    STATE(2369), 1,
      sym__literal,
    STATE(5447), 1,
      sym__range,
    STATE(5497), 1,
      sym_simple_range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(910), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(370), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2368), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2491), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(908), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [649] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(662), 1,
      sym_label_declaration,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(1916), 1,
      sym__expression,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    STATE(8058), 1,
      sym__label,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(371), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(924), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [778] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    ACTIONS(790), 1,
      sym_UNAFFECTED,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2284), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(6196), 1,
      sym__expression_or_unaffected,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(372), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [907] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(6718), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(373), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [1036] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(942), 1,
      sym_condition_conversion,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    ACTIONS(948), 1,
      sym_NEW,
    ACTIONS(954), 1,
      sym_based_base,
    ACTIONS(958), 1,
      sym_bit_string_length,
    ACTIONS(960), 1,
      sym_bit_string_base,
    STATE(458), 1,
      sym_unary_operator,
    STATE(459), 1,
      sym_sign,
    STATE(779), 1,
      sym__direct_name,
    STATE(1413), 1,
      sym__abstract_literal,
    STATE(1642), 1,
      sym_based_literal,
    STATE(2184), 1,
      sym__expression,
    STATE(2443), 1,
      sym_simple_expression,
    STATE(2496), 1,
      sym_bit_string_literal,
    STATE(2497), 1,
      sym__primary,
    STATE(2501), 1,
      sym__literal,
    STATE(6149), 1,
      sym__range,
    STATE(6287), 1,
      sym_simple_range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(952), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(374), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2500), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2681), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(950), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [1167] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(962), 1,
      anon_sym_LPAREN,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    ACTIONS(970), 1,
      sym_NEW,
    ACTIONS(976), 1,
      sym_based_base,
    ACTIONS(980), 1,
      sym_bit_string_length,
    ACTIONS(982), 1,
      sym_bit_string_base,
    STATE(468), 1,
      sym_unary_operator,
    STATE(469), 1,
      sym_sign,
    STATE(793), 1,
      sym__direct_name,
    STATE(1497), 1,
      sym__abstract_literal,
    STATE(1680), 1,
      sym_based_literal,
    STATE(2418), 1,
      sym__expression,
    STATE(2520), 1,
      sym_simple_expression,
    STATE(2554), 1,
      sym_bit_string_literal,
    STATE(2557), 1,
      sym__primary,
    STATE(2564), 1,
      sym__literal,
    STATE(7087), 1,
      sym__range,
    STATE(7170), 1,
      sym_simple_range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(974), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(375), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2558), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(972), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [1298] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(8260), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(376), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [1427] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    STATE(439), 1,
      sym_sign,
    STATE(720), 1,
      sym_unary_operator,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(902), 1,
      sym_based_literal,
    STATE(1050), 1,
      sym__expression,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    STATE(1837), 1,
      sym_simple_expression,
    STATE(5022), 1,
      sym_simple_range,
    STATE(5152), 1,
      sym__range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(377), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [1558] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(7756), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(378), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [1687] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(6708), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(379), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [1816] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(8282), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(380), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [1945] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(660), 1,
      sym_label_declaration,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2050), 1,
      sym__expression,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    STATE(8058), 1,
      sym__label,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(381), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(924), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [2074] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    STATE(439), 1,
      sym_sign,
    STATE(720), 1,
      sym_unary_operator,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(902), 1,
      sym_based_literal,
    STATE(1050), 1,
      sym__expression,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    STATE(1837), 1,
      sym_simple_expression,
    STATE(5022), 1,
      sym_simple_range,
    STATE(6238), 1,
      sym__range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(382), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [2205] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(7472), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(383), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [2334] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(8229), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(384), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [2463] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(854), 1,
      sym_condition_conversion,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    ACTIONS(860), 1,
      sym_NEW,
    ACTIONS(866), 1,
      sym_based_base,
    ACTIONS(870), 1,
      sym_bit_string_length,
    ACTIONS(872), 1,
      sym_bit_string_base,
    ACTIONS(984), 1,
      anon_sym_LPAREN,
    STATE(453), 1,
      sym_unary_operator,
    STATE(454), 1,
      sym_sign,
    STATE(772), 1,
      sym__direct_name,
    STATE(1407), 1,
      sym__abstract_literal,
    STATE(1615), 1,
      sym_based_literal,
    STATE(2048), 1,
      sym__expression,
    STATE(2463), 1,
      sym_simple_expression,
    STATE(2480), 1,
      sym__literal,
    STATE(2530), 1,
      sym_bit_string_literal,
    STATE(2535), 1,
      sym__primary,
    STATE(6038), 1,
      sym_simple_range,
    STATE(6356), 1,
      sym__range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(864), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(385), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2507), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2621), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(862), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [2594] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(986), 1,
      anon_sym_LPAREN,
    ACTIONS(988), 1,
      sym_condition_conversion,
    ACTIONS(990), 1,
      sym_NEW,
    ACTIONS(996), 1,
      sym_based_base,
    ACTIONS(998), 1,
      sym_bit_string_length,
    ACTIONS(1000), 1,
      sym_bit_string_base,
    STATE(499), 1,
      sym_unary_operator,
    STATE(500), 1,
      sym_sign,
    STATE(763), 1,
      sym__direct_name,
    STATE(1138), 1,
      sym__abstract_literal,
    STATE(1257), 1,
      sym_based_literal,
    STATE(1782), 1,
      sym__expression,
    STATE(2345), 1,
      sym_simple_expression,
    STATE(2452), 1,
      sym_bit_string_literal,
    STATE(2454), 1,
      sym__primary,
    STATE(2456), 1,
      sym__literal,
    STATE(5685), 1,
      sym__range,
    STATE(5757), 1,
      sym_simple_range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(994), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(386), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2455), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2565), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(992), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [2725] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1002), 1,
      anon_sym_LPAREN,
    ACTIONS(1004), 1,
      sym_condition_conversion,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    ACTIONS(1010), 1,
      sym_NEW,
    ACTIONS(1016), 1,
      sym_based_base,
    ACTIONS(1020), 1,
      sym_bit_string_length,
    ACTIONS(1022), 1,
      sym_bit_string_base,
    STATE(519), 1,
      sym_unary_operator,
    STATE(520), 1,
      sym_sign,
    STATE(746), 1,
      sym__direct_name,
    STATE(1071), 1,
      sym__abstract_literal,
    STATE(1154), 1,
      sym_based_literal,
    STATE(1556), 1,
      sym__expression,
    STATE(2015), 1,
      sym_simple_expression,
    STATE(2290), 1,
      sym__primary,
    STATE(2399), 1,
      sym__literal,
    STATE(2429), 1,
      sym_bit_string_literal,
    STATE(5489), 1,
      sym__range,
    STATE(5513), 1,
      sym_simple_range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1014), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(387), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2301), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2533), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1012), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [2856] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(8041), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(388), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [2985] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(405), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2275), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(7613), 1,
      sym_selected_expressions,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(389), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [3114] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1024), 1,
      anon_sym_LPAREN,
    ACTIONS(1026), 1,
      sym_condition_conversion,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    ACTIONS(1032), 1,
      sym_NEW,
    ACTIONS(1038), 1,
      sym_based_base,
    ACTIONS(1042), 1,
      sym_bit_string_length,
    ACTIONS(1044), 1,
      sym_bit_string_base,
    STATE(473), 1,
      sym_unary_operator,
    STATE(474), 1,
      sym_sign,
    STATE(750), 1,
      sym__direct_name,
    STATE(1140), 1,
      sym__abstract_literal,
    STATE(1326), 1,
      sym_based_literal,
    STATE(1844), 1,
      sym__expression,
    STATE(2315), 1,
      sym_simple_expression,
    STATE(2436), 1,
      sym_bit_string_literal,
    STATE(2438), 1,
      sym__primary,
    STATE(2440), 1,
      sym__literal,
    STATE(5675), 1,
      sym_simple_range,
    STATE(5707), 1,
      sym__range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1036), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(390), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2439), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2560), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1034), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [3245] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1046), 1,
      anon_sym_LPAREN,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    ACTIONS(1054), 1,
      sym_NEW,
    ACTIONS(1060), 1,
      sym_based_base,
    ACTIONS(1064), 1,
      sym_bit_string_length,
    ACTIONS(1066), 1,
      sym_bit_string_base,
    STATE(478), 1,
      sym_unary_operator,
    STATE(479), 1,
      sym_sign,
    STATE(788), 1,
      sym__direct_name,
    STATE(1526), 1,
      sym__abstract_literal,
    STATE(1709), 1,
      sym_based_literal,
    STATE(2332), 1,
      sym__expression,
    STATE(2486), 1,
      sym_simple_expression,
    STATE(2541), 1,
      sym__primary,
    STATE(2546), 1,
      sym__literal,
    STATE(2572), 1,
      sym_bit_string_literal,
    STATE(6886), 1,
      sym_simple_range,
    STATE(7102), 1,
      sym__range,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1058), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(391), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2543), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2738), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1056), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [3376] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2400), 1,
      sym__expression,
    STATE(7290), 1,
      sym_file_logical_name,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(392), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [3502] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(6873), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(393), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [3628] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(8212), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(394), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [3754] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2400), 1,
      sym__expression,
    STATE(7222), 1,
      sym_file_logical_name,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(395), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [3880] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1090), 1,
      anon_sym_QMARK,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2413), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(396), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [4006] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(6021), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(397), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [4132] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(6570), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(398), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [4258] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(7794), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(399), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [4384] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(6236), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(400), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [4510] = 38,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1110), 1,
      anon_sym_LPAREN,
    ACTIONS(1112), 1,
      sym_BEGIN,
    ACTIONS(1114), 1,
      sym_GENERIC,
    ACTIONS(1116), 1,
      sym_IS,
    ACTIONS(1118), 1,
      sym_PORT,
    STATE(722), 1,
      sym_guard_condition,
    STATE(737), 1,
      sym_generic_clause,
    STATE(757), 1,
      sym_port_clause,
    STATE(807), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(6018), 1,
      sym_concurrent_block,
    STATE(6020), 1,
      sym_block_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(401), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [4650] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1120), 1,
      anon_sym_LPAREN,
    ACTIONS(1123), 1,
      sym_condition_conversion,
    ACTIONS(1129), 1,
      anon_sym_LT_LT,
    ACTIONS(1138), 1,
      sym_NEW,
    ACTIONS(1147), 1,
      sym_based_base,
    ACTIONS(1153), 1,
      sym_bit_string_length,
    ACTIONS(1156), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(1973), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(1126), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1144), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1150), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1132), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(402), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_selected_expressions_repeat1,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(1135), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1141), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [4774] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(7218), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(403), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [4900] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(411), 1,
      anon_sym_LPAREN,
    ACTIONS(413), 1,
      sym_condition_conversion,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    ACTIONS(423), 1,
      sym_NEW,
    ACTIONS(433), 1,
      sym_based_base,
    ACTIONS(437), 1,
      sym_bit_string_length,
    ACTIONS(439), 1,
      sym_bit_string_base,
    STATE(704), 1,
      sym_unary_operator,
    STATE(705), 1,
      sym_sign,
    STATE(777), 1,
      sym__direct_name,
    STATE(1408), 1,
      sym__abstract_literal,
    STATE(1519), 1,
      sym__expression,
    STATE(1627), 1,
      sym_based_literal,
    STATE(2506), 1,
      sym_bit_string_literal,
    STATE(2508), 1,
      sym__primary,
    STATE(2510), 1,
      sym__literal,
    STATE(5766), 1,
      sym_waveform_element,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(431), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(404), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2509), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2536), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(425), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [5026] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(402), 1,
      aux_sym_selected_expressions_repeat1,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(1941), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(405), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [5152] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(7122), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(406), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [5278] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(7338), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(407), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [5404] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(7681), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(408), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [5530] = 38,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1110), 1,
      anon_sym_LPAREN,
    ACTIONS(1112), 1,
      sym_BEGIN,
    ACTIONS(1114), 1,
      sym_GENERIC,
    ACTIONS(1116), 1,
      sym_IS,
    ACTIONS(1118), 1,
      sym_PORT,
    STATE(726), 1,
      sym_guard_condition,
    STATE(737), 1,
      sym_generic_clause,
    STATE(757), 1,
      sym_port_clause,
    STATE(807), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(6387), 1,
      sym_block_head,
    STATE(6565), 1,
      sym_concurrent_block,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(409), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [5670] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1201), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    STATE(6710), 1,
      sym_conditional_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(410), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [5796] = 38,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1110), 1,
      anon_sym_LPAREN,
    ACTIONS(1112), 1,
      sym_BEGIN,
    ACTIONS(1114), 1,
      sym_GENERIC,
    ACTIONS(1116), 1,
      sym_IS,
    ACTIONS(1118), 1,
      sym_PORT,
    STATE(727), 1,
      sym_guard_condition,
    STATE(737), 1,
      sym_generic_clause,
    STATE(757), 1,
      sym_port_clause,
    STATE(807), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(6040), 1,
      sym_concurrent_block,
    STATE(6419), 1,
      sym_block_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(411), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [5936] = 38,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1110), 1,
      anon_sym_LPAREN,
    ACTIONS(1112), 1,
      sym_BEGIN,
    ACTIONS(1114), 1,
      sym_GENERIC,
    ACTIONS(1116), 1,
      sym_IS,
    ACTIONS(1118), 1,
      sym_PORT,
    STATE(724), 1,
      sym_guard_condition,
    STATE(737), 1,
      sym_generic_clause,
    STATE(757), 1,
      sym_port_clause,
    STATE(807), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(6103), 1,
      sym_concurrent_block,
    STATE(6447), 1,
      sym_block_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(412), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [6076] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1159), 1,
      anon_sym_LPAREN,
    ACTIONS(1161), 1,
      sym_condition_conversion,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    ACTIONS(1167), 1,
      sym_NEW,
    ACTIONS(1173), 1,
      sym_based_base,
    ACTIONS(1177), 1,
      sym_bit_string_length,
    ACTIONS(1179), 1,
      sym_bit_string_base,
    STATE(661), 1,
      sym_sign,
    STATE(780), 1,
      sym__direct_name,
    STATE(1531), 1,
      sym__abstract_literal,
    STATE(1730), 1,
      sym_based_literal,
    STATE(1751), 1,
      sym__expression,
    STATE(2567), 1,
      sym__primary,
    STATE(2610), 1,
      sym_bit_string_literal,
    STATE(2612), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1171), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2584), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(413), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2547), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1169), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [6197] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1010), 1,
      sym__expression,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(414), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [6318] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(998), 1,
      sym__expression,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(415), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [6441] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1059), 1,
      sym__expression,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(416), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [6564] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1064), 1,
      sym__expression,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(417), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [6687] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2310), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(418), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [6808] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2311), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(419), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [6929] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2320), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(420), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [7052] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2325), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(421), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [7175] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2327), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(422), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [7298] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(424), 1,
      sym_sign,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2343), 1,
      sym__expression,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(423), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [7419] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(423), 1,
      sym_unary_operator,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2344), 1,
      sym__expression,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(424), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [7540] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(423), 1,
      sym_unary_operator,
    STATE(424), 1,
      sym_sign,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2346), 1,
      sym__expression,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(425), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [7663] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(423), 1,
      sym_unary_operator,
    STATE(424), 1,
      sym_sign,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2350), 1,
      sym__expression,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(426), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [7786] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(423), 1,
      sym_unary_operator,
    STATE(424), 1,
      sym_sign,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2351), 1,
      sym__expression,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(427), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [7909] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(429), 1,
      sym_sign,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2359), 1,
      sym__expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(428), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2738), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [8030] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(428), 1,
      sym_unary_operator,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2360), 1,
      sym__expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(429), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2738), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [8151] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(428), 1,
      sym_unary_operator,
    STATE(429), 1,
      sym_sign,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2361), 1,
      sym__expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(430), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2738), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [8274] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(428), 1,
      sym_unary_operator,
    STATE(429), 1,
      sym_sign,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2365), 1,
      sym__expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(431), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2738), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [8397] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(428), 1,
      sym_unary_operator,
    STATE(429), 1,
      sym_sign,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2370), 1,
      sym__expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(432), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2738), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [8520] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1223), 1,
      anon_sym_LPAREN,
    ACTIONS(1225), 1,
      sym_condition_conversion,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    ACTIONS(1231), 1,
      sym_NEW,
    ACTIONS(1237), 1,
      sym_based_base,
    ACTIONS(1241), 1,
      sym_bit_string_length,
    ACTIONS(1243), 1,
      sym_bit_string_base,
    STATE(434), 1,
      sym_sign,
    STATE(830), 1,
      sym__direct_name,
    STATE(1678), 1,
      sym__abstract_literal,
    STATE(2019), 1,
      sym_based_literal,
    STATE(2027), 1,
      sym__expression,
    STATE(2648), 1,
      sym_bit_string_literal,
    STATE(2659), 1,
      sym__primary,
    STATE(2667), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1235), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2662), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(433), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2650), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1233), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [8641] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1223), 1,
      anon_sym_LPAREN,
    ACTIONS(1225), 1,
      sym_condition_conversion,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    ACTIONS(1231), 1,
      sym_NEW,
    ACTIONS(1237), 1,
      sym_based_base,
    ACTIONS(1241), 1,
      sym_bit_string_length,
    ACTIONS(1243), 1,
      sym_bit_string_base,
    STATE(433), 1,
      sym_unary_operator,
    STATE(830), 1,
      sym__direct_name,
    STATE(1678), 1,
      sym__abstract_literal,
    STATE(2019), 1,
      sym_based_literal,
    STATE(2028), 1,
      sym__expression,
    STATE(2648), 1,
      sym_bit_string_literal,
    STATE(2659), 1,
      sym__primary,
    STATE(2667), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1235), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2662), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(434), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2650), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1233), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [8762] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1223), 1,
      anon_sym_LPAREN,
    ACTIONS(1225), 1,
      sym_condition_conversion,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    ACTIONS(1231), 1,
      sym_NEW,
    ACTIONS(1237), 1,
      sym_based_base,
    ACTIONS(1241), 1,
      sym_bit_string_length,
    ACTIONS(1243), 1,
      sym_bit_string_base,
    STATE(433), 1,
      sym_unary_operator,
    STATE(434), 1,
      sym_sign,
    STATE(830), 1,
      sym__direct_name,
    STATE(1678), 1,
      sym__abstract_literal,
    STATE(2019), 1,
      sym_based_literal,
    STATE(2034), 1,
      sym__expression,
    STATE(2648), 1,
      sym_bit_string_literal,
    STATE(2659), 1,
      sym__primary,
    STATE(2667), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1235), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(435), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2662), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2650), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1233), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [8885] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1223), 1,
      anon_sym_LPAREN,
    ACTIONS(1225), 1,
      sym_condition_conversion,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    ACTIONS(1231), 1,
      sym_NEW,
    ACTIONS(1237), 1,
      sym_based_base,
    ACTIONS(1241), 1,
      sym_bit_string_length,
    ACTIONS(1243), 1,
      sym_bit_string_base,
    STATE(433), 1,
      sym_unary_operator,
    STATE(434), 1,
      sym_sign,
    STATE(830), 1,
      sym__direct_name,
    STATE(1678), 1,
      sym__abstract_literal,
    STATE(2019), 1,
      sym_based_literal,
    STATE(2038), 1,
      sym__expression,
    STATE(2648), 1,
      sym_bit_string_literal,
    STATE(2659), 1,
      sym__primary,
    STATE(2667), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1235), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(436), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2662), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2650), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1233), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [9008] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1223), 1,
      anon_sym_LPAREN,
    ACTIONS(1225), 1,
      sym_condition_conversion,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    ACTIONS(1231), 1,
      sym_NEW,
    ACTIONS(1237), 1,
      sym_based_base,
    ACTIONS(1241), 1,
      sym_bit_string_length,
    ACTIONS(1243), 1,
      sym_bit_string_base,
    STATE(433), 1,
      sym_unary_operator,
    STATE(434), 1,
      sym_sign,
    STATE(830), 1,
      sym__direct_name,
    STATE(1678), 1,
      sym__abstract_literal,
    STATE(2019), 1,
      sym_based_literal,
    STATE(2039), 1,
      sym__expression,
    STATE(2648), 1,
      sym_bit_string_literal,
    STATE(2659), 1,
      sym__primary,
    STATE(2667), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1235), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(437), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2662), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2650), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1233), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [9131] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2430), 1,
      sym__expression,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(438), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [9254] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    ACTIONS(1247), 1,
      sym_condition_conversion,
    STATE(720), 1,
      sym_unary_operator,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(902), 1,
      sym_based_literal,
    STATE(920), 1,
      sym__expression,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(439), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1644), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [9375] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    ACTIONS(1247), 1,
      sym_condition_conversion,
    STATE(439), 1,
      sym_sign,
    STATE(720), 1,
      sym_unary_operator,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(902), 1,
      sym_based_literal,
    STATE(918), 1,
      sym__expression,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(440), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1644), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [9498] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    ACTIONS(1247), 1,
      sym_condition_conversion,
    STATE(439), 1,
      sym_sign,
    STATE(720), 1,
      sym_unary_operator,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(902), 1,
      sym_based_literal,
    STATE(909), 1,
      sym__expression,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(441), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1644), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [9621] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    ACTIONS(1247), 1,
      sym_condition_conversion,
    STATE(439), 1,
      sym_sign,
    STATE(720), 1,
      sym_unary_operator,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(901), 1,
      sym__expression,
    STATE(902), 1,
      sym_based_literal,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(442), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1644), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [9744] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1249), 1,
      anon_sym_LPAREN,
    ACTIONS(1251), 1,
      sym_condition_conversion,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    ACTIONS(1257), 1,
      sym_NEW,
    ACTIONS(1263), 1,
      sym_based_base,
    ACTIONS(1267), 1,
      sym_bit_string_length,
    ACTIONS(1269), 1,
      sym_bit_string_base,
    STATE(444), 1,
      sym_sign,
    STATE(842), 1,
      sym__direct_name,
    STATE(2054), 1,
      sym__abstract_literal,
    STATE(2296), 1,
      sym_based_literal,
    STATE(2299), 1,
      sym__expression,
    STATE(2735), 1,
      sym_bit_string_literal,
    STATE(2746), 1,
      sym__primary,
    STATE(2753), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1261), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2752), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(443), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2734), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1259), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [9865] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1249), 1,
      anon_sym_LPAREN,
    ACTIONS(1251), 1,
      sym_condition_conversion,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    ACTIONS(1257), 1,
      sym_NEW,
    ACTIONS(1263), 1,
      sym_based_base,
    ACTIONS(1267), 1,
      sym_bit_string_length,
    ACTIONS(1269), 1,
      sym_bit_string_base,
    STATE(443), 1,
      sym_unary_operator,
    STATE(842), 1,
      sym__direct_name,
    STATE(2054), 1,
      sym__abstract_literal,
    STATE(2296), 1,
      sym_based_literal,
    STATE(2300), 1,
      sym__expression,
    STATE(2735), 1,
      sym_bit_string_literal,
    STATE(2746), 1,
      sym__primary,
    STATE(2753), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1261), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2752), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(444), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2734), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1259), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [9986] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1249), 1,
      anon_sym_LPAREN,
    ACTIONS(1251), 1,
      sym_condition_conversion,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    ACTIONS(1257), 1,
      sym_NEW,
    ACTIONS(1263), 1,
      sym_based_base,
    ACTIONS(1267), 1,
      sym_bit_string_length,
    ACTIONS(1269), 1,
      sym_bit_string_base,
    STATE(443), 1,
      sym_unary_operator,
    STATE(444), 1,
      sym_sign,
    STATE(842), 1,
      sym__direct_name,
    STATE(2054), 1,
      sym__abstract_literal,
    STATE(2296), 1,
      sym_based_literal,
    STATE(2433), 1,
      sym__expression,
    STATE(2735), 1,
      sym_bit_string_literal,
    STATE(2746), 1,
      sym__primary,
    STATE(2753), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1261), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(445), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2752), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2734), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1259), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [10109] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1249), 1,
      anon_sym_LPAREN,
    ACTIONS(1251), 1,
      sym_condition_conversion,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    ACTIONS(1257), 1,
      sym_NEW,
    ACTIONS(1263), 1,
      sym_based_base,
    ACTIONS(1267), 1,
      sym_bit_string_length,
    ACTIONS(1269), 1,
      sym_bit_string_base,
    STATE(443), 1,
      sym_unary_operator,
    STATE(444), 1,
      sym_sign,
    STATE(842), 1,
      sym__direct_name,
    STATE(2054), 1,
      sym__abstract_literal,
    STATE(2296), 1,
      sym_based_literal,
    STATE(2302), 1,
      sym__expression,
    STATE(2735), 1,
      sym_bit_string_literal,
    STATE(2746), 1,
      sym__primary,
    STATE(2753), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1261), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(446), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2752), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2734), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1259), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [10232] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1249), 1,
      anon_sym_LPAREN,
    ACTIONS(1251), 1,
      sym_condition_conversion,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    ACTIONS(1257), 1,
      sym_NEW,
    ACTIONS(1263), 1,
      sym_based_base,
    ACTIONS(1267), 1,
      sym_bit_string_length,
    ACTIONS(1269), 1,
      sym_bit_string_base,
    STATE(443), 1,
      sym_unary_operator,
    STATE(444), 1,
      sym_sign,
    STATE(842), 1,
      sym__direct_name,
    STATE(2054), 1,
      sym__abstract_literal,
    STATE(2296), 1,
      sym_based_literal,
    STATE(2303), 1,
      sym__expression,
    STATE(2735), 1,
      sym_bit_string_literal,
    STATE(2746), 1,
      sym__primary,
    STATE(2753), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1261), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(447), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2752), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2734), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1259), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [10355] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1271), 1,
      anon_sym_LPAREN,
    ACTIONS(1273), 1,
      sym_condition_conversion,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    ACTIONS(1279), 1,
      sym_NEW,
    ACTIONS(1285), 1,
      sym_based_base,
    ACTIONS(1289), 1,
      sym_bit_string_length,
    ACTIONS(1291), 1,
      sym_bit_string_base,
    STATE(449), 1,
      sym_sign,
    STATE(812), 1,
      sym__direct_name,
    STATE(1769), 1,
      sym__abstract_literal,
    STATE(2061), 1,
      sym_based_literal,
    STATE(2064), 1,
      sym__expression,
    STATE(2622), 1,
      sym__primary,
    STATE(2627), 1,
      sym__literal,
    STATE(2683), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2625), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(448), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2673), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1281), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [10476] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1271), 1,
      anon_sym_LPAREN,
    ACTIONS(1273), 1,
      sym_condition_conversion,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    ACTIONS(1279), 1,
      sym_NEW,
    ACTIONS(1285), 1,
      sym_based_base,
    ACTIONS(1289), 1,
      sym_bit_string_length,
    ACTIONS(1291), 1,
      sym_bit_string_base,
    STATE(448), 1,
      sym_unary_operator,
    STATE(812), 1,
      sym__direct_name,
    STATE(1769), 1,
      sym__abstract_literal,
    STATE(2061), 1,
      sym_based_literal,
    STATE(2065), 1,
      sym__expression,
    STATE(2622), 1,
      sym__primary,
    STATE(2627), 1,
      sym__literal,
    STATE(2683), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2625), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(449), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2673), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1281), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [10597] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1271), 1,
      anon_sym_LPAREN,
    ACTIONS(1273), 1,
      sym_condition_conversion,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    ACTIONS(1279), 1,
      sym_NEW,
    ACTIONS(1285), 1,
      sym_based_base,
    ACTIONS(1289), 1,
      sym_bit_string_length,
    ACTIONS(1291), 1,
      sym_bit_string_base,
    STATE(448), 1,
      sym_unary_operator,
    STATE(449), 1,
      sym_sign,
    STATE(812), 1,
      sym__direct_name,
    STATE(1769), 1,
      sym__abstract_literal,
    STATE(2061), 1,
      sym_based_literal,
    STATE(2066), 1,
      sym__expression,
    STATE(2622), 1,
      sym__primary,
    STATE(2627), 1,
      sym__literal,
    STATE(2683), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(450), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2625), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2673), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1281), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [10720] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1271), 1,
      anon_sym_LPAREN,
    ACTIONS(1273), 1,
      sym_condition_conversion,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    ACTIONS(1279), 1,
      sym_NEW,
    ACTIONS(1285), 1,
      sym_based_base,
    ACTIONS(1289), 1,
      sym_bit_string_length,
    ACTIONS(1291), 1,
      sym_bit_string_base,
    STATE(448), 1,
      sym_unary_operator,
    STATE(449), 1,
      sym_sign,
    STATE(812), 1,
      sym__direct_name,
    STATE(1769), 1,
      sym__abstract_literal,
    STATE(2061), 1,
      sym_based_literal,
    STATE(2070), 1,
      sym__expression,
    STATE(2622), 1,
      sym__primary,
    STATE(2627), 1,
      sym__literal,
    STATE(2683), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(451), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2625), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2673), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1281), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [10843] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1271), 1,
      anon_sym_LPAREN,
    ACTIONS(1273), 1,
      sym_condition_conversion,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    ACTIONS(1279), 1,
      sym_NEW,
    ACTIONS(1285), 1,
      sym_based_base,
    ACTIONS(1289), 1,
      sym_bit_string_length,
    ACTIONS(1291), 1,
      sym_bit_string_base,
    STATE(448), 1,
      sym_unary_operator,
    STATE(449), 1,
      sym_sign,
    STATE(812), 1,
      sym__direct_name,
    STATE(1769), 1,
      sym__abstract_literal,
    STATE(2061), 1,
      sym_based_literal,
    STATE(2071), 1,
      sym__expression,
    STATE(2622), 1,
      sym__primary,
    STATE(2627), 1,
      sym__literal,
    STATE(2683), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(452), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2625), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2673), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1281), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [10966] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    ACTIONS(860), 1,
      sym_NEW,
    ACTIONS(866), 1,
      sym_based_base,
    ACTIONS(870), 1,
      sym_bit_string_length,
    ACTIONS(872), 1,
      sym_bit_string_base,
    ACTIONS(984), 1,
      anon_sym_LPAREN,
    ACTIONS(1293), 1,
      sym_condition_conversion,
    STATE(454), 1,
      sym_sign,
    STATE(772), 1,
      sym__direct_name,
    STATE(1407), 1,
      sym__abstract_literal,
    STATE(1615), 1,
      sym_based_literal,
    STATE(1624), 1,
      sym__expression,
    STATE(2480), 1,
      sym__literal,
    STATE(2530), 1,
      sym_bit_string_literal,
    STATE(2535), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(864), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2507), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(453), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2481), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(862), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [11087] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    ACTIONS(860), 1,
      sym_NEW,
    ACTIONS(866), 1,
      sym_based_base,
    ACTIONS(870), 1,
      sym_bit_string_length,
    ACTIONS(872), 1,
      sym_bit_string_base,
    ACTIONS(984), 1,
      anon_sym_LPAREN,
    ACTIONS(1293), 1,
      sym_condition_conversion,
    STATE(453), 1,
      sym_unary_operator,
    STATE(772), 1,
      sym__direct_name,
    STATE(1407), 1,
      sym__abstract_literal,
    STATE(1615), 1,
      sym_based_literal,
    STATE(1626), 1,
      sym__expression,
    STATE(2480), 1,
      sym__literal,
    STATE(2530), 1,
      sym_bit_string_literal,
    STATE(2535), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(864), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2507), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(454), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2481), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(862), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [11208] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    ACTIONS(860), 1,
      sym_NEW,
    ACTIONS(866), 1,
      sym_based_base,
    ACTIONS(870), 1,
      sym_bit_string_length,
    ACTIONS(872), 1,
      sym_bit_string_base,
    ACTIONS(984), 1,
      anon_sym_LPAREN,
    ACTIONS(1293), 1,
      sym_condition_conversion,
    STATE(453), 1,
      sym_unary_operator,
    STATE(454), 1,
      sym_sign,
    STATE(772), 1,
      sym__direct_name,
    STATE(1407), 1,
      sym__abstract_literal,
    STATE(1615), 1,
      sym_based_literal,
    STATE(1639), 1,
      sym__expression,
    STATE(2480), 1,
      sym__literal,
    STATE(2530), 1,
      sym_bit_string_literal,
    STATE(2535), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(864), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(455), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2507), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2481), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(862), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [11331] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    ACTIONS(860), 1,
      sym_NEW,
    ACTIONS(866), 1,
      sym_based_base,
    ACTIONS(870), 1,
      sym_bit_string_length,
    ACTIONS(872), 1,
      sym_bit_string_base,
    ACTIONS(984), 1,
      anon_sym_LPAREN,
    ACTIONS(1293), 1,
      sym_condition_conversion,
    STATE(453), 1,
      sym_unary_operator,
    STATE(454), 1,
      sym_sign,
    STATE(772), 1,
      sym__direct_name,
    STATE(1407), 1,
      sym__abstract_literal,
    STATE(1615), 1,
      sym_based_literal,
    STATE(1640), 1,
      sym__expression,
    STATE(2480), 1,
      sym__literal,
    STATE(2530), 1,
      sym_bit_string_literal,
    STATE(2535), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(864), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(456), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2507), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2481), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(862), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [11454] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    ACTIONS(860), 1,
      sym_NEW,
    ACTIONS(866), 1,
      sym_based_base,
    ACTIONS(870), 1,
      sym_bit_string_length,
    ACTIONS(872), 1,
      sym_bit_string_base,
    ACTIONS(984), 1,
      anon_sym_LPAREN,
    ACTIONS(1293), 1,
      sym_condition_conversion,
    STATE(453), 1,
      sym_unary_operator,
    STATE(454), 1,
      sym_sign,
    STATE(772), 1,
      sym__direct_name,
    STATE(1407), 1,
      sym__abstract_literal,
    STATE(1615), 1,
      sym_based_literal,
    STATE(1641), 1,
      sym__expression,
    STATE(2480), 1,
      sym__literal,
    STATE(2530), 1,
      sym_bit_string_literal,
    STATE(2535), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(864), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(457), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2507), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2481), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(862), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [11577] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    ACTIONS(948), 1,
      sym_NEW,
    ACTIONS(954), 1,
      sym_based_base,
    ACTIONS(958), 1,
      sym_bit_string_length,
    ACTIONS(960), 1,
      sym_bit_string_base,
    ACTIONS(1295), 1,
      sym_condition_conversion,
    STATE(459), 1,
      sym_sign,
    STATE(779), 1,
      sym__direct_name,
    STATE(1413), 1,
      sym__abstract_literal,
    STATE(1469), 1,
      sym__expression,
    STATE(1642), 1,
      sym_based_literal,
    STATE(2496), 1,
      sym_bit_string_literal,
    STATE(2497), 1,
      sym__primary,
    STATE(2501), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(952), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2500), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(458), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2521), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(950), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [11698] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    ACTIONS(948), 1,
      sym_NEW,
    ACTIONS(954), 1,
      sym_based_base,
    ACTIONS(958), 1,
      sym_bit_string_length,
    ACTIONS(960), 1,
      sym_bit_string_base,
    ACTIONS(1295), 1,
      sym_condition_conversion,
    STATE(458), 1,
      sym_unary_operator,
    STATE(779), 1,
      sym__direct_name,
    STATE(1413), 1,
      sym__abstract_literal,
    STATE(1492), 1,
      sym__expression,
    STATE(1642), 1,
      sym_based_literal,
    STATE(2496), 1,
      sym_bit_string_literal,
    STATE(2497), 1,
      sym__primary,
    STATE(2501), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(952), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2500), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(459), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2521), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(950), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [11819] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    ACTIONS(948), 1,
      sym_NEW,
    ACTIONS(954), 1,
      sym_based_base,
    ACTIONS(958), 1,
      sym_bit_string_length,
    ACTIONS(960), 1,
      sym_bit_string_base,
    ACTIONS(1295), 1,
      sym_condition_conversion,
    STATE(458), 1,
      sym_unary_operator,
    STATE(459), 1,
      sym_sign,
    STATE(779), 1,
      sym__direct_name,
    STATE(1413), 1,
      sym__abstract_literal,
    STATE(1464), 1,
      sym__expression,
    STATE(1642), 1,
      sym_based_literal,
    STATE(2496), 1,
      sym_bit_string_literal,
    STATE(2497), 1,
      sym__primary,
    STATE(2501), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(952), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(460), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2500), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2521), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(950), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [11942] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    ACTIONS(948), 1,
      sym_NEW,
    ACTIONS(954), 1,
      sym_based_base,
    ACTIONS(958), 1,
      sym_bit_string_length,
    ACTIONS(960), 1,
      sym_bit_string_base,
    ACTIONS(1295), 1,
      sym_condition_conversion,
    STATE(458), 1,
      sym_unary_operator,
    STATE(459), 1,
      sym_sign,
    STATE(779), 1,
      sym__direct_name,
    STATE(1413), 1,
      sym__abstract_literal,
    STATE(1518), 1,
      sym__expression,
    STATE(1642), 1,
      sym_based_literal,
    STATE(2496), 1,
      sym_bit_string_literal,
    STATE(2497), 1,
      sym__primary,
    STATE(2501), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(952), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(461), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2500), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2521), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(950), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [12065] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    ACTIONS(948), 1,
      sym_NEW,
    ACTIONS(954), 1,
      sym_based_base,
    ACTIONS(958), 1,
      sym_bit_string_length,
    ACTIONS(960), 1,
      sym_bit_string_base,
    ACTIONS(1295), 1,
      sym_condition_conversion,
    STATE(458), 1,
      sym_unary_operator,
    STATE(459), 1,
      sym_sign,
    STATE(779), 1,
      sym__direct_name,
    STATE(1413), 1,
      sym__abstract_literal,
    STATE(1546), 1,
      sym__expression,
    STATE(1642), 1,
      sym_based_literal,
    STATE(2496), 1,
      sym_bit_string_literal,
    STATE(2497), 1,
      sym__primary,
    STATE(2501), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(952), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(462), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2500), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2521), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(950), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [12188] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(876), 1,
      anon_sym_LPAREN,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    ACTIONS(884), 1,
      sym_NEW,
    ACTIONS(890), 1,
      sym_based_base,
    ACTIONS(894), 1,
      sym_bit_string_length,
    ACTIONS(896), 1,
      sym_bit_string_base,
    ACTIONS(1297), 1,
      sym_condition_conversion,
    STATE(464), 1,
      sym_sign,
    STATE(786), 1,
      sym__direct_name,
    STATE(1609), 1,
      sym__abstract_literal,
    STATE(1675), 1,
      sym_based_literal,
    STATE(1701), 1,
      sym__expression,
    STATE(2559), 1,
      sym_bit_string_literal,
    STATE(2561), 1,
      sym__primary,
    STATE(2563), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(888), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2562), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(463), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2592), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(886), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [12309] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(876), 1,
      anon_sym_LPAREN,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    ACTIONS(884), 1,
      sym_NEW,
    ACTIONS(890), 1,
      sym_based_base,
    ACTIONS(894), 1,
      sym_bit_string_length,
    ACTIONS(896), 1,
      sym_bit_string_base,
    ACTIONS(1297), 1,
      sym_condition_conversion,
    STATE(463), 1,
      sym_unary_operator,
    STATE(786), 1,
      sym__direct_name,
    STATE(1609), 1,
      sym__abstract_literal,
    STATE(1675), 1,
      sym_based_literal,
    STATE(1702), 1,
      sym__expression,
    STATE(2559), 1,
      sym_bit_string_literal,
    STATE(2561), 1,
      sym__primary,
    STATE(2563), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(888), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2562), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(464), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2592), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(886), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [12430] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(876), 1,
      anon_sym_LPAREN,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    ACTIONS(884), 1,
      sym_NEW,
    ACTIONS(890), 1,
      sym_based_base,
    ACTIONS(894), 1,
      sym_bit_string_length,
    ACTIONS(896), 1,
      sym_bit_string_base,
    ACTIONS(1297), 1,
      sym_condition_conversion,
    STATE(463), 1,
      sym_unary_operator,
    STATE(464), 1,
      sym_sign,
    STATE(786), 1,
      sym__direct_name,
    STATE(1609), 1,
      sym__abstract_literal,
    STATE(1675), 1,
      sym_based_literal,
    STATE(1712), 1,
      sym__expression,
    STATE(2559), 1,
      sym_bit_string_literal,
    STATE(2561), 1,
      sym__primary,
    STATE(2563), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(888), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(465), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2562), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2592), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(886), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [12553] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(876), 1,
      anon_sym_LPAREN,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    ACTIONS(884), 1,
      sym_NEW,
    ACTIONS(890), 1,
      sym_based_base,
    ACTIONS(894), 1,
      sym_bit_string_length,
    ACTIONS(896), 1,
      sym_bit_string_base,
    ACTIONS(1297), 1,
      sym_condition_conversion,
    STATE(463), 1,
      sym_unary_operator,
    STATE(464), 1,
      sym_sign,
    STATE(786), 1,
      sym__direct_name,
    STATE(1609), 1,
      sym__abstract_literal,
    STATE(1675), 1,
      sym_based_literal,
    STATE(1839), 1,
      sym__expression,
    STATE(2559), 1,
      sym_bit_string_literal,
    STATE(2561), 1,
      sym__primary,
    STATE(2563), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(888), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(466), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2562), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2592), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(886), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [12676] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(876), 1,
      anon_sym_LPAREN,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    ACTIONS(884), 1,
      sym_NEW,
    ACTIONS(890), 1,
      sym_based_base,
    ACTIONS(894), 1,
      sym_bit_string_length,
    ACTIONS(896), 1,
      sym_bit_string_base,
    ACTIONS(1297), 1,
      sym_condition_conversion,
    STATE(463), 1,
      sym_unary_operator,
    STATE(464), 1,
      sym_sign,
    STATE(786), 1,
      sym__direct_name,
    STATE(1609), 1,
      sym__abstract_literal,
    STATE(1675), 1,
      sym_based_literal,
    STATE(1886), 1,
      sym__expression,
    STATE(2559), 1,
      sym_bit_string_literal,
    STATE(2561), 1,
      sym__primary,
    STATE(2563), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(888), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(467), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2562), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2592), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(886), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [12799] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(962), 1,
      anon_sym_LPAREN,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    ACTIONS(970), 1,
      sym_NEW,
    ACTIONS(976), 1,
      sym_based_base,
    ACTIONS(980), 1,
      sym_bit_string_length,
    ACTIONS(982), 1,
      sym_bit_string_base,
    ACTIONS(1299), 1,
      sym_condition_conversion,
    STATE(469), 1,
      sym_sign,
    STATE(793), 1,
      sym__direct_name,
    STATE(1497), 1,
      sym__abstract_literal,
    STATE(1680), 1,
      sym_based_literal,
    STATE(1861), 1,
      sym__expression,
    STATE(2554), 1,
      sym_bit_string_literal,
    STATE(2557), 1,
      sym__primary,
    STATE(2564), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(974), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2558), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(468), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2544), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(972), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [12920] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(962), 1,
      anon_sym_LPAREN,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    ACTIONS(970), 1,
      sym_NEW,
    ACTIONS(976), 1,
      sym_based_base,
    ACTIONS(980), 1,
      sym_bit_string_length,
    ACTIONS(982), 1,
      sym_bit_string_base,
    ACTIONS(1299), 1,
      sym_condition_conversion,
    STATE(468), 1,
      sym_unary_operator,
    STATE(793), 1,
      sym__direct_name,
    STATE(1497), 1,
      sym__abstract_literal,
    STATE(1680), 1,
      sym_based_literal,
    STATE(1866), 1,
      sym__expression,
    STATE(2554), 1,
      sym_bit_string_literal,
    STATE(2557), 1,
      sym__primary,
    STATE(2564), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(974), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2558), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(469), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2544), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(972), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [13041] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(962), 1,
      anon_sym_LPAREN,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    ACTIONS(970), 1,
      sym_NEW,
    ACTIONS(976), 1,
      sym_based_base,
    ACTIONS(980), 1,
      sym_bit_string_length,
    ACTIONS(982), 1,
      sym_bit_string_base,
    ACTIONS(1299), 1,
      sym_condition_conversion,
    STATE(468), 1,
      sym_unary_operator,
    STATE(469), 1,
      sym_sign,
    STATE(793), 1,
      sym__direct_name,
    STATE(1497), 1,
      sym__abstract_literal,
    STATE(1680), 1,
      sym_based_literal,
    STATE(1681), 1,
      sym__expression,
    STATE(2554), 1,
      sym_bit_string_literal,
    STATE(2557), 1,
      sym__primary,
    STATE(2564), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(974), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(470), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2558), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2544), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(972), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [13164] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(962), 1,
      anon_sym_LPAREN,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    ACTIONS(970), 1,
      sym_NEW,
    ACTIONS(976), 1,
      sym_based_base,
    ACTIONS(980), 1,
      sym_bit_string_length,
    ACTIONS(982), 1,
      sym_bit_string_base,
    ACTIONS(1299), 1,
      sym_condition_conversion,
    STATE(468), 1,
      sym_unary_operator,
    STATE(469), 1,
      sym_sign,
    STATE(793), 1,
      sym__direct_name,
    STATE(1497), 1,
      sym__abstract_literal,
    STATE(1680), 1,
      sym_based_literal,
    STATE(1688), 1,
      sym__expression,
    STATE(2554), 1,
      sym_bit_string_literal,
    STATE(2557), 1,
      sym__primary,
    STATE(2564), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(974), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(471), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2558), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2544), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(972), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [13287] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(962), 1,
      anon_sym_LPAREN,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    ACTIONS(970), 1,
      sym_NEW,
    ACTIONS(976), 1,
      sym_based_base,
    ACTIONS(980), 1,
      sym_bit_string_length,
    ACTIONS(982), 1,
      sym_bit_string_base,
    ACTIONS(1299), 1,
      sym_condition_conversion,
    STATE(468), 1,
      sym_unary_operator,
    STATE(469), 1,
      sym_sign,
    STATE(793), 1,
      sym__direct_name,
    STATE(1497), 1,
      sym__abstract_literal,
    STATE(1680), 1,
      sym_based_literal,
    STATE(1692), 1,
      sym__expression,
    STATE(2554), 1,
      sym_bit_string_literal,
    STATE(2557), 1,
      sym__primary,
    STATE(2564), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(974), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(472), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2558), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2544), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(972), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [13410] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1024), 1,
      anon_sym_LPAREN,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    ACTIONS(1032), 1,
      sym_NEW,
    ACTIONS(1038), 1,
      sym_based_base,
    ACTIONS(1042), 1,
      sym_bit_string_length,
    ACTIONS(1044), 1,
      sym_bit_string_base,
    ACTIONS(1301), 1,
      sym_condition_conversion,
    STATE(474), 1,
      sym_sign,
    STATE(750), 1,
      sym__direct_name,
    STATE(1140), 1,
      sym__abstract_literal,
    STATE(1326), 1,
      sym_based_literal,
    STATE(1332), 1,
      sym__expression,
    STATE(2436), 1,
      sym_bit_string_literal,
    STATE(2438), 1,
      sym__primary,
    STATE(2440), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1036), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2439), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(473), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2453), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1034), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [13531] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1024), 1,
      anon_sym_LPAREN,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    ACTIONS(1032), 1,
      sym_NEW,
    ACTIONS(1038), 1,
      sym_based_base,
    ACTIONS(1042), 1,
      sym_bit_string_length,
    ACTIONS(1044), 1,
      sym_bit_string_base,
    ACTIONS(1301), 1,
      sym_condition_conversion,
    STATE(473), 1,
      sym_unary_operator,
    STATE(750), 1,
      sym__direct_name,
    STATE(1140), 1,
      sym__abstract_literal,
    STATE(1326), 1,
      sym_based_literal,
    STATE(1333), 1,
      sym__expression,
    STATE(2436), 1,
      sym_bit_string_literal,
    STATE(2438), 1,
      sym__primary,
    STATE(2440), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1036), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2439), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(474), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2453), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1034), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [13652] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1024), 1,
      anon_sym_LPAREN,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    ACTIONS(1032), 1,
      sym_NEW,
    ACTIONS(1038), 1,
      sym_based_base,
    ACTIONS(1042), 1,
      sym_bit_string_length,
    ACTIONS(1044), 1,
      sym_bit_string_base,
    ACTIONS(1301), 1,
      sym_condition_conversion,
    STATE(473), 1,
      sym_unary_operator,
    STATE(474), 1,
      sym_sign,
    STATE(750), 1,
      sym__direct_name,
    STATE(1140), 1,
      sym__abstract_literal,
    STATE(1326), 1,
      sym_based_literal,
    STATE(1338), 1,
      sym__expression,
    STATE(2436), 1,
      sym_bit_string_literal,
    STATE(2438), 1,
      sym__primary,
    STATE(2440), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1036), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(475), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2439), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2453), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1034), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [13775] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1024), 1,
      anon_sym_LPAREN,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    ACTIONS(1032), 1,
      sym_NEW,
    ACTIONS(1038), 1,
      sym_based_base,
    ACTIONS(1042), 1,
      sym_bit_string_length,
    ACTIONS(1044), 1,
      sym_bit_string_base,
    ACTIONS(1301), 1,
      sym_condition_conversion,
    STATE(473), 1,
      sym_unary_operator,
    STATE(474), 1,
      sym_sign,
    STATE(750), 1,
      sym__direct_name,
    STATE(1140), 1,
      sym__abstract_literal,
    STATE(1326), 1,
      sym_based_literal,
    STATE(1365), 1,
      sym__expression,
    STATE(2436), 1,
      sym_bit_string_literal,
    STATE(2438), 1,
      sym__primary,
    STATE(2440), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1036), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(476), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2439), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2453), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1034), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [13898] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1024), 1,
      anon_sym_LPAREN,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    ACTIONS(1032), 1,
      sym_NEW,
    ACTIONS(1038), 1,
      sym_based_base,
    ACTIONS(1042), 1,
      sym_bit_string_length,
    ACTIONS(1044), 1,
      sym_bit_string_base,
    ACTIONS(1301), 1,
      sym_condition_conversion,
    STATE(473), 1,
      sym_unary_operator,
    STATE(474), 1,
      sym_sign,
    STATE(750), 1,
      sym__direct_name,
    STATE(1140), 1,
      sym__abstract_literal,
    STATE(1326), 1,
      sym_based_literal,
    STATE(1366), 1,
      sym__expression,
    STATE(2436), 1,
      sym_bit_string_literal,
    STATE(2438), 1,
      sym__primary,
    STATE(2440), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1036), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(477), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2439), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2453), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1034), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [14021] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1046), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    ACTIONS(1054), 1,
      sym_NEW,
    ACTIONS(1060), 1,
      sym_based_base,
    ACTIONS(1064), 1,
      sym_bit_string_length,
    ACTIONS(1066), 1,
      sym_bit_string_base,
    ACTIONS(1303), 1,
      sym_condition_conversion,
    STATE(479), 1,
      sym_sign,
    STATE(788), 1,
      sym__direct_name,
    STATE(1526), 1,
      sym__abstract_literal,
    STATE(1709), 1,
      sym_based_literal,
    STATE(1797), 1,
      sym__expression,
    STATE(2541), 1,
      sym__primary,
    STATE(2546), 1,
      sym__literal,
    STATE(2572), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1058), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2543), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(478), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2571), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1056), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [14142] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1046), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    ACTIONS(1054), 1,
      sym_NEW,
    ACTIONS(1060), 1,
      sym_based_base,
    ACTIONS(1064), 1,
      sym_bit_string_length,
    ACTIONS(1066), 1,
      sym_bit_string_base,
    ACTIONS(1303), 1,
      sym_condition_conversion,
    STATE(478), 1,
      sym_unary_operator,
    STATE(788), 1,
      sym__direct_name,
    STATE(1526), 1,
      sym__abstract_literal,
    STATE(1709), 1,
      sym_based_literal,
    STATE(1804), 1,
      sym__expression,
    STATE(2541), 1,
      sym__primary,
    STATE(2546), 1,
      sym__literal,
    STATE(2572), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1058), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2543), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(479), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2571), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1056), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [14263] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1046), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    ACTIONS(1054), 1,
      sym_NEW,
    ACTIONS(1060), 1,
      sym_based_base,
    ACTIONS(1064), 1,
      sym_bit_string_length,
    ACTIONS(1066), 1,
      sym_bit_string_base,
    ACTIONS(1303), 1,
      sym_condition_conversion,
    STATE(478), 1,
      sym_unary_operator,
    STATE(479), 1,
      sym_sign,
    STATE(788), 1,
      sym__direct_name,
    STATE(1526), 1,
      sym__abstract_literal,
    STATE(1709), 1,
      sym_based_literal,
    STATE(1819), 1,
      sym__expression,
    STATE(2541), 1,
      sym__primary,
    STATE(2546), 1,
      sym__literal,
    STATE(2572), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1058), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(480), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2543), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2571), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1056), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [14386] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1046), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    ACTIONS(1054), 1,
      sym_NEW,
    ACTIONS(1060), 1,
      sym_based_base,
    ACTIONS(1064), 1,
      sym_bit_string_length,
    ACTIONS(1066), 1,
      sym_bit_string_base,
    ACTIONS(1303), 1,
      sym_condition_conversion,
    STATE(478), 1,
      sym_unary_operator,
    STATE(479), 1,
      sym_sign,
    STATE(788), 1,
      sym__direct_name,
    STATE(1526), 1,
      sym__abstract_literal,
    STATE(1709), 1,
      sym_based_literal,
    STATE(1845), 1,
      sym__expression,
    STATE(2541), 1,
      sym__primary,
    STATE(2546), 1,
      sym__literal,
    STATE(2572), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1058), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(481), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2543), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2571), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1056), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [14509] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1046), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    ACTIONS(1054), 1,
      sym_NEW,
    ACTIONS(1060), 1,
      sym_based_base,
    ACTIONS(1064), 1,
      sym_bit_string_length,
    ACTIONS(1066), 1,
      sym_bit_string_base,
    ACTIONS(1303), 1,
      sym_condition_conversion,
    STATE(478), 1,
      sym_unary_operator,
    STATE(479), 1,
      sym_sign,
    STATE(788), 1,
      sym__direct_name,
    STATE(1526), 1,
      sym__abstract_literal,
    STATE(1709), 1,
      sym_based_literal,
    STATE(1875), 1,
      sym__expression,
    STATE(2541), 1,
      sym__primary,
    STATE(2546), 1,
      sym__literal,
    STATE(2572), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1058), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(482), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2543), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2571), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1056), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [14632] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      anon_sym_LPAREN,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    ACTIONS(906), 1,
      sym_NEW,
    ACTIONS(912), 1,
      sym_based_base,
    ACTIONS(916), 1,
      sym_bit_string_length,
    ACTIONS(918), 1,
      sym_bit_string_base,
    ACTIONS(1305), 1,
      sym_condition_conversion,
    STATE(484), 1,
      sym_sign,
    STATE(743), 1,
      sym__direct_name,
    STATE(1060), 1,
      sym__abstract_literal,
    STATE(1160), 1,
      sym__expression,
    STATE(1236), 1,
      sym_based_literal,
    STATE(2366), 1,
      sym_bit_string_literal,
    STATE(2367), 1,
      sym__primary,
    STATE(2369), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(910), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2368), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(483), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2417), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(908), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [14753] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      anon_sym_LPAREN,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    ACTIONS(906), 1,
      sym_NEW,
    ACTIONS(912), 1,
      sym_based_base,
    ACTIONS(916), 1,
      sym_bit_string_length,
    ACTIONS(918), 1,
      sym_bit_string_base,
    ACTIONS(1305), 1,
      sym_condition_conversion,
    STATE(483), 1,
      sym_unary_operator,
    STATE(743), 1,
      sym__direct_name,
    STATE(1060), 1,
      sym__abstract_literal,
    STATE(1128), 1,
      sym__expression,
    STATE(1236), 1,
      sym_based_literal,
    STATE(2366), 1,
      sym_bit_string_literal,
    STATE(2367), 1,
      sym__primary,
    STATE(2369), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(910), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2368), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(484), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2417), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(908), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [14874] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      anon_sym_LPAREN,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    ACTIONS(906), 1,
      sym_NEW,
    ACTIONS(912), 1,
      sym_based_base,
    ACTIONS(916), 1,
      sym_bit_string_length,
    ACTIONS(918), 1,
      sym_bit_string_base,
    ACTIONS(1305), 1,
      sym_condition_conversion,
    STATE(483), 1,
      sym_unary_operator,
    STATE(484), 1,
      sym_sign,
    STATE(743), 1,
      sym__direct_name,
    STATE(1060), 1,
      sym__abstract_literal,
    STATE(1164), 1,
      sym__expression,
    STATE(1236), 1,
      sym_based_literal,
    STATE(2366), 1,
      sym_bit_string_literal,
    STATE(2367), 1,
      sym__primary,
    STATE(2369), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(910), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(485), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2368), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2417), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(908), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [14997] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      anon_sym_LPAREN,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    ACTIONS(906), 1,
      sym_NEW,
    ACTIONS(912), 1,
      sym_based_base,
    ACTIONS(916), 1,
      sym_bit_string_length,
    ACTIONS(918), 1,
      sym_bit_string_base,
    ACTIONS(1305), 1,
      sym_condition_conversion,
    STATE(483), 1,
      sym_unary_operator,
    STATE(484), 1,
      sym_sign,
    STATE(743), 1,
      sym__direct_name,
    STATE(1060), 1,
      sym__abstract_literal,
    STATE(1223), 1,
      sym__expression,
    STATE(1236), 1,
      sym_based_literal,
    STATE(2366), 1,
      sym_bit_string_literal,
    STATE(2367), 1,
      sym__primary,
    STATE(2369), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(910), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(486), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2368), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2417), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(908), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [15120] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      anon_sym_LPAREN,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    ACTIONS(906), 1,
      sym_NEW,
    ACTIONS(912), 1,
      sym_based_base,
    ACTIONS(916), 1,
      sym_bit_string_length,
    ACTIONS(918), 1,
      sym_bit_string_base,
    ACTIONS(1305), 1,
      sym_condition_conversion,
    STATE(483), 1,
      sym_unary_operator,
    STATE(484), 1,
      sym_sign,
    STATE(743), 1,
      sym__direct_name,
    STATE(1060), 1,
      sym__abstract_literal,
    STATE(1228), 1,
      sym__expression,
    STATE(1236), 1,
      sym_based_literal,
    STATE(2366), 1,
      sym_bit_string_literal,
    STATE(2367), 1,
      sym__primary,
    STATE(2369), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(910), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(487), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2368), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2417), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(908), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [15243] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(854), 1,
      sym_condition_conversion,
    ACTIONS(1307), 1,
      anon_sym_LPAREN,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    ACTIONS(1313), 1,
      sym_NEW,
    ACTIONS(1319), 1,
      sym_based_base,
    ACTIONS(1323), 1,
      sym_bit_string_length,
    ACTIONS(1325), 1,
      sym_bit_string_base,
    STATE(489), 1,
      sym_sign,
    STATE(803), 1,
      sym__direct_name,
    STATE(1772), 1,
      sym__abstract_literal,
    STATE(2207), 1,
      sym_based_literal,
    STATE(2210), 1,
      sym__expression,
    STATE(2616), 1,
      sym__literal,
    STATE(2651), 1,
      sym_bit_string_literal,
    STATE(2653), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1317), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2654), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(488), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2621), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1315), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [15364] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(854), 1,
      sym_condition_conversion,
    ACTIONS(1307), 1,
      anon_sym_LPAREN,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    ACTIONS(1313), 1,
      sym_NEW,
    ACTIONS(1319), 1,
      sym_based_base,
    ACTIONS(1323), 1,
      sym_bit_string_length,
    ACTIONS(1325), 1,
      sym_bit_string_base,
    STATE(488), 1,
      sym_unary_operator,
    STATE(803), 1,
      sym__direct_name,
    STATE(1772), 1,
      sym__abstract_literal,
    STATE(2207), 1,
      sym_based_literal,
    STATE(2211), 1,
      sym__expression,
    STATE(2616), 1,
      sym__literal,
    STATE(2651), 1,
      sym_bit_string_literal,
    STATE(2653), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1317), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2654), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(489), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2621), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1315), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [15485] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(854), 1,
      sym_condition_conversion,
    ACTIONS(1307), 1,
      anon_sym_LPAREN,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    ACTIONS(1313), 1,
      sym_NEW,
    ACTIONS(1319), 1,
      sym_based_base,
    ACTIONS(1323), 1,
      sym_bit_string_length,
    ACTIONS(1325), 1,
      sym_bit_string_base,
    STATE(488), 1,
      sym_unary_operator,
    STATE(489), 1,
      sym_sign,
    STATE(803), 1,
      sym__direct_name,
    STATE(1772), 1,
      sym__abstract_literal,
    STATE(2207), 1,
      sym_based_literal,
    STATE(2214), 1,
      sym__expression,
    STATE(2616), 1,
      sym__literal,
    STATE(2651), 1,
      sym_bit_string_literal,
    STATE(2653), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1317), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(490), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2654), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2621), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1315), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [15608] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(854), 1,
      sym_condition_conversion,
    ACTIONS(1307), 1,
      anon_sym_LPAREN,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    ACTIONS(1313), 1,
      sym_NEW,
    ACTIONS(1319), 1,
      sym_based_base,
    ACTIONS(1323), 1,
      sym_bit_string_length,
    ACTIONS(1325), 1,
      sym_bit_string_base,
    STATE(488), 1,
      sym_unary_operator,
    STATE(489), 1,
      sym_sign,
    STATE(803), 1,
      sym__direct_name,
    STATE(1772), 1,
      sym__abstract_literal,
    STATE(2207), 1,
      sym_based_literal,
    STATE(2217), 1,
      sym__expression,
    STATE(2616), 1,
      sym__literal,
    STATE(2651), 1,
      sym_bit_string_literal,
    STATE(2653), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1317), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(491), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2654), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2621), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1315), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [15731] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(854), 1,
      sym_condition_conversion,
    ACTIONS(1307), 1,
      anon_sym_LPAREN,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    ACTIONS(1313), 1,
      sym_NEW,
    ACTIONS(1319), 1,
      sym_based_base,
    ACTIONS(1323), 1,
      sym_bit_string_length,
    ACTIONS(1325), 1,
      sym_bit_string_base,
    STATE(488), 1,
      sym_unary_operator,
    STATE(489), 1,
      sym_sign,
    STATE(803), 1,
      sym__direct_name,
    STATE(1772), 1,
      sym__abstract_literal,
    STATE(2207), 1,
      sym_based_literal,
    STATE(2218), 1,
      sym__expression,
    STATE(2616), 1,
      sym__literal,
    STATE(2651), 1,
      sym_bit_string_literal,
    STATE(2653), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1317), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(492), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2654), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2621), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1315), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [15854] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(942), 1,
      sym_condition_conversion,
    ACTIONS(1327), 1,
      anon_sym_LPAREN,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    ACTIONS(1333), 1,
      sym_NEW,
    ACTIONS(1339), 1,
      sym_based_base,
    ACTIONS(1343), 1,
      sym_bit_string_length,
    ACTIONS(1345), 1,
      sym_bit_string_base,
    STATE(494), 1,
      sym_sign,
    STATE(808), 1,
      sym__direct_name,
    STATE(1862), 1,
      sym__abstract_literal,
    STATE(2224), 1,
      sym_based_literal,
    STATE(2231), 1,
      sym__expression,
    STATE(2631), 1,
      sym_bit_string_literal,
    STATE(2649), 1,
      sym__primary,
    STATE(2663), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1337), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2661), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(493), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2681), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1335), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [15975] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(942), 1,
      sym_condition_conversion,
    ACTIONS(1327), 1,
      anon_sym_LPAREN,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    ACTIONS(1333), 1,
      sym_NEW,
    ACTIONS(1339), 1,
      sym_based_base,
    ACTIONS(1343), 1,
      sym_bit_string_length,
    ACTIONS(1345), 1,
      sym_bit_string_base,
    STATE(493), 1,
      sym_unary_operator,
    STATE(808), 1,
      sym__direct_name,
    STATE(1862), 1,
      sym__abstract_literal,
    STATE(2224), 1,
      sym_based_literal,
    STATE(2232), 1,
      sym__expression,
    STATE(2631), 1,
      sym_bit_string_literal,
    STATE(2649), 1,
      sym__primary,
    STATE(2663), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1337), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2661), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(494), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2681), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1335), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [16096] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2432), 1,
      sym__expression,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(495), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [16219] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(942), 1,
      sym_condition_conversion,
    ACTIONS(1327), 1,
      anon_sym_LPAREN,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    ACTIONS(1333), 1,
      sym_NEW,
    ACTIONS(1339), 1,
      sym_based_base,
    ACTIONS(1343), 1,
      sym_bit_string_length,
    ACTIONS(1345), 1,
      sym_bit_string_base,
    STATE(493), 1,
      sym_unary_operator,
    STATE(494), 1,
      sym_sign,
    STATE(808), 1,
      sym__direct_name,
    STATE(1862), 1,
      sym__abstract_literal,
    STATE(2224), 1,
      sym_based_literal,
    STATE(2239), 1,
      sym__expression,
    STATE(2631), 1,
      sym_bit_string_literal,
    STATE(2649), 1,
      sym__primary,
    STATE(2663), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1337), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(496), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2661), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2681), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1335), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [16342] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(942), 1,
      sym_condition_conversion,
    ACTIONS(1327), 1,
      anon_sym_LPAREN,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    ACTIONS(1333), 1,
      sym_NEW,
    ACTIONS(1339), 1,
      sym_based_base,
    ACTIONS(1343), 1,
      sym_bit_string_length,
    ACTIONS(1345), 1,
      sym_bit_string_base,
    STATE(493), 1,
      sym_unary_operator,
    STATE(494), 1,
      sym_sign,
    STATE(808), 1,
      sym__direct_name,
    STATE(1862), 1,
      sym__abstract_literal,
    STATE(2224), 1,
      sym_based_literal,
    STATE(2240), 1,
      sym__expression,
    STATE(2631), 1,
      sym_bit_string_literal,
    STATE(2649), 1,
      sym__primary,
    STATE(2663), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1337), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(497), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2661), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2681), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1335), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [16465] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(942), 1,
      sym_condition_conversion,
    ACTIONS(1327), 1,
      anon_sym_LPAREN,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    ACTIONS(1333), 1,
      sym_NEW,
    ACTIONS(1339), 1,
      sym_based_base,
    ACTIONS(1343), 1,
      sym_bit_string_length,
    ACTIONS(1345), 1,
      sym_bit_string_base,
    STATE(493), 1,
      sym_unary_operator,
    STATE(494), 1,
      sym_sign,
    STATE(808), 1,
      sym__direct_name,
    STATE(1862), 1,
      sym__abstract_literal,
    STATE(2224), 1,
      sym_based_literal,
    STATE(2241), 1,
      sym__expression,
    STATE(2631), 1,
      sym_bit_string_literal,
    STATE(2649), 1,
      sym__primary,
    STATE(2663), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1337), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(498), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2661), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2681), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1335), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [16588] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(986), 1,
      anon_sym_LPAREN,
    ACTIONS(990), 1,
      sym_NEW,
    ACTIONS(996), 1,
      sym_based_base,
    ACTIONS(998), 1,
      sym_bit_string_length,
    ACTIONS(1000), 1,
      sym_bit_string_base,
    ACTIONS(1347), 1,
      sym_condition_conversion,
    STATE(500), 1,
      sym_sign,
    STATE(763), 1,
      sym__direct_name,
    STATE(1138), 1,
      sym__abstract_literal,
    STATE(1257), 1,
      sym_based_literal,
    STATE(1263), 1,
      sym__expression,
    STATE(2452), 1,
      sym_bit_string_literal,
    STATE(2454), 1,
      sym__primary,
    STATE(2456), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(994), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2455), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(499), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2467), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(992), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [16709] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(986), 1,
      anon_sym_LPAREN,
    ACTIONS(990), 1,
      sym_NEW,
    ACTIONS(996), 1,
      sym_based_base,
    ACTIONS(998), 1,
      sym_bit_string_length,
    ACTIONS(1000), 1,
      sym_bit_string_base,
    ACTIONS(1347), 1,
      sym_condition_conversion,
    STATE(499), 1,
      sym_unary_operator,
    STATE(763), 1,
      sym__direct_name,
    STATE(1138), 1,
      sym__abstract_literal,
    STATE(1257), 1,
      sym_based_literal,
    STATE(1264), 1,
      sym__expression,
    STATE(2452), 1,
      sym_bit_string_literal,
    STATE(2454), 1,
      sym__primary,
    STATE(2456), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(994), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2455), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(500), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2467), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(992), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [16830] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(986), 1,
      anon_sym_LPAREN,
    ACTIONS(990), 1,
      sym_NEW,
    ACTIONS(996), 1,
      sym_based_base,
    ACTIONS(998), 1,
      sym_bit_string_length,
    ACTIONS(1000), 1,
      sym_bit_string_base,
    ACTIONS(1347), 1,
      sym_condition_conversion,
    STATE(499), 1,
      sym_unary_operator,
    STATE(500), 1,
      sym_sign,
    STATE(763), 1,
      sym__direct_name,
    STATE(1138), 1,
      sym__abstract_literal,
    STATE(1257), 1,
      sym_based_literal,
    STATE(1266), 1,
      sym__expression,
    STATE(2452), 1,
      sym_bit_string_literal,
    STATE(2454), 1,
      sym__primary,
    STATE(2456), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(994), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(501), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2455), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2467), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(992), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [16953] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(986), 1,
      anon_sym_LPAREN,
    ACTIONS(990), 1,
      sym_NEW,
    ACTIONS(996), 1,
      sym_based_base,
    ACTIONS(998), 1,
      sym_bit_string_length,
    ACTIONS(1000), 1,
      sym_bit_string_base,
    ACTIONS(1347), 1,
      sym_condition_conversion,
    STATE(499), 1,
      sym_unary_operator,
    STATE(500), 1,
      sym_sign,
    STATE(763), 1,
      sym__direct_name,
    STATE(1138), 1,
      sym__abstract_literal,
    STATE(1257), 1,
      sym_based_literal,
    STATE(1270), 1,
      sym__expression,
    STATE(2452), 1,
      sym_bit_string_literal,
    STATE(2454), 1,
      sym__primary,
    STATE(2456), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(994), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(502), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2455), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2467), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(992), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [17076] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(986), 1,
      anon_sym_LPAREN,
    ACTIONS(990), 1,
      sym_NEW,
    ACTIONS(996), 1,
      sym_based_base,
    ACTIONS(998), 1,
      sym_bit_string_length,
    ACTIONS(1000), 1,
      sym_bit_string_base,
    ACTIONS(1347), 1,
      sym_condition_conversion,
    STATE(499), 1,
      sym_unary_operator,
    STATE(500), 1,
      sym_sign,
    STATE(763), 1,
      sym__direct_name,
    STATE(1138), 1,
      sym__abstract_literal,
    STATE(1257), 1,
      sym_based_literal,
    STATE(1271), 1,
      sym__expression,
    STATE(2452), 1,
      sym_bit_string_literal,
    STATE(2454), 1,
      sym__primary,
    STATE(2456), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(994), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(503), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2455), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2467), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(992), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [17199] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1026), 1,
      sym_condition_conversion,
    ACTIONS(1349), 1,
      anon_sym_LPAREN,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    ACTIONS(1355), 1,
      sym_NEW,
    ACTIONS(1361), 1,
      sym_based_base,
    ACTIONS(1365), 1,
      sym_bit_string_length,
    ACTIONS(1367), 1,
      sym_bit_string_base,
    STATE(505), 1,
      sym_sign,
    STATE(795), 1,
      sym__direct_name,
    STATE(1455), 1,
      sym__abstract_literal,
    STATE(1683), 1,
      sym_based_literal,
    STATE(1720), 1,
      sym__expression,
    STATE(2570), 1,
      sym_bit_string_literal,
    STATE(2574), 1,
      sym__literal,
    STATE(2615), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1359), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2573), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(504), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2560), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1357), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [17320] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1026), 1,
      sym_condition_conversion,
    ACTIONS(1349), 1,
      anon_sym_LPAREN,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    ACTIONS(1355), 1,
      sym_NEW,
    ACTIONS(1361), 1,
      sym_based_base,
    ACTIONS(1365), 1,
      sym_bit_string_length,
    ACTIONS(1367), 1,
      sym_bit_string_base,
    STATE(504), 1,
      sym_unary_operator,
    STATE(795), 1,
      sym__direct_name,
    STATE(1455), 1,
      sym__abstract_literal,
    STATE(1683), 1,
      sym_based_literal,
    STATE(1721), 1,
      sym__expression,
    STATE(2570), 1,
      sym_bit_string_literal,
    STATE(2574), 1,
      sym__literal,
    STATE(2615), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1359), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2573), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(505), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2560), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1357), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [17441] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1026), 1,
      sym_condition_conversion,
    ACTIONS(1349), 1,
      anon_sym_LPAREN,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    ACTIONS(1355), 1,
      sym_NEW,
    ACTIONS(1361), 1,
      sym_based_base,
    ACTIONS(1365), 1,
      sym_bit_string_length,
    ACTIONS(1367), 1,
      sym_bit_string_base,
    STATE(504), 1,
      sym_unary_operator,
    STATE(505), 1,
      sym_sign,
    STATE(795), 1,
      sym__direct_name,
    STATE(1455), 1,
      sym__abstract_literal,
    STATE(1683), 1,
      sym_based_literal,
    STATE(1722), 1,
      sym__expression,
    STATE(2570), 1,
      sym_bit_string_literal,
    STATE(2574), 1,
      sym__literal,
    STATE(2615), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1359), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(506), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2573), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2560), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1357), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [17564] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1026), 1,
      sym_condition_conversion,
    ACTIONS(1349), 1,
      anon_sym_LPAREN,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    ACTIONS(1355), 1,
      sym_NEW,
    ACTIONS(1361), 1,
      sym_based_base,
    ACTIONS(1365), 1,
      sym_bit_string_length,
    ACTIONS(1367), 1,
      sym_bit_string_base,
    STATE(504), 1,
      sym_unary_operator,
    STATE(505), 1,
      sym_sign,
    STATE(795), 1,
      sym__direct_name,
    STATE(1455), 1,
      sym__abstract_literal,
    STATE(1683), 1,
      sym_based_literal,
    STATE(1775), 1,
      sym__expression,
    STATE(2570), 1,
      sym_bit_string_literal,
    STATE(2574), 1,
      sym__literal,
    STATE(2615), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1359), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(507), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2573), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2560), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1357), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [17687] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1026), 1,
      sym_condition_conversion,
    ACTIONS(1349), 1,
      anon_sym_LPAREN,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    ACTIONS(1355), 1,
      sym_NEW,
    ACTIONS(1361), 1,
      sym_based_base,
    ACTIONS(1365), 1,
      sym_bit_string_length,
    ACTIONS(1367), 1,
      sym_bit_string_base,
    STATE(504), 1,
      sym_unary_operator,
    STATE(505), 1,
      sym_sign,
    STATE(795), 1,
      sym__direct_name,
    STATE(1455), 1,
      sym__abstract_literal,
    STATE(1683), 1,
      sym_based_literal,
    STATE(1778), 1,
      sym__expression,
    STATE(2570), 1,
      sym_bit_string_literal,
    STATE(2574), 1,
      sym__literal,
    STATE(2615), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1359), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(508), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2573), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2560), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1357), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [17810] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(900), 1,
      sym_condition_conversion,
    ACTIONS(1369), 1,
      anon_sym_LPAREN,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    ACTIONS(1375), 1,
      sym_NEW,
    ACTIONS(1381), 1,
      sym_based_base,
    ACTIONS(1385), 1,
      sym_bit_string_length,
    ACTIONS(1387), 1,
      sym_bit_string_base,
    STATE(510), 1,
      sym_sign,
    STATE(767), 1,
      sym__direct_name,
    STATE(1289), 1,
      sym__abstract_literal,
    STATE(1467), 1,
      sym_based_literal,
    STATE(1498), 1,
      sym__expression,
    STATE(2513), 1,
      sym_bit_string_literal,
    STATE(2516), 1,
      sym__primary,
    STATE(2518), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1379), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2517), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(509), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2491), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1377), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [17931] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(900), 1,
      sym_condition_conversion,
    ACTIONS(1369), 1,
      anon_sym_LPAREN,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    ACTIONS(1375), 1,
      sym_NEW,
    ACTIONS(1381), 1,
      sym_based_base,
    ACTIONS(1385), 1,
      sym_bit_string_length,
    ACTIONS(1387), 1,
      sym_bit_string_base,
    STATE(509), 1,
      sym_unary_operator,
    STATE(767), 1,
      sym__direct_name,
    STATE(1289), 1,
      sym__abstract_literal,
    STATE(1467), 1,
      sym_based_literal,
    STATE(1505), 1,
      sym__expression,
    STATE(2513), 1,
      sym_bit_string_literal,
    STATE(2516), 1,
      sym__primary,
    STATE(2518), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1379), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2517), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(510), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2491), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1377), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [18052] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(900), 1,
      sym_condition_conversion,
    ACTIONS(1369), 1,
      anon_sym_LPAREN,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    ACTIONS(1375), 1,
      sym_NEW,
    ACTIONS(1381), 1,
      sym_based_base,
    ACTIONS(1385), 1,
      sym_bit_string_length,
    ACTIONS(1387), 1,
      sym_bit_string_base,
    STATE(509), 1,
      sym_unary_operator,
    STATE(510), 1,
      sym_sign,
    STATE(767), 1,
      sym__direct_name,
    STATE(1289), 1,
      sym__abstract_literal,
    STATE(1467), 1,
      sym_based_literal,
    STATE(1512), 1,
      sym__expression,
    STATE(2513), 1,
      sym_bit_string_literal,
    STATE(2516), 1,
      sym__primary,
    STATE(2518), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1379), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(511), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2517), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2491), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1377), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [18175] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(900), 1,
      sym_condition_conversion,
    ACTIONS(1369), 1,
      anon_sym_LPAREN,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    ACTIONS(1375), 1,
      sym_NEW,
    ACTIONS(1381), 1,
      sym_based_base,
    ACTIONS(1385), 1,
      sym_bit_string_length,
    ACTIONS(1387), 1,
      sym_bit_string_base,
    STATE(509), 1,
      sym_unary_operator,
    STATE(510), 1,
      sym_sign,
    STATE(767), 1,
      sym__direct_name,
    STATE(1289), 1,
      sym__abstract_literal,
    STATE(1467), 1,
      sym_based_literal,
    STATE(1525), 1,
      sym__expression,
    STATE(2513), 1,
      sym_bit_string_literal,
    STATE(2516), 1,
      sym__primary,
    STATE(2518), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1379), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(512), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2517), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2491), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1377), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [18298] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(900), 1,
      sym_condition_conversion,
    ACTIONS(1369), 1,
      anon_sym_LPAREN,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    ACTIONS(1375), 1,
      sym_NEW,
    ACTIONS(1381), 1,
      sym_based_base,
    ACTIONS(1385), 1,
      sym_bit_string_length,
    ACTIONS(1387), 1,
      sym_bit_string_base,
    STATE(509), 1,
      sym_unary_operator,
    STATE(510), 1,
      sym_sign,
    STATE(767), 1,
      sym__direct_name,
    STATE(1289), 1,
      sym__abstract_literal,
    STATE(1467), 1,
      sym_based_literal,
    STATE(1535), 1,
      sym__expression,
    STATE(2513), 1,
      sym_bit_string_literal,
    STATE(2516), 1,
      sym__primary,
    STATE(2518), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1379), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(513), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2517), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2491), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1377), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [18421] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(988), 1,
      sym_condition_conversion,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1389), 1,
      anon_sym_LPAREN,
    ACTIONS(1391), 1,
      sym_NEW,
    ACTIONS(1397), 1,
      sym_based_base,
    ACTIONS(1399), 1,
      sym_bit_string_length,
    ACTIONS(1401), 1,
      sym_bit_string_base,
    STATE(515), 1,
      sym_sign,
    STATE(822), 1,
      sym__direct_name,
    STATE(1610), 1,
      sym__abstract_literal,
    STATE(1868), 1,
      sym__expression,
    STATE(1887), 1,
      sym_based_literal,
    STATE(2598), 1,
      sym_bit_string_literal,
    STATE(2599), 1,
      sym__primary,
    STATE(2601), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1395), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2600), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(514), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2565), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1393), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [18542] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(988), 1,
      sym_condition_conversion,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1389), 1,
      anon_sym_LPAREN,
    ACTIONS(1391), 1,
      sym_NEW,
    ACTIONS(1397), 1,
      sym_based_base,
    ACTIONS(1399), 1,
      sym_bit_string_length,
    ACTIONS(1401), 1,
      sym_bit_string_base,
    STATE(514), 1,
      sym_unary_operator,
    STATE(822), 1,
      sym__direct_name,
    STATE(1610), 1,
      sym__abstract_literal,
    STATE(1870), 1,
      sym__expression,
    STATE(1887), 1,
      sym_based_literal,
    STATE(2598), 1,
      sym_bit_string_literal,
    STATE(2599), 1,
      sym__primary,
    STATE(2601), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1395), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2600), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(515), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2565), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1393), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [18663] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(988), 1,
      sym_condition_conversion,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1389), 1,
      anon_sym_LPAREN,
    ACTIONS(1391), 1,
      sym_NEW,
    ACTIONS(1397), 1,
      sym_based_base,
    ACTIONS(1399), 1,
      sym_bit_string_length,
    ACTIONS(1401), 1,
      sym_bit_string_base,
    STATE(514), 1,
      sym_unary_operator,
    STATE(515), 1,
      sym_sign,
    STATE(822), 1,
      sym__direct_name,
    STATE(1610), 1,
      sym__abstract_literal,
    STATE(1887), 1,
      sym_based_literal,
    STATE(1901), 1,
      sym__expression,
    STATE(2598), 1,
      sym_bit_string_literal,
    STATE(2599), 1,
      sym__primary,
    STATE(2601), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1395), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(516), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2600), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2565), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1393), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [18786] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(988), 1,
      sym_condition_conversion,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1389), 1,
      anon_sym_LPAREN,
    ACTIONS(1391), 1,
      sym_NEW,
    ACTIONS(1397), 1,
      sym_based_base,
    ACTIONS(1399), 1,
      sym_bit_string_length,
    ACTIONS(1401), 1,
      sym_bit_string_base,
    STATE(514), 1,
      sym_unary_operator,
    STATE(515), 1,
      sym_sign,
    STATE(822), 1,
      sym__direct_name,
    STATE(1610), 1,
      sym__abstract_literal,
    STATE(1803), 1,
      sym__expression,
    STATE(1887), 1,
      sym_based_literal,
    STATE(2598), 1,
      sym_bit_string_literal,
    STATE(2599), 1,
      sym__primary,
    STATE(2601), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1395), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(517), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2600), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2565), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1393), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [18909] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(988), 1,
      sym_condition_conversion,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1389), 1,
      anon_sym_LPAREN,
    ACTIONS(1391), 1,
      sym_NEW,
    ACTIONS(1397), 1,
      sym_based_base,
    ACTIONS(1399), 1,
      sym_bit_string_length,
    ACTIONS(1401), 1,
      sym_bit_string_base,
    STATE(514), 1,
      sym_unary_operator,
    STATE(515), 1,
      sym_sign,
    STATE(822), 1,
      sym__direct_name,
    STATE(1610), 1,
      sym__abstract_literal,
    STATE(1677), 1,
      sym__expression,
    STATE(1887), 1,
      sym_based_literal,
    STATE(2598), 1,
      sym_bit_string_literal,
    STATE(2599), 1,
      sym__primary,
    STATE(2601), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1395), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(518), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2600), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2565), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1393), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [19032] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1002), 1,
      anon_sym_LPAREN,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    ACTIONS(1010), 1,
      sym_NEW,
    ACTIONS(1016), 1,
      sym_based_base,
    ACTIONS(1020), 1,
      sym_bit_string_length,
    ACTIONS(1022), 1,
      sym_bit_string_base,
    ACTIONS(1403), 1,
      sym_condition_conversion,
    STATE(520), 1,
      sym_sign,
    STATE(746), 1,
      sym__direct_name,
    STATE(1071), 1,
      sym__abstract_literal,
    STATE(1154), 1,
      sym_based_literal,
    STATE(1165), 1,
      sym__expression,
    STATE(2290), 1,
      sym__primary,
    STATE(2399), 1,
      sym__literal,
    STATE(2429), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1014), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2301), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(519), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2387), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1012), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [19153] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1002), 1,
      anon_sym_LPAREN,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    ACTIONS(1010), 1,
      sym_NEW,
    ACTIONS(1016), 1,
      sym_based_base,
    ACTIONS(1020), 1,
      sym_bit_string_length,
    ACTIONS(1022), 1,
      sym_bit_string_base,
    ACTIONS(1403), 1,
      sym_condition_conversion,
    STATE(519), 1,
      sym_unary_operator,
    STATE(746), 1,
      sym__direct_name,
    STATE(1071), 1,
      sym__abstract_literal,
    STATE(1154), 1,
      sym_based_literal,
    STATE(1168), 1,
      sym__expression,
    STATE(2290), 1,
      sym__primary,
    STATE(2399), 1,
      sym__literal,
    STATE(2429), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1014), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2301), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(520), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2387), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1012), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [19274] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1002), 1,
      anon_sym_LPAREN,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    ACTIONS(1010), 1,
      sym_NEW,
    ACTIONS(1016), 1,
      sym_based_base,
    ACTIONS(1020), 1,
      sym_bit_string_length,
    ACTIONS(1022), 1,
      sym_bit_string_base,
    ACTIONS(1403), 1,
      sym_condition_conversion,
    STATE(519), 1,
      sym_unary_operator,
    STATE(520), 1,
      sym_sign,
    STATE(746), 1,
      sym__direct_name,
    STATE(1071), 1,
      sym__abstract_literal,
    STATE(1154), 1,
      sym_based_literal,
    STATE(1205), 1,
      sym__expression,
    STATE(2290), 1,
      sym__primary,
    STATE(2399), 1,
      sym__literal,
    STATE(2429), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1014), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(521), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2301), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2387), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1012), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [19397] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1002), 1,
      anon_sym_LPAREN,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    ACTIONS(1010), 1,
      sym_NEW,
    ACTIONS(1016), 1,
      sym_based_base,
    ACTIONS(1020), 1,
      sym_bit_string_length,
    ACTIONS(1022), 1,
      sym_bit_string_base,
    ACTIONS(1403), 1,
      sym_condition_conversion,
    STATE(519), 1,
      sym_unary_operator,
    STATE(520), 1,
      sym_sign,
    STATE(746), 1,
      sym__direct_name,
    STATE(1071), 1,
      sym__abstract_literal,
    STATE(1154), 1,
      sym_based_literal,
    STATE(1221), 1,
      sym__expression,
    STATE(2290), 1,
      sym__primary,
    STATE(2399), 1,
      sym__literal,
    STATE(2429), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1014), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(522), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2301), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2387), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1012), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [19520] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1002), 1,
      anon_sym_LPAREN,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    ACTIONS(1010), 1,
      sym_NEW,
    ACTIONS(1016), 1,
      sym_based_base,
    ACTIONS(1020), 1,
      sym_bit_string_length,
    ACTIONS(1022), 1,
      sym_bit_string_base,
    ACTIONS(1403), 1,
      sym_condition_conversion,
    STATE(519), 1,
      sym_unary_operator,
    STATE(520), 1,
      sym_sign,
    STATE(746), 1,
      sym__direct_name,
    STATE(1071), 1,
      sym__abstract_literal,
    STATE(1154), 1,
      sym_based_literal,
    STATE(1180), 1,
      sym__expression,
    STATE(2290), 1,
      sym__primary,
    STATE(2399), 1,
      sym__literal,
    STATE(2429), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1014), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(523), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2301), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2387), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1012), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [19643] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1004), 1,
      sym_condition_conversion,
    ACTIONS(1405), 1,
      anon_sym_LPAREN,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    ACTIONS(1411), 1,
      sym_NEW,
    ACTIONS(1417), 1,
      sym_based_base,
    ACTIONS(1421), 1,
      sym_bit_string_length,
    ACTIONS(1423), 1,
      sym_bit_string_base,
    STATE(525), 1,
      sym_sign,
    STATE(769), 1,
      sym__direct_name,
    STATE(1331), 1,
      sym__abstract_literal,
    STATE(1445), 1,
      sym_based_literal,
    STATE(1451), 1,
      sym__expression,
    STATE(2477), 1,
      sym_bit_string_literal,
    STATE(2482), 1,
      sym__primary,
    STATE(2485), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1415), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2484), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(524), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2533), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1413), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [19764] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1004), 1,
      sym_condition_conversion,
    ACTIONS(1405), 1,
      anon_sym_LPAREN,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    ACTIONS(1411), 1,
      sym_NEW,
    ACTIONS(1417), 1,
      sym_based_base,
    ACTIONS(1421), 1,
      sym_bit_string_length,
    ACTIONS(1423), 1,
      sym_bit_string_base,
    STATE(524), 1,
      sym_unary_operator,
    STATE(769), 1,
      sym__direct_name,
    STATE(1331), 1,
      sym__abstract_literal,
    STATE(1445), 1,
      sym_based_literal,
    STATE(1452), 1,
      sym__expression,
    STATE(2477), 1,
      sym_bit_string_literal,
    STATE(2482), 1,
      sym__primary,
    STATE(2485), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1415), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2484), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(525), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2533), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1413), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [19885] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1004), 1,
      sym_condition_conversion,
    ACTIONS(1405), 1,
      anon_sym_LPAREN,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    ACTIONS(1411), 1,
      sym_NEW,
    ACTIONS(1417), 1,
      sym_based_base,
    ACTIONS(1421), 1,
      sym_bit_string_length,
    ACTIONS(1423), 1,
      sym_bit_string_base,
    STATE(524), 1,
      sym_unary_operator,
    STATE(525), 1,
      sym_sign,
    STATE(769), 1,
      sym__direct_name,
    STATE(1331), 1,
      sym__abstract_literal,
    STATE(1445), 1,
      sym_based_literal,
    STATE(1453), 1,
      sym__expression,
    STATE(2477), 1,
      sym_bit_string_literal,
    STATE(2482), 1,
      sym__primary,
    STATE(2485), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1415), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(526), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2484), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2533), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1413), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [20008] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1004), 1,
      sym_condition_conversion,
    ACTIONS(1405), 1,
      anon_sym_LPAREN,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    ACTIONS(1411), 1,
      sym_NEW,
    ACTIONS(1417), 1,
      sym_based_base,
    ACTIONS(1421), 1,
      sym_bit_string_length,
    ACTIONS(1423), 1,
      sym_bit_string_base,
    STATE(524), 1,
      sym_unary_operator,
    STATE(525), 1,
      sym_sign,
    STATE(769), 1,
      sym__direct_name,
    STATE(1331), 1,
      sym__abstract_literal,
    STATE(1445), 1,
      sym_based_literal,
    STATE(1456), 1,
      sym__expression,
    STATE(2477), 1,
      sym_bit_string_literal,
    STATE(2482), 1,
      sym__primary,
    STATE(2485), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1415), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(527), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2484), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2533), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1413), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [20131] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1004), 1,
      sym_condition_conversion,
    ACTIONS(1405), 1,
      anon_sym_LPAREN,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    ACTIONS(1411), 1,
      sym_NEW,
    ACTIONS(1417), 1,
      sym_based_base,
    ACTIONS(1421), 1,
      sym_bit_string_length,
    ACTIONS(1423), 1,
      sym_bit_string_base,
    STATE(524), 1,
      sym_unary_operator,
    STATE(525), 1,
      sym_sign,
    STATE(769), 1,
      sym__direct_name,
    STATE(1331), 1,
      sym__abstract_literal,
    STATE(1445), 1,
      sym_based_literal,
    STATE(1457), 1,
      sym__expression,
    STATE(2477), 1,
      sym_bit_string_literal,
    STATE(2482), 1,
      sym__primary,
    STATE(2485), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1415), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(528), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2484), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2533), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1413), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [20254] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(1517), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(529), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [20377] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2291), 1,
      sym__expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(530), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [20500] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1307), 1,
      anon_sym_LPAREN,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    ACTIONS(1313), 1,
      sym_NEW,
    ACTIONS(1319), 1,
      sym_based_base,
    ACTIONS(1323), 1,
      sym_bit_string_length,
    ACTIONS(1325), 1,
      sym_bit_string_base,
    STATE(488), 1,
      sym_unary_operator,
    STATE(489), 1,
      sym_sign,
    STATE(803), 1,
      sym__direct_name,
    STATE(1772), 1,
      sym__abstract_literal,
    STATE(2207), 1,
      sym_based_literal,
    STATE(2464), 1,
      sym__expression,
    STATE(2616), 1,
      sym__literal,
    STATE(2651), 1,
      sym_bit_string_literal,
    STATE(2653), 1,
      sym__primary,
    STATE(2675), 1,
      sym_simple_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1317), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(531), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2654), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1315), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [20625] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2379), 1,
      sym__expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(532), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [20748] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2381), 1,
      sym__expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(533), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [20871] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(411), 1,
      anon_sym_LPAREN,
    ACTIONS(413), 1,
      sym_condition_conversion,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    ACTIONS(423), 1,
      sym_NEW,
    ACTIONS(433), 1,
      sym_based_base,
    ACTIONS(437), 1,
      sym_bit_string_length,
    ACTIONS(439), 1,
      sym_bit_string_base,
    STATE(704), 1,
      sym_unary_operator,
    STATE(705), 1,
      sym_sign,
    STATE(777), 1,
      sym__direct_name,
    STATE(1408), 1,
      sym__abstract_literal,
    STATE(1627), 1,
      sym_based_literal,
    STATE(1630), 1,
      sym__expression,
    STATE(2506), 1,
      sym_bit_string_literal,
    STATE(2508), 1,
      sym__primary,
    STATE(2510), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(431), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(534), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2509), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2536), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(425), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [20994] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(411), 1,
      anon_sym_LPAREN,
    ACTIONS(413), 1,
      sym_condition_conversion,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    ACTIONS(423), 1,
      sym_NEW,
    ACTIONS(433), 1,
      sym_based_base,
    ACTIONS(437), 1,
      sym_bit_string_length,
    ACTIONS(439), 1,
      sym_bit_string_base,
    STATE(704), 1,
      sym_unary_operator,
    STATE(705), 1,
      sym_sign,
    STATE(777), 1,
      sym__direct_name,
    STATE(1408), 1,
      sym__abstract_literal,
    STATE(1627), 1,
      sym_based_literal,
    STATE(1634), 1,
      sym__expression,
    STATE(2506), 1,
      sym_bit_string_literal,
    STATE(2508), 1,
      sym__primary,
    STATE(2510), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(431), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(535), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2509), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2536), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(425), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [21117] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(411), 1,
      anon_sym_LPAREN,
    ACTIONS(413), 1,
      sym_condition_conversion,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    ACTIONS(423), 1,
      sym_NEW,
    ACTIONS(433), 1,
      sym_based_base,
    ACTIONS(437), 1,
      sym_bit_string_length,
    ACTIONS(439), 1,
      sym_bit_string_base,
    STATE(704), 1,
      sym_unary_operator,
    STATE(705), 1,
      sym_sign,
    STATE(777), 1,
      sym__direct_name,
    STATE(1408), 1,
      sym__abstract_literal,
    STATE(1627), 1,
      sym_based_literal,
    STATE(1635), 1,
      sym__expression,
    STATE(2506), 1,
      sym_bit_string_literal,
    STATE(2508), 1,
      sym__primary,
    STATE(2510), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(431), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(536), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2509), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2536), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(425), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [21240] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(411), 1,
      anon_sym_LPAREN,
    ACTIONS(413), 1,
      sym_condition_conversion,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    ACTIONS(423), 1,
      sym_NEW,
    ACTIONS(433), 1,
      sym_based_base,
    ACTIONS(437), 1,
      sym_bit_string_length,
    ACTIONS(439), 1,
      sym_bit_string_base,
    STATE(704), 1,
      sym_unary_operator,
    STATE(705), 1,
      sym_sign,
    STATE(777), 1,
      sym__direct_name,
    STATE(1408), 1,
      sym__abstract_literal,
    STATE(1627), 1,
      sym_based_literal,
    STATE(1636), 1,
      sym__expression,
    STATE(2506), 1,
      sym_bit_string_literal,
    STATE(2508), 1,
      sym__primary,
    STATE(2510), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(431), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(537), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2509), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2536), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(425), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [21363] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1327), 1,
      anon_sym_LPAREN,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    ACTIONS(1333), 1,
      sym_NEW,
    ACTIONS(1339), 1,
      sym_based_base,
    ACTIONS(1343), 1,
      sym_bit_string_length,
    ACTIONS(1345), 1,
      sym_bit_string_base,
    STATE(493), 1,
      sym_unary_operator,
    STATE(494), 1,
      sym_sign,
    STATE(808), 1,
      sym__direct_name,
    STATE(1862), 1,
      sym__abstract_literal,
    STATE(2224), 1,
      sym_based_literal,
    STATE(2466), 1,
      sym__expression,
    STATE(2631), 1,
      sym_bit_string_literal,
    STATE(2649), 1,
      sym__primary,
    STATE(2663), 1,
      sym__literal,
    STATE(2685), 1,
      sym_simple_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1337), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(538), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2661), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1335), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [21488] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2428), 1,
      sym__expression,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(539), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [21611] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2319), 1,
      sym__expression,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(540), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [21734] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2328), 1,
      sym__expression,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(541), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [21857] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2329), 1,
      sym__expression,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(542), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [21980] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1229), 1,
      sym_based_literal,
    STATE(1756), 1,
      sym__expression,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(543), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [22103] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2468), 1,
      sym__expression,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    STATE(2772), 1,
      sym_simple_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(544), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [22228] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1425), 1,
      anon_sym_LPAREN,
    ACTIONS(1427), 1,
      sym_condition_conversion,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    ACTIONS(1433), 1,
      sym_NEW,
    ACTIONS(1439), 1,
      sym_based_base,
    ACTIONS(1443), 1,
      sym_bit_string_length,
    ACTIONS(1445), 1,
      sym_bit_string_base,
    STATE(714), 1,
      sym_unary_operator,
    STATE(715), 1,
      sym_sign,
    STATE(820), 1,
      sym__direct_name,
    STATE(1860), 1,
      sym__abstract_literal,
    STATE(1927), 1,
      sym_based_literal,
    STATE(1931), 1,
      sym__expression,
    STATE(2640), 1,
      sym_bit_string_literal,
    STATE(2652), 1,
      sym__literal,
    STATE(2677), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1437), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(545), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2643), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2688), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1435), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [22351] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1425), 1,
      anon_sym_LPAREN,
    ACTIONS(1427), 1,
      sym_condition_conversion,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    ACTIONS(1433), 1,
      sym_NEW,
    ACTIONS(1439), 1,
      sym_based_base,
    ACTIONS(1443), 1,
      sym_bit_string_length,
    ACTIONS(1445), 1,
      sym_bit_string_base,
    STATE(714), 1,
      sym_unary_operator,
    STATE(715), 1,
      sym_sign,
    STATE(820), 1,
      sym__direct_name,
    STATE(1860), 1,
      sym__abstract_literal,
    STATE(1927), 1,
      sym_based_literal,
    STATE(1936), 1,
      sym__expression,
    STATE(2640), 1,
      sym_bit_string_literal,
    STATE(2652), 1,
      sym__literal,
    STATE(2677), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1437), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(546), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2643), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2688), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1435), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [22474] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1425), 1,
      anon_sym_LPAREN,
    ACTIONS(1427), 1,
      sym_condition_conversion,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    ACTIONS(1433), 1,
      sym_NEW,
    ACTIONS(1439), 1,
      sym_based_base,
    ACTIONS(1443), 1,
      sym_bit_string_length,
    ACTIONS(1445), 1,
      sym_bit_string_base,
    STATE(714), 1,
      sym_unary_operator,
    STATE(715), 1,
      sym_sign,
    STATE(820), 1,
      sym__direct_name,
    STATE(1860), 1,
      sym__abstract_literal,
    STATE(1927), 1,
      sym_based_literal,
    STATE(1937), 1,
      sym__expression,
    STATE(2640), 1,
      sym_bit_string_literal,
    STATE(2652), 1,
      sym__literal,
    STATE(2677), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1437), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(547), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2643), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2688), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1435), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [22597] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1425), 1,
      anon_sym_LPAREN,
    ACTIONS(1427), 1,
      sym_condition_conversion,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    ACTIONS(1433), 1,
      sym_NEW,
    ACTIONS(1439), 1,
      sym_based_base,
    ACTIONS(1443), 1,
      sym_bit_string_length,
    ACTIONS(1445), 1,
      sym_bit_string_base,
    STATE(714), 1,
      sym_unary_operator,
    STATE(715), 1,
      sym_sign,
    STATE(820), 1,
      sym__direct_name,
    STATE(1860), 1,
      sym__abstract_literal,
    STATE(1927), 1,
      sym_based_literal,
    STATE(1938), 1,
      sym__expression,
    STATE(2640), 1,
      sym_bit_string_literal,
    STATE(2652), 1,
      sym__literal,
    STATE(2677), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1437), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(548), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2643), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2688), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1435), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [22720] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2449), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2751), 1,
      sym_simple_expression,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(549), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [22845] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1022), 1,
      sym__expression,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(550), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [22968] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1039), 1,
      sym__expression,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(551), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [23091] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1040), 1,
      sym__expression,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(552), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [23214] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1057), 1,
      sym__expression,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(553), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [23337] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1349), 1,
      anon_sym_LPAREN,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    ACTIONS(1355), 1,
      sym_NEW,
    ACTIONS(1361), 1,
      sym_based_base,
    ACTIONS(1365), 1,
      sym_bit_string_length,
    ACTIONS(1367), 1,
      sym_bit_string_base,
    STATE(504), 1,
      sym_unary_operator,
    STATE(505), 1,
      sym_sign,
    STATE(795), 1,
      sym__direct_name,
    STATE(1455), 1,
      sym__abstract_literal,
    STATE(1683), 1,
      sym_based_literal,
    STATE(2465), 1,
      sym__expression,
    STATE(2570), 1,
      sym_bit_string_literal,
    STATE(2574), 1,
      sym__literal,
    STATE(2609), 1,
      sym_simple_expression,
    STATE(2615), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1359), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(554), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2573), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1357), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [23462] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2308), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(555), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [23585] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2321), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(556), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [23708] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2323), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(557), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [23831] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2324), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(558), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [23954] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(428), 1,
      sym_unary_operator,
    STATE(429), 1,
      sym_sign,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2450), 1,
      sym__expression,
    STATE(2707), 1,
      sym_simple_expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(559), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [24079] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(423), 1,
      sym_unary_operator,
    STATE(424), 1,
      sym_sign,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2342), 1,
      sym__expression,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(560), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [24202] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(423), 1,
      sym_unary_operator,
    STATE(424), 1,
      sym_sign,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2347), 1,
      sym__expression,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(561), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [24325] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(423), 1,
      sym_unary_operator,
    STATE(424), 1,
      sym_sign,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2348), 1,
      sym__expression,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(562), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [24448] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(423), 1,
      sym_unary_operator,
    STATE(424), 1,
      sym_sign,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2349), 1,
      sym__expression,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(563), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [24571] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1369), 1,
      anon_sym_LPAREN,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    ACTIONS(1375), 1,
      sym_NEW,
    ACTIONS(1381), 1,
      sym_based_base,
    ACTIONS(1385), 1,
      sym_bit_string_length,
    ACTIONS(1387), 1,
      sym_bit_string_base,
    STATE(509), 1,
      sym_unary_operator,
    STATE(510), 1,
      sym_sign,
    STATE(767), 1,
      sym__direct_name,
    STATE(1289), 1,
      sym__abstract_literal,
    STATE(1467), 1,
      sym_based_literal,
    STATE(2435), 1,
      sym__expression,
    STATE(2513), 1,
      sym_bit_string_literal,
    STATE(2516), 1,
      sym__primary,
    STATE(2518), 1,
      sym__literal,
    STATE(2534), 1,
      sym_simple_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1379), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(564), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2517), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1377), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [24696] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(428), 1,
      sym_unary_operator,
    STATE(429), 1,
      sym_sign,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2356), 1,
      sym__expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(565), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2738), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [24819] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(428), 1,
      sym_unary_operator,
    STATE(429), 1,
      sym_sign,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2362), 1,
      sym__expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(566), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2738), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [24942] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(428), 1,
      sym_unary_operator,
    STATE(429), 1,
      sym_sign,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2363), 1,
      sym__expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(567), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2738), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [25065] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(428), 1,
      sym_unary_operator,
    STATE(429), 1,
      sym_sign,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2364), 1,
      sym__expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(568), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2738), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [25188] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1389), 1,
      anon_sym_LPAREN,
    ACTIONS(1391), 1,
      sym_NEW,
    ACTIONS(1397), 1,
      sym_based_base,
    ACTIONS(1399), 1,
      sym_bit_string_length,
    ACTIONS(1401), 1,
      sym_bit_string_base,
    STATE(514), 1,
      sym_unary_operator,
    STATE(515), 1,
      sym_sign,
    STATE(822), 1,
      sym__direct_name,
    STATE(1610), 1,
      sym__abstract_literal,
    STATE(1887), 1,
      sym_based_literal,
    STATE(2437), 1,
      sym__expression,
    STATE(2575), 1,
      sym_simple_expression,
    STATE(2598), 1,
      sym_bit_string_literal,
    STATE(2599), 1,
      sym__primary,
    STATE(2601), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1395), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(569), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2600), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1393), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [25313] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1223), 1,
      anon_sym_LPAREN,
    ACTIONS(1225), 1,
      sym_condition_conversion,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    ACTIONS(1231), 1,
      sym_NEW,
    ACTIONS(1237), 1,
      sym_based_base,
    ACTIONS(1241), 1,
      sym_bit_string_length,
    ACTIONS(1243), 1,
      sym_bit_string_base,
    STATE(433), 1,
      sym_unary_operator,
    STATE(434), 1,
      sym_sign,
    STATE(830), 1,
      sym__direct_name,
    STATE(1678), 1,
      sym__abstract_literal,
    STATE(2019), 1,
      sym_based_literal,
    STATE(2024), 1,
      sym__expression,
    STATE(2648), 1,
      sym_bit_string_literal,
    STATE(2659), 1,
      sym__primary,
    STATE(2667), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1235), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(570), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2662), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2650), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1233), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [25436] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1223), 1,
      anon_sym_LPAREN,
    ACTIONS(1225), 1,
      sym_condition_conversion,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    ACTIONS(1231), 1,
      sym_NEW,
    ACTIONS(1237), 1,
      sym_based_base,
    ACTIONS(1241), 1,
      sym_bit_string_length,
    ACTIONS(1243), 1,
      sym_bit_string_base,
    STATE(433), 1,
      sym_unary_operator,
    STATE(434), 1,
      sym_sign,
    STATE(830), 1,
      sym__direct_name,
    STATE(1678), 1,
      sym__abstract_literal,
    STATE(2019), 1,
      sym_based_literal,
    STATE(2035), 1,
      sym__expression,
    STATE(2648), 1,
      sym_bit_string_literal,
    STATE(2659), 1,
      sym__primary,
    STATE(2667), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1235), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(571), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2662), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2650), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1233), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [25559] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1223), 1,
      anon_sym_LPAREN,
    ACTIONS(1225), 1,
      sym_condition_conversion,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    ACTIONS(1231), 1,
      sym_NEW,
    ACTIONS(1237), 1,
      sym_based_base,
    ACTIONS(1241), 1,
      sym_bit_string_length,
    ACTIONS(1243), 1,
      sym_bit_string_base,
    STATE(433), 1,
      sym_unary_operator,
    STATE(434), 1,
      sym_sign,
    STATE(830), 1,
      sym__direct_name,
    STATE(1678), 1,
      sym__abstract_literal,
    STATE(2019), 1,
      sym_based_literal,
    STATE(2036), 1,
      sym__expression,
    STATE(2648), 1,
      sym_bit_string_literal,
    STATE(2659), 1,
      sym__primary,
    STATE(2667), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1235), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(572), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2662), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2650), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1233), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [25682] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1223), 1,
      anon_sym_LPAREN,
    ACTIONS(1225), 1,
      sym_condition_conversion,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    ACTIONS(1231), 1,
      sym_NEW,
    ACTIONS(1237), 1,
      sym_based_base,
    ACTIONS(1241), 1,
      sym_bit_string_length,
    ACTIONS(1243), 1,
      sym_bit_string_base,
    STATE(433), 1,
      sym_unary_operator,
    STATE(434), 1,
      sym_sign,
    STATE(830), 1,
      sym__direct_name,
    STATE(1678), 1,
      sym__abstract_literal,
    STATE(2019), 1,
      sym_based_literal,
    STATE(2037), 1,
      sym__expression,
    STATE(2648), 1,
      sym_bit_string_literal,
    STATE(2659), 1,
      sym__primary,
    STATE(2667), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1235), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(573), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2662), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2650), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1233), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [25805] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1405), 1,
      anon_sym_LPAREN,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    ACTIONS(1411), 1,
      sym_NEW,
    ACTIONS(1417), 1,
      sym_based_base,
    ACTIONS(1421), 1,
      sym_bit_string_length,
    ACTIONS(1423), 1,
      sym_bit_string_base,
    STATE(524), 1,
      sym_unary_operator,
    STATE(525), 1,
      sym_sign,
    STATE(769), 1,
      sym__direct_name,
    STATE(1331), 1,
      sym__abstract_literal,
    STATE(1445), 1,
      sym_based_literal,
    STATE(2441), 1,
      sym__expression,
    STATE(2477), 1,
      sym_bit_string_literal,
    STATE(2482), 1,
      sym__primary,
    STATE(2485), 1,
      sym__literal,
    STATE(2498), 1,
      sym_simple_expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1415), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(574), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2484), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1413), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [25930] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1249), 1,
      anon_sym_LPAREN,
    ACTIONS(1251), 1,
      sym_condition_conversion,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    ACTIONS(1257), 1,
      sym_NEW,
    ACTIONS(1263), 1,
      sym_based_base,
    ACTIONS(1267), 1,
      sym_bit_string_length,
    ACTIONS(1269), 1,
      sym_bit_string_base,
    STATE(443), 1,
      sym_unary_operator,
    STATE(444), 1,
      sym_sign,
    STATE(842), 1,
      sym__direct_name,
    STATE(2054), 1,
      sym__abstract_literal,
    STATE(2296), 1,
      sym_based_literal,
    STATE(2419), 1,
      sym__expression,
    STATE(2735), 1,
      sym_bit_string_literal,
    STATE(2746), 1,
      sym__primary,
    STATE(2753), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1261), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(575), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2752), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2734), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1259), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [26053] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1249), 1,
      anon_sym_LPAREN,
    ACTIONS(1251), 1,
      sym_condition_conversion,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    ACTIONS(1257), 1,
      sym_NEW,
    ACTIONS(1263), 1,
      sym_based_base,
    ACTIONS(1267), 1,
      sym_bit_string_length,
    ACTIONS(1269), 1,
      sym_bit_string_base,
    STATE(443), 1,
      sym_unary_operator,
    STATE(444), 1,
      sym_sign,
    STATE(842), 1,
      sym__direct_name,
    STATE(2054), 1,
      sym__abstract_literal,
    STATE(2292), 1,
      sym__expression,
    STATE(2296), 1,
      sym_based_literal,
    STATE(2735), 1,
      sym_bit_string_literal,
    STATE(2746), 1,
      sym__primary,
    STATE(2753), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1261), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(576), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2752), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2734), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1259), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [26176] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1249), 1,
      anon_sym_LPAREN,
    ACTIONS(1251), 1,
      sym_condition_conversion,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    ACTIONS(1257), 1,
      sym_NEW,
    ACTIONS(1263), 1,
      sym_based_base,
    ACTIONS(1267), 1,
      sym_bit_string_length,
    ACTIONS(1269), 1,
      sym_bit_string_base,
    STATE(443), 1,
      sym_unary_operator,
    STATE(444), 1,
      sym_sign,
    STATE(842), 1,
      sym__direct_name,
    STATE(2054), 1,
      sym__abstract_literal,
    STATE(2293), 1,
      sym__expression,
    STATE(2296), 1,
      sym_based_literal,
    STATE(2735), 1,
      sym_bit_string_literal,
    STATE(2746), 1,
      sym__primary,
    STATE(2753), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1261), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(577), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2752), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2734), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1259), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [26299] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1249), 1,
      anon_sym_LPAREN,
    ACTIONS(1251), 1,
      sym_condition_conversion,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    ACTIONS(1257), 1,
      sym_NEW,
    ACTIONS(1263), 1,
      sym_based_base,
    ACTIONS(1267), 1,
      sym_bit_string_length,
    ACTIONS(1269), 1,
      sym_bit_string_base,
    STATE(443), 1,
      sym_unary_operator,
    STATE(444), 1,
      sym_sign,
    STATE(842), 1,
      sym__direct_name,
    STATE(2054), 1,
      sym__abstract_literal,
    STATE(2294), 1,
      sym__expression,
    STATE(2296), 1,
      sym_based_literal,
    STATE(2735), 1,
      sym_bit_string_literal,
    STATE(2746), 1,
      sym__primary,
    STATE(2753), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1261), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(578), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2752), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2734), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1259), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [26422] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1271), 1,
      anon_sym_LPAREN,
    ACTIONS(1273), 1,
      sym_condition_conversion,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    ACTIONS(1279), 1,
      sym_NEW,
    ACTIONS(1285), 1,
      sym_based_base,
    ACTIONS(1289), 1,
      sym_bit_string_length,
    ACTIONS(1291), 1,
      sym_bit_string_base,
    STATE(448), 1,
      sym_unary_operator,
    STATE(449), 1,
      sym_sign,
    STATE(812), 1,
      sym__direct_name,
    STATE(1769), 1,
      sym__abstract_literal,
    STATE(2055), 1,
      sym__expression,
    STATE(2061), 1,
      sym_based_literal,
    STATE(2622), 1,
      sym__primary,
    STATE(2627), 1,
      sym__literal,
    STATE(2683), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(579), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2625), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2673), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1281), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [26545] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1271), 1,
      anon_sym_LPAREN,
    ACTIONS(1273), 1,
      sym_condition_conversion,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    ACTIONS(1279), 1,
      sym_NEW,
    ACTIONS(1285), 1,
      sym_based_base,
    ACTIONS(1289), 1,
      sym_bit_string_length,
    ACTIONS(1291), 1,
      sym_bit_string_base,
    STATE(448), 1,
      sym_unary_operator,
    STATE(449), 1,
      sym_sign,
    STATE(812), 1,
      sym__direct_name,
    STATE(1769), 1,
      sym__abstract_literal,
    STATE(2057), 1,
      sym__expression,
    STATE(2061), 1,
      sym_based_literal,
    STATE(2622), 1,
      sym__primary,
    STATE(2627), 1,
      sym__literal,
    STATE(2683), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(580), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2625), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2673), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1281), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [26668] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1271), 1,
      anon_sym_LPAREN,
    ACTIONS(1273), 1,
      sym_condition_conversion,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    ACTIONS(1279), 1,
      sym_NEW,
    ACTIONS(1285), 1,
      sym_based_base,
    ACTIONS(1289), 1,
      sym_bit_string_length,
    ACTIONS(1291), 1,
      sym_bit_string_base,
    STATE(448), 1,
      sym_unary_operator,
    STATE(449), 1,
      sym_sign,
    STATE(812), 1,
      sym__direct_name,
    STATE(1769), 1,
      sym__abstract_literal,
    STATE(2058), 1,
      sym__expression,
    STATE(2061), 1,
      sym_based_literal,
    STATE(2622), 1,
      sym__primary,
    STATE(2627), 1,
      sym__literal,
    STATE(2683), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(581), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2625), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2673), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1281), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [26791] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1271), 1,
      anon_sym_LPAREN,
    ACTIONS(1273), 1,
      sym_condition_conversion,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    ACTIONS(1279), 1,
      sym_NEW,
    ACTIONS(1285), 1,
      sym_based_base,
    ACTIONS(1289), 1,
      sym_bit_string_length,
    ACTIONS(1291), 1,
      sym_bit_string_base,
    STATE(448), 1,
      sym_unary_operator,
    STATE(449), 1,
      sym_sign,
    STATE(812), 1,
      sym__direct_name,
    STATE(1769), 1,
      sym__abstract_literal,
    STATE(2059), 1,
      sym__expression,
    STATE(2061), 1,
      sym_based_literal,
    STATE(2622), 1,
      sym__primary,
    STATE(2627), 1,
      sym__literal,
    STATE(2683), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(582), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2625), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2673), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1281), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [26914] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(854), 1,
      sym_condition_conversion,
    ACTIONS(1307), 1,
      anon_sym_LPAREN,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    ACTIONS(1313), 1,
      sym_NEW,
    ACTIONS(1319), 1,
      sym_based_base,
    ACTIONS(1323), 1,
      sym_bit_string_length,
    ACTIONS(1325), 1,
      sym_bit_string_base,
    STATE(488), 1,
      sym_unary_operator,
    STATE(489), 1,
      sym_sign,
    STATE(803), 1,
      sym__direct_name,
    STATE(1772), 1,
      sym__abstract_literal,
    STATE(2063), 1,
      sym__expression,
    STATE(2207), 1,
      sym_based_literal,
    STATE(2616), 1,
      sym__literal,
    STATE(2651), 1,
      sym_bit_string_literal,
    STATE(2653), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1317), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(583), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2654), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2621), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1315), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [27037] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(854), 1,
      sym_condition_conversion,
    ACTIONS(1307), 1,
      anon_sym_LPAREN,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    ACTIONS(1313), 1,
      sym_NEW,
    ACTIONS(1319), 1,
      sym_based_base,
    ACTIONS(1323), 1,
      sym_bit_string_length,
    ACTIONS(1325), 1,
      sym_bit_string_base,
    STATE(488), 1,
      sym_unary_operator,
    STATE(489), 1,
      sym_sign,
    STATE(803), 1,
      sym__direct_name,
    STATE(1772), 1,
      sym__abstract_literal,
    STATE(2067), 1,
      sym__expression,
    STATE(2207), 1,
      sym_based_literal,
    STATE(2616), 1,
      sym__literal,
    STATE(2651), 1,
      sym_bit_string_literal,
    STATE(2653), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1317), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(584), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2654), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2621), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1315), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [27160] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(854), 1,
      sym_condition_conversion,
    ACTIONS(1307), 1,
      anon_sym_LPAREN,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    ACTIONS(1313), 1,
      sym_NEW,
    ACTIONS(1319), 1,
      sym_based_base,
    ACTIONS(1323), 1,
      sym_bit_string_length,
    ACTIONS(1325), 1,
      sym_bit_string_base,
    STATE(488), 1,
      sym_unary_operator,
    STATE(489), 1,
      sym_sign,
    STATE(803), 1,
      sym__direct_name,
    STATE(1772), 1,
      sym__abstract_literal,
    STATE(2068), 1,
      sym__expression,
    STATE(2207), 1,
      sym_based_literal,
    STATE(2616), 1,
      sym__literal,
    STATE(2651), 1,
      sym_bit_string_literal,
    STATE(2653), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1317), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(585), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2654), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2621), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1315), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [27283] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(854), 1,
      sym_condition_conversion,
    ACTIONS(1307), 1,
      anon_sym_LPAREN,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    ACTIONS(1313), 1,
      sym_NEW,
    ACTIONS(1319), 1,
      sym_based_base,
    ACTIONS(1323), 1,
      sym_bit_string_length,
    ACTIONS(1325), 1,
      sym_bit_string_base,
    STATE(488), 1,
      sym_unary_operator,
    STATE(489), 1,
      sym_sign,
    STATE(803), 1,
      sym__direct_name,
    STATE(1772), 1,
      sym__abstract_literal,
    STATE(2069), 1,
      sym__expression,
    STATE(2207), 1,
      sym_based_literal,
    STATE(2616), 1,
      sym__literal,
    STATE(2651), 1,
      sym_bit_string_literal,
    STATE(2653), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1317), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(586), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2654), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2621), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1315), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [27406] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(942), 1,
      sym_condition_conversion,
    ACTIONS(1327), 1,
      anon_sym_LPAREN,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    ACTIONS(1333), 1,
      sym_NEW,
    ACTIONS(1339), 1,
      sym_based_base,
    ACTIONS(1343), 1,
      sym_bit_string_length,
    ACTIONS(1345), 1,
      sym_bit_string_base,
    STATE(493), 1,
      sym_unary_operator,
    STATE(494), 1,
      sym_sign,
    STATE(808), 1,
      sym__direct_name,
    STATE(1862), 1,
      sym__abstract_literal,
    STATE(2076), 1,
      sym__expression,
    STATE(2224), 1,
      sym_based_literal,
    STATE(2631), 1,
      sym_bit_string_literal,
    STATE(2649), 1,
      sym__primary,
    STATE(2663), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1337), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(587), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2661), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2681), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1335), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [27529] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(942), 1,
      sym_condition_conversion,
    ACTIONS(1327), 1,
      anon_sym_LPAREN,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    ACTIONS(1333), 1,
      sym_NEW,
    ACTIONS(1339), 1,
      sym_based_base,
    ACTIONS(1343), 1,
      sym_bit_string_length,
    ACTIONS(1345), 1,
      sym_bit_string_base,
    STATE(493), 1,
      sym_unary_operator,
    STATE(494), 1,
      sym_sign,
    STATE(808), 1,
      sym__direct_name,
    STATE(1862), 1,
      sym__abstract_literal,
    STATE(2077), 1,
      sym__expression,
    STATE(2224), 1,
      sym_based_literal,
    STATE(2631), 1,
      sym_bit_string_literal,
    STATE(2649), 1,
      sym__primary,
    STATE(2663), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1337), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(588), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2661), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2681), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1335), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [27652] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(942), 1,
      sym_condition_conversion,
    ACTIONS(1327), 1,
      anon_sym_LPAREN,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    ACTIONS(1333), 1,
      sym_NEW,
    ACTIONS(1339), 1,
      sym_based_base,
    ACTIONS(1343), 1,
      sym_bit_string_length,
    ACTIONS(1345), 1,
      sym_bit_string_base,
    STATE(493), 1,
      sym_unary_operator,
    STATE(494), 1,
      sym_sign,
    STATE(808), 1,
      sym__direct_name,
    STATE(1862), 1,
      sym__abstract_literal,
    STATE(2078), 1,
      sym__expression,
    STATE(2224), 1,
      sym_based_literal,
    STATE(2631), 1,
      sym_bit_string_literal,
    STATE(2649), 1,
      sym__primary,
    STATE(2663), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1337), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(589), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2661), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2681), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1335), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [27775] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(942), 1,
      sym_condition_conversion,
    ACTIONS(1327), 1,
      anon_sym_LPAREN,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    ACTIONS(1333), 1,
      sym_NEW,
    ACTIONS(1339), 1,
      sym_based_base,
    ACTIONS(1343), 1,
      sym_bit_string_length,
    ACTIONS(1345), 1,
      sym_bit_string_base,
    STATE(493), 1,
      sym_unary_operator,
    STATE(494), 1,
      sym_sign,
    STATE(808), 1,
      sym__direct_name,
    STATE(1862), 1,
      sym__abstract_literal,
    STATE(2079), 1,
      sym__expression,
    STATE(2224), 1,
      sym_based_literal,
    STATE(2631), 1,
      sym_bit_string_literal,
    STATE(2649), 1,
      sym__primary,
    STATE(2663), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1337), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(590), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2661), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2681), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1335), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [27898] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(423), 1,
      sym_unary_operator,
    STATE(424), 1,
      sym_sign,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2416), 1,
      sym__expression,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(591), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [28021] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    ACTIONS(1247), 1,
      sym_condition_conversion,
    STATE(439), 1,
      sym_sign,
    STATE(720), 1,
      sym_unary_operator,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(895), 1,
      sym__expression,
    STATE(902), 1,
      sym_based_literal,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(592), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1644), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [28144] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    ACTIONS(1247), 1,
      sym_condition_conversion,
    STATE(439), 1,
      sym_sign,
    STATE(720), 1,
      sym_unary_operator,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(898), 1,
      sym__expression,
    STATE(902), 1,
      sym_based_literal,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(593), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1644), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [28267] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    ACTIONS(1247), 1,
      sym_condition_conversion,
    STATE(439), 1,
      sym_sign,
    STATE(720), 1,
      sym_unary_operator,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(902), 1,
      sym_based_literal,
    STATE(919), 1,
      sym__expression,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(594), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1644), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [28390] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    ACTIONS(1247), 1,
      sym_condition_conversion,
    STATE(439), 1,
      sym_sign,
    STATE(720), 1,
      sym_unary_operator,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(902), 1,
      sym_based_literal,
    STATE(927), 1,
      sym__expression,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(595), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1644), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [28513] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1026), 1,
      sym_condition_conversion,
    ACTIONS(1349), 1,
      anon_sym_LPAREN,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    ACTIONS(1355), 1,
      sym_NEW,
    ACTIONS(1361), 1,
      sym_based_base,
    ACTIONS(1365), 1,
      sym_bit_string_length,
    ACTIONS(1367), 1,
      sym_bit_string_base,
    STATE(504), 1,
      sym_unary_operator,
    STATE(505), 1,
      sym_sign,
    STATE(795), 1,
      sym__direct_name,
    STATE(1455), 1,
      sym__abstract_literal,
    STATE(1683), 1,
      sym_based_literal,
    STATE(1698), 1,
      sym__expression,
    STATE(2570), 1,
      sym_bit_string_literal,
    STATE(2574), 1,
      sym__literal,
    STATE(2615), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1359), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(596), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2573), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2560), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1357), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [28636] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1026), 1,
      sym_condition_conversion,
    ACTIONS(1349), 1,
      anon_sym_LPAREN,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    ACTIONS(1355), 1,
      sym_NEW,
    ACTIONS(1361), 1,
      sym_based_base,
    ACTIONS(1365), 1,
      sym_bit_string_length,
    ACTIONS(1367), 1,
      sym_bit_string_base,
    STATE(504), 1,
      sym_unary_operator,
    STATE(505), 1,
      sym_sign,
    STATE(795), 1,
      sym__direct_name,
    STATE(1455), 1,
      sym__abstract_literal,
    STATE(1683), 1,
      sym_based_literal,
    STATE(1714), 1,
      sym__expression,
    STATE(2570), 1,
      sym_bit_string_literal,
    STATE(2574), 1,
      sym__literal,
    STATE(2615), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1359), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(597), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2573), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2560), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1357), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [28759] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1026), 1,
      sym_condition_conversion,
    ACTIONS(1349), 1,
      anon_sym_LPAREN,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    ACTIONS(1355), 1,
      sym_NEW,
    ACTIONS(1361), 1,
      sym_based_base,
    ACTIONS(1365), 1,
      sym_bit_string_length,
    ACTIONS(1367), 1,
      sym_bit_string_base,
    STATE(504), 1,
      sym_unary_operator,
    STATE(505), 1,
      sym_sign,
    STATE(795), 1,
      sym__direct_name,
    STATE(1455), 1,
      sym__abstract_literal,
    STATE(1683), 1,
      sym_based_literal,
    STATE(1719), 1,
      sym__expression,
    STATE(2570), 1,
      sym_bit_string_literal,
    STATE(2574), 1,
      sym__literal,
    STATE(2615), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1359), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(598), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2573), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2560), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1357), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [28882] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1026), 1,
      sym_condition_conversion,
    ACTIONS(1349), 1,
      anon_sym_LPAREN,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    ACTIONS(1355), 1,
      sym_NEW,
    ACTIONS(1361), 1,
      sym_based_base,
    ACTIONS(1365), 1,
      sym_bit_string_length,
    ACTIONS(1367), 1,
      sym_bit_string_base,
    STATE(504), 1,
      sym_unary_operator,
    STATE(505), 1,
      sym_sign,
    STATE(795), 1,
      sym__direct_name,
    STATE(1455), 1,
      sym__abstract_literal,
    STATE(1683), 1,
      sym_based_literal,
    STATE(1808), 1,
      sym__expression,
    STATE(2570), 1,
      sym_bit_string_literal,
    STATE(2574), 1,
      sym__literal,
    STATE(2615), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1359), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(599), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2573), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2560), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1357), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [29005] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(900), 1,
      sym_condition_conversion,
    ACTIONS(1369), 1,
      anon_sym_LPAREN,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    ACTIONS(1375), 1,
      sym_NEW,
    ACTIONS(1381), 1,
      sym_based_base,
    ACTIONS(1385), 1,
      sym_bit_string_length,
    ACTIONS(1387), 1,
      sym_bit_string_base,
    STATE(509), 1,
      sym_unary_operator,
    STATE(510), 1,
      sym_sign,
    STATE(767), 1,
      sym__direct_name,
    STATE(1289), 1,
      sym__abstract_literal,
    STATE(1467), 1,
      sym_based_literal,
    STATE(1524), 1,
      sym__expression,
    STATE(2513), 1,
      sym_bit_string_literal,
    STATE(2516), 1,
      sym__primary,
    STATE(2518), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1379), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(600), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2517), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2491), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1377), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [29128] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(900), 1,
      sym_condition_conversion,
    ACTIONS(1369), 1,
      anon_sym_LPAREN,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    ACTIONS(1375), 1,
      sym_NEW,
    ACTIONS(1381), 1,
      sym_based_base,
    ACTIONS(1385), 1,
      sym_bit_string_length,
    ACTIONS(1387), 1,
      sym_bit_string_base,
    STATE(509), 1,
      sym_unary_operator,
    STATE(510), 1,
      sym_sign,
    STATE(767), 1,
      sym__direct_name,
    STATE(1289), 1,
      sym__abstract_literal,
    STATE(1429), 1,
      sym__expression,
    STATE(1467), 1,
      sym_based_literal,
    STATE(2513), 1,
      sym_bit_string_literal,
    STATE(2516), 1,
      sym__primary,
    STATE(2518), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1379), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(601), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2517), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2491), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1377), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [29251] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(900), 1,
      sym_condition_conversion,
    ACTIONS(1369), 1,
      anon_sym_LPAREN,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    ACTIONS(1375), 1,
      sym_NEW,
    ACTIONS(1381), 1,
      sym_based_base,
    ACTIONS(1385), 1,
      sym_bit_string_length,
    ACTIONS(1387), 1,
      sym_bit_string_base,
    STATE(509), 1,
      sym_unary_operator,
    STATE(510), 1,
      sym_sign,
    STATE(767), 1,
      sym__direct_name,
    STATE(1289), 1,
      sym__abstract_literal,
    STATE(1432), 1,
      sym__expression,
    STATE(1467), 1,
      sym_based_literal,
    STATE(2513), 1,
      sym_bit_string_literal,
    STATE(2516), 1,
      sym__primary,
    STATE(2518), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1379), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(602), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2517), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2491), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1377), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [29374] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(900), 1,
      sym_condition_conversion,
    ACTIONS(1369), 1,
      anon_sym_LPAREN,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    ACTIONS(1375), 1,
      sym_NEW,
    ACTIONS(1381), 1,
      sym_based_base,
    ACTIONS(1385), 1,
      sym_bit_string_length,
    ACTIONS(1387), 1,
      sym_bit_string_base,
    STATE(509), 1,
      sym_unary_operator,
    STATE(510), 1,
      sym_sign,
    STATE(767), 1,
      sym__direct_name,
    STATE(1289), 1,
      sym__abstract_literal,
    STATE(1435), 1,
      sym__expression,
    STATE(1467), 1,
      sym_based_literal,
    STATE(2513), 1,
      sym_bit_string_literal,
    STATE(2516), 1,
      sym__primary,
    STATE(2518), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1379), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(603), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2517), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2491), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1377), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [29497] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    ACTIONS(860), 1,
      sym_NEW,
    ACTIONS(866), 1,
      sym_based_base,
    ACTIONS(870), 1,
      sym_bit_string_length,
    ACTIONS(872), 1,
      sym_bit_string_base,
    ACTIONS(984), 1,
      anon_sym_LPAREN,
    ACTIONS(1293), 1,
      sym_condition_conversion,
    STATE(453), 1,
      sym_unary_operator,
    STATE(454), 1,
      sym_sign,
    STATE(772), 1,
      sym__direct_name,
    STATE(1407), 1,
      sym__abstract_literal,
    STATE(1448), 1,
      sym__expression,
    STATE(1615), 1,
      sym_based_literal,
    STATE(2480), 1,
      sym__literal,
    STATE(2530), 1,
      sym_bit_string_literal,
    STATE(2535), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(864), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(604), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2507), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2481), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(862), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [29620] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    ACTIONS(860), 1,
      sym_NEW,
    ACTIONS(866), 1,
      sym_based_base,
    ACTIONS(870), 1,
      sym_bit_string_length,
    ACTIONS(872), 1,
      sym_bit_string_base,
    ACTIONS(984), 1,
      anon_sym_LPAREN,
    ACTIONS(1293), 1,
      sym_condition_conversion,
    STATE(453), 1,
      sym_unary_operator,
    STATE(454), 1,
      sym_sign,
    STATE(772), 1,
      sym__direct_name,
    STATE(1407), 1,
      sym__abstract_literal,
    STATE(1459), 1,
      sym__expression,
    STATE(1615), 1,
      sym_based_literal,
    STATE(2480), 1,
      sym__literal,
    STATE(2530), 1,
      sym_bit_string_literal,
    STATE(2535), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(864), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(605), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2507), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2481), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(862), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [29743] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    ACTIONS(860), 1,
      sym_NEW,
    ACTIONS(866), 1,
      sym_based_base,
    ACTIONS(870), 1,
      sym_bit_string_length,
    ACTIONS(872), 1,
      sym_bit_string_base,
    ACTIONS(984), 1,
      anon_sym_LPAREN,
    ACTIONS(1293), 1,
      sym_condition_conversion,
    STATE(453), 1,
      sym_unary_operator,
    STATE(454), 1,
      sym_sign,
    STATE(772), 1,
      sym__direct_name,
    STATE(1407), 1,
      sym__abstract_literal,
    STATE(1461), 1,
      sym__expression,
    STATE(1615), 1,
      sym_based_literal,
    STATE(2480), 1,
      sym__literal,
    STATE(2530), 1,
      sym_bit_string_literal,
    STATE(2535), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(864), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(606), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2507), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2481), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(862), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [29866] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    ACTIONS(860), 1,
      sym_NEW,
    ACTIONS(866), 1,
      sym_based_base,
    ACTIONS(870), 1,
      sym_bit_string_length,
    ACTIONS(872), 1,
      sym_bit_string_base,
    ACTIONS(984), 1,
      anon_sym_LPAREN,
    ACTIONS(1293), 1,
      sym_condition_conversion,
    STATE(453), 1,
      sym_unary_operator,
    STATE(454), 1,
      sym_sign,
    STATE(772), 1,
      sym__direct_name,
    STATE(1407), 1,
      sym__abstract_literal,
    STATE(1466), 1,
      sym__expression,
    STATE(1615), 1,
      sym_based_literal,
    STATE(2480), 1,
      sym__literal,
    STATE(2530), 1,
      sym_bit_string_literal,
    STATE(2535), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(864), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(607), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2507), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2481), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(862), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [29989] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    ACTIONS(948), 1,
      sym_NEW,
    ACTIONS(954), 1,
      sym_based_base,
    ACTIONS(958), 1,
      sym_bit_string_length,
    ACTIONS(960), 1,
      sym_bit_string_base,
    ACTIONS(1295), 1,
      sym_condition_conversion,
    STATE(458), 1,
      sym_unary_operator,
    STATE(459), 1,
      sym_sign,
    STATE(779), 1,
      sym__direct_name,
    STATE(1413), 1,
      sym__abstract_literal,
    STATE(1528), 1,
      sym__expression,
    STATE(1642), 1,
      sym_based_literal,
    STATE(2496), 1,
      sym_bit_string_literal,
    STATE(2497), 1,
      sym__primary,
    STATE(2501), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(952), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(608), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2500), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2521), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(950), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [30112] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    ACTIONS(948), 1,
      sym_NEW,
    ACTIONS(954), 1,
      sym_based_base,
    ACTIONS(958), 1,
      sym_bit_string_length,
    ACTIONS(960), 1,
      sym_bit_string_base,
    ACTIONS(1295), 1,
      sym_condition_conversion,
    STATE(458), 1,
      sym_unary_operator,
    STATE(459), 1,
      sym_sign,
    STATE(779), 1,
      sym__direct_name,
    STATE(1413), 1,
      sym__abstract_literal,
    STATE(1533), 1,
      sym__expression,
    STATE(1642), 1,
      sym_based_literal,
    STATE(2496), 1,
      sym_bit_string_literal,
    STATE(2497), 1,
      sym__primary,
    STATE(2501), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(952), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(609), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2500), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2521), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(950), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [30235] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    ACTIONS(948), 1,
      sym_NEW,
    ACTIONS(954), 1,
      sym_based_base,
    ACTIONS(958), 1,
      sym_bit_string_length,
    ACTIONS(960), 1,
      sym_bit_string_base,
    ACTIONS(1295), 1,
      sym_condition_conversion,
    STATE(458), 1,
      sym_unary_operator,
    STATE(459), 1,
      sym_sign,
    STATE(779), 1,
      sym__direct_name,
    STATE(1413), 1,
      sym__abstract_literal,
    STATE(1534), 1,
      sym__expression,
    STATE(1642), 1,
      sym_based_literal,
    STATE(2496), 1,
      sym_bit_string_literal,
    STATE(2497), 1,
      sym__primary,
    STATE(2501), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(952), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(610), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2500), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2521), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(950), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [30358] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(940), 1,
      anon_sym_LPAREN,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    ACTIONS(948), 1,
      sym_NEW,
    ACTIONS(954), 1,
      sym_based_base,
    ACTIONS(958), 1,
      sym_bit_string_length,
    ACTIONS(960), 1,
      sym_bit_string_base,
    ACTIONS(1295), 1,
      sym_condition_conversion,
    STATE(458), 1,
      sym_unary_operator,
    STATE(459), 1,
      sym_sign,
    STATE(779), 1,
      sym__direct_name,
    STATE(1413), 1,
      sym__abstract_literal,
    STATE(1536), 1,
      sym__expression,
    STATE(1642), 1,
      sym_based_literal,
    STATE(2496), 1,
      sym_bit_string_literal,
    STATE(2497), 1,
      sym__primary,
    STATE(2501), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(952), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(611), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2500), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2521), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(950), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [30481] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2401), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(612), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [30604] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(988), 1,
      sym_condition_conversion,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1389), 1,
      anon_sym_LPAREN,
    ACTIONS(1391), 1,
      sym_NEW,
    ACTIONS(1397), 1,
      sym_based_base,
    ACTIONS(1399), 1,
      sym_bit_string_length,
    ACTIONS(1401), 1,
      sym_bit_string_base,
    STATE(514), 1,
      sym_unary_operator,
    STATE(515), 1,
      sym_sign,
    STATE(822), 1,
      sym__direct_name,
    STATE(1610), 1,
      sym__abstract_literal,
    STATE(1725), 1,
      sym__expression,
    STATE(1887), 1,
      sym_based_literal,
    STATE(2598), 1,
      sym_bit_string_literal,
    STATE(2599), 1,
      sym__primary,
    STATE(2601), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1395), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(613), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2600), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2565), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1393), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [30727] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(988), 1,
      sym_condition_conversion,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1389), 1,
      anon_sym_LPAREN,
    ACTIONS(1391), 1,
      sym_NEW,
    ACTIONS(1397), 1,
      sym_based_base,
    ACTIONS(1399), 1,
      sym_bit_string_length,
    ACTIONS(1401), 1,
      sym_bit_string_base,
    STATE(514), 1,
      sym_unary_operator,
    STATE(515), 1,
      sym_sign,
    STATE(822), 1,
      sym__direct_name,
    STATE(1610), 1,
      sym__abstract_literal,
    STATE(1750), 1,
      sym__expression,
    STATE(1887), 1,
      sym_based_literal,
    STATE(2598), 1,
      sym_bit_string_literal,
    STATE(2599), 1,
      sym__primary,
    STATE(2601), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1395), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(614), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2600), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2565), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1393), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [30850] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(988), 1,
      sym_condition_conversion,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1389), 1,
      anon_sym_LPAREN,
    ACTIONS(1391), 1,
      sym_NEW,
    ACTIONS(1397), 1,
      sym_based_base,
    ACTIONS(1399), 1,
      sym_bit_string_length,
    ACTIONS(1401), 1,
      sym_bit_string_base,
    STATE(514), 1,
      sym_unary_operator,
    STATE(515), 1,
      sym_sign,
    STATE(822), 1,
      sym__direct_name,
    STATE(1610), 1,
      sym__abstract_literal,
    STATE(1753), 1,
      sym__expression,
    STATE(1887), 1,
      sym_based_literal,
    STATE(2598), 1,
      sym_bit_string_literal,
    STATE(2599), 1,
      sym__primary,
    STATE(2601), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1395), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(615), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2600), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2565), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1393), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [30973] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(988), 1,
      sym_condition_conversion,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1389), 1,
      anon_sym_LPAREN,
    ACTIONS(1391), 1,
      sym_NEW,
    ACTIONS(1397), 1,
      sym_based_base,
    ACTIONS(1399), 1,
      sym_bit_string_length,
    ACTIONS(1401), 1,
      sym_bit_string_base,
    STATE(514), 1,
      sym_unary_operator,
    STATE(515), 1,
      sym_sign,
    STATE(822), 1,
      sym__direct_name,
    STATE(1610), 1,
      sym__abstract_literal,
    STATE(1755), 1,
      sym__expression,
    STATE(1887), 1,
      sym_based_literal,
    STATE(2598), 1,
      sym_bit_string_literal,
    STATE(2599), 1,
      sym__primary,
    STATE(2601), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1395), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(616), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2600), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2565), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1393), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [31096] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(876), 1,
      anon_sym_LPAREN,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    ACTIONS(884), 1,
      sym_NEW,
    ACTIONS(890), 1,
      sym_based_base,
    ACTIONS(894), 1,
      sym_bit_string_length,
    ACTIONS(896), 1,
      sym_bit_string_base,
    ACTIONS(1297), 1,
      sym_condition_conversion,
    STATE(463), 1,
      sym_unary_operator,
    STATE(464), 1,
      sym_sign,
    STATE(786), 1,
      sym__direct_name,
    STATE(1609), 1,
      sym__abstract_literal,
    STATE(1675), 1,
      sym_based_literal,
    STATE(1777), 1,
      sym__expression,
    STATE(2559), 1,
      sym_bit_string_literal,
    STATE(2561), 1,
      sym__primary,
    STATE(2563), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(888), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(617), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2562), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2592), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(886), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [31219] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(876), 1,
      anon_sym_LPAREN,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    ACTIONS(884), 1,
      sym_NEW,
    ACTIONS(890), 1,
      sym_based_base,
    ACTIONS(894), 1,
      sym_bit_string_length,
    ACTIONS(896), 1,
      sym_bit_string_base,
    ACTIONS(1297), 1,
      sym_condition_conversion,
    STATE(463), 1,
      sym_unary_operator,
    STATE(464), 1,
      sym_sign,
    STATE(786), 1,
      sym__direct_name,
    STATE(1609), 1,
      sym__abstract_literal,
    STATE(1675), 1,
      sym_based_literal,
    STATE(1800), 1,
      sym__expression,
    STATE(2559), 1,
      sym_bit_string_literal,
    STATE(2561), 1,
      sym__primary,
    STATE(2563), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(888), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(618), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2562), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2592), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(886), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [31342] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(876), 1,
      anon_sym_LPAREN,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    ACTIONS(884), 1,
      sym_NEW,
    ACTIONS(890), 1,
      sym_based_base,
    ACTIONS(894), 1,
      sym_bit_string_length,
    ACTIONS(896), 1,
      sym_bit_string_base,
    ACTIONS(1297), 1,
      sym_condition_conversion,
    STATE(463), 1,
      sym_unary_operator,
    STATE(464), 1,
      sym_sign,
    STATE(786), 1,
      sym__direct_name,
    STATE(1609), 1,
      sym__abstract_literal,
    STATE(1675), 1,
      sym_based_literal,
    STATE(1805), 1,
      sym__expression,
    STATE(2559), 1,
      sym_bit_string_literal,
    STATE(2561), 1,
      sym__primary,
    STATE(2563), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(888), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(619), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2562), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2592), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(886), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [31465] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(876), 1,
      anon_sym_LPAREN,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    ACTIONS(884), 1,
      sym_NEW,
    ACTIONS(890), 1,
      sym_based_base,
    ACTIONS(894), 1,
      sym_bit_string_length,
    ACTIONS(896), 1,
      sym_bit_string_base,
    ACTIONS(1297), 1,
      sym_condition_conversion,
    STATE(463), 1,
      sym_unary_operator,
    STATE(464), 1,
      sym_sign,
    STATE(786), 1,
      sym__direct_name,
    STATE(1609), 1,
      sym__abstract_literal,
    STATE(1675), 1,
      sym_based_literal,
    STATE(1812), 1,
      sym__expression,
    STATE(2559), 1,
      sym_bit_string_literal,
    STATE(2561), 1,
      sym__primary,
    STATE(2563), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(888), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(620), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2562), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2592), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(886), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [31588] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(962), 1,
      anon_sym_LPAREN,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    ACTIONS(970), 1,
      sym_NEW,
    ACTIONS(976), 1,
      sym_based_base,
    ACTIONS(980), 1,
      sym_bit_string_length,
    ACTIONS(982), 1,
      sym_bit_string_base,
    ACTIONS(1299), 1,
      sym_condition_conversion,
    STATE(468), 1,
      sym_unary_operator,
    STATE(469), 1,
      sym_sign,
    STATE(793), 1,
      sym__direct_name,
    STATE(1497), 1,
      sym__abstract_literal,
    STATE(1680), 1,
      sym_based_literal,
    STATE(1889), 1,
      sym__expression,
    STATE(2554), 1,
      sym_bit_string_literal,
    STATE(2557), 1,
      sym__primary,
    STATE(2564), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(974), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(621), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2558), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2544), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(972), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [31711] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(962), 1,
      anon_sym_LPAREN,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    ACTIONS(970), 1,
      sym_NEW,
    ACTIONS(976), 1,
      sym_based_base,
    ACTIONS(980), 1,
      sym_bit_string_length,
    ACTIONS(982), 1,
      sym_bit_string_base,
    ACTIONS(1299), 1,
      sym_condition_conversion,
    STATE(468), 1,
      sym_unary_operator,
    STATE(469), 1,
      sym_sign,
    STATE(793), 1,
      sym__direct_name,
    STATE(1497), 1,
      sym__abstract_literal,
    STATE(1680), 1,
      sym_based_literal,
    STATE(1718), 1,
      sym__expression,
    STATE(2554), 1,
      sym_bit_string_literal,
    STATE(2557), 1,
      sym__primary,
    STATE(2564), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(974), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(622), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2558), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2544), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(972), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [31834] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(962), 1,
      anon_sym_LPAREN,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    ACTIONS(970), 1,
      sym_NEW,
    ACTIONS(976), 1,
      sym_based_base,
    ACTIONS(980), 1,
      sym_bit_string_length,
    ACTIONS(982), 1,
      sym_bit_string_base,
    ACTIONS(1299), 1,
      sym_condition_conversion,
    STATE(468), 1,
      sym_unary_operator,
    STATE(469), 1,
      sym_sign,
    STATE(793), 1,
      sym__direct_name,
    STATE(1497), 1,
      sym__abstract_literal,
    STATE(1680), 1,
      sym_based_literal,
    STATE(1883), 1,
      sym__expression,
    STATE(2554), 1,
      sym_bit_string_literal,
    STATE(2557), 1,
      sym__primary,
    STATE(2564), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(974), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(623), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2558), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2544), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(972), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [31957] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(962), 1,
      anon_sym_LPAREN,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    ACTIONS(970), 1,
      sym_NEW,
    ACTIONS(976), 1,
      sym_based_base,
    ACTIONS(980), 1,
      sym_bit_string_length,
    ACTIONS(982), 1,
      sym_bit_string_base,
    ACTIONS(1299), 1,
      sym_condition_conversion,
    STATE(468), 1,
      sym_unary_operator,
    STATE(469), 1,
      sym_sign,
    STATE(793), 1,
      sym__direct_name,
    STATE(1497), 1,
      sym__abstract_literal,
    STATE(1680), 1,
      sym_based_literal,
    STATE(1733), 1,
      sym__expression,
    STATE(2554), 1,
      sym_bit_string_literal,
    STATE(2557), 1,
      sym__primary,
    STATE(2564), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(974), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(624), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2558), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2544), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(972), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [32080] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1024), 1,
      anon_sym_LPAREN,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    ACTIONS(1032), 1,
      sym_NEW,
    ACTIONS(1038), 1,
      sym_based_base,
    ACTIONS(1042), 1,
      sym_bit_string_length,
    ACTIONS(1044), 1,
      sym_bit_string_base,
    ACTIONS(1301), 1,
      sym_condition_conversion,
    STATE(473), 1,
      sym_unary_operator,
    STATE(474), 1,
      sym_sign,
    STATE(750), 1,
      sym__direct_name,
    STATE(1140), 1,
      sym__abstract_literal,
    STATE(1261), 1,
      sym__expression,
    STATE(1326), 1,
      sym_based_literal,
    STATE(2436), 1,
      sym_bit_string_literal,
    STATE(2438), 1,
      sym__primary,
    STATE(2440), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1036), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(625), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2439), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2453), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1034), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [32203] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1024), 1,
      anon_sym_LPAREN,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    ACTIONS(1032), 1,
      sym_NEW,
    ACTIONS(1038), 1,
      sym_based_base,
    ACTIONS(1042), 1,
      sym_bit_string_length,
    ACTIONS(1044), 1,
      sym_bit_string_base,
    ACTIONS(1301), 1,
      sym_condition_conversion,
    STATE(473), 1,
      sym_unary_operator,
    STATE(474), 1,
      sym_sign,
    STATE(750), 1,
      sym__direct_name,
    STATE(1140), 1,
      sym__abstract_literal,
    STATE(1267), 1,
      sym__expression,
    STATE(1326), 1,
      sym_based_literal,
    STATE(2436), 1,
      sym_bit_string_literal,
    STATE(2438), 1,
      sym__primary,
    STATE(2440), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1036), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(626), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2439), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2453), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1034), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [32326] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1024), 1,
      anon_sym_LPAREN,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    ACTIONS(1032), 1,
      sym_NEW,
    ACTIONS(1038), 1,
      sym_based_base,
    ACTIONS(1042), 1,
      sym_bit_string_length,
    ACTIONS(1044), 1,
      sym_bit_string_base,
    ACTIONS(1301), 1,
      sym_condition_conversion,
    STATE(473), 1,
      sym_unary_operator,
    STATE(474), 1,
      sym_sign,
    STATE(750), 1,
      sym__direct_name,
    STATE(1140), 1,
      sym__abstract_literal,
    STATE(1268), 1,
      sym__expression,
    STATE(1326), 1,
      sym_based_literal,
    STATE(2436), 1,
      sym_bit_string_literal,
    STATE(2438), 1,
      sym__primary,
    STATE(2440), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1036), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(627), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2439), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2453), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1034), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [32449] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1024), 1,
      anon_sym_LPAREN,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    ACTIONS(1032), 1,
      sym_NEW,
    ACTIONS(1038), 1,
      sym_based_base,
    ACTIONS(1042), 1,
      sym_bit_string_length,
    ACTIONS(1044), 1,
      sym_bit_string_base,
    ACTIONS(1301), 1,
      sym_condition_conversion,
    STATE(473), 1,
      sym_unary_operator,
    STATE(474), 1,
      sym_sign,
    STATE(750), 1,
      sym__direct_name,
    STATE(1140), 1,
      sym__abstract_literal,
    STATE(1269), 1,
      sym__expression,
    STATE(1326), 1,
      sym_based_literal,
    STATE(2436), 1,
      sym_bit_string_literal,
    STATE(2438), 1,
      sym__primary,
    STATE(2440), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1036), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(628), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2439), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2453), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1034), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [32572] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1046), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    ACTIONS(1054), 1,
      sym_NEW,
    ACTIONS(1060), 1,
      sym_based_base,
    ACTIONS(1064), 1,
      sym_bit_string_length,
    ACTIONS(1066), 1,
      sym_bit_string_base,
    ACTIONS(1303), 1,
      sym_condition_conversion,
    STATE(478), 1,
      sym_unary_operator,
    STATE(479), 1,
      sym_sign,
    STATE(788), 1,
      sym__direct_name,
    STATE(1526), 1,
      sym__abstract_literal,
    STATE(1709), 1,
      sym_based_literal,
    STATE(1711), 1,
      sym__expression,
    STATE(2541), 1,
      sym__primary,
    STATE(2546), 1,
      sym__literal,
    STATE(2572), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1058), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(629), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2543), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2571), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1056), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [32695] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1046), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    ACTIONS(1054), 1,
      sym_NEW,
    ACTIONS(1060), 1,
      sym_based_base,
    ACTIONS(1064), 1,
      sym_bit_string_length,
    ACTIONS(1066), 1,
      sym_bit_string_base,
    ACTIONS(1303), 1,
      sym_condition_conversion,
    STATE(478), 1,
      sym_unary_operator,
    STATE(479), 1,
      sym_sign,
    STATE(788), 1,
      sym__direct_name,
    STATE(1526), 1,
      sym__abstract_literal,
    STATE(1709), 1,
      sym_based_literal,
    STATE(1723), 1,
      sym__expression,
    STATE(2541), 1,
      sym__primary,
    STATE(2546), 1,
      sym__literal,
    STATE(2572), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1058), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(630), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2543), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2571), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1056), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [32818] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1046), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    ACTIONS(1054), 1,
      sym_NEW,
    ACTIONS(1060), 1,
      sym_based_base,
    ACTIONS(1064), 1,
      sym_bit_string_length,
    ACTIONS(1066), 1,
      sym_bit_string_base,
    ACTIONS(1303), 1,
      sym_condition_conversion,
    STATE(478), 1,
      sym_unary_operator,
    STATE(479), 1,
      sym_sign,
    STATE(788), 1,
      sym__direct_name,
    STATE(1526), 1,
      sym__abstract_literal,
    STATE(1709), 1,
      sym_based_literal,
    STATE(1771), 1,
      sym__expression,
    STATE(2541), 1,
      sym__primary,
    STATE(2546), 1,
      sym__literal,
    STATE(2572), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1058), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(631), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2543), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2571), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1056), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [32941] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1046), 1,
      anon_sym_LPAREN,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    ACTIONS(1054), 1,
      sym_NEW,
    ACTIONS(1060), 1,
      sym_based_base,
    ACTIONS(1064), 1,
      sym_bit_string_length,
    ACTIONS(1066), 1,
      sym_bit_string_base,
    ACTIONS(1303), 1,
      sym_condition_conversion,
    STATE(478), 1,
      sym_unary_operator,
    STATE(479), 1,
      sym_sign,
    STATE(788), 1,
      sym__direct_name,
    STATE(1526), 1,
      sym__abstract_literal,
    STATE(1709), 1,
      sym_based_literal,
    STATE(1774), 1,
      sym__expression,
    STATE(2541), 1,
      sym__primary,
    STATE(2546), 1,
      sym__literal,
    STATE(2572), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1058), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(632), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2543), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2571), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1056), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [33064] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1048), 1,
      sym_condition_conversion,
    ACTIONS(1203), 1,
      anon_sym_LPAREN,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    ACTIONS(1209), 1,
      sym_NEW,
    ACTIONS(1215), 1,
      sym_based_base,
    ACTIONS(1219), 1,
      sym_bit_string_length,
    ACTIONS(1221), 1,
      sym_bit_string_base,
    STATE(428), 1,
      sym_unary_operator,
    STATE(429), 1,
      sym_sign,
    STATE(837), 1,
      sym__direct_name,
    STATE(2014), 1,
      sym__abstract_literal,
    STATE(2354), 1,
      sym_based_literal,
    STATE(2388), 1,
      sym__expression,
    STATE(2737), 1,
      sym_bit_string_literal,
    STATE(2740), 1,
      sym__primary,
    STATE(2742), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1213), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(633), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2741), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2738), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1211), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [33187] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      anon_sym_LPAREN,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    ACTIONS(906), 1,
      sym_NEW,
    ACTIONS(912), 1,
      sym_based_base,
    ACTIONS(916), 1,
      sym_bit_string_length,
    ACTIONS(918), 1,
      sym_bit_string_base,
    ACTIONS(1305), 1,
      sym_condition_conversion,
    STATE(483), 1,
      sym_unary_operator,
    STATE(484), 1,
      sym_sign,
    STATE(743), 1,
      sym__direct_name,
    STATE(1060), 1,
      sym__abstract_literal,
    STATE(1162), 1,
      sym__expression,
    STATE(1236), 1,
      sym_based_literal,
    STATE(2366), 1,
      sym_bit_string_literal,
    STATE(2367), 1,
      sym__primary,
    STATE(2369), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(910), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(634), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2368), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2417), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(908), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [33310] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      anon_sym_LPAREN,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    ACTIONS(906), 1,
      sym_NEW,
    ACTIONS(912), 1,
      sym_based_base,
    ACTIONS(916), 1,
      sym_bit_string_length,
    ACTIONS(918), 1,
      sym_bit_string_base,
    ACTIONS(1305), 1,
      sym_condition_conversion,
    STATE(483), 1,
      sym_unary_operator,
    STATE(484), 1,
      sym_sign,
    STATE(743), 1,
      sym__direct_name,
    STATE(1060), 1,
      sym__abstract_literal,
    STATE(1123), 1,
      sym__expression,
    STATE(1236), 1,
      sym_based_literal,
    STATE(2366), 1,
      sym_bit_string_literal,
    STATE(2367), 1,
      sym__primary,
    STATE(2369), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(910), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(635), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2368), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2417), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(908), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [33433] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      anon_sym_LPAREN,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    ACTIONS(906), 1,
      sym_NEW,
    ACTIONS(912), 1,
      sym_based_base,
    ACTIONS(916), 1,
      sym_bit_string_length,
    ACTIONS(918), 1,
      sym_bit_string_base,
    ACTIONS(1305), 1,
      sym_condition_conversion,
    STATE(483), 1,
      sym_unary_operator,
    STATE(484), 1,
      sym_sign,
    STATE(743), 1,
      sym__direct_name,
    STATE(1060), 1,
      sym__abstract_literal,
    STATE(1124), 1,
      sym__expression,
    STATE(1236), 1,
      sym_based_literal,
    STATE(2366), 1,
      sym_bit_string_literal,
    STATE(2367), 1,
      sym__primary,
    STATE(2369), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(910), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(636), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2368), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2417), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(908), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [33556] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(898), 1,
      anon_sym_LPAREN,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    ACTIONS(906), 1,
      sym_NEW,
    ACTIONS(912), 1,
      sym_based_base,
    ACTIONS(916), 1,
      sym_bit_string_length,
    ACTIONS(918), 1,
      sym_bit_string_base,
    ACTIONS(1305), 1,
      sym_condition_conversion,
    STATE(483), 1,
      sym_unary_operator,
    STATE(484), 1,
      sym_sign,
    STATE(743), 1,
      sym__direct_name,
    STATE(1060), 1,
      sym__abstract_literal,
    STATE(1137), 1,
      sym__expression,
    STATE(1236), 1,
      sym_based_literal,
    STATE(2366), 1,
      sym_bit_string_literal,
    STATE(2367), 1,
      sym__primary,
    STATE(2369), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(910), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(637), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2368), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2417), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(908), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [33679] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(986), 1,
      anon_sym_LPAREN,
    ACTIONS(990), 1,
      sym_NEW,
    ACTIONS(996), 1,
      sym_based_base,
    ACTIONS(998), 1,
      sym_bit_string_length,
    ACTIONS(1000), 1,
      sym_bit_string_base,
    ACTIONS(1347), 1,
      sym_condition_conversion,
    STATE(499), 1,
      sym_unary_operator,
    STATE(500), 1,
      sym_sign,
    STATE(763), 1,
      sym__direct_name,
    STATE(1138), 1,
      sym__abstract_literal,
    STATE(1257), 1,
      sym_based_literal,
    STATE(1311), 1,
      sym__expression,
    STATE(2452), 1,
      sym_bit_string_literal,
    STATE(2454), 1,
      sym__primary,
    STATE(2456), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(994), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(638), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2455), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2467), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(992), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [33802] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(986), 1,
      anon_sym_LPAREN,
    ACTIONS(990), 1,
      sym_NEW,
    ACTIONS(996), 1,
      sym_based_base,
    ACTIONS(998), 1,
      sym_bit_string_length,
    ACTIONS(1000), 1,
      sym_bit_string_base,
    ACTIONS(1347), 1,
      sym_condition_conversion,
    STATE(499), 1,
      sym_unary_operator,
    STATE(500), 1,
      sym_sign,
    STATE(763), 1,
      sym__direct_name,
    STATE(1138), 1,
      sym__abstract_literal,
    STATE(1257), 1,
      sym_based_literal,
    STATE(1316), 1,
      sym__expression,
    STATE(2452), 1,
      sym_bit_string_literal,
    STATE(2454), 1,
      sym__primary,
    STATE(2456), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(994), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(639), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2455), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2467), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(992), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [33925] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(986), 1,
      anon_sym_LPAREN,
    ACTIONS(990), 1,
      sym_NEW,
    ACTIONS(996), 1,
      sym_based_base,
    ACTIONS(998), 1,
      sym_bit_string_length,
    ACTIONS(1000), 1,
      sym_bit_string_base,
    ACTIONS(1347), 1,
      sym_condition_conversion,
    STATE(499), 1,
      sym_unary_operator,
    STATE(500), 1,
      sym_sign,
    STATE(763), 1,
      sym__direct_name,
    STATE(1138), 1,
      sym__abstract_literal,
    STATE(1257), 1,
      sym_based_literal,
    STATE(1317), 1,
      sym__expression,
    STATE(2452), 1,
      sym_bit_string_literal,
    STATE(2454), 1,
      sym__primary,
    STATE(2456), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(994), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(640), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2455), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2467), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(992), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [34048] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(986), 1,
      anon_sym_LPAREN,
    ACTIONS(990), 1,
      sym_NEW,
    ACTIONS(996), 1,
      sym_based_base,
    ACTIONS(998), 1,
      sym_bit_string_length,
    ACTIONS(1000), 1,
      sym_bit_string_base,
    ACTIONS(1347), 1,
      sym_condition_conversion,
    STATE(499), 1,
      sym_unary_operator,
    STATE(500), 1,
      sym_sign,
    STATE(763), 1,
      sym__direct_name,
    STATE(1138), 1,
      sym__abstract_literal,
    STATE(1257), 1,
      sym_based_literal,
    STATE(1318), 1,
      sym__expression,
    STATE(2452), 1,
      sym_bit_string_literal,
    STATE(2454), 1,
      sym__primary,
    STATE(2456), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(994), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(641), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2455), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2467), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(992), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [34171] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1004), 1,
      sym_condition_conversion,
    ACTIONS(1405), 1,
      anon_sym_LPAREN,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    ACTIONS(1411), 1,
      sym_NEW,
    ACTIONS(1417), 1,
      sym_based_base,
    ACTIONS(1421), 1,
      sym_bit_string_length,
    ACTIONS(1423), 1,
      sym_bit_string_base,
    STATE(524), 1,
      sym_unary_operator,
    STATE(525), 1,
      sym_sign,
    STATE(769), 1,
      sym__direct_name,
    STATE(1331), 1,
      sym__abstract_literal,
    STATE(1436), 1,
      sym__expression,
    STATE(1445), 1,
      sym_based_literal,
    STATE(2477), 1,
      sym_bit_string_literal,
    STATE(2482), 1,
      sym__primary,
    STATE(2485), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1415), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(642), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2484), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2533), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1413), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [34294] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1004), 1,
      sym_condition_conversion,
    ACTIONS(1405), 1,
      anon_sym_LPAREN,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    ACTIONS(1411), 1,
      sym_NEW,
    ACTIONS(1417), 1,
      sym_based_base,
    ACTIONS(1421), 1,
      sym_bit_string_length,
    ACTIONS(1423), 1,
      sym_bit_string_base,
    STATE(524), 1,
      sym_unary_operator,
    STATE(525), 1,
      sym_sign,
    STATE(769), 1,
      sym__direct_name,
    STATE(1331), 1,
      sym__abstract_literal,
    STATE(1440), 1,
      sym__expression,
    STATE(1445), 1,
      sym_based_literal,
    STATE(2477), 1,
      sym_bit_string_literal,
    STATE(2482), 1,
      sym__primary,
    STATE(2485), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1415), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(643), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2484), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2533), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1413), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [34417] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1004), 1,
      sym_condition_conversion,
    ACTIONS(1405), 1,
      anon_sym_LPAREN,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    ACTIONS(1411), 1,
      sym_NEW,
    ACTIONS(1417), 1,
      sym_based_base,
    ACTIONS(1421), 1,
      sym_bit_string_length,
    ACTIONS(1423), 1,
      sym_bit_string_base,
    STATE(524), 1,
      sym_unary_operator,
    STATE(525), 1,
      sym_sign,
    STATE(769), 1,
      sym__direct_name,
    STATE(1331), 1,
      sym__abstract_literal,
    STATE(1441), 1,
      sym__expression,
    STATE(1445), 1,
      sym_based_literal,
    STATE(2477), 1,
      sym_bit_string_literal,
    STATE(2482), 1,
      sym__primary,
    STATE(2485), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1415), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(644), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2484), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2533), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1413), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [34540] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1004), 1,
      sym_condition_conversion,
    ACTIONS(1405), 1,
      anon_sym_LPAREN,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    ACTIONS(1411), 1,
      sym_NEW,
    ACTIONS(1417), 1,
      sym_based_base,
    ACTIONS(1421), 1,
      sym_bit_string_length,
    ACTIONS(1423), 1,
      sym_bit_string_base,
    STATE(524), 1,
      sym_unary_operator,
    STATE(525), 1,
      sym_sign,
    STATE(769), 1,
      sym__direct_name,
    STATE(1331), 1,
      sym__abstract_literal,
    STATE(1442), 1,
      sym__expression,
    STATE(1445), 1,
      sym_based_literal,
    STATE(2477), 1,
      sym_bit_string_literal,
    STATE(2482), 1,
      sym__primary,
    STATE(2485), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1415), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(645), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2484), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2533), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1413), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [34663] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1223), 1,
      anon_sym_LPAREN,
    ACTIONS(1225), 1,
      sym_condition_conversion,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    ACTIONS(1231), 1,
      sym_NEW,
    ACTIONS(1237), 1,
      sym_based_base,
    ACTIONS(1241), 1,
      sym_bit_string_length,
    ACTIONS(1243), 1,
      sym_bit_string_base,
    STATE(433), 1,
      sym_unary_operator,
    STATE(434), 1,
      sym_sign,
    STATE(830), 1,
      sym__direct_name,
    STATE(1678), 1,
      sym__abstract_literal,
    STATE(1951), 1,
      sym__expression,
    STATE(2019), 1,
      sym_based_literal,
    STATE(2648), 1,
      sym_bit_string_literal,
    STATE(2659), 1,
      sym__primary,
    STATE(2667), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1235), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(646), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2662), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2650), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1233), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [34786] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1002), 1,
      anon_sym_LPAREN,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    ACTIONS(1010), 1,
      sym_NEW,
    ACTIONS(1016), 1,
      sym_based_base,
    ACTIONS(1020), 1,
      sym_bit_string_length,
    ACTIONS(1022), 1,
      sym_bit_string_base,
    ACTIONS(1403), 1,
      sym_condition_conversion,
    STATE(519), 1,
      sym_unary_operator,
    STATE(520), 1,
      sym_sign,
    STATE(746), 1,
      sym__direct_name,
    STATE(1071), 1,
      sym__abstract_literal,
    STATE(1154), 1,
      sym_based_literal,
    STATE(1218), 1,
      sym__expression,
    STATE(2290), 1,
      sym__primary,
    STATE(2399), 1,
      sym__literal,
    STATE(2429), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1014), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(647), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2301), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2387), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1012), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [34909] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1002), 1,
      anon_sym_LPAREN,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    ACTIONS(1010), 1,
      sym_NEW,
    ACTIONS(1016), 1,
      sym_based_base,
    ACTIONS(1020), 1,
      sym_bit_string_length,
    ACTIONS(1022), 1,
      sym_bit_string_base,
    ACTIONS(1403), 1,
      sym_condition_conversion,
    STATE(519), 1,
      sym_unary_operator,
    STATE(520), 1,
      sym_sign,
    STATE(746), 1,
      sym__direct_name,
    STATE(1071), 1,
      sym__abstract_literal,
    STATE(1154), 1,
      sym_based_literal,
    STATE(1211), 1,
      sym__expression,
    STATE(2290), 1,
      sym__primary,
    STATE(2399), 1,
      sym__literal,
    STATE(2429), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1014), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(648), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2301), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2387), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1012), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [35032] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1002), 1,
      anon_sym_LPAREN,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    ACTIONS(1010), 1,
      sym_NEW,
    ACTIONS(1016), 1,
      sym_based_base,
    ACTIONS(1020), 1,
      sym_bit_string_length,
    ACTIONS(1022), 1,
      sym_bit_string_base,
    ACTIONS(1403), 1,
      sym_condition_conversion,
    STATE(519), 1,
      sym_unary_operator,
    STATE(520), 1,
      sym_sign,
    STATE(746), 1,
      sym__direct_name,
    STATE(1071), 1,
      sym__abstract_literal,
    STATE(1134), 1,
      sym__expression,
    STATE(1154), 1,
      sym_based_literal,
    STATE(2290), 1,
      sym__primary,
    STATE(2399), 1,
      sym__literal,
    STATE(2429), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1014), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(649), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2301), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2387), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1012), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [35155] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1002), 1,
      anon_sym_LPAREN,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    ACTIONS(1010), 1,
      sym_NEW,
    ACTIONS(1016), 1,
      sym_based_base,
    ACTIONS(1020), 1,
      sym_bit_string_length,
    ACTIONS(1022), 1,
      sym_bit_string_base,
    ACTIONS(1403), 1,
      sym_condition_conversion,
    STATE(519), 1,
      sym_unary_operator,
    STATE(520), 1,
      sym_sign,
    STATE(746), 1,
      sym__direct_name,
    STATE(1071), 1,
      sym__abstract_literal,
    STATE(1139), 1,
      sym__expression,
    STATE(1154), 1,
      sym_based_literal,
    STATE(2290), 1,
      sym__primary,
    STATE(2399), 1,
      sym__literal,
    STATE(2429), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1014), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(650), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2301), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2387), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1012), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [35278] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(1958), 1,
      sym_simple_expression,
    STATE(2434), 1,
      sym__expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(651), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 4,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [35403] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    ACTIONS(1449), 1,
      sym_condition_conversion,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    ACTIONS(1455), 1,
      sym_NEW,
    ACTIONS(1461), 1,
      sym_based_base,
    ACTIONS(1465), 1,
      sym_bit_string_length,
    ACTIONS(1467), 1,
      sym_bit_string_base,
    STATE(673), 1,
      sym_unary_operator,
    STATE(674), 1,
      sym_sign,
    STATE(853), 1,
      sym__direct_name,
    STATE(2020), 1,
      sym__abstract_literal,
    STATE(2357), 1,
      sym_based_literal,
    STATE(2382), 1,
      sym__expression,
    STATE(2700), 1,
      sym__primary,
    STATE(2705), 1,
      sym__literal,
    STATE(2739), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1459), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(652), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2701), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2750), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1457), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [35526] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1159), 1,
      anon_sym_LPAREN,
    ACTIONS(1161), 1,
      sym_condition_conversion,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    ACTIONS(1167), 1,
      sym_NEW,
    ACTIONS(1173), 1,
      sym_based_base,
    ACTIONS(1177), 1,
      sym_bit_string_length,
    ACTIONS(1179), 1,
      sym_bit_string_base,
    STATE(413), 1,
      sym_unary_operator,
    STATE(661), 1,
      sym_sign,
    STATE(780), 1,
      sym__direct_name,
    STATE(1531), 1,
      sym__abstract_literal,
    STATE(1730), 1,
      sym_based_literal,
    STATE(1874), 1,
      sym__expression,
    STATE(2567), 1,
      sym__primary,
    STATE(2610), 1,
      sym_bit_string_literal,
    STATE(2612), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1171), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(653), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2584), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2547), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1169), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [35649] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2295), 1,
      sym__expression,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(654), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [35772] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1159), 1,
      anon_sym_LPAREN,
    ACTIONS(1161), 1,
      sym_condition_conversion,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    ACTIONS(1167), 1,
      sym_NEW,
    ACTIONS(1173), 1,
      sym_based_base,
    ACTIONS(1177), 1,
      sym_bit_string_length,
    ACTIONS(1179), 1,
      sym_bit_string_base,
    STATE(413), 1,
      sym_unary_operator,
    STATE(661), 1,
      sym_sign,
    STATE(780), 1,
      sym__direct_name,
    STATE(1386), 1,
      sym__expression,
    STATE(1531), 1,
      sym__abstract_literal,
    STATE(1730), 1,
      sym_based_literal,
    STATE(2567), 1,
      sym__primary,
    STATE(2610), 1,
      sym_bit_string_literal,
    STATE(2612), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1171), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(655), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2584), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2547), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1169), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [35895] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1249), 1,
      anon_sym_LPAREN,
    ACTIONS(1251), 1,
      sym_condition_conversion,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    ACTIONS(1257), 1,
      sym_NEW,
    ACTIONS(1263), 1,
      sym_based_base,
    ACTIONS(1267), 1,
      sym_bit_string_length,
    ACTIONS(1269), 1,
      sym_bit_string_base,
    STATE(443), 1,
      sym_unary_operator,
    STATE(444), 1,
      sym_sign,
    STATE(842), 1,
      sym__direct_name,
    STATE(2054), 1,
      sym__abstract_literal,
    STATE(2296), 1,
      sym_based_literal,
    STATE(2339), 1,
      sym__expression,
    STATE(2735), 1,
      sym_bit_string_literal,
    STATE(2746), 1,
      sym__primary,
    STATE(2753), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1261), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(656), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2752), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2734), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1259), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [36018] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2377), 1,
      sym__expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(657), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [36141] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1181), 1,
      anon_sym_LPAREN,
    ACTIONS(1183), 1,
      sym_condition_conversion,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    ACTIONS(1189), 1,
      sym_NEW,
    ACTIONS(1195), 1,
      sym_based_base,
    ACTIONS(1199), 1,
      sym_bit_string_length,
    ACTIONS(1201), 1,
      sym_bit_string_base,
    STATE(423), 1,
      sym_unary_operator,
    STATE(424), 1,
      sym_sign,
    STATE(883), 1,
      sym__direct_name,
    STATE(2008), 1,
      sym__abstract_literal,
    STATE(2305), 1,
      sym__expression,
    STATE(2337), 1,
      sym_based_literal,
    STATE(2709), 1,
      sym_bit_string_literal,
    STATE(2713), 1,
      sym__primary,
    STATE(2717), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1193), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(658), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2716), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2710), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1191), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [36264] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2326), 1,
      sym__expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(659), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [36387] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2022), 1,
      sym__expression,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(660), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [36510] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1159), 1,
      anon_sym_LPAREN,
    ACTIONS(1161), 1,
      sym_condition_conversion,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    ACTIONS(1167), 1,
      sym_NEW,
    ACTIONS(1173), 1,
      sym_based_base,
    ACTIONS(1177), 1,
      sym_bit_string_length,
    ACTIONS(1179), 1,
      sym_bit_string_base,
    STATE(413), 1,
      sym_unary_operator,
    STATE(780), 1,
      sym__direct_name,
    STATE(1531), 1,
      sym__abstract_literal,
    STATE(1693), 1,
      sym__expression,
    STATE(1730), 1,
      sym_based_literal,
    STATE(2567), 1,
      sym__primary,
    STATE(2610), 1,
      sym_bit_string_literal,
    STATE(2612), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1171), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2584), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(661), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2547), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1169), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [36631] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(1963), 1,
      sym__expression,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(662), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [36754] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1425), 1,
      anon_sym_LPAREN,
    ACTIONS(1427), 1,
      sym_condition_conversion,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    ACTIONS(1433), 1,
      sym_NEW,
    ACTIONS(1439), 1,
      sym_based_base,
    ACTIONS(1443), 1,
      sym_bit_string_length,
    ACTIONS(1445), 1,
      sym_bit_string_base,
    STATE(714), 1,
      sym_unary_operator,
    STATE(715), 1,
      sym_sign,
    STATE(820), 1,
      sym__direct_name,
    STATE(1860), 1,
      sym__abstract_literal,
    STATE(1927), 1,
      sym_based_literal,
    STATE(2116), 1,
      sym__expression,
    STATE(2640), 1,
      sym_bit_string_literal,
    STATE(2652), 1,
      sym__literal,
    STATE(2677), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1437), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(663), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2643), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2688), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1435), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [36877] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(964), 1,
      sym_condition_conversion,
    ACTIONS(1088), 1,
      anon_sym_LPAREN,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    ACTIONS(1096), 1,
      sym_NEW,
    ACTIONS(1102), 1,
      sym_based_base,
    ACTIONS(1106), 1,
      sym_bit_string_length,
    ACTIONS(1108), 1,
      sym_bit_string_base,
    STATE(418), 1,
      sym_unary_operator,
    STATE(419), 1,
      sym_sign,
    STATE(873), 1,
      sym__direct_name,
    STATE(1978), 1,
      sym__abstract_literal,
    STATE(2304), 1,
      sym_based_literal,
    STATE(2312), 1,
      sym__expression,
    STATE(2694), 1,
      sym__primary,
    STATE(2698), 1,
      sym__literal,
    STATE(2758), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1100), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(664), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2697), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2768), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1098), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [37000] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2335), 1,
      sym__expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(665), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [37123] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1159), 1,
      anon_sym_LPAREN,
    ACTIONS(1161), 1,
      sym_condition_conversion,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    ACTIONS(1167), 1,
      sym_NEW,
    ACTIONS(1173), 1,
      sym_based_base,
    ACTIONS(1177), 1,
      sym_bit_string_length,
    ACTIONS(1179), 1,
      sym_bit_string_base,
    STATE(413), 1,
      sym_unary_operator,
    STATE(661), 1,
      sym_sign,
    STATE(780), 1,
      sym__direct_name,
    STATE(1531), 1,
      sym__abstract_literal,
    STATE(1730), 1,
      sym_based_literal,
    STATE(1809), 1,
      sym__expression,
    STATE(2567), 1,
      sym__primary,
    STATE(2610), 1,
      sym_bit_string_literal,
    STATE(2612), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1171), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(666), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2584), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2547), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1169), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [37246] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2322), 1,
      sym__expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(667), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [37369] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1271), 1,
      anon_sym_LPAREN,
    ACTIONS(1273), 1,
      sym_condition_conversion,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    ACTIONS(1279), 1,
      sym_NEW,
    ACTIONS(1285), 1,
      sym_based_base,
    ACTIONS(1289), 1,
      sym_bit_string_length,
    ACTIONS(1291), 1,
      sym_bit_string_base,
    STATE(448), 1,
      sym_unary_operator,
    STATE(449), 1,
      sym_sign,
    STATE(812), 1,
      sym__direct_name,
    STATE(1769), 1,
      sym__abstract_literal,
    STATE(2061), 1,
      sym_based_literal,
    STATE(2110), 1,
      sym__expression,
    STATE(2622), 1,
      sym__primary,
    STATE(2627), 1,
      sym__literal,
    STATE(2683), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(668), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2625), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2673), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1281), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [37492] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1159), 1,
      anon_sym_LPAREN,
    ACTIONS(1161), 1,
      sym_condition_conversion,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    ACTIONS(1167), 1,
      sym_NEW,
    ACTIONS(1173), 1,
      sym_based_base,
    ACTIONS(1177), 1,
      sym_bit_string_length,
    ACTIONS(1179), 1,
      sym_bit_string_base,
    STATE(413), 1,
      sym_unary_operator,
    STATE(661), 1,
      sym_sign,
    STATE(780), 1,
      sym__direct_name,
    STATE(1531), 1,
      sym__abstract_literal,
    STATE(1730), 1,
      sym_based_literal,
    STATE(1813), 1,
      sym__expression,
    STATE(2567), 1,
      sym__primary,
    STATE(2610), 1,
      sym_bit_string_literal,
    STATE(2612), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1171), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(669), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2584), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2547), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1169), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [37615] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1159), 1,
      anon_sym_LPAREN,
    ACTIONS(1161), 1,
      sym_condition_conversion,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    ACTIONS(1167), 1,
      sym_NEW,
    ACTIONS(1173), 1,
      sym_based_base,
    ACTIONS(1177), 1,
      sym_bit_string_length,
    ACTIONS(1179), 1,
      sym_bit_string_base,
    STATE(413), 1,
      sym_unary_operator,
    STATE(661), 1,
      sym_sign,
    STATE(780), 1,
      sym__direct_name,
    STATE(1531), 1,
      sym__abstract_literal,
    STATE(1730), 1,
      sym_based_literal,
    STATE(1814), 1,
      sym__expression,
    STATE(2567), 1,
      sym__primary,
    STATE(2610), 1,
      sym_bit_string_literal,
    STATE(2612), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1171), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(670), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2584), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2547), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1169), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [37738] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2330), 1,
      sym__expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(671), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [37861] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    ACTIONS(1449), 1,
      sym_condition_conversion,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    ACTIONS(1455), 1,
      sym_NEW,
    ACTIONS(1461), 1,
      sym_based_base,
    ACTIONS(1465), 1,
      sym_bit_string_length,
    ACTIONS(1467), 1,
      sym_bit_string_base,
    STATE(673), 1,
      sym_unary_operator,
    STATE(674), 1,
      sym_sign,
    STATE(853), 1,
      sym__direct_name,
    STATE(2020), 1,
      sym__abstract_literal,
    STATE(2357), 1,
      sym_based_literal,
    STATE(2394), 1,
      sym__expression,
    STATE(2700), 1,
      sym__primary,
    STATE(2705), 1,
      sym__literal,
    STATE(2739), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1459), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(672), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2701), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2750), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1457), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [37984] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    ACTIONS(1449), 1,
      sym_condition_conversion,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    ACTIONS(1455), 1,
      sym_NEW,
    ACTIONS(1461), 1,
      sym_based_base,
    ACTIONS(1465), 1,
      sym_bit_string_length,
    ACTIONS(1467), 1,
      sym_bit_string_base,
    STATE(674), 1,
      sym_sign,
    STATE(853), 1,
      sym__direct_name,
    STATE(2020), 1,
      sym__abstract_literal,
    STATE(2357), 1,
      sym_based_literal,
    STATE(2402), 1,
      sym__expression,
    STATE(2700), 1,
      sym__primary,
    STATE(2705), 1,
      sym__literal,
    STATE(2739), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1459), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2701), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(673), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2750), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1457), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [38105] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    ACTIONS(1449), 1,
      sym_condition_conversion,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    ACTIONS(1455), 1,
      sym_NEW,
    ACTIONS(1461), 1,
      sym_based_base,
    ACTIONS(1465), 1,
      sym_bit_string_length,
    ACTIONS(1467), 1,
      sym_bit_string_base,
    STATE(673), 1,
      sym_unary_operator,
    STATE(853), 1,
      sym__direct_name,
    STATE(2020), 1,
      sym__abstract_literal,
    STATE(2357), 1,
      sym_based_literal,
    STATE(2403), 1,
      sym__expression,
    STATE(2700), 1,
      sym__primary,
    STATE(2705), 1,
      sym__literal,
    STATE(2739), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1459), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2701), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(674), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2750), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1457), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [38226] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    ACTIONS(1449), 1,
      sym_condition_conversion,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    ACTIONS(1455), 1,
      sym_NEW,
    ACTIONS(1461), 1,
      sym_based_base,
    ACTIONS(1465), 1,
      sym_bit_string_length,
    ACTIONS(1467), 1,
      sym_bit_string_base,
    STATE(673), 1,
      sym_unary_operator,
    STATE(674), 1,
      sym_sign,
    STATE(853), 1,
      sym__direct_name,
    STATE(2020), 1,
      sym__abstract_literal,
    STATE(2357), 1,
      sym_based_literal,
    STATE(2414), 1,
      sym__expression,
    STATE(2700), 1,
      sym__primary,
    STATE(2705), 1,
      sym__literal,
    STATE(2739), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1459), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(675), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2701), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2750), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1457), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [38349] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    ACTIONS(1449), 1,
      sym_condition_conversion,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    ACTIONS(1455), 1,
      sym_NEW,
    ACTIONS(1461), 1,
      sym_based_base,
    ACTIONS(1465), 1,
      sym_bit_string_length,
    ACTIONS(1467), 1,
      sym_bit_string_base,
    STATE(673), 1,
      sym_unary_operator,
    STATE(674), 1,
      sym_sign,
    STATE(853), 1,
      sym__direct_name,
    STATE(2020), 1,
      sym__abstract_literal,
    STATE(2357), 1,
      sym_based_literal,
    STATE(2415), 1,
      sym__expression,
    STATE(2700), 1,
      sym__primary,
    STATE(2705), 1,
      sym__literal,
    STATE(2739), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1459), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(676), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2701), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2750), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1457), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [38472] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    ACTIONS(1449), 1,
      sym_condition_conversion,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    ACTIONS(1455), 1,
      sym_NEW,
    ACTIONS(1461), 1,
      sym_based_base,
    ACTIONS(1465), 1,
      sym_bit_string_length,
    ACTIONS(1467), 1,
      sym_bit_string_base,
    STATE(673), 1,
      sym_unary_operator,
    STATE(674), 1,
      sym_sign,
    STATE(853), 1,
      sym__direct_name,
    STATE(2020), 1,
      sym__abstract_literal,
    STATE(2357), 1,
      sym_based_literal,
    STATE(2420), 1,
      sym__expression,
    STATE(2700), 1,
      sym__primary,
    STATE(2705), 1,
      sym__literal,
    STATE(2739), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1459), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(677), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2701), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2750), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1457), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [38595] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    ACTIONS(1449), 1,
      sym_condition_conversion,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    ACTIONS(1455), 1,
      sym_NEW,
    ACTIONS(1461), 1,
      sym_based_base,
    ACTIONS(1465), 1,
      sym_bit_string_length,
    ACTIONS(1467), 1,
      sym_bit_string_base,
    STATE(673), 1,
      sym_unary_operator,
    STATE(674), 1,
      sym_sign,
    STATE(853), 1,
      sym__direct_name,
    STATE(2020), 1,
      sym__abstract_literal,
    STATE(2357), 1,
      sym_based_literal,
    STATE(2423), 1,
      sym__expression,
    STATE(2700), 1,
      sym__primary,
    STATE(2705), 1,
      sym__literal,
    STATE(2739), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1459), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(678), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2701), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2750), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1457), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [38718] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    ACTIONS(1449), 1,
      sym_condition_conversion,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    ACTIONS(1455), 1,
      sym_NEW,
    ACTIONS(1461), 1,
      sym_based_base,
    ACTIONS(1465), 1,
      sym_bit_string_length,
    ACTIONS(1467), 1,
      sym_bit_string_base,
    STATE(673), 1,
      sym_unary_operator,
    STATE(674), 1,
      sym_sign,
    STATE(853), 1,
      sym__direct_name,
    STATE(2020), 1,
      sym__abstract_literal,
    STATE(2357), 1,
      sym_based_literal,
    STATE(2424), 1,
      sym__expression,
    STATE(2700), 1,
      sym__primary,
    STATE(2705), 1,
      sym__literal,
    STATE(2739), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1459), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(679), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2701), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2750), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1457), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [38841] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
    ACTIONS(1449), 1,
      sym_condition_conversion,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    ACTIONS(1455), 1,
      sym_NEW,
    ACTIONS(1461), 1,
      sym_based_base,
    ACTIONS(1465), 1,
      sym_bit_string_length,
    ACTIONS(1467), 1,
      sym_bit_string_base,
    STATE(673), 1,
      sym_unary_operator,
    STATE(674), 1,
      sym_sign,
    STATE(853), 1,
      sym__direct_name,
    STATE(2020), 1,
      sym__abstract_literal,
    STATE(2357), 1,
      sym_based_literal,
    STATE(2425), 1,
      sym__expression,
    STATE(2700), 1,
      sym__primary,
    STATE(2705), 1,
      sym__literal,
    STATE(2739), 1,
      sym_bit_string_literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1459), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(680), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2701), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2750), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1457), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [38964] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1159), 1,
      anon_sym_LPAREN,
    ACTIONS(1161), 1,
      sym_condition_conversion,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    ACTIONS(1167), 1,
      sym_NEW,
    ACTIONS(1173), 1,
      sym_based_base,
    ACTIONS(1177), 1,
      sym_bit_string_length,
    ACTIONS(1179), 1,
      sym_bit_string_base,
    STATE(413), 1,
      sym_unary_operator,
    STATE(661), 1,
      sym_sign,
    STATE(780), 1,
      sym__direct_name,
    STATE(1531), 1,
      sym__abstract_literal,
    STATE(1730), 1,
      sym_based_literal,
    STATE(1815), 1,
      sym__expression,
    STATE(2567), 1,
      sym__primary,
    STATE(2610), 1,
      sym_bit_string_literal,
    STATE(2612), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1171), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(681), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2584), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2547), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1169), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [39087] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1159), 1,
      anon_sym_LPAREN,
    ACTIONS(1161), 1,
      sym_condition_conversion,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    ACTIONS(1167), 1,
      sym_NEW,
    ACTIONS(1173), 1,
      sym_based_base,
    ACTIONS(1177), 1,
      sym_bit_string_length,
    ACTIONS(1179), 1,
      sym_bit_string_base,
    STATE(413), 1,
      sym_unary_operator,
    STATE(661), 1,
      sym_sign,
    STATE(780), 1,
      sym__direct_name,
    STATE(1531), 1,
      sym__abstract_literal,
    STATE(1730), 1,
      sym_based_literal,
    STATE(1816), 1,
      sym__expression,
    STATE(2567), 1,
      sym__primary,
    STATE(2610), 1,
      sym_bit_string_literal,
    STATE(2612), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1171), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(682), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2584), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2547), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1169), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [39210] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(719), 1,
      sym_unary_operator,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    STATE(2333), 1,
      sym__expression,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(683), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [39333] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1159), 1,
      anon_sym_LPAREN,
    ACTIONS(1161), 1,
      sym_condition_conversion,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    ACTIONS(1167), 1,
      sym_NEW,
    ACTIONS(1173), 1,
      sym_based_base,
    ACTIONS(1177), 1,
      sym_bit_string_length,
    ACTIONS(1179), 1,
      sym_bit_string_base,
    STATE(413), 1,
      sym_unary_operator,
    STATE(661), 1,
      sym_sign,
    STATE(780), 1,
      sym__direct_name,
    STATE(1531), 1,
      sym__abstract_literal,
    STATE(1730), 1,
      sym_based_literal,
    STATE(1818), 1,
      sym__expression,
    STATE(2567), 1,
      sym__primary,
    STATE(2610), 1,
      sym_bit_string_literal,
    STATE(2612), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1171), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(684), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2584), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2547), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1169), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [39456] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1198), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(685), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [39579] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(263), 1,
      anon_sym_LPAREN,
    ACTIONS(275), 1,
      sym_NEW,
    ACTIONS(285), 1,
      sym_based_base,
    ACTIONS(289), 1,
      sym_bit_string_length,
    ACTIONS(291), 1,
      sym_bit_string_base,
    ACTIONS(401), 1,
      anon_sym_LT_LT,
    ACTIONS(1469), 1,
      sym_condition_conversion,
    STATE(687), 1,
      sym_sign,
    STATE(734), 1,
      sym__direct_name,
    STATE(997), 1,
      sym__abstract_literal,
    STATE(1044), 1,
      sym_based_literal,
    STATE(1051), 1,
      sym__expression,
    STATE(2089), 1,
      sym_bit_string_literal,
    STATE(2097), 1,
      sym__primary,
    STATE(2099), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(966), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(409), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(967), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2098), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(403), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(686), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2216), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(277), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [39700] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(263), 1,
      anon_sym_LPAREN,
    ACTIONS(275), 1,
      sym_NEW,
    ACTIONS(285), 1,
      sym_based_base,
    ACTIONS(289), 1,
      sym_bit_string_length,
    ACTIONS(291), 1,
      sym_bit_string_base,
    ACTIONS(401), 1,
      anon_sym_LT_LT,
    ACTIONS(1469), 1,
      sym_condition_conversion,
    STATE(686), 1,
      sym_unary_operator,
    STATE(734), 1,
      sym__direct_name,
    STATE(997), 1,
      sym__abstract_literal,
    STATE(1044), 1,
      sym_based_literal,
    STATE(1052), 1,
      sym__expression,
    STATE(2089), 1,
      sym_bit_string_literal,
    STATE(2097), 1,
      sym__primary,
    STATE(2099), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(966), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(409), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(967), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2098), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(403), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(687), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2216), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(277), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [39821] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(263), 1,
      anon_sym_LPAREN,
    ACTIONS(275), 1,
      sym_NEW,
    ACTIONS(285), 1,
      sym_based_base,
    ACTIONS(289), 1,
      sym_bit_string_length,
    ACTIONS(291), 1,
      sym_bit_string_base,
    ACTIONS(401), 1,
      anon_sym_LT_LT,
    ACTIONS(1469), 1,
      sym_condition_conversion,
    STATE(686), 1,
      sym_unary_operator,
    STATE(687), 1,
      sym_sign,
    STATE(734), 1,
      sym__direct_name,
    STATE(997), 1,
      sym__abstract_literal,
    STATE(1044), 1,
      sym_based_literal,
    STATE(1053), 1,
      sym__expression,
    STATE(2089), 1,
      sym_bit_string_literal,
    STATE(2097), 1,
      sym__primary,
    STATE(2099), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(966), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(409), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(688), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(967), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2098), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(403), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2216), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(277), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [39944] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1214), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(689), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [40067] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1215), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(690), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [40190] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1216), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(691), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [40313] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(263), 1,
      anon_sym_LPAREN,
    ACTIONS(275), 1,
      sym_NEW,
    ACTIONS(285), 1,
      sym_based_base,
    ACTIONS(289), 1,
      sym_bit_string_length,
    ACTIONS(291), 1,
      sym_bit_string_base,
    ACTIONS(401), 1,
      anon_sym_LT_LT,
    ACTIONS(1469), 1,
      sym_condition_conversion,
    STATE(686), 1,
      sym_unary_operator,
    STATE(687), 1,
      sym_sign,
    STATE(734), 1,
      sym__direct_name,
    STATE(997), 1,
      sym__abstract_literal,
    STATE(1044), 1,
      sym_based_literal,
    STATE(1054), 1,
      sym__expression,
    STATE(2089), 1,
      sym_bit_string_literal,
    STATE(2097), 1,
      sym__primary,
    STATE(2099), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(966), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(409), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(692), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(967), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2098), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(403), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2216), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(277), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [40436] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(263), 1,
      anon_sym_LPAREN,
    ACTIONS(275), 1,
      sym_NEW,
    ACTIONS(285), 1,
      sym_based_base,
    ACTIONS(289), 1,
      sym_bit_string_length,
    ACTIONS(291), 1,
      sym_bit_string_base,
    ACTIONS(401), 1,
      anon_sym_LT_LT,
    ACTIONS(1469), 1,
      sym_condition_conversion,
    STATE(686), 1,
      sym_unary_operator,
    STATE(687), 1,
      sym_sign,
    STATE(734), 1,
      sym__direct_name,
    STATE(997), 1,
      sym__abstract_literal,
    STATE(1044), 1,
      sym_based_literal,
    STATE(1055), 1,
      sym__expression,
    STATE(2089), 1,
      sym_bit_string_literal,
    STATE(2097), 1,
      sym__primary,
    STATE(2099), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(966), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(409), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(693), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(967), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2098), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(403), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2216), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(277), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [40559] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(263), 1,
      anon_sym_LPAREN,
    ACTIONS(275), 1,
      sym_NEW,
    ACTIONS(285), 1,
      sym_based_base,
    ACTIONS(289), 1,
      sym_bit_string_length,
    ACTIONS(291), 1,
      sym_bit_string_base,
    ACTIONS(401), 1,
      anon_sym_LT_LT,
    ACTIONS(1469), 1,
      sym_condition_conversion,
    STATE(686), 1,
      sym_unary_operator,
    STATE(687), 1,
      sym_sign,
    STATE(734), 1,
      sym__direct_name,
    STATE(997), 1,
      sym__abstract_literal,
    STATE(1044), 1,
      sym_based_literal,
    STATE(1058), 1,
      sym__expression,
    STATE(2089), 1,
      sym_bit_string_literal,
    STATE(2097), 1,
      sym__primary,
    STATE(2099), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(966), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(409), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(694), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(967), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2098), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(403), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2216), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(277), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [40682] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1135), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(695), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [40803] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1136), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(696), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [40924] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1146), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(697), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [41047] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(263), 1,
      anon_sym_LPAREN,
    ACTIONS(275), 1,
      sym_NEW,
    ACTIONS(285), 1,
      sym_based_base,
    ACTIONS(289), 1,
      sym_bit_string_length,
    ACTIONS(291), 1,
      sym_bit_string_base,
    ACTIONS(401), 1,
      anon_sym_LT_LT,
    ACTIONS(1469), 1,
      sym_condition_conversion,
    STATE(686), 1,
      sym_unary_operator,
    STATE(687), 1,
      sym_sign,
    STATE(734), 1,
      sym__direct_name,
    STATE(997), 1,
      sym__abstract_literal,
    STATE(1044), 1,
      sym_based_literal,
    STATE(1065), 1,
      sym__expression,
    STATE(2089), 1,
      sym_bit_string_literal,
    STATE(2097), 1,
      sym__primary,
    STATE(2099), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(966), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(409), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(698), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(967), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2098), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(403), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2216), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(277), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [41170] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(263), 1,
      anon_sym_LPAREN,
    ACTIONS(275), 1,
      sym_NEW,
    ACTIONS(285), 1,
      sym_based_base,
    ACTIONS(289), 1,
      sym_bit_string_length,
    ACTIONS(291), 1,
      sym_bit_string_base,
    ACTIONS(401), 1,
      anon_sym_LT_LT,
    ACTIONS(1469), 1,
      sym_condition_conversion,
    STATE(686), 1,
      sym_unary_operator,
    STATE(687), 1,
      sym_sign,
    STATE(734), 1,
      sym__direct_name,
    STATE(997), 1,
      sym__abstract_literal,
    STATE(1044), 1,
      sym_based_literal,
    STATE(1066), 1,
      sym__expression,
    STATE(2089), 1,
      sym_bit_string_literal,
    STATE(2097), 1,
      sym__primary,
    STATE(2099), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(966), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(409), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(699), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(967), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2098), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(403), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2216), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(277), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [41293] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(263), 1,
      anon_sym_LPAREN,
    ACTIONS(275), 1,
      sym_NEW,
    ACTIONS(285), 1,
      sym_based_base,
    ACTIONS(289), 1,
      sym_bit_string_length,
    ACTIONS(291), 1,
      sym_bit_string_base,
    ACTIONS(401), 1,
      anon_sym_LT_LT,
    ACTIONS(1469), 1,
      sym_condition_conversion,
    STATE(686), 1,
      sym_unary_operator,
    STATE(687), 1,
      sym_sign,
    STATE(734), 1,
      sym__direct_name,
    STATE(997), 1,
      sym__abstract_literal,
    STATE(1044), 1,
      sym_based_literal,
    STATE(1067), 1,
      sym__expression,
    STATE(2089), 1,
      sym_bit_string_literal,
    STATE(2097), 1,
      sym__primary,
    STATE(2099), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(283), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(966), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(409), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(700), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(967), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2098), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(403), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2216), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(277), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [41416] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1150), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(701), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [41539] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(265), 1,
      sym_condition_conversion,
    ACTIONS(736), 1,
      anon_sym_LPAREN,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    ACTIONS(744), 1,
      sym_NEW,
    ACTIONS(752), 1,
      sym_based_base,
    ACTIONS(756), 1,
      sym_bit_string_length,
    ACTIONS(758), 1,
      sym_bit_string_base,
    STATE(695), 1,
      sym_unary_operator,
    STATE(696), 1,
      sym_sign,
    STATE(748), 1,
      sym__direct_name,
    STATE(1056), 1,
      sym__abstract_literal,
    STATE(1151), 1,
      sym__expression,
    STATE(1229), 1,
      sym_based_literal,
    STATE(2380), 1,
      sym_bit_string_literal,
    STATE(2383), 1,
      sym__primary,
    STATE(2385), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(750), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(702), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2384), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2313), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(746), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [41662] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2431), 1,
      sym__expression,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(703), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [41785] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(411), 1,
      anon_sym_LPAREN,
    ACTIONS(413), 1,
      sym_condition_conversion,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    ACTIONS(423), 1,
      sym_NEW,
    ACTIONS(433), 1,
      sym_based_base,
    ACTIONS(437), 1,
      sym_bit_string_length,
    ACTIONS(439), 1,
      sym_bit_string_base,
    STATE(705), 1,
      sym_sign,
    STATE(777), 1,
      sym__direct_name,
    STATE(1408), 1,
      sym__abstract_literal,
    STATE(1627), 1,
      sym_based_literal,
    STATE(1631), 1,
      sym__expression,
    STATE(2506), 1,
      sym_bit_string_literal,
    STATE(2508), 1,
      sym__primary,
    STATE(2510), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(431), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2509), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(704), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2536), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(425), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [41906] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(411), 1,
      anon_sym_LPAREN,
    ACTIONS(413), 1,
      sym_condition_conversion,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    ACTIONS(423), 1,
      sym_NEW,
    ACTIONS(433), 1,
      sym_based_base,
    ACTIONS(437), 1,
      sym_bit_string_length,
    ACTIONS(439), 1,
      sym_bit_string_base,
    STATE(704), 1,
      sym_unary_operator,
    STATE(777), 1,
      sym__direct_name,
    STATE(1408), 1,
      sym__abstract_literal,
    STATE(1627), 1,
      sym_based_literal,
    STATE(1632), 1,
      sym__expression,
    STATE(2506), 1,
      sym_bit_string_literal,
    STATE(2508), 1,
      sym__primary,
    STATE(2510), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(431), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2509), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(705), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2536), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(425), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [42027] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(411), 1,
      anon_sym_LPAREN,
    ACTIONS(413), 1,
      sym_condition_conversion,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    ACTIONS(423), 1,
      sym_NEW,
    ACTIONS(433), 1,
      sym_based_base,
    ACTIONS(437), 1,
      sym_bit_string_length,
    ACTIONS(439), 1,
      sym_bit_string_base,
    STATE(704), 1,
      sym_unary_operator,
    STATE(705), 1,
      sym_sign,
    STATE(777), 1,
      sym__direct_name,
    STATE(1408), 1,
      sym__abstract_literal,
    STATE(1627), 1,
      sym_based_literal,
    STATE(1633), 1,
      sym__expression,
    STATE(2506), 1,
      sym_bit_string_literal,
    STATE(2508), 1,
      sym__primary,
    STATE(2510), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(431), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(706), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2509), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2536), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(425), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [42150] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(411), 1,
      anon_sym_LPAREN,
    ACTIONS(413), 1,
      sym_condition_conversion,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    ACTIONS(423), 1,
      sym_NEW,
    ACTIONS(433), 1,
      sym_based_base,
    ACTIONS(437), 1,
      sym_bit_string_length,
    ACTIONS(439), 1,
      sym_bit_string_base,
    STATE(704), 1,
      sym_unary_operator,
    STATE(705), 1,
      sym_sign,
    STATE(777), 1,
      sym__direct_name,
    STATE(1408), 1,
      sym__abstract_literal,
    STATE(1627), 1,
      sym_based_literal,
    STATE(1637), 1,
      sym__expression,
    STATE(2506), 1,
      sym_bit_string_literal,
    STATE(2508), 1,
      sym__primary,
    STATE(2510), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(431), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(707), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2509), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2536), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(425), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [42273] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(411), 1,
      anon_sym_LPAREN,
    ACTIONS(413), 1,
      sym_condition_conversion,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    ACTIONS(423), 1,
      sym_NEW,
    ACTIONS(433), 1,
      sym_based_base,
    ACTIONS(437), 1,
      sym_bit_string_length,
    ACTIONS(439), 1,
      sym_bit_string_base,
    STATE(704), 1,
      sym_unary_operator,
    STATE(705), 1,
      sym_sign,
    STATE(777), 1,
      sym__direct_name,
    STATE(1408), 1,
      sym__abstract_literal,
    STATE(1627), 1,
      sym_based_literal,
    STATE(1638), 1,
      sym__expression,
    STATE(2506), 1,
      sym_bit_string_literal,
    STATE(2508), 1,
      sym__primary,
    STATE(2510), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(431), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(708), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2509), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2536), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(425), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [42396] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2412), 1,
      sym__expression,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(709), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [42517] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2378), 1,
      sym__expression,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(710), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [42638] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2316), 1,
      sym__expression,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(711), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [42761] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2331), 1,
      sym__expression,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(712), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [42884] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(878), 1,
      sym_condition_conversion,
    ACTIONS(920), 1,
      anon_sym_LPAREN,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    ACTIONS(926), 1,
      sym_NEW,
    ACTIONS(932), 1,
      sym_based_base,
    ACTIONS(936), 1,
      sym_bit_string_length,
    ACTIONS(938), 1,
      sym_bit_string_base,
    STATE(709), 1,
      sym_unary_operator,
    STATE(710), 1,
      sym_sign,
    STATE(868), 1,
      sym__direct_name,
    STATE(2283), 1,
      sym__abstract_literal,
    STATE(2334), 1,
      sym__expression,
    STATE(2422), 1,
      sym_based_literal,
    STATE(2756), 1,
      sym_bit_string_literal,
    STATE(2762), 1,
      sym__primary,
    STATE(2764), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(930), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(713), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2763), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2761), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(928), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [43007] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1425), 1,
      anon_sym_LPAREN,
    ACTIONS(1427), 1,
      sym_condition_conversion,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    ACTIONS(1433), 1,
      sym_NEW,
    ACTIONS(1439), 1,
      sym_based_base,
    ACTIONS(1443), 1,
      sym_bit_string_length,
    ACTIONS(1445), 1,
      sym_bit_string_base,
    STATE(715), 1,
      sym_sign,
    STATE(820), 1,
      sym__direct_name,
    STATE(1860), 1,
      sym__abstract_literal,
    STATE(1927), 1,
      sym_based_literal,
    STATE(1933), 1,
      sym__expression,
    STATE(2640), 1,
      sym_bit_string_literal,
    STATE(2652), 1,
      sym__literal,
    STATE(2677), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1437), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2643), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(714), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2688), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1435), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [43128] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1425), 1,
      anon_sym_LPAREN,
    ACTIONS(1427), 1,
      sym_condition_conversion,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    ACTIONS(1433), 1,
      sym_NEW,
    ACTIONS(1439), 1,
      sym_based_base,
    ACTIONS(1443), 1,
      sym_bit_string_length,
    ACTIONS(1445), 1,
      sym_bit_string_base,
    STATE(714), 1,
      sym_unary_operator,
    STATE(820), 1,
      sym__direct_name,
    STATE(1860), 1,
      sym__abstract_literal,
    STATE(1927), 1,
      sym_based_literal,
    STATE(1934), 1,
      sym__expression,
    STATE(2640), 1,
      sym_bit_string_literal,
    STATE(2652), 1,
      sym__literal,
    STATE(2677), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1437), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2643), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(715), 4,
      sym_sign,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2688), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1435), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [43249] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1425), 1,
      anon_sym_LPAREN,
    ACTIONS(1427), 1,
      sym_condition_conversion,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    ACTIONS(1433), 1,
      sym_NEW,
    ACTIONS(1439), 1,
      sym_based_base,
    ACTIONS(1443), 1,
      sym_bit_string_length,
    ACTIONS(1445), 1,
      sym_bit_string_base,
    STATE(714), 1,
      sym_unary_operator,
    STATE(715), 1,
      sym_sign,
    STATE(820), 1,
      sym__direct_name,
    STATE(1860), 1,
      sym__abstract_literal,
    STATE(1927), 1,
      sym_based_literal,
    STATE(1935), 1,
      sym__expression,
    STATE(2640), 1,
      sym_bit_string_literal,
    STATE(2652), 1,
      sym__literal,
    STATE(2677), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1437), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(716), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2643), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2688), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1435), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [43372] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1425), 1,
      anon_sym_LPAREN,
    ACTIONS(1427), 1,
      sym_condition_conversion,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    ACTIONS(1433), 1,
      sym_NEW,
    ACTIONS(1439), 1,
      sym_based_base,
    ACTIONS(1443), 1,
      sym_bit_string_length,
    ACTIONS(1445), 1,
      sym_bit_string_base,
    STATE(714), 1,
      sym_unary_operator,
    STATE(715), 1,
      sym_sign,
    STATE(820), 1,
      sym__direct_name,
    STATE(1860), 1,
      sym__abstract_literal,
    STATE(1927), 1,
      sym_based_literal,
    STATE(1939), 1,
      sym__expression,
    STATE(2640), 1,
      sym_bit_string_literal,
    STATE(2652), 1,
      sym__literal,
    STATE(2677), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1437), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(717), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2643), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2688), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1435), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [43495] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1425), 1,
      anon_sym_LPAREN,
    ACTIONS(1427), 1,
      sym_condition_conversion,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    ACTIONS(1433), 1,
      sym_NEW,
    ACTIONS(1439), 1,
      sym_based_base,
    ACTIONS(1443), 1,
      sym_bit_string_length,
    ACTIONS(1445), 1,
      sym_bit_string_base,
    STATE(714), 1,
      sym_unary_operator,
    STATE(715), 1,
      sym_sign,
    STATE(820), 1,
      sym__direct_name,
    STATE(1860), 1,
      sym__abstract_literal,
    STATE(1927), 1,
      sym_based_literal,
    STATE(1940), 1,
      sym__expression,
    STATE(2640), 1,
      sym_bit_string_literal,
    STATE(2652), 1,
      sym__literal,
    STATE(2677), 1,
      sym__primary,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1437), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(718), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(2643), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2688), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1435), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [43618] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(762), 1,
      sym_condition_conversion,
    ACTIONS(1068), 1,
      anon_sym_LPAREN,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    ACTIONS(1074), 1,
      sym_NEW,
    ACTIONS(1080), 1,
      sym_based_base,
    ACTIONS(1084), 1,
      sym_bit_string_length,
    ACTIONS(1086), 1,
      sym_bit_string_base,
    STATE(414), 1,
      sym_sign,
    STATE(733), 1,
      sym__direct_name,
    STATE(965), 1,
      sym__abstract_literal,
    STATE(1076), 1,
      sym__expression,
    STATE(1110), 1,
      sym_based_literal,
    STATE(1943), 1,
      sym_bit_string_literal,
    STATE(1946), 1,
      sym__primary,
    STATE(1948), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1078), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1947), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(719), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1944), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(1076), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [43739] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(760), 1,
      anon_sym_LPAREN,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    ACTIONS(768), 1,
      sym_NEW,
    ACTIONS(774), 1,
      sym_based_base,
    ACTIONS(778), 1,
      sym_bit_string_length,
    ACTIONS(780), 1,
      sym_bit_string_base,
    ACTIONS(1247), 1,
      sym_condition_conversion,
    STATE(439), 1,
      sym_sign,
    STATE(721), 1,
      sym__direct_name,
    STATE(846), 1,
      sym__abstract_literal,
    STATE(902), 1,
      sym_based_literal,
    STATE(916), 1,
      sym__expression,
    STATE(1612), 1,
      sym_bit_string_literal,
    STATE(1620), 1,
      sym__primary,
    STATE(1622), 1,
      sym__literal,
    ACTIONS(267), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(772), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(1621), 3,
      sym_name,
      sym_allocator,
      sym_parenthesis_expression,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(720), 4,
      sym_unary_operator,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1644), 5,
      sym_condition_expression,
      sym_logical_expression,
      sym_relational_expression,
      sym_shift_expression,
      sym_simple_expression,
    ACTIONS(273), 8,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_NOT,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(770), 8,
      sym_NULL,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
  [43860] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1473), 1,
      anon_sym_DOT,
    ACTIONS(1475), 1,
      anon_sym_LPAREN,
    ACTIONS(1479), 1,
      anon_sym_SQUOTE,
    ACTIONS(1481), 1,
      anon_sym_LBRACK,
    ACTIONS(1483), 1,
      sym_GENERIC,
    ACTIONS(1485), 1,
      sym_PARAMETER,
    STATE(723), 1,
      aux_sym_name_repeat1,
    STATE(805), 1,
      sym_generic_map_aspect,
    STATE(847), 1,
      sym__name_selector,
    STATE(721), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(848), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [43954] = 36,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1112), 1,
      sym_BEGIN,
    ACTIONS(1114), 1,
      sym_GENERIC,
    ACTIONS(1116), 1,
      sym_IS,
    ACTIONS(1118), 1,
      sym_PORT,
    STATE(737), 1,
      sym_generic_clause,
    STATE(757), 1,
      sym_port_clause,
    STATE(807), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(6261), 1,
      sym_concurrent_block,
    STATE(6268), 1,
      sym_block_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(722), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [44088] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1473), 1,
      anon_sym_DOT,
    ACTIONS(1475), 1,
      anon_sym_LPAREN,
    ACTIONS(1479), 1,
      anon_sym_SQUOTE,
    ACTIONS(1481), 1,
      anon_sym_LBRACK,
    ACTIONS(1483), 1,
      sym_GENERIC,
    ACTIONS(1485), 1,
      sym_PARAMETER,
    STATE(725), 1,
      aux_sym_name_repeat1,
    STATE(805), 1,
      sym_generic_map_aspect,
    STATE(847), 1,
      sym__name_selector,
    STATE(723), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(848), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [44182] = 36,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1112), 1,
      sym_BEGIN,
    ACTIONS(1114), 1,
      sym_GENERIC,
    ACTIONS(1116), 1,
      sym_IS,
    ACTIONS(1118), 1,
      sym_PORT,
    STATE(737), 1,
      sym_generic_clause,
    STATE(757), 1,
      sym_port_clause,
    STATE(807), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(6117), 1,
      sym_concurrent_block,
    STATE(6452), 1,
      sym_block_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(724), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [44316] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1493), 1,
      anon_sym_DOT,
    ACTIONS(1496), 1,
      anon_sym_LPAREN,
    ACTIONS(1501), 1,
      anon_sym_SQUOTE,
    ACTIONS(1504), 1,
      anon_sym_LBRACK,
    ACTIONS(1507), 1,
      sym_GENERIC,
    ACTIONS(1510), 1,
      sym_PARAMETER,
    STATE(805), 1,
      sym_generic_map_aspect,
    STATE(847), 1,
      sym__name_selector,
    STATE(725), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(848), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [44408] = 36,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1112), 1,
      sym_BEGIN,
    ACTIONS(1114), 1,
      sym_GENERIC,
    ACTIONS(1116), 1,
      sym_IS,
    ACTIONS(1118), 1,
      sym_PORT,
    STATE(737), 1,
      sym_generic_clause,
    STATE(757), 1,
      sym_port_clause,
    STATE(807), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(6397), 1,
      sym_block_head,
    STATE(6584), 1,
      sym_concurrent_block,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(726), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [44542] = 36,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1112), 1,
      sym_BEGIN,
    ACTIONS(1114), 1,
      sym_GENERIC,
    ACTIONS(1116), 1,
      sym_IS,
    ACTIONS(1118), 1,
      sym_PORT,
    STATE(737), 1,
      sym_generic_clause,
    STATE(757), 1,
      sym_port_clause,
    STATE(807), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(6057), 1,
      sym_concurrent_block,
    STATE(6428), 1,
      sym_block_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(727), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [44676] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1513), 1,
      anon_sym_DOT,
    ACTIONS(1516), 1,
      anon_sym_LPAREN,
    ACTIONS(1519), 1,
      anon_sym_SQUOTE,
    ACTIONS(1522), 1,
      anon_sym_LBRACK,
    ACTIONS(1525), 1,
      sym_GENERIC,
    ACTIONS(1528), 1,
      sym_PARAMETER,
    STATE(850), 1,
      sym_generic_map_aspect,
    STATE(924), 1,
      sym__name_selector,
    STATE(728), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(926), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 33,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [44767] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1531), 1,
      anon_sym_DOT,
    ACTIONS(1533), 1,
      anon_sym_LPAREN,
    ACTIONS(1535), 1,
      anon_sym_SQUOTE,
    ACTIONS(1537), 1,
      anon_sym_LBRACK,
    ACTIONS(1539), 1,
      sym_GENERIC,
    ACTIONS(1541), 1,
      sym_PARAMETER,
    STATE(728), 1,
      aux_sym_name_repeat1,
    STATE(850), 1,
      sym_generic_map_aspect,
    STATE(924), 1,
      sym__name_selector,
    STATE(729), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(926), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 33,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [44860] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1531), 1,
      anon_sym_DOT,
    ACTIONS(1533), 1,
      anon_sym_LPAREN,
    ACTIONS(1535), 1,
      anon_sym_SQUOTE,
    ACTIONS(1537), 1,
      anon_sym_LBRACK,
    ACTIONS(1539), 1,
      sym_GENERIC,
    ACTIONS(1541), 1,
      sym_PARAMETER,
    STATE(729), 1,
      aux_sym_name_repeat1,
    STATE(850), 1,
      sym_generic_map_aspect,
    STATE(924), 1,
      sym__name_selector,
    STATE(730), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(926), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 33,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [44953] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1543), 1,
      anon_sym_DOT,
    ACTIONS(1546), 1,
      anon_sym_LPAREN,
    ACTIONS(1549), 1,
      anon_sym_SQUOTE,
    ACTIONS(1552), 1,
      anon_sym_LBRACK,
    ACTIONS(1555), 1,
      sym_GENERIC,
    ACTIONS(1558), 1,
      sym_PARAMETER,
    STATE(888), 1,
      sym_generic_map_aspect,
    STATE(962), 1,
      sym__name_selector,
    STATE(731), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(963), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [45043] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1561), 1,
      anon_sym_DOT,
    ACTIONS(1563), 1,
      anon_sym_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_SQUOTE,
    ACTIONS(1567), 1,
      anon_sym_LBRACK,
    ACTIONS(1569), 1,
      sym_GENERIC,
    ACTIONS(1571), 1,
      sym_PARAMETER,
    STATE(731), 1,
      aux_sym_name_repeat1,
    STATE(888), 1,
      sym_generic_map_aspect,
    STATE(962), 1,
      sym__name_selector,
    STATE(732), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(963), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [45135] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1561), 1,
      anon_sym_DOT,
    ACTIONS(1563), 1,
      anon_sym_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_SQUOTE,
    ACTIONS(1567), 1,
      anon_sym_LBRACK,
    ACTIONS(1569), 1,
      sym_GENERIC,
    ACTIONS(1571), 1,
      sym_PARAMETER,
    STATE(732), 1,
      aux_sym_name_repeat1,
    STATE(888), 1,
      sym_generic_map_aspect,
    STATE(962), 1,
      sym__name_selector,
    STATE(733), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(963), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [45227] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1573), 1,
      anon_sym_DOT,
    ACTIONS(1575), 1,
      anon_sym_LPAREN,
    ACTIONS(1577), 1,
      anon_sym_SQUOTE,
    ACTIONS(1579), 1,
      anon_sym_LBRACK,
    ACTIONS(1581), 1,
      sym_GENERIC,
    ACTIONS(1583), 1,
      sym_PARAMETER,
    STATE(735), 1,
      aux_sym_name_repeat1,
    STATE(889), 1,
      sym_generic_map_aspect,
    STATE(968), 1,
      sym__name_selector,
    STATE(734), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(969), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [45319] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1573), 1,
      anon_sym_DOT,
    ACTIONS(1575), 1,
      anon_sym_LPAREN,
    ACTIONS(1577), 1,
      anon_sym_SQUOTE,
    ACTIONS(1579), 1,
      anon_sym_LBRACK,
    ACTIONS(1581), 1,
      sym_GENERIC,
    ACTIONS(1583), 1,
      sym_PARAMETER,
    STATE(736), 1,
      aux_sym_name_repeat1,
    STATE(889), 1,
      sym_generic_map_aspect,
    STATE(968), 1,
      sym__name_selector,
    STATE(735), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(969), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [45411] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1585), 1,
      anon_sym_DOT,
    ACTIONS(1588), 1,
      anon_sym_LPAREN,
    ACTIONS(1591), 1,
      anon_sym_SQUOTE,
    ACTIONS(1594), 1,
      anon_sym_LBRACK,
    ACTIONS(1597), 1,
      sym_GENERIC,
    ACTIONS(1600), 1,
      sym_PARAMETER,
    STATE(889), 1,
      sym_generic_map_aspect,
    STATE(968), 1,
      sym__name_selector,
    STATE(736), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(969), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [45501] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1118), 1,
      sym_PORT,
    ACTIONS(1603), 1,
      sym_BEGIN,
    ACTIONS(1605), 1,
      sym_GENERIC,
    STATE(773), 1,
      sym_port_clause,
    STATE(829), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(6658), 1,
      sym_generic_map_aspect,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(737), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [45626] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1607), 1,
      anon_sym_DOT,
    ACTIONS(1610), 1,
      anon_sym_LPAREN,
    ACTIONS(1613), 1,
      anon_sym_SQUOTE,
    ACTIONS(1616), 1,
      anon_sym_LBRACK,
    ACTIONS(1619), 1,
      sym_GENERIC,
    ACTIONS(1622), 1,
      sym_PARAMETER,
    STATE(992), 1,
      sym_generic_map_aspect,
    STATE(1079), 1,
      sym__name_selector,
    STATE(738), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1080), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [45715] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1625), 1,
      anon_sym_DOT,
    ACTIONS(1628), 1,
      anon_sym_LPAREN,
    ACTIONS(1631), 1,
      anon_sym_SQUOTE,
    ACTIONS(1634), 1,
      anon_sym_LBRACK,
    ACTIONS(1637), 1,
      sym_GENERIC,
    ACTIONS(1640), 1,
      sym_PARAMETER,
    STATE(995), 1,
      sym_generic_map_aspect,
    STATE(1091), 1,
      sym__name_selector,
    STATE(739), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1104), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [45804] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(17), 1,
      sym_ARCHITECTURE,
    ACTIONS(25), 1,
      sym_CONFIGURATION,
    ACTIONS(29), 1,
      sym_CONTEXT,
    ACTIONS(33), 1,
      sym_ENTITY,
    ACTIONS(45), 1,
      sym_LIBRARY,
    ACTIONS(47), 1,
      sym_PACKAGE,
    ACTIONS(63), 1,
      sym_USE,
    STATE(1019), 1,
      aux_sym_design_unit_repeat1,
    STATE(1974), 1,
      sym__library_unit,
    STATE(2270), 1,
      sym__context_item,
    STATE(740), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1917), 3,
      sym_library_clause,
      sym_use_clause,
      sym_context_reference,
    STATE(1979), 7,
      sym_entity_declaration,
      sym_configuration_declaration,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_context_declaration,
      sym_architecture_definition,
      sym_package_definition,
    ACTIONS(672), 32,
      sym_identifier,
      sym_ALIAS,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [45897] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1643), 1,
      anon_sym_DOT,
    ACTIONS(1645), 1,
      anon_sym_LPAREN,
    ACTIONS(1647), 1,
      anon_sym_SQUOTE,
    ACTIONS(1649), 1,
      anon_sym_LBRACK,
    ACTIONS(1651), 1,
      sym_GENERIC,
    ACTIONS(1653), 1,
      sym_PARAMETER,
    STATE(739), 1,
      aux_sym_name_repeat1,
    STATE(995), 1,
      sym_generic_map_aspect,
    STATE(1091), 1,
      sym__name_selector,
    STATE(741), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1104), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [45988] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1655), 1,
      anon_sym_DOT,
    ACTIONS(1658), 1,
      anon_sym_LPAREN,
    ACTIONS(1661), 1,
      anon_sym_SQUOTE,
    ACTIONS(1664), 1,
      anon_sym_LBRACK,
    ACTIONS(1667), 1,
      sym_GENERIC,
    ACTIONS(1670), 1,
      sym_PARAMETER,
    STATE(956), 1,
      sym_generic_map_aspect,
    STATE(1001), 1,
      sym__name_selector,
    STATE(742), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1002), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [46077] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1643), 1,
      anon_sym_DOT,
    ACTIONS(1645), 1,
      anon_sym_LPAREN,
    ACTIONS(1647), 1,
      anon_sym_SQUOTE,
    ACTIONS(1649), 1,
      anon_sym_LBRACK,
    ACTIONS(1651), 1,
      sym_GENERIC,
    ACTIONS(1653), 1,
      sym_PARAMETER,
    STATE(741), 1,
      aux_sym_name_repeat1,
    STATE(995), 1,
      sym_generic_map_aspect,
    STATE(1091), 1,
      sym__name_selector,
    STATE(743), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1104), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [46168] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1114), 1,
      sym_GENERIC,
    ACTIONS(1118), 1,
      sym_PORT,
    ACTIONS(1603), 1,
      sym_BEGIN,
    STATE(745), 1,
      sym_generic_clause,
    STATE(773), 1,
      sym_port_clause,
    STATE(829), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(744), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [46293] = 33,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1118), 1,
      sym_PORT,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(1673), 1,
      sym_BEGIN,
    STATE(775), 1,
      sym_port_clause,
    STATE(801), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(7138), 1,
      sym_generic_map_aspect,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(745), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [46418] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1675), 1,
      anon_sym_DOT,
    ACTIONS(1677), 1,
      anon_sym_LPAREN,
    ACTIONS(1679), 1,
      anon_sym_SQUOTE,
    ACTIONS(1681), 1,
      anon_sym_LBRACK,
    ACTIONS(1683), 1,
      sym_GENERIC,
    ACTIONS(1685), 1,
      sym_PARAMETER,
    STATE(747), 1,
      aux_sym_name_repeat1,
    STATE(956), 1,
      sym_generic_map_aspect,
    STATE(1001), 1,
      sym__name_selector,
    STATE(746), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1002), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [46509] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1675), 1,
      anon_sym_DOT,
    ACTIONS(1677), 1,
      anon_sym_LPAREN,
    ACTIONS(1679), 1,
      anon_sym_SQUOTE,
    ACTIONS(1681), 1,
      anon_sym_LBRACK,
    ACTIONS(1683), 1,
      sym_GENERIC,
    ACTIONS(1685), 1,
      sym_PARAMETER,
    STATE(742), 1,
      aux_sym_name_repeat1,
    STATE(956), 1,
      sym_generic_map_aspect,
    STATE(1001), 1,
      sym__name_selector,
    STATE(747), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1002), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [46600] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1687), 1,
      anon_sym_DOT,
    ACTIONS(1689), 1,
      anon_sym_LPAREN,
    ACTIONS(1691), 1,
      anon_sym_SQUOTE,
    ACTIONS(1693), 1,
      anon_sym_LBRACK,
    ACTIONS(1695), 1,
      sym_GENERIC,
    ACTIONS(1697), 1,
      sym_PARAMETER,
    STATE(749), 1,
      aux_sym_name_repeat1,
    STATE(992), 1,
      sym_generic_map_aspect,
    STATE(1079), 1,
      sym__name_selector,
    STATE(748), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1080), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [46691] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1687), 1,
      anon_sym_DOT,
    ACTIONS(1689), 1,
      anon_sym_LPAREN,
    ACTIONS(1691), 1,
      anon_sym_SQUOTE,
    ACTIONS(1693), 1,
      anon_sym_LBRACK,
    ACTIONS(1695), 1,
      sym_GENERIC,
    ACTIONS(1697), 1,
      sym_PARAMETER,
    STATE(738), 1,
      aux_sym_name_repeat1,
    STATE(992), 1,
      sym_generic_map_aspect,
    STATE(1079), 1,
      sym__name_selector,
    STATE(749), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1080), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [46782] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1699), 1,
      anon_sym_DOT,
    ACTIONS(1701), 1,
      anon_sym_LPAREN,
    ACTIONS(1703), 1,
      anon_sym_SQUOTE,
    ACTIONS(1705), 1,
      anon_sym_LBRACK,
    ACTIONS(1707), 1,
      sym_GENERIC,
    ACTIONS(1709), 1,
      sym_PARAMETER,
    STATE(751), 1,
      aux_sym_name_repeat1,
    STATE(1049), 1,
      sym_generic_map_aspect,
    STATE(1199), 1,
      sym__name_selector,
    STATE(750), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1202), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [46872] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1699), 1,
      anon_sym_DOT,
    ACTIONS(1701), 1,
      anon_sym_LPAREN,
    ACTIONS(1703), 1,
      anon_sym_SQUOTE,
    ACTIONS(1705), 1,
      anon_sym_LBRACK,
    ACTIONS(1707), 1,
      sym_GENERIC,
    ACTIONS(1709), 1,
      sym_PARAMETER,
    STATE(753), 1,
      aux_sym_name_repeat1,
    STATE(1049), 1,
      sym_generic_map_aspect,
    STATE(1199), 1,
      sym__name_selector,
    STATE(751), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1202), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [46962] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1711), 1,
      anon_sym_DOT,
    ACTIONS(1713), 1,
      anon_sym_LPAREN,
    ACTIONS(1715), 1,
      anon_sym_SQUOTE,
    ACTIONS(1717), 1,
      anon_sym_LBRACK,
    ACTIONS(1719), 1,
      sym_GENERIC,
    ACTIONS(1721), 1,
      sym_PARAMETER,
    STATE(755), 1,
      aux_sym_name_repeat1,
    STATE(1018), 1,
      sym_generic_map_aspect,
    STATE(1159), 1,
      sym__name_selector,
    STATE(752), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1186), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 31,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [47052] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1723), 1,
      anon_sym_DOT,
    ACTIONS(1726), 1,
      anon_sym_LPAREN,
    ACTIONS(1729), 1,
      anon_sym_SQUOTE,
    ACTIONS(1732), 1,
      anon_sym_LBRACK,
    ACTIONS(1735), 1,
      sym_GENERIC,
    ACTIONS(1738), 1,
      sym_PARAMETER,
    STATE(1049), 1,
      sym_generic_map_aspect,
    STATE(1199), 1,
      sym__name_selector,
    STATE(753), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1202), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [47140] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1741), 1,
      anon_sym_DOT,
    ACTIONS(1744), 1,
      anon_sym_LPAREN,
    ACTIONS(1747), 1,
      anon_sym_SQUOTE,
    ACTIONS(1750), 1,
      anon_sym_LBRACK,
    ACTIONS(1753), 1,
      sym_GENERIC,
    ACTIONS(1756), 1,
      sym_PARAMETER,
    STATE(1018), 1,
      sym_generic_map_aspect,
    STATE(1159), 1,
      sym__name_selector,
    STATE(754), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1186), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 31,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [47228] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1711), 1,
      anon_sym_DOT,
    ACTIONS(1713), 1,
      anon_sym_LPAREN,
    ACTIONS(1715), 1,
      anon_sym_SQUOTE,
    ACTIONS(1717), 1,
      anon_sym_LBRACK,
    ACTIONS(1719), 1,
      sym_GENERIC,
    ACTIONS(1721), 1,
      sym_PARAMETER,
    STATE(754), 1,
      aux_sym_name_repeat1,
    STATE(1018), 1,
      sym_generic_map_aspect,
    STATE(1159), 1,
      sym__name_selector,
    STATE(755), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1186), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 31,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [47318] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1473), 1,
      anon_sym_DOT,
    ACTIONS(1475), 1,
      anon_sym_LPAREN,
    ACTIONS(1479), 1,
      anon_sym_SQUOTE,
    ACTIONS(1481), 1,
      anon_sym_LBRACK,
    ACTIONS(1485), 1,
      sym_PARAMETER,
    STATE(725), 1,
      aux_sym_name_repeat1,
    STATE(805), 1,
      sym_generic_map_aspect,
    STATE(847), 1,
      sym__name_selector,
    STATE(756), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(848), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [47405] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1603), 1,
      sym_BEGIN,
    ACTIONS(1759), 1,
      sym_PORT,
    STATE(829), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(8349), 1,
      sym_port_map_aspect,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(757), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [47524] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1761), 1,
      anon_sym_DOT,
    ACTIONS(1764), 1,
      anon_sym_LPAREN,
    ACTIONS(1767), 1,
      anon_sym_SQUOTE,
    ACTIONS(1770), 1,
      anon_sym_LBRACK,
    ACTIONS(1773), 1,
      sym_GENERIC,
    ACTIONS(1776), 1,
      sym_PARAMETER,
    STATE(1161), 1,
      sym_generic_map_aspect,
    STATE(1339), 1,
      sym__name_selector,
    STATE(758), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1340), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [47611] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1779), 1,
      anon_sym_DOT,
    ACTIONS(1782), 1,
      anon_sym_LPAREN,
    ACTIONS(1785), 1,
      anon_sym_SQUOTE,
    ACTIONS(1788), 1,
      anon_sym_LBRACK,
    ACTIONS(1791), 1,
      sym_GENERIC,
    ACTIONS(1794), 1,
      sym_PARAMETER,
    STATE(1200), 1,
      sym_generic_map_aspect,
    STATE(1243), 1,
      sym__name_selector,
    STATE(759), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1244), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [47698] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1797), 1,
      anon_sym_DOT,
    ACTIONS(1800), 1,
      anon_sym_LPAREN,
    ACTIONS(1803), 1,
      anon_sym_SQUOTE,
    ACTIONS(1806), 1,
      anon_sym_LBRACK,
    ACTIONS(1809), 1,
      sym_GENERIC,
    ACTIONS(1812), 1,
      sym_PARAMETER,
    STATE(1172), 1,
      sym_generic_map_aspect,
    STATE(1356), 1,
      sym__name_selector,
    STATE(760), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1357), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [47785] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1118), 1,
      sym_PORT,
    ACTIONS(1815), 1,
      sym_BEGIN,
    STATE(765), 1,
      sym_port_clause,
    STATE(814), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(761), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [47904] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(1815), 1,
      sym_BEGIN,
    STATE(814), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(6985), 1,
      sym_port_map_aspect,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(762), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [48023] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1473), 1,
      anon_sym_DOT,
    ACTIONS(1475), 1,
      anon_sym_LPAREN,
    ACTIONS(1479), 1,
      anon_sym_SQUOTE,
    ACTIONS(1481), 1,
      anon_sym_LBRACK,
    ACTIONS(1485), 1,
      sym_PARAMETER,
    STATE(756), 1,
      aux_sym_name_repeat1,
    STATE(805), 1,
      sym_generic_map_aspect,
    STATE(847), 1,
      sym__name_selector,
    STATE(763), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(848), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [48110] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1817), 1,
      anon_sym_DOT,
    ACTIONS(1819), 1,
      anon_sym_LPAREN,
    ACTIONS(1821), 1,
      anon_sym_SQUOTE,
    ACTIONS(1823), 1,
      anon_sym_LBRACK,
    ACTIONS(1825), 1,
      sym_GENERIC,
    ACTIONS(1827), 1,
      sym_PARAMETER,
    STATE(760), 1,
      aux_sym_name_repeat1,
    STATE(1172), 1,
      sym_generic_map_aspect,
    STATE(1356), 1,
      sym__name_selector,
    STATE(764), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1357), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [48199] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(1829), 1,
      sym_BEGIN,
    STATE(836), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(8191), 1,
      sym_port_map_aspect,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(765), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [48318] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1831), 1,
      anon_sym_DOT,
    ACTIONS(1834), 1,
      anon_sym_LPAREN,
    ACTIONS(1837), 1,
      anon_sym_SQUOTE,
    ACTIONS(1840), 1,
      anon_sym_LBRACK,
    ACTIONS(1843), 1,
      sym_GENERIC,
    ACTIONS(1846), 1,
      sym_PARAMETER,
    STATE(1195), 1,
      sym_generic_map_aspect,
    STATE(1282), 1,
      sym__name_selector,
    STATE(766), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1283), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [48405] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1849), 1,
      anon_sym_DOT,
    ACTIONS(1851), 1,
      anon_sym_LPAREN,
    ACTIONS(1853), 1,
      anon_sym_SQUOTE,
    ACTIONS(1855), 1,
      anon_sym_LBRACK,
    ACTIONS(1857), 1,
      sym_GENERIC,
    ACTIONS(1859), 1,
      sym_PARAMETER,
    STATE(768), 1,
      aux_sym_name_repeat1,
    STATE(1200), 1,
      sym_generic_map_aspect,
    STATE(1243), 1,
      sym__name_selector,
    STATE(767), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1244), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [48494] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1849), 1,
      anon_sym_DOT,
    ACTIONS(1851), 1,
      anon_sym_LPAREN,
    ACTIONS(1853), 1,
      anon_sym_SQUOTE,
    ACTIONS(1855), 1,
      anon_sym_LBRACK,
    ACTIONS(1857), 1,
      sym_GENERIC,
    ACTIONS(1859), 1,
      sym_PARAMETER,
    STATE(759), 1,
      aux_sym_name_repeat1,
    STATE(1200), 1,
      sym_generic_map_aspect,
    STATE(1243), 1,
      sym__name_selector,
    STATE(768), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1244), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [48583] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1861), 1,
      anon_sym_DOT,
    ACTIONS(1863), 1,
      anon_sym_LPAREN,
    ACTIONS(1865), 1,
      anon_sym_SQUOTE,
    ACTIONS(1867), 1,
      anon_sym_LBRACK,
    ACTIONS(1869), 1,
      sym_GENERIC,
    ACTIONS(1871), 1,
      sym_PARAMETER,
    STATE(770), 1,
      aux_sym_name_repeat1,
    STATE(1195), 1,
      sym_generic_map_aspect,
    STATE(1282), 1,
      sym__name_selector,
    STATE(769), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1283), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [48672] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1861), 1,
      anon_sym_DOT,
    ACTIONS(1863), 1,
      anon_sym_LPAREN,
    ACTIONS(1865), 1,
      anon_sym_SQUOTE,
    ACTIONS(1867), 1,
      anon_sym_LBRACK,
    ACTIONS(1869), 1,
      sym_GENERIC,
    ACTIONS(1871), 1,
      sym_PARAMETER,
    STATE(766), 1,
      aux_sym_name_repeat1,
    STATE(1195), 1,
      sym_generic_map_aspect,
    STATE(1282), 1,
      sym__name_selector,
    STATE(770), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1283), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [48761] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1873), 1,
      anon_sym_DOT,
    ACTIONS(1875), 1,
      anon_sym_LPAREN,
    ACTIONS(1877), 1,
      anon_sym_SQUOTE,
    ACTIONS(1879), 1,
      anon_sym_LBRACK,
    ACTIONS(1881), 1,
      sym_GENERIC,
    ACTIONS(1883), 1,
      sym_PARAMETER,
    STATE(774), 1,
      aux_sym_name_repeat1,
    STATE(1121), 1,
      sym_generic_map_aspect,
    STATE(1334), 1,
      sym__name_selector,
    STATE(771), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1239), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [48850] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1885), 1,
      anon_sym_DOT,
    ACTIONS(1887), 1,
      anon_sym_LPAREN,
    ACTIONS(1889), 1,
      anon_sym_SQUOTE,
    ACTIONS(1891), 1,
      anon_sym_LBRACK,
    ACTIONS(1893), 1,
      sym_GENERIC,
    ACTIONS(1895), 1,
      sym_PARAMETER,
    STATE(778), 1,
      aux_sym_name_repeat1,
    STATE(1161), 1,
      sym_generic_map_aspect,
    STATE(1339), 1,
      sym__name_selector,
    STATE(772), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1340), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [48939] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1673), 1,
      sym_BEGIN,
    ACTIONS(1759), 1,
      sym_PORT,
    STATE(801), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(7133), 1,
      sym_port_map_aspect,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(773), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [49058] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1897), 1,
      anon_sym_DOT,
    ACTIONS(1900), 1,
      anon_sym_LPAREN,
    ACTIONS(1903), 1,
      anon_sym_SQUOTE,
    ACTIONS(1906), 1,
      anon_sym_LBRACK,
    ACTIONS(1909), 1,
      sym_GENERIC,
    ACTIONS(1912), 1,
      sym_PARAMETER,
    STATE(1121), 1,
      sym_generic_map_aspect,
    STATE(1334), 1,
      sym__name_selector,
    STATE(774), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1239), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [49145] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(1915), 1,
      sym_BEGIN,
    STATE(811), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    STATE(7031), 1,
      sym_port_map_aspect,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(775), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [49264] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1118), 1,
      sym_PORT,
    ACTIONS(1915), 1,
      sym_BEGIN,
    STATE(762), 1,
      sym_port_clause,
    STATE(811), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(776), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [49383] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1873), 1,
      anon_sym_DOT,
    ACTIONS(1875), 1,
      anon_sym_LPAREN,
    ACTIONS(1877), 1,
      anon_sym_SQUOTE,
    ACTIONS(1879), 1,
      anon_sym_LBRACK,
    ACTIONS(1881), 1,
      sym_GENERIC,
    ACTIONS(1883), 1,
      sym_PARAMETER,
    STATE(771), 1,
      aux_sym_name_repeat1,
    STATE(1121), 1,
      sym_generic_map_aspect,
    STATE(1334), 1,
      sym__name_selector,
    STATE(777), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1239), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [49472] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1885), 1,
      anon_sym_DOT,
    ACTIONS(1887), 1,
      anon_sym_LPAREN,
    ACTIONS(1889), 1,
      anon_sym_SQUOTE,
    ACTIONS(1891), 1,
      anon_sym_LBRACK,
    ACTIONS(1893), 1,
      sym_GENERIC,
    ACTIONS(1895), 1,
      sym_PARAMETER,
    STATE(758), 1,
      aux_sym_name_repeat1,
    STATE(1161), 1,
      sym_generic_map_aspect,
    STATE(1339), 1,
      sym__name_selector,
    STATE(778), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1340), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [49561] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1817), 1,
      anon_sym_DOT,
    ACTIONS(1819), 1,
      anon_sym_LPAREN,
    ACTIONS(1821), 1,
      anon_sym_SQUOTE,
    ACTIONS(1823), 1,
      anon_sym_LBRACK,
    ACTIONS(1825), 1,
      sym_GENERIC,
    ACTIONS(1827), 1,
      sym_PARAMETER,
    STATE(764), 1,
      aux_sym_name_repeat1,
    STATE(1172), 1,
      sym_generic_map_aspect,
    STATE(1356), 1,
      sym__name_selector,
    STATE(779), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1357), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [49650] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1917), 1,
      anon_sym_DOT,
    ACTIONS(1919), 1,
      anon_sym_LPAREN,
    ACTIONS(1921), 1,
      anon_sym_SQUOTE,
    ACTIONS(1923), 1,
      anon_sym_LBRACK,
    ACTIONS(1925), 1,
      sym_GENERIC,
    ACTIONS(1927), 1,
      sym_PARAMETER,
    STATE(789), 1,
      aux_sym_name_repeat1,
    STATE(1399), 1,
      sym_generic_map_aspect,
    STATE(1554), 1,
      sym__name_selector,
    STATE(780), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1555), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [49738] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1929), 1,
      anon_sym_DOT,
    ACTIONS(1931), 1,
      anon_sym_LPAREN,
    ACTIONS(1933), 1,
      anon_sym_SQUOTE,
    ACTIONS(1935), 1,
      anon_sym_LBRACK,
    ACTIONS(1937), 1,
      sym_GENERIC,
    ACTIONS(1939), 1,
      sym_PARAMETER,
    STATE(791), 1,
      aux_sym_name_repeat1,
    STATE(1418), 1,
      sym_generic_map_aspect,
    STATE(1665), 1,
      sym__name_selector,
    STATE(781), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1667), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [49826] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1941), 1,
      anon_sym_DOT,
    ACTIONS(1944), 1,
      anon_sym_LPAREN,
    ACTIONS(1947), 1,
      anon_sym_SQUOTE,
    ACTIONS(1950), 1,
      anon_sym_LBRACK,
    ACTIONS(1953), 1,
      sym_GENERIC,
    ACTIONS(1956), 1,
      sym_PARAMETER,
    STATE(1399), 1,
      sym_generic_map_aspect,
    STATE(1554), 1,
      sym__name_selector,
    STATE(782), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1555), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [49912] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1959), 1,
      anon_sym_DOT,
    ACTIONS(1962), 1,
      anon_sym_LPAREN,
    ACTIONS(1965), 1,
      anon_sym_SQUOTE,
    ACTIONS(1968), 1,
      anon_sym_LBRACK,
    ACTIONS(1971), 1,
      sym_GENERIC,
    ACTIONS(1974), 1,
      sym_PARAMETER,
    STATE(1374), 1,
      sym_generic_map_aspect,
    STATE(1478), 1,
      sym__name_selector,
    STATE(783), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1479), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [49998] = 31,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(1977), 1,
      sym_ALIAS,
    ACTIONS(1979), 1,
      sym_ATTRIBUTE,
    ACTIONS(1983), 1,
      sym_CONSTANT,
    ACTIONS(1985), 1,
      sym_DISCONNECT,
    ACTIONS(1987), 1,
      sym_FILE,
    ACTIONS(1989), 1,
      sym_FUNCTION,
    ACTIONS(1991), 1,
      sym_GENERIC,
    ACTIONS(1993), 1,
      sym_GROUP,
    ACTIONS(1995), 1,
      sym_PACKAGE,
    ACTIONS(1997), 1,
      sym_PORT,
    ACTIONS(1999), 1,
      sym_SIGNAL,
    ACTIONS(2001), 1,
      sym_SHARED,
    ACTIONS(2003), 1,
      sym_SUBTYPE,
    ACTIONS(2005), 1,
      sym_TYPE,
    ACTIONS(2007), 1,
      sym_USE,
    ACTIONS(2009), 1,
      sym_VARIABLE,
    ACTIONS(2011), 1,
      sym_VIEW,
    STATE(861), 1,
      sym_generic_clause,
    STATE(960), 1,
      sym_port_clause,
    STATE(961), 1,
      aux_sym_entity_head_repeat1,
    STATE(3244), 1,
      sym__entity_declarative_item,
    STATE(5961), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    ACTIONS(1981), 2,
      sym_BEGIN,
      sym_END,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(784), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3241), 20,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [50116] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2013), 1,
      anon_sym_DOT,
    ACTIONS(2016), 1,
      anon_sym_LPAREN,
    ACTIONS(2019), 1,
      anon_sym_SQUOTE,
    ACTIONS(2022), 1,
      anon_sym_LBRACK,
    ACTIONS(2025), 1,
      sym_GENERIC,
    ACTIONS(2028), 1,
      sym_PARAMETER,
    STATE(1383), 1,
      sym_generic_map_aspect,
    STATE(1503), 1,
      sym__name_selector,
    STATE(785), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1504), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [50202] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2031), 1,
      anon_sym_DOT,
    ACTIONS(2033), 1,
      anon_sym_LPAREN,
    ACTIONS(2035), 1,
      anon_sym_SQUOTE,
    ACTIONS(2037), 1,
      anon_sym_LBRACK,
    ACTIONS(2039), 1,
      sym_GENERIC,
    ACTIONS(2041), 1,
      sym_PARAMETER,
    STATE(792), 1,
      aux_sym_name_repeat1,
    STATE(1374), 1,
      sym_generic_map_aspect,
    STATE(1478), 1,
      sym__name_selector,
    STATE(786), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1479), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [50290] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2043), 1,
      anon_sym_DOT,
    ACTIONS(2046), 1,
      anon_sym_LPAREN,
    ACTIONS(2049), 1,
      anon_sym_SQUOTE,
    ACTIONS(2052), 1,
      anon_sym_LBRACK,
    ACTIONS(2055), 1,
      sym_GENERIC,
    ACTIONS(2058), 1,
      sym_PARAMETER,
    STATE(1400), 1,
      sym_generic_map_aspect,
    STATE(1558), 1,
      sym__name_selector,
    STATE(787), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1559), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [50376] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2061), 1,
      anon_sym_DOT,
    ACTIONS(2063), 1,
      anon_sym_LPAREN,
    ACTIONS(2065), 1,
      anon_sym_SQUOTE,
    ACTIONS(2067), 1,
      anon_sym_LBRACK,
    ACTIONS(2069), 1,
      sym_GENERIC,
    ACTIONS(2071), 1,
      sym_PARAMETER,
    STATE(790), 1,
      aux_sym_name_repeat1,
    STATE(1400), 1,
      sym_generic_map_aspect,
    STATE(1558), 1,
      sym__name_selector,
    STATE(788), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1559), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [50464] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1917), 1,
      anon_sym_DOT,
    ACTIONS(1919), 1,
      anon_sym_LPAREN,
    ACTIONS(1921), 1,
      anon_sym_SQUOTE,
    ACTIONS(1923), 1,
      anon_sym_LBRACK,
    ACTIONS(1925), 1,
      sym_GENERIC,
    ACTIONS(1927), 1,
      sym_PARAMETER,
    STATE(782), 1,
      aux_sym_name_repeat1,
    STATE(1399), 1,
      sym_generic_map_aspect,
    STATE(1554), 1,
      sym__name_selector,
    STATE(789), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1555), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [50552] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2061), 1,
      anon_sym_DOT,
    ACTIONS(2063), 1,
      anon_sym_LPAREN,
    ACTIONS(2065), 1,
      anon_sym_SQUOTE,
    ACTIONS(2067), 1,
      anon_sym_LBRACK,
    ACTIONS(2069), 1,
      sym_GENERIC,
    ACTIONS(2071), 1,
      sym_PARAMETER,
    STATE(787), 1,
      aux_sym_name_repeat1,
    STATE(1400), 1,
      sym_generic_map_aspect,
    STATE(1558), 1,
      sym__name_selector,
    STATE(790), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1559), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [50640] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2073), 1,
      anon_sym_DOT,
    ACTIONS(2076), 1,
      anon_sym_LPAREN,
    ACTIONS(2079), 1,
      anon_sym_SQUOTE,
    ACTIONS(2082), 1,
      anon_sym_LBRACK,
    ACTIONS(2085), 1,
      sym_GENERIC,
    ACTIONS(2088), 1,
      sym_PARAMETER,
    STATE(1418), 1,
      sym_generic_map_aspect,
    STATE(1665), 1,
      sym__name_selector,
    STATE(791), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1667), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [50726] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2031), 1,
      anon_sym_DOT,
    ACTIONS(2033), 1,
      anon_sym_LPAREN,
    ACTIONS(2035), 1,
      anon_sym_SQUOTE,
    ACTIONS(2037), 1,
      anon_sym_LBRACK,
    ACTIONS(2039), 1,
      sym_GENERIC,
    ACTIONS(2041), 1,
      sym_PARAMETER,
    STATE(783), 1,
      aux_sym_name_repeat1,
    STATE(1374), 1,
      sym_generic_map_aspect,
    STATE(1478), 1,
      sym__name_selector,
    STATE(792), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1479), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [50814] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2091), 1,
      anon_sym_DOT,
    ACTIONS(2093), 1,
      anon_sym_LPAREN,
    ACTIONS(2095), 1,
      anon_sym_SQUOTE,
    ACTIONS(2097), 1,
      anon_sym_LBRACK,
    ACTIONS(2099), 1,
      sym_GENERIC,
    ACTIONS(2101), 1,
      sym_PARAMETER,
    STATE(794), 1,
      aux_sym_name_repeat1,
    STATE(1383), 1,
      sym_generic_map_aspect,
    STATE(1503), 1,
      sym__name_selector,
    STATE(793), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1504), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [50902] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2091), 1,
      anon_sym_DOT,
    ACTIONS(2093), 1,
      anon_sym_LPAREN,
    ACTIONS(2095), 1,
      anon_sym_SQUOTE,
    ACTIONS(2097), 1,
      anon_sym_LBRACK,
    ACTIONS(2099), 1,
      sym_GENERIC,
    ACTIONS(2101), 1,
      sym_PARAMETER,
    STATE(785), 1,
      aux_sym_name_repeat1,
    STATE(1383), 1,
      sym_generic_map_aspect,
    STATE(1503), 1,
      sym__name_selector,
    STATE(794), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1504), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [50990] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1929), 1,
      anon_sym_DOT,
    ACTIONS(1931), 1,
      anon_sym_LPAREN,
    ACTIONS(1933), 1,
      anon_sym_SQUOTE,
    ACTIONS(1935), 1,
      anon_sym_LBRACK,
    ACTIONS(1937), 1,
      sym_GENERIC,
    ACTIONS(1939), 1,
      sym_PARAMETER,
    STATE(781), 1,
      aux_sym_name_repeat1,
    STATE(1418), 1,
      sym_generic_map_aspect,
    STATE(1665), 1,
      sym__name_selector,
    STATE(795), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1667), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [51078] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(2103), 1,
      sym_BEGIN,
    STATE(824), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(796), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [51191] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2105), 1,
      anon_sym_DOT,
    ACTIONS(2108), 1,
      anon_sym_LPAREN,
    ACTIONS(2111), 1,
      anon_sym_SQUOTE,
    ACTIONS(2114), 1,
      anon_sym_LBRACK,
    ACTIONS(2117), 1,
      sym_GENERIC,
    ACTIONS(2120), 1,
      sym_PARAMETER,
    STATE(1628), 1,
      sym_generic_map_aspect,
    STATE(1801), 1,
      sym__name_selector,
    STATE(797), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1912), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [51276] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2123), 1,
      anon_sym_DOT,
    ACTIONS(2126), 1,
      anon_sym_LPAREN,
    ACTIONS(2129), 1,
      anon_sym_SQUOTE,
    ACTIONS(2132), 1,
      anon_sym_LBRACK,
    ACTIONS(2135), 1,
      sym_GENERIC,
    ACTIONS(2138), 1,
      sym_PARAMETER,
    STATE(1643), 1,
      sym_generic_map_aspect,
    STATE(1842), 1,
      sym__name_selector,
    STATE(798), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1843), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [51361] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2141), 1,
      anon_sym_DOT,
    ACTIONS(2144), 1,
      anon_sym_LPAREN,
    ACTIONS(2147), 1,
      anon_sym_SQUOTE,
    ACTIONS(2150), 1,
      anon_sym_LBRACK,
    ACTIONS(2153), 1,
      sym_GENERIC,
    ACTIONS(2156), 1,
      sym_PARAMETER,
    STATE(1574), 1,
      sym_generic_map_aspect,
    STATE(1679), 1,
      sym__name_selector,
    STATE(799), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1689), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [51446] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2159), 1,
      anon_sym_DOT,
    ACTIONS(2162), 1,
      anon_sym_LPAREN,
    ACTIONS(2165), 1,
      anon_sym_SQUOTE,
    ACTIONS(2168), 1,
      anon_sym_LBRACK,
    ACTIONS(2171), 1,
      sym_GENERIC,
    ACTIONS(2174), 1,
      sym_PARAMETER,
    STATE(1434), 1,
      sym_generic_map_aspect,
    STATE(1820), 1,
      sym__name_selector,
    STATE(800), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1835), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [51531] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1915), 1,
      sym_BEGIN,
    STATE(824), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(801), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [51644] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1915), 1,
      sym_BEGIN,
    STATE(811), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(802), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [51757] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    ACTIONS(2179), 1,
      anon_sym_LPAREN,
    ACTIONS(2181), 1,
      anon_sym_SQUOTE,
    ACTIONS(2183), 1,
      anon_sym_LBRACK,
    ACTIONS(2185), 1,
      sym_GENERIC,
    ACTIONS(2187), 1,
      sym_PARAMETER,
    STATE(806), 1,
      aux_sym_name_repeat1,
    STATE(1628), 1,
      sym_generic_map_aspect,
    STATE(1801), 1,
      sym__name_selector,
    STATE(803), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1912), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [51844] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(2189), 1,
      sym_BEGIN,
    STATE(824), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(804), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [51957] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1475), 1,
      anon_sym_LPAREN,
    ACTIONS(2195), 1,
      sym_PARAMETER,
    STATE(859), 1,
      sym_parenthesis_group,
    STATE(805), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [52030] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2177), 1,
      anon_sym_DOT,
    ACTIONS(2179), 1,
      anon_sym_LPAREN,
    ACTIONS(2181), 1,
      anon_sym_SQUOTE,
    ACTIONS(2183), 1,
      anon_sym_LBRACK,
    ACTIONS(2185), 1,
      sym_GENERIC,
    ACTIONS(2187), 1,
      sym_PARAMETER,
    STATE(797), 1,
      aux_sym_name_repeat1,
    STATE(1628), 1,
      sym_generic_map_aspect,
    STATE(1801), 1,
      sym__name_selector,
    STATE(806), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1912), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [52117] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1603), 1,
      sym_BEGIN,
    STATE(824), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(807), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [52230] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2197), 1,
      anon_sym_DOT,
    ACTIONS(2199), 1,
      anon_sym_LPAREN,
    ACTIONS(2201), 1,
      anon_sym_SQUOTE,
    ACTIONS(2203), 1,
      anon_sym_LBRACK,
    ACTIONS(2205), 1,
      sym_GENERIC,
    ACTIONS(2207), 1,
      sym_PARAMETER,
    STATE(809), 1,
      aux_sym_name_repeat1,
    STATE(1643), 1,
      sym_generic_map_aspect,
    STATE(1842), 1,
      sym__name_selector,
    STATE(808), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1843), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [52317] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2197), 1,
      anon_sym_DOT,
    ACTIONS(2199), 1,
      anon_sym_LPAREN,
    ACTIONS(2201), 1,
      anon_sym_SQUOTE,
    ACTIONS(2203), 1,
      anon_sym_LBRACK,
    ACTIONS(2205), 1,
      sym_GENERIC,
    ACTIONS(2207), 1,
      sym_PARAMETER,
    STATE(798), 1,
      aux_sym_name_repeat1,
    STATE(1643), 1,
      sym_generic_map_aspect,
    STATE(1842), 1,
      sym__name_selector,
    STATE(809), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1843), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [52404] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1815), 1,
      sym_BEGIN,
    STATE(814), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(810), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [52517] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1815), 1,
      sym_BEGIN,
    STATE(824), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(811), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [52630] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2209), 1,
      anon_sym_DOT,
    ACTIONS(2211), 1,
      anon_sym_LPAREN,
    ACTIONS(2213), 1,
      anon_sym_SQUOTE,
    ACTIONS(2215), 1,
      anon_sym_LBRACK,
    ACTIONS(2217), 1,
      sym_GENERIC,
    ACTIONS(2219), 1,
      sym_PARAMETER,
    STATE(813), 1,
      aux_sym_name_repeat1,
    STATE(1434), 1,
      sym_generic_map_aspect,
    STATE(1820), 1,
      sym__name_selector,
    STATE(812), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1835), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [52717] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2209), 1,
      anon_sym_DOT,
    ACTIONS(2211), 1,
      anon_sym_LPAREN,
    ACTIONS(2213), 1,
      anon_sym_SQUOTE,
    ACTIONS(2215), 1,
      anon_sym_LBRACK,
    ACTIONS(2217), 1,
      sym_GENERIC,
    ACTIONS(2219), 1,
      sym_PARAMETER,
    STATE(800), 1,
      aux_sym_name_repeat1,
    STATE(1434), 1,
      sym_generic_map_aspect,
    STATE(1820), 1,
      sym__name_selector,
    STATE(813), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1835), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [52804] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1829), 1,
      sym_BEGIN,
    STATE(824), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(814), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [52917] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1829), 1,
      sym_BEGIN,
    STATE(836), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(815), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [53030] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2221), 1,
      sym_ALIAS,
    ACTIONS(2223), 1,
      sym_ATTRIBUTE,
    ACTIONS(2225), 1,
      sym_COMPONENT,
    ACTIONS(2227), 1,
      sym_CONSTANT,
    ACTIONS(2229), 1,
      sym_DISCONNECT,
    ACTIONS(2231), 1,
      sym_END,
    ACTIONS(2233), 1,
      sym_FILE,
    ACTIONS(2235), 1,
      sym_FUNCTION,
    ACTIONS(2237), 1,
      sym_GENERIC,
    ACTIONS(2239), 1,
      sym_GROUP,
    ACTIONS(2243), 1,
      sym_NEW,
    ACTIONS(2245), 1,
      sym_PACKAGE,
    ACTIONS(2247), 1,
      sym_PROCEDURE,
    ACTIONS(2249), 1,
      sym_SIGNAL,
    ACTIONS(2251), 1,
      sym_SHARED,
    ACTIONS(2253), 1,
      sym_SUBTYPE,
    ACTIONS(2255), 1,
      sym_TYPE,
    ACTIONS(2257), 1,
      sym_USE,
    ACTIONS(2259), 1,
      sym_VARIABLE,
    ACTIONS(2261), 1,
      sym_VIEW,
    STATE(1025), 1,
      sym_package_header,
    STATE(1026), 1,
      aux_sym_package_declaration_body_repeat1,
    STATE(3015), 1,
      sym_generic_clause,
    STATE(3274), 1,
      sym__package_declarative_item,
    STATE(7256), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(816), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3273), 19,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [53149] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(2263), 1,
      sym_BEGIN,
    STATE(819), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(817), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [53262] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(2265), 1,
      sym_BEGIN,
    STATE(821), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(818), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [53375] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(2265), 1,
      sym_BEGIN,
    STATE(824), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(819), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [53488] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2267), 1,
      anon_sym_DOT,
    ACTIONS(2269), 1,
      anon_sym_LPAREN,
    ACTIONS(2271), 1,
      anon_sym_SQUOTE,
    ACTIONS(2273), 1,
      anon_sym_LBRACK,
    ACTIONS(2275), 1,
      sym_GENERIC,
    ACTIONS(2277), 1,
      sym_PARAMETER,
    STATE(826), 1,
      aux_sym_name_repeat1,
    STATE(1449), 1,
      sym_generic_map_aspect,
    STATE(1690), 1,
      sym__name_selector,
    STATE(820), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1691), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [53575] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(2279), 1,
      sym_BEGIN,
    STATE(824), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(821), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [53688] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1561), 1,
      anon_sym_DOT,
    ACTIONS(1563), 1,
      anon_sym_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_SQUOTE,
    ACTIONS(1567), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      sym_PARAMETER,
    STATE(823), 1,
      aux_sym_name_repeat1,
    STATE(888), 1,
      sym_generic_map_aspect,
    STATE(962), 1,
      sym__name_selector,
    STATE(822), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(963), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [53773] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1561), 1,
      anon_sym_DOT,
    ACTIONS(1563), 1,
      anon_sym_LPAREN,
    ACTIONS(1565), 1,
      anon_sym_SQUOTE,
    ACTIONS(1567), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      sym_PARAMETER,
    STATE(731), 1,
      aux_sym_name_repeat1,
    STATE(888), 1,
      sym_generic_map_aspect,
    STATE(962), 1,
      sym__name_selector,
    STATE(823), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(963), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [53858] = 28,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(447), 1,
      sym_BEGIN,
    ACTIONS(482), 1,
      sym_PROCEDURE,
    ACTIONS(2281), 1,
      sym_ALIAS,
    ACTIONS(2284), 1,
      sym_ATTRIBUTE,
    ACTIONS(2287), 1,
      sym_COMPONENT,
    ACTIONS(2290), 1,
      sym_CONSTANT,
    ACTIONS(2293), 1,
      sym_DISCONNECT,
    ACTIONS(2296), 1,
      sym_FILE,
    ACTIONS(2299), 1,
      sym_FOR,
    ACTIONS(2302), 1,
      sym_FUNCTION,
    ACTIONS(2305), 1,
      sym_GROUP,
    ACTIONS(2308), 1,
      sym_PACKAGE,
    ACTIONS(2311), 1,
      sym_SIGNAL,
    ACTIONS(2314), 1,
      sym_SHARED,
    ACTIONS(2317), 1,
      sym_SUBTYPE,
    ACTIONS(2320), 1,
      sym_TYPE,
    ACTIONS(2323), 1,
      sym_USE,
    ACTIONS(2326), 1,
      sym_VARIABLE,
    ACTIONS(2329), 1,
      sym_VIEW,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(476), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(824), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_design_unit_repeat2,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [53969] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(2332), 1,
      sym_BEGIN,
    STATE(804), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(825), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [54082] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2267), 1,
      anon_sym_DOT,
    ACTIONS(2269), 1,
      anon_sym_LPAREN,
    ACTIONS(2271), 1,
      anon_sym_SQUOTE,
    ACTIONS(2273), 1,
      anon_sym_LBRACK,
    ACTIONS(2275), 1,
      sym_GENERIC,
    ACTIONS(2277), 1,
      sym_PARAMETER,
    STATE(828), 1,
      aux_sym_name_repeat1,
    STATE(1449), 1,
      sym_generic_map_aspect,
    STATE(1690), 1,
      sym__name_selector,
    STATE(826), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1691), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [54169] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2221), 1,
      sym_ALIAS,
    ACTIONS(2223), 1,
      sym_ATTRIBUTE,
    ACTIONS(2225), 1,
      sym_COMPONENT,
    ACTIONS(2227), 1,
      sym_CONSTANT,
    ACTIONS(2229), 1,
      sym_DISCONNECT,
    ACTIONS(2231), 1,
      sym_END,
    ACTIONS(2233), 1,
      sym_FILE,
    ACTIONS(2235), 1,
      sym_FUNCTION,
    ACTIONS(2237), 1,
      sym_GENERIC,
    ACTIONS(2239), 1,
      sym_GROUP,
    ACTIONS(2245), 1,
      sym_PACKAGE,
    ACTIONS(2247), 1,
      sym_PROCEDURE,
    ACTIONS(2249), 1,
      sym_SIGNAL,
    ACTIONS(2251), 1,
      sym_SHARED,
    ACTIONS(2253), 1,
      sym_SUBTYPE,
    ACTIONS(2255), 1,
      sym_TYPE,
    ACTIONS(2257), 1,
      sym_USE,
    ACTIONS(2259), 1,
      sym_VARIABLE,
    ACTIONS(2261), 1,
      sym_VIEW,
    ACTIONS(2334), 1,
      sym_NEW,
    STATE(1025), 1,
      sym_package_header,
    STATE(1026), 1,
      aux_sym_package_declaration_body_repeat1,
    STATE(3015), 1,
      sym_generic_clause,
    STATE(3274), 1,
      sym__package_declarative_item,
    STATE(7256), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(827), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3273), 19,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [54288] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2336), 1,
      anon_sym_DOT,
    ACTIONS(2339), 1,
      anon_sym_LPAREN,
    ACTIONS(2342), 1,
      anon_sym_SQUOTE,
    ACTIONS(2345), 1,
      anon_sym_LBRACK,
    ACTIONS(2348), 1,
      sym_GENERIC,
    ACTIONS(2351), 1,
      sym_PARAMETER,
    STATE(1449), 1,
      sym_generic_map_aspect,
    STATE(1690), 1,
      sym__name_selector,
    STATE(828), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1691), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [54373] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(1673), 1,
      sym_BEGIN,
    STATE(824), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(829), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [54486] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2354), 1,
      anon_sym_DOT,
    ACTIONS(2356), 1,
      anon_sym_LPAREN,
    ACTIONS(2358), 1,
      anon_sym_SQUOTE,
    ACTIONS(2360), 1,
      anon_sym_LBRACK,
    ACTIONS(2362), 1,
      sym_GENERIC,
    ACTIONS(2364), 1,
      sym_PARAMETER,
    STATE(832), 1,
      aux_sym_name_repeat1,
    STATE(1574), 1,
      sym_generic_map_aspect,
    STATE(1679), 1,
      sym__name_selector,
    STATE(830), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1689), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [54573] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2221), 1,
      sym_ALIAS,
    ACTIONS(2223), 1,
      sym_ATTRIBUTE,
    ACTIONS(2225), 1,
      sym_COMPONENT,
    ACTIONS(2227), 1,
      sym_CONSTANT,
    ACTIONS(2229), 1,
      sym_DISCONNECT,
    ACTIONS(2231), 1,
      sym_END,
    ACTIONS(2233), 1,
      sym_FILE,
    ACTIONS(2235), 1,
      sym_FUNCTION,
    ACTIONS(2237), 1,
      sym_GENERIC,
    ACTIONS(2239), 1,
      sym_GROUP,
    ACTIONS(2245), 1,
      sym_PACKAGE,
    ACTIONS(2247), 1,
      sym_PROCEDURE,
    ACTIONS(2249), 1,
      sym_SIGNAL,
    ACTIONS(2251), 1,
      sym_SHARED,
    ACTIONS(2253), 1,
      sym_SUBTYPE,
    ACTIONS(2255), 1,
      sym_TYPE,
    ACTIONS(2257), 1,
      sym_USE,
    ACTIONS(2259), 1,
      sym_VARIABLE,
    ACTIONS(2261), 1,
      sym_VIEW,
    ACTIONS(2366), 1,
      sym_NEW,
    STATE(1025), 1,
      sym_package_header,
    STATE(1026), 1,
      aux_sym_package_declaration_body_repeat1,
    STATE(3015), 1,
      sym_generic_clause,
    STATE(3274), 1,
      sym__package_declarative_item,
    STATE(7256), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(831), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3273), 19,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [54692] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2354), 1,
      anon_sym_DOT,
    ACTIONS(2356), 1,
      anon_sym_LPAREN,
    ACTIONS(2358), 1,
      anon_sym_SQUOTE,
    ACTIONS(2360), 1,
      anon_sym_LBRACK,
    ACTIONS(2362), 1,
      sym_GENERIC,
    ACTIONS(2364), 1,
      sym_PARAMETER,
    STATE(799), 1,
      aux_sym_name_repeat1,
    STATE(1574), 1,
      sym_generic_map_aspect,
    STATE(1679), 1,
      sym__name_selector,
    STATE(832), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1689), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [54779] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2221), 1,
      sym_ALIAS,
    ACTIONS(2223), 1,
      sym_ATTRIBUTE,
    ACTIONS(2225), 1,
      sym_COMPONENT,
    ACTIONS(2227), 1,
      sym_CONSTANT,
    ACTIONS(2229), 1,
      sym_DISCONNECT,
    ACTIONS(2231), 1,
      sym_END,
    ACTIONS(2233), 1,
      sym_FILE,
    ACTIONS(2235), 1,
      sym_FUNCTION,
    ACTIONS(2237), 1,
      sym_GENERIC,
    ACTIONS(2239), 1,
      sym_GROUP,
    ACTIONS(2245), 1,
      sym_PACKAGE,
    ACTIONS(2247), 1,
      sym_PROCEDURE,
    ACTIONS(2249), 1,
      sym_SIGNAL,
    ACTIONS(2251), 1,
      sym_SHARED,
    ACTIONS(2253), 1,
      sym_SUBTYPE,
    ACTIONS(2255), 1,
      sym_TYPE,
    ACTIONS(2257), 1,
      sym_USE,
    ACTIONS(2259), 1,
      sym_VARIABLE,
    ACTIONS(2261), 1,
      sym_VIEW,
    ACTIONS(2368), 1,
      sym_NEW,
    STATE(1025), 1,
      sym_package_header,
    STATE(1026), 1,
      aux_sym_package_declaration_body_repeat1,
    STATE(3015), 1,
      sym_generic_clause,
    STATE(3274), 1,
      sym__package_declarative_item,
    STATE(7256), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(833), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3273), 19,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [54898] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2221), 1,
      sym_ALIAS,
    ACTIONS(2223), 1,
      sym_ATTRIBUTE,
    ACTIONS(2225), 1,
      sym_COMPONENT,
    ACTIONS(2227), 1,
      sym_CONSTANT,
    ACTIONS(2229), 1,
      sym_DISCONNECT,
    ACTIONS(2231), 1,
      sym_END,
    ACTIONS(2233), 1,
      sym_FILE,
    ACTIONS(2235), 1,
      sym_FUNCTION,
    ACTIONS(2237), 1,
      sym_GENERIC,
    ACTIONS(2239), 1,
      sym_GROUP,
    ACTIONS(2245), 1,
      sym_PACKAGE,
    ACTIONS(2247), 1,
      sym_PROCEDURE,
    ACTIONS(2249), 1,
      sym_SIGNAL,
    ACTIONS(2251), 1,
      sym_SHARED,
    ACTIONS(2253), 1,
      sym_SUBTYPE,
    ACTIONS(2255), 1,
      sym_TYPE,
    ACTIONS(2257), 1,
      sym_USE,
    ACTIONS(2259), 1,
      sym_VARIABLE,
    ACTIONS(2261), 1,
      sym_VIEW,
    ACTIONS(2370), 1,
      sym_NEW,
    STATE(1025), 1,
      sym_package_header,
    STATE(1026), 1,
      aux_sym_package_declaration_body_repeat1,
    STATE(3015), 1,
      sym_generic_clause,
    STATE(3274), 1,
      sym__package_declarative_item,
    STATE(7256), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(834), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3273), 19,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [55017] = 32,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2221), 1,
      sym_ALIAS,
    ACTIONS(2223), 1,
      sym_ATTRIBUTE,
    ACTIONS(2225), 1,
      sym_COMPONENT,
    ACTIONS(2227), 1,
      sym_CONSTANT,
    ACTIONS(2229), 1,
      sym_DISCONNECT,
    ACTIONS(2231), 1,
      sym_END,
    ACTIONS(2233), 1,
      sym_FILE,
    ACTIONS(2235), 1,
      sym_FUNCTION,
    ACTIONS(2237), 1,
      sym_GENERIC,
    ACTIONS(2239), 1,
      sym_GROUP,
    ACTIONS(2245), 1,
      sym_PACKAGE,
    ACTIONS(2247), 1,
      sym_PROCEDURE,
    ACTIONS(2249), 1,
      sym_SIGNAL,
    ACTIONS(2251), 1,
      sym_SHARED,
    ACTIONS(2253), 1,
      sym_SUBTYPE,
    ACTIONS(2255), 1,
      sym_TYPE,
    ACTIONS(2257), 1,
      sym_USE,
    ACTIONS(2259), 1,
      sym_VARIABLE,
    ACTIONS(2261), 1,
      sym_VIEW,
    ACTIONS(2372), 1,
      sym_NEW,
    STATE(1025), 1,
      sym_package_header,
    STATE(1026), 1,
      aux_sym_package_declaration_body_repeat1,
    STATE(3015), 1,
      sym_generic_clause,
    STATE(3274), 1,
      sym__package_declarative_item,
    STATE(7256), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(835), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3273), 19,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [55136] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(175), 1,
      sym_ALIAS,
    ACTIONS(177), 1,
      sym_ATTRIBUTE,
    ACTIONS(181), 1,
      sym_COMPONENT,
    ACTIONS(183), 1,
      sym_CONSTANT,
    ACTIONS(185), 1,
      sym_DISCONNECT,
    ACTIONS(189), 1,
      sym_FILE,
    ACTIONS(191), 1,
      sym_FOR,
    ACTIONS(193), 1,
      sym_FUNCTION,
    ACTIONS(195), 1,
      sym_GROUP,
    ACTIONS(197), 1,
      sym_PACKAGE,
    ACTIONS(203), 1,
      sym_SIGNAL,
    ACTIONS(205), 1,
      sym_SHARED,
    ACTIONS(207), 1,
      sym_SUBTYPE,
    ACTIONS(209), 1,
      sym_TYPE,
    ACTIONS(211), 1,
      sym_USE,
    ACTIONS(213), 1,
      sym_VARIABLE,
    ACTIONS(215), 1,
      sym_VIEW,
    ACTIONS(2263), 1,
      sym_BEGIN,
    STATE(824), 1,
      aux_sym_design_unit_repeat2,
    STATE(3070), 1,
      sym__block_declarative_item,
    STATE(5909), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(836), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3071), 22,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
      sym_configuration_specification,
  [55249] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2374), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_LPAREN,
    ACTIONS(2378), 1,
      anon_sym_SQUOTE,
    ACTIONS(2380), 1,
      anon_sym_LBRACK,
    ACTIONS(2382), 1,
      sym_GENERIC,
    ACTIONS(2384), 1,
      sym_PARAMETER,
    STATE(884), 1,
      aux_sym_name_repeat1,
    STATE(1836), 1,
      sym_generic_map_aspect,
    STATE(2144), 1,
      sym__name_selector,
    STATE(837), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2145), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [55335] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(838), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [55401] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(839), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [55467] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(840), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [55533] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2398), 1,
      anon_sym_DOT,
    ACTIONS(2401), 1,
      anon_sym_LPAREN,
    ACTIONS(2404), 1,
      anon_sym_SQUOTE,
    ACTIONS(2407), 1,
      anon_sym_LBRACK,
    ACTIONS(2410), 1,
      sym_GENERIC,
    ACTIONS(2413), 1,
      sym_PARAMETER,
    STATE(1836), 1,
      sym_generic_map_aspect,
    STATE(2144), 1,
      sym__name_selector,
    STATE(841), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(2145), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [55617] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2416), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_LPAREN,
    ACTIONS(2420), 1,
      anon_sym_SQUOTE,
    ACTIONS(2422), 1,
      anon_sym_LBRACK,
    ACTIONS(2424), 1,
      sym_GENERIC,
    ACTIONS(2426), 1,
      sym_PARAMETER,
    STATE(845), 1,
      aux_sym_name_repeat1,
    STATE(1882), 1,
      sym_generic_map_aspect,
    STATE(1913), 1,
      sym__name_selector,
    STATE(842), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2226), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [55703] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2428), 1,
      anon_sym_DOT,
    ACTIONS(2430), 1,
      anon_sym_LPAREN,
    ACTIONS(2432), 1,
      anon_sym_SQUOTE,
    ACTIONS(2434), 1,
      anon_sym_LBRACK,
    ACTIONS(2436), 1,
      sym_GENERIC,
    ACTIONS(2438), 1,
      sym_PARAMETER,
    STATE(872), 1,
      aux_sym_name_repeat1,
    STATE(1747), 1,
      sym_generic_map_aspect,
    STATE(2112), 1,
      sym__name_selector,
    STATE(843), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2113), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [55789] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(844), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [55855] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2416), 1,
      anon_sym_DOT,
    ACTIONS(2418), 1,
      anon_sym_LPAREN,
    ACTIONS(2420), 1,
      anon_sym_SQUOTE,
    ACTIONS(2422), 1,
      anon_sym_LBRACK,
    ACTIONS(2424), 1,
      sym_GENERIC,
    ACTIONS(2426), 1,
      sym_PARAMETER,
    STATE(876), 1,
      aux_sym_name_repeat1,
    STATE(1882), 1,
      sym_generic_map_aspect,
    STATE(1913), 1,
      sym__name_selector,
    STATE(845), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2226), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [55941] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2450), 1,
      sym_library_constant_unit,
    STATE(1483), 1,
      sym__unit,
    STATE(846), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2448), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [56013] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(847), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [56079] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(848), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [56145] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2460), 1,
      anon_sym_DOT,
    ACTIONS(2462), 1,
      anon_sym_LPAREN,
    ACTIONS(2464), 1,
      anon_sym_SQUOTE,
    ACTIONS(2466), 1,
      anon_sym_LBRACK,
    ACTIONS(2468), 1,
      sym_GENERIC,
    ACTIONS(2470), 1,
      sym_PARAMETER,
    STATE(851), 1,
      aux_sym_name_repeat1,
    STATE(1806), 1,
      sym_generic_map_aspect,
    STATE(2084), 1,
      sym__name_selector,
    STATE(849), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2085), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [56231] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1533), 1,
      anon_sym_LPAREN,
    ACTIONS(2472), 1,
      sym_PARAMETER,
    STATE(890), 1,
      sym_parenthesis_group,
    STATE(850), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [56303] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2474), 1,
      anon_sym_DOT,
    ACTIONS(2477), 1,
      anon_sym_LPAREN,
    ACTIONS(2480), 1,
      anon_sym_SQUOTE,
    ACTIONS(2483), 1,
      anon_sym_LBRACK,
    ACTIONS(2486), 1,
      sym_GENERIC,
    ACTIONS(2489), 1,
      sym_PARAMETER,
    STATE(1806), 1,
      sym_generic_map_aspect,
    STATE(2084), 1,
      sym__name_selector,
    STATE(851), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(2085), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [56387] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(852), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [56453] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2496), 1,
      anon_sym_DOT,
    ACTIONS(2498), 1,
      anon_sym_LPAREN,
    ACTIONS(2500), 1,
      anon_sym_SQUOTE,
    ACTIONS(2502), 1,
      anon_sym_LBRACK,
    ACTIONS(2504), 1,
      sym_GENERIC,
    ACTIONS(2506), 1,
      sym_PARAMETER,
    STATE(857), 1,
      aux_sym_name_repeat1,
    STATE(1731), 1,
      sym_generic_map_aspect,
    STATE(2108), 1,
      sym__name_selector,
    STATE(853), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2109), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [56539] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(854), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [56605] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(855), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [56671] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(856), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [56737] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2496), 1,
      anon_sym_DOT,
    ACTIONS(2498), 1,
      anon_sym_LPAREN,
    ACTIONS(2500), 1,
      anon_sym_SQUOTE,
    ACTIONS(2502), 1,
      anon_sym_LBRACK,
    ACTIONS(2504), 1,
      sym_GENERIC,
    ACTIONS(2506), 1,
      sym_PARAMETER,
    STATE(862), 1,
      aux_sym_name_repeat1,
    STATE(1731), 1,
      sym_generic_map_aspect,
    STATE(2108), 1,
      sym__name_selector,
    STATE(857), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2109), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [56823] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(858), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [56889] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(859), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [56955] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(860), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [57021] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(1977), 1,
      sym_ALIAS,
    ACTIONS(1979), 1,
      sym_ATTRIBUTE,
    ACTIONS(1983), 1,
      sym_CONSTANT,
    ACTIONS(1985), 1,
      sym_DISCONNECT,
    ACTIONS(1987), 1,
      sym_FILE,
    ACTIONS(1989), 1,
      sym_FUNCTION,
    ACTIONS(1993), 1,
      sym_GROUP,
    ACTIONS(1995), 1,
      sym_PACKAGE,
    ACTIONS(1997), 1,
      sym_PORT,
    ACTIONS(1999), 1,
      sym_SIGNAL,
    ACTIONS(2001), 1,
      sym_SHARED,
    ACTIONS(2003), 1,
      sym_SUBTYPE,
    ACTIONS(2005), 1,
      sym_TYPE,
    ACTIONS(2007), 1,
      sym_USE,
    ACTIONS(2009), 1,
      sym_VARIABLE,
    ACTIONS(2011), 1,
      sym_VIEW,
    STATE(941), 1,
      aux_sym_entity_head_repeat1,
    STATE(944), 1,
      sym_port_clause,
    STATE(3244), 1,
      sym__entity_declarative_item,
    STATE(5961), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    ACTIONS(2532), 2,
      sym_BEGIN,
      sym_END,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(861), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3241), 20,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [57133] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2534), 1,
      anon_sym_DOT,
    ACTIONS(2537), 1,
      anon_sym_LPAREN,
    ACTIONS(2540), 1,
      anon_sym_SQUOTE,
    ACTIONS(2543), 1,
      anon_sym_LBRACK,
    ACTIONS(2546), 1,
      sym_GENERIC,
    ACTIONS(2549), 1,
      sym_PARAMETER,
    STATE(1731), 1,
      sym_generic_map_aspect,
    STATE(2108), 1,
      sym__name_selector,
    STATE(862), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(2109), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [57217] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(863), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [57283] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(864), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [57349] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(865), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [57415] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(866), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [57481] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(867), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [57547] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2572), 1,
      anon_sym_DOT,
    ACTIONS(2574), 1,
      anon_sym_LPAREN,
    ACTIONS(2576), 1,
      anon_sym_SQUOTE,
    ACTIONS(2578), 1,
      anon_sym_LBRACK,
    ACTIONS(2580), 1,
      sym_GENERIC,
    ACTIONS(2582), 1,
      sym_PARAMETER,
    STATE(870), 1,
      aux_sym_name_repeat1,
    STATE(1896), 1,
      sym_generic_map_aspect,
    STATE(1980), 1,
      sym__name_selector,
    STATE(868), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1981), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [57633] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(869), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [57699] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2572), 1,
      anon_sym_DOT,
    ACTIONS(2574), 1,
      anon_sym_LPAREN,
    ACTIONS(2576), 1,
      anon_sym_SQUOTE,
    ACTIONS(2578), 1,
      anon_sym_LBRACK,
    ACTIONS(2580), 1,
      sym_GENERIC,
    ACTIONS(2582), 1,
      sym_PARAMETER,
    STATE(882), 1,
      aux_sym_name_repeat1,
    STATE(1896), 1,
      sym_generic_map_aspect,
    STATE(1980), 1,
      sym__name_selector,
    STATE(870), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(1981), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [57785] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(871), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [57851] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2592), 1,
      anon_sym_DOT,
    ACTIONS(2595), 1,
      anon_sym_LPAREN,
    ACTIONS(2598), 1,
      anon_sym_SQUOTE,
    ACTIONS(2601), 1,
      anon_sym_LBRACK,
    ACTIONS(2604), 1,
      sym_GENERIC,
    ACTIONS(2607), 1,
      sym_PARAMETER,
    STATE(1747), 1,
      sym_generic_map_aspect,
    STATE(2112), 1,
      sym__name_selector,
    STATE(872), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(2113), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [57935] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2460), 1,
      anon_sym_DOT,
    ACTIONS(2462), 1,
      anon_sym_LPAREN,
    ACTIONS(2464), 1,
      anon_sym_SQUOTE,
    ACTIONS(2466), 1,
      anon_sym_LBRACK,
    ACTIONS(2468), 1,
      sym_GENERIC,
    ACTIONS(2470), 1,
      sym_PARAMETER,
    STATE(849), 1,
      aux_sym_name_repeat1,
    STATE(1806), 1,
      sym_generic_map_aspect,
    STATE(2084), 1,
      sym__name_selector,
    STATE(873), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2085), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [58021] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(874), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [58087] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(875), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [58153] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2618), 1,
      anon_sym_DOT,
    ACTIONS(2621), 1,
      anon_sym_LPAREN,
    ACTIONS(2624), 1,
      anon_sym_SQUOTE,
    ACTIONS(2627), 1,
      anon_sym_LBRACK,
    ACTIONS(2630), 1,
      sym_GENERIC,
    ACTIONS(2633), 1,
      sym_PARAMETER,
    STATE(1882), 1,
      sym_generic_map_aspect,
    STATE(1913), 1,
      sym__name_selector,
    STATE(876), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(2226), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [58237] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(877), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [58303] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(878), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [58369] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(879), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [58435] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(880), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [58501] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(881), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [58567] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2656), 1,
      anon_sym_DOT,
    ACTIONS(2659), 1,
      anon_sym_LPAREN,
    ACTIONS(2662), 1,
      anon_sym_SQUOTE,
    ACTIONS(2665), 1,
      anon_sym_LBRACK,
    ACTIONS(2668), 1,
      sym_GENERIC,
    ACTIONS(2671), 1,
      sym_PARAMETER,
    STATE(1896), 1,
      sym_generic_map_aspect,
    STATE(1980), 1,
      sym__name_selector,
    STATE(882), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(1981), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1499), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1491), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [58651] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2428), 1,
      anon_sym_DOT,
    ACTIONS(2430), 1,
      anon_sym_LPAREN,
    ACTIONS(2432), 1,
      anon_sym_SQUOTE,
    ACTIONS(2434), 1,
      anon_sym_LBRACK,
    ACTIONS(2436), 1,
      sym_GENERIC,
    ACTIONS(2438), 1,
      sym_PARAMETER,
    STATE(843), 1,
      aux_sym_name_repeat1,
    STATE(1747), 1,
      sym_generic_map_aspect,
    STATE(2112), 1,
      sym__name_selector,
    STATE(883), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2113), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1477), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1471), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [58737] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2374), 1,
      anon_sym_DOT,
    ACTIONS(2376), 1,
      anon_sym_LPAREN,
    ACTIONS(2378), 1,
      anon_sym_SQUOTE,
    ACTIONS(2380), 1,
      anon_sym_LBRACK,
    ACTIONS(2382), 1,
      sym_GENERIC,
    ACTIONS(2384), 1,
      sym_PARAMETER,
    STATE(841), 1,
      aux_sym_name_repeat1,
    STATE(1836), 1,
      sym_generic_map_aspect,
    STATE(2144), 1,
      sym__name_selector,
    STATE(884), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(2145), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1489), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1487), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [58823] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(885), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [58889] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(886), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 40,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [58955] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(887), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59020] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1563), 1,
      anon_sym_LPAREN,
    ACTIONS(2682), 1,
      sym_PARAMETER,
    STATE(938), 1,
      sym_parenthesis_group,
    STATE(888), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 36,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59091] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1575), 1,
      anon_sym_LPAREN,
    ACTIONS(2684), 1,
      sym_PARAMETER,
    STATE(976), 1,
      sym_parenthesis_group,
    STATE(889), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59162] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(890), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59227] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(891), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59292] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(892), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59357] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(893), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 39,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [59422] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(894), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59487] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2700), 1,
      sym_exponentiate,
    STATE(441), 1,
      sym_adding_operator,
    STATE(442), 1,
      sym_multiplying_operator,
    STATE(593), 1,
      sym_logical_operator,
    STATE(594), 1,
      sym_relational_operator,
    STATE(595), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(895), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 7,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2690), 9,
      sym_DOWNTO,
      sym_TO,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [59576] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(896), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59641] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(897), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59706] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2700), 1,
      sym_exponentiate,
    STATE(441), 1,
      sym_adding_operator,
    STATE(442), 1,
      sym_multiplying_operator,
    STATE(593), 1,
      sym_logical_operator,
    STATE(594), 1,
      sym_relational_operator,
    STATE(595), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(898), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 7,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 15,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [59793] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(899), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59858] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(900), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [59923] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2700), 1,
      sym_exponentiate,
    STATE(441), 1,
      sym_adding_operator,
    STATE(442), 1,
      sym_multiplying_operator,
    STATE(593), 1,
      sym_logical_operator,
    STATE(594), 1,
      sym_relational_operator,
    STATE(595), 1,
      sym_shift_operator,
    STATE(901), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 33,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [60000] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(902), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 39,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [60065] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(903), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60130] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(904), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60195] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(905), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60260] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(906), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60325] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(907), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60390] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(908), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60455] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2700), 1,
      sym_exponentiate,
    STATE(441), 1,
      sym_adding_operator,
    STATE(442), 1,
      sym_multiplying_operator,
    STATE(593), 1,
      sym_logical_operator,
    STATE(594), 1,
      sym_relational_operator,
    STATE(595), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(909), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 31,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [60536] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(910), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60601] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(911), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60666] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(912), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60731] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(913), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60796] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(914), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60861] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(915), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [60926] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(441), 1,
      sym_adding_operator,
    STATE(442), 1,
      sym_multiplying_operator,
    STATE(593), 1,
      sym_logical_operator,
    STATE(594), 1,
      sym_relational_operator,
    STATE(595), 1,
      sym_shift_operator,
    STATE(916), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [61001] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(917), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [61066] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(441), 1,
      sym_adding_operator,
    STATE(442), 1,
      sym_multiplying_operator,
    STATE(593), 1,
      sym_logical_operator,
    STATE(594), 1,
      sym_relational_operator,
    STATE(595), 1,
      sym_shift_operator,
    STATE(918), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [61141] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2700), 1,
      sym_exponentiate,
    STATE(441), 1,
      sym_adding_operator,
    STATE(442), 1,
      sym_multiplying_operator,
    STATE(593), 1,
      sym_logical_operator,
    STATE(594), 1,
      sym_relational_operator,
    STATE(595), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(919), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 22,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [61226] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2700), 1,
      sym_exponentiate,
    STATE(441), 1,
      sym_adding_operator,
    STATE(442), 1,
      sym_multiplying_operator,
    STATE(593), 1,
      sym_logical_operator,
    STATE(594), 1,
      sym_relational_operator,
    STATE(595), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(920), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 31,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [61307] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(921), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [61372] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(922), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [61437] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(923), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [61502] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(924), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [61567] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(925), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [61632] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(926), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [61697] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2700), 1,
      sym_exponentiate,
    STATE(441), 1,
      sym_adding_operator,
    STATE(442), 1,
      sym_multiplying_operator,
    STATE(593), 1,
      sym_logical_operator,
    STATE(594), 1,
      sym_relational_operator,
    STATE(595), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(927), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 28,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [61780] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(928), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 39,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [61845] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(929), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [61909] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(930), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [61973] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(931), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62037] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(932), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62101] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(933), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62165] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(934), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62229] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(935), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62293] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(936), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62357] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(937), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62421] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(938), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62485] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(939), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62549] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(940), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62613] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(1977), 1,
      sym_ALIAS,
    ACTIONS(1979), 1,
      sym_ATTRIBUTE,
    ACTIONS(1983), 1,
      sym_CONSTANT,
    ACTIONS(1985), 1,
      sym_DISCONNECT,
    ACTIONS(1987), 1,
      sym_FILE,
    ACTIONS(1989), 1,
      sym_FUNCTION,
    ACTIONS(1993), 1,
      sym_GROUP,
    ACTIONS(1995), 1,
      sym_PACKAGE,
    ACTIONS(1999), 1,
      sym_SIGNAL,
    ACTIONS(2001), 1,
      sym_SHARED,
    ACTIONS(2003), 1,
      sym_SUBTYPE,
    ACTIONS(2005), 1,
      sym_TYPE,
    ACTIONS(2007), 1,
      sym_USE,
    ACTIONS(2009), 1,
      sym_VARIABLE,
    ACTIONS(2011), 1,
      sym_VIEW,
    STATE(945), 1,
      aux_sym_entity_head_repeat1,
    STATE(3244), 1,
      sym__entity_declarative_item,
    STATE(5961), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    ACTIONS(2730), 2,
      sym_BEGIN,
      sym_END,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(941), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3241), 20,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [62719] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(942), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62783] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(943), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [62847] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(1977), 1,
      sym_ALIAS,
    ACTIONS(1979), 1,
      sym_ATTRIBUTE,
    ACTIONS(1983), 1,
      sym_CONSTANT,
    ACTIONS(1985), 1,
      sym_DISCONNECT,
    ACTIONS(1987), 1,
      sym_FILE,
    ACTIONS(1989), 1,
      sym_FUNCTION,
    ACTIONS(1993), 1,
      sym_GROUP,
    ACTIONS(1995), 1,
      sym_PACKAGE,
    ACTIONS(1999), 1,
      sym_SIGNAL,
    ACTIONS(2001), 1,
      sym_SHARED,
    ACTIONS(2003), 1,
      sym_SUBTYPE,
    ACTIONS(2005), 1,
      sym_TYPE,
    ACTIONS(2007), 1,
      sym_USE,
    ACTIONS(2009), 1,
      sym_VARIABLE,
    ACTIONS(2011), 1,
      sym_VIEW,
    STATE(958), 1,
      aux_sym_entity_head_repeat1,
    STATE(3244), 1,
      sym__entity_declarative_item,
    STATE(5961), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    ACTIONS(2730), 2,
      sym_BEGIN,
      sym_END,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(944), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3241), 20,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [62953] = 26,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2732), 1,
      sym_ALIAS,
    ACTIONS(2735), 1,
      sym_ATTRIBUTE,
    ACTIONS(2740), 1,
      sym_CONSTANT,
    ACTIONS(2743), 1,
      sym_DISCONNECT,
    ACTIONS(2746), 1,
      sym_FILE,
    ACTIONS(2749), 1,
      sym_FUNCTION,
    ACTIONS(2752), 1,
      sym_GROUP,
    ACTIONS(2758), 1,
      sym_PACKAGE,
    ACTIONS(2761), 1,
      sym_PROCEDURE,
    ACTIONS(2764), 1,
      sym_SIGNAL,
    ACTIONS(2767), 1,
      sym_SHARED,
    ACTIONS(2770), 1,
      sym_SUBTYPE,
    ACTIONS(2773), 1,
      sym_TYPE,
    ACTIONS(2776), 1,
      sym_USE,
    ACTIONS(2779), 1,
      sym_VARIABLE,
    ACTIONS(2782), 1,
      sym_VIEW,
    STATE(3244), 1,
      sym__entity_declarative_item,
    STATE(5961), 1,
      sym__subprogram_specification,
    ACTIONS(2738), 2,
      sym_BEGIN,
      sym_END,
    ACTIONS(2755), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(945), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_entity_head_repeat1,
    STATE(3241), 20,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [63057] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(946), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63121] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(947), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63185] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(948), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63249] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(949), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63313] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(950), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63377] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(951), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63441] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(952), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63505] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(953), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63569] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(954), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63633] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(955), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63697] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1677), 1,
      anon_sym_LPAREN,
    ACTIONS(2785), 1,
      sym_PARAMETER,
    STATE(1009), 1,
      sym_parenthesis_group,
    STATE(956), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 36,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [63767] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(957), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [63831] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(1977), 1,
      sym_ALIAS,
    ACTIONS(1979), 1,
      sym_ATTRIBUTE,
    ACTIONS(1983), 1,
      sym_CONSTANT,
    ACTIONS(1985), 1,
      sym_DISCONNECT,
    ACTIONS(1987), 1,
      sym_FILE,
    ACTIONS(1989), 1,
      sym_FUNCTION,
    ACTIONS(1993), 1,
      sym_GROUP,
    ACTIONS(1995), 1,
      sym_PACKAGE,
    ACTIONS(1999), 1,
      sym_SIGNAL,
    ACTIONS(2001), 1,
      sym_SHARED,
    ACTIONS(2003), 1,
      sym_SUBTYPE,
    ACTIONS(2005), 1,
      sym_TYPE,
    ACTIONS(2007), 1,
      sym_USE,
    ACTIONS(2009), 1,
      sym_VARIABLE,
    ACTIONS(2011), 1,
      sym_VIEW,
    STATE(945), 1,
      aux_sym_entity_head_repeat1,
    STATE(3244), 1,
      sym__entity_declarative_item,
    STATE(5961), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    ACTIONS(2787), 2,
      sym_BEGIN,
      sym_END,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(958), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3241), 20,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [63937] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(959), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64001] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(1977), 1,
      sym_ALIAS,
    ACTIONS(1979), 1,
      sym_ATTRIBUTE,
    ACTIONS(1983), 1,
      sym_CONSTANT,
    ACTIONS(1985), 1,
      sym_DISCONNECT,
    ACTIONS(1987), 1,
      sym_FILE,
    ACTIONS(1989), 1,
      sym_FUNCTION,
    ACTIONS(1993), 1,
      sym_GROUP,
    ACTIONS(1995), 1,
      sym_PACKAGE,
    ACTIONS(1999), 1,
      sym_SIGNAL,
    ACTIONS(2001), 1,
      sym_SHARED,
    ACTIONS(2003), 1,
      sym_SUBTYPE,
    ACTIONS(2005), 1,
      sym_TYPE,
    ACTIONS(2007), 1,
      sym_USE,
    ACTIONS(2009), 1,
      sym_VARIABLE,
    ACTIONS(2011), 1,
      sym_VIEW,
    STATE(941), 1,
      aux_sym_entity_head_repeat1,
    STATE(3244), 1,
      sym__entity_declarative_item,
    STATE(5961), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    ACTIONS(2532), 2,
      sym_BEGIN,
      sym_END,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(960), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3241), 20,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [64107] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(1977), 1,
      sym_ALIAS,
    ACTIONS(1979), 1,
      sym_ATTRIBUTE,
    ACTIONS(1983), 1,
      sym_CONSTANT,
    ACTIONS(1985), 1,
      sym_DISCONNECT,
    ACTIONS(1987), 1,
      sym_FILE,
    ACTIONS(1989), 1,
      sym_FUNCTION,
    ACTIONS(1993), 1,
      sym_GROUP,
    ACTIONS(1995), 1,
      sym_PACKAGE,
    ACTIONS(1999), 1,
      sym_SIGNAL,
    ACTIONS(2001), 1,
      sym_SHARED,
    ACTIONS(2003), 1,
      sym_SUBTYPE,
    ACTIONS(2005), 1,
      sym_TYPE,
    ACTIONS(2007), 1,
      sym_USE,
    ACTIONS(2009), 1,
      sym_VARIABLE,
    ACTIONS(2011), 1,
      sym_VIEW,
    STATE(945), 1,
      aux_sym_entity_head_repeat1,
    STATE(3244), 1,
      sym__entity_declarative_item,
    STATE(5961), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    ACTIONS(2532), 2,
      sym_BEGIN,
      sym_END,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(961), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3241), 20,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [64213] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(962), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64277] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(963), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64341] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(964), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64405] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2791), 1,
      sym_library_constant_unit,
    STATE(1956), 1,
      sym__unit,
    STATE(965), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2789), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [64475] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(966), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64539] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(967), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64603] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(968), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64667] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(969), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64731] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(970), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64795] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(971), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64859] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(972), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64923] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(973), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [64987] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(974), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65051] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(975), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65115] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(976), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65179] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(977), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65243] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(978), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65307] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(979), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65371] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(980), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65435] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(981), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65499] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(982), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65563] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(983), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65627] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(984), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65691] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(985), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65755] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(986), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65819] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(987), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65883] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(988), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [65947] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(989), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [66011] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(990), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [66075] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(991), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [66139] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1689), 1,
      anon_sym_LPAREN,
    ACTIONS(2793), 1,
      sym_PARAMETER,
    STATE(1090), 1,
      sym_parenthesis_group,
    STATE(992), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [66209] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(993), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [66273] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(994), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 38,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [66337] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1645), 1,
      anon_sym_LPAREN,
    ACTIONS(2795), 1,
      sym_PARAMETER,
    STATE(1113), 1,
      sym_parenthesis_group,
    STATE(995), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 36,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [66407] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(996), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 38,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [66471] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2799), 1,
      sym_library_constant_unit,
    STATE(2141), 1,
      sym__unit,
    STATE(997), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2797), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [66541] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    STATE(998), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [66614] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(999), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [66677] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1000), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [66740] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1001), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [66803] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1002), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [66866] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1003), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [66929] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1004), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [66992] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1005), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67055] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1006), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67118] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1007), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67181] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1008), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67244] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1009), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67307] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1010), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 29,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [67386] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1011), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67449] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1012), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67512] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1013), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67575] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1014), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67638] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1015), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67701] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1016), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67764] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1017), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67827] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1713), 1,
      anon_sym_LPAREN,
    ACTIONS(2803), 1,
      sym_PARAMETER,
    STATE(1157), 1,
      sym_parenthesis_group,
    STATE(1018), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [67896] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2807), 1,
      sym_CONTEXT,
    ACTIONS(2810), 1,
      sym_LIBRARY,
    ACTIONS(2813), 1,
      sym_USE,
    STATE(2270), 1,
      sym__context_item,
    STATE(1917), 3,
      sym_library_clause,
      sym_use_clause,
      sym_context_reference,
    STATE(1019), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_design_unit_repeat1,
    ACTIONS(2805), 36,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [67967] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1020), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68030] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1021), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68093] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1022), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2690), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_PIPE,
    ACTIONS(2692), 7,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [68180] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1023), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68243] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1024), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68306] = 28,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2221), 1,
      sym_ALIAS,
    ACTIONS(2223), 1,
      sym_ATTRIBUTE,
    ACTIONS(2225), 1,
      sym_COMPONENT,
    ACTIONS(2227), 1,
      sym_CONSTANT,
    ACTIONS(2229), 1,
      sym_DISCONNECT,
    ACTIONS(2233), 1,
      sym_FILE,
    ACTIONS(2235), 1,
      sym_FUNCTION,
    ACTIONS(2239), 1,
      sym_GROUP,
    ACTIONS(2245), 1,
      sym_PACKAGE,
    ACTIONS(2247), 1,
      sym_PROCEDURE,
    ACTIONS(2249), 1,
      sym_SIGNAL,
    ACTIONS(2251), 1,
      sym_SHARED,
    ACTIONS(2253), 1,
      sym_SUBTYPE,
    ACTIONS(2255), 1,
      sym_TYPE,
    ACTIONS(2257), 1,
      sym_USE,
    ACTIONS(2259), 1,
      sym_VARIABLE,
    ACTIONS(2261), 1,
      sym_VIEW,
    ACTIONS(2816), 1,
      sym_END,
    STATE(1072), 1,
      aux_sym_package_declaration_body_repeat1,
    STATE(3274), 1,
      sym__package_declarative_item,
    STATE(7256), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1025), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3273), 19,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [68413] = 28,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2221), 1,
      sym_ALIAS,
    ACTIONS(2223), 1,
      sym_ATTRIBUTE,
    ACTIONS(2225), 1,
      sym_COMPONENT,
    ACTIONS(2227), 1,
      sym_CONSTANT,
    ACTIONS(2229), 1,
      sym_DISCONNECT,
    ACTIONS(2233), 1,
      sym_FILE,
    ACTIONS(2235), 1,
      sym_FUNCTION,
    ACTIONS(2239), 1,
      sym_GROUP,
    ACTIONS(2245), 1,
      sym_PACKAGE,
    ACTIONS(2247), 1,
      sym_PROCEDURE,
    ACTIONS(2249), 1,
      sym_SIGNAL,
    ACTIONS(2251), 1,
      sym_SHARED,
    ACTIONS(2253), 1,
      sym_SUBTYPE,
    ACTIONS(2255), 1,
      sym_TYPE,
    ACTIONS(2257), 1,
      sym_USE,
    ACTIONS(2259), 1,
      sym_VARIABLE,
    ACTIONS(2261), 1,
      sym_VIEW,
    ACTIONS(2816), 1,
      sym_END,
    STATE(1075), 1,
      aux_sym_package_declaration_body_repeat1,
    STATE(3274), 1,
      sym__package_declarative_item,
    STATE(7256), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1026), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3273), 19,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [68520] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1027), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68583] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1028), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68646] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1029), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68709] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1030), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68772] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1031), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68835] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1032), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68898] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1033), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [68961] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1034), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [69024] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1035), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [69087] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1036), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [69150] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1037), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [69213] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1038), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [69276] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1039), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 7,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 13,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [69361] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1040), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 20,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [69444] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(199), 1,
      sym_POSTPONED,
    ACTIONS(201), 1,
      sym_PROCESS,
    ACTIONS(2820), 1,
      sym_END,
    STATE(1401), 1,
      aux_sym_design_unit_repeat3,
    STATE(2658), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(3725), 1,
      sym__concurrent_statement,
    STATE(5504), 1,
      sym_generate_block,
    STATE(5761), 1,
      sym_generate_block_end,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6403), 1,
      sym__target,
    STATE(6480), 1,
      sym_with_expression,
    STATE(6481), 1,
      sym_name,
    STATE(7169), 1,
      sym__label,
    STATE(8292), 1,
      sym_assertion,
    ACTIONS(2818), 2,
      sym_ELSE,
      sym_ELSIF,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1041), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(3814), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [69553] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1042), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [69616] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(199), 1,
      sym_POSTPONED,
    ACTIONS(201), 1,
      sym_PROCESS,
    ACTIONS(2825), 1,
      sym_END,
    STATE(1401), 1,
      aux_sym_design_unit_repeat3,
    STATE(2658), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(3725), 1,
      sym__concurrent_statement,
    STATE(5512), 1,
      sym_generate_block,
    STATE(5778), 1,
      sym_generate_block_end,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6403), 1,
      sym__target,
    STATE(6480), 1,
      sym_with_expression,
    STATE(6481), 1,
      sym_name,
    STATE(7169), 1,
      sym__label,
    STATE(8292), 1,
      sym_assertion,
    ACTIONS(2823), 2,
      sym_ELSE,
      sym_ELSIF,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1043), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(3814), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [69725] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1044), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 37,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [69788] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1045), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [69851] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1046), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [69914] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1047), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [69977] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1048), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 37,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [70040] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1701), 1,
      anon_sym_LPAREN,
    ACTIONS(2828), 1,
      sym_PARAMETER,
    STATE(1181), 1,
      sym_parenthesis_group,
    STATE(1049), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [70109] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2700), 1,
      sym_exponentiate,
    STATE(441), 1,
      sym_adding_operator,
    STATE(442), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1050), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 7,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2830), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [70196] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(692), 1,
      sym_adding_operator,
    STATE(693), 1,
      sym_multiplying_operator,
    STATE(698), 1,
      sym_logical_operator,
    STATE(699), 1,
      sym_relational_operator,
    STATE(700), 1,
      sym_shift_operator,
    STATE(1051), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [70269] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2832), 1,
      sym_exponentiate,
    STATE(692), 1,
      sym_adding_operator,
    STATE(693), 1,
      sym_multiplying_operator,
    STATE(698), 1,
      sym_logical_operator,
    STATE(699), 1,
      sym_relational_operator,
    STATE(700), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1052), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 29,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_PIPE,
  [70348] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(692), 1,
      sym_adding_operator,
    STATE(693), 1,
      sym_multiplying_operator,
    STATE(698), 1,
      sym_logical_operator,
    STATE(699), 1,
      sym_relational_operator,
    STATE(700), 1,
      sym_shift_operator,
    STATE(1053), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [70421] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2832), 1,
      sym_exponentiate,
    STATE(692), 1,
      sym_adding_operator,
    STATE(693), 1,
      sym_multiplying_operator,
    STATE(698), 1,
      sym_logical_operator,
    STATE(699), 1,
      sym_relational_operator,
    STATE(700), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1054), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 29,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_PIPE,
  [70500] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2832), 1,
      sym_exponentiate,
    STATE(692), 1,
      sym_adding_operator,
    STATE(693), 1,
      sym_multiplying_operator,
    STATE(698), 1,
      sym_logical_operator,
    STATE(699), 1,
      sym_relational_operator,
    STATE(700), 1,
      sym_shift_operator,
    STATE(1055), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 31,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_PIPE,
  [70575] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2836), 1,
      sym_library_constant_unit,
    STATE(2397), 1,
      sym__unit,
    STATE(1056), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2834), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [70644] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1057), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 26,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [70725] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2832), 1,
      sym_exponentiate,
    STATE(692), 1,
      sym_adding_operator,
    STATE(693), 1,
      sym_multiplying_operator,
    STATE(698), 1,
      sym_logical_operator,
    STATE(699), 1,
      sym_relational_operator,
    STATE(700), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1058), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2690), 7,
      sym_DOWNTO,
      sym_TO,
      sym_WHEN,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_PIPE,
    ACTIONS(2692), 7,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [70812] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1059), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 29,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [70891] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2840), 1,
      sym_library_constant_unit,
    STATE(2375), 1,
      sym__unit,
    STATE(1060), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [70960] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1061), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [71023] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1062), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [71086] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1063), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [71149] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    STATE(1064), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [71224] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2832), 1,
      sym_exponentiate,
    STATE(692), 1,
      sym_adding_operator,
    STATE(693), 1,
      sym_multiplying_operator,
    STATE(698), 1,
      sym_logical_operator,
    STATE(699), 1,
      sym_relational_operator,
    STATE(700), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1065), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 7,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 13,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_PIPE,
  [71309] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2832), 1,
      sym_exponentiate,
    STATE(692), 1,
      sym_adding_operator,
    STATE(693), 1,
      sym_multiplying_operator,
    STATE(698), 1,
      sym_logical_operator,
    STATE(699), 1,
      sym_relational_operator,
    STATE(700), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1066), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 20,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PIPE,
  [71392] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2832), 1,
      sym_exponentiate,
    STATE(692), 1,
      sym_adding_operator,
    STATE(693), 1,
      sym_multiplying_operator,
    STATE(698), 1,
      sym_logical_operator,
    STATE(699), 1,
      sym_relational_operator,
    STATE(700), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1067), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 26,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PIPE,
  [71473] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1068), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [71536] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1069), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [71599] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1070), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 37,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [71662] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2844), 1,
      sym_library_constant_unit,
    STATE(2409), 1,
      sym__unit,
    STATE(1071), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2842), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [71731] = 28,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2221), 1,
      sym_ALIAS,
    ACTIONS(2223), 1,
      sym_ATTRIBUTE,
    ACTIONS(2225), 1,
      sym_COMPONENT,
    ACTIONS(2227), 1,
      sym_CONSTANT,
    ACTIONS(2229), 1,
      sym_DISCONNECT,
    ACTIONS(2233), 1,
      sym_FILE,
    ACTIONS(2235), 1,
      sym_FUNCTION,
    ACTIONS(2239), 1,
      sym_GROUP,
    ACTIONS(2245), 1,
      sym_PACKAGE,
    ACTIONS(2247), 1,
      sym_PROCEDURE,
    ACTIONS(2249), 1,
      sym_SIGNAL,
    ACTIONS(2251), 1,
      sym_SHARED,
    ACTIONS(2253), 1,
      sym_SUBTYPE,
    ACTIONS(2255), 1,
      sym_TYPE,
    ACTIONS(2257), 1,
      sym_USE,
    ACTIONS(2259), 1,
      sym_VARIABLE,
    ACTIONS(2261), 1,
      sym_VIEW,
    ACTIONS(2846), 1,
      sym_END,
    STATE(1075), 1,
      aux_sym_package_declaration_body_repeat1,
    STATE(3274), 1,
      sym__package_declarative_item,
    STATE(7256), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1072), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3273), 19,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [71838] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1073), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [71901] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1074), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [71964] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2848), 1,
      sym_ALIAS,
    ACTIONS(2851), 1,
      sym_ATTRIBUTE,
    ACTIONS(2854), 1,
      sym_COMPONENT,
    ACTIONS(2857), 1,
      sym_CONSTANT,
    ACTIONS(2860), 1,
      sym_DISCONNECT,
    ACTIONS(2863), 1,
      sym_END,
    ACTIONS(2865), 1,
      sym_FILE,
    ACTIONS(2868), 1,
      sym_FUNCTION,
    ACTIONS(2871), 1,
      sym_GROUP,
    ACTIONS(2877), 1,
      sym_PACKAGE,
    ACTIONS(2880), 1,
      sym_PROCEDURE,
    ACTIONS(2883), 1,
      sym_SIGNAL,
    ACTIONS(2886), 1,
      sym_SHARED,
    ACTIONS(2889), 1,
      sym_SUBTYPE,
    ACTIONS(2892), 1,
      sym_TYPE,
    ACTIONS(2895), 1,
      sym_USE,
    ACTIONS(2898), 1,
      sym_VARIABLE,
    ACTIONS(2901), 1,
      sym_VIEW,
    STATE(3274), 1,
      sym__package_declarative_item,
    STATE(7256), 1,
      sym__subprogram_specification,
    ACTIONS(2874), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1075), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_package_declaration_body_repeat1,
    STATE(3273), 19,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_mode_view_declaration,
      sym_constant_declaration,
      sym_signal_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_component_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_disconnection_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [72069] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    STATE(1076), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [72142] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1077), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72205] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1078), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72268] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1079), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72331] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1080), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72394] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1081), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [72457] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1082), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [72520] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1083), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72583] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1084), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72646] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1085), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72709] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1086), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72772] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1087), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72835] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1088), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72898] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1089), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [72961] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1090), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73024] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1091), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [73087] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1092), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73150] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1093), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73213] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1094), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73276] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1095), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73339] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1096), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73402] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1097), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73465] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1098), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73528] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1099), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73591] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1100), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73654] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1101), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73717] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2832), 1,
      sym_exponentiate,
    ACTIONS(2906), 1,
      sym_WHEN,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(692), 1,
      sym_adding_operator,
    STATE(693), 1,
      sym_multiplying_operator,
    STATE(5145), 1,
      aux_sym_conditional_expression_repeat1,
    STATE(6496), 1,
      sym_when_expression,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2830), 2,
      anon_sym_EQ_GT,
      anon_sym_PIPE,
    ACTIONS(2904), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1102), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 7,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [73812] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1103), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [73875] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1104), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [73938] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1105), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [74001] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1106), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74064] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1107), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74127] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1108), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [74190] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1109), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74253] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1110), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 37,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [74316] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1111), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74379] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1112), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74442] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1113), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74505] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1114), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74568] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1115), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74631] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1116), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74694] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1117), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74757] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1118), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74820] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1119), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 38,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [74883] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1120), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 37,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [74946] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1875), 1,
      anon_sym_LPAREN,
    ACTIONS(2908), 1,
      sym_PARAMETER,
    STATE(1308), 1,
      sym_parenthesis_group,
    STATE(1121), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 34,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [75014] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1122), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [75076] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2910), 1,
      sym_exponentiate,
    STATE(486), 1,
      sym_adding_operator,
    STATE(487), 1,
      sym_multiplying_operator,
    STATE(635), 1,
      sym_logical_operator,
    STATE(636), 1,
      sym_relational_operator,
    STATE(637), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1123), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 12,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      sym_variable_assignment,
  [75160] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2910), 1,
      sym_exponentiate,
    STATE(486), 1,
      sym_adding_operator,
    STATE(487), 1,
      sym_multiplying_operator,
    STATE(635), 1,
      sym_logical_operator,
    STATE(636), 1,
      sym_relational_operator,
    STATE(637), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1124), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 20,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [75242] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1125), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [75304] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1126), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [75366] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1127), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 36,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [75428] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2910), 1,
      sym_exponentiate,
    STATE(486), 1,
      sym_adding_operator,
    STATE(487), 1,
      sym_multiplying_operator,
    STATE(635), 1,
      sym_logical_operator,
    STATE(636), 1,
      sym_relational_operator,
    STATE(637), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1128), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 29,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [75506] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1129), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [75568] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1130), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [75630] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1131), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [75692] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1132), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [75754] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1133), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [75816] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2912), 1,
      sym_exponentiate,
    STATE(522), 1,
      sym_adding_operator,
    STATE(523), 1,
      sym_multiplying_operator,
    STATE(648), 1,
      sym_logical_operator,
    STATE(649), 1,
      sym_relational_operator,
    STATE(650), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1134), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 20,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [75898] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    STATE(1135), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [75970] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1136), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 28,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_PIPE,
  [76048] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2910), 1,
      sym_exponentiate,
    STATE(486), 1,
      sym_adding_operator,
    STATE(487), 1,
      sym_multiplying_operator,
    STATE(635), 1,
      sym_logical_operator,
    STATE(636), 1,
      sym_relational_operator,
    STATE(637), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1137), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 26,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [76128] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2918), 1,
      sym_library_constant_unit,
    STATE(2461), 1,
      sym__unit,
    STATE(1138), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2916), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [76196] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2912), 1,
      sym_exponentiate,
    STATE(522), 1,
      sym_adding_operator,
    STATE(523), 1,
      sym_multiplying_operator,
    STATE(648), 1,
      sym_logical_operator,
    STATE(649), 1,
      sym_relational_operator,
    STATE(650), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1139), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 26,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [76276] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2922), 1,
      sym_library_constant_unit,
    STATE(2447), 1,
      sym__unit,
    STATE(1140), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2920), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [76344] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1141), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [76406] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1142), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [76468] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1143), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [76530] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1144), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [76592] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1145), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [76654] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    STATE(1146), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [76726] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(221), 1,
      sym_POSTPONED,
    ACTIONS(223), 1,
      sym_PROCESS,
    ACTIONS(2924), 1,
      sym_END,
    ACTIONS(2927), 1,
      sym_WHEN,
    STATE(1603), 1,
      aux_sym_design_unit_repeat3,
    STATE(2636), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4169), 1,
      sym__concurrent_statement,
    STATE(5679), 1,
      sym_generate_block,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6435), 1,
      sym__target,
    STATE(6493), 1,
      sym_with_expression,
    STATE(6555), 1,
      sym_name,
    STATE(6573), 1,
      sym_generate_block_end,
    STATE(6638), 1,
      sym_assertion,
    STATE(7169), 1,
      sym__label,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1147), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4170), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [76834] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1148), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [76896] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1149), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [76958] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1150), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 28,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_PIPE,
  [77036] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    STATE(1151), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 30,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      anon_sym_PIPE,
  [77110] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1152), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [77172] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(221), 1,
      sym_POSTPONED,
    ACTIONS(223), 1,
      sym_PROCESS,
    ACTIONS(2929), 1,
      sym_END,
    ACTIONS(2932), 1,
      sym_WHEN,
    STATE(1603), 1,
      aux_sym_design_unit_repeat3,
    STATE(2636), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4169), 1,
      sym__concurrent_statement,
    STATE(5682), 1,
      sym_generate_block,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6435), 1,
      sym__target,
    STATE(6493), 1,
      sym_with_expression,
    STATE(6555), 1,
      sym_name,
    STATE(6586), 1,
      sym_generate_block_end,
    STATE(6638), 1,
      sym_assertion,
    STATE(7169), 1,
      sym__label,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1153), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4170), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [77280] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1154), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 37,
      sym_identifier,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [77342] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1155), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [77404] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1156), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [77466] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1157), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [77528] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1158), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 37,
      sym_identifier,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [77590] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1159), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [77652] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(486), 1,
      sym_adding_operator,
    STATE(487), 1,
      sym_multiplying_operator,
    STATE(635), 1,
      sym_logical_operator,
    STATE(636), 1,
      sym_relational_operator,
    STATE(637), 1,
      sym_shift_operator,
    STATE(1160), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [77724] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1887), 1,
      anon_sym_LPAREN,
    ACTIONS(2934), 1,
      sym_PARAMETER,
    STATE(1346), 1,
      sym_parenthesis_group,
    STATE(1161), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 34,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [77792] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2910), 1,
      sym_exponentiate,
    STATE(486), 1,
      sym_adding_operator,
    STATE(487), 1,
      sym_multiplying_operator,
    STATE(635), 1,
      sym_logical_operator,
    STATE(636), 1,
      sym_relational_operator,
    STATE(637), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1162), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2690), 6,
      sym_BUS,
      sym_DOWNTO,
      sym_REGISTER,
      sym_TO,
      anon_sym_SEMI,
      sym_variable_assignment,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [77878] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1163), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [77940] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(486), 1,
      sym_adding_operator,
    STATE(487), 1,
      sym_multiplying_operator,
    STATE(635), 1,
      sym_logical_operator,
    STATE(636), 1,
      sym_relational_operator,
    STATE(637), 1,
      sym_shift_operator,
    STATE(1164), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [78012] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(522), 1,
      sym_adding_operator,
    STATE(523), 1,
      sym_multiplying_operator,
    STATE(648), 1,
      sym_logical_operator,
    STATE(649), 1,
      sym_relational_operator,
    STATE(650), 1,
      sym_shift_operator,
    STATE(1165), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [78084] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1166), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78146] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1167), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78208] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2912), 1,
      sym_exponentiate,
    STATE(522), 1,
      sym_adding_operator,
    STATE(523), 1,
      sym_multiplying_operator,
    STATE(648), 1,
      sym_logical_operator,
    STATE(649), 1,
      sym_relational_operator,
    STATE(650), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1168), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 29,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [78286] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1169), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78348] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1170), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78410] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1171), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78472] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1819), 1,
      anon_sym_LPAREN,
    ACTIONS(2936), 1,
      sym_PARAMETER,
    STATE(1364), 1,
      sym_parenthesis_group,
    STATE(1172), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 34,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78540] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1173), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78602] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1174), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78664] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1175), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78726] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1176), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78788] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1177), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78850] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1178), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78912] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1179), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [78974] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2912), 1,
      sym_exponentiate,
    STATE(522), 1,
      sym_adding_operator,
    STATE(523), 1,
      sym_multiplying_operator,
    STATE(648), 1,
      sym_logical_operator,
    STATE(649), 1,
      sym_relational_operator,
    STATE(650), 1,
      sym_shift_operator,
    STATE(1180), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 31,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [79048] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1181), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [79110] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1473), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6539), 1,
      sym_sequential_block,
    STATE(6597), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1182), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [79218] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1450), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6249), 1,
      sym_sequential_block,
    STATE(6250), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1183), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [79326] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1184), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [79388] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1185), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [79450] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1186), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [79512] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1187), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [79574] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1188), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [79636] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2968), 1,
      anon_sym_COLON,
    STATE(1189), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_PIPE,
  [79700] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1190), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [79762] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1191), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [79824] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1192), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [79886] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1193), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [79948] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1194), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [80010] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1863), 1,
      anon_sym_LPAREN,
    ACTIONS(2970), 1,
      sym_PARAMETER,
    STATE(1290), 1,
      sym_parenthesis_group,
    STATE(1195), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 34,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [80078] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1196), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [80140] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1197), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [80202] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1198), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2690), 6,
      sym_WHEN,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_PIPE,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 7,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [80288] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1199), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [80350] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1851), 1,
      anon_sym_LPAREN,
    ACTIONS(2972), 1,
      sym_PARAMETER,
    STATE(1250), 1,
      sym_parenthesis_group,
    STATE(1200), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 34,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [80418] = 21,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    STATE(5145), 1,
      aux_sym_conditional_expression_repeat1,
    STATE(6496), 1,
      sym_when_expression,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    ACTIONS(2904), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(1201), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [80510] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1202), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [80572] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1203), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [80634] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1204), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 37,
      sym_identifier,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [80696] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(522), 1,
      sym_adding_operator,
    STATE(523), 1,
      sym_multiplying_operator,
    STATE(648), 1,
      sym_logical_operator,
    STATE(649), 1,
      sym_relational_operator,
    STATE(650), 1,
      sym_shift_operator,
    STATE(1205), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [80768] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1206), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [80830] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1207), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [80892] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1443), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6201), 1,
      sym_sequential_block,
    STATE(6216), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1208), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [81000] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1209), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [81062] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1210), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [81124] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2912), 1,
      sym_exponentiate,
    STATE(522), 1,
      sym_adding_operator,
    STATE(523), 1,
      sym_multiplying_operator,
    STATE(648), 1,
      sym_logical_operator,
    STATE(649), 1,
      sym_relational_operator,
    STATE(650), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1211), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 12,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [81208] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1647), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6377), 1,
      sym_process_head,
    STATE(6544), 1,
      sym_sequential_block,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1212), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [81316] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1650), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6382), 1,
      sym_process_head,
    STATE(6559), 1,
      sym_sequential_block,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1213), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [81424] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1214), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 7,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 12,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_PIPE,
  [81508] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1215), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 19,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PIPE,
  [81590] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1216), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 25,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PIPE,
  [81670] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1654), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6389), 1,
      sym_process_head,
    STATE(6574), 1,
      sym_sequential_block,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1217), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [81778] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2912), 1,
      sym_exponentiate,
    STATE(522), 1,
      sym_adding_operator,
    STATE(523), 1,
      sym_multiplying_operator,
    STATE(648), 1,
      sym_logical_operator,
    STATE(649), 1,
      sym_relational_operator,
    STATE(650), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1218), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2690), 6,
      sym_BUS,
      sym_DOWNTO,
      sym_TO,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [81864] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1655), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6017), 1,
      sym_sequential_block,
    STATE(6408), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1219), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [81972] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1656), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6027), 1,
      sym_sequential_block,
    STATE(6414), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1220), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [82080] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2912), 1,
      sym_exponentiate,
    STATE(522), 1,
      sym_adding_operator,
    STATE(523), 1,
      sym_multiplying_operator,
    STATE(648), 1,
      sym_logical_operator,
    STATE(649), 1,
      sym_relational_operator,
    STATE(650), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1221), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 29,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [82158] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1222), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [82220] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2910), 1,
      sym_exponentiate,
    STATE(486), 1,
      sym_adding_operator,
    STATE(487), 1,
      sym_multiplying_operator,
    STATE(635), 1,
      sym_logical_operator,
    STATE(636), 1,
      sym_relational_operator,
    STATE(637), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1223), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 29,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [82298] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1659), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6046), 1,
      sym_sequential_block,
    STATE(6421), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1224), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [82406] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1660), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6088), 1,
      sym_sequential_block,
    STATE(6439), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1225), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [82514] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1662), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6094), 1,
      sym_sequential_block,
    STATE(6445), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1226), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [82622] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1227), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [82684] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2910), 1,
      sym_exponentiate,
    STATE(486), 1,
      sym_adding_operator,
    STATE(487), 1,
      sym_multiplying_operator,
    STATE(635), 1,
      sym_logical_operator,
    STATE(636), 1,
      sym_relational_operator,
    STATE(637), 1,
      sym_shift_operator,
    STATE(1228), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 31,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [82758] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1229), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 36,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [82820] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1664), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6109), 1,
      sym_sequential_block,
    STATE(6450), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1230), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [82928] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1666), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6132), 1,
      sym_sequential_block,
    STATE(6460), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1231), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [83036] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1668), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6136), 1,
      sym_sequential_block,
    STATE(6464), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1232), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [83144] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1233), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83206] = 29,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2938), 1,
      anon_sym_LPAREN,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(1670), 1,
      sym_sensitivity_specification,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6144), 1,
      sym_sequential_block,
    STATE(6465), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1234), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [83314] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1235), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83376] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1236), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 37,
      sym_identifier,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [83438] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1237), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 37,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_RANGE,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83500] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1238), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83561] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1239), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83622] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1240), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83683] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1241), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83744] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1242), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83805] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1243), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83866] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1244), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83927] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1245), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [83988] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1246), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84049] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1247), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84110] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1248), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84171] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1249), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84232] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1250), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84293] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1251), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84354] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1252), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84415] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1253), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84476] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1254), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84537] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1255), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84598] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1256), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84659] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1257), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 36,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [84720] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1258), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84781] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1259), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [84842] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1260), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 36,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [84903] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2974), 1,
      sym_exponentiate,
    STATE(476), 1,
      sym_adding_operator,
    STATE(477), 1,
      sym_multiplying_operator,
    STATE(626), 1,
      sym_logical_operator,
    STATE(627), 1,
      sym_relational_operator,
    STATE(628), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1261), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2690), 5,
      sym_DOWNTO,
      sym_IS,
      sym_OPEN,
      sym_TO,
      anon_sym_SEMI,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [84988] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1262), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [85049] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(502), 1,
      sym_adding_operator,
    STATE(503), 1,
      sym_multiplying_operator,
    STATE(639), 1,
      sym_logical_operator,
    STATE(640), 1,
      sym_relational_operator,
    STATE(641), 1,
      sym_shift_operator,
    STATE(1263), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [85120] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2976), 1,
      sym_exponentiate,
    STATE(502), 1,
      sym_adding_operator,
    STATE(503), 1,
      sym_multiplying_operator,
    STATE(639), 1,
      sym_logical_operator,
    STATE(640), 1,
      sym_relational_operator,
    STATE(641), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1264), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 28,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [85197] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1265), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [85258] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(502), 1,
      sym_adding_operator,
    STATE(503), 1,
      sym_multiplying_operator,
    STATE(639), 1,
      sym_logical_operator,
    STATE(640), 1,
      sym_relational_operator,
    STATE(641), 1,
      sym_shift_operator,
    STATE(1266), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [85329] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2974), 1,
      sym_exponentiate,
    STATE(476), 1,
      sym_adding_operator,
    STATE(477), 1,
      sym_multiplying_operator,
    STATE(626), 1,
      sym_logical_operator,
    STATE(627), 1,
      sym_relational_operator,
    STATE(628), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1267), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 11,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
  [85412] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2974), 1,
      sym_exponentiate,
    STATE(476), 1,
      sym_adding_operator,
    STATE(477), 1,
      sym_multiplying_operator,
    STATE(626), 1,
      sym_logical_operator,
    STATE(627), 1,
      sym_relational_operator,
    STATE(628), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1268), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 19,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [85493] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2974), 1,
      sym_exponentiate,
    STATE(476), 1,
      sym_adding_operator,
    STATE(477), 1,
      sym_multiplying_operator,
    STATE(626), 1,
      sym_logical_operator,
    STATE(627), 1,
      sym_relational_operator,
    STATE(628), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1269), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 25,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [85572] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2976), 1,
      sym_exponentiate,
    STATE(502), 1,
      sym_adding_operator,
    STATE(503), 1,
      sym_multiplying_operator,
    STATE(639), 1,
      sym_logical_operator,
    STATE(640), 1,
      sym_relational_operator,
    STATE(641), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1270), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 28,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [85649] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2976), 1,
      sym_exponentiate,
    STATE(502), 1,
      sym_adding_operator,
    STATE(503), 1,
      sym_multiplying_operator,
    STATE(639), 1,
      sym_logical_operator,
    STATE(640), 1,
      sym_relational_operator,
    STATE(641), 1,
      sym_shift_operator,
    STATE(1271), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 30,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [85722] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1272), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [85783] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1273), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [85844] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1274), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [85905] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1275), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [85966] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1276), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86027] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1277), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86088] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1278), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86149] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1279), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86210] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1280), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86271] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1281), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86332] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1282), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86393] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1283), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86454] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1284), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86515] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1285), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86576] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1286), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86637] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1287), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86698] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1288), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86759] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2980), 1,
      sym_library_constant_unit,
    STATE(2526), 1,
      sym__unit,
    STATE(1289), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2978), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [86826] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1290), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86887] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1291), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [86948] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1292), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87009] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1293), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87070] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1294), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87131] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1295), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87192] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1296), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87253] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1297), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87314] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1298), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87375] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1299), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87436] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1300), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87497] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1301), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87558] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1302), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87619] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1303), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87680] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1304), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87741] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1305), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87802] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1306), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87863] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1307), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87924] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1308), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [87985] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1309), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88046] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1310), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88107] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2976), 1,
      sym_exponentiate,
    STATE(502), 1,
      sym_adding_operator,
    STATE(503), 1,
      sym_multiplying_operator,
    STATE(639), 1,
      sym_logical_operator,
    STATE(640), 1,
      sym_relational_operator,
    STATE(641), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1311), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2690), 5,
      sym_DOWNTO,
      sym_GENERIC,
      sym_TO,
      anon_sym_SEMI,
      sym_variable_assignment,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [88192] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1312), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88253] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1313), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88314] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1314), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88375] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1315), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88436] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2976), 1,
      sym_exponentiate,
    STATE(502), 1,
      sym_adding_operator,
    STATE(503), 1,
      sym_multiplying_operator,
    STATE(639), 1,
      sym_logical_operator,
    STATE(640), 1,
      sym_relational_operator,
    STATE(641), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1316), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 11,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      sym_variable_assignment,
  [88519] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2976), 1,
      sym_exponentiate,
    STATE(502), 1,
      sym_adding_operator,
    STATE(503), 1,
      sym_multiplying_operator,
    STATE(639), 1,
      sym_logical_operator,
    STATE(640), 1,
      sym_relational_operator,
    STATE(641), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1317), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 19,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [88600] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2976), 1,
      sym_exponentiate,
    STATE(502), 1,
      sym_adding_operator,
    STATE(503), 1,
      sym_multiplying_operator,
    STATE(639), 1,
      sym_logical_operator,
    STATE(640), 1,
      sym_relational_operator,
    STATE(641), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1318), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 25,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [88679] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1319), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88740] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1320), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88801] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1321), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88862] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1322), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88923] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1323), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [88984] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1324), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [89045] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1325), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 36,
      sym_AND,
      sym_BUS,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [89106] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1326), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 36,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [89167] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    STATE(7068), 1,
      sym_end_for,
    STATE(7081), 1,
      sym_verification_unit_binding_indication,
    STATE(1327), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2982), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [89232] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1328), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [89293] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1329), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [89354] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1330), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 36,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [89415] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2988), 1,
      sym_library_constant_unit,
    STATE(2474), 1,
      sym__unit,
    STATE(1331), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2986), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [89482] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(476), 1,
      sym_adding_operator,
    STATE(477), 1,
      sym_multiplying_operator,
    STATE(626), 1,
      sym_logical_operator,
    STATE(627), 1,
      sym_relational_operator,
    STATE(628), 1,
      sym_shift_operator,
    STATE(1332), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [89553] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2974), 1,
      sym_exponentiate,
    STATE(476), 1,
      sym_adding_operator,
    STATE(477), 1,
      sym_multiplying_operator,
    STATE(626), 1,
      sym_logical_operator,
    STATE(627), 1,
      sym_relational_operator,
    STATE(628), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1333), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 28,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [89630] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1334), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [89691] = 28,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(227), 1,
      sym_POSTPONED,
    ACTIONS(229), 1,
      sym_PROCESS,
    ACTIONS(2990), 1,
      sym_END,
    STATE(1727), 1,
      aux_sym_design_unit_repeat3,
    STATE(2626), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4386), 1,
      sym__concurrent_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6203), 1,
      sym_generate_block,
    STATE(6317), 1,
      sym_name,
    STATE(6373), 1,
      sym__target,
    STATE(6470), 1,
      sym_with_expression,
    STATE(7169), 1,
      sym__label,
    STATE(7183), 1,
      sym_generate_block_end,
    STATE(7448), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1335), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4354), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [89796] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1336), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [89857] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1337), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [89918] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(476), 1,
      sym_adding_operator,
    STATE(477), 1,
      sym_multiplying_operator,
    STATE(626), 1,
      sym_logical_operator,
    STATE(627), 1,
      sym_relational_operator,
    STATE(628), 1,
      sym_shift_operator,
    STATE(1338), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [89989] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1339), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90050] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1340), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90111] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1341), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90172] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1342), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90233] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1343), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90294] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1344), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90355] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1345), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90416] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1346), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90477] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1347), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90538] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1348), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90599] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1349), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90660] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1350), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90721] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1351), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90782] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1352), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90843] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1353), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90904] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1354), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [90965] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1355), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91026] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1356), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91087] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1357), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91148] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1358), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91209] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1359), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91270] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1360), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91331] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1361), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91392] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1362), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91453] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1363), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91514] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1364), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91575] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2974), 1,
      sym_exponentiate,
    STATE(476), 1,
      sym_adding_operator,
    STATE(477), 1,
      sym_multiplying_operator,
    STATE(626), 1,
      sym_logical_operator,
    STATE(627), 1,
      sym_relational_operator,
    STATE(628), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1365), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 28,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [91652] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2974), 1,
      sym_exponentiate,
    STATE(476), 1,
      sym_adding_operator,
    STATE(477), 1,
      sym_multiplying_operator,
    STATE(626), 1,
      sym_logical_operator,
    STATE(627), 1,
      sym_relational_operator,
    STATE(628), 1,
      sym_shift_operator,
    STATE(1366), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 30,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [91725] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1367), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91786] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1368), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91847] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1369), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91908] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1370), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [91969] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1371), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92030] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1372), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92091] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1373), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92152] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2033), 1,
      anon_sym_LPAREN,
    ACTIONS(2993), 1,
      sym_PARAMETER,
    STATE(1487), 1,
      sym_parenthesis_group,
    STATE(1374), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 33,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92219] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1375), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92280] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1376), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92341] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1377), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92402] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1378), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92463] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1379), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92524] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1380), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92585] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1381), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92646] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1382), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92707] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2093), 1,
      anon_sym_LPAREN,
    ACTIONS(2995), 1,
      sym_PARAMETER,
    STATE(1513), 1,
      sym_parenthesis_group,
    STATE(1383), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 33,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92774] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1384), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92835] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1385), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [92896] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(253), 1,
      sym_REPORT,
    ACTIONS(2997), 1,
      anon_sym_SEMI,
    ACTIONS(2999), 1,
      sym_exponentiate,
    ACTIONS(3001), 1,
      sym_SEVERITY,
    STATE(669), 1,
      sym_logical_operator,
    STATE(670), 1,
      sym_relational_operator,
    STATE(681), 1,
      sym_shift_operator,
    STATE(682), 1,
      sym_adding_operator,
    STATE(684), 1,
      sym_multiplying_operator,
    STATE(5726), 1,
      sym_report_expression,
    STATE(6867), 1,
      sym_severity_expression,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1386), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [92989] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1387), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93050] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1388), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93111] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1389), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93172] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1390), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93233] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1391), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93294] = 28,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(227), 1,
      sym_POSTPONED,
    ACTIONS(229), 1,
      sym_PROCESS,
    ACTIONS(3003), 1,
      sym_END,
    STATE(1727), 1,
      aux_sym_design_unit_repeat3,
    STATE(2626), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4386), 1,
      sym__concurrent_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6317), 1,
      sym_name,
    STATE(6373), 1,
      sym__target,
    STATE(6470), 1,
      sym_with_expression,
    STATE(6510), 1,
      sym_generate_block,
    STATE(6615), 1,
      sym_generate_block_end,
    STATE(7169), 1,
      sym__label,
    STATE(7448), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1392), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4354), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [93399] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1393), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93460] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1394), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93521] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1395), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93582] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1396), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93643] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1397), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93704] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1398), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93765] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1919), 1,
      anon_sym_LPAREN,
    ACTIONS(3006), 1,
      sym_PARAMETER,
    STATE(1571), 1,
      sym_parenthesis_group,
    STATE(1399), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93832] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2063), 1,
      anon_sym_LPAREN,
    ACTIONS(3008), 1,
      sym_PARAMETER,
    STATE(1579), 1,
      sym_parenthesis_group,
    STATE(1400), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 33,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [93899] = 26,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(199), 1,
      sym_POSTPONED,
    ACTIONS(201), 1,
      sym_PROCESS,
    STATE(1402), 1,
      aux_sym_design_unit_repeat3,
    STATE(2658), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(3725), 1,
      sym__concurrent_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6403), 1,
      sym__target,
    STATE(6480), 1,
      sym_with_expression,
    STATE(6481), 1,
      sym_name,
    STATE(7169), 1,
      sym__label,
    STATE(8292), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    ACTIONS(3010), 3,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
    STATE(1401), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(3814), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [94000] = 25,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(702), 1,
      anon_sym_LPAREN,
    ACTIONS(705), 1,
      anon_sym_LT_LT,
    ACTIONS(713), 1,
      sym_ASSERT,
    ACTIONS(722), 1,
      sym_WITH,
    ACTIONS(3012), 1,
      sym_POSTPONED,
    ACTIONS(3015), 1,
      sym_PROCESS,
    STATE(2658), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(3725), 1,
      sym__concurrent_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6403), 1,
      sym__target,
    STATE(6480), 1,
      sym_with_expression,
    STATE(6481), 1,
      sym_name,
    STATE(7169), 1,
      sym__label,
    STATE(8292), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(711), 3,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
    ACTIONS(725), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(708), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1402), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_design_unit_repeat3,
    STATE(3814), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [94099] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1403), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94160] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1404), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94221] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1405), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94282] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1406), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94343] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3020), 1,
      sym_library_constant_unit,
    STATE(2483), 1,
      sym__unit,
    STATE(1407), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3018), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [94410] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3024), 1,
      sym_library_constant_unit,
    STATE(2519), 1,
      sym__unit,
    STATE(1408), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3022), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [94477] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1409), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94538] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1410), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94599] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1411), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94660] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1412), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94721] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3028), 1,
      sym_library_constant_unit,
    STATE(2529), 1,
      sym__unit,
    STATE(1413), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3026), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [94788] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1414), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94849] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1415), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94910] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1416), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [94971] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1417), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 36,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95032] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1931), 1,
      anon_sym_LPAREN,
    ACTIONS(3030), 1,
      sym_PARAMETER,
    STATE(1465), 1,
      sym_parenthesis_group,
    STATE(1418), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95099] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1419), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95160] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1420), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 36,
      sym_AFTER,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95221] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1421), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95281] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1422), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95341] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1423), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [95401] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1424), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95461] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1425), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95521] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1426), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95581] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1427), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95641] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1428), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95701] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3032), 1,
      sym_exponentiate,
    STATE(512), 1,
      sym_adding_operator,
    STATE(513), 1,
      sym_multiplying_operator,
    STATE(601), 1,
      sym_logical_operator,
    STATE(602), 1,
      sym_relational_operator,
    STATE(603), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1429), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 10,
      sym_AND,
      sym_BUS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      sym_variable_assignment,
  [95783] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1430), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95843] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1431), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [95903] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3032), 1,
      sym_exponentiate,
    STATE(512), 1,
      sym_adding_operator,
    STATE(513), 1,
      sym_multiplying_operator,
    STATE(601), 1,
      sym_logical_operator,
    STATE(602), 1,
      sym_relational_operator,
    STATE(603), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1432), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 18,
      sym_AND,
      sym_BUS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [95983] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1433), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [96043] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2211), 1,
      anon_sym_LPAREN,
    ACTIONS(3034), 1,
      sym_PARAMETER,
    STATE(1734), 1,
      sym_parenthesis_group,
    STATE(1434), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 32,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [96109] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3032), 1,
      sym_exponentiate,
    STATE(512), 1,
      sym_adding_operator,
    STATE(513), 1,
      sym_multiplying_operator,
    STATE(601), 1,
      sym_logical_operator,
    STATE(602), 1,
      sym_relational_operator,
    STATE(603), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1435), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 24,
      sym_AND,
      sym_BUS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [96187] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3036), 1,
      sym_exponentiate,
    STATE(527), 1,
      sym_adding_operator,
    STATE(528), 1,
      sym_multiplying_operator,
    STATE(643), 1,
      sym_logical_operator,
    STATE(644), 1,
      sym_relational_operator,
    STATE(645), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1436), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2690), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [96271] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1437), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [96331] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1438), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [96391] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1439), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [96451] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3036), 1,
      sym_exponentiate,
    STATE(527), 1,
      sym_adding_operator,
    STATE(528), 1,
      sym_multiplying_operator,
    STATE(643), 1,
      sym_logical_operator,
    STATE(644), 1,
      sym_relational_operator,
    STATE(645), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1440), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 10,
      sym_AND,
      sym_BUS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [96533] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3036), 1,
      sym_exponentiate,
    STATE(527), 1,
      sym_adding_operator,
    STATE(528), 1,
      sym_multiplying_operator,
    STATE(643), 1,
      sym_logical_operator,
    STATE(644), 1,
      sym_relational_operator,
    STATE(645), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1441), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 18,
      sym_AND,
      sym_BUS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [96613] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3036), 1,
      sym_exponentiate,
    STATE(527), 1,
      sym_adding_operator,
    STATE(528), 1,
      sym_multiplying_operator,
    STATE(643), 1,
      sym_logical_operator,
    STATE(644), 1,
      sym_relational_operator,
    STATE(645), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1442), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 24,
      sym_AND,
      sym_BUS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [96691] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6249), 1,
      sym_sequential_block,
    STATE(6250), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1443), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [96793] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1444), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [96853] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1445), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 35,
      sym_identifier,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [96913] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1446), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [96973] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1447), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 35,
      sym_identifier,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [97033] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3046), 1,
      sym_exponentiate,
    STATE(456), 1,
      sym_adding_operator,
    STATE(457), 1,
      sym_multiplying_operator,
    STATE(605), 1,
      sym_logical_operator,
    STATE(606), 1,
      sym_relational_operator,
    STATE(607), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1448), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2690), 4,
      sym_DOWNTO,
      sym_OF,
      sym_TO,
      anon_sym_COMMA,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [97117] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2269), 1,
      anon_sym_LPAREN,
    ACTIONS(3048), 1,
      sym_PARAMETER,
    STATE(1894), 1,
      sym_parenthesis_group,
    STATE(1449), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 32,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [97183] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6056), 1,
      sym_sequential_block,
    STATE(6199), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1450), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [97285] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(527), 1,
      sym_adding_operator,
    STATE(528), 1,
      sym_multiplying_operator,
    STATE(643), 1,
      sym_logical_operator,
    STATE(644), 1,
      sym_relational_operator,
    STATE(645), 1,
      sym_shift_operator,
    STATE(1451), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [97355] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3036), 1,
      sym_exponentiate,
    STATE(527), 1,
      sym_adding_operator,
    STATE(528), 1,
      sym_multiplying_operator,
    STATE(643), 1,
      sym_logical_operator,
    STATE(644), 1,
      sym_relational_operator,
    STATE(645), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1452), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 27,
      sym_AND,
      sym_BUS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [97431] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(527), 1,
      sym_adding_operator,
    STATE(528), 1,
      sym_multiplying_operator,
    STATE(643), 1,
      sym_logical_operator,
    STATE(644), 1,
      sym_relational_operator,
    STATE(645), 1,
      sym_shift_operator,
    STATE(1453), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [97501] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3050), 1,
      sym_IS,
    STATE(2411), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6266), 1,
      sym_sequential_block,
    STATE(6275), 1,
      sym_sequential_block_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1454), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [97603] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3054), 1,
      sym_library_constant_unit,
    STATE(2582), 1,
      sym__unit,
    STATE(1455), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [97669] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3036), 1,
      sym_exponentiate,
    STATE(527), 1,
      sym_adding_operator,
    STATE(528), 1,
      sym_multiplying_operator,
    STATE(643), 1,
      sym_logical_operator,
    STATE(644), 1,
      sym_relational_operator,
    STATE(645), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1456), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 27,
      sym_AND,
      sym_BUS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [97745] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3036), 1,
      sym_exponentiate,
    STATE(527), 1,
      sym_adding_operator,
    STATE(528), 1,
      sym_multiplying_operator,
    STATE(643), 1,
      sym_logical_operator,
    STATE(644), 1,
      sym_relational_operator,
    STATE(645), 1,
      sym_shift_operator,
    STATE(1457), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [97817] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1458), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [97877] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3046), 1,
      sym_exponentiate,
    STATE(456), 1,
      sym_adding_operator,
    STATE(457), 1,
      sym_multiplying_operator,
    STATE(605), 1,
      sym_logical_operator,
    STATE(606), 1,
      sym_relational_operator,
    STATE(607), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1459), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 10,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
  [97959] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1460), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [98019] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3046), 1,
      sym_exponentiate,
    STATE(456), 1,
      sym_adding_operator,
    STATE(457), 1,
      sym_multiplying_operator,
    STATE(605), 1,
      sym_logical_operator,
    STATE(606), 1,
      sym_relational_operator,
    STATE(607), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1461), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 18,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [98099] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1462), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [98159] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1463), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [98219] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(461), 1,
      sym_adding_operator,
    STATE(462), 1,
      sym_multiplying_operator,
    STATE(609), 1,
      sym_logical_operator,
    STATE(610), 1,
      sym_relational_operator,
    STATE(611), 1,
      sym_shift_operator,
    STATE(1464), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [98289] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1465), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [98349] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3046), 1,
      sym_exponentiate,
    STATE(456), 1,
      sym_adding_operator,
    STATE(457), 1,
      sym_multiplying_operator,
    STATE(605), 1,
      sym_logical_operator,
    STATE(606), 1,
      sym_relational_operator,
    STATE(607), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1466), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 24,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [98427] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1467), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 35,
      sym_identifier,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [98487] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1468), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [98547] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(461), 1,
      sym_adding_operator,
    STATE(462), 1,
      sym_multiplying_operator,
    STATE(609), 1,
      sym_logical_operator,
    STATE(610), 1,
      sym_relational_operator,
    STATE(611), 1,
      sym_shift_operator,
    STATE(1469), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [98617] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1470), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [98677] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1471), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [98737] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1472), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [98797] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6201), 1,
      sym_sequential_block,
    STATE(6216), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1473), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [98899] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1474), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [98959] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1475), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99019] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1476), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99079] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1477), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 35,
      sym_identifier,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [99139] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1478), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99199] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1479), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99259] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1480), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99319] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1481), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99379] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1482), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99439] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1483), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [99499] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1484), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99559] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1485), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99619] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1486), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99679] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1487), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99739] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1488), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99799] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1489), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99859] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1490), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99919] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1491), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [99979] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3072), 1,
      sym_exponentiate,
    STATE(461), 1,
      sym_adding_operator,
    STATE(462), 1,
      sym_multiplying_operator,
    STATE(609), 1,
      sym_logical_operator,
    STATE(610), 1,
      sym_relational_operator,
    STATE(611), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1492), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 27,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [100055] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1493), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [100115] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1494), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [100175] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1495), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [100235] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1496), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [100295] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3076), 1,
      sym_library_constant_unit,
    STATE(2597), 1,
      sym__unit,
    STATE(1497), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3074), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [100361] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(512), 1,
      sym_adding_operator,
    STATE(513), 1,
      sym_multiplying_operator,
    STATE(601), 1,
      sym_logical_operator,
    STATE(602), 1,
      sym_relational_operator,
    STATE(603), 1,
      sym_shift_operator,
    STATE(1498), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [100431] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1499), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [100491] = 36,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(233), 1,
      sym_CASE,
    ACTIONS(243), 1,
      sym_FOR,
    ACTIONS(245), 1,
      sym_IF,
    ACTIONS(247), 1,
      sym_LOOP,
    ACTIONS(253), 1,
      sym_REPORT,
    ACTIONS(259), 1,
      sym_WHILE,
    ACTIONS(3080), 1,
      sym_BLOCK,
    ACTIONS(3082), 1,
      sym_EXIT,
    ACTIONS(3084), 1,
      sym_NEXT,
    ACTIONS(3086), 1,
      sym_NULL,
    ACTIONS(3088), 1,
      sym_RETURN,
    ACTIONS(3090), 1,
      sym_WAIT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5855), 1,
      sym_report_expression,
    STATE(5870), 1,
      sym_name,
    STATE(5874), 1,
      sym__target,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6315), 1,
      sym_loop_body,
    STATE(6342), 1,
      sym_case_expression,
    STATE(6350), 1,
      sym_if_statement,
    STATE(6351), 1,
      sym__iteration_scheme,
    STATE(6355), 1,
      sym_with_expression,
    STATE(7505), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    STATE(7407), 2,
      sym_while_loop,
      sym_for_loop,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1500), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [100611] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1501), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [100671] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1502), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [100731] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1503), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [100791] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1504), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [100851] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3032), 1,
      sym_exponentiate,
    STATE(512), 1,
      sym_adding_operator,
    STATE(513), 1,
      sym_multiplying_operator,
    STATE(601), 1,
      sym_logical_operator,
    STATE(602), 1,
      sym_relational_operator,
    STATE(603), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1505), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 27,
      sym_AND,
      sym_BUS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [100927] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1506), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [100987] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1507), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101047] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1508), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101107] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1509), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101167] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1510), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101227] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1511), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101287] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(512), 1,
      sym_adding_operator,
    STATE(513), 1,
      sym_multiplying_operator,
    STATE(601), 1,
      sym_logical_operator,
    STATE(602), 1,
      sym_relational_operator,
    STATE(603), 1,
      sym_shift_operator,
    STATE(1512), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [101357] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1513), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101417] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1514), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101477] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1515), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101537] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1516), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101597] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1517), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(3092), 4,
      sym_WHEN,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [101681] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3072), 1,
      sym_exponentiate,
    STATE(461), 1,
      sym_adding_operator,
    STATE(462), 1,
      sym_multiplying_operator,
    STATE(609), 1,
      sym_logical_operator,
    STATE(610), 1,
      sym_relational_operator,
    STATE(611), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1518), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 27,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [101757] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3096), 1,
      sym_exponentiate,
    ACTIONS(3098), 1,
      sym_AFTER,
    STATE(535), 1,
      sym_logical_operator,
    STATE(536), 1,
      sym_relational_operator,
    STATE(537), 1,
      sym_shift_operator,
    STATE(707), 1,
      sym_adding_operator,
    STATE(708), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    ACTIONS(3094), 3,
      sym_WHEN,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(1519), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [101843] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1520), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101903] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1521), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [101963] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1522), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [102023] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1523), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [102083] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3032), 1,
      sym_exponentiate,
    STATE(512), 1,
      sym_adding_operator,
    STATE(513), 1,
      sym_multiplying_operator,
    STATE(601), 1,
      sym_logical_operator,
    STATE(602), 1,
      sym_relational_operator,
    STATE(603), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1524), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2690), 4,
      sym_BUS,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [102167] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3032), 1,
      sym_exponentiate,
    STATE(512), 1,
      sym_adding_operator,
    STATE(513), 1,
      sym_multiplying_operator,
    STATE(601), 1,
      sym_logical_operator,
    STATE(602), 1,
      sym_relational_operator,
    STATE(603), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1525), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 27,
      sym_AND,
      sym_BUS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [102243] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3102), 1,
      sym_library_constant_unit,
    STATE(2553), 1,
      sym__unit,
    STATE(1526), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3100), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [102309] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1527), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [102369] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3072), 1,
      sym_exponentiate,
    STATE(461), 1,
      sym_adding_operator,
    STATE(462), 1,
      sym_multiplying_operator,
    STATE(609), 1,
      sym_logical_operator,
    STATE(610), 1,
      sym_relational_operator,
    STATE(611), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1528), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2690), 4,
      sym_DOWNTO,
      sym_TO,
      sym_UNITS,
      anon_sym_SEMI,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [102453] = 36,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(233), 1,
      sym_CASE,
    ACTIONS(243), 1,
      sym_FOR,
    ACTIONS(245), 1,
      sym_IF,
    ACTIONS(247), 1,
      sym_LOOP,
    ACTIONS(253), 1,
      sym_REPORT,
    ACTIONS(259), 1,
      sym_WHILE,
    ACTIONS(3104), 1,
      sym_BLOCK,
    ACTIONS(3106), 1,
      sym_EXIT,
    ACTIONS(3108), 1,
      sym_NEXT,
    ACTIONS(3110), 1,
      sym_NULL,
    ACTIONS(3112), 1,
      sym_RETURN,
    ACTIONS(3114), 1,
      sym_WAIT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5866), 1,
      sym_report_expression,
    STATE(5867), 1,
      sym_name,
    STATE(5887), 1,
      sym__target,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6392), 1,
      sym_case_expression,
    STATE(6393), 1,
      sym__iteration_scheme,
    STATE(6476), 1,
      sym_with_expression,
    STATE(6578), 1,
      sym_loop_body,
    STATE(6580), 1,
      sym_if_statement,
    STATE(7580), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    STATE(7407), 2,
      sym_while_loop,
      sym_for_loop,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1529), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [102573] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1530), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [102633] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3118), 1,
      sym_library_constant_unit,
    STATE(2556), 1,
      sym__unit,
    STATE(1531), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3116), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [102699] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1532), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [102759] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3072), 1,
      sym_exponentiate,
    STATE(461), 1,
      sym_adding_operator,
    STATE(462), 1,
      sym_multiplying_operator,
    STATE(609), 1,
      sym_logical_operator,
    STATE(610), 1,
      sym_relational_operator,
    STATE(611), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1533), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 10,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
  [102841] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3072), 1,
      sym_exponentiate,
    STATE(461), 1,
      sym_adding_operator,
    STATE(462), 1,
      sym_multiplying_operator,
    STATE(609), 1,
      sym_logical_operator,
    STATE(610), 1,
      sym_relational_operator,
    STATE(611), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1534), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 18,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [102921] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3032), 1,
      sym_exponentiate,
    STATE(512), 1,
      sym_adding_operator,
    STATE(513), 1,
      sym_multiplying_operator,
    STATE(601), 1,
      sym_logical_operator,
    STATE(602), 1,
      sym_relational_operator,
    STATE(603), 1,
      sym_shift_operator,
    STATE(1535), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [102993] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3072), 1,
      sym_exponentiate,
    STATE(461), 1,
      sym_adding_operator,
    STATE(462), 1,
      sym_multiplying_operator,
    STATE(609), 1,
      sym_logical_operator,
    STATE(610), 1,
      sym_relational_operator,
    STATE(611), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1536), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 24,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [103071] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1537), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [103131] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1538), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [103191] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1539), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [103251] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1540), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [103311] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1541), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [103371] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1542), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [103431] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1543), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [103491] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1544), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [103551] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1545), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [103611] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3072), 1,
      sym_exponentiate,
    STATE(461), 1,
      sym_adding_operator,
    STATE(462), 1,
      sym_multiplying_operator,
    STATE(609), 1,
      sym_logical_operator,
    STATE(610), 1,
      sym_relational_operator,
    STATE(611), 1,
      sym_shift_operator,
    STATE(1546), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [103683] = 36,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(233), 1,
      sym_CASE,
    ACTIONS(243), 1,
      sym_FOR,
    ACTIONS(245), 1,
      sym_IF,
    ACTIONS(247), 1,
      sym_LOOP,
    ACTIONS(253), 1,
      sym_REPORT,
    ACTIONS(259), 1,
      sym_WHILE,
    ACTIONS(3120), 1,
      sym_BLOCK,
    ACTIONS(3122), 1,
      sym_EXIT,
    ACTIONS(3124), 1,
      sym_NEXT,
    ACTIONS(3126), 1,
      sym_NULL,
    ACTIONS(3128), 1,
      sym_RETURN,
    ACTIONS(3130), 1,
      sym_WAIT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5895), 1,
      sym__target,
    STATE(5928), 1,
      sym_report_expression,
    STATE(5929), 1,
      sym_name,
    STATE(6006), 1,
      sym_loop_body,
    STATE(6053), 1,
      sym_if_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6424), 1,
      sym_case_expression,
    STATE(6425), 1,
      sym__iteration_scheme,
    STATE(6489), 1,
      sym_with_expression,
    STATE(8148), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    STATE(7407), 2,
      sym_while_loop,
      sym_for_loop,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1547), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [103803] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1548), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [103863] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3050), 1,
      sym_IS,
    STATE(2411), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6066), 1,
      sym_sequential_block,
    STATE(6108), 1,
      sym_sequential_block_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1549), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [103965] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1550), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104025] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1551), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104085] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1552), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104145] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1553), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104205] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1554), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104265] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1555), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104325] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2912), 1,
      sym_exponentiate,
    STATE(522), 1,
      sym_adding_operator,
    STATE(523), 1,
      sym_multiplying_operator,
    STATE(643), 1,
      sym_logical_operator,
    STATE(644), 1,
      sym_relational_operator,
    STATE(645), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1556), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2830), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [104409] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1557), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104469] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1558), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104529] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1559), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104589] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1560), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104649] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1561), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104709] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1562), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104769] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1563), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104829] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1564), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104889] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1565), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [104949] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1566), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105009] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1567), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105069] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1568), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105129] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1569), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105189] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1570), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105249] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1571), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105309] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1572), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105369] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1573), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105429] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2356), 1,
      anon_sym_LPAREN,
    ACTIONS(3132), 1,
      sym_PARAMETER,
    STATE(1746), 1,
      sym_parenthesis_group,
    STATE(1574), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 32,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105495] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1575), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105555] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1576), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105615] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1577), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105675] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1578), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105735] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1579), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105795] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1580), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105855] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1581), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105915] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1582), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [105975] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1583), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106035] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1584), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106095] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1585), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106155] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1586), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106215] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1587), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106275] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1588), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106335] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1589), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106395] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1590), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106455] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1591), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106515] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1592), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106575] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1593), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106635] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1594), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106695] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1595), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106755] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1596), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106815] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1597), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106875] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1598), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106935] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1599), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [106995] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1600), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [107055] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1601), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [107115] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1602), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [107175] = 26,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(221), 1,
      sym_POSTPONED,
    ACTIONS(223), 1,
      sym_PROCESS,
    STATE(1608), 1,
      aux_sym_design_unit_repeat3,
    STATE(2636), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4169), 1,
      sym__concurrent_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6435), 1,
      sym__target,
    STATE(6493), 1,
      sym_with_expression,
    STATE(6555), 1,
      sym_name,
    STATE(6638), 1,
      sym_assertion,
    STATE(7169), 1,
      sym__label,
    ACTIONS(3010), 2,
      sym_END,
      sym_WHEN,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1603), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4170), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [107275] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1604), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [107335] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1605), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 35,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [107395] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1606), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [107455] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1607), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [107515] = 25,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(702), 1,
      anon_sym_LPAREN,
    ACTIONS(705), 1,
      anon_sym_LT_LT,
    ACTIONS(713), 1,
      sym_ASSERT,
    ACTIONS(722), 1,
      sym_WITH,
    ACTIONS(3134), 1,
      sym_POSTPONED,
    ACTIONS(3137), 1,
      sym_PROCESS,
    STATE(2636), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4169), 1,
      sym__concurrent_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6435), 1,
      sym__target,
    STATE(6493), 1,
      sym_with_expression,
    STATE(6555), 1,
      sym_name,
    STATE(6638), 1,
      sym_assertion,
    STATE(7169), 1,
      sym__label,
    ACTIONS(711), 2,
      sym_END,
      sym_WHEN,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(725), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(708), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1608), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_design_unit_repeat3,
    STATE(4170), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [107613] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3142), 1,
      sym_library_constant_unit,
    STATE(2591), 1,
      sym__unit,
    STATE(1609), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3140), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [107679] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3146), 1,
      sym_library_constant_unit,
    STATE(2606), 1,
      sym__unit,
    STATE(1610), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3144), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [107745] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1611), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [107805] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1612), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [107865] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2910), 1,
      sym_exponentiate,
    STATE(486), 1,
      sym_adding_operator,
    STATE(487), 1,
      sym_multiplying_operator,
    STATE(601), 1,
      sym_logical_operator,
    STATE(602), 1,
      sym_relational_operator,
    STATE(603), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1613), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2830), 4,
      sym_BUS,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [107949] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1614), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [108009] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1615), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 35,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [108069] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1616), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 35,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [108129] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1617), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [108189] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1618), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [108249] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1619), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [108309] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1620), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [108369] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1621), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [108429] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1622), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [108489] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1623), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [108549] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(456), 1,
      sym_adding_operator,
    STATE(457), 1,
      sym_multiplying_operator,
    STATE(605), 1,
      sym_logical_operator,
    STATE(606), 1,
      sym_relational_operator,
    STATE(607), 1,
      sym_shift_operator,
    STATE(1624), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [108619] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1625), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [108679] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3046), 1,
      sym_exponentiate,
    STATE(456), 1,
      sym_adding_operator,
    STATE(457), 1,
      sym_multiplying_operator,
    STATE(605), 1,
      sym_logical_operator,
    STATE(606), 1,
      sym_relational_operator,
    STATE(607), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1626), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 27,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [108755] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1627), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 35,
      sym_identifier,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [108815] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2179), 1,
      anon_sym_LPAREN,
    ACTIONS(3160), 1,
      sym_PARAMETER,
    STATE(1824), 1,
      sym_parenthesis_group,
    STATE(1628), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 32,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [108881] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1629), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 35,
      sym_identifier,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [108941] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3096), 1,
      sym_exponentiate,
    STATE(535), 1,
      sym_logical_operator,
    STATE(536), 1,
      sym_relational_operator,
    STATE(537), 1,
      sym_shift_operator,
    STATE(707), 1,
      sym_adding_operator,
    STATE(708), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1630), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2690), 4,
      sym_AFTER,
      sym_WHEN,
      anon_sym_SEMI,
      anon_sym_COMMA,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [109025] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(535), 1,
      sym_logical_operator,
    STATE(536), 1,
      sym_relational_operator,
    STATE(537), 1,
      sym_shift_operator,
    STATE(707), 1,
      sym_adding_operator,
    STATE(708), 1,
      sym_multiplying_operator,
    STATE(1631), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [109095] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3096), 1,
      sym_exponentiate,
    STATE(535), 1,
      sym_logical_operator,
    STATE(536), 1,
      sym_relational_operator,
    STATE(537), 1,
      sym_shift_operator,
    STATE(707), 1,
      sym_adding_operator,
    STATE(708), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1632), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 27,
      sym_AFTER,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [109171] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(535), 1,
      sym_logical_operator,
    STATE(536), 1,
      sym_relational_operator,
    STATE(537), 1,
      sym_shift_operator,
    STATE(707), 1,
      sym_adding_operator,
    STATE(708), 1,
      sym_multiplying_operator,
    STATE(1633), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [109241] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3096), 1,
      sym_exponentiate,
    STATE(535), 1,
      sym_logical_operator,
    STATE(536), 1,
      sym_relational_operator,
    STATE(537), 1,
      sym_shift_operator,
    STATE(707), 1,
      sym_adding_operator,
    STATE(708), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1634), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 10,
      sym_AFTER,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [109323] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3096), 1,
      sym_exponentiate,
    STATE(535), 1,
      sym_logical_operator,
    STATE(536), 1,
      sym_relational_operator,
    STATE(537), 1,
      sym_shift_operator,
    STATE(707), 1,
      sym_adding_operator,
    STATE(708), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1635), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 18,
      sym_AFTER,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [109403] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3096), 1,
      sym_exponentiate,
    STATE(535), 1,
      sym_logical_operator,
    STATE(536), 1,
      sym_relational_operator,
    STATE(537), 1,
      sym_shift_operator,
    STATE(707), 1,
      sym_adding_operator,
    STATE(708), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1636), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 24,
      sym_AFTER,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [109481] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3096), 1,
      sym_exponentiate,
    STATE(535), 1,
      sym_logical_operator,
    STATE(536), 1,
      sym_relational_operator,
    STATE(537), 1,
      sym_shift_operator,
    STATE(707), 1,
      sym_adding_operator,
    STATE(708), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1637), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 27,
      sym_AFTER,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [109557] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3096), 1,
      sym_exponentiate,
    STATE(535), 1,
      sym_logical_operator,
    STATE(536), 1,
      sym_relational_operator,
    STATE(537), 1,
      sym_shift_operator,
    STATE(707), 1,
      sym_adding_operator,
    STATE(708), 1,
      sym_multiplying_operator,
    STATE(1638), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [109629] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(456), 1,
      sym_adding_operator,
    STATE(457), 1,
      sym_multiplying_operator,
    STATE(605), 1,
      sym_logical_operator,
    STATE(606), 1,
      sym_relational_operator,
    STATE(607), 1,
      sym_shift_operator,
    STATE(1639), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [109699] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3046), 1,
      sym_exponentiate,
    STATE(456), 1,
      sym_adding_operator,
    STATE(457), 1,
      sym_multiplying_operator,
    STATE(605), 1,
      sym_logical_operator,
    STATE(606), 1,
      sym_relational_operator,
    STATE(607), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1640), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 27,
      sym_AND,
      sym_DOWNTO,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [109775] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3046), 1,
      sym_exponentiate,
    STATE(456), 1,
      sym_adding_operator,
    STATE(457), 1,
      sym_multiplying_operator,
    STATE(605), 1,
      sym_logical_operator,
    STATE(606), 1,
      sym_relational_operator,
    STATE(607), 1,
      sym_shift_operator,
    STATE(1641), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [109847] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1642), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 35,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [109907] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2199), 1,
      anon_sym_LPAREN,
    ACTIONS(3162), 1,
      sym_PARAMETER,
    STATE(1854), 1,
      sym_parenthesis_group,
    STATE(1643), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 32,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [109973] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1644), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 34,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [110033] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1645), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [110093] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1646), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [110153] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6382), 1,
      sym_process_head,
    STATE(6559), 1,
      sym_sequential_block,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1647), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [110255] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1648), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [110315] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1649), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 35,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [110375] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6389), 1,
      sym_process_head,
    STATE(6574), 1,
      sym_sequential_block,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1650), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [110477] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3050), 1,
      sym_IS,
    STATE(2411), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6390), 1,
      sym_sequential_block_head,
    STATE(6576), 1,
      sym_sequential_block,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1651), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [110579] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1652), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [110639] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3050), 1,
      sym_IS,
    STATE(2411), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6400), 1,
      sym_sequential_block_head,
    STATE(6587), 1,
      sym_sequential_block,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1653), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [110741] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6402), 1,
      sym_process_head,
    STATE(6585), 1,
      sym_sequential_block,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1654), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [110843] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6027), 1,
      sym_sequential_block,
    STATE(6414), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1655), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [110945] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6046), 1,
      sym_sequential_block,
    STATE(6421), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1656), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [111047] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3050), 1,
      sym_IS,
    STATE(2411), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6049), 1,
      sym_sequential_block,
    STATE(6422), 1,
      sym_sequential_block_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1657), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [111149] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3050), 1,
      sym_IS,
    STATE(2411), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6061), 1,
      sym_sequential_block,
    STATE(6432), 1,
      sym_sequential_block_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1658), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [111251] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6059), 1,
      sym_sequential_block,
    STATE(6434), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1659), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [111353] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6094), 1,
      sym_sequential_block,
    STATE(6445), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1660), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [111455] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1661), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [111515] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6109), 1,
      sym_sequential_block,
    STATE(6450), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1662), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [111617] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1663), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [111677] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6120), 1,
      sym_sequential_block,
    STATE(6456), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1664), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [111779] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1665), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [111839] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6136), 1,
      sym_sequential_block,
    STATE(6464), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1666), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [111941] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1667), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [112001] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6144), 1,
      sym_sequential_block,
    STATE(6465), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1668), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [112103] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1669), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [112163] = 27,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2944), 1,
      sym_BEGIN,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2954), 1,
      sym_IS,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    STATE(2336), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    STATE(6150), 1,
      sym_sequential_block,
    STATE(6466), 1,
      sym_process_head,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(1670), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [112265] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1671), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [112325] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1672), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 35,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [112385] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1673), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 35,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [112445] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1674), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [112504] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1675), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 34,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [112563] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1676), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [112622] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3168), 1,
      sym_exponentiate,
    STATE(517), 1,
      sym_adding_operator,
    STATE(518), 1,
      sym_multiplying_operator,
    STATE(614), 1,
      sym_logical_operator,
    STATE(615), 1,
      sym_relational_operator,
    STATE(616), 1,
      sym_shift_operator,
    STATE(1677), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [112693] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3172), 1,
      sym_library_constant_unit,
    STATE(2668), 1,
      sym__unit,
    STATE(1678), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3170), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [112758] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1679), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [112817] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1680), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 34,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [112876] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(471), 1,
      sym_adding_operator,
    STATE(472), 1,
      sym_multiplying_operator,
    STATE(622), 1,
      sym_logical_operator,
    STATE(623), 1,
      sym_relational_operator,
    STATE(624), 1,
      sym_shift_operator,
    STATE(1681), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [112945] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1682), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3174), 40,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_END,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [113002] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1683), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 34,
      sym_identifier,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [113061] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1684), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3176), 40,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_FORCE,
      sym_INERTIAL,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_REJECT,
      sym_RELEASE,
      sym_TRANSPORT,
      sym_UNAFFECTED,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [113118] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1685), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [113177] = 25,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(702), 1,
      anon_sym_LPAREN,
    ACTIONS(705), 1,
      anon_sym_LT_LT,
    ACTIONS(711), 1,
      sym_END,
    ACTIONS(713), 1,
      sym_ASSERT,
    ACTIONS(722), 1,
      sym_WITH,
    ACTIONS(3178), 1,
      sym_POSTPONED,
    ACTIONS(3181), 1,
      sym_PROCESS,
    STATE(2626), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4386), 1,
      sym__concurrent_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6317), 1,
      sym_name,
    STATE(6373), 1,
      sym__target,
    STATE(6470), 1,
      sym_with_expression,
    STATE(7169), 1,
      sym__label,
    STATE(7448), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(725), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(708), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(1686), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_design_unit_repeat3,
    STATE(4354), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [113274] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1687), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [113333] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3184), 1,
      sym_exponentiate,
    STATE(471), 1,
      sym_adding_operator,
    STATE(472), 1,
      sym_multiplying_operator,
    STATE(622), 1,
      sym_logical_operator,
    STATE(623), 1,
      sym_relational_operator,
    STATE(624), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1688), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 26,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [113408] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1689), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [113467] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1690), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [113526] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1691), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [113585] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3184), 1,
      sym_exponentiate,
    STATE(471), 1,
      sym_adding_operator,
    STATE(472), 1,
      sym_multiplying_operator,
    STATE(622), 1,
      sym_logical_operator,
    STATE(623), 1,
      sym_relational_operator,
    STATE(624), 1,
      sym_shift_operator,
    STATE(1692), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [113656] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2999), 1,
      sym_exponentiate,
    STATE(669), 1,
      sym_logical_operator,
    STATE(670), 1,
      sym_relational_operator,
    STATE(681), 1,
      sym_shift_operator,
    STATE(682), 1,
      sym_adding_operator,
    STATE(684), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1693), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 26,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [113731] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1694), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [113790] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1695), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [113849] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1696), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [113908] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1697), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [113967] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3186), 1,
      sym_exponentiate,
    STATE(507), 1,
      sym_adding_operator,
    STATE(508), 1,
      sym_multiplying_operator,
    STATE(597), 1,
      sym_logical_operator,
    STATE(598), 1,
      sym_relational_operator,
    STATE(599), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2690), 3,
      sym_IS,
      sym_OPEN,
      anon_sym_SEMI,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1698), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [114050] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1699), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [114109] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1700), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [114168] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(466), 1,
      sym_adding_operator,
    STATE(467), 1,
      sym_multiplying_operator,
    STATE(618), 1,
      sym_logical_operator,
    STATE(619), 1,
      sym_relational_operator,
    STATE(620), 1,
      sym_shift_operator,
    STATE(1701), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [114237] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3188), 1,
      sym_exponentiate,
    STATE(466), 1,
      sym_adding_operator,
    STATE(467), 1,
      sym_multiplying_operator,
    STATE(618), 1,
      sym_logical_operator,
    STATE(619), 1,
      sym_relational_operator,
    STATE(620), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1702), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 26,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [114312] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1703), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [114371] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1704), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [114430] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1705), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [114489] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1706), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [114548] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1707), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3190), 40,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_END,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [114605] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1708), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [114664] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1709), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 34,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [114723] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1710), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 34,
      sym_identifier,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [114782] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3192), 1,
      sym_exponentiate,
    STATE(481), 1,
      sym_adding_operator,
    STATE(482), 1,
      sym_multiplying_operator,
    STATE(630), 1,
      sym_logical_operator,
    STATE(631), 1,
      sym_relational_operator,
    STATE(632), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2690), 3,
      sym_DOWNTO,
      sym_LOOP,
      sym_TO,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1711), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [114865] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(466), 1,
      sym_adding_operator,
    STATE(467), 1,
      sym_multiplying_operator,
    STATE(618), 1,
      sym_logical_operator,
    STATE(619), 1,
      sym_relational_operator,
    STATE(620), 1,
      sym_shift_operator,
    STATE(1712), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [114934] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1713), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [114993] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3186), 1,
      sym_exponentiate,
    STATE(507), 1,
      sym_adding_operator,
    STATE(508), 1,
      sym_multiplying_operator,
    STATE(597), 1,
      sym_logical_operator,
    STATE(598), 1,
      sym_relational_operator,
    STATE(599), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1714), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 9,
      sym_AND,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
  [115074] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1715), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [115133] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(651), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(1716), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 30,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [115196] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1717), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [115255] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3184), 1,
      sym_exponentiate,
    STATE(471), 1,
      sym_adding_operator,
    STATE(472), 1,
      sym_multiplying_operator,
    STATE(622), 1,
      sym_logical_operator,
    STATE(623), 1,
      sym_relational_operator,
    STATE(624), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1718), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 9,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
  [115336] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3186), 1,
      sym_exponentiate,
    STATE(507), 1,
      sym_adding_operator,
    STATE(508), 1,
      sym_multiplying_operator,
    STATE(597), 1,
      sym_logical_operator,
    STATE(598), 1,
      sym_relational_operator,
    STATE(599), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1719), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 17,
      sym_AND,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [115415] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(507), 1,
      sym_adding_operator,
    STATE(508), 1,
      sym_multiplying_operator,
    STATE(597), 1,
      sym_logical_operator,
    STATE(598), 1,
      sym_relational_operator,
    STATE(599), 1,
      sym_shift_operator,
    STATE(1720), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [115484] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3186), 1,
      sym_exponentiate,
    STATE(507), 1,
      sym_adding_operator,
    STATE(508), 1,
      sym_multiplying_operator,
    STATE(597), 1,
      sym_logical_operator,
    STATE(598), 1,
      sym_relational_operator,
    STATE(599), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1721), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 26,
      sym_AND,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [115559] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(507), 1,
      sym_adding_operator,
    STATE(508), 1,
      sym_multiplying_operator,
    STATE(597), 1,
      sym_logical_operator,
    STATE(598), 1,
      sym_relational_operator,
    STATE(599), 1,
      sym_shift_operator,
    STATE(1722), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [115628] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3192), 1,
      sym_exponentiate,
    STATE(481), 1,
      sym_adding_operator,
    STATE(482), 1,
      sym_multiplying_operator,
    STATE(630), 1,
      sym_logical_operator,
    STATE(631), 1,
      sym_relational_operator,
    STATE(632), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1723), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 9,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
  [115709] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1724), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [115768] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3168), 1,
      sym_exponentiate,
    STATE(517), 1,
      sym_adding_operator,
    STATE(518), 1,
      sym_multiplying_operator,
    STATE(614), 1,
      sym_logical_operator,
    STATE(615), 1,
      sym_relational_operator,
    STATE(616), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2690), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      sym_variable_assignment,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1725), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [115851] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1726), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [115910] = 26,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(227), 1,
      sym_POSTPONED,
    ACTIONS(229), 1,
      sym_PROCESS,
    ACTIONS(3010), 1,
      sym_END,
    STATE(1686), 1,
      aux_sym_design_unit_repeat3,
    STATE(2626), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4386), 1,
      sym__concurrent_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6317), 1,
      sym_name,
    STATE(6373), 1,
      sym__target,
    STATE(6470), 1,
      sym_with_expression,
    STATE(7169), 1,
      sym__label,
    STATE(7448), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1727), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4354), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [116009] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1728), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116068] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1729), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116127] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1730), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 34,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [116186] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2498), 1,
      anon_sym_LPAREN,
    ACTIONS(3196), 1,
      sym_PARAMETER,
    STATE(2152), 1,
      sym_parenthesis_group,
    STATE(1731), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 31,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116251] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1732), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116310] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3184), 1,
      sym_exponentiate,
    STATE(471), 1,
      sym_adding_operator,
    STATE(472), 1,
      sym_multiplying_operator,
    STATE(622), 1,
      sym_logical_operator,
    STATE(623), 1,
      sym_relational_operator,
    STATE(624), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1733), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 23,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [116387] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1734), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116446] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1735), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116505] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1736), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116564] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1737), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116623] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1738), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116682] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1739), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116741] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1740), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116800] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1741), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116859] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1742), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116918] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1743), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [116977] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1744), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [117036] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1745), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [117095] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1746), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [117154] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2430), 1,
      anon_sym_LPAREN,
    ACTIONS(3198), 1,
      sym_PARAMETER,
    STATE(2125), 1,
      sym_parenthesis_group,
    STATE(1747), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 31,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [117219] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1748), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [117278] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1749), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [117337] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3168), 1,
      sym_exponentiate,
    STATE(517), 1,
      sym_adding_operator,
    STATE(518), 1,
      sym_multiplying_operator,
    STATE(614), 1,
      sym_logical_operator,
    STATE(615), 1,
      sym_relational_operator,
    STATE(616), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1750), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 9,
      sym_AND,
      sym_GENERIC,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      sym_variable_assignment,
  [117418] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(669), 1,
      sym_logical_operator,
    STATE(670), 1,
      sym_relational_operator,
    STATE(681), 1,
      sym_shift_operator,
    STATE(682), 1,
      sym_adding_operator,
    STATE(684), 1,
      sym_multiplying_operator,
    STATE(1751), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [117487] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1752), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [117546] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3168), 1,
      sym_exponentiate,
    STATE(517), 1,
      sym_adding_operator,
    STATE(518), 1,
      sym_multiplying_operator,
    STATE(614), 1,
      sym_logical_operator,
    STATE(615), 1,
      sym_relational_operator,
    STATE(616), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1753), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 17,
      sym_AND,
      sym_GENERIC,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [117625] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1754), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [117684] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3168), 1,
      sym_exponentiate,
    STATE(517), 1,
      sym_adding_operator,
    STATE(518), 1,
      sym_multiplying_operator,
    STATE(614), 1,
      sym_logical_operator,
    STATE(615), 1,
      sym_relational_operator,
    STATE(616), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1755), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 23,
      sym_AND,
      sym_GENERIC,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [117761] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    ACTIONS(3200), 3,
      sym_WHEN,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(1756), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [117844] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1757), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [117903] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1758), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [117962] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1759), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118021] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1760), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118080] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1761), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118139] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1762), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118198] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1763), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 34,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [118257] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1764), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118316] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1765), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118375] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1766), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118434] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1767), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118493] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1768), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118552] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3204), 1,
      sym_library_constant_unit,
    STATE(2641), 1,
      sym__unit,
    STATE(1769), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3202), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [118617] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1770), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118676] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3192), 1,
      sym_exponentiate,
    STATE(481), 1,
      sym_adding_operator,
    STATE(482), 1,
      sym_multiplying_operator,
    STATE(630), 1,
      sym_logical_operator,
    STATE(631), 1,
      sym_relational_operator,
    STATE(632), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1771), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 17,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [118755] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3208), 1,
      sym_library_constant_unit,
    STATE(2684), 1,
      sym__unit,
    STATE(1772), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3206), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [118820] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1773), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [118879] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3192), 1,
      sym_exponentiate,
    STATE(481), 1,
      sym_adding_operator,
    STATE(482), 1,
      sym_multiplying_operator,
    STATE(630), 1,
      sym_logical_operator,
    STATE(631), 1,
      sym_relational_operator,
    STATE(632), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1774), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 23,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [118956] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3186), 1,
      sym_exponentiate,
    STATE(507), 1,
      sym_adding_operator,
    STATE(508), 1,
      sym_multiplying_operator,
    STATE(597), 1,
      sym_logical_operator,
    STATE(598), 1,
      sym_relational_operator,
    STATE(599), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1775), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 26,
      sym_AND,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [119031] = 26,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(227), 1,
      sym_POSTPONED,
    ACTIONS(229), 1,
      sym_PROCESS,
    ACTIONS(3210), 1,
      sym_END,
    STATE(1890), 1,
      aux_sym_design_unit_repeat3,
    STATE(2626), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4386), 1,
      sym__concurrent_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6317), 1,
      sym_name,
    STATE(6373), 1,
      sym__target,
    STATE(6470), 1,
      sym_with_expression,
    STATE(7169), 1,
      sym__label,
    STATE(7448), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1776), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4354), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [119130] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3188), 1,
      sym_exponentiate,
    STATE(466), 1,
      sym_adding_operator,
    STATE(467), 1,
      sym_multiplying_operator,
    STATE(618), 1,
      sym_logical_operator,
    STATE(619), 1,
      sym_relational_operator,
    STATE(620), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2690), 3,
      sym_DOWNTO,
      sym_GENERATE,
      sym_TO,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1777), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [119213] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3186), 1,
      sym_exponentiate,
    STATE(507), 1,
      sym_adding_operator,
    STATE(508), 1,
      sym_multiplying_operator,
    STATE(597), 1,
      sym_logical_operator,
    STATE(598), 1,
      sym_relational_operator,
    STATE(599), 1,
      sym_shift_operator,
    STATE(1778), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [119284] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1779), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [119343] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1780), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [119402] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1781), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [119461] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2976), 1,
      sym_exponentiate,
    STATE(502), 1,
      sym_adding_operator,
    STATE(503), 1,
      sym_multiplying_operator,
    STATE(614), 1,
      sym_logical_operator,
    STATE(615), 1,
      sym_relational_operator,
    STATE(616), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    ACTIONS(2830), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(1782), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [119544] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1783), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 34,
      sym_identifier,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [119603] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1784), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [119662] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1785), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [119721] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1786), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3212), 40,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_END,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [119778] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1787), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [119837] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1788), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [119896] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1789), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 34,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [119955] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1790), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120014] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1791), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120073] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1792), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120132] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1793), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120191] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1794), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120250] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1795), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120309] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1796), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120368] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(481), 1,
      sym_adding_operator,
    STATE(482), 1,
      sym_multiplying_operator,
    STATE(630), 1,
      sym_logical_operator,
    STATE(631), 1,
      sym_relational_operator,
    STATE(632), 1,
      sym_shift_operator,
    STATE(1797), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [120437] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1798), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120496] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1799), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120555] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3188), 1,
      sym_exponentiate,
    STATE(466), 1,
      sym_adding_operator,
    STATE(467), 1,
      sym_multiplying_operator,
    STATE(618), 1,
      sym_logical_operator,
    STATE(619), 1,
      sym_relational_operator,
    STATE(620), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1800), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 9,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
  [120636] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1801), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120695] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1802), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [120754] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3168), 1,
      sym_exponentiate,
    STATE(517), 1,
      sym_adding_operator,
    STATE(518), 1,
      sym_multiplying_operator,
    STATE(614), 1,
      sym_logical_operator,
    STATE(615), 1,
      sym_relational_operator,
    STATE(616), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1803), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 26,
      sym_AND,
      sym_GENERIC,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [120829] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3192), 1,
      sym_exponentiate,
    STATE(481), 1,
      sym_adding_operator,
    STATE(482), 1,
      sym_multiplying_operator,
    STATE(630), 1,
      sym_logical_operator,
    STATE(631), 1,
      sym_relational_operator,
    STATE(632), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1804), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 26,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [120904] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3188), 1,
      sym_exponentiate,
    STATE(466), 1,
      sym_adding_operator,
    STATE(467), 1,
      sym_multiplying_operator,
    STATE(618), 1,
      sym_logical_operator,
    STATE(619), 1,
      sym_relational_operator,
    STATE(620), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1805), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 17,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [120983] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2462), 1,
      anon_sym_LPAREN,
    ACTIONS(3214), 1,
      sym_PARAMETER,
    STATE(2096), 1,
      sym_parenthesis_group,
    STATE(1806), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 31,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [121048] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1807), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [121107] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3186), 1,
      sym_exponentiate,
    STATE(507), 1,
      sym_adding_operator,
    STATE(508), 1,
      sym_multiplying_operator,
    STATE(597), 1,
      sym_logical_operator,
    STATE(598), 1,
      sym_relational_operator,
    STATE(599), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1808), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 23,
      sym_AND,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [121184] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(669), 1,
      sym_logical_operator,
    STATE(670), 1,
      sym_relational_operator,
    STATE(681), 1,
      sym_shift_operator,
    STATE(682), 1,
      sym_adding_operator,
    STATE(684), 1,
      sym_multiplying_operator,
    STATE(1809), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [121253] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1810), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [121312] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1811), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [121371] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3188), 1,
      sym_exponentiate,
    STATE(466), 1,
      sym_adding_operator,
    STATE(467), 1,
      sym_multiplying_operator,
    STATE(618), 1,
      sym_logical_operator,
    STATE(619), 1,
      sym_relational_operator,
    STATE(620), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1812), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 23,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [121448] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2999), 1,
      sym_exponentiate,
    STATE(669), 1,
      sym_logical_operator,
    STATE(670), 1,
      sym_relational_operator,
    STATE(681), 1,
      sym_shift_operator,
    STATE(682), 1,
      sym_adding_operator,
    STATE(684), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1813), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 9,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REPORT,
      sym_SEVERITY,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
  [121529] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2999), 1,
      sym_exponentiate,
    STATE(669), 1,
      sym_logical_operator,
    STATE(670), 1,
      sym_relational_operator,
    STATE(681), 1,
      sym_shift_operator,
    STATE(682), 1,
      sym_adding_operator,
    STATE(684), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1814), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 17,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REPORT,
      sym_SEVERITY,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [121608] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2999), 1,
      sym_exponentiate,
    STATE(669), 1,
      sym_logical_operator,
    STATE(670), 1,
      sym_relational_operator,
    STATE(681), 1,
      sym_shift_operator,
    STATE(682), 1,
      sym_adding_operator,
    STATE(684), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1815), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 23,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [121685] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2999), 1,
      sym_exponentiate,
    STATE(669), 1,
      sym_logical_operator,
    STATE(670), 1,
      sym_relational_operator,
    STATE(681), 1,
      sym_shift_operator,
    STATE(682), 1,
      sym_adding_operator,
    STATE(684), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1816), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 26,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [121760] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3216), 1,
      sym_RANGE,
    STATE(1817), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [121821] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2999), 1,
      sym_exponentiate,
    STATE(669), 1,
      sym_logical_operator,
    STATE(670), 1,
      sym_relational_operator,
    STATE(681), 1,
      sym_shift_operator,
    STATE(682), 1,
      sym_adding_operator,
    STATE(684), 1,
      sym_multiplying_operator,
    STATE(1818), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [121892] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(481), 1,
      sym_adding_operator,
    STATE(482), 1,
      sym_multiplying_operator,
    STATE(630), 1,
      sym_logical_operator,
    STATE(631), 1,
      sym_relational_operator,
    STATE(632), 1,
      sym_shift_operator,
    STATE(1819), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [121961] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1820), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122020] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1821), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122079] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1822), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122138] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1823), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 34,
      sym_identifier,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [122197] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1824), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122256] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1825), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122315] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1826), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122374] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1827), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122433] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1828), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122492] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1829), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122551] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1830), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122610] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1831), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122669] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1832), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122728] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1833), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122787] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1834), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122846] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1835), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122905] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2376), 1,
      anon_sym_LPAREN,
    ACTIONS(3218), 1,
      sym_PARAMETER,
    STATE(2160), 1,
      sym_parenthesis_group,
    STATE(1836), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 31,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [122970] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(651), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(1837), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
  [123033] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1838), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123092] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3188), 1,
      sym_exponentiate,
    STATE(466), 1,
      sym_adding_operator,
    STATE(467), 1,
      sym_multiplying_operator,
    STATE(618), 1,
      sym_logical_operator,
    STATE(619), 1,
      sym_relational_operator,
    STATE(620), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1839), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 26,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [123167] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1840), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123226] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1841), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123285] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1842), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123344] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1843), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123403] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2974), 1,
      sym_exponentiate,
    STATE(476), 1,
      sym_adding_operator,
    STATE(477), 1,
      sym_multiplying_operator,
    STATE(597), 1,
      sym_logical_operator,
    STATE(598), 1,
      sym_relational_operator,
    STATE(599), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    ACTIONS(2830), 3,
      sym_IS,
      sym_OPEN,
      anon_sym_SEMI,
    STATE(1844), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [123486] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3192), 1,
      sym_exponentiate,
    STATE(481), 1,
      sym_adding_operator,
    STATE(482), 1,
      sym_multiplying_operator,
    STATE(630), 1,
      sym_logical_operator,
    STATE(631), 1,
      sym_relational_operator,
    STATE(632), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1845), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 26,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [123561] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1846), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123620] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1847), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123679] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1848), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123738] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1849), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123797] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1850), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123856] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1851), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123915] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1852), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [123974] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1853), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [124033] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1854), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [124092] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1855), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [124151] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1856), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3220), 40,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_END,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [124208] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1857), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [124267] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1858), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [124326] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1859), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [124385] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3224), 1,
      sym_library_constant_unit,
    STATE(2670), 1,
      sym__unit,
    STATE(1860), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3222), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [124450] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(471), 1,
      sym_adding_operator,
    STATE(472), 1,
      sym_multiplying_operator,
    STATE(622), 1,
      sym_logical_operator,
    STATE(623), 1,
      sym_relational_operator,
    STATE(624), 1,
      sym_shift_operator,
    STATE(1861), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [124519] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3228), 1,
      sym_library_constant_unit,
    STATE(2618), 1,
      sym__unit,
    STATE(1862), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3226), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [124584] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1863), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [124643] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1864), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [124702] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1865), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [124761] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3184), 1,
      sym_exponentiate,
    STATE(471), 1,
      sym_adding_operator,
    STATE(472), 1,
      sym_multiplying_operator,
    STATE(622), 1,
      sym_logical_operator,
    STATE(623), 1,
      sym_relational_operator,
    STATE(624), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1866), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 26,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [124836] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1867), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [124895] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(517), 1,
      sym_adding_operator,
    STATE(518), 1,
      sym_multiplying_operator,
    STATE(614), 1,
      sym_logical_operator,
    STATE(615), 1,
      sym_relational_operator,
    STATE(616), 1,
      sym_shift_operator,
    STATE(1868), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [124964] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1869), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [125023] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3168), 1,
      sym_exponentiate,
    STATE(517), 1,
      sym_adding_operator,
    STATE(518), 1,
      sym_multiplying_operator,
    STATE(614), 1,
      sym_logical_operator,
    STATE(615), 1,
      sym_relational_operator,
    STATE(616), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1870), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 26,
      sym_AND,
      sym_GENERIC,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [125098] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1871), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3230), 40,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_END,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [125155] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1872), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [125214] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1873), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 34,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [125273] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2999), 1,
      sym_exponentiate,
    STATE(669), 1,
      sym_logical_operator,
    STATE(670), 1,
      sym_relational_operator,
    STATE(681), 1,
      sym_shift_operator,
    STATE(682), 1,
      sym_adding_operator,
    STATE(684), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2690), 3,
      sym_REPORT,
      sym_SEVERITY,
      anon_sym_SEMI,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1874), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [125356] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3192), 1,
      sym_exponentiate,
    STATE(481), 1,
      sym_adding_operator,
    STATE(482), 1,
      sym_multiplying_operator,
    STATE(630), 1,
      sym_logical_operator,
    STATE(631), 1,
      sym_relational_operator,
    STATE(632), 1,
      sym_shift_operator,
    STATE(1875), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [125427] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1876), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [125486] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1877), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [125545] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1878), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [125604] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1879), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [125663] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1880), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [125722] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1881), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [125781] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2418), 1,
      anon_sym_LPAREN,
    ACTIONS(3232), 1,
      sym_PARAMETER,
    STATE(2238), 1,
      sym_parenthesis_group,
    STATE(1882), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 31,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [125846] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3184), 1,
      sym_exponentiate,
    STATE(471), 1,
      sym_adding_operator,
    STATE(472), 1,
      sym_multiplying_operator,
    STATE(622), 1,
      sym_logical_operator,
    STATE(623), 1,
      sym_relational_operator,
    STATE(624), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1883), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 17,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [125925] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1884), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [125984] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1885), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126043] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3188), 1,
      sym_exponentiate,
    STATE(466), 1,
      sym_adding_operator,
    STATE(467), 1,
      sym_multiplying_operator,
    STATE(618), 1,
      sym_logical_operator,
    STATE(619), 1,
      sym_relational_operator,
    STATE(620), 1,
      sym_shift_operator,
    STATE(1886), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [126114] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1887), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 34,
      sym_identifier,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [126173] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1888), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126232] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3184), 1,
      sym_exponentiate,
    STATE(471), 1,
      sym_adding_operator,
    STATE(472), 1,
      sym_multiplying_operator,
    STATE(622), 1,
      sym_logical_operator,
    STATE(623), 1,
      sym_relational_operator,
    STATE(624), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2690), 3,
      sym_DOWNTO,
      sym_IS,
      sym_TO,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1889), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [126315] = 26,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(227), 1,
      sym_POSTPONED,
    ACTIONS(229), 1,
      sym_PROCESS,
    ACTIONS(3234), 1,
      sym_END,
    STATE(1686), 1,
      aux_sym_design_unit_repeat3,
    STATE(2626), 1,
      sym_label_declaration,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4386), 1,
      sym__concurrent_statement,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6317), 1,
      sym_name,
    STATE(6373), 1,
      sym__target,
    STATE(6470), 1,
      sym_with_expression,
    STATE(7169), 1,
      sym__label,
    STATE(7448), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1890), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4354), 11,
      sym_block_statement,
      sym_component_instantiation_statement,
      sym_process_statement,
      sym_case_generate_statement,
      sym_for_generate_statement,
      sym_if_generate_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
      sym_concurrent_simple_signal_assignment,
      sym_concurrent_conditional_signal_assignment,
      sym_concurrent_selected_signal_assignment,
  [126414] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1891), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126473] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1892), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126532] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1893), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126591] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1894), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126650] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1895), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126709] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2574), 1,
      anon_sym_LPAREN,
    ACTIONS(3236), 1,
      sym_PARAMETER,
    STATE(1994), 1,
      sym_parenthesis_group,
    STATE(1896), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2193), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2191), 31,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126774] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1897), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126833] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1898), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126892] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1899), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [126951] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1900), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127010] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(517), 1,
      sym_adding_operator,
    STATE(518), 1,
      sym_multiplying_operator,
    STATE(614), 1,
      sym_logical_operator,
    STATE(615), 1,
      sym_relational_operator,
    STATE(616), 1,
      sym_shift_operator,
    STATE(1901), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [127079] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1902), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127138] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1903), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127197] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1904), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127256] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1905), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127315] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1906), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127374] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1907), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 34,
      sym_AND,
      sym_ELSE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127433] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1908), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127492] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1909), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127551] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2968), 1,
      anon_sym_COLON,
    STATE(1910), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127612] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1911), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 34,
      sym_AND,
      sym_FOR,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127671] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1912), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 34,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127730] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1913), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [127788] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1914), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3176), 39,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_GUARDED,
      sym_INERTIAL,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_REJECT,
      sym_TRANSPORT,
      sym_UNAFFECTED,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [127844] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1915), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3238), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [127900] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    ACTIONS(3242), 1,
      sym_GENERATE,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    STATE(5129), 1,
      sym_generate_body,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1916), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [127984] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1917), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3244), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128040] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1918), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3246), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128096] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1919), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3248), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128152] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1920), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3250), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128208] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1921), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3252), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128264] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1922), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3254), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128320] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1923), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3256), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128376] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1924), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3258), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128432] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1925), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3260), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128488] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1926), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3262), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128544] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1927), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 33,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [128602] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1928), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3264), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128658] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1929), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3266), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128714] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1930), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 33,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [128772] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3268), 1,
      sym_exponentiate,
    STATE(546), 1,
      sym_logical_operator,
    STATE(547), 1,
      sym_relational_operator,
    STATE(548), 1,
      sym_shift_operator,
    STATE(717), 1,
      sym_adding_operator,
    STATE(718), 1,
      sym_multiplying_operator,
    ACTIONS(2690), 2,
      sym_SEVERITY,
      anon_sym_SEMI,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1931), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [128854] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1932), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3270), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [128910] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(546), 1,
      sym_logical_operator,
    STATE(547), 1,
      sym_relational_operator,
    STATE(548), 1,
      sym_shift_operator,
    STATE(717), 1,
      sym_adding_operator,
    STATE(718), 1,
      sym_multiplying_operator,
    STATE(1933), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [128978] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3268), 1,
      sym_exponentiate,
    STATE(546), 1,
      sym_logical_operator,
    STATE(547), 1,
      sym_relational_operator,
    STATE(548), 1,
      sym_shift_operator,
    STATE(717), 1,
      sym_adding_operator,
    STATE(718), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1934), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 25,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [129052] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(546), 1,
      sym_logical_operator,
    STATE(547), 1,
      sym_relational_operator,
    STATE(548), 1,
      sym_shift_operator,
    STATE(717), 1,
      sym_adding_operator,
    STATE(718), 1,
      sym_multiplying_operator,
    STATE(1935), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [129120] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3268), 1,
      sym_exponentiate,
    STATE(546), 1,
      sym_logical_operator,
    STATE(547), 1,
      sym_relational_operator,
    STATE(548), 1,
      sym_shift_operator,
    STATE(717), 1,
      sym_adding_operator,
    STATE(718), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1936), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 8,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_SEVERITY,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
  [129200] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3268), 1,
      sym_exponentiate,
    STATE(546), 1,
      sym_logical_operator,
    STATE(547), 1,
      sym_relational_operator,
    STATE(548), 1,
      sym_shift_operator,
    STATE(717), 1,
      sym_adding_operator,
    STATE(718), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1937), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 16,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_SEVERITY,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [129278] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3268), 1,
      sym_exponentiate,
    STATE(546), 1,
      sym_logical_operator,
    STATE(547), 1,
      sym_relational_operator,
    STATE(548), 1,
      sym_shift_operator,
    STATE(717), 1,
      sym_adding_operator,
    STATE(718), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1938), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 22,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [129354] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3268), 1,
      sym_exponentiate,
    STATE(546), 1,
      sym_logical_operator,
    STATE(547), 1,
      sym_relational_operator,
    STATE(548), 1,
      sym_shift_operator,
    STATE(717), 1,
      sym_adding_operator,
    STATE(718), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(1939), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 25,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [129428] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3268), 1,
      sym_exponentiate,
    STATE(546), 1,
      sym_logical_operator,
    STATE(547), 1,
      sym_relational_operator,
    STATE(548), 1,
      sym_shift_operator,
    STATE(717), 1,
      sym_adding_operator,
    STATE(718), 1,
      sym_multiplying_operator,
    STATE(1940), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [129498] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    ACTIONS(3272), 1,
      sym_WHEN,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    STATE(6237), 1,
      sym_when_element,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1941), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [129582] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1942), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3274), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [129638] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1943), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [129696] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1944), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [129754] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1945), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [129812] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1946), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [129870] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1947), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [129928] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1948), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [129986] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1949), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [130044] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1950), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [130102] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3278), 1,
      sym_exponentiate,
    STATE(436), 1,
      sym_adding_operator,
    STATE(437), 1,
      sym_multiplying_operator,
    STATE(571), 1,
      sym_logical_operator,
    STATE(572), 1,
      sym_relational_operator,
    STATE(573), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(3276), 2,
      sym_ELSE,
      anon_sym_SEMI,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1951), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [130184] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1952), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3280), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [130240] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1953), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3282), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [130296] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1954), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [130354] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1955), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [130412] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1956), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [130470] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1957), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3284), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [130526] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1958), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3286), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_PIPE,
    ACTIONS(3164), 25,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [130586] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1959), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3288), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [130642] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1960), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3290), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [130698] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1961), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3292), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [130754] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1962), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3294), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [130810] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    ACTIONS(3242), 1,
      sym_GENERATE,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    STATE(5056), 1,
      sym_generate_body,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1963), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [130894] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1964), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [130952] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1965), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(672), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131008] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1966), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3296), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131064] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1967), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3298), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131120] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1968), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3300), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131176] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1969), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3302), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131232] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1970), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3304), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131288] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1971), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3306), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131344] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1972), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3308), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131400] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    ACTIONS(3272), 1,
      sym_WHEN,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    STATE(6717), 1,
      sym_when_element,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(1973), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [131484] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1974), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3310), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131540] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1975), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [131598] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1976), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [131656] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1977), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3312), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131712] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3316), 1,
      sym_library_constant_unit,
    STATE(2749), 1,
      sym__unit,
    STATE(1978), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3314), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [131776] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1979), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3318), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [131832] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1980), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [131890] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1981), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [131948] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1982), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3274), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [132004] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1983), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3320), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [132060] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1984), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3322), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [132116] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1985), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3324), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [132172] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1986), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132230] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1987), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132288] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1988), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132346] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(564), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(1989), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [132408] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1990), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132466] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1991), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132524] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1992), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132582] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1993), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132640] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1994), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132698] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1995), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 32,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [132756] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1996), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132814] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1997), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132872] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1998), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132930] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1999), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [132988] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2000), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133046] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2001), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133104] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2002), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133162] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2003), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133220] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2004), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133278] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2005), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133336] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2006), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133394] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2007), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133452] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3328), 1,
      sym_library_constant_unit,
    STATE(2730), 1,
      sym__unit,
    STATE(2008), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3326), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [133516] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2009), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133574] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2010), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3330), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [133630] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2011), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3332), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [133686] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2012), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3334), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [133742] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2013), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133800] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3338), 1,
      sym_library_constant_unit,
    STATE(2760), 1,
      sym__unit,
    STATE(2014), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3336), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [133864] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(574), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(2015), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [133926] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2016), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [133984] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2017), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [134042] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2018), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [134100] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2019), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 33,
      sym_identifier,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [134158] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3342), 1,
      sym_library_constant_unit,
    STATE(2731), 1,
      sym__unit,
    STATE(2020), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3340), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [134222] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(651), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(2021), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [134284] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    ACTIONS(3242), 1,
      sym_GENERATE,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    STATE(5111), 1,
      sym_generate_body,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2022), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [134368] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2023), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 33,
      sym_identifier,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [134426] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3278), 1,
      sym_exponentiate,
    STATE(436), 1,
      sym_adding_operator,
    STATE(437), 1,
      sym_multiplying_operator,
    STATE(571), 1,
      sym_logical_operator,
    STATE(572), 1,
      sym_relational_operator,
    STATE(573), 1,
      sym_shift_operator,
    ACTIONS(2690), 2,
      sym_ELSE,
      anon_sym_SEMI,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2024), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [134508] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2025), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3344), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [134564] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2026), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3346), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [134620] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(436), 1,
      sym_adding_operator,
    STATE(437), 1,
      sym_multiplying_operator,
    STATE(571), 1,
      sym_logical_operator,
    STATE(572), 1,
      sym_relational_operator,
    STATE(573), 1,
      sym_shift_operator,
    STATE(2027), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [134688] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3278), 1,
      sym_exponentiate,
    STATE(436), 1,
      sym_adding_operator,
    STATE(437), 1,
      sym_multiplying_operator,
    STATE(571), 1,
      sym_logical_operator,
    STATE(572), 1,
      sym_relational_operator,
    STATE(573), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2028), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 25,
      sym_AND,
      sym_ELSE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [134762] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2029), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3348), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [134818] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2030), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3350), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [134874] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2031), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3352), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [134930] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2032), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [134988] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2033), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3354), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [135044] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(436), 1,
      sym_adding_operator,
    STATE(437), 1,
      sym_multiplying_operator,
    STATE(571), 1,
      sym_logical_operator,
    STATE(572), 1,
      sym_relational_operator,
    STATE(573), 1,
      sym_shift_operator,
    STATE(2034), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [135112] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3278), 1,
      sym_exponentiate,
    STATE(436), 1,
      sym_adding_operator,
    STATE(437), 1,
      sym_multiplying_operator,
    STATE(571), 1,
      sym_logical_operator,
    STATE(572), 1,
      sym_relational_operator,
    STATE(573), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2035), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 8,
      sym_AND,
      sym_ELSE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
  [135192] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3278), 1,
      sym_exponentiate,
    STATE(436), 1,
      sym_adding_operator,
    STATE(437), 1,
      sym_multiplying_operator,
    STATE(571), 1,
      sym_logical_operator,
    STATE(572), 1,
      sym_relational_operator,
    STATE(573), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2036), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 16,
      sym_AND,
      sym_ELSE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [135270] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3278), 1,
      sym_exponentiate,
    STATE(436), 1,
      sym_adding_operator,
    STATE(437), 1,
      sym_multiplying_operator,
    STATE(571), 1,
      sym_logical_operator,
    STATE(572), 1,
      sym_relational_operator,
    STATE(573), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2037), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 22,
      sym_AND,
      sym_ELSE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [135346] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3278), 1,
      sym_exponentiate,
    STATE(436), 1,
      sym_adding_operator,
    STATE(437), 1,
      sym_multiplying_operator,
    STATE(571), 1,
      sym_logical_operator,
    STATE(572), 1,
      sym_relational_operator,
    STATE(573), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2038), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 25,
      sym_AND,
      sym_ELSE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [135420] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3278), 1,
      sym_exponentiate,
    STATE(436), 1,
      sym_adding_operator,
    STATE(437), 1,
      sym_multiplying_operator,
    STATE(571), 1,
      sym_logical_operator,
    STATE(572), 1,
      sym_relational_operator,
    STATE(573), 1,
      sym_shift_operator,
    STATE(2039), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [135490] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2040), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3274), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [135546] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2041), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3356), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [135602] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2042), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [135660] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2043), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3358), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [135716] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2044), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3360), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [135772] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2045), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3362), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [135828] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2046), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3364), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [135884] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2047), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3366), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [135940] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3046), 1,
      sym_exponentiate,
    STATE(456), 1,
      sym_adding_operator,
    STATE(457), 1,
      sym_multiplying_operator,
    STATE(584), 1,
      sym_logical_operator,
    STATE(585), 1,
      sym_relational_operator,
    STATE(586), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2830), 2,
      sym_OF,
      anon_sym_COMMA,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2048), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [136022] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2049), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3368), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [136078] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    ACTIONS(3242), 1,
      sym_GENERATE,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    STATE(5147), 1,
      sym_generate_body,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2050), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [136162] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2051), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3370), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [136218] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2052), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3372), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [136274] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2053), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3374), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [136330] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3378), 1,
      sym_library_constant_unit,
    STATE(2704), 1,
      sym__unit,
    STATE(2054), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3376), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [136394] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3380), 1,
      sym_exponentiate,
    STATE(451), 1,
      sym_adding_operator,
    STATE(452), 1,
      sym_multiplying_operator,
    STATE(580), 1,
      sym_logical_operator,
    STATE(581), 1,
      sym_relational_operator,
    STATE(582), 1,
      sym_shift_operator,
    ACTIONS(2690), 2,
      sym_FOR,
      anon_sym_SEMI,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2055), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [136476] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2056), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [136534] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3380), 1,
      sym_exponentiate,
    STATE(451), 1,
      sym_adding_operator,
    STATE(452), 1,
      sym_multiplying_operator,
    STATE(580), 1,
      sym_logical_operator,
    STATE(581), 1,
      sym_relational_operator,
    STATE(582), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2057), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 8,
      sym_AND,
      sym_FOR,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
  [136614] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3380), 1,
      sym_exponentiate,
    STATE(451), 1,
      sym_adding_operator,
    STATE(452), 1,
      sym_multiplying_operator,
    STATE(580), 1,
      sym_logical_operator,
    STATE(581), 1,
      sym_relational_operator,
    STATE(582), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2058), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 16,
      sym_AND,
      sym_FOR,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [136692] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3380), 1,
      sym_exponentiate,
    STATE(451), 1,
      sym_adding_operator,
    STATE(452), 1,
      sym_multiplying_operator,
    STATE(580), 1,
      sym_logical_operator,
    STATE(581), 1,
      sym_relational_operator,
    STATE(582), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2059), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 22,
      sym_AND,
      sym_FOR,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [136768] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2060), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3382), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [136824] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2061), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 33,
      sym_identifier,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [136882] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2062), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 33,
      sym_identifier,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [136940] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3384), 1,
      sym_exponentiate,
    STATE(491), 1,
      sym_adding_operator,
    STATE(492), 1,
      sym_multiplying_operator,
    STATE(584), 1,
      sym_logical_operator,
    STATE(585), 1,
      sym_relational_operator,
    STATE(586), 1,
      sym_shift_operator,
    ACTIONS(2690), 2,
      sym_OF,
      anon_sym_COMMA,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2063), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [137022] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(451), 1,
      sym_adding_operator,
    STATE(452), 1,
      sym_multiplying_operator,
    STATE(580), 1,
      sym_logical_operator,
    STATE(581), 1,
      sym_relational_operator,
    STATE(582), 1,
      sym_shift_operator,
    STATE(2064), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [137090] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3380), 1,
      sym_exponentiate,
    STATE(451), 1,
      sym_adding_operator,
    STATE(452), 1,
      sym_multiplying_operator,
    STATE(580), 1,
      sym_logical_operator,
    STATE(581), 1,
      sym_relational_operator,
    STATE(582), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2065), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 25,
      sym_AND,
      sym_FOR,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [137164] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(451), 1,
      sym_adding_operator,
    STATE(452), 1,
      sym_multiplying_operator,
    STATE(580), 1,
      sym_logical_operator,
    STATE(581), 1,
      sym_relational_operator,
    STATE(582), 1,
      sym_shift_operator,
    STATE(2066), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [137232] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3384), 1,
      sym_exponentiate,
    STATE(491), 1,
      sym_adding_operator,
    STATE(492), 1,
      sym_multiplying_operator,
    STATE(584), 1,
      sym_logical_operator,
    STATE(585), 1,
      sym_relational_operator,
    STATE(586), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2067), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 8,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
  [137312] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3384), 1,
      sym_exponentiate,
    STATE(491), 1,
      sym_adding_operator,
    STATE(492), 1,
      sym_multiplying_operator,
    STATE(584), 1,
      sym_logical_operator,
    STATE(585), 1,
      sym_relational_operator,
    STATE(586), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2068), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 16,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [137390] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3384), 1,
      sym_exponentiate,
    STATE(491), 1,
      sym_adding_operator,
    STATE(492), 1,
      sym_multiplying_operator,
    STATE(584), 1,
      sym_logical_operator,
    STATE(585), 1,
      sym_relational_operator,
    STATE(586), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2069), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 22,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [137466] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3380), 1,
      sym_exponentiate,
    STATE(451), 1,
      sym_adding_operator,
    STATE(452), 1,
      sym_multiplying_operator,
    STATE(580), 1,
      sym_logical_operator,
    STATE(581), 1,
      sym_relational_operator,
    STATE(582), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2070), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 25,
      sym_AND,
      sym_FOR,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [137540] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3380), 1,
      sym_exponentiate,
    STATE(451), 1,
      sym_adding_operator,
    STATE(452), 1,
      sym_multiplying_operator,
    STATE(580), 1,
      sym_logical_operator,
    STATE(581), 1,
      sym_relational_operator,
    STATE(582), 1,
      sym_shift_operator,
    STATE(2071), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [137610] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2072), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [137668] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2073), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3386), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [137724] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2074), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3388), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [137780] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2075), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3390), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [137836] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3392), 1,
      sym_exponentiate,
    STATE(497), 1,
      sym_adding_operator,
    STATE(498), 1,
      sym_multiplying_operator,
    STATE(588), 1,
      sym_logical_operator,
    STATE(589), 1,
      sym_relational_operator,
    STATE(590), 1,
      sym_shift_operator,
    ACTIONS(2690), 2,
      sym_UNITS,
      anon_sym_SEMI,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2076), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [137918] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3392), 1,
      sym_exponentiate,
    STATE(497), 1,
      sym_adding_operator,
    STATE(498), 1,
      sym_multiplying_operator,
    STATE(588), 1,
      sym_logical_operator,
    STATE(589), 1,
      sym_relational_operator,
    STATE(590), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2077), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
    ACTIONS(2708), 8,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
  [137998] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3392), 1,
      sym_exponentiate,
    STATE(497), 1,
      sym_adding_operator,
    STATE(498), 1,
      sym_multiplying_operator,
    STATE(588), 1,
      sym_logical_operator,
    STATE(589), 1,
      sym_relational_operator,
    STATE(590), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2078), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 16,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [138076] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3392), 1,
      sym_exponentiate,
    STATE(497), 1,
      sym_adding_operator,
    STATE(498), 1,
      sym_multiplying_operator,
    STATE(588), 1,
      sym_logical_operator,
    STATE(589), 1,
      sym_relational_operator,
    STATE(590), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2079), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 22,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [138152] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2080), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138210] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2081), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3394), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [138266] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2082), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138324] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2083), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138382] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2084), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138440] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2085), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138498] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2086), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3396), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [138554] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2087), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138612] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2088), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138670] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2089), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [138728] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2090), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138786] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2091), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138844] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2092), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138902] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2093), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [138960] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2094), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139018] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2095), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139076] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2096), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139134] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2097), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [139192] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2098), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [139250] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2099), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [139308] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2100), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139366] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2101), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139424] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2102), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139482] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2103), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139540] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2104), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139598] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2105), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139656] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2106), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139714] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2107), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139772] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2108), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139830] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2109), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [139888] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3380), 1,
      sym_exponentiate,
    STATE(451), 1,
      sym_adding_operator,
    STATE(452), 1,
      sym_multiplying_operator,
    STATE(580), 1,
      sym_logical_operator,
    STATE(581), 1,
      sym_relational_operator,
    STATE(582), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(3398), 2,
      sym_FOR,
      anon_sym_SEMI,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2110), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [139970] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2111), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140028] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2112), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140086] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2113), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140144] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2114), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3176), 39,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_FORCE,
      sym_INERTIAL,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_REJECT,
      sym_TRANSPORT,
      sym_UNAFFECTED,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [140200] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2115), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [140258] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3268), 1,
      sym_exponentiate,
    STATE(546), 1,
      sym_logical_operator,
    STATE(547), 1,
      sym_relational_operator,
    STATE(548), 1,
      sym_shift_operator,
    STATE(717), 1,
      sym_adding_operator,
    STATE(718), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(3400), 2,
      sym_SEVERITY,
      anon_sym_SEMI,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2116), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [140340] = 25,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(3402), 1,
      sym_ALIAS,
    ACTIONS(3404), 1,
      sym_ATTRIBUTE,
    ACTIONS(3406), 1,
      sym_CONSTANT,
    ACTIONS(3408), 1,
      sym_END,
    ACTIONS(3410), 1,
      sym_FILE,
    ACTIONS(3412), 1,
      sym_FUNCTION,
    ACTIONS(3414), 1,
      sym_GROUP,
    ACTIONS(3416), 1,
      sym_PACKAGE,
    ACTIONS(3418), 1,
      sym_SHARED,
    ACTIONS(3420), 1,
      sym_SUBTYPE,
    ACTIONS(3422), 1,
      sym_TYPE,
    ACTIONS(3424), 1,
      sym_USE,
    ACTIONS(3426), 1,
      sym_VARIABLE,
    STATE(2392), 1,
      aux_sym_protected_type_body_repeat1,
    STATE(3692), 1,
      sym__protected_type_body_declarative_item,
    STATE(5901), 1,
      sym__subprogram_specification,
    STATE(6815), 1,
      sym_protected_type_body_end,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2117), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3687), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [140436] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2118), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140494] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2119), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140552] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2120), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140610] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2121), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140668] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2122), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140726] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2123), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140784] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2124), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140842] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2125), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140900] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2126), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [140958] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2127), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141016] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2128), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141074] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2129), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141132] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2130), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141190] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2131), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141248] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2132), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141306] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2133), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141364] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2134), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141422] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2135), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141480] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2136), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141538] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2137), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141596] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2138), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [141654] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2139), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141712] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2140), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [141770] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2141), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [141828] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2142), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141886] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2143), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [141944] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2144), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2454), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2452), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142002] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2145), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142060] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2146), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142118] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2147), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142176] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2148), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142234] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2149), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142292] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2150), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142350] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2151), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142408] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2152), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142466] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2153), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142524] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2154), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142582] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2155), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142640] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2156), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142698] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2157), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142756] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2158), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142814] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2159), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142872] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2160), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [142930] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2161), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [142988] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2162), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [143046] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2163), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3428), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [143102] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2164), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3430), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [143158] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2165), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3432), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [143214] = 25,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(3402), 1,
      sym_ALIAS,
    ACTIONS(3404), 1,
      sym_ATTRIBUTE,
    ACTIONS(3406), 1,
      sym_CONSTANT,
    ACTIONS(3408), 1,
      sym_END,
    ACTIONS(3410), 1,
      sym_FILE,
    ACTIONS(3412), 1,
      sym_FUNCTION,
    ACTIONS(3414), 1,
      sym_GROUP,
    ACTIONS(3416), 1,
      sym_PACKAGE,
    ACTIONS(3418), 1,
      sym_SHARED,
    ACTIONS(3420), 1,
      sym_SUBTYPE,
    ACTIONS(3422), 1,
      sym_TYPE,
    ACTIONS(3424), 1,
      sym_USE,
    ACTIONS(3426), 1,
      sym_VARIABLE,
    STATE(2117), 1,
      aux_sym_protected_type_body_repeat1,
    STATE(3692), 1,
      sym__protected_type_body_declarative_item,
    STATE(5901), 1,
      sym__subprogram_specification,
    STATE(8089), 1,
      sym_protected_type_body_end,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2166), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3687), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [143310] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2167), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143368] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2168), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143426] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2169), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143484] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2170), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143542] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2171), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143600] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2172), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143658] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2173), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143716] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2174), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143774] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2175), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143832] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2176), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143890] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2177), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [143948] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2178), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144006] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2179), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144064] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2180), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144122] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2181), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3434), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [144178] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2182), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144236] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2183), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3436), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [144292] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3072), 1,
      sym_exponentiate,
    STATE(461), 1,
      sym_adding_operator,
    STATE(462), 1,
      sym_multiplying_operator,
    STATE(588), 1,
      sym_logical_operator,
    STATE(589), 1,
      sym_relational_operator,
    STATE(590), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2830), 2,
      sym_UNITS,
      anon_sym_SEMI,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2184), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [144374] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2185), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144432] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2186), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144490] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2187), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144548] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2188), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144606] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2189), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144664] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2190), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144722] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2191), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144780] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2192), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144838] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2193), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144896] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2194), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [144954] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2195), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145012] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2196), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145070] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2197), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145128] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2198), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145186] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2199), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145244] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2200), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145302] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2201), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145360] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2202), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145418] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2203), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145476] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2204), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145534] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2205), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145592] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2206), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145650] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2207), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 33,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [145708] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2208), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [145766] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2209), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 33,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [145824] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(491), 1,
      sym_adding_operator,
    STATE(492), 1,
      sym_multiplying_operator,
    STATE(584), 1,
      sym_logical_operator,
    STATE(585), 1,
      sym_relational_operator,
    STATE(586), 1,
      sym_shift_operator,
    STATE(2210), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [145892] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3384), 1,
      sym_exponentiate,
    STATE(491), 1,
      sym_adding_operator,
    STATE(492), 1,
      sym_multiplying_operator,
    STATE(584), 1,
      sym_logical_operator,
    STATE(585), 1,
      sym_relational_operator,
    STATE(586), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2211), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 25,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [145966] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2212), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3438), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [146022] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2213), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3440), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [146078] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(491), 1,
      sym_adding_operator,
    STATE(492), 1,
      sym_multiplying_operator,
    STATE(584), 1,
      sym_logical_operator,
    STATE(585), 1,
      sym_relational_operator,
    STATE(586), 1,
      sym_shift_operator,
    STATE(2214), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [146146] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2215), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3442), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [146202] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2216), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [146260] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3384), 1,
      sym_exponentiate,
    STATE(491), 1,
      sym_adding_operator,
    STATE(492), 1,
      sym_multiplying_operator,
    STATE(584), 1,
      sym_logical_operator,
    STATE(585), 1,
      sym_relational_operator,
    STATE(586), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2217), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 25,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [146334] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3384), 1,
      sym_exponentiate,
    STATE(491), 1,
      sym_adding_operator,
    STATE(492), 1,
      sym_multiplying_operator,
    STATE(584), 1,
      sym_logical_operator,
    STATE(585), 1,
      sym_relational_operator,
    STATE(586), 1,
      sym_shift_operator,
    STATE(2218), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [146404] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2219), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [146462] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2220), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3444), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [146518] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2221), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3446), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [146574] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2222), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2388), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2386), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [146632] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2223), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2396), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2394), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [146690] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2224), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 33,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [146748] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2225), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3448), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [146804] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2226), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2458), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2456), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [146862] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2227), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 33,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [146920] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2228), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3450), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [146976] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2229), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3452), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [147032] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2230), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3454), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [147088] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(497), 1,
      sym_adding_operator,
    STATE(498), 1,
      sym_multiplying_operator,
    STATE(588), 1,
      sym_logical_operator,
    STATE(589), 1,
      sym_relational_operator,
    STATE(590), 1,
      sym_shift_operator,
    STATE(2231), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [147156] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3392), 1,
      sym_exponentiate,
    STATE(497), 1,
      sym_adding_operator,
    STATE(498), 1,
      sym_multiplying_operator,
    STATE(588), 1,
      sym_logical_operator,
    STATE(589), 1,
      sym_relational_operator,
    STATE(590), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2232), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 25,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [147230] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2233), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2494), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2492), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [147288] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2234), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2510), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2508), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [147346] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2235), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2514), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2512), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [147404] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2236), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2518), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2516), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [147462] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2237), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2522), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2520), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [147520] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2238), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2526), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2524), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [147578] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(497), 1,
      sym_adding_operator,
    STATE(498), 1,
      sym_multiplying_operator,
    STATE(588), 1,
      sym_logical_operator,
    STATE(589), 1,
      sym_relational_operator,
    STATE(590), 1,
      sym_shift_operator,
    STATE(2239), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [147646] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3392), 1,
      sym_exponentiate,
    STATE(497), 1,
      sym_adding_operator,
    STATE(498), 1,
      sym_multiplying_operator,
    STATE(588), 1,
      sym_logical_operator,
    STATE(589), 1,
      sym_relational_operator,
    STATE(590), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2240), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 25,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [147720] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3392), 1,
      sym_exponentiate,
    STATE(497), 1,
      sym_adding_operator,
    STATE(498), 1,
      sym_multiplying_operator,
    STATE(588), 1,
      sym_logical_operator,
    STATE(589), 1,
      sym_relational_operator,
    STATE(590), 1,
      sym_shift_operator,
    STATE(2241), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [147790] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2242), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3456), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [147846] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2243), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2530), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2528), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [147904] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2244), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2554), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2552), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [147962] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2245), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2558), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2556), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [148020] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2246), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2562), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2560), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [148078] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2247), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2566), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2564), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [148136] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2248), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2570), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2568), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [148194] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2249), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3458), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [148250] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2250), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3460), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [148306] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2251), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3462), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [148362] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2252), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3464), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [148418] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2253), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2586), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2584), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [148476] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2254), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [148534] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2255), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2612), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2610), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [148592] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2256), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2616), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2614), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [148650] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2257), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3466), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [148706] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2258), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3468), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [148762] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2259), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3470), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [148818] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2260), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3472), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [148874] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2261), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3474), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [148930] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2262), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3476), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [148986] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2263), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3478), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [149042] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2264), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3480), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [149098] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2265), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2638), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2636), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149156] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2266), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2642), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2640), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149214] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2267), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2646), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2644), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149272] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2268), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2650), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2648), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149330] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2269), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2654), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2652), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149388] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2270), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3482), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [149444] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2271), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 33,
      sym_AND,
      sym_GENERIC,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149502] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2272), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3484), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [149558] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2273), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 33,
      sym_AND,
      sym_GENERATE,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149616] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2274), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3486), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [149672] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    ACTIONS(3272), 1,
      sym_WHEN,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    STATE(6207), 1,
      sym_when_element,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2275), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [149756] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2276), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2590), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2588), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149814] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2277), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149872] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2278), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 33,
      sym_AND,
      sym_GENERIC,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149930] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2279), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [149988] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2280), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2676), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2674), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [150046] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2281), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2442), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2440), 33,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [150104] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2282), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2680), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2678), 33,
      sym_AND,
      sym_GENERIC,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_PARAMETER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [150162] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3490), 1,
      sym_library_constant_unit,
    STATE(2708), 1,
      sym__unit,
    STATE(2283), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3488), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [150226] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2914), 1,
      sym_exponentiate,
    STATE(689), 1,
      sym_logical_operator,
    STATE(690), 1,
      sym_relational_operator,
    STATE(691), 1,
      sym_shift_operator,
    STATE(701), 1,
      sym_adding_operator,
    STATE(702), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(3492), 2,
      sym_WHEN,
      anon_sym_SEMI,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2284), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [150308] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2285), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3494), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [150364] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2286), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3496), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [150420] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2287), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3498), 39,
      sym_identifier,
      sym_ALIAS,
      sym_ARCHITECTURE,
      sym_ASSERT,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_CONTEXT,
      sym_DISCONNECT,
      sym_ENTITY,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_LIBRARY,
      sym_PACKAGE,
      sym_POSTPONED,
      sym_PROCEDURE,
      sym_PROCESS,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      sym__end_of_file,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [150476] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2288), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 32,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [150534] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2289), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [150591] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2290), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [150648] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    ACTIONS(3276), 1,
      anon_sym_SEMI,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2291), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [150729] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3500), 1,
      sym_exponentiate,
    STATE(446), 1,
      sym_adding_operator,
    STATE(447), 1,
      sym_multiplying_operator,
    STATE(576), 1,
      sym_logical_operator,
    STATE(577), 1,
      sym_relational_operator,
    STATE(578), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2292), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2708), 7,
      sym_AND,
      sym_INERTIAL,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [150808] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3500), 1,
      sym_exponentiate,
    STATE(446), 1,
      sym_adding_operator,
    STATE(447), 1,
      sym_multiplying_operator,
    STATE(576), 1,
      sym_logical_operator,
    STATE(577), 1,
      sym_relational_operator,
    STATE(578), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2293), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 15,
      sym_AND,
      sym_INERTIAL,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [150885] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3500), 1,
      sym_exponentiate,
    STATE(446), 1,
      sym_adding_operator,
    STATE(447), 1,
      sym_multiplying_operator,
    STATE(576), 1,
      sym_logical_operator,
    STATE(577), 1,
      sym_relational_operator,
    STATE(578), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2294), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 21,
      sym_AND,
      sym_INERTIAL,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [150960] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    ACTIONS(3502), 1,
      sym_GENERATE,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2295), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [151041] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2296), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 32,
      sym_identifier,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [151098] = 24,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(3402), 1,
      sym_ALIAS,
    ACTIONS(3404), 1,
      sym_ATTRIBUTE,
    ACTIONS(3406), 1,
      sym_CONSTANT,
    ACTIONS(3410), 1,
      sym_FILE,
    ACTIONS(3412), 1,
      sym_FUNCTION,
    ACTIONS(3414), 1,
      sym_GROUP,
    ACTIONS(3416), 1,
      sym_PACKAGE,
    ACTIONS(3418), 1,
      sym_SHARED,
    ACTIONS(3420), 1,
      sym_SUBTYPE,
    ACTIONS(3422), 1,
      sym_TYPE,
    ACTIONS(3424), 1,
      sym_USE,
    ACTIONS(3426), 1,
      sym_VARIABLE,
    ACTIONS(3504), 1,
      sym_END,
    STATE(2406), 1,
      aux_sym_package_definition_body_repeat1,
    STATE(3786), 1,
      sym__package_body_declarative_item,
    STATE(5901), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2297), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3785), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [151191] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2298), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 32,
      sym_identifier,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [151248] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(446), 1,
      sym_adding_operator,
    STATE(447), 1,
      sym_multiplying_operator,
    STATE(576), 1,
      sym_logical_operator,
    STATE(577), 1,
      sym_relational_operator,
    STATE(578), 1,
      sym_shift_operator,
    STATE(2299), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [151315] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3500), 1,
      sym_exponentiate,
    STATE(446), 1,
      sym_adding_operator,
    STATE(447), 1,
      sym_multiplying_operator,
    STATE(576), 1,
      sym_logical_operator,
    STATE(577), 1,
      sym_relational_operator,
    STATE(578), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2300), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 24,
      sym_AND,
      sym_INERTIAL,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [151388] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2301), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [151445] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3500), 1,
      sym_exponentiate,
    STATE(446), 1,
      sym_adding_operator,
    STATE(447), 1,
      sym_multiplying_operator,
    STATE(576), 1,
      sym_logical_operator,
    STATE(577), 1,
      sym_relational_operator,
    STATE(578), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2302), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 24,
      sym_AND,
      sym_INERTIAL,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [151518] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3500), 1,
      sym_exponentiate,
    STATE(446), 1,
      sym_adding_operator,
    STATE(447), 1,
      sym_multiplying_operator,
    STATE(576), 1,
      sym_logical_operator,
    STATE(577), 1,
      sym_relational_operator,
    STATE(578), 1,
      sym_shift_operator,
    STATE(2303), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 26,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [151587] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2304), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 32,
      sym_identifier,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [151644] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3506), 1,
      sym_exponentiate,
    ACTIONS(3508), 1,
      sym_THEN,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2305), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [151725] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2306), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [151782] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2307), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 32,
      sym_identifier,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [151839] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2690), 1,
      sym_IS,
    ACTIONS(3510), 1,
      sym_exponentiate,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2308), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [151920] = 24,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3512), 1,
      sym_BEGIN,
    STATE(2398), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2309), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [152013] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    STATE(2310), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [152080] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3510), 1,
      sym_exponentiate,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2311), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 24,
      sym_AND,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [152153] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3510), 1,
      sym_exponentiate,
    ACTIONS(3514), 1,
      sym_IS,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2312), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [152234] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2313), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [152291] = 24,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3516), 1,
      sym_BEGIN,
    STATE(2386), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2314), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [152384] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(554), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(2315), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [152445] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    STATE(2316), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [152512] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2317), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [152569] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2318), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [152626] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2319), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2708), 7,
      sym_AND,
      sym_GENERATE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [152705] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    STATE(2320), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [152772] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3510), 1,
      sym_exponentiate,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2321), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2708), 7,
      sym_AND,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [152851] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    ACTIONS(3518), 1,
      anon_sym_SEMI,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2322), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [152932] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3510), 1,
      sym_exponentiate,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2323), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 15,
      sym_AND,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [153009] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3510), 1,
      sym_exponentiate,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2324), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 21,
      sym_AND,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [153084] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3510), 1,
      sym_exponentiate,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2325), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 24,
      sym_AND,
      sym_IS,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [153157] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    ACTIONS(3520), 1,
      anon_sym_SEMI,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2326), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [153238] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3510), 1,
      sym_exponentiate,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    STATE(2327), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 26,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [153307] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2328), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 15,
      sym_AND,
      sym_GENERATE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [153384] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2329), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 21,
      sym_AND,
      sym_GENERATE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [153459] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    ACTIONS(3522), 1,
      anon_sym_SEMI,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2330), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [153540] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2331), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 24,
      sym_AND,
      sym_GENERATE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [153613] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2830), 1,
      sym_LOOP,
    ACTIONS(3192), 1,
      sym_exponentiate,
    STATE(481), 1,
      sym_adding_operator,
    STATE(482), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2332), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [153694] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    ACTIONS(3524), 1,
      anon_sym_SEMI,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2333), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [153775] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    STATE(2334), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 26,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [153844] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    ACTIONS(3526), 1,
      anon_sym_SEMI,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2335), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [153925] = 24,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3516), 1,
      sym_BEGIN,
    STATE(2398), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2336), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [154018] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2337), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 32,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [154075] = 23,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3528), 1,
      sym_ALIAS,
    ACTIONS(3531), 1,
      sym_ATTRIBUTE,
    ACTIONS(3534), 1,
      sym_BEGIN,
    ACTIONS(3536), 1,
      sym_CONSTANT,
    ACTIONS(3539), 1,
      sym_FILE,
    ACTIONS(3542), 1,
      sym_FUNCTION,
    ACTIONS(3545), 1,
      sym_GROUP,
    ACTIONS(3551), 1,
      sym_PACKAGE,
    ACTIONS(3554), 1,
      sym_PROCEDURE,
    ACTIONS(3557), 1,
      sym_SHARED,
    ACTIONS(3560), 1,
      sym_SUBTYPE,
    ACTIONS(3563), 1,
      sym_TYPE,
    ACTIONS(3566), 1,
      sym_USE,
    ACTIONS(3569), 1,
      sym_VARIABLE,
    STATE(3700), 1,
      sym__subprogram_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    ACTIONS(3548), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2338), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_subprogram_head_repeat1,
    STATE(3686), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [154166] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3500), 1,
      sym_exponentiate,
    ACTIONS(3572), 1,
      sym_INERTIAL,
    STATE(446), 1,
      sym_adding_operator,
    STATE(447), 1,
      sym_multiplying_operator,
    STATE(576), 1,
      sym_logical_operator,
    STATE(577), 1,
      sym_relational_operator,
    STATE(578), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2339), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [154247] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2340), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [154304] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2341), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 32,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [154361] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2690), 1,
      sym_THEN,
    ACTIONS(3506), 1,
      sym_exponentiate,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2342), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [154442] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    STATE(2343), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [154509] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3506), 1,
      sym_exponentiate,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2344), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 24,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [154582] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(569), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(2345), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [154643] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    STATE(2346), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [154710] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3506), 1,
      sym_exponentiate,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2347), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2708), 7,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [154789] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3506), 1,
      sym_exponentiate,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2348), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 15,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [154866] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3506), 1,
      sym_exponentiate,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2349), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 21,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [154941] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3506), 1,
      sym_exponentiate,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2350), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 24,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [155014] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3506), 1,
      sym_exponentiate,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    STATE(2351), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [155083] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(651), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(2352), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [155144] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3574), 1,
      anon_sym_EQ_GT,
    STATE(2353), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [155203] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2354), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 32,
      sym_identifier,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [155260] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2355), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 32,
      sym_identifier,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [155317] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2690), 1,
      sym_LOOP,
    ACTIONS(3576), 1,
      sym_exponentiate,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2356), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [155398] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2357), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 32,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [155455] = 24,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3578), 1,
      sym_BEGIN,
    STATE(2338), 1,
      aux_sym_subprogram_head_repeat1,
    STATE(3700), 1,
      sym__subprogram_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2358), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3686), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [155548] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    STATE(2359), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [155615] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3576), 1,
      sym_exponentiate,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2360), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 24,
      sym_AND,
      sym_LOOP,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [155688] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    STATE(2361), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [155755] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3576), 1,
      sym_exponentiate,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2362), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2708), 7,
      sym_AND,
      sym_LOOP,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [155834] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3576), 1,
      sym_exponentiate,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2363), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 15,
      sym_AND,
      sym_LOOP,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [155911] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3576), 1,
      sym_exponentiate,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2364), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 21,
      sym_AND,
      sym_LOOP,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [155986] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3576), 1,
      sym_exponentiate,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2365), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 24,
      sym_AND,
      sym_LOOP,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [156059] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2366), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [156116] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2367), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [156173] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2368), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [156230] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2369), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [156287] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3576), 1,
      sym_exponentiate,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    STATE(2370), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 26,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [156356] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2371), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [156413] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2372), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [156470] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2373), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [156527] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2374), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [156584] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2375), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [156641] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2376), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [156698] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    ACTIONS(3580), 1,
      anon_sym_RPAREN,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2377), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [156779] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2378), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 24,
      sym_AND,
      sym_GENERATE,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [156852] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    ACTIONS(3582), 1,
      anon_sym_RPAREN,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2379), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [156933] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2380), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [156990] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    ACTIONS(3584), 1,
      anon_sym_SEMI,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2381), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [157071] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3586), 1,
      sym_exponentiate,
    ACTIONS(3588), 1,
      sym_SELECT,
    STATE(676), 1,
      sym_logical_operator,
    STATE(677), 1,
      sym_relational_operator,
    STATE(678), 1,
      sym_shift_operator,
    STATE(679), 1,
      sym_adding_operator,
    STATE(680), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2382), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [157152] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2383), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [157209] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2384), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [157266] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2385), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [157323] = 24,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3590), 1,
      sym_BEGIN,
    STATE(2398), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2386), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [157416] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2387), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [157473] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3576), 1,
      sym_exponentiate,
    ACTIONS(3592), 1,
      sym_LOOP,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(566), 1,
      sym_logical_operator,
    STATE(567), 1,
      sym_relational_operator,
    STATE(568), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2388), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [157554] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2389), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [157611] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2390), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [157668] = 24,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(3402), 1,
      sym_ALIAS,
    ACTIONS(3404), 1,
      sym_ATTRIBUTE,
    ACTIONS(3406), 1,
      sym_CONSTANT,
    ACTIONS(3410), 1,
      sym_FILE,
    ACTIONS(3412), 1,
      sym_FUNCTION,
    ACTIONS(3414), 1,
      sym_GROUP,
    ACTIONS(3416), 1,
      sym_PACKAGE,
    ACTIONS(3418), 1,
      sym_SHARED,
    ACTIONS(3420), 1,
      sym_SUBTYPE,
    ACTIONS(3422), 1,
      sym_TYPE,
    ACTIONS(3424), 1,
      sym_USE,
    ACTIONS(3426), 1,
      sym_VARIABLE,
    ACTIONS(3594), 1,
      sym_END,
    STATE(2297), 1,
      aux_sym_package_definition_body_repeat1,
    STATE(3786), 1,
      sym__package_body_declarative_item,
    STATE(5901), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2391), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3785), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [157761] = 23,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3596), 1,
      sym_ALIAS,
    ACTIONS(3599), 1,
      sym_ATTRIBUTE,
    ACTIONS(3602), 1,
      sym_CONSTANT,
    ACTIONS(3605), 1,
      sym_END,
    ACTIONS(3607), 1,
      sym_FILE,
    ACTIONS(3610), 1,
      sym_FUNCTION,
    ACTIONS(3613), 1,
      sym_GROUP,
    ACTIONS(3619), 1,
      sym_PACKAGE,
    ACTIONS(3622), 1,
      sym_PROCEDURE,
    ACTIONS(3625), 1,
      sym_SHARED,
    ACTIONS(3628), 1,
      sym_SUBTYPE,
    ACTIONS(3631), 1,
      sym_TYPE,
    ACTIONS(3634), 1,
      sym_USE,
    ACTIONS(3637), 1,
      sym_VARIABLE,
    STATE(3692), 1,
      sym__protected_type_body_declarative_item,
    STATE(5901), 1,
      sym__subprogram_specification,
    ACTIONS(3616), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2392), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_protected_type_body_repeat1,
    STATE(3687), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [157852] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2393), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 32,
      sym_identifier,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [157909] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2690), 1,
      sym_SELECT,
    ACTIONS(3586), 1,
      sym_exponentiate,
    STATE(676), 1,
      sym_logical_operator,
    STATE(677), 1,
      sym_relational_operator,
    STATE(678), 1,
      sym_shift_operator,
    STATE(679), 1,
      sym_adding_operator,
    STATE(680), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2394), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [157990] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2395), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [158047] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2396), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [158104] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2397), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [158161] = 23,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3640), 1,
      sym_ALIAS,
    ACTIONS(3643), 1,
      sym_ATTRIBUTE,
    ACTIONS(3646), 1,
      sym_BEGIN,
    ACTIONS(3648), 1,
      sym_CONSTANT,
    ACTIONS(3651), 1,
      sym_FILE,
    ACTIONS(3654), 1,
      sym_FUNCTION,
    ACTIONS(3657), 1,
      sym_GROUP,
    ACTIONS(3663), 1,
      sym_PACKAGE,
    ACTIONS(3666), 1,
      sym_PROCEDURE,
    ACTIONS(3669), 1,
      sym_SHARED,
    ACTIONS(3672), 1,
      sym_SUBTYPE,
    ACTIONS(3675), 1,
      sym_TYPE,
    ACTIONS(3678), 1,
      sym_USE,
    ACTIONS(3681), 1,
      sym_VARIABLE,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    ACTIONS(3660), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2398), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_sequential_block_head_repeat1,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [158252] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2399), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [158309] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    ACTIONS(3684), 1,
      anon_sym_SEMI,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2400), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [158390] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3510), 1,
      sym_exponentiate,
    ACTIONS(3686), 1,
      sym_IS,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2401), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [158471] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(676), 1,
      sym_logical_operator,
    STATE(677), 1,
      sym_relational_operator,
    STATE(678), 1,
      sym_shift_operator,
    STATE(679), 1,
      sym_adding_operator,
    STATE(680), 1,
      sym_multiplying_operator,
    STATE(2402), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [158538] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3586), 1,
      sym_exponentiate,
    STATE(676), 1,
      sym_logical_operator,
    STATE(677), 1,
      sym_relational_operator,
    STATE(678), 1,
      sym_shift_operator,
    STATE(679), 1,
      sym_adding_operator,
    STATE(680), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2403), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2718), 24,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [158611] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2404), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [158668] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2405), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 31,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
      anon_sym_PIPE,
  [158725] = 23,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3688), 1,
      sym_ALIAS,
    ACTIONS(3691), 1,
      sym_ATTRIBUTE,
    ACTIONS(3694), 1,
      sym_CONSTANT,
    ACTIONS(3697), 1,
      sym_END,
    ACTIONS(3699), 1,
      sym_FILE,
    ACTIONS(3702), 1,
      sym_FUNCTION,
    ACTIONS(3705), 1,
      sym_GROUP,
    ACTIONS(3711), 1,
      sym_PACKAGE,
    ACTIONS(3714), 1,
      sym_PROCEDURE,
    ACTIONS(3717), 1,
      sym_SHARED,
    ACTIONS(3720), 1,
      sym_SUBTYPE,
    ACTIONS(3723), 1,
      sym_TYPE,
    ACTIONS(3726), 1,
      sym_USE,
    ACTIONS(3729), 1,
      sym_VARIABLE,
    STATE(3786), 1,
      sym__package_body_declarative_item,
    STATE(5901), 1,
      sym__subprogram_specification,
    ACTIONS(3708), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2406), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_package_definition_body_repeat1,
    STATE(3785), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [158816] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2830), 1,
      sym_GENERATE,
    ACTIONS(3188), 1,
      sym_exponentiate,
    STATE(466), 1,
      sym_adding_operator,
    STATE(467), 1,
      sym_multiplying_operator,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2407), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [158897] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2408), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [158954] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2409), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [159011] = 24,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3732), 1,
      sym_BEGIN,
    STATE(2309), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2410), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [159104] = 24,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3732), 1,
      sym_BEGIN,
    STATE(2398), 1,
      aux_sym_sequential_block_head_repeat1,
    STATE(3793), 1,
      sym__process_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2411), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3792), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [159197] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    STATE(2412), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2720), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2718), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [159264] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3510), 1,
      sym_exponentiate,
    ACTIONS(3734), 1,
      sym_IS,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2413), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [159345] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(676), 1,
      sym_logical_operator,
    STATE(677), 1,
      sym_relational_operator,
    STATE(678), 1,
      sym_shift_operator,
    STATE(679), 1,
      sym_adding_operator,
    STATE(680), 1,
      sym_multiplying_operator,
    STATE(2414), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [159412] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3586), 1,
      sym_exponentiate,
    STATE(676), 1,
      sym_logical_operator,
    STATE(677), 1,
      sym_relational_operator,
    STATE(678), 1,
      sym_shift_operator,
    STATE(679), 1,
      sym_adding_operator,
    STATE(680), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2415), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2708), 7,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_SELECT,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [159491] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3506), 1,
      sym_exponentiate,
    ACTIONS(3736), 1,
      sym_THEN,
    STATE(426), 1,
      sym_adding_operator,
    STATE(427), 1,
      sym_multiplying_operator,
    STATE(561), 1,
      sym_logical_operator,
    STATE(562), 1,
      sym_relational_operator,
    STATE(563), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2416), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [159572] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2417), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [159629] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2830), 1,
      sym_IS,
    ACTIONS(3184), 1,
      sym_exponentiate,
    STATE(471), 1,
      sym_adding_operator,
    STATE(472), 1,
      sym_multiplying_operator,
    STATE(556), 1,
      sym_logical_operator,
    STATE(557), 1,
      sym_relational_operator,
    STATE(558), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2418), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [159710] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2690), 1,
      sym_INERTIAL,
    ACTIONS(3500), 1,
      sym_exponentiate,
    STATE(446), 1,
      sym_adding_operator,
    STATE(447), 1,
      sym_multiplying_operator,
    STATE(576), 1,
      sym_logical_operator,
    STATE(577), 1,
      sym_relational_operator,
    STATE(578), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2419), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [159791] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3586), 1,
      sym_exponentiate,
    STATE(676), 1,
      sym_logical_operator,
    STATE(677), 1,
      sym_relational_operator,
    STATE(678), 1,
      sym_shift_operator,
    STATE(679), 1,
      sym_adding_operator,
    STATE(680), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2420), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2724), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2722), 15,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_SELECT,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [159868] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2421), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [159925] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2422), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2716), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2714), 32,
      sym_identifier,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [159982] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3586), 1,
      sym_exponentiate,
    STATE(676), 1,
      sym_logical_operator,
    STATE(677), 1,
      sym_relational_operator,
    STATE(678), 1,
      sym_shift_operator,
    STATE(679), 1,
      sym_adding_operator,
    STATE(680), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2423), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2728), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2726), 21,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [160057] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3586), 1,
      sym_exponentiate,
    STATE(676), 1,
      sym_logical_operator,
    STATE(677), 1,
      sym_relational_operator,
    STATE(678), 1,
      sym_shift_operator,
    STATE(679), 1,
      sym_adding_operator,
    STATE(680), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    STATE(2424), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2710), 24,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [160130] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3586), 1,
      sym_exponentiate,
    STATE(676), 1,
      sym_logical_operator,
    STATE(677), 1,
      sym_relational_operator,
    STATE(678), 1,
      sym_shift_operator,
    STATE(679), 1,
      sym_adding_operator,
    STATE(680), 1,
      sym_multiplying_operator,
    STATE(2425), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
  [160199] = 24,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(51), 1,
      sym_PROCEDURE,
    ACTIONS(2940), 1,
      sym_ALIAS,
    ACTIONS(2942), 1,
      sym_ATTRIBUTE,
    ACTIONS(2946), 1,
      sym_CONSTANT,
    ACTIONS(2948), 1,
      sym_FILE,
    ACTIONS(2950), 1,
      sym_FUNCTION,
    ACTIONS(2952), 1,
      sym_GROUP,
    ACTIONS(2956), 1,
      sym_PACKAGE,
    ACTIONS(2958), 1,
      sym_SHARED,
    ACTIONS(2960), 1,
      sym_SUBTYPE,
    ACTIONS(2962), 1,
      sym_TYPE,
    ACTIONS(2964), 1,
      sym_USE,
    ACTIONS(2966), 1,
      sym_VARIABLE,
    ACTIONS(3738), 1,
      sym_BEGIN,
    STATE(2358), 1,
      aux_sym_subprogram_head_repeat1,
    STATE(3700), 1,
      sym__subprogram_declarative_item,
    STATE(5844), 1,
      sym__subprogram_specification,
    ACTIONS(43), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6069), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2426), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3686), 17,
      sym_use_clause,
      sym_package_declaration,
      sym_package_instantiation_declaration,
      sym_package_definition,
      sym_subprogram_declaration,
      sym_subprogram_definition,
      sym_subprogram_instantiation_declaration,
      sym_type_declaration,
      sym_subtype_declaration,
      sym_constant_declaration,
      sym_variable_declaration,
      sym_file_declaration,
      sym_alias_declaration,
      sym_attribute_declaration,
      sym_attribute_specification,
      sym_group_template_declaration,
      sym_group_declaration,
  [160292] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2427), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2688), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2686), 32,
      sym_identifier,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      sym_library_constant,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [160349] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2690), 1,
      sym_GENERATE,
    ACTIONS(3240), 1,
      sym_exponentiate,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2428), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [160430] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2429), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 32,
      sym_AND,
      sym_BUS,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [160487] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    ACTIONS(3740), 1,
      sym_GENERATE,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2430), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [160568] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    ACTIONS(3742), 1,
      sym_GENERATE,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2431), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [160649] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    ACTIONS(3744), 1,
      sym_GENERATE,
    STATE(540), 1,
      sym_logical_operator,
    STATE(541), 1,
      sym_relational_operator,
    STATE(542), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2432), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [160730] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(446), 1,
      sym_adding_operator,
    STATE(447), 1,
      sym_multiplying_operator,
    STATE(576), 1,
      sym_logical_operator,
    STATE(577), 1,
      sym_relational_operator,
    STATE(578), 1,
      sym_shift_operator,
    STATE(2433), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2712), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2710), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [160797] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2801), 1,
      sym_exponentiate,
    STATE(416), 1,
      sym_adding_operator,
    STATE(417), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2434), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [160875] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3032), 1,
      sym_exponentiate,
    STATE(512), 1,
      sym_adding_operator,
    STATE(513), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2435), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [160953] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2436), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161009] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3168), 1,
      sym_exponentiate,
    STATE(517), 1,
      sym_adding_operator,
    STATE(518), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2437), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [161087] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2438), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161143] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2439), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161199] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2440), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161255] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3036), 1,
      sym_exponentiate,
    STATE(527), 1,
      sym_adding_operator,
    STATE(528), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2441), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [161333] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2442), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161389] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(538), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(2443), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161449] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2444), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161505] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2445), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161561] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2446), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161617] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2447), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161673] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2448), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161729] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3510), 1,
      sym_exponentiate,
    STATE(421), 1,
      sym_adding_operator,
    STATE(422), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2449), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [161807] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3576), 1,
      sym_exponentiate,
    STATE(431), 1,
      sym_adding_operator,
    STATE(432), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2450), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [161885] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2451), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161941] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2452), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [161997] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2453), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162053] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2454), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162109] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2455), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162165] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2456), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162221] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2457), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162277] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2458), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162333] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2459), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162389] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2460), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162445] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2461), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162501] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2462), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162557] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(531), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(2463), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162617] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3384), 1,
      sym_exponentiate,
    STATE(491), 1,
      sym_adding_operator,
    STATE(492), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2464), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [162695] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3186), 1,
      sym_exponentiate,
    STATE(507), 1,
      sym_adding_operator,
    STATE(508), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2465), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [162773] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3392), 1,
      sym_exponentiate,
    STATE(497), 1,
      sym_adding_operator,
    STATE(498), 1,
      sym_multiplying_operator,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2466), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [162851] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2467), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 31,
      sym_AND,
      sym_DOWNTO,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [162907] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3240), 1,
      sym_exponentiate,
    STATE(551), 1,
      sym_logical_operator,
    STATE(552), 1,
      sym_relational_operator,
    STATE(553), 1,
      sym_shift_operator,
    STATE(712), 1,
      sym_adding_operator,
    STATE(713), 1,
      sym_multiplying_operator,
    ACTIONS(2698), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2704), 2,
      sym_MOD,
      sym_REM,
    ACTIONS(2696), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
    STATE(2468), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2694), 4,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
    ACTIONS(2702), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
    ACTIONS(2706), 6,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
    ACTIONS(2692), 8,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
  [162985] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3216), 1,
      sym_RANGE,
    STATE(2469), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163043] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2470), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 31,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163099] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2471), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163154] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2472), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163209] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2473), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163264] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2474), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163319] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2475), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163374] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2476), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163429] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2477), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163484] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2478), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163539] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2479), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163594] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2480), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163649] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2481), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163704] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2482), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163759] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2483), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163814] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2484), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163869] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2485), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163924] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(559), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(2486), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [163983] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2487), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164038] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2488), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164093] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(544), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(2489), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164152] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2490), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164207] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2491), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164262] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2492), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164317] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2493), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164372] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2494), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164427] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2495), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164482] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2496), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164537] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2497), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164592] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2498), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3286), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164649] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2499), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164704] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2500), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164759] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2501), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164814] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2502), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164869] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2503), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164924] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3574), 1,
      anon_sym_EQ_GT,
    STATE(2504), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 7,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [164981] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2505), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165036] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2506), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165091] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2507), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165146] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2508), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165201] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2509), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165256] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2510), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165311] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2511), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165366] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2512), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165421] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2513), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165476] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2514), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165531] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2515), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165586] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2516), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165641] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2517), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165696] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2518), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165751] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2519), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165806] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(549), 1,
      sym__direction,
    ACTIONS(3194), 2,
      sym_DOWNTO,
      sym_TO,
    STATE(2520), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165865] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2521), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165920] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2522), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [165975] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2523), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166030] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2524), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166085] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2525), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166140] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2526), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166195] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2527), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REGISTER,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166250] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2528), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166305] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2529), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166360] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2530), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166415] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2531), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166470] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2532), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166525] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2533), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 30,
      sym_AND,
      sym_BUS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166580] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2534), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3286), 4,
      sym_BUS,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166637] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2535), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 30,
      sym_AND,
      sym_DOWNTO,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166692] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2536), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 30,
      sym_AFTER,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_WHEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166747] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2537), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166801] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2538), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3746), 35,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_UNAFFECTED,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [166853] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2539), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166907] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2540), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [166961] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2541), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167015] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2542), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3748), 35,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_UNAFFECTED,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [167067] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2543), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167121] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2544), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167175] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2545), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167229] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2546), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167283] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2547), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167337] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2548), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167391] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2549), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167445] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2550), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167499] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2551), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167553] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2552), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167607] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2553), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167661] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2554), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167715] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2555), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167769] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2556), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167823] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2557), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167877] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2558), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167931] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2559), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [167985] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2560), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168039] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2561), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168093] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2562), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168147] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2563), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168201] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2564), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168255] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2565), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168309] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2566), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168363] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2567), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168417] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2568), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168471] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2569), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168525] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2570), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168579] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2571), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168633] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2572), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_DOWNTO,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168687] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2573), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168741] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2574), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168795] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3286), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(2575), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168851] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2576), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168905] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2577), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [168959] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2578), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2579), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3750), 35,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_OTHERS,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [169065] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2580), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169119] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2581), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169173] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2582), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169227] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2583), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169281] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2584), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169335] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2585), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169389] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2586), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169443] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2587), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169497] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2588), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169551] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2589), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169605] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2590), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169659] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2591), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169713] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2592), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169767] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2593), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3752), 35,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_UNAFFECTED,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [169819] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2594), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 29,
      sym_AND,
      sym_DOWNTO,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169873] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2595), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169927] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2596), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [169981] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2597), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170035] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2598), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170089] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2599), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170143] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2600), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170197] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2601), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170251] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2602), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170305] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2603), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170359] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2604), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170413] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2605), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170467] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2606), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170521] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2607), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 29,
      sym_AND,
      sym_GENERIC,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170575] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2608), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170629] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3286), 3,
      sym_IS,
      sym_OPEN,
      anon_sym_SEMI,
    STATE(2609), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170685] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2610), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170739] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2611), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 29,
      sym_AND,
      sym_DOWNTO,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_TO,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170793] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2612), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170847] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2613), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3754), 35,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_UNAFFECTED,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [170899] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2614), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 29,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_REPORT,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [170953] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2615), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 29,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OPEN,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171007] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2616), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171060] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2617), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3756), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [171111] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2618), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171164] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2619), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3758), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [171215] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2620), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3752), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [171266] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2621), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171319] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2622), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171372] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2623), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171425] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2624), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171478] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2625), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171531] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(3760), 1,
      anon_sym_LT_LT,
    ACTIONS(3764), 1,
      sym_BLOCK,
    ACTIONS(3766), 1,
      sym_CASE,
    ACTIONS(3768), 1,
      sym_COMPONENT,
    ACTIONS(3770), 1,
      sym_CONFIGURATION,
    ACTIONS(3772), 1,
      sym_ENTITY,
    ACTIONS(3774), 1,
      sym_FOR,
    ACTIONS(3776), 1,
      sym_IF,
    ACTIONS(3778), 1,
      sym_POSTPONED,
    ACTIONS(3780), 1,
      sym_PROCESS,
    STATE(3636), 1,
      sym__direct_name,
    STATE(5081), 1,
      sym_name,
    STATE(5178), 1,
      sym_instantiated_unit,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6376), 1,
      sym__target,
    STATE(6380), 1,
      sym_for_loop,
    STATE(6471), 1,
      sym_with_expression,
    STATE(6545), 1,
      sym_if_generate,
    STATE(7462), 1,
      sym_assertion,
    STATE(4717), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3782), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2626), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4725), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3762), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [171632] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2627), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171685] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2628), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171738] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2629), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3784), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [171789] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2630), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3750), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [171840] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2631), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171893] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2632), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171946] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2633), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [171999] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2634), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172052] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2635), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172105] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(3760), 1,
      anon_sym_LT_LT,
    ACTIONS(3768), 1,
      sym_COMPONENT,
    ACTIONS(3770), 1,
      sym_CONFIGURATION,
    ACTIONS(3772), 1,
      sym_ENTITY,
    ACTIONS(3774), 1,
      sym_FOR,
    ACTIONS(3776), 1,
      sym_IF,
    ACTIONS(3786), 1,
      sym_BLOCK,
    ACTIONS(3788), 1,
      sym_CASE,
    ACTIONS(3790), 1,
      sym_POSTPONED,
    ACTIONS(3792), 1,
      sym_PROCESS,
    STATE(3636), 1,
      sym__direct_name,
    STATE(5122), 1,
      sym_name,
    STATE(5212), 1,
      sym_instantiated_unit,
    STATE(6089), 1,
      sym_if_generate,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6438), 1,
      sym__target,
    STATE(6443), 1,
      sym_for_loop,
    STATE(6495), 1,
      sym_with_expression,
    STATE(6711), 1,
      sym_assertion,
    STATE(4717), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3782), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2636), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4725), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3762), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [172206] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2637), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3794), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [172257] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2638), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3796), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [172308] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2639), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172361] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2640), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172414] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2641), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172467] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2642), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172520] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2643), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172573] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2644), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172626] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2645), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3798), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [172677] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2646), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172730] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2647), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172783] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2648), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172836] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2649), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172889] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2650), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172942] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2651), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [172995] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2652), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173048] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2653), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173101] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2654), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173154] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2655), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173207] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2656), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173260] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2657), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173313] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(3760), 1,
      anon_sym_LT_LT,
    ACTIONS(3768), 1,
      sym_COMPONENT,
    ACTIONS(3770), 1,
      sym_CONFIGURATION,
    ACTIONS(3772), 1,
      sym_ENTITY,
    ACTIONS(3774), 1,
      sym_FOR,
    ACTIONS(3776), 1,
      sym_IF,
    ACTIONS(3800), 1,
      sym_BLOCK,
    ACTIONS(3802), 1,
      sym_CASE,
    ACTIONS(3804), 1,
      sym_POSTPONED,
    ACTIONS(3806), 1,
      sym_PROCESS,
    STATE(3636), 1,
      sym__direct_name,
    STATE(5120), 1,
      sym_name,
    STATE(5257), 1,
      sym_instantiated_unit,
    STATE(6019), 1,
      sym_if_generate,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6406), 1,
      sym__target,
    STATE(6411), 1,
      sym_for_loop,
    STATE(6482), 1,
      sym_with_expression,
    STATE(8355), 1,
      sym_assertion,
    STATE(4717), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3782), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2658), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4725), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3762), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [173414] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2659), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173467] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2660), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173520] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2661), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173573] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2662), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173626] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2663), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173679] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2664), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173732] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2665), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173785] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2666), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173838] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2667), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173891] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2668), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 28,
      sym_AND,
      sym_ELSE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173944] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2669), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [173997] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2670), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174050] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2671), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174103] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2672), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174156] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2673), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174209] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2674), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174262] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3286), 2,
      sym_OF,
      anon_sym_COMMA,
    STATE(2675), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174317] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2676), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3808), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [174368] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2677), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174421] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2678), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3810), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [174472] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2679), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174525] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2680), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174578] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2681), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174631] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2682), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174684] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2683), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 28,
      sym_AND,
      sym_FOR,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174737] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2684), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174790] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3286), 2,
      sym_UNITS,
      anon_sym_SEMI,
    STATE(2685), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174845] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2686), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_UNITS,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174898] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2687), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OF,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_COMMA,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [174951] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2688), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175004] = 30,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(3760), 1,
      anon_sym_LT_LT,
    ACTIONS(3768), 1,
      sym_COMPONENT,
    ACTIONS(3770), 1,
      sym_CONFIGURATION,
    ACTIONS(3772), 1,
      sym_ENTITY,
    ACTIONS(3774), 1,
      sym_FOR,
    ACTIONS(3776), 1,
      sym_IF,
    ACTIONS(3812), 1,
      sym_BLOCK,
    ACTIONS(3814), 1,
      sym_CASE,
    ACTIONS(3816), 1,
      sym_POSTPONED,
    ACTIONS(3818), 1,
      sym_PROCESS,
    STATE(3636), 1,
      sym__direct_name,
    STATE(5090), 1,
      sym_name,
    STATE(5207), 1,
      sym_instantiated_unit,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6243), 1,
      sym_with_expression,
    STATE(6303), 1,
      sym_if_generate,
    STATE(6308), 1,
      sym_for_loop,
    STATE(6309), 1,
      sym__target,
    STATE(8323), 1,
      sym_assertion,
    STATE(4717), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3782), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2689), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4725), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3762), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [175105] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2690), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175158] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2691), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3820), 34,
      sym_identifier,
      sym_ABS,
      sym_AND,
      sym_NAND,
      sym_NEW,
      sym_NOR,
      sym_NOT,
      sym_NULL,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      sym_decimal_integer,
      sym_decimal_float,
      sym_based_base,
      sym_character_literal,
      sym_string_literal,
      sym_string_literal_std_logic,
      sym_bit_string_length,
      sym_bit_string_base,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_boolean,
      sym_library_constant_character,
      sym_library_constant_debug,
      sym_library_constant_env,
      sym_library_constant_standard,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      sym_condition_conversion,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_LT_LT,
  [175209] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2692), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 28,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SEVERITY,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175262] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2693), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175314] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2694), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175366] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2695), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175418] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2696), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175470] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2697), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175522] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2698), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175574] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2699), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175626] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2700), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175678] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2701), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175730] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2702), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175782] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2703), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175834] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2704), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175886] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2705), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175938] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2706), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [175990] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3286), 1,
      sym_LOOP,
    STATE(2707), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176044] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2708), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176096] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2709), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176148] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2710), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176200] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2711), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176252] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2712), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176304] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2713), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176356] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(3824), 1,
      anon_sym_LT_LT,
    ACTIONS(3830), 1,
      sym_TYPE,
    ACTIONS(3832), 1,
      sym_VIEW,
    STATE(2806), 1,
      sym__direct_name,
    STATE(3112), 1,
      sym_mode,
    STATE(4177), 1,
      sym_resolution_indication,
    STATE(4365), 1,
      sym_name,
    STATE(5288), 1,
      sym__interface_type_indication,
    STATE(6180), 1,
      sym__mode_indication,
    STATE(3364), 2,
      sym__identifier,
      sym__external_name,
    STATE(5501), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    STATE(6284), 2,
      sym_simple_mode_indication,
      sym__mode_view_indication,
    STATE(6285), 2,
      sym_record_mode_view_indication,
      sym_array_mode_view_indication,
    ACTIONS(3834), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2714), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3367), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3826), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(3828), 5,
      sym_BUFFER,
      sym_IN,
      sym_INOUT,
      sym_LINKAGE,
      sym_OUT,
  [176440] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2715), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176492] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2716), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176544] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2717), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176596] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2718), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176648] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2719), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176700] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2720), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176752] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2721), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176804] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(3824), 1,
      anon_sym_LT_LT,
    ACTIONS(3830), 1,
      sym_TYPE,
    ACTIONS(3832), 1,
      sym_VIEW,
    STATE(2806), 1,
      sym__direct_name,
    STATE(3112), 1,
      sym_mode,
    STATE(4177), 1,
      sym_resolution_indication,
    STATE(4365), 1,
      sym_name,
    STATE(5288), 1,
      sym__interface_type_indication,
    STATE(6337), 1,
      sym__mode_indication,
    STATE(3364), 2,
      sym__identifier,
      sym__external_name,
    STATE(5501), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    STATE(6284), 2,
      sym_simple_mode_indication,
      sym__mode_view_indication,
    STATE(6285), 2,
      sym_record_mode_view_indication,
      sym_array_mode_view_indication,
    ACTIONS(3834), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2722), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3367), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3826), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(3828), 5,
      sym_BUFFER,
      sym_IN,
      sym_INOUT,
      sym_LINKAGE,
      sym_OUT,
  [176888] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2723), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [176940] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(3824), 1,
      anon_sym_LT_LT,
    ACTIONS(3830), 1,
      sym_TYPE,
    ACTIONS(3832), 1,
      sym_VIEW,
    STATE(2806), 1,
      sym__direct_name,
    STATE(3112), 1,
      sym_mode,
    STATE(4177), 1,
      sym_resolution_indication,
    STATE(4365), 1,
      sym_name,
    STATE(5288), 1,
      sym__interface_type_indication,
    STATE(6473), 1,
      sym__mode_indication,
    STATE(3364), 2,
      sym__identifier,
      sym__external_name,
    STATE(5501), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    STATE(6284), 2,
      sym_simple_mode_indication,
      sym__mode_view_indication,
    STATE(6285), 2,
      sym_record_mode_view_indication,
      sym_array_mode_view_indication,
    ACTIONS(3834), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2724), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3367), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3826), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(3828), 5,
      sym_BUFFER,
      sym_IN,
      sym_INOUT,
      sym_LINKAGE,
      sym_OUT,
  [177024] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2725), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177076] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(3824), 1,
      anon_sym_LT_LT,
    ACTIONS(3830), 1,
      sym_TYPE,
    ACTIONS(3832), 1,
      sym_VIEW,
    STATE(2806), 1,
      sym__direct_name,
    STATE(3112), 1,
      sym_mode,
    STATE(4177), 1,
      sym_resolution_indication,
    STATE(4365), 1,
      sym_name,
    STATE(5288), 1,
      sym__interface_type_indication,
    STATE(6569), 1,
      sym__mode_indication,
    STATE(3364), 2,
      sym__identifier,
      sym__external_name,
    STATE(5501), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    STATE(6284), 2,
      sym_simple_mode_indication,
      sym__mode_view_indication,
    STATE(6285), 2,
      sym_record_mode_view_indication,
      sym_array_mode_view_indication,
    ACTIONS(3834), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2726), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3367), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3826), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(3828), 5,
      sym_BUFFER,
      sym_IN,
      sym_INOUT,
      sym_LINKAGE,
      sym_OUT,
  [177160] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2727), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177212] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2728), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177264] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2729), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177316] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2730), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177368] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2731), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177420] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2732), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177472] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2733), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177524] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2734), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177576] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2735), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177628] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2736), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177680] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2737), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177732] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2738), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177784] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2739), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177836] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2740), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177888] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2741), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177940] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2742), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [177992] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2743), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178044] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2744), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178096] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2745), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178148] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2746), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178200] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2747), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178252] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2748), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178304] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2749), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178356] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2750), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 27,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SELECT,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178408] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3286), 1,
      sym_IS,
    STATE(2751), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178462] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2752), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178514] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2753), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178566] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(3824), 1,
      anon_sym_LT_LT,
    ACTIONS(3830), 1,
      sym_TYPE,
    ACTIONS(3832), 1,
      sym_VIEW,
    STATE(2806), 1,
      sym__direct_name,
    STATE(3112), 1,
      sym_mode,
    STATE(4177), 1,
      sym_resolution_indication,
    STATE(4365), 1,
      sym_name,
    STATE(5288), 1,
      sym__interface_type_indication,
    STATE(6283), 1,
      sym__mode_indication,
    STATE(3364), 2,
      sym__identifier,
      sym__external_name,
    STATE(5501), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    STATE(6284), 2,
      sym_simple_mode_indication,
      sym__mode_view_indication,
    STATE(6285), 2,
      sym_record_mode_view_indication,
      sym_array_mode_view_indication,
    ACTIONS(3834), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2754), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3367), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3826), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(3828), 5,
      sym_BUFFER,
      sym_IN,
      sym_INOUT,
      sym_LINKAGE,
      sym_OUT,
  [178650] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2755), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3044), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3042), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178702] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2756), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178754] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2757), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 27,
      sym_AND,
      sym_INERTIAL,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178806] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2758), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2446), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2444), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178858] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2759), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3058), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3056), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178910] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2760), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3070), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3068), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [178962] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2761), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179014] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2762), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3150), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3148), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179066] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2763), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3154), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3152), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179118] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2764), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3158), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3156), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179170] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2765), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3040), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3038), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179222] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2766), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179274] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2767), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 27,
      sym_AND,
      sym_LOOP,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179326] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2768), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 27,
      sym_AND,
      sym_IS,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179378] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2769), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3062), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3060), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179430] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2770), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3066), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3064), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179482] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2771), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2392), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2390), 27,
      sym_AND,
      sym_GENERATE,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179534] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3286), 1,
      sym_GENERATE,
    STATE(2772), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3166), 6,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_QMARK_LT,
      anon_sym_QMARK_GT,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(3164), 26,
      sym_AND,
      sym_MOD,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_REM,
      sym_ROL,
      sym_ROR,
      sym_SLA,
      sym_SLL,
      sym_SRA,
      sym_SRL,
      sym_XNOR,
      sym_XOR,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_EQ,
      anon_sym_QMARK_SLASH_EQ,
      anon_sym_QMARK_LT_EQ,
      anon_sym_QMARK_GT_EQ,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_AMP,
      sym_exponentiate,
  [179588] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3836), 1,
      anon_sym_RPAREN,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    STATE(5772), 1,
      sym__interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(7171), 1,
      sym_identifier_list,
    STATE(8174), 1,
      sym_interface_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2773), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [179669] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3858), 1,
      anon_sym_RPAREN,
    STATE(5582), 1,
      sym__identifier,
    STATE(5587), 1,
      sym__generic_interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(6896), 1,
      sym_generic_interface_list,
    STATE(7667), 1,
      sym_generic_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2774), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [179750] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3860), 1,
      anon_sym_RPAREN,
    STATE(5772), 1,
      sym__interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(7171), 1,
      sym_identifier_list,
    STATE(7545), 1,
      sym_interface_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2775), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [179831] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3862), 1,
      anon_sym_RPAREN,
    STATE(5582), 1,
      sym__identifier,
    STATE(5587), 1,
      sym__generic_interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(6990), 1,
      sym_generic_interface_list,
    STATE(7667), 1,
      sym_generic_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2776), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [179912] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3864), 1,
      anon_sym_RPAREN,
    STATE(5582), 1,
      sym__identifier,
    STATE(5587), 1,
      sym__generic_interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(7041), 1,
      sym_generic_interface_list,
    STATE(7667), 1,
      sym_generic_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2777), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [179993] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3866), 1,
      anon_sym_RPAREN,
    STATE(5582), 1,
      sym__identifier,
    STATE(5587), 1,
      sym__generic_interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(6775), 1,
      sym_generic_interface_list,
    STATE(7667), 1,
      sym_generic_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2778), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180074] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3868), 1,
      anon_sym_RPAREN,
    STATE(5772), 1,
      sym__interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(6897), 1,
      sym_interface_list,
    STATE(7171), 1,
      sym_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2779), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180155] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3870), 1,
      anon_sym_RPAREN,
    STATE(5772), 1,
      sym__interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(6778), 1,
      sym_interface_list,
    STATE(7171), 1,
      sym_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2780), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180236] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3872), 1,
      anon_sym_RPAREN,
    STATE(5772), 1,
      sym__interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(6797), 1,
      sym_interface_list,
    STATE(7171), 1,
      sym_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2781), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180317] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3874), 1,
      anon_sym_RPAREN,
    STATE(5772), 1,
      sym__interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(7150), 1,
      sym_interface_list,
    STATE(7171), 1,
      sym_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2782), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180398] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3876), 1,
      anon_sym_RPAREN,
    STATE(5582), 1,
      sym__identifier,
    STATE(5587), 1,
      sym__generic_interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(6777), 1,
      sym_generic_interface_list,
    STATE(7667), 1,
      sym_generic_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2783), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180479] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3878), 1,
      anon_sym_RPAREN,
    STATE(5582), 1,
      sym__identifier,
    STATE(5587), 1,
      sym__generic_interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(7663), 1,
      sym_generic_interface_list,
    STATE(7667), 1,
      sym_generic_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2784), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180560] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3880), 1,
      anon_sym_RPAREN,
    STATE(5772), 1,
      sym__interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(7171), 1,
      sym_identifier_list,
    STATE(8126), 1,
      sym_interface_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2785), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180641] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3882), 1,
      anon_sym_RPAREN,
    STATE(5772), 1,
      sym__interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(7171), 1,
      sym_identifier_list,
    STATE(7707), 1,
      sym_interface_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2786), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180722] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3884), 1,
      anon_sym_RPAREN,
    STATE(5772), 1,
      sym__interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(7171), 1,
      sym_identifier_list,
    STATE(7618), 1,
      sym_interface_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2787), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180803] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3886), 1,
      anon_sym_RPAREN,
    STATE(5582), 1,
      sym__identifier,
    STATE(5587), 1,
      sym__generic_interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(7667), 1,
      sym_generic_identifier_list,
    STATE(8172), 1,
      sym_generic_interface_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2788), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180884] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3888), 1,
      anon_sym_RPAREN,
    STATE(5582), 1,
      sym__identifier,
    STATE(5587), 1,
      sym__generic_interface_declaration,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(7616), 1,
      sym_generic_interface_list,
    STATE(7667), 1,
      sym_generic_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2789), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [180965] = 21,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3890), 1,
      anon_sym_RPAREN,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(6277), 1,
      sym__interface_declaration,
    STATE(7171), 1,
      sym_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2790), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [181043] = 21,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3892), 1,
      anon_sym_RPAREN,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(6277), 1,
      sym__interface_declaration,
    STATE(7171), 1,
      sym_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2791), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [181121] = 21,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3894), 1,
      anon_sym_RPAREN,
    STATE(5582), 1,
      sym__identifier,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(6106), 1,
      sym__generic_interface_declaration,
    STATE(7667), 1,
      sym_generic_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2792), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [181199] = 21,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    ACTIONS(3896), 1,
      anon_sym_RPAREN,
    STATE(5582), 1,
      sym__identifier,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(6106), 1,
      sym__generic_interface_declaration,
    STATE(7667), 1,
      sym_generic_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2793), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [181277] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3898), 1,
      anon_sym_DOT,
    ACTIONS(3900), 1,
      anon_sym_LPAREN,
    ACTIONS(3902), 1,
      anon_sym_SQUOTE,
    ACTIONS(3904), 1,
      anon_sym_LBRACK,
    ACTIONS(3906), 1,
      sym_GENERIC,
    ACTIONS(3908), 1,
      sym_PARAMETER,
    STATE(2798), 1,
      aux_sym_name_repeat1,
    STATE(3059), 1,
      sym_generic_map_aspect,
    STATE(3285), 1,
      sym__name_selector,
    STATE(2794), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3286), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1471), 14,
      sym_identifier,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [181342] = 20,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(5848), 1,
      sym__identifier,
    STATE(6277), 1,
      sym__interface_declaration,
    STATE(7171), 1,
      sym_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2795), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6191), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [181417] = 20,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3840), 1,
      sym_CONSTANT,
    ACTIONS(3842), 1,
      sym_FILE,
    ACTIONS(3844), 1,
      sym_FUNCTION,
    ACTIONS(3848), 1,
      sym_PACKAGE,
    ACTIONS(3850), 1,
      sym_PROCEDURE,
    ACTIONS(3852), 1,
      sym_SIGNAL,
    ACTIONS(3854), 1,
      sym_TYPE,
    ACTIONS(3856), 1,
      sym_VARIABLE,
    STATE(5582), 1,
      sym__identifier,
    STATE(5773), 1,
      sym__interface_subprogram_specification,
    STATE(6106), 1,
      sym__generic_interface_declaration,
    STATE(7667), 1,
      sym_generic_identifier_list,
    ACTIONS(3846), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(5774), 2,
      sym_interface_procedure_specification,
      sym_interface_function_specification,
    STATE(2796), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6235), 8,
      sym_interface_package_declaration,
      sym_interface_file_declaration,
      sym_interface_type_declaration,
      sym_generic_interface_declaration,
      sym_interface_constant_declaration,
      sym_interface_signal_declaration,
      sym_interface_variable_declaration,
      sym_interface_subprogram_declaration,
  [181492] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3910), 1,
      anon_sym_DOT,
    ACTIONS(3913), 1,
      anon_sym_LPAREN,
    ACTIONS(3916), 1,
      anon_sym_SQUOTE,
    ACTIONS(3919), 1,
      anon_sym_LBRACK,
    ACTIONS(3922), 1,
      sym_GENERIC,
    ACTIONS(3925), 1,
      sym_PARAMETER,
    STATE(3059), 1,
      sym_generic_map_aspect,
    STATE(3285), 1,
      sym__name_selector,
    STATE(2797), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(3286), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1491), 14,
      sym_identifier,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [181555] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3898), 1,
      anon_sym_DOT,
    ACTIONS(3900), 1,
      anon_sym_LPAREN,
    ACTIONS(3902), 1,
      anon_sym_SQUOTE,
    ACTIONS(3904), 1,
      anon_sym_LBRACK,
    ACTIONS(3906), 1,
      sym_GENERIC,
    ACTIONS(3908), 1,
      sym_PARAMETER,
    STATE(2797), 1,
      aux_sym_name_repeat1,
    STATE(3059), 1,
      sym_generic_map_aspect,
    STATE(3285), 1,
      sym__name_selector,
    STATE(2798), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3286), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1487), 14,
      sym_identifier,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
  [181620] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3928), 1,
      anon_sym_DOT,
    ACTIONS(3930), 1,
      anon_sym_LPAREN,
    ACTIONS(3932), 1,
      anon_sym_SQUOTE,
    ACTIONS(3934), 1,
      anon_sym_LBRACK,
    ACTIONS(3936), 1,
      sym_GENERIC,
    ACTIONS(3938), 1,
      sym_PARAMETER,
    STATE(2804), 1,
      aux_sym_name_repeat1,
    STATE(3270), 1,
      sym_generic_map_aspect,
    STATE(3395), 1,
      sym__name_selector,
    STATE(2799), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3396), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1471), 13,
      sym_identifier,
      sym_BUS,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      sym_variable_assignment,
      anon_sym_LT_LT,
  [181684] = 21,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(3940), 1,
      sym_END,
    ACTIONS(3942), 1,
      sym_POSTPONED,
    ACTIONS(3944), 1,
      sym_PROCESS,
    STATE(2803), 1,
      aux_sym_entity_body_repeat1,
    STATE(3023), 1,
      sym__direct_name,
    STATE(3405), 1,
      sym_label_declaration,
    STATE(4437), 1,
      sym__entity_statement,
    STATE(8055), 1,
      sym_assertion,
    STATE(8302), 1,
      sym__label,
    STATE(8334), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2800), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4438), 3,
      sym_process_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [181760] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3946), 1,
      anon_sym_DOT,
    ACTIONS(3948), 1,
      anon_sym_LPAREN,
    ACTIONS(3950), 1,
      anon_sym_SQUOTE,
    ACTIONS(3952), 1,
      anon_sym_LBRACK,
    ACTIONS(3954), 1,
      sym_GENERIC,
    ACTIONS(3956), 1,
      sym_PARAMETER,
    STATE(2802), 1,
      aux_sym_name_repeat1,
    STATE(3279), 1,
      sym_generic_map_aspect,
    STATE(3368), 1,
      sym__name_selector,
    STATE(2801), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3369), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1487), 13,
      sym_identifier,
      sym_BUS,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
  [181824] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3958), 1,
      anon_sym_DOT,
    ACTIONS(3961), 1,
      anon_sym_LPAREN,
    ACTIONS(3964), 1,
      anon_sym_SQUOTE,
    ACTIONS(3967), 1,
      anon_sym_LBRACK,
    ACTIONS(3970), 1,
      sym_GENERIC,
    ACTIONS(3973), 1,
      sym_PARAMETER,
    STATE(3279), 1,
      sym_generic_map_aspect,
    STATE(3368), 1,
      sym__name_selector,
    STATE(2802), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(3369), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1491), 13,
      sym_identifier,
      sym_BUS,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
  [181886] = 21,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(3942), 1,
      sym_POSTPONED,
    ACTIONS(3944), 1,
      sym_PROCESS,
    ACTIONS(3976), 1,
      sym_END,
    STATE(2805), 1,
      aux_sym_entity_body_repeat1,
    STATE(3023), 1,
      sym__direct_name,
    STATE(3405), 1,
      sym_label_declaration,
    STATE(4437), 1,
      sym__entity_statement,
    STATE(8055), 1,
      sym_assertion,
    STATE(8302), 1,
      sym__label,
    STATE(8334), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2803), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4438), 3,
      sym_process_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
    ACTIONS(13), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [181962] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3928), 1,
      anon_sym_DOT,
    ACTIONS(3930), 1,
      anon_sym_LPAREN,
    ACTIONS(3932), 1,
      anon_sym_SQUOTE,
    ACTIONS(3934), 1,
      anon_sym_LBRACK,
    ACTIONS(3936), 1,
      sym_GENERIC,
    ACTIONS(3938), 1,
      sym_PARAMETER,
    STATE(2807), 1,
      aux_sym_name_repeat1,
    STATE(3270), 1,
      sym_generic_map_aspect,
    STATE(3395), 1,
      sym__name_selector,
    STATE(2804), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3396), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1487), 13,
      sym_identifier,
      sym_BUS,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      sym_variable_assignment,
      anon_sym_LT_LT,
  [182026] = 20,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3978), 1,
      anon_sym_LT_LT,
    ACTIONS(3984), 1,
      sym_ASSERT,
    ACTIONS(3987), 1,
      sym_END,
    ACTIONS(3989), 1,
      sym_POSTPONED,
    ACTIONS(3992), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(3405), 1,
      sym_label_declaration,
    STATE(4437), 1,
      sym__entity_statement,
    STATE(8055), 1,
      sym_assertion,
    STATE(8302), 1,
      sym__label,
    STATE(8334), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3995), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4438), 3,
      sym_process_statement,
      sym_concurrent_assertion_statement,
      sym_concurrent_procedure_call_statement,
    ACTIONS(3981), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(2805), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_entity_body_repeat1,
  [182100] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3946), 1,
      anon_sym_DOT,
    ACTIONS(3948), 1,
      anon_sym_LPAREN,
    ACTIONS(3950), 1,
      anon_sym_SQUOTE,
    ACTIONS(3952), 1,
      anon_sym_LBRACK,
    ACTIONS(3954), 1,
      sym_GENERIC,
    ACTIONS(3956), 1,
      sym_PARAMETER,
    STATE(2801), 1,
      aux_sym_name_repeat1,
    STATE(3279), 1,
      sym_generic_map_aspect,
    STATE(3368), 1,
      sym__name_selector,
    STATE(2806), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3369), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1471), 13,
      sym_identifier,
      sym_BUS,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
  [182164] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3998), 1,
      anon_sym_DOT,
    ACTIONS(4001), 1,
      anon_sym_LPAREN,
    ACTIONS(4004), 1,
      anon_sym_SQUOTE,
    ACTIONS(4007), 1,
      anon_sym_LBRACK,
    ACTIONS(4010), 1,
      sym_GENERIC,
    ACTIONS(4013), 1,
      sym_PARAMETER,
    STATE(3270), 1,
      sym_generic_map_aspect,
    STATE(3395), 1,
      sym__name_selector,
    STATE(2807), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(3396), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1491), 13,
      sym_identifier,
      sym_BUS,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      sym_variable_assignment,
      anon_sym_LT_LT,
  [182226] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2808), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4016), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182269] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2809), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4018), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182312] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2810), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4020), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182355] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2811), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4022), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182398] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2812), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4024), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182441] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2813), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4026), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182484] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2814), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4028), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182527] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2815), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4030), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182570] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2816), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4032), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182613] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2817), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4034), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182656] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2818), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4036), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182699] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2819), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4038), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182742] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2820), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4040), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182785] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2821), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4042), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182828] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2822), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4044), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182871] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2823), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4046), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182914] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2824), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4048), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [182957] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2825), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4050), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183000] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2826), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4052), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183043] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2827), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4054), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183086] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2828), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4056), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183129] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2829), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4058), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183172] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2830), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4060), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183215] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2831), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4062), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183258] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2832), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4064), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183301] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2833), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4066), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183344] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2834), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4068), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183387] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2835), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4070), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183430] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2836), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4072), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183473] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4074), 1,
      anon_sym_DOT,
    ACTIONS(4077), 1,
      anon_sym_LPAREN,
    ACTIONS(4080), 1,
      anon_sym_SQUOTE,
    ACTIONS(4083), 1,
      anon_sym_LBRACK,
    ACTIONS(4086), 1,
      sym_GENERIC,
    ACTIONS(4089), 1,
      sym_PARAMETER,
    STATE(3415), 1,
      sym_generic_map_aspect,
    STATE(3522), 1,
      sym__name_selector,
    STATE(2837), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(3523), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1491), 12,
      sym_identifier,
      sym_IS,
      sym_OPEN,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_LT,
  [183534] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2838), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4092), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183577] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2839), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4094), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183620] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2840), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4096), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183663] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2841), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4098), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183706] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2842), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4100), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183749] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2843), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4102), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183792] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2844), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4104), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183835] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2845), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4106), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183878] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2846), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4108), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183921] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2847), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4110), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [183964] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2848), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4112), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184007] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4114), 1,
      anon_sym_DOT,
    ACTIONS(4116), 1,
      anon_sym_LPAREN,
    ACTIONS(4118), 1,
      anon_sym_SQUOTE,
    ACTIONS(4120), 1,
      anon_sym_LBRACK,
    ACTIONS(4122), 1,
      sym_GENERIC,
    ACTIONS(4124), 1,
      sym_PARAMETER,
    STATE(2859), 1,
      aux_sym_name_repeat1,
    STATE(3415), 1,
      sym_generic_map_aspect,
    STATE(3522), 1,
      sym__name_selector,
    STATE(2849), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3523), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1471), 12,
      sym_identifier,
      sym_IS,
      sym_OPEN,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_LT,
  [184070] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2850), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4126), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184113] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2851), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4128), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184156] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2852), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4130), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184199] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2853), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4132), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184242] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2854), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4134), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184285] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2855), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4136), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184328] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2856), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4138), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184371] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2857), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4140), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184414] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2858), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4142), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184457] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4114), 1,
      anon_sym_DOT,
    ACTIONS(4116), 1,
      anon_sym_LPAREN,
    ACTIONS(4118), 1,
      anon_sym_SQUOTE,
    ACTIONS(4120), 1,
      anon_sym_LBRACK,
    ACTIONS(4122), 1,
      sym_GENERIC,
    ACTIONS(4124), 1,
      sym_PARAMETER,
    STATE(2837), 1,
      aux_sym_name_repeat1,
    STATE(3415), 1,
      sym_generic_map_aspect,
    STATE(3522), 1,
      sym__name_selector,
    STATE(2859), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3523), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1487), 12,
      sym_identifier,
      sym_IS,
      sym_OPEN,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_LT_LT,
  [184520] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2860), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4144), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184563] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2861), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4146), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184606] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2862), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4148), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184649] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2863), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4150), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184692] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2864), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4152), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184735] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2865), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4154), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184778] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2866), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4156), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184821] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2867), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4158), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184864] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2868), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4160), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184907] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2869), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4162), 26,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184950] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2870), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4134), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [184992] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2871), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4066), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185034] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2872), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4030), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185076] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2873), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4106), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185118] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3898), 1,
      anon_sym_DOT,
    ACTIONS(3900), 1,
      anon_sym_LPAREN,
    ACTIONS(3902), 1,
      anon_sym_SQUOTE,
    ACTIONS(3904), 1,
      anon_sym_LBRACK,
    ACTIONS(3908), 1,
      sym_PARAMETER,
    STATE(2900), 1,
      aux_sym_name_repeat1,
    STATE(3059), 1,
      sym_generic_map_aspect,
    STATE(3285), 1,
      sym__name_selector,
    STATE(2874), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3286), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1471), 12,
      sym_identifier,
      sym_GENERIC,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      sym_variable_assignment,
      anon_sym_LT_LT,
  [185178] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2875), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4072), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185220] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2876), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4126), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185262] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2877), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4050), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185304] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2878), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4160), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185346] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2879), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4156), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185388] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2880), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4158), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185430] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2881), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4138), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185472] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2882), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4036), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185514] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2883), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4040), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185556] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2884), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4068), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185598] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2885), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4064), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185640] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2886), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4092), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185682] = 22,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4164), 1,
      sym_ALIAS,
    ACTIONS(4166), 1,
      sym_ATTRIBUTE,
    ACTIONS(4168), 1,
      sym_BODY,
    ACTIONS(4170), 1,
      sym_END,
    ACTIONS(4172), 1,
      sym_FUNCTION,
    ACTIONS(4174), 1,
      sym_GENERIC,
    ACTIONS(4176), 1,
      sym_PROCEDURE,
    ACTIONS(4178), 1,
      sym_PRIVATE,
    ACTIONS(4180), 1,
      sym_USE,
    STATE(3074), 1,
      sym_protected_type_header,
    STATE(3075), 1,
      aux_sym_protected_type_declaration_repeat1,
    STATE(4466), 1,
      sym_generic_clause,
    STATE(4770), 1,
      sym__protected_type_declarative_item,
    STATE(7034), 1,
      sym_protected_type_declaration_end,
    STATE(7550), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(2887), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4768), 6,
      sym_use_clause,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_alias_declaration,
      sym_attribute_specification,
      sym_private_variable_declaration,
  [185758] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2888), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4096), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185800] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2889), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4042), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185842] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2890), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4132), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185884] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2891), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4094), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185926] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2892), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4152), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [185968] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2893), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4044), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186010] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2894), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4046), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186052] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2895), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4128), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186094] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2896), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4130), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186136] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2897), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4048), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186178] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2898), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4104), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186220] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2899), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4154), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186262] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3898), 1,
      anon_sym_DOT,
    ACTIONS(3900), 1,
      anon_sym_LPAREN,
    ACTIONS(3902), 1,
      anon_sym_SQUOTE,
    ACTIONS(3904), 1,
      anon_sym_LBRACK,
    ACTIONS(3908), 1,
      sym_PARAMETER,
    STATE(2797), 1,
      aux_sym_name_repeat1,
    STATE(3059), 1,
      sym_generic_map_aspect,
    STATE(3285), 1,
      sym__name_selector,
    STATE(2900), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3286), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1487), 12,
      sym_identifier,
      sym_GENERIC,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      sym_variable_assignment,
      anon_sym_LT_LT,
  [186322] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2901), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4034), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186364] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2902), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4112), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186406] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2903), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4038), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186448] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2904), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4142), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186490] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2905), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4028), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186532] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2906), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4146), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186574] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2907), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4162), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186616] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2908), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4150), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186658] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2909), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4054), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186700] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2910), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4056), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186742] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2911), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4018), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186784] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2912), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4026), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186826] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2913), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4144), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186868] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2914), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4148), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186910] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2915), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4062), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186952] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2916), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4070), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [186994] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2917), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4020), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187036] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2918), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4022), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187078] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2919), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4098), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187120] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2920), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4108), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187162] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2921), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4058), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187204] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2922), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4102), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187246] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2923), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4100), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187288] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2924), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4136), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187330] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2925), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4060), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187372] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2926), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4140), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187414] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2927), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4024), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187456] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2928), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4016), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187498] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2929), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4032), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187540] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2930), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4052), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187582] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2931), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4110), 25,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHEN,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187624] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2932), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4154), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187665] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2933), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4050), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187706] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2934), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4134), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187747] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4182), 1,
      anon_sym_DOT,
    ACTIONS(4184), 1,
      anon_sym_LPAREN,
    ACTIONS(4186), 1,
      anon_sym_SQUOTE,
    ACTIONS(4188), 1,
      anon_sym_LBRACK,
    ACTIONS(4190), 1,
      sym_GENERIC,
    ACTIONS(4192), 1,
      sym_PARAMETER,
    STATE(2962), 1,
      aux_sym_name_repeat1,
    STATE(3647), 1,
      sym_generic_map_aspect,
    STATE(3808), 1,
      sym__name_selector,
    STATE(2935), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3809), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1487), 10,
      sym_identifier,
      sym_IS,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [187808] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2936), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4092), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187849] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2937), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4062), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187890] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2938), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4070), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187931] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2939), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4130), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [187972] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2940), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4072), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2941), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4022), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188054] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2942), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4044), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188095] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2943), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4156), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188136] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2944), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4032), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188177] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2945), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4052), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188218] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2946), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4064), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188259] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2947), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4026), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188300] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2948), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4126), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188341] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2949), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4140), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188382] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2950), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4102), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188423] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2951), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4046), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188464] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2952), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4016), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188505] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2953), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4024), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188546] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2954), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4108), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188587] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2955), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4110), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188628] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2956), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4112), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188669] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2957), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4028), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188710] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2958), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4034), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188751] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2959), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4194), 24,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GENERIC,
      sym_GROUP,
      sym_IMPURE,
      sym_IS,
      sym_PACKAGE,
      sym_PORT,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [188792] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2960), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4042), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188833] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2961), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4150), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [188874] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4196), 1,
      anon_sym_DOT,
    ACTIONS(4199), 1,
      anon_sym_LPAREN,
    ACTIONS(4202), 1,
      anon_sym_SQUOTE,
    ACTIONS(4205), 1,
      anon_sym_LBRACK,
    ACTIONS(4208), 1,
      sym_GENERIC,
    ACTIONS(4211), 1,
      sym_PARAMETER,
    STATE(3647), 1,
      sym_generic_map_aspect,
    STATE(3808), 1,
      sym__name_selector,
    STATE(2962), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(3809), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1491), 10,
      sym_identifier,
      sym_IS,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [188933] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    STATE(7540), 1,
      sym_end_for,
    STATE(7698), 1,
      sym_verification_unit_binding_indication,
    STATE(2963), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2982), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [188980] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2964), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4162), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189021] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2965), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4100), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189062] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2966), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4160), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189103] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2967), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4066), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189144] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2968), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4098), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189185] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2969), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4058), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189226] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2970), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4136), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189267] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2971), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4158), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189308] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2972), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4018), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189349] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2973), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4036), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189390] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2974), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4146), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189431] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2975), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4128), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189472] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2976), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4152), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189513] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2977), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4138), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189554] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2978), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4068), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189595] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2979), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4094), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189636] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2980), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4132), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189677] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2981), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4142), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189718] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2982), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4106), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189759] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2983), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4144), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189800] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2984), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4040), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189841] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2985), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4048), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189882] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2986), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4054), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189923] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2987), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4104), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [189964] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2988), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4030), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [190005] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2989), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4060), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [190046] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4182), 1,
      anon_sym_DOT,
    ACTIONS(4184), 1,
      anon_sym_LPAREN,
    ACTIONS(4186), 1,
      anon_sym_SQUOTE,
    ACTIONS(4188), 1,
      anon_sym_LBRACK,
    ACTIONS(4190), 1,
      sym_GENERIC,
    ACTIONS(4192), 1,
      sym_PARAMETER,
    STATE(2935), 1,
      aux_sym_name_repeat1,
    STATE(3647), 1,
      sym_generic_map_aspect,
    STATE(3808), 1,
      sym__name_selector,
    STATE(2990), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3809), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1471), 10,
      sym_identifier,
      sym_IS,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [190107] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2991), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4038), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [190148] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2992), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4020), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [190189] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2993), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4056), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [190230] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2994), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4148), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [190271] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2995), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4096), 24,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_END,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [190312] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(3806), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6406), 1,
      sym__target,
    STATE(6482), 1,
      sym_with_expression,
    STATE(6486), 1,
      sym_name,
    STATE(8355), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2996), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [190380] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5695), 1,
      sym_entity_class_entry,
    STATE(5696), 1,
      sym__entity_class,
    STATE(6898), 1,
      sym_entity_class_entry_list,
    STATE(2997), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4214), 20,
      sym_ARCHITECTURE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_ENTITY,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_LABEL,
      sym_LITERAL,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PROPERTY,
      sym_SEQUENCE,
      sym_SIGNAL,
      sym_SUBTYPE,
      sym_TYPE,
      sym_UNITS,
      sym_VARIABLE,
      sym_VIEW,
  [190426] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2998), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4216), 23,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GENERIC,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PORT,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [190466] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(4218), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6226), 1,
      sym_name,
    STATE(6449), 1,
      sym__target,
    STATE(6497), 1,
      sym_with_expression,
    STATE(6841), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2999), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [190534] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5695), 1,
      sym_entity_class_entry,
    STATE(5696), 1,
      sym__entity_class,
    STATE(6780), 1,
      sym_entity_class_entry_list,
    STATE(3000), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4214), 20,
      sym_ARCHITECTURE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_ENTITY,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_LABEL,
      sym_LITERAL,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PROPERTY,
      sym_SEQUENCE,
      sym_SIGNAL,
      sym_SUBTYPE,
      sym_TYPE,
      sym_UNITS,
      sym_VARIABLE,
      sym_VIEW,
  [190580] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5695), 1,
      sym_entity_class_entry,
    STATE(5696), 1,
      sym__entity_class,
    STATE(6991), 1,
      sym_entity_class_entry_list,
    STATE(3001), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4214), 20,
      sym_ARCHITECTURE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_ENTITY,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_LABEL,
      sym_LITERAL,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PROPERTY,
      sym_SEQUENCE,
      sym_SIGNAL,
      sym_SUBTYPE,
      sym_TYPE,
      sym_UNITS,
      sym_VARIABLE,
      sym_VIEW,
  [190626] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(3818), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6243), 1,
      sym_with_expression,
    STATE(6245), 1,
      sym_name,
    STATE(6309), 1,
      sym__target,
    STATE(8323), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3002), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [190694] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3003), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4220), 23,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GENERIC,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PORT,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [190734] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(3792), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6438), 1,
      sym__target,
    STATE(6495), 1,
      sym_with_expression,
    STATE(6605), 1,
      sym_name,
    STATE(6711), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3004), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [190802] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(4222), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6338), 1,
      sym_with_expression,
    STATE(6352), 1,
      sym_name,
    STATE(6362), 1,
      sym__target,
    STATE(8026), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3005), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [190870] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(4224), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6335), 1,
      sym_name,
    STATE(6388), 1,
      sym__target,
    STATE(6475), 1,
      sym_with_expression,
    STATE(7515), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3006), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [190938] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(3780), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6321), 1,
      sym_name,
    STATE(6376), 1,
      sym__target,
    STATE(6471), 1,
      sym_with_expression,
    STATE(7462), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3007), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [191006] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5695), 1,
      sym_entity_class_entry,
    STATE(5696), 1,
      sym__entity_class,
    STATE(6854), 1,
      sym_entity_class_entry_list,
    STATE(3008), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4214), 20,
      sym_ARCHITECTURE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_ENTITY,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_LABEL,
      sym_LITERAL,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PROPERTY,
      sym_SEQUENCE,
      sym_SIGNAL,
      sym_SUBTYPE,
      sym_TYPE,
      sym_UNITS,
      sym_VARIABLE,
      sym_VIEW,
  [191052] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5695), 1,
      sym_entity_class_entry,
    STATE(5696), 1,
      sym__entity_class,
    STATE(7042), 1,
      sym_entity_class_entry_list,
    STATE(3009), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4214), 20,
      sym_ARCHITECTURE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_ENTITY,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_LABEL,
      sym_LITERAL,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PROPERTY,
      sym_SEQUENCE,
      sym_SIGNAL,
      sym_SUBTYPE,
      sym_TYPE,
      sym_UNITS,
      sym_VARIABLE,
      sym_VIEW,
  [191098] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3010), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3750), 23,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_EXIT,
      sym_FOR,
      sym_IF,
      sym_LOOP,
      sym_NEXT,
      sym_NULL,
      sym_REPORT,
      sym_RETURN,
      sym_WAIT,
      sym_WHILE,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [191138] = 19,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(69), 1,
      sym_WITH,
    ACTIONS(4226), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6098), 1,
      sym_aggregate,
    STATE(6420), 1,
      sym__target,
    STATE(6488), 1,
      sym_with_expression,
    STATE(6504), 1,
      sym_name,
    STATE(7395), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3011), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [191206] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5695), 1,
      sym_entity_class_entry,
    STATE(5696), 1,
      sym__entity_class,
    STATE(7073), 1,
      sym_entity_class_entry_list,
    STATE(3012), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4214), 20,
      sym_ARCHITECTURE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_ENTITY,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_LABEL,
      sym_LITERAL,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PROPERTY,
      sym_SEQUENCE,
      sym_SIGNAL,
      sym_SUBTYPE,
      sym_TYPE,
      sym_UNITS,
      sym_VARIABLE,
      sym_VIEW,
  [191252] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6060), 1,
      sym_name,
    STATE(6212), 1,
      sym_resolution_indication,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3013), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(2386), 7,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [191308] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3014), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3230), 22,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [191347] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    STATE(6881), 1,
      sym_generic_map_aspect,
    STATE(3015), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4228), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [191390] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3016), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3174), 22,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [191429] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3017), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4230), 22,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PORT,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [191468] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3018), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3212), 22,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [191507] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3019), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3190), 22,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [191546] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3020), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3220), 22,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [191585] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1489), 1,
      anon_sym_COLON,
    ACTIONS(4232), 1,
      anon_sym_DOT,
    ACTIONS(4234), 1,
      anon_sym_LPAREN,
    ACTIONS(4236), 1,
      anon_sym_SQUOTE,
    ACTIONS(4238), 1,
      anon_sym_LBRACK,
    ACTIONS(4240), 1,
      sym_GENERIC,
    ACTIONS(4242), 1,
      sym_PARAMETER,
    STATE(3025), 1,
      aux_sym_name_repeat1,
    STATE(4272), 1,
      sym_generic_map_aspect,
    STATE(4331), 1,
      sym__name_selector,
    STATE(3021), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4337), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1487), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_RBRACK,
  [191646] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5696), 1,
      sym__entity_class,
    STATE(6077), 1,
      sym_entity_class_entry,
    STATE(3022), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4214), 20,
      sym_ARCHITECTURE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_ENTITY,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_LABEL,
      sym_LITERAL,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PROPERTY,
      sym_SEQUENCE,
      sym_SIGNAL,
      sym_SUBTYPE,
      sym_TYPE,
      sym_UNITS,
      sym_VARIABLE,
      sym_VIEW,
  [191689] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1477), 1,
      anon_sym_COLON,
    ACTIONS(4232), 1,
      anon_sym_DOT,
    ACTIONS(4234), 1,
      anon_sym_LPAREN,
    ACTIONS(4236), 1,
      anon_sym_SQUOTE,
    ACTIONS(4238), 1,
      anon_sym_LBRACK,
    ACTIONS(4240), 1,
      sym_GENERIC,
    ACTIONS(4242), 1,
      sym_PARAMETER,
    STATE(3021), 1,
      aux_sym_name_repeat1,
    STATE(4272), 1,
      sym_generic_map_aspect,
    STATE(4331), 1,
      sym__name_selector,
    STATE(3023), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4337), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1471), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_RBRACK,
  [191750] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3024), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4244), 22,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PORT,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [191789] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1499), 1,
      anon_sym_COLON,
    ACTIONS(4246), 1,
      anon_sym_DOT,
    ACTIONS(4249), 1,
      anon_sym_LPAREN,
    ACTIONS(4252), 1,
      anon_sym_SQUOTE,
    ACTIONS(4255), 1,
      anon_sym_LBRACK,
    ACTIONS(4258), 1,
      sym_GENERIC,
    ACTIONS(4261), 1,
      sym_PARAMETER,
    STATE(4272), 1,
      sym_generic_map_aspect,
    STATE(4331), 1,
      sym__name_selector,
    STATE(3025), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4337), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1491), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_RBRACK,
  [191848] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8311), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3026), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [191908] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6637), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3027), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [191968] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7074), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3028), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [192028] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7078), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3029), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [192088] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3030), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3498), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192126] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4164), 1,
      sym_ALIAS,
    ACTIONS(4166), 1,
      sym_ATTRIBUTE,
    ACTIONS(4170), 1,
      sym_END,
    ACTIONS(4172), 1,
      sym_FUNCTION,
    ACTIONS(4176), 1,
      sym_PROCEDURE,
    ACTIONS(4178), 1,
      sym_PRIVATE,
    ACTIONS(4180), 1,
      sym_USE,
    STATE(3352), 1,
      aux_sym_protected_type_declaration_repeat1,
    STATE(4770), 1,
      sym__protected_type_declarative_item,
    STATE(6826), 1,
      sym_protected_type_declaration_end,
    STATE(7550), 1,
      sym__subprogram_specification,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(3031), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4768), 6,
      sym_use_clause,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_alias_declaration,
      sym_attribute_specification,
      sym_private_variable_declaration,
  [192190] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3032), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3288), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192228] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3033), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3292), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192266] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(5417), 1,
      sym__incomplete_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(5416), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3034), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [192326] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3035), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3356), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192364] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3036), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3358), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192402] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3037), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3362), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192440] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3038), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3370), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192478] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3039), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3388), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192516] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3040), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3430), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192554] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3041), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3434), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192592] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3042), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3436), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192630] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3043), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3440), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192668] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3044), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4220), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PORT,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192706] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3045), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3456), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192744] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3046), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3460), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192782] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3047), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3464), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192820] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3048), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3466), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192858] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3049), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3468), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192896] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3050), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3480), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [192934] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7220), 1,
      sym__entity_class,
    STATE(3051), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4270), 20,
      sym_ARCHITECTURE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_CONSTANT,
      sym_ENTITY,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_LABEL,
      sym_LITERAL,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PROPERTY,
      sym_SEQUENCE,
      sym_SIGNAL,
      sym_SUBTYPE,
      sym_TYPE,
      sym_UNITS,
      sym_VARIABLE,
      sym_VIEW,
  [192974] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3052), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3496), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193012] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3053), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3246), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193050] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3054), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3248), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193088] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3055), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4216), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PORT,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193126] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3056), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3252), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193164] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3057), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3254), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193202] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3058), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3256), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193240] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3900), 1,
      anon_sym_LPAREN,
    ACTIONS(4272), 1,
      sym_PARAMETER,
    STATE(3295), 1,
      sym_parenthesis_group,
    STATE(3059), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 18,
      sym_identifier,
      sym_GENERIC,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [193284] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3060), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3266), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193322] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3061), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3282), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193360] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3062), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3302), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193398] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3063), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3304), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193436] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3064), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3308), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193474] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3065), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3312), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193512] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(5472), 1,
      sym__incomplete_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(5416), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3066), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [193572] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(4276), 1,
      sym_based_base,
    ACTIONS(4278), 1,
      sym_library_constant_unit,
    STATE(3023), 1,
      sym__direct_name,
    STATE(3789), 1,
      sym__abstract_literal,
    STATE(4879), 1,
      sym_based_literal,
    STATE(6732), 1,
      sym_name,
    ACTIONS(4274), 2,
      sym_decimal_integer,
      sym_decimal_float,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3067), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [193632] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3068), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4220), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GENERIC,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193670] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3069), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4216), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GENERIC,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193708] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3070), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3348), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193746] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3071), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3274), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [193784] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(3824), 1,
      anon_sym_LT_LT,
    ACTIONS(3830), 1,
      sym_TYPE,
    STATE(2806), 1,
      sym__direct_name,
    STATE(4177), 1,
      sym_resolution_indication,
    STATE(4365), 1,
      sym_name,
    STATE(5526), 1,
      sym__incomplete_subtype_indication,
    STATE(3364), 2,
      sym__identifier,
      sym__external_name,
    STATE(5525), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(3834), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3072), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3367), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3826), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [193844] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(3824), 1,
      anon_sym_LT_LT,
    ACTIONS(3830), 1,
      sym_TYPE,
    STATE(2806), 1,
      sym__direct_name,
    STATE(4177), 1,
      sym_resolution_indication,
    STATE(4365), 1,
      sym_name,
    STATE(5532), 1,
      sym__incomplete_subtype_indication,
    STATE(3364), 2,
      sym__identifier,
      sym__external_name,
    STATE(5525), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(3834), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3073), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3367), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3826), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [193904] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4164), 1,
      sym_ALIAS,
    ACTIONS(4166), 1,
      sym_ATTRIBUTE,
    ACTIONS(4170), 1,
      sym_END,
    ACTIONS(4172), 1,
      sym_FUNCTION,
    ACTIONS(4176), 1,
      sym_PROCEDURE,
    ACTIONS(4178), 1,
      sym_PRIVATE,
    ACTIONS(4180), 1,
      sym_USE,
    STATE(3031), 1,
      aux_sym_protected_type_declaration_repeat1,
    STATE(4770), 1,
      sym__protected_type_declarative_item,
    STATE(7550), 1,
      sym__subprogram_specification,
    STATE(8122), 1,
      sym_protected_type_declaration_end,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(3074), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4768), 6,
      sym_use_clause,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_alias_declaration,
      sym_attribute_specification,
      sym_private_variable_declaration,
  [193968] = 18,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4164), 1,
      sym_ALIAS,
    ACTIONS(4166), 1,
      sym_ATTRIBUTE,
    ACTIONS(4170), 1,
      sym_END,
    ACTIONS(4172), 1,
      sym_FUNCTION,
    ACTIONS(4176), 1,
      sym_PROCEDURE,
    ACTIONS(4178), 1,
      sym_PRIVATE,
    ACTIONS(4180), 1,
      sym_USE,
    STATE(3352), 1,
      aux_sym_protected_type_declaration_repeat1,
    STATE(4770), 1,
      sym__protected_type_declarative_item,
    STATE(7550), 1,
      sym__subprogram_specification,
    STATE(8122), 1,
      sym_protected_type_declaration_end,
    ACTIONS(2241), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(3075), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4768), 6,
      sym_use_clause,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_alias_declaration,
      sym_attribute_specification,
      sym_private_variable_declaration,
  [194032] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3076), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3442), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [194070] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4282), 1,
      sym_ELSE,
    ACTIONS(4284), 1,
      sym_ELSIF,
    ACTIONS(4286), 1,
      sym_END,
    ACTIONS(4288), 1,
      sym_IF,
    ACTIONS(4290), 1,
      sym_directive_error,
    ACTIONS(4292), 1,
      sym_directive_protect,
    ACTIONS(4294), 1,
      sym_directive_warning,
    STATE(5618), 1,
      sym__identifier,
    STATE(8225), 1,
      sym__directive,
    STATE(3077), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4280), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(6716), 8,
      sym_user_directive,
      sym_protect_directive,
      sym_warning_directive,
      sym_error_directive,
      sym_if_conditional_analysis,
      sym_elsif_conditional_analysis,
      sym_else_conditional_analysis,
      sym_end_conditional_analysis,
  [194128] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3078), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3360), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [194166] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7652), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3079), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194226] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7653), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3080), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194286] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7654), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3081), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194346] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3082), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3366), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [194384] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3083), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3374), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [194422] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8204), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3084), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194482] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8205), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3085), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194542] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8206), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3086), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194602] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3087), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3428), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [194640] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7129), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3088), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194700] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7195), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3089), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194760] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7225), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3090), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194820] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7053), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3091), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194880] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7065), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3092), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [194940] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7072), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3093), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195000] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7755), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3094), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195060] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7763), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3095), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195120] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7765), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3096), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195180] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6667), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3097), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195240] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6669), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3098), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195300] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6672), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3099), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195360] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6855), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3100), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195420] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6858), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3101), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195480] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6860), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3102), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195540] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6957), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3103), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195600] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6961), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3104), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195660] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6963), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3105), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195720] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7085), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3106), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195780] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7088), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3107), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195840] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7095), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3108), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195900] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7234), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3109), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [195960] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7237), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3110), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196020] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7244), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3111), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196080] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(3824), 1,
      anon_sym_LT_LT,
    ACTIONS(3830), 1,
      sym_TYPE,
    STATE(2806), 1,
      sym__direct_name,
    STATE(4177), 1,
      sym_resolution_indication,
    STATE(4365), 1,
      sym_name,
    STATE(5234), 1,
      sym__interface_type_indication,
    STATE(3364), 2,
      sym__identifier,
      sym__external_name,
    STATE(5501), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(3834), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3112), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3367), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3826), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196140] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7339), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3113), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196200] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7348), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3114), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196260] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7352), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3115), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196320] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3116), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3494), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [196358] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7485), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3117), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196418] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7490), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3118), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196478] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7498), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3119), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196538] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7686), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3120), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196598] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7687), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3121), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196658] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7701), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3122), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196718] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7808), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3123), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196778] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7810), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3124), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196838] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7815), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3125), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196898] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8106), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3126), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [196958] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8114), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3127), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197018] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8132), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3128), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197078] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8301), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3129), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197138] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8309), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3130), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197198] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6628), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3131), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197258] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6629), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3132), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197318] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6631), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3133), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197378] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6699), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3134), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197438] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6700), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3135), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197498] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6701), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3136), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197558] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6748), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3137), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197618] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6749), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3138), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197678] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6751), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3139), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197738] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6800), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3140), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197798] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6801), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3141), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197858] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6804), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3142), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197918] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6849), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3143), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [197978] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6850), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3144), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198038] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6851), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3145), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198098] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6883), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3146), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198158] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6884), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3147), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198218] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6888), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3148), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198278] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3149), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3250), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [198316] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6930), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3150), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198376] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6931), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3151), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198436] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6933), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3152), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198496] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6971), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3153), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198556] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6972), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3154), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198616] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6973), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3155), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198676] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7018), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3156), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198736] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7019), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3157), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198796] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7020), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3158), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198856] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7076), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3159), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198916] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7082), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3160), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [198976] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7083), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3161), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199036] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7141), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3162), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199096] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7143), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3163), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199156] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7147), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3164), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199216] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7198), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3165), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199276] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7201), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3166), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199336] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7203), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3167), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199396] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7253), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3168), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199456] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7254), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3169), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199516] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7255), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3170), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199576] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7300), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3171), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199636] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7304), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3172), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199696] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7309), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3173), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199756] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7340), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3174), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199816] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7345), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3175), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199876] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7347), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3176), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199936] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7384), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3177), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [199996] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7385), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3178), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200056] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7386), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3179), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200116] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7432), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3180), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200176] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7435), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3181), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200236] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7437), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3182), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200296] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7501), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3183), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200356] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7506), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3184), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200416] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7507), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3185), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200476] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7572), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3186), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200536] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7573), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3187), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200596] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7578), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3188), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200656] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7644), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3189), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200716] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7647), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3190), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200776] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7649), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3191), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200836] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7693), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3192), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200896] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7695), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3193), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [200956] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7700), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3194), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201016] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7728), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3195), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201076] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7729), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3196), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201136] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7730), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3197), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201196] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7766), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3198), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201256] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7767), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3199), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201316] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7768), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3200), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201376] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7798), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3201), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201436] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7801), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3202), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201496] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7802), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3203), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201556] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7838), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3204), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201616] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7839), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3205), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201676] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7849), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3206), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201736] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8057), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3207), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201796] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8059), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3208), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201856] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8061), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3209), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201916] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8100), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3210), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [201976] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8102), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3211), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202036] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8103), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3212), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202096] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3213), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3280), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [202134] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8159), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3214), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202194] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8160), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3215), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202254] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8170), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3216), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202314] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3217), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3284), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [202352] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8233), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3218), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202412] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8237), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3219), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202472] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8241), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3220), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202532] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8304), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3221), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202592] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8306), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3222), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202652] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8307), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3223), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202712] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3224), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3300), 21,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_FILE,
      sym_FOR,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [202750] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8325), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3225), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202810] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8326), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3226), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202870] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8327), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3227), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202930] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8344), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3228), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [202990] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8345), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3229), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [203050] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8346), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3230), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [203110] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6616), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3231), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [203170] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6617), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3232), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [203230] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6618), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3233), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [203290] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6635), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3234), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [203350] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6636), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3235), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [203410] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(281), 1,
      sym_TYPE,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7070), 1,
      sym__interface_type_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    STATE(7071), 2,
      sym_unspecified_type_indication,
      sym_subtype_indication,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3236), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [203470] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3237), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3280), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203507] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3238), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3308), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203544] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3239), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3312), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203581] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3240), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3498), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203618] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3241), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4296), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203655] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3242), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3288), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203692] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3243), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3292), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203729] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3244), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4298), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203766] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3245), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3356), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203803] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3246), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3358), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203840] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3247), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3362), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203877] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3248), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3370), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203914] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3249), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3388), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203951] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3250), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3430), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [203988] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3251), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3434), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204025] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3252), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3436), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204062] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3253), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3456), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204099] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3254), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3460), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204136] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3255), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3464), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204173] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3256), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3468), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204210] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3257), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3480), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204247] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3258), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3496), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204284] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3259), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3246), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204321] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3260), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3248), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204358] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3261), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3252), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204395] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3262), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [204432] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3263), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3254), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204469] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3264), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3256), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204506] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3265), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3266), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204543] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3266), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3282), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204580] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3267), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3302), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204617] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3268), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3304), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204654] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3269), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3308), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204691] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3930), 1,
      anon_sym_LPAREN,
    ACTIONS(4300), 1,
      sym_PARAMETER,
    STATE(3414), 1,
      sym_parenthesis_group,
    STATE(3270), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 17,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [204734] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3271), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3312), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204771] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3272), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3292), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204808] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3273), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4302), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204845] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3274), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4304), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204882] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(4306), 1,
      anon_sym_LPAREN,
    STATE(3013), 1,
      sym__identifier,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4362), 1,
      sym__external_name,
    STATE(5815), 1,
      sym_record_element_resolution,
    STATE(6060), 1,
      sym_name,
    STATE(7607), 1,
      sym__element_resolution,
    STATE(7090), 2,
      sym_resolution_indication,
      sym_record_resolution,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3275), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4308), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [204941] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3276), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3442), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [204978] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3277), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3494), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [205015] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3278), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205052] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3948), 1,
      anon_sym_LPAREN,
    ACTIONS(4310), 1,
      sym_PARAMETER,
    STATE(3378), 1,
      sym_parenthesis_group,
    STATE(3279), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 17,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205095] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3280), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205132] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4312), 1,
      anon_sym_DOT,
    ACTIONS(4314), 1,
      anon_sym_LPAREN,
    ACTIONS(4316), 1,
      anon_sym_SQUOTE,
    ACTIONS(4318), 1,
      anon_sym_LBRACK,
    ACTIONS(4320), 1,
      sym_GENERIC,
    ACTIONS(4322), 1,
      sym_PARAMETER,
    STATE(3287), 1,
      aux_sym_name_repeat1,
    STATE(4399), 1,
      sym_generic_map_aspect,
    STATE(4412), 1,
      sym__name_selector,
    STATE(3281), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4425), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1471), 6,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
  [205189] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3282), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4230), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [205226] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3283), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205263] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3284), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3360), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [205300] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3285), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205337] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3286), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205374] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4312), 1,
      anon_sym_DOT,
    ACTIONS(4314), 1,
      anon_sym_LPAREN,
    ACTIONS(4316), 1,
      anon_sym_SQUOTE,
    ACTIONS(4318), 1,
      anon_sym_LBRACK,
    ACTIONS(4320), 1,
      sym_GENERIC,
    ACTIONS(4322), 1,
      sym_PARAMETER,
    STATE(3306), 1,
      aux_sym_name_repeat1,
    STATE(4399), 1,
      sym_generic_map_aspect,
    STATE(4412), 1,
      sym__name_selector,
    STATE(3287), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4425), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1487), 6,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
  [205431] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3288), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3366), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [205468] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3289), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205505] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3290), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205542] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3291), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205579] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3292), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205616] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3293), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205653] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3294), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205690] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3295), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205727] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3296), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3302), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [205764] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3297), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3304), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [205801] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3298), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3284), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [205838] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3299), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3428), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [205875] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3300), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205912] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3301), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205949] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3302), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [205986] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3303), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206023] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3304), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206060] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3305), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206097] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4324), 1,
      anon_sym_DOT,
    ACTIONS(4327), 1,
      anon_sym_LPAREN,
    ACTIONS(4330), 1,
      anon_sym_SQUOTE,
    ACTIONS(4333), 1,
      anon_sym_LBRACK,
    ACTIONS(4336), 1,
      sym_GENERIC,
    ACTIONS(4339), 1,
      sym_PARAMETER,
    STATE(4399), 1,
      sym_generic_map_aspect,
    STATE(4412), 1,
      sym__name_selector,
    STATE(3306), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4425), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1491), 6,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
  [206152] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3307), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206189] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3308), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3750), 20,
      sym_identifier,
      sym_ASSERT,
      sym_BLOCK,
      sym_CASE,
      sym_COMPONENT,
      sym_CONFIGURATION,
      sym_ENTITY,
      sym_FOR,
      sym_IF,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [206226] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4342), 1,
      anon_sym_LT_LT,
    STATE(3641), 1,
      sym__direct_name,
    STATE(4388), 1,
      sym_component_specification,
    STATE(4642), 1,
      sym_name,
    STATE(5910), 1,
      sym__label,
    STATE(6709), 1,
      sym_instantiation_list,
    ACTIONS(4346), 2,
      sym_ALL,
      sym_OTHERS,
    STATE(4861), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4348), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3309), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4866), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4344), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [206283] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3310), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206320] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3311), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206357] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3312), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206394] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3313), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206431] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3314), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206468] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3315), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206505] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3316), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3498), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [206542] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(4306), 1,
      anon_sym_LPAREN,
    STATE(3013), 1,
      sym__identifier,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4362), 1,
      sym__external_name,
    STATE(5815), 1,
      sym_record_element_resolution,
    STATE(6060), 1,
      sym_name,
    STATE(7092), 1,
      sym__element_resolution,
    STATE(7090), 2,
      sym_resolution_indication,
      sym_record_resolution,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3317), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4308), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [206601] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3318), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3288), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [206638] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3319), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206675] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3320), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206712] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3321), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4350), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [206749] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3322), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 20,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [206786] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3323), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3356), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [206823] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3324), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3358), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [206860] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3325), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3362), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [206897] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3326), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3366), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [206934] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3327), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3370), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [206971] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3328), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3388), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207008] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3329), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3428), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207045] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3330), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3430), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207082] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3331), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3434), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207119] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3332), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3436), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207156] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3333), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3440), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207193] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3334), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3456), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207230] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3335), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3460), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207267] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3336), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3464), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207304] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3337), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3466), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207341] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3338), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3468), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207378] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3339), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3480), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207415] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3340), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3494), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207452] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3341), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3496), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207489] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3342), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3246), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207526] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3343), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3248), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207563] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3344), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4244), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207600] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3345), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3252), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207637] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3346), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3254), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207674] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3347), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3256), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207711] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3348), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3266), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207748] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3349), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3280), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207785] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3350), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3282), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207822] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3351), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3284), 20,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_COMPONENT,
      sym_CONSTANT,
      sym_DISCONNECT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SIGNAL,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
      sym_VIEW,
  [207859] = 16,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4352), 1,
      sym_ALIAS,
    ACTIONS(4355), 1,
      sym_ATTRIBUTE,
    ACTIONS(4358), 1,
      sym_END,
    ACTIONS(4360), 1,
      sym_FUNCTION,
    ACTIONS(4366), 1,
      sym_PROCEDURE,
    ACTIONS(4369), 1,
      sym_PRIVATE,
    ACTIONS(4372), 1,
      sym_USE,
    STATE(4770), 1,
      sym__protected_type_declarative_item,
    STATE(7550), 1,
      sym__subprogram_specification,
    ACTIONS(4363), 2,
      sym_IMPURE,
      sym_PURE,
    STATE(6833), 2,
      sym_procedure_specification,
      sym_function_specification,
    STATE(3352), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_protected_type_declaration_repeat1,
    STATE(4768), 6,
      sym_use_clause,
      sym_subprogram_declaration,
      sym_subprogram_instantiation_declaration,
      sym_alias_declaration,
      sym_attribute_specification,
      sym_private_variable_declaration,
  [207918] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3353), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [207954] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3354), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [207990] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3355), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208026] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3356), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208062] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3357), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208098] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3358), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208134] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3359), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208170] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3360), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208206] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3361), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208242] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4375), 1,
      anon_sym_LPAREN,
    ACTIONS(4377), 1,
      sym_ACCESS,
    ACTIONS(4379), 1,
      sym_ARRAY,
    ACTIONS(4381), 1,
      sym_FILE,
    ACTIONS(4383), 1,
      sym_NEW,
    ACTIONS(4385), 1,
      sym_PROTECTED,
    ACTIONS(4387), 1,
      sym_RANGE,
    ACTIONS(4389), 1,
      sym_RECORD,
    STATE(6532), 1,
      sym_range_constraint,
    STATE(7586), 1,
      sym__type_definition,
    STATE(3362), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(8156), 9,
      sym_array_type_definition,
      sym_access_type_definition,
      sym_file_type_definition,
      sym_protected_type_instantiation_definition,
      sym_protected_type_declaration,
      sym_protected_type_body,
      sym_physical_type_definition,
      sym_enumeration_type_definition,
      sym_record_type_definition,
  [208298] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5825), 1,
      sym_name,
    STATE(7360), 1,
      sym_guarded_signal_specification,
    STATE(7372), 1,
      sym_signal_list,
    ACTIONS(4391), 2,
      sym_ALL,
      sym_OTHERS,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3363), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [208352] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3364), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208388] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4375), 1,
      anon_sym_LPAREN,
    ACTIONS(4377), 1,
      sym_ACCESS,
    ACTIONS(4379), 1,
      sym_ARRAY,
    ACTIONS(4381), 1,
      sym_FILE,
    ACTIONS(4383), 1,
      sym_NEW,
    ACTIONS(4385), 1,
      sym_PROTECTED,
    ACTIONS(4387), 1,
      sym_RANGE,
    ACTIONS(4389), 1,
      sym_RECORD,
    STATE(6532), 1,
      sym_range_constraint,
    STATE(8154), 1,
      sym__type_definition,
    STATE(3365), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(8156), 9,
      sym_array_type_definition,
      sym_access_type_definition,
      sym_file_type_definition,
      sym_protected_type_instantiation_definition,
      sym_protected_type_declaration,
      sym_protected_type_body,
      sym_physical_type_definition,
      sym_enumeration_type_definition,
      sym_record_type_definition,
  [208444] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(4393), 1,
      anon_sym_QMARK,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6256), 1,
      sym__target,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    STATE(6098), 2,
      sym_name,
      sym_aggregate,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3366), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [208498] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3367), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208534] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3368), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208570] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3369), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208606] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4375), 1,
      anon_sym_LPAREN,
    ACTIONS(4377), 1,
      sym_ACCESS,
    ACTIONS(4379), 1,
      sym_ARRAY,
    ACTIONS(4381), 1,
      sym_FILE,
    ACTIONS(4383), 1,
      sym_NEW,
    ACTIONS(4385), 1,
      sym_PROTECTED,
    ACTIONS(4387), 1,
      sym_RANGE,
    ACTIONS(4389), 1,
      sym_RECORD,
    STATE(6532), 1,
      sym_range_constraint,
    STATE(8270), 1,
      sym__type_definition,
    STATE(3370), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(8156), 9,
      sym_array_type_definition,
      sym_access_type_definition,
      sym_file_type_definition,
      sym_protected_type_instantiation_definition,
      sym_protected_type_declaration,
      sym_protected_type_body,
      sym_physical_type_definition,
      sym_enumeration_type_definition,
      sym_record_type_definition,
  [208662] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3371), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208698] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3372), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208734] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4375), 1,
      anon_sym_LPAREN,
    ACTIONS(4377), 1,
      sym_ACCESS,
    ACTIONS(4379), 1,
      sym_ARRAY,
    ACTIONS(4381), 1,
      sym_FILE,
    ACTIONS(4383), 1,
      sym_NEW,
    ACTIONS(4385), 1,
      sym_PROTECTED,
    ACTIONS(4387), 1,
      sym_RANGE,
    ACTIONS(4389), 1,
      sym_RECORD,
    STATE(6532), 1,
      sym_range_constraint,
    STATE(6913), 1,
      sym__type_definition,
    STATE(3373), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(8156), 9,
      sym_array_type_definition,
      sym_access_type_definition,
      sym_file_type_definition,
      sym_protected_type_instantiation_definition,
      sym_protected_type_declaration,
      sym_protected_type_body,
      sym_physical_type_definition,
      sym_enumeration_type_definition,
      sym_record_type_definition,
  [208790] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3374), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208826] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3375), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208862] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4375), 1,
      anon_sym_LPAREN,
    ACTIONS(4377), 1,
      sym_ACCESS,
    ACTIONS(4379), 1,
      sym_ARRAY,
    ACTIONS(4381), 1,
      sym_FILE,
    ACTIONS(4383), 1,
      sym_NEW,
    ACTIONS(4385), 1,
      sym_PROTECTED,
    ACTIONS(4387), 1,
      sym_RANGE,
    ACTIONS(4389), 1,
      sym_RECORD,
    STATE(6532), 1,
      sym_range_constraint,
    STATE(7595), 1,
      sym__type_definition,
    STATE(3376), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(8156), 9,
      sym_array_type_definition,
      sym_access_type_definition,
      sym_file_type_definition,
      sym_protected_type_instantiation_definition,
      sym_protected_type_declaration,
      sym_protected_type_body,
      sym_physical_type_definition,
      sym_enumeration_type_definition,
      sym_record_type_definition,
  [208918] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3377), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208954] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3378), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [208990] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3379), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209026] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3380), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209062] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4395), 1,
      anon_sym_DOT,
    ACTIONS(4397), 1,
      anon_sym_LPAREN,
    ACTIONS(4399), 1,
      anon_sym_SQUOTE,
    ACTIONS(4401), 1,
      anon_sym_LBRACK,
    ACTIONS(4403), 1,
      sym_GENERIC,
    ACTIONS(4405), 1,
      sym_PARAMETER,
    STATE(3428), 1,
      aux_sym_name_repeat1,
    STATE(4432), 1,
      sym_generic_map_aspect,
    STATE(4447), 1,
      sym__name_selector,
    STATE(3381), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1471), 5,
      sym_BUS,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(4450), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [209118] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3382), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209154] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3383), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209190] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3384), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209226] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3385), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209262] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4407), 1,
      anon_sym_DOT,
    ACTIONS(4409), 1,
      anon_sym_LPAREN,
    ACTIONS(4411), 1,
      anon_sym_SQUOTE,
    ACTIONS(4413), 1,
      anon_sym_LBRACK,
    ACTIONS(4415), 1,
      sym_PARAMETER,
    STATE(3389), 1,
      aux_sym_name_repeat1,
    STATE(4431), 1,
      sym_generic_map_aspect,
    STATE(4515), 1,
      sym__name_selector,
    STATE(3386), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4517), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1471), 6,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
  [209316] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3387), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209352] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3388), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209388] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4407), 1,
      anon_sym_DOT,
    ACTIONS(4409), 1,
      anon_sym_LPAREN,
    ACTIONS(4411), 1,
      anon_sym_SQUOTE,
    ACTIONS(4413), 1,
      anon_sym_LBRACK,
    ACTIONS(4415), 1,
      sym_PARAMETER,
    STATE(3432), 1,
      aux_sym_name_repeat1,
    STATE(4431), 1,
      sym_generic_map_aspect,
    STATE(4515), 1,
      sym__name_selector,
    STATE(3389), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4517), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
    ACTIONS(1487), 6,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
  [209442] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3390), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209478] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3391), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209514] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3392), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209550] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5825), 1,
      sym_name,
    STATE(7372), 1,
      sym_signal_list,
    STATE(7782), 1,
      sym_guarded_signal_specification,
    ACTIONS(4391), 2,
      sym_ALL,
      sym_OTHERS,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3393), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [209604] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4417), 1,
      anon_sym_DOT,
    ACTIONS(4419), 1,
      anon_sym_LPAREN,
    ACTIONS(4421), 1,
      anon_sym_SQUOTE,
    ACTIONS(4423), 1,
      anon_sym_LBRACK,
    ACTIONS(4425), 1,
      sym_GENERIC,
    ACTIONS(4427), 1,
      sym_PARAMETER,
    STATE(3397), 1,
      aux_sym_name_repeat1,
    STATE(4441), 1,
      sym_generic_map_aspect,
    STATE(4455), 1,
      sym__name_selector,
    STATE(3394), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1471), 5,
      sym_BUS,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    STATE(4456), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [209660] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3395), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209696] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3396), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209732] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4417), 1,
      anon_sym_DOT,
    ACTIONS(4419), 1,
      anon_sym_LPAREN,
    ACTIONS(4421), 1,
      anon_sym_SQUOTE,
    ACTIONS(4423), 1,
      anon_sym_LBRACK,
    ACTIONS(4425), 1,
      sym_GENERIC,
    ACTIONS(4427), 1,
      sym_PARAMETER,
    STATE(3420), 1,
      aux_sym_name_repeat1,
    STATE(4441), 1,
      sym_generic_map_aspect,
    STATE(4455), 1,
      sym__name_selector,
    STATE(3397), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1487), 5,
      sym_BUS,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    STATE(4456), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [209788] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3398), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209824] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3399), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209860] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3400), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209896] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3401), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209932] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3402), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [209968] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3403), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210004] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3404), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210040] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(4429), 1,
      sym_POSTPONED,
    ACTIONS(4431), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7270), 1,
      sym_name,
    STATE(8063), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3405), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [210096] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3406), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210132] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3407), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210168] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3408), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210204] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3409), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210240] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3410), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210276] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3411), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210312] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3412), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210348] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3413), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210384] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3414), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210420] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4116), 1,
      anon_sym_LPAREN,
    ACTIONS(4433), 1,
      sym_PARAMETER,
    STATE(3532), 1,
      sym_parenthesis_group,
    STATE(3415), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210462] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4435), 1,
      anon_sym_DOT,
    ACTIONS(4438), 1,
      anon_sym_LPAREN,
    ACTIONS(4441), 1,
      anon_sym_SQUOTE,
    ACTIONS(4444), 1,
      anon_sym_LBRACK,
    ACTIONS(4447), 1,
      sym_GENERIC,
    ACTIONS(4450), 1,
      sym_PARAMETER,
    STATE(4432), 1,
      sym_generic_map_aspect,
    STATE(4447), 1,
      sym__name_selector,
    STATE(3416), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    ACTIONS(1491), 5,
      sym_BUS,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(4450), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [210516] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3417), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210552] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3418), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210588] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3419), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210624] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4453), 1,
      anon_sym_DOT,
    ACTIONS(4456), 1,
      anon_sym_LPAREN,
    ACTIONS(4459), 1,
      anon_sym_SQUOTE,
    ACTIONS(4462), 1,
      anon_sym_LBRACK,
    ACTIONS(4465), 1,
      sym_GENERIC,
    ACTIONS(4468), 1,
      sym_PARAMETER,
    STATE(4441), 1,
      sym_generic_map_aspect,
    STATE(4455), 1,
      sym__name_selector,
    STATE(3420), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    ACTIONS(1491), 5,
      sym_BUS,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    STATE(4456), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [210678] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3421), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210714] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3422), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210750] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3423), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210786] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3424), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210822] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3425), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210858] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3426), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210894] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3427), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 19,
      sym_identifier,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [210930] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4395), 1,
      anon_sym_DOT,
    ACTIONS(4397), 1,
      anon_sym_LPAREN,
    ACTIONS(4399), 1,
      anon_sym_SQUOTE,
    ACTIONS(4401), 1,
      anon_sym_LBRACK,
    ACTIONS(4403), 1,
      sym_GENERIC,
    ACTIONS(4405), 1,
      sym_PARAMETER,
    STATE(3416), 1,
      aux_sym_name_repeat1,
    STATE(4432), 1,
      sym_generic_map_aspect,
    STATE(4447), 1,
      sym__name_selector,
    STATE(3428), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1487), 5,
      sym_BUS,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(4450), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [210986] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4375), 1,
      anon_sym_LPAREN,
    ACTIONS(4377), 1,
      sym_ACCESS,
    ACTIONS(4379), 1,
      sym_ARRAY,
    ACTIONS(4381), 1,
      sym_FILE,
    ACTIONS(4383), 1,
      sym_NEW,
    ACTIONS(4385), 1,
      sym_PROTECTED,
    ACTIONS(4387), 1,
      sym_RANGE,
    ACTIONS(4389), 1,
      sym_RECORD,
    STATE(6532), 1,
      sym_range_constraint,
    STATE(8283), 1,
      sym__type_definition,
    STATE(3429), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(8156), 9,
      sym_array_type_definition,
      sym_access_type_definition,
      sym_file_type_definition,
      sym_protected_type_instantiation_definition,
      sym_protected_type_declaration,
      sym_protected_type_body,
      sym_physical_type_definition,
      sym_enumeration_type_definition,
      sym_record_type_definition,
  [211042] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5825), 1,
      sym_name,
    STATE(7372), 1,
      sym_signal_list,
    STATE(7661), 1,
      sym_guarded_signal_specification,
    ACTIONS(4391), 2,
      sym_ALL,
      sym_OTHERS,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3430), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211096] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5825), 1,
      sym_name,
    STATE(7372), 1,
      sym_signal_list,
    STATE(7724), 1,
      sym_guarded_signal_specification,
    ACTIONS(4391), 2,
      sym_ALL,
      sym_OTHERS,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3431), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211150] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4471), 1,
      anon_sym_DOT,
    ACTIONS(4474), 1,
      anon_sym_LPAREN,
    ACTIONS(4477), 1,
      anon_sym_SQUOTE,
    ACTIONS(4480), 1,
      anon_sym_LBRACK,
    ACTIONS(4483), 1,
      sym_GENERIC,
    ACTIONS(4486), 1,
      sym_PARAMETER,
    STATE(4431), 1,
      sym_generic_map_aspect,
    STATE(4515), 1,
      sym__name_selector,
    STATE(3432), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    ACTIONS(1491), 5,
      sym_END,
      sym_FOR,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
    STATE(4517), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [211204] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4489), 1,
      anon_sym_LPAREN,
    STATE(2149), 1,
      sym__attribute_designator,
    STATE(2150), 1,
      sym_parenthesis_expression,
    STATE(2219), 1,
      sym__attribute,
    STATE(3433), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4491), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4493), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [211249] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(7878), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3434), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211302] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(5857), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3435), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211355] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4501), 1,
      anon_sym_LT_LT,
    STATE(2849), 1,
      sym__direct_name,
    STATE(4135), 1,
      sym_resolution_indication,
    STATE(4389), 1,
      sym_name,
    STATE(5541), 1,
      sym_subtype_indication,
    STATE(3518), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4505), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3436), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3520), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4503), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211408] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4507), 1,
      anon_sym_LPAREN,
    STATE(2157), 1,
      sym__attribute_designator,
    STATE(2158), 1,
      sym_parenthesis_expression,
    STATE(2182), 1,
      sym__attribute,
    STATE(3437), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4509), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4511), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [211453] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4513), 1,
      anon_sym_LT_LT,
    STATE(2799), 1,
      sym__direct_name,
    STATE(4143), 1,
      sym_resolution_indication,
    STATE(4344), 1,
      sym_name,
    STATE(5109), 1,
      sym_subtype_indication,
    STATE(3388), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4517), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3391), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3438), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4515), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211506] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7584), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3439), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211559] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5253), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3440), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211612] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(6838), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3441), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211665] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5306), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3442), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211718] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(5719), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3443), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211771] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4519), 1,
      anon_sym_DOT,
    ACTIONS(4522), 1,
      anon_sym_LPAREN,
    ACTIONS(4525), 1,
      anon_sym_SQUOTE,
    ACTIONS(4528), 1,
      anon_sym_LBRACK,
    ACTIONS(4531), 1,
      sym_GENERIC,
    ACTIONS(4534), 1,
      sym_PARAMETER,
    STATE(4481), 1,
      sym_generic_map_aspect,
    STATE(4588), 1,
      sym__name_selector,
    ACTIONS(1491), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    STATE(3444), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4591), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [211824] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(5389), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3445), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211877] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(5919), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3446), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211930] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4501), 1,
      anon_sym_LT_LT,
    STATE(2849), 1,
      sym__direct_name,
    STATE(4135), 1,
      sym_resolution_indication,
    STATE(4389), 1,
      sym_name,
    STATE(5551), 1,
      sym_subtype_indication,
    STATE(3518), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4505), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3447), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3520), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4503), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [211983] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4513), 1,
      anon_sym_LT_LT,
    STATE(2799), 1,
      sym__direct_name,
    STATE(4143), 1,
      sym_resolution_indication,
    STATE(4344), 1,
      sym_name,
    STATE(5138), 1,
      sym_subtype_indication,
    STATE(3388), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4517), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3391), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3448), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4515), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212036] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8152), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3449), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212089] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3450), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [212124] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5164), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3451), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212177] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4537), 1,
      anon_sym_LPAREN,
    STATE(856), 1,
      sym__attribute_designator,
    STATE(858), 1,
      sym_parenthesis_expression,
    STATE(869), 1,
      sym__attribute,
    STATE(3452), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4539), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4541), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [212222] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5183), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3453), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212275] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4543), 1,
      anon_sym_DOT,
    ACTIONS(4545), 1,
      anon_sym_LPAREN,
    ACTIONS(4547), 1,
      anon_sym_SQUOTE,
    ACTIONS(4549), 1,
      anon_sym_LBRACK,
    ACTIONS(4551), 1,
      sym_GENERIC,
    ACTIONS(4553), 1,
      sym_PARAMETER,
    STATE(3490), 1,
      aux_sym_name_repeat1,
    STATE(4481), 1,
      sym_generic_map_aspect,
    STATE(4588), 1,
      sym__name_selector,
    STATE(3454), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1471), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    STATE(4591), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [212330] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5260), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3455), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212383] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(4555), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6988), 1,
      sym_name,
    STATE(8105), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3456), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212436] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4557), 1,
      anon_sym_LPAREN,
    STATE(2236), 1,
      sym__attribute_designator,
    STATE(2237), 1,
      sym_parenthesis_expression,
    STATE(2253), 1,
      sym__attribute,
    STATE(3457), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4559), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4561), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [212481] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4563), 1,
      anon_sym_LPAREN,
    STATE(4341), 1,
      sym__attribute_designator,
    STATE(4343), 1,
      sym_parenthesis_expression,
    STATE(4380), 1,
      sym__attribute,
    STATE(3458), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4565), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4567), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [212526] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(5563), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3459), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212579] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4501), 1,
      anon_sym_LT_LT,
    STATE(2849), 1,
      sym__direct_name,
    STATE(4135), 1,
      sym_resolution_indication,
    STATE(4389), 1,
      sym_name,
    STATE(5333), 1,
      sym_subtype_indication,
    STATE(3518), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4505), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3460), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3520), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4503), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212632] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4513), 1,
      anon_sym_LT_LT,
    STATE(2799), 1,
      sym__direct_name,
    STATE(4143), 1,
      sym_resolution_indication,
    STATE(4344), 1,
      sym_name,
    STATE(5094), 1,
      sym_subtype_indication,
    STATE(3388), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4517), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3391), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3461), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4515), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212685] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7651), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3462), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212738] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5167), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3463), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212791] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5209), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3464), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212844] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(3824), 1,
      anon_sym_LT_LT,
    STATE(2806), 1,
      sym__direct_name,
    STATE(4177), 1,
      sym_resolution_indication,
    STATE(4365), 1,
      sym_name,
    STATE(5522), 1,
      sym_subtype_indication,
    STATE(3364), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3834), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3367), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3465), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3826), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212897] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4569), 1,
      anon_sym_LPAREN,
    STATE(4894), 1,
      sym_parenthesis_expression,
    STATE(4927), 1,
      sym__attribute_designator,
    STATE(4949), 1,
      sym__attribute,
    STATE(3466), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4571), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4573), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [212942] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(5562), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3467), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [212995] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4501), 1,
      anon_sym_LT_LT,
    STATE(2849), 1,
      sym__direct_name,
    STATE(4135), 1,
      sym_resolution_indication,
    STATE(4389), 1,
      sym_name,
    STATE(5336), 1,
      sym_subtype_indication,
    STATE(3518), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4505), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3468), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3520), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4503), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213048] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6895), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3469), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213101] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5270), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3470), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213154] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5208), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3471), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213207] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(6051), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3472), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213260] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4575), 1,
      anon_sym_LPAREN,
    ACTIONS(4577), 1,
      sym_ACCESS,
    ACTIONS(4579), 1,
      sym_ARRAY,
    ACTIONS(4581), 1,
      sym_FILE,
    ACTIONS(4583), 1,
      sym_PRIVATE,
    ACTIONS(4585), 1,
      sym_RANGE,
    ACTIONS(4587), 1,
      sym_UNITS,
    ACTIONS(4589), 1,
      sym_box,
    STATE(5523), 1,
      sym__incomplete_type_definition,
    STATE(3473), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(5518), 9,
      sym_private_incomplete_type_definition,
      sym_scalar_incomplete_type_definition,
      sym_discrete_incomplete_type_definition,
      sym_integer_incomplete_type_definition,
      sym_physical_incomplete_type_definition,
      sym_floating_incomplete_type_definition,
      sym_array_type_definition,
      sym_access_incomplete_type_definition,
      sym_file_incomplete_type_definition,
  [213313] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4501), 1,
      anon_sym_LT_LT,
    STATE(2849), 1,
      sym__direct_name,
    STATE(4135), 1,
      sym_resolution_indication,
    STATE(4389), 1,
      sym_name,
    STATE(5498), 1,
      sym_subtype_indication,
    STATE(3518), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4505), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3474), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3520), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4503), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213366] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(5637), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3475), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213419] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4501), 1,
      anon_sym_LT_LT,
    STATE(2849), 1,
      sym__direct_name,
    STATE(4135), 1,
      sym_resolution_indication,
    STATE(4389), 1,
      sym_name,
    STATE(5341), 1,
      sym_subtype_indication,
    STATE(3518), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4505), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3476), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3520), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4503), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213472] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(7581), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3477), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213525] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4591), 1,
      anon_sym_LPAREN,
    STATE(4406), 1,
      sym__attribute_designator,
    STATE(4422), 1,
      sym_parenthesis_expression,
    STATE(4436), 1,
      sym__attribute,
    STATE(3478), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4593), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4595), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [213570] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5160), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3479), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213623] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5311), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3480), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213676] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4597), 1,
      anon_sym_DOT,
    ACTIONS(4599), 1,
      anon_sym_LPAREN,
    ACTIONS(4601), 1,
      anon_sym_SQUOTE,
    ACTIONS(4603), 1,
      anon_sym_LBRACK,
    ACTIONS(4605), 1,
      sym_GENERIC,
    ACTIONS(4607), 1,
      sym_PARAMETER,
    STATE(3486), 1,
      aux_sym_name_repeat1,
    STATE(4491), 1,
      sym_generic_map_aspect,
    STATE(4640), 1,
      sym__name_selector,
    STATE(3481), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1471), 4,
      sym_FOR,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(4641), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [213731] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4609), 1,
      anon_sym_LPAREN,
    STATE(4594), 1,
      sym__attribute_designator,
    STATE(4595), 1,
      sym_parenthesis_expression,
    STATE(4637), 1,
      sym__attribute,
    STATE(3482), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4611), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4613), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [213776] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6339), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3483), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213829] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5300), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3484), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213882] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5172), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3485), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [213935] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4597), 1,
      anon_sym_DOT,
    ACTIONS(4599), 1,
      anon_sym_LPAREN,
    ACTIONS(4601), 1,
      anon_sym_SQUOTE,
    ACTIONS(4603), 1,
      anon_sym_LBRACK,
    ACTIONS(4605), 1,
      sym_GENERIC,
    ACTIONS(4607), 1,
      sym_PARAMETER,
    STATE(3576), 1,
      aux_sym_name_repeat1,
    STATE(4491), 1,
      sym_generic_map_aspect,
    STATE(4640), 1,
      sym__name_selector,
    STATE(3486), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1487), 4,
      sym_FOR,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(4641), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [213990] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3487), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [214025] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4379), 1,
      sym_ARRAY,
    ACTIONS(4615), 1,
      anon_sym_LPAREN,
    ACTIONS(4617), 1,
      sym_ACCESS,
    ACTIONS(4619), 1,
      sym_FILE,
    ACTIONS(4621), 1,
      sym_PRIVATE,
    ACTIONS(4623), 1,
      sym_RANGE,
    ACTIONS(4625), 1,
      sym_UNITS,
    ACTIONS(4627), 1,
      sym_box,
    STATE(6536), 1,
      sym__incomplete_type_definition,
    STATE(3488), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(5364), 9,
      sym_private_incomplete_type_definition,
      sym_scalar_incomplete_type_definition,
      sym_discrete_incomplete_type_definition,
      sym_integer_incomplete_type_definition,
      sym_physical_incomplete_type_definition,
      sym_floating_incomplete_type_definition,
      sym_array_type_definition,
      sym_access_incomplete_type_definition,
      sym_file_incomplete_type_definition,
  [214078] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4629), 1,
      anon_sym_LPAREN,
    STATE(4557), 1,
      sym__attribute_designator,
    STATE(4558), 1,
      sym_parenthesis_expression,
    STATE(4609), 1,
      sym__attribute,
    STATE(3489), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4631), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4633), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214123] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4543), 1,
      anon_sym_DOT,
    ACTIONS(4545), 1,
      anon_sym_LPAREN,
    ACTIONS(4547), 1,
      anon_sym_SQUOTE,
    ACTIONS(4549), 1,
      anon_sym_LBRACK,
    ACTIONS(4551), 1,
      sym_GENERIC,
    ACTIONS(4553), 1,
      sym_PARAMETER,
    STATE(3444), 1,
      aux_sym_name_repeat1,
    STATE(4481), 1,
      sym_generic_map_aspect,
    STATE(4588), 1,
      sym__name_selector,
    STATE(3490), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1487), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    STATE(4591), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [214178] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4635), 1,
      anon_sym_LPAREN,
    STATE(1685), 1,
      sym__attribute,
    STATE(1708), 1,
      sym__attribute_designator,
    STATE(1729), 1,
      sym_parenthesis_expression,
    STATE(3491), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4637), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4639), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214223] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4641), 1,
      anon_sym_LPAREN,
    STATE(4488), 1,
      sym__attribute_designator,
    STATE(4494), 1,
      sym_parenthesis_expression,
    STATE(4544), 1,
      sym__attribute,
    STATE(3492), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4643), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4645), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214268] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4647), 1,
      anon_sym_DOT,
    ACTIONS(4649), 1,
      anon_sym_LPAREN,
    ACTIONS(4651), 1,
      anon_sym_SQUOTE,
    ACTIONS(4653), 1,
      anon_sym_LBRACK,
    ACTIONS(4655), 1,
      sym_GENERIC,
    ACTIONS(4657), 1,
      sym_PARAMETER,
    STATE(3511), 1,
      aux_sym_name_repeat1,
    STATE(4461), 1,
      sym_generic_map_aspect,
    STATE(4629), 1,
      sym__name_selector,
    STATE(3493), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1471), 4,
      sym_IS,
      sym_OPEN,
      sym_RANGE,
      anon_sym_SEMI,
    STATE(4633), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [214323] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4659), 1,
      anon_sym_LPAREN,
    STATE(2122), 1,
      sym__attribute_designator,
    STATE(2123), 1,
      sym_parenthesis_expression,
    STATE(2143), 1,
      sym__attribute,
    STATE(3494), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4661), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4663), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214368] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4665), 1,
      anon_sym_LPAREN,
    STATE(1344), 1,
      sym__attribute_designator,
    STATE(1345), 1,
      sym_parenthesis_expression,
    STATE(1355), 1,
      sym__attribute,
    STATE(3495), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4669), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214413] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4671), 1,
      anon_sym_LPAREN,
    STATE(1362), 1,
      sym__attribute_designator,
    STATE(1363), 1,
      sym_parenthesis_expression,
    STATE(1373), 1,
      sym__attribute,
    STATE(3496), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4673), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4675), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214458] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4677), 1,
      anon_sym_LPAREN,
    STATE(1484), 1,
      sym__attribute_designator,
    STATE(1485), 1,
      sym_parenthesis_expression,
    STATE(1502), 1,
      sym__attribute,
    STATE(3497), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4679), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4681), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214503] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4683), 1,
      anon_sym_LPAREN,
    STATE(1510), 1,
      sym__attribute_designator,
    STATE(1511), 1,
      sym_parenthesis_expression,
    STATE(1527), 1,
      sym__attribute,
    STATE(3498), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4685), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4687), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214548] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4689), 1,
      anon_sym_LPAREN,
    STATE(929), 1,
      sym__attribute_designator,
    STATE(974), 1,
      sym_parenthesis_expression,
    STATE(988), 1,
      sym__attribute,
    STATE(3499), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4691), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4693), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214593] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6857), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3500), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [214646] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4695), 1,
      anon_sym_LPAREN,
    STATE(1141), 1,
      sym__attribute,
    STATE(1233), 1,
      sym__attribute_designator,
    STATE(1235), 1,
      sym_parenthesis_expression,
    STATE(3501), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4697), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4699), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214691] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4701), 1,
      anon_sym_LPAREN,
    STATE(3375), 1,
      sym__attribute_designator,
    STATE(3377), 1,
      sym_parenthesis_expression,
    STATE(3387), 1,
      sym__attribute,
    STATE(3502), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4703), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4705), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214736] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4707), 1,
      anon_sym_LPAREN,
    STATE(1568), 1,
      sym__attribute_designator,
    STATE(1569), 1,
      sym_parenthesis_expression,
    STATE(1597), 1,
      sym__attribute,
    STATE(3503), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4709), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4711), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214781] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4713), 1,
      anon_sym_LPAREN,
    STATE(4718), 1,
      sym__attribute_designator,
    STATE(4738), 1,
      sym_parenthesis_expression,
    STATE(4758), 1,
      sym__attribute,
    STATE(3504), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4715), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4717), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214826] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4379), 1,
      sym_ARRAY,
    ACTIONS(4615), 1,
      anon_sym_LPAREN,
    ACTIONS(4617), 1,
      sym_ACCESS,
    ACTIONS(4619), 1,
      sym_FILE,
    ACTIONS(4621), 1,
      sym_PRIVATE,
    ACTIONS(4623), 1,
      sym_RANGE,
    ACTIONS(4625), 1,
      sym_UNITS,
    ACTIONS(4627), 1,
      sym_box,
    STATE(5398), 1,
      sym__incomplete_type_definition,
    STATE(3505), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(5364), 9,
      sym_private_incomplete_type_definition,
      sym_scalar_incomplete_type_definition,
      sym_discrete_incomplete_type_definition,
      sym_integer_incomplete_type_definition,
      sym_physical_incomplete_type_definition,
      sym_floating_incomplete_type_definition,
      sym_array_type_definition,
      sym_access_incomplete_type_definition,
      sym_file_incomplete_type_definition,
  [214879] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4719), 1,
      anon_sym_LPAREN,
    STATE(1572), 1,
      sym__attribute_designator,
    STATE(1573), 1,
      sym_parenthesis_expression,
    STATE(1601), 1,
      sym__attribute,
    STATE(3506), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4721), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4723), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214924] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4725), 1,
      anon_sym_LPAREN,
    STATE(1061), 1,
      sym__attribute,
    STATE(1111), 1,
      sym__attribute_designator,
    STATE(1112), 1,
      sym_parenthesis_expression,
    STATE(3507), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4727), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4729), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [214969] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(19), 1,
      sym_ASSERT,
    ACTIONS(4431), 1,
      sym_PROCESS,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7270), 1,
      sym_name,
    STATE(8063), 1,
      sym_assertion,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3508), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [215022] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4731), 1,
      anon_sym_LPAREN,
    STATE(1087), 1,
      sym__attribute_designator,
    STATE(1088), 1,
      sym_parenthesis_expression,
    STATE(1105), 1,
      sym__attribute,
    STATE(3509), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4733), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4735), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [215067] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4737), 1,
      anon_sym_LPAREN,
    STATE(887), 1,
      sym_parenthesis_expression,
    STATE(913), 1,
      sym__attribute_designator,
    STATE(923), 1,
      sym__attribute,
    STATE(3510), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4739), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4741), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [215112] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4647), 1,
      anon_sym_DOT,
    ACTIONS(4649), 1,
      anon_sym_LPAREN,
    ACTIONS(4651), 1,
      anon_sym_SQUOTE,
    ACTIONS(4653), 1,
      anon_sym_LBRACK,
    ACTIONS(4655), 1,
      sym_GENERIC,
    ACTIONS(4657), 1,
      sym_PARAMETER,
    STATE(3570), 1,
      aux_sym_name_repeat1,
    STATE(4461), 1,
      sym_generic_map_aspect,
    STATE(4629), 1,
      sym__name_selector,
    STATE(3511), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1487), 4,
      sym_IS,
      sym_OPEN,
      sym_RANGE,
      anon_sym_SEMI,
    STATE(4633), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [215167] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4743), 1,
      anon_sym_LPAREN,
    STATE(1821), 1,
      sym__attribute_designator,
    STATE(1822), 1,
      sym_parenthesis_expression,
    STATE(1841), 1,
      sym__attribute,
    STATE(3512), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4745), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4747), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [215212] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4749), 1,
      anon_sym_LPAREN,
    STATE(1851), 1,
      sym__attribute_designator,
    STATE(1852), 1,
      sym_parenthesis_expression,
    STATE(1877), 1,
      sym__attribute,
    STATE(3513), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4751), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4753), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [215257] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(281), 1,
      sym_TYPE,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5468), 1,
      sym_unspecified_type_indication,
    STATE(5469), 1,
      sym__incomplete_type_mark,
    STATE(5470), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3514), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [215310] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6541), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3515), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [215363] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4755), 1,
      anon_sym_LPAREN,
    STATE(4659), 1,
      sym__attribute_designator,
    STATE(4661), 1,
      sym_parenthesis_expression,
    STATE(4690), 1,
      sym__attribute,
    STATE(3516), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4757), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4759), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [215408] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4761), 1,
      anon_sym_LPAREN,
    STATE(4764), 1,
      sym__attribute,
    STATE(4769), 1,
      sym__attribute_designator,
    STATE(4783), 1,
      sym_parenthesis_expression,
    STATE(3517), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4763), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4765), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [215453] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3518), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [215488] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4767), 1,
      anon_sym_LPAREN,
    STATE(1462), 1,
      sym__attribute_designator,
    STATE(1463), 1,
      sym_parenthesis_expression,
    STATE(1580), 1,
      sym__attribute,
    STATE(3519), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4769), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4771), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [215533] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3520), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [215568] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4773), 1,
      anon_sym_LPAREN,
    STATE(4733), 1,
      sym__attribute_designator,
    STATE(4734), 1,
      sym_parenthesis_expression,
    STATE(4765), 1,
      sym__attribute,
    STATE(3521), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4775), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4777), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [215613] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3522), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [215648] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3523), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [215683] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4779), 1,
      anon_sym_LPAREN,
    STATE(4445), 1,
      sym__attribute,
    STATE(4465), 1,
      sym__attribute_designator,
    STATE(4467), 1,
      sym_parenthesis_expression,
    STATE(3524), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4781), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4783), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [215728] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4785), 1,
      anon_sym_LPAREN,
    STATE(1384), 1,
      sym__attribute,
    STATE(1403), 1,
      sym__attribute_designator,
    STATE(1404), 1,
      sym_parenthesis_expression,
    STATE(3525), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4787), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4789), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [215773] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3526), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [215808] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3527), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [215843] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3528), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [215878] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3529), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [215913] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3530), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [215948] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3531), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [215983] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3532), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216018] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4791), 1,
      anon_sym_LPAREN,
    STATE(3409), 1,
      sym__attribute_designator,
    STATE(3410), 1,
      sym_parenthesis_expression,
    STATE(3426), 1,
      sym__attribute,
    STATE(3533), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4793), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4795), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [216063] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3534), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216098] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3535), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216133] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3536), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216168] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3537), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216203] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3538), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216238] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3539), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216273] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4797), 1,
      anon_sym_LPAREN,
    STATE(1248), 1,
      sym__attribute_designator,
    STATE(1249), 1,
      sym_parenthesis_expression,
    STATE(1258), 1,
      sym__attribute,
    STATE(3540), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4799), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4801), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [216318] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9), 1,
      anon_sym_LPAREN,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6253), 1,
      sym__target,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    STATE(6098), 2,
      sym_name,
      sym_aggregate,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3541), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [216369] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3542), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216404] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(4803), 1,
      sym_ALL,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5846), 1,
      sym_name,
    STATE(7424), 1,
      sym_sensitivity_list,
    STATE(7471), 1,
      sym__process_sensitivity_list,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3543), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [216457] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3544), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216492] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4805), 1,
      anon_sym_LPAREN,
    STATE(1007), 1,
      sym__attribute_designator,
    STATE(1008), 1,
      sym_parenthesis_expression,
    STATE(1017), 1,
      sym__attribute,
    STATE(3545), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4807), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4809), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [216537] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3546), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216572] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3547), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216607] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3548), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216642] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3549), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216677] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3550), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [216712] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4811), 1,
      anon_sym_LPAREN,
    STATE(1991), 1,
      sym__attribute_designator,
    STATE(1992), 1,
      sym_parenthesis_expression,
    STATE(2009), 1,
      sym__attribute,
    STATE(3551), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4813), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4815), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [216757] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4817), 1,
      anon_sym_LPAREN,
    STATE(1125), 1,
      sym__attribute_designator,
    STATE(1156), 1,
      sym_parenthesis_expression,
    STATE(1194), 1,
      sym__attribute,
    STATE(3552), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4819), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4821), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [216802] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4823), 1,
      anon_sym_LPAREN,
    STATE(1287), 1,
      sym__attribute_designator,
    STATE(1288), 1,
      sym_parenthesis_expression,
    STATE(1298), 1,
      sym__attribute,
    STATE(3553), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4825), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4827), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [216847] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4829), 1,
      anon_sym_LPAREN,
    STATE(3694), 1,
      sym__attribute,
    STATE(3703), 1,
      sym_parenthesis_expression,
    STATE(3782), 1,
      sym__attribute_designator,
    STATE(3554), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4831), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4833), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [216892] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4835), 1,
      anon_sym_LPAREN,
    STATE(4656), 1,
      sym__attribute_designator,
    STATE(4850), 1,
      sym__attribute,
    STATE(4891), 1,
      sym_parenthesis_expression,
    STATE(3555), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4837), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4839), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [216937] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4841), 1,
      anon_sym_LPAREN,
    STATE(1795), 1,
      sym__attribute_designator,
    STATE(1859), 1,
      sym__attribute,
    STATE(1878), 1,
      sym_parenthesis_expression,
    STATE(3556), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4843), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4845), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [216982] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4847), 1,
      anon_sym_LPAREN,
    STATE(4564), 1,
      sym__attribute_designator,
    STATE(4574), 1,
      sym_parenthesis_expression,
    STATE(4603), 1,
      sym__attribute,
    STATE(3557), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4849), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4851), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [217027] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4853), 1,
      anon_sym_LPAREN,
    STATE(935), 1,
      sym__attribute_designator,
    STATE(936), 1,
      sym_parenthesis_expression,
    STATE(959), 1,
      sym__attribute,
    STATE(3558), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4855), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4857), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [217072] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4859), 1,
      anon_sym_LPAREN,
    STATE(3293), 1,
      sym__attribute_designator,
    STATE(3294), 1,
      sym_parenthesis_expression,
    STATE(3307), 1,
      sym__attribute,
    STATE(3559), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4861), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4863), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [217117] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5193), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3560), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [217170] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4865), 1,
      anon_sym_LPAREN,
    STATE(4975), 1,
      sym__attribute,
    STATE(4981), 1,
      sym__attribute_designator,
    STATE(4983), 1,
      sym_parenthesis_expression,
    STATE(3561), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4867), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4869), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [217215] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(6596), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3562), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [217268] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3563), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [217303] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4871), 1,
      anon_sym_LPAREN,
    STATE(3530), 1,
      sym__attribute_designator,
    STATE(3531), 1,
      sym_parenthesis_expression,
    STATE(3542), 1,
      sym__attribute,
    STATE(3564), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4873), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4875), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [217348] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4877), 1,
      anon_sym_LPAREN,
    STATE(4740), 1,
      sym__attribute_designator,
    STATE(4741), 1,
      sym_parenthesis_expression,
    STATE(4844), 1,
      sym__attribute,
    STATE(3565), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4879), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4881), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [217393] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3566), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [217428] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3567), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 18,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [217463] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4883), 1,
      anon_sym_LPAREN,
    STATE(2094), 1,
      sym__attribute_designator,
    STATE(2095), 1,
      sym_parenthesis_expression,
    STATE(2111), 1,
      sym__attribute,
    STATE(3568), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4885), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4887), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [217508] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4889), 1,
      anon_sym_LPAREN,
    STATE(4521), 1,
      sym__attribute_designator,
    STATE(4522), 1,
      sym_parenthesis_expression,
    STATE(4540), 1,
      sym__attribute,
    STATE(3569), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4891), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4893), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [217553] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4895), 1,
      anon_sym_DOT,
    ACTIONS(4898), 1,
      anon_sym_LPAREN,
    ACTIONS(4901), 1,
      anon_sym_SQUOTE,
    ACTIONS(4904), 1,
      anon_sym_LBRACK,
    ACTIONS(4907), 1,
      sym_GENERIC,
    ACTIONS(4910), 1,
      sym_PARAMETER,
    STATE(4461), 1,
      sym_generic_map_aspect,
    STATE(4629), 1,
      sym__name_selector,
    ACTIONS(1491), 4,
      sym_IS,
      sym_OPEN,
      sym_RANGE,
      anon_sym_SEMI,
    STATE(3570), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4633), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [217606] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3830), 1,
      sym_TYPE,
    ACTIONS(4913), 1,
      anon_sym_LT_LT,
    STATE(3454), 1,
      sym__direct_name,
    STATE(5528), 1,
      sym_unspecified_type_indication,
    STATE(5529), 1,
      sym__incomplete_type_mark,
    STATE(5530), 1,
      sym_name,
    STATE(4623), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4917), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3571), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4562), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4915), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [217659] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5738), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3572), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [217712] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2874), 1,
      sym__direct_name,
    STATE(4163), 1,
      sym_resolution_indication,
    STATE(4391), 1,
      sym_name,
    STATE(5740), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3573), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [217765] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4513), 1,
      anon_sym_LT_LT,
    STATE(2799), 1,
      sym__direct_name,
    STATE(4143), 1,
      sym_resolution_indication,
    STATE(4344), 1,
      sym_name,
    STATE(5116), 1,
      sym_subtype_indication,
    STATE(3388), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4517), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3391), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3574), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4515), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [217818] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4264), 1,
      anon_sym_LT_LT,
    STATE(2794), 1,
      sym__direct_name,
    STATE(4124), 1,
      sym_resolution_indication,
    STATE(4252), 1,
      sym_name,
    STATE(8087), 1,
      sym_subtype_indication,
    STATE(3278), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4268), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3280), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(3575), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4266), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [217871] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4919), 1,
      anon_sym_DOT,
    ACTIONS(4922), 1,
      anon_sym_LPAREN,
    ACTIONS(4925), 1,
      anon_sym_SQUOTE,
    ACTIONS(4928), 1,
      anon_sym_LBRACK,
    ACTIONS(4931), 1,
      sym_GENERIC,
    ACTIONS(4934), 1,
      sym_PARAMETER,
    STATE(4491), 1,
      sym_generic_map_aspect,
    STATE(4640), 1,
      sym__name_selector,
    ACTIONS(1491), 4,
      sym_FOR,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(3576), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4641), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [217924] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(7696), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3577), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [217977] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(6396), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3578), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218030] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(7758), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3579), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218083] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(6427), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3580), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218136] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(7811), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3581), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218189] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(6451), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3582), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218242] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(7846), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3583), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218295] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3822), 1,
      anon_sym_LPAREN,
    ACTIONS(4495), 1,
      anon_sym_LT_LT,
    STATE(2990), 1,
      sym__direct_name,
    STATE(4236), 1,
      sym_resolution_indication,
    STATE(4506), 1,
      sym_name,
    STATE(7866), 1,
      sym_subtype_indication,
    STATE(3800), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4499), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3584), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(3802), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4497), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218348] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4937), 1,
      anon_sym_LPAREN,
    STATE(1744), 1,
      sym__attribute_designator,
    STATE(1745), 1,
      sym_parenthesis_expression,
    STATE(1773), 1,
      sym__attribute,
    STATE(3585), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4939), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(4941), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [218393] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(4945), 1,
      anon_sym_RBRACK,
    ACTIONS(4949), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5448), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3586), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218443] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4953), 1,
      anon_sym_DOT,
    ACTIONS(4956), 1,
      anon_sym_LPAREN,
    ACTIONS(4959), 1,
      anon_sym_SQUOTE,
    ACTIONS(4962), 1,
      anon_sym_LBRACK,
    ACTIONS(4965), 1,
      sym_GENERIC,
    ACTIONS(4968), 1,
      sym_PARAMETER,
    STATE(4561), 1,
      sym_generic_map_aspect,
    STATE(4797), 1,
      sym__name_selector,
    ACTIONS(1491), 3,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
    STATE(3587), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4805), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [218495] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(4306), 1,
      anon_sym_LPAREN,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6060), 1,
      sym_name,
    STATE(6212), 1,
      sym_resolution_indication,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3588), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218545] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(4971), 1,
      anon_sym_RBRACK,
    ACTIONS(4973), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5401), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3589), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218595] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(4975), 1,
      anon_sym_RBRACK,
    ACTIONS(4977), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5433), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3590), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218645] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(4979), 1,
      anon_sym_RBRACK,
    ACTIONS(4981), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5452), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3591), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218695] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(4983), 1,
      anon_sym_RBRACK,
    ACTIONS(4985), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5384), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3592), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218745] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4987), 1,
      anon_sym_DOT,
    ACTIONS(4990), 1,
      anon_sym_LPAREN,
    ACTIONS(4993), 1,
      anon_sym_SQUOTE,
    ACTIONS(4996), 1,
      anon_sym_LBRACK,
    ACTIONS(4999), 1,
      sym_GENERIC,
    ACTIONS(5002), 1,
      sym_PARAMETER,
    STATE(4646), 1,
      sym_generic_map_aspect,
    STATE(4723), 1,
      sym__name_selector,
    ACTIONS(1491), 3,
      sym_OF,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(3593), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4726), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [218797] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5005), 1,
      anon_sym_RBRACK,
    ACTIONS(5007), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5550), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3594), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218847] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5009), 1,
      anon_sym_DOT,
    ACTIONS(5012), 1,
      anon_sym_LPAREN,
    ACTIONS(5015), 1,
      anon_sym_SQUOTE,
    ACTIONS(5018), 1,
      anon_sym_LBRACK,
    ACTIONS(5021), 1,
      sym_GENERIC,
    ACTIONS(5024), 1,
      sym_PARAMETER,
    STATE(4655), 1,
      sym_generic_map_aspect,
    STATE(4706), 1,
      sym__name_selector,
    ACTIONS(1491), 3,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
    STATE(3595), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4708), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [218899] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5027), 1,
      anon_sym_RBRACK,
    ACTIONS(5029), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5412), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3596), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218949] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5031), 1,
      anon_sym_RBRACK,
    ACTIONS(5033), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5339), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3597), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [218999] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5035), 1,
      anon_sym_RBRACK,
    ACTIONS(5037), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5435), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3598), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219049] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5039), 1,
      anon_sym_RBRACK,
    ACTIONS(5041), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5393), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3599), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219099] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5043), 1,
      anon_sym_RBRACK,
    ACTIONS(5045), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5437), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3600), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219149] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5047), 1,
      anon_sym_RBRACK,
    ACTIONS(5049), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5414), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3601), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219199] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5051), 1,
      anon_sym_RBRACK,
    ACTIONS(5053), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5399), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3602), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219249] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5055), 1,
      anon_sym_DOT,
    ACTIONS(5057), 1,
      anon_sym_LPAREN,
    ACTIONS(5059), 1,
      anon_sym_SQUOTE,
    ACTIONS(5061), 1,
      anon_sym_LBRACK,
    ACTIONS(5063), 1,
      sym_PARAMETER,
    STATE(3615), 1,
      aux_sym_name_repeat1,
    STATE(4655), 1,
      sym_generic_map_aspect,
    STATE(4706), 1,
      sym__name_selector,
    STATE(3603), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1471), 4,
      sym_GENERIC,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
    STATE(4708), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [219301] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5065), 1,
      anon_sym_RBRACK,
    ACTIONS(5067), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5351), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3604), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219351] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5069), 1,
      anon_sym_RBRACK,
    ACTIONS(5071), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5454), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3605), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219401] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5073), 1,
      anon_sym_RBRACK,
    ACTIONS(5075), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5418), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3606), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219451] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5077), 1,
      anon_sym_RBRACK,
    ACTIONS(5079), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5456), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3607), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219501] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5081), 1,
      anon_sym_RBRACK,
    ACTIONS(5083), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5439), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3608), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219551] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5085), 1,
      anon_sym_DOT,
    ACTIONS(5087), 1,
      anon_sym_LPAREN,
    ACTIONS(5089), 1,
      anon_sym_SQUOTE,
    ACTIONS(5091), 1,
      anon_sym_LBRACK,
    ACTIONS(5093), 1,
      sym_GENERIC,
    ACTIONS(5095), 1,
      sym_PARAMETER,
    STATE(3667), 1,
      aux_sym_name_repeat1,
    STATE(4639), 1,
      sym_generic_map_aspect,
    STATE(4869), 1,
      sym__name_selector,
    ACTIONS(1471), 3,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    STATE(3609), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4871), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [219605] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3610), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5097), 17,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_IS,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [219639] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5099), 1,
      anon_sym_RBRACK,
    ACTIONS(5101), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5353), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3611), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219689] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5103), 1,
      anon_sym_RBRACK,
    ACTIONS(5105), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5458), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3612), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219739] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5107), 1,
      anon_sym_RBRACK,
    ACTIONS(5109), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5540), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3613), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219789] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5111), 1,
      anon_sym_RBRACK,
    ACTIONS(5113), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5443), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3614), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219839] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5055), 1,
      anon_sym_DOT,
    ACTIONS(5057), 1,
      anon_sym_LPAREN,
    ACTIONS(5059), 1,
      anon_sym_SQUOTE,
    ACTIONS(5061), 1,
      anon_sym_LBRACK,
    ACTIONS(5063), 1,
      sym_PARAMETER,
    STATE(3595), 1,
      aux_sym_name_repeat1,
    STATE(4655), 1,
      sym_generic_map_aspect,
    STATE(4706), 1,
      sym__name_selector,
    STATE(3615), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1487), 4,
      sym_GENERIC,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
    STATE(4708), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [219891] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5115), 1,
      anon_sym_RBRACK,
    ACTIONS(5117), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5356), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3616), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [219941] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7144), 1,
      sym__attribute,
    STATE(7184), 1,
      sym__identifier,
    STATE(7232), 1,
      sym__attribute_designator,
    STATE(3617), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5119), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(5121), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [219983] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5123), 1,
      anon_sym_RBRACK,
    ACTIONS(5125), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5421), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3618), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220033] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5127), 1,
      anon_sym_RBRACK,
    ACTIONS(5129), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5361), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3619), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220083] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5131), 1,
      anon_sym_RBRACK,
    ACTIONS(5133), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5344), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3620), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220133] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5135), 1,
      anon_sym_DOT,
    ACTIONS(5137), 1,
      anon_sym_LPAREN,
    ACTIONS(5139), 1,
      anon_sym_SQUOTE,
    ACTIONS(5141), 1,
      anon_sym_LBRACK,
    ACTIONS(5143), 1,
      sym_PARAMETER,
    STATE(3587), 1,
      aux_sym_name_repeat1,
    STATE(4561), 1,
      sym_generic_map_aspect,
    STATE(4797), 1,
      sym__name_selector,
    STATE(3621), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1487), 4,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
    STATE(4805), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [220185] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5145), 1,
      anon_sym_RBRACK,
    ACTIONS(5147), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5365), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3622), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220235] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5149), 1,
      anon_sym_DOT,
    ACTIONS(5151), 1,
      anon_sym_LPAREN,
    ACTIONS(5153), 1,
      anon_sym_SQUOTE,
    ACTIONS(5155), 1,
      anon_sym_LBRACK,
    ACTIONS(5157), 1,
      sym_GENERIC,
    ACTIONS(5159), 1,
      sym_PARAMETER,
    STATE(3662), 1,
      aux_sym_name_repeat1,
    STATE(4618), 1,
      sym_generic_map_aspect,
    STATE(4875), 1,
      sym__name_selector,
    ACTIONS(1487), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(3623), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4878), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [220289] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5161), 1,
      anon_sym_RBRACK,
    ACTIONS(5163), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5368), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3624), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220339] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5165), 1,
      anon_sym_DOT,
    ACTIONS(5167), 1,
      anon_sym_LPAREN,
    ACTIONS(5169), 1,
      anon_sym_SQUOTE,
    ACTIONS(5171), 1,
      anon_sym_LBRACK,
    ACTIONS(5173), 1,
      sym_GENERIC,
    ACTIONS(5175), 1,
      sym_PARAMETER,
    STATE(3627), 1,
      aux_sym_name_repeat1,
    STATE(4646), 1,
      sym_generic_map_aspect,
    STATE(4723), 1,
      sym__name_selector,
    ACTIONS(1471), 3,
      sym_OF,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(3625), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4726), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [220393] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5177), 1,
      anon_sym_RBRACK,
    ACTIONS(5179), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5445), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3626), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220443] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5165), 1,
      anon_sym_DOT,
    ACTIONS(5167), 1,
      anon_sym_LPAREN,
    ACTIONS(5169), 1,
      anon_sym_SQUOTE,
    ACTIONS(5171), 1,
      anon_sym_LBRACK,
    ACTIONS(5173), 1,
      sym_GENERIC,
    ACTIONS(5175), 1,
      sym_PARAMETER,
    STATE(3593), 1,
      aux_sym_name_repeat1,
    STATE(4646), 1,
      sym_generic_map_aspect,
    STATE(4723), 1,
      sym__name_selector,
    ACTIONS(1487), 3,
      sym_OF,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(3627), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4726), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [220497] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5181), 1,
      anon_sym_RBRACK,
    ACTIONS(5183), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5372), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3628), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220547] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5185), 1,
      anon_sym_RBRACK,
    ACTIONS(5187), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5390), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3629), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220597] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5189), 1,
      anon_sym_RBRACK,
    ACTIONS(5191), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5431), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3630), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220647] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5193), 1,
      anon_sym_RBRACK,
    ACTIONS(5195), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5374), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3631), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220697] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5197), 1,
      anon_sym_RBRACK,
    ACTIONS(5199), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5395), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3632), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220747] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5201), 1,
      anon_sym_RBRACK,
    ACTIONS(5203), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5378), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3633), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220797] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5205), 1,
      anon_sym_RBRACK,
    ACTIONS(5207), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5460), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3634), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220847] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5209), 1,
      anon_sym_RBRACK,
    ACTIONS(5211), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5332), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3635), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220897] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5135), 1,
      anon_sym_DOT,
    ACTIONS(5137), 1,
      anon_sym_LPAREN,
    ACTIONS(5139), 1,
      anon_sym_SQUOTE,
    ACTIONS(5141), 1,
      anon_sym_LBRACK,
    ACTIONS(5143), 1,
      sym_PARAMETER,
    STATE(3621), 1,
      aux_sym_name_repeat1,
    STATE(4561), 1,
      sym_generic_map_aspect,
    STATE(4797), 1,
      sym__name_selector,
    STATE(3636), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1471), 4,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_LT_EQ,
    STATE(4805), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [220949] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(5213), 1,
      sym_box,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6280), 1,
      sym_name,
    STATE(6281), 1,
      sym__interface_subprogram_default,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3637), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [220999] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5215), 1,
      anon_sym_RBRACK,
    ACTIONS(5217), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5381), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3638), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221049] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5219), 1,
      anon_sym_DOT,
    ACTIONS(5222), 1,
      anon_sym_LPAREN,
    ACTIONS(5225), 1,
      anon_sym_SQUOTE,
    ACTIONS(5228), 1,
      anon_sym_LBRACK,
    ACTIONS(5231), 1,
      sym_GENERIC,
    ACTIONS(5234), 1,
      sym_PARAMETER,
    STATE(4587), 1,
      sym_generic_map_aspect,
    STATE(4827), 1,
      sym__name_selector,
    ACTIONS(1491), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(3639), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4847), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [221101] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5237), 1,
      anon_sym_DOT,
    ACTIONS(5239), 1,
      anon_sym_LPAREN,
    ACTIONS(5241), 1,
      anon_sym_SQUOTE,
    ACTIONS(5243), 1,
      anon_sym_LBRACK,
    ACTIONS(5245), 1,
      sym_GENERIC,
    ACTIONS(5247), 1,
      sym_PARAMETER,
    STATE(3639), 1,
      aux_sym_name_repeat1,
    STATE(4587), 1,
      sym_generic_map_aspect,
    STATE(4827), 1,
      sym__name_selector,
    ACTIONS(1487), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(3640), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4847), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [221155] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5149), 1,
      anon_sym_DOT,
    ACTIONS(5151), 1,
      anon_sym_LPAREN,
    ACTIONS(5153), 1,
      anon_sym_SQUOTE,
    ACTIONS(5155), 1,
      anon_sym_LBRACK,
    ACTIONS(5157), 1,
      sym_GENERIC,
    ACTIONS(5159), 1,
      sym_PARAMETER,
    STATE(3623), 1,
      aux_sym_name_repeat1,
    STATE(4618), 1,
      sym_generic_map_aspect,
    STATE(4875), 1,
      sym__name_selector,
    ACTIONS(1471), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(3641), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4878), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [221209] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4312), 1,
      anon_sym_DOT,
    ACTIONS(4314), 1,
      anon_sym_LPAREN,
    ACTIONS(4316), 1,
      anon_sym_SQUOTE,
    ACTIONS(4318), 1,
      anon_sym_LBRACK,
    ACTIONS(4322), 1,
      sym_PARAMETER,
    STATE(3646), 1,
      aux_sym_name_repeat1,
    STATE(4399), 1,
      sym_generic_map_aspect,
    STATE(4412), 1,
      sym__name_selector,
    STATE(3642), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1471), 4,
      sym_GENERIC,
      sym_RANGE,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(4425), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [221261] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5249), 1,
      anon_sym_RBRACK,
    ACTIONS(5251), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5406), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3643), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221311] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5253), 1,
      anon_sym_RBRACK,
    ACTIONS(5255), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5424), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3644), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221361] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5257), 1,
      anon_sym_RBRACK,
    ACTIONS(5259), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5380), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3645), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221411] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4312), 1,
      anon_sym_DOT,
    ACTIONS(4314), 1,
      anon_sym_LPAREN,
    ACTIONS(4316), 1,
      anon_sym_SQUOTE,
    ACTIONS(4318), 1,
      anon_sym_LBRACK,
    ACTIONS(4322), 1,
      sym_PARAMETER,
    STATE(3306), 1,
      aux_sym_name_repeat1,
    STATE(4399), 1,
      sym_generic_map_aspect,
    STATE(4412), 1,
      sym__name_selector,
    STATE(3646), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1487), 4,
      sym_GENERIC,
      sym_RANGE,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(4425), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [221463] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4184), 1,
      anon_sym_LPAREN,
    ACTIONS(5261), 1,
      sym_PARAMETER,
    STATE(3701), 1,
      sym_parenthesis_group,
    STATE(3647), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 14,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [221503] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5263), 1,
      anon_sym_RBRACK,
    ACTIONS(5265), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5427), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3648), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221553] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5267), 1,
      anon_sym_RBRACK,
    ACTIONS(5269), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5408), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3649), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221603] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5271), 1,
      anon_sym_RBRACK,
    ACTIONS(5273), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5462), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3650), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221653] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7144), 1,
      sym__attribute,
    STATE(7659), 1,
      sym__identifier,
    STATE(8271), 1,
      sym__attribute_designator,
    STATE(3651), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5119), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(5121), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [221695] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5275), 1,
      anon_sym_RBRACK,
    ACTIONS(5277), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5464), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3652), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221745] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5279), 1,
      anon_sym_RBRACK,
    ACTIONS(5281), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5450), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3653), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221795] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5283), 1,
      anon_sym_RBRACK,
    ACTIONS(5285), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5347), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3654), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221845] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7144), 1,
      sym__attribute,
    STATE(7722), 1,
      sym__identifier,
    STATE(8279), 1,
      sym__attribute_designator,
    STATE(3655), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5119), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(5121), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [221887] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5287), 1,
      anon_sym_RBRACK,
    ACTIONS(5289), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5466), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3656), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221937] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5291), 1,
      anon_sym_RBRACK,
    ACTIONS(5293), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5403), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3657), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [221987] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5295), 1,
      anon_sym_RBRACK,
    ACTIONS(5297), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5387), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3658), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [222037] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7144), 1,
      sym__attribute,
    STATE(7780), 1,
      sym__identifier,
    STATE(8287), 1,
      sym__attribute_designator,
    STATE(3659), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5119), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(5121), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [222079] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5299), 1,
      anon_sym_DOT,
    ACTIONS(5302), 1,
      anon_sym_LPAREN,
    ACTIONS(5305), 1,
      anon_sym_SQUOTE,
    ACTIONS(5308), 1,
      anon_sym_LBRACK,
    ACTIONS(5311), 1,
      sym_GENERIC,
    ACTIONS(5314), 1,
      sym_PARAMETER,
    STATE(4639), 1,
      sym_generic_map_aspect,
    STATE(4869), 1,
      sym__name_selector,
    ACTIONS(1491), 3,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    STATE(3660), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4871), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [222131] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5317), 1,
      anon_sym_RBRACK,
    ACTIONS(5319), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5335), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3661), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [222181] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5321), 1,
      anon_sym_DOT,
    ACTIONS(5324), 1,
      anon_sym_LPAREN,
    ACTIONS(5327), 1,
      anon_sym_SQUOTE,
    ACTIONS(5330), 1,
      anon_sym_LBRACK,
    ACTIONS(5333), 1,
      sym_GENERIC,
    ACTIONS(5336), 1,
      sym_PARAMETER,
    STATE(4618), 1,
      sym_generic_map_aspect,
    STATE(4875), 1,
      sym__name_selector,
    ACTIONS(1491), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(3662), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4878), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [222233] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7144), 1,
      sym__attribute,
    STATE(7824), 1,
      sym__identifier,
    STATE(8295), 1,
      sym__attribute_designator,
    STATE(3663), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5119), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(5121), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [222275] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5339), 1,
      anon_sym_RBRACK,
    ACTIONS(5341), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5429), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3664), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [222325] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7144), 1,
      sym__attribute,
    STATE(7854), 1,
      sym__identifier,
    STATE(8300), 1,
      sym__attribute_designator,
    STATE(3665), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5119), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(5121), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [222367] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    ACTIONS(5343), 1,
      anon_sym_RBRACK,
    ACTIONS(5345), 1,
      sym_RETURN,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5410), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3666), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [222417] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5085), 1,
      anon_sym_DOT,
    ACTIONS(5087), 1,
      anon_sym_LPAREN,
    ACTIONS(5089), 1,
      anon_sym_SQUOTE,
    ACTIONS(5091), 1,
      anon_sym_LBRACK,
    ACTIONS(5093), 1,
      sym_GENERIC,
    ACTIONS(5095), 1,
      sym_PARAMETER,
    STATE(3660), 1,
      aux_sym_name_repeat1,
    STATE(4639), 1,
      sym_generic_map_aspect,
    STATE(4869), 1,
      sym__name_selector,
    ACTIONS(1487), 3,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    STATE(3667), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4871), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [222471] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5237), 1,
      anon_sym_DOT,
    ACTIONS(5239), 1,
      anon_sym_LPAREN,
    ACTIONS(5241), 1,
      anon_sym_SQUOTE,
    ACTIONS(5243), 1,
      anon_sym_LBRACK,
    ACTIONS(5245), 1,
      sym_GENERIC,
    ACTIONS(5247), 1,
      sym_PARAMETER,
    STATE(3640), 1,
      aux_sym_name_repeat1,
    STATE(4587), 1,
      sym_generic_map_aspect,
    STATE(4827), 1,
      sym__name_selector,
    ACTIONS(1471), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(3668), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4847), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [222525] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3669), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3456), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [222558] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3670), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [222591] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3671), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3350), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [222624] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3672), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3352), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [222657] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3673), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3354), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [222690] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3674), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [222723] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3675), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3470), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [222756] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3676), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [222789] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3677), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [222822] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3678), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [222855] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3679), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3258), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [222888] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3680), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3394), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [222921] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3681), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [222954] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3682), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3396), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [222987] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3683), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3320), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223020] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3684), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [223053] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3685), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3438), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223086] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3686), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5347), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223119] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3687), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5349), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223152] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3688), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3446), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223185] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3689), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3450), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223218] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3690), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3452), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223251] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3691), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3454), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223284] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3692), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5351), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223317] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3693), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3498), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223350] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3694), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [223383] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3695), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3440), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223416] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3696), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3474), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223449] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3697), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3476), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223482] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3698), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3466), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223515] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3699), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3478), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223548] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3700), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5353), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223581] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3701), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [223614] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3702), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3238), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223647] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3703), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [223680] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7144), 1,
      sym__attribute,
    STATE(8305), 1,
      sym__attribute_designator,
    STATE(3704), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5355), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(5121), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [223719] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3705), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3498), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223752] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3706), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3288), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223785] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3707), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3288), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223818] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3708), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3260), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223851] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3709), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3262), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223884] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3710), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3292), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [223917] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3711), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3264), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [223950] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7324), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3712), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [224003] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3713), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3356), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224036] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3714), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3358), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224069] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3715), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3362), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224102] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3716), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3370), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224135] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3717), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3388), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224168] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3718), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3430), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224201] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3719), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3434), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224234] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3720), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3436), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224267] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3721), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3460), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224300] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3722), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3464), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224333] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3723), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3468), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224366] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3724), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3480), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224399] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3725), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3372), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [224432] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3726), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3496), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224465] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3727), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [224498] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3728), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [224531] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3729), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3246), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224564] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3730), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3248), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224597] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3731), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3252), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224630] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3732), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3254), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224663] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3733), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3290), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [224696] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3734), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3256), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224729] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3735), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3266), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224762] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3736), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3282), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224795] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3737), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3294), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [224828] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3738), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3302), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224861] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3739), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3304), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224894] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3740), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3308), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224927] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3741), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3312), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [224960] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(6843), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3742), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [225013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3743), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3306), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [225046] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(6914), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3744), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [225099] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3745), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [225132] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3746), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3322), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [225165] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3747), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3324), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [225198] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3748), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3330), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [225231] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3749), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3292), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225264] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3750), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3356), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225297] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3751), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3358), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225330] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3752), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3362), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225363] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3753), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3370), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225396] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3754), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3388), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225429] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3755), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [225462] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5810), 1,
      sym_name,
    STATE(7301), 1,
      sym_verification_unit_list,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3756), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [225509] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3757), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3448), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [225542] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3758), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3472), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [225575] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3759), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3430), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225608] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3760), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3434), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225641] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3761), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3436), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225674] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3762), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3440), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225707] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3763), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3456), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225740] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3764), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3460), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225773] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3765), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3464), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225806] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5363), 1,
      anon_sym_DOT,
    ACTIONS(5366), 1,
      anon_sym_LPAREN,
    ACTIONS(5369), 1,
      anon_sym_SQUOTE,
    ACTIONS(5372), 1,
      anon_sym_LBRACK,
    ACTIONS(5375), 1,
      sym_GENERIC,
    ACTIONS(5378), 1,
      sym_PARAMETER,
    STATE(4792), 1,
      sym_generic_map_aspect,
    STATE(4920), 1,
      sym__name_selector,
    ACTIONS(1491), 2,
      sym_IS,
      sym_RANGE,
    STATE(3766), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(4926), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [225857] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3767), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3466), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225890] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3768), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3468), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225923] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3769), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3480), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [225956] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5831), 1,
      sym_name,
    STATE(7383), 1,
      sym_group_constituent_list,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3770), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [226003] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3771), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3386), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [226036] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5381), 1,
      anon_sym_LT_LT,
    STATE(3481), 1,
      sym__direct_name,
    STATE(5309), 1,
      sym_name,
    STATE(5714), 1,
      sym_sensitivity_list,
    STATE(4628), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5385), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3772), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4630), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5383), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [226083] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3773), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [226116] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3774), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [226149] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3775), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [226182] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(5387), 1,
      anon_sym_LPAREN,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6737), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3776), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [226229] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3777), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3484), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [226262] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3778), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3486), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [226295] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5389), 1,
      anon_sym_LPAREN,
    ACTIONS(5391), 1,
      anon_sym_LT_LT,
    STATE(3625), 1,
      sym__direct_name,
    STATE(5592), 1,
      sym_name,
    STATE(4719), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5395), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3779), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4720), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5393), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [226342] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3780), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3496), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226375] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3781), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [226408] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3782), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [226441] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3783), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3246), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226474] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3784), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3248), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226507] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3785), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5397), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226540] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3786), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5399), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_CONSTANT,
      sym_END,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226573] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3787), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3252), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226606] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3788), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3254), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226639] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    ACTIONS(5401), 1,
      sym_library_constant_unit,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6776), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3789), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [226686] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3790), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3270), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [226719] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3791), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3256), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226752] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3792), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5403), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226785] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3793), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5405), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226818] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3794), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3266), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226851] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3795), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3344), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [226884] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3796), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3346), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [226917] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3797), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3282), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [226950] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3798), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [226983] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3799), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3302), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [227016] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3800), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [227049] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5135), 1,
      anon_sym_DOT,
    ACTIONS(5139), 1,
      anon_sym_SQUOTE,
    ACTIONS(5141), 1,
      anon_sym_LBRACK,
    ACTIONS(5143), 1,
      sym_PARAMETER,
    STATE(3810), 1,
      aux_sym_name_repeat1,
    STATE(4561), 1,
      sym_generic_map_aspect,
    STATE(4797), 1,
      sym__name_selector,
    STATE(3801), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1471), 4,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_LPAREN,
    STATE(4805), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [227098] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3802), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [227131] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5407), 1,
      anon_sym_DOT,
    ACTIONS(5409), 1,
      anon_sym_LPAREN,
    ACTIONS(5411), 1,
      anon_sym_SQUOTE,
    ACTIONS(5413), 1,
      anon_sym_LBRACK,
    ACTIONS(5415), 1,
      sym_GENERIC,
    ACTIONS(5417), 1,
      sym_PARAMETER,
    STATE(3807), 1,
      aux_sym_name_repeat1,
    STATE(4792), 1,
      sym_generic_map_aspect,
    STATE(4920), 1,
      sym__name_selector,
    ACTIONS(1471), 2,
      sym_IS,
      sym_RANGE,
    STATE(3803), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4926), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [227184] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3804), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3304), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [227217] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3805), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3308), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [227250] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3806), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [227283] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5407), 1,
      anon_sym_DOT,
    ACTIONS(5409), 1,
      anon_sym_LPAREN,
    ACTIONS(5411), 1,
      anon_sym_SQUOTE,
    ACTIONS(5413), 1,
      anon_sym_LBRACK,
    ACTIONS(5415), 1,
      sym_GENERIC,
    ACTIONS(5417), 1,
      sym_PARAMETER,
    STATE(3766), 1,
      aux_sym_name_repeat1,
    STATE(4792), 1,
      sym_generic_map_aspect,
    STATE(4920), 1,
      sym__name_selector,
    ACTIONS(1487), 2,
      sym_IS,
      sym_RANGE,
    STATE(3807), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4926), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [227336] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3808), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [227369] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3809), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [227402] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5135), 1,
      anon_sym_DOT,
    ACTIONS(5139), 1,
      anon_sym_SQUOTE,
    ACTIONS(5141), 1,
      anon_sym_LBRACK,
    ACTIONS(5143), 1,
      sym_PARAMETER,
    STATE(3587), 1,
      aux_sym_name_repeat1,
    STATE(4561), 1,
      sym_generic_map_aspect,
    STATE(4797), 1,
      sym__name_selector,
    STATE(3810), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1487), 4,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_LPAREN,
    STATE(4805), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [227451] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5831), 1,
      sym_name,
    STATE(6803), 1,
      sym_group_constituent_list,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3811), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [227498] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3812), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3312), 16,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_BEGIN,
      sym_CONSTANT,
      sym_FILE,
      sym_FUNCTION,
      sym_GROUP,
      sym_IMPURE,
      sym_PACKAGE,
      sym_PROCEDURE,
      sym_PURE,
      sym_SHARED,
      sym_SUBTYPE,
      sym_TYPE,
      sym_USE,
      sym_VARIABLE,
  [227531] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5831), 1,
      sym_name,
    STATE(6918), 1,
      sym_group_constituent_list,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3813), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [227578] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3814), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3382), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [227611] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3815), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [227644] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3816), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [227677] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3817), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [227710] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5831), 1,
      sym_name,
    STATE(7002), 1,
      sym_group_constituent_list,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3818), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [227757] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3760), 1,
      anon_sym_LT_LT,
    ACTIONS(5419), 1,
      sym_library_namespace,
    STATE(3801), 1,
      sym__direct_name,
    STATE(5426), 1,
      sym_name,
    STATE(4717), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3782), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3819), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4725), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3762), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [227804] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5831), 1,
      sym_name,
    STATE(7045), 1,
      sym_group_constituent_list,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3820), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [227851] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5831), 1,
      sym_name,
    STATE(7075), 1,
      sym_group_constituent_list,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3821), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [227898] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5831), 1,
      sym_name,
    STATE(7096), 1,
      sym_group_constituent_list,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3822), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [227945] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7689), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3823), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [227998] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7690), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3824), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228051] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7691), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3825), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228104] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3826), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3390), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [228137] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7751), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3827), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228190] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7752), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3828), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228243] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7753), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3829), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228296] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7804), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3830), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228349] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7805), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3831), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228402] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7806), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3832), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228455] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7842), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3833), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228508] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7843), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3834), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228561] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7844), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3835), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228614] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7863), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3836), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228667] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7864), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3837), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228720] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7865), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3838), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228773] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3839), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3332), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [228806] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7875), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3840), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228859] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7876), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3841), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228912] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7877), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3842), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [228965] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3843), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3334), 16,
      sym_identifier,
      sym_ASSERT,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [228998] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7881), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3844), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229051] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7882), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3845), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229104] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7883), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3846), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229157] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7885), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3847), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229210] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7886), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3848), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229263] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7887), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3849), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229316] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7888), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3850), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229369] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7889), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3851), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229422] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7890), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3852), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229475] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7891), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3853), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229528] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7892), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3854), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229581] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7893), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3855), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229634] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7894), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3856), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229687] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7895), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3857), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229740] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7896), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3858), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229793] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7897), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3859), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229846] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7898), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3860), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229899] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7899), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3861), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [229952] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7900), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3862), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230005] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7901), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3863), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230058] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7902), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3864), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230111] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7903), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3865), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230164] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7904), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3866), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230217] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7905), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3867), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230270] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7906), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3868), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230323] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7907), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3869), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230376] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7908), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3870), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230429] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7909), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3871), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230482] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7910), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3872), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230535] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7911), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3873), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230588] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7912), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3874), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230641] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7913), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3875), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230694] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7914), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3876), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230747] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7915), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3877), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230800] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7916), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3878), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230853] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7917), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3879), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230906] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7918), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3880), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [230959] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7919), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3881), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231012] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7920), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3882), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231065] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7921), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3883), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231118] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7922), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3884), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231171] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7923), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3885), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231224] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7924), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3886), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231277] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7925), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3887), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231330] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7926), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3888), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231383] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7927), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3889), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231436] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7928), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3890), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231489] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7929), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3891), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231542] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7930), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3892), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231595] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7931), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3893), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231648] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7932), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3894), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231701] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7933), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3895), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231754] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7934), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3896), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231807] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7935), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3897), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231860] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7936), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3898), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231913] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7937), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3899), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [231966] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7938), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3900), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232019] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7939), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3901), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232072] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7940), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3902), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232125] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7941), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3903), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232178] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7942), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3904), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232231] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7943), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3905), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232284] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7944), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3906), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232337] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7945), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3907), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232390] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7946), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3908), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232443] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7947), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3909), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232496] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7948), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3910), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232549] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7949), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3911), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232602] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7950), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3912), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232655] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7951), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3913), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232708] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7952), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3914), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232761] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7953), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3915), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232814] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7954), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3916), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232867] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7955), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3917), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232920] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7956), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3918), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [232973] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7957), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3919), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233026] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7958), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3920), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233079] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7959), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3921), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233132] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7960), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3922), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233185] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7961), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3923), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233238] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7962), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3924), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233291] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7963), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3925), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233344] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7964), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3926), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233397] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7965), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3927), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233450] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7966), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3928), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233503] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7967), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3929), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233556] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7968), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3930), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233609] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7969), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3931), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233662] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7970), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3932), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233715] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7971), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3933), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233768] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7972), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3934), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233821] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7973), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3935), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233874] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7974), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3936), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233927] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7975), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3937), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [233980] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7976), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3938), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234033] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7977), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3939), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234086] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7978), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3940), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234139] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7979), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3941), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234192] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7980), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3942), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234245] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7981), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3943), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234298] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7982), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3944), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234351] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7983), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3945), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234404] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7984), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3946), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234457] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7985), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3947), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234510] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7986), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3948), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234563] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7987), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3949), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234616] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7988), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3950), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234669] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7989), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3951), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234722] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7990), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3952), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234775] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7991), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3953), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234828] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7992), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3954), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234881] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7993), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3955), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234934] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7994), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3956), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [234987] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7995), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3957), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235040] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7996), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3958), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235093] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7997), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3959), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235146] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7998), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3960), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235199] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7999), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3961), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235252] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8000), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3962), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235305] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8001), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3963), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235358] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8002), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3964), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235411] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8003), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3965), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235464] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8004), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3966), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235517] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8005), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3967), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235570] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8006), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3968), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235623] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8007), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3969), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235676] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8008), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3970), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235729] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8009), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3971), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235782] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8010), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3972), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235835] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8011), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3973), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235888] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8012), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3974), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235941] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5357), 1,
      anon_sym_DOT,
    ACTIONS(5359), 1,
      anon_sym_AT,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4632), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(8013), 1,
      sym__external_pathname,
    STATE(8181), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(3975), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(7408), 3,
      sym_package_pathname,
      sym_absolute_pathname,
      sym_relative_pathname,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [235994] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3976), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 16,
      sym_identifier,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [236027] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7144), 1,
      sym__attribute,
    STATE(8308), 1,
      sym__attribute_designator,
    STATE(3977), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5355), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    ACTIONS(5121), 10,
      sym_attribute_function,
      sym_attribute_impure_function,
      sym_attribute_mode_view,
      sym_attribute_pure_function,
      sym_attribute_range,
      sym_attribute_signal,
      sym_attribute_subtype,
      sym_attribute_type,
      sym_attribute_value,
      sym_library_attribute,
  [236066] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(6206), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3978), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236110] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3979), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3262), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [236142] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8288), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3980), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236186] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3981), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3264), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [236218] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8318), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3982), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236262] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8321), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3983), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236306] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6619), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3984), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236350] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6634), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3985), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236394] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6640), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3986), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236438] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6671), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3987), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236482] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3988), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3270), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [236514] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6704), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3989), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236558] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6707), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3990), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236602] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6745), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3991), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236646] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3992), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3478), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [236678] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6753), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3993), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236722] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6755), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3994), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236766] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6795), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3995), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236810] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6807), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3996), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236854] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6809), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3997), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236898] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6844), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(3998), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [236942] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3999), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3386), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [236974] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(5880), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4000), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237018] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6853), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4001), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237062] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6856), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4002), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237106] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6879), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4003), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237150] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6892), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4004), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237194] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6899), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4005), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237238] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(880), 1,
      anon_sym_LT_LT,
    STATE(786), 1,
      sym__direct_name,
    STATE(2566), 1,
      sym_name,
    STATE(1475), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(892), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1476), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4006), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(882), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237282] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6927), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4007), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237326] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7287), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4008), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237370] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6936), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4009), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237414] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6939), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4010), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237458] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4011), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3344), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [237490] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6967), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4012), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237534] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(5889), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4013), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237578] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4014), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3346), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [237610] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7289), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4015), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237654] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6975), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4016), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237698] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6978), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4017), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237742] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7011), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4018), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237786] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6779), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4019), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237830] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4251), 1,
      sym__direct_name,
    STATE(5700), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4020), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237874] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(966), 1,
      anon_sym_LT_LT,
    STATE(793), 1,
      sym__direct_name,
    STATE(2585), 1,
      sym_name,
    STATE(1499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(978), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1501), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4021), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(968), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237918] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7023), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4022), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [237962] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7027), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4023), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238006] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7063), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4024), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238050] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4025), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3390), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [238082] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4026), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3290), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [238114] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7086), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4027), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238158] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7089), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4028), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238202] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7126), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4029), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238246] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7153), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4030), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238290] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7156), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4031), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238334] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7194), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4032), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238378] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7207), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4033), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238422] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7210), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4034), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238466] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7249), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4035), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238510] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7261), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4036), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238554] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7264), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4037), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238598] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7292), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4038), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238642] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7314), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4039), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238686] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7316), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4040), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238730] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7337), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4041), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238774] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7351), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4042), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238818] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7354), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4043), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238862] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7377), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4044), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238906] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4251), 1,
      sym__direct_name,
    STATE(5166), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4045), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238950] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7388), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4046), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [238994] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7397), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4047), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239038] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7421), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4048), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239082] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7441), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4049), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239126] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7449), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4050), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239170] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7486), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4051), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239214] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1451), 1,
      anon_sym_LT_LT,
    STATE(853), 1,
      sym__direct_name,
    STATE(2711), 1,
      sym_name,
    STATE(2087), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1463), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2088), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4052), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239258] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7512), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4053), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239302] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7517), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4054), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239346] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7561), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4055), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239390] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7582), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4056), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239434] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7589), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4057), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239478] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7639), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4058), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239522] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1163), 1,
      anon_sym_LT_LT,
    STATE(780), 1,
      sym__direct_name,
    STATE(2537), 1,
      sym_name,
    STATE(1550), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1175), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1551), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4059), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1165), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239566] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7656), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4060), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239610] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7658), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4061), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239654] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7685), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4062), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239698] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7702), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4063), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239742] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7705), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4064), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239786] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7720), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4065), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239830] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7735), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4066), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239874] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7741), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4067), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239918] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7762), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4068), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [239962] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7770), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4069), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240006] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7773), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4070), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240050] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7793), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4071), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240094] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4072), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3332), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [240126] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7809), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4073), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240170] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7814), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4074), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240214] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7830), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4075), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240258] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7853), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4076), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240302] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7870), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4077), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240346] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8053), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4078), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240390] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4079), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3334), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [240422] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8066), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4080), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240466] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8068), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4081), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240510] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8095), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4082), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240554] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5427), 1,
      anon_sym_LT_LT,
    STATE(3386), 1,
      sym__direct_name,
    STATE(5088), 1,
      sym_name,
    STATE(4511), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5431), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4083), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4512), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5429), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240598] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1028), 1,
      anon_sym_LT_LT,
    STATE(750), 1,
      sym__direct_name,
    STATE(2442), 1,
      sym_name,
    STATE(1184), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1040), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1193), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4084), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1030), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240642] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8107), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4085), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240686] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8117), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4086), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240730] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4087), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3294), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [240762] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8150), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4088), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240806] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4089), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3350), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [240838] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4090), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3352), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [240870] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4091), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3354), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [240902] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8179), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4092), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240946] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8185), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4093), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [240990] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8217), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4094), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241034] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8246), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4095), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241078] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8266), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4096), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241122] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8297), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4097), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241166] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8310), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4098), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241210] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8312), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4099), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241254] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4100), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3306), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [241286] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8322), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4101), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241330] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8329), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4102), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241374] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8331), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4103), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241418] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8341), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4104), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241462] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8348), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4105), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241506] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8350), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4106), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241550] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6613), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4107), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241594] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7362), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4108), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241638] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6620), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4109), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241682] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6622), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4110), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241726] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6632), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4111), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241770] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6639), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4112), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241814] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6641), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4113), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241858] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6646), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4114), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241902] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6649), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4115), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241946] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6651), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4116), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [241990] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6656), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4117), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242034] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1050), 1,
      anon_sym_LT_LT,
    STATE(788), 1,
      sym__direct_name,
    STATE(2548), 1,
      sym_name,
    STATE(1553), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1062), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1673), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4118), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1052), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242078] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6659), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4119), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242122] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6661), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4120), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242166] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4362), 1,
      sym__external_name,
    STATE(4928), 1,
      sym__identifier,
    STATE(7054), 1,
      sym_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4121), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5393), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242212] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4362), 1,
      sym__external_name,
    STATE(4957), 1,
      sym__identifier,
    STATE(7059), 1,
      sym_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4122), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5393), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242258] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(902), 1,
      anon_sym_LT_LT,
    STATE(743), 1,
      sym__direct_name,
    STATE(2371), 1,
      sym_name,
    STATE(1081), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(914), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1082), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4123), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(904), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242302] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5433), 1,
      anon_sym_LT_LT,
    STATE(3281), 1,
      sym__direct_name,
    STATE(5007), 1,
      sym_name,
    STATE(4434), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5437), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4124), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4435), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5435), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242346] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7734), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4125), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242390] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7323), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4126), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242434] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4127), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3450), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [242466] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3760), 1,
      anon_sym_LT_LT,
    STATE(3636), 1,
      sym__direct_name,
    STATE(5950), 1,
      sym_name,
    STATE(4717), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3782), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4128), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4725), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3762), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242510] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4129), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3322), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [242542] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4130), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3470), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [242574] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1309), 1,
      anon_sym_LT_LT,
    STATE(803), 1,
      sym__direct_name,
    STATE(2657), 1,
      sym_name,
    STATE(1796), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1321), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1798), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4131), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1311), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242618] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4132), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3324), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [242650] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(401), 1,
      anon_sym_LT_LT,
    STATE(734), 1,
      sym__direct_name,
    STATE(2115), 1,
      sym_name,
    STATE(966), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(409), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(967), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4133), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(403), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242694] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6827), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4134), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242738] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5439), 1,
      anon_sym_LT_LT,
    STATE(3493), 1,
      sym__direct_name,
    STATE(5173), 1,
      sym_name,
    STATE(4604), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5443), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4135), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4608), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5441), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242782] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7205), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4136), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242826] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1329), 1,
      anon_sym_LT_LT,
    STATE(808), 1,
      sym__direct_name,
    STATE(2666), 1,
      sym_name,
    STATE(1838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1341), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4137), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1331), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242870] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3760), 1,
      anon_sym_LT_LT,
    STATE(3636), 1,
      sym__direct_name,
    STATE(5808), 1,
      sym_name,
    STATE(4717), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3782), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4138), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4725), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3762), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242914] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(6024), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4139), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [242958] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4140), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3258), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [242990] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3760), 1,
      anon_sym_LT_LT,
    STATE(3636), 1,
      sym__direct_name,
    STATE(5677), 1,
      sym_name,
    STATE(4717), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3782), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4141), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4725), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3762), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243034] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(738), 1,
      anon_sym_LT_LT,
    STATE(748), 1,
      sym__direct_name,
    STATE(2389), 1,
      sym_name,
    STATE(1077), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(754), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1078), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4142), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(740), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243078] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5445), 1,
      anon_sym_LT_LT,
    STATE(3381), 1,
      sym__direct_name,
    STATE(5112), 1,
      sym_name,
    STATE(4499), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5449), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4143), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4508), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5447), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243122] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    STATE(763), 1,
      sym__direct_name,
    STATE(2457), 1,
      sym_name,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4144), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243166] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1351), 1,
      anon_sym_LT_LT,
    STATE(795), 1,
      sym__direct_name,
    STATE(2576), 1,
      sym_name,
    STATE(1661), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1363), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1663), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4145), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1353), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243210] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4146), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3320), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [243242] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(6188), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4147), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243286] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1371), 1,
      anon_sym_LT_LT,
    STATE(767), 1,
      sym__direct_name,
    STATE(2471), 1,
      sym_name,
    STATE(1240), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1383), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1241), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4148), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1373), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243330] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(415), 1,
      anon_sym_LT_LT,
    STATE(777), 1,
      sym__direct_name,
    STATE(2511), 1,
      sym_name,
    STATE(1411), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(435), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1412), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4149), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(417), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243374] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4150), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3330), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [243406] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4151), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 15,
      sym_identifier,
      sym_GENERIC,
      sym_PARAMETER,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_LT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [243438] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    STATE(822), 1,
      sym__direct_name,
    STATE(2602), 1,
      sym_name,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4152), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243482] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3760), 1,
      anon_sym_LT_LT,
    STATE(3801), 1,
      sym__direct_name,
    STATE(5475), 1,
      sym_name,
    STATE(4717), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3782), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4153), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4725), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3762), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243526] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1006), 1,
      anon_sym_LT_LT,
    STATE(746), 1,
      sym__direct_name,
    STATE(2306), 1,
      sym_name,
    STATE(1116), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1018), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1000), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4154), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1008), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243570] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1407), 1,
      anon_sym_LT_LT,
    STATE(769), 1,
      sym__direct_name,
    STATE(2488), 1,
      sym_name,
    STATE(1280), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1419), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1281), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4155), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1409), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243614] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4156), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3394), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [243646] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4157), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3396), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [243678] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4232), 1,
      anon_sym_DOT,
    ACTIONS(4234), 1,
      anon_sym_LPAREN,
    ACTIONS(4236), 1,
      anon_sym_SQUOTE,
    ACTIONS(4242), 1,
      sym_PARAMETER,
    STATE(4272), 1,
      sym_generic_map_aspect,
    STATE(4314), 1,
      aux_sym_name_repeat1,
    STATE(4331), 1,
      sym__name_selector,
    ACTIONS(1471), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      anon_sym_LBRACK,
    STATE(4158), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4337), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [243726] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5451), 1,
      anon_sym_LT_LT,
    STATE(3603), 1,
      sym__direct_name,
    STATE(5500), 1,
      sym_name,
    STATE(4687), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5455), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4159), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4691), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5453), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243770] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(4715), 1,
      sym__external_name,
    STATE(4757), 1,
      sym__identifier,
    STATE(6177), 1,
      sym_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4160), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5457), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243816] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3760), 1,
      anon_sym_LT_LT,
    STATE(3801), 1,
      sym__direct_name,
    STATE(5510), 1,
      sym_name,
    STATE(4717), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(3782), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4161), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4725), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3762), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243860] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5275), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4162), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243904] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5433), 1,
      anon_sym_LT_LT,
    STATE(3642), 1,
      sym__direct_name,
    STATE(5218), 1,
      sym_name,
    STATE(4434), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5437), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4163), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4435), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5435), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243948] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(5494), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4164), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [243992] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4165), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3438), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [244024] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8149), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4166), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244068] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(922), 1,
      anon_sym_LT_LT,
    STATE(868), 1,
      sym__direct_name,
    STATE(2769), 1,
      sym_name,
    STATE(1975), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(934), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1976), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4167), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1245), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244112] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5202), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4168), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244156] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4169), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3372), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [244188] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4170), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3382), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [244220] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1491), 1,
      sym_AFTER,
    ACTIONS(5459), 1,
      anon_sym_DOT,
    ACTIONS(5462), 1,
      anon_sym_LPAREN,
    ACTIONS(5465), 1,
      anon_sym_SQUOTE,
    ACTIONS(5468), 1,
      anon_sym_LBRACK,
    ACTIONS(5471), 1,
      sym_GENERIC,
    ACTIONS(5474), 1,
      sym_PARAMETER,
    STATE(4951), 1,
      sym_generic_map_aspect,
    STATE(5042), 1,
      sym__name_selector,
    STATE(4171), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_name_repeat1,
    STATE(5043), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [244270] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(5648), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4172), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244314] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7528), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4173), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244358] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7530), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4174), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244402] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(6543), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4175), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244446] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7593), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4176), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244490] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5477), 1,
      anon_sym_LT_LT,
    STATE(3394), 1,
      sym__direct_name,
    STATE(5140), 1,
      sym_name,
    STATE(4452), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5481), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4177), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4454), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5479), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244534] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5290), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4178), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244578] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4251), 1,
      sym__direct_name,
    STATE(5872), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4179), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244622] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6891), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4180), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244666] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5291), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4181), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244710] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1429), 1,
      anon_sym_LT_LT,
    STATE(820), 1,
      sym__direct_name,
    STATE(2632), 1,
      sym_name,
    STATE(1695), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1441), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1696), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4182), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1431), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244754] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7664), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4183), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244798] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7680), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4184), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244842] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7845), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4185), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244886] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5483), 1,
      anon_sym_LT_LT,
    STATE(4195), 1,
      sym__direct_name,
    STATE(6993), 1,
      sym_name,
    STATE(5029), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5487), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4186), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(5030), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5485), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244930] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4336), 1,
      sym__direct_name,
    STATE(7181), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4187), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [244974] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7131), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4188), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245018] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4362), 1,
      sym__external_name,
    STATE(4909), 1,
      sym__identifier,
    STATE(7131), 1,
      sym_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4189), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5393), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245064] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8113), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4190), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245108] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8115), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4191), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245152] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4362), 1,
      sym__external_name,
    STATE(4917), 1,
      sym__identifier,
    STATE(7136), 1,
      sym_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4192), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5393), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245198] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7094), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4193), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245242] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4194), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3448), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [245274] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1471), 1,
      sym_AFTER,
    ACTIONS(5489), 1,
      anon_sym_DOT,
    ACTIONS(5491), 1,
      anon_sym_LPAREN,
    ACTIONS(5493), 1,
      anon_sym_SQUOTE,
    ACTIONS(5495), 1,
      anon_sym_LBRACK,
    ACTIONS(5497), 1,
      sym_GENERIC,
    ACTIONS(5499), 1,
      sym_PARAMETER,
    STATE(4227), 1,
      aux_sym_name_repeat1,
    STATE(4951), 1,
      sym_generic_map_aspect,
    STATE(5042), 1,
      sym__name_selector,
    STATE(4195), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(5043), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [245326] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8076), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4196), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245370] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4197), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3472), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [245402] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1070), 1,
      anon_sym_LT_LT,
    STATE(733), 1,
      sym__direct_name,
    STATE(1949), 1,
      sym_name,
    STATE(939), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1082), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(940), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4198), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1072), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245446] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8158), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4199), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245490] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5151), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4200), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245534] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4251), 1,
      sym__direct_name,
    STATE(5933), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4201), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245578] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4202), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3238), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [245610] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5252), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4203), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245654] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4342), 1,
      anon_sym_LT_LT,
    STATE(3641), 1,
      sym__direct_name,
    STATE(4634), 1,
      sym_name,
    STATE(4861), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4348), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4204), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4866), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5501), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245698] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8208), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4205), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245742] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8211), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4206), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245786] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4943), 1,
      anon_sym_LT_LT,
    STATE(3609), 1,
      sym__direct_name,
    STATE(5756), 1,
      sym_name,
    STATE(4862), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(4951), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4207), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4863), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245830] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5381), 1,
      anon_sym_LT_LT,
    STATE(3481), 1,
      sym__direct_name,
    STATE(5493), 1,
      sym_name,
    STATE(4628), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5385), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4208), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4630), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5383), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245874] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1092), 1,
      anon_sym_LT_LT,
    STATE(873), 1,
      sym__direct_name,
    STATE(2721), 1,
      sym_name,
    STATE(2082), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1104), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2083), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4209), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1094), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245918] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4362), 1,
      sym__external_name,
    STATE(4913), 1,
      sym__identifier,
    STATE(6976), 1,
      sym_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4210), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5393), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [245964] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8247), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4211), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246008] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1185), 1,
      anon_sym_LT_LT,
    STATE(883), 1,
      sym__direct_name,
    STATE(2718), 1,
      sym_name,
    STATE(2106), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1197), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2107), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4212), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1187), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246052] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(6084), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4213), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246096] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4214), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3452), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [246128] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6665), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4215), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246172] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6928), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4216), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246216] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(6263), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4217), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246260] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1205), 1,
      anon_sym_LT_LT,
    STATE(837), 1,
      sym__direct_name,
    STATE(2743), 1,
      sym_name,
    STATE(2137), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1217), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2139), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4218), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1207), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246304] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4219), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3484), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [246336] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6652), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4220), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246380] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7216), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4221), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246424] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5185), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4222), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246468] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4251), 1,
      sym__direct_name,
    STATE(5627), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4223), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246512] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7219), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4224), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246556] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(4362), 1,
      sym__external_name,
    STATE(4943), 1,
      sym__identifier,
    STATE(7219), 1,
      sym_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4225), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5393), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246602] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5206), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4226), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246646] = 15,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1487), 1,
      sym_AFTER,
    ACTIONS(5489), 1,
      anon_sym_DOT,
    ACTIONS(5491), 1,
      anon_sym_LPAREN,
    ACTIONS(5493), 1,
      anon_sym_SQUOTE,
    ACTIONS(5495), 1,
      anon_sym_LBRACK,
    ACTIONS(5497), 1,
      sym_GENERIC,
    ACTIONS(5499), 1,
      sym_PARAMETER,
    STATE(4171), 1,
      aux_sym_name_repeat1,
    STATE(4951), 1,
      sym_generic_map_aspect,
    STATE(5042), 1,
      sym__name_selector,
    STATE(4227), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(5043), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [246698] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4228), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3486), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [246730] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4229), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3454), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [246762] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7294), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4230), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246806] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7335), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4231), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246850] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7726), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4232), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246894] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1227), 1,
      anon_sym_LT_LT,
    STATE(830), 1,
      sym__direct_name,
    STATE(2635), 1,
      sym_name,
    STATE(1876), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1239), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1880), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4233), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1229), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246938] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6846), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4234), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [246982] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6862), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4235), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247026] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5503), 1,
      anon_sym_LT_LT,
    STATE(3803), 1,
      sym__direct_name,
    STATE(5828), 1,
      sym_name,
    STATE(4947), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5507), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4236), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4952), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5505), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247070] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(4671), 1,
      sym__identifier,
    STATE(4715), 1,
      sym__external_name,
    STATE(6263), 1,
      sym_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4237), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5457), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247116] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(764), 1,
      anon_sym_LT_LT,
    STATE(721), 1,
      sym__direct_name,
    STATE(1468), 1,
      sym_name,
    STATE(838), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(776), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(840), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4238), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(808), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247160] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6929), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4239), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247204] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5184), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4240), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247248] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4251), 1,
      sym__direct_name,
    STATE(5588), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4241), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247292] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5199), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4242), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247336] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4243), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3446), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [247368] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7106), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4244), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247412] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7121), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4245), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247456] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6633), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4246), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247500] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7251), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4247), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247544] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7509), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4248), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247588] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7548), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4249), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247632] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1253), 1,
      anon_sym_LT_LT,
    STATE(842), 1,
      sym__direct_name,
    STATE(2727), 1,
      sym_name,
    STATE(2222), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1265), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(2223), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4250), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1255), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247676] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4232), 1,
      anon_sym_DOT,
    ACTIONS(4234), 1,
      anon_sym_LPAREN,
    ACTIONS(4236), 1,
      anon_sym_SQUOTE,
    ACTIONS(4238), 1,
      anon_sym_LBRACK,
    ACTIONS(4242), 1,
      sym_PARAMETER,
    STATE(4272), 1,
      sym_generic_map_aspect,
    STATE(4282), 1,
      aux_sym_name_repeat1,
    STATE(4331), 1,
      sym__name_selector,
    ACTIONS(1471), 2,
      sym_GENERIC,
      anon_sym_SEMI,
    STATE(4251), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4337), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [247726] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5513), 1,
      sym_RANGE,
    STATE(5186), 1,
      sym_range_constraint,
    STATE(4252), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5509), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
    ACTIONS(5511), 8,
      sym_identifier,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [247764] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7629), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4253), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247808] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5220), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4254), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247852] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4251), 1,
      sym__direct_name,
    STATE(5653), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4255), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247896] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5240), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4256), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247940] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1275), 1,
      anon_sym_LT_LT,
    STATE(812), 1,
      sym__direct_name,
    STATE(2628), 1,
      sym_name,
    STATE(1825), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(1287), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1687), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4257), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(1277), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [247984] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7778), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4258), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248028] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7786), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4259), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248072] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8075), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4260), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248116] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(5558), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4261), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248160] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6610), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4262), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248204] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(4715), 1,
      sym__external_name,
    STATE(4826), 1,
      sym__identifier,
    STATE(6023), 1,
      sym_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4263), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5457), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248250] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6626), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4264), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248294] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5168), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4265), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248338] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4158), 1,
      sym__direct_name,
    STATE(5169), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4266), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248382] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4267), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3474), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [248414] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(856), 1,
      anon_sym_LT_LT,
    STATE(772), 1,
      sym__direct_name,
    STATE(2499), 1,
      sym_name,
    STATE(1336), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(868), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1337), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4268), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(858), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248458] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6702), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4269), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248502] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6705), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4270), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248546] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6746), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4271), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248590] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2193), 1,
      anon_sym_COLON,
    ACTIONS(4234), 1,
      anon_sym_LPAREN,
    ACTIONS(5515), 1,
      sym_PARAMETER,
    STATE(4352), 1,
      sym_parenthesis_group,
    STATE(4272), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 11,
      sym_GENERIC,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [248630] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6845), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4273), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248674] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(4677), 1,
      sym__identifier,
    STATE(4715), 1,
      sym__external_name,
    STATE(6293), 1,
      sym_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4274), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5457), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248720] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6863), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4275), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248764] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6878), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4276), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248808] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(4715), 1,
      sym__external_name,
    STATE(4867), 1,
      sym__identifier,
    STATE(6035), 1,
      sym_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4277), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5457), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248854] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6951), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4278), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248898] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6968), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4279), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248942] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(6974), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4280), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [248986] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7056), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4281), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249030] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4232), 1,
      anon_sym_DOT,
    ACTIONS(4234), 1,
      anon_sym_LPAREN,
    ACTIONS(4236), 1,
      anon_sym_SQUOTE,
    ACTIONS(4238), 1,
      anon_sym_LBRACK,
    ACTIONS(4242), 1,
      sym_PARAMETER,
    STATE(3025), 1,
      aux_sym_name_repeat1,
    STATE(4272), 1,
      sym_generic_map_aspect,
    STATE(4331), 1,
      sym__name_selector,
    ACTIONS(1487), 2,
      sym_GENERIC,
      anon_sym_SEMI,
    STATE(4282), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4337), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [249080] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7104), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4283), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249124] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7107), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4284), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249168] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7215), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4285), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249212] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7252), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4286), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249256] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7263), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4287), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249300] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7330), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4288), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249344] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7366), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4289), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249388] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7374), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4290), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249432] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4291), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3476), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [249464] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7459), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4292), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249508] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(6055), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4293), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249552] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7516), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4294), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249596] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7542), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4295), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249640] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7674), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4296), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249684] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4297), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3260), 15,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WHEN,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [249716] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(944), 1,
      anon_sym_LT_LT,
    STATE(779), 1,
      sym__direct_name,
    STATE(2502), 1,
      sym_name,
    STATE(1353), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(956), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(1354), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    STATE(4298), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(946), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249760] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7712), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4299), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249804] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7716), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4300), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249848] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(6114), 1,
      sym_name,
    STATE(4715), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4301), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249892] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7792), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4302), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249936] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7819), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4303), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [249980] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(7822), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4304), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [250024] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5421), 1,
      anon_sym_LT_LT,
    STATE(3668), 1,
      sym__direct_name,
    STATE(4699), 1,
      sym__identifier,
    STATE(4715), 1,
      sym__external_name,
    STATE(6114), 1,
      sym_name,
    ACTIONS(5425), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4305), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4833), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(5457), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [250070] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4336), 1,
      sym__direct_name,
    STATE(8038), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4306), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [250114] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8093), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4307), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [250158] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4336), 1,
      sym__direct_name,
    STATE(8069), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4308), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [250202] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(3023), 1,
      sym__direct_name,
    STATE(8138), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4309), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [250246] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4336), 1,
      sym__direct_name,
    STATE(8094), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4310), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [250290] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4336), 1,
      sym__direct_name,
    STATE(8111), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4311), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [250334] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4336), 1,
      sym__direct_name,
    STATE(8125), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4312), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [250378] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11), 1,
      anon_sym_LT_LT,
    STATE(4336), 1,
      sym__direct_name,
    STATE(8134), 1,
      sym_name,
    STATE(4362), 2,
      sym__identifier,
      sym__external_name,
    ACTIONS(71), 3,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant_std_logic,
    STATE(4313), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4351), 3,
      sym_external_constant_name,
      sym_external_signal_name,
      sym_external_variable_name,
    ACTIONS(3078), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [250422] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4232), 1,
      anon_sym_DOT,
    ACTIONS(4234), 1,
      anon_sym_LPAREN,
    ACTIONS(4236), 1,
      anon_sym_SQUOTE,
    ACTIONS(4242), 1,
      sym_PARAMETER,
    STATE(3025), 1,
      aux_sym_name_repeat1,
    STATE(4272), 1,
      sym_generic_map_aspect,
    STATE(4331), 1,
      sym__name_selector,
    ACTIONS(1487), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      anon_sym_LBRACK,
    STATE(4314), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4337), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [250470] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2510), 1,
      anon_sym_COLON,
    STATE(4315), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [250503] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4316), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3352), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250534] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4317), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3354), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250565] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2554), 1,
      anon_sym_COLON,
    STATE(4318), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [250598] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4319), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3390), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250629] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2392), 1,
      anon_sym_COLON,
    STATE(4320), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [250662] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4321), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3448), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250693] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4322), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3394), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250724] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4323), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3396), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250755] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4324), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3472), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250786] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4325), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3438), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250817] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4326), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3446), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250848] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4327), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3450), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250879] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4328), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3452), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250910] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2638), 1,
      anon_sym_COLON,
    STATE(4329), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [250943] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4330), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3454), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [250974] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2454), 1,
      anon_sym_COLON,
    STATE(4331), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251007] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4332), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3470), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251038] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4333), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3474), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251069] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4334), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3476), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251100] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4335), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3478), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251131] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1471), 1,
      anon_sym_LPAREN,
    ACTIONS(4232), 1,
      anon_sym_DOT,
    ACTIONS(4236), 1,
      anon_sym_SQUOTE,
    ACTIONS(4238), 1,
      anon_sym_LBRACK,
    ACTIONS(4240), 1,
      sym_GENERIC,
    ACTIONS(4242), 1,
      sym_PARAMETER,
    STATE(4272), 1,
      sym_generic_map_aspect,
    STATE(4331), 1,
      sym__name_selector,
    STATE(4345), 1,
      aux_sym_name_repeat1,
    STATE(4336), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4337), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [251180] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2458), 1,
      anon_sym_COLON,
    STATE(4337), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251213] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2612), 1,
      anon_sym_COLON,
    STATE(4338), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251246] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2494), 1,
      anon_sym_COLON,
    STATE(4339), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251279] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4340), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3238), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251310] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2518), 1,
      anon_sym_COLON,
    STATE(4341), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251343] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2514), 1,
      anon_sym_COLON,
    STATE(4342), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251376] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2522), 1,
      anon_sym_COLON,
    STATE(4343), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251409] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5517), 1,
      sym_RANGE,
    STATE(5383), 1,
      sym_range_constraint,
    STATE(4344), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5509), 4,
      sym_BUS,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
    ACTIONS(5511), 8,
      sym_identifier,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [251446] = 14,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1487), 1,
      anon_sym_LPAREN,
    ACTIONS(4232), 1,
      anon_sym_DOT,
    ACTIONS(4236), 1,
      anon_sym_SQUOTE,
    ACTIONS(4238), 1,
      anon_sym_LBRACK,
    ACTIONS(4240), 1,
      sym_GENERIC,
    ACTIONS(4242), 1,
      sym_PARAMETER,
    STATE(3025), 1,
      aux_sym_name_repeat1,
    STATE(4272), 1,
      sym_generic_map_aspect,
    STATE(4331), 1,
      sym__name_selector,
    STATE(4345), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(4337), 5,
      sym_function_call,
      sym_parenthesis_group,
      sym_attribute,
      sym_signature,
      sym_selection,
  [251495] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2616), 1,
      anon_sym_COLON,
    STATE(4346), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251528] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4347), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3258), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251559] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4348), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3260), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251590] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4349), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3262), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251621] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4350), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3264), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251652] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2396), 1,
      anon_sym_COLON,
    STATE(4351), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251685] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2526), 1,
      anon_sym_COLON,
    STATE(4352), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251718] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4353), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3350), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251749] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4354), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3382), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251780] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2562), 1,
      anon_sym_COLON,
    STATE(4355), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251813] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2566), 1,
      anon_sym_COLON,
    STATE(4356), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251846] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2570), 1,
      anon_sym_COLON,
    STATE(4357), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [251879] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4358), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3484), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251910] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4359), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3486), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251941] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4360), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3290), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [251972] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4361), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3294), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252003] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2388), 1,
      anon_sym_COLON,
    STATE(4362), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252036] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2590), 1,
      anon_sym_COLON,
    STATE(4363), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252069] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4364), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3306), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252100] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5519), 1,
      sym_RANGE,
    STATE(5477), 1,
      sym_range_constraint,
    STATE(4365), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5509), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    ACTIONS(5511), 8,
      sym_identifier,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [252137] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2642), 1,
      anon_sym_COLON,
    STATE(4366), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252170] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4367), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3386), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252201] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2646), 1,
      anon_sym_COLON,
    STATE(4368), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252234] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4369), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3320), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252265] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4370), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3322), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252296] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4371), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3324), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252327] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2558), 1,
      anon_sym_COLON,
    STATE(4372), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252360] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4373), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3330), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252391] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2530), 1,
      anon_sym_COLON,
    STATE(4374), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252424] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2676), 1,
      anon_sym_COLON,
    STATE(4375), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252457] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2442), 1,
      anon_sym_COLON,
    STATE(4376), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252490] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2654), 1,
      anon_sym_COLON,
    STATE(4377), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252523] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2650), 1,
      anon_sym_COLON,
    STATE(4378), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252556] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2680), 1,
      anon_sym_COLON,
    STATE(4379), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252589] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2586), 1,
      anon_sym_COLON,
    STATE(4380), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 13,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [252622] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4381), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3270), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252653] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4382), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3332), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252684] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4383), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3334), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252715] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4384), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3344), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252746] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4385), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3346), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252777] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4386), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3372), 14,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_WITH,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [252808] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5483), 1,
      sym__entity_tag,
    STATE(5704), 1,
      sym__identifier,
    STATE(5705), 1,
      sym_entity_designator,
    STATE(6902), 1,
      sym_entity_specification,
    STATE(6904), 1,
      sym_entity_name_list,
    ACTIONS(5521), 2,
      sym_ALL,
      sym_OTHERS,
    ACTIONS(5523), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4387), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [252852] = 17,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5525), 1,
      anon_sym_SEMI,
    ACTIONS(5527), 1,
      sym_FOR,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(5531), 1,
      sym_USE,
    STATE(4987), 1,
      sym_binding_indication,
    STATE(5028), 1,
      aux_sym_configuration_head_repeat2,
    STATE(5912), 1,
      sym_generic_map_aspect,
    STATE(6579), 1,
      sym_block_configuration,
    STATE(6763), 1,
      sym_end_for,
    STATE(8084), 1,
      sym_port_map_aspect,
    STATE(8215), 1,
      sym_verification_unit_binding_indication,
    STATE(4388), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [252906] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5533), 1,
      sym_RANGE,
    STATE(5863), 1,
      sym_range_constraint,
    ACTIONS(5509), 3,
      sym_IS,
      sym_OPEN,
      anon_sym_SEMI,
    STATE(4389), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5511), 8,
      sym_identifier,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [252942] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2680), 2,
      anon_sym_LT,
      anon_sym_GT,
    STATE(4390), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 11,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
      anon_sym_LBRACK,
  [252974] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5535), 1,
      sym_RANGE,
    STATE(5611), 1,
      sym_range_constraint,
    ACTIONS(5509), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(4391), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5511), 8,
      sym_identifier,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [253010] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5483), 1,
      sym__entity_tag,
    STATE(5704), 1,
      sym__identifier,
    STATE(5705), 1,
      sym_entity_designator,
    STATE(6904), 1,
      sym_entity_name_list,
    STATE(7697), 1,
      sym_entity_specification,
    ACTIONS(5521), 2,
      sym_ALL,
      sym_OTHERS,
    ACTIONS(5523), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4392), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [253054] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5483), 1,
      sym__entity_tag,
    STATE(5704), 1,
      sym__identifier,
    STATE(5705), 1,
      sym_entity_designator,
    STATE(6904), 1,
      sym_entity_name_list,
    STATE(7759), 1,
      sym_entity_specification,
    ACTIONS(5521), 2,
      sym_ALL,
      sym_OTHERS,
    ACTIONS(5523), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4393), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [253098] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5483), 1,
      sym__entity_tag,
    STATE(5704), 1,
      sym__identifier,
    STATE(5705), 1,
      sym_entity_designator,
    STATE(6904), 1,
      sym_entity_name_list,
    STATE(7812), 1,
      sym_entity_specification,
    ACTIONS(5521), 2,
      sym_ALL,
      sym_OTHERS,
    ACTIONS(5523), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4394), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [253142] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5483), 1,
      sym__entity_tag,
    STATE(5704), 1,
      sym__identifier,
    STATE(5705), 1,
      sym_entity_designator,
    STATE(6904), 1,
      sym_entity_name_list,
    STATE(7847), 1,
      sym_entity_specification,
    ACTIONS(5521), 2,
      sym_ALL,
      sym_OTHERS,
    ACTIONS(5523), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4395), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [253186] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5483), 1,
      sym__entity_tag,
    STATE(5704), 1,
      sym__identifier,
    STATE(5705), 1,
      sym_entity_designator,
    STATE(6904), 1,
      sym_entity_name_list,
    STATE(7867), 1,
      sym_entity_specification,
    ACTIONS(5521), 2,
      sym_ALL,
      sym_OTHERS,
    ACTIONS(5523), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4396), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [253230] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5483), 1,
      sym__entity_tag,
    STATE(5704), 1,
      sym__identifier,
    STATE(5705), 1,
      sym_entity_designator,
    STATE(6904), 1,
      sym_entity_name_list,
    STATE(7879), 1,
      sym_entity_specification,
    ACTIONS(5521), 2,
      sym_ALL,
      sym_OTHERS,
    ACTIONS(5523), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4397), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [253274] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5483), 1,
      sym__entity_tag,
    STATE(5704), 1,
      sym__identifier,
    STATE(5705), 1,
      sym_entity_designator,
    STATE(6904), 1,
      sym_entity_name_list,
    STATE(7884), 1,
      sym_entity_specification,
    ACTIONS(5521), 2,
      sym_ALL,
      sym_OTHERS,
    ACTIONS(5523), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4398), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [253318] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4314), 1,
      anon_sym_LPAREN,
    ACTIONS(5537), 1,
      sym_PARAMETER,
    STATE(4411), 1,
      sym_parenthesis_group,
    STATE(4399), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 10,
      sym_GENERIC,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253354] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4400), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253383] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4401), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253412] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4402), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253441] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4403), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253470] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4404), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3390), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [253499] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4405), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253528] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4406), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253557] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4407), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253586] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4408), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253615] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4409), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253644] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4410), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253673] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4411), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253702] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4412), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253731] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4413), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253760] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4414), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3270), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [253789] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4415), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253818] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4416), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3448), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [253847] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4417), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3470), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [253876] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4418), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3258), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [253905] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4419), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253934] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4420), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3344), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [253963] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4421), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [253992] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4422), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254021] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4423), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254050] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4424), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3472), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [254079] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4425), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254108] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4426), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254137] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4427), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3484), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [254166] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4428), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3486), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [254195] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4429), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254224] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4430), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3346), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [254253] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4409), 1,
      anon_sym_LPAREN,
    ACTIONS(5539), 1,
      sym_PARAMETER,
    STATE(4524), 1,
      sym_parenthesis_group,
    STATE(4431), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254288] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4397), 1,
      anon_sym_LPAREN,
    ACTIONS(5541), 1,
      sym_PARAMETER,
    STATE(4510), 1,
      sym_parenthesis_group,
    STATE(4432), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 9,
      sym_BUS,
      sym_GENERIC,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254323] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4433), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254352] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4434), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254381] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4435), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254410] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4436), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254439] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4437), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5543), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [254468] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4438), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5545), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [254497] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4439), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3320), 12,
      sym_identifier,
      sym_ASSERT,
      sym_END,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [254526] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4440), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254555] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4419), 1,
      anon_sym_LPAREN,
    ACTIONS(5547), 1,
      sym_PARAMETER,
    STATE(4468), 1,
      sym_parenthesis_group,
    STATE(4441), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 9,
      sym_BUS,
      sym_GENERIC,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254590] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4442), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254619] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4443), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254648] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4444), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 12,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254677] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4445), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254705] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4446), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254733] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4447), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254761] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5549), 1,
      anon_sym_LPAREN,
    ACTIONS(5551), 1,
      sym_NOT,
    ACTIONS(5553), 1,
      sym_directive_constant_builtin,
    STATE(4660), 1,
      sym_conditional_analysis_relation,
    STATE(4958), 1,
      sym__conditional_analysis_identifier,
    STATE(5067), 1,
      sym__identifier,
    STATE(8291), 1,
      sym_conditional_analysis_expression,
    STATE(4448), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [254803] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4449), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254831] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4450), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254859] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4451), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254887] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4452), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254915] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4453), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254943] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4454), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254971] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4455), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [254999] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4456), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255027] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4457), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255055] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4458), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255083] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4459), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255111] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4460), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255139] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4649), 1,
      anon_sym_LPAREN,
    ACTIONS(5555), 1,
      sym_PARAMETER,
    STATE(4605), 1,
      sym_parenthesis_group,
    STATE(4461), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 8,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255173] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2968), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(4462), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255203] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4463), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255231] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4464), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255259] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4465), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255287] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    STATE(8120), 1,
      sym_generic_map_aspect,
    STATE(4466), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5557), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [255319] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4467), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255347] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4468), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255375] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4469), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255403] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4470), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255431] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4471), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255459] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4472), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255487] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4473), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255515] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4474), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255543] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4475), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255571] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4476), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255599] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4477), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255627] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4478), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255655] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4479), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255683] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4480), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255711] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4545), 1,
      anon_sym_LPAREN,
    ACTIONS(5559), 1,
      sym_PARAMETER,
    STATE(4575), 1,
      sym_parenthesis_group,
    STATE(4481), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 8,
      sym_BUS,
      sym_GENERIC,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255745] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4482), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255773] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4483), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255801] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4484), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255829] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4485), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255857] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5553), 1,
      sym_directive_constant_builtin,
    ACTIONS(5561), 1,
      anon_sym_LPAREN,
    ACTIONS(5563), 1,
      sym_NOT,
    STATE(4716), 1,
      sym_conditional_analysis_relation,
    STATE(4973), 1,
      sym__conditional_analysis_identifier,
    STATE(5067), 1,
      sym__identifier,
    STATE(7525), 1,
      sym_conditional_analysis_expression,
    STATE(4486), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [255899] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5553), 1,
      sym_directive_constant_builtin,
    ACTIONS(5561), 1,
      anon_sym_LPAREN,
    ACTIONS(5563), 1,
      sym_NOT,
    STATE(4716), 1,
      sym_conditional_analysis_relation,
    STATE(4973), 1,
      sym__conditional_analysis_identifier,
    STATE(5067), 1,
      sym__identifier,
    STATE(6735), 1,
      sym_conditional_analysis_expression,
    STATE(4487), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [255941] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4488), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255969] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4489), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [255997] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4490), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256025] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4599), 1,
      anon_sym_LPAREN,
    ACTIONS(5565), 1,
      sym_PARAMETER,
    STATE(4560), 1,
      sym_parenthesis_group,
    STATE(4491), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 8,
      sym_FOR,
      sym_GENERIC,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256059] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4492), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256087] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4493), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256115] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4494), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256143] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5549), 1,
      anon_sym_LPAREN,
    ACTIONS(5551), 1,
      sym_NOT,
    ACTIONS(5553), 1,
      sym_directive_constant_builtin,
    STATE(4660), 1,
      sym_conditional_analysis_relation,
    STATE(4958), 1,
      sym__conditional_analysis_identifier,
    STATE(5067), 1,
      sym__identifier,
    STATE(8098), 1,
      sym_conditional_analysis_expression,
    STATE(4495), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [256185] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4496), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256213] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4497), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256241] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4498), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256269] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4499), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256297] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4500), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256325] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5567), 1,
      sym_ATTRIBUTE,
    ACTIONS(5569), 1,
      sym_FOR,
    ACTIONS(5571), 1,
      sym_GROUP,
    ACTIONS(5573), 1,
      sym_USE,
    STATE(4658), 1,
      aux_sym_configuration_head_repeat1,
    STATE(5349), 1,
      aux_sym_configuration_head_repeat2,
    STATE(5520), 1,
      sym__configuration_declarative_item,
    STATE(7246), 1,
      sym_verification_unit_binding_indication,
    STATE(4501), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(5519), 3,
      sym_use_clause,
      sym_attribute_specification,
      sym_group_declaration,
  [256369] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4502), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256397] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4503), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256425] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4504), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256453] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5577), 1,
      sym_VIEW,
    STATE(8072), 1,
      sym__element_mode_indication,
    STATE(8074), 2,
      sym__element_mode_view_indication,
      sym_mode,
    STATE(8119), 2,
      sym_element_record_mode_view_indication,
      sym_element_array_mode_view_indication,
    STATE(4505), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5575), 5,
      sym_BUFFER,
      sym_IN,
      sym_INOUT,
      sym_LINKAGE,
      sym_OUT,
  [256489] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5509), 1,
      sym_IS,
    ACTIONS(5579), 1,
      sym_RANGE,
    STATE(7213), 1,
      sym_range_constraint,
    STATE(4506), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5511), 8,
      sym_identifier,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [256523] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4507), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3750), 11,
      sym_identifier,
      sym_ASSERT,
      sym_POSTPONED,
      sym_PROCESS,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [256551] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4508), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256579] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4509), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256607] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4510), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256635] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4511), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256663] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4512), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256691] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4513), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256719] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5567), 1,
      sym_ATTRIBUTE,
    ACTIONS(5571), 1,
      sym_GROUP,
    ACTIONS(5573), 1,
      sym_USE,
    ACTIONS(5581), 1,
      sym_FOR,
    STATE(4501), 1,
      aux_sym_configuration_head_repeat1,
    STATE(5520), 1,
      sym__configuration_declarative_item,
    STATE(5521), 1,
      aux_sym_configuration_head_repeat2,
    STATE(7246), 1,
      sym_verification_unit_binding_indication,
    STATE(4514), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(5519), 3,
      sym_use_clause,
      sym_attribute_specification,
      sym_group_declaration,
  [256763] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4515), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256791] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4516), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256819] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4517), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256847] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4518), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256875] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4519), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256903] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4520), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256931] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4521), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256959] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4522), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [256987] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4523), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257015] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4524), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257043] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5553), 1,
      sym_directive_constant_builtin,
    ACTIONS(5561), 1,
      anon_sym_LPAREN,
    ACTIONS(5563), 1,
      sym_NOT,
    STATE(4716), 1,
      sym_conditional_analysis_relation,
    STATE(4973), 1,
      sym__conditional_analysis_identifier,
    STATE(5067), 1,
      sym__identifier,
    STATE(7606), 1,
      sym_conditional_analysis_expression,
    STATE(4525), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [257085] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4526), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257113] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4527), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257141] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4528), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257169] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4529), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257197] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4530), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257225] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4531), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257253] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4532), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257281] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4533), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257309] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4534), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257337] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4535), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257365] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4536), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257393] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4537), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257421] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4538), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257449] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4539), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257477] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4540), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257505] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4541), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257533] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4542), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257561] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5553), 1,
      sym_directive_constant_builtin,
    ACTIONS(5561), 1,
      anon_sym_LPAREN,
    ACTIONS(5563), 1,
      sym_NOT,
    STATE(4716), 1,
      sym_conditional_analysis_relation,
    STATE(4973), 1,
      sym__conditional_analysis_identifier,
    STATE(5067), 1,
      sym__identifier,
    STATE(6986), 1,
      sym_conditional_analysis_expression,
    STATE(4543), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [257603] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4544), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257631] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4545), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 11,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RANGE,
      sym_REGISTER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257659] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4546), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257687] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4547), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 11,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257715] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4548), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257742] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5583), 1,
      sym_END,
    STATE(4731), 1,
      aux_sym_record_type_definition_repeat1,
    STATE(5287), 1,
      sym_element_declaration,
    STATE(5848), 1,
      sym__identifier,
    STATE(7039), 1,
      sym_identifier_list,
    STATE(8153), 1,
      sym_end_record,
    STATE(4549), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [257781] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4550), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257808] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4551), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4216), 10,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_GENERIC,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [257835] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5585), 1,
      anon_sym_COLON,
    STATE(4552), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 9,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257864] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4553), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257891] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4554), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257918] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4555), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257945] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4556), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257972] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4557), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [257999] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4558), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258026] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4559), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258053] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4560), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258080] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5137), 1,
      anon_sym_LPAREN,
    ACTIONS(5587), 1,
      sym_PARAMETER,
    STATE(4748), 1,
      sym_parenthesis_group,
    STATE(4561), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 7,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258113] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4562), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258140] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4563), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258167] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4564), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258194] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4565), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258221] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4566), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258248] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4567), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258275] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4568), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258302] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4569), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258329] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4570), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258356] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4571), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258383] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4572), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258410] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4573), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258437] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4574), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258464] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4575), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258491] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5589), 1,
      sym_VUNIT,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(6908), 1,
      sym_selected_name_list,
    STATE(4576), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [258530] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4577), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258557] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4578), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4220), 10,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_GENERIC,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [258584] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4579), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258611] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4580), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258638] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4581), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258665] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4582), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258692] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4583), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258719] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4584), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258746] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4585), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258773] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5553), 1,
      sym_directive_constant_builtin,
    ACTIONS(5561), 1,
      anon_sym_LPAREN,
    ACTIONS(5563), 1,
      sym_NOT,
    STATE(4973), 1,
      sym__conditional_analysis_identifier,
    STATE(4985), 1,
      sym_conditional_analysis_relation,
    STATE(5067), 1,
      sym__identifier,
    STATE(4586), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [258812] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5239), 1,
      anon_sym_LPAREN,
    ACTIONS(5593), 1,
      sym_PARAMETER,
    STATE(4781), 1,
      sym_parenthesis_group,
    STATE(4587), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 7,
      sym_GENERIC,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258845] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4588), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258872] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4589), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258899] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4590), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258926] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4591), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258953] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4592), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [258980] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4593), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259007] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4594), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259034] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4595), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259061] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4596), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259088] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5595), 1,
      sym_FOR,
    ACTIONS(5597), 1,
      sym_USE,
    STATE(5040), 1,
      aux_sym_block_configuration_repeat2,
    STATE(5243), 1,
      aux_sym_block_configuration_repeat1,
    STATE(5816), 1,
      sym_use_clause,
    STATE(6067), 1,
      sym__configuration_item,
    STATE(8046), 1,
      sym_end_for,
    STATE(6072), 2,
      sym_block_configuration,
      sym_component_configuration,
    STATE(4597), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [259131] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5595), 1,
      sym_FOR,
    ACTIONS(5597), 1,
      sym_USE,
    STATE(5050), 1,
      aux_sym_block_configuration_repeat2,
    STATE(5243), 1,
      aux_sym_block_configuration_repeat1,
    STATE(5816), 1,
      sym_use_clause,
    STATE(6067), 1,
      sym__configuration_item,
    STATE(8356), 1,
      sym_end_for,
    STATE(6072), 2,
      sym_block_configuration,
      sym_component_configuration,
    STATE(4598), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [259174] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4599), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259201] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4600), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259228] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4601), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259255] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4602), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259282] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4603), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259309] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4604), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259336] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4605), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259363] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4606), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259390] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4607), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259417] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4608), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259444] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4609), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259471] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4610), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259498] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4611), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259525] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4612), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259552] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4613), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259579] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4614), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259606] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4615), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259633] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4616), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5599), 10,
      sym_identifier,
      sym_TYPE,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [259660] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4617), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259687] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5151), 1,
      anon_sym_LPAREN,
    ACTIONS(5601), 1,
      sym_PARAMETER,
    STATE(4663), 1,
      sym_parenthesis_group,
    STATE(4618), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 7,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_USE,
      anon_sym_DOT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259720] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4619), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259747] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4620), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259774] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4621), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259801] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4622), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259828] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4623), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259855] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4624), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259882] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4625), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259909] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4626), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259936] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4627), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259963] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4628), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [259990] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4629), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260017] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4630), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260044] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4631), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260071] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5361), 1,
      anon_sym_CARET,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5096), 1,
      aux_sym_relative_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(6766), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(4632), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [260110] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4633), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260137] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5595), 1,
      sym_FOR,
    ACTIONS(5597), 1,
      sym_USE,
    STATE(4598), 1,
      aux_sym_block_configuration_repeat1,
    STATE(4966), 1,
      aux_sym_block_configuration_repeat2,
    STATE(5816), 1,
      sym_use_clause,
    STATE(6067), 1,
      sym__configuration_item,
    STATE(7111), 1,
      sym_end_for,
    STATE(6072), 2,
      sym_block_configuration,
      sym_component_configuration,
    STATE(4634), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [260180] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4635), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260207] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4636), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260234] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4637), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260261] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4638), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260288] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5087), 1,
      anon_sym_LPAREN,
    ACTIONS(5603), 1,
      sym_PARAMETER,
    STATE(4794), 1,
      sym_parenthesis_group,
    STATE(4639), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 7,
      sym_GENERIC,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [260321] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4640), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260348] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4641), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 10,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260375] = 13,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5595), 1,
      sym_FOR,
    ACTIONS(5597), 1,
      sym_USE,
    STATE(4597), 1,
      aux_sym_block_configuration_repeat1,
    STATE(5034), 1,
      aux_sym_block_configuration_repeat2,
    STATE(5816), 1,
      sym_use_clause,
    STATE(6067), 1,
      sym__configuration_item,
    STATE(8036), 1,
      sym_end_for,
    STATE(6072), 2,
      sym_block_configuration,
      sym_component_configuration,
    STATE(4642), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [260418] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4643), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260445] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4644), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260472] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5549), 1,
      anon_sym_LPAREN,
    ACTIONS(5551), 1,
      sym_NOT,
    ACTIONS(5553), 1,
      sym_directive_constant_builtin,
    STATE(4958), 1,
      sym__conditional_analysis_identifier,
    STATE(5048), 1,
      sym_conditional_analysis_relation,
    STATE(5067), 1,
      sym__identifier,
    STATE(4645), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [260511] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5167), 1,
      anon_sym_LPAREN,
    ACTIONS(5605), 1,
      sym_PARAMETER,
    STATE(4737), 1,
      sym_parenthesis_group,
    STATE(4646), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 7,
      sym_GENERIC,
      sym_OF,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260544] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4647), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260571] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4648), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260598] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5583), 1,
      sym_END,
    STATE(4549), 1,
      aux_sym_record_type_definition_repeat1,
    STATE(5287), 1,
      sym_element_declaration,
    STATE(5848), 1,
      sym__identifier,
    STATE(7039), 1,
      sym_identifier_list,
    STATE(7040), 1,
      sym_end_record,
    STATE(4649), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [260637] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4650), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260664] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5607), 1,
      anon_sym_SEMI,
    ACTIONS(5611), 1,
      sym_operator_symbol,
    STATE(6733), 1,
      sym__identifier,
    STATE(7110), 1,
      sym__designator,
    ACTIONS(5609), 2,
      sym_FUNCTION,
      sym_PROCEDURE,
    STATE(4651), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [260701] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4652), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 10,
      sym_GENERIC,
      sym_IS,
      sym_OPEN,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260728] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4653), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260755] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4654), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 10,
      sym_BUS,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260782] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5057), 1,
      anon_sym_LPAREN,
    ACTIONS(5613), 1,
      sym_PARAMETER,
    STATE(4759), 1,
      sym_parenthesis_group,
    STATE(4655), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 7,
      sym_GENERIC,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260815] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4656), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260841] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4657), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260867] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5615), 1,
      sym_ATTRIBUTE,
    ACTIONS(5618), 1,
      sym_FOR,
    ACTIONS(5620), 1,
      sym_GROUP,
    ACTIONS(5623), 1,
      sym_USE,
    STATE(5520), 1,
      sym__configuration_declarative_item,
    STATE(5519), 3,
      sym_use_clause,
      sym_attribute_specification,
      sym_group_declaration,
    STATE(4658), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_configuration_head_repeat1,
  [260903] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4659), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260929] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5628), 1,
      sym_THEN,
    STATE(4645), 1,
      sym_logical_operator,
    STATE(4774), 1,
      aux_sym_conditional_analysis_expression_repeat1,
    STATE(4660), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5626), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
  [260961] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4661), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [260987] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4662), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3356), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [261013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4663), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261039] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4664), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261065] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4665), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261091] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2805), 1,
      sym_END,
    ACTIONS(5630), 1,
      sym_CONTEXT,
    ACTIONS(5633), 1,
      sym_LIBRARY,
    ACTIONS(5636), 1,
      sym_USE,
    STATE(5478), 1,
      sym__context_item,
    STATE(5480), 3,
      sym_library_clause,
      sym_use_clause,
      sym_context_reference,
    STATE(4666), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_design_unit_repeat1,
  [261127] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4667), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261153] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4668), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261179] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4669), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261205] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4670), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261231] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5639), 1,
      sym_OF,
    STATE(4671), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261259] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5644), 1,
      sym_THEN,
    STATE(4645), 1,
      sym_logical_operator,
    STATE(4672), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_conditional_analysis_expression_repeat1,
    ACTIONS(5641), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
  [261289] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4673), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3436), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [261315] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4674), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261341] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4675), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261367] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4676), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261393] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5646), 1,
      sym_OF,
    STATE(4677), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261421] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4678), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [261447] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4679), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261473] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4680), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261499] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4681), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [261525] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4682), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [261551] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4683), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3464), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [261577] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4684), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3480), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [261603] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4685), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3496), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [261629] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4686), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [261655] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4687), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261681] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4688), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [261707] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4689), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [261733] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4690), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261759] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4691), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261785] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4692), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261811] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4693), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261837] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4694), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [261863] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4695), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3246), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [261889] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4696), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3248), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [261915] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4697), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3252), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [261941] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4698), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5648), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [261967] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5650), 1,
      sym_OF,
    STATE(4699), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [261995] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4700), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262021] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4701), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262047] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4702), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3266), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [262073] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4703), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3282), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [262099] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5483), 1,
      sym__entity_tag,
    STATE(5704), 1,
      sym__identifier,
    STATE(6222), 1,
      sym_entity_designator,
    ACTIONS(5523), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4704), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [262133] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4705), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262159] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4706), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262185] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4707), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262211] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4708), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262237] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4709), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [262263] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4710), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3302), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [262289] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4711), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3308), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [262315] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4712), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262341] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4713), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3312), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [262367] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4714), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [262393] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4715), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262419] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5628), 1,
      anon_sym_RPAREN,
    STATE(4586), 1,
      sym_logical_operator,
    STATE(4820), 1,
      aux_sym_conditional_analysis_expression_repeat1,
    STATE(4716), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5626), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
  [262451] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4717), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262477] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4718), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262503] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4719), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262529] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4720), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262555] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4721), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262581] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4722), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262607] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4723), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262633] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4724), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [262659] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4725), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262685] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4726), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262711] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4727), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262737] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4728), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262763] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4729), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262789] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4730), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262815] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5655), 1,
      sym_END,
    STATE(5287), 1,
      sym_element_declaration,
    STATE(5848), 1,
      sym__identifier,
    STATE(7039), 1,
      sym_identifier_list,
    ACTIONS(5652), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4731), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_record_type_definition_repeat1,
  [262849] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4732), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262875] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4733), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262901] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4734), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262927] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(8071), 1,
      sym_selected_name_list,
    STATE(4735), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [262963] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4736), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [262989] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4737), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263015] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4738), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263041] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5657), 1,
      sym_END,
    STATE(4842), 1,
      aux_sym_physical_type_definition_repeat1,
    STATE(5298), 1,
      sym_secondary_unit_declaration,
    STATE(6866), 1,
      sym__identifier,
    STATE(6869), 1,
      sym_end_units,
    STATE(4739), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [263077] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4740), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263103] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4741), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263129] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4742), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263155] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4743), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263181] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5659), 1,
      sym_END,
    STATE(4846), 1,
      aux_sym_mode_view_body_repeat1,
    STATE(5323), 1,
      sym_mode_view_element_definition,
    STATE(5944), 1,
      sym__identifier,
    STATE(8290), 1,
      sym_record_element_list,
    STATE(4744), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [263217] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4745), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263243] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4746), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263269] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4747), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263295] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4748), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263321] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4749), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263347] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4750), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263373] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4751), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263399] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4752), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263425] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4753), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [263451] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4754), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263477] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4755), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263503] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4756), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263529] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5661), 1,
      sym_OF,
    STATE(4757), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263557] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4758), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263583] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4759), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263609] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4760), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263635] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(7774), 1,
      sym_selected_name_list,
    STATE(4761), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [263671] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4762), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263697] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4763), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263723] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4764), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [263749] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4765), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263775] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4766), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [263801] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4767), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 9,
      sym_identifier,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
      anon_sym_LT_LT,
  [263827] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4768), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5663), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [263853] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4769), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [263879] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4770), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5665), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [263905] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4771), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263931] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5002), 1,
      sym_component_specification,
    STATE(5910), 1,
      sym__label,
    STATE(8086), 1,
      sym_instantiation_list,
    ACTIONS(4346), 2,
      sym_ALL,
      sym_OTHERS,
    STATE(4772), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [263965] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4773), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [263991] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5669), 1,
      sym_THEN,
    STATE(4645), 1,
      sym_logical_operator,
    STATE(4672), 1,
      aux_sym_conditional_analysis_expression_repeat1,
    STATE(4774), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5626), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
  [264023] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4775), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264049] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4776), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264075] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4777), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264101] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4778), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264127] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4779), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264153] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4780), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264179] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4781), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264205] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4782), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 9,
      sym_GENERIC,
      sym_IS,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264231] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4783), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [264257] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4784), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264283] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4785), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3292), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [264309] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(7465), 1,
      sym_selected_name_list,
    STATE(4786), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [264345] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4787), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264371] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4788), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264397] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4789), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264423] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5238), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(7306), 1,
      sym_selected_name_list,
    STATE(4790), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [264459] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4791), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264485] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5409), 1,
      anon_sym_LPAREN,
    ACTIONS(5673), 1,
      sym_PARAMETER,
    STATE(4898), 1,
      sym_parenthesis_group,
    STATE(4792), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 6,
      sym_GENERIC,
      sym_IS,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264517] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4793), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [264543] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4794), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [264569] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4795), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264595] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4796), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264621] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4797), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264647] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4798), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264673] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4799), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264699] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4800), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264725] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5675), 1,
      sym_END,
    STATE(4744), 1,
      aux_sym_mode_view_body_repeat1,
    STATE(5323), 1,
      sym_mode_view_element_definition,
    STATE(5944), 1,
      sym__identifier,
    STATE(8290), 1,
      sym_record_element_list,
    STATE(4801), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [264761] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(8358), 1,
      sym_selected_name_list,
    STATE(4802), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [264797] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4803), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [264823] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4804), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [264849] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4805), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264875] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5677), 1,
      sym_CONTEXT,
    ACTIONS(5679), 1,
      sym_END,
    ACTIONS(5681), 1,
      sym_LIBRARY,
    ACTIONS(5683), 1,
      sym_USE,
    STATE(4852), 1,
      aux_sym_design_unit_repeat1,
    STATE(5478), 1,
      sym__context_item,
    STATE(4806), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(5480), 3,
      sym_library_clause,
      sym_use_clause,
      sym_context_reference,
  [264913] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4807), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264939] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4808), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2686), 9,
      sym_identifier,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [264965] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4809), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [264991] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4810), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265017] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(7306), 1,
      sym_selected_name_list,
    STATE(4811), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [265053] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4812), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265079] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4813), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [265105] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4814), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265131] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4815), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265157] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4816), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265183] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4817), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265209] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4818), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265235] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4819), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [265261] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5669), 1,
      anon_sym_RPAREN,
    STATE(4586), 1,
      sym_logical_operator,
    STATE(4828), 1,
      aux_sym_conditional_analysis_expression_repeat1,
    STATE(4820), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5626), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
  [265293] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4821), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265319] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4822), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265345] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4823), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5685), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [265371] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(6740), 1,
      sym_selected_name_list,
    STATE(4824), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [265407] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4825), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265433] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5687), 1,
      sym_OF,
    STATE(4826), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265461] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4827), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265487] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5644), 1,
      anon_sym_RPAREN,
    STATE(4586), 1,
      sym_logical_operator,
    STATE(4828), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_conditional_analysis_expression_repeat1,
    ACTIONS(5641), 6,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
  [265517] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4829), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265543] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4830), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265569] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4831), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265595] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4832), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265621] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4833), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265647] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4834), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265673] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4835), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265699] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4836), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265725] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4837), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265751] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4838), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265777] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4839), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265803] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4840), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265829] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4841), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265855] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5657), 1,
      sym_END,
    STATE(4914), 1,
      aux_sym_physical_type_definition_repeat1,
    STATE(5298), 1,
      sym_secondary_unit_declaration,
    STATE(6866), 1,
      sym__identifier,
    STATE(8028), 1,
      sym_end_units,
    STATE(4842), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [265891] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4843), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265917] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4844), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [265943] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(7382), 1,
      sym_selected_name_list,
    STATE(4845), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [265979] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5692), 1,
      sym_END,
    STATE(5323), 1,
      sym_mode_view_element_definition,
    STATE(5944), 1,
      sym__identifier,
    STATE(8290), 1,
      sym_record_element_list,
    ACTIONS(5689), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4846), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_mode_view_body_repeat1,
  [266013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4847), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266039] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4848), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266065] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4849), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266091] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4850), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266117] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4851), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266143] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5677), 1,
      sym_CONTEXT,
    ACTIONS(5681), 1,
      sym_LIBRARY,
    ACTIONS(5683), 1,
      sym_USE,
    ACTIONS(5694), 1,
      sym_END,
    STATE(4666), 1,
      aux_sym_design_unit_repeat1,
    STATE(5478), 1,
      sym__context_item,
    STATE(4852), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(5480), 3,
      sym_library_clause,
      sym_use_clause,
      sym_context_reference,
  [266181] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4853), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266207] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4854), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266233] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4855), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266259] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4856), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266285] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(8223), 1,
      sym_selected_name_list,
    STATE(4857), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [266321] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(7457), 1,
      sym_selected_name_list,
    STATE(4858), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [266357] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4859), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_LT_EQ,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266383] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4860), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 9,
      sym_GENERIC,
      sym_OF,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266409] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4861), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266435] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4862), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [266461] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4863), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [266487] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(6791), 1,
      sym_selected_name_list,
    STATE(4864), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [266523] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4865), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [266549] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4866), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266575] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5696), 1,
      sym_OF,
    STATE(4867), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266603] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(6908), 1,
      sym_selected_name_list,
    STATE(4868), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [266639] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4869), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [266665] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4870), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266691] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4871), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [266717] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4872), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266743] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4873), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266769] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4874), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3498), 9,
      sym_ALIAS,
      sym_ATTRIBUTE,
      sym_END,
      sym_FUNCTION,
      sym_IMPURE,
      sym_PROCEDURE,
      sym_PRIVATE,
      sym_PURE,
      sym_USE,
  [266795] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4875), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266821] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4876), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266847] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4877), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266873] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4878), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266899] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4879), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2714), 9,
      sym_identifier,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_constant_unit,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [266925] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4880), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266951] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4881), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [266977] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4882), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267003] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(5812), 1,
      sym_selected_name,
    STATE(6005), 1,
      sym__logical_name,
    STATE(7005), 1,
      sym_selected_name_list,
    STATE(4883), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267039] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4884), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267065] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4885), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267091] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4886), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 9,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PARAMETER,
      sym_USE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267117] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4887), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [267143] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4888), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [267169] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4889), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 9,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
  [267195] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5014), 1,
      sym_component_specification,
    STATE(5910), 1,
      sym__label,
    STATE(8086), 1,
      sym_instantiation_list,
    ACTIONS(4346), 2,
      sym_ALL,
      sym_OTHERS,
    STATE(4890), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267229] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4891), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 9,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267255] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4892), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267280] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4893), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267305] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4894), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267330] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6344), 1,
      sym__identifier,
    STATE(6458), 1,
      sym__alias_designator,
    ACTIONS(5698), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4895), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267361] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5700), 1,
      anon_sym_SEMI,
    STATE(5923), 1,
      sym__label,
    STATE(8133), 1,
      sym_when_expression,
    STATE(4896), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267394] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4897), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267419] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4898), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267444] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6344), 1,
      sym__identifier,
    STATE(6467), 1,
      sym__alias_designator,
    ACTIONS(5698), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4899), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267475] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5605), 1,
      sym__logical_name,
    STATE(5807), 1,
      sym__identifier,
    STATE(7461), 1,
      sym_logical_name_list,
    STATE(4900), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267508] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4901), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267533] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4902), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267558] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6344), 1,
      sym__identifier,
    STATE(6469), 1,
      sym__alias_designator,
    ACTIONS(5698), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4903), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267589] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5704), 1,
      anon_sym_SEMI,
    STATE(5728), 1,
      sym__label,
    STATE(6922), 1,
      sym_when_expression,
    STATE(4904), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267622] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4986), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5645), 1,
      sym__identifier,
    STATE(7714), 1,
      sym_partial_pathname,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(4905), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267655] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5706), 1,
      sym_END,
    ACTIONS(5708), 1,
      sym_GENERIC,
    ACTIONS(5710), 1,
      sym_IS,
    ACTIONS(5712), 1,
      sym_PORT,
    STATE(5692), 1,
      sym_generic_clause,
    STATE(6022), 1,
      sym_component_body,
    STATE(6765), 1,
      sym_port_clause,
    STATE(8081), 1,
      sym_end_component,
    STATE(4906), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [267694] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4907), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267719] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4908), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267744] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5714), 1,
      sym_OF,
    STATE(4909), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 7,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267771] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4910), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5716), 8,
      sym_identifier,
      sym_character_literal,
      sym_operator_symbol,
      sym_library_constant,
      sym_library_constant_std_logic,
      sym_library_function,
      sym_library_type,
      anon_sym_LT_LT,
  [267796] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5718), 1,
      anon_sym_SEMI,
    STATE(5861), 1,
      sym__label,
    STATE(7558), 1,
      sym_when_expression,
    STATE(4911), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267829] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5720), 1,
      anon_sym_SEMI,
    STATE(5862), 1,
      sym__label,
    STATE(7560), 1,
      sym_when_expression,
    STATE(4912), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267862] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5722), 1,
      sym_OF,
    STATE(4913), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 7,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [267889] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5727), 1,
      sym_END,
    STATE(5298), 1,
      sym_secondary_unit_declaration,
    STATE(6866), 1,
      sym__identifier,
    ACTIONS(5724), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(4914), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_physical_type_definition_repeat1,
  [267920] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5729), 1,
      anon_sym_SEMI,
    STATE(5832), 1,
      sym__label,
    STATE(7376), 1,
      sym_when_expression,
    STATE(4915), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267953] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(6005), 1,
      sym__logical_name,
    STATE(6096), 1,
      sym_selected_name,
    STATE(4916), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [267986] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5731), 1,
      sym_OF,
    STATE(4917), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 7,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268013] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6344), 1,
      sym__identifier,
    STATE(6520), 1,
      sym__alias_designator,
    ACTIONS(5698), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4918), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268044] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4919), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268069] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4920), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268094] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4921), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268119] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5733), 1,
      anon_sym_SEMI,
    STATE(5827), 1,
      sym__label,
    STATE(7361), 1,
      sym_when_expression,
    STATE(4922), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268152] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5605), 1,
      sym__logical_name,
    STATE(5807), 1,
      sym__identifier,
    STATE(7415), 1,
      sym_logical_name_list,
    STATE(4923), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268185] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5735), 1,
      anon_sym_SEMI,
    STATE(5878), 1,
      sym__label,
    STATE(7631), 1,
      sym_when_expression,
    STATE(4924), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268218] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5737), 1,
      anon_sym_SEMI,
    STATE(5879), 1,
      sym__label,
    STATE(7632), 1,
      sym_when_expression,
    STATE(4925), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268251] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4926), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268276] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4927), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268301] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5739), 1,
      sym_OF,
    STATE(4928), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 7,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268328] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4929), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268353] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4930), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268378] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5741), 1,
      sym_GENERATE,
    STATE(6348), 1,
      sym_label_declaration,
    STATE(7328), 1,
      sym__label,
    STATE(8143), 1,
      sym_generate_body,
    STATE(4931), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268411] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4932), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268436] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4933), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268461] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5743), 1,
      anon_sym_SEMI,
    STATE(5939), 1,
      sym__label,
    STATE(8183), 1,
      sym_when_expression,
    STATE(4934), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268494] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5611), 1,
      sym_operator_symbol,
    ACTIONS(5745), 1,
      anon_sym_SEMI,
    STATE(6733), 1,
      sym__identifier,
    STATE(6764), 1,
      sym__designator,
    STATE(4935), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268527] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4936), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268552] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4937), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268577] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5706), 1,
      sym_END,
    ACTIONS(5708), 1,
      sym_GENERIC,
    ACTIONS(5710), 1,
      sym_IS,
    ACTIONS(5712), 1,
      sym_PORT,
    STATE(5692), 1,
      sym_generic_clause,
    STATE(6551), 1,
      sym_component_body,
    STATE(6765), 1,
      sym_port_clause,
    STATE(7481), 1,
      sym_end_component,
    STATE(4938), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [268616] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5747), 1,
      anon_sym_SEMI,
    STATE(5940), 1,
      sym__label,
    STATE(8184), 1,
      sym_when_expression,
    STATE(4939), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268649] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4940), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268674] = 12,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5706), 1,
      sym_END,
    ACTIONS(5708), 1,
      sym_GENERIC,
    ACTIONS(5710), 1,
      sym_IS,
    ACTIONS(5712), 1,
      sym_PORT,
    STATE(5692), 1,
      sym_generic_clause,
    STATE(6581), 1,
      sym_component_body,
    STATE(6765), 1,
      sym_port_clause,
    STATE(6829), 1,
      sym_end_component,
    STATE(4941), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [268713] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4942), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268738] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5749), 1,
      sym_OF,
    STATE(4943), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 7,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268765] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6344), 1,
      sym__identifier,
    STATE(6374), 1,
      sym__alias_designator,
    ACTIONS(5698), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4944), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268796] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4945), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268821] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4946), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268846] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4947), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268871] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5751), 1,
      anon_sym_SEMI,
    STATE(5922), 1,
      sym__label,
    STATE(8131), 1,
      sym_when_expression,
    STATE(4948), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [268904] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4949), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268929] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4950), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268954] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5491), 1,
      anon_sym_LPAREN,
    ACTIONS(5753), 1,
      sym_PARAMETER,
    STATE(4989), 1,
      sym_parenthesis_group,
    STATE(4951), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2191), 5,
      sym_AFTER,
      sym_GENERIC,
      anon_sym_DOT,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [268985] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4952), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269010] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6344), 1,
      sym__identifier,
    STATE(6404), 1,
      sym__alias_designator,
    ACTIONS(5698), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4953), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269041] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5755), 1,
      anon_sym_SEMI,
    STATE(5727), 1,
      sym__label,
    STATE(6917), 1,
      sym_when_expression,
    STATE(4954), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5702), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269074] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4955), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 8,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RANGE,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269099] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6344), 1,
      sym__identifier,
    STATE(6436), 1,
      sym__alias_designator,
    ACTIONS(5698), 2,
      sym_character_literal,
      sym_operator_symbol,
    STATE(4956), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269130] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5757), 1,
      sym_OF,
    STATE(4957), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 7,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269157] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6774), 1,
      sym__conditional_analysis_operator,
    ACTIONS(5761), 2,
      anon_sym_LT,
      anon_sym_GT,
    STATE(4958), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5759), 4,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
  [269185] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5763), 1,
      sym_character_literal,
    STATE(6153), 1,
      sym__identifier,
    STATE(6463), 1,
      sym_enumeration_literal,
    STATE(4959), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269215] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5765), 1,
      sym_operator_symbol,
    STATE(5082), 1,
      sym__designator,
    STATE(5535), 1,
      sym__identifier,
    STATE(4960), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269245] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5767), 1,
      anon_sym_SEMI,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5771), 1,
      sym_ON,
    ACTIONS(5773), 1,
      sym_UNTIL,
    STATE(5247), 1,
      sym_sensitivity_clause,
    STATE(5837), 1,
      sym_condition_clause,
    STATE(7418), 1,
      sym_timeout_clause,
    STATE(4961), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [269281] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4962), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5775), 7,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_RPAREN,
  [269305] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4963), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269329] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5779), 1,
      sym_operator_symbol,
    STATE(5481), 1,
      sym__designator,
    STATE(5569), 1,
      sym__identifier,
    STATE(4964), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5777), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269359] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4965), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269383] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5595), 1,
      sym_FOR,
    STATE(5066), 1,
      aux_sym_block_configuration_repeat2,
    STATE(6067), 1,
      sym__configuration_item,
    STATE(8356), 1,
      sym_end_for,
    STATE(6072), 2,
      sym_block_configuration,
      sym_component_configuration,
    STATE(4966), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [269417] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4967), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2652), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269441] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5781), 1,
      anon_sym_SEMI,
    ACTIONS(5783), 1,
      sym_CONFIGURATION,
    STATE(7179), 1,
      sym__identifier,
    STATE(4968), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269471] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5771), 1,
      sym_ON,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(5785), 1,
      anon_sym_SEMI,
    STATE(5282), 1,
      sym_sensitivity_clause,
    STATE(5925), 1,
      sym_condition_clause,
    STATE(8139), 1,
      sym_timeout_clause,
    STATE(4969), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [269507] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5771), 1,
      sym_ON,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(5787), 1,
      anon_sym_SEMI,
    STATE(5171), 1,
      sym_sensitivity_clause,
    STATE(5936), 1,
      sym_condition_clause,
    STATE(8186), 1,
      sym_timeout_clause,
    STATE(4970), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [269543] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5791), 1,
      sym_operator_symbol,
    STATE(5123), 1,
      sym__designator,
    STATE(5237), 1,
      sym__identifier,
    STATE(4971), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5789), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269573] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4972), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2644), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269597] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7527), 1,
      sym__conditional_analysis_operator,
    ACTIONS(5761), 2,
      anon_sym_LT,
      anon_sym_GT,
    STATE(4973), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5759), 4,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
  [269625] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5793), 1,
      anon_sym_SEMI,
    ACTIONS(5795), 1,
      sym_ENTITY,
    STATE(6624), 1,
      sym__identifier,
    STATE(4974), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269655] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4975), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2584), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269679] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5797), 1,
      anon_sym_SEMI,
    ACTIONS(5799), 1,
      anon_sym_QMARK,
    STATE(8203), 1,
      sym__label,
    STATE(4976), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269709] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4977), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2528), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269733] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5779), 1,
      sym_operator_symbol,
    STATE(5534), 1,
      sym__designator,
    STATE(5569), 1,
      sym__identifier,
    STATE(4978), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5777), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269763] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4979), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2552), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269787] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5763), 1,
      sym_character_literal,
    STATE(5747), 1,
      sym_enumeration_literal,
    STATE(6153), 1,
      sym__identifier,
    STATE(4980), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269817] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4981), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2516), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269841] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4982), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5801), 7,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_RPAREN,
  [269865] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4983), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2520), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269889] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4984), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2640), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [269913] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4985), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5644), 7,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_XNOR,
      sym_XOR,
      anon_sym_RPAREN,
  [269937] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5037), 1,
      aux_sym_partial_pathname_repeat1,
    STATE(5765), 1,
      sym__identifier,
    STATE(8278), 1,
      sym_pathname_element,
    STATE(4986), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [269967] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5527), 1,
      sym_FOR,
    ACTIONS(5803), 1,
      sym_USE,
    STATE(5023), 1,
      aux_sym_configuration_head_repeat2,
    STATE(6590), 1,
      sym_block_configuration,
    STATE(6787), 1,
      sym_end_for,
    STATE(8215), 1,
      sym_verification_unit_binding_indication,
    STATE(4987), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270003] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4988), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2512), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [270027] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4989), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2524), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [270051] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5807), 1,
      sym_GENERIC,
    ACTIONS(5809), 1,
      sym_IS,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(5813), 1,
      sym_RETURN,
    STATE(5502), 1,
      sym_subprogram_header,
    STATE(7109), 1,
      sym_parameter_list_specification,
    STATE(4990), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270087] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5771), 1,
      sym_ON,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(5815), 1,
      anon_sym_SEMI,
    STATE(5292), 1,
      sym_sensitivity_clause,
    STATE(5716), 1,
      sym_condition_clause,
    STATE(6935), 1,
      sym_timeout_clause,
    STATE(4991), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270123] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4992), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2556), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [270147] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5817), 1,
      sym_operator_symbol,
    STATE(5049), 1,
      sym__designator,
    STATE(5256), 1,
      sym__identifier,
    STATE(4993), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [270177] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5819), 1,
      anon_sym_SEMI,
    ACTIONS(5821), 1,
      anon_sym_LPAREN,
    ACTIONS(5823), 1,
      sym_GENERIC,
    ACTIONS(5825), 1,
      sym_IS,
    ACTIONS(5827), 1,
      sym_PARAMETER,
    STATE(5486), 1,
      sym_subprogram_header,
    STATE(6925), 1,
      sym_parameter_list_specification,
    STATE(4994), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270213] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5817), 1,
      sym_operator_symbol,
    STATE(4994), 1,
      sym__designator,
    STATE(5256), 1,
      sym__identifier,
    STATE(4995), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [270243] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5807), 1,
      sym_GENERIC,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(5829), 1,
      sym_IS,
    ACTIONS(5831), 1,
      sym_RETURN,
    STATE(5539), 1,
      sym_subprogram_header,
    STATE(7487), 1,
      sym_parameter_list_specification,
    STATE(4996), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270279] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5835), 1,
      sym_operator_symbol,
    STATE(4996), 1,
      sym__designator,
    STATE(5267), 1,
      sym__identifier,
    STATE(4997), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5833), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [270309] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4998), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2610), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [270333] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5807), 1,
      sym_GENERIC,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(5813), 1,
      sym_RETURN,
    ACTIONS(5837), 1,
      sym_IS,
    STATE(5502), 1,
      sym_subprogram_header,
    STATE(7109), 1,
      sym_parameter_list_specification,
    STATE(4999), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270369] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5839), 1,
      anon_sym_LPAREN,
    ACTIONS(5841), 1,
      sym_GENERIC,
    ACTIONS(5843), 1,
      sym_PARAMETER,
    STATE(5271), 1,
      sym_subprogram_header,
    STATE(6233), 1,
      sym_parameter_list_specification,
    ACTIONS(5819), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(5000), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270403] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5807), 1,
      sym_GENERIC,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(5813), 1,
      sym_RETURN,
    ACTIONS(5845), 1,
      sym_IS,
    STATE(5502), 1,
      sym_subprogram_header,
    STATE(7109), 1,
      sym_parameter_list_specification,
    STATE(5001), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270439] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(5847), 1,
      anon_sym_SEMI,
    ACTIONS(5849), 1,
      sym_USE,
    STATE(2963), 1,
      sym_binding_indication,
    STATE(5847), 1,
      sym_generic_map_aspect,
    STATE(7483), 1,
      sym_port_map_aspect,
    STATE(5002), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270475] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5807), 1,
      sym_GENERIC,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(5831), 1,
      sym_RETURN,
    ACTIONS(5851), 1,
      sym_IS,
    STATE(5539), 1,
      sym_subprogram_header,
    STATE(7487), 1,
      sym_parameter_list_specification,
    STATE(5003), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270511] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5807), 1,
      sym_GENERIC,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(5813), 1,
      sym_RETURN,
    ACTIONS(5853), 1,
      sym_IS,
    STATE(5502), 1,
      sym_subprogram_header,
    STATE(7109), 1,
      sym_parameter_list_specification,
    STATE(5004), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270547] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5835), 1,
      sym_operator_symbol,
    STATE(5031), 1,
      sym__designator,
    STATE(5267), 1,
      sym__identifier,
    STATE(5005), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5833), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [270577] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5855), 1,
      anon_sym_SEMI,
    ACTIONS(5857), 1,
      sym_COMPONENT,
    STATE(6916), 1,
      sym__identifier,
    STATE(5006), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [270607] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5513), 1,
      sym_RANGE,
    STATE(5249), 1,
      sym_range_constraint,
    STATE(5007), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5859), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
  [270635] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5861), 1,
      anon_sym_SEMI,
    ACTIONS(5863), 1,
      sym_ARCHITECTURE,
    STATE(7091), 1,
      sym__identifier,
    STATE(5008), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [270665] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5865), 1,
      anon_sym_SEMI,
    ACTIONS(5867), 1,
      sym_CONTEXT,
    STATE(8032), 1,
      sym__identifier,
    STATE(5009), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [270695] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5010), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2614), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [270719] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5771), 1,
      sym_ON,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(5869), 1,
      anon_sym_SEMI,
    STATE(5231), 1,
      sym_sensitivity_clause,
    STATE(5864), 1,
      sym_condition_clause,
    STATE(7567), 1,
      sym_timeout_clause,
    STATE(5011), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270755] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5012), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2492), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [270779] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5591), 1,
      sym_library_namespace,
    STATE(5807), 1,
      sym__identifier,
    STATE(6208), 1,
      sym__logical_name,
    STATE(5013), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [270809] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(5871), 1,
      anon_sym_SEMI,
    ACTIONS(5873), 1,
      sym_USE,
    STATE(1327), 1,
      sym_binding_indication,
    STATE(5737), 1,
      sym_generic_map_aspect,
    STATE(7009), 1,
      sym_port_map_aspect,
    STATE(5014), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [270845] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5015), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2560), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [270869] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5016), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2508), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [270893] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5017), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2564), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [270917] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5835), 1,
      sym_operator_symbol,
    STATE(5003), 1,
      sym__designator,
    STATE(5267), 1,
      sym__identifier,
    STATE(5018), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5833), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [270947] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5875), 1,
      anon_sym_SEMI,
    ACTIONS(5877), 1,
      sym_PACKAGE,
    STATE(7393), 1,
      sym__identifier,
    STATE(5019), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [270977] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5020), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2636), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [271001] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5771), 1,
      sym_ON,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(5879), 1,
      anon_sym_SEMI,
    STATE(5303), 1,
      sym_sensitivity_clause,
    STATE(5875), 1,
      sym_condition_clause,
    STATE(7634), 1,
      sym_timeout_clause,
    STATE(5021), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [271037] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5022), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2830), 7,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_EQ_GT,
      sym_variable_assignment,
      anon_sym_GT_GT,
      anon_sym_PIPE,
  [271061] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5527), 1,
      sym_FOR,
    ACTIONS(5803), 1,
      sym_USE,
    STATE(5233), 1,
      aux_sym_configuration_head_repeat2,
    STATE(6600), 1,
      sym_block_configuration,
    STATE(6806), 1,
      sym_end_for,
    STATE(8215), 1,
      sym_verification_unit_binding_indication,
    STATE(5023), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [271097] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5881), 1,
      anon_sym_SEMI,
    ACTIONS(5883), 1,
      sym_PACKAGE,
    STATE(6885), 1,
      sym__identifier,
    STATE(5024), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271127] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5025), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2568), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [271151] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5026), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3794), 7,
      sym_identifier,
      sym_NOT,
      sym_directive_constant_builtin,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
      anon_sym_LPAREN,
  [271175] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5027), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5801), 7,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
  [271199] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5527), 1,
      sym_FOR,
    ACTIONS(5803), 1,
      sym_USE,
    STATE(5233), 1,
      aux_sym_configuration_head_repeat2,
    STATE(6590), 1,
      sym_block_configuration,
    STATE(6787), 1,
      sym_end_for,
    STATE(8215), 1,
      sym_verification_unit_binding_indication,
    STATE(5028), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [271235] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5029), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2386), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [271259] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5030), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2394), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [271283] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5807), 1,
      sym_GENERIC,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(5813), 1,
      sym_RETURN,
    ACTIONS(5885), 1,
      sym_IS,
    STATE(5502), 1,
      sym_subprogram_header,
    STATE(7109), 1,
      sym_parameter_list_specification,
    STATE(5031), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [271319] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5765), 1,
      sym_operator_symbol,
    STATE(5058), 1,
      sym__designator,
    STATE(5535), 1,
      sym__identifier,
    STATE(5032), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(4947), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271349] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5817), 1,
      sym_operator_symbol,
    STATE(5000), 1,
      sym__designator,
    STATE(5256), 1,
      sym__identifier,
    STATE(5033), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5423), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271379] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5595), 1,
      sym_FOR,
    STATE(5066), 1,
      aux_sym_block_configuration_repeat2,
    STATE(6067), 1,
      sym__configuration_item,
    STATE(8046), 1,
      sym_end_for,
    STATE(6072), 2,
      sym_block_configuration,
      sym_component_configuration,
    STATE(5034), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [271413] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5035), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2648), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [271437] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5036), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2390), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [271461] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6160), 1,
      sym__identifier,
    STATE(8278), 1,
      sym_pathname_element,
    ACTIONS(5887), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(5037), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_partial_pathname_repeat1,
  [271489] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5038), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [271513] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5039), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5775), 7,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
  [271537] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5595), 1,
      sym_FOR,
    STATE(5066), 1,
      aux_sym_block_configuration_repeat2,
    STATE(6067), 1,
      sym__configuration_item,
    STATE(8049), 1,
      sym_end_for,
    STATE(6072), 2,
      sym_block_configuration,
      sym_component_configuration,
    STATE(5040), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [271571] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5041), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [271595] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5042), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2452), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [271619] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5043), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2456), 7,
      sym_AFTER,
      sym_GENERIC,
      sym_PARAMETER,
      anon_sym_DOT,
      anon_sym_LPAREN,
      anon_sym_SQUOTE,
      anon_sym_LBRACK,
  [271643] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5835), 1,
      sym_operator_symbol,
    STATE(4990), 1,
      sym__designator,
    STATE(5267), 1,
      sym__identifier,
    STATE(5044), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5833), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271673] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5835), 1,
      sym_operator_symbol,
    STATE(4999), 1,
      sym__designator,
    STATE(5267), 1,
      sym__identifier,
    STATE(5045), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5833), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271703] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5835), 1,
      sym_operator_symbol,
    STATE(5001), 1,
      sym__designator,
    STATE(5267), 1,
      sym__identifier,
    STATE(5046), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5833), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271733] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5835), 1,
      sym_operator_symbol,
    STATE(5004), 1,
      sym__designator,
    STATE(5267), 1,
      sym__identifier,
    STATE(5047), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5833), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271763] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5048), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5644), 7,
      sym_AND,
      sym_NAND,
      sym_NOR,
      sym_OR,
      sym_THEN,
      sym_XNOR,
      sym_XOR,
  [271787] = 11,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5819), 1,
      anon_sym_SEMI,
    ACTIONS(5821), 1,
      anon_sym_LPAREN,
    ACTIONS(5823), 1,
      sym_GENERIC,
    ACTIONS(5827), 1,
      sym_PARAMETER,
    ACTIONS(5890), 1,
      sym_IS,
    STATE(5486), 1,
      sym_subprogram_header,
    STATE(6925), 1,
      sym_parameter_list_specification,
    STATE(5049), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [271823] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5595), 1,
      sym_FOR,
    STATE(5066), 1,
      aux_sym_block_configuration_repeat2,
    STATE(6067), 1,
      sym__configuration_item,
    STATE(6768), 1,
      sym_end_for,
    STATE(6072), 2,
      sym_block_configuration,
      sym_component_configuration,
    STATE(5050), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [271857] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5892), 1,
      anon_sym_SEMI,
    STATE(7460), 1,
      sym__label,
    STATE(5051), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271884] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7791), 1,
      sym_identifier_list,
    STATE(5052), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271911] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(6821), 1,
      sym_identifier_list,
    STATE(5053), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271938] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7803), 1,
      sym_identifier_list,
    STATE(5054), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271965] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7496), 1,
      sym_identifier_list,
    STATE(5055), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [271992] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5894), 1,
      sym_ELSE,
    ACTIONS(5896), 1,
      sym_ELSIF,
    ACTIONS(5898), 1,
      sym_END,
    STATE(6756), 1,
      sym__elsif_else_generate,
    STATE(7168), 2,
      sym_elsif_generate,
      sym_else_generate,
    STATE(5056), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [272023] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7708), 1,
      sym__identifier,
    STATE(7744), 1,
      sym_parameter_specification,
    STATE(5057), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5900), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272050] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5807), 1,
      sym_GENERIC,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(5902), 1,
      sym_RETURN,
    STATE(5511), 1,
      sym_subprogram_header,
    STATE(7189), 1,
      sym_parameter_list_specification,
    STATE(5058), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [272083] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5801), 1,
      sym__identifier,
    STATE(7825), 1,
      sym_constant_identifier_list,
    STATE(5059), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272110] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7826), 1,
      sym_identifier_list,
    STATE(5060), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272137] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5904), 1,
      sym_BODY,
    STATE(6468), 1,
      sym__identifier,
    STATE(5061), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272164] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7832), 1,
      sym_identifier_list,
    STATE(5062), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272191] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5801), 1,
      sym__identifier,
    STATE(7276), 1,
      sym_constant_identifier_list,
    STATE(5063), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272218] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7841), 1,
      sym_identifier_list,
    STATE(5064), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272245] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5065), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 6,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [272268] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5906), 1,
      sym_END,
    ACTIONS(5908), 1,
      sym_FOR,
    STATE(6067), 1,
      sym__configuration_item,
    STATE(6072), 2,
      sym_block_configuration,
      sym_component_configuration,
    STATE(5066), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_block_configuration_repeat2,
  [272297] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5913), 2,
      anon_sym_LT,
      anon_sym_GT,
    STATE(5067), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5911), 4,
      anon_sym_LT_EQ,
      anon_sym_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_GT_EQ,
  [272322] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5801), 1,
      sym__identifier,
    STATE(7855), 1,
      sym_constant_identifier_list,
    STATE(5068), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272349] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7856), 1,
      sym_identifier_list,
    STATE(5069), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272376] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7859), 1,
      sym_identifier_list,
    STATE(5070), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272403] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7862), 1,
      sym_identifier_list,
    STATE(5071), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272430] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5915), 1,
      anon_sym_SEMI,
    STATE(7280), 1,
      sym__identifier,
    STATE(5072), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272457] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7816), 1,
      sym_identifier_list,
    STATE(5073), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272484] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7871), 1,
      sym_identifier_list,
    STATE(5074), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272511] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7874), 1,
      sym_identifier_list,
    STATE(5075), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272538] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5917), 1,
      anon_sym_SEMI,
    STATE(8129), 1,
      sym__identifier,
    STATE(5076), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272565] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7493), 1,
      sym_identifier_list,
    STATE(5077), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272592] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6744), 1,
      sym__identifier,
    STATE(6949), 1,
      sym_parameter_specification,
    STATE(5078), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5900), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272619] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5919), 1,
      anon_sym_SEMI,
    STATE(7267), 1,
      sym__identifier,
    STATE(5079), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272646] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5104), 1,
      aux_sym_package_pathname_repeat1,
    STATE(6349), 1,
      sym__identifier,
    STATE(5080), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272673] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(5921), 1,
      anon_sym_SEMI,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    STATE(6566), 1,
      sym_generic_map_aspect,
    STATE(7520), 1,
      sym_port_map_aspect,
    STATE(5081), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [272706] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5807), 1,
      sym_GENERIC,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(5925), 1,
      sym_RETURN,
    STATE(5542), 1,
      sym_subprogram_header,
    STATE(7549), 1,
      sym_parameter_list_specification,
    STATE(5082), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [272739] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5927), 1,
      anon_sym_SEMI,
    STATE(7526), 1,
      sym__identifier,
    STATE(5083), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272766] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5929), 1,
      anon_sym_SEMI,
    STATE(7000), 1,
      sym__label,
    STATE(5084), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272793] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5933), 1,
      sym_WHEN,
    STATE(6496), 1,
      sym_when_expression,
    ACTIONS(5931), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(5085), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_conditional_expression_repeat1,
  [272820] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5936), 1,
      sym_ALL,
    STATE(5798), 1,
      sym__identifier,
    STATE(5086), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272847] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5938), 1,
      anon_sym_SEMI,
    STATE(6831), 1,
      sym__identifier,
    STATE(5087), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272874] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5088), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5940), 6,
      sym_END,
      sym_FOR,
      sym_GENERIC,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
  [272897] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5942), 1,
      anon_sym_SEMI,
    STATE(8250), 1,
      sym__label,
    STATE(5089), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [272924] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(5944), 1,
      anon_sym_SEMI,
    STATE(6479), 1,
      sym_generic_map_aspect,
    STATE(8357), 1,
      sym_port_map_aspect,
    STATE(5090), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [272957] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(235), 1,
      sym_ELSE,
    ACTIONS(237), 1,
      sym_ELSIF,
    ACTIONS(5946), 1,
      sym_END,
    STATE(8056), 1,
      sym__elsif_else_statement,
    STATE(7427), 2,
      sym_elsif_statement,
      sym_else_statement,
    STATE(5091), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [272988] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(235), 1,
      sym_ELSE,
    ACTIONS(237), 1,
      sym_ELSIF,
    ACTIONS(5948), 1,
      sym_END,
    STATE(6772), 1,
      sym__elsif_else_statement,
    STATE(7427), 2,
      sym_elsif_statement,
      sym_else_statement,
    STATE(5092), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [273019] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5950), 1,
      anon_sym_SEMI,
    STATE(7463), 1,
      sym__identifier,
    STATE(5093), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273046] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5952), 1,
      anon_sym_SEMI,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    STATE(5650), 1,
      sym_signal_kind,
    STATE(6996), 1,
      sym_initialiser,
    ACTIONS(5956), 2,
      sym_BUS,
      sym_REGISTER,
    STATE(5094), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [273077] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(6834), 1,
      sym_identifier_list,
    STATE(5095), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273104] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5958), 1,
      anon_sym_CARET,
    ACTIONS(5961), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(5096), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_relative_pathname_repeat1,
  [273129] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6782), 1,
      sym__identifier,
    ACTIONS(5963), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
    STATE(5097), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_package_pathname_repeat1,
  [273154] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5966), 1,
      anon_sym_SEMI,
    STATE(7321), 1,
      sym__label,
    STATE(5098), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273181] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5801), 1,
      sym__identifier,
    STATE(7660), 1,
      sym_constant_identifier_list,
    STATE(5099), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273208] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7157), 1,
      sym_identifier_list,
    STATE(5100), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273235] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5968), 1,
      anon_sym_SEMI,
    STATE(7177), 1,
      sym__label,
    STATE(5101), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273262] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5970), 1,
      sym_BODY,
    STATE(6437), 1,
      sym__identifier,
    STATE(5102), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273289] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7789), 1,
      sym_identifier_list,
    STATE(5103), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273316] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5097), 1,
      aux_sym_package_pathname_repeat1,
    STATE(6568), 1,
      sym__identifier,
    STATE(5104), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273343] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7662), 1,
      sym_identifier_list,
    STATE(5105), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273370] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5972), 1,
      sym_BODY,
    STATE(6375), 1,
      sym__identifier,
    STATE(5106), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273397] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7668), 1,
      sym_identifier_list,
    STATE(5107), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273424] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5974), 1,
      anon_sym_SEMI,
    STATE(7470), 1,
      sym__label,
    STATE(5108), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273451] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(5976), 1,
      anon_sym_SEMI,
    STATE(5881), 1,
      sym_signal_kind,
    STATE(7648), 1,
      sym_initialiser,
    ACTIONS(5956), 2,
      sym_BUS,
      sym_REGISTER,
    STATE(5109), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [273482] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7670), 1,
      sym_identifier_list,
    STATE(5110), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273509] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5894), 1,
      sym_ELSE,
    ACTIONS(5896), 1,
      sym_ELSIF,
    ACTIONS(5978), 1,
      sym_END,
    STATE(8190), 1,
      sym__elsif_else_generate,
    STATE(7168), 2,
      sym_elsif_generate,
      sym_else_generate,
    STATE(5111), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [273540] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5517), 1,
      sym_RANGE,
    STATE(5392), 1,
      sym_range_constraint,
    STATE(5112), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5859), 4,
      sym_BUS,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
  [273567] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7688), 1,
      sym_identifier_list,
    STATE(5113), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273594] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5980), 1,
      anon_sym_SEMI,
    STATE(6625), 1,
      sym__identifier,
    STATE(5114), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273621] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5982), 1,
      sym_BODY,
    STATE(6172), 1,
      sym__identifier,
    STATE(5115), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273648] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(5984), 1,
      anon_sym_SEMI,
    STATE(5736), 1,
      sym_signal_kind,
    STATE(7007), 1,
      sym_initialiser,
    ACTIONS(5956), 2,
      sym_BUS,
      sym_REGISTER,
    STATE(5116), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [273679] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5986), 1,
      anon_sym_SEMI,
    STATE(7398), 1,
      sym__identifier,
    STATE(5117), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273706] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5988), 1,
      anon_sym_SEMI,
    STATE(7536), 1,
      sym__label,
    STATE(5118), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273733] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4739), 1,
      sym_primary_unit_declaration,
    STATE(8196), 1,
      sym__identifier,
    STATE(5119), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273760] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(5990), 1,
      anon_sym_SEMI,
    STATE(6041), 1,
      sym_generic_map_aspect,
    STATE(8112), 1,
      sym_port_map_aspect,
    STATE(5120), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [273793] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5994), 1,
      sym_GENERIC,
    STATE(5515), 1,
      sym_generic_map_aspect,
    STATE(5121), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5992), 4,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
  [273820] = 10,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(5996), 1,
      anon_sym_SEMI,
    STATE(6104), 1,
      sym_generic_map_aspect,
    STATE(6655), 1,
      sym_port_map_aspect,
    STATE(5122), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [273853] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5839), 1,
      anon_sym_LPAREN,
    ACTIONS(5843), 1,
      sym_PARAMETER,
    STATE(5824), 1,
      sym_parameter_list_specification,
    ACTIONS(5998), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5123), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [273882] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3588), 1,
      sym__identifier,
    STATE(6522), 1,
      sym_record_element_resolution,
    STATE(5124), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6000), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273909] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6002), 1,
      anon_sym_SEMI,
    STATE(6824), 1,
      sym__identifier,
    STATE(5125), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273936] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(6830), 1,
      sym_identifier_list,
    STATE(5126), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273963] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5801), 1,
      sym__identifier,
    STATE(7723), 1,
      sym_constant_identifier_list,
    STATE(5127), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [273990] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5801), 1,
      sym__identifier,
    STATE(6820), 1,
      sym_constant_identifier_list,
    STATE(5128), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274017] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5894), 1,
      sym_ELSE,
    ACTIONS(5896), 1,
      sym_ELSIF,
    ACTIONS(6004), 1,
      sym_END,
    STATE(8314), 1,
      sym__elsif_else_generate,
    STATE(7168), 2,
      sym_elsif_generate,
      sym_else_generate,
    STATE(5129), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [274048] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6006), 1,
      anon_sym_SEMI,
    STATE(8319), 1,
      sym__label,
    STATE(5130), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274075] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6008), 1,
      anon_sym_SEMI,
    STATE(8335), 1,
      sym__identifier,
    STATE(5131), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274102] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7725), 1,
      sym_identifier_list,
    STATE(5132), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274129] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6010), 1,
      anon_sym_SEMI,
    STATE(7673), 1,
      sym__label,
    STATE(5133), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274156] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6012), 1,
      sym_BODY,
    STATE(6405), 1,
      sym__identifier,
    STATE(5134), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274183] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7731), 1,
      sym_identifier_list,
    STATE(5135), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274210] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7733), 1,
      sym_identifier_list,
    STATE(5136), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274237] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5994), 1,
      sym_GENERIC,
    STATE(5506), 1,
      sym_generic_map_aspect,
    STATE(5137), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6014), 4,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
  [274264] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6016), 1,
      anon_sym_SEMI,
    STATE(5941), 1,
      sym_signal_kind,
    STATE(8200), 1,
      sym_initialiser,
    ACTIONS(5956), 2,
      sym_BUS,
      sym_REGISTER,
    STATE(5138), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [274295] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7750), 1,
      sym_identifier_list,
    STATE(5139), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274322] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5519), 1,
      sym_RANGE,
    STATE(5479), 1,
      sym_range_constraint,
    STATE(5140), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5859), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [274349] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6018), 1,
      anon_sym_SEMI,
    STATE(6723), 1,
      sym__identifier,
    STATE(5141), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274376] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6020), 1,
      anon_sym_SEMI,
    STATE(6729), 1,
      sym__identifier,
    STATE(5142), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274403] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6022), 1,
      anon_sym_SEMI,
    STATE(7851), 1,
      sym__label,
    STATE(5143), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274430] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5801), 1,
      sym__identifier,
    STATE(7781), 1,
      sym_constant_identifier_list,
    STATE(5144), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274457] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    STATE(5085), 1,
      aux_sym_conditional_expression_repeat1,
    STATE(6496), 1,
      sym_when_expression,
    ACTIONS(6024), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(5145), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [274486] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5848), 1,
      sym__identifier,
    STATE(7783), 1,
      sym_identifier_list,
    STATE(5146), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274513] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5894), 1,
      sym_ELSE,
    ACTIONS(5896), 1,
      sym_ELSIF,
    ACTIONS(6026), 1,
      sym_END,
    STATE(7167), 1,
      sym__elsif_else_generate,
    STATE(7168), 2,
      sym_elsif_generate,
      sym_else_generate,
    STATE(5147), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [274544] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7227), 1,
      sym__identifier,
    STATE(5148), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274568] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4848), 1,
      sym__suffix,
    STATE(5149), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6028), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [274592] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4519), 1,
      sym__suffix,
    STATE(5150), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6030), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [274616] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6032), 1,
      anon_sym_SEMI,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    STATE(5945), 1,
      sym_signature,
    STATE(8218), 1,
      sym_generic_map_aspect,
    STATE(5151), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [274646] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5152), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6036), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
  [274668] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(6038), 1,
      anon_sym_SEMI,
    STATE(5680), 1,
      sym_generic_map_aspect,
    STATE(6757), 1,
      sym_port_map_aspect,
    STATE(5153), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [274698] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6058), 1,
      sym__identifier,
    STATE(5154), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274722] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6040), 1,
      sym_GENERIC,
    STATE(5900), 1,
      sym_generic_map_aspect,
    ACTIONS(5992), 3,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
    STATE(5155), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [274748] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(8155), 1,
      sym__identifier,
    STATE(5156), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274772] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6369), 1,
      sym__identifier,
    STATE(5157), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274796] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5158), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5961), 5,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
      anon_sym_CARET,
  [274818] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7858), 1,
      sym__identifier,
    STATE(5159), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274842] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6042), 1,
      anon_sym_SEMI,
    STATE(5667), 1,
      sym_generic_map_aspect,
    STATE(7737), 1,
      sym_initialiser,
    STATE(5160), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [274872] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4729), 1,
      sym__suffix,
    STATE(5161), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6044), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [274896] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6909), 1,
      sym__identifier,
    STATE(5162), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [274920] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2091), 1,
      sym__suffix,
    STATE(5163), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6046), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [274944] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6048), 1,
      anon_sym_SEMI,
    STATE(5942), 1,
      sym_generic_map_aspect,
    STATE(8202), 1,
      sym_initialiser,
    STATE(5164), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [274974] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1557), 1,
      sym__suffix,
    STATE(5165), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6050), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [274998] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6052), 1,
      sym_GENERIC,
    STATE(6517), 1,
      sym__interface_package_generic_map_aspect,
    STATE(5166), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    STATE(6535), 3,
      sym_generic_map_any,
      sym_generic_map_default,
      sym_generic_map_aspect,
  [275024] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6054), 1,
      anon_sym_SEMI,
    STATE(5688), 1,
      sym_generic_map_aspect,
    STATE(7033), 1,
      sym_initialiser,
    STATE(5167), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275054] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6056), 1,
      anon_sym_SEMI,
    STATE(5715), 1,
      sym_signature,
    STATE(6719), 1,
      sym_generic_map_aspect,
    STATE(5168), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275084] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6058), 1,
      anon_sym_SEMI,
    STATE(5724), 1,
      sym_signature,
    STATE(6728), 1,
      sym_generic_map_aspect,
    STATE(5169), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275114] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5170), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6060), 5,
      sym_identifier,
      sym_END,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [275136] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(6062), 1,
      anon_sym_SEMI,
    STATE(5949), 1,
      sym_condition_clause,
    STATE(8228), 1,
      sym_timeout_clause,
    STATE(5171), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275166] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6064), 1,
      anon_sym_SEMI,
    STATE(5729), 1,
      sym_generic_map_aspect,
    STATE(6734), 1,
      sym_initialiser,
    STATE(5172), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275196] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5533), 1,
      sym_RANGE,
    STATE(5918), 1,
      sym_range_constraint,
    ACTIONS(5859), 3,
      sym_IS,
      sym_OPEN,
      anon_sym_SEMI,
    STATE(5173), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275222] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6994), 1,
      sym__identifier,
    STATE(5174), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [275246] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6068), 1,
      anon_sym_COMMA,
    ACTIONS(6066), 3,
      sym_FOR,
      sym_UNTIL,
      anon_sym_SEMI,
    STATE(5175), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_group_constituent_list_repeat1,
  [275270] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5708), 1,
      sym_GENERIC,
    ACTIONS(5712), 1,
      sym_PORT,
    ACTIONS(6071), 1,
      sym_END,
    STATE(5711), 1,
      sym_generic_clause,
    STATE(6921), 1,
      sym_port_clause,
    STATE(5176), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275300] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2119), 1,
      sym__suffix,
    STATE(5177), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6073), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [275324] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(6075), 1,
      anon_sym_SEMI,
    STATE(5854), 1,
      sym_generic_map_aspect,
    STATE(7518), 1,
      sym_port_map_aspect,
    STATE(5178), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275354] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5690), 1,
      sym__identifier,
    STATE(5179), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [275378] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6828), 1,
      sym__identifier,
    STATE(5180), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [275402] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5803), 1,
      sym_USE,
    STATE(5514), 1,
      aux_sym_configuration_head_repeat2,
    STATE(7694), 1,
      sym_verification_unit_binding_indication,
    STATE(7861), 1,
      sym_end_for,
    STATE(5181), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275432] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5775), 1,
      aux_sym_index_constraint_repeat1,
    ACTIONS(6077), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    ACTIONS(6079), 2,
      anon_sym_EQ_GT,
      anon_sym_PIPE,
    STATE(5182), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275458] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6081), 1,
      anon_sym_SEMI,
    STATE(5951), 1,
      sym_generic_map_aspect,
    STATE(8243), 1,
      sym_initialiser,
    STATE(5183), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275488] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6083), 1,
      anon_sym_SEMI,
    STATE(5596), 1,
      sym_signature,
    STATE(7135), 1,
      sym_generic_map_aspect,
    STATE(5184), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275518] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6085), 1,
      anon_sym_SEMI,
    STATE(5770), 1,
      sym_signature,
    STATE(7399), 1,
      sym_generic_map_aspect,
    STATE(5185), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275548] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5186), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6087), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
  [275570] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2154), 1,
      sym__suffix,
    STATE(5187), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6089), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [275594] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5803), 1,
      sym_USE,
    STATE(5307), 1,
      aux_sym_configuration_head_repeat2,
    STATE(7302), 1,
      sym_end_for,
    STATE(7694), 1,
      sym_verification_unit_binding_indication,
    STATE(5188), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275624] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(8360), 1,
      sym__identifier,
    STATE(5189), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [275648] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(971), 1,
      sym__suffix,
    STATE(5190), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6091), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [275672] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7669), 1,
      sym__identifier,
    STATE(5191), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [275696] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6093), 1,
      sym_END,
    ACTIONS(6095), 1,
      sym_WHEN,
    STATE(6086), 1,
      sym_case_statement_alternative,
    STATE(6940), 1,
      sym_when_element,
    STATE(5192), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_case_body_repeat1,
  [275724] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6098), 1,
      anon_sym_SEMI,
    STATE(5907), 1,
      sym_generic_map_aspect,
    STATE(7869), 1,
      sym_initialiser,
    STATE(5193), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275754] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6459), 1,
      sym__identifier,
    STATE(5194), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [275778] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6381), 1,
      sym__identifier,
    STATE(5195), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [275802] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3406), 1,
      sym__suffix,
    STATE(5196), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6100), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [275826] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4938), 1,
      sym__identifier,
    STATE(5197), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6102), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [275850] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4941), 1,
      sym__identifier,
    STATE(5198), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6102), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [275874] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6104), 1,
      anon_sym_SEMI,
    STATE(5602), 1,
      sym_signature,
    STATE(7196), 1,
      sym_generic_map_aspect,
    STATE(5199), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275904] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6040), 1,
      sym_GENERIC,
    STATE(5811), 1,
      sym_generic_map_aspect,
    ACTIONS(6014), 3,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
    STATE(5200), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275930] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4714), 1,
      sym__suffix,
    STATE(5201), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6106), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [275954] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6108), 1,
      anon_sym_SEMI,
    STATE(5813), 1,
      sym_signature,
    STATE(7312), 1,
      sym_generic_map_aspect,
    STATE(5202), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [275984] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3527), 1,
      sym__suffix,
    STATE(5203), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6110), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [276008] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5589), 1,
      sym_VUNIT,
    ACTIONS(6112), 1,
      sym_CONFIGURATION,
    ACTIONS(6114), 1,
      sym_ENTITY,
    ACTIONS(6116), 1,
      sym_OPEN,
    STATE(5226), 1,
      sym_entity_aspect,
    STATE(5204), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276038] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1742), 1,
      sym__suffix,
    STATE(5205), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6118), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [276062] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6120), 1,
      anon_sym_SEMI,
    STATE(5805), 1,
      sym_signature,
    STATE(7600), 1,
      sym_generic_map_aspect,
    STATE(5206), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276092] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(6122), 1,
      anon_sym_SEMI,
    STATE(5630), 1,
      sym_generic_map_aspect,
    STATE(8315), 1,
      sym_port_map_aspect,
    STATE(5207), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276122] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6124), 1,
      anon_sym_SEMI,
    STATE(5612), 1,
      sym_generic_map_aspect,
    STATE(7208), 1,
      sym_initialiser,
    STATE(5208), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276152] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6126), 1,
      anon_sym_SEMI,
    STATE(5817), 1,
      sym_generic_map_aspect,
    STATE(7642), 1,
      sym_initialiser,
    STATE(5209), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276182] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4636), 1,
      sym__suffix,
    STATE(5210), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6128), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [276206] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(8194), 1,
      sym__identifier,
    STATE(5211), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6130), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [276230] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(6132), 1,
      anon_sym_SEMI,
    STATE(5626), 1,
      sym_generic_map_aspect,
    STATE(8347), 1,
      sym_port_map_aspect,
    STATE(5212), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276260] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1987), 1,
      sym__suffix,
    STATE(5213), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6134), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [276284] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(854), 1,
      sym__suffix,
    STATE(5214), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6136), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [276308] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7158), 1,
      sym__identifier,
    STATE(5215), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [276332] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(6138), 1,
      anon_sym_SEMI,
    STATE(5871), 1,
      sym_generic_map_aspect,
    STATE(7614), 1,
      sym_port_map_aspect,
    STATE(5216), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276362] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6140), 1,
      anon_sym_COMMA,
    ACTIONS(6066), 3,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
    STATE(5217), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_group_constituent_list_repeat1,
  [276386] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5535), 1,
      sym_RANGE,
    STATE(5625), 1,
      sym_range_constraint,
    ACTIONS(5859), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(5218), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276412] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6143), 1,
      sym_GENERIC,
    STATE(5926), 1,
      sym_generic_map_aspect,
    ACTIONS(6014), 3,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
    STATE(5219), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276438] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6145), 1,
      anon_sym_SEMI,
    STATE(5669), 1,
      sym_signature,
    STATE(7817), 1,
      sym_generic_map_aspect,
    STATE(5220), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276468] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5221), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 5,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [276490] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6085), 1,
      sym__identifier,
    STATE(5222), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [276514] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2234), 1,
      sym__suffix,
    STATE(5223), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6147), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [276538] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3290), 1,
      sym__suffix,
    STATE(5224), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6149), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [276562] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5225), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6151), 5,
      sym_identifier,
      sym_END,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [276584] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(6153), 1,
      anon_sym_SEMI,
    STATE(5932), 1,
      sym_generic_map_aspect,
    STATE(8171), 1,
      sym_port_map_aspect,
    STATE(5226), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276614] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(931), 1,
      sym__suffix,
    STATE(5227), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6155), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [276638] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6159), 1,
      anon_sym_DOT,
    STATE(5228), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6157), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [276662] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3272), 1,
      sym_WHEN,
    ACTIONS(6161), 1,
      sym_END,
    STATE(5192), 1,
      aux_sym_case_body_repeat1,
    STATE(6086), 1,
      sym_case_statement_alternative,
    STATE(6940), 1,
      sym_when_element,
    STATE(5229), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276692] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5230), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6163), 5,
      sym_identifier,
      sym_END,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [276714] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(5879), 1,
      anon_sym_SEMI,
    STATE(5875), 1,
      sym_condition_clause,
    STATE(7634), 1,
      sym_timeout_clause,
    STATE(5231), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276744] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4946), 1,
      sym__suffix,
    STATE(5232), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6165), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [276768] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6169), 1,
      sym_USE,
    STATE(8215), 1,
      sym_verification_unit_binding_indication,
    ACTIONS(6167), 2,
      sym_END,
      sym_FOR,
    STATE(5233), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_configuration_head_repeat2,
  [276794] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6174), 1,
      sym_BUS,
    STATE(6184), 1,
      sym_initialiser,
    ACTIONS(6172), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5234), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276822] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6176), 1,
      anon_sym_DOT,
    STATE(5235), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6157), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [276846] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5803), 1,
      sym_USE,
    STATE(5181), 1,
      aux_sym_configuration_head_repeat2,
    STATE(7694), 1,
      sym_verification_unit_binding_indication,
    STATE(7703), 1,
      sym_end_for,
    STATE(5236), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276876] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5237), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6178), 5,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
  [276898] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6182), 1,
      sym_IS,
    STATE(6054), 1,
      sym_context_declaration_body,
    ACTIONS(6180), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
    STATE(5238), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276924] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4906), 1,
      sym__identifier,
    STATE(5239), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6102), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [276948] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6184), 1,
      anon_sym_SEMI,
    STATE(5674), 1,
      sym_signature,
    STATE(8023), 1,
      sym_generic_map_aspect,
    STATE(5240), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [276978] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4442), 1,
      sym__suffix,
    STATE(5241), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6186), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277002] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7732), 1,
      sym__identifier,
    STATE(5242), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277026] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6190), 1,
      sym_USE,
    STATE(5816), 1,
      sym_use_clause,
    ACTIONS(6188), 2,
      sym_END,
      sym_FOR,
    STATE(5243), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_block_configuration_repeat1,
  [277052] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6462), 1,
      sym__identifier,
    STATE(5244), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277076] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6146), 1,
      sym__label,
    STATE(5245), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5667), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277100] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6143), 1,
      sym_GENERIC,
    STATE(5579), 1,
      sym_generic_map_aspect,
    ACTIONS(5992), 3,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
    STATE(5246), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [277126] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(5815), 1,
      anon_sym_SEMI,
    STATE(5716), 1,
      sym_condition_clause,
    STATE(6935), 1,
      sym_timeout_clause,
    STATE(5247), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [277156] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6412), 1,
      sym__identifier,
    STATE(5248), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277180] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5249), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6193), 5,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      sym_variable_assignment,
      anon_sym_GT_GT,
  [277202] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4580), 1,
      sym__suffix,
    STATE(5250), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6195), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277226] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6138), 1,
      sym__identifier,
    STATE(5251), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277250] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6197), 1,
      anon_sym_SEMI,
    STATE(5947), 1,
      sym_signature,
    STATE(8226), 1,
      sym_generic_map_aspect,
    STATE(5252), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [277280] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6199), 1,
      anon_sym_SEMI,
    STATE(5883), 1,
      sym_generic_map_aspect,
    STATE(7650), 1,
      sym_initialiser,
    STATE(5253), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [277310] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6203), 1,
      anon_sym_COMMA,
    STATE(5175), 1,
      aux_sym_group_constituent_list_repeat1,
    ACTIONS(6201), 3,
      sym_FOR,
      sym_UNTIL,
      anon_sym_SEMI,
    STATE(5254), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [277336] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1209), 1,
      sym__suffix,
    STATE(5255), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6205), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277360] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5256), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6178), 5,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
  [277382] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(5529), 1,
      sym_GENERIC,
    ACTIONS(6207), 1,
      anon_sym_SEMI,
    STATE(5917), 1,
      sym_generic_map_aspect,
    STATE(8108), 1,
      sym_port_map_aspect,
    STATE(5257), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [277412] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3816), 1,
      sym__suffix,
    STATE(5258), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6209), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277436] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4554), 1,
      sym__suffix,
    STATE(5259), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6211), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277460] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6213), 1,
      anon_sym_SEMI,
    STATE(5749), 1,
      sym_generic_map_aspect,
    STATE(7062), 1,
      sym_initialiser,
    STATE(5260), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [277490] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1005), 1,
      sym__suffix,
    STATE(5261), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6215), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277514] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5262), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6217), 5,
      sym_identifier,
      sym_END,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277536] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7242), 1,
      sym__identifier,
    STATE(5263), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6130), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277560] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1703), 1,
      sym__suffix,
    STATE(5264), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6219), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277584] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1246), 1,
      sym__suffix,
    STATE(5265), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6221), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277608] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5266), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 5,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
  [277630] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5267), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6178), 5,
      sym_GENERIC,
      sym_IS,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
  [277652] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6322), 1,
      sym__identifier,
    STATE(5268), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277676] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4483), 1,
      sym__suffix,
    STATE(5269), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6223), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277700] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6225), 1,
      anon_sym_SEMI,
    STATE(5589), 1,
      sym_generic_map_aspect,
    STATE(7032), 1,
      sym_initialiser,
    STATE(5270), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [277730] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5839), 1,
      anon_sym_LPAREN,
    ACTIONS(5843), 1,
      sym_PARAMETER,
    STATE(6264), 1,
      sym_parameter_list_specification,
    ACTIONS(6227), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(5271), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [277758] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6271), 1,
      sym__identifier,
    STATE(5272), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277782] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1342), 1,
      sym__suffix,
    STATE(5273), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6229), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277806] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1785), 1,
      sym__suffix,
    STATE(5274), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6231), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277830] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6233), 1,
      anon_sym_SEMI,
    STATE(5838), 1,
      sym_signature,
    STATE(7417), 1,
      sym_generic_map_aspect,
    STATE(5275), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [277860] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6307), 1,
      sym__identifier,
    STATE(5276), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277884] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1359), 1,
      sym__suffix,
    STATE(5277), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6235), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [277908] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7790), 1,
      sym__identifier,
    STATE(5278), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277932] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6606), 1,
      sym__identifier,
    STATE(5279), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277956] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6189), 1,
      sym__identifier,
    STATE(5280), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [277980] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1481), 1,
      sym__suffix,
    STATE(5281), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6237), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278004] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(5787), 1,
      anon_sym_SEMI,
    STATE(5936), 1,
      sym_condition_clause,
    STATE(8186), 1,
      sym_timeout_clause,
    STATE(5282), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278034] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4315), 1,
      sym__suffix,
    STATE(5283), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6239), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278058] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1285), 1,
      sym__suffix,
    STATE(5284), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6241), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278082] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1507), 1,
      sym__suffix,
    STATE(5285), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6243), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278106] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6444), 1,
      sym__identifier,
    STATE(5286), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [278130] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5287), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6245), 5,
      sym_identifier,
      sym_END,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [278152] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6249), 1,
      sym_BUS,
    STATE(6045), 1,
      sym_initialiser,
    ACTIONS(6247), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5288), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278180] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1210), 1,
      sym__suffix,
    STATE(5289), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6251), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278204] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6253), 1,
      anon_sym_SEMI,
    STATE(5888), 1,
      sym_signature,
    STATE(7711), 1,
      sym_generic_map_aspect,
    STATE(5290), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278234] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6034), 1,
      anon_sym_LBRACK,
    ACTIONS(6255), 1,
      anon_sym_SEMI,
    STATE(5892), 1,
      sym_signature,
    STATE(7743), 1,
      sym_generic_map_aspect,
    STATE(5291), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278264] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(6257), 1,
      anon_sym_SEMI,
    STATE(5865), 1,
      sym_condition_clause,
    STATE(7439), 1,
      sym_timeout_clause,
    STATE(5292), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278294] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(3372), 1,
      sym__suffix,
    STATE(5293), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6259), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278318] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6948), 1,
      sym__identifier,
    STATE(5294), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6130), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [278342] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4722), 1,
      sym__suffix,
    STATE(5295), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6261), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278366] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4776), 1,
      sym__suffix,
    STATE(5296), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6263), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278390] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1562), 1,
      sym__suffix,
    STATE(5297), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6265), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278414] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5298), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6267), 5,
      sym_identifier,
      sym_END,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [278436] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6997), 1,
      sym__identifier,
    STATE(5299), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [278460] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6269), 1,
      anon_sym_SEMI,
    STATE(5710), 1,
      sym_generic_map_aspect,
    STATE(6657), 1,
      sym_initialiser,
    STATE(5300), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278490] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1107), 1,
      sym__suffix,
    STATE(5301), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6271), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278514] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6082), 1,
      sym__identifier,
    STATE(5302), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3838), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [278538] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5773), 1,
      sym_UNTIL,
    ACTIONS(6273), 1,
      anon_sym_SEMI,
    STATE(5899), 1,
      sym_condition_clause,
    STATE(7754), 1,
      sym_timeout_clause,
    STATE(5303), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278568] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1565), 1,
      sym__suffix,
    STATE(5304), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6275), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278592] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(892), 1,
      sym__suffix,
    STATE(5305), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6277), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278616] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6279), 1,
      anon_sym_SEMI,
    STATE(5902), 1,
      sym_generic_map_aspect,
    STATE(7818), 1,
      sym_initialiser,
    STATE(5306), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278646] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    ACTIONS(5803), 1,
      sym_USE,
    STATE(5514), 1,
      aux_sym_configuration_head_repeat2,
    STATE(7021), 1,
      sym_end_for,
    STATE(7694), 1,
      sym_verification_unit_binding_indication,
    STATE(5307), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278676] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1329), 1,
      sym__suffix,
    STATE(5308), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6281), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278700] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6203), 1,
      anon_sym_COMMA,
    STATE(5254), 1,
      aux_sym_group_constituent_list_repeat1,
    ACTIONS(6283), 3,
      sym_FOR,
      sym_UNTIL,
      anon_sym_SEMI,
    STATE(5309), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278726] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1084), 1,
      sym__suffix,
    STATE(5310), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6285), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278750] = 9,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6287), 1,
      anon_sym_SEMI,
    STATE(5676), 1,
      sym_generic_map_aspect,
    STATE(8039), 1,
      sym_initialiser,
    STATE(5311), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [278780] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1810), 1,
      sym__suffix,
    STATE(5312), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6289), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278804] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4886), 1,
      sym__suffix,
    STATE(5313), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6291), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278828] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6571), 1,
      sym__identifier,
    STATE(5314), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [278852] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(2146), 1,
      sym__suffix,
    STATE(5315), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6293), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278876] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5016), 1,
      sym__suffix,
    STATE(5316), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6295), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278900] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(7831), 1,
      sym__identifier,
    STATE(5317), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [278924] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(1848), 1,
      sym__suffix,
    STATE(5318), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6297), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [278948] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5319), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6299), 5,
      sym_identifier,
      sym_END,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [278970] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6487), 1,
      sym__identifier,
    STATE(5320), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [278994] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6547), 1,
      sym__identifier,
    STATE(5321), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [279018] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(8273), 1,
      sym__identifier,
    STATE(5322), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6130), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [279042] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5323), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6301), 5,
      sym_identifier,
      sym_END,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [279064] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6549), 1,
      sym__identifier,
    STATE(5324), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [279088] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(8281), 1,
      sym__identifier,
    STATE(5325), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6130), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [279112] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6552), 1,
      sym__identifier,
    STATE(5326), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [279136] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(8289), 1,
      sym__identifier,
    STATE(5327), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6130), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [279160] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6554), 1,
      sym__identifier,
    STATE(5328), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [279184] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(6557), 1,
      sym__identifier,
    STATE(5329), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5671), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [279208] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(4463), 1,
      sym__suffix,
    STATE(5330), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6303), 4,
      sym_identifier,
      sym_ALL,
      sym_character_literal,
      sym_operator_symbol,
  [279232] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(6305), 1,
      anon_sym_SEMI,
    STATE(5487), 1,
      aux_sym_conditional_or_unaffected_expression_repeat1,
    STATE(5712), 1,
      sym_when_expression,
    STATE(5331), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279259] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6309), 1,
      anon_sym_RBRACK,
    ACTIONS(6311), 1,
      sym_RETURN,
    STATE(5334), 1,
      aux_sym_signature_repeat1,
    STATE(5332), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279286] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6313), 1,
      anon_sym_SEMI,
    ACTIONS(6315), 1,
      sym_IS,
    ACTIONS(6317), 1,
      sym_OPEN,
    STATE(6825), 1,
      sym_file_open_information,
    STATE(5333), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279313] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6319), 1,
      anon_sym_RBRACK,
    ACTIONS(6321), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5334), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279340] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6323), 1,
      anon_sym_RBRACK,
    ACTIONS(6325), 1,
      sym_RETURN,
    STATE(5338), 1,
      aux_sym_signature_repeat1,
    STATE(5335), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279367] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6315), 1,
      sym_IS,
    ACTIONS(6317), 1,
      sym_OPEN,
    ACTIONS(6327), 1,
      anon_sym_SEMI,
    STATE(6970), 1,
      sym_file_open_information,
    STATE(5336), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279394] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6329), 1,
      anon_sym_SEMI,
    ACTIONS(6331), 1,
      sym_ELSE,
    STATE(6538), 1,
      sym_else_waveform,
    STATE(5337), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_conditional_waveforms_repeat1,
  [279419] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6334), 1,
      anon_sym_RBRACK,
    ACTIONS(6336), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5338), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279446] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6338), 1,
      anon_sym_RBRACK,
    ACTIONS(6340), 1,
      sym_RETURN,
    STATE(5343), 1,
      aux_sym_signature_repeat1,
    STATE(5339), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279473] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5340), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6342), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
      anon_sym_PIPE,
  [279494] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6315), 1,
      sym_IS,
    ACTIONS(6317), 1,
      sym_OPEN,
    ACTIONS(6344), 1,
      anon_sym_SEMI,
    STATE(7679), 1,
      sym_file_open_information,
    STATE(5341), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279521] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6346), 1,
      sym_END,
    ACTIONS(6348), 1,
      sym_WHEN,
    STATE(6192), 1,
      sym_case_generate_alternative,
    STATE(5342), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_case_generate_block_repeat1,
  [279546] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6351), 1,
      anon_sym_RBRACK,
    ACTIONS(6353), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5343), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279573] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6355), 1,
      anon_sym_RBRACK,
    ACTIONS(6357), 1,
      sym_RETURN,
    STATE(5345), 1,
      aux_sym_signature_repeat1,
    STATE(5344), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279600] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6359), 1,
      anon_sym_RBRACK,
    ACTIONS(6361), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5345), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279627] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5346), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6363), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [279648] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6365), 1,
      anon_sym_RBRACK,
    ACTIONS(6367), 1,
      sym_RETURN,
    STATE(5348), 1,
      aux_sym_signature_repeat1,
    STATE(5347), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279675] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6369), 1,
      anon_sym_RBRACK,
    ACTIONS(6371), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5348), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279702] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5803), 1,
      sym_USE,
    ACTIONS(6373), 1,
      sym_FOR,
    STATE(5350), 1,
      aux_sym_configuration_head_repeat2,
    STATE(7246), 1,
      sym_verification_unit_binding_indication,
    STATE(5349), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279729] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6167), 1,
      sym_FOR,
    ACTIONS(6169), 1,
      sym_USE,
    STATE(7246), 1,
      sym_verification_unit_binding_indication,
    STATE(5350), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_configuration_head_repeat2,
  [279754] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6375), 1,
      anon_sym_RBRACK,
    ACTIONS(6377), 1,
      sym_RETURN,
    STATE(5352), 1,
      aux_sym_signature_repeat1,
    STATE(5351), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279781] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6379), 1,
      anon_sym_RBRACK,
    ACTIONS(6381), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5352), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279808] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6383), 1,
      anon_sym_RBRACK,
    ACTIONS(6385), 1,
      sym_RETURN,
    STATE(5354), 1,
      aux_sym_signature_repeat1,
    STATE(5353), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279835] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6387), 1,
      anon_sym_RBRACK,
    ACTIONS(6389), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5354), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279862] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5355), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2588), 4,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
  [279883] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6391), 1,
      anon_sym_RBRACK,
    ACTIONS(6393), 1,
      sym_RETURN,
    STATE(5357), 1,
      aux_sym_signature_repeat1,
    STATE(5356), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279910] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6395), 1,
      anon_sym_RBRACK,
    ACTIONS(6397), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5357), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [279937] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5358), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2674), 4,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
  [279958] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5359), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6399), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [279979] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5360), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2440), 4,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
  [280000] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6401), 1,
      anon_sym_RBRACK,
    ACTIONS(6403), 1,
      sym_RETURN,
    STATE(5362), 1,
      aux_sym_signature_repeat1,
    STATE(5361), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280027] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6405), 1,
      anon_sym_RBRACK,
    ACTIONS(6407), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5362), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280054] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5363), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6409), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [280075] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5364), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6411), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [280096] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6413), 1,
      anon_sym_RBRACK,
    ACTIONS(6415), 1,
      sym_RETURN,
    STATE(5366), 1,
      aux_sym_signature_repeat1,
    STATE(5365), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280123] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6417), 1,
      anon_sym_RBRACK,
    ACTIONS(6419), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5366), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280150] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    STATE(6184), 1,
      sym_initialiser,
    ACTIONS(6172), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5367), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280175] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6421), 1,
      anon_sym_RBRACK,
    ACTIONS(6423), 1,
      sym_RETURN,
    STATE(5369), 1,
      aux_sym_signature_repeat1,
    STATE(5368), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280202] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6425), 1,
      anon_sym_RBRACK,
    ACTIONS(6427), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5369), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280229] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5370), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3292), 4,
      sym_CONTEXT,
      sym_END,
      sym_LIBRARY,
      sym_USE,
  [280250] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6429), 1,
      anon_sym_RBRACK,
    ACTIONS(6431), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5371), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280277] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6433), 1,
      anon_sym_RBRACK,
    ACTIONS(6435), 1,
      sym_RETURN,
    STATE(5373), 1,
      aux_sym_signature_repeat1,
    STATE(5372), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280304] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6437), 1,
      anon_sym_RBRACK,
    ACTIONS(6439), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5373), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280331] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6441), 1,
      anon_sym_RBRACK,
    ACTIONS(6443), 1,
      sym_RETURN,
    STATE(5375), 1,
      aux_sym_signature_repeat1,
    STATE(5374), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280358] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6445), 1,
      anon_sym_RBRACK,
    ACTIONS(6447), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5375), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280385] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6449), 1,
      anon_sym_SEMI,
    ACTIONS(6451), 1,
      sym_ELSE,
    STATE(5505), 1,
      aux_sym_conditional_waveforms_repeat1,
    STATE(5762), 1,
      sym_else_waveform,
    STATE(5376), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280412] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6455), 1,
      anon_sym_COMMA,
    STATE(5508), 1,
      aux_sym_waveform_repeat1,
    ACTIONS(6453), 2,
      sym_WHEN,
      anon_sym_SEMI,
    STATE(5377), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280437] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6457), 1,
      anon_sym_RBRACK,
    ACTIONS(6459), 1,
      sym_RETURN,
    STATE(5379), 1,
      aux_sym_signature_repeat1,
    STATE(5378), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280464] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6461), 1,
      anon_sym_RBRACK,
    ACTIONS(6463), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5379), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280491] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6465), 1,
      anon_sym_RBRACK,
    ACTIONS(6467), 1,
      sym_RETURN,
    STATE(5371), 1,
      aux_sym_signature_repeat1,
    STATE(5380), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280518] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6469), 1,
      anon_sym_RBRACK,
    ACTIONS(6471), 1,
      sym_RETURN,
    STATE(5382), 1,
      aux_sym_signature_repeat1,
    STATE(5381), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280545] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6473), 1,
      anon_sym_RBRACK,
    ACTIONS(6475), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5382), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280572] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5383), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6087), 4,
      sym_BUS,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
  [280593] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6477), 1,
      anon_sym_RBRACK,
    ACTIONS(6479), 1,
      sym_RETURN,
    STATE(5385), 1,
      aux_sym_signature_repeat1,
    STATE(5384), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280620] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6481), 1,
      anon_sym_RBRACK,
    ACTIONS(6483), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5385), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280647] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6485), 1,
      sym_WHEN,
    STATE(5509), 1,
      aux_sym_case_generate_block_repeat1,
    STATE(6192), 1,
      sym_case_generate_alternative,
    STATE(6193), 1,
      sym_case_generate_block,
    STATE(5386), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280674] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6487), 1,
      anon_sym_RBRACK,
    ACTIONS(6489), 1,
      sym_RETURN,
    STATE(5388), 1,
      aux_sym_signature_repeat1,
    STATE(5387), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280701] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6491), 1,
      anon_sym_RBRACK,
    ACTIONS(6493), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5388), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280728] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5389), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6495), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [280749] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6497), 1,
      anon_sym_RBRACK,
    ACTIONS(6499), 1,
      sym_RETURN,
    STATE(5391), 1,
      aux_sym_signature_repeat1,
    STATE(5390), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280776] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6501), 1,
      anon_sym_RBRACK,
    ACTIONS(6503), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5391), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280803] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5392), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6193), 4,
      sym_BUS,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
  [280824] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6505), 1,
      anon_sym_RBRACK,
    ACTIONS(6507), 1,
      sym_RETURN,
    STATE(5394), 1,
      aux_sym_signature_repeat1,
    STATE(5393), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280851] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6509), 1,
      anon_sym_RBRACK,
    ACTIONS(6511), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5394), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280878] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6513), 1,
      anon_sym_RBRACK,
    ACTIONS(6515), 1,
      sym_RETURN,
    STATE(5396), 1,
      aux_sym_signature_repeat1,
    STATE(5395), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280905] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6517), 1,
      anon_sym_RBRACK,
    ACTIONS(6519), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5396), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280932] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6455), 1,
      anon_sym_COMMA,
    STATE(5377), 1,
      aux_sym_waveform_repeat1,
    ACTIONS(6521), 2,
      sym_WHEN,
      anon_sym_SEMI,
    STATE(5397), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [280957] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5398), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6523), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [280978] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6525), 1,
      anon_sym_RBRACK,
    ACTIONS(6527), 1,
      sym_RETURN,
    STATE(5400), 1,
      aux_sym_signature_repeat1,
    STATE(5399), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281005] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6529), 1,
      anon_sym_RBRACK,
    ACTIONS(6531), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5400), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281032] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6533), 1,
      anon_sym_RBRACK,
    ACTIONS(6535), 1,
      sym_RETURN,
    STATE(5402), 1,
      aux_sym_signature_repeat1,
    STATE(5401), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281059] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6537), 1,
      anon_sym_RBRACK,
    ACTIONS(6539), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5402), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281086] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6541), 1,
      anon_sym_RBRACK,
    ACTIONS(6543), 1,
      sym_RETURN,
    STATE(5404), 1,
      aux_sym_signature_repeat1,
    STATE(5403), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281113] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6545), 1,
      anon_sym_RBRACK,
    ACTIONS(6547), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5404), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281140] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5405), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2678), 4,
      sym_END,
      sym_GENERIC,
      sym_IS,
      sym_PORT,
  [281161] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6549), 1,
      anon_sym_RBRACK,
    ACTIONS(6551), 1,
      sym_RETURN,
    STATE(5407), 1,
      aux_sym_signature_repeat1,
    STATE(5406), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281188] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6553), 1,
      anon_sym_RBRACK,
    ACTIONS(6555), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5407), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281215] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6557), 1,
      anon_sym_RBRACK,
    ACTIONS(6559), 1,
      sym_RETURN,
    STATE(5409), 1,
      aux_sym_signature_repeat1,
    STATE(5408), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281242] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6561), 1,
      anon_sym_RBRACK,
    ACTIONS(6563), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5409), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281269] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6565), 1,
      anon_sym_RBRACK,
    ACTIONS(6567), 1,
      sym_RETURN,
    STATE(5411), 1,
      aux_sym_signature_repeat1,
    STATE(5410), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281296] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6569), 1,
      anon_sym_RBRACK,
    ACTIONS(6571), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5411), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281323] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6573), 1,
      anon_sym_RBRACK,
    ACTIONS(6575), 1,
      sym_RETURN,
    STATE(5413), 1,
      aux_sym_signature_repeat1,
    STATE(5412), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281350] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6577), 1,
      anon_sym_RBRACK,
    ACTIONS(6579), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5413), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281377] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6581), 1,
      anon_sym_RBRACK,
    ACTIONS(6583), 1,
      sym_RETURN,
    STATE(5415), 1,
      aux_sym_signature_repeat1,
    STATE(5414), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281404] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6585), 1,
      anon_sym_RBRACK,
    ACTIONS(6587), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5415), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281431] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5416), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6589), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [281452] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5417), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6591), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [281473] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6593), 1,
      anon_sym_RBRACK,
    ACTIONS(6595), 1,
      sym_RETURN,
    STATE(5419), 1,
      aux_sym_signature_repeat1,
    STATE(5418), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281500] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6597), 1,
      anon_sym_RBRACK,
    ACTIONS(6599), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5419), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281527] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5420), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6601), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [281548] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6603), 1,
      anon_sym_RBRACK,
    ACTIONS(6605), 1,
      sym_RETURN,
    STATE(5422), 1,
      aux_sym_signature_repeat1,
    STATE(5421), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281575] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6607), 1,
      anon_sym_RBRACK,
    ACTIONS(6609), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5422), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281602] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    STATE(6542), 1,
      sym_initialiser,
    ACTIONS(6611), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5423), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281627] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6613), 1,
      anon_sym_RBRACK,
    ACTIONS(6615), 1,
      sym_RETURN,
    STATE(5425), 1,
      aux_sym_signature_repeat1,
    STATE(5424), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281654] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6617), 1,
      anon_sym_RBRACK,
    ACTIONS(6619), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5425), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281681] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6623), 1,
      anon_sym_LPAREN,
    ACTIONS(6621), 3,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
    STATE(5426), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281704] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6625), 1,
      anon_sym_RBRACK,
    ACTIONS(6627), 1,
      sym_RETURN,
    STATE(5428), 1,
      aux_sym_signature_repeat1,
    STATE(5427), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281731] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6629), 1,
      anon_sym_RBRACK,
    ACTIONS(6631), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5428), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281758] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6633), 1,
      anon_sym_RBRACK,
    ACTIONS(6635), 1,
      sym_RETURN,
    STATE(5430), 1,
      aux_sym_signature_repeat1,
    STATE(5429), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281785] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6637), 1,
      anon_sym_RBRACK,
    ACTIONS(6639), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5430), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281812] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6641), 1,
      anon_sym_RBRACK,
    ACTIONS(6643), 1,
      sym_RETURN,
    STATE(5432), 1,
      aux_sym_signature_repeat1,
    STATE(5431), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281839] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6645), 1,
      anon_sym_RBRACK,
    ACTIONS(6647), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5432), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281866] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6649), 1,
      anon_sym_RBRACK,
    ACTIONS(6651), 1,
      sym_RETURN,
    STATE(5434), 1,
      aux_sym_signature_repeat1,
    STATE(5433), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281893] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6653), 1,
      anon_sym_RBRACK,
    ACTIONS(6655), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5434), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281920] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6657), 1,
      anon_sym_RBRACK,
    ACTIONS(6659), 1,
      sym_RETURN,
    STATE(5436), 1,
      aux_sym_signature_repeat1,
    STATE(5435), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281947] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6661), 1,
      anon_sym_RBRACK,
    ACTIONS(6663), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5436), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [281974] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6665), 1,
      anon_sym_RBRACK,
    ACTIONS(6667), 1,
      sym_RETURN,
    STATE(5438), 1,
      aux_sym_signature_repeat1,
    STATE(5437), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282001] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6669), 1,
      anon_sym_RBRACK,
    ACTIONS(6671), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5438), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282028] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6673), 1,
      anon_sym_RBRACK,
    ACTIONS(6675), 1,
      sym_RETURN,
    STATE(5440), 1,
      aux_sym_signature_repeat1,
    STATE(5439), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282055] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6677), 1,
      anon_sym_RBRACK,
    ACTIONS(6679), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5440), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282082] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6683), 1,
      anon_sym_DOT,
    STATE(5524), 1,
      aux_sym_selected_name_repeat1,
    ACTIONS(6681), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(5441), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282107] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6683), 1,
      anon_sym_DOT,
    STATE(5524), 1,
      aux_sym_selected_name_repeat1,
    ACTIONS(6685), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(5442), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282132] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6687), 1,
      anon_sym_RBRACK,
    ACTIONS(6689), 1,
      sym_RETURN,
    STATE(5444), 1,
      aux_sym_signature_repeat1,
    STATE(5443), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282159] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6691), 1,
      anon_sym_RBRACK,
    ACTIONS(6693), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5444), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282186] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6695), 1,
      anon_sym_RBRACK,
    ACTIONS(6697), 1,
      sym_RETURN,
    STATE(5446), 1,
      aux_sym_signature_repeat1,
    STATE(5445), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282213] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6699), 1,
      anon_sym_RBRACK,
    ACTIONS(6701), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5446), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282240] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5447), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6036), 4,
      sym_BUS,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
  [282261] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6703), 1,
      anon_sym_RBRACK,
    ACTIONS(6705), 1,
      sym_RETURN,
    STATE(5449), 1,
      aux_sym_signature_repeat1,
    STATE(5448), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282288] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6707), 1,
      anon_sym_RBRACK,
    ACTIONS(6709), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5449), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282315] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6711), 1,
      anon_sym_RBRACK,
    ACTIONS(6713), 1,
      sym_RETURN,
    STATE(5451), 1,
      aux_sym_signature_repeat1,
    STATE(5450), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282342] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6715), 1,
      anon_sym_RBRACK,
    ACTIONS(6717), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5451), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282369] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6719), 1,
      anon_sym_RBRACK,
    ACTIONS(6721), 1,
      sym_RETURN,
    STATE(5453), 1,
      aux_sym_signature_repeat1,
    STATE(5452), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282396] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6723), 1,
      anon_sym_RBRACK,
    ACTIONS(6725), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5453), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282423] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6727), 1,
      anon_sym_RBRACK,
    ACTIONS(6729), 1,
      sym_RETURN,
    STATE(5555), 1,
      aux_sym_signature_repeat1,
    STATE(5454), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282450] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5455), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5931), 4,
      sym_WHEN,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [282471] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6731), 1,
      anon_sym_RBRACK,
    ACTIONS(6733), 1,
      sym_RETURN,
    STATE(5457), 1,
      aux_sym_signature_repeat1,
    STATE(5456), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282498] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6735), 1,
      anon_sym_RBRACK,
    ACTIONS(6737), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5457), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282525] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6739), 1,
      anon_sym_RBRACK,
    ACTIONS(6741), 1,
      sym_RETURN,
    STATE(5459), 1,
      aux_sym_signature_repeat1,
    STATE(5458), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282552] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6743), 1,
      anon_sym_RBRACK,
    ACTIONS(6745), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5459), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282579] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6747), 1,
      anon_sym_RBRACK,
    ACTIONS(6749), 1,
      sym_RETURN,
    STATE(5461), 1,
      aux_sym_signature_repeat1,
    STATE(5460), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282606] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6751), 1,
      anon_sym_RBRACK,
    ACTIONS(6753), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5461), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282633] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6755), 1,
      anon_sym_RBRACK,
    ACTIONS(6757), 1,
      sym_RETURN,
    STATE(5463), 1,
      aux_sym_signature_repeat1,
    STATE(5462), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282660] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6759), 1,
      anon_sym_RBRACK,
    ACTIONS(6761), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5463), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282687] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6763), 1,
      anon_sym_RBRACK,
    ACTIONS(6765), 1,
      sym_RETURN,
    STATE(5465), 1,
      aux_sym_signature_repeat1,
    STATE(5464), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282714] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6767), 1,
      anon_sym_RBRACK,
    ACTIONS(6769), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5465), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282741] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6771), 1,
      anon_sym_RBRACK,
    ACTIONS(6773), 1,
      sym_RETURN,
    STATE(5467), 1,
      aux_sym_signature_repeat1,
    STATE(5466), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282768] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6775), 1,
      anon_sym_RBRACK,
    ACTIONS(6777), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5467), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282795] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5468), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6779), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [282816] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5469), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6781), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [282837] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5470), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6783), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [282858] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5471), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6785), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [282879] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5472), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6787), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [282900] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6112), 1,
      sym_CONFIGURATION,
    ACTIONS(6114), 1,
      sym_ENTITY,
    ACTIONS(6116), 1,
      sym_OPEN,
    STATE(5216), 1,
      sym_entity_aspect,
    STATE(5473), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282927] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5474), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6789), 4,
      sym_identifier,
      sym_library_constant,
      sym_library_function,
      sym_library_type,
  [282948] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6793), 1,
      anon_sym_LPAREN,
    ACTIONS(6791), 3,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
    STATE(5475), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [282971] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5476), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6795), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_GT_GT,
  [282992] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5477), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6087), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [283013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5478), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3482), 4,
      sym_CONTEXT,
      sym_END,
      sym_LIBRARY,
      sym_USE,
  [283034] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5479), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6193), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [283055] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5480), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3244), 4,
      sym_CONTEXT,
      sym_END,
      sym_LIBRARY,
      sym_USE,
  [283076] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(6797), 1,
      sym_RETURN,
    STATE(7336), 1,
      sym_parameter_list_specification,
    STATE(5481), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283103] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5482), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6079), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
      anon_sym_PIPE,
  [283124] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6801), 1,
      anon_sym_LBRACK,
    STATE(6219), 1,
      sym_signature,
    ACTIONS(6799), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(5483), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283149] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5484), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3432), 4,
      sym_CONTEXT,
      sym_END,
      sym_LIBRARY,
      sym_USE,
  [283170] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5485), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3462), 4,
      sym_CONTEXT,
      sym_END,
      sym_LIBRARY,
      sym_USE,
  [283191] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5821), 1,
      anon_sym_LPAREN,
    ACTIONS(5827), 1,
      sym_PARAMETER,
    ACTIONS(6227), 1,
      anon_sym_SEMI,
    STATE(6982), 1,
      sym_parameter_list_specification,
    STATE(5486), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283218] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(6803), 1,
      anon_sym_SEMI,
    STATE(5537), 1,
      aux_sym_conditional_or_unaffected_expression_repeat1,
    STATE(5840), 1,
      sym_when_expression,
    STATE(5487), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283245] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6805), 1,
      sym_BEGIN,
    ACTIONS(6807), 1,
      sym_END,
    STATE(6134), 1,
      sym_entity_body,
    STATE(7022), 1,
      sym_end_entity,
    STATE(5488), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283272] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5489), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6036), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [283293] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6683), 1,
      anon_sym_DOT,
    STATE(5441), 1,
      aux_sym_selected_name_repeat1,
    ACTIONS(6809), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(5490), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283318] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6683), 1,
      anon_sym_DOT,
    STATE(5442), 1,
      aux_sym_selected_name_repeat1,
    ACTIONS(6811), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(5491), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283343] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6813), 1,
      anon_sym_SEMI,
    STATE(7504), 1,
      sym_force_mode,
    ACTIONS(6815), 2,
      sym_IN,
      sym_OUT,
    STATE(5492), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283368] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5493), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6066), 4,
      sym_FOR,
      sym_UNTIL,
      anon_sym_SEMI,
      anon_sym_COMMA,
  [283389] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5494), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6066), 4,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_COLON,
  [283410] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5495), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3304), 4,
      sym_ATTRIBUTE,
      sym_FOR,
      sym_GROUP,
      sym_USE,
  [283431] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6112), 1,
      sym_CONFIGURATION,
    ACTIONS(6114), 1,
      sym_ENTITY,
    ACTIONS(6116), 1,
      sym_OPEN,
    STATE(5153), 1,
      sym_entity_aspect,
    STATE(5496), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283458] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5497), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2830), 4,
      sym_BUS,
      sym_REGISTER,
      anon_sym_SEMI,
      sym_variable_assignment,
  [283479] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6315), 1,
      sym_IS,
    ACTIONS(6317), 1,
      sym_OPEN,
    ACTIONS(6817), 1,
      anon_sym_SEMI,
    STATE(6736), 1,
      sym_file_open_information,
    STATE(5498), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283506] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5499), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3248), 4,
      sym_ATTRIBUTE,
      sym_FOR,
      sym_GROUP,
      sym_USE,
  [283527] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5500), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5940), 4,
      sym_GENERIC,
      sym_PORT,
      sym_USE,
      anon_sym_SEMI,
  [283548] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5501), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6819), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [283569] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(6821), 1,
      sym_RETURN,
    STATE(7174), 1,
      sym_parameter_list_specification,
    STATE(5502), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283596] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6823), 1,
      anon_sym_COMMA,
    ACTIONS(6826), 2,
      sym_RETURN,
      anon_sym_RBRACK,
    STATE(5503), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_signature_repeat1,
  [283619] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2825), 1,
      sym_END,
    STATE(5778), 1,
      sym_generate_block_end,
    ACTIONS(2823), 2,
      sym_ELSE,
      sym_ELSIF,
    STATE(5504), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283644] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6451), 1,
      sym_ELSE,
    ACTIONS(6828), 1,
      anon_sym_SEMI,
    STATE(5337), 1,
      aux_sym_conditional_waveforms_repeat1,
    STATE(5561), 1,
      sym_else_waveform,
    STATE(5505), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283671] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5506), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(5992), 4,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
  [283692] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5507), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(3292), 4,
      sym_ATTRIBUTE,
      sym_FOR,
      sym_GROUP,
      sym_USE,
  [283713] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6832), 1,
      anon_sym_COMMA,
    ACTIONS(6830), 2,
      sym_WHEN,
      anon_sym_SEMI,
    STATE(5508), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_waveform_repeat1,
  [283736] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6485), 1,
      sym_WHEN,
    ACTIONS(6835), 1,
      sym_END,
    STATE(5342), 1,
      aux_sym_case_generate_block_repeat1,
    STATE(6192), 1,
      sym_case_generate_alternative,
    STATE(5509), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283763] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6839), 1,
      anon_sym_LPAREN,
    ACTIONS(6837), 3,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
    STATE(5510), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283786] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(6841), 1,
      sym_RETURN,
    STATE(6868), 1,
      sym_parameter_list_specification,
    STATE(5511), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283813] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6845), 1,
      sym_END,
    STATE(5803), 1,
      sym_generate_block_end,
    ACTIONS(6843), 2,
      sym_ELSE,
      sym_ELSIF,
    STATE(5512), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [283838] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5513), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(2830), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [283859] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6167), 1,
      sym_END,
    ACTIONS(6169), 1,
      sym_USE,
    STATE(7694), 1,
      sym_verification_unit_binding_indication,
    STATE(5514), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_configuration_head_repeat2,
  [283884] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5515), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6848), 4,
      sym_IS,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
  [283905] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5516), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6399), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [283926] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5517), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6409), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [283947] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5518), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6411), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [283968] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5519), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6850), 4,
      sym_ATTRIBUTE,
      sym_FOR,
      sym_GROUP,
      sym_USE,
  [283989] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5520), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6852), 4,
      sym_ATTRIBUTE,
      sym_FOR,
      sym_GROUP,
      sym_USE,
  [284010] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5569), 1,
      sym_FOR,
    ACTIONS(5803), 1,
      sym_USE,
    STATE(5350), 1,
      aux_sym_configuration_head_repeat2,
    STATE(7246), 1,
      sym_verification_unit_binding_indication,
    STATE(5521), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284037] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5522), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6495), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284058] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5523), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6523), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284079] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6856), 1,
      anon_sym_DOT,
    ACTIONS(6854), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(5524), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_selected_name_repeat1,
  [284102] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5525), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6589), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284123] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5526), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6591), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284144] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5527), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6601), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284165] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5528), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6779), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284186] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5529), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6781), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284207] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5530), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6783), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284228] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5531), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6785), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284249] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5532), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6787), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284270] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5533), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6795), 4,
      sym_BUS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
  [284291] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(6859), 1,
      sym_RETURN,
    STATE(7172), 1,
      sym_parameter_list_specification,
    STATE(5534), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284318] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5535), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
    ACTIONS(6178), 4,
      sym_GENERIC,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
  [284339] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3272), 1,
      sym_WHEN,
    STATE(5229), 1,
      aux_sym_case_body_repeat1,
    STATE(6086), 1,
      sym_case_statement_alternative,
    STATE(6940), 1,
      sym_when_element,
    STATE(5536), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284366] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6861), 1,
      anon_sym_SEMI,
    ACTIONS(6863), 1,
      sym_WHEN,
    STATE(6197), 1,
      sym_when_expression,
    STATE(5537), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_conditional_or_unaffected_expression_repeat1,
  [284391] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6868), 1,
      anon_sym_PIPE,
    ACTIONS(6866), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_EQ_GT,
    STATE(5538), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284414] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(6870), 1,
      sym_RETURN,
    STATE(7547), 1,
      sym_parameter_list_specification,
    STATE(5539), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284441] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6872), 1,
      anon_sym_RBRACK,
    ACTIONS(6874), 1,
      sym_RETURN,
    STATE(5543), 1,
      aux_sym_signature_repeat1,
    STATE(5540), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284468] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6315), 1,
      sym_IS,
    ACTIONS(6317), 1,
      sym_OPEN,
    ACTIONS(6876), 1,
      anon_sym_SEMI,
    STATE(7612), 1,
      sym_file_open_information,
    STATE(5541), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284495] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5805), 1,
      anon_sym_LPAREN,
    ACTIONS(5811), 1,
      sym_PARAMETER,
    ACTIONS(6878), 1,
      sym_RETURN,
    STATE(7619), 1,
      sym_parameter_list_specification,
    STATE(5542), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284522] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6880), 1,
      anon_sym_RBRACK,
    ACTIONS(6882), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5543), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284549] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6884), 1,
      anon_sym_SEMI,
    STATE(7771), 1,
      sym_force_mode,
    ACTIONS(6815), 2,
      sym_IN,
      sym_OUT,
    STATE(5544), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284574] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6886), 1,
      anon_sym_SEMI,
    STATE(6722), 1,
      sym_force_mode,
    ACTIONS(6815), 2,
      sym_IN,
      sym_OUT,
    STATE(5545), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284599] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6485), 1,
      sym_WHEN,
    STATE(5509), 1,
      aux_sym_case_generate_block_repeat1,
    STATE(6192), 1,
      sym_case_generate_alternative,
    STATE(6595), 1,
      sym_case_generate_block,
    STATE(5546), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284626] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6485), 1,
      sym_WHEN,
    STATE(5509), 1,
      aux_sym_case_generate_block_repeat1,
    STATE(6071), 1,
      sym_case_generate_block,
    STATE(6192), 1,
      sym_case_generate_alternative,
    STATE(5547), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284653] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6888), 1,
      anon_sym_SEMI,
    STATE(8027), 1,
      sym_force_mode,
    ACTIONS(6815), 2,
      sym_IN,
      sym_OUT,
    STATE(5548), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284678] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6485), 1,
      sym_WHEN,
    STATE(5509), 1,
      aux_sym_case_generate_block_repeat1,
    STATE(6125), 1,
      sym_case_generate_block,
    STATE(6192), 1,
      sym_case_generate_alternative,
    STATE(5549), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284705] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6890), 1,
      anon_sym_RBRACK,
    ACTIONS(6892), 1,
      sym_RETURN,
    STATE(5552), 1,
      aux_sym_signature_repeat1,
    STATE(5550), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284732] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6315), 1,
      sym_IS,
    ACTIONS(6317), 1,
      sym_OPEN,
    ACTIONS(6894), 1,
      anon_sym_SEMI,
    STATE(8169), 1,
      sym_file_open_information,
    STATE(5551), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284759] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6896), 1,
      anon_sym_RBRACK,
    ACTIONS(6898), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5552), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284786] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6900), 1,
      anon_sym_SEMI,
    STATE(8232), 1,
      sym_force_mode,
    ACTIONS(6815), 2,
      sym_IN,
      sym_OUT,
    STATE(5553), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284811] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6902), 1,
      anon_sym_SEMI,
    STATE(8261), 1,
      sym_force_mode,
    ACTIONS(6815), 2,
      sym_IN,
      sym_OUT,
    STATE(5554), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284836] = 8,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6307), 1,
      anon_sym_COMMA,
    ACTIONS(6904), 1,
      anon_sym_RBRACK,
    ACTIONS(6906), 1,
      sym_RETURN,
    STATE(5503), 1,
      aux_sym_signature_repeat1,
    STATE(5555), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284863] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2818), 3,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
    STATE(5556), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284883] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6910), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5557), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284907] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6912), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5558), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284927] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6914), 1,
      anon_sym_RPAREN,
    STATE(5560), 1,
      aux_sym_association_list_repeat1,
    STATE(5559), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284951] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6916), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5560), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284975] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(6918), 1,
      anon_sym_SEMI,
    STATE(6156), 1,
      sym_when_expression,
    STATE(5561), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [284999] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6920), 1,
      anon_sym_SEMI,
    STATE(6966), 1,
      sym_initialiser,
    STATE(5562), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285023] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6922), 1,
      anon_sym_SEMI,
    STATE(6818), 1,
      sym_initialiser,
    STATE(5563), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285047] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6924), 1,
      anon_sym_RPAREN,
    STATE(5565), 1,
      aux_sym_association_list_repeat1,
    STATE(5564), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285071] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6926), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5565), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285095] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6928), 1,
      anon_sym_RPAREN,
    STATE(5568), 1,
      aux_sym_association_list_repeat1,
    STATE(5566), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285119] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2678), 3,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
    STATE(5567), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285139] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6930), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5568), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285163] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6178), 3,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
    STATE(5569), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285183] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6932), 1,
      anon_sym_RPAREN,
    STATE(5571), 1,
      aux_sym_association_list_repeat1,
    STATE(5570), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285207] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6934), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5571), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285231] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6167), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(5572), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285251] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6936), 1,
      anon_sym_COMMA,
    ACTIONS(6938), 1,
      anon_sym_COLON,
    STATE(5789), 1,
      aux_sym_entity_name_list_repeat1,
    STATE(5573), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285275] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(6940), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5574), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285299] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3230), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(5575), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285319] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6942), 1,
      anon_sym_RPAREN,
    STATE(5577), 1,
      aux_sym_association_list_repeat1,
    STATE(5576), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285343] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6944), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5577), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285367] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(6946), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5578), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285391] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6848), 3,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
    STATE(5579), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285411] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6948), 1,
      anon_sym_RPAREN,
    STATE(5581), 1,
      aux_sym_association_list_repeat1,
    STATE(5580), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285435] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6950), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5581), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285459] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6952), 1,
      anon_sym_COMMA,
    ACTIONS(6954), 1,
      anon_sym_COLON,
    STATE(5790), 1,
      aux_sym_generic_identifier_list_repeat1,
    STATE(5582), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285483] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6956), 1,
      anon_sym_RPAREN,
    STATE(5584), 1,
      aux_sym_association_list_repeat1,
    STATE(5583), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285507] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6958), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5584), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285531] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6960), 1,
      anon_sym_RPAREN,
    STATE(5586), 1,
      aux_sym_association_list_repeat1,
    STATE(5585), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285555] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6962), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5586), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285579] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6964), 1,
      anon_sym_SEMI,
    ACTIONS(6966), 1,
      anon_sym_RPAREN,
    STATE(5791), 1,
      aux_sym_generic_interface_list_repeat1,
    STATE(5587), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285603] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6968), 1,
      anon_sym_SEMI,
    STATE(7160), 1,
      sym_generic_map_aspect,
    STATE(5588), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285627] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6124), 1,
      anon_sym_SEMI,
    STATE(7208), 1,
      sym_initialiser,
    STATE(5589), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285651] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6970), 1,
      anon_sym_RPAREN,
    STATE(5591), 1,
      aux_sym_association_list_repeat1,
    STATE(5590), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285675] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6972), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5591), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285699] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6976), 1,
      sym_OF,
    ACTIONS(6974), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5592), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285721] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6978), 1,
      anon_sym_RPAREN,
    STATE(5613), 1,
      aux_sym_association_list_repeat1,
    STATE(5593), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285745] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6980), 1,
      anon_sym_RPAREN,
    STATE(5595), 1,
      aux_sym_association_list_repeat1,
    STATE(5594), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285769] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6982), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5595), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285793] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6984), 1,
      anon_sym_SEMI,
    STATE(7258), 1,
      sym_generic_map_aspect,
    STATE(5596), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285817] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6986), 1,
      anon_sym_RPAREN,
    STATE(5601), 1,
      aux_sym_association_list_repeat1,
    STATE(5597), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285841] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6988), 1,
      anon_sym_RPAREN,
    STATE(5599), 1,
      aux_sym_association_list_repeat1,
    STATE(5598), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285865] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6990), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5599), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285889] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6992), 1,
      anon_sym_COMMA,
    ACTIONS(6995), 1,
      anon_sym_RPAREN,
    STATE(5600), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_array_index_incomplete_type_list_repeat1,
  [285911] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(6997), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5601), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285935] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(6999), 1,
      anon_sym_SEMI,
    STATE(7266), 1,
      sym_generic_map_aspect,
    STATE(5602), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285959] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7001), 1,
      anon_sym_RPAREN,
    STATE(5604), 1,
      aux_sym_association_list_repeat1,
    STATE(5603), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [285983] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7003), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5604), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286007] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7005), 1,
      anon_sym_SEMI,
    ACTIONS(7007), 1,
      anon_sym_COMMA,
    STATE(5769), 1,
      aux_sym_logical_name_list_repeat1,
    STATE(5605), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286031] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7009), 1,
      anon_sym_COMMA,
    ACTIONS(7011), 1,
      anon_sym_RPAREN,
    STATE(5755), 1,
      aux_sym_association_or_range_list_repeat1,
    STATE(5606), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286055] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7013), 1,
      anon_sym_COMMA,
    ACTIONS(7016), 1,
      anon_sym_RPAREN,
    STATE(5607), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_entity_class_entry_list_repeat1,
  [286077] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7018), 1,
      anon_sym_COMMA,
    ACTIONS(7020), 1,
      anon_sym_RPAREN,
    STATE(5217), 1,
      aux_sym_group_constituent_list_repeat1,
    STATE(5608), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286101] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7022), 1,
      anon_sym_RPAREN,
    STATE(5610), 1,
      aux_sym_association_list_repeat1,
    STATE(5609), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286125] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7024), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5610), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286149] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6087), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(5611), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286169] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7026), 1,
      anon_sym_SEMI,
    STATE(7279), 1,
      sym_initialiser,
    STATE(5612), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286193] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7028), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5613), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286217] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7030), 1,
      anon_sym_RPAREN,
    STATE(5615), 1,
      aux_sym_association_list_repeat1,
    STATE(5614), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286241] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7032), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5615), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286265] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7034), 1,
      anon_sym_RPAREN,
    STATE(5617), 1,
      aux_sym_association_list_repeat1,
    STATE(5616), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286289] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7036), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5617), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286313] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7038), 1,
      sym_directive_body,
    ACTIONS(7040), 1,
      sym__directive_newline,
    STATE(5764), 1,
      aux_sym_user_directive_repeat1,
    STATE(5618), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286337] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7042), 1,
      anon_sym_RPAREN,
    STATE(5620), 1,
      aux_sym_association_list_repeat1,
    STATE(5619), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286361] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7044), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5620), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286385] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7046), 1,
      anon_sym_COMMA,
    ACTIONS(7049), 1,
      anon_sym_RPAREN,
    STATE(5621), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_aggregate_repeat1,
  [286407] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6868), 1,
      anon_sym_PIPE,
    ACTIONS(7051), 1,
      anon_sym_EQ_GT,
    STATE(6575), 1,
      sym_case_generate_body,
    STATE(5622), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286431] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7053), 1,
      anon_sym_RPAREN,
    STATE(5624), 1,
      aux_sym_association_list_repeat1,
    STATE(5623), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286455] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7055), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5624), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286479] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6193), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(5625), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286499] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(7057), 1,
      anon_sym_SEMI,
    STATE(6770), 1,
      sym_port_map_aspect,
    STATE(5626), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286523] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7059), 1,
      anon_sym_SEMI,
    STATE(7443), 1,
      sym_generic_map_aspect,
    STATE(5627), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286547] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7061), 1,
      anon_sym_RPAREN,
    STATE(5629), 1,
      aux_sym_association_list_repeat1,
    STATE(5628), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286571] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7063), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5629), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286595] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(7065), 1,
      anon_sym_SEMI,
    STATE(6906), 1,
      sym_port_map_aspect,
    STATE(5630), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286619] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7067), 1,
      anon_sym_RPAREN,
    STATE(5632), 1,
      aux_sym_association_list_repeat1,
    STATE(5631), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286643] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7069), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5632), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286667] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7071), 1,
      anon_sym_RPAREN,
    STATE(5634), 1,
      aux_sym_association_list_repeat1,
    STATE(5633), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286691] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7073), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5634), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286715] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7075), 3,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
    STATE(5635), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286735] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7009), 1,
      anon_sym_COMMA,
    ACTIONS(7077), 1,
      anon_sym_RPAREN,
    STATE(5606), 1,
      aux_sym_association_or_range_list_repeat1,
    STATE(5636), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286759] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7079), 1,
      anon_sym_SEMI,
    STATE(7665), 1,
      sym_initialiser,
    STATE(5637), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286783] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7081), 1,
      anon_sym_RPAREN,
    STATE(5639), 1,
      aux_sym_association_list_repeat1,
    STATE(5638), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286807] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7083), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5639), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286831] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7085), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5640), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286855] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7087), 1,
      anon_sym_RPAREN,
    STATE(5644), 1,
      aux_sym_association_list_repeat1,
    STATE(5641), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286879] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7089), 1,
      anon_sym_RPAREN,
    STATE(5643), 1,
      aux_sym_association_list_repeat1,
    STATE(5642), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286903] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7091), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5643), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286927] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7093), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5644), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286951] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7095), 1,
      anon_sym_DOT,
    ACTIONS(7097), 1,
      anon_sym_LPAREN,
    ACTIONS(7099), 1,
      anon_sym_COLON,
    STATE(5645), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286975] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7101), 1,
      anon_sym_RPAREN,
    STATE(5647), 1,
      aux_sym_association_list_repeat1,
    STATE(5646), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [286999] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7103), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5647), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287023] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7105), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5648), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287043] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7107), 1,
      anon_sym_COMMA,
    ACTIONS(7109), 1,
      anon_sym_RPAREN,
    STATE(5698), 1,
      aux_sym_element_association_list_repeat1,
    STATE(5649), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287067] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7111), 1,
      anon_sym_SEMI,
    STATE(7626), 1,
      sym_initialiser,
    STATE(5650), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287091] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7113), 1,
      anon_sym_RPAREN,
    STATE(5652), 1,
      aux_sym_association_list_repeat1,
    STATE(5651), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287115] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7115), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5652), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287139] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7117), 1,
      anon_sym_SEMI,
    STATE(7833), 1,
      sym_generic_map_aspect,
    STATE(5653), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287163] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7119), 3,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
    STATE(5654), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287183] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7121), 1,
      anon_sym_RPAREN,
    STATE(5656), 1,
      aux_sym_association_list_repeat1,
    STATE(5655), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287207] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7123), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5656), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287231] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7125), 1,
      anon_sym_RPAREN,
    STATE(5658), 1,
      aux_sym_association_list_repeat1,
    STATE(5657), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287255] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7127), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5658), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287279] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7129), 1,
      anon_sym_RPAREN,
    STATE(5660), 1,
      aux_sym_association_list_repeat1,
    STATE(5659), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287303] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7131), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5660), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287327] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7133), 1,
      anon_sym_RPAREN,
    STATE(5662), 1,
      aux_sym_association_list_repeat1,
    STATE(5661), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287351] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7135), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5662), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287375] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7137), 1,
      anon_sym_RPAREN,
    STATE(5664), 1,
      aux_sym_association_list_repeat1,
    STATE(5663), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287399] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7139), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5664), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287423] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7141), 1,
      anon_sym_RPAREN,
    STATE(5666), 1,
      aux_sym_association_list_repeat1,
    STATE(5665), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287447] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7143), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5666), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287471] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6287), 1,
      anon_sym_SEMI,
    STATE(8039), 1,
      sym_initialiser,
    STATE(5667), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287495] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7145), 1,
      anon_sym_RPAREN,
    STATE(5678), 1,
      aux_sym_association_list_repeat1,
    STATE(5668), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287519] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7147), 1,
      anon_sym_SEMI,
    STATE(8083), 1,
      sym_generic_map_aspect,
    STATE(5669), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287543] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7149), 1,
      anon_sym_RPAREN,
    STATE(5673), 1,
      aux_sym_association_list_repeat1,
    STATE(5670), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287567] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7153), 1,
      sym_OF,
    ACTIONS(7151), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5671), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287589] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7155), 1,
      sym_CONSTANT,
    ACTIONS(7157), 1,
      sym_SIGNAL,
    ACTIONS(7159), 1,
      sym_VARIABLE,
    STATE(5672), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287613] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7161), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5673), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287637] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7163), 1,
      anon_sym_SEMI,
    STATE(8099), 1,
      sym_generic_map_aspect,
    STATE(5674), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287661] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2830), 3,
      sym_IS,
      sym_OPEN,
      anon_sym_SEMI,
    STATE(5675), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287681] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7165), 1,
      anon_sym_SEMI,
    STATE(8135), 1,
      sym_initialiser,
    STATE(5676), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287705] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7167), 3,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
    STATE(5677), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287725] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7169), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5678), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287749] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2929), 1,
      sym_END,
    ACTIONS(2932), 1,
      sym_WHEN,
    STATE(6586), 1,
      sym_generate_block_end,
    STATE(5679), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287773] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(7171), 1,
      anon_sym_SEMI,
    STATE(7299), 1,
      sym_port_map_aspect,
    STATE(5680), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287797] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7173), 1,
      anon_sym_RPAREN,
    STATE(5684), 1,
      aux_sym_association_list_repeat1,
    STATE(5681), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287821] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7175), 1,
      sym_END,
    ACTIONS(7178), 1,
      sym_WHEN,
    STATE(6598), 1,
      sym_generate_block_end,
    STATE(5682), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287845] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7180), 1,
      sym_directive_body,
    ACTIONS(7183), 1,
      sym__directive_newline,
    STATE(5683), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_user_directive_repeat1,
  [287867] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7185), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5684), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287891] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6036), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(5685), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287911] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7187), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5686), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287935] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7189), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5687), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287959] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6126), 1,
      anon_sym_SEMI,
    STATE(7642), 1,
      sym_initialiser,
    STATE(5688), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [287983] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7191), 1,
      anon_sym_RPAREN,
    STATE(5818), 1,
      aux_sym_association_list_repeat1,
    STATE(5689), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288007] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7195), 1,
      sym_IS,
    ACTIONS(7193), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5690), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288029] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3890), 1,
      anon_sym_RPAREN,
    ACTIONS(7197), 1,
      anon_sym_SEMI,
    STATE(5826), 1,
      aux_sym_interface_list_repeat1,
    STATE(5691), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288053] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5712), 1,
      sym_PORT,
    ACTIONS(6071), 1,
      sym_END,
    STATE(6921), 1,
      sym_port_clause,
    STATE(5692), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288077] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7199), 1,
      anon_sym_RPAREN,
    STATE(5706), 1,
      aux_sym_association_list_repeat1,
    STATE(5693), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288101] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7201), 3,
      sym_box,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(5694), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288121] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7203), 1,
      anon_sym_COMMA,
    ACTIONS(7205), 1,
      anon_sym_RPAREN,
    STATE(5830), 1,
      aux_sym_entity_class_entry_list_repeat1,
    STATE(5695), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288145] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7209), 1,
      sym_box,
    ACTIONS(7207), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(5696), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288167] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7211), 1,
      anon_sym_COMMA,
    ACTIONS(7213), 1,
      anon_sym_COLON,
    STATE(5717), 1,
      aux_sym_constant_identifier_list_repeat1,
    STATE(5697), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288191] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7107), 1,
      anon_sym_COMMA,
    ACTIONS(7215), 1,
      anon_sym_RPAREN,
    STATE(5896), 1,
      aux_sym_element_association_list_repeat1,
    STATE(5698), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288215] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7217), 1,
      anon_sym_COMMA,
    ACTIONS(7219), 1,
      anon_sym_RPAREN,
    STATE(5930), 1,
      aux_sym_aggregate_repeat1,
    STATE(5699), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288239] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7221), 1,
      anon_sym_SEMI,
    STATE(7396), 1,
      sym_generic_map_aspect,
    STATE(5700), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288263] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7223), 1,
      anon_sym_SEMI,
    ACTIONS(7225), 1,
      anon_sym_COMMA,
    STATE(5722), 1,
      aux_sym_selected_name_list_repeat1,
    STATE(5701), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288287] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7018), 1,
      anon_sym_COMMA,
    ACTIONS(7227), 1,
      anon_sym_COLON,
    STATE(5217), 1,
      aux_sym_group_constituent_list_repeat1,
    STATE(5702), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288311] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7229), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5703), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288335] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7231), 3,
      anon_sym_COMMA,
      anon_sym_COLON,
      anon_sym_LBRACK,
    STATE(5704), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288355] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6936), 1,
      anon_sym_COMMA,
    ACTIONS(7233), 1,
      anon_sym_COLON,
    STATE(5573), 1,
      aux_sym_entity_name_list_repeat1,
    STATE(5705), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288379] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7235), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5706), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288403] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6036), 3,
      sym_IS,
      sym_OPEN,
      anon_sym_SEMI,
    STATE(5707), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288423] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7237), 1,
      anon_sym_SEMI,
    ACTIONS(7239), 1,
      sym_IS,
    STATE(6231), 1,
      sym_subprogram_head,
    STATE(5708), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288447] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7241), 1,
      anon_sym_COMMA,
    ACTIONS(7243), 1,
      anon_sym_RPAREN,
    STATE(5600), 1,
      aux_sym_array_index_incomplete_type_list_repeat1,
    STATE(5709), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288471] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6064), 1,
      anon_sym_SEMI,
    STATE(6734), 1,
      sym_initialiser,
    STATE(5710), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288495] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5712), 1,
      sym_PORT,
    ACTIONS(7245), 1,
      sym_END,
    STATE(7671), 1,
      sym_port_clause,
    STATE(5711), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288519] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6803), 1,
      anon_sym_SEMI,
    ACTIONS(7247), 1,
      sym_ELSE,
    STATE(6311), 1,
      sym_else_expression_or_unaffected,
    STATE(5712), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288543] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7249), 1,
      anon_sym_RPAREN,
    STATE(5731), 1,
      aux_sym_association_list_repeat1,
    STATE(5713), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288567] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7251), 3,
      sym_FOR,
      sym_UNTIL,
      anon_sym_SEMI,
    STATE(5714), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288587] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7253), 1,
      anon_sym_SEMI,
    STATE(6747), 1,
      sym_generic_map_aspect,
    STATE(5715), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288611] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(6257), 1,
      anon_sym_SEMI,
    STATE(7439), 1,
      sym_timeout_clause,
    STATE(5716), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288635] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7255), 1,
      anon_sym_COMMA,
    ACTIONS(7258), 1,
      anon_sym_COLON,
    STATE(5717), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_constant_identifier_list_repeat1,
  [288657] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7260), 1,
      anon_sym_RPAREN,
    STATE(5720), 1,
      aux_sym_association_list_repeat1,
    STATE(5718), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288681] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7262), 1,
      anon_sym_SEMI,
    STATE(7775), 1,
      sym_initialiser,
    STATE(5719), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288705] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7264), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5720), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288729] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7266), 1,
      anon_sym_COMMA,
    ACTIONS(7269), 1,
      sym_OF,
    STATE(5721), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_index_constraint_repeat1,
  [288751] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7271), 1,
      anon_sym_SEMI,
    ACTIONS(7273), 1,
      anon_sym_COMMA,
    STATE(5722), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_selected_name_list_repeat1,
  [288773] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7276), 1,
      anon_sym_COMMA,
    ACTIONS(7278), 1,
      anon_sym_COLON,
    STATE(5743), 1,
      aux_sym_identifier_list_repeat1,
    STATE(5723), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288797] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7280), 1,
      anon_sym_SEMI,
    STATE(6754), 1,
      sym_generic_map_aspect,
    STATE(5724), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288821] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7282), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5725), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288845] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3001), 1,
      sym_SEVERITY,
    ACTIONS(7284), 1,
      anon_sym_SEMI,
    STATE(6874), 1,
      sym_severity_expression,
    STATE(5726), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288869] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(7286), 1,
      anon_sym_SEMI,
    STATE(7538), 1,
      sym_when_expression,
    STATE(5727), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288893] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(7288), 1,
      anon_sym_SEMI,
    STATE(7541), 1,
      sym_when_expression,
    STATE(5728), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288917] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7290), 1,
      anon_sym_SEMI,
    STATE(6759), 1,
      sym_initialiser,
    STATE(5729), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288941] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7292), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5730), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288961] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7294), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5731), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [288985] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7296), 1,
      anon_sym_LT_EQ,
    ACTIONS(7298), 1,
      sym_variable_assignment,
    STATE(22), 1,
      sym_signal_assignment,
    STATE(5732), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289009] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7300), 1,
      anon_sym_COMMA,
    ACTIONS(7302), 1,
      anon_sym_COLON,
    STATE(5745), 1,
      aux_sym_instantiation_list_repeat1,
    STATE(5733), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289033] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7304), 1,
      anon_sym_RPAREN,
    STATE(5735), 1,
      aux_sym_association_list_repeat1,
    STATE(5734), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289057] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7306), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5735), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289081] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7308), 1,
      anon_sym_SEMI,
    STATE(7828), 1,
      sym_initialiser,
    STATE(5736), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289105] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(7310), 1,
      anon_sym_SEMI,
    STATE(7064), 1,
      sym_port_map_aspect,
    STATE(5737), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289129] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7312), 1,
      anon_sym_SEMI,
    STATE(8050), 1,
      sym_generic_map_aspect,
    STATE(5738), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289153] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6077), 1,
      sym_OF,
    ACTIONS(7314), 1,
      anon_sym_COMMA,
    STATE(5911), 1,
      aux_sym_index_constraint_repeat1,
    STATE(5739), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289177] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7316), 1,
      anon_sym_SEMI,
    STATE(8077), 1,
      sym_generic_map_aspect,
    STATE(5740), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289201] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7318), 1,
      sym_CONSTANT,
    ACTIONS(7320), 1,
      sym_SIGNAL,
    ACTIONS(7322), 1,
      sym_VARIABLE,
    STATE(5741), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289225] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7324), 1,
      sym_SHARED,
    ACTIONS(7326), 1,
      sym_VARIABLE,
    STATE(4698), 1,
      sym_variable_declaration,
    STATE(5742), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289249] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7328), 1,
      anon_sym_COMMA,
    ACTIONS(7331), 1,
      anon_sym_COLON,
    STATE(5743), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_identifier_list_repeat1,
  [289271] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7333), 1,
      anon_sym_COMMA,
    ACTIONS(7336), 1,
      anon_sym_RPAREN,
    STATE(5744), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_enumeration_type_definition_repeat1,
  [289293] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7338), 1,
      anon_sym_COMMA,
    ACTIONS(7341), 1,
      anon_sym_COLON,
    STATE(5745), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_instantiation_list_repeat1,
  [289315] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7343), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5746), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289335] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7345), 1,
      anon_sym_COMMA,
    ACTIONS(7347), 1,
      anon_sym_RPAREN,
    STATE(5935), 1,
      aux_sym_enumeration_type_definition_repeat1,
    STATE(5747), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289359] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7349), 3,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
    STATE(5748), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289379] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6098), 1,
      anon_sym_SEMI,
    STATE(7869), 1,
      sym_initialiser,
    STATE(5749), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289403] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7351), 1,
      anon_sym_COMMA,
    ACTIONS(7354), 1,
      anon_sym_RPAREN,
    STATE(5750), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_record_resolution_repeat1,
  [289425] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2588), 3,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
    STATE(5751), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289445] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5775), 1,
      aux_sym_index_constraint_repeat1,
    ACTIONS(6077), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(5752), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289467] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7356), 1,
      anon_sym_RPAREN,
    STATE(5686), 1,
      aux_sym_association_list_repeat1,
    STATE(5753), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289491] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7358), 1,
      anon_sym_RPAREN,
    STATE(5758), 1,
      aux_sym_association_list_repeat1,
    STATE(5754), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289515] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7360), 1,
      anon_sym_COMMA,
    ACTIONS(7363), 1,
      anon_sym_RPAREN,
    STATE(5755), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_association_or_range_list_repeat1,
  [289537] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7365), 3,
      sym_RETURN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
    STATE(5756), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289557] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2830), 3,
      sym_GENERIC,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(5757), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289577] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7367), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5758), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289601] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7369), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5759), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289625] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7276), 1,
      anon_sym_COMMA,
    ACTIONS(7371), 1,
      anon_sym_COLON,
    STATE(5743), 1,
      aux_sym_identifier_list_repeat1,
    STATE(5760), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289649] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2823), 3,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
    STATE(5761), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289669] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(6828), 1,
      anon_sym_SEMI,
    STATE(6156), 1,
      sym_when_expression,
    STATE(5762), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289693] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2674), 3,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    STATE(5763), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289713] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7038), 1,
      sym_directive_body,
    ACTIONS(7373), 1,
      sym__directive_newline,
    STATE(5683), 1,
      aux_sym_user_directive_repeat1,
    STATE(5764), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289737] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7095), 1,
      anon_sym_DOT,
    ACTIONS(7097), 1,
      anon_sym_LPAREN,
    ACTIONS(7375), 1,
      anon_sym_COLON,
    STATE(5765), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289761] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6830), 3,
      sym_WHEN,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(5766), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289781] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7377), 1,
      anon_sym_COMMA,
    ACTIONS(7380), 1,
      anon_sym_COLON,
    STATE(5767), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_generic_identifier_list_repeat1,
  [289803] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7382), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5768), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289823] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7007), 1,
      anon_sym_COMMA,
    ACTIONS(7384), 1,
      anon_sym_SEMI,
    STATE(5780), 1,
      aux_sym_logical_name_list_repeat1,
    STATE(5769), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289847] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7386), 1,
      anon_sym_SEMI,
    STATE(7795), 1,
      sym_generic_map_aspect,
    STATE(5770), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289871] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7388), 1,
      anon_sym_RPAREN,
    STATE(5776), 1,
      aux_sym_association_list_repeat1,
    STATE(5771), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289895] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7390), 1,
      anon_sym_SEMI,
    ACTIONS(7392), 1,
      anon_sym_RPAREN,
    STATE(5691), 1,
      aux_sym_interface_list_repeat1,
    STATE(5772), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289919] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7396), 1,
      sym_IS,
    ACTIONS(7394), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5773), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289941] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7398), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5774), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289961] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    STATE(5797), 1,
      aux_sym_index_constraint_repeat1,
    ACTIONS(7400), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(5775), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [289983] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7402), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5776), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290007] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7404), 3,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
    STATE(5777), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290027] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6843), 3,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
    STATE(5778), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290047] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7406), 1,
      anon_sym_COMMA,
    ACTIONS(7409), 1,
      anon_sym_RPAREN,
    STATE(5779), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_association_list_repeat1,
  [290069] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7411), 1,
      anon_sym_SEMI,
    ACTIONS(7413), 1,
      anon_sym_COMMA,
    STATE(5780), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_logical_name_list_repeat1,
  [290091] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7416), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5781), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290115] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7418), 1,
      anon_sym_COMMA,
    ACTIONS(7420), 1,
      anon_sym_RPAREN,
    STATE(5750), 1,
      aux_sym_record_resolution_repeat1,
    STATE(5782), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290139] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7422), 1,
      anon_sym_RPAREN,
    STATE(5792), 1,
      aux_sym_association_list_repeat1,
    STATE(5783), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290163] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7424), 1,
      anon_sym_RPAREN,
    STATE(5785), 1,
      aux_sym_association_list_repeat1,
    STATE(5784), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290187] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7426), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5785), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290211] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2440), 3,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    STATE(5786), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290231] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6868), 1,
      anon_sym_PIPE,
    ACTIONS(7051), 1,
      anon_sym_EQ_GT,
    STATE(6386), 1,
      sym_case_generate_body,
    STATE(5787), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290255] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3292), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(5788), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290275] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7428), 1,
      anon_sym_COMMA,
    ACTIONS(7431), 1,
      anon_sym_COLON,
    STATE(5789), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_entity_name_list_repeat1,
  [290297] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6952), 1,
      anon_sym_COMMA,
    ACTIONS(7433), 1,
      anon_sym_COLON,
    STATE(5767), 1,
      aux_sym_generic_identifier_list_repeat1,
    STATE(5790), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290321] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3894), 1,
      anon_sym_RPAREN,
    ACTIONS(7435), 1,
      anon_sym_SEMI,
    STATE(5793), 1,
      aux_sym_generic_interface_list_repeat1,
    STATE(5791), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290345] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7437), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5792), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290369] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7439), 1,
      anon_sym_SEMI,
    ACTIONS(7442), 1,
      anon_sym_RPAREN,
    STATE(5793), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_generic_interface_list_repeat1,
  [290391] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7444), 1,
      anon_sym_RPAREN,
    STATE(5796), 1,
      aux_sym_association_list_repeat1,
    STATE(5794), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290415] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2588), 3,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      sym_variable_assignment,
    STATE(5795), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290435] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7446), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5796), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290459] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7269), 1,
      anon_sym_RPAREN,
    ACTIONS(7448), 1,
      anon_sym_COMMA,
    STATE(5797), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_index_constraint_repeat1,
  [290481] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6854), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
    STATE(5798), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290501] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7451), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5799), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290525] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2674), 3,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
    STATE(5800), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290545] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7211), 1,
      anon_sym_COMMA,
    ACTIONS(7453), 1,
      anon_sym_COLON,
    STATE(5697), 1,
      aux_sym_constant_identifier_list_repeat1,
    STATE(5801), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290569] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7455), 3,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
    STATE(5802), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290589] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7457), 3,
      sym_ELSE,
      sym_ELSIF,
      sym_END,
    STATE(5803), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290609] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2440), 3,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
    STATE(5804), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290629] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7459), 1,
      anon_sym_SEMI,
    STATE(8070), 1,
      sym_generic_map_aspect,
    STATE(5805), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290653] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7461), 1,
      anon_sym_RPAREN,
    STATE(5814), 1,
      aux_sym_association_list_repeat1,
    STATE(5806), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290677] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6180), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_DOT,
    STATE(5807), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290697] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7463), 3,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
    STATE(5808), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290717] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7465), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5809), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290741] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7018), 1,
      anon_sym_COMMA,
    ACTIONS(7467), 1,
      anon_sym_SEMI,
    STATE(5859), 1,
      aux_sym_group_constituent_list_repeat1,
    STATE(5810), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290765] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5992), 3,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
    STATE(5811), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290785] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7225), 1,
      anon_sym_COMMA,
    ACTIONS(7469), 1,
      anon_sym_SEMI,
    STATE(5701), 1,
      aux_sym_selected_name_list_repeat1,
    STATE(5812), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290809] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7471), 1,
      anon_sym_SEMI,
    STATE(7108), 1,
      sym_generic_map_aspect,
    STATE(5813), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290833] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7473), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5814), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290857] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7418), 1,
      anon_sym_COMMA,
    ACTIONS(7475), 1,
      anon_sym_RPAREN,
    STATE(5782), 1,
      aux_sym_record_resolution_repeat1,
    STATE(5815), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290881] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7477), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(5816), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290901] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7479), 1,
      anon_sym_SEMI,
    STATE(8151), 1,
      sym_initialiser,
    STATE(5817), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290925] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7481), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5818), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290949] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2588), 3,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
    STATE(5819), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290969] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7483), 1,
      anon_sym_RPAREN,
    STATE(5822), 1,
      aux_sym_association_list_repeat1,
    STATE(5820), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [290993] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2674), 3,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
    STATE(5821), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291013] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7485), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5822), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291037] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2440), 3,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
    STATE(5823), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291057] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7487), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5824), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291077] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7018), 1,
      anon_sym_COMMA,
    ACTIONS(7489), 1,
      anon_sym_COLON,
    STATE(5702), 1,
      aux_sym_group_constituent_list_repeat1,
    STATE(5825), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291101] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7491), 1,
      anon_sym_SEMI,
    ACTIONS(7494), 1,
      anon_sym_RPAREN,
    STATE(5826), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_interface_list_repeat1,
  [291123] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5755), 1,
      anon_sym_SEMI,
    STATE(6917), 1,
      sym_when_expression,
    STATE(5827), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291147] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5579), 1,
      sym_RANGE,
    ACTIONS(5859), 1,
      sym_IS,
    STATE(7206), 1,
      sym_range_constraint,
    STATE(5828), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291171] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3001), 1,
      sym_SEVERITY,
    ACTIONS(7496), 1,
      anon_sym_SEMI,
    STATE(7428), 1,
      sym_severity_expression,
    STATE(5829), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291195] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7203), 1,
      anon_sym_COMMA,
    ACTIONS(7498), 1,
      anon_sym_RPAREN,
    STATE(5607), 1,
      aux_sym_entity_class_entry_list_repeat1,
    STATE(5830), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291219] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7018), 1,
      anon_sym_COMMA,
    ACTIONS(7500), 1,
      anon_sym_RPAREN,
    STATE(5608), 1,
      aux_sym_group_constituent_list_repeat1,
    STATE(5831), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291243] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5704), 1,
      anon_sym_SEMI,
    STATE(6922), 1,
      sym_when_expression,
    STATE(5832), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291267] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7502), 1,
      anon_sym_SEMI,
    ACTIONS(5923), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(5833), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291289] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3190), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(5834), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291309] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7504), 1,
      anon_sym_LT_EQ,
    ACTIONS(7506), 1,
      sym_variable_assignment,
    STATE(98), 1,
      sym_signal_assignment,
    STATE(5835), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291333] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7508), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5836), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291357] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5815), 1,
      anon_sym_SEMI,
    STATE(6935), 1,
      sym_timeout_clause,
    STATE(5837), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291381] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7510), 1,
      anon_sym_SEMI,
    STATE(7857), 1,
      sym_generic_map_aspect,
    STATE(5838), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291405] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7512), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5839), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291429] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7247), 1,
      sym_ELSE,
    ACTIONS(7514), 1,
      anon_sym_SEMI,
    STATE(6311), 1,
      sym_else_expression_or_unaffected,
    STATE(5840), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291453] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7516), 1,
      anon_sym_RPAREN,
    STATE(5858), 1,
      aux_sym_association_list_repeat1,
    STATE(5841), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291477] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7518), 1,
      anon_sym_RPAREN,
    STATE(5849), 1,
      aux_sym_association_list_repeat1,
    STATE(5842), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291501] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3220), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(5843), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291521] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7239), 1,
      sym_IS,
    ACTIONS(7520), 1,
      anon_sym_SEMI,
    STATE(6379), 1,
      sym_subprogram_head,
    STATE(5844), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291545] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7296), 1,
      anon_sym_LT_EQ,
    ACTIONS(7522), 1,
      sym_variable_assignment,
    STATE(18), 1,
      sym_signal_assignment,
    STATE(5845), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291569] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6283), 1,
      anon_sym_RPAREN,
    ACTIONS(7018), 1,
      anon_sym_COMMA,
    STATE(5903), 1,
      aux_sym_group_constituent_list_repeat1,
    STATE(5846), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291593] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(7524), 1,
      anon_sym_SEMI,
    STATE(7539), 1,
      sym_port_map_aspect,
    STATE(5847), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291617] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7276), 1,
      anon_sym_COMMA,
    ACTIONS(7526), 1,
      anon_sym_COLON,
    STATE(5723), 1,
      aux_sym_identifier_list_repeat1,
    STATE(5848), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291641] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7528), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5849), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291665] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3001), 1,
      sym_SEVERITY,
    ACTIONS(7530), 1,
      anon_sym_SEMI,
    STATE(7569), 1,
      sym_severity_expression,
    STATE(5850), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291689] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7532), 1,
      anon_sym_SEMI,
    ACTIONS(5923), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(5851), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291711] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7534), 3,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
    STATE(5852), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291731] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7536), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5853), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291755] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(7538), 1,
      anon_sym_SEMI,
    STATE(7602), 1,
      sym_port_map_aspect,
    STATE(5854), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291779] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3001), 1,
      sym_SEVERITY,
    ACTIONS(7540), 1,
      anon_sym_SEMI,
    STATE(6984), 1,
      sym_severity_expression,
    STATE(5855), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291803] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7542), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5856), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291827] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7544), 1,
      anon_sym_SEMI,
    STATE(7611), 1,
      sym_initialiser,
    STATE(5857), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291851] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7546), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5858), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291875] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7018), 1,
      anon_sym_COMMA,
    ACTIONS(7548), 1,
      anon_sym_SEMI,
    STATE(5217), 1,
      aux_sym_group_constituent_list_repeat1,
    STATE(5859), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291899] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7550), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5860), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291923] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5735), 1,
      anon_sym_SEMI,
    STATE(7631), 1,
      sym_when_expression,
    STATE(5861), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291947] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5737), 1,
      anon_sym_SEMI,
    STATE(7632), 1,
      sym_when_expression,
    STATE(5862), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291971] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6087), 3,
      sym_IS,
      sym_OPEN,
      anon_sym_SEMI,
    STATE(5863), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [291991] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5879), 1,
      anon_sym_SEMI,
    STATE(7634), 1,
      sym_timeout_clause,
    STATE(5864), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292015] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(7552), 1,
      anon_sym_SEMI,
    STATE(6650), 1,
      sym_timeout_clause,
    STATE(5865), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292039] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3001), 1,
      sym_SEVERITY,
    ACTIONS(7554), 1,
      anon_sym_SEMI,
    STATE(7645), 1,
      sym_severity_expression,
    STATE(5866), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292063] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7556), 1,
      anon_sym_SEMI,
    ACTIONS(5923), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(5867), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292085] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3174), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(5868), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292105] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7558), 1,
      anon_sym_RPAREN,
    STATE(5876), 1,
      aux_sym_association_list_repeat1,
    STATE(5869), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292129] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7560), 1,
      anon_sym_SEMI,
    ACTIONS(5923), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(5870), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292151] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(7562), 1,
      anon_sym_SEMI,
    STATE(7699), 1,
      sym_port_map_aspect,
    STATE(5871), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292175] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7564), 1,
      anon_sym_SEMI,
    STATE(7718), 1,
      sym_generic_map_aspect,
    STATE(5872), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292199] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7566), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5873), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292223] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7504), 1,
      anon_sym_LT_EQ,
    ACTIONS(7568), 1,
      sym_variable_assignment,
    STATE(19), 1,
      sym_signal_assignment,
    STATE(5874), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292247] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(6273), 1,
      anon_sym_SEMI,
    STATE(7754), 1,
      sym_timeout_clause,
    STATE(5875), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292271] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7570), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5876), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292295] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7572), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5877), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292319] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(7574), 1,
      anon_sym_SEMI,
    STATE(7785), 1,
      sym_when_expression,
    STATE(5878), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292343] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(7576), 1,
      anon_sym_SEMI,
    STATE(7787), 1,
      sym_when_expression,
    STATE(5879), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292367] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7578), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5880), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292387] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7580), 1,
      anon_sym_SEMI,
    STATE(7813), 1,
      sym_initialiser,
    STATE(5881), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292411] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2968), 3,
      anon_sym_SEMI,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(5882), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292431] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6279), 1,
      anon_sym_SEMI,
    STATE(7818), 1,
      sym_initialiser,
    STATE(5883), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292455] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7582), 1,
      anon_sym_RPAREN,
    STATE(5904), 1,
      aux_sym_association_list_repeat1,
    STATE(5884), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292479] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7504), 1,
      anon_sym_LT_EQ,
    ACTIONS(7584), 1,
      sym_variable_assignment,
    STATE(82), 1,
      sym_signal_assignment,
    STATE(5885), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292503] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7296), 1,
      anon_sym_LT_EQ,
    ACTIONS(7586), 1,
      sym_variable_assignment,
    STATE(83), 1,
      sym_signal_assignment,
    STATE(5886), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292527] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7504), 1,
      anon_sym_LT_EQ,
    ACTIONS(7588), 1,
      sym_variable_assignment,
    STATE(84), 1,
      sym_signal_assignment,
    STATE(5887), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292551] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7590), 1,
      anon_sym_SEMI,
    STATE(7868), 1,
      sym_generic_map_aspect,
    STATE(5888), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292575] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7592), 3,
      sym_IS,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(5889), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292595] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7296), 1,
      anon_sym_LT_EQ,
    ACTIONS(7594), 1,
      sym_variable_assignment,
    STATE(86), 1,
      sym_signal_assignment,
    STATE(5890), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292619] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7596), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5891), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292643] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7598), 1,
      anon_sym_SEMI,
    STATE(8018), 1,
      sym_generic_map_aspect,
    STATE(5892), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292667] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7504), 1,
      anon_sym_LT_EQ,
    ACTIONS(7600), 1,
      sym_variable_assignment,
    STATE(89), 1,
      sym_signal_assignment,
    STATE(5893), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292691] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7296), 1,
      anon_sym_LT_EQ,
    ACTIONS(7602), 1,
      sym_variable_assignment,
    STATE(90), 1,
      sym_signal_assignment,
    STATE(5894), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292715] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7504), 1,
      anon_sym_LT_EQ,
    ACTIONS(7604), 1,
      sym_variable_assignment,
    STATE(91), 1,
      sym_signal_assignment,
    STATE(5895), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292739] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7606), 1,
      anon_sym_COMMA,
    ACTIONS(7609), 1,
      anon_sym_RPAREN,
    STATE(5896), 4,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
      aux_sym_element_association_list_repeat1,
  [292761] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7611), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5897), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292785] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7296), 1,
      anon_sym_LT_EQ,
    ACTIONS(7613), 1,
      sym_variable_assignment,
    STATE(93), 1,
      sym_signal_assignment,
    STATE(5898), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292809] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(7615), 1,
      anon_sym_SEMI,
    STATE(8024), 1,
      sym_timeout_clause,
    STATE(5899), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292833] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6848), 3,
      sym_PARAMETER,
      sym_RETURN,
      anon_sym_LPAREN,
    STATE(5900), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292853] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7239), 1,
      sym_IS,
    ACTIONS(7617), 1,
      anon_sym_SEMI,
    STATE(6461), 1,
      sym_subprogram_head,
    STATE(5901), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292877] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7619), 1,
      anon_sym_SEMI,
    STATE(8030), 1,
      sym_initialiser,
    STATE(5902), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292901] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6201), 1,
      anon_sym_RPAREN,
    ACTIONS(7018), 1,
      anon_sym_COMMA,
    STATE(5217), 1,
      aux_sym_group_constituent_list_repeat1,
    STATE(5903), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292925] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7621), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5904), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292949] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7623), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5905), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292973] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7625), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5906), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [292997] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7627), 1,
      anon_sym_SEMI,
    STATE(6761), 1,
      sym_initialiser,
    STATE(5907), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293021] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7241), 1,
      anon_sym_COMMA,
    ACTIONS(7629), 1,
      anon_sym_RPAREN,
    STATE(5709), 1,
      aux_sym_array_index_incomplete_type_list_repeat1,
    STATE(5908), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293045] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7239), 1,
      sym_IS,
    ACTIONS(7631), 1,
      anon_sym_SEMI,
    STATE(6410), 1,
      sym_subprogram_head,
    STATE(5909), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293069] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7300), 1,
      anon_sym_COMMA,
    ACTIONS(7633), 1,
      anon_sym_COLON,
    STATE(5733), 1,
      aux_sym_instantiation_list_repeat1,
    STATE(5910), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293093] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7314), 1,
      anon_sym_COMMA,
    ACTIONS(7400), 1,
      sym_OF,
    STATE(5721), 1,
      aux_sym_index_constraint_repeat1,
    STATE(5911), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293117] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(7635), 1,
      anon_sym_SEMI,
    STATE(8121), 1,
      sym_port_map_aspect,
    STATE(5912), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293141] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3001), 1,
      sym_SEVERITY,
    ACTIONS(7637), 1,
      anon_sym_SEMI,
    STATE(8140), 1,
      sym_severity_expression,
    STATE(5913), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293165] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7639), 1,
      anon_sym_SEMI,
    ACTIONS(5923), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(5914), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293187] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7641), 1,
      anon_sym_RPAREN,
    STATE(5920), 1,
      aux_sym_association_list_repeat1,
    STATE(5915), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293211] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7643), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5916), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293235] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(7645), 1,
      anon_sym_SEMI,
    STATE(8163), 1,
      sym_port_map_aspect,
    STATE(5917), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293259] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6193), 3,
      sym_IS,
      sym_OPEN,
      anon_sym_SEMI,
    STATE(5918), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293279] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7647), 1,
      anon_sym_SEMI,
    STATE(8168), 1,
      sym_initialiser,
    STATE(5919), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293303] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7649), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5920), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293327] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7651), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5921), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293351] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5743), 1,
      anon_sym_SEMI,
    STATE(8183), 1,
      sym_when_expression,
    STATE(5922), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293375] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(5747), 1,
      anon_sym_SEMI,
    STATE(8184), 1,
      sym_when_expression,
    STATE(5923), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293399] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3212), 3,
      sym_END,
      sym_FOR,
      sym_USE,
    STATE(5924), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293419] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(5787), 1,
      anon_sym_SEMI,
    STATE(8186), 1,
      sym_timeout_clause,
    STATE(5925), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293443] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5992), 3,
      sym_PARAMETER,
      anon_sym_SEMI,
      anon_sym_LPAREN,
    STATE(5926), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293463] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7038), 1,
      sym_directive_body,
    ACTIONS(7653), 1,
      sym__directive_newline,
    STATE(5931), 1,
      aux_sym_user_directive_repeat1,
    STATE(5927), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293487] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3001), 1,
      sym_SEVERITY,
    ACTIONS(7655), 1,
      anon_sym_SEMI,
    STATE(8197), 1,
      sym_severity_expression,
    STATE(5928), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293511] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7657), 1,
      anon_sym_SEMI,
    ACTIONS(5923), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(5929), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293533] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7217), 1,
      anon_sym_COMMA,
    ACTIONS(7659), 1,
      anon_sym_RPAREN,
    STATE(5621), 1,
      aux_sym_aggregate_repeat1,
    STATE(5930), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293557] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7038), 1,
      sym_directive_body,
    ACTIONS(7661), 1,
      sym__directive_newline,
    STATE(5683), 1,
      aux_sym_user_directive_repeat1,
    STATE(5931), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293581] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    ACTIONS(7663), 1,
      anon_sym_SEMI,
    STATE(8216), 1,
      sym_port_map_aspect,
    STATE(5932), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293605] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7665), 1,
      anon_sym_SEMI,
    STATE(8220), 1,
      sym_generic_map_aspect,
    STATE(5933), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293629] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7667), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5934), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293653] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7345), 1,
      anon_sym_COMMA,
    ACTIONS(7669), 1,
      anon_sym_RPAREN,
    STATE(5744), 1,
      aux_sym_enumeration_type_definition_repeat1,
    STATE(5935), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293677] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(6062), 1,
      anon_sym_SEMI,
    STATE(8228), 1,
      sym_timeout_clause,
    STATE(5936), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293701] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7671), 1,
      anon_sym_RPAREN,
    STATE(5557), 1,
      aux_sym_association_list_repeat1,
    STATE(5937), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293725] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7673), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5938), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293749] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(7675), 1,
      anon_sym_SEMI,
    STATE(8235), 1,
      sym_when_expression,
    STATE(5939), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293773] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(792), 1,
      sym_WHEN,
    ACTIONS(7677), 1,
      anon_sym_SEMI,
    STATE(8236), 1,
      sym_when_expression,
    STATE(5940), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293797] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7679), 1,
      anon_sym_SEMI,
    STATE(8242), 1,
      sym_initialiser,
    STATE(5941), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293821] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(6081), 1,
      anon_sym_SEMI,
    STATE(8243), 1,
      sym_initialiser,
    STATE(5942), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293845] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7681), 1,
      anon_sym_RPAREN,
    STATE(5952), 1,
      aux_sym_association_list_repeat1,
    STATE(5943), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293869] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7276), 1,
      anon_sym_COMMA,
    ACTIONS(7683), 1,
      anon_sym_COLON,
    STATE(5760), 1,
      aux_sym_identifier_list_repeat1,
    STATE(5944), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293893] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7685), 1,
      anon_sym_SEMI,
    STATE(8251), 1,
      sym_generic_map_aspect,
    STATE(5945), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293917] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7687), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5946), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293941] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1605), 1,
      sym_GENERIC,
    ACTIONS(7689), 1,
      anon_sym_SEMI,
    STATE(8254), 1,
      sym_generic_map_aspect,
    STATE(5947), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293965] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7691), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5948), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [293989] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5769), 1,
      sym_FOR,
    ACTIONS(7693), 1,
      anon_sym_SEMI,
    STATE(8259), 1,
      sym_timeout_clause,
    STATE(5949), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7695), 3,
      sym_GENERIC,
      sym_PORT,
      anon_sym_SEMI,
    STATE(5950), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294033] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5954), 1,
      sym_variable_assignment,
    ACTIONS(7697), 1,
      anon_sym_SEMI,
    STATE(8264), 1,
      sym_initialiser,
    STATE(5951), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294057] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6908), 1,
      anon_sym_COMMA,
    ACTIONS(7699), 1,
      anon_sym_RPAREN,
    STATE(5779), 1,
      aux_sym_association_list_repeat1,
    STATE(5952), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294081] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7701), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5953), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294105] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    ACTIONS(7703), 1,
      anon_sym_SEMI,
    STATE(5376), 1,
      sym_when_expression,
    STATE(5954), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294129] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7705), 1,
      sym_CONSTANT,
    ACTIONS(7707), 1,
      sym_SIGNAL,
    ACTIONS(7709), 1,
      sym_VARIABLE,
    STATE(5955), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294153] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7711), 1,
      sym_CONSTANT,
    ACTIONS(7713), 1,
      sym_SIGNAL,
    ACTIONS(7715), 1,
      sym_VARIABLE,
    STATE(5956), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294177] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7717), 1,
      sym_CONSTANT,
    ACTIONS(7719), 1,
      sym_SIGNAL,
    ACTIONS(7721), 1,
      sym_VARIABLE,
    STATE(5957), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294201] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7723), 1,
      sym_CONSTANT,
    ACTIONS(7725), 1,
      sym_SIGNAL,
    ACTIONS(7727), 1,
      sym_VARIABLE,
    STATE(5958), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294225] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7729), 1,
      sym_CONSTANT,
    ACTIONS(7731), 1,
      sym_SIGNAL,
    ACTIONS(7733), 1,
      sym_VARIABLE,
    STATE(5959), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294249] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7735), 1,
      sym_CONSTANT,
    ACTIONS(7737), 1,
      sym_SIGNAL,
    ACTIONS(7739), 1,
      sym_VARIABLE,
    STATE(5960), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294273] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7239), 1,
      sym_IS,
    ACTIONS(7741), 1,
      anon_sym_SEMI,
    STATE(6441), 1,
      sym_subprogram_head,
    STATE(5961), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294297] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7743), 1,
      sym_CONSTANT,
    ACTIONS(7745), 1,
      sym_SIGNAL,
    ACTIONS(7747), 1,
      sym_VARIABLE,
    STATE(5962), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294321] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7749), 1,
      sym_CONSTANT,
    ACTIONS(7751), 1,
      sym_SIGNAL,
    ACTIONS(7753), 1,
      sym_VARIABLE,
    STATE(5963), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294345] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7755), 1,
      sym_CONSTANT,
    ACTIONS(7757), 1,
      sym_SIGNAL,
    ACTIONS(7759), 1,
      sym_VARIABLE,
    STATE(5964), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294369] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7761), 1,
      sym_CONSTANT,
    ACTIONS(7763), 1,
      sym_SIGNAL,
    ACTIONS(7765), 1,
      sym_VARIABLE,
    STATE(5965), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294393] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7767), 1,
      sym_CONSTANT,
    ACTIONS(7769), 1,
      sym_SIGNAL,
    ACTIONS(7771), 1,
      sym_VARIABLE,
    STATE(5966), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294417] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7773), 1,
      sym_CONSTANT,
    ACTIONS(7775), 1,
      sym_SIGNAL,
    ACTIONS(7777), 1,
      sym_VARIABLE,
    STATE(5967), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294441] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7779), 1,
      sym_CONSTANT,
    ACTIONS(7781), 1,
      sym_SIGNAL,
    ACTIONS(7783), 1,
      sym_VARIABLE,
    STATE(5968), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294465] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7785), 1,
      sym_CONSTANT,
    ACTIONS(7787), 1,
      sym_SIGNAL,
    ACTIONS(7789), 1,
      sym_VARIABLE,
    STATE(5969), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294489] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7791), 1,
      sym_CONSTANT,
    ACTIONS(7793), 1,
      sym_SIGNAL,
    ACTIONS(7795), 1,
      sym_VARIABLE,
    STATE(5970), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294513] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7797), 1,
      sym_CONSTANT,
    ACTIONS(7799), 1,
      sym_SIGNAL,
    ACTIONS(7801), 1,
      sym_VARIABLE,
    STATE(5971), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294537] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7803), 1,
      sym_CONSTANT,
    ACTIONS(7805), 1,
      sym_SIGNAL,
    ACTIONS(7807), 1,
      sym_VARIABLE,
    STATE(5972), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294561] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7809), 1,
      sym_CONSTANT,
    ACTIONS(7811), 1,
      sym_SIGNAL,
    ACTIONS(7813), 1,
      sym_VARIABLE,
    STATE(5973), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294585] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7815), 1,
      sym_CONSTANT,
    ACTIONS(7817), 1,
      sym_SIGNAL,
    ACTIONS(7819), 1,
      sym_VARIABLE,
    STATE(5974), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294609] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7821), 1,
      sym_CONSTANT,
    ACTIONS(7823), 1,
      sym_SIGNAL,
    ACTIONS(7825), 1,
      sym_VARIABLE,
    STATE(5975), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294633] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7827), 1,
      sym_CONSTANT,
    ACTIONS(7829), 1,
      sym_SIGNAL,
    ACTIONS(7831), 1,
      sym_VARIABLE,
    STATE(5976), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294657] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7833), 1,
      sym_CONSTANT,
    ACTIONS(7835), 1,
      sym_SIGNAL,
    ACTIONS(7837), 1,
      sym_VARIABLE,
    STATE(5977), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294681] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7839), 1,
      sym_CONSTANT,
    ACTIONS(7841), 1,
      sym_SIGNAL,
    ACTIONS(7843), 1,
      sym_VARIABLE,
    STATE(5978), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294705] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7845), 1,
      sym_CONSTANT,
    ACTIONS(7847), 1,
      sym_SIGNAL,
    ACTIONS(7849), 1,
      sym_VARIABLE,
    STATE(5979), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294729] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7851), 1,
      sym_CONSTANT,
    ACTIONS(7853), 1,
      sym_SIGNAL,
    ACTIONS(7855), 1,
      sym_VARIABLE,
    STATE(5980), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294753] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7857), 1,
      sym_CONSTANT,
    ACTIONS(7859), 1,
      sym_SIGNAL,
    ACTIONS(7861), 1,
      sym_VARIABLE,
    STATE(5981), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294777] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7863), 1,
      sym_CONSTANT,
    ACTIONS(7865), 1,
      sym_SIGNAL,
    ACTIONS(7867), 1,
      sym_VARIABLE,
    STATE(5982), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294801] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7869), 1,
      sym_CONSTANT,
    ACTIONS(7871), 1,
      sym_SIGNAL,
    ACTIONS(7873), 1,
      sym_VARIABLE,
    STATE(5983), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294825] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7875), 1,
      sym_CONSTANT,
    ACTIONS(7877), 1,
      sym_SIGNAL,
    ACTIONS(7879), 1,
      sym_VARIABLE,
    STATE(5984), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294849] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7881), 1,
      sym_CONSTANT,
    ACTIONS(7883), 1,
      sym_SIGNAL,
    ACTIONS(7885), 1,
      sym_VARIABLE,
    STATE(5985), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294873] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7887), 1,
      sym_CONSTANT,
    ACTIONS(7889), 1,
      sym_SIGNAL,
    ACTIONS(7891), 1,
      sym_VARIABLE,
    STATE(5986), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294897] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7893), 1,
      sym_CONSTANT,
    ACTIONS(7895), 1,
      sym_SIGNAL,
    ACTIONS(7897), 1,
      sym_VARIABLE,
    STATE(5987), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294921] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7899), 1,
      sym_CONSTANT,
    ACTIONS(7901), 1,
      sym_SIGNAL,
    ACTIONS(7903), 1,
      sym_VARIABLE,
    STATE(5988), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294945] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7905), 1,
      sym_CONSTANT,
    ACTIONS(7907), 1,
      sym_SIGNAL,
    ACTIONS(7909), 1,
      sym_VARIABLE,
    STATE(5989), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294969] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7911), 1,
      sym_CONSTANT,
    ACTIONS(7913), 1,
      sym_SIGNAL,
    ACTIONS(7915), 1,
      sym_VARIABLE,
    STATE(5990), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [294993] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7917), 1,
      sym_CONSTANT,
    ACTIONS(7919), 1,
      sym_SIGNAL,
    ACTIONS(7921), 1,
      sym_VARIABLE,
    STATE(5991), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295017] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7923), 1,
      sym_CONSTANT,
    ACTIONS(7925), 1,
      sym_SIGNAL,
    ACTIONS(7927), 1,
      sym_VARIABLE,
    STATE(5992), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295041] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7929), 1,
      sym_CONSTANT,
    ACTIONS(7931), 1,
      sym_SIGNAL,
    ACTIONS(7933), 1,
      sym_VARIABLE,
    STATE(5993), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295065] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7935), 1,
      sym_CONSTANT,
    ACTIONS(7937), 1,
      sym_SIGNAL,
    ACTIONS(7939), 1,
      sym_VARIABLE,
    STATE(5994), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295089] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7941), 1,
      sym_CONSTANT,
    ACTIONS(7943), 1,
      sym_SIGNAL,
    ACTIONS(7945), 1,
      sym_VARIABLE,
    STATE(5995), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295113] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7947), 1,
      sym_CONSTANT,
    ACTIONS(7949), 1,
      sym_SIGNAL,
    ACTIONS(7951), 1,
      sym_VARIABLE,
    STATE(5996), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295137] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7953), 1,
      sym_CONSTANT,
    ACTIONS(7955), 1,
      sym_SIGNAL,
    ACTIONS(7957), 1,
      sym_VARIABLE,
    STATE(5997), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295161] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7959), 1,
      sym_CONSTANT,
    ACTIONS(7961), 1,
      sym_SIGNAL,
    ACTIONS(7963), 1,
      sym_VARIABLE,
    STATE(5998), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295185] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7965), 1,
      sym_CONSTANT,
    ACTIONS(7967), 1,
      sym_SIGNAL,
    ACTIONS(7969), 1,
      sym_VARIABLE,
    STATE(5999), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295209] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7971), 1,
      sym_CONSTANT,
    ACTIONS(7973), 1,
      sym_SIGNAL,
    ACTIONS(7975), 1,
      sym_VARIABLE,
    STATE(6000), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295233] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7977), 1,
      sym_CONSTANT,
    ACTIONS(7979), 1,
      sym_SIGNAL,
    ACTIONS(7981), 1,
      sym_VARIABLE,
    STATE(6001), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295257] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7983), 1,
      sym_CONSTANT,
    ACTIONS(7985), 1,
      sym_SIGNAL,
    ACTIONS(7987), 1,
      sym_VARIABLE,
    STATE(6002), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295281] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7989), 1,
      sym_CONSTANT,
    ACTIONS(7991), 1,
      sym_SIGNAL,
    ACTIONS(7993), 1,
      sym_VARIABLE,
    STATE(6003), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295305] = 7,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7995), 1,
      sym_CONSTANT,
    ACTIONS(7997), 1,
      sym_SIGNAL,
    ACTIONS(7999), 1,
      sym_VARIABLE,
    STATE(6004), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295329] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8003), 1,
      anon_sym_DOT,
    ACTIONS(8001), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(6005), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295351] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8005), 1,
      sym_END,
    STATE(8188), 1,
      sym_end_loop,
    STATE(6006), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295372] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(8031), 1,
      sym_end_process,
    STATE(6007), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295393] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8009), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6008), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295412] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4314), 1,
      anon_sym_LPAREN,
    STATE(4403), 1,
      sym_parenthesis_group,
    STATE(6009), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295433] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8011), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6010), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295452] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8013), 2,
      sym_END,
      sym_FOR,
    STATE(6011), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295471] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1563), 1,
      anon_sym_LPAREN,
    STATE(943), 1,
      sym_parenthesis_group,
    STATE(6012), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295492] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7404), 2,
      sym_END,
      sym_WHEN,
    STATE(6013), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295511] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4314), 1,
      anon_sym_LPAREN,
    STATE(4407), 1,
      sym_parenthesis_group,
    STATE(6014), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295532] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2614), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6015), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295551] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1563), 1,
      anon_sym_LPAREN,
    STATE(950), 1,
      sym_parenthesis_group,
    STATE(6016), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295572] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(8097), 1,
      sym_end_process,
    STATE(6017), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295593] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(6647), 1,
      sym_end_block,
    STATE(6018), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295614] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(8109), 1,
      sym_end_generate,
    STATE(6019), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295635] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1112), 1,
      sym_BEGIN,
    STATE(6261), 1,
      sym_concurrent_block,
    STATE(6020), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295656] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8019), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6021), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295675] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5706), 1,
      sym_END,
    STATE(8116), 1,
      sym_end_component,
    STATE(6022), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295696] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8021), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6023), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295715] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8023), 1,
      sym_IS,
    STATE(6110), 1,
      sym_architecture_head,
    STATE(6024), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295736] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8025), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6025), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295755] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8027), 1,
      sym_END,
    STATE(8127), 1,
      sym_end_package,
    STATE(6026), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295776] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(8128), 1,
      sym_end_process,
    STATE(6027), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295797] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8029), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6028), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295816] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8031), 1,
      sym_END,
    STATE(8141), 1,
      sym_end_if,
    STATE(6029), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295837] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8005), 1,
      sym_END,
    STATE(8142), 1,
      sym_end_loop,
    STATE(6030), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295858] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2636), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6031), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295877] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8033), 1,
      anon_sym_LPAREN,
    STATE(4893), 1,
      sym_association_list,
    STATE(6032), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295898] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2968), 2,
      sym_WHEN,
      anon_sym_SEMI,
    STATE(6033), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295917] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4649), 1,
      anon_sym_LPAREN,
    STATE(4613), 1,
      sym_parenthesis_group,
    STATE(6034), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295938] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8035), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6035), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295957] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8037), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6036), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295976] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8039), 1,
      sym_END,
    STATE(8161), 1,
      sym_subprogram_end,
    STATE(6037), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [295997] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2830), 2,
      sym_OF,
      anon_sym_COMMA,
    STATE(6038), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296016] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8041), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6039), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296035] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(8162), 1,
      sym_end_block,
    STATE(6040), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296056] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    STATE(8164), 1,
      sym_port_map_aspect,
    STATE(6041), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296077] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(8165), 1,
      sym_end_generate,
    STATE(6042), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296098] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8043), 1,
      sym_identifier,
    ACTIONS(8045), 1,
      sym_ALL,
    STATE(6043), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296119] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8047), 1,
      sym_END,
    STATE(8175), 1,
      sym_end_package_body,
    STATE(6044), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296140] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6172), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6045), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296159] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(8177), 1,
      sym_end_process,
    STATE(6046), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296180] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4649), 1,
      anon_sym_LPAREN,
    STATE(4619), 1,
      sym_parenthesis_group,
    STATE(6047), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296201] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8049), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6048), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296220] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(8182), 1,
      sym_end_block,
    STATE(6049), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296241] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8051), 1,
      sym_END,
    STATE(8187), 1,
      sym_end_case,
    STATE(6050), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296262] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8053), 1,
      sym_IS,
    STATE(6159), 1,
      sym_mode_view_body,
    STATE(6051), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296283] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6868), 1,
      anon_sym_PIPE,
    ACTIONS(8055), 1,
      anon_sym_EQ_GT,
    STATE(6052), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296304] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8031), 1,
      sym_END,
    STATE(8198), 1,
      sym_end_if,
    STATE(6053), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296325] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8057), 1,
      sym_END,
    STATE(6969), 1,
      sym_end_context,
    STATE(6054), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296346] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8059), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6055), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296365] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7402), 1,
      sym_end_process,
    STATE(6056), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296386] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(8209), 1,
      sym_end_block,
    STATE(6057), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296407] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8061), 1,
      anon_sym_SEMI,
    ACTIONS(8063), 1,
      sym_IS,
    STATE(6058), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296428] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(8221), 1,
      sym_end_process,
    STATE(6059), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296449] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5511), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6060), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296468] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(8227), 1,
      sym_end_block,
    STATE(6061), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296489] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8065), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6062), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296508] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8067), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6063), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296527] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8051), 1,
      sym_END,
    STATE(8238), 1,
      sym_end_case,
    STATE(6064), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296548] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8005), 1,
      sym_END,
    STATE(8239), 1,
      sym_end_loop,
    STATE(6065), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296569] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(7419), 1,
      sym_end_block,
    STATE(6066), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296590] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8069), 2,
      sym_END,
      sym_FOR,
    STATE(6067), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296609] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8071), 1,
      sym_END,
    STATE(8245), 1,
      sym_end_view,
    STATE(6068), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296630] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8073), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6069), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296649] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(8248), 1,
      sym_end_block,
    STATE(6070), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296670] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(8249), 1,
      sym_end_generate,
    STATE(6071), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296691] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8075), 2,
      sym_END,
      sym_FOR,
    STATE(6072), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296710] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8077), 1,
      anon_sym_LPAREN,
    STATE(4413), 1,
      sym_association_list,
    STATE(6073), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296731] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(8258), 1,
      sym_end_block,
    STATE(6074), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296752] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4599), 1,
      anon_sym_LPAREN,
    STATE(4566), 1,
      sym_parenthesis_group,
    STATE(6075), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296773] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8079), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6076), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296792] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7016), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6077), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296811] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(8265), 1,
      sym_end_process,
    STATE(6078), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296832] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4599), 1,
      anon_sym_LPAREN,
    STATE(4571), 1,
      sym_parenthesis_group,
    STATE(6079), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296853] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(106), 1,
      sym_signal_assignment,
    STATE(6080), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296874] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2640), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6081), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296893] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8083), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6082), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296912] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8085), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6083), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296931] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8087), 1,
      sym_IS,
    STATE(6269), 1,
      sym_configuration_head,
    STATE(6084), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296952] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8089), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6085), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296971] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8091), 2,
      sym_END,
      sym_WHEN,
    STATE(6086), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [296990] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8093), 1,
      anon_sym_LPAREN,
    STATE(1739), 1,
      sym_association_list,
    STATE(6087), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297011] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(6943), 1,
      sym_end_process,
    STATE(6088), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297032] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(6612), 1,
      sym_end_generate,
    STATE(6089), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297053] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8095), 1,
      anon_sym_LPAREN,
    STATE(4606), 1,
      sym_association_list,
    STATE(6090), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297074] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2211), 1,
      anon_sym_LPAREN,
    STATE(1869), 1,
      sym_parenthesis_group,
    STATE(6091), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297095] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3900), 1,
      anon_sym_LPAREN,
    STATE(3300), 1,
      sym_parenthesis_group,
    STATE(6092), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297116] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8027), 1,
      sym_END,
    STATE(7099), 1,
      sym_end_package,
    STATE(6093), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297137] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7228), 1,
      sym_end_process,
    STATE(6094), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297158] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2644), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6095), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297177] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7271), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(6096), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297196] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2211), 1,
      anon_sym_LPAREN,
    STATE(1905), 1,
      sym_parenthesis_group,
    STATE(6097), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297217] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(6098), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297236] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4216), 2,
      sym_END,
      sym_PORT,
    STATE(6099), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297255] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3900), 1,
      anon_sym_LPAREN,
    STATE(3302), 1,
      sym_parenthesis_group,
    STATE(6100), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297276] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8039), 1,
      sym_END,
    STATE(6724), 1,
      sym_subprogram_end,
    STATE(6101), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297297] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8097), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6102), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297316] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(6743), 1,
      sym_end_block,
    STATE(6103), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297337] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    STATE(6781), 1,
      sym_port_map_aspect,
    STATE(6104), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297358] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(6805), 1,
      sym_end_generate,
    STATE(6105), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297379] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7442), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6106), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297398] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8047), 1,
      sym_END,
    STATE(6876), 1,
      sym_end_package_body,
    STATE(6107), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297419] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6340), 1,
      sym_sequential_block,
    STATE(6108), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297440] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(6903), 1,
      sym_end_process,
    STATE(6109), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297461] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1112), 1,
      sym_BEGIN,
    STATE(6220), 1,
      sym_concurrent_block,
    STATE(6110), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297482] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8099), 1,
      anon_sym_LPAREN,
    STATE(5751), 1,
      sym_association_list,
    STATE(6111), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297503] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4397), 1,
      anon_sym_LPAREN,
    STATE(4523), 1,
      sym_parenthesis_group,
    STATE(6112), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297524] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8101), 1,
      anon_sym_LPAREN,
    STATE(4813), 1,
      sym_association_list,
    STATE(6113), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297545] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8103), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6114), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297564] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(25), 1,
      sym_signal_assignment,
    STATE(6115), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297585] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8051), 1,
      sym_END,
    STATE(7575), 1,
      sym_end_case,
    STATE(6116), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297606] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(7297), 1,
      sym_end_block,
    STATE(6117), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297627] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8005), 1,
      sym_END,
    STATE(7583), 1,
      sym_end_loop,
    STATE(6118), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297648] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4397), 1,
      anon_sym_LPAREN,
    STATE(4531), 1,
      sym_parenthesis_group,
    STATE(6119), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297669] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7453), 1,
      sym_end_process,
    STATE(6120), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297690] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1875), 1,
      anon_sym_LPAREN,
    STATE(1406), 1,
      sym_parenthesis_group,
    STATE(6121), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297711] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8071), 1,
      sym_END,
    STATE(7683), 1,
      sym_end_view,
    STATE(6122), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297732] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8105), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6123), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297751] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(7747), 1,
      sym_end_block,
    STATE(6124), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297772] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(7776), 1,
      sym_end_generate,
    STATE(6125), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297793] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8107), 2,
      anon_sym_SEMI,
      sym_variable_assignment,
    STATE(6126), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297812] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(8193), 1,
      sym_end_process,
    STATE(6127), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297833] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8109), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6128), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297852] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8111), 2,
      sym_WHEN,
      anon_sym_SEMI,
    STATE(6129), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297871] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8113), 1,
      anon_sym_LPAREN,
    STATE(4830), 1,
      sym_association_list,
    STATE(6130), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297892] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7609), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6131), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297911] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(6799), 1,
      sym_end_process,
    STATE(6132), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297932] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4234), 1,
      anon_sym_LPAREN,
    STATE(4372), 1,
      sym_parenthesis_group,
    STATE(6133), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297953] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6807), 1,
      sym_END,
    STATE(6715), 1,
      sym_end_entity,
    STATE(6134), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297974] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8027), 1,
      sym_END,
    STATE(6882), 1,
      sym_end_package,
    STATE(6135), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [297995] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(6890), 1,
      sym_end_process,
    STATE(6136), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298016] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8115), 1,
      anon_sym_LPAREN,
    STATE(5355), 1,
      sym_association_list,
    STATE(6137), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298037] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7331), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6138), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298056] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8039), 1,
      sym_END,
    STATE(6954), 1,
      sym_subprogram_end,
    STATE(6139), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298077] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1887), 1,
      anon_sym_LPAREN,
    STATE(1347), 1,
      sym_parenthesis_group,
    STATE(6140), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298098] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8117), 1,
      anon_sym_LPAREN,
    STATE(989), 1,
      sym_association_list,
    STATE(6141), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298119] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8047), 1,
      sym_END,
    STATE(6989), 1,
      sym_end_package_body,
    STATE(6142), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298140] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5491), 1,
      anon_sym_LPAREN,
    STATE(4977), 1,
      sym_parenthesis_group,
    STATE(6143), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298161] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7012), 1,
      sym_end_process,
    STATE(6144), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298182] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1887), 1,
      anon_sym_LPAREN,
    STATE(1349), 1,
      sym_parenthesis_group,
    STATE(6145), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298203] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7341), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6146), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298222] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1875), 1,
      anon_sym_LPAREN,
    STATE(1242), 1,
      sym_parenthesis_group,
    STATE(6147), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298243] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8119), 1,
      anon_sym_LPAREN,
    STATE(6738), 1,
      sym_association_list,
    STATE(6148), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298264] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6036), 2,
      sym_UNITS,
      anon_sym_SEMI,
    STATE(6149), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298283] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7186), 1,
      sym_end_process,
    STATE(6150), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298304] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7291), 1,
      sym_end_process,
    STATE(6151), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298325] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8121), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6152), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298344] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8123), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6153), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298363] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1689), 1,
      anon_sym_LPAREN,
    STATE(1092), 1,
      sym_parenthesis_group,
    STATE(6154), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298384] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5491), 1,
      anon_sym_LPAREN,
    STATE(4992), 1,
      sym_parenthesis_group,
    STATE(6155), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298405] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6329), 2,
      sym_ELSE,
      anon_sym_SEMI,
    STATE(6156), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298424] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8027), 1,
      sym_END,
    STATE(7564), 1,
      sym_end_package,
    STATE(6157), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298445] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8125), 1,
      anon_sym_LPAREN,
    STATE(991), 1,
      sym_association_list,
    STATE(6158), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298466] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8071), 1,
      sym_END,
    STATE(8359), 1,
      sym_end_view,
    STATE(6159), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298487] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7095), 1,
      anon_sym_DOT,
    ACTIONS(7097), 1,
      anon_sym_LPAREN,
    STATE(6160), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298508] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5137), 1,
      anon_sym_LPAREN,
    STATE(4807), 1,
      sym_parenthesis_group,
    STATE(6161), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298529] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8127), 1,
      anon_sym_SEMI,
    ACTIONS(8129), 1,
      anon_sym_COMMA,
    STATE(6162), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298550] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3272), 1,
      sym_WHEN,
    STATE(6274), 1,
      sym_when_element,
    STATE(6163), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298571] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8131), 1,
      anon_sym_LPAREN,
    STATE(4602), 1,
      sym_association_list,
    STATE(6164), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298592] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1819), 1,
      anon_sym_LPAREN,
    STATE(1367), 1,
      sym_parenthesis_group,
    STATE(6165), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298613] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8133), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6166), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298632] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8135), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6167), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298651] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8137), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6168), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298670] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7363), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6169), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298689] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1819), 1,
      anon_sym_LPAREN,
    STATE(1369), 1,
      sym_parenthesis_group,
    STATE(6170), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298710] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8139), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6171), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298729] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8141), 1,
      sym_IS,
    STATE(6200), 1,
      sym_package_declaration_body,
    STATE(6172), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298750] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8143), 1,
      anon_sym_LPAREN,
    STATE(3319), 1,
      sym_association_list,
    STATE(6173), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298771] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8145), 1,
      anon_sym_LPAREN,
    STATE(1694), 1,
      sym_association_list,
    STATE(6174), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298792] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2033), 1,
      anon_sym_LPAREN,
    STATE(1489), 1,
      sym_parenthesis_group,
    STATE(6175), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298813] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4116), 1,
      anon_sym_LPAREN,
    STATE(3534), 1,
      sym_parenthesis_group,
    STATE(6176), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298834] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8147), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6177), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298853] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5716), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6178), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298872] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2678), 2,
      sym_directive_body,
      sym__directive_newline,
    STATE(6179), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298891] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8149), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6180), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298910] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2033), 1,
      anon_sym_LPAREN,
    STATE(1493), 1,
      sym_parenthesis_group,
    STATE(6181), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298931] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2640), 2,
      sym_GENERIC,
      anon_sym_SEMI,
    STATE(6182), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298950] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4116), 1,
      anon_sym_LPAREN,
    STATE(3536), 1,
      sym_parenthesis_group,
    STATE(6183), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298971] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6611), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6184), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [298990] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8151), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6185), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299009] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6167), 2,
      sym_FOR,
      sym_USE,
    STATE(6186), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299028] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2610), 1,
      sym_OF,
    ACTIONS(2678), 1,
      anon_sym_COLON,
    STATE(6187), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299049] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8153), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6188), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299068] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8155), 1,
      sym_IS,
    STATE(6225), 1,
      sym_package_definition_body,
    STATE(6189), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299089] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(7405), 1,
      sym_end_block,
    STATE(6190), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299110] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8157), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6191), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299129] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8159), 2,
      sym_END,
      sym_WHEN,
    STATE(6192), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299148] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(7585), 1,
      sym_end_generate,
    STATE(6193), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299169] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8161), 1,
      anon_sym_LPAREN,
    STATE(4485), 1,
      sym_association_list,
    STATE(6194), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299190] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2093), 1,
      anon_sym_LPAREN,
    STATE(1514), 1,
      sym_parenthesis_group,
    STATE(6195), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299211] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8163), 2,
      sym_WHEN,
      anon_sym_SEMI,
    STATE(6196), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299230] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7247), 1,
      sym_ELSE,
    STATE(6311), 1,
      sym_else_expression_or_unaffected,
    STATE(6197), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299251] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2093), 1,
      anon_sym_LPAREN,
    STATE(1520), 1,
      sym_parenthesis_group,
    STATE(6198), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299272] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6429), 1,
      sym_sequential_block,
    STATE(6199), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299293] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8027), 1,
      sym_END,
    STATE(7277), 1,
      sym_end_package,
    STATE(6200), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299314] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7282), 1,
      sym_end_process,
    STATE(6201), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299335] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2648), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6202), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299354] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3003), 1,
      sym_END,
    STATE(6615), 1,
      sym_generate_block_end,
    STATE(6203), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299375] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2644), 2,
      sym_GENERIC,
      anon_sym_SEMI,
    STATE(6204), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299394] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8165), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6205), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299413] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8167), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6206), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299432] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8169), 1,
      anon_sym_SEMI,
    ACTIONS(8171), 1,
      anon_sym_COMMA,
    STATE(6207), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299453] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8173), 2,
      anon_sym_SEMI,
      anon_sym_COMMA,
    STATE(6208), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299472] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8175), 1,
      anon_sym_LPAREN,
    STATE(1393), 1,
      sym_association_list,
    STATE(6209), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299493] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1701), 1,
      anon_sym_LPAREN,
    STATE(1126), 1,
      sym_parenthesis_group,
    STATE(6210), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299514] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8177), 1,
      anon_sym_LPAREN,
    STATE(1409), 1,
      sym_association_list,
    STATE(6211), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299535] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8179), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6212), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299554] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6167), 2,
      sym_END,
      sym_USE,
    STATE(6213), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299573] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8181), 1,
      anon_sym_LPAREN,
    STATE(5038), 1,
      sym_association_list,
    STATE(6214), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299594] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1701), 1,
      anon_sym_LPAREN,
    STATE(1130), 1,
      sym_parenthesis_group,
    STATE(6215), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299615] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6249), 1,
      sym_sequential_block,
    STATE(6216), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299636] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5057), 1,
      anon_sym_LPAREN,
    STATE(4787), 1,
      sym_parenthesis_group,
    STATE(6217), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299657] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2574), 1,
      anon_sym_LPAREN,
    STATE(1996), 1,
      sym_parenthesis_group,
    STATE(6218), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299678] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8183), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6219), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299697] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8185), 1,
      sym_END,
    STATE(6784), 1,
      sym_end_architecture,
    STATE(6220), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299718] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8187), 1,
      anon_sym_LPAREN,
    STATE(1396), 1,
      sym_association_list,
    STATE(6221), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299739] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7431), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6222), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299758] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3948), 1,
      anon_sym_LPAREN,
    STATE(3379), 1,
      sym_parenthesis_group,
    STATE(6223), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299779] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4220), 2,
      sym_END,
      sym_PORT,
    STATE(6224), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299798] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8047), 1,
      sym_END,
    STATE(6875), 1,
      sym_end_package_body,
    STATE(6225), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299819] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(8189), 1,
      anon_sym_SEMI,
    STATE(6226), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299840] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5057), 1,
      anon_sym_LPAREN,
    STATE(4791), 1,
      sym_parenthesis_group,
    STATE(6227), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299861] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(27), 1,
      sym_signal_assignment,
    STATE(6228), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299882] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3948), 1,
      anon_sym_LPAREN,
    STATE(3382), 1,
      sym_parenthesis_group,
    STATE(6229), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299903] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1689), 1,
      anon_sym_LPAREN,
    STATE(1097), 1,
      sym_parenthesis_group,
    STATE(6230), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299924] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6318), 1,
      sym_sequential_block,
    STATE(6231), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299945] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2574), 1,
      anon_sym_LPAREN,
    STATE(2001), 1,
      sym_parenthesis_group,
    STATE(6232), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299966] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6227), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6233), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [299985] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8191), 1,
      sym_END,
    STATE(6983), 1,
      sym_end_configuration,
    STATE(6234), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300006] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8193), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6235), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300025] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8195), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6236), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300044] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8171), 1,
      anon_sym_COMMA,
    ACTIONS(8197), 1,
      anon_sym_SEMI,
    STATE(6237), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300065] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7269), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6238), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300084] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2652), 2,
      anon_sym_COMMA,
      anon_sym_COLON,
    STATE(6239), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300103] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8199), 1,
      anon_sym_LPAREN,
    STATE(1584), 1,
      sym_association_list,
    STATE(6240), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300124] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2063), 1,
      anon_sym_LPAREN,
    STATE(1585), 1,
      sym_parenthesis_group,
    STATE(6241), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300145] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8201), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6242), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300164] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6228), 1,
      sym_select_target,
    STATE(6243), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300185] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2063), 1,
      anon_sym_LPAREN,
    STATE(1594), 1,
      sym_parenthesis_group,
    STATE(6244), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300206] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(5944), 1,
      anon_sym_SEMI,
    STATE(6245), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300227] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8205), 1,
      anon_sym_LPAREN,
    STATE(3563), 1,
      sym_association_list,
    STATE(6246), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300248] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2462), 1,
      anon_sym_LPAREN,
    STATE(2100), 1,
      sym_parenthesis_group,
    STATE(6247), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300269] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8207), 1,
      anon_sym_LPAREN,
    STATE(1604), 1,
      sym_association_list,
    STATE(6248), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300290] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(6889), 1,
      sym_end_process,
    STATE(6249), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300311] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6056), 1,
      sym_sequential_block,
    STATE(6250), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300332] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1645), 1,
      anon_sym_LPAREN,
    STATE(1114), 1,
      sym_parenthesis_group,
    STATE(6251), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300353] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1575), 1,
      anon_sym_LPAREN,
    STATE(982), 1,
      sym_parenthesis_group,
    STATE(6252), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300374] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8209), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(6253), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300393] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1645), 1,
      anon_sym_LPAREN,
    STATE(1117), 1,
      sym_parenthesis_group,
    STATE(6254), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300414] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2462), 1,
      anon_sym_LPAREN,
    STATE(2102), 1,
      sym_parenthesis_group,
    STATE(6255), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300435] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8211), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(6256), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300454] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8213), 1,
      anon_sym_LPAREN,
    STATE(1203), 1,
      sym_association_list,
    STATE(6257), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300475] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8215), 2,
      sym_END,
      sym_FOR,
    STATE(6258), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300494] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1533), 1,
      anon_sym_LPAREN,
    STATE(900), 1,
      sym_parenthesis_group,
    STATE(6259), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300515] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8217), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6260), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300534] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(7146), 1,
      sym_end_block,
    STATE(6261), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300555] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1533), 1,
      anon_sym_LPAREN,
    STATE(906), 1,
      sym_parenthesis_group,
    STATE(6262), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300576] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8219), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6263), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300595] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8221), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6264), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300614] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8223), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6265), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300633] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(6907), 1,
      sym_end_block,
    STATE(6266), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300654] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8225), 2,
      sym_END,
      sym_FOR,
    STATE(6267), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300673] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1112), 1,
      sym_BEGIN,
    STATE(6190), 1,
      sym_concurrent_block,
    STATE(6268), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300694] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5527), 1,
      sym_FOR,
    STATE(6234), 1,
      sym_block_configuration,
    STATE(6269), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300715] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8227), 2,
      sym_END,
      sym_FOR,
    STATE(6270), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300734] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8229), 1,
      anon_sym_SEMI,
    ACTIONS(8231), 1,
      sym_IS,
    STATE(6271), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300755] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8233), 1,
      anon_sym_LPAREN,
    STATE(3411), 1,
      sym_association_list,
    STATE(6272), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300776] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2179), 1,
      anon_sym_LPAREN,
    STATE(1826), 1,
      sym_parenthesis_group,
    STATE(6273), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300797] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8129), 1,
      anon_sym_COMMA,
    ACTIONS(8235), 1,
      anon_sym_SEMI,
    STATE(6274), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300818] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6066), 1,
      sym_sequential_block,
    STATE(6275), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300839] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8237), 1,
      sym_IF,
    ACTIONS(8239), 1,
      sym__directive_newline,
    STATE(6276), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300860] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7494), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6277), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300879] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8241), 1,
      anon_sym_LPAREN,
    STATE(4851), 1,
      sym_association_list,
    STATE(6278), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300900] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2179), 1,
      anon_sym_LPAREN,
    STATE(1831), 1,
      sym_parenthesis_group,
    STATE(6279), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300921] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8243), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6280), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300940] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8245), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6281), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300959] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2430), 1,
      anon_sym_LPAREN,
    STATE(2126), 1,
      sym_parenthesis_group,
    STATE(6282), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300980] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8247), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6283), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [300999] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8249), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6284), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301018] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8251), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6285), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301037] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3272), 1,
      sym_WHEN,
    STATE(6162), 1,
      sym_when_element,
    STATE(6286), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301058] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2830), 2,
      sym_UNITS,
      anon_sym_SEMI,
    STATE(6287), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301077] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6995), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6288), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301096] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8253), 1,
      anon_sym_LPAREN,
    STATE(2254), 1,
      sym_association_list,
    STATE(6289), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301117] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8255), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6290), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301136] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8257), 1,
      anon_sym_LPAREN,
    STATE(1645), 1,
      sym_association_list,
    STATE(6291), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301157] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2199), 1,
      anon_sym_LPAREN,
    STATE(1857), 1,
      sym_parenthesis_group,
    STATE(6292), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301178] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8259), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6293), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301197] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8261), 1,
      sym_IS,
    STATE(6310), 1,
      sym_case_body,
    STATE(6294), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301218] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3492), 2,
      sym_WHEN,
      anon_sym_SEMI,
    STATE(6295), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301237] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8031), 1,
      sym_END,
    STATE(7438), 1,
      sym_end_if,
    STATE(6296), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301258] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym_LOOP,
    STATE(6315), 1,
      sym_loop_body,
    STATE(6297), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301279] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8005), 1,
      sym_END,
    STATE(7452), 1,
      sym_end_loop,
    STATE(6298), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301300] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(5845), 1,
      sym_select_target,
    STATE(6299), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301321] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2430), 1,
      anon_sym_LPAREN,
    STATE(2131), 1,
      sym_parenthesis_group,
    STATE(6300), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301342] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8119), 1,
      anon_sym_LPAREN,
    STATE(5795), 1,
      sym_association_list,
    STATE(6301), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301363] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2199), 1,
      anon_sym_LPAREN,
    STATE(1863), 1,
      sym_parenthesis_group,
    STATE(6302), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301384] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(8333), 1,
      sym_end_generate,
    STATE(6303), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301405] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4184), 1,
      anon_sym_LPAREN,
    STATE(3727), 1,
      sym_parenthesis_group,
    STATE(6304), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301426] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1575), 1,
      anon_sym_LPAREN,
    STATE(977), 1,
      sym_parenthesis_group,
    STATE(6305), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301447] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8263), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6306), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301466] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8265), 1,
      sym_IS,
    STATE(5488), 1,
      sym_entity_head,
    STATE(6307), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301487] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5741), 1,
      sym_GENERATE,
    STATE(6546), 1,
      sym_generate_body,
    STATE(6308), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301508] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(117), 1,
      sym_signal_assignment,
    STATE(6309), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301529] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8051), 1,
      sym_END,
    STATE(6947), 1,
      sym_end_case,
    STATE(6310), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301550] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6861), 2,
      sym_WHEN,
      anon_sym_SEMI,
    STATE(6311), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301569] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8267), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6312), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301588] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8269), 1,
      anon_sym_LPAREN,
    STATE(1063), 1,
      sym_association_list,
    STATE(6313), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301609] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5151), 1,
      anon_sym_LPAREN,
    STATE(4668), 1,
      sym_parenthesis_group,
    STATE(6314), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301630] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8005), 1,
      sym_END,
    STATE(6952), 1,
      sym_end_loop,
    STATE(6315), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301651] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3272), 1,
      sym_WHEN,
    STATE(6942), 1,
      sym_when_element,
    STATE(6316), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301672] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(8271), 1,
      anon_sym_SEMI,
    STATE(6317), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301693] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8039), 1,
      sym_END,
    STATE(7852), 1,
      sym_subprogram_end,
    STATE(6318), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301714] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8273), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6319), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301733] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5151), 1,
      anon_sym_LPAREN,
    STATE(4674), 1,
      sym_parenthesis_group,
    STATE(6320), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301754] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5921), 1,
      anon_sym_SEMI,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    STATE(6321), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301775] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8275), 1,
      anon_sym_SEMI,
    ACTIONS(8277), 1,
      sym_IS,
    STATE(6322), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301796] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8279), 1,
      anon_sym_LPAREN,
    STATE(2167), 1,
      sym_association_list,
    STATE(6323), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301817] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8281), 1,
      sym_POSTPONED,
    ACTIONS(8283), 1,
      sym_PROCESS,
    STATE(6324), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301838] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2376), 1,
      anon_sym_LPAREN,
    STATE(2168), 1,
      sym_parenthesis_group,
    STATE(6325), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301859] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4184), 1,
      anon_sym_LPAREN,
    STATE(3774), 1,
      sym_parenthesis_group,
    STATE(6326), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301880] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8285), 1,
      anon_sym_LPAREN,
    STATE(921), 1,
      sym_association_list,
    STATE(6327), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301901] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2588), 2,
      sym_PORT,
      anon_sym_SEMI,
    STATE(6328), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301920] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1931), 1,
      anon_sym_LPAREN,
    STATE(1471), 1,
      sym_parenthesis_group,
    STATE(6329), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301941] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5137), 1,
      anon_sym_LPAREN,
    STATE(4853), 1,
      sym_parenthesis_group,
    STATE(6330), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301962] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8287), 1,
      anon_sym_LPAREN,
    STATE(1575), 1,
      sym_association_list,
    STATE(6331), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [301983] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1919), 1,
      anon_sym_LPAREN,
    STATE(1576), 1,
      sym_parenthesis_group,
    STATE(6332), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302004] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2376), 1,
      anon_sym_LPAREN,
    STATE(2173), 1,
      sym_parenthesis_group,
    STATE(6333), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302025] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1931), 1,
      anon_sym_LPAREN,
    STATE(1474), 1,
      sym_parenthesis_group,
    STATE(6334), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302046] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(8289), 1,
      anon_sym_SEMI,
    STATE(6335), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302067] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2674), 2,
      sym_PORT,
      anon_sym_SEMI,
    STATE(6336), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302086] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8291), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6337), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302105] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6407), 1,
      sym_select_target,
    STATE(6338), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302126] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8293), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6339), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302145] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(6645), 1,
      sym_end_block,
    STATE(6340), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302166] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8295), 1,
      anon_sym_LPAREN,
    STATE(1674), 1,
      sym_association_list,
    STATE(6341), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302187] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8261), 1,
      sym_IS,
    STATE(6116), 1,
      sym_case_body,
    STATE(6342), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302208] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5167), 1,
      anon_sym_LPAREN,
    STATE(4742), 1,
      sym_parenthesis_group,
    STATE(6343), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302229] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8297), 2,
      sym_IS,
      anon_sym_COLON,
    STATE(6344), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302248] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8299), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6345), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302267] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8301), 2,
      sym_directive_body,
      sym__directive_newline,
    STATE(6346), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302286] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7049), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6347), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302305] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5741), 1,
      sym_GENERATE,
    STATE(6998), 1,
      sym_generate_body,
    STATE(6348), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302326] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8303), 1,
      anon_sym_DOT,
    ACTIONS(8305), 1,
      anon_sym_COLON,
    STATE(6349), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302347] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8031), 1,
      sym_END,
    STATE(6987), 1,
      sym_end_if,
    STATE(6350), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302368] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym_LOOP,
    STATE(6118), 1,
      sym_loop_body,
    STATE(6351), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302389] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(8307), 1,
      anon_sym_SEMI,
    STATE(6352), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302410] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8309), 1,
      anon_sym_LPAREN,
    STATE(1787), 1,
      sym_association_list,
    STATE(6353), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302431] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4419), 1,
      anon_sym_LPAREN,
    STATE(4471), 1,
      sym_parenthesis_group,
    STATE(6354), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302452] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(5732), 1,
      sym_select_target,
    STATE(6355), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302473] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7269), 2,
      sym_OF,
      anon_sym_COMMA,
    STATE(6356), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302492] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1919), 1,
      anon_sym_LPAREN,
    STATE(1583), 1,
      sym_parenthesis_group,
    STATE(6357), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302513] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8311), 1,
      anon_sym_LPAREN,
    STATE(2189), 1,
      sym_association_list,
    STATE(6358), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302534] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4419), 1,
      anon_sym_LPAREN,
    STATE(4474), 1,
      sym_parenthesis_group,
    STATE(6359), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302555] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3930), 1,
      anon_sym_LPAREN,
    STATE(3417), 1,
      sym_parenthesis_group,
    STATE(6360), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302576] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8313), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6361), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302595] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(110), 1,
      sym_signal_assignment,
    STATE(6362), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302616] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8315), 1,
      anon_sym_LPAREN,
    STATE(4363), 1,
      sym_association_list,
    STATE(6363), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302637] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8317), 1,
      anon_sym_LPAREN,
    STATE(4821), 1,
      sym_association_list,
    STATE(6364), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302658] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1851), 1,
      anon_sym_LPAREN,
    STATE(1251), 1,
      sym_parenthesis_group,
    STATE(6365), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302679] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3930), 1,
      anon_sym_LPAREN,
    STATE(3421), 1,
      sym_parenthesis_group,
    STATE(6366), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302700] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4234), 1,
      anon_sym_LPAREN,
    STATE(4374), 1,
      sym_parenthesis_group,
    STATE(6367), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302721] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1851), 1,
      anon_sym_LPAREN,
    STATE(1253), 1,
      sym_parenthesis_group,
    STATE(6368), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302742] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8319), 1,
      anon_sym_COLON,
    ACTIONS(8321), 1,
      sym_IS,
    STATE(6369), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302763] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8323), 1,
      anon_sym_LPAREN,
    STATE(2016), 1,
      sym_association_list,
    STATE(6370), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302784] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5239), 1,
      anon_sym_LPAREN,
    STATE(4876), 1,
      sym_parenthesis_group,
    STATE(6371), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302805] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2927), 2,
      sym_END,
      sym_WHEN,
    STATE(6372), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302824] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(104), 1,
      sym_signal_assignment,
    STATE(6373), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302845] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8325), 1,
      anon_sym_COLON,
    ACTIONS(8327), 1,
      sym_IS,
    STATE(6374), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302866] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8329), 1,
      sym_IS,
    STATE(6558), 1,
      sym_package_declaration_body,
    STATE(6375), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302887] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(105), 1,
      sym_signal_assignment,
    STATE(6376), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302908] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6559), 1,
      sym_sequential_block,
    STATE(6377), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302929] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(80), 1,
      sym_signal_assignment,
    STATE(6378), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302950] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6563), 1,
      sym_sequential_block,
    STATE(6379), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302971] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5741), 1,
      sym_GENERATE,
    STATE(6567), 1,
      sym_generate_body,
    STATE(6380), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [302992] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8155), 1,
      sym_IS,
    STATE(6572), 1,
      sym_package_definition_body,
    STATE(6381), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303013] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6574), 1,
      sym_sequential_block,
    STATE(6382), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303034] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(81), 1,
      sym_signal_assignment,
    STATE(6383), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303055] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8261), 1,
      sym_IS,
    STATE(6577), 1,
      sym_case_body,
    STATE(6384), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303076] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym_LOOP,
    STATE(6578), 1,
      sym_loop_body,
    STATE(6385), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303097] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8331), 2,
      sym_END,
      sym_WHEN,
    STATE(6386), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303116] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1112), 1,
      sym_BEGIN,
    STATE(6584), 1,
      sym_concurrent_block,
    STATE(6387), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303137] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(113), 1,
      sym_signal_assignment,
    STATE(6388), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303158] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6585), 1,
      sym_sequential_block,
    STATE(6389), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303179] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6587), 1,
      sym_sequential_block,
    STATE(6390), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303200] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8333), 1,
      anon_sym_LPAREN,
    STATE(1671), 1,
      sym_association_list,
    STATE(6391), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303221] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8261), 1,
      sym_IS,
    STATE(6591), 1,
      sym_case_body,
    STATE(6392), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303242] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym_LOOP,
    STATE(6592), 1,
      sym_loop_body,
    STATE(6393), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303263] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1677), 1,
      anon_sym_LPAREN,
    STATE(1011), 1,
      sym_parenthesis_group,
    STATE(6394), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303284] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8335), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6395), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303303] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8053), 1,
      sym_IS,
    STATE(6593), 1,
      sym_mode_view_body,
    STATE(6396), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303324] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1112), 1,
      sym_BEGIN,
    STATE(6594), 1,
      sym_concurrent_block,
    STATE(6397), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303345] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8337), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6398), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303364] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(85), 1,
      sym_signal_assignment,
    STATE(6399), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303385] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6601), 1,
      sym_sequential_block,
    STATE(6400), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303406] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1677), 1,
      anon_sym_LPAREN,
    STATE(1013), 1,
      sym_parenthesis_group,
    STATE(6401), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303427] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6007), 1,
      sym_sequential_block,
    STATE(6402), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303448] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(102), 1,
      sym_signal_assignment,
    STATE(6403), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303469] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8339), 1,
      anon_sym_COLON,
    ACTIONS(8341), 1,
      sym_IS,
    STATE(6404), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303490] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8343), 1,
      sym_IS,
    STATE(6026), 1,
      sym_package_declaration_body,
    STATE(6405), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303511] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(103), 1,
      sym_signal_assignment,
    STATE(6406), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303532] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(20), 1,
      sym_signal_assignment,
    STATE(6407), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303553] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6027), 1,
      sym_sequential_block,
    STATE(6408), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303574] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(87), 1,
      sym_signal_assignment,
    STATE(6409), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303595] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6037), 1,
      sym_sequential_block,
    STATE(6410), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303616] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5741), 1,
      sym_GENERATE,
    STATE(6042), 1,
      sym_generate_body,
    STATE(6411), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303637] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8155), 1,
      sym_IS,
    STATE(6044), 1,
      sym_package_definition_body,
    STATE(6412), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303658] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8345), 1,
      anon_sym_LPAREN,
    STATE(2204), 1,
      sym_association_list,
    STATE(6413), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303679] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6046), 1,
      sym_sequential_block,
    STATE(6414), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303700] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(88), 1,
      sym_signal_assignment,
    STATE(6415), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303721] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8261), 1,
      sym_IS,
    STATE(6050), 1,
      sym_case_body,
    STATE(6416), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303742] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym_LOOP,
    STATE(6006), 1,
      sym_loop_body,
    STATE(6417), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303763] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2356), 1,
      anon_sym_LPAREN,
    STATE(1759), 1,
      sym_parenthesis_group,
    STATE(6418), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303784] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1112), 1,
      sym_BEGIN,
    STATE(6057), 1,
      sym_concurrent_block,
    STATE(6419), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303805] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(108), 1,
      sym_signal_assignment,
    STATE(6420), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303826] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6059), 1,
      sym_sequential_block,
    STATE(6421), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303847] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6061), 1,
      sym_sequential_block,
    STATE(6422), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303868] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8347), 1,
      anon_sym_LPAREN,
    STATE(4784), 1,
      sym_association_list,
    STATE(6423), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303889] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8261), 1,
      sym_IS,
    STATE(6064), 1,
      sym_case_body,
    STATE(6424), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303910] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(247), 1,
      sym_LOOP,
    STATE(6065), 1,
      sym_loop_body,
    STATE(6425), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303931] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1713), 1,
      anon_sym_LPAREN,
    STATE(1175), 1,
      sym_parenthesis_group,
    STATE(6426), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303952] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8053), 1,
      sym_IS,
    STATE(6068), 1,
      sym_mode_view_body,
    STATE(6427), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303973] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1112), 1,
      sym_BEGIN,
    STATE(6070), 1,
      sym_concurrent_block,
    STATE(6428), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [303994] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7030), 1,
      sym_end_process,
    STATE(6429), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304015] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(92), 1,
      sym_signal_assignment,
    STATE(6430), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304036] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5239), 1,
      anon_sym_LPAREN,
    STATE(4762), 1,
      sym_parenthesis_group,
    STATE(6431), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304057] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6074), 1,
      sym_sequential_block,
    STATE(6432), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304078] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1713), 1,
      anon_sym_LPAREN,
    STATE(1237), 1,
      sym_parenthesis_group,
    STATE(6433), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304099] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6078), 1,
      sym_sequential_block,
    STATE(6434), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304120] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(115), 1,
      sym_signal_assignment,
    STATE(6435), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304141] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8349), 1,
      anon_sym_COLON,
    ACTIONS(8351), 1,
      sym_IS,
    STATE(6436), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304162] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8353), 1,
      sym_IS,
    STATE(6093), 1,
      sym_package_declaration_body,
    STATE(6437), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304183] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(100), 1,
      sym_signal_assignment,
    STATE(6438), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304204] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6094), 1,
      sym_sequential_block,
    STATE(6439), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304225] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(95), 1,
      sym_signal_assignment,
    STATE(6440), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304246] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6101), 1,
      sym_sequential_block,
    STATE(6441), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304267] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2356), 1,
      anon_sym_LPAREN,
    STATE(1761), 1,
      sym_parenthesis_group,
    STATE(6442), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304288] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5741), 1,
      sym_GENERATE,
    STATE(6105), 1,
      sym_generate_body,
    STATE(6443), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304309] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8155), 1,
      sym_IS,
    STATE(6107), 1,
      sym_package_definition_body,
    STATE(6444), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304330] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6109), 1,
      sym_sequential_block,
    STATE(6445), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304351] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(96), 1,
      sym_signal_assignment,
    STATE(6446), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304372] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1112), 1,
      sym_BEGIN,
    STATE(6117), 1,
      sym_concurrent_block,
    STATE(6447), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304393] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8355), 1,
      anon_sym_LPAREN,
    STATE(1108), 1,
      sym_association_list,
    STATE(6448), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304414] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(120), 1,
      sym_signal_assignment,
    STATE(6449), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304435] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6120), 1,
      sym_sequential_block,
    STATE(6450), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304456] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8053), 1,
      sym_IS,
    STATE(6122), 1,
      sym_mode_view_body,
    STATE(6451), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304477] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1112), 1,
      sym_BEGIN,
    STATE(6124), 1,
      sym_concurrent_block,
    STATE(6452), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304498] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8357), 1,
      anon_sym_LPAREN,
    STATE(4497), 1,
      sym_association_list,
    STATE(6453), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304519] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8081), 1,
      anon_sym_LT_EQ,
    STATE(97), 1,
      sym_signal_assignment,
    STATE(6454), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304540] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1863), 1,
      anon_sym_LPAREN,
    STATE(1291), 1,
      sym_parenthesis_group,
    STATE(6455), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304561] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6127), 1,
      sym_sequential_block,
    STATE(6456), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304582] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5087), 1,
      anon_sym_LPAREN,
    STATE(4887), 1,
      sym_parenthesis_group,
    STATE(6457), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304603] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8359), 1,
      anon_sym_COLON,
    ACTIONS(8361), 1,
      sym_IS,
    STATE(6458), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304624] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8363), 1,
      sym_IS,
    STATE(6135), 1,
      sym_package_declaration_body,
    STATE(6459), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304645] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6136), 1,
      sym_sequential_block,
    STATE(6460), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304666] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6139), 1,
      sym_sequential_block,
    STATE(6461), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304687] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8155), 1,
      sym_IS,
    STATE(6142), 1,
      sym_package_definition_body,
    STATE(6462), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304708] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7336), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6463), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304727] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6144), 1,
      sym_sequential_block,
    STATE(6464), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304748] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6150), 1,
      sym_sequential_block,
    STATE(6465), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304769] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6151), 1,
      sym_sequential_block,
    STATE(6466), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304790] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8365), 1,
      anon_sym_COLON,
    ACTIONS(8367), 1,
      sym_IS,
    STATE(6467), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304811] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8369), 1,
      sym_IS,
    STATE(6157), 1,
      sym_package_declaration_body,
    STATE(6468), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304832] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8371), 1,
      anon_sym_COLON,
    ACTIONS(8373), 1,
      sym_IS,
    STATE(6469), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304853] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6378), 1,
      sym_select_target,
    STATE(6470), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304874] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6383), 1,
      sym_select_target,
    STATE(6471), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304895] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1863), 1,
      anon_sym_LPAREN,
    STATE(1293), 1,
      sym_parenthesis_group,
    STATE(6472), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304916] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8375), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6473), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304935] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(5886), 1,
      sym_select_target,
    STATE(6474), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304956] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6399), 1,
      sym_select_target,
    STATE(6475), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304977] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(5890), 1,
      sym_select_target,
    STATE(6476), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [304998] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6521), 2,
      sym_WHEN,
      anon_sym_SEMI,
    STATE(6477), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305017] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8377), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6478), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305036] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    STATE(6912), 1,
      sym_port_map_aspect,
    STATE(6479), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305057] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6409), 1,
      sym_select_target,
    STATE(6480), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305078] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(8379), 1,
      anon_sym_SEMI,
    STATE(6481), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305099] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6415), 1,
      sym_select_target,
    STATE(6482), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305120] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8381), 1,
      anon_sym_LPAREN,
    STATE(1300), 1,
      sym_association_list,
    STATE(6483), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305141] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8383), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6484), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305160] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(5894), 1,
      sym_select_target,
    STATE(6485), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305181] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(5990), 1,
      anon_sym_SEMI,
    STATE(6486), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305202] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8385), 1,
      anon_sym_SEMI,
    ACTIONS(8387), 1,
      sym_IS,
    STATE(6487), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305223] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6430), 1,
      sym_select_target,
    STATE(6488), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305244] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(5898), 1,
      sym_select_target,
    STATE(6489), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305265] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4545), 1,
      anon_sym_LPAREN,
    STATE(4582), 1,
      sym_parenthesis_group,
    STATE(6490), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305286] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6115), 1,
      sym_select_target,
    STATE(6491), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305307] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8389), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6492), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305326] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6440), 1,
      sym_select_target,
    STATE(6493), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305347] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8391), 1,
      anon_sym_LPAREN,
    STATE(3359), 1,
      sym_association_list,
    STATE(6494), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305368] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6446), 1,
      sym_select_target,
    STATE(6495), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305389] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8393), 1,
      sym_ELSE,
    STATE(5455), 1,
      sym_else_expression,
    STATE(6496), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305410] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8203), 1,
      sym_SELECT,
    STATE(6454), 1,
      sym_select_target,
    STATE(6497), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305431] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4545), 1,
      anon_sym_LPAREN,
    STATE(4589), 1,
      sym_parenthesis_group,
    STATE(6498), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305452] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1475), 1,
      anon_sym_LPAREN,
    STATE(860), 1,
      sym_parenthesis_group,
    STATE(6499), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305473] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8395), 1,
      anon_sym_LPAREN,
    STATE(6328), 1,
      sym_association_list,
    STATE(6500), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305494] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2498), 1,
      anon_sym_LPAREN,
    STATE(2185), 1,
      sym_parenthesis_group,
    STATE(6501), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305515] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8397), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(6502), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305534] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8399), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6503), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305553] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(8401), 1,
      anon_sym_SEMI,
    STATE(6504), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305574] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8403), 1,
      anon_sym_LPAREN,
    STATE(1033), 1,
      sym_association_list,
    STATE(6505), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305595] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4409), 1,
      anon_sym_LPAREN,
    STATE(4526), 1,
      sym_parenthesis_group,
    STATE(6506), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305616] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1475), 1,
      anon_sym_LPAREN,
    STATE(864), 1,
      sym_parenthesis_group,
    STATE(6507), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305637] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2440), 2,
      sym_PORT,
      anon_sym_SEMI,
    STATE(6508), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305656] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4409), 1,
      anon_sym_LPAREN,
    STATE(4529), 1,
      sym_parenthesis_group,
    STATE(6509), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305677] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8405), 1,
      sym_END,
    STATE(7043), 1,
      sym_generate_block_end,
    STATE(6510), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305698] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2648), 2,
      sym_GENERIC,
      anon_sym_SEMI,
    STATE(6511), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305717] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8408), 1,
      anon_sym_LPAREN,
    STATE(5819), 1,
      sym_association_list,
    STATE(6512), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305738] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8410), 1,
      anon_sym_LPAREN,
    STATE(3674), 1,
      sym_association_list,
    STATE(6513), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305759] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2269), 1,
      anon_sym_LPAREN,
    STATE(1897), 1,
      sym_parenthesis_group,
    STATE(6514), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305780] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5087), 1,
      anon_sym_LPAREN,
    STATE(4682), 1,
      sym_parenthesis_group,
    STATE(6515), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305801] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8412), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6516), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305820] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8414), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6517), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305839] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8416), 2,
      anon_sym_LT_EQ,
      sym_variable_assignment,
    STATE(6518), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305858] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8418), 1,
      anon_sym_LPAREN,
    STATE(1166), 1,
      sym_association_list,
    STATE(6519), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305879] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8420), 1,
      anon_sym_COLON,
    ACTIONS(8422), 1,
      sym_IS,
    STATE(6520), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305900] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2498), 1,
      anon_sym_LPAREN,
    STATE(2199), 1,
      sym_parenthesis_group,
    STATE(6521), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305921] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7354), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6522), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305940] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8424), 1,
      anon_sym_LPAREN,
    STATE(1790), 1,
      sym_association_list,
    STATE(6523), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305961] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2418), 1,
      anon_sym_LPAREN,
    STATE(2243), 1,
      sym_parenthesis_group,
    STATE(6524), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [305982] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8426), 1,
      anon_sym_LPAREN,
    STATE(1313), 1,
      sym_association_list,
    STATE(6525), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306003] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2269), 1,
      anon_sym_LPAREN,
    STATE(1737), 1,
      sym_parenthesis_group,
    STATE(6526), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306024] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7455), 2,
      sym_END,
      sym_WHEN,
    STATE(6527), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306043] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8428), 1,
      anon_sym_LPAREN,
    STATE(4611), 1,
      sym_association_list,
    STATE(6528), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306064] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2418), 1,
      anon_sym_LPAREN,
    STATE(2245), 1,
      sym_parenthesis_group,
    STATE(6529), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306085] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2636), 2,
      sym_GENERIC,
      anon_sym_SEMI,
    STATE(6530), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306104] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8430), 1,
      anon_sym_LPAREN,
    STATE(4547), 1,
      sym_association_list,
    STATE(6531), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306125] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8432), 1,
      anon_sym_SEMI,
    ACTIONS(8434), 1,
      sym_UNITS,
    STATE(6532), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306146] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2652), 2,
      sym_GENERIC,
      anon_sym_SEMI,
    STATE(6533), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306165] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7409), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6534), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306184] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8436), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6535), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306203] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8438), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6536), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306222] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8440), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6537), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306241] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2906), 1,
      sym_WHEN,
    STATE(6156), 1,
      sym_when_expression,
    STATE(6538), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306262] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7474), 1,
      sym_end_process,
    STATE(6539), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306283] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(8442), 1,
      anon_sym_SEMI,
    STATE(6540), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306304] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8444), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6541), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306323] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8446), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6542), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306342] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8448), 2,
      sym_IS,
      anon_sym_SEMI,
    STATE(6543), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306361] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7503), 1,
      sym_end_process,
    STATE(6544), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306382] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(7519), 1,
      sym_end_generate,
    STATE(6545), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306403] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(6979), 1,
      sym_end_generate,
    STATE(6546), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306424] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8450), 1,
      anon_sym_COLON,
    ACTIONS(8452), 1,
      sym_IS,
    STATE(6547), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306445] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8454), 1,
      anon_sym_LPAREN,
    STATE(871), 1,
      sym_association_list,
    STATE(6548), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306466] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8456), 1,
      anon_sym_COLON,
    ACTIONS(8458), 1,
      sym_IS,
    STATE(6549), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306487] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2614), 2,
      sym_GENERIC,
      anon_sym_SEMI,
    STATE(6550), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306506] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5706), 1,
      sym_END,
    STATE(7531), 1,
      sym_end_component,
    STATE(6551), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306527] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8460), 1,
      anon_sym_COLON,
    ACTIONS(8462), 1,
      sym_IS,
    STATE(6552), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306548] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5409), 1,
      anon_sym_LPAREN,
    STATE(4919), 1,
      sym_parenthesis_group,
    STATE(6553), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306569] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8464), 1,
      anon_sym_COLON,
    ACTIONS(8466), 1,
      sym_IS,
    STATE(6554), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306590] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(8468), 1,
      anon_sym_SEMI,
    STATE(6555), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306611] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8470), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6556), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306630] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8472), 1,
      anon_sym_COLON,
    ACTIONS(8474), 1,
      sym_IS,
    STATE(6557), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306651] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8027), 1,
      sym_END,
    STATE(7551), 1,
      sym_end_package,
    STATE(6558), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306672] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7552), 1,
      sym_end_process,
    STATE(6559), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306693] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8031), 1,
      sym_END,
    STATE(7570), 1,
      sym_end_if,
    STATE(6560), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306714] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8005), 1,
      sym_END,
    STATE(7571), 1,
      sym_end_loop,
    STATE(6561), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306735] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8476), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6562), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306754] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8039), 1,
      sym_END,
    STATE(7596), 1,
      sym_subprogram_end,
    STATE(6563), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306775] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5409), 1,
      anon_sym_LPAREN,
    STATE(4892), 1,
      sym_parenthesis_group,
    STATE(6564), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306796] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(7597), 1,
      sym_end_block,
    STATE(6565), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306817] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(1759), 1,
      sym_PORT,
    STATE(7603), 1,
      sym_port_map_aspect,
    STATE(6566), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306838] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(7605), 1,
      sym_end_generate,
    STATE(6567), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306859] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8303), 1,
      anon_sym_DOT,
    ACTIONS(8478), 1,
      anon_sym_COLON,
    STATE(6568), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306880] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8480), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6569), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306899] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8482), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
    STATE(6570), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306918] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8484), 1,
      anon_sym_SEMI,
    ACTIONS(8486), 1,
      sym_IS,
    STATE(6571), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306939] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8047), 1,
      sym_END,
    STATE(7620), 1,
      sym_end_package_body,
    STATE(6572), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306960] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2932), 2,
      sym_END,
      sym_WHEN,
    STATE(6573), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [306979] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7623), 1,
      sym_end_process,
    STATE(6574), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307000] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8488), 2,
      sym_END,
      sym_WHEN,
    STATE(6575), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307019] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(7630), 1,
      sym_end_block,
    STATE(6576), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307040] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8051), 1,
      sym_END,
    STATE(7635), 1,
      sym_end_case,
    STATE(6577), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307061] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8005), 1,
      sym_END,
    STATE(7636), 1,
      sym_end_loop,
    STATE(6578), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307082] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    STATE(6787), 1,
      sym_end_for,
    STATE(6579), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307103] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8031), 1,
      sym_END,
    STATE(7646), 1,
      sym_end_if,
    STATE(6580), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307124] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5706), 1,
      sym_END,
    STATE(6932), 1,
      sym_end_component,
    STATE(6581), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307145] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8490), 1,
      anon_sym_LPAREN,
    STATE(5795), 1,
      sym_association_list,
    STATE(6582), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307166] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8492), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6583), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307185] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(7666), 1,
      sym_end_block,
    STATE(6584), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307206] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8007), 1,
      sym_END,
    STATE(7719), 1,
      sym_end_process,
    STATE(6585), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307227] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7178), 2,
      sym_END,
      sym_WHEN,
    STATE(6586), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307246] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(7746), 1,
      sym_end_block,
    STATE(6587), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307267] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8494), 1,
      anon_sym_LPAREN,
    STATE(4679), 1,
      sym_association_list,
    STATE(6588), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307288] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8496), 2,
      sym_END,
      sym_FOR,
    STATE(6589), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307307] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    STATE(6806), 1,
      sym_end_for,
    STATE(6590), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307328] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8051), 1,
      sym_END,
    STATE(7799), 1,
      sym_end_case,
    STATE(6591), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307349] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8005), 1,
      sym_END,
    STATE(7800), 1,
      sym_end_loop,
    STATE(6592), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307370] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8071), 1,
      sym_END,
    STATE(7829), 1,
      sym_end_view,
    STATE(6593), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307391] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(7848), 1,
      sym_end_block,
    STATE(6594), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307412] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8017), 1,
      sym_END,
    STATE(7850), 1,
      sym_end_generate,
    STATE(6595), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307433] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8498), 2,
      anon_sym_SEMI,
      anon_sym_RPAREN,
    STATE(6596), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307452] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2944), 1,
      sym_BEGIN,
    STATE(6201), 1,
      sym_sequential_block,
    STATE(6597), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307473] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8500), 2,
      sym_END,
      sym_WHEN,
    STATE(6598), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307492] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8502), 2,
      sym_END,
      sym_FOR,
    STATE(6599), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307511] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2984), 1,
      sym_END,
    STATE(6814), 1,
      sym_end_for,
    STATE(6600), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307532] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8015), 1,
      sym_END,
    STATE(8022), 1,
      sym_end_block,
    STATE(6601), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307553] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8504), 1,
      anon_sym_LPAREN,
    STATE(2276), 1,
      sym_association_list,
    STATE(6602), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307574] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8506), 2,
      sym_END,
      sym_FOR,
    STATE(6603), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307593] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8508), 2,
      sym_based_integer,
      sym_based_float,
    STATE(6604), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307612] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5923), 1,
      anon_sym_LT_EQ,
    ACTIONS(5996), 1,
      anon_sym_SEMI,
    STATE(6605), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307633] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8510), 1,
      anon_sym_SEMI,
    ACTIONS(8512), 1,
      sym_IS,
    STATE(6606), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307654] = 6,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5167), 1,
      anon_sym_LPAREN,
    STATE(4750), 1,
      sym_parenthesis_group,
    STATE(6607), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307675] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8514), 1,
      anon_sym_SEMI,
    STATE(6608), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307693] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8516), 1,
      anon_sym_RPAREN,
    STATE(6609), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307711] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8518), 1,
      anon_sym_SEMI,
    STATE(6610), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307729] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8520), 1,
      anon_sym_RPAREN,
    STATE(6611), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307747] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8522), 1,
      anon_sym_SEMI,
    STATE(6612), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307765] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8524), 1,
      anon_sym_RBRACK,
    STATE(6613), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307783] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8526), 1,
      anon_sym_RPAREN,
    STATE(6614), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307801] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6843), 1,
      sym_END,
    STATE(6615), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307819] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8528), 1,
      anon_sym_GT_GT,
    STATE(6616), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307837] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8530), 1,
      anon_sym_GT_GT,
    STATE(6617), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307855] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8532), 1,
      anon_sym_GT_GT,
    STATE(6618), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307873] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8534), 1,
      anon_sym_RBRACK,
    STATE(6619), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307891] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8536), 1,
      anon_sym_RBRACK,
    STATE(6620), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307909] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8538), 1,
      anon_sym_SEMI,
    STATE(6621), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307927] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8540), 1,
      anon_sym_RBRACK,
    STATE(6622), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307945] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8542), 1,
      anon_sym_LPAREN,
    STATE(6623), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307963] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8544), 1,
      anon_sym_SEMI,
    STATE(6624), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307981] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8546), 1,
      anon_sym_SEMI,
    STATE(6625), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [307999] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8548), 1,
      anon_sym_RBRACK,
    STATE(6626), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308017] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8550), 1,
      anon_sym_RPAREN,
    STATE(6627), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308035] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8552), 1,
      anon_sym_GT_GT,
    STATE(6628), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308053] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8554), 1,
      anon_sym_GT_GT,
    STATE(6629), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308071] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8556), 1,
      anon_sym_RPAREN,
    STATE(6630), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308089] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8558), 1,
      anon_sym_GT_GT,
    STATE(6631), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308107] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8560), 1,
      anon_sym_RBRACK,
    STATE(6632), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308125] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8562), 1,
      anon_sym_RBRACK,
    STATE(6633), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308143] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8564), 1,
      anon_sym_RBRACK,
    STATE(6634), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308161] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8566), 1,
      anon_sym_GT_GT,
    STATE(6635), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308179] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8568), 1,
      anon_sym_GT_GT,
    STATE(6636), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308197] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8570), 1,
      anon_sym_GT_GT,
    STATE(6637), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308215] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8572), 1,
      anon_sym_SEMI,
    STATE(6638), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308233] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8574), 1,
      anon_sym_RBRACK,
    STATE(6639), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308251] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8576), 1,
      anon_sym_RBRACK,
    STATE(6640), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308269] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8578), 1,
      anon_sym_RBRACK,
    STATE(6641), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308287] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8580), 1,
      sym_MAP,
    STATE(6642), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308305] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8582), 1,
      anon_sym_RPAREN,
    STATE(6643), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308323] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8584), 1,
      sym_BLOCK,
    STATE(6644), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308341] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8586), 1,
      anon_sym_SEMI,
    STATE(6645), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308359] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8588), 1,
      anon_sym_RBRACK,
    STATE(6646), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308377] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8590), 1,
      anon_sym_SEMI,
    STATE(6647), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308395] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8592), 1,
      anon_sym_COLON,
    STATE(6648), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308413] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8594), 1,
      anon_sym_RBRACK,
    STATE(6649), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308431] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8596), 1,
      anon_sym_SEMI,
    STATE(6650), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308449] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8598), 1,
      anon_sym_RBRACK,
    STATE(6651), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308467] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8600), 1,
      anon_sym_RBRACK,
    STATE(6652), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308485] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8602), 1,
      anon_sym_RPAREN,
    STATE(6653), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308503] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8604), 1,
      sym_bit_string_value,
    STATE(6654), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308521] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8606), 1,
      anon_sym_SEMI,
    STATE(6655), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308539] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8608), 1,
      anon_sym_RBRACK,
    STATE(6656), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308557] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6064), 1,
      anon_sym_SEMI,
    STATE(6657), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308575] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8610), 1,
      anon_sym_SEMI,
    STATE(6658), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308593] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8612), 1,
      anon_sym_RBRACK,
    STATE(6659), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308611] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2678), 1,
      sym_IN,
    STATE(6660), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308629] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8614), 1,
      anon_sym_RBRACK,
    STATE(6661), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308647] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8616), 1,
      anon_sym_RPAREN,
    STATE(6662), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308665] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8618), 1,
      sym_bit_string_value,
    STATE(6663), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308683] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8620), 1,
      anon_sym_RPAREN,
    STATE(6664), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308701] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8622), 1,
      anon_sym_SEMI,
    STATE(6665), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308719] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8624), 1,
      anon_sym_RPAREN,
    STATE(6666), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308737] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8626), 1,
      anon_sym_GT_GT,
    STATE(6667), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308755] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8628), 1,
      anon_sym_RPAREN,
    STATE(6668), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308773] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8630), 1,
      anon_sym_GT_GT,
    STATE(6669), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308791] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8632), 1,
      anon_sym_RPAREN,
    STATE(6670), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308809] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8634), 1,
      anon_sym_RBRACK,
    STATE(6671), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308827] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8636), 1,
      anon_sym_GT_GT,
    STATE(6672), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308845] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8638), 1,
      sym_bit_string_value,
    STATE(6673), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308863] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8640), 1,
      anon_sym_RPAREN,
    STATE(6674), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308881] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8642), 1,
      anon_sym_RPAREN,
    STATE(6675), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308899] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8644), 1,
      anon_sym_RPAREN,
    STATE(6676), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308917] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8646), 1,
      anon_sym_RPAREN,
    STATE(6677), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308935] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8648), 1,
      anon_sym_RPAREN,
    STATE(6678), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308953] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8650), 1,
      anon_sym_RPAREN,
    STATE(6679), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308971] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8652), 1,
      anon_sym_RPAREN,
    STATE(6680), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [308989] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8654), 1,
      anon_sym_RPAREN,
    STATE(6681), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309007] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8656), 1,
      anon_sym_RPAREN,
    STATE(6682), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309025] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8658), 1,
      anon_sym_RPAREN,
    STATE(6683), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309043] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8660), 1,
      anon_sym_RPAREN,
    STATE(6684), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309061] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8662), 1,
      anon_sym_RPAREN,
    STATE(6685), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309079] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8664), 1,
      anon_sym_RPAREN,
    STATE(6686), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309097] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8666), 1,
      anon_sym_RPAREN,
    STATE(6687), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309115] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8668), 1,
      anon_sym_RPAREN,
    STATE(6688), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309133] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8670), 1,
      anon_sym_RPAREN,
    STATE(6689), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309151] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8672), 1,
      anon_sym_RPAREN,
    STATE(6690), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309169] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8674), 1,
      anon_sym_RPAREN,
    STATE(6691), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309187] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8676), 1,
      anon_sym_RPAREN,
    STATE(6692), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309205] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8678), 1,
      anon_sym_RPAREN,
    STATE(6693), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309223] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8680), 1,
      anon_sym_RPAREN,
    STATE(6694), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309241] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8682), 1,
      anon_sym_RPAREN,
    STATE(6695), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309259] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8684), 1,
      anon_sym_RPAREN,
    STATE(6696), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309277] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8686), 1,
      anon_sym_RPAREN,
    STATE(6697), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309295] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8688), 1,
      anon_sym_RPAREN,
    STATE(6698), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309313] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8690), 1,
      anon_sym_GT_GT,
    STATE(6699), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309331] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8692), 1,
      anon_sym_GT_GT,
    STATE(6700), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309349] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8694), 1,
      anon_sym_GT_GT,
    STATE(6701), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309367] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8696), 1,
      anon_sym_RBRACK,
    STATE(6702), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309385] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8698), 1,
      sym_box,
    STATE(6703), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309403] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8700), 1,
      anon_sym_RBRACK,
    STATE(6704), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309421] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8702), 1,
      anon_sym_SEMI,
    STATE(6705), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309439] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8704), 1,
      sym_bit_string_base,
    STATE(6706), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309457] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8706), 1,
      anon_sym_RBRACK,
    STATE(6707), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309475] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8708), 1,
      anon_sym_SEMI,
    STATE(6708), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309493] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8710), 1,
      anon_sym_COLON,
    STATE(6709), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309511] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8712), 1,
      anon_sym_SEMI,
    STATE(6710), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309529] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8714), 1,
      anon_sym_SEMI,
    STATE(6711), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309547] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8716), 1,
      sym_MAP,
    STATE(6712), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309565] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8718), 1,
      sym_MAP,
    STATE(6713), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309583] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5948), 1,
      sym_END,
    STATE(6714), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309601] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8720), 1,
      anon_sym_SEMI,
    STATE(6715), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309619] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8722), 1,
      sym__directive_newline,
    STATE(6716), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309637] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8171), 1,
      anon_sym_COMMA,
    STATE(6717), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309655] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8724), 1,
      anon_sym_SEMI,
    STATE(6718), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309673] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7253), 1,
      anon_sym_SEMI,
    STATE(6719), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309691] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8726), 1,
      sym_bit_string_value,
    STATE(6720), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309709] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8728), 1,
      anon_sym_LPAREN,
    STATE(6721), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309727] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8730), 1,
      anon_sym_SEMI,
    STATE(6722), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309745] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8732), 1,
      anon_sym_SEMI,
    STATE(6723), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309763] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8734), 1,
      anon_sym_SEMI,
    STATE(6724), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309781] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8736), 1,
      anon_sym_LPAREN,
    STATE(6725), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309799] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8738), 1,
      anon_sym_LPAREN,
    STATE(6726), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309817] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8740), 1,
      anon_sym_SEMI,
    STATE(6727), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309835] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7280), 1,
      anon_sym_SEMI,
    STATE(6728), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309853] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8742), 1,
      anon_sym_SEMI,
    STATE(6729), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309871] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8744), 1,
      sym_bit_string_value,
    STATE(6730), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309889] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8746), 1,
      anon_sym_RPAREN,
    STATE(6731), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309907] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8748), 1,
      anon_sym_SEMI,
    STATE(6732), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309925] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6178), 1,
      anon_sym_SEMI,
    STATE(6733), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309943] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7290), 1,
      anon_sym_SEMI,
    STATE(6734), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309961] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8750), 1,
      anon_sym_RPAREN,
    STATE(6735), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309979] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8752), 1,
      anon_sym_SEMI,
    STATE(6736), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [309997] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8754), 1,
      anon_sym_SEMI,
    STATE(6737), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310015] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8756), 1,
      anon_sym_SEMI,
    STATE(6738), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310033] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8758), 1,
      sym_MAP,
    STATE(6739), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310051] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8760), 1,
      anon_sym_SEMI,
    STATE(6740), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310069] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3750), 1,
      sym_GENERATE,
    STATE(6741), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310087] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8762), 1,
      anon_sym_RPAREN,
    STATE(6742), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310105] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8764), 1,
      anon_sym_SEMI,
    STATE(6743), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310123] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8766), 1,
      sym_IN,
    STATE(6744), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310141] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8768), 1,
      anon_sym_RBRACK,
    STATE(6745), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310159] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8770), 1,
      anon_sym_RBRACK,
    STATE(6746), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310177] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8772), 1,
      anon_sym_SEMI,
    STATE(6747), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310195] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8774), 1,
      anon_sym_GT_GT,
    STATE(6748), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310213] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8776), 1,
      anon_sym_GT_GT,
    STATE(6749), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310231] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8778), 1,
      sym__directive_newline,
    STATE(6750), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310249] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8780), 1,
      anon_sym_GT_GT,
    STATE(6751), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310267] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8782), 1,
      anon_sym_SEMI,
    STATE(6752), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310285] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8784), 1,
      anon_sym_RBRACK,
    STATE(6753), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310303] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8786), 1,
      anon_sym_SEMI,
    STATE(6754), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310321] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8788), 1,
      anon_sym_RBRACK,
    STATE(6755), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310339] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8790), 1,
      sym_END,
    STATE(6756), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310357] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7171), 1,
      anon_sym_SEMI,
    STATE(6757), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310375] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8792), 1,
      anon_sym_SEMI,
    STATE(6758), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310393] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8794), 1,
      anon_sym_SEMI,
    STATE(6759), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310411] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8796), 1,
      sym__block_comment_end,
    STATE(6760), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310429] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8798), 1,
      anon_sym_SEMI,
    STATE(6761), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310447] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8800), 1,
      sym_OF,
    STATE(6762), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310465] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8802), 1,
      anon_sym_SEMI,
    STATE(6763), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310483] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8804), 1,
      anon_sym_SEMI,
    STATE(6764), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310501] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6071), 1,
      sym_END,
    STATE(6765), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310519] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8806), 1,
      anon_sym_COLON,
    STATE(6766), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310537] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8225), 1,
      sym_END,
    STATE(6767), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310555] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8808), 1,
      anon_sym_SEMI,
    STATE(6768), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310573] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8810), 1,
      sym_bit_string_value,
    STATE(6769), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310591] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8812), 1,
      anon_sym_SEMI,
    STATE(6770), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310609] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8814), 1,
      sym_string_literal,
    STATE(6771), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310627] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8816), 1,
      sym_END,
    STATE(6772), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310645] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8818), 1,
      sym_box,
    STATE(6773), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310663] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8750), 1,
      sym_string_literal,
    STATE(6774), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310681] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8820), 1,
      anon_sym_RPAREN,
    STATE(6775), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310699] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8822), 1,
      anon_sym_SEMI,
    STATE(6776), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310717] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8824), 1,
      anon_sym_RPAREN,
    STATE(6777), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310735] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8826), 1,
      anon_sym_RPAREN,
    STATE(6778), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310753] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8828), 1,
      anon_sym_RPAREN,
    STATE(6779), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310771] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8830), 1,
      anon_sym_RPAREN,
    STATE(6780), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310789] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8832), 1,
      anon_sym_SEMI,
    STATE(6781), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310807] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8303), 1,
      anon_sym_DOT,
    STATE(6782), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310825] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8834), 1,
      sym__directive_newline,
    STATE(6783), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310843] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8836), 1,
      anon_sym_SEMI,
    STATE(6784), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310861] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8838), 1,
      sym_bit_string_value,
    STATE(6785), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310879] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8840), 1,
      anon_sym_RPAREN,
    STATE(6786), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310897] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8842), 1,
      anon_sym_SEMI,
    STATE(6787), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310915] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8844), 1,
      sym_bit_string_value,
    STATE(6788), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310933] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8846), 1,
      sym_bit_string_value,
    STATE(6789), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310951] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8848), 1,
      anon_sym_RPAREN,
    STATE(6790), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310969] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8850), 1,
      anon_sym_SEMI,
    STATE(6791), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [310987] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7343), 1,
      sym_RETURN,
    STATE(6792), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311005] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8227), 1,
      sym_END,
    STATE(6793), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311023] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8852), 1,
      sym_OF,
    STATE(6794), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311041] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8854), 1,
      anon_sym_RBRACK,
    STATE(6795), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311059] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8856), 1,
      sym_IS,
    STATE(6796), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311077] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8858), 1,
      anon_sym_RPAREN,
    STATE(6797), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311095] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8860), 1,
      anon_sym_RPAREN,
    STATE(6798), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311113] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8862), 1,
      anon_sym_SEMI,
    STATE(6799), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311131] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8864), 1,
      anon_sym_GT_GT,
    STATE(6800), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311149] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8866), 1,
      anon_sym_GT_GT,
    STATE(6801), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311167] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8868), 1,
      anon_sym_SEMI,
    STATE(6802), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311185] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8870), 1,
      anon_sym_RPAREN,
    STATE(6803), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311203] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8872), 1,
      anon_sym_GT_GT,
    STATE(6804), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311221] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8874), 1,
      anon_sym_SEMI,
    STATE(6805), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311239] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8876), 1,
      anon_sym_SEMI,
    STATE(6806), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311257] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8878), 1,
      anon_sym_RBRACK,
    STATE(6807), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311275] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8880), 1,
      anon_sym_SEMI,
    STATE(6808), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311293] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8882), 1,
      anon_sym_RBRACK,
    STATE(6809), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311311] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8884), 1,
      anon_sym_RPAREN,
    STATE(6810), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311329] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8886), 1,
      anon_sym_RPAREN,
    STATE(6811), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311347] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8888), 1,
      sym_bit_string_value,
    STATE(6812), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311365] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8890), 1,
      sym_BODY,
    STATE(6813), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311383] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8892), 1,
      anon_sym_SEMI,
    STATE(6814), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311401] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8894), 1,
      anon_sym_SEMI,
    STATE(6815), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311419] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8896), 1,
      anon_sym_RPAREN,
    STATE(6816), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311437] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8898), 1,
      sym_box,
    STATE(6817), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311455] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8900), 1,
      anon_sym_SEMI,
    STATE(6818), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311473] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8902), 1,
      anon_sym_RPAREN,
    STATE(6819), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311491] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8904), 1,
      anon_sym_COLON,
    STATE(6820), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311509] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8906), 1,
      anon_sym_COLON,
    STATE(6821), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311527] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8908), 1,
      sym_OF,
    STATE(6822), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311545] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8910), 1,
      sym_bit_string_value,
    STATE(6823), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311563] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8912), 1,
      anon_sym_SEMI,
    STATE(6824), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311581] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8914), 1,
      anon_sym_SEMI,
    STATE(6825), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311599] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8916), 1,
      anon_sym_SEMI,
    STATE(6826), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311617] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8918), 1,
      anon_sym_SEMI,
    STATE(6827), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311635] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8920), 1,
      sym_IS,
    STATE(6828), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311653] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8922), 1,
      anon_sym_SEMI,
    STATE(6829), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311671] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8924), 1,
      anon_sym_COLON,
    STATE(6830), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311689] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8926), 1,
      anon_sym_SEMI,
    STATE(6831), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311707] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8928), 1,
      sym_bit_string_base,
    STATE(6832), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311725] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8073), 1,
      anon_sym_SEMI,
    STATE(6833), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311743] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8930), 1,
      anon_sym_COLON,
    STATE(6834), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311761] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8932), 1,
      sym_bit_string_value,
    STATE(6835), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311779] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8934), 1,
      sym_MAP,
    STATE(6836), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311797] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8936), 1,
      sym_MAP,
    STATE(6837), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311815] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8938), 1,
      sym_IS,
    STATE(6838), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311833] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8940), 1,
      anon_sym_RPAREN,
    STATE(6839), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311851] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8942), 1,
      anon_sym_RPAREN,
    STATE(6840), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311869] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8944), 1,
      anon_sym_SEMI,
    STATE(6841), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311887] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8946), 1,
      anon_sym_RPAREN,
    STATE(6842), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311905] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8948), 1,
      anon_sym_COLON,
    STATE(6843), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311923] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8950), 1,
      anon_sym_RBRACK,
    STATE(6844), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311941] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8952), 1,
      anon_sym_RBRACK,
    STATE(6845), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311959] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8954), 1,
      anon_sym_SEMI,
    STATE(6846), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311977] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8956), 1,
      anon_sym_LPAREN,
    STATE(6847), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [311995] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8958), 1,
      anon_sym_LPAREN,
    STATE(6848), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8960), 1,
      anon_sym_GT_GT,
    STATE(6849), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312031] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8962), 1,
      anon_sym_GT_GT,
    STATE(6850), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312049] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8964), 1,
      anon_sym_GT_GT,
    STATE(6851), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312067] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8966), 1,
      anon_sym_DOT,
    STATE(6852), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312085] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8968), 1,
      anon_sym_RBRACK,
    STATE(6853), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312103] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8970), 1,
      anon_sym_RPAREN,
    STATE(6854), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312121] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8972), 1,
      anon_sym_GT_GT,
    STATE(6855), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312139] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8974), 1,
      anon_sym_RBRACK,
    STATE(6856), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312157] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8976), 1,
      anon_sym_SEMI,
    STATE(6857), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312175] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8978), 1,
      anon_sym_GT_GT,
    STATE(6858), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312193] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8980), 1,
      anon_sym_SEMI,
    STATE(6859), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312211] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8982), 1,
      anon_sym_GT_GT,
    STATE(6860), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312229] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8984), 1,
      sym_MAP,
    STATE(6861), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312247] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8986), 1,
      anon_sym_SEMI,
    STATE(6862), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312265] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8988), 1,
      anon_sym_RBRACK,
    STATE(6863), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312283] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8990), 1,
      sym_bit_string_value,
    STATE(6864), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312301] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8992), 1,
      sym_UNITS,
    STATE(6865), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312319] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8994), 1,
      anon_sym_EQ,
    STATE(6866), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312337] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7284), 1,
      anon_sym_SEMI,
    STATE(6867), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312355] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8996), 1,
      sym_RETURN,
    STATE(6868), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312373] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8998), 1,
      anon_sym_SEMI,
    STATE(6869), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312391] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9000), 1,
      sym_bit_string_base,
    STATE(6870), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312409] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9002), 1,
      sym_bit_string_value,
    STATE(6871), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312427] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9004), 1,
      anon_sym_RPAREN,
    STATE(6872), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312445] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9006), 1,
      anon_sym_SEMI,
    STATE(6873), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312463] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9008), 1,
      anon_sym_SEMI,
    STATE(6874), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312481] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9010), 1,
      anon_sym_SEMI,
    STATE(6875), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312499] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9012), 1,
      anon_sym_SEMI,
    STATE(6876), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312517] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9014), 1,
      anon_sym_RPAREN,
    STATE(6877), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312535] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9016), 1,
      anon_sym_RBRACK,
    STATE(6878), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312553] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9018), 1,
      anon_sym_RBRACK,
    STATE(6879), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312571] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9020), 1,
      anon_sym_RPAREN,
    STATE(6880), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312589] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9022), 1,
      anon_sym_SEMI,
    STATE(6881), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312607] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9024), 1,
      anon_sym_SEMI,
    STATE(6882), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312625] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9026), 1,
      anon_sym_GT_GT,
    STATE(6883), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312643] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9028), 1,
      anon_sym_GT_GT,
    STATE(6884), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312661] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9030), 1,
      anon_sym_SEMI,
    STATE(6885), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312679] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2830), 1,
      sym_LOOP,
    STATE(6886), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312697] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9032), 1,
      sym_bit_string_value,
    STATE(6887), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312715] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9034), 1,
      anon_sym_GT_GT,
    STATE(6888), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312733] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9036), 1,
      anon_sym_SEMI,
    STATE(6889), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312751] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9038), 1,
      anon_sym_SEMI,
    STATE(6890), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312769] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9040), 1,
      anon_sym_SEMI,
    STATE(6891), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312787] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9042), 1,
      anon_sym_RBRACK,
    STATE(6892), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312805] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9044), 1,
      anon_sym_SEMI,
    STATE(6893), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312823] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7233), 1,
      anon_sym_COLON,
    STATE(6894), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312841] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9046), 1,
      anon_sym_SEMI,
    STATE(6895), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312859] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9048), 1,
      anon_sym_RPAREN,
    STATE(6896), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312877] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9050), 1,
      anon_sym_RPAREN,
    STATE(6897), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312895] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9052), 1,
      anon_sym_RPAREN,
    STATE(6898), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312913] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9054), 1,
      anon_sym_RBRACK,
    STATE(6899), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312931] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9056), 1,
      anon_sym_SEMI,
    STATE(6900), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312949] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4230), 1,
      sym_END,
    STATE(6901), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312967] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9058), 1,
      sym_IS,
    STATE(6902), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [312985] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9060), 1,
      anon_sym_SEMI,
    STATE(6903), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313003] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9062), 1,
      anon_sym_COLON,
    STATE(6904), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313021] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9064), 1,
      sym_bit_string_value,
    STATE(6905), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313039] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9066), 1,
      anon_sym_SEMI,
    STATE(6906), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313057] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9068), 1,
      anon_sym_SEMI,
    STATE(6907), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313075] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9070), 1,
      anon_sym_SEMI,
    STATE(6908), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313093] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9072), 1,
      anon_sym_COLON,
    STATE(6909), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313111] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9074), 1,
      sym_bit_string_value,
    STATE(6910), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313129] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9076), 1,
      anon_sym_SEMI,
    STATE(6911), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313147] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9078), 1,
      anon_sym_SEMI,
    STATE(6912), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313165] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9080), 1,
      anon_sym_SEMI,
    STATE(6913), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313183] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9082), 1,
      anon_sym_COLON,
    STATE(6914), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313201] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9084), 1,
      sym_bit_string_value,
    STATE(6915), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313219] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9086), 1,
      anon_sym_SEMI,
    STATE(6916), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313237] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7286), 1,
      anon_sym_SEMI,
    STATE(6917), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313255] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9088), 1,
      anon_sym_RPAREN,
    STATE(6918), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313273] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9090), 1,
      anon_sym_RPAREN,
    STATE(6919), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313291] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9092), 1,
      anon_sym_SEMI,
    STATE(6920), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313309] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7245), 1,
      sym_END,
    STATE(6921), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313327] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7288), 1,
      anon_sym_SEMI,
    STATE(6922), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313345] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9094), 1,
      sym_bit_string_value,
    STATE(6923), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313363] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9096), 1,
      anon_sym_RPAREN,
    STATE(6924), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313381] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6227), 1,
      anon_sym_SEMI,
    STATE(6925), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313399] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9098), 1,
      anon_sym_RPAREN,
    STATE(6926), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313417] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9100), 1,
      anon_sym_RBRACK,
    STATE(6927), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313435] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9102), 1,
      anon_sym_SEMI,
    STATE(6928), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313453] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9104), 1,
      anon_sym_RBRACK,
    STATE(6929), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313471] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9106), 1,
      anon_sym_GT_GT,
    STATE(6930), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313489] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9108), 1,
      anon_sym_GT_GT,
    STATE(6931), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313507] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9110), 1,
      anon_sym_SEMI,
    STATE(6932), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313525] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9112), 1,
      anon_sym_GT_GT,
    STATE(6933), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313543] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9114), 1,
      anon_sym_SEMI,
    STATE(6934), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313561] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6257), 1,
      anon_sym_SEMI,
    STATE(6935), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313579] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9116), 1,
      anon_sym_RBRACK,
    STATE(6936), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313597] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2610), 1,
      sym_OF,
    STATE(6937), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313615] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9118), 1,
      anon_sym_RPAREN,
    STATE(6938), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313633] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9120), 1,
      anon_sym_RBRACK,
    STATE(6939), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313651] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9122), 1,
      anon_sym_EQ_GT,
    STATE(6940), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313669] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9124), 1,
      sym_bit_string_base,
    STATE(6941), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313687] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8129), 1,
      anon_sym_COMMA,
    STATE(6942), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313705] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9126), 1,
      anon_sym_SEMI,
    STATE(6943), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313723] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9128), 1,
      sym_MAP,
    STATE(6944), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313741] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9130), 1,
      sym_MAP,
    STATE(6945), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313759] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9132), 1,
      sym_CASE,
    STATE(6946), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313777] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9134), 1,
      anon_sym_SEMI,
    STATE(6947), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313795] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9136), 1,
      sym_OF,
    STATE(6948), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313813] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9138), 1,
      sym_LOOP,
    STATE(6949), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313831] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9140), 1,
      sym_BEGIN,
    STATE(6950), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313849] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9142), 1,
      anon_sym_RBRACK,
    STATE(6951), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313867] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9144), 1,
      anon_sym_SEMI,
    STATE(6952), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313885] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9146), 1,
      sym_bit_string_value,
    STATE(6953), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313903] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9148), 1,
      anon_sym_SEMI,
    STATE(6954), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313921] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2818), 1,
      sym_END,
    STATE(6955), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313939] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9150), 1,
      anon_sym_SEMI,
    STATE(6956), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313957] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9152), 1,
      anon_sym_GT_GT,
    STATE(6957), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313975] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9154), 1,
      sym_bit_string_value,
    STATE(6958), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [313993] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9156), 1,
      anon_sym_SEMI,
    STATE(6959), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314011] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9158), 1,
      anon_sym_RPAREN,
    STATE(6960), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314029] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9160), 1,
      anon_sym_GT_GT,
    STATE(6961), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314047] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9162), 1,
      sym_MAP,
    STATE(6962), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314065] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9164), 1,
      anon_sym_GT_GT,
    STATE(6963), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314083] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9166), 1,
      anon_sym_RPAREN,
    STATE(6964), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314101] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9168), 1,
      anon_sym_SEMI,
    STATE(6965), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314119] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9170), 1,
      anon_sym_SEMI,
    STATE(6966), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314137] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9172), 1,
      anon_sym_RBRACK,
    STATE(6967), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314155] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9174), 1,
      anon_sym_RBRACK,
    STATE(6968), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314173] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9176), 1,
      anon_sym_SEMI,
    STATE(6969), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314191] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9178), 1,
      anon_sym_SEMI,
    STATE(6970), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314209] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9180), 1,
      anon_sym_GT_GT,
    STATE(6971), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314227] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9182), 1,
      anon_sym_GT_GT,
    STATE(6972), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314245] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9184), 1,
      anon_sym_GT_GT,
    STATE(6973), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314263] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9186), 1,
      anon_sym_RBRACK,
    STATE(6974), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314281] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9188), 1,
      anon_sym_RBRACK,
    STATE(6975), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314299] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8147), 1,
      anon_sym_SEMI,
    STATE(6976), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314317] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9190), 1,
      sym__directive_newline,
    STATE(6977), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314335] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9192), 1,
      anon_sym_RBRACK,
    STATE(6978), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314353] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9194), 1,
      anon_sym_SEMI,
    STATE(6979), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314371] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9196), 1,
      anon_sym_SEMI,
    STATE(6980), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314389] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9198), 1,
      ts_builtin_sym_end,
    STATE(6981), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314407] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8221), 1,
      anon_sym_SEMI,
    STATE(6982), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314425] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9200), 1,
      anon_sym_SEMI,
    STATE(6983), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314443] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9202), 1,
      anon_sym_SEMI,
    STATE(6984), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314461] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9204), 1,
      anon_sym_SEMI,
    STATE(6985), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314479] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9206), 1,
      anon_sym_RPAREN,
    STATE(6986), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314497] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9208), 1,
      anon_sym_SEMI,
    STATE(6987), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314515] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9210), 1,
      anon_sym_SEMI,
    STATE(6988), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314533] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9212), 1,
      anon_sym_SEMI,
    STATE(6989), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314551] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9214), 1,
      anon_sym_RPAREN,
    STATE(6990), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314569] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9216), 1,
      anon_sym_RPAREN,
    STATE(6991), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314587] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9218), 1,
      sym_bit_string_value,
    STATE(6992), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314605] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9220), 1,
      sym_AFTER,
    STATE(6993), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314623] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9222), 1,
      anon_sym_RPAREN,
    STATE(6994), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314641] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9224), 1,
      sym_MAP,
    STATE(6995), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314659] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7111), 1,
      anon_sym_SEMI,
    STATE(6996), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314677] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9226), 1,
      anon_sym_RPAREN,
    STATE(6997), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314695] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9228), 1,
      sym_END,
    STATE(6998), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314713] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9230), 1,
      sym_bit_string_value,
    STATE(6999), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314731] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6010), 1,
      anon_sym_SEMI,
    STATE(7000), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314749] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9232), 1,
      anon_sym_RPAREN,
    STATE(7001), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314767] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9234), 1,
      anon_sym_RPAREN,
    STATE(7002), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314785] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9236), 1,
      sym_comment_content,
    STATE(7003), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314803] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9238), 1,
      sym_bit_string_value,
    STATE(7004), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314821] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9240), 1,
      anon_sym_SEMI,
    STATE(7005), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314839] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9242), 1,
      anon_sym_RPAREN,
    STATE(7006), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314857] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7308), 1,
      anon_sym_SEMI,
    STATE(7007), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314875] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9244), 1,
      anon_sym_SEMI,
    STATE(7008), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314893] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7310), 1,
      anon_sym_SEMI,
    STATE(7009), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314911] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9246), 1,
      sym_bit_string_base,
    STATE(7010), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314929] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9248), 1,
      anon_sym_RBRACK,
    STATE(7011), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314947] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9250), 1,
      anon_sym_SEMI,
    STATE(7012), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314965] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9252), 1,
      sym_MAP,
    STATE(7013), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [314983] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9254), 1,
      sym_MAP,
    STATE(7014), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315001] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9256), 1,
      sym_MAP,
    STATE(7015), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315019] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9258), 1,
      sym_OF,
    STATE(7016), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315037] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9260), 1,
      anon_sym_SEMI,
    STATE(7017), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315055] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9262), 1,
      anon_sym_GT_GT,
    STATE(7018), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315073] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9264), 1,
      anon_sym_GT_GT,
    STATE(7019), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315091] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9266), 1,
      anon_sym_GT_GT,
    STATE(7020), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315109] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9268), 1,
      anon_sym_SEMI,
    STATE(7021), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315127] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9270), 1,
      anon_sym_SEMI,
    STATE(7022), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315145] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9272), 1,
      anon_sym_RBRACK,
    STATE(7023), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315163] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9274), 1,
      sym_PROTECTED,
    STATE(7024), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315181] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9276), 1,
      sym_MAP,
    STATE(7025), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315199] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9278), 1,
      sym_bit_string_value,
    STATE(7026), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315217] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9280), 1,
      anon_sym_RBRACK,
    STATE(7027), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315235] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9282), 1,
      anon_sym_RPAREN,
    STATE(7028), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315253] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9284), 1,
      sym_VARIABLE,
    STATE(7029), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315271] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9286), 1,
      anon_sym_SEMI,
    STATE(7030), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315289] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9288), 1,
      anon_sym_SEMI,
    STATE(7031), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315307] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6124), 1,
      anon_sym_SEMI,
    STATE(7032), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315325] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6126), 1,
      anon_sym_SEMI,
    STATE(7033), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315343] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9290), 1,
      anon_sym_SEMI,
    STATE(7034), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315361] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9292), 1,
      sym_bit_string_value,
    STATE(7035), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315379] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9294), 1,
      anon_sym_SEMI,
    STATE(7036), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315397] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7455), 1,
      sym_END,
    STATE(7037), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315415] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9296), 1,
      sym_RECORD,
    STATE(7038), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315433] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9298), 1,
      anon_sym_COLON,
    STATE(7039), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315451] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9300), 1,
      anon_sym_SEMI,
    STATE(7040), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315469] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9302), 1,
      anon_sym_RPAREN,
    STATE(7041), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315487] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9304), 1,
      anon_sym_RPAREN,
    STATE(7042), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315505] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7457), 1,
      sym_END,
    STATE(7043), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315523] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9306), 1,
      sym_MAP,
    STATE(7044), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315541] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9308), 1,
      anon_sym_RPAREN,
    STATE(7045), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315559] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9310), 1,
      anon_sym_RPAREN,
    STATE(7046), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315577] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9312), 1,
      anon_sym_LPAREN,
    STATE(7047), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315595] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9314), 1,
      sym_bit_string_value,
    STATE(7048), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315613] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9316), 1,
      sym_bit_string_base,
    STATE(7049), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315631] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9318), 1,
      anon_sym_RPAREN,
    STATE(7050), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315649] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9320), 1,
      sym_MAP,
    STATE(7051), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315667] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9322), 1,
      sym_MAP,
    STATE(7052), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315685] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9324), 1,
      anon_sym_GT_GT,
    STATE(7053), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315703] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8021), 1,
      anon_sym_SEMI,
    STATE(7054), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315721] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9326), 1,
      anon_sym_RPAREN,
    STATE(7055), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315739] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9328), 1,
      anon_sym_RBRACK,
    STATE(7056), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315757] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9330), 1,
      sym_NEW,
    STATE(7057), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315775] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7201), 1,
      sym_IS,
    STATE(7058), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315793] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8035), 1,
      anon_sym_SEMI,
    STATE(7059), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315811] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9332), 1,
      anon_sym_LPAREN,
    STATE(7060), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315829] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9334), 1,
      sym_MAP,
    STATE(7061), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315847] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6098), 1,
      anon_sym_SEMI,
    STATE(7062), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315865] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9336), 1,
      anon_sym_RBRACK,
    STATE(7063), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315883] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6038), 1,
      anon_sym_SEMI,
    STATE(7064), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315901] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9338), 1,
      anon_sym_GT_GT,
    STATE(7065), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315919] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9340), 1,
      sym_FOR,
    STATE(7066), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315937] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5589), 1,
      sym_VUNIT,
    STATE(7067), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315955] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9342), 1,
      anon_sym_SEMI,
    STATE(7068), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315973] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9344), 1,
      sym_IS,
    STATE(7069), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [315991] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9346), 1,
      anon_sym_GT_GT,
    STATE(7070), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316009] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6819), 1,
      anon_sym_GT_GT,
    STATE(7071), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316027] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9348), 1,
      anon_sym_GT_GT,
    STATE(7072), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316045] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9350), 1,
      anon_sym_RPAREN,
    STATE(7073), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316063] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9352), 1,
      anon_sym_GT_GT,
    STATE(7074), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316081] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9354), 1,
      anon_sym_RPAREN,
    STATE(7075), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316099] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9356), 1,
      anon_sym_GT_GT,
    STATE(7076), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316117] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9358), 1,
      sym_bit_string_base,
    STATE(7077), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316135] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9360), 1,
      anon_sym_GT_GT,
    STATE(7078), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316153] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9362), 1,
      sym_MAP,
    STATE(7079), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316171] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9364), 1,
      sym_MAP,
    STATE(7080), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316189] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9366), 1,
      anon_sym_SEMI,
    STATE(7081), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316207] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9368), 1,
      anon_sym_GT_GT,
    STATE(7082), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316225] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9370), 1,
      anon_sym_GT_GT,
    STATE(7083), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316243] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9372), 1,
      sym_MAP,
    STATE(7084), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316261] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9374), 1,
      anon_sym_GT_GT,
    STATE(7085), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316279] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9376), 1,
      anon_sym_RBRACK,
    STATE(7086), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316297] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6036), 1,
      sym_IS,
    STATE(7087), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316315] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9378), 1,
      anon_sym_GT_GT,
    STATE(7088), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316333] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9380), 1,
      anon_sym_RBRACK,
    STATE(7089), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316351] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9382), 1,
      anon_sym_RPAREN,
    STATE(7090), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316369] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9384), 1,
      anon_sym_SEMI,
    STATE(7091), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316387] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9386), 1,
      anon_sym_RPAREN,
    STATE(7092), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316405] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9388), 1,
      sym_comment_content,
    STATE(7093), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316423] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9390), 1,
      anon_sym_RBRACK,
    STATE(7094), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316441] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9392), 1,
      anon_sym_GT_GT,
    STATE(7095), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316459] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9394), 1,
      anon_sym_RPAREN,
    STATE(7096), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316477] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9396), 1,
      sym_FUNCTION,
    STATE(7097), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316495] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9398), 1,
      sym_bit_string_base,
    STATE(7098), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316513] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9400), 1,
      anon_sym_SEMI,
    STATE(7099), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316531] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9402), 1,
      sym_MAP,
    STATE(7100), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316549] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9404), 1,
      sym_MAP,
    STATE(7101), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316567] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9406), 1,
      sym_LOOP,
    STATE(7102), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316585] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9408), 1,
      sym_MAP,
    STATE(7103), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316603] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9410), 1,
      anon_sym_RBRACK,
    STATE(7104), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316621] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9412), 1,
      sym_bit_string_value,
    STATE(7105), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316639] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9414), 1,
      anon_sym_RBRACK,
    STATE(7106), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316657] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9416), 1,
      anon_sym_RBRACK,
    STATE(7107), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316675] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9418), 1,
      anon_sym_SEMI,
    STATE(7108), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316693] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6821), 1,
      sym_RETURN,
    STATE(7109), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316711] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9420), 1,
      anon_sym_SEMI,
    STATE(7110), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316729] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9422), 1,
      anon_sym_SEMI,
    STATE(7111), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316747] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9424), 1,
      sym_bit_string_base,
    STATE(7112), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316765] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9426), 1,
      anon_sym_LPAREN,
    STATE(7113), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316783] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9428), 1,
      sym_MAP,
    STATE(7114), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316801] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9430), 1,
      sym_MAP,
    STATE(7115), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316819] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9432), 1,
      ts_builtin_sym_end,
    STATE(7116), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316837] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9434), 1,
      sym_MAP,
    STATE(7117), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316855] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9436), 1,
      sym_bit_string_value,
    STATE(7118), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316873] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9438), 1,
      sym_FUNCTION,
    STATE(7119), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316891] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9440), 1,
      anon_sym_RPAREN,
    STATE(7120), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316909] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9442), 1,
      anon_sym_SEMI,
    STATE(7121), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316927] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9444), 1,
      anon_sym_SEMI,
    STATE(7122), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316945] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9446), 1,
      anon_sym_RPAREN,
    STATE(7123), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316963] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9448), 1,
      anon_sym_SEMI,
    STATE(7124), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316981] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9450), 1,
      sym_bit_string_base,
    STATE(7125), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [316999] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9452), 1,
      anon_sym_RBRACK,
    STATE(7126), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317017] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9454), 1,
      sym_MAP,
    STATE(7127), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317035] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9456), 1,
      sym_MAP,
    STATE(7128), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317053] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9458), 1,
      anon_sym_GT_GT,
    STATE(7129), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317071] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9460), 1,
      sym_MAP,
    STATE(7130), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317089] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8219), 1,
      anon_sym_SEMI,
    STATE(7131), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317107] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5599), 1,
      anon_sym_SEMI,
    STATE(7132), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317125] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9462), 1,
      anon_sym_SEMI,
    STATE(7133), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317143] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9464), 1,
      sym__directive_newline,
    STATE(7134), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317161] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6984), 1,
      anon_sym_SEMI,
    STATE(7135), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317179] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8259), 1,
      anon_sym_SEMI,
    STATE(7136), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317197] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9466), 1,
      sym_bit_string_base,
    STATE(7137), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317215] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9468), 1,
      anon_sym_SEMI,
    STATE(7138), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317233] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9470), 1,
      sym_MAP,
    STATE(7139), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317251] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9472), 1,
      sym_MAP,
    STATE(7140), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317269] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9474), 1,
      anon_sym_GT_GT,
    STATE(7141), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317287] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9476), 1,
      sym_MAP,
    STATE(7142), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317305] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9478), 1,
      anon_sym_GT_GT,
    STATE(7143), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317323] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2584), 1,
      sym_OF,
    STATE(7144), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317341] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7292), 1,
      sym_RETURN,
    STATE(7145), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317359] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9480), 1,
      anon_sym_SEMI,
    STATE(7146), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317377] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9482), 1,
      anon_sym_GT_GT,
    STATE(7147), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317395] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9484), 1,
      sym_bit_string_value,
    STATE(7148), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317413] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9486), 1,
      sym_bit_string_base,
    STATE(7149), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317431] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3872), 1,
      anon_sym_RPAREN,
    STATE(7150), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317449] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9488), 1,
      sym_MAP,
    STATE(7151), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317467] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9490), 1,
      sym_MAP,
    STATE(7152), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317485] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9492), 1,
      anon_sym_RBRACK,
    STATE(7153), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317503] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9494), 1,
      sym_MAP,
    STATE(7154), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317521] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9496), 1,
      anon_sym_SEMI,
    STATE(7155), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317539] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9498), 1,
      anon_sym_RBRACK,
    STATE(7156), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317557] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9500), 1,
      anon_sym_COLON,
    STATE(7157), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317575] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9502), 1,
      anon_sym_RPAREN,
    STATE(7158), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317593] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9406), 1,
      sym_GENERATE,
    STATE(7159), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317611] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9504), 1,
      anon_sym_SEMI,
    STATE(7160), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317629] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9506), 1,
      sym_bit_string_base,
    STATE(7161), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317647] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9508), 1,
      sym_bit_string_value,
    STATE(7162), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317665] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9510), 1,
      sym_MAP,
    STATE(7163), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317683] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9512), 1,
      sym_MAP,
    STATE(7164), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317701] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3752), 1,
      anon_sym_SEMI,
    STATE(7165), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317719] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9514), 1,
      sym_MAP,
    STATE(7166), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317737] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5978), 1,
      sym_END,
    STATE(7167), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317755] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9516), 1,
      sym_END,
    STATE(7168), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317773] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9518), 1,
      anon_sym_COLON,
    STATE(7169), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317791] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2830), 1,
      sym_IS,
    STATE(7170), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317809] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9520), 1,
      anon_sym_COLON,
    STATE(7171), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317827] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9522), 1,
      sym_RETURN,
    STATE(7172), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317845] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9524), 1,
      sym_bit_string_base,
    STATE(7173), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317863] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9526), 1,
      sym_RETURN,
    STATE(7174), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317881] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9528), 1,
      sym_MAP,
    STATE(7175), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317899] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9530), 1,
      sym_MAP,
    STATE(7176), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317917] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9532), 1,
      anon_sym_SEMI,
    STATE(7177), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317935] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9534), 1,
      sym_MAP,
    STATE(7178), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317953] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9536), 1,
      anon_sym_SEMI,
    STATE(7179), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317971] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9538), 1,
      sym_bit_string_value,
    STATE(7180), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [317989] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9540), 1,
      anon_sym_LPAREN,
    STATE(7181), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318007] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9542), 1,
      anon_sym_RPAREN,
    STATE(7182), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318025] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2823), 1,
      sym_END,
    STATE(7183), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318043] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9544), 1,
      anon_sym_COLON,
    STATE(7184), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318061] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9546), 1,
      sym_bit_string_base,
    STATE(7185), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318079] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9548), 1,
      anon_sym_SEMI,
    STATE(7186), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318097] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9550), 1,
      sym_MAP,
    STATE(7187), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318115] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9552), 1,
      sym_MAP,
    STATE(7188), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318133] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6841), 1,
      sym_RETURN,
    STATE(7189), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318151] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9554), 1,
      sym_MAP,
    STATE(7190), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318169] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9556), 1,
      sym_bit_string_value,
    STATE(7191), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318187] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9558), 1,
      anon_sym_RPAREN,
    STATE(7192), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318205] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9560), 1,
      anon_sym_RPAREN,
    STATE(7193), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318223] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9562), 1,
      anon_sym_RBRACK,
    STATE(7194), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318241] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9564), 1,
      anon_sym_GT_GT,
    STATE(7195), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318259] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6999), 1,
      anon_sym_SEMI,
    STATE(7196), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318277] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9566), 1,
      sym_bit_string_base,
    STATE(7197), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318295] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9568), 1,
      anon_sym_GT_GT,
    STATE(7198), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318313] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9570), 1,
      sym_MAP,
    STATE(7199), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318331] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9572), 1,
      sym_MAP,
    STATE(7200), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318349] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9574), 1,
      anon_sym_GT_GT,
    STATE(7201), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318367] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9576), 1,
      sym_MAP,
    STATE(7202), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318385] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9578), 1,
      anon_sym_GT_GT,
    STATE(7203), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318403] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9580), 1,
      anon_sym_RPAREN,
    STATE(7204), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318421] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9582), 1,
      anon_sym_SEMI,
    STATE(7205), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318439] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6193), 1,
      sym_IS,
    STATE(7206), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318457] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9584), 1,
      anon_sym_RBRACK,
    STATE(7207), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318475] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7026), 1,
      anon_sym_SEMI,
    STATE(7208), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318493] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9586), 1,
      sym_bit_string_base,
    STATE(7209), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318511] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9588), 1,
      anon_sym_RBRACK,
    STATE(7210), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318529] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9590), 1,
      sym_MAP,
    STATE(7211), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318547] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9592), 1,
      sym_MAP,
    STATE(7212), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318565] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6087), 1,
      sym_IS,
    STATE(7213), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318583] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9594), 1,
      sym_MAP,
    STATE(7214), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318601] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9596), 1,
      anon_sym_RBRACK,
    STATE(7215), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318619] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8059), 1,
      anon_sym_SEMI,
    STATE(7216), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318637] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9598), 1,
      sym_bit_string_value,
    STATE(7217), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318655] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9600), 1,
      anon_sym_SEMI,
    STATE(7218), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318673] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8103), 1,
      anon_sym_SEMI,
    STATE(7219), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318691] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9602), 1,
      sym_IS,
    STATE(7220), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318709] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9604), 1,
      sym_bit_string_base,
    STATE(7221), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318727] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9606), 1,
      anon_sym_SEMI,
    STATE(7222), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318745] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9608), 1,
      sym_MAP,
    STATE(7223), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318763] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9610), 1,
      sym_MAP,
    STATE(7224), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318781] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9612), 1,
      anon_sym_GT_GT,
    STATE(7225), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318799] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9614), 1,
      sym_MAP,
    STATE(7226), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318817] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9616), 1,
      sym_IS,
    STATE(7227), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318835] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9618), 1,
      anon_sym_SEMI,
    STATE(7228), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318853] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9620), 1,
      sym_bit_string_value,
    STATE(7229), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318871] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9622), 1,
      anon_sym_RPAREN,
    STATE(7230), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318889] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9624), 1,
      anon_sym_SEMI,
    STATE(7231), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318907] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9626), 1,
      sym_OF,
    STATE(7232), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318925] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9628), 1,
      sym_bit_string_base,
    STATE(7233), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318943] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9630), 1,
      anon_sym_GT_GT,
    STATE(7234), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318961] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9632), 1,
      sym_MAP,
    STATE(7235), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318979] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9634), 1,
      sym_MAP,
    STATE(7236), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [318997] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9636), 1,
      anon_sym_GT_GT,
    STATE(7237), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319015] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9638), 1,
      sym_MAP,
    STATE(7238), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319033] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(4244), 1,
      sym_END,
    STATE(7239), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319051] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9640), 1,
      anon_sym_SEMI,
    STATE(7240), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319069] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9642), 1,
      anon_sym_SEMI,
    STATE(7241), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319087] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9644), 1,
      sym_OF,
    STATE(7242), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319105] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9646), 1,
      anon_sym_RPAREN,
    STATE(7243), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319123] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9648), 1,
      anon_sym_GT_GT,
    STATE(7244), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319141] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9650), 1,
      sym_bit_string_base,
    STATE(7245), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319159] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9652), 1,
      anon_sym_SEMI,
    STATE(7246), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319177] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9654), 1,
      sym_MAP,
    STATE(7247), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319195] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9656), 1,
      sym_MAP,
    STATE(7248), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319213] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9658), 1,
      anon_sym_RBRACK,
    STATE(7249), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319231] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9660), 1,
      sym_MAP,
    STATE(7250), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319249] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9662), 1,
      anon_sym_RBRACK,
    STATE(7251), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319267] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9664), 1,
      anon_sym_RBRACK,
    STATE(7252), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319285] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9666), 1,
      anon_sym_GT_GT,
    STATE(7253), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319303] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9668), 1,
      anon_sym_GT_GT,
    STATE(7254), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319321] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9670), 1,
      anon_sym_GT_GT,
    STATE(7255), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319339] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9672), 1,
      anon_sym_SEMI,
    STATE(7256), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319357] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9674), 1,
      sym_bit_string_base,
    STATE(7257), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319375] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9676), 1,
      anon_sym_SEMI,
    STATE(7258), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319393] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9678), 1,
      sym_MAP,
    STATE(7259), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319411] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9680), 1,
      sym_MAP,
    STATE(7260), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319429] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9682), 1,
      anon_sym_RBRACK,
    STATE(7261), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319447] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9684), 1,
      sym_MAP,
    STATE(7262), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319465] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9686), 1,
      anon_sym_RBRACK,
    STATE(7263), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319483] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9688), 1,
      anon_sym_RBRACK,
    STATE(7264), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319501] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9690), 1,
      anon_sym_SEMI,
    STATE(7265), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319519] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9692), 1,
      anon_sym_SEMI,
    STATE(7266), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319537] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9694), 1,
      anon_sym_SEMI,
    STATE(7267), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319555] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2830), 1,
      sym_GENERATE,
    STATE(7268), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319573] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9696), 1,
      sym_bit_string_base,
    STATE(7269), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319591] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9698), 1,
      anon_sym_SEMI,
    STATE(7270), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319609] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9700), 1,
      sym_MAP,
    STATE(7271), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319627] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9702), 1,
      sym_MAP,
    STATE(7272), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319645] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9704), 1,
      sym_string_literal,
    STATE(7273), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319663] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9706), 1,
      sym_MAP,
    STATE(7274), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319681] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9708), 1,
      sym_OF,
    STATE(7275), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319699] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9710), 1,
      anon_sym_COLON,
    STATE(7276), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319717] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9712), 1,
      anon_sym_SEMI,
    STATE(7277), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319735] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9714), 1,
      anon_sym_RPAREN,
    STATE(7278), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319753] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9716), 1,
      anon_sym_SEMI,
    STATE(7279), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319771] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9718), 1,
      anon_sym_SEMI,
    STATE(7280), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319789] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9720), 1,
      sym_bit_string_base,
    STATE(7281), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319807] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9722), 1,
      anon_sym_SEMI,
    STATE(7282), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319825] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9724), 1,
      sym_MAP,
    STATE(7283), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319843] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9726), 1,
      sym_MAP,
    STATE(7284), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319861] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9728), 1,
      sym_bit_string_value,
    STATE(7285), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319879] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9730), 1,
      sym_MAP,
    STATE(7286), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319897] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8153), 1,
      anon_sym_SEMI,
    STATE(7287), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319915] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9732), 1,
      anon_sym_RPAREN,
    STATE(7288), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319933] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8167), 1,
      anon_sym_SEMI,
    STATE(7289), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319951] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9734), 1,
      anon_sym_SEMI,
    STATE(7290), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319969] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9736), 1,
      anon_sym_SEMI,
    STATE(7291), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [319987] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9738), 1,
      anon_sym_RBRACK,
    STATE(7292), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320005] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9740), 1,
      sym_bit_string_base,
    STATE(7293), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320023] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9742), 1,
      anon_sym_RBRACK,
    STATE(7294), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320041] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9744), 1,
      sym_MAP,
    STATE(7295), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320059] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9746), 1,
      sym_MAP,
    STATE(7296), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320077] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9748), 1,
      anon_sym_SEMI,
    STATE(7297), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320095] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9750), 1,
      sym_MAP,
    STATE(7298), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320113] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9752), 1,
      anon_sym_SEMI,
    STATE(7299), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320131] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9754), 1,
      anon_sym_GT_GT,
    STATE(7300), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320149] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9756), 1,
      anon_sym_SEMI,
    STATE(7301), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320167] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9758), 1,
      anon_sym_SEMI,
    STATE(7302), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320185] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9760), 1,
      anon_sym_SEMI,
    STATE(7303), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320203] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9762), 1,
      anon_sym_GT_GT,
    STATE(7304), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320221] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9764), 1,
      sym_bit_string_base,
    STATE(7305), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320239] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9766), 1,
      anon_sym_SEMI,
    STATE(7306), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320257] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9768), 1,
      sym_MAP,
    STATE(7307), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320275] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9770), 1,
      sym_MAP,
    STATE(7308), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320293] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9772), 1,
      anon_sym_GT_GT,
    STATE(7309), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320311] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9774), 1,
      sym_MAP,
    STATE(7310), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320329] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9776), 1,
      sym_box,
    STATE(7311), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320347] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7471), 1,
      anon_sym_SEMI,
    STATE(7312), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320365] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7382), 1,
      sym_RETURN,
    STATE(7313), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320383] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9778), 1,
      anon_sym_RBRACK,
    STATE(7314), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320401] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9780), 1,
      sym_bit_string_value,
    STATE(7315), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320419] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9782), 1,
      anon_sym_RBRACK,
    STATE(7316), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320437] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9784), 1,
      sym_bit_string_base,
    STATE(7317), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320455] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9786), 1,
      anon_sym_RPAREN,
    STATE(7318), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320473] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9788), 1,
      sym_MAP,
    STATE(7319), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320491] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9790), 1,
      sym_MAP,
    STATE(7320), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320509] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9792), 1,
      anon_sym_SEMI,
    STATE(7321), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320527] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9794), 1,
      sym_MAP,
    STATE(7322), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320545] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8448), 1,
      anon_sym_SEMI,
    STATE(7323), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320563] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9796), 1,
      anon_sym_COLON,
    STATE(7324), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320581] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9798), 1,
      anon_sym_RPAREN,
    STATE(7325), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320599] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9800), 1,
      anon_sym_SEMI,
    STATE(7326), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320617] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9802), 1,
      anon_sym_RPAREN,
    STATE(7327), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320635] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9804), 1,
      anon_sym_COLON,
    STATE(7328), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320653] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9806), 1,
      sym_bit_string_base,
    STATE(7329), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320671] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9808), 1,
      anon_sym_RBRACK,
    STATE(7330), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320689] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9810), 1,
      sym_MAP,
    STATE(7331), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320707] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9812), 1,
      sym_MAP,
    STATE(7332), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320725] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9814), 1,
      anon_sym_RPAREN,
    STATE(7333), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320743] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9816), 1,
      sym_MAP,
    STATE(7334), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320761] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9818), 1,
      anon_sym_SEMI,
    STATE(7335), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320779] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9820), 1,
      sym_RETURN,
    STATE(7336), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320797] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9822), 1,
      anon_sym_RBRACK,
    STATE(7337), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320815] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9824), 1,
      anon_sym_SEMI,
    STATE(7338), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320833] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9826), 1,
      anon_sym_GT_GT,
    STATE(7339), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320851] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9828), 1,
      anon_sym_GT_GT,
    STATE(7340), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320869] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9830), 1,
      sym_bit_string_base,
    STATE(7341), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320887] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9832), 1,
      sym_NEW,
    STATE(7342), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320905] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9834), 1,
      sym_MAP,
    STATE(7343), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320923] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9836), 1,
      sym_MAP,
    STATE(7344), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320941] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9838), 1,
      anon_sym_GT_GT,
    STATE(7345), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320959] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9840), 1,
      sym_MAP,
    STATE(7346), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320977] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9842), 1,
      anon_sym_GT_GT,
    STATE(7347), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [320995] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9844), 1,
      anon_sym_GT_GT,
    STATE(7348), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7489), 1,
      anon_sym_COLON,
    STATE(7349), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321031] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(2678), 1,
      sym_OF,
    STATE(7350), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321049] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9846), 1,
      anon_sym_RBRACK,
    STATE(7351), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321067] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9848), 1,
      anon_sym_GT_GT,
    STATE(7352), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321085] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9850), 1,
      sym_bit_string_base,
    STATE(7353), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321103] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9852), 1,
      anon_sym_RBRACK,
    STATE(7354), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321121] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9854), 1,
      sym_MAP,
    STATE(7355), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321139] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9856), 1,
      sym_MAP,
    STATE(7356), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321157] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9858), 1,
      sym_bit_string_value,
    STATE(7357), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321175] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9860), 1,
      sym_MAP,
    STATE(7358), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321193] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9862), 1,
      anon_sym_RPAREN,
    STATE(7359), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321211] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9864), 1,
      sym_AFTER,
    STATE(7360), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321229] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5755), 1,
      anon_sym_SEMI,
    STATE(7361), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321247] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9866), 1,
      anon_sym_RPAREN,
    STATE(7362), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321265] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9868), 1,
      anon_sym_DOT,
    STATE(7363), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321283] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9870), 1,
      anon_sym_RPAREN,
    STATE(7364), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321301] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9872), 1,
      sym_bit_string_base,
    STATE(7365), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321319] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9874), 1,
      anon_sym_RBRACK,
    STATE(7366), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321337] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9876), 1,
      sym_MAP,
    STATE(7367), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321355] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9878), 1,
      sym_MAP,
    STATE(7368), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321373] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9880), 1,
      anon_sym_SEMI,
    STATE(7369), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321391] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9882), 1,
      sym_MAP,
    STATE(7370), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321409] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9884), 1,
      anon_sym_RPAREN,
    STATE(7371), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321427] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9886), 1,
      anon_sym_COLON,
    STATE(7372), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321445] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9888), 1,
      anon_sym_SEMI,
    STATE(7373), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321463] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9890), 1,
      anon_sym_RBRACK,
    STATE(7374), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321481] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9892), 1,
      anon_sym_SEMI,
    STATE(7375), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321499] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5704), 1,
      anon_sym_SEMI,
    STATE(7376), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321517] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9894), 1,
      anon_sym_RBRACK,
    STATE(7377), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321535] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9896), 1,
      sym_MAP,
    STATE(7378), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321553] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9898), 1,
      sym_MAP,
    STATE(7379), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321571] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9900), 1,
      sym_box,
    STATE(7380), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321589] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9902), 1,
      sym_MAP,
    STATE(7381), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321607] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9904), 1,
      anon_sym_SEMI,
    STATE(7382), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321625] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9906), 1,
      anon_sym_RPAREN,
    STATE(7383), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321643] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9908), 1,
      anon_sym_GT_GT,
    STATE(7384), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321661] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9910), 1,
      anon_sym_GT_GT,
    STATE(7385), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321679] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9912), 1,
      anon_sym_GT_GT,
    STATE(7386), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321697] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9914), 1,
      sym_box,
    STATE(7387), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321715] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9916), 1,
      anon_sym_RBRACK,
    STATE(7388), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321733] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9918), 1,
      sym_MAP,
    STATE(7389), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321751] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9920), 1,
      sym_MAP,
    STATE(7390), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321769] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9922), 1,
      sym_BODY,
    STATE(7391), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321787] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9924), 1,
      sym_MAP,
    STATE(7392), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321805] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9926), 1,
      anon_sym_SEMI,
    STATE(7393), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321823] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(804), 1,
      anon_sym_SEMI,
    STATE(7394), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321841] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9928), 1,
      anon_sym_SEMI,
    STATE(7395), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321859] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9930), 1,
      anon_sym_SEMI,
    STATE(7396), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321877] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9932), 1,
      anon_sym_RBRACK,
    STATE(7397), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321895] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9934), 1,
      anon_sym_SEMI,
    STATE(7398), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321913] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7386), 1,
      anon_sym_SEMI,
    STATE(7399), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321931] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9936), 1,
      sym_MAP,
    STATE(7400), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321949] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9938), 1,
      sym_MAP,
    STATE(7401), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321967] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9940), 1,
      anon_sym_SEMI,
    STATE(7402), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [321985] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9942), 1,
      sym_MAP,
    STATE(7403), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322003] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9944), 1,
      anon_sym_SEMI,
    STATE(7404), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322021] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9946), 1,
      anon_sym_SEMI,
    STATE(7405), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322039] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9948), 1,
      sym_bit_string_value,
    STATE(7406), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322057] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9950), 1,
      sym_LOOP,
    STATE(7407), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322075] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9952), 1,
      anon_sym_COLON,
    STATE(7408), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322093] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9954), 1,
      anon_sym_RPAREN,
    STATE(7409), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322111] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9956), 1,
      sym_bit_string_value,
    STATE(7410), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322129] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9958), 1,
      sym_MAP,
    STATE(7411), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322147] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9960), 1,
      sym_MAP,
    STATE(7412), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322165] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9962), 1,
      anon_sym_SEMI,
    STATE(7413), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322183] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9964), 1,
      sym_MAP,
    STATE(7414), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322201] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9966), 1,
      anon_sym_SEMI,
    STATE(7415), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322219] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9968), 1,
      anon_sym_RPAREN,
    STATE(7416), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322237] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7510), 1,
      anon_sym_SEMI,
    STATE(7417), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322255] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5815), 1,
      anon_sym_SEMI,
    STATE(7418), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322273] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9970), 1,
      anon_sym_SEMI,
    STATE(7419), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322291] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9972), 1,
      anon_sym_RPAREN,
    STATE(7420), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322309] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9974), 1,
      anon_sym_RBRACK,
    STATE(7421), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322327] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9976), 1,
      sym_MAP,
    STATE(7422), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322345] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9978), 1,
      sym_MAP,
    STATE(7423), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322363] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9980), 1,
      anon_sym_RPAREN,
    STATE(7424), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322381] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9982), 1,
      sym_MAP,
    STATE(7425), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322399] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5946), 1,
      sym_END,
    STATE(7426), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322417] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9984), 1,
      sym_END,
    STATE(7427), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322435] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7540), 1,
      anon_sym_SEMI,
    STATE(7428), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322453] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9986), 1,
      sym_bit_string_value,
    STATE(7429), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322471] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9988), 1,
      anon_sym_RPAREN,
    STATE(7430), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322489] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9990), 1,
      sym_IF,
    STATE(7431), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322507] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9992), 1,
      anon_sym_GT_GT,
    STATE(7432), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322525] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9994), 1,
      sym_MAP,
    STATE(7433), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322543] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9996), 1,
      sym_MAP,
    STATE(7434), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322561] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9998), 1,
      anon_sym_GT_GT,
    STATE(7435), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322579] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10000), 1,
      sym_MAP,
    STATE(7436), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322597] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10002), 1,
      anon_sym_GT_GT,
    STATE(7437), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322615] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10004), 1,
      anon_sym_SEMI,
    STATE(7438), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322633] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7552), 1,
      anon_sym_SEMI,
    STATE(7439), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322651] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10006), 1,
      anon_sym_SEMI,
    STATE(7440), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322669] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10008), 1,
      anon_sym_RBRACK,
    STATE(7441), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322687] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10010), 1,
      sym_LOOP,
    STATE(7442), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322705] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10012), 1,
      anon_sym_SEMI,
    STATE(7443), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322723] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10014), 1,
      sym_MAP,
    STATE(7444), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322741] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10016), 1,
      sym_MAP,
    STATE(7445), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322759] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7292), 1,
      anon_sym_SEMI,
    STATE(7446), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322777] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10018), 1,
      sym_MAP,
    STATE(7447), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322795] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10020), 1,
      anon_sym_SEMI,
    STATE(7448), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322813] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10022), 1,
      anon_sym_RBRACK,
    STATE(7449), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322831] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10024), 1,
      anon_sym_COLON,
    STATE(7450), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322849] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10026), 1,
      anon_sym_RPAREN,
    STATE(7451), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322867] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10028), 1,
      anon_sym_SEMI,
    STATE(7452), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322885] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10030), 1,
      anon_sym_SEMI,
    STATE(7453), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322903] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10032), 1,
      sym_bit_string_value,
    STATE(7454), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322921] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10034), 1,
      sym_MAP,
    STATE(7455), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322939] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10036), 1,
      sym_MAP,
    STATE(7456), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322957] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10038), 1,
      anon_sym_SEMI,
    STATE(7457), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322975] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10040), 1,
      sym_MAP,
    STATE(7458), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [322993] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10042), 1,
      anon_sym_RBRACK,
    STATE(7459), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323011] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10044), 1,
      anon_sym_SEMI,
    STATE(7460), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323029] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10046), 1,
      anon_sym_SEMI,
    STATE(7461), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323047] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10048), 1,
      anon_sym_SEMI,
    STATE(7462), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323065] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10050), 1,
      anon_sym_SEMI,
    STATE(7463), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323083] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10052), 1,
      anon_sym_RPAREN,
    STATE(7464), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323101] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10054), 1,
      anon_sym_SEMI,
    STATE(7465), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323119] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10056), 1,
      sym_MAP,
    STATE(7466), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323137] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10058), 1,
      sym_MAP,
    STATE(7467), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323155] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10060), 1,
      anon_sym_RPAREN,
    STATE(7468), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323173] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10062), 1,
      sym_MAP,
    STATE(7469), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323191] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10064), 1,
      anon_sym_SEMI,
    STATE(7470), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323209] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10066), 1,
      anon_sym_RPAREN,
    STATE(7471), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323227] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10068), 1,
      anon_sym_SEMI,
    STATE(7472), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323245] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10070), 1,
      anon_sym_SEMI,
    STATE(7473), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323263] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10072), 1,
      anon_sym_SEMI,
    STATE(7474), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323281] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10074), 1,
      anon_sym_SEMI,
    STATE(7475), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323299] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10076), 1,
      sym_bit_string_value,
    STATE(7476), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323317] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10078), 1,
      sym_MAP,
    STATE(7477), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323335] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10080), 1,
      sym_MAP,
    STATE(7478), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323353] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10082), 1,
      anon_sym_RPAREN,
    STATE(7479), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323371] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10084), 1,
      sym_MAP,
    STATE(7480), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323389] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10086), 1,
      anon_sym_SEMI,
    STATE(7481), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323407] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10088), 1,
      anon_sym_RPAREN,
    STATE(7482), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323425] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7524), 1,
      anon_sym_SEMI,
    STATE(7483), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323443] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10090), 1,
      ts_builtin_sym_end,
    STATE(7484), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323461] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10092), 1,
      anon_sym_GT_GT,
    STATE(7485), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323479] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10094), 1,
      anon_sym_RBRACK,
    STATE(7486), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323497] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6870), 1,
      sym_RETURN,
    STATE(7487), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323515] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10096), 1,
      sym_MAP,
    STATE(7488), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323533] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10098), 1,
      sym_MAP,
    STATE(7489), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323551] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10100), 1,
      anon_sym_GT_GT,
    STATE(7490), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323569] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10102), 1,
      sym_MAP,
    STATE(7491), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323587] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10104), 1,
      anon_sym_SEMI,
    STATE(7492), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323605] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10106), 1,
      anon_sym_COLON,
    STATE(7493), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323623] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10108), 1,
      anon_sym_SEMI,
    STATE(7494), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323641] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10110), 1,
      anon_sym_SEMI,
    STATE(7495), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323659] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10112), 1,
      anon_sym_COLON,
    STATE(7496), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323677] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10114), 1,
      anon_sym_SEMI,
    STATE(7497), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323695] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10116), 1,
      anon_sym_GT_GT,
    STATE(7498), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323713] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10118), 1,
      sym_MAP,
    STATE(7499), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323731] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10120), 1,
      sym_MAP,
    STATE(7500), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323749] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10122), 1,
      anon_sym_GT_GT,
    STATE(7501), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323767] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10124), 1,
      sym_MAP,
    STATE(7502), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323785] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10126), 1,
      anon_sym_SEMI,
    STATE(7503), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323803] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6886), 1,
      anon_sym_SEMI,
    STATE(7504), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323821] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10128), 1,
      anon_sym_SEMI,
    STATE(7505), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323839] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10130), 1,
      anon_sym_GT_GT,
    STATE(7506), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323857] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10132), 1,
      anon_sym_GT_GT,
    STATE(7507), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323875] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10134), 1,
      anon_sym_RPAREN,
    STATE(7508), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323893] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10136), 1,
      anon_sym_SEMI,
    STATE(7509), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323911] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10138), 1,
      sym_MAP,
    STATE(7510), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323929] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10140), 1,
      sym_MAP,
    STATE(7511), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323947] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10142), 1,
      anon_sym_RBRACK,
    STATE(7512), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323965] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10144), 1,
      sym_MAP,
    STATE(7513), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [323983] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10146), 1,
      anon_sym_SEMI,
    STATE(7514), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324001] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10148), 1,
      anon_sym_SEMI,
    STATE(7515), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324019] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10150), 1,
      anon_sym_RBRACK,
    STATE(7516), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324037] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10152), 1,
      anon_sym_RBRACK,
    STATE(7517), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324055] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7538), 1,
      anon_sym_SEMI,
    STATE(7518), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324073] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10154), 1,
      anon_sym_SEMI,
    STATE(7519), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324091] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10156), 1,
      anon_sym_SEMI,
    STATE(7520), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324109] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10158), 1,
      sym_MAP,
    STATE(7521), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324127] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10160), 1,
      sym_MAP,
    STATE(7522), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324145] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10162), 1,
      anon_sym_SEMI,
    STATE(7523), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324163] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10164), 1,
      sym_MAP,
    STATE(7524), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324181] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10166), 1,
      anon_sym_RPAREN,
    STATE(7525), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324199] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10168), 1,
      anon_sym_SEMI,
    STATE(7526), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324217] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10166), 1,
      sym_string_literal,
    STATE(7527), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324235] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10170), 1,
      anon_sym_SEMI,
    STATE(7528), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324253] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10172), 1,
      anon_sym_SEMI,
    STATE(7529), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324271] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10174), 1,
      anon_sym_SEMI,
    STATE(7530), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324289] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10176), 1,
      anon_sym_SEMI,
    STATE(7531), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324307] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10178), 1,
      sym_MAP,
    STATE(7532), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324325] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10180), 1,
      sym_MAP,
    STATE(7533), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324343] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10182), 1,
      anon_sym_SEMI,
    STATE(7534), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324361] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10184), 1,
      sym_MAP,
    STATE(7535), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324379] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10186), 1,
      anon_sym_SEMI,
    STATE(7536), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324397] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10188), 1,
      anon_sym_RPAREN,
    STATE(7537), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324415] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10190), 1,
      anon_sym_SEMI,
    STATE(7538), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324433] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6138), 1,
      anon_sym_SEMI,
    STATE(7539), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324451] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10192), 1,
      anon_sym_SEMI,
    STATE(7540), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324469] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10194), 1,
      anon_sym_SEMI,
    STATE(7541), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324487] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10196), 1,
      anon_sym_RBRACK,
    STATE(7542), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324505] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10198), 1,
      sym_MAP,
    STATE(7543), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324523] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10200), 1,
      sym_MAP,
    STATE(7544), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324541] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3884), 1,
      anon_sym_RPAREN,
    STATE(7545), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324559] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10202), 1,
      sym_MAP,
    STATE(7546), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324577] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10204), 1,
      sym_RETURN,
    STATE(7547), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324595] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10206), 1,
      anon_sym_SEMI,
    STATE(7548), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324613] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6878), 1,
      sym_RETURN,
    STATE(7549), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324631] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10208), 1,
      anon_sym_SEMI,
    STATE(7550), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324649] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10210), 1,
      anon_sym_SEMI,
    STATE(7551), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324667] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10212), 1,
      anon_sym_SEMI,
    STATE(7552), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324685] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10214), 1,
      anon_sym_RPAREN,
    STATE(7553), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324703] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10216), 1,
      sym_MAP,
    STATE(7554), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324721] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10218), 1,
      sym_MAP,
    STATE(7555), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324739] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10220), 1,
      anon_sym_SEMI,
    STATE(7556), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324757] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10222), 1,
      sym_MAP,
    STATE(7557), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324775] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5735), 1,
      anon_sym_SEMI,
    STATE(7558), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324793] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10224), 1,
      anon_sym_SEMI,
    STATE(7559), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324811] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5737), 1,
      anon_sym_SEMI,
    STATE(7560), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324829] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10226), 1,
      anon_sym_RBRACK,
    STATE(7561), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324847] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(796), 1,
      anon_sym_SEMI,
    STATE(7562), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324865] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10228), 1,
      sym_bit_string_value,
    STATE(7563), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324883] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10230), 1,
      anon_sym_SEMI,
    STATE(7564), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324901] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10232), 1,
      sym_MAP,
    STATE(7565), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324919] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10234), 1,
      sym_MAP,
    STATE(7566), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324937] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5879), 1,
      anon_sym_SEMI,
    STATE(7567), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324955] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10236), 1,
      sym_MAP,
    STATE(7568), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324973] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7554), 1,
      anon_sym_SEMI,
    STATE(7569), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [324991] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10238), 1,
      anon_sym_SEMI,
    STATE(7570), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325009] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10240), 1,
      anon_sym_SEMI,
    STATE(7571), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325027] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10242), 1,
      anon_sym_GT_GT,
    STATE(7572), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325045] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10244), 1,
      anon_sym_GT_GT,
    STATE(7573), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325063] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10246), 1,
      anon_sym_SEMI,
    STATE(7574), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325081] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10248), 1,
      anon_sym_SEMI,
    STATE(7575), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325099] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10250), 1,
      sym_MAP,
    STATE(7576), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325117] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10252), 1,
      sym_MAP,
    STATE(7577), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325135] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10254), 1,
      anon_sym_GT_GT,
    STATE(7578), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325153] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10256), 1,
      sym_MAP,
    STATE(7579), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325171] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10258), 1,
      anon_sym_SEMI,
    STATE(7580), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325189] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10260), 1,
      anon_sym_SEMI,
    STATE(7581), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325207] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10262), 1,
      anon_sym_RBRACK,
    STATE(7582), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325225] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10264), 1,
      anon_sym_SEMI,
    STATE(7583), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325243] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10266), 1,
      anon_sym_SEMI,
    STATE(7584), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325261] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10268), 1,
      anon_sym_SEMI,
    STATE(7585), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325279] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10270), 1,
      anon_sym_SEMI,
    STATE(7586), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325297] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10272), 1,
      sym_MAP,
    STATE(7587), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325315] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10274), 1,
      sym_MAP,
    STATE(7588), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325333] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10276), 1,
      anon_sym_RBRACK,
    STATE(7589), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325351] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10278), 1,
      sym_MAP,
    STATE(7590), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325369] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7343), 1,
      anon_sym_SEMI,
    STATE(7591), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325387] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10280), 1,
      anon_sym_SEMI,
    STATE(7592), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325405] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10282), 1,
      anon_sym_RBRACK,
    STATE(7593), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325423] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10284), 1,
      sym_bit_string_value,
    STATE(7594), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325441] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10286), 1,
      anon_sym_SEMI,
    STATE(7595), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325459] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10288), 1,
      anon_sym_SEMI,
    STATE(7596), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325477] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10290), 1,
      anon_sym_SEMI,
    STATE(7597), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325495] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10292), 1,
      sym_MAP,
    STATE(7598), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325513] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10294), 1,
      sym_MAP,
    STATE(7599), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325531] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7459), 1,
      anon_sym_SEMI,
    STATE(7600), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325549] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10296), 1,
      anon_sym_RPAREN,
    STATE(7601), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325567] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10298), 1,
      anon_sym_SEMI,
    STATE(7602), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325585] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10300), 1,
      anon_sym_SEMI,
    STATE(7603), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325603] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10302), 1,
      sym_BEGIN,
    STATE(7604), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325621] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10304), 1,
      anon_sym_SEMI,
    STATE(7605), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325639] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10306), 1,
      anon_sym_RPAREN,
    STATE(7606), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325657] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10308), 1,
      anon_sym_RPAREN,
    STATE(7607), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325675] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10310), 1,
      anon_sym_SEMI,
    STATE(7608), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325693] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10312), 1,
      sym_MAP,
    STATE(7609), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325711] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10314), 1,
      anon_sym_SEMI,
    STATE(7610), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325729] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10316), 1,
      anon_sym_SEMI,
    STATE(7611), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325747] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10318), 1,
      anon_sym_SEMI,
    STATE(7612), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325765] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10320), 1,
      anon_sym_SEMI,
    STATE(7613), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325783] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7562), 1,
      anon_sym_SEMI,
    STATE(7614), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325801] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10322), 1,
      sym_MAP,
    STATE(7615), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325819] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10324), 1,
      anon_sym_RPAREN,
    STATE(7616), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325837] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10326), 1,
      anon_sym_SEMI,
    STATE(7617), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325855] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10328), 1,
      anon_sym_RPAREN,
    STATE(7618), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325873] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10330), 1,
      sym_RETURN,
    STATE(7619), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325891] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10332), 1,
      anon_sym_SEMI,
    STATE(7620), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325909] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10334), 1,
      sym_MAP,
    STATE(7621), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325927] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10336), 1,
      anon_sym_RPAREN,
    STATE(7622), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325945] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10338), 1,
      anon_sym_SEMI,
    STATE(7623), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325963] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10340), 1,
      sym_MAP,
    STATE(7624), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325981] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10342), 1,
      anon_sym_SEMI,
    STATE(7625), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [325999] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10344), 1,
      anon_sym_SEMI,
    STATE(7626), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326017] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10346), 1,
      sym_MAP,
    STATE(7627), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326035] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10348), 1,
      anon_sym_SEMI,
    STATE(7628), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326053] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10350), 1,
      anon_sym_RBRACK,
    STATE(7629), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326071] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10352), 1,
      anon_sym_SEMI,
    STATE(7630), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326089] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7574), 1,
      anon_sym_SEMI,
    STATE(7631), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326107] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7576), 1,
      anon_sym_SEMI,
    STATE(7632), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326125] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10354), 1,
      anon_sym_RPAREN,
    STATE(7633), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326143] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6273), 1,
      anon_sym_SEMI,
    STATE(7634), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326161] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10356), 1,
      anon_sym_SEMI,
    STATE(7635), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326179] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10358), 1,
      anon_sym_SEMI,
    STATE(7636), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326197] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10360), 1,
      anon_sym_SEMI,
    STATE(7637), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326215] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10362), 1,
      sym_bit_string_value,
    STATE(7638), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326233] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10364), 1,
      anon_sym_RBRACK,
    STATE(7639), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326251] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10366), 1,
      anon_sym_SEMI,
    STATE(7640), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326269] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10368), 1,
      anon_sym_RPAREN,
    STATE(7641), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326287] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7479), 1,
      anon_sym_SEMI,
    STATE(7642), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326305] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10370), 1,
      anon_sym_SEMI,
    STATE(7643), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326323] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10372), 1,
      anon_sym_GT_GT,
    STATE(7644), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326341] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10374), 1,
      anon_sym_SEMI,
    STATE(7645), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326359] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10376), 1,
      anon_sym_SEMI,
    STATE(7646), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326377] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10378), 1,
      anon_sym_GT_GT,
    STATE(7647), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326395] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7580), 1,
      anon_sym_SEMI,
    STATE(7648), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326413] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10380), 1,
      anon_sym_GT_GT,
    STATE(7649), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326431] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6279), 1,
      anon_sym_SEMI,
    STATE(7650), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326449] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10382), 1,
      anon_sym_SEMI,
    STATE(7651), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326467] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10384), 1,
      anon_sym_GT_GT,
    STATE(7652), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326485] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10386), 1,
      anon_sym_GT_GT,
    STATE(7653), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326503] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10388), 1,
      anon_sym_GT_GT,
    STATE(7654), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326521] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10390), 1,
      sym_FUNCTION,
    STATE(7655), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326539] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10392), 1,
      anon_sym_RBRACK,
    STATE(7656), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326557] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10394), 1,
      anon_sym_RPAREN,
    STATE(7657), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326575] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10396), 1,
      anon_sym_RBRACK,
    STATE(7658), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326593] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10398), 1,
      anon_sym_COLON,
    STATE(7659), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326611] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10400), 1,
      anon_sym_COLON,
    STATE(7660), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326629] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10402), 1,
      sym_AFTER,
    STATE(7661), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326647] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10404), 1,
      anon_sym_COLON,
    STATE(7662), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326665] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10406), 1,
      anon_sym_RPAREN,
    STATE(7663), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326683] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10408), 1,
      anon_sym_RBRACK,
    STATE(7664), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326701] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10410), 1,
      anon_sym_SEMI,
    STATE(7665), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326719] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10412), 1,
      anon_sym_SEMI,
    STATE(7666), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326737] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10414), 1,
      anon_sym_COLON,
    STATE(7667), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326755] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10416), 1,
      anon_sym_COLON,
    STATE(7668), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326773] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10418), 1,
      sym_IS,
    STATE(7669), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326791] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10420), 1,
      anon_sym_COLON,
    STATE(7670), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326809] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10422), 1,
      sym_END,
    STATE(7671), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326827] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10424), 1,
      anon_sym_RPAREN,
    STATE(7672), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326845] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10426), 1,
      anon_sym_SEMI,
    STATE(7673), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326863] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10428), 1,
      anon_sym_RBRACK,
    STATE(7674), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326881] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10430), 1,
      anon_sym_LPAREN,
    STATE(7675), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326899] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10432), 1,
      anon_sym_LPAREN,
    STATE(7676), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326917] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10434), 1,
      anon_sym_LPAREN,
    STATE(7677), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326935] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10436), 1,
      sym_NEW,
    STATE(7678), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326953] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10438), 1,
      anon_sym_SEMI,
    STATE(7679), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326971] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10440), 1,
      anon_sym_SEMI,
    STATE(7680), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [326989] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10442), 1,
      anon_sym_SEMI,
    STATE(7681), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327007] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10444), 1,
      anon_sym_RPAREN,
    STATE(7682), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327025] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10446), 1,
      anon_sym_SEMI,
    STATE(7683), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327043] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10448), 1,
      anon_sym_SEMI,
    STATE(7684), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327061] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10450), 1,
      anon_sym_RBRACK,
    STATE(7685), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327079] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10452), 1,
      anon_sym_GT_GT,
    STATE(7686), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327097] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10454), 1,
      anon_sym_GT_GT,
    STATE(7687), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327115] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10456), 1,
      anon_sym_COLON,
    STATE(7688), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327133] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10458), 1,
      anon_sym_COLON,
    STATE(7689), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327151] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10460), 1,
      anon_sym_COLON,
    STATE(7690), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327169] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10462), 1,
      anon_sym_COLON,
    STATE(7691), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327187] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10464), 1,
      anon_sym_SEMI,
    STATE(7692), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327205] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10466), 1,
      anon_sym_GT_GT,
    STATE(7693), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327223] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10468), 1,
      anon_sym_SEMI,
    STATE(7694), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327241] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10470), 1,
      anon_sym_GT_GT,
    STATE(7695), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327259] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10472), 1,
      sym_IS,
    STATE(7696), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327277] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10474), 1,
      sym_IS,
    STATE(7697), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327295] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10476), 1,
      anon_sym_SEMI,
    STATE(7698), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327313] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10478), 1,
      anon_sym_SEMI,
    STATE(7699), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327331] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10480), 1,
      anon_sym_GT_GT,
    STATE(7700), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327349] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10482), 1,
      anon_sym_GT_GT,
    STATE(7701), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327367] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10484), 1,
      anon_sym_RBRACK,
    STATE(7702), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327385] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10486), 1,
      anon_sym_SEMI,
    STATE(7703), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327403] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10488), 1,
      anon_sym_RPAREN,
    STATE(7704), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327421] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10490), 1,
      anon_sym_RBRACK,
    STATE(7705), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327439] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10492), 1,
      anon_sym_SEMI,
    STATE(7706), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327457] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10494), 1,
      anon_sym_RPAREN,
    STATE(7707), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327475] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10496), 1,
      sym_IN,
    STATE(7708), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327493] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10498), 1,
      anon_sym_SEMI,
    STATE(7709), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327511] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10500), 1,
      anon_sym_RPAREN,
    STATE(7710), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327529] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7590), 1,
      anon_sym_SEMI,
    STATE(7711), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327547] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10502), 1,
      anon_sym_RBRACK,
    STATE(7712), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327565] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10504), 1,
      anon_sym_SEMI,
    STATE(7713), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327583] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10506), 1,
      anon_sym_COLON,
    STATE(7714), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327601] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10508), 1,
      anon_sym_RPAREN,
    STATE(7715), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327619] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10510), 1,
      anon_sym_RBRACK,
    STATE(7716), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327637] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10512), 1,
      anon_sym_RPAREN,
    STATE(7717), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327655] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10514), 1,
      anon_sym_SEMI,
    STATE(7718), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327673] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10516), 1,
      anon_sym_SEMI,
    STATE(7719), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327691] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10518), 1,
      anon_sym_RBRACK,
    STATE(7720), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327709] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10520), 1,
      anon_sym_SEMI,
    STATE(7721), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327727] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10522), 1,
      anon_sym_COLON,
    STATE(7722), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327745] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10524), 1,
      anon_sym_COLON,
    STATE(7723), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327763] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10526), 1,
      sym_AFTER,
    STATE(7724), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327781] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10528), 1,
      anon_sym_COLON,
    STATE(7725), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327799] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10530), 1,
      anon_sym_RBRACK,
    STATE(7726), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327817] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10532), 1,
      anon_sym_SEMI,
    STATE(7727), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327835] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10534), 1,
      anon_sym_GT_GT,
    STATE(7728), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327853] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10536), 1,
      anon_sym_GT_GT,
    STATE(7729), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327871] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10538), 1,
      anon_sym_GT_GT,
    STATE(7730), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327889] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10540), 1,
      anon_sym_COLON,
    STATE(7731), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327907] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10542), 1,
      sym_IS,
    STATE(7732), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327925] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10544), 1,
      anon_sym_COLON,
    STATE(7733), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327943] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10546), 1,
      anon_sym_RBRACK,
    STATE(7734), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327961] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10548), 1,
      anon_sym_RBRACK,
    STATE(7735), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327979] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10550), 1,
      anon_sym_SEMI,
    STATE(7736), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [327997] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6287), 1,
      anon_sym_SEMI,
    STATE(7737), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328015] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10552), 1,
      anon_sym_LPAREN,
    STATE(7738), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328033] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10554), 1,
      anon_sym_LPAREN,
    STATE(7739), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328051] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10556), 1,
      sym_NEW,
    STATE(7740), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328069] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10558), 1,
      anon_sym_RBRACK,
    STATE(7741), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328087] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10560), 1,
      anon_sym_SEMI,
    STATE(7742), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328105] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7598), 1,
      anon_sym_SEMI,
    STATE(7743), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328123] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(9138), 1,
      sym_GENERATE,
    STATE(7744), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328141] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10562), 1,
      sym_bit_string_value,
    STATE(7745), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328159] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10564), 1,
      anon_sym_SEMI,
    STATE(7746), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328177] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10566), 1,
      anon_sym_SEMI,
    STATE(7747), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328195] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10568), 1,
      anon_sym_RPAREN,
    STATE(7748), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328213] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10570), 1,
      sym_bit_string_value,
    STATE(7749), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328231] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10572), 1,
      anon_sym_COLON,
    STATE(7750), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328249] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10574), 1,
      anon_sym_COLON,
    STATE(7751), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328267] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10576), 1,
      anon_sym_COLON,
    STATE(7752), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328285] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10578), 1,
      anon_sym_COLON,
    STATE(7753), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328303] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7615), 1,
      anon_sym_SEMI,
    STATE(7754), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328321] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10580), 1,
      anon_sym_GT_GT,
    STATE(7755), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328339] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10582), 1,
      anon_sym_SEMI,
    STATE(7756), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328357] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10584), 1,
      anon_sym_RPAREN,
    STATE(7757), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328375] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10586), 1,
      sym_IS,
    STATE(7758), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328393] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10588), 1,
      sym_IS,
    STATE(7759), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328411] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10590), 1,
      anon_sym_SEMI,
    STATE(7760), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328429] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10592), 1,
      anon_sym_RPAREN,
    STATE(7761), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328447] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10594), 1,
      anon_sym_RBRACK,
    STATE(7762), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328465] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10596), 1,
      anon_sym_GT_GT,
    STATE(7763), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328483] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10598), 1,
      anon_sym_SEMI,
    STATE(7764), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328501] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10600), 1,
      anon_sym_GT_GT,
    STATE(7765), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328519] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10602), 1,
      anon_sym_GT_GT,
    STATE(7766), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328537] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10604), 1,
      anon_sym_GT_GT,
    STATE(7767), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328555] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10606), 1,
      anon_sym_GT_GT,
    STATE(7768), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328573] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10608), 1,
      sym_bit_string_value,
    STATE(7769), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328591] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10610), 1,
      anon_sym_RBRACK,
    STATE(7770), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328609] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6888), 1,
      anon_sym_SEMI,
    STATE(7771), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328627] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10612), 1,
      anon_sym_RPAREN,
    STATE(7772), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328645] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10614), 1,
      anon_sym_RBRACK,
    STATE(7773), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328663] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10616), 1,
      anon_sym_SEMI,
    STATE(7774), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328681] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10618), 1,
      anon_sym_SEMI,
    STATE(7775), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328699] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10620), 1,
      anon_sym_SEMI,
    STATE(7776), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328717] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10622), 1,
      anon_sym_SEMI,
    STATE(7777), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328735] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10624), 1,
      anon_sym_RBRACK,
    STATE(7778), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328753] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10626), 1,
      anon_sym_RPAREN,
    STATE(7779), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328771] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10628), 1,
      anon_sym_COLON,
    STATE(7780), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328789] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10630), 1,
      anon_sym_COLON,
    STATE(7781), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328807] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10632), 1,
      sym_AFTER,
    STATE(7782), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328825] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10634), 1,
      anon_sym_COLON,
    STATE(7783), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328843] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10636), 1,
      anon_sym_RPAREN,
    STATE(7784), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328861] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10638), 1,
      anon_sym_SEMI,
    STATE(7785), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328879] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10640), 1,
      anon_sym_SEMI,
    STATE(7786), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328897] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10642), 1,
      anon_sym_SEMI,
    STATE(7787), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328915] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10644), 1,
      anon_sym_RPAREN,
    STATE(7788), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328933] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10646), 1,
      anon_sym_COLON,
    STATE(7789), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328951] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10648), 1,
      sym_IS,
    STATE(7790), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328969] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10650), 1,
      anon_sym_COLON,
    STATE(7791), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [328987] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10652), 1,
      anon_sym_RBRACK,
    STATE(7792), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329005] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10654), 1,
      anon_sym_RBRACK,
    STATE(7793), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329023] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10656), 1,
      anon_sym_SEMI,
    STATE(7794), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329041] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10658), 1,
      anon_sym_SEMI,
    STATE(7795), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329059] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10660), 1,
      anon_sym_LPAREN,
    STATE(7796), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329077] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10662), 1,
      sym_NEW,
    STATE(7797), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329095] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10664), 1,
      anon_sym_GT_GT,
    STATE(7798), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329113] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10666), 1,
      anon_sym_SEMI,
    STATE(7799), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329131] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10668), 1,
      anon_sym_SEMI,
    STATE(7800), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329149] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10670), 1,
      anon_sym_GT_GT,
    STATE(7801), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329167] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10672), 1,
      anon_sym_GT_GT,
    STATE(7802), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329185] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10674), 1,
      anon_sym_COLON,
    STATE(7803), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329203] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10676), 1,
      anon_sym_COLON,
    STATE(7804), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329221] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10678), 1,
      anon_sym_COLON,
    STATE(7805), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329239] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10680), 1,
      anon_sym_COLON,
    STATE(7806), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329257] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10682), 1,
      anon_sym_SEMI,
    STATE(7807), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329275] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10684), 1,
      anon_sym_GT_GT,
    STATE(7808), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329293] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10686), 1,
      anon_sym_RBRACK,
    STATE(7809), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329311] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10688), 1,
      anon_sym_GT_GT,
    STATE(7810), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329329] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10690), 1,
      sym_IS,
    STATE(7811), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329347] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10692), 1,
      sym_IS,
    STATE(7812), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329365] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10694), 1,
      anon_sym_SEMI,
    STATE(7813), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329383] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10696), 1,
      anon_sym_RBRACK,
    STATE(7814), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329401] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10698), 1,
      anon_sym_GT_GT,
    STATE(7815), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329419] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10700), 1,
      anon_sym_COLON,
    STATE(7816), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329437] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7147), 1,
      anon_sym_SEMI,
    STATE(7817), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329455] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7619), 1,
      anon_sym_SEMI,
    STATE(7818), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329473] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10702), 1,
      anon_sym_RBRACK,
    STATE(7819), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329491] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10704), 1,
      anon_sym_RPAREN,
    STATE(7820), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329509] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10706), 1,
      anon_sym_SEMI,
    STATE(7821), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329527] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10708), 1,
      anon_sym_RBRACK,
    STATE(7822), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329545] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10710), 1,
      anon_sym_RPAREN,
    STATE(7823), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329563] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10712), 1,
      anon_sym_COLON,
    STATE(7824), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329581] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10714), 1,
      anon_sym_COLON,
    STATE(7825), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329599] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10716), 1,
      anon_sym_COLON,
    STATE(7826), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329617] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10718), 1,
      anon_sym_SEMI,
    STATE(7827), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329635] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10720), 1,
      anon_sym_SEMI,
    STATE(7828), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329653] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10722), 1,
      anon_sym_SEMI,
    STATE(7829), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329671] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10724), 1,
      anon_sym_RBRACK,
    STATE(7830), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329689] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10726), 1,
      sym_IS,
    STATE(7831), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329707] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10728), 1,
      anon_sym_COLON,
    STATE(7832), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329725] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10730), 1,
      anon_sym_SEMI,
    STATE(7833), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329743] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10732), 1,
      anon_sym_LPAREN,
    STATE(7834), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329761] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10734), 1,
      sym_NEW,
    STATE(7835), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329779] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10736), 1,
      anon_sym_SEMI,
    STATE(7836), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329797] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7633), 1,
      anon_sym_COLON,
    STATE(7837), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329815] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10738), 1,
      anon_sym_GT_GT,
    STATE(7838), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329833] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10740), 1,
      anon_sym_GT_GT,
    STATE(7839), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329851] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10742), 1,
      sym_NEW,
    STATE(7840), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329869] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10744), 1,
      anon_sym_COLON,
    STATE(7841), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329887] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10746), 1,
      anon_sym_COLON,
    STATE(7842), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329905] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10748), 1,
      anon_sym_COLON,
    STATE(7843), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329923] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10750), 1,
      anon_sym_COLON,
    STATE(7844), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329941] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10752), 1,
      anon_sym_RBRACK,
    STATE(7845), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329959] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10754), 1,
      sym_IS,
    STATE(7846), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329977] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10756), 1,
      sym_IS,
    STATE(7847), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [329995] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10758), 1,
      anon_sym_SEMI,
    STATE(7848), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10760), 1,
      anon_sym_GT_GT,
    STATE(7849), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330031] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10762), 1,
      anon_sym_SEMI,
    STATE(7850), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330049] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10764), 1,
      anon_sym_SEMI,
    STATE(7851), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330067] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10766), 1,
      anon_sym_SEMI,
    STATE(7852), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330085] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10768), 1,
      anon_sym_RBRACK,
    STATE(7853), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330103] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10770), 1,
      anon_sym_COLON,
    STATE(7854), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330121] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10772), 1,
      anon_sym_COLON,
    STATE(7855), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330139] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10774), 1,
      anon_sym_COLON,
    STATE(7856), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330157] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10776), 1,
      anon_sym_SEMI,
    STATE(7857), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330175] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10778), 1,
      sym_IS,
    STATE(7858), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330193] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10780), 1,
      anon_sym_COLON,
    STATE(7859), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330211] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10782), 1,
      sym_NEW,
    STATE(7860), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330229] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10784), 1,
      anon_sym_SEMI,
    STATE(7861), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330247] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10786), 1,
      anon_sym_COLON,
    STATE(7862), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330265] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10788), 1,
      anon_sym_COLON,
    STATE(7863), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330283] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10790), 1,
      anon_sym_COLON,
    STATE(7864), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330301] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10792), 1,
      anon_sym_COLON,
    STATE(7865), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330319] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10794), 1,
      sym_IS,
    STATE(7866), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330337] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10796), 1,
      sym_IS,
    STATE(7867), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330355] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10798), 1,
      anon_sym_SEMI,
    STATE(7868), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330373] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7627), 1,
      anon_sym_SEMI,
    STATE(7869), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330391] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10800), 1,
      anon_sym_RBRACK,
    STATE(7870), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330409] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10802), 1,
      anon_sym_COLON,
    STATE(7871), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330427] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10804), 1,
      sym_NEW,
    STATE(7872), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330445] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10806), 1,
      sym_NEW,
    STATE(7873), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330463] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10808), 1,
      anon_sym_COLON,
    STATE(7874), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330481] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10810), 1,
      anon_sym_COLON,
    STATE(7875), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330499] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10812), 1,
      anon_sym_COLON,
    STATE(7876), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330517] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10814), 1,
      anon_sym_COLON,
    STATE(7877), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330535] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10816), 1,
      sym_IS,
    STATE(7878), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330553] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10818), 1,
      sym_IS,
    STATE(7879), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330571] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10820), 1,
      sym_OF,
    STATE(7880), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330589] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10822), 1,
      anon_sym_COLON,
    STATE(7881), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330607] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10824), 1,
      anon_sym_COLON,
    STATE(7882), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330625] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10826), 1,
      anon_sym_COLON,
    STATE(7883), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330643] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10828), 1,
      sym_IS,
    STATE(7884), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330661] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10830), 1,
      anon_sym_COLON,
    STATE(7885), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330679] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10832), 1,
      anon_sym_COLON,
    STATE(7886), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330697] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10834), 1,
      anon_sym_COLON,
    STATE(7887), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330715] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10836), 1,
      anon_sym_COLON,
    STATE(7888), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330733] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10838), 1,
      anon_sym_COLON,
    STATE(7889), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330751] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10840), 1,
      anon_sym_COLON,
    STATE(7890), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330769] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10842), 1,
      anon_sym_COLON,
    STATE(7891), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330787] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10844), 1,
      anon_sym_COLON,
    STATE(7892), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330805] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10846), 1,
      anon_sym_COLON,
    STATE(7893), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330823] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10848), 1,
      anon_sym_COLON,
    STATE(7894), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330841] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10850), 1,
      anon_sym_COLON,
    STATE(7895), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330859] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10852), 1,
      anon_sym_COLON,
    STATE(7896), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330877] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10854), 1,
      anon_sym_COLON,
    STATE(7897), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330895] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10856), 1,
      anon_sym_COLON,
    STATE(7898), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330913] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10858), 1,
      anon_sym_COLON,
    STATE(7899), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330931] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10860), 1,
      anon_sym_COLON,
    STATE(7900), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330949] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10862), 1,
      anon_sym_COLON,
    STATE(7901), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330967] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10864), 1,
      anon_sym_COLON,
    STATE(7902), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [330985] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10866), 1,
      anon_sym_COLON,
    STATE(7903), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331003] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10868), 1,
      anon_sym_COLON,
    STATE(7904), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331021] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10870), 1,
      anon_sym_COLON,
    STATE(7905), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331039] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10872), 1,
      anon_sym_COLON,
    STATE(7906), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331057] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10874), 1,
      anon_sym_COLON,
    STATE(7907), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331075] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10876), 1,
      anon_sym_COLON,
    STATE(7908), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331093] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10878), 1,
      anon_sym_COLON,
    STATE(7909), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331111] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10880), 1,
      anon_sym_COLON,
    STATE(7910), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331129] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10882), 1,
      anon_sym_COLON,
    STATE(7911), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331147] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10884), 1,
      anon_sym_COLON,
    STATE(7912), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331165] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10886), 1,
      anon_sym_COLON,
    STATE(7913), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331183] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10888), 1,
      anon_sym_COLON,
    STATE(7914), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331201] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10890), 1,
      anon_sym_COLON,
    STATE(7915), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331219] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10892), 1,
      anon_sym_COLON,
    STATE(7916), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331237] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10894), 1,
      anon_sym_COLON,
    STATE(7917), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331255] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10896), 1,
      anon_sym_COLON,
    STATE(7918), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331273] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10898), 1,
      anon_sym_COLON,
    STATE(7919), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331291] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10900), 1,
      anon_sym_COLON,
    STATE(7920), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331309] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10902), 1,
      anon_sym_COLON,
    STATE(7921), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331327] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10904), 1,
      anon_sym_COLON,
    STATE(7922), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331345] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10906), 1,
      anon_sym_COLON,
    STATE(7923), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331363] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10908), 1,
      anon_sym_COLON,
    STATE(7924), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331381] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10910), 1,
      anon_sym_COLON,
    STATE(7925), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331399] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10912), 1,
      anon_sym_COLON,
    STATE(7926), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331417] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10914), 1,
      anon_sym_COLON,
    STATE(7927), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331435] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10916), 1,
      anon_sym_COLON,
    STATE(7928), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331453] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10918), 1,
      anon_sym_COLON,
    STATE(7929), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331471] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10920), 1,
      anon_sym_COLON,
    STATE(7930), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331489] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10922), 1,
      anon_sym_COLON,
    STATE(7931), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331507] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10924), 1,
      anon_sym_COLON,
    STATE(7932), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331525] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10926), 1,
      anon_sym_COLON,
    STATE(7933), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331543] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10928), 1,
      anon_sym_COLON,
    STATE(7934), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331561] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10930), 1,
      anon_sym_COLON,
    STATE(7935), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331579] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10932), 1,
      anon_sym_COLON,
    STATE(7936), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331597] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10934), 1,
      anon_sym_COLON,
    STATE(7937), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331615] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10936), 1,
      anon_sym_COLON,
    STATE(7938), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331633] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10938), 1,
      anon_sym_COLON,
    STATE(7939), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331651] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10940), 1,
      anon_sym_COLON,
    STATE(7940), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331669] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10942), 1,
      anon_sym_COLON,
    STATE(7941), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331687] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10944), 1,
      anon_sym_COLON,
    STATE(7942), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331705] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10946), 1,
      anon_sym_COLON,
    STATE(7943), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331723] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10948), 1,
      anon_sym_COLON,
    STATE(7944), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331741] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10950), 1,
      anon_sym_COLON,
    STATE(7945), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331759] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10952), 1,
      anon_sym_COLON,
    STATE(7946), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331777] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10954), 1,
      anon_sym_COLON,
    STATE(7947), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331795] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10956), 1,
      anon_sym_COLON,
    STATE(7948), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331813] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10958), 1,
      anon_sym_COLON,
    STATE(7949), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331831] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10960), 1,
      anon_sym_COLON,
    STATE(7950), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331849] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10962), 1,
      anon_sym_COLON,
    STATE(7951), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331867] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10964), 1,
      anon_sym_COLON,
    STATE(7952), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331885] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10966), 1,
      anon_sym_COLON,
    STATE(7953), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331903] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10968), 1,
      anon_sym_COLON,
    STATE(7954), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331921] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10970), 1,
      anon_sym_COLON,
    STATE(7955), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331939] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10972), 1,
      anon_sym_COLON,
    STATE(7956), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331957] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10974), 1,
      anon_sym_COLON,
    STATE(7957), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331975] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10976), 1,
      anon_sym_COLON,
    STATE(7958), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [331993] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10978), 1,
      anon_sym_COLON,
    STATE(7959), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332011] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10980), 1,
      anon_sym_COLON,
    STATE(7960), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332029] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10982), 1,
      anon_sym_COLON,
    STATE(7961), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332047] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10984), 1,
      anon_sym_COLON,
    STATE(7962), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332065] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10986), 1,
      anon_sym_COLON,
    STATE(7963), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332083] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10988), 1,
      anon_sym_COLON,
    STATE(7964), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332101] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10990), 1,
      anon_sym_COLON,
    STATE(7965), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332119] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10992), 1,
      anon_sym_COLON,
    STATE(7966), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332137] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10994), 1,
      anon_sym_COLON,
    STATE(7967), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332155] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10996), 1,
      anon_sym_COLON,
    STATE(7968), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332173] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(10998), 1,
      anon_sym_COLON,
    STATE(7969), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332191] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11000), 1,
      anon_sym_COLON,
    STATE(7970), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332209] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11002), 1,
      anon_sym_COLON,
    STATE(7971), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332227] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11004), 1,
      anon_sym_COLON,
    STATE(7972), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332245] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11006), 1,
      anon_sym_COLON,
    STATE(7973), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332263] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11008), 1,
      anon_sym_COLON,
    STATE(7974), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332281] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11010), 1,
      anon_sym_COLON,
    STATE(7975), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332299] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11012), 1,
      anon_sym_COLON,
    STATE(7976), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332317] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11014), 1,
      anon_sym_COLON,
    STATE(7977), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332335] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11016), 1,
      anon_sym_COLON,
    STATE(7978), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332353] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11018), 1,
      anon_sym_COLON,
    STATE(7979), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332371] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11020), 1,
      anon_sym_COLON,
    STATE(7980), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332389] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11022), 1,
      anon_sym_COLON,
    STATE(7981), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332407] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11024), 1,
      anon_sym_COLON,
    STATE(7982), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332425] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11026), 1,
      anon_sym_COLON,
    STATE(7983), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332443] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11028), 1,
      anon_sym_COLON,
    STATE(7984), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332461] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11030), 1,
      anon_sym_COLON,
    STATE(7985), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332479] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11032), 1,
      anon_sym_COLON,
    STATE(7986), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332497] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11034), 1,
      anon_sym_COLON,
    STATE(7987), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332515] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11036), 1,
      anon_sym_COLON,
    STATE(7988), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332533] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11038), 1,
      anon_sym_COLON,
    STATE(7989), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332551] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11040), 1,
      anon_sym_COLON,
    STATE(7990), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332569] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11042), 1,
      anon_sym_COLON,
    STATE(7991), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332587] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11044), 1,
      anon_sym_COLON,
    STATE(7992), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332605] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11046), 1,
      anon_sym_COLON,
    STATE(7993), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332623] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11048), 1,
      anon_sym_COLON,
    STATE(7994), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332641] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11050), 1,
      anon_sym_COLON,
    STATE(7995), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332659] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11052), 1,
      anon_sym_COLON,
    STATE(7996), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332677] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11054), 1,
      anon_sym_COLON,
    STATE(7997), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332695] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11056), 1,
      anon_sym_COLON,
    STATE(7998), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332713] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11058), 1,
      anon_sym_COLON,
    STATE(7999), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332731] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11060), 1,
      anon_sym_COLON,
    STATE(8000), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332749] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11062), 1,
      anon_sym_COLON,
    STATE(8001), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332767] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11064), 1,
      anon_sym_COLON,
    STATE(8002), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332785] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11066), 1,
      anon_sym_COLON,
    STATE(8003), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332803] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11068), 1,
      anon_sym_COLON,
    STATE(8004), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332821] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11070), 1,
      anon_sym_COLON,
    STATE(8005), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332839] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11072), 1,
      anon_sym_COLON,
    STATE(8006), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332857] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11074), 1,
      anon_sym_COLON,
    STATE(8007), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332875] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11076), 1,
      anon_sym_COLON,
    STATE(8008), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332893] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11078), 1,
      anon_sym_COLON,
    STATE(8009), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332911] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11080), 1,
      anon_sym_COLON,
    STATE(8010), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332929] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11082), 1,
      anon_sym_COLON,
    STATE(8011), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332947] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11084), 1,
      anon_sym_COLON,
    STATE(8012), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332965] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11086), 1,
      anon_sym_COLON,
    STATE(8013), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [332983] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11088), 1,
      sym_box,
    STATE(8014), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333001] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11090), 1,
      sym_box,
    STATE(8015), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333019] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11092), 1,
      anon_sym_SEMI,
    STATE(8016), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333037] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11094), 1,
      anon_sym_SEMI,
    STATE(8017), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333055] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11096), 1,
      anon_sym_SEMI,
    STATE(8018), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333073] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11098), 1,
      anon_sym_SEMI,
    STATE(8019), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333091] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11100), 1,
      anon_sym_SEMI,
    STATE(8020), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333109] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11102), 1,
      anon_sym_SEMI,
    STATE(8021), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333127] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11104), 1,
      anon_sym_SEMI,
    STATE(8022), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333145] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7163), 1,
      anon_sym_SEMI,
    STATE(8023), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333163] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11106), 1,
      anon_sym_SEMI,
    STATE(8024), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333181] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11108), 1,
      anon_sym_SEMI,
    STATE(8025), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333199] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11110), 1,
      anon_sym_SEMI,
    STATE(8026), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333217] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11112), 1,
      anon_sym_SEMI,
    STATE(8027), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333235] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11114), 1,
      anon_sym_SEMI,
    STATE(8028), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333253] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11116), 1,
      anon_sym_SEMI,
    STATE(8029), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333271] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11118), 1,
      anon_sym_SEMI,
    STATE(8030), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333289] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11120), 1,
      anon_sym_SEMI,
    STATE(8031), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333307] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5919), 1,
      anon_sym_SEMI,
    STATE(8032), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333325] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11122), 1,
      anon_sym_LPAREN,
    STATE(8033), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333343] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11124), 1,
      anon_sym_RPAREN,
    STATE(8034), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333361] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11126), 1,
      anon_sym_SEMI,
    STATE(8035), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333379] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11128), 1,
      anon_sym_SEMI,
    STATE(8036), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333397] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11130), 1,
      sym_bit_string_value,
    STATE(8037), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333415] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11132), 1,
      anon_sym_LPAREN,
    STATE(8038), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333433] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7165), 1,
      anon_sym_SEMI,
    STATE(8039), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333451] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11134), 1,
      anon_sym_RPAREN,
    STATE(8040), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333469] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11136), 1,
      anon_sym_SEMI,
    STATE(8041), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333487] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11138), 1,
      anon_sym_SEMI,
    STATE(8042), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333505] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11140), 1,
      anon_sym_SEMI,
    STATE(8043), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333523] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11142), 1,
      anon_sym_RPAREN,
    STATE(8044), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333541] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11144), 1,
      anon_sym_SEMI,
    STATE(8045), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333559] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11146), 1,
      anon_sym_SEMI,
    STATE(8046), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333577] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11148), 1,
      sym_box,
    STATE(8047), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333595] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11150), 1,
      anon_sym_SEMI,
    STATE(8048), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333613] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11152), 1,
      anon_sym_SEMI,
    STATE(8049), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333631] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11154), 1,
      anon_sym_SEMI,
    STATE(8050), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333649] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11156), 1,
      sym_END,
    STATE(8051), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333667] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11158), 1,
      anon_sym_RPAREN,
    STATE(8052), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333685] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11160), 1,
      anon_sym_RBRACK,
    STATE(8053), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333703] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11162), 1,
      anon_sym_SEMI,
    STATE(8054), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333721] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11164), 1,
      anon_sym_SEMI,
    STATE(8055), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333739] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11166), 1,
      sym_END,
    STATE(8056), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333757] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11168), 1,
      anon_sym_GT_GT,
    STATE(8057), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333775] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11170), 1,
      anon_sym_COLON,
    STATE(8058), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333793] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11172), 1,
      anon_sym_GT_GT,
    STATE(8059), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333811] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11174), 1,
      sym_PROCESS,
    STATE(8060), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333829] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11176), 1,
      anon_sym_GT_GT,
    STATE(8061), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333847] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11178), 1,
      sym_bit_string_value,
    STATE(8062), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333865] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11180), 1,
      anon_sym_SEMI,
    STATE(8063), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333883] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11182), 1,
      anon_sym_LPAREN,
    STATE(8064), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333901] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11184), 1,
      sym_bit_string_value,
    STATE(8065), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333919] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11186), 1,
      anon_sym_RBRACK,
    STATE(8066), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333937] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11188), 1,
      anon_sym_RPAREN,
    STATE(8067), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333955] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11190), 1,
      anon_sym_RBRACK,
    STATE(8068), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333973] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11192), 1,
      anon_sym_LPAREN,
    STATE(8069), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [333991] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11194), 1,
      anon_sym_SEMI,
    STATE(8070), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334009] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11196), 1,
      anon_sym_SEMI,
    STATE(8071), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334027] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11198), 1,
      anon_sym_SEMI,
    STATE(8072), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334045] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11200), 1,
      anon_sym_LPAREN,
    STATE(8073), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334063] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11202), 1,
      anon_sym_SEMI,
    STATE(8074), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334081] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11204), 1,
      anon_sym_RBRACK,
    STATE(8075), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334099] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11206), 1,
      anon_sym_SEMI,
    STATE(8076), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334117] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11208), 1,
      anon_sym_SEMI,
    STATE(8077), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334135] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11210), 1,
      sym_bit_string_value,
    STATE(8078), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334153] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11212), 1,
      anon_sym_RPAREN,
    STATE(8079), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334171] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11214), 1,
      anon_sym_RPAREN,
    STATE(8080), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334189] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11216), 1,
      anon_sym_SEMI,
    STATE(8081), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334207] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11218), 1,
      anon_sym_RPAREN,
    STATE(8082), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334225] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11220), 1,
      anon_sym_SEMI,
    STATE(8083), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334243] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7635), 1,
      anon_sym_SEMI,
    STATE(8084), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334261] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11222), 1,
      sym_PROTECTED,
    STATE(8085), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334279] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11224), 1,
      anon_sym_COLON,
    STATE(8086), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334297] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11226), 1,
      anon_sym_SEMI,
    STATE(8087), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334315] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11228), 1,
      anon_sym_SEMI,
    STATE(8088), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334333] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11230), 1,
      anon_sym_SEMI,
    STATE(8089), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334351] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11232), 1,
      anon_sym_LPAREN,
    STATE(8090), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334369] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11234), 1,
      anon_sym_RPAREN,
    STATE(8091), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334387] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11236), 1,
      anon_sym_SEMI,
    STATE(8092), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334405] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11238), 1,
      anon_sym_RBRACK,
    STATE(8093), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334423] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11240), 1,
      anon_sym_LPAREN,
    STATE(8094), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334441] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11242), 1,
      anon_sym_RBRACK,
    STATE(8095), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334459] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11244), 1,
      anon_sym_SEMI,
    STATE(8096), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334477] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11246), 1,
      anon_sym_SEMI,
    STATE(8097), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334495] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11248), 1,
      sym_THEN,
    STATE(8098), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334513] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11250), 1,
      anon_sym_SEMI,
    STATE(8099), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334531] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11252), 1,
      anon_sym_GT_GT,
    STATE(8100), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334549] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11254), 1,
      anon_sym_RPAREN,
    STATE(8101), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334567] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11256), 1,
      anon_sym_GT_GT,
    STATE(8102), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334585] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11258), 1,
      anon_sym_GT_GT,
    STATE(8103), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334603] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11260), 1,
      anon_sym_SEMI,
    STATE(8104), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334621] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11262), 1,
      anon_sym_SEMI,
    STATE(8105), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334639] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11264), 1,
      anon_sym_GT_GT,
    STATE(8106), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334657] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11266), 1,
      anon_sym_RBRACK,
    STATE(8107), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334675] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7645), 1,
      anon_sym_SEMI,
    STATE(8108), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334693] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11268), 1,
      anon_sym_SEMI,
    STATE(8109), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334711] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11270), 1,
      anon_sym_LPAREN,
    STATE(8110), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334729] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11272), 1,
      anon_sym_LPAREN,
    STATE(8111), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334747] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11274), 1,
      anon_sym_SEMI,
    STATE(8112), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334765] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11276), 1,
      anon_sym_SEMI,
    STATE(8113), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334783] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11278), 1,
      anon_sym_GT_GT,
    STATE(8114), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334801] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11280), 1,
      anon_sym_SEMI,
    STATE(8115), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334819] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11282), 1,
      anon_sym_SEMI,
    STATE(8116), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334837] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11284), 1,
      anon_sym_RBRACK,
    STATE(8117), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334855] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11286), 1,
      sym_OF,
    STATE(8118), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334873] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11288), 1,
      anon_sym_SEMI,
    STATE(8119), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334891] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11290), 1,
      anon_sym_SEMI,
    STATE(8120), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334909] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6153), 1,
      anon_sym_SEMI,
    STATE(8121), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334927] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11292), 1,
      anon_sym_SEMI,
    STATE(8122), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334945] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11294), 1,
      sym_string_literal,
    STATE(8123), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334963] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11296), 1,
      anon_sym_LPAREN,
    STATE(8124), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334981] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11298), 1,
      anon_sym_LPAREN,
    STATE(8125), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [334999] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(3836), 1,
      anon_sym_RPAREN,
    STATE(8126), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335017] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11300), 1,
      anon_sym_SEMI,
    STATE(8127), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335035] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11302), 1,
      anon_sym_SEMI,
    STATE(8128), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335053] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11304), 1,
      anon_sym_SEMI,
    STATE(8129), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335071] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11306), 1,
      anon_sym_SEMI,
    STATE(8130), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335089] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5743), 1,
      anon_sym_SEMI,
    STATE(8131), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335107] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11308), 1,
      anon_sym_GT_GT,
    STATE(8132), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335125] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5747), 1,
      anon_sym_SEMI,
    STATE(8133), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335143] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11310), 1,
      anon_sym_LPAREN,
    STATE(8134), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335161] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11312), 1,
      anon_sym_SEMI,
    STATE(8135), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335179] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(788), 1,
      anon_sym_SEMI,
    STATE(8136), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335197] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11314), 1,
      anon_sym_RPAREN,
    STATE(8137), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335215] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11316), 1,
      anon_sym_RBRACK,
    STATE(8138), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335233] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5787), 1,
      anon_sym_SEMI,
    STATE(8139), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335251] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7655), 1,
      anon_sym_SEMI,
    STATE(8140), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335269] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11318), 1,
      anon_sym_SEMI,
    STATE(8141), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335287] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11320), 1,
      anon_sym_SEMI,
    STATE(8142), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335305] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11322), 1,
      sym_END,
    STATE(8143), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335323] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11324), 1,
      anon_sym_RPAREN,
    STATE(8144), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335341] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11326), 1,
      anon_sym_SEMI,
    STATE(8145), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335359] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11328), 1,
      sym__directive_newline,
    STATE(8146), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335377] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7382), 1,
      anon_sym_SEMI,
    STATE(8147), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335395] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11330), 1,
      anon_sym_SEMI,
    STATE(8148), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335413] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11332), 1,
      anon_sym_RBRACK,
    STATE(8149), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335431] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11334), 1,
      anon_sym_RBRACK,
    STATE(8150), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335449] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11336), 1,
      anon_sym_SEMI,
    STATE(8151), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335467] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11338), 1,
      anon_sym_SEMI,
    STATE(8152), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335485] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11340), 1,
      anon_sym_SEMI,
    STATE(8153), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335503] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11342), 1,
      anon_sym_SEMI,
    STATE(8154), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335521] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11344), 1,
      anon_sym_DOT,
    STATE(8155), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335539] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8432), 1,
      anon_sym_SEMI,
    STATE(8156), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335557] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11346), 1,
      anon_sym_SEMI,
    STATE(8157), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335575] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11348), 1,
      anon_sym_RBRACK,
    STATE(8158), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335593] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11350), 1,
      anon_sym_GT_GT,
    STATE(8159), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335611] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11352), 1,
      anon_sym_GT_GT,
    STATE(8160), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335629] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11354), 1,
      anon_sym_SEMI,
    STATE(8161), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335647] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11356), 1,
      anon_sym_SEMI,
    STATE(8162), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335665] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11358), 1,
      anon_sym_SEMI,
    STATE(8163), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335683] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11360), 1,
      anon_sym_SEMI,
    STATE(8164), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335701] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11362), 1,
      anon_sym_SEMI,
    STATE(8165), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335719] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11364), 1,
      anon_sym_SEMI,
    STATE(8166), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335737] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11366), 1,
      anon_sym_SEMI,
    STATE(8167), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335755] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11368), 1,
      anon_sym_SEMI,
    STATE(8168), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335773] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11370), 1,
      anon_sym_SEMI,
    STATE(8169), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335791] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11372), 1,
      anon_sym_GT_GT,
    STATE(8170), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335809] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7663), 1,
      anon_sym_SEMI,
    STATE(8171), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335827] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11374), 1,
      anon_sym_RPAREN,
    STATE(8172), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335845] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11376), 1,
      anon_sym_SEMI,
    STATE(8173), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335863] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11378), 1,
      anon_sym_RPAREN,
    STATE(8174), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335881] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11380), 1,
      anon_sym_SEMI,
    STATE(8175), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335899] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11382), 1,
      anon_sym_RPAREN,
    STATE(8176), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335917] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11384), 1,
      anon_sym_SEMI,
    STATE(8177), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335935] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11386), 1,
      anon_sym_SEMI,
    STATE(8178), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335953] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11388), 1,
      anon_sym_RBRACK,
    STATE(8179), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335971] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11390), 1,
      anon_sym_SEMI,
    STATE(8180), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [335989] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11392), 1,
      anon_sym_COLON,
    STATE(8181), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336007] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11394), 1,
      anon_sym_SEMI,
    STATE(8182), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336025] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7675), 1,
      anon_sym_SEMI,
    STATE(8183), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336043] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7677), 1,
      anon_sym_SEMI,
    STATE(8184), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336061] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11396), 1,
      anon_sym_RBRACK,
    STATE(8185), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336079] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6062), 1,
      anon_sym_SEMI,
    STATE(8186), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336097] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11398), 1,
      anon_sym_SEMI,
    STATE(8187), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336115] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11400), 1,
      anon_sym_SEMI,
    STATE(8188), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336133] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11402), 1,
      anon_sym_DOT,
    STATE(8189), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336151] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11404), 1,
      sym_END,
    STATE(8190), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336169] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11406), 1,
      anon_sym_SEMI,
    STATE(8191), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336187] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11408), 1,
      anon_sym_SEMI,
    STATE(8192), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336205] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11410), 1,
      anon_sym_SEMI,
    STATE(8193), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336223] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11412), 1,
      sym_OF,
    STATE(8194), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336241] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11414), 1,
      anon_sym_SEMI,
    STATE(8195), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336259] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11416), 1,
      anon_sym_SEMI,
    STATE(8196), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336277] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11418), 1,
      anon_sym_SEMI,
    STATE(8197), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336295] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11420), 1,
      anon_sym_SEMI,
    STATE(8198), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336313] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11422), 1,
      anon_sym_RPAREN,
    STATE(8199), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336331] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7679), 1,
      anon_sym_SEMI,
    STATE(8200), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336349] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11424), 1,
      sym_bit_string_value,
    STATE(8201), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336367] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6081), 1,
      anon_sym_SEMI,
    STATE(8202), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336385] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5988), 1,
      anon_sym_SEMI,
    STATE(8203), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336403] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11426), 1,
      anon_sym_GT_GT,
    STATE(8204), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336421] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11428), 1,
      anon_sym_GT_GT,
    STATE(8205), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336439] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11430), 1,
      anon_sym_GT_GT,
    STATE(8206), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336457] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11432), 1,
      anon_sym_RPAREN,
    STATE(8207), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336475] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11434), 1,
      anon_sym_RBRACK,
    STATE(8208), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336493] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11436), 1,
      anon_sym_SEMI,
    STATE(8209), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336511] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11438), 1,
      sym_bit_string_value,
    STATE(8210), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336529] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11440), 1,
      anon_sym_SEMI,
    STATE(8211), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336547] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11442), 1,
      anon_sym_SEMI,
    STATE(8212), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336565] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11444), 1,
      anon_sym_SEMI,
    STATE(8213), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336583] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11446), 1,
      anon_sym_SEMI,
    STATE(8214), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336601] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11448), 1,
      anon_sym_SEMI,
    STATE(8215), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336619] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11450), 1,
      anon_sym_SEMI,
    STATE(8216), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336637] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11452), 1,
      anon_sym_RBRACK,
    STATE(8217), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336655] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7685), 1,
      anon_sym_SEMI,
    STATE(8218), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336673] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11454), 1,
      anon_sym_SEMI,
    STATE(8219), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336691] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11456), 1,
      anon_sym_SEMI,
    STATE(8220), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336709] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11458), 1,
      anon_sym_SEMI,
    STATE(8221), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336727] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11460), 1,
      anon_sym_SEMI,
    STATE(8222), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336745] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11462), 1,
      anon_sym_SEMI,
    STATE(8223), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336763] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11464), 1,
      anon_sym_SEMI,
    STATE(8224), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336781] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11466), 1,
      sym__directive_newline,
    STATE(8225), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336799] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7689), 1,
      anon_sym_SEMI,
    STATE(8226), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336817] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11468), 1,
      anon_sym_SEMI,
    STATE(8227), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336835] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7693), 1,
      anon_sym_SEMI,
    STATE(8228), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336853] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11470), 1,
      anon_sym_SEMI,
    STATE(8229), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336871] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11472), 1,
      anon_sym_SEMI,
    STATE(8230), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336889] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11474), 1,
      anon_sym_SEMI,
    STATE(8231), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336907] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(6902), 1,
      anon_sym_SEMI,
    STATE(8232), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336925] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11476), 1,
      anon_sym_GT_GT,
    STATE(8233), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336943] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11478), 1,
      anon_sym_SEMI,
    STATE(8234), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336961] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11480), 1,
      anon_sym_SEMI,
    STATE(8235), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336979] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11482), 1,
      anon_sym_SEMI,
    STATE(8236), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [336997] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11484), 1,
      anon_sym_GT_GT,
    STATE(8237), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337015] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11486), 1,
      anon_sym_SEMI,
    STATE(8238), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337033] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11488), 1,
      anon_sym_SEMI,
    STATE(8239), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337051] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11490), 1,
      anon_sym_SEMI,
    STATE(8240), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337069] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11492), 1,
      anon_sym_GT_GT,
    STATE(8241), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337087] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11494), 1,
      anon_sym_SEMI,
    STATE(8242), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337105] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7697), 1,
      anon_sym_SEMI,
    STATE(8243), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337123] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11496), 1,
      sym_bit_string_value,
    STATE(8244), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337141] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11498), 1,
      anon_sym_SEMI,
    STATE(8245), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337159] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11500), 1,
      anon_sym_RBRACK,
    STATE(8246), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337177] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11502), 1,
      anon_sym_RBRACK,
    STATE(8247), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337195] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11504), 1,
      anon_sym_SEMI,
    STATE(8248), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337213] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11506), 1,
      anon_sym_SEMI,
    STATE(8249), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337231] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11508), 1,
      anon_sym_SEMI,
    STATE(8250), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337249] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11510), 1,
      anon_sym_SEMI,
    STATE(8251), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337267] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11512), 1,
      anon_sym_SEMI,
    STATE(8252), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337285] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11514), 1,
      anon_sym_SEMI,
    STATE(8253), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337303] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11516), 1,
      anon_sym_SEMI,
    STATE(8254), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337321] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11518), 1,
      anon_sym_SEMI,
    STATE(8255), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337339] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11520), 1,
      anon_sym_SEMI,
    STATE(8256), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337357] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11522), 1,
      anon_sym_SEMI,
    STATE(8257), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337375] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11524), 1,
      anon_sym_SEMI,
    STATE(8258), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337393] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11526), 1,
      anon_sym_SEMI,
    STATE(8259), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337411] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11528), 1,
      anon_sym_SEMI,
    STATE(8260), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337429] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11530), 1,
      anon_sym_SEMI,
    STATE(8261), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337447] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11532), 1,
      anon_sym_RPAREN,
    STATE(8262), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337465] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11534), 1,
      anon_sym_SEMI,
    STATE(8263), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337483] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11536), 1,
      anon_sym_SEMI,
    STATE(8264), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337501] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11538), 1,
      anon_sym_SEMI,
    STATE(8265), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337519] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11540), 1,
      anon_sym_RBRACK,
    STATE(8266), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337537] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11542), 1,
      anon_sym_SEMI,
    STATE(8267), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337555] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11544), 1,
      anon_sym_SEMI,
    STATE(8268), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337573] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11546), 1,
      sym_VARIABLE,
    STATE(8269), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337591] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11548), 1,
      anon_sym_SEMI,
    STATE(8270), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337609] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11550), 1,
      sym_OF,
    STATE(8271), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337627] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11552), 1,
      sym_BEGIN,
    STATE(8272), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337645] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11554), 1,
      sym_OF,
    STATE(8273), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337663] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11556), 1,
      anon_sym_SEMI,
    STATE(8274), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337681] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11558), 1,
      anon_sym_SEMI,
    STATE(8275), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337699] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11560), 1,
      anon_sym_SEMI,
    STATE(8276), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337717] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11562), 1,
      sym_VARIABLE,
    STATE(8277), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337735] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11564), 1,
      anon_sym_DOT,
    STATE(8278), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337753] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11566), 1,
      sym_OF,
    STATE(8279), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337771] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11568), 1,
      anon_sym_RPAREN,
    STATE(8280), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337789] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11570), 1,
      sym_OF,
    STATE(8281), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337807] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11572), 1,
      anon_sym_SEMI,
    STATE(8282), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337825] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11574), 1,
      anon_sym_SEMI,
    STATE(8283), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337843] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11576), 1,
      anon_sym_RPAREN,
    STATE(8284), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337861] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11578), 1,
      sym_VARIABLE,
    STATE(8285), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337879] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11580), 1,
      sym__directive_newline,
    STATE(8286), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337897] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11582), 1,
      sym_OF,
    STATE(8287), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337915] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11584), 1,
      anon_sym_RBRACK,
    STATE(8288), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337933] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11586), 1,
      sym_OF,
    STATE(8289), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337951] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11588), 1,
      anon_sym_COLON,
    STATE(8290), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337969] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11590), 1,
      sym_THEN,
    STATE(8291), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [337987] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11592), 1,
      anon_sym_SEMI,
    STATE(8292), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338005] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11594), 1,
      sym_VARIABLE,
    STATE(8293), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338023] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11596), 1,
      anon_sym_RPAREN,
    STATE(8294), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338041] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11598), 1,
      sym_OF,
    STATE(8295), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338059] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7404), 1,
      sym_END,
    STATE(8296), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338077] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11600), 1,
      anon_sym_RBRACK,
    STATE(8297), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338095] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11602), 1,
      sym_VARIABLE,
    STATE(8298), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338113] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11604), 1,
      sym_bit_string_value,
    STATE(8299), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338131] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11606), 1,
      sym_OF,
    STATE(8300), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338149] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11608), 1,
      anon_sym_GT_GT,
    STATE(8301), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338167] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11610), 1,
      anon_sym_COLON,
    STATE(8302), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338185] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11612), 1,
      sym_VARIABLE,
    STATE(8303), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338203] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11614), 1,
      anon_sym_GT_GT,
    STATE(8304), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338221] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11616), 1,
      sym_OF,
    STATE(8305), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338239] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11618), 1,
      anon_sym_GT_GT,
    STATE(8306), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338257] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11620), 1,
      anon_sym_GT_GT,
    STATE(8307), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338275] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11622), 1,
      sym_OF,
    STATE(8308), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338293] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11624), 1,
      anon_sym_GT_GT,
    STATE(8309), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338311] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11626), 1,
      anon_sym_RBRACK,
    STATE(8310), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338329] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11628), 1,
      anon_sym_GT_GT,
    STATE(8311), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338347] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11630), 1,
      anon_sym_RBRACK,
    STATE(8312), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338365] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11632), 1,
      anon_sym_RPAREN,
    STATE(8313), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338383] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(5898), 1,
      sym_END,
    STATE(8314), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338401] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7065), 1,
      anon_sym_SEMI,
    STATE(8315), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338419] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11634), 1,
      sym_GENERATE,
    STATE(8316), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338437] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11636), 1,
      anon_sym_RPAREN,
    STATE(8317), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338455] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11638), 1,
      anon_sym_RBRACK,
    STATE(8318), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338473] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11640), 1,
      anon_sym_SEMI,
    STATE(8319), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338491] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11642), 1,
      anon_sym_RPAREN,
    STATE(8320), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338509] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11644), 1,
      anon_sym_RBRACK,
    STATE(8321), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338527] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11646), 1,
      anon_sym_RBRACK,
    STATE(8322), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338545] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11648), 1,
      anon_sym_SEMI,
    STATE(8323), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338563] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11650), 1,
      anon_sym_SEMI,
    STATE(8324), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338581] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11652), 1,
      anon_sym_GT_GT,
    STATE(8325), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338599] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11654), 1,
      anon_sym_GT_GT,
    STATE(8326), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338617] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11656), 1,
      anon_sym_GT_GT,
    STATE(8327), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338635] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11658), 1,
      anon_sym_SEMI,
    STATE(8328), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338653] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11660), 1,
      anon_sym_RBRACK,
    STATE(8329), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338671] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11662), 1,
      sym_MAP,
    STATE(8330), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338689] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11664), 1,
      anon_sym_RBRACK,
    STATE(8331), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338707] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11666), 1,
      anon_sym_SEMI,
    STATE(8332), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338725] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11668), 1,
      anon_sym_SEMI,
    STATE(8333), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338743] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11670), 1,
      anon_sym_SEMI,
    STATE(8334), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338761] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11672), 1,
      anon_sym_SEMI,
    STATE(8335), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338779] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11674), 1,
      anon_sym_RPAREN,
    STATE(8336), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338797] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11676), 1,
      anon_sym_SEMI,
    STATE(8337), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338815] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11678), 1,
      anon_sym_SEMI,
    STATE(8338), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338833] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11680), 1,
      anon_sym_RPAREN,
    STATE(8339), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338851] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11682), 1,
      sym_bit_string_value,
    STATE(8340), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338869] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11684), 1,
      anon_sym_RBRACK,
    STATE(8341), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338887] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11686), 1,
      anon_sym_SEMI,
    STATE(8342), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338905] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11688), 1,
      anon_sym_RPAREN,
    STATE(8343), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338923] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11690), 1,
      anon_sym_GT_GT,
    STATE(8344), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338941] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11692), 1,
      anon_sym_GT_GT,
    STATE(8345), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338959] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11694), 1,
      anon_sym_GT_GT,
    STATE(8346), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338977] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(7057), 1,
      anon_sym_SEMI,
    STATE(8347), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [338995] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11696), 1,
      anon_sym_RBRACK,
    STATE(8348), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339013] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11698), 1,
      anon_sym_SEMI,
    STATE(8349), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339031] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11700), 1,
      anon_sym_RBRACK,
    STATE(8350), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339049] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11702), 1,
      sym_bit_string_value,
    STATE(8351), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339067] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11704), 1,
      sym_VIEW,
    STATE(8352), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339085] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(8215), 1,
      sym_END,
    STATE(8353), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339103] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11706), 1,
      sym_bit_string_value,
    STATE(8354), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339121] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11708), 1,
      anon_sym_SEMI,
    STATE(8355), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339139] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11710), 1,
      anon_sym_SEMI,
    STATE(8356), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339157] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11712), 1,
      anon_sym_SEMI,
    STATE(8357), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339175] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11714), 1,
      anon_sym_SEMI,
    STATE(8358), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339193] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11716), 1,
      anon_sym_SEMI,
    STATE(8359), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339211] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11718), 1,
      anon_sym_DOT,
    STATE(8360), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339229] = 5,
    ACTIONS(3), 1,
      sym__grave_accent,
    ACTIONS(5), 1,
      sym__line_comment_start,
    ACTIONS(7), 1,
      sym__block_comment_start,
    ACTIONS(11720), 1,
      sym_MAP,
    STATE(8361), 3,
      sym__tool_directive,
      sym_line_comment,
      sym_block_comment,
  [339247] = 1,
    ACTIONS(11722), 1,
      ts_builtin_sym_end,
  [339251] = 1,
    ACTIONS(11724), 1,
      ts_builtin_sym_end,
  [339255] = 1,
    ACTIONS(11726), 1,
      ts_builtin_sym_end,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(366)] = 0,
  [SMALL_STATE(367)] = 131,
  [SMALL_STATE(368)] = 260,
  [SMALL_STATE(369)] = 389,
  [SMALL_STATE(370)] = 518,
  [SMALL_STATE(371)] = 649,
  [SMALL_STATE(372)] = 778,
  [SMALL_STATE(373)] = 907,
  [SMALL_STATE(374)] = 1036,
  [SMALL_STATE(375)] = 1167,
  [SMALL_STATE(376)] = 1298,
  [SMALL_STATE(377)] = 1427,
  [SMALL_STATE(378)] = 1558,
  [SMALL_STATE(379)] = 1687,
  [SMALL_STATE(380)] = 1816,
  [SMALL_STATE(381)] = 1945,
  [SMALL_STATE(382)] = 2074,
  [SMALL_STATE(383)] = 2205,
  [SMALL_STATE(384)] = 2334,
  [SMALL_STATE(385)] = 2463,
  [SMALL_STATE(386)] = 2594,
  [SMALL_STATE(387)] = 2725,
  [SMALL_STATE(388)] = 2856,
  [SMALL_STATE(389)] = 2985,
  [SMALL_STATE(390)] = 3114,
  [SMALL_STATE(391)] = 3245,
  [SMALL_STATE(392)] = 3376,
  [SMALL_STATE(393)] = 3502,
  [SMALL_STATE(394)] = 3628,
  [SMALL_STATE(395)] = 3754,
  [SMALL_STATE(396)] = 3880,
  [SMALL_STATE(397)] = 4006,
  [SMALL_STATE(398)] = 4132,
  [SMALL_STATE(399)] = 4258,
  [SMALL_STATE(400)] = 4384,
  [SMALL_STATE(401)] = 4510,
  [SMALL_STATE(402)] = 4650,
  [SMALL_STATE(403)] = 4774,
  [SMALL_STATE(404)] = 4900,
  [SMALL_STATE(405)] = 5026,
  [SMALL_STATE(406)] = 5152,
  [SMALL_STATE(407)] = 5278,
  [SMALL_STATE(408)] = 5404,
  [SMALL_STATE(409)] = 5530,
  [SMALL_STATE(410)] = 5670,
  [SMALL_STATE(411)] = 5796,
  [SMALL_STATE(412)] = 5936,
  [SMALL_STATE(413)] = 6076,
  [SMALL_STATE(414)] = 6197,
  [SMALL_STATE(415)] = 6318,
  [SMALL_STATE(416)] = 6441,
  [SMALL_STATE(417)] = 6564,
  [SMALL_STATE(418)] = 6687,
  [SMALL_STATE(419)] = 6808,
  [SMALL_STATE(420)] = 6929,
  [SMALL_STATE(421)] = 7052,
  [SMALL_STATE(422)] = 7175,
  [SMALL_STATE(423)] = 7298,
  [SMALL_STATE(424)] = 7419,
  [SMALL_STATE(425)] = 7540,
  [SMALL_STATE(426)] = 7663,
  [SMALL_STATE(427)] = 7786,
  [SMALL_STATE(428)] = 7909,
  [SMALL_STATE(429)] = 8030,
  [SMALL_STATE(430)] = 8151,
  [SMALL_STATE(431)] = 8274,
  [SMALL_STATE(432)] = 8397,
  [SMALL_STATE(433)] = 8520,
  [SMALL_STATE(434)] = 8641,
  [SMALL_STATE(435)] = 8762,
  [SMALL_STATE(436)] = 8885,
  [SMALL_STATE(437)] = 9008,
  [SMALL_STATE(438)] = 9131,
  [SMALL_STATE(439)] = 9254,
  [SMALL_STATE(440)] = 9375,
  [SMALL_STATE(441)] = 9498,
  [SMALL_STATE(442)] = 9621,
  [SMALL_STATE(443)] = 9744,
  [SMALL_STATE(444)] = 9865,
  [SMALL_STATE(445)] = 9986,
  [SMALL_STATE(446)] = 10109,
  [SMALL_STATE(447)] = 10232,
  [SMALL_STATE(448)] = 10355,
  [SMALL_STATE(449)] = 10476,
  [SMALL_STATE(450)] = 10597,
  [SMALL_STATE(451)] = 10720,
  [SMALL_STATE(452)] = 10843,
  [SMALL_STATE(453)] = 10966,
  [SMALL_STATE(454)] = 11087,
  [SMALL_STATE(455)] = 11208,
  [SMALL_STATE(456)] = 11331,
  [SMALL_STATE(457)] = 11454,
  [SMALL_STATE(458)] = 11577,
  [SMALL_STATE(459)] = 11698,
  [SMALL_STATE(460)] = 11819,
  [SMALL_STATE(461)] = 11942,
  [SMALL_STATE(462)] = 12065,
  [SMALL_STATE(463)] = 12188,
  [SMALL_STATE(464)] = 12309,
  [SMALL_STATE(465)] = 12430,
  [SMALL_STATE(466)] = 12553,
  [SMALL_STATE(467)] = 12676,
  [SMALL_STATE(468)] = 12799,
  [SMALL_STATE(469)] = 12920,
  [SMALL_STATE(470)] = 13041,
  [SMALL_STATE(471)] = 13164,
  [SMALL_STATE(472)] = 13287,
  [SMALL_STATE(473)] = 13410,
  [SMALL_STATE(474)] = 13531,
  [SMALL_STATE(475)] = 13652,
  [SMALL_STATE(476)] = 13775,
  [SMALL_STATE(477)] = 13898,
  [SMALL_STATE(478)] = 14021,
  [SMALL_STATE(479)] = 14142,
  [SMALL_STATE(480)] = 14263,
  [SMALL_STATE(481)] = 14386,
  [SMALL_STATE(482)] = 14509,
  [SMALL_STATE(483)] = 14632,
  [SMALL_STATE(484)] = 14753,
  [SMALL_STATE(485)] = 14874,
  [SMALL_STATE(486)] = 14997,
  [SMALL_STATE(487)] = 15120,
  [SMALL_STATE(488)] = 15243,
  [SMALL_STATE(489)] = 15364,
  [SMALL_STATE(490)] = 15485,
  [SMALL_STATE(491)] = 15608,
  [SMALL_STATE(492)] = 15731,
  [SMALL_STATE(493)] = 15854,
  [SMALL_STATE(494)] = 15975,
  [SMALL_STATE(495)] = 16096,
  [SMALL_STATE(496)] = 16219,
  [SMALL_STATE(497)] = 16342,
  [SMALL_STATE(498)] = 16465,
  [SMALL_STATE(499)] = 16588,
  [SMALL_STATE(500)] = 16709,
  [SMALL_STATE(501)] = 16830,
  [SMALL_STATE(502)] = 16953,
  [SMALL_STATE(503)] = 17076,
  [SMALL_STATE(504)] = 17199,
  [SMALL_STATE(505)] = 17320,
  [SMALL_STATE(506)] = 17441,
  [SMALL_STATE(507)] = 17564,
  [SMALL_STATE(508)] = 17687,
  [SMALL_STATE(509)] = 17810,
  [SMALL_STATE(510)] = 17931,
  [SMALL_STATE(511)] = 18052,
  [SMALL_STATE(512)] = 18175,
  [SMALL_STATE(513)] = 18298,
  [SMALL_STATE(514)] = 18421,
  [SMALL_STATE(515)] = 18542,
  [SMALL_STATE(516)] = 18663,
  [SMALL_STATE(517)] = 18786,
  [SMALL_STATE(518)] = 18909,
  [SMALL_STATE(519)] = 19032,
  [SMALL_STATE(520)] = 19153,
  [SMALL_STATE(521)] = 19274,
  [SMALL_STATE(522)] = 19397,
  [SMALL_STATE(523)] = 19520,
  [SMALL_STATE(524)] = 19643,
  [SMALL_STATE(525)] = 19764,
  [SMALL_STATE(526)] = 19885,
  [SMALL_STATE(527)] = 20008,
  [SMALL_STATE(528)] = 20131,
  [SMALL_STATE(529)] = 20254,
  [SMALL_STATE(530)] = 20377,
  [SMALL_STATE(531)] = 20500,
  [SMALL_STATE(532)] = 20625,
  [SMALL_STATE(533)] = 20748,
  [SMALL_STATE(534)] = 20871,
  [SMALL_STATE(535)] = 20994,
  [SMALL_STATE(536)] = 21117,
  [SMALL_STATE(537)] = 21240,
  [SMALL_STATE(538)] = 21363,
  [SMALL_STATE(539)] = 21488,
  [SMALL_STATE(540)] = 21611,
  [SMALL_STATE(541)] = 21734,
  [SMALL_STATE(542)] = 21857,
  [SMALL_STATE(543)] = 21980,
  [SMALL_STATE(544)] = 22103,
  [SMALL_STATE(545)] = 22228,
  [SMALL_STATE(546)] = 22351,
  [SMALL_STATE(547)] = 22474,
  [SMALL_STATE(548)] = 22597,
  [SMALL_STATE(549)] = 22720,
  [SMALL_STATE(550)] = 22845,
  [SMALL_STATE(551)] = 22968,
  [SMALL_STATE(552)] = 23091,
  [SMALL_STATE(553)] = 23214,
  [SMALL_STATE(554)] = 23337,
  [SMALL_STATE(555)] = 23462,
  [SMALL_STATE(556)] = 23585,
  [SMALL_STATE(557)] = 23708,
  [SMALL_STATE(558)] = 23831,
  [SMALL_STATE(559)] = 23954,
  [SMALL_STATE(560)] = 24079,
  [SMALL_STATE(561)] = 24202,
  [SMALL_STATE(562)] = 24325,
  [SMALL_STATE(563)] = 24448,
  [SMALL_STATE(564)] = 24571,
  [SMALL_STATE(565)] = 24696,
  [SMALL_STATE(566)] = 24819,
  [SMALL_STATE(567)] = 24942,
  [SMALL_STATE(568)] = 25065,
  [SMALL_STATE(569)] = 25188,
  [SMALL_STATE(570)] = 25313,
  [SMALL_STATE(571)] = 25436,
  [SMALL_STATE(572)] = 25559,
  [SMALL_STATE(573)] = 25682,
  [SMALL_STATE(574)] = 25805,
  [SMALL_STATE(575)] = 25930,
  [SMALL_STATE(576)] = 26053,
  [SMALL_STATE(577)] = 26176,
  [SMALL_STATE(578)] = 26299,
  [SMALL_STATE(579)] = 26422,
  [SMALL_STATE(580)] = 26545,
  [SMALL_STATE(581)] = 26668,
  [SMALL_STATE(582)] = 26791,
  [SMALL_STATE(583)] = 26914,
  [SMALL_STATE(584)] = 27037,
  [SMALL_STATE(585)] = 27160,
  [SMALL_STATE(586)] = 27283,
  [SMALL_STATE(587)] = 27406,
  [SMALL_STATE(588)] = 27529,
  [SMALL_STATE(589)] = 27652,
  [SMALL_STATE(590)] = 27775,
  [SMALL_STATE(591)] = 27898,
  [SMALL_STATE(592)] = 28021,
  [SMALL_STATE(593)] = 28144,
  [SMALL_STATE(594)] = 28267,
  [SMALL_STATE(595)] = 28390,
  [SMALL_STATE(596)] = 28513,
  [SMALL_STATE(597)] = 28636,
  [SMALL_STATE(598)] = 28759,
  [SMALL_STATE(599)] = 28882,
  [SMALL_STATE(600)] = 29005,
  [SMALL_STATE(601)] = 29128,
  [SMALL_STATE(602)] = 29251,
  [SMALL_STATE(603)] = 29374,
  [SMALL_STATE(604)] = 29497,
  [SMALL_STATE(605)] = 29620,
  [SMALL_STATE(606)] = 29743,
  [SMALL_STATE(607)] = 29866,
  [SMALL_STATE(608)] = 29989,
  [SMALL_STATE(609)] = 30112,
  [SMALL_STATE(610)] = 30235,
  [SMALL_STATE(611)] = 30358,
  [SMALL_STATE(612)] = 30481,
  [SMALL_STATE(613)] = 30604,
  [SMALL_STATE(614)] = 30727,
  [SMALL_STATE(615)] = 30850,
  [SMALL_STATE(616)] = 30973,
  [SMALL_STATE(617)] = 31096,
  [SMALL_STATE(618)] = 31219,
  [SMALL_STATE(619)] = 31342,
  [SMALL_STATE(620)] = 31465,
  [SMALL_STATE(621)] = 31588,
  [SMALL_STATE(622)] = 31711,
  [SMALL_STATE(623)] = 31834,
  [SMALL_STATE(624)] = 31957,
  [SMALL_STATE(625)] = 32080,
  [SMALL_STATE(626)] = 32203,
  [SMALL_STATE(627)] = 32326,
  [SMALL_STATE(628)] = 32449,
  [SMALL_STATE(629)] = 32572,
  [SMALL_STATE(630)] = 32695,
  [SMALL_STATE(631)] = 32818,
  [SMALL_STATE(632)] = 32941,
  [SMALL_STATE(633)] = 33064,
  [SMALL_STATE(634)] = 33187,
  [SMALL_STATE(635)] = 33310,
  [SMALL_STATE(636)] = 33433,
  [SMALL_STATE(637)] = 33556,
  [SMALL_STATE(638)] = 33679,
  [SMALL_STATE(639)] = 33802,
  [SMALL_STATE(640)] = 33925,
  [SMALL_STATE(641)] = 34048,
  [SMALL_STATE(642)] = 34171,
  [SMALL_STATE(643)] = 34294,
  [SMALL_STATE(644)] = 34417,
  [SMALL_STATE(645)] = 34540,
  [SMALL_STATE(646)] = 34663,
  [SMALL_STATE(647)] = 34786,
  [SMALL_STATE(648)] = 34909,
  [SMALL_STATE(649)] = 35032,
  [SMALL_STATE(650)] = 35155,
  [SMALL_STATE(651)] = 35278,
  [SMALL_STATE(652)] = 35403,
  [SMALL_STATE(653)] = 35526,
  [SMALL_STATE(654)] = 35649,
  [SMALL_STATE(655)] = 35772,
  [SMALL_STATE(656)] = 35895,
  [SMALL_STATE(657)] = 36018,
  [SMALL_STATE(658)] = 36141,
  [SMALL_STATE(659)] = 36264,
  [SMALL_STATE(660)] = 36387,
  [SMALL_STATE(661)] = 36510,
  [SMALL_STATE(662)] = 36631,
  [SMALL_STATE(663)] = 36754,
  [SMALL_STATE(664)] = 36877,
  [SMALL_STATE(665)] = 37000,
  [SMALL_STATE(666)] = 37123,
  [SMALL_STATE(667)] = 37246,
  [SMALL_STATE(668)] = 37369,
  [SMALL_STATE(669)] = 37492,
  [SMALL_STATE(670)] = 37615,
  [SMALL_STATE(671)] = 37738,
  [SMALL_STATE(672)] = 37861,
  [SMALL_STATE(673)] = 37984,
  [SMALL_STATE(674)] = 38105,
  [SMALL_STATE(675)] = 38226,
  [SMALL_STATE(676)] = 38349,
  [SMALL_STATE(677)] = 38472,
  [SMALL_STATE(678)] = 38595,
  [SMALL_STATE(679)] = 38718,
  [SMALL_STATE(680)] = 38841,
  [SMALL_STATE(681)] = 38964,
  [SMALL_STATE(682)] = 39087,
  [SMALL_STATE(683)] = 39210,
  [SMALL_STATE(684)] = 39333,
  [SMALL_STATE(685)] = 39456,
  [SMALL_STATE(686)] = 39579,
  [SMALL_STATE(687)] = 39700,
  [SMALL_STATE(688)] = 39821,
  [SMALL_STATE(689)] = 39944,
  [SMALL_STATE(690)] = 40067,
  [SMALL_STATE(691)] = 40190,
  [SMALL_STATE(692)] = 40313,
  [SMALL_STATE(693)] = 40436,
  [SMALL_STATE(694)] = 40559,
  [SMALL_STATE(695)] = 40682,
  [SMALL_STATE(696)] = 40803,
  [SMALL_STATE(697)] = 40924,
  [SMALL_STATE(698)] = 41047,
  [SMALL_STATE(699)] = 41170,
  [SMALL_STATE(700)] = 41293,
  [SMALL_STATE(701)] = 41416,
  [SMALL_STATE(702)] = 41539,
  [SMALL_STATE(703)] = 41662,
  [SMALL_STATE(704)] = 41785,
  [SMALL_STATE(705)] = 41906,
  [SMALL_STATE(706)] = 42027,
  [SMALL_STATE(707)] = 42150,
  [SMALL_STATE(708)] = 42273,
  [SMALL_STATE(709)] = 42396,
  [SMALL_STATE(710)] = 42517,
  [SMALL_STATE(711)] = 42638,
  [SMALL_STATE(712)] = 42761,
  [SMALL_STATE(713)] = 42884,
  [SMALL_STATE(714)] = 43007,
  [SMALL_STATE(715)] = 43128,
  [SMALL_STATE(716)] = 43249,
  [SMALL_STATE(717)] = 43372,
  [SMALL_STATE(718)] = 43495,
  [SMALL_STATE(719)] = 43618,
  [SMALL_STATE(720)] = 43739,
  [SMALL_STATE(721)] = 43860,
  [SMALL_STATE(722)] = 43954,
  [SMALL_STATE(723)] = 44088,
  [SMALL_STATE(724)] = 44182,
  [SMALL_STATE(725)] = 44316,
  [SMALL_STATE(726)] = 44408,
  [SMALL_STATE(727)] = 44542,
  [SMALL_STATE(728)] = 44676,
  [SMALL_STATE(729)] = 44767,
  [SMALL_STATE(730)] = 44860,
  [SMALL_STATE(731)] = 44953,
  [SMALL_STATE(732)] = 45043,
  [SMALL_STATE(733)] = 45135,
  [SMALL_STATE(734)] = 45227,
  [SMALL_STATE(735)] = 45319,
  [SMALL_STATE(736)] = 45411,
  [SMALL_STATE(737)] = 45501,
  [SMALL_STATE(738)] = 45626,
  [SMALL_STATE(739)] = 45715,
  [SMALL_STATE(740)] = 45804,
  [SMALL_STATE(741)] = 45897,
  [SMALL_STATE(742)] = 45988,
  [SMALL_STATE(743)] = 46077,
  [SMALL_STATE(744)] = 46168,
  [SMALL_STATE(745)] = 46293,
  [SMALL_STATE(746)] = 46418,
  [SMALL_STATE(747)] = 46509,
  [SMALL_STATE(748)] = 46600,
  [SMALL_STATE(749)] = 46691,
  [SMALL_STATE(750)] = 46782,
  [SMALL_STATE(751)] = 46872,
  [SMALL_STATE(752)] = 46962,
  [SMALL_STATE(753)] = 47052,
  [SMALL_STATE(754)] = 47140,
  [SMALL_STATE(755)] = 47228,
  [SMALL_STATE(756)] = 47318,
  [SMALL_STATE(757)] = 47405,
  [SMALL_STATE(758)] = 47524,
  [SMALL_STATE(759)] = 47611,
  [SMALL_STATE(760)] = 47698,
  [SMALL_STATE(761)] = 47785,
  [SMALL_STATE(762)] = 47904,
  [SMALL_STATE(763)] = 48023,
  [SMALL_STATE(764)] = 48110,
  [SMALL_STATE(765)] = 48199,
  [SMALL_STATE(766)] = 48318,
  [SMALL_STATE(767)] = 48405,
  [SMALL_STATE(768)] = 48494,
  [SMALL_STATE(769)] = 48583,
  [SMALL_STATE(770)] = 48672,
  [SMALL_STATE(771)] = 48761,
  [SMALL_STATE(772)] = 48850,
  [SMALL_STATE(773)] = 48939,
  [SMALL_STATE(774)] = 49058,
  [SMALL_STATE(775)] = 49145,
  [SMALL_STATE(776)] = 49264,
  [SMALL_STATE(777)] = 49383,
  [SMALL_STATE(778)] = 49472,
  [SMALL_STATE(779)] = 49561,
  [SMALL_STATE(780)] = 49650,
  [SMALL_STATE(781)] = 49738,
  [SMALL_STATE(782)] = 49826,
  [SMALL_STATE(783)] = 49912,
  [SMALL_STATE(784)] = 49998,
  [SMALL_STATE(785)] = 50116,
  [SMALL_STATE(786)] = 50202,
  [SMALL_STATE(787)] = 50290,
  [SMALL_STATE(788)] = 50376,
  [SMALL_STATE(789)] = 50464,
  [SMALL_STATE(790)] = 50552,
  [SMALL_STATE(791)] = 50640,
  [SMALL_STATE(792)] = 50726,
  [SMALL_STATE(793)] = 50814,
  [SMALL_STATE(794)] = 50902,
  [SMALL_STATE(795)] = 50990,
  [SMALL_STATE(796)] = 51078,
  [SMALL_STATE(797)] = 51191,
  [SMALL_STATE(798)] = 51276,
  [SMALL_STATE(799)] = 51361,
  [SMALL_STATE(800)] = 51446,
  [SMALL_STATE(801)] = 51531,
  [SMALL_STATE(802)] = 51644,
  [SMALL_STATE(803)] = 51757,
  [SMALL_STATE(804)] = 51844,
  [SMALL_STATE(805)] = 51957,
  [SMALL_STATE(806)] = 52030,
  [SMALL_STATE(807)] = 52117,
  [SMALL_STATE(808)] = 52230,
  [SMALL_STATE(809)] = 52317,
  [SMALL_STATE(810)] = 52404,
  [SMALL_STATE(811)] = 52517,
  [SMALL_STATE(812)] = 52630,
  [SMALL_STATE(813)] = 52717,
  [SMALL_STATE(814)] = 52804,
  [SMALL_STATE(815)] = 52917,
  [SMALL_STATE(816)] = 53030,
  [SMALL_STATE(817)] = 53149,
  [SMALL_STATE(818)] = 53262,
  [SMALL_STATE(819)] = 53375,
  [SMALL_STATE(820)] = 53488,
  [SMALL_STATE(821)] = 53575,
  [SMALL_STATE(822)] = 53688,
  [SMALL_STATE(823)] = 53773,
  [SMALL_STATE(824)] = 53858,
  [SMALL_STATE(825)] = 53969,
  [SMALL_STATE(826)] = 54082,
  [SMALL_STATE(827)] = 54169,
  [SMALL_STATE(828)] = 54288,
  [SMALL_STATE(829)] = 54373,
  [SMALL_STATE(830)] = 54486,
  [SMALL_STATE(831)] = 54573,
  [SMALL_STATE(832)] = 54692,
  [SMALL_STATE(833)] = 54779,
  [SMALL_STATE(834)] = 54898,
  [SMALL_STATE(835)] = 55017,
  [SMALL_STATE(836)] = 55136,
  [SMALL_STATE(837)] = 55249,
  [SMALL_STATE(838)] = 55335,
  [SMALL_STATE(839)] = 55401,
  [SMALL_STATE(840)] = 55467,
  [SMALL_STATE(841)] = 55533,
  [SMALL_STATE(842)] = 55617,
  [SMALL_STATE(843)] = 55703,
  [SMALL_STATE(844)] = 55789,
  [SMALL_STATE(845)] = 55855,
  [SMALL_STATE(846)] = 55941,
  [SMALL_STATE(847)] = 56013,
  [SMALL_STATE(848)] = 56079,
  [SMALL_STATE(849)] = 56145,
  [SMALL_STATE(850)] = 56231,
  [SMALL_STATE(851)] = 56303,
  [SMALL_STATE(852)] = 56387,
  [SMALL_STATE(853)] = 56453,
  [SMALL_STATE(854)] = 56539,
  [SMALL_STATE(855)] = 56605,
  [SMALL_STATE(856)] = 56671,
  [SMALL_STATE(857)] = 56737,
  [SMALL_STATE(858)] = 56823,
  [SMALL_STATE(859)] = 56889,
  [SMALL_STATE(860)] = 56955,
  [SMALL_STATE(861)] = 57021,
  [SMALL_STATE(862)] = 57133,
  [SMALL_STATE(863)] = 57217,
  [SMALL_STATE(864)] = 57283,
  [SMALL_STATE(865)] = 57349,
  [SMALL_STATE(866)] = 57415,
  [SMALL_STATE(867)] = 57481,
  [SMALL_STATE(868)] = 57547,
  [SMALL_STATE(869)] = 57633,
  [SMALL_STATE(870)] = 57699,
  [SMALL_STATE(871)] = 57785,
  [SMALL_STATE(872)] = 57851,
  [SMALL_STATE(873)] = 57935,
  [SMALL_STATE(874)] = 58021,
  [SMALL_STATE(875)] = 58087,
  [SMALL_STATE(876)] = 58153,
  [SMALL_STATE(877)] = 58237,
  [SMALL_STATE(878)] = 58303,
  [SMALL_STATE(879)] = 58369,
  [SMALL_STATE(880)] = 58435,
  [SMALL_STATE(881)] = 58501,
  [SMALL_STATE(882)] = 58567,
  [SMALL_STATE(883)] = 58651,
  [SMALL_STATE(884)] = 58737,
  [SMALL_STATE(885)] = 58823,
  [SMALL_STATE(886)] = 58889,
  [SMALL_STATE(887)] = 58955,
  [SMALL_STATE(888)] = 59020,
  [SMALL_STATE(889)] = 59091,
  [SMALL_STATE(890)] = 59162,
  [SMALL_STATE(891)] = 59227,
  [SMALL_STATE(892)] = 59292,
  [SMALL_STATE(893)] = 59357,
  [SMALL_STATE(894)] = 59422,
  [SMALL_STATE(895)] = 59487,
  [SMALL_STATE(896)] = 59576,
  [SMALL_STATE(897)] = 59641,
  [SMALL_STATE(898)] = 59706,
  [SMALL_STATE(899)] = 59793,
  [SMALL_STATE(900)] = 59858,
  [SMALL_STATE(901)] = 59923,
  [SMALL_STATE(902)] = 60000,
  [SMALL_STATE(903)] = 60065,
  [SMALL_STATE(904)] = 60130,
  [SMALL_STATE(905)] = 60195,
  [SMALL_STATE(906)] = 60260,
  [SMALL_STATE(907)] = 60325,
  [SMALL_STATE(908)] = 60390,
  [SMALL_STATE(909)] = 60455,
  [SMALL_STATE(910)] = 60536,
  [SMALL_STATE(911)] = 60601,
  [SMALL_STATE(912)] = 60666,
  [SMALL_STATE(913)] = 60731,
  [SMALL_STATE(914)] = 60796,
  [SMALL_STATE(915)] = 60861,
  [SMALL_STATE(916)] = 60926,
  [SMALL_STATE(917)] = 61001,
  [SMALL_STATE(918)] = 61066,
  [SMALL_STATE(919)] = 61141,
  [SMALL_STATE(920)] = 61226,
  [SMALL_STATE(921)] = 61307,
  [SMALL_STATE(922)] = 61372,
  [SMALL_STATE(923)] = 61437,
  [SMALL_STATE(924)] = 61502,
  [SMALL_STATE(925)] = 61567,
  [SMALL_STATE(926)] = 61632,
  [SMALL_STATE(927)] = 61697,
  [SMALL_STATE(928)] = 61780,
  [SMALL_STATE(929)] = 61845,
  [SMALL_STATE(930)] = 61909,
  [SMALL_STATE(931)] = 61973,
  [SMALL_STATE(932)] = 62037,
  [SMALL_STATE(933)] = 62101,
  [SMALL_STATE(934)] = 62165,
  [SMALL_STATE(935)] = 62229,
  [SMALL_STATE(936)] = 62293,
  [SMALL_STATE(937)] = 62357,
  [SMALL_STATE(938)] = 62421,
  [SMALL_STATE(939)] = 62485,
  [SMALL_STATE(940)] = 62549,
  [SMALL_STATE(941)] = 62613,
  [SMALL_STATE(942)] = 62719,
  [SMALL_STATE(943)] = 62783,
  [SMALL_STATE(944)] = 62847,
  [SMALL_STATE(945)] = 62953,
  [SMALL_STATE(946)] = 63057,
  [SMALL_STATE(947)] = 63121,
  [SMALL_STATE(948)] = 63185,
  [SMALL_STATE(949)] = 63249,
  [SMALL_STATE(950)] = 63313,
  [SMALL_STATE(951)] = 63377,
  [SMALL_STATE(952)] = 63441,
  [SMALL_STATE(953)] = 63505,
  [SMALL_STATE(954)] = 63569,
  [SMALL_STATE(955)] = 63633,
  [SMALL_STATE(956)] = 63697,
  [SMALL_STATE(957)] = 63767,
  [SMALL_STATE(958)] = 63831,
  [SMALL_STATE(959)] = 63937,
  [SMALL_STATE(960)] = 64001,
  [SMALL_STATE(961)] = 64107,
  [SMALL_STATE(962)] = 64213,
  [SMALL_STATE(963)] = 64277,
  [SMALL_STATE(964)] = 64341,
  [SMALL_STATE(965)] = 64405,
  [SMALL_STATE(966)] = 64475,
  [SMALL_STATE(967)] = 64539,
  [SMALL_STATE(968)] = 64603,
  [SMALL_STATE(969)] = 64667,
  [SMALL_STATE(970)] = 64731,
  [SMALL_STATE(971)] = 64795,
  [SMALL_STATE(972)] = 64859,
  [SMALL_STATE(973)] = 64923,
  [SMALL_STATE(974)] = 64987,
  [SMALL_STATE(975)] = 65051,
  [SMALL_STATE(976)] = 65115,
  [SMALL_STATE(977)] = 65179,
  [SMALL_STATE(978)] = 65243,
  [SMALL_STATE(979)] = 65307,
  [SMALL_STATE(980)] = 65371,
  [SMALL_STATE(981)] = 65435,
  [SMALL_STATE(982)] = 65499,
  [SMALL_STATE(983)] = 65563,
  [SMALL_STATE(984)] = 65627,
  [SMALL_STATE(985)] = 65691,
  [SMALL_STATE(986)] = 65755,
  [SMALL_STATE(987)] = 65819,
  [SMALL_STATE(988)] = 65883,
  [SMALL_STATE(989)] = 65947,
  [SMALL_STATE(990)] = 66011,
  [SMALL_STATE(991)] = 66075,
  [SMALL_STATE(992)] = 66139,
  [SMALL_STATE(993)] = 66209,
  [SMALL_STATE(994)] = 66273,
  [SMALL_STATE(995)] = 66337,
  [SMALL_STATE(996)] = 66407,
  [SMALL_STATE(997)] = 66471,
  [SMALL_STATE(998)] = 66541,
  [SMALL_STATE(999)] = 66614,
  [SMALL_STATE(1000)] = 66677,
  [SMALL_STATE(1001)] = 66740,
  [SMALL_STATE(1002)] = 66803,
  [SMALL_STATE(1003)] = 66866,
  [SMALL_STATE(1004)] = 66929,
  [SMALL_STATE(1005)] = 66992,
  [SMALL_STATE(1006)] = 67055,
  [SMALL_STATE(1007)] = 67118,
  [SMALL_STATE(1008)] = 67181,
  [SMALL_STATE(1009)] = 67244,
  [SMALL_STATE(1010)] = 67307,
  [SMALL_STATE(1011)] = 67386,
  [SMALL_STATE(1012)] = 67449,
  [SMALL_STATE(1013)] = 67512,
  [SMALL_STATE(1014)] = 67575,
  [SMALL_STATE(1015)] = 67638,
  [SMALL_STATE(1016)] = 67701,
  [SMALL_STATE(1017)] = 67764,
  [SMALL_STATE(1018)] = 67827,
  [SMALL_STATE(1019)] = 67896,
  [SMALL_STATE(1020)] = 67967,
  [SMALL_STATE(1021)] = 68030,
  [SMALL_STATE(1022)] = 68093,
  [SMALL_STATE(1023)] = 68180,
  [SMALL_STATE(1024)] = 68243,
  [SMALL_STATE(1025)] = 68306,
  [SMALL_STATE(1026)] = 68413,
  [SMALL_STATE(1027)] = 68520,
  [SMALL_STATE(1028)] = 68583,
  [SMALL_STATE(1029)] = 68646,
  [SMALL_STATE(1030)] = 68709,
  [SMALL_STATE(1031)] = 68772,
  [SMALL_STATE(1032)] = 68835,
  [SMALL_STATE(1033)] = 68898,
  [SMALL_STATE(1034)] = 68961,
  [SMALL_STATE(1035)] = 69024,
  [SMALL_STATE(1036)] = 69087,
  [SMALL_STATE(1037)] = 69150,
  [SMALL_STATE(1038)] = 69213,
  [SMALL_STATE(1039)] = 69276,
  [SMALL_STATE(1040)] = 69361,
  [SMALL_STATE(1041)] = 69444,
  [SMALL_STATE(1042)] = 69553,
  [SMALL_STATE(1043)] = 69616,
  [SMALL_STATE(1044)] = 69725,
  [SMALL_STATE(1045)] = 69788,
  [SMALL_STATE(1046)] = 69851,
  [SMALL_STATE(1047)] = 69914,
  [SMALL_STATE(1048)] = 69977,
  [SMALL_STATE(1049)] = 70040,
  [SMALL_STATE(1050)] = 70109,
  [SMALL_STATE(1051)] = 70196,
  [SMALL_STATE(1052)] = 70269,
  [SMALL_STATE(1053)] = 70348,
  [SMALL_STATE(1054)] = 70421,
  [SMALL_STATE(1055)] = 70500,
  [SMALL_STATE(1056)] = 70575,
  [SMALL_STATE(1057)] = 70644,
  [SMALL_STATE(1058)] = 70725,
  [SMALL_STATE(1059)] = 70812,
  [SMALL_STATE(1060)] = 70891,
  [SMALL_STATE(1061)] = 70960,
  [SMALL_STATE(1062)] = 71023,
  [SMALL_STATE(1063)] = 71086,
  [SMALL_STATE(1064)] = 71149,
  [SMALL_STATE(1065)] = 71224,
  [SMALL_STATE(1066)] = 71309,
  [SMALL_STATE(1067)] = 71392,
  [SMALL_STATE(1068)] = 71473,
  [SMALL_STATE(1069)] = 71536,
  [SMALL_STATE(1070)] = 71599,
  [SMALL_STATE(1071)] = 71662,
  [SMALL_STATE(1072)] = 71731,
  [SMALL_STATE(1073)] = 71838,
  [SMALL_STATE(1074)] = 71901,
  [SMALL_STATE(1075)] = 71964,
  [SMALL_STATE(1076)] = 72069,
  [SMALL_STATE(1077)] = 72142,
  [SMALL_STATE(1078)] = 72205,
  [SMALL_STATE(1079)] = 72268,
  [SMALL_STATE(1080)] = 72331,
  [SMALL_STATE(1081)] = 72394,
  [SMALL_STATE(1082)] = 72457,
  [SMALL_STATE(1083)] = 72520,
  [SMALL_STATE(1084)] = 72583,
  [SMALL_STATE(1085)] = 72646,
  [SMALL_STATE(1086)] = 72709,
  [SMALL_STATE(1087)] = 72772,
  [SMALL_STATE(1088)] = 72835,
  [SMALL_STATE(1089)] = 72898,
  [SMALL_STATE(1090)] = 72961,
  [SMALL_STATE(1091)] = 73024,
  [SMALL_STATE(1092)] = 73087,
  [SMALL_STATE(1093)] = 73150,
  [SMALL_STATE(1094)] = 73213,
  [SMALL_STATE(1095)] = 73276,
  [SMALL_STATE(1096)] = 73339,
  [SMALL_STATE(1097)] = 73402,
  [SMALL_STATE(1098)] = 73465,
  [SMALL_STATE(1099)] = 73528,
  [SMALL_STATE(1100)] = 73591,
  [SMALL_STATE(1101)] = 73654,
  [SMALL_STATE(1102)] = 73717,
  [SMALL_STATE(1103)] = 73812,
  [SMALL_STATE(1104)] = 73875,
  [SMALL_STATE(1105)] = 73938,
  [SMALL_STATE(1106)] = 74001,
  [SMALL_STATE(1107)] = 74064,
  [SMALL_STATE(1108)] = 74127,
  [SMALL_STATE(1109)] = 74190,
  [SMALL_STATE(1110)] = 74253,
  [SMALL_STATE(1111)] = 74316,
  [SMALL_STATE(1112)] = 74379,
  [SMALL_STATE(1113)] = 74442,
  [SMALL_STATE(1114)] = 74505,
  [SMALL_STATE(1115)] = 74568,
  [SMALL_STATE(1116)] = 74631,
  [SMALL_STATE(1117)] = 74694,
  [SMALL_STATE(1118)] = 74757,
  [SMALL_STATE(1119)] = 74820,
  [SMALL_STATE(1120)] = 74883,
  [SMALL_STATE(1121)] = 74946,
  [SMALL_STATE(1122)] = 75014,
  [SMALL_STATE(1123)] = 75076,
  [SMALL_STATE(1124)] = 75160,
  [SMALL_STATE(1125)] = 75242,
  [SMALL_STATE(1126)] = 75304,
  [SMALL_STATE(1127)] = 75366,
  [SMALL_STATE(1128)] = 75428,
  [SMALL_STATE(1129)] = 75506,
  [SMALL_STATE(1130)] = 75568,
  [SMALL_STATE(1131)] = 75630,
  [SMALL_STATE(1132)] = 75692,
  [SMALL_STATE(1133)] = 75754,
  [SMALL_STATE(1134)] = 75816,
  [SMALL_STATE(1135)] = 75898,
  [SMALL_STATE(1136)] = 75970,
  [SMALL_STATE(1137)] = 76048,
  [SMALL_STATE(1138)] = 76128,
  [SMALL_STATE(1139)] = 76196,
  [SMALL_STATE(1140)] = 76276,
  [SMALL_STATE(1141)] = 76344,
  [SMALL_STATE(1142)] = 76406,
  [SMALL_STATE(1143)] = 76468,
  [SMALL_STATE(1144)] = 76530,
  [SMALL_STATE(1145)] = 76592,
  [SMALL_STATE(1146)] = 76654,
  [SMALL_STATE(1147)] = 76726,
  [SMALL_STATE(1148)] = 76834,
  [SMALL_STATE(1149)] = 76896,
  [SMALL_STATE(1150)] = 76958,
  [SMALL_STATE(1151)] = 77036,
  [SMALL_STATE(1152)] = 77110,
  [SMALL_STATE(1153)] = 77172,
  [SMALL_STATE(1154)] = 77280,
  [SMALL_STATE(1155)] = 77342,
  [SMALL_STATE(1156)] = 77404,
  [SMALL_STATE(1157)] = 77466,
  [SMALL_STATE(1158)] = 77528,
  [SMALL_STATE(1159)] = 77590,
  [SMALL_STATE(1160)] = 77652,
  [SMALL_STATE(1161)] = 77724,
  [SMALL_STATE(1162)] = 77792,
  [SMALL_STATE(1163)] = 77878,
  [SMALL_STATE(1164)] = 77940,
  [SMALL_STATE(1165)] = 78012,
  [SMALL_STATE(1166)] = 78084,
  [SMALL_STATE(1167)] = 78146,
  [SMALL_STATE(1168)] = 78208,
  [SMALL_STATE(1169)] = 78286,
  [SMALL_STATE(1170)] = 78348,
  [SMALL_STATE(1171)] = 78410,
  [SMALL_STATE(1172)] = 78472,
  [SMALL_STATE(1173)] = 78540,
  [SMALL_STATE(1174)] = 78602,
  [SMALL_STATE(1175)] = 78664,
  [SMALL_STATE(1176)] = 78726,
  [SMALL_STATE(1177)] = 78788,
  [SMALL_STATE(1178)] = 78850,
  [SMALL_STATE(1179)] = 78912,
  [SMALL_STATE(1180)] = 78974,
  [SMALL_STATE(1181)] = 79048,
  [SMALL_STATE(1182)] = 79110,
  [SMALL_STATE(1183)] = 79218,
  [SMALL_STATE(1184)] = 79326,
  [SMALL_STATE(1185)] = 79388,
  [SMALL_STATE(1186)] = 79450,
  [SMALL_STATE(1187)] = 79512,
  [SMALL_STATE(1188)] = 79574,
  [SMALL_STATE(1189)] = 79636,
  [SMALL_STATE(1190)] = 79700,
  [SMALL_STATE(1191)] = 79762,
  [SMALL_STATE(1192)] = 79824,
  [SMALL_STATE(1193)] = 79886,
  [SMALL_STATE(1194)] = 79948,
  [SMALL_STATE(1195)] = 80010,
  [SMALL_STATE(1196)] = 80078,
  [SMALL_STATE(1197)] = 80140,
  [SMALL_STATE(1198)] = 80202,
  [SMALL_STATE(1199)] = 80288,
  [SMALL_STATE(1200)] = 80350,
  [SMALL_STATE(1201)] = 80418,
  [SMALL_STATE(1202)] = 80510,
  [SMALL_STATE(1203)] = 80572,
  [SMALL_STATE(1204)] = 80634,
  [SMALL_STATE(1205)] = 80696,
  [SMALL_STATE(1206)] = 80768,
  [SMALL_STATE(1207)] = 80830,
  [SMALL_STATE(1208)] = 80892,
  [SMALL_STATE(1209)] = 81000,
  [SMALL_STATE(1210)] = 81062,
  [SMALL_STATE(1211)] = 81124,
  [SMALL_STATE(1212)] = 81208,
  [SMALL_STATE(1213)] = 81316,
  [SMALL_STATE(1214)] = 81424,
  [SMALL_STATE(1215)] = 81508,
  [SMALL_STATE(1216)] = 81590,
  [SMALL_STATE(1217)] = 81670,
  [SMALL_STATE(1218)] = 81778,
  [SMALL_STATE(1219)] = 81864,
  [SMALL_STATE(1220)] = 81972,
  [SMALL_STATE(1221)] = 82080,
  [SMALL_STATE(1222)] = 82158,
  [SMALL_STATE(1223)] = 82220,
  [SMALL_STATE(1224)] = 82298,
  [SMALL_STATE(1225)] = 82406,
  [SMALL_STATE(1226)] = 82514,
  [SMALL_STATE(1227)] = 82622,
  [SMALL_STATE(1228)] = 82684,
  [SMALL_STATE(1229)] = 82758,
  [SMALL_STATE(1230)] = 82820,
  [SMALL_STATE(1231)] = 82928,
  [SMALL_STATE(1232)] = 83036,
  [SMALL_STATE(1233)] = 83144,
  [SMALL_STATE(1234)] = 83206,
  [SMALL_STATE(1235)] = 83314,
  [SMALL_STATE(1236)] = 83376,
  [SMALL_STATE(1237)] = 83438,
  [SMALL_STATE(1238)] = 83500,
  [SMALL_STATE(1239)] = 83561,
  [SMALL_STATE(1240)] = 83622,
  [SMALL_STATE(1241)] = 83683,
  [SMALL_STATE(1242)] = 83744,
  [SMALL_STATE(1243)] = 83805,
  [SMALL_STATE(1244)] = 83866,
  [SMALL_STATE(1245)] = 83927,
  [SMALL_STATE(1246)] = 83988,
  [SMALL_STATE(1247)] = 84049,
  [SMALL_STATE(1248)] = 84110,
  [SMALL_STATE(1249)] = 84171,
  [SMALL_STATE(1250)] = 84232,
  [SMALL_STATE(1251)] = 84293,
  [SMALL_STATE(1252)] = 84354,
  [SMALL_STATE(1253)] = 84415,
  [SMALL_STATE(1254)] = 84476,
  [SMALL_STATE(1255)] = 84537,
  [SMALL_STATE(1256)] = 84598,
  [SMALL_STATE(1257)] = 84659,
  [SMALL_STATE(1258)] = 84720,
  [SMALL_STATE(1259)] = 84781,
  [SMALL_STATE(1260)] = 84842,
  [SMALL_STATE(1261)] = 84903,
  [SMALL_STATE(1262)] = 84988,
  [SMALL_STATE(1263)] = 85049,
  [SMALL_STATE(1264)] = 85120,
  [SMALL_STATE(1265)] = 85197,
  [SMALL_STATE(1266)] = 85258,
  [SMALL_STATE(1267)] = 85329,
  [SMALL_STATE(1268)] = 85412,
  [SMALL_STATE(1269)] = 85493,
  [SMALL_STATE(1270)] = 85572,
  [SMALL_STATE(1271)] = 85649,
  [SMALL_STATE(1272)] = 85722,
  [SMALL_STATE(1273)] = 85783,
  [SMALL_STATE(1274)] = 85844,
  [SMALL_STATE(1275)] = 85905,
  [SMALL_STATE(1276)] = 85966,
  [SMALL_STATE(1277)] = 86027,
  [SMALL_STATE(1278)] = 86088,
  [SMALL_STATE(1279)] = 86149,
  [SMALL_STATE(1280)] = 86210,
  [SMALL_STATE(1281)] = 86271,
  [SMALL_STATE(1282)] = 86332,
  [SMALL_STATE(1283)] = 86393,
  [SMALL_STATE(1284)] = 86454,
  [SMALL_STATE(1285)] = 86515,
  [SMALL_STATE(1286)] = 86576,
  [SMALL_STATE(1287)] = 86637,
  [SMALL_STATE(1288)] = 86698,
  [SMALL_STATE(1289)] = 86759,
  [SMALL_STATE(1290)] = 86826,
  [SMALL_STATE(1291)] = 86887,
  [SMALL_STATE(1292)] = 86948,
  [SMALL_STATE(1293)] = 87009,
  [SMALL_STATE(1294)] = 87070,
  [SMALL_STATE(1295)] = 87131,
  [SMALL_STATE(1296)] = 87192,
  [SMALL_STATE(1297)] = 87253,
  [SMALL_STATE(1298)] = 87314,
  [SMALL_STATE(1299)] = 87375,
  [SMALL_STATE(1300)] = 87436,
  [SMALL_STATE(1301)] = 87497,
  [SMALL_STATE(1302)] = 87558,
  [SMALL_STATE(1303)] = 87619,
  [SMALL_STATE(1304)] = 87680,
  [SMALL_STATE(1305)] = 87741,
  [SMALL_STATE(1306)] = 87802,
  [SMALL_STATE(1307)] = 87863,
  [SMALL_STATE(1308)] = 87924,
  [SMALL_STATE(1309)] = 87985,
  [SMALL_STATE(1310)] = 88046,
  [SMALL_STATE(1311)] = 88107,
  [SMALL_STATE(1312)] = 88192,
  [SMALL_STATE(1313)] = 88253,
  [SMALL_STATE(1314)] = 88314,
  [SMALL_STATE(1315)] = 88375,
  [SMALL_STATE(1316)] = 88436,
  [SMALL_STATE(1317)] = 88519,
  [SMALL_STATE(1318)] = 88600,
  [SMALL_STATE(1319)] = 88679,
  [SMALL_STATE(1320)] = 88740,
  [SMALL_STATE(1321)] = 88801,
  [SMALL_STATE(1322)] = 88862,
  [SMALL_STATE(1323)] = 88923,
  [SMALL_STATE(1324)] = 88984,
  [SMALL_STATE(1325)] = 89045,
  [SMALL_STATE(1326)] = 89106,
  [SMALL_STATE(1327)] = 89167,
  [SMALL_STATE(1328)] = 89232,
  [SMALL_STATE(1329)] = 89293,
  [SMALL_STATE(1330)] = 89354,
  [SMALL_STATE(1331)] = 89415,
  [SMALL_STATE(1332)] = 89482,
  [SMALL_STATE(1333)] = 89553,
  [SMALL_STATE(1334)] = 89630,
  [SMALL_STATE(1335)] = 89691,
  [SMALL_STATE(1336)] = 89796,
  [SMALL_STATE(1337)] = 89857,
  [SMALL_STATE(1338)] = 89918,
  [SMALL_STATE(1339)] = 89989,
  [SMALL_STATE(1340)] = 90050,
  [SMALL_STATE(1341)] = 90111,
  [SMALL_STATE(1342)] = 90172,
  [SMALL_STATE(1343)] = 90233,
  [SMALL_STATE(1344)] = 90294,
  [SMALL_STATE(1345)] = 90355,
  [SMALL_STATE(1346)] = 90416,
  [SMALL_STATE(1347)] = 90477,
  [SMALL_STATE(1348)] = 90538,
  [SMALL_STATE(1349)] = 90599,
  [SMALL_STATE(1350)] = 90660,
  [SMALL_STATE(1351)] = 90721,
  [SMALL_STATE(1352)] = 90782,
  [SMALL_STATE(1353)] = 90843,
  [SMALL_STATE(1354)] = 90904,
  [SMALL_STATE(1355)] = 90965,
  [SMALL_STATE(1356)] = 91026,
  [SMALL_STATE(1357)] = 91087,
  [SMALL_STATE(1358)] = 91148,
  [SMALL_STATE(1359)] = 91209,
  [SMALL_STATE(1360)] = 91270,
  [SMALL_STATE(1361)] = 91331,
  [SMALL_STATE(1362)] = 91392,
  [SMALL_STATE(1363)] = 91453,
  [SMALL_STATE(1364)] = 91514,
  [SMALL_STATE(1365)] = 91575,
  [SMALL_STATE(1366)] = 91652,
  [SMALL_STATE(1367)] = 91725,
  [SMALL_STATE(1368)] = 91786,
  [SMALL_STATE(1369)] = 91847,
  [SMALL_STATE(1370)] = 91908,
  [SMALL_STATE(1371)] = 91969,
  [SMALL_STATE(1372)] = 92030,
  [SMALL_STATE(1373)] = 92091,
  [SMALL_STATE(1374)] = 92152,
  [SMALL_STATE(1375)] = 92219,
  [SMALL_STATE(1376)] = 92280,
  [SMALL_STATE(1377)] = 92341,
  [SMALL_STATE(1378)] = 92402,
  [SMALL_STATE(1379)] = 92463,
  [SMALL_STATE(1380)] = 92524,
  [SMALL_STATE(1381)] = 92585,
  [SMALL_STATE(1382)] = 92646,
  [SMALL_STATE(1383)] = 92707,
  [SMALL_STATE(1384)] = 92774,
  [SMALL_STATE(1385)] = 92835,
  [SMALL_STATE(1386)] = 92896,
  [SMALL_STATE(1387)] = 92989,
  [SMALL_STATE(1388)] = 93050,
  [SMALL_STATE(1389)] = 93111,
  [SMALL_STATE(1390)] = 93172,
  [SMALL_STATE(1391)] = 93233,
  [SMALL_STATE(1392)] = 93294,
  [SMALL_STATE(1393)] = 93399,
  [SMALL_STATE(1394)] = 93460,
  [SMALL_STATE(1395)] = 93521,
  [SMALL_STATE(1396)] = 93582,
  [SMALL_STATE(1397)] = 93643,
  [SMALL_STATE(1398)] = 93704,
  [SMALL_STATE(1399)] = 93765,
  [SMALL_STATE(1400)] = 93832,
  [SMALL_STATE(1401)] = 93899,
  [SMALL_STATE(1402)] = 94000,
  [SMALL_STATE(1403)] = 94099,
  [SMALL_STATE(1404)] = 94160,
  [SMALL_STATE(1405)] = 94221,
  [SMALL_STATE(1406)] = 94282,
  [SMALL_STATE(1407)] = 94343,
  [SMALL_STATE(1408)] = 94410,
  [SMALL_STATE(1409)] = 94477,
  [SMALL_STATE(1410)] = 94538,
  [SMALL_STATE(1411)] = 94599,
  [SMALL_STATE(1412)] = 94660,
  [SMALL_STATE(1413)] = 94721,
  [SMALL_STATE(1414)] = 94788,
  [SMALL_STATE(1415)] = 94849,
  [SMALL_STATE(1416)] = 94910,
  [SMALL_STATE(1417)] = 94971,
  [SMALL_STATE(1418)] = 95032,
  [SMALL_STATE(1419)] = 95099,
  [SMALL_STATE(1420)] = 95160,
  [SMALL_STATE(1421)] = 95221,
  [SMALL_STATE(1422)] = 95281,
  [SMALL_STATE(1423)] = 95341,
  [SMALL_STATE(1424)] = 95401,
  [SMALL_STATE(1425)] = 95461,
  [SMALL_STATE(1426)] = 95521,
  [SMALL_STATE(1427)] = 95581,
  [SMALL_STATE(1428)] = 95641,
  [SMALL_STATE(1429)] = 95701,
  [SMALL_STATE(1430)] = 95783,
  [SMALL_STATE(1431)] = 95843,
  [SMALL_STATE(1432)] = 95903,
  [SMALL_STATE(1433)] = 95983,
  [SMALL_STATE(1434)] = 96043,
  [SMALL_STATE(1435)] = 96109,
  [SMALL_STATE(1436)] = 96187,
  [SMALL_STATE(1437)] = 96271,
  [SMALL_STATE(1438)] = 96331,
  [SMALL_STATE(1439)] = 96391,
  [SMALL_STATE(1440)] = 96451,
  [SMALL_STATE(1441)] = 96533,
  [SMALL_STATE(1442)] = 96613,
  [SMALL_STATE(1443)] = 96691,
  [SMALL_STATE(1444)] = 96793,
  [SMALL_STATE(1445)] = 96853,
  [SMALL_STATE(1446)] = 96913,
  [SMALL_STATE(1447)] = 96973,
  [SMALL_STATE(1448)] = 97033,
  [SMALL_STATE(1449)] = 97117,
  [SMALL_STATE(1450)] = 97183,
  [SMALL_STATE(1451)] = 97285,
  [SMALL_STATE(1452)] = 97355,
  [SMALL_STATE(1453)] = 97431,
  [SMALL_STATE(1454)] = 97501,
  [SMALL_STATE(1455)] = 97603,
  [SMALL_STATE(1456)] = 97669,
  [SMALL_STATE(1457)] = 97745,
  [SMALL_STATE(1458)] = 97817,
  [SMALL_STATE(1459)] = 97877,
  [SMALL_STATE(1460)] = 97959,
  [SMALL_STATE(1461)] = 98019,
  [SMALL_STATE(1462)] = 98099,
  [SMALL_STATE(1463)] = 98159,
  [SMALL_STATE(1464)] = 98219,
  [SMALL_STATE(1465)] = 98289,
  [SMALL_STATE(1466)] = 98349,
  [SMALL_STATE(1467)] = 98427,
  [SMALL_STATE(1468)] = 98487,
  [SMALL_STATE(1469)] = 98547,
  [SMALL_STATE(1470)] = 98617,
  [SMALL_STATE(1471)] = 98677,
  [SMALL_STATE(1472)] = 98737,
  [SMALL_STATE(1473)] = 98797,
  [SMALL_STATE(1474)] = 98899,
  [SMALL_STATE(1475)] = 98959,
  [SMALL_STATE(1476)] = 99019,
  [SMALL_STATE(1477)] = 99079,
  [SMALL_STATE(1478)] = 99139,
  [SMALL_STATE(1479)] = 99199,
  [SMALL_STATE(1480)] = 99259,
  [SMALL_STATE(1481)] = 99319,
  [SMALL_STATE(1482)] = 99379,
  [SMALL_STATE(1483)] = 99439,
  [SMALL_STATE(1484)] = 99499,
  [SMALL_STATE(1485)] = 99559,
  [SMALL_STATE(1486)] = 99619,
  [SMALL_STATE(1487)] = 99679,
  [SMALL_STATE(1488)] = 99739,
  [SMALL_STATE(1489)] = 99799,
  [SMALL_STATE(1490)] = 99859,
  [SMALL_STATE(1491)] = 99919,
  [SMALL_STATE(1492)] = 99979,
  [SMALL_STATE(1493)] = 100055,
  [SMALL_STATE(1494)] = 100115,
  [SMALL_STATE(1495)] = 100175,
  [SMALL_STATE(1496)] = 100235,
  [SMALL_STATE(1497)] = 100295,
  [SMALL_STATE(1498)] = 100361,
  [SMALL_STATE(1499)] = 100431,
  [SMALL_STATE(1500)] = 100491,
  [SMALL_STATE(1501)] = 100611,
  [SMALL_STATE(1502)] = 100671,
  [SMALL_STATE(1503)] = 100731,
  [SMALL_STATE(1504)] = 100791,
  [SMALL_STATE(1505)] = 100851,
  [SMALL_STATE(1506)] = 100927,
  [SMALL_STATE(1507)] = 100987,
  [SMALL_STATE(1508)] = 101047,
  [SMALL_STATE(1509)] = 101107,
  [SMALL_STATE(1510)] = 101167,
  [SMALL_STATE(1511)] = 101227,
  [SMALL_STATE(1512)] = 101287,
  [SMALL_STATE(1513)] = 101357,
  [SMALL_STATE(1514)] = 101417,
  [SMALL_STATE(1515)] = 101477,
  [SMALL_STATE(1516)] = 101537,
  [SMALL_STATE(1517)] = 101597,
  [SMALL_STATE(1518)] = 101681,
  [SMALL_STATE(1519)] = 101757,
  [SMALL_STATE(1520)] = 101843,
  [SMALL_STATE(1521)] = 101903,
  [SMALL_STATE(1522)] = 101963,
  [SMALL_STATE(1523)] = 102023,
  [SMALL_STATE(1524)] = 102083,
  [SMALL_STATE(1525)] = 102167,
  [SMALL_STATE(1526)] = 102243,
  [SMALL_STATE(1527)] = 102309,
  [SMALL_STATE(1528)] = 102369,
  [SMALL_STATE(1529)] = 102453,
  [SMALL_STATE(1530)] = 102573,
  [SMALL_STATE(1531)] = 102633,
  [SMALL_STATE(1532)] = 102699,
  [SMALL_STATE(1533)] = 102759,
  [SMALL_STATE(1534)] = 102841,
  [SMALL_STATE(1535)] = 102921,
  [SMALL_STATE(1536)] = 102993,
  [SMALL_STATE(1537)] = 103071,
  [SMALL_STATE(1538)] = 103131,
  [SMALL_STATE(1539)] = 103191,
  [SMALL_STATE(1540)] = 103251,
  [SMALL_STATE(1541)] = 103311,
  [SMALL_STATE(1542)] = 103371,
  [SMALL_STATE(1543)] = 103431,
  [SMALL_STATE(1544)] = 103491,
  [SMALL_STATE(1545)] = 103551,
  [SMALL_STATE(1546)] = 103611,
  [SMALL_STATE(1547)] = 103683,
  [SMALL_STATE(1548)] = 103803,
  [SMALL_STATE(1549)] = 103863,
  [SMALL_STATE(1550)] = 103965,
  [SMALL_STATE(1551)] = 104025,
  [SMALL_STATE(1552)] = 104085,
  [SMALL_STATE(1553)] = 104145,
  [SMALL_STATE(1554)] = 104205,
  [SMALL_STATE(1555)] = 104265,
  [SMALL_STATE(1556)] = 104325,
  [SMALL_STATE(1557)] = 104409,
  [SMALL_STATE(1558)] = 104469,
  [SMALL_STATE(1559)] = 104529,
  [SMALL_STATE(1560)] = 104589,
  [SMALL_STATE(1561)] = 104649,
  [SMALL_STATE(1562)] = 104709,
  [SMALL_STATE(1563)] = 104769,
  [SMALL_STATE(1564)] = 104829,
  [SMALL_STATE(1565)] = 104889,
  [SMALL_STATE(1566)] = 104949,
  [SMALL_STATE(1567)] = 105009,
  [SMALL_STATE(1568)] = 105069,
  [SMALL_STATE(1569)] = 105129,
  [SMALL_STATE(1570)] = 105189,
  [SMALL_STATE(1571)] = 105249,
  [SMALL_STATE(1572)] = 105309,
  [SMALL_STATE(1573)] = 105369,
  [SMALL_STATE(1574)] = 105429,
  [SMALL_STATE(1575)] = 105495,
  [SMALL_STATE(1576)] = 105555,
  [SMALL_STATE(1577)] = 105615,
  [SMALL_STATE(1578)] = 105675,
  [SMALL_STATE(1579)] = 105735,
  [SMALL_STATE(1580)] = 105795,
  [SMALL_STATE(1581)] = 105855,
  [SMALL_STATE(1582)] = 105915,
  [SMALL_STATE(1583)] = 105975,
  [SMALL_STATE(1584)] = 106035,
  [SMALL_STATE(1585)] = 106095,
  [SMALL_STATE(1586)] = 106155,
  [SMALL_STATE(1587)] = 106215,
  [SMALL_STATE(1588)] = 106275,
  [SMALL_STATE(1589)] = 106335,
  [SMALL_STATE(1590)] = 106395,
  [SMALL_STATE(1591)] = 106455,
  [SMALL_STATE(1592)] = 106515,
  [SMALL_STATE(1593)] = 106575,
  [SMALL_STATE(1594)] = 106635,
  [SMALL_STATE(1595)] = 106695,
  [SMALL_STATE(1596)] = 106755,
  [SMALL_STATE(1597)] = 106815,
  [SMALL_STATE(1598)] = 106875,
  [SMALL_STATE(1599)] = 106935,
  [SMALL_STATE(1600)] = 106995,
  [SMALL_STATE(1601)] = 107055,
  [SMALL_STATE(1602)] = 107115,
  [SMALL_STATE(1603)] = 107175,
  [SMALL_STATE(1604)] = 107275,
  [SMALL_STATE(1605)] = 107335,
  [SMALL_STATE(1606)] = 107395,
  [SMALL_STATE(1607)] = 107455,
  [SMALL_STATE(1608)] = 107515,
  [SMALL_STATE(1609)] = 107613,
  [SMALL_STATE(1610)] = 107679,
  [SMALL_STATE(1611)] = 107745,
  [SMALL_STATE(1612)] = 107805,
  [SMALL_STATE(1613)] = 107865,
  [SMALL_STATE(1614)] = 107949,
  [SMALL_STATE(1615)] = 108009,
  [SMALL_STATE(1616)] = 108069,
  [SMALL_STATE(1617)] = 108129,
  [SMALL_STATE(1618)] = 108189,
  [SMALL_STATE(1619)] = 108249,
  [SMALL_STATE(1620)] = 108309,
  [SMALL_STATE(1621)] = 108369,
  [SMALL_STATE(1622)] = 108429,
  [SMALL_STATE(1623)] = 108489,
  [SMALL_STATE(1624)] = 108549,
  [SMALL_STATE(1625)] = 108619,
  [SMALL_STATE(1626)] = 108679,
  [SMALL_STATE(1627)] = 108755,
  [SMALL_STATE(1628)] = 108815,
  [SMALL_STATE(1629)] = 108881,
  [SMALL_STATE(1630)] = 108941,
  [SMALL_STATE(1631)] = 109025,
  [SMALL_STATE(1632)] = 109095,
  [SMALL_STATE(1633)] = 109171,
  [SMALL_STATE(1634)] = 109241,
  [SMALL_STATE(1635)] = 109323,
  [SMALL_STATE(1636)] = 109403,
  [SMALL_STATE(1637)] = 109481,
  [SMALL_STATE(1638)] = 109557,
  [SMALL_STATE(1639)] = 109629,
  [SMALL_STATE(1640)] = 109699,
  [SMALL_STATE(1641)] = 109775,
  [SMALL_STATE(1642)] = 109847,
  [SMALL_STATE(1643)] = 109907,
  [SMALL_STATE(1644)] = 109973,
  [SMALL_STATE(1645)] = 110033,
  [SMALL_STATE(1646)] = 110093,
  [SMALL_STATE(1647)] = 110153,
  [SMALL_STATE(1648)] = 110255,
  [SMALL_STATE(1649)] = 110315,
  [SMALL_STATE(1650)] = 110375,
  [SMALL_STATE(1651)] = 110477,
  [SMALL_STATE(1652)] = 110579,
  [SMALL_STATE(1653)] = 110639,
  [SMALL_STATE(1654)] = 110741,
  [SMALL_STATE(1655)] = 110843,
  [SMALL_STATE(1656)] = 110945,
  [SMALL_STATE(1657)] = 111047,
  [SMALL_STATE(1658)] = 111149,
  [SMALL_STATE(1659)] = 111251,
  [SMALL_STATE(1660)] = 111353,
  [SMALL_STATE(1661)] = 111455,
  [SMALL_STATE(1662)] = 111515,
  [SMALL_STATE(1663)] = 111617,
  [SMALL_STATE(1664)] = 111677,
  [SMALL_STATE(1665)] = 111779,
  [SMALL_STATE(1666)] = 111839,
  [SMALL_STATE(1667)] = 111941,
  [SMALL_STATE(1668)] = 112001,
  [SMALL_STATE(1669)] = 112103,
  [SMALL_STATE(1670)] = 112163,
  [SMALL_STATE(1671)] = 112265,
  [SMALL_STATE(1672)] = 112325,
  [SMALL_STATE(1673)] = 112385,
  [SMALL_STATE(1674)] = 112445,
  [SMALL_STATE(1675)] = 112504,
  [SMALL_STATE(1676)] = 112563,
  [SMALL_STATE(1677)] = 112622,
  [SMALL_STATE(1678)] = 112693,
  [SMALL_STATE(1679)] = 112758,
  [SMALL_STATE(1680)] = 112817,
  [SMALL_STATE(1681)] = 112876,
  [SMALL_STATE(1682)] = 112945,
  [SMALL_STATE(1683)] = 113002,
  [SMALL_STATE(1684)] = 113061,
  [SMALL_STATE(1685)] = 113118,
  [SMALL_STATE(1686)] = 113177,
  [SMALL_STATE(1687)] = 113274,
  [SMALL_STATE(1688)] = 113333,
  [SMALL_STATE(1689)] = 113408,
  [SMALL_STATE(1690)] = 113467,
  [SMALL_STATE(1691)] = 113526,
  [SMALL_STATE(1692)] = 113585,
  [SMALL_STATE(1693)] = 113656,
  [SMALL_STATE(1694)] = 113731,
  [SMALL_STATE(1695)] = 113790,
  [SMALL_STATE(1696)] = 113849,
  [SMALL_STATE(1697)] = 113908,
  [SMALL_STATE(1698)] = 113967,
  [SMALL_STATE(1699)] = 114050,
  [SMALL_STATE(1700)] = 114109,
  [SMALL_STATE(1701)] = 114168,
  [SMALL_STATE(1702)] = 114237,
  [SMALL_STATE(1703)] = 114312,
  [SMALL_STATE(1704)] = 114371,
  [SMALL_STATE(1705)] = 114430,
  [SMALL_STATE(1706)] = 114489,
  [SMALL_STATE(1707)] = 114548,
  [SMALL_STATE(1708)] = 114605,
  [SMALL_STATE(1709)] = 114664,
  [SMALL_STATE(1710)] = 114723,
  [SMALL_STATE(1711)] = 114782,
  [SMALL_STATE(1712)] = 114865,
  [SMALL_STATE(1713)] = 114934,
  [SMALL_STATE(1714)] = 114993,
  [SMALL_STATE(1715)] = 115074,
  [SMALL_STATE(1716)] = 115133,
  [SMALL_STATE(1717)] = 115196,
  [SMALL_STATE(1718)] = 115255,
  [SMALL_STATE(1719)] = 115336,
  [SMALL_STATE(1720)] = 115415,
  [SMALL_STATE(1721)] = 115484,
  [SMALL_STATE(1722)] = 115559,
  [SMALL_STATE(1723)] = 115628,
  [SMALL_STATE(1724)] = 115709,
  [SMALL_STATE(1725)] = 115768,
  [SMALL_STATE(1726)] = 115851,
  [SMALL_STATE(1727)] = 115910,
  [SMALL_STATE(1728)] = 116009,
  [SMALL_STATE(1729)] = 116068,
  [SMALL_STATE(1730)] = 116127,
  [SMALL_STATE(1731)] = 116186,
  [SMALL_STATE(1732)] = 116251,
  [SMALL_STATE(1733)] = 116310,
  [SMALL_STATE(1734)] = 116387,
  [SMALL_STATE(1735)] = 116446,
  [SMALL_STATE(1736)] = 116505,
  [SMALL_STATE(1737)] = 116564,
  [SMALL_STATE(1738)] = 116623,
  [SMALL_STATE(1739)] = 116682,
  [SMALL_STATE(1740)] = 116741,
  [SMALL_STATE(1741)] = 116800,
  [SMALL_STATE(1742)] = 116859,
  [SMALL_STATE(1743)] = 116918,
  [SMALL_STATE(1744)] = 116977,
  [SMALL_STATE(1745)] = 117036,
  [SMALL_STATE(1746)] = 117095,
  [SMALL_STATE(1747)] = 117154,
  [SMALL_STATE(1748)] = 117219,
  [SMALL_STATE(1749)] = 117278,
  [SMALL_STATE(1750)] = 117337,
  [SMALL_STATE(1751)] = 117418,
  [SMALL_STATE(1752)] = 117487,
  [SMALL_STATE(1753)] = 117546,
  [SMALL_STATE(1754)] = 117625,
  [SMALL_STATE(1755)] = 117684,
  [SMALL_STATE(1756)] = 117761,
  [SMALL_STATE(1757)] = 117844,
  [SMALL_STATE(1758)] = 117903,
  [SMALL_STATE(1759)] = 117962,
  [SMALL_STATE(1760)] = 118021,
  [SMALL_STATE(1761)] = 118080,
  [SMALL_STATE(1762)] = 118139,
  [SMALL_STATE(1763)] = 118198,
  [SMALL_STATE(1764)] = 118257,
  [SMALL_STATE(1765)] = 118316,
  [SMALL_STATE(1766)] = 118375,
  [SMALL_STATE(1767)] = 118434,
  [SMALL_STATE(1768)] = 118493,
  [SMALL_STATE(1769)] = 118552,
  [SMALL_STATE(1770)] = 118617,
  [SMALL_STATE(1771)] = 118676,
  [SMALL_STATE(1772)] = 118755,
  [SMALL_STATE(1773)] = 118820,
  [SMALL_STATE(1774)] = 118879,
  [SMALL_STATE(1775)] = 118956,
  [SMALL_STATE(1776)] = 119031,
  [SMALL_STATE(1777)] = 119130,
  [SMALL_STATE(1778)] = 119213,
  [SMALL_STATE(1779)] = 119284,
  [SMALL_STATE(1780)] = 119343,
  [SMALL_STATE(1781)] = 119402,
  [SMALL_STATE(1782)] = 119461,
  [SMALL_STATE(1783)] = 119544,
  [SMALL_STATE(1784)] = 119603,
  [SMALL_STATE(1785)] = 119662,
  [SMALL_STATE(1786)] = 119721,
  [SMALL_STATE(1787)] = 119778,
  [SMALL_STATE(1788)] = 119837,
  [SMALL_STATE(1789)] = 119896,
  [SMALL_STATE(1790)] = 119955,
  [SMALL_STATE(1791)] = 120014,
  [SMALL_STATE(1792)] = 120073,
  [SMALL_STATE(1793)] = 120132,
  [SMALL_STATE(1794)] = 120191,
  [SMALL_STATE(1795)] = 120250,
  [SMALL_STATE(1796)] = 120309,
  [SMALL_STATE(1797)] = 120368,
  [SMALL_STATE(1798)] = 120437,
  [SMALL_STATE(1799)] = 120496,
  [SMALL_STATE(1800)] = 120555,
  [SMALL_STATE(1801)] = 120636,
  [SMALL_STATE(1802)] = 120695,
  [SMALL_STATE(1803)] = 120754,
  [SMALL_STATE(1804)] = 120829,
  [SMALL_STATE(1805)] = 120904,
  [SMALL_STATE(1806)] = 120983,
  [SMALL_STATE(1807)] = 121048,
  [SMALL_STATE(1808)] = 121107,
  [SMALL_STATE(1809)] = 121184,
  [SMALL_STATE(1810)] = 121253,
  [SMALL_STATE(1811)] = 121312,
  [SMALL_STATE(1812)] = 121371,
  [SMALL_STATE(1813)] = 121448,
  [SMALL_STATE(1814)] = 121529,
  [SMALL_STATE(1815)] = 121608,
  [SMALL_STATE(1816)] = 121685,
  [SMALL_STATE(1817)] = 121760,
  [SMALL_STATE(1818)] = 121821,
  [SMALL_STATE(1819)] = 121892,
  [SMALL_STATE(1820)] = 121961,
  [SMALL_STATE(1821)] = 122020,
  [SMALL_STATE(1822)] = 122079,
  [SMALL_STATE(1823)] = 122138,
  [SMALL_STATE(1824)] = 122197,
  [SMALL_STATE(1825)] = 122256,
  [SMALL_STATE(1826)] = 122315,
  [SMALL_STATE(1827)] = 122374,
  [SMALL_STATE(1828)] = 122433,
  [SMALL_STATE(1829)] = 122492,
  [SMALL_STATE(1830)] = 122551,
  [SMALL_STATE(1831)] = 122610,
  [SMALL_STATE(1832)] = 122669,
  [SMALL_STATE(1833)] = 122728,
  [SMALL_STATE(1834)] = 122787,
  [SMALL_STATE(1835)] = 122846,
  [SMALL_STATE(1836)] = 122905,
  [SMALL_STATE(1837)] = 122970,
  [SMALL_STATE(1838)] = 123033,
  [SMALL_STATE(1839)] = 123092,
  [SMALL_STATE(1840)] = 123167,
  [SMALL_STATE(1841)] = 123226,
  [SMALL_STATE(1842)] = 123285,
  [SMALL_STATE(1843)] = 123344,
  [SMALL_STATE(1844)] = 123403,
  [SMALL_STATE(1845)] = 123486,
  [SMALL_STATE(1846)] = 123561,
  [SMALL_STATE(1847)] = 123620,
  [SMALL_STATE(1848)] = 123679,
  [SMALL_STATE(1849)] = 123738,
  [SMALL_STATE(1850)] = 123797,
  [SMALL_STATE(1851)] = 123856,
  [SMALL_STATE(1852)] = 123915,
  [SMALL_STATE(1853)] = 123974,
  [SMALL_STATE(1854)] = 124033,
  [SMALL_STATE(1855)] = 124092,
  [SMALL_STATE(1856)] = 124151,
  [SMALL_STATE(1857)] = 124208,
  [SMALL_STATE(1858)] = 124267,
  [SMALL_STATE(1859)] = 124326,
  [SMALL_STATE(1860)] = 124385,
  [SMALL_STATE(1861)] = 124450,
  [SMALL_STATE(1862)] = 124519,
  [SMALL_STATE(1863)] = 124584,
  [SMALL_STATE(1864)] = 124643,
  [SMALL_STATE(1865)] = 124702,
  [SMALL_STATE(1866)] = 124761,
  [SMALL_STATE(1867)] = 124836,
  [SMALL_STATE(1868)] = 124895,
  [SMALL_STATE(1869)] = 124964,
  [SMALL_STATE(1870)] = 125023,
  [SMALL_STATE(1871)] = 125098,
  [SMALL_STATE(1872)] = 125155,
  [SMALL_STATE(1873)] = 125214,
  [SMALL_STATE(1874)] = 125273,
  [SMALL_STATE(1875)] = 125356,
  [SMALL_STATE(1876)] = 125427,
  [SMALL_STATE(1877)] = 125486,
  [SMALL_STATE(1878)] = 125545,
  [SMALL_STATE(1879)] = 125604,
  [SMALL_STATE(1880)] = 125663,
  [SMALL_STATE(1881)] = 125722,
  [SMALL_STATE(1882)] = 125781,
  [SMALL_STATE(1883)] = 125846,
  [SMALL_STATE(1884)] = 125925,
  [SMALL_STATE(1885)] = 125984,
  [SMALL_STATE(1886)] = 126043,
  [SMALL_STATE(1887)] = 126114,
  [SMALL_STATE(1888)] = 126173,
  [SMALL_STATE(1889)] = 126232,
  [SMALL_STATE(1890)] = 126315,
  [SMALL_STATE(1891)] = 126414,
  [SMALL_STATE(1892)] = 126473,
  [SMALL_STATE(1893)] = 126532,
  [SMALL_STATE(1894)] = 126591,
  [SMALL_STATE(1895)] = 126650,
  [SMALL_STATE(1896)] = 126709,
  [SMALL_STATE(1897)] = 126774,
  [SMALL_STATE(1898)] = 126833,
  [SMALL_STATE(1899)] = 126892,
  [SMALL_STATE(1900)] = 126951,
  [SMALL_STATE(1901)] = 127010,
  [SMALL_STATE(1902)] = 127079,
  [SMALL_STATE(1903)] = 127138,
  [SMALL_STATE(1904)] = 127197,
  [SMALL_STATE(1905)] = 127256,
  [SMALL_STATE(1906)] = 127315,
  [SMALL_STATE(1907)] = 127374,
  [SMALL_STATE(1908)] = 127433,
  [SMALL_STATE(1909)] = 127492,
  [SMALL_STATE(1910)] = 127551,
  [SMALL_STATE(1911)] = 127612,
  [SMALL_STATE(1912)] = 127671,
  [SMALL_STATE(1913)] = 127730,
  [SMALL_STATE(1914)] = 127788,
  [SMALL_STATE(1915)] = 127844,
  [SMALL_STATE(1916)] = 127900,
  [SMALL_STATE(1917)] = 127984,
  [SMALL_STATE(1918)] = 128040,
  [SMALL_STATE(1919)] = 128096,
  [SMALL_STATE(1920)] = 128152,
  [SMALL_STATE(1921)] = 128208,
  [SMALL_STATE(1922)] = 128264,
  [SMALL_STATE(1923)] = 128320,
  [SMALL_STATE(1924)] = 128376,
  [SMALL_STATE(1925)] = 128432,
  [SMALL_STATE(1926)] = 128488,
  [SMALL_STATE(1927)] = 128544,
  [SMALL_STATE(1928)] = 128602,
  [SMALL_STATE(1929)] = 128658,
  [SMALL_STATE(1930)] = 128714,
  [SMALL_STATE(1931)] = 128772,
  [SMALL_STATE(1932)] = 128854,
  [SMALL_STATE(1933)] = 128910,
  [SMALL_STATE(1934)] = 128978,
  [SMALL_STATE(1935)] = 129052,
  [SMALL_STATE(1936)] = 129120,
  [SMALL_STATE(1937)] = 129200,
  [SMALL_STATE(1938)] = 129278,
  [SMALL_STATE(1939)] = 129354,
  [SMALL_STATE(1940)] = 129428,
  [SMALL_STATE(1941)] = 129498,
  [SMALL_STATE(1942)] = 129582,
  [SMALL_STATE(1943)] = 129638,
  [SMALL_STATE(1944)] = 129696,
  [SMALL_STATE(1945)] = 129754,
  [SMALL_STATE(1946)] = 129812,
  [SMALL_STATE(1947)] = 129870,
  [SMALL_STATE(1948)] = 129928,
  [SMALL_STATE(1949)] = 129986,
  [SMALL_STATE(1950)] = 130044,
  [SMALL_STATE(1951)] = 130102,
  [SMALL_STATE(1952)] = 130184,
  [SMALL_STATE(1953)] = 130240,
  [SMALL_STATE(1954)] = 130296,
  [SMALL_STATE(1955)] = 130354,
  [SMALL_STATE(1956)] = 130412,
  [SMALL_STATE(1957)] = 130470,
  [SMALL_STATE(1958)] = 130526,
  [SMALL_STATE(1959)] = 130586,
  [SMALL_STATE(1960)] = 130642,
  [SMALL_STATE(1961)] = 130698,
  [SMALL_STATE(1962)] = 130754,
  [SMALL_STATE(1963)] = 130810,
  [SMALL_STATE(1964)] = 130894,
  [SMALL_STATE(1965)] = 130952,
  [SMALL_STATE(1966)] = 131008,
  [SMALL_STATE(1967)] = 131064,
  [SMALL_STATE(1968)] = 131120,
  [SMALL_STATE(1969)] = 131176,
  [SMALL_STATE(1970)] = 131232,
  [SMALL_STATE(1971)] = 131288,
  [SMALL_STATE(1972)] = 131344,
  [SMALL_STATE(1973)] = 131400,
  [SMALL_STATE(1974)] = 131484,
  [SMALL_STATE(1975)] = 131540,
  [SMALL_STATE(1976)] = 131598,
  [SMALL_STATE(1977)] = 131656,
  [SMALL_STATE(1978)] = 131712,
  [SMALL_STATE(1979)] = 131776,
  [SMALL_STATE(1980)] = 131832,
  [SMALL_STATE(1981)] = 131890,
  [SMALL_STATE(1982)] = 131948,
  [SMALL_STATE(1983)] = 132004,
  [SMALL_STATE(1984)] = 132060,
  [SMALL_STATE(1985)] = 132116,
  [SMALL_STATE(1986)] = 132172,
  [SMALL_STATE(1987)] = 132230,
  [SMALL_STATE(1988)] = 132288,
  [SMALL_STATE(1989)] = 132346,
  [SMALL_STATE(1990)] = 132408,
  [SMALL_STATE(1991)] = 132466,
  [SMALL_STATE(1992)] = 132524,
  [SMALL_STATE(1993)] = 132582,
  [SMALL_STATE(1994)] = 132640,
  [SMALL_STATE(1995)] = 132698,
  [SMALL_STATE(1996)] = 132756,
  [SMALL_STATE(1997)] = 132814,
  [SMALL_STATE(1998)] = 132872,
  [SMALL_STATE(1999)] = 132930,
  [SMALL_STATE(2000)] = 132988,
  [SMALL_STATE(2001)] = 133046,
  [SMALL_STATE(2002)] = 133104,
  [SMALL_STATE(2003)] = 133162,
  [SMALL_STATE(2004)] = 133220,
  [SMALL_STATE(2005)] = 133278,
  [SMALL_STATE(2006)] = 133336,
  [SMALL_STATE(2007)] = 133394,
  [SMALL_STATE(2008)] = 133452,
  [SMALL_STATE(2009)] = 133516,
  [SMALL_STATE(2010)] = 133574,
  [SMALL_STATE(2011)] = 133630,
  [SMALL_STATE(2012)] = 133686,
  [SMALL_STATE(2013)] = 133742,
  [SMALL_STATE(2014)] = 133800,
  [SMALL_STATE(2015)] = 133864,
  [SMALL_STATE(2016)] = 133926,
  [SMALL_STATE(2017)] = 133984,
  [SMALL_STATE(2018)] = 134042,
  [SMALL_STATE(2019)] = 134100,
  [SMALL_STATE(2020)] = 134158,
  [SMALL_STATE(2021)] = 134222,
  [SMALL_STATE(2022)] = 134284,
  [SMALL_STATE(2023)] = 134368,
  [SMALL_STATE(2024)] = 134426,
  [SMALL_STATE(2025)] = 134508,
  [SMALL_STATE(2026)] = 134564,
  [SMALL_STATE(2027)] = 134620,
  [SMALL_STATE(2028)] = 134688,
  [SMALL_STATE(2029)] = 134762,
  [SMALL_STATE(2030)] = 134818,
  [SMALL_STATE(2031)] = 134874,
  [SMALL_STATE(2032)] = 134930,
  [SMALL_STATE(2033)] = 134988,
  [SMALL_STATE(2034)] = 135044,
  [SMALL_STATE(2035)] = 135112,
  [SMALL_STATE(2036)] = 135192,
  [SMALL_STATE(2037)] = 135270,
  [SMALL_STATE(2038)] = 135346,
  [SMALL_STATE(2039)] = 135420,
  [SMALL_STATE(2040)] = 135490,
  [SMALL_STATE(2041)] = 135546,
  [SMALL_STATE(2042)] = 135602,
  [SMALL_STATE(2043)] = 135660,
  [SMALL_STATE(2044)] = 135716,
  [SMALL_STATE(2045)] = 135772,
  [SMALL_STATE(2046)] = 135828,
  [SMALL_STATE(2047)] = 135884,
  [SMALL_STATE(2048)] = 135940,
  [SMALL_STATE(2049)] = 136022,
  [SMALL_STATE(2050)] = 136078,
  [SMALL_STATE(2051)] = 136162,
  [SMALL_STATE(2052)] = 136218,
  [SMALL_STATE(2053)] = 136274,
  [SMALL_STATE(2054)] = 136330,
  [SMALL_STATE(2055)] = 136394,
  [SMALL_STATE(2056)] = 136476,
  [SMALL_STATE(2057)] = 136534,
  [SMALL_STATE(2058)] = 136614,
  [SMALL_STATE(2059)] = 136692,
  [SMALL_STATE(2060)] = 136768,
  [SMALL_STATE(2061)] = 136824,
  [SMALL_STATE(2062)] = 136882,
  [SMALL_STATE(2063)] = 136940,
  [SMALL_STATE(2064)] = 137022,
  [SMALL_STATE(2065)] = 137090,
  [SMALL_STATE(2066)] = 137164,
  [SMALL_STATE(2067)] = 137232,
  [SMALL_STATE(2068)] = 137312,
  [SMALL_STATE(2069)] = 137390,
  [SMALL_STATE(2070)] = 137466,
  [SMALL_STATE(2071)] = 137540,
  [SMALL_STATE(2072)] = 137610,
  [SMALL_STATE(2073)] = 137668,
  [SMALL_STATE(2074)] = 137724,
  [SMALL_STATE(2075)] = 137780,
  [SMALL_STATE(2076)] = 137836,
  [SMALL_STATE(2077)] = 137918,
  [SMALL_STATE(2078)] = 137998,
  [SMALL_STATE(2079)] = 138076,
  [SMALL_STATE(2080)] = 138152,
  [SMALL_STATE(2081)] = 138210,
  [SMALL_STATE(2082)] = 138266,
  [SMALL_STATE(2083)] = 138324,
  [SMALL_STATE(2084)] = 138382,
  [SMALL_STATE(2085)] = 138440,
  [SMALL_STATE(2086)] = 138498,
  [SMALL_STATE(2087)] = 138554,
  [SMALL_STATE(2088)] = 138612,
  [SMALL_STATE(2089)] = 138670,
  [SMALL_STATE(2090)] = 138728,
  [SMALL_STATE(2091)] = 138786,
  [SMALL_STATE(2092)] = 138844,
  [SMALL_STATE(2093)] = 138902,
  [SMALL_STATE(2094)] = 138960,
  [SMALL_STATE(2095)] = 139018,
  [SMALL_STATE(2096)] = 139076,
  [SMALL_STATE(2097)] = 139134,
  [SMALL_STATE(2098)] = 139192,
  [SMALL_STATE(2099)] = 139250,
  [SMALL_STATE(2100)] = 139308,
  [SMALL_STATE(2101)] = 139366,
  [SMALL_STATE(2102)] = 139424,
  [SMALL_STATE(2103)] = 139482,
  [SMALL_STATE(2104)] = 139540,
  [SMALL_STATE(2105)] = 139598,
  [SMALL_STATE(2106)] = 139656,
  [SMALL_STATE(2107)] = 139714,
  [SMALL_STATE(2108)] = 139772,
  [SMALL_STATE(2109)] = 139830,
  [SMALL_STATE(2110)] = 139888,
  [SMALL_STATE(2111)] = 139970,
  [SMALL_STATE(2112)] = 140028,
  [SMALL_STATE(2113)] = 140086,
  [SMALL_STATE(2114)] = 140144,
  [SMALL_STATE(2115)] = 140200,
  [SMALL_STATE(2116)] = 140258,
  [SMALL_STATE(2117)] = 140340,
  [SMALL_STATE(2118)] = 140436,
  [SMALL_STATE(2119)] = 140494,
  [SMALL_STATE(2120)] = 140552,
  [SMALL_STATE(2121)] = 140610,
  [SMALL_STATE(2122)] = 140668,
  [SMALL_STATE(2123)] = 140726,
  [SMALL_STATE(2124)] = 140784,
  [SMALL_STATE(2125)] = 140842,
  [SMALL_STATE(2126)] = 140900,
  [SMALL_STATE(2127)] = 140958,
  [SMALL_STATE(2128)] = 141016,
  [SMALL_STATE(2129)] = 141074,
  [SMALL_STATE(2130)] = 141132,
  [SMALL_STATE(2131)] = 141190,
  [SMALL_STATE(2132)] = 141248,
  [SMALL_STATE(2133)] = 141306,
  [SMALL_STATE(2134)] = 141364,
  [SMALL_STATE(2135)] = 141422,
  [SMALL_STATE(2136)] = 141480,
  [SMALL_STATE(2137)] = 141538,
  [SMALL_STATE(2138)] = 141596,
  [SMALL_STATE(2139)] = 141654,
  [SMALL_STATE(2140)] = 141712,
  [SMALL_STATE(2141)] = 141770,
  [SMALL_STATE(2142)] = 141828,
  [SMALL_STATE(2143)] = 141886,
  [SMALL_STATE(2144)] = 141944,
  [SMALL_STATE(2145)] = 142002,
  [SMALL_STATE(2146)] = 142060,
  [SMALL_STATE(2147)] = 142118,
  [SMALL_STATE(2148)] = 142176,
  [SMALL_STATE(2149)] = 142234,
  [SMALL_STATE(2150)] = 142292,
  [SMALL_STATE(2151)] = 142350,
  [SMALL_STATE(2152)] = 142408,
  [SMALL_STATE(2153)] = 142466,
  [SMALL_STATE(2154)] = 142524,
  [SMALL_STATE(2155)] = 142582,
  [SMALL_STATE(2156)] = 142640,
  [SMALL_STATE(2157)] = 142698,
  [SMALL_STATE(2158)] = 142756,
  [SMALL_STATE(2159)] = 142814,
  [SMALL_STATE(2160)] = 142872,
  [SMALL_STATE(2161)] = 142930,
  [SMALL_STATE(2162)] = 142988,
  [SMALL_STATE(2163)] = 143046,
  [SMALL_STATE(2164)] = 143102,
  [SMALL_STATE(2165)] = 143158,
  [SMALL_STATE(2166)] = 143214,
  [SMALL_STATE(2167)] = 143310,
  [SMALL_STATE(2168)] = 143368,
  [SMALL_STATE(2169)] = 143426,
  [SMALL_STATE(2170)] = 143484,
  [SMALL_STATE(2171)] = 143542,
  [SMALL_STATE(2172)] = 143600,
  [SMALL_STATE(2173)] = 143658,
  [SMALL_STATE(2174)] = 143716,
  [SMALL_STATE(2175)] = 143774,
  [SMALL_STATE(2176)] = 143832,
  [SMALL_STATE(2177)] = 143890,
  [SMALL_STATE(2178)] = 143948,
  [SMALL_STATE(2179)] = 144006,
  [SMALL_STATE(2180)] = 144064,
  [SMALL_STATE(2181)] = 144122,
  [SMALL_STATE(2182)] = 144178,
  [SMALL_STATE(2183)] = 144236,
  [SMALL_STATE(2184)] = 144292,
  [SMALL_STATE(2185)] = 144374,
  [SMALL_STATE(2186)] = 144432,
  [SMALL_STATE(2187)] = 144490,
  [SMALL_STATE(2188)] = 144548,
  [SMALL_STATE(2189)] = 144606,
  [SMALL_STATE(2190)] = 144664,
  [SMALL_STATE(2191)] = 144722,
  [SMALL_STATE(2192)] = 144780,
  [SMALL_STATE(2193)] = 144838,
  [SMALL_STATE(2194)] = 144896,
  [SMALL_STATE(2195)] = 144954,
  [SMALL_STATE(2196)] = 145012,
  [SMALL_STATE(2197)] = 145070,
  [SMALL_STATE(2198)] = 145128,
  [SMALL_STATE(2199)] = 145186,
  [SMALL_STATE(2200)] = 145244,
  [SMALL_STATE(2201)] = 145302,
  [SMALL_STATE(2202)] = 145360,
  [SMALL_STATE(2203)] = 145418,
  [SMALL_STATE(2204)] = 145476,
  [SMALL_STATE(2205)] = 145534,
  [SMALL_STATE(2206)] = 145592,
  [SMALL_STATE(2207)] = 145650,
  [SMALL_STATE(2208)] = 145708,
  [SMALL_STATE(2209)] = 145766,
  [SMALL_STATE(2210)] = 145824,
  [SMALL_STATE(2211)] = 145892,
  [SMALL_STATE(2212)] = 145966,
  [SMALL_STATE(2213)] = 146022,
  [SMALL_STATE(2214)] = 146078,
  [SMALL_STATE(2215)] = 146146,
  [SMALL_STATE(2216)] = 146202,
  [SMALL_STATE(2217)] = 146260,
  [SMALL_STATE(2218)] = 146334,
  [SMALL_STATE(2219)] = 146404,
  [SMALL_STATE(2220)] = 146462,
  [SMALL_STATE(2221)] = 146518,
  [SMALL_STATE(2222)] = 146574,
  [SMALL_STATE(2223)] = 146632,
  [SMALL_STATE(2224)] = 146690,
  [SMALL_STATE(2225)] = 146748,
  [SMALL_STATE(2226)] = 146804,
  [SMALL_STATE(2227)] = 146862,
  [SMALL_STATE(2228)] = 146920,
  [SMALL_STATE(2229)] = 146976,
  [SMALL_STATE(2230)] = 147032,
  [SMALL_STATE(2231)] = 147088,
  [SMALL_STATE(2232)] = 147156,
  [SMALL_STATE(2233)] = 147230,
  [SMALL_STATE(2234)] = 147288,
  [SMALL_STATE(2235)] = 147346,
  [SMALL_STATE(2236)] = 147404,
  [SMALL_STATE(2237)] = 147462,
  [SMALL_STATE(2238)] = 147520,
  [SMALL_STATE(2239)] = 147578,
  [SMALL_STATE(2240)] = 147646,
  [SMALL_STATE(2241)] = 147720,
  [SMALL_STATE(2242)] = 147790,
  [SMALL_STATE(2243)] = 147846,
  [SMALL_STATE(2244)] = 147904,
  [SMALL_STATE(2245)] = 147962,
  [SMALL_STATE(2246)] = 148020,
  [SMALL_STATE(2247)] = 148078,
  [SMALL_STATE(2248)] = 148136,
  [SMALL_STATE(2249)] = 148194,
  [SMALL_STATE(2250)] = 148250,
  [SMALL_STATE(2251)] = 148306,
  [SMALL_STATE(2252)] = 148362,
  [SMALL_STATE(2253)] = 148418,
  [SMALL_STATE(2254)] = 148476,
  [SMALL_STATE(2255)] = 148534,
  [SMALL_STATE(2256)] = 148592,
  [SMALL_STATE(2257)] = 148650,
  [SMALL_STATE(2258)] = 148706,
  [SMALL_STATE(2259)] = 148762,
  [SMALL_STATE(2260)] = 148818,
  [SMALL_STATE(2261)] = 148874,
  [SMALL_STATE(2262)] = 148930,
  [SMALL_STATE(2263)] = 148986,
  [SMALL_STATE(2264)] = 149042,
  [SMALL_STATE(2265)] = 149098,
  [SMALL_STATE(2266)] = 149156,
  [SMALL_STATE(2267)] = 149214,
  [SMALL_STATE(2268)] = 149272,
  [SMALL_STATE(2269)] = 149330,
  [SMALL_STATE(2270)] = 149388,
  [SMALL_STATE(2271)] = 149444,
  [SMALL_STATE(2272)] = 149502,
  [SMALL_STATE(2273)] = 149558,
  [SMALL_STATE(2274)] = 149616,
  [SMALL_STATE(2275)] = 149672,
  [SMALL_STATE(2276)] = 149756,
  [SMALL_STATE(2277)] = 149814,
  [SMALL_STATE(2278)] = 149872,
  [SMALL_STATE(2279)] = 149930,
  [SMALL_STATE(2280)] = 149988,
  [SMALL_STATE(2281)] = 150046,
  [SMALL_STATE(2282)] = 150104,
  [SMALL_STATE(2283)] = 150162,
  [SMALL_STATE(2284)] = 150226,
  [SMALL_STATE(2285)] = 150308,
  [SMALL_STATE(2286)] = 150364,
  [SMALL_STATE(2287)] = 150420,
  [SMALL_STATE(2288)] = 150476,
  [SMALL_STATE(2289)] = 150534,
  [SMALL_STATE(2290)] = 150591,
  [SMALL_STATE(2291)] = 150648,
  [SMALL_STATE(2292)] = 150729,
  [SMALL_STATE(2293)] = 150808,
  [SMALL_STATE(2294)] = 150885,
  [SMALL_STATE(2295)] = 150960,
  [SMALL_STATE(2296)] = 151041,
  [SMALL_STATE(2297)] = 151098,
  [SMALL_STATE(2298)] = 151191,
  [SMALL_STATE(2299)] = 151248,
  [SMALL_STATE(2300)] = 151315,
  [SMALL_STATE(2301)] = 151388,
  [SMALL_STATE(2302)] = 151445,
  [SMALL_STATE(2303)] = 151518,
  [SMALL_STATE(2304)] = 151587,
  [SMALL_STATE(2305)] = 151644,
  [SMALL_STATE(2306)] = 151725,
  [SMALL_STATE(2307)] = 151782,
  [SMALL_STATE(2308)] = 151839,
  [SMALL_STATE(2309)] = 151920,
  [SMALL_STATE(2310)] = 152013,
  [SMALL_STATE(2311)] = 152080,
  [SMALL_STATE(2312)] = 152153,
  [SMALL_STATE(2313)] = 152234,
  [SMALL_STATE(2314)] = 152291,
  [SMALL_STATE(2315)] = 152384,
  [SMALL_STATE(2316)] = 152445,
  [SMALL_STATE(2317)] = 152512,
  [SMALL_STATE(2318)] = 152569,
  [SMALL_STATE(2319)] = 152626,
  [SMALL_STATE(2320)] = 152705,
  [SMALL_STATE(2321)] = 152772,
  [SMALL_STATE(2322)] = 152851,
  [SMALL_STATE(2323)] = 152932,
  [SMALL_STATE(2324)] = 153009,
  [SMALL_STATE(2325)] = 153084,
  [SMALL_STATE(2326)] = 153157,
  [SMALL_STATE(2327)] = 153238,
  [SMALL_STATE(2328)] = 153307,
  [SMALL_STATE(2329)] = 153384,
  [SMALL_STATE(2330)] = 153459,
  [SMALL_STATE(2331)] = 153540,
  [SMALL_STATE(2332)] = 153613,
  [SMALL_STATE(2333)] = 153694,
  [SMALL_STATE(2334)] = 153775,
  [SMALL_STATE(2335)] = 153844,
  [SMALL_STATE(2336)] = 153925,
  [SMALL_STATE(2337)] = 154018,
  [SMALL_STATE(2338)] = 154075,
  [SMALL_STATE(2339)] = 154166,
  [SMALL_STATE(2340)] = 154247,
  [SMALL_STATE(2341)] = 154304,
  [SMALL_STATE(2342)] = 154361,
  [SMALL_STATE(2343)] = 154442,
  [SMALL_STATE(2344)] = 154509,
  [SMALL_STATE(2345)] = 154582,
  [SMALL_STATE(2346)] = 154643,
  [SMALL_STATE(2347)] = 154710,
  [SMALL_STATE(2348)] = 154789,
  [SMALL_STATE(2349)] = 154866,
  [SMALL_STATE(2350)] = 154941,
  [SMALL_STATE(2351)] = 155014,
  [SMALL_STATE(2352)] = 155083,
  [SMALL_STATE(2353)] = 155144,
  [SMALL_STATE(2354)] = 155203,
  [SMALL_STATE(2355)] = 155260,
  [SMALL_STATE(2356)] = 155317,
  [SMALL_STATE(2357)] = 155398,
  [SMALL_STATE(2358)] = 155455,
  [SMALL_STATE(2359)] = 155548,
  [SMALL_STATE(2360)] = 155615,
  [SMALL_STATE(2361)] = 155688,
  [SMALL_STATE(2362)] = 155755,
  [SMALL_STATE(2363)] = 155834,
  [SMALL_STATE(2364)] = 155911,
  [SMALL_STATE(2365)] = 155986,
  [SMALL_STATE(2366)] = 156059,
  [SMALL_STATE(2367)] = 156116,
  [SMALL_STATE(2368)] = 156173,
  [SMALL_STATE(2369)] = 156230,
  [SMALL_STATE(2370)] = 156287,
  [SMALL_STATE(2371)] = 156356,
  [SMALL_STATE(2372)] = 156413,
  [SMALL_STATE(2373)] = 156470,
  [SMALL_STATE(2374)] = 156527,
  [SMALL_STATE(2375)] = 156584,
  [SMALL_STATE(2376)] = 156641,
  [SMALL_STATE(2377)] = 156698,
  [SMALL_STATE(2378)] = 156779,
  [SMALL_STATE(2379)] = 156852,
  [SMALL_STATE(2380)] = 156933,
  [SMALL_STATE(2381)] = 156990,
  [SMALL_STATE(2382)] = 157071,
  [SMALL_STATE(2383)] = 157152,
  [SMALL_STATE(2384)] = 157209,
  [SMALL_STATE(2385)] = 157266,
  [SMALL_STATE(2386)] = 157323,
  [SMALL_STATE(2387)] = 157416,
  [SMALL_STATE(2388)] = 157473,
  [SMALL_STATE(2389)] = 157554,
  [SMALL_STATE(2390)] = 157611,
  [SMALL_STATE(2391)] = 157668,
  [SMALL_STATE(2392)] = 157761,
  [SMALL_STATE(2393)] = 157852,
  [SMALL_STATE(2394)] = 157909,
  [SMALL_STATE(2395)] = 157990,
  [SMALL_STATE(2396)] = 158047,
  [SMALL_STATE(2397)] = 158104,
  [SMALL_STATE(2398)] = 158161,
  [SMALL_STATE(2399)] = 158252,
  [SMALL_STATE(2400)] = 158309,
  [SMALL_STATE(2401)] = 158390,
  [SMALL_STATE(2402)] = 158471,
  [SMALL_STATE(2403)] = 158538,
  [SMALL_STATE(2404)] = 158611,
  [SMALL_STATE(2405)] = 158668,
  [SMALL_STATE(2406)] = 158725,
  [SMALL_STATE(2407)] = 158816,
  [SMALL_STATE(2408)] = 158897,
  [SMALL_STATE(2409)] = 158954,
  [SMALL_STATE(2410)] = 159011,
  [SMALL_STATE(2411)] = 159104,
  [SMALL_STATE(2412)] = 159197,
  [SMALL_STATE(2413)] = 159264,
  [SMALL_STATE(2414)] = 159345,
  [SMALL_STATE(2415)] = 159412,
  [SMALL_STATE(2416)] = 159491,
  [SMALL_STATE(2417)] = 159572,
  [SMALL_STATE(2418)] = 159629,
  [SMALL_STATE(2419)] = 159710,
  [SMALL_STATE(2420)] = 159791,
  [SMALL_STATE(2421)] = 159868,
  [SMALL_STATE(2422)] = 159925,
  [SMALL_STATE(2423)] = 159982,
  [SMALL_STATE(2424)] = 160057,
  [SMALL_STATE(2425)] = 160130,
  [SMALL_STATE(2426)] = 160199,
  [SMALL_STATE(2427)] = 160292,
  [SMALL_STATE(2428)] = 160349,
  [SMALL_STATE(2429)] = 160430,
  [SMALL_STATE(2430)] = 160487,
  [SMALL_STATE(2431)] = 160568,
  [SMALL_STATE(2432)] = 160649,
  [SMALL_STATE(2433)] = 160730,
  [SMALL_STATE(2434)] = 160797,
  [SMALL_STATE(2435)] = 160875,
  [SMALL_STATE(2436)] = 160953,
  [SMALL_STATE(2437)] = 161009,
  [SMALL_STATE(2438)] = 161087,
  [SMALL_STATE(2439)] = 161143,
  [SMALL_STATE(2440)] = 161199,
  [SMALL_STATE(2441)] = 161255,
  [SMALL_STATE(2442)] = 161333,
  [SMALL_STATE(2443)] = 161389,
  [SMALL_STATE(2444)] = 161449,
  [SMALL_STATE(2445)] = 161505,
  [SMALL_STATE(2446)] = 161561,
  [SMALL_STATE(2447)] = 161617,
  [SMALL_STATE(2448)] = 161673,
  [SMALL_STATE(2449)] = 161729,
  [SMALL_STATE(2450)] = 161807,
  [SMALL_STATE(2451)] = 161885,
  [SMALL_STATE(2452)] = 161941,
  [SMALL_STATE(2453)] = 161997,
  [SMALL_STATE(2454)] = 162053,
  [SMALL_STATE(2455)] = 162109,
  [SMALL_STATE(2456)] = 162165,
  [SMALL_STATE(2457)] = 162221,
  [SMALL_STATE(2458)] = 162277,
  [SMALL_STATE(2459)] = 162333,
  [SMALL_STATE(2460)] = 162389,
  [SMALL_STATE(2461)] = 162445,
  [SMALL_STATE(2462)] = 162501,
  [SMALL_STATE(2463)] = 162557,
  [SMALL_STATE(2464)] = 162617,
  [SMALL_STATE(2465)] = 162695,
  [SMALL_STATE(2466)] = 162773,
  [SMALL_STATE(2467)] = 162851,
  [SMALL_STATE(2468)] = 162907,
  [SMALL_STATE(2469)] = 162985,
  [SMALL_STATE(2470)] = 163043,
  [SMALL_STATE(2471)] = 163099,
  [SMALL_STATE(2472)] = 163154,
  [SMALL_STATE(2473)] = 163209,
  [SMALL_STATE(2474)] = 163264,
  [SMALL_STATE(2475)] = 163319,
  [SMALL_STATE(2476)] = 163374,
  [SMALL_STATE(2477)] = 163429,
  [SMALL_STATE(2478)] = 163484,
  [SMALL_STATE(2479)] = 163539,
  [SMALL_STATE(2480)] = 163594,
  [SMALL_STATE(2481)] = 163649,
  [SMALL_STATE(2482)] = 163704,
  [SMALL_STATE(2483)] = 163759,
  [SMALL_STATE(2484)] = 163814,
  [SMALL_STATE(2485)] = 163869,
  [SMALL_STATE(2486)] = 163924,
  [SMALL_STATE(2487)] = 163983,
  [SMALL_STATE(2488)] = 164038,
  [SMALL_STATE(2489)] = 164093,
  [SMALL_STATE(2490)] = 164152,
  [SMALL_STATE(2491)] = 164207,
  [SMALL_STATE(2492)] = 164262,
  [SMALL_STATE(2493)] = 164317,
  [SMALL_STATE(2494)] = 164372,
  [SMALL_STATE(2495)] = 164427,
  [SMALL_STATE(2496)] = 164482,
  [SMALL_STATE(2497)] = 164537,
  [SMALL_STATE(2498)] = 164592,
  [SMALL_STATE(2499)] = 164649,
  [SMALL_STATE(2500)] = 164704,
  [SMALL_STATE(2501)] = 164759,
  [SMALL_STATE(2502)] = 164814,
  [SMALL_STATE(2503)] = 164869,
  [SMALL_STATE(2504)] = 164924,
  [SMALL_STATE(2505)] = 164981,
  [SMALL_STATE(2506)] = 165036,
  [SMALL_STATE(2507)] = 165091,
  [SMALL_STATE(2508)] = 165146,
  [SMALL_STATE(2509)] = 165201,
  [SMALL_STATE(2510)] = 165256,
  [SMALL_STATE(2511)] = 165311,
  [SMALL_STATE(2512)] = 165366,
  [SMALL_STATE(2513)] = 165421,
  [SMALL_STATE(2514)] = 165476,
  [SMALL_STATE(2515)] = 165531,
  [SMALL_STATE(2516)] = 165586,
  [SMALL_STATE(2517)] = 165641,
  [SMALL_STATE(2518)] = 165696,
  [SMALL_STATE(2519)] = 165751,
  [SMALL_STATE(2520)] = 165806,
  [SMALL_STATE(2521)] = 165865,
  [SMALL_STATE(2522)] = 165920,
  [SMALL_STATE(2523)] = 165975,
  [SMALL_STATE(2524)] = 166030,
  [SMALL_STATE(2525)] = 166085,
  [SMALL_STATE(2526)] = 166140,
  [SMALL_STATE(2527)] = 166195,
  [SMALL_STATE(2528)] = 166250,
  [SMALL_STATE(2529)] = 166305,
  [SMALL_STATE(2530)] = 166360,
  [SMALL_STATE(2531)] = 166415,
  [SMALL_STATE(2532)] = 166470,
  [SMALL_STATE(2533)] = 166525,
  [SMALL_STATE(2534)] = 166580,
  [SMALL_STATE(2535)] = 166637,
  [SMALL_STATE(2536)] = 166692,
  [SMALL_STATE(2537)] = 166747,
  [SMALL_STATE(2538)] = 166801,
  [SMALL_STATE(2539)] = 166853,
  [SMALL_STATE(2540)] = 166907,
  [SMALL_STATE(2541)] = 166961,
  [SMALL_STATE(2542)] = 167015,
  [SMALL_STATE(2543)] = 167067,
  [SMALL_STATE(2544)] = 167121,
  [SMALL_STATE(2545)] = 167175,
  [SMALL_STATE(2546)] = 167229,
  [SMALL_STATE(2547)] = 167283,
  [SMALL_STATE(2548)] = 167337,
  [SMALL_STATE(2549)] = 167391,
  [SMALL_STATE(2550)] = 167445,
  [SMALL_STATE(2551)] = 167499,
  [SMALL_STATE(2552)] = 167553,
  [SMALL_STATE(2553)] = 167607,
  [SMALL_STATE(2554)] = 167661,
  [SMALL_STATE(2555)] = 167715,
  [SMALL_STATE(2556)] = 167769,
  [SMALL_STATE(2557)] = 167823,
  [SMALL_STATE(2558)] = 167877,
  [SMALL_STATE(2559)] = 167931,
  [SMALL_STATE(2560)] = 167985,
  [SMALL_STATE(2561)] = 168039,
  [SMALL_STATE(2562)] = 168093,
  [SMALL_STATE(2563)] = 168147,
  [SMALL_STATE(2564)] = 168201,
  [SMALL_STATE(2565)] = 168255,
  [SMALL_STATE(2566)] = 168309,
  [SMALL_STATE(2567)] = 168363,
  [SMALL_STATE(2568)] = 168417,
  [SMALL_STATE(2569)] = 168471,
  [SMALL_STATE(2570)] = 168525,
  [SMALL_STATE(2571)] = 168579,
  [SMALL_STATE(2572)] = 168633,
  [SMALL_STATE(2573)] = 168687,
  [SMALL_STATE(2574)] = 168741,
  [SMALL_STATE(2575)] = 168795,
  [SMALL_STATE(2576)] = 168851,
  [SMALL_STATE(2577)] = 168905,
  [SMALL_STATE(2578)] = 168959,
  [SMALL_STATE(2579)] = 169013,
  [SMALL_STATE(2580)] = 169065,
  [SMALL_STATE(2581)] = 169119,
  [SMALL_STATE(2582)] = 169173,
  [SMALL_STATE(2583)] = 169227,
  [SMALL_STATE(2584)] = 169281,
  [SMALL_STATE(2585)] = 169335,
  [SMALL_STATE(2586)] = 169389,
  [SMALL_STATE(2587)] = 169443,
  [SMALL_STATE(2588)] = 169497,
  [SMALL_STATE(2589)] = 169551,
  [SMALL_STATE(2590)] = 169605,
  [SMALL_STATE(2591)] = 169659,
  [SMALL_STATE(2592)] = 169713,
  [SMALL_STATE(2593)] = 169767,
  [SMALL_STATE(2594)] = 169819,
  [SMALL_STATE(2595)] = 169873,
  [SMALL_STATE(2596)] = 169927,
  [SMALL_STATE(2597)] = 169981,
  [SMALL_STATE(2598)] = 170035,
  [SMALL_STATE(2599)] = 170089,
  [SMALL_STATE(2600)] = 170143,
  [SMALL_STATE(2601)] = 170197,
  [SMALL_STATE(2602)] = 170251,
  [SMALL_STATE(2603)] = 170305,
  [SMALL_STATE(2604)] = 170359,
  [SMALL_STATE(2605)] = 170413,
  [SMALL_STATE(2606)] = 170467,
  [SMALL_STATE(2607)] = 170521,
  [SMALL_STATE(2608)] = 170575,
  [SMALL_STATE(2609)] = 170629,
  [SMALL_STATE(2610)] = 170685,
  [SMALL_STATE(2611)] = 170739,
  [SMALL_STATE(2612)] = 170793,
  [SMALL_STATE(2613)] = 170847,
  [SMALL_STATE(2614)] = 170899,
  [SMALL_STATE(2615)] = 170953,
  [SMALL_STATE(2616)] = 171007,
  [SMALL_STATE(2617)] = 171060,
  [SMALL_STATE(2618)] = 171111,
  [SMALL_STATE(2619)] = 171164,
  [SMALL_STATE(2620)] = 171215,
  [SMALL_STATE(2621)] = 171266,
  [SMALL_STATE(2622)] = 171319,
  [SMALL_STATE(2623)] = 171372,
  [SMALL_STATE(2624)] = 171425,
  [SMALL_STATE(2625)] = 171478,
  [SMALL_STATE(2626)] = 171531,
  [SMALL_STATE(2627)] = 171632,
  [SMALL_STATE(2628)] = 171685,
  [SMALL_STATE(2629)] = 171738,
  [SMALL_STATE(2630)] = 171789,
  [SMALL_STATE(2631)] = 171840,
  [SMALL_STATE(2632)] = 171893,
  [SMALL_STATE(2633)] = 171946,
  [SMALL_STATE(2634)] = 171999,
  [SMALL_STATE(2635)] = 172052,
  [SMALL_STATE(2636)] = 172105,
  [SMALL_STATE(2637)] = 172206,
  [SMALL_STATE(2638)] = 172257,
  [SMALL_STATE(2639)] = 172308,
  [SMALL_STATE(2640)] = 172361,
  [SMALL_STATE(2641)] = 172414,
  [SMALL_STATE(2642)] = 172467,
  [SMALL_STATE(2643)] = 172520,
  [SMALL_STATE(2644)] = 172573,
  [SMALL_STATE(2645)] = 172626,
  [SMALL_STATE(2646)] = 172677,
  [SMALL_STATE(2647)] = 172730,
  [SMALL_STATE(2648)] = 172783,
  [SMALL_STATE(2649)] = 172836,
  [SMALL_STATE(2650)] = 172889,
  [SMALL_STATE(2651)] = 172942,
  [SMALL_STATE(2652)] = 172995,
  [SMALL_STATE(2653)] = 173048,
  [SMALL_STATE(2654)] = 173101,
  [SMALL_STATE(2655)] = 173154,
  [SMALL_STATE(2656)] = 173207,
  [SMALL_STATE(2657)] = 173260,
  [SMALL_STATE(2658)] = 173313,
  [SMALL_STATE(2659)] = 173414,
  [SMALL_STATE(2660)] = 173467,
  [SMALL_STATE(2661)] = 173520,
  [SMALL_STATE(2662)] = 173573,
  [SMALL_STATE(2663)] = 173626,
  [SMALL_STATE(2664)] = 173679,
  [SMALL_STATE(2665)] = 173732,
  [SMALL_STATE(2666)] = 173785,
  [SMALL_STATE(2667)] = 173838,
  [SMALL_STATE(2668)] = 173891,
  [SMALL_STATE(2669)] = 173944,
  [SMALL_STATE(2670)] = 173997,
  [SMALL_STATE(2671)] = 174050,
  [SMALL_STATE(2672)] = 174103,
  [SMALL_STATE(2673)] = 174156,
  [SMALL_STATE(2674)] = 174209,
  [SMALL_STATE(2675)] = 174262,
  [SMALL_STATE(2676)] = 174317,
  [SMALL_STATE(2677)] = 174368,
  [SMALL_STATE(2678)] = 174421,
  [SMALL_STATE(2679)] = 174472,
  [SMALL_STATE(2680)] = 174525,
  [SMALL_STATE(2681)] = 174578,
  [SMALL_STATE(2682)] = 174631,
  [SMALL_STATE(2683)] = 174684,
  [SMALL_STATE(2684)] = 174737,
  [SMALL_STATE(2685)] = 174790,
  [SMALL_STATE(2686)] = 174845,
  [SMALL_STATE(2687)] = 174898,
  [SMALL_STATE(2688)] = 174951,
  [SMALL_STATE(2689)] = 175004,
  [SMALL_STATE(2690)] = 175105,
  [SMALL_STATE(2691)] = 175158,
  [SMALL_STATE(2692)] = 175209,
  [SMALL_STATE(2693)] = 175262,
  [SMALL_STATE(2694)] = 175314,
  [SMALL_STATE(2695)] = 175366,
  [SMALL_STATE(2696)] = 175418,
  [SMALL_STATE(2697)] = 175470,
  [SMALL_STATE(2698)] = 175522,
  [SMALL_STATE(2699)] = 175574,
  [SMALL_STATE(2700)] = 175626,
  [SMALL_STATE(2701)] = 175678,
  [SMALL_STATE(2702)] = 175730,
  [SMALL_STATE(2703)] = 175782,
  [SMALL_STATE(2704)] = 175834,
  [SMALL_STATE(2705)] = 175886,
  [SMALL_STATE(2706)] = 175938,
  [SMALL_STATE(2707)] = 175990,
  [SMALL_STATE(2708)] = 176044,
  [SMALL_STATE(2709)] = 176096,
  [SMALL_STATE(2710)] = 176148,
  [SMALL_STATE(2711)] = 176200,
  [SMALL_STATE(2712)] = 176252,
  [SMALL_STATE(2713)] = 176304,
  [SMALL_STATE(2714)] = 176356,
  [SMALL_STATE(2715)] = 176440,
  [SMALL_STATE(2716)] = 176492,
  [SMALL_STATE(2717)] = 176544,
  [SMALL_STATE(2718)] = 176596,
  [SMALL_STATE(2719)] = 176648,
  [SMALL_STATE(2720)] = 176700,
  [SMALL_STATE(2721)] = 176752,
  [SMALL_STATE(2722)] = 176804,
  [SMALL_STATE(2723)] = 176888,
  [SMALL_STATE(2724)] = 176940,
  [SMALL_STATE(2725)] = 177024,
  [SMALL_STATE(2726)] = 177076,
  [SMALL_STATE(2727)] = 177160,
  [SMALL_STATE(2728)] = 177212,
  [SMALL_STATE(2729)] = 177264,
  [SMALL_STATE(2730)] = 177316,
  [SMALL_STATE(2731)] = 177368,
  [SMALL_STATE(2732)] = 177420,
  [SMALL_STATE(2733)] = 177472,
  [SMALL_STATE(2734)] = 177524,
  [SMALL_STATE(2735)] = 177576,
  [SMALL_STATE(2736)] = 177628,
  [SMALL_STATE(2737)] = 177680,
  [SMALL_STATE(2738)] = 177732,
  [SMALL_STATE(2739)] = 177784,
  [SMALL_STATE(2740)] = 177836,
  [SMALL_STATE(2741)] = 177888,
  [SMALL_STATE(2742)] = 177940,
  [SMALL_STATE(2743)] = 177992,
  [SMALL_STATE(2744)] = 178044,
  [SMALL_STATE(2745)] = 178096,
  [SMALL_STATE(2746)] = 178148,
  [SMALL_STATE(2747)] = 178200,
  [SMALL_STATE(2748)] = 178252,
  [SMALL_STATE(2749)] = 178304,
  [SMALL_STATE(2750)] = 178356,
  [SMALL_STATE(2751)] = 178408,
  [SMALL_STATE(2752)] = 178462,
  [SMALL_STATE(2753)] = 178514,
  [SMALL_STATE(2754)] = 178566,
  [SMALL_STATE(2755)] = 178650,
  [SMALL_STATE(2756)] = 178702,
  [SMALL_STATE(2757)] = 178754,
  [SMALL_STATE(2758)] = 178806,
  [SMALL_STATE(2759)] = 178858,
  [SMALL_STATE(2760)] = 178910,
  [SMALL_STATE(2761)] = 178962,
  [SMALL_STATE(2762)] = 179014,
  [SMALL_STATE(2763)] = 179066,
  [SMALL_STATE(2764)] = 179118,
  [SMALL_STATE(2765)] = 179170,
  [SMALL_STATE(2766)] = 179222,
  [SMALL_STATE(2767)] = 179274,
  [SMALL_STATE(2768)] = 179326,
  [SMALL_STATE(2769)] = 179378,
  [SMALL_STATE(2770)] = 179430,
  [SMALL_STATE(2771)] = 179482,
  [SMALL_STATE(2772)] = 179534,
  [SMALL_STATE(2773)] = 179588,
  [SMALL_STATE(2774)] = 179669,
  [SMALL_STATE(2775)] = 179750,
  [SMALL_STATE(2776)] = 179831,
  [SMALL_STATE(2777)] = 179912,
  [SMALL_STATE(2778)] = 179993,
  [SMALL_STATE(2779)] = 180074,
  [SMALL_STATE(2780)] = 180155,
  [SMALL_STATE(2781)] = 180236,
  [SMALL_STATE(2782)] = 180317,
  [SMALL_STATE(2783)] = 180398,
  [SMALL_STATE(2784)] = 180479,
  [SMALL_STATE(2785)] = 180560,
  [SMALL_STATE(2786)] = 180641,
  [SMALL_STATE(2787)] = 180722,
  [SMALL_STATE(2788)] = 180803,
  [SMALL_STATE(2789)] = 180884,
  [SMALL_STATE(2790)] = 180965,
  [SMALL_STATE(2791)] = 181043,
  [SMALL_STATE(2792)] = 181121,
  [SMALL_STATE(2793)] = 181199,
  [SMALL_STATE(2794)] = 181277,
  [SMALL_STATE(2795)] = 181342,
  [SMALL_STATE(2796)] = 181417,
  [SMALL_STATE(2797)] = 181492,
  [SMALL_STATE(2798)] = 181555,
  [SMALL_STATE(2799)] = 181620,
  [SMALL_STATE(2800)] = 181684,
  [SMALL_STATE(2801)] = 181760,
  [SMALL_STATE(2802)] = 181824,
  [SMALL_STATE(2803)] = 181886,
  [SMALL_STATE(2804)] = 181962,
  [SMALL_STATE(2805)] = 182026,
  [SMALL_STATE(2806)] = 182100,
  [SMALL_STATE(2807)] = 182164,
  [SMALL_STATE(2808)] = 182226,
  [SMALL_STATE(2809)] = 182269,
  [SMALL_STATE(2810)] = 182312,
  [SMALL_STATE(2811)] = 182355,
  [SMALL_STATE(2812)] = 182398,
  [SMALL_STATE(2813)] = 182441,
  [SMALL_STATE(2814)] = 182484,
  [SMALL_STATE(2815)] = 182527,
  [SMALL_STATE(2816)] = 182570,
  [SMALL_STATE(2817)] = 182613,
  [SMALL_STATE(2818)] = 182656,
  [SMALL_STATE(2819)] = 182699,
  [SMALL_STATE(2820)] = 182742,
  [SMALL_STATE(2821)] = 182785,
  [SMALL_STATE(2822)] = 182828,
  [SMALL_STATE(2823)] = 182871,
  [SMALL_STATE(2824)] = 182914,
  [SMALL_STATE(2825)] = 182957,
  [SMALL_STATE(2826)] = 183000,
  [SMALL_STATE(2827)] = 183043,
  [SMALL_STATE(2828)] = 183086,
  [SMALL_STATE(2829)] = 183129,
  [SMALL_STATE(2830)] = 183172,
  [SMALL_STATE(2831)] = 183215,
  [SMALL_STATE(2832)] = 183258,
  [SMALL_STATE(2833)] = 183301,
  [SMALL_STATE(2834)] = 183344,
  [SMALL_STATE(2835)] = 183387,
  [SMALL_STATE(2836)] = 183430,
  [SMALL_STATE(2837)] = 183473,
  [SMALL_STATE(2838)] = 183534,
  [SMALL_STATE(2839)] = 183577,
  [SMALL_STATE(2840)] = 183620,
  [SMALL_STATE(2841)] = 183663,
  [SMALL_STATE(2842)] = 183706,
  [SMALL_STATE(2843)] = 183749,
  [SMALL_STATE(2844)] = 183792,
  [SMALL_STATE(2845)] = 183835,
  [SMALL_STATE(2846)] = 183878,
  [SMALL_STATE(2847)] = 183921,
  [SMALL_STATE(2848)] = 183964,
  [SMALL_STATE(2849)] = 184007,
  [SMALL_STATE(2850)] = 184070,
  [SMALL_STATE(2851)] = 184113,
  [SMALL_STATE(2852)] = 184156,
  [SMALL_STATE(2853)] = 184199,
  [SMALL_STATE(2854)] = 184242,
  [SMALL_STATE(2855)] = 184285,
  [SMALL_STATE(2856)] = 184328,
  [SMALL_STATE(2857)] = 184371,
  [SMALL_STATE(2858)] = 184414,
  [SMALL_STATE(2859)] = 184457,
  [SMALL_STATE(2860)] = 184520,
  [SMALL_STATE(2861)] = 184563,
  [SMALL_STATE(2862)] = 184606,
  [SMALL_STATE(2863)] = 184649,
  [SMALL_STATE(2864)] = 184692,
  [SMALL_STATE(2865)] = 184735,
  [SMALL_STATE(2866)] = 184778,
  [SMALL_STATE(2867)] = 184821,
  [SMALL_STATE(2868)] = 184864,
  [SMALL_STATE(2869)] = 184907,
  [SMALL_STATE(2870)] = 184950,
  [SMALL_STATE(2871)] = 184992,
  [SMALL_STATE(2872)] = 185034,
  [SMALL_STATE(2873)] = 185076,
  [SMALL_STATE(2874)] = 185118,
  [SMALL_STATE(2875)] = 185178,
  [SMALL_STATE(2876)] = 185220,
  [SMALL_STATE(2877)] = 185262,
  [SMALL_STATE(2878)] = 185304,
  [SMALL_STATE(2879)] = 185346,
  [SMALL_STATE(2880)] = 185388,
  [SMALL_STATE(2881)] = 185430,
  [SMALL_STATE(2882)] = 185472,
  [SMALL_STATE(2883)] = 185514,
  [SMALL_STATE(2884)] = 185556,
  [SMALL_STATE(2885)] = 185598,
  [SMALL_STATE(2886)] = 185640,
  [SMALL_STATE(2887)] = 185682,
  [SMALL_STATE(2888)] = 185758,
  [SMALL_STATE(2889)] = 185800,
  [SMALL_STATE(2890)] = 185842,
  [SMALL_STATE(2891)] = 185884,
  [SMALL_STATE(2892)] = 185926,
  [SMALL_STATE(2893)] = 185968,
  [SMALL_STATE(2894)] = 186010,
  [SMALL_STATE(2895)] = 186052,
  [SMALL_STATE(2896)] = 186094,
  [SMALL_STATE(2897)] = 186136,
  [SMALL_STATE(2898)] = 186178,
  [SMALL_STATE(2899)] = 186220,
  [SMALL_STATE(2900)] = 186262,
  [SMALL_STATE(2901)] = 186322,
  [SMALL_STATE(2902)] = 186364,
  [SMALL_STATE(2903)] = 186406,
  [SMALL_STATE(2904)] = 186448,
  [SMALL_STATE(2905)] = 186490,
  [SMALL_STATE(2906)] = 186532,
  [SMALL_STATE(2907)] = 186574,
  [SMALL_STATE(2908)] = 186616,
  [SMALL_STATE(2909)] = 186658,
  [SMALL_STATE(2910)] = 186700,
  [SMALL_STATE(2911)] = 186742,
  [SMALL_STATE(2912)] = 186784,
  [SMALL_STATE(2913)] = 186826,
  [SMALL_STATE(2914)] = 186868,
  [SMALL_STATE(2915)] = 186910,
  [SMALL_STATE(2916)] = 186952,
  [SMALL_STATE(2917)] = 186994,
  [SMALL_STATE(2918)] = 187036,
  [SMALL_STATE(2919)] = 187078,
  [SMALL_STATE(2920)] = 187120,
  [SMALL_STATE(2921)] = 187162,
  [SMALL_STATE(2922)] = 187204,
  [SMALL_STATE(2923)] = 187246,
  [SMALL_STATE(2924)] = 187288,
  [SMALL_STATE(2925)] = 187330,
  [SMALL_STATE(2926)] = 187372,
  [SMALL_STATE(2927)] = 187414,
  [SMALL_STATE(2928)] = 187456,
  [SMALL_STATE(2929)] = 187498,
  [SMALL_STATE(2930)] = 187540,
  [SMALL_STATE(2931)] = 187582,
  [SMALL_STATE(2932)] = 187624,
  [SMALL_STATE(2933)] = 187665,
  [SMALL_STATE(2934)] = 187706,
  [SMALL_STATE(2935)] = 187747,
  [SMALL_STATE(2936)] = 187808,
  [SMALL_STATE(2937)] = 187849,
  [SMALL_STATE(2938)] = 187890,
  [SMALL_STATE(2939)] = 187931,
  [SMALL_STATE(2940)] = 187972,
  [SMALL_STATE(2941)] = 188013,
  [SMALL_STATE(2942)] = 188054,
  [SMALL_STATE(2943)] = 188095,
  [SMALL_STATE(2944)] = 188136,
  [SMALL_STATE(2945)] = 188177,
  [SMALL_STATE(2946)] = 188218,
  [SMALL_STATE(2947)] = 188259,
  [SMALL_STATE(2948)] = 188300,
  [SMALL_STATE(2949)] = 188341,
  [SMALL_STATE(2950)] = 188382,
  [SMALL_STATE(2951)] = 188423,
  [SMALL_STATE(2952)] = 188464,
  [SMALL_STATE(2953)] = 188505,
  [SMALL_STATE(2954)] = 188546,
  [SMALL_STATE(2955)] = 188587,
  [SMALL_STATE(2956)] = 188628,
  [SMALL_STATE(2957)] = 188669,
  [SMALL_STATE(2958)] = 188710,
  [SMALL_STATE(2959)] = 188751,
  [SMALL_STATE(2960)] = 188792,
  [SMALL_STATE(2961)] = 188833,
  [SMALL_STATE(2962)] = 188874,
  [SMALL_STATE(2963)] = 188933,
  [SMALL_STATE(2964)] = 188980,
  [SMALL_STATE(2965)] = 189021,
  [SMALL_STATE(2966)] = 189062,
  [SMALL_STATE(2967)] = 189103,
  [SMALL_STATE(2968)] = 189144,
  [SMALL_STATE(2969)] = 189185,
  [SMALL_STATE(2970)] = 189226,
  [SMALL_STATE(2971)] = 189267,
  [SMALL_STATE(2972)] = 189308,
  [SMALL_STATE(2973)] = 189349,
  [SMALL_STATE(2974)] = 189390,
  [SMALL_STATE(2975)] = 189431,
  [SMALL_STATE(2976)] = 189472,
  [SMALL_STATE(2977)] = 189513,
  [SMALL_STATE(2978)] = 189554,
  [SMALL_STATE(2979)] = 189595,
  [SMALL_STATE(2980)] = 189636,
  [SMALL_STATE(2981)] = 189677,
  [SMALL_STATE(2982)] = 189718,
  [SMALL_STATE(2983)] = 189759,
  [SMALL_STATE(2984)] = 189800,
  [SMALL_STATE(2985)] = 189841,
  [SMALL_STATE(2986)] = 189882,
  [SMALL_STATE(2987)] = 189923,
  [SMALL_STATE(2988)] = 189964,
  [SMALL_STATE(2989)] = 190005,
  [SMALL_STATE(2990)] = 190046,
  [SMALL_STATE(2991)] = 190107,
  [SMALL_STATE(2992)] = 190148,
  [SMALL_STATE(2993)] = 190189,
  [SMALL_STATE(2994)] = 190230,
  [SMALL_STATE(2995)] = 190271,
  [SMALL_STATE(2996)] = 190312,
  [SMALL_STATE(2997)] = 190380,
  [SMALL_STATE(2998)] = 190426,
  [SMALL_STATE(2999)] = 190466,
  [SMALL_STATE(3000)] = 190534,
  [SMALL_STATE(3001)] = 190580,
  [SMALL_STATE(3002)] = 190626,
  [SMALL_STATE(3003)] = 190694,
  [SMALL_STATE(3004)] = 190734,
  [SMALL_STATE(3005)] = 190802,
  [SMALL_STATE(3006)] = 190870,
  [SMALL_STATE(3007)] = 190938,
  [SMALL_STATE(3008)] = 191006,
  [SMALL_STATE(3009)] = 191052,
  [SMALL_STATE(3010)] = 191098,
  [SMALL_STATE(3011)] = 191138,
  [SMALL_STATE(3012)] = 191206,
  [SMALL_STATE(3013)] = 191252,
  [SMALL_STATE(3014)] = 191308,
  [SMALL_STATE(3015)] = 191347,
  [SMALL_STATE(3016)] = 191390,
  [SMALL_STATE(3017)] = 191429,
  [SMALL_STATE(3018)] = 191468,
  [SMALL_STATE(3019)] = 191507,
  [SMALL_STATE(3020)] = 191546,
  [SMALL_STATE(3021)] = 191585,
  [SMALL_STATE(3022)] = 191646,
  [SMALL_STATE(3023)] = 191689,
  [SMALL_STATE(3024)] = 191750,
  [SMALL_STATE(3025)] = 191789,
  [SMALL_STATE(3026)] = 191848,
  [SMALL_STATE(3027)] = 191908,
  [SMALL_STATE(3028)] = 191968,
  [SMALL_STATE(3029)] = 192028,
  [SMALL_STATE(3030)] = 192088,
  [SMALL_STATE(3031)] = 192126,
  [SMALL_STATE(3032)] = 192190,
  [SMALL_STATE(3033)] = 192228,
  [SMALL_STATE(3034)] = 192266,
  [SMALL_STATE(3035)] = 192326,
  [SMALL_STATE(3036)] = 192364,
  [SMALL_STATE(3037)] = 192402,
  [SMALL_STATE(3038)] = 192440,
  [SMALL_STATE(3039)] = 192478,
  [SMALL_STATE(3040)] = 192516,
  [SMALL_STATE(3041)] = 192554,
  [SMALL_STATE(3042)] = 192592,
  [SMALL_STATE(3043)] = 192630,
  [SMALL_STATE(3044)] = 192668,
  [SMALL_STATE(3045)] = 192706,
  [SMALL_STATE(3046)] = 192744,
  [SMALL_STATE(3047)] = 192782,
  [SMALL_STATE(3048)] = 192820,
  [SMALL_STATE(3049)] = 192858,
  [SMALL_STATE(3050)] = 192896,
  [SMALL_STATE(3051)] = 192934,
  [SMALL_STATE(3052)] = 192974,
  [SMALL_STATE(3053)] = 193012,
  [SMALL_STATE(3054)] = 193050,
  [SMALL_STATE(3055)] = 193088,
  [SMALL_STATE(3056)] = 193126,
  [SMALL_STATE(3057)] = 193164,
  [SMALL_STATE(3058)] = 193202,
  [SMALL_STATE(3059)] = 193240,
  [SMALL_STATE(3060)] = 193284,
  [SMALL_STATE(3061)] = 193322,
  [SMALL_STATE(3062)] = 193360,
  [SMALL_STATE(3063)] = 193398,
  [SMALL_STATE(3064)] = 193436,
  [SMALL_STATE(3065)] = 193474,
  [SMALL_STATE(3066)] = 193512,
  [SMALL_STATE(3067)] = 193572,
  [SMALL_STATE(3068)] = 193632,
  [SMALL_STATE(3069)] = 193670,
  [SMALL_STATE(3070)] = 193708,
  [SMALL_STATE(3071)] = 193746,
  [SMALL_STATE(3072)] = 193784,
  [SMALL_STATE(3073)] = 193844,
  [SMALL_STATE(3074)] = 193904,
  [SMALL_STATE(3075)] = 193968,
  [SMALL_STATE(3076)] = 194032,
  [SMALL_STATE(3077)] = 194070,
  [SMALL_STATE(3078)] = 194128,
  [SMALL_STATE(3079)] = 194166,
  [SMALL_STATE(3080)] = 194226,
  [SMALL_STATE(3081)] = 194286,
  [SMALL_STATE(3082)] = 194346,
  [SMALL_STATE(3083)] = 194384,
  [SMALL_STATE(3084)] = 194422,
  [SMALL_STATE(3085)] = 194482,
  [SMALL_STATE(3086)] = 194542,
  [SMALL_STATE(3087)] = 194602,
  [SMALL_STATE(3088)] = 194640,
  [SMALL_STATE(3089)] = 194700,
  [SMALL_STATE(3090)] = 194760,
  [SMALL_STATE(3091)] = 194820,
  [SMALL_STATE(3092)] = 194880,
  [SMALL_STATE(3093)] = 194940,
  [SMALL_STATE(3094)] = 195000,
  [SMALL_STATE(3095)] = 195060,
  [SMALL_STATE(3096)] = 195120,
  [SMALL_STATE(3097)] = 195180,
  [SMALL_STATE(3098)] = 195240,
  [SMALL_STATE(3099)] = 195300,
  [SMALL_STATE(3100)] = 195360,
  [SMALL_STATE(3101)] = 195420,
  [SMALL_STATE(3102)] = 195480,
  [SMALL_STATE(3103)] = 195540,
  [SMALL_STATE(3104)] = 195600,
  [SMALL_STATE(3105)] = 195660,
  [SMALL_STATE(3106)] = 195720,
  [SMALL_STATE(3107)] = 195780,
  [SMALL_STATE(3108)] = 195840,
  [SMALL_STATE(3109)] = 195900,
  [SMALL_STATE(3110)] = 195960,
  [SMALL_STATE(3111)] = 196020,
  [SMALL_STATE(3112)] = 196080,
  [SMALL_STATE(3113)] = 196140,
  [SMALL_STATE(3114)] = 196200,
  [SMALL_STATE(3115)] = 196260,
  [SMALL_STATE(3116)] = 196320,
  [SMALL_STATE(3117)] = 196358,
  [SMALL_STATE(3118)] = 196418,
  [SMALL_STATE(3119)] = 196478,
  [SMALL_STATE(3120)] = 196538,
  [SMALL_STATE(3121)] = 196598,
  [SMALL_STATE(3122)] = 196658,
  [SMALL_STATE(3123)] = 196718,
  [SMALL_STATE(3124)] = 196778,
  [SMALL_STATE(3125)] = 196838,
  [SMALL_STATE(3126)] = 196898,
  [SMALL_STATE(3127)] = 196958,
  [SMALL_STATE(3128)] = 197018,
  [SMALL_STATE(3129)] = 197078,
  [SMALL_STATE(3130)] = 197138,
  [SMALL_STATE(3131)] = 197198,
  [SMALL_STATE(3132)] = 197258,
  [SMALL_STATE(3133)] = 197318,
  [SMALL_STATE(3134)] = 197378,
  [SMALL_STATE(3135)] = 197438,
  [SMALL_STATE(3136)] = 197498,
  [SMALL_STATE(3137)] = 197558,
  [SMALL_STATE(3138)] = 197618,
  [SMALL_STATE(3139)] = 197678,
  [SMALL_STATE(3140)] = 197738,
  [SMALL_STATE(3141)] = 197798,
  [SMALL_STATE(3142)] = 197858,
  [SMALL_STATE(3143)] = 197918,
  [SMALL_STATE(3144)] = 197978,
  [SMALL_STATE(3145)] = 198038,
  [SMALL_STATE(3146)] = 198098,
  [SMALL_STATE(3147)] = 198158,
  [SMALL_STATE(3148)] = 198218,
  [SMALL_STATE(3149)] = 198278,
  [SMALL_STATE(3150)] = 198316,
  [SMALL_STATE(3151)] = 198376,
  [SMALL_STATE(3152)] = 198436,
  [SMALL_STATE(3153)] = 198496,
  [SMALL_STATE(3154)] = 198556,
  [SMALL_STATE(3155)] = 198616,
  [SMALL_STATE(3156)] = 198676,
  [SMALL_STATE(3157)] = 198736,
  [SMALL_STATE(3158)] = 198796,
  [SMALL_STATE(3159)] = 198856,
  [SMALL_STATE(3160)] = 198916,
  [SMALL_STATE(3161)] = 198976,
  [SMALL_STATE(3162)] = 199036,
  [SMALL_STATE(3163)] = 199096,
  [SMALL_STATE(3164)] = 199156,
  [SMALL_STATE(3165)] = 199216,
  [SMALL_STATE(3166)] = 199276,
  [SMALL_STATE(3167)] = 199336,
  [SMALL_STATE(3168)] = 199396,
  [SMALL_STATE(3169)] = 199456,
  [SMALL_STATE(3170)] = 199516,
  [SMALL_STATE(3171)] = 199576,
  [SMALL_STATE(3172)] = 199636,
  [SMALL_STATE(3173)] = 199696,
  [SMALL_STATE(3174)] = 199756,
  [SMALL_STATE(3175)] = 199816,
  [SMALL_STATE(3176)] = 199876,
  [SMALL_STATE(3177)] = 199936,
  [SMALL_STATE(3178)] = 199996,
  [SMALL_STATE(3179)] = 200056,
  [SMALL_STATE(3180)] = 200116,
  [SMALL_STATE(3181)] = 200176,
  [SMALL_STATE(3182)] = 200236,
  [SMALL_STATE(3183)] = 200296,
  [SMALL_STATE(3184)] = 200356,
  [SMALL_STATE(3185)] = 200416,
  [SMALL_STATE(3186)] = 200476,
  [SMALL_STATE(3187)] = 200536,
  [SMALL_STATE(3188)] = 200596,
  [SMALL_STATE(3189)] = 200656,
  [SMALL_STATE(3190)] = 200716,
  [SMALL_STATE(3191)] = 200776,
  [SMALL_STATE(3192)] = 200836,
  [SMALL_STATE(3193)] = 200896,
  [SMALL_STATE(3194)] = 200956,
  [SMALL_STATE(3195)] = 201016,
  [SMALL_STATE(3196)] = 201076,
  [SMALL_STATE(3197)] = 201136,
  [SMALL_STATE(3198)] = 201196,
  [SMALL_STATE(3199)] = 201256,
  [SMALL_STATE(3200)] = 201316,
  [SMALL_STATE(3201)] = 201376,
  [SMALL_STATE(3202)] = 201436,
  [SMALL_STATE(3203)] = 201496,
  [SMALL_STATE(3204)] = 201556,
  [SMALL_STATE(3205)] = 201616,
  [SMALL_STATE(3206)] = 201676,
  [SMALL_STATE(3207)] = 201736,
  [SMALL_STATE(3208)] = 201796,
  [SMALL_STATE(3209)] = 201856,
  [SMALL_STATE(3210)] = 201916,
  [SMALL_STATE(3211)] = 201976,
  [SMALL_STATE(3212)] = 202036,
  [SMALL_STATE(3213)] = 202096,
  [SMALL_STATE(3214)] = 202134,
  [SMALL_STATE(3215)] = 202194,
  [SMALL_STATE(3216)] = 202254,
  [SMALL_STATE(3217)] = 202314,
  [SMALL_STATE(3218)] = 202352,
  [SMALL_STATE(3219)] = 202412,
  [SMALL_STATE(3220)] = 202472,
  [SMALL_STATE(3221)] = 202532,
  [SMALL_STATE(3222)] = 202592,
  [SMALL_STATE(3223)] = 202652,
  [SMALL_STATE(3224)] = 202712,
  [SMALL_STATE(3225)] = 202750,
  [SMALL_STATE(3226)] = 202810,
  [SMALL_STATE(3227)] = 202870,
  [SMALL_STATE(3228)] = 202930,
  [SMALL_STATE(3229)] = 202990,
  [SMALL_STATE(3230)] = 203050,
  [SMALL_STATE(3231)] = 203110,
  [SMALL_STATE(3232)] = 203170,
  [SMALL_STATE(3233)] = 203230,
  [SMALL_STATE(3234)] = 203290,
  [SMALL_STATE(3235)] = 203350,
  [SMALL_STATE(3236)] = 203410,
  [SMALL_STATE(3237)] = 203470,
  [SMALL_STATE(3238)] = 203507,
  [SMALL_STATE(3239)] = 203544,
  [SMALL_STATE(3240)] = 203581,
  [SMALL_STATE(3241)] = 203618,
  [SMALL_STATE(3242)] = 203655,
  [SMALL_STATE(3243)] = 203692,
  [SMALL_STATE(3244)] = 203729,
  [SMALL_STATE(3245)] = 203766,
  [SMALL_STATE(3246)] = 203803,
  [SMALL_STATE(3247)] = 203840,
  [SMALL_STATE(3248)] = 203877,
  [SMALL_STATE(3249)] = 203914,
  [SMALL_STATE(3250)] = 203951,
  [SMALL_STATE(3251)] = 203988,
  [SMALL_STATE(3252)] = 204025,
  [SMALL_STATE(3253)] = 204062,
  [SMALL_STATE(3254)] = 204099,
  [SMALL_STATE(3255)] = 204136,
  [SMALL_STATE(3256)] = 204173,
  [SMALL_STATE(3257)] = 204210,
  [SMALL_STATE(3258)] = 204247,
  [SMALL_STATE(3259)] = 204284,
  [SMALL_STATE(3260)] = 204321,
  [SMALL_STATE(3261)] = 204358,
  [SMALL_STATE(3262)] = 204395,
  [SMALL_STATE(3263)] = 204432,
  [SMALL_STATE(3264)] = 204469,
  [SMALL_STATE(3265)] = 204506,
  [SMALL_STATE(3266)] = 204543,
  [SMALL_STATE(3267)] = 204580,
  [SMALL_STATE(3268)] = 204617,
  [SMALL_STATE(3269)] = 204654,
  [SMALL_STATE(3270)] = 204691,
  [SMALL_STATE(3271)] = 204734,
  [SMALL_STATE(3272)] = 204771,
  [SMALL_STATE(3273)] = 204808,
  [SMALL_STATE(3274)] = 204845,
  [SMALL_STATE(3275)] = 204882,
  [SMALL_STATE(3276)] = 204941,
  [SMALL_STATE(3277)] = 204978,
  [SMALL_STATE(3278)] = 205015,
  [SMALL_STATE(3279)] = 205052,
  [SMALL_STATE(3280)] = 205095,
  [SMALL_STATE(3281)] = 205132,
  [SMALL_STATE(3282)] = 205189,
  [SMALL_STATE(3283)] = 205226,
  [SMALL_STATE(3284)] = 205263,
  [SMALL_STATE(3285)] = 205300,
  [SMALL_STATE(3286)] = 205337,
  [SMALL_STATE(3287)] = 205374,
  [SMALL_STATE(3288)] = 205431,
  [SMALL_STATE(3289)] = 205468,
  [SMALL_STATE(3290)] = 205505,
  [SMALL_STATE(3291)] = 205542,
  [SMALL_STATE(3292)] = 205579,
  [SMALL_STATE(3293)] = 205616,
  [SMALL_STATE(3294)] = 205653,
  [SMALL_STATE(3295)] = 205690,
  [SMALL_STATE(3296)] = 205727,
  [SMALL_STATE(3297)] = 205764,
  [SMALL_STATE(3298)] = 205801,
  [SMALL_STATE(3299)] = 205838,
  [SMALL_STATE(3300)] = 205875,
  [SMALL_STATE(3301)] = 205912,
  [SMALL_STATE(3302)] = 205949,
  [SMALL_STATE(3303)] = 205986,
  [SMALL_STATE(3304)] = 206023,
  [SMALL_STATE(3305)] = 206060,
  [SMALL_STATE(3306)] = 206097,
  [SMALL_STATE(3307)] = 206152,
  [SMALL_STATE(3308)] = 206189,
  [SMALL_STATE(3309)] = 206226,
  [SMALL_STATE(3310)] = 206283,
  [SMALL_STATE(3311)] = 206320,
  [SMALL_STATE(3312)] = 206357,
  [SMALL_STATE(3313)] = 206394,
  [SMALL_STATE(3314)] = 206431,
  [SMALL_STATE(3315)] = 206468,
  [SMALL_STATE(3316)] = 206505,
  [SMALL_STATE(3317)] = 206542,
  [SMALL_STATE(3318)] = 206601,
  [SMALL_STATE(3319)] = 206638,
  [SMALL_STATE(3320)] = 206675,
  [SMALL_STATE(3321)] = 206712,
  [SMALL_STATE(3322)] = 206749,
  [SMALL_STATE(3323)] = 206786,
  [SMALL_STATE(3324)] = 206823,
  [SMALL_STATE(3325)] = 206860,
  [SMALL_STATE(3326)] = 206897,
  [SMALL_STATE(3327)] = 206934,
  [SMALL_STATE(3328)] = 206971,
  [SMALL_STATE(3329)] = 207008,
  [SMALL_STATE(3330)] = 207045,
  [SMALL_STATE(3331)] = 207082,
  [SMALL_STATE(3332)] = 207119,
  [SMALL_STATE(3333)] = 207156,
  [SMALL_STATE(3334)] = 207193,
  [SMALL_STATE(3335)] = 207230,
  [SMALL_STATE(3336)] = 207267,
  [SMALL_STATE(3337)] = 207304,
  [SMALL_STATE(3338)] = 207341,
  [SMALL_STATE(3339)] = 207378,
  [SMALL_STATE(3340)] = 207415,
  [SMALL_STATE(3341)] = 207452,
  [SMALL_STATE(3342)] = 207489,
  [SMALL_STATE(3343)] = 207526,
  [SMALL_STATE(3344)] = 207563,
  [SMALL_STATE(3345)] = 207600,
  [SMALL_STATE(3346)] = 207637,
  [SMALL_STATE(3347)] = 207674,
  [SMALL_STATE(3348)] = 207711,
  [SMALL_STATE(3349)] = 207748,
  [SMALL_STATE(3350)] = 207785,
  [SMALL_STATE(3351)] = 207822,
  [SMALL_STATE(3352)] = 207859,
  [SMALL_STATE(3353)] = 207918,
  [SMALL_STATE(3354)] = 207954,
  [SMALL_STATE(3355)] = 207990,
  [SMALL_STATE(3356)] = 208026,
  [SMALL_STATE(3357)] = 208062,
  [SMALL_STATE(3358)] = 208098,
  [SMALL_STATE(3359)] = 208134,
  [SMALL_STATE(3360)] = 208170,
  [SMALL_STATE(3361)] = 208206,
  [SMALL_STATE(3362)] = 208242,
  [SMALL_STATE(3363)] = 208298,
  [SMALL_STATE(3364)] = 208352,
  [SMALL_STATE(3365)] = 208388,
  [SMALL_STATE(3366)] = 208444,
  [SMALL_STATE(3367)] = 208498,
  [SMALL_STATE(3368)] = 208534,
  [SMALL_STATE(3369)] = 208570,
  [SMALL_STATE(3370)] = 208606,
  [SMALL_STATE(3371)] = 208662,
  [SMALL_STATE(3372)] = 208698,
  [SMALL_STATE(3373)] = 208734,
  [SMALL_STATE(3374)] = 208790,
  [SMALL_STATE(3375)] = 208826,
  [SMALL_STATE(3376)] = 208862,
  [SMALL_STATE(3377)] = 208918,
  [SMALL_STATE(3378)] = 208954,
  [SMALL_STATE(3379)] = 208990,
  [SMALL_STATE(3380)] = 209026,
  [SMALL_STATE(3381)] = 209062,
  [SMALL_STATE(3382)] = 209118,
  [SMALL_STATE(3383)] = 209154,
  [SMALL_STATE(3384)] = 209190,
  [SMALL_STATE(3385)] = 209226,
  [SMALL_STATE(3386)] = 209262,
  [SMALL_STATE(3387)] = 209316,
  [SMALL_STATE(3388)] = 209352,
  [SMALL_STATE(3389)] = 209388,
  [SMALL_STATE(3390)] = 209442,
  [SMALL_STATE(3391)] = 209478,
  [SMALL_STATE(3392)] = 209514,
  [SMALL_STATE(3393)] = 209550,
  [SMALL_STATE(3394)] = 209604,
  [SMALL_STATE(3395)] = 209660,
  [SMALL_STATE(3396)] = 209696,
  [SMALL_STATE(3397)] = 209732,
  [SMALL_STATE(3398)] = 209788,
  [SMALL_STATE(3399)] = 209824,
  [SMALL_STATE(3400)] = 209860,
  [SMALL_STATE(3401)] = 209896,
  [SMALL_STATE(3402)] = 209932,
  [SMALL_STATE(3403)] = 209968,
  [SMALL_STATE(3404)] = 210004,
  [SMALL_STATE(3405)] = 210040,
  [SMALL_STATE(3406)] = 210096,
  [SMALL_STATE(3407)] = 210132,
  [SMALL_STATE(3408)] = 210168,
  [SMALL_STATE(3409)] = 210204,
  [SMALL_STATE(3410)] = 210240,
  [SMALL_STATE(3411)] = 210276,
  [SMALL_STATE(3412)] = 210312,
  [SMALL_STATE(3413)] = 210348,
  [SMALL_STATE(3414)] = 210384,
  [SMALL_STATE(3415)] = 210420,
  [SMALL_STATE(3416)] = 210462,
  [SMALL_STATE(3417)] = 210516,
  [SMALL_STATE(3418)] = 210552,
  [SMALL_STATE(3419)] = 210588,
  [SMALL_STATE(3420)] = 210624,
  [SMALL_STATE(3421)] = 210678,
  [SMALL_STATE(3422)] = 210714,
  [SMALL_STATE(3423)] = 210750,
  [SMALL_STATE(3424)] = 210786,
  [SMALL_STATE(3425)] = 210822,
  [SMALL_STATE(3426)] = 210858,
  [SMALL_STATE(3427)] = 210894,
  [SMALL_STATE(3428)] = 210930,
  [SMALL_STATE(3429)] = 210986,
  [SMALL_STATE(3430)] = 211042,
  [SMALL_STATE(3431)] = 211096,
  [SMALL_STATE(3432)] = 211150,
  [SMALL_STATE(3433)] = 211204,
  [SMALL_STATE(3434)] = 211249,
  [SMALL_STATE(3435)] = 211302,
  [SMALL_STATE(3436)] = 211355,
  [SMALL_STATE(3437)] = 211408,
  [SMALL_STATE(3438)] = 211453,
  [SMALL_STATE(3439)] = 211506,
  [SMALL_STATE(3440)] = 211559,
  [SMALL_STATE(3441)] = 211612,
  [SMALL_STATE(3442)] = 211665,
  [SMALL_STATE(3443)] = 211718,
  [SMALL_STATE(3444)] = 211771,
  [SMALL_STATE(3445)] = 211824,
  [SMALL_STATE(3446)] = 211877,
  [SMALL_STATE(3447)] = 211930,
  [SMALL_STATE(3448)] = 211983,
  [SMALL_STATE(3449)] = 212036,
  [SMALL_STATE(3450)] = 212089,
  [SMALL_STATE(3451)] = 212124,
  [SMALL_STATE(3452)] = 212177,
  [SMALL_STATE(3453)] = 212222,
  [SMALL_STATE(3454)] = 212275,
  [SMALL_STATE(3455)] = 212330,
  [SMALL_STATE(3456)] = 212383,
  [SMALL_STATE(3457)] = 212436,
  [SMALL_STATE(3458)] = 212481,
  [SMALL_STATE(3459)] = 212526,
  [SMALL_STATE(3460)] = 212579,
  [SMALL_STATE(3461)] = 212632,
  [SMALL_STATE(3462)] = 212685,
  [SMALL_STATE(3463)] = 212738,
  [SMALL_STATE(3464)] = 212791,
  [SMALL_STATE(3465)] = 212844,
  [SMALL_STATE(3466)] = 212897,
  [SMALL_STATE(3467)] = 212942,
  [SMALL_STATE(3468)] = 212995,
  [SMALL_STATE(3469)] = 213048,
  [SMALL_STATE(3470)] = 213101,
  [SMALL_STATE(3471)] = 213154,
  [SMALL_STATE(3472)] = 213207,
  [SMALL_STATE(3473)] = 213260,
  [SMALL_STATE(3474)] = 213313,
  [SMALL_STATE(3475)] = 213366,
  [SMALL_STATE(3476)] = 213419,
  [SMALL_STATE(3477)] = 213472,
  [SMALL_STATE(3478)] = 213525,
  [SMALL_STATE(3479)] = 213570,
  [SMALL_STATE(3480)] = 213623,
  [SMALL_STATE(3481)] = 213676,
  [SMALL_STATE(3482)] = 213731,
  [SMALL_STATE(3483)] = 213776,
  [SMALL_STATE(3484)] = 213829,
  [SMALL_STATE(3485)] = 213882,
  [SMALL_STATE(3486)] = 213935,
  [SMALL_STATE(3487)] = 213990,
  [SMALL_STATE(3488)] = 214025,
  [SMALL_STATE(3489)] = 214078,
  [SMALL_STATE(3490)] = 214123,
  [SMALL_STATE(3491)] = 214178,
  [SMALL_STATE(3492)] = 214223,
  [SMALL_STATE(3493)] = 214268,
  [SMALL_STATE(3494)] = 214323,
  [SMALL_STATE(3495)] = 214368,
  [SMALL_STATE(3496)] = 214413,
  [SMALL_STATE(3497)] = 214458,
  [SMALL_STATE(3498)] = 214503,
  [SMALL_STATE(3499)] = 214548,
  [SMALL_STATE(3500)] = 214593,
  [SMALL_STATE(3501)] = 214646,
  [SMALL_STATE(3502)] = 214691,
  [SMALL_STATE(3503)] = 214736,
  [SMALL_STATE(3504)] = 214781,
  [SMALL_STATE(3505)] = 214826,
  [SMALL_STATE(3506)] = 214879,
  [SMALL_STATE(3507)] = 214924,
  [SMALL_STATE(3508)] = 214969,
  [SMALL_STATE(3509)] = 215022,
  [SMALL_STATE(3510)] = 215067,
  [SMALL_STATE(3511)] = 215112,
  [SMALL_STATE(3512)] = 215167,
  [SMALL_STATE(3513)] = 215212,
  [SMALL_STATE(3514)] = 215257,
  [SMALL_STATE(3515)] = 215310,
  [SMALL_STATE(3516)] = 215363,
  [SMALL_STATE(3517)] = 215408,
  [SMALL_STATE(3518)] = 215453,
  [SMALL_STATE(3519)] = 215488,
  [SMALL_STATE(3520)] = 215533,
  [SMALL_STATE(3521)] = 215568,
  [SMALL_STATE(3522)] = 215613,
  [SMALL_STATE(3523)] = 215648,
  [SMALL_STATE(3524)] = 215683,
  [SMALL_STATE(3525)] = 215728,
  [SMALL_STATE(3526)] = 215773,
  [SMALL_STATE(3527)] = 215808,
  [SMALL_STATE(3528)] = 215843,
  [SMALL_STATE(3529)] = 215878,
  [SMALL_STATE(3530)] = 215913,
  [SMALL_STATE(3531)] = 215948,
  [SMALL_STATE(3532)] = 215983,
  [SMALL_STATE(3533)] = 216018,
  [SMALL_STATE(3534)] = 216063,
  [SMALL_STATE(3535)] = 216098,
  [SMALL_STATE(3536)] = 216133,
  [SMALL_STATE(3537)] = 216168,
  [SMALL_STATE(3538)] = 216203,
  [SMALL_STATE(3539)] = 216238,
  [SMALL_STATE(3540)] = 216273,
  [SMALL_STATE(3541)] = 216318,
  [SMALL_STATE(3542)] = 216369,
  [SMALL_STATE(3543)] = 216404,
  [SMALL_STATE(3544)] = 216457,
  [SMALL_STATE(3545)] = 216492,
  [SMALL_STATE(3546)] = 216537,
  [SMALL_STATE(3547)] = 216572,
  [SMALL_STATE(3548)] = 216607,
  [SMALL_STATE(3549)] = 216642,
  [SMALL_STATE(3550)] = 216677,
  [SMALL_STATE(3551)] = 216712,
  [SMALL_STATE(3552)] = 216757,
  [SMALL_STATE(3553)] = 216802,
  [SMALL_STATE(3554)] = 216847,
  [SMALL_STATE(3555)] = 216892,
  [SMALL_STATE(3556)] = 216937,
  [SMALL_STATE(3557)] = 216982,
  [SMALL_STATE(3558)] = 217027,
  [SMALL_STATE(3559)] = 217072,
  [SMALL_STATE(3560)] = 217117,
  [SMALL_STATE(3561)] = 217170,
  [SMALL_STATE(3562)] = 217215,
  [SMALL_STATE(3563)] = 217268,
  [SMALL_STATE(3564)] = 217303,
  [SMALL_STATE(3565)] = 217348,
  [SMALL_STATE(3566)] = 217393,
  [SMALL_STATE(3567)] = 217428,
  [SMALL_STATE(3568)] = 217463,
  [SMALL_STATE(3569)] = 217508,
  [SMALL_STATE(3570)] = 217553,
  [SMALL_STATE(3571)] = 217606,
  [SMALL_STATE(3572)] = 217659,
  [SMALL_STATE(3573)] = 217712,
  [SMALL_STATE(3574)] = 217765,
  [SMALL_STATE(3575)] = 217818,
  [SMALL_STATE(3576)] = 217871,
  [SMALL_STATE(3577)] = 217924,
  [SMALL_STATE(3578)] = 217977,
  [SMALL_STATE(3579)] = 218030,
  [SMALL_STATE(3580)] = 218083,
  [SMALL_STATE(3581)] = 218136,
  [SMALL_STATE(3582)] = 218189,
  [SMALL_STATE(3583)] = 218242,
  [SMALL_STATE(3584)] = 218295,
  [SMALL_STATE(3585)] = 218348,
  [SMALL_STATE(3586)] = 218393,
  [SMALL_STATE(3587)] = 218443,
  [SMALL_STATE(3588)] = 218495,
  [SMALL_STATE(3589)] = 218545,
  [SMALL_STATE(3590)] = 218595,
  [SMALL_STATE(3591)] = 218645,
  [SMALL_STATE(3592)] = 218695,
  [SMALL_STATE(3593)] = 218745,
  [SMALL_STATE(3594)] = 218797,
  [SMALL_STATE(3595)] = 218847,
  [SMALL_STATE(3596)] = 218899,
  [SMALL_STATE(3597)] = 218949,
  [SMALL_STATE(3598)] = 218999,
  [SMALL_STATE(3599)] = 219049,
  [SMALL_STATE(3600)] = 219099,
  [SMALL_STATE(3601)] = 219149,
  [SMALL_STATE(3602)] = 219199,
  [SMALL_STATE(3603)] = 219249,
  [SMALL_STATE(3604)] = 219301,
  [SMALL_STATE(3605)] = 219351,
  [SMALL_STATE(3606)] = 219401,
  [SMALL_STATE(3607)] = 219451,
  [SMALL_STATE(3608)] = 219501,
  [SMALL_STATE(3609)] = 219551,
  [SMALL_STATE(3610)] = 219605,
  [SMALL_STATE(3611)] = 219639,
  [SMALL_STATE(3612)] = 219689,
  [SMALL_STATE(3613)] = 219739,
  [SMALL_STATE(3614)] = 219789,
  [SMALL_STATE(3615)] = 219839,
  [SMALL_STATE(3616)] = 219891,
  [SMALL_STATE(3617)] = 219941,
  [SMALL_STATE(3618)] = 219983,
  [SMALL_STATE(3619)] = 220033,
  [SMALL_STATE(3620)] = 220083,
  [SMALL_STATE(3621)] = 220133,
  [SMALL_STATE(3622)] = 220185,
  [SMALL_STATE(3623)] = 220235,
  [SMALL_STATE(3624)] = 220289,
  [SMALL_STATE(3625)] = 220339,
  [SMALL_STATE(3626)] = 220393,
  [SMALL_STATE(3627)] = 220443,
  [SMALL_STATE(3628)] = 220497,
  [SMALL_STATE(3629)] = 220547,
  [SMALL_STATE(3630)] = 220597,
  [SMALL_STATE(3631)] = 220647,
  [SMALL_STATE(3632)] = 220697,
  [SMALL_STATE(3633)] = 220747,
  [SMALL_STATE(3634)] = 220797,
  [SMALL_STATE(3635)] = 220847,
  [SMALL_STATE(3636)] = 220897,
  [SMALL_STATE(3637)] = 220949,
  [SMALL_STATE(3638)] = 220999,
  [SMALL_STATE(3639)] = 221049,
  [SMALL_STATE(3640)] = 221101,
  [SMALL_STATE(3641)] = 221155,
  [SMALL_STATE(3642)] = 221209,
  [SMALL_STATE(3643)] = 221261,
  [SMALL_STATE(3644)] = 221311,
  [SMALL_STATE(3645)] = 221361,
  [SMALL_STATE(3646)] = 221411,
  [SMALL_STATE(3647)] = 221463,
  [SMALL_STATE(3648)] = 221503,
  [SMALL_STATE(3649)] = 221553,
  [SMALL_STATE(3650)] = 221603,
  [SMALL_STATE(3651)] = 221653,
  [SMALL_STATE(3652)] = 221695,
  [SMALL_STATE(3653)] = 221745,
  [SMALL_STATE(3654)] = 221795,
  [SMALL_STATE(3655)] = 221845,
  [SMALL_STATE(3656)] = 221887,
  [SMALL_STATE(3657)] = 221937,
  [SMALL_STATE(3658)] = 221987,
  [SMALL_STATE(3659)] = 222037,
  [SMALL_STATE(3660)] = 222079,
  [SMALL_STATE(3661)] = 222131,
  [SMALL_STATE(3662)] = 222181,
  [SMALL_STATE(3663)] = 222233,
  [SMALL_STATE(3664)] = 222275,
  [SMALL_STATE(3665)] = 222325,
  [SMALL_STATE(3666)] = 222367,
  [SMALL_STATE(3667)] = 222417,
  [SMALL_STATE(3668)] = 222471,
  [SMALL_STATE(3669)] = 222525,
  [SMALL_STATE(3670)] = 222558,
  [SMALL_STATE(3671)] = 222591,
  [SMALL_STATE(3672)] = 222624,
  [SMALL_STATE(3673)] = 222657,
  [SMALL_STATE(3674)] = 222690,
  [SMALL_STATE(3675)] = 222723,
  [SMALL_STATE(3676)] = 222756,
  [SMALL_STATE(3677)] = 222789,
  [SMALL_STATE(3678)] = 222822,
  [SMALL_STATE(3679)] = 222855,
  [SMALL_STATE(3680)] = 222888,
  [SMALL_STATE(3681)] = 222921,
  [SMALL_STATE(3682)] = 222954,
  [SMALL_STATE(3683)] = 222987,
  [SMALL_STATE(3684)] = 223020,
  [SMALL_STATE(3685)] = 223053,
  [SMALL_STATE(3686)] = 223086,
  [SMALL_STATE(3687)] = 223119,
  [SMALL_STATE(3688)] = 223152,
  [SMALL_STATE(3689)] = 223185,
  [SMALL_STATE(3690)] = 223218,
  [SMALL_STATE(3691)] = 223251,
  [SMALL_STATE(3692)] = 223284,
  [SMALL_STATE(3693)] = 223317,
  [SMALL_STATE(3694)] = 223350,
  [SMALL_STATE(3695)] = 223383,
  [SMALL_STATE(3696)] = 223416,
  [SMALL_STATE(3697)] = 223449,
  [SMALL_STATE(3698)] = 223482,
  [SMALL_STATE(3699)] = 223515,
  [SMALL_STATE(3700)] = 223548,
  [SMALL_STATE(3701)] = 223581,
  [SMALL_STATE(3702)] = 223614,
  [SMALL_STATE(3703)] = 223647,
  [SMALL_STATE(3704)] = 223680,
  [SMALL_STATE(3705)] = 223719,
  [SMALL_STATE(3706)] = 223752,
  [SMALL_STATE(3707)] = 223785,
  [SMALL_STATE(3708)] = 223818,
  [SMALL_STATE(3709)] = 223851,
  [SMALL_STATE(3710)] = 223884,
  [SMALL_STATE(3711)] = 223917,
  [SMALL_STATE(3712)] = 223950,
  [SMALL_STATE(3713)] = 224003,
  [SMALL_STATE(3714)] = 224036,
  [SMALL_STATE(3715)] = 224069,
  [SMALL_STATE(3716)] = 224102,
  [SMALL_STATE(3717)] = 224135,
  [SMALL_STATE(3718)] = 224168,
  [SMALL_STATE(3719)] = 224201,
  [SMALL_STATE(3720)] = 224234,
  [SMALL_STATE(3721)] = 224267,
  [SMALL_STATE(3722)] = 224300,
  [SMALL_STATE(3723)] = 224333,
  [SMALL_STATE(3724)] = 224366,
  [SMALL_STATE(3725)] = 224399,
  [SMALL_STATE(3726)] = 224432,
  [SMALL_STATE(3727)] = 224465,
  [SMALL_STATE(3728)] = 224498,
  [SMALL_STATE(3729)] = 224531,
  [SMALL_STATE(3730)] = 224564,
  [SMALL_STATE(3731)] = 224597,
  [SMALL_STATE(3732)] = 224630,
  [SMALL_STATE(3733)] = 224663,
  [SMALL_STATE(3734)] = 224696,
  [SMALL_STATE(3735)] = 224729,
  [SMALL_STATE(3736)] = 224762,
  [SMALL_STATE(3737)] = 224795,
  [SMALL_STATE(3738)] = 224828,
  [SMALL_STATE(3739)] = 224861,
  [SMALL_STATE(3740)] = 224894,
  [SMALL_STATE(3741)] = 224927,
  [SMALL_STATE(3742)] = 224960,
  [SMALL_STATE(3743)] = 225013,
  [SMALL_STATE(3744)] = 225046,
  [SMALL_STATE(3745)] = 225099,
  [SMALL_STATE(3746)] = 225132,
  [SMALL_STATE(3747)] = 225165,
  [SMALL_STATE(3748)] = 225198,
  [SMALL_STATE(3749)] = 225231,
  [SMALL_STATE(3750)] = 225264,
  [SMALL_STATE(3751)] = 225297,
  [SMALL_STATE(3752)] = 225330,
  [SMALL_STATE(3753)] = 225363,
  [SMALL_STATE(3754)] = 225396,
  [SMALL_STATE(3755)] = 225429,
  [SMALL_STATE(3756)] = 225462,
  [SMALL_STATE(3757)] = 225509,
  [SMALL_STATE(3758)] = 225542,
  [SMALL_STATE(3759)] = 225575,
  [SMALL_STATE(3760)] = 225608,
  [SMALL_STATE(3761)] = 225641,
  [SMALL_STATE(3762)] = 225674,
  [SMALL_STATE(3763)] = 225707,
  [SMALL_STATE(3764)] = 225740,
  [SMALL_STATE(3765)] = 225773,
  [SMALL_STATE(3766)] = 225806,
  [SMALL_STATE(3767)] = 225857,
  [SMALL_STATE(3768)] = 225890,
  [SMALL_STATE(3769)] = 225923,
  [SMALL_STATE(3770)] = 225956,
  [SMALL_STATE(3771)] = 226003,
  [SMALL_STATE(3772)] = 226036,
  [SMALL_STATE(3773)] = 226083,
  [SMALL_STATE(3774)] = 226116,
  [SMALL_STATE(3775)] = 226149,
  [SMALL_STATE(3776)] = 226182,
  [SMALL_STATE(3777)] = 226229,
  [SMALL_STATE(3778)] = 226262,
  [SMALL_STATE(3779)] = 226295,
  [SMALL_STATE(3780)] = 226342,
  [SMALL_STATE(3781)] = 226375,
  [SMALL_STATE(3782)] = 226408,
  [SMALL_STATE(3783)] = 226441,
  [SMALL_STATE(3784)] = 226474,
  [SMALL_STATE(3785)] = 226507,
  [SMALL_STATE(3786)] = 226540,
  [SMALL_STATE(3787)] = 226573,
  [SMALL_STATE(3788)] = 226606,
  [SMALL_STATE(3789)] = 226639,
  [SMALL_STATE(3790)] = 226686,
  [SMALL_STATE(3791)] = 226719,
  [SMALL_STATE(3792)] = 226752,
  [SMALL_STATE(3793)] = 226785,
  [SMALL_STATE(3794)] = 226818,
  [SMALL_STATE(3795)] = 226851,
  [SMALL_STATE(3796)] = 226884,
  [SMALL_STATE(3797)] = 226917,
  [SMALL_STATE(3798)] = 226950,
  [SMALL_STATE(3799)] = 226983,
  [SMALL_STATE(3800)] = 227016,
  [SMALL_STATE(3801)] = 227049,
  [SMALL_STATE(3802)] = 227098,
  [SMALL_STATE(3803)] = 227131,
  [SMALL_STATE(3804)] = 227184,
  [SMALL_STATE(3805)] = 227217,
  [SMALL_STATE(3806)] = 227250,
  [SMALL_STATE(3807)] = 227283,
  [SMALL_STATE(3808)] = 227336,
  [SMALL_STATE(3809)] = 227369,
  [SMALL_STATE(3810)] = 227402,
  [SMALL_STATE(3811)] = 227451,
  [SMALL_STATE(3812)] = 227498,
  [SMALL_STATE(3813)] = 227531,
  [SMALL_STATE(3814)] = 227578,
  [SMALL_STATE(3815)] = 227611,
  [SMALL_STATE(3816)] = 227644,
  [SMALL_STATE(3817)] = 227677,
  [SMALL_STATE(3818)] = 227710,
  [SMALL_STATE(3819)] = 227757,
  [SMALL_STATE(3820)] = 227804,
  [SMALL_STATE(3821)] = 227851,
  [SMALL_STATE(3822)] = 227898,
  [SMALL_STATE(3823)] = 227945,
  [SMALL_STATE(3824)] = 227998,
  [SMALL_STATE(3825)] = 228051,
  [SMALL_STATE(3826)] = 228104,
  [SMALL_STATE(3827)] = 228137,
  [SMALL_STATE(3828)] = 228190,
  [SMALL_STATE(3829)] = 228243,
  [SMALL_STATE(3830)] = 228296,
  [SMALL_STATE(3831)] = 228349,
  [SMALL_STATE(3832)] = 228402,
  [SMALL_STATE(3833)] = 228455,
  [SMALL_STATE(3834)] = 228508,
  [SMALL_STATE(3835)] = 228561,
  [SMALL_STATE(3836)] = 228614,
  [SMALL_STATE(3837)] = 228667,
  [SMALL_STATE(3838)] = 228720,
  [SMALL_STATE(3839)] = 228773,
  [SMALL_STATE(3840)] = 228806,
  [SMALL_STATE(3841)] = 228859,
  [SMALL_STATE(3842)] = 228912,
  [SMALL_STATE(3843)] = 228965,
  [SMALL_STATE(3844)] = 228998,
  [SMALL_STATE(3845)] = 229051,
  [SMALL_STATE(3846)] = 229104,
  [SMALL_STATE(3847)] = 229157,
  [SMALL_STATE(3848)] = 229210,
  [SMALL_STATE(3849)] = 229263,
  [SMALL_STATE(3850)] = 229316,
  [SMALL_STATE(3851)] = 229369,
  [SMALL_STATE(3852)] = 229422,
  [SMALL_STATE(3853)] = 229475,
  [SMALL_STATE(3854)] = 229528,
  [SMALL_STATE(3855)] = 229581,
  [SMALL_STATE(3856)] = 229634,
  [SMALL_STATE(3857)] = 229687,
  [SMALL_STATE(3858)] = 229740,
  [SMALL_STATE(3859)] = 229793,
  [SMALL_STATE(3860)] = 229846,
  [SMALL_STATE(3861)] = 229899,
  [SMALL_STATE(3862)] = 229952,
  [SMALL_STATE(3863)] = 230005,
  [SMALL_STATE(3864)] = 230058,
  [SMALL_STATE(3865)] = 230111,
  [SMALL_STATE(3866)] = 230164,
  [SMALL_STATE(3867)] = 230217,
  [SMALL_STATE(3868)] = 230270,
  [SMALL_STATE(3869)] = 230323,
  [SMALL_STATE(3870)] = 230376,
  [SMALL_STATE(3871)] = 230429,
  [SMALL_STATE(3872)] = 230482,
  [SMALL_STATE(3873)] = 230535,
  [SMALL_STATE(3874)] = 230588,
  [SMALL_STATE(3875)] = 230641,
  [SMALL_STATE(3876)] = 230694,
  [SMALL_STATE(3877)] = 230747,
  [SMALL_STATE(3878)] = 230800,
  [SMALL_STATE(3879)] = 230853,
  [SMALL_STATE(3880)] = 230906,
  [SMALL_STATE(3881)] = 230959,
  [SMALL_STATE(3882)] = 231012,
  [SMALL_STATE(3883)] = 231065,
  [SMALL_STATE(3884)] = 231118,
  [SMALL_STATE(3885)] = 231171,
  [SMALL_STATE(3886)] = 231224,
  [SMALL_STATE(3887)] = 231277,
  [SMALL_STATE(3888)] = 231330,
  [SMALL_STATE(3889)] = 231383,
  [SMALL_STATE(3890)] = 231436,
  [SMALL_STATE(3891)] = 231489,
  [SMALL_STATE(3892)] = 231542,
  [SMALL_STATE(3893)] = 231595,
  [SMALL_STATE(3894)] = 231648,
  [SMALL_STATE(3895)] = 231701,
  [SMALL_STATE(3896)] = 231754,
  [SMALL_STATE(3897)] = 231807,
  [SMALL_STATE(3898)] = 231860,
  [SMALL_STATE(3899)] = 231913,
  [SMALL_STATE(3900)] = 231966,
  [SMALL_STATE(3901)] = 232019,
  [SMALL_STATE(3902)] = 232072,
  [SMALL_STATE(3903)] = 232125,
  [SMALL_STATE(3904)] = 232178,
  [SMALL_STATE(3905)] = 232231,
  [SMALL_STATE(3906)] = 232284,
  [SMALL_STATE(3907)] = 232337,
  [SMALL_STATE(3908)] = 232390,
  [SMALL_STATE(3909)] = 232443,
  [SMALL_STATE(3910)] = 232496,
  [SMALL_STATE(3911)] = 232549,
  [SMALL_STATE(3912)] = 232602,
  [SMALL_STATE(3913)] = 232655,
  [SMALL_STATE(3914)] = 232708,
  [SMALL_STATE(3915)] = 232761,
  [SMALL_STATE(3916)] = 232814,
  [SMALL_STATE(3917)] = 232867,
  [SMALL_STATE(3918)] = 232920,
  [SMALL_STATE(3919)] = 232973,
  [SMALL_STATE(3920)] = 233026,
  [SMALL_STATE(3921)] = 233079,
  [SMALL_STATE(3922)] = 233132,
  [SMALL_STATE(3923)] = 233185,
  [SMALL_STATE(3924)] = 233238,
  [SMALL_STATE(3925)] = 233291,
  [SMALL_STATE(3926)] = 233344,
  [SMALL_STATE(3927)] = 233397,
  [SMALL_STATE(3928)] = 233450,
  [SMALL_STATE(3929)] = 233503,
  [SMALL_STATE(3930)] = 233556,
  [SMALL_STATE(3931)] = 233609,
  [SMALL_STATE(3932)] = 233662,
  [SMALL_STATE(3933)] = 233715,
  [SMALL_STATE(3934)] = 233768,
  [SMALL_STATE(3935)] = 233821,
  [SMALL_STATE(3936)] = 233874,
  [SMALL_STATE(3937)] = 233927,
  [SMALL_STATE(3938)] = 233980,
  [SMALL_STATE(3939)] = 234033,
  [SMALL_STATE(3940)] = 234086,
  [SMALL_STATE(3941)] = 234139,
  [SMALL_STATE(3942)] = 234192,
  [SMALL_STATE(3943)] = 234245,
  [SMALL_STATE(3944)] = 234298,
  [SMALL_STATE(3945)] = 234351,
  [SMALL_STATE(3946)] = 234404,
  [SMALL_STATE(3947)] = 234457,
  [SMALL_STATE(3948)] = 234510,
  [SMALL_STATE(3949)] = 234563,
  [SMALL_STATE(3950)] = 234616,
  [SMALL_STATE(3951)] = 234669,
  [SMALL_STATE(3952)] = 234722,
  [SMALL_STATE(3953)] = 234775,
  [SMALL_STATE(3954)] = 234828,
  [SMALL_STATE(3955)] = 234881,
  [SMALL_STATE(3956)] = 234934,
  [SMALL_STATE(3957)] = 234987,
  [SMALL_STATE(3958)] = 235040,
  [SMALL_STATE(3959)] = 235093,
  [SMALL_STATE(3960)] = 235146,
  [SMALL_STATE(3961)] = 235199,
  [SMALL_STATE(3962)] = 235252,
  [SMALL_STATE(3963)] = 235305,
  [SMALL_STATE(3964)] = 235358,
  [SMALL_STATE(3965)] = 235411,
  [SMALL_STATE(3966)] = 235464,
  [SMALL_STATE(3967)] = 235517,
  [SMALL_STATE(3968)] = 235570,
  [SMALL_STATE(3969)] = 235623,
  [SMALL_STATE(3970)] = 235676,
  [SMALL_STATE(3971)] = 235729,
  [SMALL_STATE(3972)] = 235782,
  [SMALL_STATE(3973)] = 235835,
  [SMALL_STATE(3974)] = 235888,
  [SMALL_STATE(3975)] = 235941,
  [SMALL_STATE(3976)] = 235994,
  [SMALL_STATE(3977)] = 236027,
  [SMALL_STATE(3978)] = 236066,
  [SMALL_STATE(3979)] = 236110,
  [SMALL_STATE(3980)] = 236142,
  [SMALL_STATE(3981)] = 236186,
  [SMALL_STATE(3982)] = 236218,
  [SMALL_STATE(3983)] = 236262,
  [SMALL_STATE(3984)] = 236306,
  [SMALL_STATE(3985)] = 236350,
  [SMALL_STATE(3986)] = 236394,
  [SMALL_STATE(3987)] = 236438,
  [SMALL_STATE(3988)] = 236482,
  [SMALL_STATE(3989)] = 236514,
  [SMALL_STATE(3990)] = 236558,
  [SMALL_STATE(3991)] = 236602,
  [SMALL_STATE(3992)] = 236646,
  [SMALL_STATE(3993)] = 236678,
  [SMALL_STATE(3994)] = 236722,
  [SMALL_STATE(3995)] = 236766,
  [SMALL_STATE(3996)] = 236810,
  [SMALL_STATE(3997)] = 236854,
  [SMALL_STATE(3998)] = 236898,
  [SMALL_STATE(3999)] = 236942,
  [SMALL_STATE(4000)] = 236974,
  [SMALL_STATE(4001)] = 237018,
  [SMALL_STATE(4002)] = 237062,
  [SMALL_STATE(4003)] = 237106,
  [SMALL_STATE(4004)] = 237150,
  [SMALL_STATE(4005)] = 237194,
  [SMALL_STATE(4006)] = 237238,
  [SMALL_STATE(4007)] = 237282,
  [SMALL_STATE(4008)] = 237326,
  [SMALL_STATE(4009)] = 237370,
  [SMALL_STATE(4010)] = 237414,
  [SMALL_STATE(4011)] = 237458,
  [SMALL_STATE(4012)] = 237490,
  [SMALL_STATE(4013)] = 237534,
  [SMALL_STATE(4014)] = 237578,
  [SMALL_STATE(4015)] = 237610,
  [SMALL_STATE(4016)] = 237654,
  [SMALL_STATE(4017)] = 237698,
  [SMALL_STATE(4018)] = 237742,
  [SMALL_STATE(4019)] = 237786,
  [SMALL_STATE(4020)] = 237830,
  [SMALL_STATE(4021)] = 237874,
  [SMALL_STATE(4022)] = 237918,
  [SMALL_STATE(4023)] = 237962,
  [SMALL_STATE(4024)] = 238006,
  [SMALL_STATE(4025)] = 238050,
  [SMALL_STATE(4026)] = 238082,
  [SMALL_STATE(4027)] = 238114,
  [SMALL_STATE(4028)] = 238158,
  [SMALL_STATE(4029)] = 238202,
  [SMALL_STATE(4030)] = 238246,
  [SMALL_STATE(4031)] = 238290,
  [SMALL_STATE(4032)] = 238334,
  [SMALL_STATE(4033)] = 238378,
  [SMALL_STATE(4034)] = 238422,
  [SMALL_STATE(4035)] = 238466,
  [SMALL_STATE(4036)] = 238510,
  [SMALL_STATE(4037)] = 238554,
  [SMALL_STATE(4038)] = 238598,
  [SMALL_STATE(4039)] = 238642,
  [SMALL_STATE(4040)] = 238686,
  [SMALL_STATE(4041)] = 238730,
  [SMALL_STATE(4042)] = 238774,
  [SMALL_STATE(4043)] = 238818,
  [SMALL_STATE(4044)] = 238862,
  [SMALL_STATE(4045)] = 238906,
  [SMALL_STATE(4046)] = 238950,
  [SMALL_STATE(4047)] = 238994,
  [SMALL_STATE(4048)] = 239038,
  [SMALL_STATE(4049)] = 239082,
  [SMALL_STATE(4050)] = 239126,
  [SMALL_STATE(4051)] = 239170,
  [SMALL_STATE(4052)] = 239214,
  [SMALL_STATE(4053)] = 239258,
  [SMALL_STATE(4054)] = 239302,
  [SMALL_STATE(4055)] = 239346,
  [SMALL_STATE(4056)] = 239390,
  [SMALL_STATE(4057)] = 239434,
  [SMALL_STATE(4058)] = 239478,
  [SMALL_STATE(4059)] = 239522,
  [SMALL_STATE(4060)] = 239566,
  [SMALL_STATE(4061)] = 239610,
  [SMALL_STATE(4062)] = 239654,
  [SMALL_STATE(4063)] = 239698,
  [SMALL_STATE(4064)] = 239742,
  [SMALL_STATE(4065)] = 239786,
  [SMALL_STATE(4066)] = 239830,
  [SMALL_STATE(4067)] = 239874,
  [SMALL_STATE(4068)] = 239918,
  [SMALL_STATE(4069)] = 239962,
  [SMALL_STATE(4070)] = 240006,
  [SMALL_STATE(4071)] = 240050,
  [SMALL_STATE(4072)] = 240094,
  [SMALL_STATE(4073)] = 240126,
  [SMALL_STATE(4074)] = 240170,
  [SMALL_STATE(4075)] = 240214,
  [SMALL_STATE(4076)] = 240258,
  [SMALL_STATE(4077)] = 240302,
  [SMALL_STATE(4078)] = 240346,
  [SMALL_STATE(4079)] = 240390,
  [SMALL_STATE(4080)] = 240422,
  [SMALL_STATE(4081)] = 240466,
  [SMALL_STATE(4082)] = 240510,
  [SMALL_STATE(4083)] = 240554,
  [SMALL_STATE(4084)] = 240598,
  [SMALL_STATE(4085)] = 240642,
  [SMALL_STATE(4086)] = 240686,
  [SMALL_STATE(4087)] = 240730,
  [SMALL_STATE(4088)] = 240762,
  [SMALL_STATE(4089)] = 240806,
  [SMALL_STATE(4090)] = 240838,
  [SMALL_STATE(4091)] = 240870,
  [SMALL_STATE(4092)] = 240902,
  [SMALL_STATE(4093)] = 240946,
  [SMALL_STATE(4094)] = 240990,
  [SMALL_STATE(4095)] = 241034,
  [SMALL_STATE(4096)] = 241078,
  [SMALL_STATE(4097)] = 241122,
  [SMALL_STATE(4098)] = 241166,
  [SMALL_STATE(4099)] = 241210,
  [SMALL_STATE(4100)] = 241254,
  [SMALL_STATE(4101)] = 241286,
  [SMALL_STATE(4102)] = 241330,
  [SMALL_STATE(4103)] = 241374,
  [SMALL_STATE(4104)] = 241418,
  [SMALL_STATE(4105)] = 241462,
  [SMALL_STATE(4106)] = 241506,
  [SMALL_STATE(4107)] = 241550,
  [SMALL_STATE(4108)] = 241594,
  [SMALL_STATE(4109)] = 241638,
  [SMALL_STATE(4110)] = 241682,
  [SMALL_STATE(4111)] = 241726,
  [SMALL_STATE(4112)] = 241770,
  [SMALL_STATE(4113)] = 241814,
  [SMALL_STATE(4114)] = 241858,
  [SMALL_STATE(4115)] = 241902,
  [SMALL_STATE(4116)] = 241946,
  [SMALL_STATE(4117)] = 241990,
  [SMALL_STATE(4118)] = 242034,
  [SMALL_STATE(4119)] = 242078,
  [SMALL_STATE(4120)] = 242122,
  [SMALL_STATE(4121)] = 242166,
  [SMALL_STATE(4122)] = 242212,
  [SMALL_STATE(4123)] = 242258,
  [SMALL_STATE(4124)] = 242302,
  [SMALL_STATE(4125)] = 242346,
  [SMALL_STATE(4126)] = 242390,
  [SMALL_STATE(4127)] = 242434,
  [SMALL_STATE(4128)] = 242466,
  [SMALL_STATE(4129)] = 242510,
  [SMALL_STATE(4130)] = 242542,
  [SMALL_STATE(4131)] = 242574,
  [SMALL_STATE(4132)] = 242618,
  [SMALL_STATE(4133)] = 242650,
  [SMALL_STATE(4134)] = 242694,
  [SMALL_STATE(4135)] = 242738,
  [SMALL_STATE(4136)] = 242782,
  [SMALL_STATE(4137)] = 242826,
  [SMALL_STATE(4138)] = 242870,
  [SMALL_STATE(4139)] = 242914,
  [SMALL_STATE(4140)] = 242958,
  [SMALL_STATE(4141)] = 242990,
  [SMALL_STATE(4142)] = 243034,
  [SMALL_STATE(4143)] = 243078,
  [SMALL_STATE(4144)] = 243122,
  [SMALL_STATE(4145)] = 243166,
  [SMALL_STATE(4146)] = 243210,
  [SMALL_STATE(4147)] = 243242,
  [SMALL_STATE(4148)] = 243286,
  [SMALL_STATE(4149)] = 243330,
  [SMALL_STATE(4150)] = 243374,
  [SMALL_STATE(4151)] = 243406,
  [SMALL_STATE(4152)] = 243438,
  [SMALL_STATE(4153)] = 243482,
  [SMALL_STATE(4154)] = 243526,
  [SMALL_STATE(4155)] = 243570,
  [SMALL_STATE(4156)] = 243614,
  [SMALL_STATE(4157)] = 243646,
  [SMALL_STATE(4158)] = 243678,
  [SMALL_STATE(4159)] = 243726,
  [SMALL_STATE(4160)] = 243770,
  [SMALL_STATE(4161)] = 243816,
  [SMALL_STATE(4162)] = 243860,
  [SMALL_STATE(4163)] = 243904,
  [SMALL_STATE(4164)] = 243948,
  [SMALL_STATE(4165)] = 243992,
  [SMALL_STATE(4166)] = 244024,
  [SMALL_STATE(4167)] = 244068,
  [SMALL_STATE(4168)] = 244112,
  [SMALL_STATE(4169)] = 244156,
  [SMALL_STATE(4170)] = 244188,
  [SMALL_STATE(4171)] = 244220,
  [SMALL_STATE(4172)] = 244270,
  [SMALL_STATE(4173)] = 244314,
  [SMALL_STATE(4174)] = 244358,
  [SMALL_STATE(4175)] = 244402,
  [SMALL_STATE(4176)] = 244446,
  [SMALL_STATE(4177)] = 244490,
  [SMALL_STATE(4178)] = 244534,
  [SMALL_STATE(4179)] = 244578,
  [SMALL_STATE(4180)] = 244622,
  [SMALL_STATE(4181)] = 244666,
  [SMALL_STATE(4182)] = 244710,
  [SMALL_STATE(4183)] = 244754,
  [SMALL_STATE(4184)] = 244798,
  [SMALL_STATE(4185)] = 244842,
  [SMALL_STATE(4186)] = 244886,
  [SMALL_STATE(4187)] = 244930,
  [SMALL_STATE(4188)] = 244974,
  [SMALL_STATE(4189)] = 245018,
  [SMALL_STATE(4190)] = 245064,
  [SMALL_STATE(4191)] = 245108,
  [SMALL_STATE(4192)] = 245152,
  [SMALL_STATE(4193)] = 245198,
  [SMALL_STATE(4194)] = 245242,
  [SMALL_STATE(4195)] = 245274,
  [SMALL_STATE(4196)] = 245326,
  [SMALL_STATE(4197)] = 245370,
  [SMALL_STATE(4198)] = 245402,
  [SMALL_STATE(4199)] = 245446,
  [SMALL_STATE(4200)] = 245490,
  [SMALL_STATE(4201)] = 245534,
  [SMALL_STATE(4202)] = 245578,
  [SMALL_STATE(4203)] = 245610,
  [SMALL_STATE(4204)] = 245654,
  [SMALL_STATE(4205)] = 245698,
  [SMALL_STATE(4206)] = 245742,
  [SMALL_STATE(4207)] = 245786,
  [SMALL_STATE(4208)] = 245830,
  [SMALL_STATE(4209)] = 245874,
  [SMALL_STATE(4210)] = 245918,
  [SMALL_STATE(4211)] = 245964,
  [SMALL_STATE(4212)] = 246008,
  [SMALL_STATE(4213)] = 246052,
  [SMALL_STATE(4214)] = 246096,
  [SMALL_STATE(4215)] = 246128,
  [SMALL_STATE(4216)] = 246172,
  [SMALL_STATE(4217)] = 246216,
  [SMALL_STATE(4218)] = 246260,
  [SMALL_STATE(4219)] = 246304,
  [SMALL_STATE(4220)] = 246336,
  [SMALL_STATE(4221)] = 246380,
  [SMALL_STATE(4222)] = 246424,
  [SMALL_STATE(4223)] = 246468,
  [SMALL_STATE(4224)] = 246512,
  [SMALL_STATE(4225)] = 246556,
  [SMALL_STATE(4226)] = 246602,
  [SMALL_STATE(4227)] = 246646,
  [SMALL_STATE(4228)] = 246698,
  [SMALL_STATE(4229)] = 246730,
  [SMALL_STATE(4230)] = 246762,
  [SMALL_STATE(4231)] = 246806,
  [SMALL_STATE(4232)] = 246850,
  [SMALL_STATE(4233)] = 246894,
  [SMALL_STATE(4234)] = 246938,
  [SMALL_STATE(4235)] = 246982,
  [SMALL_STATE(4236)] = 247026,
  [SMALL_STATE(4237)] = 247070,
  [SMALL_STATE(4238)] = 247116,
  [SMALL_STATE(4239)] = 247160,
  [SMALL_STATE(4240)] = 247204,
  [SMALL_STATE(4241)] = 247248,
  [SMALL_STATE(4242)] = 247292,
  [SMALL_STATE(4243)] = 247336,
  [SMALL_STATE(4244)] = 247368,
  [SMALL_STATE(4245)] = 247412,
  [SMALL_STATE(4246)] = 247456,
  [SMALL_STATE(4247)] = 247500,
  [SMALL_STATE(4248)] = 247544,
  [SMALL_STATE(4249)] = 247588,
  [SMALL_STATE(4250)] = 247632,
  [SMALL_STATE(4251)] = 247676,
  [SMALL_STATE(4252)] = 247726,
  [SMALL_STATE(4253)] = 247764,
  [SMALL_STATE(4254)] = 247808,
  [SMALL_STATE(4255)] = 247852,
  [SMALL_STATE(4256)] = 247896,
  [SMALL_STATE(4257)] = 247940,
  [SMALL_STATE(4258)] = 247984,
  [SMALL_STATE(4259)] = 248028,
  [SMALL_STATE(4260)] = 248072,
  [SMALL_STATE(4261)] = 248116,
  [SMALL_STATE(4262)] = 248160,
  [SMALL_STATE(4263)] = 248204,
  [SMALL_STATE(4264)] = 248250,
  [SMALL_STATE(4265)] = 248294,
  [SMALL_STATE(4266)] = 248338,
  [SMALL_STATE(4267)] = 248382,
  [SMALL_STATE(4268)] = 248414,
  [SMALL_STATE(4269)] = 248458,
  [SMALL_STATE(4270)] = 248502,
  [SMALL_STATE(4271)] = 248546,
  [SMALL_STATE(4272)] = 248590,
  [SMALL_STATE(4273)] = 248630,
  [SMALL_STATE(4274)] = 248674,
  [SMALL_STATE(4275)] = 248720,
  [SMALL_STATE(4276)] = 248764,
  [SMALL_STATE(4277)] = 248808,
  [SMALL_STATE(4278)] = 248854,
  [SMALL_STATE(4279)] = 248898,
  [SMALL_STATE(4280)] = 248942,
  [SMALL_STATE(4281)] = 248986,
  [SMALL_STATE(4282)] = 249030,
  [SMALL_STATE(4283)] = 249080,
  [SMALL_STATE(4284)] = 249124,
  [SMALL_STATE(4285)] = 249168,
  [SMALL_STATE(4286)] = 249212,
  [SMALL_STATE(4287)] = 249256,
  [SMALL_STATE(4288)] = 249300,
  [SMALL_STATE(4289)] = 249344,
  [SMALL_STATE(4290)] = 249388,
  [SMALL_STATE(4291)] = 249432,
  [SMALL_STATE(4292)] = 249464,
  [SMALL_STATE(4293)] = 249508,
  [SMALL_STATE(4294)] = 249552,
  [SMALL_STATE(4295)] = 249596,
  [SMALL_STATE(4296)] = 249640,
  [SMALL_STATE(4297)] = 249684,
  [SMALL_STATE(4298)] = 249716,
  [SMALL_STATE(4299)] = 249760,
  [SMALL_STATE(4300)] = 249804,
  [SMALL_STATE(4301)] = 249848,
  [SMALL_STATE(4302)] = 249892,
  [SMALL_STATE(4303)] = 249936,
  [SMALL_STATE(4304)] = 249980,
  [SMALL_STATE(4305)] = 250024,
  [SMALL_STATE(4306)] = 250070,
  [SMALL_STATE(4307)] = 250114,
  [SMALL_STATE(4308)] = 250158,
  [SMALL_STATE(4309)] = 250202,
  [SMALL_STATE(4310)] = 250246,
  [SMALL_STATE(4311)] = 250290,
  [SMALL_STATE(4312)] = 250334,
  [SMALL_STATE(4313)] = 250378,
  [SMALL_STATE(4314)] = 250422,
  [SMALL_STATE(4315)] = 250470,
  [SMALL_STATE(4316)] = 250503,
  [SMALL_STATE(4317)] = 250534,
  [SMALL_STATE(4318)] = 250565,
  [SMALL_STATE(4319)] = 250598,
  [SMALL_STATE(4320)] = 250629,
  [SMALL_STATE(4321)] = 250662,
  [SMALL_STATE(4322)] = 250693,
  [SMALL_STATE(4323)] = 250724,
  [SMALL_STATE(4324)] = 250755,
  [SMALL_STATE(4325)] = 250786,
  [SMALL_STATE(4326)] = 250817,
  [SMALL_STATE(4327)] = 250848,
  [SMALL_STATE(4328)] = 250879,
  [SMALL_STATE(4329)] = 250910,
  [SMALL_STATE(4330)] = 250943,
  [SMALL_STATE(4331)] = 250974,
  [SMALL_STATE(4332)] = 251007,
  [SMALL_STATE(4333)] = 251038,
  [SMALL_STATE(4334)] = 251069,
  [SMALL_STATE(4335)] = 251100,
  [SMALL_STATE(4336)] = 251131,
  [SMALL_STATE(4337)] = 251180,
  [SMALL_STATE(4338)] = 251213,
  [SMALL_STATE(4339)] = 251246,
  [SMALL_STATE(4340)] = 251279,
  [SMALL_STATE(4341)] = 251310,
  [SMALL_STATE(4342)] = 251343,
  [SMALL_STATE(4343)] = 251376,
  [SMALL_STATE(4344)] = 251409,
  [SMALL_STATE(4345)] = 251446,
  [SMALL_STATE(4346)] = 251495,
  [SMALL_STATE(4347)] = 251528,
  [SMALL_STATE(4348)] = 251559,
  [SMALL_STATE(4349)] = 251590,
  [SMALL_STATE(4350)] = 251621,
  [SMALL_STATE(4351)] = 251652,
  [SMALL_STATE(4352)] = 251685,
  [SMALL_STATE(4353)] = 251718,
  [SMALL_STATE(4354)] = 251749,
  [SMALL_STATE(4355)] = 251780,
  [SMALL_STATE(4356)] = 251813,
  [SMALL_STATE(4357)] = 251846,
  [SMALL_STATE(4358)] = 251879,
  [SMALL_STATE(4359)] = 251910,
  [SMALL_STATE(4360)] = 251941,
  [SMALL_STATE(4361)] = 251972,
  [SMALL_STATE(4362)] = 252003,
  [SMALL_STATE(4363)] = 252036,
  [SMALL_STATE(4364)] = 252069,
  [SMALL_STATE(4365)] = 252100,
  [SMALL_STATE(4366)] = 252137,
  [SMALL_STATE(4367)] = 252170,
  [SMALL_STATE(4368)] = 252201,
  [SMALL_STATE(4369)] = 252234,
  [SMALL_STATE(4370)] = 252265,
  [SMALL_STATE(4371)] = 252296,
  [SMALL_STATE(4372)] = 252327,
  [SMALL_STATE(4373)] = 252360,
  [SMALL_STATE(4374)] = 252391,
  [SMALL_STATE(4375)] = 252424,
  [SMALL_STATE(4376)] = 252457,
  [SMALL_STATE(4377)] = 252490,
  [SMALL_STATE(4378)] = 252523,
  [SMALL_STATE(4379)] = 252556,
  [SMALL_STATE(4380)] = 252589,
  [SMALL_STATE(4381)] = 252622,
  [SMALL_STATE(4382)] = 252653,
  [SMALL_STATE(4383)] = 252684,
  [SMALL_STATE(4384)] = 252715,
  [SMALL_STATE(4385)] = 252746,
  [SMALL_STATE(4386)] = 252777,
  [SMALL_STATE(4387)] = 252808,
  [SMALL_STATE(4388)] = 252852,
  [SMALL_STATE(4389)] = 252906,
  [SMALL_STATE(4390)] = 252942,
  [SMALL_STATE(4391)] = 252974,
  [SMALL_STATE(4392)] = 253010,
  [SMALL_STATE(4393)] = 253054,
  [SMALL_STATE(4394)] = 253098,
  [SMALL_STATE(4395)] = 253142,
  [SMALL_STATE(4396)] = 253186,
  [SMALL_STATE(4397)] = 253230,
  [SMALL_STATE(4398)] = 253274,
  [SMALL_STATE(4399)] = 253318,
  [SMALL_STATE(4400)] = 253354,
  [SMALL_STATE(4401)] = 253383,
  [SMALL_STATE(4402)] = 253412,
  [SMALL_STATE(4403)] = 253441,
  [SMALL_STATE(4404)] = 253470,
  [SMALL_STATE(4405)] = 253499,
  [SMALL_STATE(4406)] = 253528,
  [SMALL_STATE(4407)] = 253557,
  [SMALL_STATE(4408)] = 253586,
  [SMALL_STATE(4409)] = 253615,
  [SMALL_STATE(4410)] = 253644,
  [SMALL_STATE(4411)] = 253673,
  [SMALL_STATE(4412)] = 253702,
  [SMALL_STATE(4413)] = 253731,
  [SMALL_STATE(4414)] = 253760,
  [SMALL_STATE(4415)] = 253789,
  [SMALL_STATE(4416)] = 253818,
  [SMALL_STATE(4417)] = 253847,
  [SMALL_STATE(4418)] = 253876,
  [SMALL_STATE(4419)] = 253905,
  [SMALL_STATE(4420)] = 253934,
  [SMALL_STATE(4421)] = 253963,
  [SMALL_STATE(4422)] = 253992,
  [SMALL_STATE(4423)] = 254021,
  [SMALL_STATE(4424)] = 254050,
  [SMALL_STATE(4425)] = 254079,
  [SMALL_STATE(4426)] = 254108,
  [SMALL_STATE(4427)] = 254137,
  [SMALL_STATE(4428)] = 254166,
  [SMALL_STATE(4429)] = 254195,
  [SMALL_STATE(4430)] = 254224,
  [SMALL_STATE(4431)] = 254253,
  [SMALL_STATE(4432)] = 254288,
  [SMALL_STATE(4433)] = 254323,
  [SMALL_STATE(4434)] = 254352,
  [SMALL_STATE(4435)] = 254381,
  [SMALL_STATE(4436)] = 254410,
  [SMALL_STATE(4437)] = 254439,
  [SMALL_STATE(4438)] = 254468,
  [SMALL_STATE(4439)] = 254497,
  [SMALL_STATE(4440)] = 254526,
  [SMALL_STATE(4441)] = 254555,
  [SMALL_STATE(4442)] = 254590,
  [SMALL_STATE(4443)] = 254619,
  [SMALL_STATE(4444)] = 254648,
  [SMALL_STATE(4445)] = 254677,
  [SMALL_STATE(4446)] = 254705,
  [SMALL_STATE(4447)] = 254733,
  [SMALL_STATE(4448)] = 254761,
  [SMALL_STATE(4449)] = 254803,
  [SMALL_STATE(4450)] = 254831,
  [SMALL_STATE(4451)] = 254859,
  [SMALL_STATE(4452)] = 254887,
  [SMALL_STATE(4453)] = 254915,
  [SMALL_STATE(4454)] = 254943,
  [SMALL_STATE(4455)] = 254971,
  [SMALL_STATE(4456)] = 254999,
  [SMALL_STATE(4457)] = 255027,
  [SMALL_STATE(4458)] = 255055,
  [SMALL_STATE(4459)] = 255083,
  [SMALL_STATE(4460)] = 255111,
  [SMALL_STATE(4461)] = 255139,
  [SMALL_STATE(4462)] = 255173,
  [SMALL_STATE(4463)] = 255203,
  [SMALL_STATE(4464)] = 255231,
  [SMALL_STATE(4465)] = 255259,
  [SMALL_STATE(4466)] = 255287,
  [SMALL_STATE(4467)] = 255319,
  [SMALL_STATE(4468)] = 255347,
  [SMALL_STATE(4469)] = 255375,
  [SMALL_STATE(4470)] = 255403,
  [SMALL_STATE(4471)] = 255431,
  [SMALL_STATE(4472)] = 255459,
  [SMALL_STATE(4473)] = 255487,
  [SMALL_STATE(4474)] = 255515,
  [SMALL_STATE(4475)] = 255543,
  [SMALL_STATE(4476)] = 255571,
  [SMALL_STATE(4477)] = 255599,
  [SMALL_STATE(4478)] = 255627,
  [SMALL_STATE(4479)] = 255655,
  [SMALL_STATE(4480)] = 255683,
  [SMALL_STATE(4481)] = 255711,
  [SMALL_STATE(4482)] = 255745,
  [SMALL_STATE(4483)] = 255773,
  [SMALL_STATE(4484)] = 255801,
  [SMALL_STATE(4485)] = 255829,
  [SMALL_STATE(4486)] = 255857,
  [SMALL_STATE(4487)] = 255899,
  [SMALL_STATE(4488)] = 255941,
  [SMALL_STATE(4489)] = 255969,
  [SMALL_STATE(4490)] = 255997,
  [SMALL_STATE(4491)] = 256025,
  [SMALL_STATE(4492)] = 256059,
  [SMALL_STATE(4493)] = 256087,
  [SMALL_STATE(4494)] = 256115,
  [SMALL_STATE(4495)] = 256143,
  [SMALL_STATE(4496)] = 256185,
  [SMALL_STATE(4497)] = 256213,
  [SMALL_STATE(4498)] = 256241,
  [SMALL_STATE(4499)] = 256269,
  [SMALL_STATE(4500)] = 256297,
  [SMALL_STATE(4501)] = 256325,
  [SMALL_STATE(4502)] = 256369,
  [SMALL_STATE(4503)] = 256397,
  [SMALL_STATE(4504)] = 256425,
  [SMALL_STATE(4505)] = 256453,
  [SMALL_STATE(4506)] = 256489,
  [SMALL_STATE(4507)] = 256523,
  [SMALL_STATE(4508)] = 256551,
  [SMALL_STATE(4509)] = 256579,
  [SMALL_STATE(4510)] = 256607,
  [SMALL_STATE(4511)] = 256635,
  [SMALL_STATE(4512)] = 256663,
  [SMALL_STATE(4513)] = 256691,
  [SMALL_STATE(4514)] = 256719,
  [SMALL_STATE(4515)] = 256763,
  [SMALL_STATE(4516)] = 256791,
  [SMALL_STATE(4517)] = 256819,
  [SMALL_STATE(4518)] = 256847,
  [SMALL_STATE(4519)] = 256875,
  [SMALL_STATE(4520)] = 256903,
  [SMALL_STATE(4521)] = 256931,
  [SMALL_STATE(4522)] = 256959,
  [SMALL_STATE(4523)] = 256987,
  [SMALL_STATE(4524)] = 257015,
  [SMALL_STATE(4525)] = 257043,
  [SMALL_STATE(4526)] = 257085,
  [SMALL_STATE(4527)] = 257113,
  [SMALL_STATE(4528)] = 257141,
  [SMALL_STATE(4529)] = 257169,
  [SMALL_STATE(4530)] = 257197,
  [SMALL_STATE(4531)] = 257225,
  [SMALL_STATE(4532)] = 257253,
  [SMALL_STATE(4533)] = 257281,
  [SMALL_STATE(4534)] = 257309,
  [SMALL_STATE(4535)] = 257337,
  [SMALL_STATE(4536)] = 257365,
  [SMALL_STATE(4537)] = 257393,
  [SMALL_STATE(4538)] = 257421,
  [SMALL_STATE(4539)] = 257449,
  [SMALL_STATE(4540)] = 257477,
  [SMALL_STATE(4541)] = 257505,
  [SMALL_STATE(4542)] = 257533,
  [SMALL_STATE(4543)] = 257561,
  [SMALL_STATE(4544)] = 257603,
  [SMALL_STATE(4545)] = 257631,
  [SMALL_STATE(4546)] = 257659,
  [SMALL_STATE(4547)] = 257687,
  [SMALL_STATE(4548)] = 257715,
  [SMALL_STATE(4549)] = 257742,
  [SMALL_STATE(4550)] = 257781,
  [SMALL_STATE(4551)] = 257808,
  [SMALL_STATE(4552)] = 257835,
  [SMALL_STATE(4553)] = 257864,
  [SMALL_STATE(4554)] = 257891,
  [SMALL_STATE(4555)] = 257918,
  [SMALL_STATE(4556)] = 257945,
  [SMALL_STATE(4557)] = 257972,
  [SMALL_STATE(4558)] = 257999,
  [SMALL_STATE(4559)] = 258026,
  [SMALL_STATE(4560)] = 258053,
  [SMALL_STATE(4561)] = 258080,
  [SMALL_STATE(4562)] = 258113,
  [SMALL_STATE(4563)] = 258140,
  [SMALL_STATE(4564)] = 258167,
  [SMALL_STATE(4565)] = 258194,
  [SMALL_STATE(4566)] = 258221,
  [SMALL_STATE(4567)] = 258248,
  [SMALL_STATE(4568)] = 258275,
  [SMALL_STATE(4569)] = 258302,
  [SMALL_STATE(4570)] = 258329,
  [SMALL_STATE(4571)] = 258356,
  [SMALL_STATE(4572)] = 258383,
  [SMALL_STATE(4573)] = 258410,
  [SMALL_STATE(4574)] = 258437,
  [SMALL_STATE(4575)] = 258464,
  [SMALL_STATE(4576)] = 258491,
  [SMALL_STATE(4577)] = 258530,
  [SMALL_STATE(4578)] = 258557,
  [SMALL_STATE(4579)] = 258584,
  [SMALL_STATE(4580)] = 258611,
  [SMALL_STATE(4581)] = 258638,
  [SMALL_STATE(4582)] = 258665,
  [SMALL_STATE(4583)] = 258692,
  [SMALL_STATE(4584)] = 258719,
  [SMALL_STATE(4585)] = 258746,
  [SMALL_STATE(4586)] = 258773,
  [SMALL_STATE(4587)] = 258812,
  [SMALL_STATE(4588)] = 258845,
  [SMALL_STATE(4589)] = 258872,
  [SMALL_STATE(4590)] = 258899,
  [SMALL_STATE(4591)] = 258926,
  [SMALL_STATE(4592)] = 258953,
  [SMALL_STATE(4593)] = 258980,
  [SMALL_STATE(4594)] = 259007,
  [SMALL_STATE(4595)] = 259034,
  [SMALL_STATE(4596)] = 259061,
  [SMALL_STATE(4597)] = 259088,
  [SMALL_STATE(4598)] = 259131,
  [SMALL_STATE(4599)] = 259174,
  [SMALL_STATE(4600)] = 259201,
  [SMALL_STATE(4601)] = 259228,
  [SMALL_STATE(4602)] = 259255,
  [SMALL_STATE(4603)] = 259282,
  [SMALL_STATE(4604)] = 259309,
  [SMALL_STATE(4605)] = 259336,
  [SMALL_STATE(4606)] = 259363,
  [SMALL_STATE(4607)] = 259390,
  [SMALL_STATE(4608)] = 259417,
  [SMALL_STATE(4609)] = 259444,
  [SMALL_STATE(4610)] = 259471,
  [SMALL_STATE(4611)] = 259498,
  [SMALL_STATE(4612)] = 259525,
  [SMALL_STATE(4613)] = 259552,
  [SMALL_STATE(4614)] = 259579,
  [SMALL_STATE(4615)] = 259606,
  [SMALL_STATE(4616)] = 259633,
  [SMALL_STATE(4617)] = 259660,
  [SMALL_STATE(4618)] = 259687,
  [SMALL_STATE(4619)] = 259720,
  [SMALL_STATE(4620)] = 259747,
  [SMALL_STATE(4621)] = 259774,
  [SMALL_STATE(4622)] = 259801,
  [SMALL_STATE(4623)] = 259828,
  [SMALL_STATE(4624)] = 259855,
  [SMALL_STATE(4625)] = 259882,
  [SMALL_STATE(4626)] = 259909,
  [SMALL_STATE(4627)] = 259936,
  [SMALL_STATE(4628)] = 259963,
  [SMALL_STATE(4629)] = 259990,
  [SMALL_STATE(4630)] = 260017,
  [SMALL_STATE(4631)] = 260044,
  [SMALL_STATE(4632)] = 260071,
  [SMALL_STATE(4633)] = 260110,
  [SMALL_STATE(4634)] = 260137,
  [SMALL_STATE(4635)] = 260180,
  [SMALL_STATE(4636)] = 260207,
  [SMALL_STATE(4637)] = 260234,
  [SMALL_STATE(4638)] = 260261,
  [SMALL_STATE(4639)] = 260288,
  [SMALL_STATE(4640)] = 260321,
  [SMALL_STATE(4641)] = 260348,
  [SMALL_STATE(4642)] = 260375,
  [SMALL_STATE(4643)] = 260418,
  [SMALL_STATE(4644)] = 260445,
  [SMALL_STATE(4645)] = 260472,
  [SMALL_STATE(4646)] = 260511,
  [SMALL_STATE(4647)] = 260544,
  [SMALL_STATE(4648)] = 260571,
  [SMALL_STATE(4649)] = 260598,
  [SMALL_STATE(4650)] = 260637,
  [SMALL_STATE(4651)] = 260664,
  [SMALL_STATE(4652)] = 260701,
  [SMALL_STATE(4653)] = 260728,
  [SMALL_STATE(4654)] = 260755,
  [SMALL_STATE(4655)] = 260782,
  [SMALL_STATE(4656)] = 260815,
  [SMALL_STATE(4657)] = 260841,
  [SMALL_STATE(4658)] = 260867,
  [SMALL_STATE(4659)] = 260903,
  [SMALL_STATE(4660)] = 260929,
  [SMALL_STATE(4661)] = 260961,
  [SMALL_STATE(4662)] = 260987,
  [SMALL_STATE(4663)] = 261013,
  [SMALL_STATE(4664)] = 261039,
  [SMALL_STATE(4665)] = 261065,
  [SMALL_STATE(4666)] = 261091,
  [SMALL_STATE(4667)] = 261127,
  [SMALL_STATE(4668)] = 261153,
  [SMALL_STATE(4669)] = 261179,
  [SMALL_STATE(4670)] = 261205,
  [SMALL_STATE(4671)] = 261231,
  [SMALL_STATE(4672)] = 261259,
  [SMALL_STATE(4673)] = 261289,
  [SMALL_STATE(4674)] = 261315,
  [SMALL_STATE(4675)] = 261341,
  [SMALL_STATE(4676)] = 261367,
  [SMALL_STATE(4677)] = 261393,
  [SMALL_STATE(4678)] = 261421,
  [SMALL_STATE(4679)] = 261447,
  [SMALL_STATE(4680)] = 261473,
  [SMALL_STATE(4681)] = 261499,
  [SMALL_STATE(4682)] = 261525,
  [SMALL_STATE(4683)] = 261551,
  [SMALL_STATE(4684)] = 261577,
  [SMALL_STATE(4685)] = 261603,
  [SMALL_STATE(4686)] = 261629,
  [SMALL_STATE(4687)] = 261655,
  [SMALL_STATE(4688)] = 261681,
  [SMALL_STATE(4689)] = 261707,
  [SMALL_STATE(4690)] = 261733,
  [SMALL_STATE(4691)] = 261759,
  [SMALL_STATE(4692)] = 261785,
  [SMALL_STATE(4693)] = 261811,
  [SMALL_STATE(4694)] = 261837,
  [SMALL_STATE(4695)] = 261863,
  [SMALL_STATE(4696)] = 261889,
  [SMALL_STATE(4697)] = 261915,
  [SMALL_STATE(4698)] = 261941,
  [SMALL_STATE(4699)] = 261967,
  [SMALL_STATE(4700)] = 261995,
  [SMALL_STATE(4701)] = 262021,
  [SMALL_STATE(4702)] = 262047,
  [SMALL_STATE(4703)] = 262073,
  [SMALL_STATE(4704)] = 262099,
  [SMALL_STATE(4705)] = 262133,
  [SMALL_STATE(4706)] = 262159,
  [SMALL_STATE(4707)] = 262185,
  [SMALL_STATE(4708)] = 262211,
  [SMALL_STATE(4709)] = 262237,
  [SMALL_STATE(4710)] = 262263,
  [SMALL_STATE(4711)] = 262289,
  [SMALL_STATE(4712)] = 262315,
  [SMALL_STATE(4713)] = 262341,
  [SMALL_STATE(4714)] = 262367,
  [SMALL_STATE(4715)] = 262393,
  [SMALL_STATE(4716)] = 262419,
  [SMALL_STATE(4717)] = 262451,
  [SMALL_STATE(4718)] = 262477,
  [SMALL_STATE(4719)] = 262503,
  [SMALL_STATE(4720)] = 262529,
  [SMALL_STATE(4721)] = 262555,
  [SMALL_STATE(4722)] = 262581,
  [SMALL_STATE(4723)] = 262607,
  [SMALL_STATE(4724)] = 262633,
  [SMALL_STATE(4725)] = 262659,
  [SMALL_STATE(4726)] = 262685,
  [SMALL_STATE(4727)] = 262711,
  [SMALL_STATE(4728)] = 262737,
  [SMALL_STATE(4729)] = 262763,
  [SMALL_STATE(4730)] = 262789,
  [SMALL_STATE(4731)] = 262815,
  [SMALL_STATE(4732)] = 262849,
  [SMALL_STATE(4733)] = 262875,
  [SMALL_STATE(4734)] = 262901,
  [SMALL_STATE(4735)] = 262927,
  [SMALL_STATE(4736)] = 262963,
  [SMALL_STATE(4737)] = 262989,
  [SMALL_STATE(4738)] = 263015,
  [SMALL_STATE(4739)] = 263041,
  [SMALL_STATE(4740)] = 263077,
  [SMALL_STATE(4741)] = 263103,
  [SMALL_STATE(4742)] = 263129,
  [SMALL_STATE(4743)] = 263155,
  [SMALL_STATE(4744)] = 263181,
  [SMALL_STATE(4745)] = 263217,
  [SMALL_STATE(4746)] = 263243,
  [SMALL_STATE(4747)] = 263269,
  [SMALL_STATE(4748)] = 263295,
  [SMALL_STATE(4749)] = 263321,
  [SMALL_STATE(4750)] = 263347,
  [SMALL_STATE(4751)] = 263373,
  [SMALL_STATE(4752)] = 263399,
  [SMALL_STATE(4753)] = 263425,
  [SMALL_STATE(4754)] = 263451,
  [SMALL_STATE(4755)] = 263477,
  [SMALL_STATE(4756)] = 263503,
  [SMALL_STATE(4757)] = 263529,
  [SMALL_STATE(4758)] = 263557,
  [SMALL_STATE(4759)] = 263583,
  [SMALL_STATE(4760)] = 263609,
  [SMALL_STATE(4761)] = 263635,
  [SMALL_STATE(4762)] = 263671,
  [SMALL_STATE(4763)] = 263697,
  [SMALL_STATE(4764)] = 263723,
  [SMALL_STATE(4765)] = 263749,
  [SMALL_STATE(4766)] = 263775,
  [SMALL_STATE(4767)] = 263801,
  [SMALL_STATE(4768)] = 263827,
  [SMALL_STATE(4769)] = 263853,
  [SMALL_STATE(4770)] = 263879,
  [SMALL_STATE(4771)] = 263905,
  [SMALL_STATE(4772)] = 263931,
  [SMALL_STATE(4773)] = 263965,
  [SMALL_STATE(4774)] = 263991,
  [SMALL_STATE(4775)] = 264023,
  [SMALL_STATE(4776)] = 264049,
  [SMALL_STATE(4777)] = 264075,
  [SMALL_STATE(4778)] = 264101,
  [SMALL_STATE(4779)] = 264127,
  [SMALL_STATE(4780)] = 264153,
  [SMALL_STATE(4781)] = 264179,
  [SMALL_STATE(4782)] = 264205,
  [SMALL_STATE(4783)] = 264231,
  [SMALL_STATE(4784)] = 264257,
  [SMALL_STATE(4785)] = 264283,
  [SMALL_STATE(4786)] = 264309,
  [SMALL_STATE(4787)] = 264345,
  [SMALL_STATE(4788)] = 264371,
  [SMALL_STATE(4789)] = 264397,
  [SMALL_STATE(4790)] = 264423,
  [SMALL_STATE(4791)] = 264459,
  [SMALL_STATE(4792)] = 264485,
  [SMALL_STATE(4793)] = 264517,
  [SMALL_STATE(4794)] = 264543,
  [SMALL_STATE(4795)] = 264569,
  [SMALL_STATE(4796)] = 264595,
  [SMALL_STATE(4797)] = 264621,
  [SMALL_STATE(4798)] = 264647,
  [SMALL_STATE(4799)] = 264673,
  [SMALL_STATE(4800)] = 264699,
  [SMALL_STATE(4801)] = 264725,
  [SMALL_STATE(4802)] = 264761,
  [SMALL_STATE(4803)] = 264797,
  [SMALL_STATE(4804)] = 264823,
  [SMALL_STATE(4805)] = 264849,
  [SMALL_STATE(4806)] = 264875,
  [SMALL_STATE(4807)] = 264913,
  [SMALL_STATE(4808)] = 264939,
  [SMALL_STATE(4809)] = 264965,
  [SMALL_STATE(4810)] = 264991,
  [SMALL_STATE(4811)] = 265017,
  [SMALL_STATE(4812)] = 265053,
  [SMALL_STATE(4813)] = 265079,
  [SMALL_STATE(4814)] = 265105,
  [SMALL_STATE(4815)] = 265131,
  [SMALL_STATE(4816)] = 265157,
  [SMALL_STATE(4817)] = 265183,
  [SMALL_STATE(4818)] = 265209,
  [SMALL_STATE(4819)] = 265235,
  [SMALL_STATE(4820)] = 265261,
  [SMALL_STATE(4821)] = 265293,
  [SMALL_STATE(4822)] = 265319,
  [SMALL_STATE(4823)] = 265345,
  [SMALL_STATE(4824)] = 265371,
  [SMALL_STATE(4825)] = 265407,
  [SMALL_STATE(4826)] = 265433,
  [SMALL_STATE(4827)] = 265461,
  [SMALL_STATE(4828)] = 265487,
  [SMALL_STATE(4829)] = 265517,
  [SMALL_STATE(4830)] = 265543,
  [SMALL_STATE(4831)] = 265569,
  [SMALL_STATE(4832)] = 265595,
  [SMALL_STATE(4833)] = 265621,
  [SMALL_STATE(4834)] = 265647,
  [SMALL_STATE(4835)] = 265673,
  [SMALL_STATE(4836)] = 265699,
  [SMALL_STATE(4837)] = 265725,
  [SMALL_STATE(4838)] = 265751,
  [SMALL_STATE(4839)] = 265777,
  [SMALL_STATE(4840)] = 265803,
  [SMALL_STATE(4841)] = 265829,
  [SMALL_STATE(4842)] = 265855,
  [SMALL_STATE(4843)] = 265891,
  [SMALL_STATE(4844)] = 265917,
  [SMALL_STATE(4845)] = 265943,
  [SMALL_STATE(4846)] = 265979,
  [SMALL_STATE(4847)] = 266013,
  [SMALL_STATE(4848)] = 266039,
  [SMALL_STATE(4849)] = 266065,
  [SMALL_STATE(4850)] = 266091,
  [SMALL_STATE(4851)] = 266117,
  [SMALL_STATE(4852)] = 266143,
  [SMALL_STATE(4853)] = 266181,
  [SMALL_STATE(4854)] = 266207,
  [SMALL_STATE(4855)] = 266233,
  [SMALL_STATE(4856)] = 266259,
  [SMALL_STATE(4857)] = 266285,
  [SMALL_STATE(4858)] = 266321,
  [SMALL_STATE(4859)] = 266357,
  [SMALL_STATE(4860)] = 266383,
  [SMALL_STATE(4861)] = 266409,
  [SMALL_STATE(4862)] = 266435,
  [SMALL_STATE(4863)] = 266461,
  [SMALL_STATE(4864)] = 266487,
  [SMALL_STATE(4865)] = 266523,
  [SMALL_STATE(4866)] = 266549,
  [SMALL_STATE(4867)] = 266575,
  [SMALL_STATE(4868)] = 266603,
  [SMALL_STATE(4869)] = 266639,
  [SMALL_STATE(4870)] = 266665,
  [SMALL_STATE(4871)] = 266691,
  [SMALL_STATE(4872)] = 266717,
  [SMALL_STATE(4873)] = 266743,
  [SMALL_STATE(4874)] = 266769,
  [SMALL_STATE(4875)] = 266795,
  [SMALL_STATE(4876)] = 266821,
  [SMALL_STATE(4877)] = 266847,
  [SMALL_STATE(4878)] = 266873,
  [SMALL_STATE(4879)] = 266899,
  [SMALL_STATE(4880)] = 266925,
  [SMALL_STATE(4881)] = 266951,
  [SMALL_STATE(4882)] = 266977,
  [SMALL_STATE(4883)] = 267003,
  [SMALL_STATE(4884)] = 267039,
  [SMALL_STATE(4885)] = 267065,
  [SMALL_STATE(4886)] = 267091,
  [SMALL_STATE(4887)] = 267117,
  [SMALL_STATE(4888)] = 267143,
  [SMALL_STATE(4889)] = 267169,
  [SMALL_STATE(4890)] = 267195,
  [SMALL_STATE(4891)] = 267229,
  [SMALL_STATE(4892)] = 267255,
  [SMALL_STATE(4893)] = 267280,
  [SMALL_STATE(4894)] = 267305,
  [SMALL_STATE(4895)] = 267330,
  [SMALL_STATE(4896)] = 267361,
  [SMALL_STATE(4897)] = 267394,
  [SMALL_STATE(4898)] = 267419,
  [SMALL_STATE(4899)] = 267444,
  [SMALL_STATE(4900)] = 267475,
  [SMALL_STATE(4901)] = 267508,
  [SMALL_STATE(4902)] = 267533,
  [SMALL_STATE(4903)] = 267558,
  [SMALL_STATE(4904)] = 267589,
  [SMALL_STATE(4905)] = 267622,
  [SMALL_STATE(4906)] = 267655,
  [SMALL_STATE(4907)] = 267694,
  [SMALL_STATE(4908)] = 267719,
  [SMALL_STATE(4909)] = 267744,
  [SMALL_STATE(4910)] = 267771,
  [SMALL_STATE(4911)] = 267796,
  [SMALL_STATE(4912)] = 267829,
  [SMALL_STATE(4913)] = 267862,
  [SMALL_STATE(4914)] = 267889,
  [SMALL_STATE(4915)] = 267920,
  [SMALL_STATE(4916)] = 267953,
  [SMALL_STATE(4917)] = 267986,
  [SMALL_STATE(4918)] = 268013,
  [SMALL_STATE(4919)] = 268044,
  [SMALL_STATE(4920)] = 268069,
  [SMALL_STATE(4921)] = 268094,
  [SMALL_STATE(4922)] = 268119,
  [SMALL_STATE(4923)] = 268152,
  [SMALL_STATE(4924)] = 268185,
  [SMALL_STATE(4925)] = 268218,
  [SMALL_STATE(4926)] = 268251,
  [SMALL_STATE(4927)] = 268276,
  [SMALL_STATE(4928)] = 268301,
  [SMALL_STATE(4929)] = 268328,
  [SMALL_STATE(4930)] = 268353,
  [SMALL_STATE(4931)] = 268378,
  [SMALL_STATE(4932)] = 268411,
  [SMALL_STATE(4933)] = 268436,
  [SMALL_STATE(4934)] = 268461,
  [SMALL_STATE(4935)] = 268494,
  [SMALL_STATE(4936)] = 268527,
  [SMALL_STATE(4937)] = 268552,
  [SMALL_STATE(4938)] = 268577,
  [SMALL_STATE(4939)] = 268616,
  [SMALL_STATE(4940)] = 268649,
  [SMALL_STATE(4941)] = 268674,
  [SMALL_STATE(4942)] = 268713,
  [SMALL_STATE(4943)] = 268738,
  [SMALL_STATE(4944)] = 268765,
  [SMALL_STATE(4945)] = 268796,
  [SMALL_STATE(4946)] = 268821,
  [SMALL_STATE(4947)] = 268846,
  [SMALL_STATE(4948)] = 268871,
  [SMALL_STATE(4949)] = 268904,
  [SMALL_STATE(4950)] = 268929,
  [SMALL_STATE(4951)] = 268954,
  [SMALL_STATE(4952)] = 268985,
  [SMALL_STATE(4953)] = 269010,
  [SMALL_STATE(4954)] = 269041,
  [SMALL_STATE(4955)] = 269074,
  [SMALL_STATE(4956)] = 269099,
  [SMALL_STATE(4957)] = 269130,
  [SMALL_STATE(4958)] = 269157,
  [SMALL_STATE(4959)] = 269185,
  [SMALL_STATE(4960)] = 269215,
  [SMALL_STATE(4961)] = 269245,
  [SMALL_STATE(4962)] = 269281,
  [SMALL_STATE(4963)] = 269305,
  [SMALL_STATE(4964)] = 269329,
  [SMALL_STATE(4965)] = 269359,
  [SMALL_STATE(4966)] = 269383,
  [SMALL_STATE(4967)] = 269417,
  [SMALL_STATE(4968)] = 269441,
  [SMALL_STATE(4969)] = 269471,
  [SMALL_STATE(4970)] = 269507,
  [SMALL_STATE(4971)] = 269543,
  [SMALL_STATE(4972)] = 269573,
  [SMALL_STATE(4973)] = 269597,
  [SMALL_STATE(4974)] = 269625,
  [SMALL_STATE(4975)] = 269655,
  [SMALL_STATE(4976)] = 269679,
  [SMALL_STATE(4977)] = 269709,
  [SMALL_STATE(4978)] = 269733,
  [SMALL_STATE(4979)] = 269763,
  [SMALL_STATE(4980)] = 269787,
  [SMALL_STATE(4981)] = 269817,
  [SMALL_STATE(4982)] = 269841,
  [SMALL_STATE(4983)] = 269865,
  [SMALL_STATE(4984)] = 269889,
  [SMALL_STATE(4985)] = 269913,
  [SMALL_STATE(4986)] = 269937,
  [SMALL_STATE(4987)] = 269967,
  [SMALL_STATE(4988)] = 270003,
  [SMALL_STATE(4989)] = 270027,
  [SMALL_STATE(4990)] = 270051,
  [SMALL_STATE(4991)] = 270087,
  [SMALL_STATE(4992)] = 270123,
  [SMALL_STATE(4993)] = 270147,
  [SMALL_STATE(4994)] = 270177,
  [SMALL_STATE(4995)] = 270213,
  [SMALL_STATE(4996)] = 270243,
  [SMALL_STATE(4997)] = 270279,
  [SMALL_STATE(4998)] = 270309,
  [SMALL_STATE(4999)] = 270333,
  [SMALL_STATE(5000)] = 270369,
  [SMALL_STATE(5001)] = 270403,
  [SMALL_STATE(5002)] = 270439,
  [SMALL_STATE(5003)] = 270475,
  [SMALL_STATE(5004)] = 270511,
  [SMALL_STATE(5005)] = 270547,
  [SMALL_STATE(5006)] = 270577,
  [SMALL_STATE(5007)] = 270607,
  [SMALL_STATE(5008)] = 270635,
  [SMALL_STATE(5009)] = 270665,
  [SMALL_STATE(5010)] = 270695,
  [SMALL_STATE(5011)] = 270719,
  [SMALL_STATE(5012)] = 270755,
  [SMALL_STATE(5013)] = 270779,
  [SMALL_STATE(5014)] = 270809,
  [SMALL_STATE(5015)] = 270845,
  [SMALL_STATE(5016)] = 270869,
  [SMALL_STATE(5017)] = 270893,
  [SMALL_STATE(5018)] = 270917,
  [SMALL_STATE(5019)] = 270947,
  [SMALL_STATE(5020)] = 270977,
  [SMALL_STATE(5021)] = 271001,
  [SMALL_STATE(5022)] = 271037,
  [SMALL_STATE(5023)] = 271061,
  [SMALL_STATE(5024)] = 271097,
  [SMALL_STATE(5025)] = 271127,
  [SMALL_STATE(5026)] = 271151,
  [SMALL_STATE(5027)] = 271175,
  [SMALL_STATE(5028)] = 271199,
  [SMALL_STATE(5029)] = 271235,
  [SMALL_STATE(5030)] = 271259,
  [SMALL_STATE(5031)] = 271283,
  [SMALL_STATE(5032)] = 271319,
  [SMALL_STATE(5033)] = 271349,
  [SMALL_STATE(5034)] = 271379,
  [SMALL_STATE(5035)] = 271413,
  [SMALL_STATE(5036)] = 271437,
  [SMALL_STATE(5037)] = 271461,
  [SMALL_STATE(5038)] = 271489,
  [SMALL_STATE(5039)] = 271513,
  [SMALL_STATE(5040)] = 271537,
  [SMALL_STATE(5041)] = 271571,
  [SMALL_STATE(5042)] = 271595,
  [SMALL_STATE(5043)] = 271619,
  [SMALL_STATE(5044)] = 271643,
  [SMALL_STATE(5045)] = 271673,
  [SMALL_STATE(5046)] = 271703,
  [SMALL_STATE(5047)] = 271733,
  [SMALL_STATE(5048)] = 271763,
  [SMALL_STATE(5049)] = 271787,
  [SMALL_STATE(5050)] = 271823,
  [SMALL_STATE(5051)] = 271857,
  [SMALL_STATE(5052)] = 271884,
  [SMALL_STATE(5053)] = 271911,
  [SMALL_STATE(5054)] = 271938,
  [SMALL_STATE(5055)] = 271965,
  [SMALL_STATE(5056)] = 271992,
  [SMALL_STATE(5057)] = 272023,
  [SMALL_STATE(5058)] = 272050,
  [SMALL_STATE(5059)] = 272083,
  [SMALL_STATE(5060)] = 272110,
  [SMALL_STATE(5061)] = 272137,
  [SMALL_STATE(5062)] = 272164,
  [SMALL_STATE(5063)] = 272191,
  [SMALL_STATE(5064)] = 272218,
  [SMALL_STATE(5065)] = 272245,
  [SMALL_STATE(5066)] = 272268,
  [SMALL_STATE(5067)] = 272297,
  [SMALL_STATE(5068)] = 272322,
  [SMALL_STATE(5069)] = 272349,
  [SMALL_STATE(5070)] = 272376,
  [SMALL_STATE(5071)] = 272403,
  [SMALL_STATE(5072)] = 272430,
  [SMALL_STATE(5073)] = 272457,
  [SMALL_STATE(5074)] = 272484,
  [SMALL_STATE(5075)] = 272511,
  [SMALL_STATE(5076)] = 272538,
  [SMALL_STATE(5077)] = 272565,
  [SMALL_STATE(5078)] = 272592,
  [SMALL_STATE(5079)] = 272619,
  [SMALL_STATE(5080)] = 272646,
  [SMALL_STATE(5081)] = 272673,
  [SMALL_STATE(5082)] = 272706,
  [SMALL_STATE(5083)] = 272739,
  [SMALL_STATE(5084)] = 272766,
  [SMALL_STATE(5085)] = 272793,
  [SMALL_STATE(5086)] = 272820,
  [SMALL_STATE(5087)] = 272847,
  [SMALL_STATE(5088)] = 272874,
  [SMALL_STATE(5089)] = 272897,
  [SMALL_STATE(5090)] = 272924,
  [SMALL_STATE(5091)] = 272957,
  [SMALL_STATE(5092)] = 272988,
  [SMALL_STATE(5093)] = 273019,
  [SMALL_STATE(5094)] = 273046,
  [SMALL_STATE(5095)] = 273077,
  [SMALL_STATE(5096)] = 273104,
  [SMALL_STATE(5097)] = 273129,
  [SMALL_STATE(5098)] = 273154,
  [SMALL_STATE(5099)] = 273181,
  [SMALL_STATE(5100)] = 273208,
  [SMALL_STATE(5101)] = 273235,
  [SMALL_STATE(5102)] = 273262,
  [SMALL_STATE(5103)] = 273289,
  [SMALL_STATE(5104)] = 273316,
  [SMALL_STATE(5105)] = 273343,
  [SMALL_STATE(5106)] = 273370,
  [SMALL_STATE(5107)] = 273397,
  [SMALL_STATE(5108)] = 273424,
  [SMALL_STATE(5109)] = 273451,
  [SMALL_STATE(5110)] = 273482,
  [SMALL_STATE(5111)] = 273509,
  [SMALL_STATE(5112)] = 273540,
  [SMALL_STATE(5113)] = 273567,
  [SMALL_STATE(5114)] = 273594,
  [SMALL_STATE(5115)] = 273621,
  [SMALL_STATE(5116)] = 273648,
  [SMALL_STATE(5117)] = 273679,
  [SMALL_STATE(5118)] = 273706,
  [SMALL_STATE(5119)] = 273733,
  [SMALL_STATE(5120)] = 273760,
  [SMALL_STATE(5121)] = 273793,
  [SMALL_STATE(5122)] = 273820,
  [SMALL_STATE(5123)] = 273853,
  [SMALL_STATE(5124)] = 273882,
  [SMALL_STATE(5125)] = 273909,
  [SMALL_STATE(5126)] = 273936,
  [SMALL_STATE(5127)] = 273963,
  [SMALL_STATE(5128)] = 273990,
  [SMALL_STATE(5129)] = 274017,
  [SMALL_STATE(5130)] = 274048,
  [SMALL_STATE(5131)] = 274075,
  [SMALL_STATE(5132)] = 274102,
  [SMALL_STATE(5133)] = 274129,
  [SMALL_STATE(5134)] = 274156,
  [SMALL_STATE(5135)] = 274183,
  [SMALL_STATE(5136)] = 274210,
  [SMALL_STATE(5137)] = 274237,
  [SMALL_STATE(5138)] = 274264,
  [SMALL_STATE(5139)] = 274295,
  [SMALL_STATE(5140)] = 274322,
  [SMALL_STATE(5141)] = 274349,
  [SMALL_STATE(5142)] = 274376,
  [SMALL_STATE(5143)] = 274403,
  [SMALL_STATE(5144)] = 274430,
  [SMALL_STATE(5145)] = 274457,
  [SMALL_STATE(5146)] = 274486,
  [SMALL_STATE(5147)] = 274513,
  [SMALL_STATE(5148)] = 274544,
  [SMALL_STATE(5149)] = 274568,
  [SMALL_STATE(5150)] = 274592,
  [SMALL_STATE(5151)] = 274616,
  [SMALL_STATE(5152)] = 274646,
  [SMALL_STATE(5153)] = 274668,
  [SMALL_STATE(5154)] = 274698,
  [SMALL_STATE(5155)] = 274722,
  [SMALL_STATE(5156)] = 274748,
  [SMALL_STATE(5157)] = 274772,
  [SMALL_STATE(5158)] = 274796,
  [SMALL_STATE(5159)] = 274818,
  [SMALL_STATE(5160)] = 274842,
  [SMALL_STATE(5161)] = 274872,
  [SMALL_STATE(5162)] = 274896,
  [SMALL_STATE(5163)] = 274920,
  [SMALL_STATE(5164)] = 274944,
  [SMALL_STATE(5165)] = 274974,
  [SMALL_STATE(5166)] = 274998,
  [SMALL_STATE(5167)] = 275024,
  [SMALL_STATE(5168)] = 275054,
  [SMALL_STATE(5169)] = 275084,
  [SMALL_STATE(5170)] = 275114,
  [SMALL_STATE(5171)] = 275136,
  [SMALL_STATE(5172)] = 275166,
  [SMALL_STATE(5173)] = 275196,
  [SMALL_STATE(5174)] = 275222,
  [SMALL_STATE(5175)] = 275246,
  [SMALL_STATE(5176)] = 275270,
  [SMALL_STATE(5177)] = 275300,
  [SMALL_STATE(5178)] = 275324,
  [SMALL_STATE(5179)] = 275354,
  [SMALL_STATE(5180)] = 275378,
  [SMALL_STATE(5181)] = 275402,
  [SMALL_STATE(5182)] = 275432,
  [SMALL_STATE(5183)] = 275458,
  [SMALL_STATE(5184)] = 275488,
  [SMALL_STATE(5185)] = 275518,
  [SMALL_STATE(5186)] = 275548,
  [SMALL_STATE(5187)] = 275570,
  [SMALL_STATE(5188)] = 275594,
  [SMALL_STATE(5189)] = 275624,
  [SMALL_STATE(5190)] = 275648,
  [SMALL_STATE(5191)] = 275672,
  [SMALL_STATE(5192)] = 275696,
  [SMALL_STATE(5193)] = 275724,
  [SMALL_STATE(5194)] = 275754,
  [SMALL_STATE(5195)] = 275778,
  [SMALL_STATE(5196)] = 275802,
  [SMALL_STATE(5197)] = 275826,
  [SMALL_STATE(5198)] = 275850,
  [SMALL_STATE(5199)] = 275874,
  [SMALL_STATE(5200)] = 275904,
  [SMALL_STATE(5201)] = 275930,
  [SMALL_STATE(5202)] = 275954,
  [SMALL_STATE(5203)] = 275984,
  [SMALL_STATE(5204)] = 276008,
  [SMALL_STATE(5205)] = 276038,
  [SMALL_STATE(5206)] = 276062,
  [SMALL_STATE(5207)] = 276092,
  [SMALL_STATE(5208)] = 276122,
  [SMALL_STATE(5209)] = 276152,
  [SMALL_STATE(5210)] = 276182,
  [SMALL_STATE(5211)] = 276206,
  [SMALL_STATE(5212)] = 276230,
  [SMALL_STATE(5213)] = 276260,
  [SMALL_STATE(5214)] = 276284,
  [SMALL_STATE(5215)] = 276308,
  [SMALL_STATE(5216)] = 276332,
  [SMALL_STATE(5217)] = 276362,
  [SMALL_STATE(5218)] = 276386,
  [SMALL_STATE(5219)] = 276412,
  [SMALL_STATE(5220)] = 276438,
  [SMALL_STATE(5221)] = 276468,
  [SMALL_STATE(5222)] = 276490,
  [SMALL_STATE(5223)] = 276514,
  [SMALL_STATE(5224)] = 276538,
  [SMALL_STATE(5225)] = 276562,
  [SMALL_STATE(5226)] = 276584,
  [SMALL_STATE(5227)] = 276614,
  [SMALL_STATE(5228)] = 276638,
  [SMALL_STATE(5229)] = 276662,
  [SMALL_STATE(5230)] = 276692,
  [SMALL_STATE(5231)] = 276714,
  [SMALL_STATE(5232)] = 276744,
  [SMALL_STATE(5233)] = 276768,
  [SMALL_STATE(5234)] = 276794,
  [SMALL_STATE(5235)] = 276822,
  [SMALL_STATE(5236)] = 276846,
  [SMALL_STATE(5237)] = 276876,
  [SMALL_STATE(5238)] = 276898,
  [SMALL_STATE(5239)] = 276924,
  [SMALL_STATE(5240)] = 276948,
  [SMALL_STATE(5241)] = 276978,
  [SMALL_STATE(5242)] = 277002,
  [SMALL_STATE(5243)] = 277026,
  [SMALL_STATE(5244)] = 277052,
  [SMALL_STATE(5245)] = 277076,
  [SMALL_STATE(5246)] = 277100,
  [SMALL_STATE(5247)] = 277126,
  [SMALL_STATE(5248)] = 277156,
  [SMALL_STATE(5249)] = 277180,
  [SMALL_STATE(5250)] = 277202,
  [SMALL_STATE(5251)] = 277226,
  [SMALL_STATE(5252)] = 277250,
  [SMALL_STATE(5253)] = 277280,
  [SMALL_STATE(5254)] = 277310,
  [SMALL_STATE(5255)] = 277336,
  [SMALL_STATE(5256)] = 277360,
  [SMALL_STATE(5257)] = 277382,
  [SMALL_STATE(5258)] = 277412,
  [SMALL_STATE(5259)] = 277436,
  [SMALL_STATE(5260)] = 277460,
  [SMALL_STATE(5261)] = 277490,
  [SMALL_STATE(5262)] = 277514,
  [SMALL_STATE(5263)] = 277536,
  [SMALL_STATE(5264)] = 277560,
  [SMALL_STATE(5265)] = 277584,
  [SMALL_STATE(5266)] = 277608,
  [SMALL_STATE(5267)] = 277630,
  [SMALL_STATE(5268)] = 277652,
  [SMALL_STATE(5269)] = 277676,
  [SMALL_STATE(5270)] = 277700,
  [SMALL_STATE(5271)] = 277730,
  [SMALL_STATE(5272)] = 277758,
  [SMALL_STATE(5273)] = 277782,
  [SMALL_STATE(5274)] = 277806,
  [SMALL_STATE(5275)] = 277830,
  [SMALL_STATE(5276)] = 277860,
  [SMALL_STATE(5277)] = 277884,
  [SMALL_STATE(5278)] = 277908,
  [SMALL_STATE(5279)] = 277932,
  [SMALL_STATE(5280)] = 277956,
  [SMALL_STATE(5281)] = 277980,
  [SMALL_STATE(5282)] = 278004,
  [SMALL_STATE(5283)] = 278034,
  [SMALL_STATE(5284)] = 278058,
  [SMALL_STATE(5285)] = 278082,
  [SMALL_STATE(5286)] = 278106,
  [SMALL_STATE(5287)] = 278130,
  [SMALL_STATE(5288)] = 278152,
  [SMALL_STATE(5289)] = 278180,
  [SMALL_STATE(5290)] = 278204,
  [SMALL_STATE(5291)] = 278234,
  [SMALL_STATE(5292)] = 278264,
  [SMALL_STATE(5293)] = 278294,
  [SMALL_STATE(5294)] = 278318,
  [SMALL_STATE(5295)] = 278342,
  [SMALL_STATE(5296)] = 278366,
  [SMALL_STATE(5297)] = 278390,
  [SMALL_STATE(5298)] = 278414,
  [SMALL_STATE(5299)] = 278436,
  [SMALL_STATE(5300)] = 278460,
  [SMALL_STATE(5301)] = 278490,
  [SMALL_STATE(5302)] = 278514,
  [SMALL_STATE(5303)] = 278538,
  [SMALL_STATE(5304)] = 278568,
  [SMALL_STATE(5305)] = 278592,
  [SMALL_STATE(5306)] = 278616,
  [SMALL_STATE(5307)] = 278646,
  [SMALL_STATE(5308)] = 278676,
  [SMALL_STATE(5309)] = 278700,
  [SMALL_STATE(5310)] = 278726,
  [SMALL_STATE(5311)] = 278750,
  [SMALL_STATE(5312)] = 278780,
  [SMALL_STATE(5313)] = 278804,
  [SMALL_STATE(5314)] = 278828,
  [SMALL_STATE(5315)] = 278852,
  [SMALL_STATE(5316)] = 278876,
  [SMALL_STATE(5317)] = 278900,
  [SMALL_STATE(5318)] = 278924,
  [SMALL_STATE(5319)] = 278948,
  [SMALL_STATE(5320)] = 278970,
  [SMALL_STATE(5321)] = 278994,
  [SMALL_STATE(5322)] = 279018,
  [SMALL_STATE(5323)] = 279042,
  [SMALL_STATE(5324)] = 279064,
  [SMALL_STATE(5325)] = 279088,
  [SMALL_STATE(5326)] = 279112,
  [SMALL_STATE(5327)] = 279136,
  [SMALL_STATE(5328)] = 279160,
  [SMALL_STATE(5329)] = 279184,
  [SMALL_STATE(5330)] = 279208,
  [SMALL_STATE(5331)] = 279232,
  [SMALL_STATE(5332)] = 279259,
  [SMALL_STATE(5333)] = 279286,
  [SMALL_STATE(5334)] = 279313,
  [SMALL_STATE(5335)] = 279340,
  [SMALL_STATE(5336)] = 279367,
  [SMALL_STATE(5337)] = 279394,
  [SMALL_STATE(5338)] = 279419,
  [SMALL_STATE(5339)] = 279446,
  [SMALL_STATE(5340)] = 279473,
  [SMALL_STATE(5341)] = 279494,
  [SMALL_STATE(5342)] = 279521,
  [SMALL_STATE(5343)] = 279546,
  [SMALL_STATE(5344)] = 279573,
  [SMALL_STATE(5345)] = 279600,
  [SMALL_STATE(5346)] = 279627,
  [SMALL_STATE(5347)] = 279648,
  [SMALL_STATE(5348)] = 279675,
  [SMALL_STATE(5349)] = 279702,
  [SMALL_STATE(5350)] = 279729,
  [SMALL_STATE(5351)] = 279754,
  [SMALL_STATE(5352)] = 279781,
  [SMALL_STATE(5353)] = 279808,
  [SMALL_STATE(5354)] = 279835,
  [SMALL_STATE(5355)] = 279862,
  [SMALL_STATE(5356)] = 279883,
  [SMALL_STATE(5357)] = 279910,
  [SMALL_STATE(5358)] = 279937,
  [SMALL_STATE(5359)] = 279958,
  [SMALL_STATE(5360)] = 279979,
  [SMALL_STATE(5361)] = 280000,
  [SMALL_STATE(5362)] = 280027,
  [SMALL_STATE(5363)] = 280054,
  [SMALL_STATE(5364)] = 280075,
  [SMALL_STATE(5365)] = 280096,
  [SMALL_STATE(5366)] = 280123,
  [SMALL_STATE(5367)] = 280150,
  [SMALL_STATE(5368)] = 280175,
  [SMALL_STATE(5369)] = 280202,
  [SMALL_STATE(5370)] = 280229,
  [SMALL_STATE(5371)] = 280250,
  [SMALL_STATE(5372)] = 280277,
  [SMALL_STATE(5373)] = 280304,
  [SMALL_STATE(5374)] = 280331,
  [SMALL_STATE(5375)] = 280358,
  [SMALL_STATE(5376)] = 280385,
  [SMALL_STATE(5377)] = 280412,
  [SMALL_STATE(5378)] = 280437,
  [SMALL_STATE(5379)] = 280464,
  [SMALL_STATE(5380)] = 280491,
  [SMALL_STATE(5381)] = 280518,
  [SMALL_STATE(5382)] = 280545,
  [SMALL_STATE(5383)] = 280572,
  [SMALL_STATE(5384)] = 280593,
  [SMALL_STATE(5385)] = 280620,
  [SMALL_STATE(5386)] = 280647,
  [SMALL_STATE(5387)] = 280674,
  [SMALL_STATE(5388)] = 280701,
  [SMALL_STATE(5389)] = 280728,
  [SMALL_STATE(5390)] = 280749,
  [SMALL_STATE(5391)] = 280776,
  [SMALL_STATE(5392)] = 280803,
  [SMALL_STATE(5393)] = 280824,
  [SMALL_STATE(5394)] = 280851,
  [SMALL_STATE(5395)] = 280878,
  [SMALL_STATE(5396)] = 280905,
  [SMALL_STATE(5397)] = 280932,
  [SMALL_STATE(5398)] = 280957,
  [SMALL_STATE(5399)] = 280978,
  [SMALL_STATE(5400)] = 281005,
  [SMALL_STATE(5401)] = 281032,
  [SMALL_STATE(5402)] = 281059,
  [SMALL_STATE(5403)] = 281086,
  [SMALL_STATE(5404)] = 281113,
  [SMALL_STATE(5405)] = 281140,
  [SMALL_STATE(5406)] = 281161,
  [SMALL_STATE(5407)] = 281188,
  [SMALL_STATE(5408)] = 281215,
  [SMALL_STATE(5409)] = 281242,
  [SMALL_STATE(5410)] = 281269,
  [SMALL_STATE(5411)] = 281296,
  [SMALL_STATE(5412)] = 281323,
  [SMALL_STATE(5413)] = 281350,
  [SMALL_STATE(5414)] = 281377,
  [SMALL_STATE(5415)] = 281404,
  [SMALL_STATE(5416)] = 281431,
  [SMALL_STATE(5417)] = 281452,
  [SMALL_STATE(5418)] = 281473,
  [SMALL_STATE(5419)] = 281500,
  [SMALL_STATE(5420)] = 281527,
  [SMALL_STATE(5421)] = 281548,
  [SMALL_STATE(5422)] = 281575,
  [SMALL_STATE(5423)] = 281602,
  [SMALL_STATE(5424)] = 281627,
  [SMALL_STATE(5425)] = 281654,
  [SMALL_STATE(5426)] = 281681,
  [SMALL_STATE(5427)] = 281704,
  [SMALL_STATE(5428)] = 281731,
  [SMALL_STATE(5429)] = 281758,
  [SMALL_STATE(5430)] = 281785,
  [SMALL_STATE(5431)] = 281812,
  [SMALL_STATE(5432)] = 281839,
  [SMALL_STATE(5433)] = 281866,
  [SMALL_STATE(5434)] = 281893,
  [SMALL_STATE(5435)] = 281920,
  [SMALL_STATE(5436)] = 281947,
  [SMALL_STATE(5437)] = 281974,
  [SMALL_STATE(5438)] = 282001,
  [SMALL_STATE(5439)] = 282028,
  [SMALL_STATE(5440)] = 282055,
  [SMALL_STATE(5441)] = 282082,
  [SMALL_STATE(5442)] = 282107,
  [SMALL_STATE(5443)] = 282132,
  [SMALL_STATE(5444)] = 282159,
  [SMALL_STATE(5445)] = 282186,
  [SMALL_STATE(5446)] = 282213,
  [SMALL_STATE(5447)] = 282240,
  [SMALL_STATE(5448)] = 282261,
  [SMALL_STATE(5449)] = 282288,
  [SMALL_STATE(5450)] = 282315,
  [SMALL_STATE(5451)] = 282342,
  [SMALL_STATE(5452)] = 282369,
  [SMALL_STATE(5453)] = 282396,
  [SMALL_STATE(5454)] = 282423,
  [SMALL_STATE(5455)] = 282450,
  [SMALL_STATE(5456)] = 282471,
  [SMALL_STATE(5457)] = 282498,
  [SMALL_STATE(5458)] = 282525,
  [SMALL_STATE(5459)] = 282552,
  [SMALL_STATE(5460)] = 282579,
  [SMALL_STATE(5461)] = 282606,
  [SMALL_STATE(5462)] = 282633,
  [SMALL_STATE(5463)] = 282660,
  [SMALL_STATE(5464)] = 282687,
  [SMALL_STATE(5465)] = 282714,
  [SMALL_STATE(5466)] = 282741,
  [SMALL_STATE(5467)] = 282768,
  [SMALL_STATE(5468)] = 282795,
  [SMALL_STATE(5469)] = 282816,
  [SMALL_STATE(5470)] = 282837,
  [SMALL_STATE(5471)] = 282858,
  [SMALL_STATE(5472)] = 282879,
  [SMALL_STATE(5473)] = 282900,
  [SMALL_STATE(5474)] = 282927,
  [SMALL_STATE(5475)] = 282948,
  [SMALL_STATE(5476)] = 282971,
  [SMALL_STATE(5477)] = 282992,
  [SMALL_STATE(5478)] = 283013,
  [SMALL_STATE(5479)] = 283034,
  [SMALL_STATE(5480)] = 283055,
  [SMALL_STATE(5481)] = 283076,
  [SMALL_STATE(5482)] = 283103,
  [SMALL_STATE(5483)] = 283124,
  [SMALL_STATE(5484)] = 283149,
  [SMALL_STATE(5485)] = 283170,
  [SMALL_STATE(5486)] = 283191,
  [SMALL_STATE(5487)] = 283218,
  [SMALL_STATE(5488)] = 283245,
  [SMALL_STATE(5489)] = 283272,
  [SMALL_STATE(5490)] = 283293,
  [SMALL_STATE(5491)] = 283318,
  [SMALL_STATE(5492)] = 283343,
  [SMALL_STATE(5493)] = 283368,
  [SMALL_STATE(5494)] = 283389,
  [SMALL_STATE(5495)] = 283410,
  [SMALL_STATE(5496)] = 283431,
  [SMALL_STATE(5497)] = 283458,
  [SMALL_STATE(5498)] = 283479,
  [SMALL_STATE(5499)] = 283506,
  [SMALL_STATE(5500)] = 283527,
  [SMALL_STATE(5501)] = 283548,
  [SMALL_STATE(5502)] = 283569,
  [SMALL_STATE(5503)] = 283596,
  [SMALL_STATE(5504)] = 283619,
  [SMALL_STATE(5505)] = 283644,
  [SMALL_STATE(5506)] = 283671,
  [SMALL_STATE(5507)] = 283692,
  [SMALL_STATE(5508)] = 283713,
  [SMALL_STATE(5509)] = 283736,
  [SMALL_STATE(5510)] = 283763,
  [SMALL_STATE(5511)] = 283786,
  [SMALL_STATE(5512)] = 283813,
  [SMALL_STATE(5513)] = 283838,
  [SMALL_STATE(5514)] = 283859,
  [SMALL_STATE(5515)] = 283884,
  [SMALL_STATE(5516)] = 283905,
  [SMALL_STATE(5517)] = 283926,
  [SMALL_STATE(5518)] = 283947,
  [SMALL_STATE(5519)] = 283968,
  [SMALL_STATE(5520)] = 283989,
  [SMALL_STATE(5521)] = 284010,
  [SMALL_STATE(5522)] = 284037,
  [SMALL_STATE(5523)] = 284058,
  [SMALL_STATE(5524)] = 284079,
  [SMALL_STATE(5525)] = 284102,
  [SMALL_STATE(5526)] = 284123,
  [SMALL_STATE(5527)] = 284144,
  [SMALL_STATE(5528)] = 284165,
  [SMALL_STATE(5529)] = 284186,
  [SMALL_STATE(5530)] = 284207,
  [SMALL_STATE(5531)] = 284228,
  [SMALL_STATE(5532)] = 284249,
  [SMALL_STATE(5533)] = 284270,
  [SMALL_STATE(5534)] = 284291,
  [SMALL_STATE(5535)] = 284318,
  [SMALL_STATE(5536)] = 284339,
  [SMALL_STATE(5537)] = 284366,
  [SMALL_STATE(5538)] = 284391,
  [SMALL_STATE(5539)] = 284414,
  [SMALL_STATE(5540)] = 284441,
  [SMALL_STATE(5541)] = 284468,
  [SMALL_STATE(5542)] = 284495,
  [SMALL_STATE(5543)] = 284522,
  [SMALL_STATE(5544)] = 284549,
  [SMALL_STATE(5545)] = 284574,
  [SMALL_STATE(5546)] = 284599,
  [SMALL_STATE(5547)] = 284626,
  [SMALL_STATE(5548)] = 284653,
  [SMALL_STATE(5549)] = 284678,
  [SMALL_STATE(5550)] = 284705,
  [SMALL_STATE(5551)] = 284732,
  [SMALL_STATE(5552)] = 284759,
  [SMALL_STATE(5553)] = 284786,
  [SMALL_STATE(5554)] = 284811,
  [SMALL_STATE(5555)] = 284836,
  [SMALL_STATE(5556)] = 284863,
  [SMALL_STATE(5557)] = 284883,
  [SMALL_STATE(5558)] = 284907,
  [SMALL_STATE(5559)] = 284927,
  [SMALL_STATE(5560)] = 284951,
  [SMALL_STATE(5561)] = 284975,
  [SMALL_STATE(5562)] = 284999,
  [SMALL_STATE(5563)] = 285023,
  [SMALL_STATE(5564)] = 285047,
  [SMALL_STATE(5565)] = 285071,
  [SMALL_STATE(5566)] = 285095,
  [SMALL_STATE(5567)] = 285119,
  [SMALL_STATE(5568)] = 285139,
  [SMALL_STATE(5569)] = 285163,
  [SMALL_STATE(5570)] = 285183,
  [SMALL_STATE(5571)] = 285207,
  [SMALL_STATE(5572)] = 285231,
  [SMALL_STATE(5573)] = 285251,
  [SMALL_STATE(5574)] = 285275,
  [SMALL_STATE(5575)] = 285299,
  [SMALL_STATE(5576)] = 285319,
  [SMALL_STATE(5577)] = 285343,
  [SMALL_STATE(5578)] = 285367,
  [SMALL_STATE(5579)] = 285391,
  [SMALL_STATE(5580)] = 285411,
  [SMALL_STATE(5581)] = 285435,
  [SMALL_STATE(5582)] = 285459,
  [SMALL_STATE(5583)] = 285483,
  [SMALL_STATE(5584)] = 285507,
  [SMALL_STATE(5585)] = 285531,
  [SMALL_STATE(5586)] = 285555,
  [SMALL_STATE(5587)] = 285579,
  [SMALL_STATE(5588)] = 285603,
  [SMALL_STATE(5589)] = 285627,
  [SMALL_STATE(5590)] = 285651,
  [SMALL_STATE(5591)] = 285675,
  [SMALL_STATE(5592)] = 285699,
  [SMALL_STATE(5593)] = 285721,
  [SMALL_STATE(5594)] = 285745,
  [SMALL_STATE(5595)] = 285769,
  [SMALL_STATE(5596)] = 285793,
  [SMALL_STATE(5597)] = 285817,
  [SMALL_STATE(5598)] = 285841,
  [SMALL_STATE(5599)] = 285865,
  [SMALL_STATE(5600)] = 285889,
  [SMALL_STATE(5601)] = 285911,
  [SMALL_STATE(5602)] = 285935,
  [SMALL_STATE(5603)] = 285959,
  [SMALL_STATE(5604)] = 285983,
  [SMALL_STATE(5605)] = 286007,
  [SMALL_STATE(5606)] = 286031,
  [SMALL_STATE(5607)] = 286055,
  [SMALL_STATE(5608)] = 286077,
  [SMALL_STATE(5609)] = 286101,
  [SMALL_STATE(5610)] = 286125,
  [SMALL_STATE(5611)] = 286149,
  [SMALL_STATE(5612)] = 286169,
  [SMALL_STATE(5613)] = 286193,
  [SMALL_STATE(5614)] = 286217,
  [SMALL_STATE(5615)] = 286241,
  [SMALL_STATE(5616)] = 286265,
  [SMALL_STATE(5617)] = 286289,
  [SMALL_STATE(5618)] = 286313,
  [SMALL_STATE(5619)] = 286337,
  [SMALL_STATE(5620)] = 286361,
  [SMALL_STATE(5621)] = 286385,
  [SMALL_STATE(5622)] = 286407,
  [SMALL_STATE(5623)] = 286431,
  [SMALL_STATE(5624)] = 286455,
  [SMALL_STATE(5625)] = 286479,
  [SMALL_STATE(5626)] = 286499,
  [SMALL_STATE(5627)] = 286523,
  [SMALL_STATE(5628)] = 286547,
  [SMALL_STATE(5629)] = 286571,
  [SMALL_STATE(5630)] = 286595,
  [SMALL_STATE(5631)] = 286619,
  [SMALL_STATE(5632)] = 286643,
  [SMALL_STATE(5633)] = 286667,
  [SMALL_STATE(5634)] = 286691,
  [SMALL_STATE(5635)] = 286715,
  [SMALL_STATE(5636)] = 286735,
  [SMALL_STATE(5637)] = 286759,
  [SMALL_STATE(5638)] = 286783,
  [SMALL_STATE(5639)] = 286807,
  [SMALL_STATE(5640)] = 286831,
  [SMALL_STATE(5641)] = 286855,
  [SMALL_STATE(5642)] = 286879,
  [SMALL_STATE(5643)] = 286903,
  [SMALL_STATE(5644)] = 286927,
  [SMALL_STATE(5645)] = 286951,
  [SMALL_STATE(5646)] = 286975,
  [SMALL_STATE(5647)] = 286999,
  [SMALL_STATE(5648)] = 287023,
  [SMALL_STATE(5649)] = 287043,
  [SMALL_STATE(5650)] = 287067,
  [SMALL_STATE(5651)] = 287091,
  [SMALL_STATE(5652)] = 287115,
  [SMALL_STATE(5653)] = 287139,
  [SMALL_STATE(5654)] = 287163,
  [SMALL_STATE(5655)] = 287183,
  [SMALL_STATE(5656)] = 287207,
  [SMALL_STATE(5657)] = 287231,
  [SMALL_STATE(5658)] = 287255,
  [SMALL_STATE(5659)] = 287279,
  [SMALL_STATE(5660)] = 287303,
  [SMALL_STATE(5661)] = 287327,
  [SMALL_STATE(5662)] = 287351,
  [SMALL_STATE(5663)] = 287375,
  [SMALL_STATE(5664)] = 287399,
  [SMALL_STATE(5665)] = 287423,
  [SMALL_STATE(5666)] = 287447,
  [SMALL_STATE(5667)] = 287471,
  [SMALL_STATE(5668)] = 287495,
  [SMALL_STATE(5669)] = 287519,
  [SMALL_STATE(5670)] = 287543,
  [SMALL_STATE(5671)] = 287567,
  [SMALL_STATE(5672)] = 287589,
  [SMALL_STATE(5673)] = 287613,
  [SMALL_STATE(5674)] = 287637,
  [SMALL_STATE(5675)] = 287661,
  [SMALL_STATE(5676)] = 287681,
  [SMALL_STATE(5677)] = 287705,
  [SMALL_STATE(5678)] = 287725,
  [SMALL_STATE(5679)] = 287749,
  [SMALL_STATE(5680)] = 287773,
  [SMALL_STATE(5681)] = 287797,
  [SMALL_STATE(5682)] = 287821,
  [SMALL_STATE(5683)] = 287845,
  [SMALL_STATE(5684)] = 287867,
  [SMALL_STATE(5685)] = 287891,
  [SMALL_STATE(5686)] = 287911,
  [SMALL_STATE(5687)] = 287935,
  [SMALL_STATE(5688)] = 287959,
  [SMALL_STATE(5689)] = 287983,
  [SMALL_STATE(5690)] = 288007,
  [SMALL_STATE(5691)] = 288029,
  [SMALL_STATE(5692)] = 288053,
  [SMALL_STATE(5693)] = 288077,
  [SMALL_STATE(5694)] = 288101,
  [SMALL_STATE(5695)] = 288121,
  [SMALL_STATE(5696)] = 288145,
  [SMALL_STATE(5697)] = 288167,
  [SMALL_STATE(5698)] = 288191,
  [SMALL_STATE(5699)] = 288215,
  [SMALL_STATE(5700)] = 288239,
  [SMALL_STATE(5701)] = 288263,
  [SMALL_STATE(5702)] = 288287,
  [SMALL_STATE(5703)] = 288311,
  [SMALL_STATE(5704)] = 288335,
  [SMALL_STATE(5705)] = 288355,
  [SMALL_STATE(5706)] = 288379,
  [SMALL_STATE(5707)] = 288403,
  [SMALL_STATE(5708)] = 288423,
  [SMALL_STATE(5709)] = 288447,
  [SMALL_STATE(5710)] = 288471,
  [SMALL_STATE(5711)] = 288495,
  [SMALL_STATE(5712)] = 288519,
  [SMALL_STATE(5713)] = 288543,
  [SMALL_STATE(5714)] = 288567,
  [SMALL_STATE(5715)] = 288587,
  [SMALL_STATE(5716)] = 288611,
  [SMALL_STATE(5717)] = 288635,
  [SMALL_STATE(5718)] = 288657,
  [SMALL_STATE(5719)] = 288681,
  [SMALL_STATE(5720)] = 288705,
  [SMALL_STATE(5721)] = 288729,
  [SMALL_STATE(5722)] = 288751,
  [SMALL_STATE(5723)] = 288773,
  [SMALL_STATE(5724)] = 288797,
  [SMALL_STATE(5725)] = 288821,
  [SMALL_STATE(5726)] = 288845,
  [SMALL_STATE(5727)] = 288869,
  [SMALL_STATE(5728)] = 288893,
  [SMALL_STATE(5729)] = 288917,
  [SMALL_STATE(5730)] = 288941,
  [SMALL_STATE(5731)] = 288961,
  [SMALL_STATE(5732)] = 288985,
  [SMALL_STATE(5733)] = 289009,
  [SMALL_STATE(5734)] = 289033,
  [SMALL_STATE(5735)] = 289057,
  [SMALL_STATE(5736)] = 289081,
  [SMALL_STATE(5737)] = 289105,
  [SMALL_STATE(5738)] = 289129,
  [SMALL_STATE(5739)] = 289153,
  [SMALL_STATE(5740)] = 289177,
  [SMALL_STATE(5741)] = 289201,
  [SMALL_STATE(5742)] = 289225,
  [SMALL_STATE(5743)] = 289249,
  [SMALL_STATE(5744)] = 289271,
  [SMALL_STATE(5745)] = 289293,
  [SMALL_STATE(5746)] = 289315,
  [SMALL_STATE(5747)] = 289335,
  [SMALL_STATE(5748)] = 289359,
  [SMALL_STATE(5749)] = 289379,
  [SMALL_STATE(5750)] = 289403,
  [SMALL_STATE(5751)] = 289425,
  [SMALL_STATE(5752)] = 289445,
  [SMALL_STATE(5753)] = 289467,
  [SMALL_STATE(5754)] = 289491,
  [SMALL_STATE(5755)] = 289515,
  [SMALL_STATE(5756)] = 289537,
  [SMALL_STATE(5757)] = 289557,
  [SMALL_STATE(5758)] = 289577,
  [SMALL_STATE(5759)] = 289601,
  [SMALL_STATE(5760)] = 289625,
  [SMALL_STATE(5761)] = 289649,
  [SMALL_STATE(5762)] = 289669,
  [SMALL_STATE(5763)] = 289693,
  [SMALL_STATE(5764)] = 289713,
  [SMALL_STATE(5765)] = 289737,
  [SMALL_STATE(5766)] = 289761,
  [SMALL_STATE(5767)] = 289781,
  [SMALL_STATE(5768)] = 289803,
  [SMALL_STATE(5769)] = 289823,
  [SMALL_STATE(5770)] = 289847,
  [SMALL_STATE(5771)] = 289871,
  [SMALL_STATE(5772)] = 289895,
  [SMALL_STATE(5773)] = 289919,
  [SMALL_STATE(5774)] = 289941,
  [SMALL_STATE(5775)] = 289961,
  [SMALL_STATE(5776)] = 289983,
  [SMALL_STATE(5777)] = 290007,
  [SMALL_STATE(5778)] = 290027,
  [SMALL_STATE(5779)] = 290047,
  [SMALL_STATE(5780)] = 290069,
  [SMALL_STATE(5781)] = 290091,
  [SMALL_STATE(5782)] = 290115,
  [SMALL_STATE(5783)] = 290139,
  [SMALL_STATE(5784)] = 290163,
  [SMALL_STATE(5785)] = 290187,
  [SMALL_STATE(5786)] = 290211,
  [SMALL_STATE(5787)] = 290231,
  [SMALL_STATE(5788)] = 290255,
  [SMALL_STATE(5789)] = 290275,
  [SMALL_STATE(5790)] = 290297,
  [SMALL_STATE(5791)] = 290321,
  [SMALL_STATE(5792)] = 290345,
  [SMALL_STATE(5793)] = 290369,
  [SMALL_STATE(5794)] = 290391,
  [SMALL_STATE(5795)] = 290415,
  [SMALL_STATE(5796)] = 290435,
  [SMALL_STATE(5797)] = 290459,
  [SMALL_STATE(5798)] = 290481,
  [SMALL_STATE(5799)] = 290501,
  [SMALL_STATE(5800)] = 290525,
  [SMALL_STATE(5801)] = 290545,
  [SMALL_STATE(5802)] = 290569,
  [SMALL_STATE(5803)] = 290589,
  [SMALL_STATE(5804)] = 290609,
  [SMALL_STATE(5805)] = 290629,
  [SMALL_STATE(5806)] = 290653,
  [SMALL_STATE(5807)] = 290677,
  [SMALL_STATE(5808)] = 290697,
  [SMALL_STATE(5809)] = 290717,
  [SMALL_STATE(5810)] = 290741,
  [SMALL_STATE(5811)] = 290765,
  [SMALL_STATE(5812)] = 290785,
  [SMALL_STATE(5813)] = 290809,
  [SMALL_STATE(5814)] = 290833,
  [SMALL_STATE(5815)] = 290857,
  [SMALL_STATE(5816)] = 290881,
  [SMALL_STATE(5817)] = 290901,
  [SMALL_STATE(5818)] = 290925,
  [SMALL_STATE(5819)] = 290949,
  [SMALL_STATE(5820)] = 290969,
  [SMALL_STATE(5821)] = 290993,
  [SMALL_STATE(5822)] = 291013,
  [SMALL_STATE(5823)] = 291037,
  [SMALL_STATE(5824)] = 291057,
  [SMALL_STATE(5825)] = 291077,
  [SMALL_STATE(5826)] = 291101,
  [SMALL_STATE(5827)] = 291123,
  [SMALL_STATE(5828)] = 291147,
  [SMALL_STATE(5829)] = 291171,
  [SMALL_STATE(5830)] = 291195,
  [SMALL_STATE(5831)] = 291219,
  [SMALL_STATE(5832)] = 291243,
  [SMALL_STATE(5833)] = 291267,
  [SMALL_STATE(5834)] = 291289,
  [SMALL_STATE(5835)] = 291309,
  [SMALL_STATE(5836)] = 291333,
  [SMALL_STATE(5837)] = 291357,
  [SMALL_STATE(5838)] = 291381,
  [SMALL_STATE(5839)] = 291405,
  [SMALL_STATE(5840)] = 291429,
  [SMALL_STATE(5841)] = 291453,
  [SMALL_STATE(5842)] = 291477,
  [SMALL_STATE(5843)] = 291501,
  [SMALL_STATE(5844)] = 291521,
  [SMALL_STATE(5845)] = 291545,
  [SMALL_STATE(5846)] = 291569,
  [SMALL_STATE(5847)] = 291593,
  [SMALL_STATE(5848)] = 291617,
  [SMALL_STATE(5849)] = 291641,
  [SMALL_STATE(5850)] = 291665,
  [SMALL_STATE(5851)] = 291689,
  [SMALL_STATE(5852)] = 291711,
  [SMALL_STATE(5853)] = 291731,
  [SMALL_STATE(5854)] = 291755,
  [SMALL_STATE(5855)] = 291779,
  [SMALL_STATE(5856)] = 291803,
  [SMALL_STATE(5857)] = 291827,
  [SMALL_STATE(5858)] = 291851,
  [SMALL_STATE(5859)] = 291875,
  [SMALL_STATE(5860)] = 291899,
  [SMALL_STATE(5861)] = 291923,
  [SMALL_STATE(5862)] = 291947,
  [SMALL_STATE(5863)] = 291971,
  [SMALL_STATE(5864)] = 291991,
  [SMALL_STATE(5865)] = 292015,
  [SMALL_STATE(5866)] = 292039,
  [SMALL_STATE(5867)] = 292063,
  [SMALL_STATE(5868)] = 292085,
  [SMALL_STATE(5869)] = 292105,
  [SMALL_STATE(5870)] = 292129,
  [SMALL_STATE(5871)] = 292151,
  [SMALL_STATE(5872)] = 292175,
  [SMALL_STATE(5873)] = 292199,
  [SMALL_STATE(5874)] = 292223,
  [SMALL_STATE(5875)] = 292247,
  [SMALL_STATE(5876)] = 292271,
  [SMALL_STATE(5877)] = 292295,
  [SMALL_STATE(5878)] = 292319,
  [SMALL_STATE(5879)] = 292343,
  [SMALL_STATE(5880)] = 292367,
  [SMALL_STATE(5881)] = 292387,
  [SMALL_STATE(5882)] = 292411,
  [SMALL_STATE(5883)] = 292431,
  [SMALL_STATE(5884)] = 292455,
  [SMALL_STATE(5885)] = 292479,
  [SMALL_STATE(5886)] = 292503,
  [SMALL_STATE(5887)] = 292527,
  [SMALL_STATE(5888)] = 292551,
  [SMALL_STATE(5889)] = 292575,
  [SMALL_STATE(5890)] = 292595,
  [SMALL_STATE(5891)] = 292619,
  [SMALL_STATE(5892)] = 292643,
  [SMALL_STATE(5893)] = 292667,
  [SMALL_STATE(5894)] = 292691,
  [SMALL_STATE(5895)] = 292715,
  [SMALL_STATE(5896)] = 292739,
  [SMALL_STATE(5897)] = 292761,
  [SMALL_STATE(5898)] = 292785,
  [SMALL_STATE(5899)] = 292809,
  [SMALL_STATE(5900)] = 292833,
  [SMALL_STATE(5901)] = 292853,
  [SMALL_STATE(5902)] = 292877,
  [SMALL_STATE(5903)] = 292901,
  [SMALL_STATE(5904)] = 292925,
  [SMALL_STATE(5905)] = 292949,
  [SMALL_STATE(5906)] = 292973,
  [SMALL_STATE(5907)] = 292997,
  [SMALL_STATE(5908)] = 293021,
  [SMALL_STATE(5909)] = 293045,
  [SMALL_STATE(5910)] = 293069,
  [SMALL_STATE(5911)] = 293093,
  [SMALL_STATE(5912)] = 293117,
  [SMALL_STATE(5913)] = 293141,
  [SMALL_STATE(5914)] = 293165,
  [SMALL_STATE(5915)] = 293187,
  [SMALL_STATE(5916)] = 293211,
  [SMALL_STATE(5917)] = 293235,
  [SMALL_STATE(5918)] = 293259,
  [SMALL_STATE(5919)] = 293279,
  [SMALL_STATE(5920)] = 293303,
  [SMALL_STATE(5921)] = 293327,
  [SMALL_STATE(5922)] = 293351,
  [SMALL_STATE(5923)] = 293375,
  [SMALL_STATE(5924)] = 293399,
  [SMALL_STATE(5925)] = 293419,
  [SMALL_STATE(5926)] = 293443,
  [SMALL_STATE(5927)] = 293463,
  [SMALL_STATE(5928)] = 293487,
  [SMALL_STATE(5929)] = 293511,
  [SMALL_STATE(5930)] = 293533,
  [SMALL_STATE(5931)] = 293557,
  [SMALL_STATE(5932)] = 293581,
  [SMALL_STATE(5933)] = 293605,
  [SMALL_STATE(5934)] = 293629,
  [SMALL_STATE(5935)] = 293653,
  [SMALL_STATE(5936)] = 293677,
  [SMALL_STATE(5937)] = 293701,
  [SMALL_STATE(5938)] = 293725,
  [SMALL_STATE(5939)] = 293749,
  [SMALL_STATE(5940)] = 293773,
  [SMALL_STATE(5941)] = 293797,
  [SMALL_STATE(5942)] = 293821,
  [SMALL_STATE(5943)] = 293845,
  [SMALL_STATE(5944)] = 293869,
  [SMALL_STATE(5945)] = 293893,
  [SMALL_STATE(5946)] = 293917,
  [SMALL_STATE(5947)] = 293941,
  [SMALL_STATE(5948)] = 293965,
  [SMALL_STATE(5949)] = 293989,
  [SMALL_STATE(5950)] = 294013,
  [SMALL_STATE(5951)] = 294033,
  [SMALL_STATE(5952)] = 294057,
  [SMALL_STATE(5953)] = 294081,
  [SMALL_STATE(5954)] = 294105,
  [SMALL_STATE(5955)] = 294129,
  [SMALL_STATE(5956)] = 294153,
  [SMALL_STATE(5957)] = 294177,
  [SMALL_STATE(5958)] = 294201,
  [SMALL_STATE(5959)] = 294225,
  [SMALL_STATE(5960)] = 294249,
  [SMALL_STATE(5961)] = 294273,
  [SMALL_STATE(5962)] = 294297,
  [SMALL_STATE(5963)] = 294321,
  [SMALL_STATE(5964)] = 294345,
  [SMALL_STATE(5965)] = 294369,
  [SMALL_STATE(5966)] = 294393,
  [SMALL_STATE(5967)] = 294417,
  [SMALL_STATE(5968)] = 294441,
  [SMALL_STATE(5969)] = 294465,
  [SMALL_STATE(5970)] = 294489,
  [SMALL_STATE(5971)] = 294513,
  [SMALL_STATE(5972)] = 294537,
  [SMALL_STATE(5973)] = 294561,
  [SMALL_STATE(5974)] = 294585,
  [SMALL_STATE(5975)] = 294609,
  [SMALL_STATE(5976)] = 294633,
  [SMALL_STATE(5977)] = 294657,
  [SMALL_STATE(5978)] = 294681,
  [SMALL_STATE(5979)] = 294705,
  [SMALL_STATE(5980)] = 294729,
  [SMALL_STATE(5981)] = 294753,
  [SMALL_STATE(5982)] = 294777,
  [SMALL_STATE(5983)] = 294801,
  [SMALL_STATE(5984)] = 294825,
  [SMALL_STATE(5985)] = 294849,
  [SMALL_STATE(5986)] = 294873,
  [SMALL_STATE(5987)] = 294897,
  [SMALL_STATE(5988)] = 294921,
  [SMALL_STATE(5989)] = 294945,
  [SMALL_STATE(5990)] = 294969,
  [SMALL_STATE(5991)] = 294993,
  [SMALL_STATE(5992)] = 295017,
  [SMALL_STATE(5993)] = 295041,
  [SMALL_STATE(5994)] = 295065,
  [SMALL_STATE(5995)] = 295089,
  [SMALL_STATE(5996)] = 295113,
  [SMALL_STATE(5997)] = 295137,
  [SMALL_STATE(5998)] = 295161,
  [SMALL_STATE(5999)] = 295185,
  [SMALL_STATE(6000)] = 295209,
  [SMALL_STATE(6001)] = 295233,
  [SMALL_STATE(6002)] = 295257,
  [SMALL_STATE(6003)] = 295281,
  [SMALL_STATE(6004)] = 295305,
  [SMALL_STATE(6005)] = 295329,
  [SMALL_STATE(6006)] = 295351,
  [SMALL_STATE(6007)] = 295372,
  [SMALL_STATE(6008)] = 295393,
  [SMALL_STATE(6009)] = 295412,
  [SMALL_STATE(6010)] = 295433,
  [SMALL_STATE(6011)] = 295452,
  [SMALL_STATE(6012)] = 295471,
  [SMALL_STATE(6013)] = 295492,
  [SMALL_STATE(6014)] = 295511,
  [SMALL_STATE(6015)] = 295532,
  [SMALL_STATE(6016)] = 295551,
  [SMALL_STATE(6017)] = 295572,
  [SMALL_STATE(6018)] = 295593,
  [SMALL_STATE(6019)] = 295614,
  [SMALL_STATE(6020)] = 295635,
  [SMALL_STATE(6021)] = 295656,
  [SMALL_STATE(6022)] = 295675,
  [SMALL_STATE(6023)] = 295696,
  [SMALL_STATE(6024)] = 295715,
  [SMALL_STATE(6025)] = 295736,
  [SMALL_STATE(6026)] = 295755,
  [SMALL_STATE(6027)] = 295776,
  [SMALL_STATE(6028)] = 295797,
  [SMALL_STATE(6029)] = 295816,
  [SMALL_STATE(6030)] = 295837,
  [SMALL_STATE(6031)] = 295858,
  [SMALL_STATE(6032)] = 295877,
  [SMALL_STATE(6033)] = 295898,
  [SMALL_STATE(6034)] = 295917,
  [SMALL_STATE(6035)] = 295938,
  [SMALL_STATE(6036)] = 295957,
  [SMALL_STATE(6037)] = 295976,
  [SMALL_STATE(6038)] = 295997,
  [SMALL_STATE(6039)] = 296016,
  [SMALL_STATE(6040)] = 296035,
  [SMALL_STATE(6041)] = 296056,
  [SMALL_STATE(6042)] = 296077,
  [SMALL_STATE(6043)] = 296098,
  [SMALL_STATE(6044)] = 296119,
  [SMALL_STATE(6045)] = 296140,
  [SMALL_STATE(6046)] = 296159,
  [SMALL_STATE(6047)] = 296180,
  [SMALL_STATE(6048)] = 296201,
  [SMALL_STATE(6049)] = 296220,
  [SMALL_STATE(6050)] = 296241,
  [SMALL_STATE(6051)] = 296262,
  [SMALL_STATE(6052)] = 296283,
  [SMALL_STATE(6053)] = 296304,
  [SMALL_STATE(6054)] = 296325,
  [SMALL_STATE(6055)] = 296346,
  [SMALL_STATE(6056)] = 296365,
  [SMALL_STATE(6057)] = 296386,
  [SMALL_STATE(6058)] = 296407,
  [SMALL_STATE(6059)] = 296428,
  [SMALL_STATE(6060)] = 296449,
  [SMALL_STATE(6061)] = 296468,
  [SMALL_STATE(6062)] = 296489,
  [SMALL_STATE(6063)] = 296508,
  [SMALL_STATE(6064)] = 296527,
  [SMALL_STATE(6065)] = 296548,
  [SMALL_STATE(6066)] = 296569,
  [SMALL_STATE(6067)] = 296590,
  [SMALL_STATE(6068)] = 296609,
  [SMALL_STATE(6069)] = 296630,
  [SMALL_STATE(6070)] = 296649,
  [SMALL_STATE(6071)] = 296670,
  [SMALL_STATE(6072)] = 296691,
  [SMALL_STATE(6073)] = 296710,
  [SMALL_STATE(6074)] = 296731,
  [SMALL_STATE(6075)] = 296752,
  [SMALL_STATE(6076)] = 296773,
  [SMALL_STATE(6077)] = 296792,
  [SMALL_STATE(6078)] = 296811,
  [SMALL_STATE(6079)] = 296832,
  [SMALL_STATE(6080)] = 296853,
  [SMALL_STATE(6081)] = 296874,
  [SMALL_STATE(6082)] = 296893,
  [SMALL_STATE(6083)] = 296912,
  [SMALL_STATE(6084)] = 296931,
  [SMALL_STATE(6085)] = 296952,
  [SMALL_STATE(6086)] = 296971,
  [SMALL_STATE(6087)] = 296990,
  [SMALL_STATE(6088)] = 297011,
  [SMALL_STATE(6089)] = 297032,
  [SMALL_STATE(6090)] = 297053,
  [SMALL_STATE(6091)] = 297074,
  [SMALL_STATE(6092)] = 297095,
  [SMALL_STATE(6093)] = 297116,
  [SMALL_STATE(6094)] = 297137,
  [SMALL_STATE(6095)] = 297158,
  [SMALL_STATE(6096)] = 297177,
  [SMALL_STATE(6097)] = 297196,
  [SMALL_STATE(6098)] = 297217,
  [SMALL_STATE(6099)] = 297236,
  [SMALL_STATE(6100)] = 297255,
  [SMALL_STATE(6101)] = 297276,
  [SMALL_STATE(6102)] = 297297,
  [SMALL_STATE(6103)] = 297316,
  [SMALL_STATE(6104)] = 297337,
  [SMALL_STATE(6105)] = 297358,
  [SMALL_STATE(6106)] = 297379,
  [SMALL_STATE(6107)] = 297398,
  [SMALL_STATE(6108)] = 297419,
  [SMALL_STATE(6109)] = 297440,
  [SMALL_STATE(6110)] = 297461,
  [SMALL_STATE(6111)] = 297482,
  [SMALL_STATE(6112)] = 297503,
  [SMALL_STATE(6113)] = 297524,
  [SMALL_STATE(6114)] = 297545,
  [SMALL_STATE(6115)] = 297564,
  [SMALL_STATE(6116)] = 297585,
  [SMALL_STATE(6117)] = 297606,
  [SMALL_STATE(6118)] = 297627,
  [SMALL_STATE(6119)] = 297648,
  [SMALL_STATE(6120)] = 297669,
  [SMALL_STATE(6121)] = 297690,
  [SMALL_STATE(6122)] = 297711,
  [SMALL_STATE(6123)] = 297732,
  [SMALL_STATE(6124)] = 297751,
  [SMALL_STATE(6125)] = 297772,
  [SMALL_STATE(6126)] = 297793,
  [SMALL_STATE(6127)] = 297812,
  [SMALL_STATE(6128)] = 297833,
  [SMALL_STATE(6129)] = 297852,
  [SMALL_STATE(6130)] = 297871,
  [SMALL_STATE(6131)] = 297892,
  [SMALL_STATE(6132)] = 297911,
  [SMALL_STATE(6133)] = 297932,
  [SMALL_STATE(6134)] = 297953,
  [SMALL_STATE(6135)] = 297974,
  [SMALL_STATE(6136)] = 297995,
  [SMALL_STATE(6137)] = 298016,
  [SMALL_STATE(6138)] = 298037,
  [SMALL_STATE(6139)] = 298056,
  [SMALL_STATE(6140)] = 298077,
  [SMALL_STATE(6141)] = 298098,
  [SMALL_STATE(6142)] = 298119,
  [SMALL_STATE(6143)] = 298140,
  [SMALL_STATE(6144)] = 298161,
  [SMALL_STATE(6145)] = 298182,
  [SMALL_STATE(6146)] = 298203,
  [SMALL_STATE(6147)] = 298222,
  [SMALL_STATE(6148)] = 298243,
  [SMALL_STATE(6149)] = 298264,
  [SMALL_STATE(6150)] = 298283,
  [SMALL_STATE(6151)] = 298304,
  [SMALL_STATE(6152)] = 298325,
  [SMALL_STATE(6153)] = 298344,
  [SMALL_STATE(6154)] = 298363,
  [SMALL_STATE(6155)] = 298384,
  [SMALL_STATE(6156)] = 298405,
  [SMALL_STATE(6157)] = 298424,
  [SMALL_STATE(6158)] = 298445,
  [SMALL_STATE(6159)] = 298466,
  [SMALL_STATE(6160)] = 298487,
  [SMALL_STATE(6161)] = 298508,
  [SMALL_STATE(6162)] = 298529,
  [SMALL_STATE(6163)] = 298550,
  [SMALL_STATE(6164)] = 298571,
  [SMALL_STATE(6165)] = 298592,
  [SMALL_STATE(6166)] = 298613,
  [SMALL_STATE(6167)] = 298632,
  [SMALL_STATE(6168)] = 298651,
  [SMALL_STATE(6169)] = 298670,
  [SMALL_STATE(6170)] = 298689,
  [SMALL_STATE(6171)] = 298710,
  [SMALL_STATE(6172)] = 298729,
  [SMALL_STATE(6173)] = 298750,
  [SMALL_STATE(6174)] = 298771,
  [SMALL_STATE(6175)] = 298792,
  [SMALL_STATE(6176)] = 298813,
  [SMALL_STATE(6177)] = 298834,
  [SMALL_STATE(6178)] = 298853,
  [SMALL_STATE(6179)] = 298872,
  [SMALL_STATE(6180)] = 298891,
  [SMALL_STATE(6181)] = 298910,
  [SMALL_STATE(6182)] = 298931,
  [SMALL_STATE(6183)] = 298950,
  [SMALL_STATE(6184)] = 298971,
  [SMALL_STATE(6185)] = 298990,
  [SMALL_STATE(6186)] = 299009,
  [SMALL_STATE(6187)] = 299028,
  [SMALL_STATE(6188)] = 299049,
  [SMALL_STATE(6189)] = 299068,
  [SMALL_STATE(6190)] = 299089,
  [SMALL_STATE(6191)] = 299110,
  [SMALL_STATE(6192)] = 299129,
  [SMALL_STATE(6193)] = 299148,
  [SMALL_STATE(6194)] = 299169,
  [SMALL_STATE(6195)] = 299190,
  [SMALL_STATE(6196)] = 299211,
  [SMALL_STATE(6197)] = 299230,
  [SMALL_STATE(6198)] = 299251,
  [SMALL_STATE(6199)] = 299272,
  [SMALL_STATE(6200)] = 299293,
  [SMALL_STATE(6201)] = 299314,
  [SMALL_STATE(6202)] = 299335,
  [SMALL_STATE(6203)] = 299354,
  [SMALL_STATE(6204)] = 299375,
  [SMALL_STATE(6205)] = 299394,
  [SMALL_STATE(6206)] = 299413,
  [SMALL_STATE(6207)] = 299432,
  [SMALL_STATE(6208)] = 299453,
  [SMALL_STATE(6209)] = 299472,
  [SMALL_STATE(6210)] = 299493,
  [SMALL_STATE(6211)] = 299514,
  [SMALL_STATE(6212)] = 299535,
  [SMALL_STATE(6213)] = 299554,
  [SMALL_STATE(6214)] = 299573,
  [SMALL_STATE(6215)] = 299594,
  [SMALL_STATE(6216)] = 299615,
  [SMALL_STATE(6217)] = 299636,
  [SMALL_STATE(6218)] = 299657,
  [SMALL_STATE(6219)] = 299678,
  [SMALL_STATE(6220)] = 299697,
  [SMALL_STATE(6221)] = 299718,
  [SMALL_STATE(6222)] = 299739,
  [SMALL_STATE(6223)] = 299758,
  [SMALL_STATE(6224)] = 299779,
  [SMALL_STATE(6225)] = 299798,
  [SMALL_STATE(6226)] = 299819,
  [SMALL_STATE(6227)] = 299840,
  [SMALL_STATE(6228)] = 299861,
  [SMALL_STATE(6229)] = 299882,
  [SMALL_STATE(6230)] = 299903,
  [SMALL_STATE(6231)] = 299924,
  [SMALL_STATE(6232)] = 299945,
  [SMALL_STATE(6233)] = 299966,
  [SMALL_STATE(6234)] = 299985,
  [SMALL_STATE(6235)] = 300006,
  [SMALL_STATE(6236)] = 300025,
  [SMALL_STATE(6237)] = 300044,
  [SMALL_STATE(6238)] = 300065,
  [SMALL_STATE(6239)] = 300084,
  [SMALL_STATE(6240)] = 300103,
  [SMALL_STATE(6241)] = 300124,
  [SMALL_STATE(6242)] = 300145,
  [SMALL_STATE(6243)] = 300164,
  [SMALL_STATE(6244)] = 300185,
  [SMALL_STATE(6245)] = 300206,
  [SMALL_STATE(6246)] = 300227,
  [SMALL_STATE(6247)] = 300248,
  [SMALL_STATE(6248)] = 300269,
  [SMALL_STATE(6249)] = 300290,
  [SMALL_STATE(6250)] = 300311,
  [SMALL_STATE(6251)] = 300332,
  [SMALL_STATE(6252)] = 300353,
  [SMALL_STATE(6253)] = 300374,
  [SMALL_STATE(6254)] = 300393,
  [SMALL_STATE(6255)] = 300414,
  [SMALL_STATE(6256)] = 300435,
  [SMALL_STATE(6257)] = 300454,
  [SMALL_STATE(6258)] = 300475,
  [SMALL_STATE(6259)] = 300494,
  [SMALL_STATE(6260)] = 300515,
  [SMALL_STATE(6261)] = 300534,
  [SMALL_STATE(6262)] = 300555,
  [SMALL_STATE(6263)] = 300576,
  [SMALL_STATE(6264)] = 300595,
  [SMALL_STATE(6265)] = 300614,
  [SMALL_STATE(6266)] = 300633,
  [SMALL_STATE(6267)] = 300654,
  [SMALL_STATE(6268)] = 300673,
  [SMALL_STATE(6269)] = 300694,
  [SMALL_STATE(6270)] = 300715,
  [SMALL_STATE(6271)] = 300734,
  [SMALL_STATE(6272)] = 300755,
  [SMALL_STATE(6273)] = 300776,
  [SMALL_STATE(6274)] = 300797,
  [SMALL_STATE(6275)] = 300818,
  [SMALL_STATE(6276)] = 300839,
  [SMALL_STATE(6277)] = 300860,
  [SMALL_STATE(6278)] = 300879,
  [SMALL_STATE(6279)] = 300900,
  [SMALL_STATE(6280)] = 300921,
  [SMALL_STATE(6281)] = 300940,
  [SMALL_STATE(6282)] = 300959,
  [SMALL_STATE(6283)] = 300980,
  [SMALL_STATE(6284)] = 300999,
  [SMALL_STATE(6285)] = 301018,
  [SMALL_STATE(6286)] = 301037,
  [SMALL_STATE(6287)] = 301058,
  [SMALL_STATE(6288)] = 301077,
  [SMALL_STATE(6289)] = 301096,
  [SMALL_STATE(6290)] = 301117,
  [SMALL_STATE(6291)] = 301136,
  [SMALL_STATE(6292)] = 301157,
  [SMALL_STATE(6293)] = 301178,
  [SMALL_STATE(6294)] = 301197,
  [SMALL_STATE(6295)] = 301218,
  [SMALL_STATE(6296)] = 301237,
  [SMALL_STATE(6297)] = 301258,
  [SMALL_STATE(6298)] = 301279,
  [SMALL_STATE(6299)] = 301300,
  [SMALL_STATE(6300)] = 301321,
  [SMALL_STATE(6301)] = 301342,
  [SMALL_STATE(6302)] = 301363,
  [SMALL_STATE(6303)] = 301384,
  [SMALL_STATE(6304)] = 301405,
  [SMALL_STATE(6305)] = 301426,
  [SMALL_STATE(6306)] = 301447,
  [SMALL_STATE(6307)] = 301466,
  [SMALL_STATE(6308)] = 301487,
  [SMALL_STATE(6309)] = 301508,
  [SMALL_STATE(6310)] = 301529,
  [SMALL_STATE(6311)] = 301550,
  [SMALL_STATE(6312)] = 301569,
  [SMALL_STATE(6313)] = 301588,
  [SMALL_STATE(6314)] = 301609,
  [SMALL_STATE(6315)] = 301630,
  [SMALL_STATE(6316)] = 301651,
  [SMALL_STATE(6317)] = 301672,
  [SMALL_STATE(6318)] = 301693,
  [SMALL_STATE(6319)] = 301714,
  [SMALL_STATE(6320)] = 301733,
  [SMALL_STATE(6321)] = 301754,
  [SMALL_STATE(6322)] = 301775,
  [SMALL_STATE(6323)] = 301796,
  [SMALL_STATE(6324)] = 301817,
  [SMALL_STATE(6325)] = 301838,
  [SMALL_STATE(6326)] = 301859,
  [SMALL_STATE(6327)] = 301880,
  [SMALL_STATE(6328)] = 301901,
  [SMALL_STATE(6329)] = 301920,
  [SMALL_STATE(6330)] = 301941,
  [SMALL_STATE(6331)] = 301962,
  [SMALL_STATE(6332)] = 301983,
  [SMALL_STATE(6333)] = 302004,
  [SMALL_STATE(6334)] = 302025,
  [SMALL_STATE(6335)] = 302046,
  [SMALL_STATE(6336)] = 302067,
  [SMALL_STATE(6337)] = 302086,
  [SMALL_STATE(6338)] = 302105,
  [SMALL_STATE(6339)] = 302126,
  [SMALL_STATE(6340)] = 302145,
  [SMALL_STATE(6341)] = 302166,
  [SMALL_STATE(6342)] = 302187,
  [SMALL_STATE(6343)] = 302208,
  [SMALL_STATE(6344)] = 302229,
  [SMALL_STATE(6345)] = 302248,
  [SMALL_STATE(6346)] = 302267,
  [SMALL_STATE(6347)] = 302286,
  [SMALL_STATE(6348)] = 302305,
  [SMALL_STATE(6349)] = 302326,
  [SMALL_STATE(6350)] = 302347,
  [SMALL_STATE(6351)] = 302368,
  [SMALL_STATE(6352)] = 302389,
  [SMALL_STATE(6353)] = 302410,
  [SMALL_STATE(6354)] = 302431,
  [SMALL_STATE(6355)] = 302452,
  [SMALL_STATE(6356)] = 302473,
  [SMALL_STATE(6357)] = 302492,
  [SMALL_STATE(6358)] = 302513,
  [SMALL_STATE(6359)] = 302534,
  [SMALL_STATE(6360)] = 302555,
  [SMALL_STATE(6361)] = 302576,
  [SMALL_STATE(6362)] = 302595,
  [SMALL_STATE(6363)] = 302616,
  [SMALL_STATE(6364)] = 302637,
  [SMALL_STATE(6365)] = 302658,
  [SMALL_STATE(6366)] = 302679,
  [SMALL_STATE(6367)] = 302700,
  [SMALL_STATE(6368)] = 302721,
  [SMALL_STATE(6369)] = 302742,
  [SMALL_STATE(6370)] = 302763,
  [SMALL_STATE(6371)] = 302784,
  [SMALL_STATE(6372)] = 302805,
  [SMALL_STATE(6373)] = 302824,
  [SMALL_STATE(6374)] = 302845,
  [SMALL_STATE(6375)] = 302866,
  [SMALL_STATE(6376)] = 302887,
  [SMALL_STATE(6377)] = 302908,
  [SMALL_STATE(6378)] = 302929,
  [SMALL_STATE(6379)] = 302950,
  [SMALL_STATE(6380)] = 302971,
  [SMALL_STATE(6381)] = 302992,
  [SMALL_STATE(6382)] = 303013,
  [SMALL_STATE(6383)] = 303034,
  [SMALL_STATE(6384)] = 303055,
  [SMALL_STATE(6385)] = 303076,
  [SMALL_STATE(6386)] = 303097,
  [SMALL_STATE(6387)] = 303116,
  [SMALL_STATE(6388)] = 303137,
  [SMALL_STATE(6389)] = 303158,
  [SMALL_STATE(6390)] = 303179,
  [SMALL_STATE(6391)] = 303200,
  [SMALL_STATE(6392)] = 303221,
  [SMALL_STATE(6393)] = 303242,
  [SMALL_STATE(6394)] = 303263,
  [SMALL_STATE(6395)] = 303284,
  [SMALL_STATE(6396)] = 303303,
  [SMALL_STATE(6397)] = 303324,
  [SMALL_STATE(6398)] = 303345,
  [SMALL_STATE(6399)] = 303364,
  [SMALL_STATE(6400)] = 303385,
  [SMALL_STATE(6401)] = 303406,
  [SMALL_STATE(6402)] = 303427,
  [SMALL_STATE(6403)] = 303448,
  [SMALL_STATE(6404)] = 303469,
  [SMALL_STATE(6405)] = 303490,
  [SMALL_STATE(6406)] = 303511,
  [SMALL_STATE(6407)] = 303532,
  [SMALL_STATE(6408)] = 303553,
  [SMALL_STATE(6409)] = 303574,
  [SMALL_STATE(6410)] = 303595,
  [SMALL_STATE(6411)] = 303616,
  [SMALL_STATE(6412)] = 303637,
  [SMALL_STATE(6413)] = 303658,
  [SMALL_STATE(6414)] = 303679,
  [SMALL_STATE(6415)] = 303700,
  [SMALL_STATE(6416)] = 303721,
  [SMALL_STATE(6417)] = 303742,
  [SMALL_STATE(6418)] = 303763,
  [SMALL_STATE(6419)] = 303784,
  [SMALL_STATE(6420)] = 303805,
  [SMALL_STATE(6421)] = 303826,
  [SMALL_STATE(6422)] = 303847,
  [SMALL_STATE(6423)] = 303868,
  [SMALL_STATE(6424)] = 303889,
  [SMALL_STATE(6425)] = 303910,
  [SMALL_STATE(6426)] = 303931,
  [SMALL_STATE(6427)] = 303952,
  [SMALL_STATE(6428)] = 303973,
  [SMALL_STATE(6429)] = 303994,
  [SMALL_STATE(6430)] = 304015,
  [SMALL_STATE(6431)] = 304036,
  [SMALL_STATE(6432)] = 304057,
  [SMALL_STATE(6433)] = 304078,
  [SMALL_STATE(6434)] = 304099,
  [SMALL_STATE(6435)] = 304120,
  [SMALL_STATE(6436)] = 304141,
  [SMALL_STATE(6437)] = 304162,
  [SMALL_STATE(6438)] = 304183,
  [SMALL_STATE(6439)] = 304204,
  [SMALL_STATE(6440)] = 304225,
  [SMALL_STATE(6441)] = 304246,
  [SMALL_STATE(6442)] = 304267,
  [SMALL_STATE(6443)] = 304288,
  [SMALL_STATE(6444)] = 304309,
  [SMALL_STATE(6445)] = 304330,
  [SMALL_STATE(6446)] = 304351,
  [SMALL_STATE(6447)] = 304372,
  [SMALL_STATE(6448)] = 304393,
  [SMALL_STATE(6449)] = 304414,
  [SMALL_STATE(6450)] = 304435,
  [SMALL_STATE(6451)] = 304456,
  [SMALL_STATE(6452)] = 304477,
  [SMALL_STATE(6453)] = 304498,
  [SMALL_STATE(6454)] = 304519,
  [SMALL_STATE(6455)] = 304540,
  [SMALL_STATE(6456)] = 304561,
  [SMALL_STATE(6457)] = 304582,
  [SMALL_STATE(6458)] = 304603,
  [SMALL_STATE(6459)] = 304624,
  [SMALL_STATE(6460)] = 304645,
  [SMALL_STATE(6461)] = 304666,
  [SMALL_STATE(6462)] = 304687,
  [SMALL_STATE(6463)] = 304708,
  [SMALL_STATE(6464)] = 304727,
  [SMALL_STATE(6465)] = 304748,
  [SMALL_STATE(6466)] = 304769,
  [SMALL_STATE(6467)] = 304790,
  [SMALL_STATE(6468)] = 304811,
  [SMALL_STATE(6469)] = 304832,
  [SMALL_STATE(6470)] = 304853,
  [SMALL_STATE(6471)] = 304874,
  [SMALL_STATE(6472)] = 304895,
  [SMALL_STATE(6473)] = 304916,
  [SMALL_STATE(6474)] = 304935,
  [SMALL_STATE(6475)] = 304956,
  [SMALL_STATE(6476)] = 304977,
  [SMALL_STATE(6477)] = 304998,
  [SMALL_STATE(6478)] = 305017,
  [SMALL_STATE(6479)] = 305036,
  [SMALL_STATE(6480)] = 305057,
  [SMALL_STATE(6481)] = 305078,
  [SMALL_STATE(6482)] = 305099,
  [SMALL_STATE(6483)] = 305120,
  [SMALL_STATE(6484)] = 305141,
  [SMALL_STATE(6485)] = 305160,
  [SMALL_STATE(6486)] = 305181,
  [SMALL_STATE(6487)] = 305202,
  [SMALL_STATE(6488)] = 305223,
  [SMALL_STATE(6489)] = 305244,
  [SMALL_STATE(6490)] = 305265,
  [SMALL_STATE(6491)] = 305286,
  [SMALL_STATE(6492)] = 305307,
  [SMALL_STATE(6493)] = 305326,
  [SMALL_STATE(6494)] = 305347,
  [SMALL_STATE(6495)] = 305368,
  [SMALL_STATE(6496)] = 305389,
  [SMALL_STATE(6497)] = 305410,
  [SMALL_STATE(6498)] = 305431,
  [SMALL_STATE(6499)] = 305452,
  [SMALL_STATE(6500)] = 305473,
  [SMALL_STATE(6501)] = 305494,
  [SMALL_STATE(6502)] = 305515,
  [SMALL_STATE(6503)] = 305534,
  [SMALL_STATE(6504)] = 305553,
  [SMALL_STATE(6505)] = 305574,
  [SMALL_STATE(6506)] = 305595,
  [SMALL_STATE(6507)] = 305616,
  [SMALL_STATE(6508)] = 305637,
  [SMALL_STATE(6509)] = 305656,
  [SMALL_STATE(6510)] = 305677,
  [SMALL_STATE(6511)] = 305698,
  [SMALL_STATE(6512)] = 305717,
  [SMALL_STATE(6513)] = 305738,
  [SMALL_STATE(6514)] = 305759,
  [SMALL_STATE(6515)] = 305780,
  [SMALL_STATE(6516)] = 305801,
  [SMALL_STATE(6517)] = 305820,
  [SMALL_STATE(6518)] = 305839,
  [SMALL_STATE(6519)] = 305858,
  [SMALL_STATE(6520)] = 305879,
  [SMALL_STATE(6521)] = 305900,
  [SMALL_STATE(6522)] = 305921,
  [SMALL_STATE(6523)] = 305940,
  [SMALL_STATE(6524)] = 305961,
  [SMALL_STATE(6525)] = 305982,
  [SMALL_STATE(6526)] = 306003,
  [SMALL_STATE(6527)] = 306024,
  [SMALL_STATE(6528)] = 306043,
  [SMALL_STATE(6529)] = 306064,
  [SMALL_STATE(6530)] = 306085,
  [SMALL_STATE(6531)] = 306104,
  [SMALL_STATE(6532)] = 306125,
  [SMALL_STATE(6533)] = 306146,
  [SMALL_STATE(6534)] = 306165,
  [SMALL_STATE(6535)] = 306184,
  [SMALL_STATE(6536)] = 306203,
  [SMALL_STATE(6537)] = 306222,
  [SMALL_STATE(6538)] = 306241,
  [SMALL_STATE(6539)] = 306262,
  [SMALL_STATE(6540)] = 306283,
  [SMALL_STATE(6541)] = 306304,
  [SMALL_STATE(6542)] = 306323,
  [SMALL_STATE(6543)] = 306342,
  [SMALL_STATE(6544)] = 306361,
  [SMALL_STATE(6545)] = 306382,
  [SMALL_STATE(6546)] = 306403,
  [SMALL_STATE(6547)] = 306424,
  [SMALL_STATE(6548)] = 306445,
  [SMALL_STATE(6549)] = 306466,
  [SMALL_STATE(6550)] = 306487,
  [SMALL_STATE(6551)] = 306506,
  [SMALL_STATE(6552)] = 306527,
  [SMALL_STATE(6553)] = 306548,
  [SMALL_STATE(6554)] = 306569,
  [SMALL_STATE(6555)] = 306590,
  [SMALL_STATE(6556)] = 306611,
  [SMALL_STATE(6557)] = 306630,
  [SMALL_STATE(6558)] = 306651,
  [SMALL_STATE(6559)] = 306672,
  [SMALL_STATE(6560)] = 306693,
  [SMALL_STATE(6561)] = 306714,
  [SMALL_STATE(6562)] = 306735,
  [SMALL_STATE(6563)] = 306754,
  [SMALL_STATE(6564)] = 306775,
  [SMALL_STATE(6565)] = 306796,
  [SMALL_STATE(6566)] = 306817,
  [SMALL_STATE(6567)] = 306838,
  [SMALL_STATE(6568)] = 306859,
  [SMALL_STATE(6569)] = 306880,
  [SMALL_STATE(6570)] = 306899,
  [SMALL_STATE(6571)] = 306918,
  [SMALL_STATE(6572)] = 306939,
  [SMALL_STATE(6573)] = 306960,
  [SMALL_STATE(6574)] = 306979,
  [SMALL_STATE(6575)] = 307000,
  [SMALL_STATE(6576)] = 307019,
  [SMALL_STATE(6577)] = 307040,
  [SMALL_STATE(6578)] = 307061,
  [SMALL_STATE(6579)] = 307082,
  [SMALL_STATE(6580)] = 307103,
  [SMALL_STATE(6581)] = 307124,
  [SMALL_STATE(6582)] = 307145,
  [SMALL_STATE(6583)] = 307166,
  [SMALL_STATE(6584)] = 307185,
  [SMALL_STATE(6585)] = 307206,
  [SMALL_STATE(6586)] = 307227,
  [SMALL_STATE(6587)] = 307246,
  [SMALL_STATE(6588)] = 307267,
  [SMALL_STATE(6589)] = 307288,
  [SMALL_STATE(6590)] = 307307,
  [SMALL_STATE(6591)] = 307328,
  [SMALL_STATE(6592)] = 307349,
  [SMALL_STATE(6593)] = 307370,
  [SMALL_STATE(6594)] = 307391,
  [SMALL_STATE(6595)] = 307412,
  [SMALL_STATE(6596)] = 307433,
  [SMALL_STATE(6597)] = 307452,
  [SMALL_STATE(6598)] = 307473,
  [SMALL_STATE(6599)] = 307492,
  [SMALL_STATE(6600)] = 307511,
  [SMALL_STATE(6601)] = 307532,
  [SMALL_STATE(6602)] = 307553,
  [SMALL_STATE(6603)] = 307574,
  [SMALL_STATE(6604)] = 307593,
  [SMALL_STATE(6605)] = 307612,
  [SMALL_STATE(6606)] = 307633,
  [SMALL_STATE(6607)] = 307654,
  [SMALL_STATE(6608)] = 307675,
  [SMALL_STATE(6609)] = 307693,
  [SMALL_STATE(6610)] = 307711,
  [SMALL_STATE(6611)] = 307729,
  [SMALL_STATE(6612)] = 307747,
  [SMALL_STATE(6613)] = 307765,
  [SMALL_STATE(6614)] = 307783,
  [SMALL_STATE(6615)] = 307801,
  [SMALL_STATE(6616)] = 307819,
  [SMALL_STATE(6617)] = 307837,
  [SMALL_STATE(6618)] = 307855,
  [SMALL_STATE(6619)] = 307873,
  [SMALL_STATE(6620)] = 307891,
  [SMALL_STATE(6621)] = 307909,
  [SMALL_STATE(6622)] = 307927,
  [SMALL_STATE(6623)] = 307945,
  [SMALL_STATE(6624)] = 307963,
  [SMALL_STATE(6625)] = 307981,
  [SMALL_STATE(6626)] = 307999,
  [SMALL_STATE(6627)] = 308017,
  [SMALL_STATE(6628)] = 308035,
  [SMALL_STATE(6629)] = 308053,
  [SMALL_STATE(6630)] = 308071,
  [SMALL_STATE(6631)] = 308089,
  [SMALL_STATE(6632)] = 308107,
  [SMALL_STATE(6633)] = 308125,
  [SMALL_STATE(6634)] = 308143,
  [SMALL_STATE(6635)] = 308161,
  [SMALL_STATE(6636)] = 308179,
  [SMALL_STATE(6637)] = 308197,
  [SMALL_STATE(6638)] = 308215,
  [SMALL_STATE(6639)] = 308233,
  [SMALL_STATE(6640)] = 308251,
  [SMALL_STATE(6641)] = 308269,
  [SMALL_STATE(6642)] = 308287,
  [SMALL_STATE(6643)] = 308305,
  [SMALL_STATE(6644)] = 308323,
  [SMALL_STATE(6645)] = 308341,
  [SMALL_STATE(6646)] = 308359,
  [SMALL_STATE(6647)] = 308377,
  [SMALL_STATE(6648)] = 308395,
  [SMALL_STATE(6649)] = 308413,
  [SMALL_STATE(6650)] = 308431,
  [SMALL_STATE(6651)] = 308449,
  [SMALL_STATE(6652)] = 308467,
  [SMALL_STATE(6653)] = 308485,
  [SMALL_STATE(6654)] = 308503,
  [SMALL_STATE(6655)] = 308521,
  [SMALL_STATE(6656)] = 308539,
  [SMALL_STATE(6657)] = 308557,
  [SMALL_STATE(6658)] = 308575,
  [SMALL_STATE(6659)] = 308593,
  [SMALL_STATE(6660)] = 308611,
  [SMALL_STATE(6661)] = 308629,
  [SMALL_STATE(6662)] = 308647,
  [SMALL_STATE(6663)] = 308665,
  [SMALL_STATE(6664)] = 308683,
  [SMALL_STATE(6665)] = 308701,
  [SMALL_STATE(6666)] = 308719,
  [SMALL_STATE(6667)] = 308737,
  [SMALL_STATE(6668)] = 308755,
  [SMALL_STATE(6669)] = 308773,
  [SMALL_STATE(6670)] = 308791,
  [SMALL_STATE(6671)] = 308809,
  [SMALL_STATE(6672)] = 308827,
  [SMALL_STATE(6673)] = 308845,
  [SMALL_STATE(6674)] = 308863,
  [SMALL_STATE(6675)] = 308881,
  [SMALL_STATE(6676)] = 308899,
  [SMALL_STATE(6677)] = 308917,
  [SMALL_STATE(6678)] = 308935,
  [SMALL_STATE(6679)] = 308953,
  [SMALL_STATE(6680)] = 308971,
  [SMALL_STATE(6681)] = 308989,
  [SMALL_STATE(6682)] = 309007,
  [SMALL_STATE(6683)] = 309025,
  [SMALL_STATE(6684)] = 309043,
  [SMALL_STATE(6685)] = 309061,
  [SMALL_STATE(6686)] = 309079,
  [SMALL_STATE(6687)] = 309097,
  [SMALL_STATE(6688)] = 309115,
  [SMALL_STATE(6689)] = 309133,
  [SMALL_STATE(6690)] = 309151,
  [SMALL_STATE(6691)] = 309169,
  [SMALL_STATE(6692)] = 309187,
  [SMALL_STATE(6693)] = 309205,
  [SMALL_STATE(6694)] = 309223,
  [SMALL_STATE(6695)] = 309241,
  [SMALL_STATE(6696)] = 309259,
  [SMALL_STATE(6697)] = 309277,
  [SMALL_STATE(6698)] = 309295,
  [SMALL_STATE(6699)] = 309313,
  [SMALL_STATE(6700)] = 309331,
  [SMALL_STATE(6701)] = 309349,
  [SMALL_STATE(6702)] = 309367,
  [SMALL_STATE(6703)] = 309385,
  [SMALL_STATE(6704)] = 309403,
  [SMALL_STATE(6705)] = 309421,
  [SMALL_STATE(6706)] = 309439,
  [SMALL_STATE(6707)] = 309457,
  [SMALL_STATE(6708)] = 309475,
  [SMALL_STATE(6709)] = 309493,
  [SMALL_STATE(6710)] = 309511,
  [SMALL_STATE(6711)] = 309529,
  [SMALL_STATE(6712)] = 309547,
  [SMALL_STATE(6713)] = 309565,
  [SMALL_STATE(6714)] = 309583,
  [SMALL_STATE(6715)] = 309601,
  [SMALL_STATE(6716)] = 309619,
  [SMALL_STATE(6717)] = 309637,
  [SMALL_STATE(6718)] = 309655,
  [SMALL_STATE(6719)] = 309673,
  [SMALL_STATE(6720)] = 309691,
  [SMALL_STATE(6721)] = 309709,
  [SMALL_STATE(6722)] = 309727,
  [SMALL_STATE(6723)] = 309745,
  [SMALL_STATE(6724)] = 309763,
  [SMALL_STATE(6725)] = 309781,
  [SMALL_STATE(6726)] = 309799,
  [SMALL_STATE(6727)] = 309817,
  [SMALL_STATE(6728)] = 309835,
  [SMALL_STATE(6729)] = 309853,
  [SMALL_STATE(6730)] = 309871,
  [SMALL_STATE(6731)] = 309889,
  [SMALL_STATE(6732)] = 309907,
  [SMALL_STATE(6733)] = 309925,
  [SMALL_STATE(6734)] = 309943,
  [SMALL_STATE(6735)] = 309961,
  [SMALL_STATE(6736)] = 309979,
  [SMALL_STATE(6737)] = 309997,
  [SMALL_STATE(6738)] = 310015,
  [SMALL_STATE(6739)] = 310033,
  [SMALL_STATE(6740)] = 310051,
  [SMALL_STATE(6741)] = 310069,
  [SMALL_STATE(6742)] = 310087,
  [SMALL_STATE(6743)] = 310105,
  [SMALL_STATE(6744)] = 310123,
  [SMALL_STATE(6745)] = 310141,
  [SMALL_STATE(6746)] = 310159,
  [SMALL_STATE(6747)] = 310177,
  [SMALL_STATE(6748)] = 310195,
  [SMALL_STATE(6749)] = 310213,
  [SMALL_STATE(6750)] = 310231,
  [SMALL_STATE(6751)] = 310249,
  [SMALL_STATE(6752)] = 310267,
  [SMALL_STATE(6753)] = 310285,
  [SMALL_STATE(6754)] = 310303,
  [SMALL_STATE(6755)] = 310321,
  [SMALL_STATE(6756)] = 310339,
  [SMALL_STATE(6757)] = 310357,
  [SMALL_STATE(6758)] = 310375,
  [SMALL_STATE(6759)] = 310393,
  [SMALL_STATE(6760)] = 310411,
  [SMALL_STATE(6761)] = 310429,
  [SMALL_STATE(6762)] = 310447,
  [SMALL_STATE(6763)] = 310465,
  [SMALL_STATE(6764)] = 310483,
  [SMALL_STATE(6765)] = 310501,
  [SMALL_STATE(6766)] = 310519,
  [SMALL_STATE(6767)] = 310537,
  [SMALL_STATE(6768)] = 310555,
  [SMALL_STATE(6769)] = 310573,
  [SMALL_STATE(6770)] = 310591,
  [SMALL_STATE(6771)] = 310609,
  [SMALL_STATE(6772)] = 310627,
  [SMALL_STATE(6773)] = 310645,
  [SMALL_STATE(6774)] = 310663,
  [SMALL_STATE(6775)] = 310681,
  [SMALL_STATE(6776)] = 310699,
  [SMALL_STATE(6777)] = 310717,
  [SMALL_STATE(6778)] = 310735,
  [SMALL_STATE(6779)] = 310753,
  [SMALL_STATE(6780)] = 310771,
  [SMALL_STATE(6781)] = 310789,
  [SMALL_STATE(6782)] = 310807,
  [SMALL_STATE(6783)] = 310825,
  [SMALL_STATE(6784)] = 310843,
  [SMALL_STATE(6785)] = 310861,
  [SMALL_STATE(6786)] = 310879,
  [SMALL_STATE(6787)] = 310897,
  [SMALL_STATE(6788)] = 310915,
  [SMALL_STATE(6789)] = 310933,
  [SMALL_STATE(6790)] = 310951,
  [SMALL_STATE(6791)] = 310969,
  [SMALL_STATE(6792)] = 310987,
  [SMALL_STATE(6793)] = 311005,
  [SMALL_STATE(6794)] = 311023,
  [SMALL_STATE(6795)] = 311041,
  [SMALL_STATE(6796)] = 311059,
  [SMALL_STATE(6797)] = 311077,
  [SMALL_STATE(6798)] = 311095,
  [SMALL_STATE(6799)] = 311113,
  [SMALL_STATE(6800)] = 311131,
  [SMALL_STATE(6801)] = 311149,
  [SMALL_STATE(6802)] = 311167,
  [SMALL_STATE(6803)] = 311185,
  [SMALL_STATE(6804)] = 311203,
  [SMALL_STATE(6805)] = 311221,
  [SMALL_STATE(6806)] = 311239,
  [SMALL_STATE(6807)] = 311257,
  [SMALL_STATE(6808)] = 311275,
  [SMALL_STATE(6809)] = 311293,
  [SMALL_STATE(6810)] = 311311,
  [SMALL_STATE(6811)] = 311329,
  [SMALL_STATE(6812)] = 311347,
  [SMALL_STATE(6813)] = 311365,
  [SMALL_STATE(6814)] = 311383,
  [SMALL_STATE(6815)] = 311401,
  [SMALL_STATE(6816)] = 311419,
  [SMALL_STATE(6817)] = 311437,
  [SMALL_STATE(6818)] = 311455,
  [SMALL_STATE(6819)] = 311473,
  [SMALL_STATE(6820)] = 311491,
  [SMALL_STATE(6821)] = 311509,
  [SMALL_STATE(6822)] = 311527,
  [SMALL_STATE(6823)] = 311545,
  [SMALL_STATE(6824)] = 311563,
  [SMALL_STATE(6825)] = 311581,
  [SMALL_STATE(6826)] = 311599,
  [SMALL_STATE(6827)] = 311617,
  [SMALL_STATE(6828)] = 311635,
  [SMALL_STATE(6829)] = 311653,
  [SMALL_STATE(6830)] = 311671,
  [SMALL_STATE(6831)] = 311689,
  [SMALL_STATE(6832)] = 311707,
  [SMALL_STATE(6833)] = 311725,
  [SMALL_STATE(6834)] = 311743,
  [SMALL_STATE(6835)] = 311761,
  [SMALL_STATE(6836)] = 311779,
  [SMALL_STATE(6837)] = 311797,
  [SMALL_STATE(6838)] = 311815,
  [SMALL_STATE(6839)] = 311833,
  [SMALL_STATE(6840)] = 311851,
  [SMALL_STATE(6841)] = 311869,
  [SMALL_STATE(6842)] = 311887,
  [SMALL_STATE(6843)] = 311905,
  [SMALL_STATE(6844)] = 311923,
  [SMALL_STATE(6845)] = 311941,
  [SMALL_STATE(6846)] = 311959,
  [SMALL_STATE(6847)] = 311977,
  [SMALL_STATE(6848)] = 311995,
  [SMALL_STATE(6849)] = 312013,
  [SMALL_STATE(6850)] = 312031,
  [SMALL_STATE(6851)] = 312049,
  [SMALL_STATE(6852)] = 312067,
  [SMALL_STATE(6853)] = 312085,
  [SMALL_STATE(6854)] = 312103,
  [SMALL_STATE(6855)] = 312121,
  [SMALL_STATE(6856)] = 312139,
  [SMALL_STATE(6857)] = 312157,
  [SMALL_STATE(6858)] = 312175,
  [SMALL_STATE(6859)] = 312193,
  [SMALL_STATE(6860)] = 312211,
  [SMALL_STATE(6861)] = 312229,
  [SMALL_STATE(6862)] = 312247,
  [SMALL_STATE(6863)] = 312265,
  [SMALL_STATE(6864)] = 312283,
  [SMALL_STATE(6865)] = 312301,
  [SMALL_STATE(6866)] = 312319,
  [SMALL_STATE(6867)] = 312337,
  [SMALL_STATE(6868)] = 312355,
  [SMALL_STATE(6869)] = 312373,
  [SMALL_STATE(6870)] = 312391,
  [SMALL_STATE(6871)] = 312409,
  [SMALL_STATE(6872)] = 312427,
  [SMALL_STATE(6873)] = 312445,
  [SMALL_STATE(6874)] = 312463,
  [SMALL_STATE(6875)] = 312481,
  [SMALL_STATE(6876)] = 312499,
  [SMALL_STATE(6877)] = 312517,
  [SMALL_STATE(6878)] = 312535,
  [SMALL_STATE(6879)] = 312553,
  [SMALL_STATE(6880)] = 312571,
  [SMALL_STATE(6881)] = 312589,
  [SMALL_STATE(6882)] = 312607,
  [SMALL_STATE(6883)] = 312625,
  [SMALL_STATE(6884)] = 312643,
  [SMALL_STATE(6885)] = 312661,
  [SMALL_STATE(6886)] = 312679,
  [SMALL_STATE(6887)] = 312697,
  [SMALL_STATE(6888)] = 312715,
  [SMALL_STATE(6889)] = 312733,
  [SMALL_STATE(6890)] = 312751,
  [SMALL_STATE(6891)] = 312769,
  [SMALL_STATE(6892)] = 312787,
  [SMALL_STATE(6893)] = 312805,
  [SMALL_STATE(6894)] = 312823,
  [SMALL_STATE(6895)] = 312841,
  [SMALL_STATE(6896)] = 312859,
  [SMALL_STATE(6897)] = 312877,
  [SMALL_STATE(6898)] = 312895,
  [SMALL_STATE(6899)] = 312913,
  [SMALL_STATE(6900)] = 312931,
  [SMALL_STATE(6901)] = 312949,
  [SMALL_STATE(6902)] = 312967,
  [SMALL_STATE(6903)] = 312985,
  [SMALL_STATE(6904)] = 313003,
  [SMALL_STATE(6905)] = 313021,
  [SMALL_STATE(6906)] = 313039,
  [SMALL_STATE(6907)] = 313057,
  [SMALL_STATE(6908)] = 313075,
  [SMALL_STATE(6909)] = 313093,
  [SMALL_STATE(6910)] = 313111,
  [SMALL_STATE(6911)] = 313129,
  [SMALL_STATE(6912)] = 313147,
  [SMALL_STATE(6913)] = 313165,
  [SMALL_STATE(6914)] = 313183,
  [SMALL_STATE(6915)] = 313201,
  [SMALL_STATE(6916)] = 313219,
  [SMALL_STATE(6917)] = 313237,
  [SMALL_STATE(6918)] = 313255,
  [SMALL_STATE(6919)] = 313273,
  [SMALL_STATE(6920)] = 313291,
  [SMALL_STATE(6921)] = 313309,
  [SMALL_STATE(6922)] = 313327,
  [SMALL_STATE(6923)] = 313345,
  [SMALL_STATE(6924)] = 313363,
  [SMALL_STATE(6925)] = 313381,
  [SMALL_STATE(6926)] = 313399,
  [SMALL_STATE(6927)] = 313417,
  [SMALL_STATE(6928)] = 313435,
  [SMALL_STATE(6929)] = 313453,
  [SMALL_STATE(6930)] = 313471,
  [SMALL_STATE(6931)] = 313489,
  [SMALL_STATE(6932)] = 313507,
  [SMALL_STATE(6933)] = 313525,
  [SMALL_STATE(6934)] = 313543,
  [SMALL_STATE(6935)] = 313561,
  [SMALL_STATE(6936)] = 313579,
  [SMALL_STATE(6937)] = 313597,
  [SMALL_STATE(6938)] = 313615,
  [SMALL_STATE(6939)] = 313633,
  [SMALL_STATE(6940)] = 313651,
  [SMALL_STATE(6941)] = 313669,
  [SMALL_STATE(6942)] = 313687,
  [SMALL_STATE(6943)] = 313705,
  [SMALL_STATE(6944)] = 313723,
  [SMALL_STATE(6945)] = 313741,
  [SMALL_STATE(6946)] = 313759,
  [SMALL_STATE(6947)] = 313777,
  [SMALL_STATE(6948)] = 313795,
  [SMALL_STATE(6949)] = 313813,
  [SMALL_STATE(6950)] = 313831,
  [SMALL_STATE(6951)] = 313849,
  [SMALL_STATE(6952)] = 313867,
  [SMALL_STATE(6953)] = 313885,
  [SMALL_STATE(6954)] = 313903,
  [SMALL_STATE(6955)] = 313921,
  [SMALL_STATE(6956)] = 313939,
  [SMALL_STATE(6957)] = 313957,
  [SMALL_STATE(6958)] = 313975,
  [SMALL_STATE(6959)] = 313993,
  [SMALL_STATE(6960)] = 314011,
  [SMALL_STATE(6961)] = 314029,
  [SMALL_STATE(6962)] = 314047,
  [SMALL_STATE(6963)] = 314065,
  [SMALL_STATE(6964)] = 314083,
  [SMALL_STATE(6965)] = 314101,
  [SMALL_STATE(6966)] = 314119,
  [SMALL_STATE(6967)] = 314137,
  [SMALL_STATE(6968)] = 314155,
  [SMALL_STATE(6969)] = 314173,
  [SMALL_STATE(6970)] = 314191,
  [SMALL_STATE(6971)] = 314209,
  [SMALL_STATE(6972)] = 314227,
  [SMALL_STATE(6973)] = 314245,
  [SMALL_STATE(6974)] = 314263,
  [SMALL_STATE(6975)] = 314281,
  [SMALL_STATE(6976)] = 314299,
  [SMALL_STATE(6977)] = 314317,
  [SMALL_STATE(6978)] = 314335,
  [SMALL_STATE(6979)] = 314353,
  [SMALL_STATE(6980)] = 314371,
  [SMALL_STATE(6981)] = 314389,
  [SMALL_STATE(6982)] = 314407,
  [SMALL_STATE(6983)] = 314425,
  [SMALL_STATE(6984)] = 314443,
  [SMALL_STATE(6985)] = 314461,
  [SMALL_STATE(6986)] = 314479,
  [SMALL_STATE(6987)] = 314497,
  [SMALL_STATE(6988)] = 314515,
  [SMALL_STATE(6989)] = 314533,
  [SMALL_STATE(6990)] = 314551,
  [SMALL_STATE(6991)] = 314569,
  [SMALL_STATE(6992)] = 314587,
  [SMALL_STATE(6993)] = 314605,
  [SMALL_STATE(6994)] = 314623,
  [SMALL_STATE(6995)] = 314641,
  [SMALL_STATE(6996)] = 314659,
  [SMALL_STATE(6997)] = 314677,
  [SMALL_STATE(6998)] = 314695,
  [SMALL_STATE(6999)] = 314713,
  [SMALL_STATE(7000)] = 314731,
  [SMALL_STATE(7001)] = 314749,
  [SMALL_STATE(7002)] = 314767,
  [SMALL_STATE(7003)] = 314785,
  [SMALL_STATE(7004)] = 314803,
  [SMALL_STATE(7005)] = 314821,
  [SMALL_STATE(7006)] = 314839,
  [SMALL_STATE(7007)] = 314857,
  [SMALL_STATE(7008)] = 314875,
  [SMALL_STATE(7009)] = 314893,
  [SMALL_STATE(7010)] = 314911,
  [SMALL_STATE(7011)] = 314929,
  [SMALL_STATE(7012)] = 314947,
  [SMALL_STATE(7013)] = 314965,
  [SMALL_STATE(7014)] = 314983,
  [SMALL_STATE(7015)] = 315001,
  [SMALL_STATE(7016)] = 315019,
  [SMALL_STATE(7017)] = 315037,
  [SMALL_STATE(7018)] = 315055,
  [SMALL_STATE(7019)] = 315073,
  [SMALL_STATE(7020)] = 315091,
  [SMALL_STATE(7021)] = 315109,
  [SMALL_STATE(7022)] = 315127,
  [SMALL_STATE(7023)] = 315145,
  [SMALL_STATE(7024)] = 315163,
  [SMALL_STATE(7025)] = 315181,
  [SMALL_STATE(7026)] = 315199,
  [SMALL_STATE(7027)] = 315217,
  [SMALL_STATE(7028)] = 315235,
  [SMALL_STATE(7029)] = 315253,
  [SMALL_STATE(7030)] = 315271,
  [SMALL_STATE(7031)] = 315289,
  [SMALL_STATE(7032)] = 315307,
  [SMALL_STATE(7033)] = 315325,
  [SMALL_STATE(7034)] = 315343,
  [SMALL_STATE(7035)] = 315361,
  [SMALL_STATE(7036)] = 315379,
  [SMALL_STATE(7037)] = 315397,
  [SMALL_STATE(7038)] = 315415,
  [SMALL_STATE(7039)] = 315433,
  [SMALL_STATE(7040)] = 315451,
  [SMALL_STATE(7041)] = 315469,
  [SMALL_STATE(7042)] = 315487,
  [SMALL_STATE(7043)] = 315505,
  [SMALL_STATE(7044)] = 315523,
  [SMALL_STATE(7045)] = 315541,
  [SMALL_STATE(7046)] = 315559,
  [SMALL_STATE(7047)] = 315577,
  [SMALL_STATE(7048)] = 315595,
  [SMALL_STATE(7049)] = 315613,
  [SMALL_STATE(7050)] = 315631,
  [SMALL_STATE(7051)] = 315649,
  [SMALL_STATE(7052)] = 315667,
  [SMALL_STATE(7053)] = 315685,
  [SMALL_STATE(7054)] = 315703,
  [SMALL_STATE(7055)] = 315721,
  [SMALL_STATE(7056)] = 315739,
  [SMALL_STATE(7057)] = 315757,
  [SMALL_STATE(7058)] = 315775,
  [SMALL_STATE(7059)] = 315793,
  [SMALL_STATE(7060)] = 315811,
  [SMALL_STATE(7061)] = 315829,
  [SMALL_STATE(7062)] = 315847,
  [SMALL_STATE(7063)] = 315865,
  [SMALL_STATE(7064)] = 315883,
  [SMALL_STATE(7065)] = 315901,
  [SMALL_STATE(7066)] = 315919,
  [SMALL_STATE(7067)] = 315937,
  [SMALL_STATE(7068)] = 315955,
  [SMALL_STATE(7069)] = 315973,
  [SMALL_STATE(7070)] = 315991,
  [SMALL_STATE(7071)] = 316009,
  [SMALL_STATE(7072)] = 316027,
  [SMALL_STATE(7073)] = 316045,
  [SMALL_STATE(7074)] = 316063,
  [SMALL_STATE(7075)] = 316081,
  [SMALL_STATE(7076)] = 316099,
  [SMALL_STATE(7077)] = 316117,
  [SMALL_STATE(7078)] = 316135,
  [SMALL_STATE(7079)] = 316153,
  [SMALL_STATE(7080)] = 316171,
  [SMALL_STATE(7081)] = 316189,
  [SMALL_STATE(7082)] = 316207,
  [SMALL_STATE(7083)] = 316225,
  [SMALL_STATE(7084)] = 316243,
  [SMALL_STATE(7085)] = 316261,
  [SMALL_STATE(7086)] = 316279,
  [SMALL_STATE(7087)] = 316297,
  [SMALL_STATE(7088)] = 316315,
  [SMALL_STATE(7089)] = 316333,
  [SMALL_STATE(7090)] = 316351,
  [SMALL_STATE(7091)] = 316369,
  [SMALL_STATE(7092)] = 316387,
  [SMALL_STATE(7093)] = 316405,
  [SMALL_STATE(7094)] = 316423,
  [SMALL_STATE(7095)] = 316441,
  [SMALL_STATE(7096)] = 316459,
  [SMALL_STATE(7097)] = 316477,
  [SMALL_STATE(7098)] = 316495,
  [SMALL_STATE(7099)] = 316513,
  [SMALL_STATE(7100)] = 316531,
  [SMALL_STATE(7101)] = 316549,
  [SMALL_STATE(7102)] = 316567,
  [SMALL_STATE(7103)] = 316585,
  [SMALL_STATE(7104)] = 316603,
  [SMALL_STATE(7105)] = 316621,
  [SMALL_STATE(7106)] = 316639,
  [SMALL_STATE(7107)] = 316657,
  [SMALL_STATE(7108)] = 316675,
  [SMALL_STATE(7109)] = 316693,
  [SMALL_STATE(7110)] = 316711,
  [SMALL_STATE(7111)] = 316729,
  [SMALL_STATE(7112)] = 316747,
  [SMALL_STATE(7113)] = 316765,
  [SMALL_STATE(7114)] = 316783,
  [SMALL_STATE(7115)] = 316801,
  [SMALL_STATE(7116)] = 316819,
  [SMALL_STATE(7117)] = 316837,
  [SMALL_STATE(7118)] = 316855,
  [SMALL_STATE(7119)] = 316873,
  [SMALL_STATE(7120)] = 316891,
  [SMALL_STATE(7121)] = 316909,
  [SMALL_STATE(7122)] = 316927,
  [SMALL_STATE(7123)] = 316945,
  [SMALL_STATE(7124)] = 316963,
  [SMALL_STATE(7125)] = 316981,
  [SMALL_STATE(7126)] = 316999,
  [SMALL_STATE(7127)] = 317017,
  [SMALL_STATE(7128)] = 317035,
  [SMALL_STATE(7129)] = 317053,
  [SMALL_STATE(7130)] = 317071,
  [SMALL_STATE(7131)] = 317089,
  [SMALL_STATE(7132)] = 317107,
  [SMALL_STATE(7133)] = 317125,
  [SMALL_STATE(7134)] = 317143,
  [SMALL_STATE(7135)] = 317161,
  [SMALL_STATE(7136)] = 317179,
  [SMALL_STATE(7137)] = 317197,
  [SMALL_STATE(7138)] = 317215,
  [SMALL_STATE(7139)] = 317233,
  [SMALL_STATE(7140)] = 317251,
  [SMALL_STATE(7141)] = 317269,
  [SMALL_STATE(7142)] = 317287,
  [SMALL_STATE(7143)] = 317305,
  [SMALL_STATE(7144)] = 317323,
  [SMALL_STATE(7145)] = 317341,
  [SMALL_STATE(7146)] = 317359,
  [SMALL_STATE(7147)] = 317377,
  [SMALL_STATE(7148)] = 317395,
  [SMALL_STATE(7149)] = 317413,
  [SMALL_STATE(7150)] = 317431,
  [SMALL_STATE(7151)] = 317449,
  [SMALL_STATE(7152)] = 317467,
  [SMALL_STATE(7153)] = 317485,
  [SMALL_STATE(7154)] = 317503,
  [SMALL_STATE(7155)] = 317521,
  [SMALL_STATE(7156)] = 317539,
  [SMALL_STATE(7157)] = 317557,
  [SMALL_STATE(7158)] = 317575,
  [SMALL_STATE(7159)] = 317593,
  [SMALL_STATE(7160)] = 317611,
  [SMALL_STATE(7161)] = 317629,
  [SMALL_STATE(7162)] = 317647,
  [SMALL_STATE(7163)] = 317665,
  [SMALL_STATE(7164)] = 317683,
  [SMALL_STATE(7165)] = 317701,
  [SMALL_STATE(7166)] = 317719,
  [SMALL_STATE(7167)] = 317737,
  [SMALL_STATE(7168)] = 317755,
  [SMALL_STATE(7169)] = 317773,
  [SMALL_STATE(7170)] = 317791,
  [SMALL_STATE(7171)] = 317809,
  [SMALL_STATE(7172)] = 317827,
  [SMALL_STATE(7173)] = 317845,
  [SMALL_STATE(7174)] = 317863,
  [SMALL_STATE(7175)] = 317881,
  [SMALL_STATE(7176)] = 317899,
  [SMALL_STATE(7177)] = 317917,
  [SMALL_STATE(7178)] = 317935,
  [SMALL_STATE(7179)] = 317953,
  [SMALL_STATE(7180)] = 317971,
  [SMALL_STATE(7181)] = 317989,
  [SMALL_STATE(7182)] = 318007,
  [SMALL_STATE(7183)] = 318025,
  [SMALL_STATE(7184)] = 318043,
  [SMALL_STATE(7185)] = 318061,
  [SMALL_STATE(7186)] = 318079,
  [SMALL_STATE(7187)] = 318097,
  [SMALL_STATE(7188)] = 318115,
  [SMALL_STATE(7189)] = 318133,
  [SMALL_STATE(7190)] = 318151,
  [SMALL_STATE(7191)] = 318169,
  [SMALL_STATE(7192)] = 318187,
  [SMALL_STATE(7193)] = 318205,
  [SMALL_STATE(7194)] = 318223,
  [SMALL_STATE(7195)] = 318241,
  [SMALL_STATE(7196)] = 318259,
  [SMALL_STATE(7197)] = 318277,
  [SMALL_STATE(7198)] = 318295,
  [SMALL_STATE(7199)] = 318313,
  [SMALL_STATE(7200)] = 318331,
  [SMALL_STATE(7201)] = 318349,
  [SMALL_STATE(7202)] = 318367,
  [SMALL_STATE(7203)] = 318385,
  [SMALL_STATE(7204)] = 318403,
  [SMALL_STATE(7205)] = 318421,
  [SMALL_STATE(7206)] = 318439,
  [SMALL_STATE(7207)] = 318457,
  [SMALL_STATE(7208)] = 318475,
  [SMALL_STATE(7209)] = 318493,
  [SMALL_STATE(7210)] = 318511,
  [SMALL_STATE(7211)] = 318529,
  [SMALL_STATE(7212)] = 318547,
  [SMALL_STATE(7213)] = 318565,
  [SMALL_STATE(7214)] = 318583,
  [SMALL_STATE(7215)] = 318601,
  [SMALL_STATE(7216)] = 318619,
  [SMALL_STATE(7217)] = 318637,
  [SMALL_STATE(7218)] = 318655,
  [SMALL_STATE(7219)] = 318673,
  [SMALL_STATE(7220)] = 318691,
  [SMALL_STATE(7221)] = 318709,
  [SMALL_STATE(7222)] = 318727,
  [SMALL_STATE(7223)] = 318745,
  [SMALL_STATE(7224)] = 318763,
  [SMALL_STATE(7225)] = 318781,
  [SMALL_STATE(7226)] = 318799,
  [SMALL_STATE(7227)] = 318817,
  [SMALL_STATE(7228)] = 318835,
  [SMALL_STATE(7229)] = 318853,
  [SMALL_STATE(7230)] = 318871,
  [SMALL_STATE(7231)] = 318889,
  [SMALL_STATE(7232)] = 318907,
  [SMALL_STATE(7233)] = 318925,
  [SMALL_STATE(7234)] = 318943,
  [SMALL_STATE(7235)] = 318961,
  [SMALL_STATE(7236)] = 318979,
  [SMALL_STATE(7237)] = 318997,
  [SMALL_STATE(7238)] = 319015,
  [SMALL_STATE(7239)] = 319033,
  [SMALL_STATE(7240)] = 319051,
  [SMALL_STATE(7241)] = 319069,
  [SMALL_STATE(7242)] = 319087,
  [SMALL_STATE(7243)] = 319105,
  [SMALL_STATE(7244)] = 319123,
  [SMALL_STATE(7245)] = 319141,
  [SMALL_STATE(7246)] = 319159,
  [SMALL_STATE(7247)] = 319177,
  [SMALL_STATE(7248)] = 319195,
  [SMALL_STATE(7249)] = 319213,
  [SMALL_STATE(7250)] = 319231,
  [SMALL_STATE(7251)] = 319249,
  [SMALL_STATE(7252)] = 319267,
  [SMALL_STATE(7253)] = 319285,
  [SMALL_STATE(7254)] = 319303,
  [SMALL_STATE(7255)] = 319321,
  [SMALL_STATE(7256)] = 319339,
  [SMALL_STATE(7257)] = 319357,
  [SMALL_STATE(7258)] = 319375,
  [SMALL_STATE(7259)] = 319393,
  [SMALL_STATE(7260)] = 319411,
  [SMALL_STATE(7261)] = 319429,
  [SMALL_STATE(7262)] = 319447,
  [SMALL_STATE(7263)] = 319465,
  [SMALL_STATE(7264)] = 319483,
  [SMALL_STATE(7265)] = 319501,
  [SMALL_STATE(7266)] = 319519,
  [SMALL_STATE(7267)] = 319537,
  [SMALL_STATE(7268)] = 319555,
  [SMALL_STATE(7269)] = 319573,
  [SMALL_STATE(7270)] = 319591,
  [SMALL_STATE(7271)] = 319609,
  [SMALL_STATE(7272)] = 319627,
  [SMALL_STATE(7273)] = 319645,
  [SMALL_STATE(7274)] = 319663,
  [SMALL_STATE(7275)] = 319681,
  [SMALL_STATE(7276)] = 319699,
  [SMALL_STATE(7277)] = 319717,
  [SMALL_STATE(7278)] = 319735,
  [SMALL_STATE(7279)] = 319753,
  [SMALL_STATE(7280)] = 319771,
  [SMALL_STATE(7281)] = 319789,
  [SMALL_STATE(7282)] = 319807,
  [SMALL_STATE(7283)] = 319825,
  [SMALL_STATE(7284)] = 319843,
  [SMALL_STATE(7285)] = 319861,
  [SMALL_STATE(7286)] = 319879,
  [SMALL_STATE(7287)] = 319897,
  [SMALL_STATE(7288)] = 319915,
  [SMALL_STATE(7289)] = 319933,
  [SMALL_STATE(7290)] = 319951,
  [SMALL_STATE(7291)] = 319969,
  [SMALL_STATE(7292)] = 319987,
  [SMALL_STATE(7293)] = 320005,
  [SMALL_STATE(7294)] = 320023,
  [SMALL_STATE(7295)] = 320041,
  [SMALL_STATE(7296)] = 320059,
  [SMALL_STATE(7297)] = 320077,
  [SMALL_STATE(7298)] = 320095,
  [SMALL_STATE(7299)] = 320113,
  [SMALL_STATE(7300)] = 320131,
  [SMALL_STATE(7301)] = 320149,
  [SMALL_STATE(7302)] = 320167,
  [SMALL_STATE(7303)] = 320185,
  [SMALL_STATE(7304)] = 320203,
  [SMALL_STATE(7305)] = 320221,
  [SMALL_STATE(7306)] = 320239,
  [SMALL_STATE(7307)] = 320257,
  [SMALL_STATE(7308)] = 320275,
  [SMALL_STATE(7309)] = 320293,
  [SMALL_STATE(7310)] = 320311,
  [SMALL_STATE(7311)] = 320329,
  [SMALL_STATE(7312)] = 320347,
  [SMALL_STATE(7313)] = 320365,
  [SMALL_STATE(7314)] = 320383,
  [SMALL_STATE(7315)] = 320401,
  [SMALL_STATE(7316)] = 320419,
  [SMALL_STATE(7317)] = 320437,
  [SMALL_STATE(7318)] = 320455,
  [SMALL_STATE(7319)] = 320473,
  [SMALL_STATE(7320)] = 320491,
  [SMALL_STATE(7321)] = 320509,
  [SMALL_STATE(7322)] = 320527,
  [SMALL_STATE(7323)] = 320545,
  [SMALL_STATE(7324)] = 320563,
  [SMALL_STATE(7325)] = 320581,
  [SMALL_STATE(7326)] = 320599,
  [SMALL_STATE(7327)] = 320617,
  [SMALL_STATE(7328)] = 320635,
  [SMALL_STATE(7329)] = 320653,
  [SMALL_STATE(7330)] = 320671,
  [SMALL_STATE(7331)] = 320689,
  [SMALL_STATE(7332)] = 320707,
  [SMALL_STATE(7333)] = 320725,
  [SMALL_STATE(7334)] = 320743,
  [SMALL_STATE(7335)] = 320761,
  [SMALL_STATE(7336)] = 320779,
  [SMALL_STATE(7337)] = 320797,
  [SMALL_STATE(7338)] = 320815,
  [SMALL_STATE(7339)] = 320833,
  [SMALL_STATE(7340)] = 320851,
  [SMALL_STATE(7341)] = 320869,
  [SMALL_STATE(7342)] = 320887,
  [SMALL_STATE(7343)] = 320905,
  [SMALL_STATE(7344)] = 320923,
  [SMALL_STATE(7345)] = 320941,
  [SMALL_STATE(7346)] = 320959,
  [SMALL_STATE(7347)] = 320977,
  [SMALL_STATE(7348)] = 320995,
  [SMALL_STATE(7349)] = 321013,
  [SMALL_STATE(7350)] = 321031,
  [SMALL_STATE(7351)] = 321049,
  [SMALL_STATE(7352)] = 321067,
  [SMALL_STATE(7353)] = 321085,
  [SMALL_STATE(7354)] = 321103,
  [SMALL_STATE(7355)] = 321121,
  [SMALL_STATE(7356)] = 321139,
  [SMALL_STATE(7357)] = 321157,
  [SMALL_STATE(7358)] = 321175,
  [SMALL_STATE(7359)] = 321193,
  [SMALL_STATE(7360)] = 321211,
  [SMALL_STATE(7361)] = 321229,
  [SMALL_STATE(7362)] = 321247,
  [SMALL_STATE(7363)] = 321265,
  [SMALL_STATE(7364)] = 321283,
  [SMALL_STATE(7365)] = 321301,
  [SMALL_STATE(7366)] = 321319,
  [SMALL_STATE(7367)] = 321337,
  [SMALL_STATE(7368)] = 321355,
  [SMALL_STATE(7369)] = 321373,
  [SMALL_STATE(7370)] = 321391,
  [SMALL_STATE(7371)] = 321409,
  [SMALL_STATE(7372)] = 321427,
  [SMALL_STATE(7373)] = 321445,
  [SMALL_STATE(7374)] = 321463,
  [SMALL_STATE(7375)] = 321481,
  [SMALL_STATE(7376)] = 321499,
  [SMALL_STATE(7377)] = 321517,
  [SMALL_STATE(7378)] = 321535,
  [SMALL_STATE(7379)] = 321553,
  [SMALL_STATE(7380)] = 321571,
  [SMALL_STATE(7381)] = 321589,
  [SMALL_STATE(7382)] = 321607,
  [SMALL_STATE(7383)] = 321625,
  [SMALL_STATE(7384)] = 321643,
  [SMALL_STATE(7385)] = 321661,
  [SMALL_STATE(7386)] = 321679,
  [SMALL_STATE(7387)] = 321697,
  [SMALL_STATE(7388)] = 321715,
  [SMALL_STATE(7389)] = 321733,
  [SMALL_STATE(7390)] = 321751,
  [SMALL_STATE(7391)] = 321769,
  [SMALL_STATE(7392)] = 321787,
  [SMALL_STATE(7393)] = 321805,
  [SMALL_STATE(7394)] = 321823,
  [SMALL_STATE(7395)] = 321841,
  [SMALL_STATE(7396)] = 321859,
  [SMALL_STATE(7397)] = 321877,
  [SMALL_STATE(7398)] = 321895,
  [SMALL_STATE(7399)] = 321913,
  [SMALL_STATE(7400)] = 321931,
  [SMALL_STATE(7401)] = 321949,
  [SMALL_STATE(7402)] = 321967,
  [SMALL_STATE(7403)] = 321985,
  [SMALL_STATE(7404)] = 322003,
  [SMALL_STATE(7405)] = 322021,
  [SMALL_STATE(7406)] = 322039,
  [SMALL_STATE(7407)] = 322057,
  [SMALL_STATE(7408)] = 322075,
  [SMALL_STATE(7409)] = 322093,
  [SMALL_STATE(7410)] = 322111,
  [SMALL_STATE(7411)] = 322129,
  [SMALL_STATE(7412)] = 322147,
  [SMALL_STATE(7413)] = 322165,
  [SMALL_STATE(7414)] = 322183,
  [SMALL_STATE(7415)] = 322201,
  [SMALL_STATE(7416)] = 322219,
  [SMALL_STATE(7417)] = 322237,
  [SMALL_STATE(7418)] = 322255,
  [SMALL_STATE(7419)] = 322273,
  [SMALL_STATE(7420)] = 322291,
  [SMALL_STATE(7421)] = 322309,
  [SMALL_STATE(7422)] = 322327,
  [SMALL_STATE(7423)] = 322345,
  [SMALL_STATE(7424)] = 322363,
  [SMALL_STATE(7425)] = 322381,
  [SMALL_STATE(7426)] = 322399,
  [SMALL_STATE(7427)] = 322417,
  [SMALL_STATE(7428)] = 322435,
  [SMALL_STATE(7429)] = 322453,
  [SMALL_STATE(7430)] = 322471,
  [SMALL_STATE(7431)] = 322489,
  [SMALL_STATE(7432)] = 322507,
  [SMALL_STATE(7433)] = 322525,
  [SMALL_STATE(7434)] = 322543,
  [SMALL_STATE(7435)] = 322561,
  [SMALL_STATE(7436)] = 322579,
  [SMALL_STATE(7437)] = 322597,
  [SMALL_STATE(7438)] = 322615,
  [SMALL_STATE(7439)] = 322633,
  [SMALL_STATE(7440)] = 322651,
  [SMALL_STATE(7441)] = 322669,
  [SMALL_STATE(7442)] = 322687,
  [SMALL_STATE(7443)] = 322705,
  [SMALL_STATE(7444)] = 322723,
  [SMALL_STATE(7445)] = 322741,
  [SMALL_STATE(7446)] = 322759,
  [SMALL_STATE(7447)] = 322777,
  [SMALL_STATE(7448)] = 322795,
  [SMALL_STATE(7449)] = 322813,
  [SMALL_STATE(7450)] = 322831,
  [SMALL_STATE(7451)] = 322849,
  [SMALL_STATE(7452)] = 322867,
  [SMALL_STATE(7453)] = 322885,
  [SMALL_STATE(7454)] = 322903,
  [SMALL_STATE(7455)] = 322921,
  [SMALL_STATE(7456)] = 322939,
  [SMALL_STATE(7457)] = 322957,
  [SMALL_STATE(7458)] = 322975,
  [SMALL_STATE(7459)] = 322993,
  [SMALL_STATE(7460)] = 323011,
  [SMALL_STATE(7461)] = 323029,
  [SMALL_STATE(7462)] = 323047,
  [SMALL_STATE(7463)] = 323065,
  [SMALL_STATE(7464)] = 323083,
  [SMALL_STATE(7465)] = 323101,
  [SMALL_STATE(7466)] = 323119,
  [SMALL_STATE(7467)] = 323137,
  [SMALL_STATE(7468)] = 323155,
  [SMALL_STATE(7469)] = 323173,
  [SMALL_STATE(7470)] = 323191,
  [SMALL_STATE(7471)] = 323209,
  [SMALL_STATE(7472)] = 323227,
  [SMALL_STATE(7473)] = 323245,
  [SMALL_STATE(7474)] = 323263,
  [SMALL_STATE(7475)] = 323281,
  [SMALL_STATE(7476)] = 323299,
  [SMALL_STATE(7477)] = 323317,
  [SMALL_STATE(7478)] = 323335,
  [SMALL_STATE(7479)] = 323353,
  [SMALL_STATE(7480)] = 323371,
  [SMALL_STATE(7481)] = 323389,
  [SMALL_STATE(7482)] = 323407,
  [SMALL_STATE(7483)] = 323425,
  [SMALL_STATE(7484)] = 323443,
  [SMALL_STATE(7485)] = 323461,
  [SMALL_STATE(7486)] = 323479,
  [SMALL_STATE(7487)] = 323497,
  [SMALL_STATE(7488)] = 323515,
  [SMALL_STATE(7489)] = 323533,
  [SMALL_STATE(7490)] = 323551,
  [SMALL_STATE(7491)] = 323569,
  [SMALL_STATE(7492)] = 323587,
  [SMALL_STATE(7493)] = 323605,
  [SMALL_STATE(7494)] = 323623,
  [SMALL_STATE(7495)] = 323641,
  [SMALL_STATE(7496)] = 323659,
  [SMALL_STATE(7497)] = 323677,
  [SMALL_STATE(7498)] = 323695,
  [SMALL_STATE(7499)] = 323713,
  [SMALL_STATE(7500)] = 323731,
  [SMALL_STATE(7501)] = 323749,
  [SMALL_STATE(7502)] = 323767,
  [SMALL_STATE(7503)] = 323785,
  [SMALL_STATE(7504)] = 323803,
  [SMALL_STATE(7505)] = 323821,
  [SMALL_STATE(7506)] = 323839,
  [SMALL_STATE(7507)] = 323857,
  [SMALL_STATE(7508)] = 323875,
  [SMALL_STATE(7509)] = 323893,
  [SMALL_STATE(7510)] = 323911,
  [SMALL_STATE(7511)] = 323929,
  [SMALL_STATE(7512)] = 323947,
  [SMALL_STATE(7513)] = 323965,
  [SMALL_STATE(7514)] = 323983,
  [SMALL_STATE(7515)] = 324001,
  [SMALL_STATE(7516)] = 324019,
  [SMALL_STATE(7517)] = 324037,
  [SMALL_STATE(7518)] = 324055,
  [SMALL_STATE(7519)] = 324073,
  [SMALL_STATE(7520)] = 324091,
  [SMALL_STATE(7521)] = 324109,
  [SMALL_STATE(7522)] = 324127,
  [SMALL_STATE(7523)] = 324145,
  [SMALL_STATE(7524)] = 324163,
  [SMALL_STATE(7525)] = 324181,
  [SMALL_STATE(7526)] = 324199,
  [SMALL_STATE(7527)] = 324217,
  [SMALL_STATE(7528)] = 324235,
  [SMALL_STATE(7529)] = 324253,
  [SMALL_STATE(7530)] = 324271,
  [SMALL_STATE(7531)] = 324289,
  [SMALL_STATE(7532)] = 324307,
  [SMALL_STATE(7533)] = 324325,
  [SMALL_STATE(7534)] = 324343,
  [SMALL_STATE(7535)] = 324361,
  [SMALL_STATE(7536)] = 324379,
  [SMALL_STATE(7537)] = 324397,
  [SMALL_STATE(7538)] = 324415,
  [SMALL_STATE(7539)] = 324433,
  [SMALL_STATE(7540)] = 324451,
  [SMALL_STATE(7541)] = 324469,
  [SMALL_STATE(7542)] = 324487,
  [SMALL_STATE(7543)] = 324505,
  [SMALL_STATE(7544)] = 324523,
  [SMALL_STATE(7545)] = 324541,
  [SMALL_STATE(7546)] = 324559,
  [SMALL_STATE(7547)] = 324577,
  [SMALL_STATE(7548)] = 324595,
  [SMALL_STATE(7549)] = 324613,
  [SMALL_STATE(7550)] = 324631,
  [SMALL_STATE(7551)] = 324649,
  [SMALL_STATE(7552)] = 324667,
  [SMALL_STATE(7553)] = 324685,
  [SMALL_STATE(7554)] = 324703,
  [SMALL_STATE(7555)] = 324721,
  [SMALL_STATE(7556)] = 324739,
  [SMALL_STATE(7557)] = 324757,
  [SMALL_STATE(7558)] = 324775,
  [SMALL_STATE(7559)] = 324793,
  [SMALL_STATE(7560)] = 324811,
  [SMALL_STATE(7561)] = 324829,
  [SMALL_STATE(7562)] = 324847,
  [SMALL_STATE(7563)] = 324865,
  [SMALL_STATE(7564)] = 324883,
  [SMALL_STATE(7565)] = 324901,
  [SMALL_STATE(7566)] = 324919,
  [SMALL_STATE(7567)] = 324937,
  [SMALL_STATE(7568)] = 324955,
  [SMALL_STATE(7569)] = 324973,
  [SMALL_STATE(7570)] = 324991,
  [SMALL_STATE(7571)] = 325009,
  [SMALL_STATE(7572)] = 325027,
  [SMALL_STATE(7573)] = 325045,
  [SMALL_STATE(7574)] = 325063,
  [SMALL_STATE(7575)] = 325081,
  [SMALL_STATE(7576)] = 325099,
  [SMALL_STATE(7577)] = 325117,
  [SMALL_STATE(7578)] = 325135,
  [SMALL_STATE(7579)] = 325153,
  [SMALL_STATE(7580)] = 325171,
  [SMALL_STATE(7581)] = 325189,
  [SMALL_STATE(7582)] = 325207,
  [SMALL_STATE(7583)] = 325225,
  [SMALL_STATE(7584)] = 325243,
  [SMALL_STATE(7585)] = 325261,
  [SMALL_STATE(7586)] = 325279,
  [SMALL_STATE(7587)] = 325297,
  [SMALL_STATE(7588)] = 325315,
  [SMALL_STATE(7589)] = 325333,
  [SMALL_STATE(7590)] = 325351,
  [SMALL_STATE(7591)] = 325369,
  [SMALL_STATE(7592)] = 325387,
  [SMALL_STATE(7593)] = 325405,
  [SMALL_STATE(7594)] = 325423,
  [SMALL_STATE(7595)] = 325441,
  [SMALL_STATE(7596)] = 325459,
  [SMALL_STATE(7597)] = 325477,
  [SMALL_STATE(7598)] = 325495,
  [SMALL_STATE(7599)] = 325513,
  [SMALL_STATE(7600)] = 325531,
  [SMALL_STATE(7601)] = 325549,
  [SMALL_STATE(7602)] = 325567,
  [SMALL_STATE(7603)] = 325585,
  [SMALL_STATE(7604)] = 325603,
  [SMALL_STATE(7605)] = 325621,
  [SMALL_STATE(7606)] = 325639,
  [SMALL_STATE(7607)] = 325657,
  [SMALL_STATE(7608)] = 325675,
  [SMALL_STATE(7609)] = 325693,
  [SMALL_STATE(7610)] = 325711,
  [SMALL_STATE(7611)] = 325729,
  [SMALL_STATE(7612)] = 325747,
  [SMALL_STATE(7613)] = 325765,
  [SMALL_STATE(7614)] = 325783,
  [SMALL_STATE(7615)] = 325801,
  [SMALL_STATE(7616)] = 325819,
  [SMALL_STATE(7617)] = 325837,
  [SMALL_STATE(7618)] = 325855,
  [SMALL_STATE(7619)] = 325873,
  [SMALL_STATE(7620)] = 325891,
  [SMALL_STATE(7621)] = 325909,
  [SMALL_STATE(7622)] = 325927,
  [SMALL_STATE(7623)] = 325945,
  [SMALL_STATE(7624)] = 325963,
  [SMALL_STATE(7625)] = 325981,
  [SMALL_STATE(7626)] = 325999,
  [SMALL_STATE(7627)] = 326017,
  [SMALL_STATE(7628)] = 326035,
  [SMALL_STATE(7629)] = 326053,
  [SMALL_STATE(7630)] = 326071,
  [SMALL_STATE(7631)] = 326089,
  [SMALL_STATE(7632)] = 326107,
  [SMALL_STATE(7633)] = 326125,
  [SMALL_STATE(7634)] = 326143,
  [SMALL_STATE(7635)] = 326161,
  [SMALL_STATE(7636)] = 326179,
  [SMALL_STATE(7637)] = 326197,
  [SMALL_STATE(7638)] = 326215,
  [SMALL_STATE(7639)] = 326233,
  [SMALL_STATE(7640)] = 326251,
  [SMALL_STATE(7641)] = 326269,
  [SMALL_STATE(7642)] = 326287,
  [SMALL_STATE(7643)] = 326305,
  [SMALL_STATE(7644)] = 326323,
  [SMALL_STATE(7645)] = 326341,
  [SMALL_STATE(7646)] = 326359,
  [SMALL_STATE(7647)] = 326377,
  [SMALL_STATE(7648)] = 326395,
  [SMALL_STATE(7649)] = 326413,
  [SMALL_STATE(7650)] = 326431,
  [SMALL_STATE(7651)] = 326449,
  [SMALL_STATE(7652)] = 326467,
  [SMALL_STATE(7653)] = 326485,
  [SMALL_STATE(7654)] = 326503,
  [SMALL_STATE(7655)] = 326521,
  [SMALL_STATE(7656)] = 326539,
  [SMALL_STATE(7657)] = 326557,
  [SMALL_STATE(7658)] = 326575,
  [SMALL_STATE(7659)] = 326593,
  [SMALL_STATE(7660)] = 326611,
  [SMALL_STATE(7661)] = 326629,
  [SMALL_STATE(7662)] = 326647,
  [SMALL_STATE(7663)] = 326665,
  [SMALL_STATE(7664)] = 326683,
  [SMALL_STATE(7665)] = 326701,
  [SMALL_STATE(7666)] = 326719,
  [SMALL_STATE(7667)] = 326737,
  [SMALL_STATE(7668)] = 326755,
  [SMALL_STATE(7669)] = 326773,
  [SMALL_STATE(7670)] = 326791,
  [SMALL_STATE(7671)] = 326809,
  [SMALL_STATE(7672)] = 326827,
  [SMALL_STATE(7673)] = 326845,
  [SMALL_STATE(7674)] = 326863,
  [SMALL_STATE(7675)] = 326881,
  [SMALL_STATE(7676)] = 326899,
  [SMALL_STATE(7677)] = 326917,
  [SMALL_STATE(7678)] = 326935,
  [SMALL_STATE(7679)] = 326953,
  [SMALL_STATE(7680)] = 326971,
  [SMALL_STATE(7681)] = 326989,
  [SMALL_STATE(7682)] = 327007,
  [SMALL_STATE(7683)] = 327025,
  [SMALL_STATE(7684)] = 327043,
  [SMALL_STATE(7685)] = 327061,
  [SMALL_STATE(7686)] = 327079,
  [SMALL_STATE(7687)] = 327097,
  [SMALL_STATE(7688)] = 327115,
  [SMALL_STATE(7689)] = 327133,
  [SMALL_STATE(7690)] = 327151,
  [SMALL_STATE(7691)] = 327169,
  [SMALL_STATE(7692)] = 327187,
  [SMALL_STATE(7693)] = 327205,
  [SMALL_STATE(7694)] = 327223,
  [SMALL_STATE(7695)] = 327241,
  [SMALL_STATE(7696)] = 327259,
  [SMALL_STATE(7697)] = 327277,
  [SMALL_STATE(7698)] = 327295,
  [SMALL_STATE(7699)] = 327313,
  [SMALL_STATE(7700)] = 327331,
  [SMALL_STATE(7701)] = 327349,
  [SMALL_STATE(7702)] = 327367,
  [SMALL_STATE(7703)] = 327385,
  [SMALL_STATE(7704)] = 327403,
  [SMALL_STATE(7705)] = 327421,
  [SMALL_STATE(7706)] = 327439,
  [SMALL_STATE(7707)] = 327457,
  [SMALL_STATE(7708)] = 327475,
  [SMALL_STATE(7709)] = 327493,
  [SMALL_STATE(7710)] = 327511,
  [SMALL_STATE(7711)] = 327529,
  [SMALL_STATE(7712)] = 327547,
  [SMALL_STATE(7713)] = 327565,
  [SMALL_STATE(7714)] = 327583,
  [SMALL_STATE(7715)] = 327601,
  [SMALL_STATE(7716)] = 327619,
  [SMALL_STATE(7717)] = 327637,
  [SMALL_STATE(7718)] = 327655,
  [SMALL_STATE(7719)] = 327673,
  [SMALL_STATE(7720)] = 327691,
  [SMALL_STATE(7721)] = 327709,
  [SMALL_STATE(7722)] = 327727,
  [SMALL_STATE(7723)] = 327745,
  [SMALL_STATE(7724)] = 327763,
  [SMALL_STATE(7725)] = 327781,
  [SMALL_STATE(7726)] = 327799,
  [SMALL_STATE(7727)] = 327817,
  [SMALL_STATE(7728)] = 327835,
  [SMALL_STATE(7729)] = 327853,
  [SMALL_STATE(7730)] = 327871,
  [SMALL_STATE(7731)] = 327889,
  [SMALL_STATE(7732)] = 327907,
  [SMALL_STATE(7733)] = 327925,
  [SMALL_STATE(7734)] = 327943,
  [SMALL_STATE(7735)] = 327961,
  [SMALL_STATE(7736)] = 327979,
  [SMALL_STATE(7737)] = 327997,
  [SMALL_STATE(7738)] = 328015,
  [SMALL_STATE(7739)] = 328033,
  [SMALL_STATE(7740)] = 328051,
  [SMALL_STATE(7741)] = 328069,
  [SMALL_STATE(7742)] = 328087,
  [SMALL_STATE(7743)] = 328105,
  [SMALL_STATE(7744)] = 328123,
  [SMALL_STATE(7745)] = 328141,
  [SMALL_STATE(7746)] = 328159,
  [SMALL_STATE(7747)] = 328177,
  [SMALL_STATE(7748)] = 328195,
  [SMALL_STATE(7749)] = 328213,
  [SMALL_STATE(7750)] = 328231,
  [SMALL_STATE(7751)] = 328249,
  [SMALL_STATE(7752)] = 328267,
  [SMALL_STATE(7753)] = 328285,
  [SMALL_STATE(7754)] = 328303,
  [SMALL_STATE(7755)] = 328321,
  [SMALL_STATE(7756)] = 328339,
  [SMALL_STATE(7757)] = 328357,
  [SMALL_STATE(7758)] = 328375,
  [SMALL_STATE(7759)] = 328393,
  [SMALL_STATE(7760)] = 328411,
  [SMALL_STATE(7761)] = 328429,
  [SMALL_STATE(7762)] = 328447,
  [SMALL_STATE(7763)] = 328465,
  [SMALL_STATE(7764)] = 328483,
  [SMALL_STATE(7765)] = 328501,
  [SMALL_STATE(7766)] = 328519,
  [SMALL_STATE(7767)] = 328537,
  [SMALL_STATE(7768)] = 328555,
  [SMALL_STATE(7769)] = 328573,
  [SMALL_STATE(7770)] = 328591,
  [SMALL_STATE(7771)] = 328609,
  [SMALL_STATE(7772)] = 328627,
  [SMALL_STATE(7773)] = 328645,
  [SMALL_STATE(7774)] = 328663,
  [SMALL_STATE(7775)] = 328681,
  [SMALL_STATE(7776)] = 328699,
  [SMALL_STATE(7777)] = 328717,
  [SMALL_STATE(7778)] = 328735,
  [SMALL_STATE(7779)] = 328753,
  [SMALL_STATE(7780)] = 328771,
  [SMALL_STATE(7781)] = 328789,
  [SMALL_STATE(7782)] = 328807,
  [SMALL_STATE(7783)] = 328825,
  [SMALL_STATE(7784)] = 328843,
  [SMALL_STATE(7785)] = 328861,
  [SMALL_STATE(7786)] = 328879,
  [SMALL_STATE(7787)] = 328897,
  [SMALL_STATE(7788)] = 328915,
  [SMALL_STATE(7789)] = 328933,
  [SMALL_STATE(7790)] = 328951,
  [SMALL_STATE(7791)] = 328969,
  [SMALL_STATE(7792)] = 328987,
  [SMALL_STATE(7793)] = 329005,
  [SMALL_STATE(7794)] = 329023,
  [SMALL_STATE(7795)] = 329041,
  [SMALL_STATE(7796)] = 329059,
  [SMALL_STATE(7797)] = 329077,
  [SMALL_STATE(7798)] = 329095,
  [SMALL_STATE(7799)] = 329113,
  [SMALL_STATE(7800)] = 329131,
  [SMALL_STATE(7801)] = 329149,
  [SMALL_STATE(7802)] = 329167,
  [SMALL_STATE(7803)] = 329185,
  [SMALL_STATE(7804)] = 329203,
  [SMALL_STATE(7805)] = 329221,
  [SMALL_STATE(7806)] = 329239,
  [SMALL_STATE(7807)] = 329257,
  [SMALL_STATE(7808)] = 329275,
  [SMALL_STATE(7809)] = 329293,
  [SMALL_STATE(7810)] = 329311,
  [SMALL_STATE(7811)] = 329329,
  [SMALL_STATE(7812)] = 329347,
  [SMALL_STATE(7813)] = 329365,
  [SMALL_STATE(7814)] = 329383,
  [SMALL_STATE(7815)] = 329401,
  [SMALL_STATE(7816)] = 329419,
  [SMALL_STATE(7817)] = 329437,
  [SMALL_STATE(7818)] = 329455,
  [SMALL_STATE(7819)] = 329473,
  [SMALL_STATE(7820)] = 329491,
  [SMALL_STATE(7821)] = 329509,
  [SMALL_STATE(7822)] = 329527,
  [SMALL_STATE(7823)] = 329545,
  [SMALL_STATE(7824)] = 329563,
  [SMALL_STATE(7825)] = 329581,
  [SMALL_STATE(7826)] = 329599,
  [SMALL_STATE(7827)] = 329617,
  [SMALL_STATE(7828)] = 329635,
  [SMALL_STATE(7829)] = 329653,
  [SMALL_STATE(7830)] = 329671,
  [SMALL_STATE(7831)] = 329689,
  [SMALL_STATE(7832)] = 329707,
  [SMALL_STATE(7833)] = 329725,
  [SMALL_STATE(7834)] = 329743,
  [SMALL_STATE(7835)] = 329761,
  [SMALL_STATE(7836)] = 329779,
  [SMALL_STATE(7837)] = 329797,
  [SMALL_STATE(7838)] = 329815,
  [SMALL_STATE(7839)] = 329833,
  [SMALL_STATE(7840)] = 329851,
  [SMALL_STATE(7841)] = 329869,
  [SMALL_STATE(7842)] = 329887,
  [SMALL_STATE(7843)] = 329905,
  [SMALL_STATE(7844)] = 329923,
  [SMALL_STATE(7845)] = 329941,
  [SMALL_STATE(7846)] = 329959,
  [SMALL_STATE(7847)] = 329977,
  [SMALL_STATE(7848)] = 329995,
  [SMALL_STATE(7849)] = 330013,
  [SMALL_STATE(7850)] = 330031,
  [SMALL_STATE(7851)] = 330049,
  [SMALL_STATE(7852)] = 330067,
  [SMALL_STATE(7853)] = 330085,
  [SMALL_STATE(7854)] = 330103,
  [SMALL_STATE(7855)] = 330121,
  [SMALL_STATE(7856)] = 330139,
  [SMALL_STATE(7857)] = 330157,
  [SMALL_STATE(7858)] = 330175,
  [SMALL_STATE(7859)] = 330193,
  [SMALL_STATE(7860)] = 330211,
  [SMALL_STATE(7861)] = 330229,
  [SMALL_STATE(7862)] = 330247,
  [SMALL_STATE(7863)] = 330265,
  [SMALL_STATE(7864)] = 330283,
  [SMALL_STATE(7865)] = 330301,
  [SMALL_STATE(7866)] = 330319,
  [SMALL_STATE(7867)] = 330337,
  [SMALL_STATE(7868)] = 330355,
  [SMALL_STATE(7869)] = 330373,
  [SMALL_STATE(7870)] = 330391,
  [SMALL_STATE(7871)] = 330409,
  [SMALL_STATE(7872)] = 330427,
  [SMALL_STATE(7873)] = 330445,
  [SMALL_STATE(7874)] = 330463,
  [SMALL_STATE(7875)] = 330481,
  [SMALL_STATE(7876)] = 330499,
  [SMALL_STATE(7877)] = 330517,
  [SMALL_STATE(7878)] = 330535,
  [SMALL_STATE(7879)] = 330553,
  [SMALL_STATE(7880)] = 330571,
  [SMALL_STATE(7881)] = 330589,
  [SMALL_STATE(7882)] = 330607,
  [SMALL_STATE(7883)] = 330625,
  [SMALL_STATE(7884)] = 330643,
  [SMALL_STATE(7885)] = 330661,
  [SMALL_STATE(7886)] = 330679,
  [SMALL_STATE(7887)] = 330697,
  [SMALL_STATE(7888)] = 330715,
  [SMALL_STATE(7889)] = 330733,
  [SMALL_STATE(7890)] = 330751,
  [SMALL_STATE(7891)] = 330769,
  [SMALL_STATE(7892)] = 330787,
  [SMALL_STATE(7893)] = 330805,
  [SMALL_STATE(7894)] = 330823,
  [SMALL_STATE(7895)] = 330841,
  [SMALL_STATE(7896)] = 330859,
  [SMALL_STATE(7897)] = 330877,
  [SMALL_STATE(7898)] = 330895,
  [SMALL_STATE(7899)] = 330913,
  [SMALL_STATE(7900)] = 330931,
  [SMALL_STATE(7901)] = 330949,
  [SMALL_STATE(7902)] = 330967,
  [SMALL_STATE(7903)] = 330985,
  [SMALL_STATE(7904)] = 331003,
  [SMALL_STATE(7905)] = 331021,
  [SMALL_STATE(7906)] = 331039,
  [SMALL_STATE(7907)] = 331057,
  [SMALL_STATE(7908)] = 331075,
  [SMALL_STATE(7909)] = 331093,
  [SMALL_STATE(7910)] = 331111,
  [SMALL_STATE(7911)] = 331129,
  [SMALL_STATE(7912)] = 331147,
  [SMALL_STATE(7913)] = 331165,
  [SMALL_STATE(7914)] = 331183,
  [SMALL_STATE(7915)] = 331201,
  [SMALL_STATE(7916)] = 331219,
  [SMALL_STATE(7917)] = 331237,
  [SMALL_STATE(7918)] = 331255,
  [SMALL_STATE(7919)] = 331273,
  [SMALL_STATE(7920)] = 331291,
  [SMALL_STATE(7921)] = 331309,
  [SMALL_STATE(7922)] = 331327,
  [SMALL_STATE(7923)] = 331345,
  [SMALL_STATE(7924)] = 331363,
  [SMALL_STATE(7925)] = 331381,
  [SMALL_STATE(7926)] = 331399,
  [SMALL_STATE(7927)] = 331417,
  [SMALL_STATE(7928)] = 331435,
  [SMALL_STATE(7929)] = 331453,
  [SMALL_STATE(7930)] = 331471,
  [SMALL_STATE(7931)] = 331489,
  [SMALL_STATE(7932)] = 331507,
  [SMALL_STATE(7933)] = 331525,
  [SMALL_STATE(7934)] = 331543,
  [SMALL_STATE(7935)] = 331561,
  [SMALL_STATE(7936)] = 331579,
  [SMALL_STATE(7937)] = 331597,
  [SMALL_STATE(7938)] = 331615,
  [SMALL_STATE(7939)] = 331633,
  [SMALL_STATE(7940)] = 331651,
  [SMALL_STATE(7941)] = 331669,
  [SMALL_STATE(7942)] = 331687,
  [SMALL_STATE(7943)] = 331705,
  [SMALL_STATE(7944)] = 331723,
  [SMALL_STATE(7945)] = 331741,
  [SMALL_STATE(7946)] = 331759,
  [SMALL_STATE(7947)] = 331777,
  [SMALL_STATE(7948)] = 331795,
  [SMALL_STATE(7949)] = 331813,
  [SMALL_STATE(7950)] = 331831,
  [SMALL_STATE(7951)] = 331849,
  [SMALL_STATE(7952)] = 331867,
  [SMALL_STATE(7953)] = 331885,
  [SMALL_STATE(7954)] = 331903,
  [SMALL_STATE(7955)] = 331921,
  [SMALL_STATE(7956)] = 331939,
  [SMALL_STATE(7957)] = 331957,
  [SMALL_STATE(7958)] = 331975,
  [SMALL_STATE(7959)] = 331993,
  [SMALL_STATE(7960)] = 332011,
  [SMALL_STATE(7961)] = 332029,
  [SMALL_STATE(7962)] = 332047,
  [SMALL_STATE(7963)] = 332065,
  [SMALL_STATE(7964)] = 332083,
  [SMALL_STATE(7965)] = 332101,
  [SMALL_STATE(7966)] = 332119,
  [SMALL_STATE(7967)] = 332137,
  [SMALL_STATE(7968)] = 332155,
  [SMALL_STATE(7969)] = 332173,
  [SMALL_STATE(7970)] = 332191,
  [SMALL_STATE(7971)] = 332209,
  [SMALL_STATE(7972)] = 332227,
  [SMALL_STATE(7973)] = 332245,
  [SMALL_STATE(7974)] = 332263,
  [SMALL_STATE(7975)] = 332281,
  [SMALL_STATE(7976)] = 332299,
  [SMALL_STATE(7977)] = 332317,
  [SMALL_STATE(7978)] = 332335,
  [SMALL_STATE(7979)] = 332353,
  [SMALL_STATE(7980)] = 332371,
  [SMALL_STATE(7981)] = 332389,
  [SMALL_STATE(7982)] = 332407,
  [SMALL_STATE(7983)] = 332425,
  [SMALL_STATE(7984)] = 332443,
  [SMALL_STATE(7985)] = 332461,
  [SMALL_STATE(7986)] = 332479,
  [SMALL_STATE(7987)] = 332497,
  [SMALL_STATE(7988)] = 332515,
  [SMALL_STATE(7989)] = 332533,
  [SMALL_STATE(7990)] = 332551,
  [SMALL_STATE(7991)] = 332569,
  [SMALL_STATE(7992)] = 332587,
  [SMALL_STATE(7993)] = 332605,
  [SMALL_STATE(7994)] = 332623,
  [SMALL_STATE(7995)] = 332641,
  [SMALL_STATE(7996)] = 332659,
  [SMALL_STATE(7997)] = 332677,
  [SMALL_STATE(7998)] = 332695,
  [SMALL_STATE(7999)] = 332713,
  [SMALL_STATE(8000)] = 332731,
  [SMALL_STATE(8001)] = 332749,
  [SMALL_STATE(8002)] = 332767,
  [SMALL_STATE(8003)] = 332785,
  [SMALL_STATE(8004)] = 332803,
  [SMALL_STATE(8005)] = 332821,
  [SMALL_STATE(8006)] = 332839,
  [SMALL_STATE(8007)] = 332857,
  [SMALL_STATE(8008)] = 332875,
  [SMALL_STATE(8009)] = 332893,
  [SMALL_STATE(8010)] = 332911,
  [SMALL_STATE(8011)] = 332929,
  [SMALL_STATE(8012)] = 332947,
  [SMALL_STATE(8013)] = 332965,
  [SMALL_STATE(8014)] = 332983,
  [SMALL_STATE(8015)] = 333001,
  [SMALL_STATE(8016)] = 333019,
  [SMALL_STATE(8017)] = 333037,
  [SMALL_STATE(8018)] = 333055,
  [SMALL_STATE(8019)] = 333073,
  [SMALL_STATE(8020)] = 333091,
  [SMALL_STATE(8021)] = 333109,
  [SMALL_STATE(8022)] = 333127,
  [SMALL_STATE(8023)] = 333145,
  [SMALL_STATE(8024)] = 333163,
  [SMALL_STATE(8025)] = 333181,
  [SMALL_STATE(8026)] = 333199,
  [SMALL_STATE(8027)] = 333217,
  [SMALL_STATE(8028)] = 333235,
  [SMALL_STATE(8029)] = 333253,
  [SMALL_STATE(8030)] = 333271,
  [SMALL_STATE(8031)] = 333289,
  [SMALL_STATE(8032)] = 333307,
  [SMALL_STATE(8033)] = 333325,
  [SMALL_STATE(8034)] = 333343,
  [SMALL_STATE(8035)] = 333361,
  [SMALL_STATE(8036)] = 333379,
  [SMALL_STATE(8037)] = 333397,
  [SMALL_STATE(8038)] = 333415,
  [SMALL_STATE(8039)] = 333433,
  [SMALL_STATE(8040)] = 333451,
  [SMALL_STATE(8041)] = 333469,
  [SMALL_STATE(8042)] = 333487,
  [SMALL_STATE(8043)] = 333505,
  [SMALL_STATE(8044)] = 333523,
  [SMALL_STATE(8045)] = 333541,
  [SMALL_STATE(8046)] = 333559,
  [SMALL_STATE(8047)] = 333577,
  [SMALL_STATE(8048)] = 333595,
  [SMALL_STATE(8049)] = 333613,
  [SMALL_STATE(8050)] = 333631,
  [SMALL_STATE(8051)] = 333649,
  [SMALL_STATE(8052)] = 333667,
  [SMALL_STATE(8053)] = 333685,
  [SMALL_STATE(8054)] = 333703,
  [SMALL_STATE(8055)] = 333721,
  [SMALL_STATE(8056)] = 333739,
  [SMALL_STATE(8057)] = 333757,
  [SMALL_STATE(8058)] = 333775,
  [SMALL_STATE(8059)] = 333793,
  [SMALL_STATE(8060)] = 333811,
  [SMALL_STATE(8061)] = 333829,
  [SMALL_STATE(8062)] = 333847,
  [SMALL_STATE(8063)] = 333865,
  [SMALL_STATE(8064)] = 333883,
  [SMALL_STATE(8065)] = 333901,
  [SMALL_STATE(8066)] = 333919,
  [SMALL_STATE(8067)] = 333937,
  [SMALL_STATE(8068)] = 333955,
  [SMALL_STATE(8069)] = 333973,
  [SMALL_STATE(8070)] = 333991,
  [SMALL_STATE(8071)] = 334009,
  [SMALL_STATE(8072)] = 334027,
  [SMALL_STATE(8073)] = 334045,
  [SMALL_STATE(8074)] = 334063,
  [SMALL_STATE(8075)] = 334081,
  [SMALL_STATE(8076)] = 334099,
  [SMALL_STATE(8077)] = 334117,
  [SMALL_STATE(8078)] = 334135,
  [SMALL_STATE(8079)] = 334153,
  [SMALL_STATE(8080)] = 334171,
  [SMALL_STATE(8081)] = 334189,
  [SMALL_STATE(8082)] = 334207,
  [SMALL_STATE(8083)] = 334225,
  [SMALL_STATE(8084)] = 334243,
  [SMALL_STATE(8085)] = 334261,
  [SMALL_STATE(8086)] = 334279,
  [SMALL_STATE(8087)] = 334297,
  [SMALL_STATE(8088)] = 334315,
  [SMALL_STATE(8089)] = 334333,
  [SMALL_STATE(8090)] = 334351,
  [SMALL_STATE(8091)] = 334369,
  [SMALL_STATE(8092)] = 334387,
  [SMALL_STATE(8093)] = 334405,
  [SMALL_STATE(8094)] = 334423,
  [SMALL_STATE(8095)] = 334441,
  [SMALL_STATE(8096)] = 334459,
  [SMALL_STATE(8097)] = 334477,
  [SMALL_STATE(8098)] = 334495,
  [SMALL_STATE(8099)] = 334513,
  [SMALL_STATE(8100)] = 334531,
  [SMALL_STATE(8101)] = 334549,
  [SMALL_STATE(8102)] = 334567,
  [SMALL_STATE(8103)] = 334585,
  [SMALL_STATE(8104)] = 334603,
  [SMALL_STATE(8105)] = 334621,
  [SMALL_STATE(8106)] = 334639,
  [SMALL_STATE(8107)] = 334657,
  [SMALL_STATE(8108)] = 334675,
  [SMALL_STATE(8109)] = 334693,
  [SMALL_STATE(8110)] = 334711,
  [SMALL_STATE(8111)] = 334729,
  [SMALL_STATE(8112)] = 334747,
  [SMALL_STATE(8113)] = 334765,
  [SMALL_STATE(8114)] = 334783,
  [SMALL_STATE(8115)] = 334801,
  [SMALL_STATE(8116)] = 334819,
  [SMALL_STATE(8117)] = 334837,
  [SMALL_STATE(8118)] = 334855,
  [SMALL_STATE(8119)] = 334873,
  [SMALL_STATE(8120)] = 334891,
  [SMALL_STATE(8121)] = 334909,
  [SMALL_STATE(8122)] = 334927,
  [SMALL_STATE(8123)] = 334945,
  [SMALL_STATE(8124)] = 334963,
  [SMALL_STATE(8125)] = 334981,
  [SMALL_STATE(8126)] = 334999,
  [SMALL_STATE(8127)] = 335017,
  [SMALL_STATE(8128)] = 335035,
  [SMALL_STATE(8129)] = 335053,
  [SMALL_STATE(8130)] = 335071,
  [SMALL_STATE(8131)] = 335089,
  [SMALL_STATE(8132)] = 335107,
  [SMALL_STATE(8133)] = 335125,
  [SMALL_STATE(8134)] = 335143,
  [SMALL_STATE(8135)] = 335161,
  [SMALL_STATE(8136)] = 335179,
  [SMALL_STATE(8137)] = 335197,
  [SMALL_STATE(8138)] = 335215,
  [SMALL_STATE(8139)] = 335233,
  [SMALL_STATE(8140)] = 335251,
  [SMALL_STATE(8141)] = 335269,
  [SMALL_STATE(8142)] = 335287,
  [SMALL_STATE(8143)] = 335305,
  [SMALL_STATE(8144)] = 335323,
  [SMALL_STATE(8145)] = 335341,
  [SMALL_STATE(8146)] = 335359,
  [SMALL_STATE(8147)] = 335377,
  [SMALL_STATE(8148)] = 335395,
  [SMALL_STATE(8149)] = 335413,
  [SMALL_STATE(8150)] = 335431,
  [SMALL_STATE(8151)] = 335449,
  [SMALL_STATE(8152)] = 335467,
  [SMALL_STATE(8153)] = 335485,
  [SMALL_STATE(8154)] = 335503,
  [SMALL_STATE(8155)] = 335521,
  [SMALL_STATE(8156)] = 335539,
  [SMALL_STATE(8157)] = 335557,
  [SMALL_STATE(8158)] = 335575,
  [SMALL_STATE(8159)] = 335593,
  [SMALL_STATE(8160)] = 335611,
  [SMALL_STATE(8161)] = 335629,
  [SMALL_STATE(8162)] = 335647,
  [SMALL_STATE(8163)] = 335665,
  [SMALL_STATE(8164)] = 335683,
  [SMALL_STATE(8165)] = 335701,
  [SMALL_STATE(8166)] = 335719,
  [SMALL_STATE(8167)] = 335737,
  [SMALL_STATE(8168)] = 335755,
  [SMALL_STATE(8169)] = 335773,
  [SMALL_STATE(8170)] = 335791,
  [SMALL_STATE(8171)] = 335809,
  [SMALL_STATE(8172)] = 335827,
  [SMALL_STATE(8173)] = 335845,
  [SMALL_STATE(8174)] = 335863,
  [SMALL_STATE(8175)] = 335881,
  [SMALL_STATE(8176)] = 335899,
  [SMALL_STATE(8177)] = 335917,
  [SMALL_STATE(8178)] = 335935,
  [SMALL_STATE(8179)] = 335953,
  [SMALL_STATE(8180)] = 335971,
  [SMALL_STATE(8181)] = 335989,
  [SMALL_STATE(8182)] = 336007,
  [SMALL_STATE(8183)] = 336025,
  [SMALL_STATE(8184)] = 336043,
  [SMALL_STATE(8185)] = 336061,
  [SMALL_STATE(8186)] = 336079,
  [SMALL_STATE(8187)] = 336097,
  [SMALL_STATE(8188)] = 336115,
  [SMALL_STATE(8189)] = 336133,
  [SMALL_STATE(8190)] = 336151,
  [SMALL_STATE(8191)] = 336169,
  [SMALL_STATE(8192)] = 336187,
  [SMALL_STATE(8193)] = 336205,
  [SMALL_STATE(8194)] = 336223,
  [SMALL_STATE(8195)] = 336241,
  [SMALL_STATE(8196)] = 336259,
  [SMALL_STATE(8197)] = 336277,
  [SMALL_STATE(8198)] = 336295,
  [SMALL_STATE(8199)] = 336313,
  [SMALL_STATE(8200)] = 336331,
  [SMALL_STATE(8201)] = 336349,
  [SMALL_STATE(8202)] = 336367,
  [SMALL_STATE(8203)] = 336385,
  [SMALL_STATE(8204)] = 336403,
  [SMALL_STATE(8205)] = 336421,
  [SMALL_STATE(8206)] = 336439,
  [SMALL_STATE(8207)] = 336457,
  [SMALL_STATE(8208)] = 336475,
  [SMALL_STATE(8209)] = 336493,
  [SMALL_STATE(8210)] = 336511,
  [SMALL_STATE(8211)] = 336529,
  [SMALL_STATE(8212)] = 336547,
  [SMALL_STATE(8213)] = 336565,
  [SMALL_STATE(8214)] = 336583,
  [SMALL_STATE(8215)] = 336601,
  [SMALL_STATE(8216)] = 336619,
  [SMALL_STATE(8217)] = 336637,
  [SMALL_STATE(8218)] = 336655,
  [SMALL_STATE(8219)] = 336673,
  [SMALL_STATE(8220)] = 336691,
  [SMALL_STATE(8221)] = 336709,
  [SMALL_STATE(8222)] = 336727,
  [SMALL_STATE(8223)] = 336745,
  [SMALL_STATE(8224)] = 336763,
  [SMALL_STATE(8225)] = 336781,
  [SMALL_STATE(8226)] = 336799,
  [SMALL_STATE(8227)] = 336817,
  [SMALL_STATE(8228)] = 336835,
  [SMALL_STATE(8229)] = 336853,
  [SMALL_STATE(8230)] = 336871,
  [SMALL_STATE(8231)] = 336889,
  [SMALL_STATE(8232)] = 336907,
  [SMALL_STATE(8233)] = 336925,
  [SMALL_STATE(8234)] = 336943,
  [SMALL_STATE(8235)] = 336961,
  [SMALL_STATE(8236)] = 336979,
  [SMALL_STATE(8237)] = 336997,
  [SMALL_STATE(8238)] = 337015,
  [SMALL_STATE(8239)] = 337033,
  [SMALL_STATE(8240)] = 337051,
  [SMALL_STATE(8241)] = 337069,
  [SMALL_STATE(8242)] = 337087,
  [SMALL_STATE(8243)] = 337105,
  [SMALL_STATE(8244)] = 337123,
  [SMALL_STATE(8245)] = 337141,
  [SMALL_STATE(8246)] = 337159,
  [SMALL_STATE(8247)] = 337177,
  [SMALL_STATE(8248)] = 337195,
  [SMALL_STATE(8249)] = 337213,
  [SMALL_STATE(8250)] = 337231,
  [SMALL_STATE(8251)] = 337249,
  [SMALL_STATE(8252)] = 337267,
  [SMALL_STATE(8253)] = 337285,
  [SMALL_STATE(8254)] = 337303,
  [SMALL_STATE(8255)] = 337321,
  [SMALL_STATE(8256)] = 337339,
  [SMALL_STATE(8257)] = 337357,
  [SMALL_STATE(8258)] = 337375,
  [SMALL_STATE(8259)] = 337393,
  [SMALL_STATE(8260)] = 337411,
  [SMALL_STATE(8261)] = 337429,
  [SMALL_STATE(8262)] = 337447,
  [SMALL_STATE(8263)] = 337465,
  [SMALL_STATE(8264)] = 337483,
  [SMALL_STATE(8265)] = 337501,
  [SMALL_STATE(8266)] = 337519,
  [SMALL_STATE(8267)] = 337537,
  [SMALL_STATE(8268)] = 337555,
  [SMALL_STATE(8269)] = 337573,
  [SMALL_STATE(8270)] = 337591,
  [SMALL_STATE(8271)] = 337609,
  [SMALL_STATE(8272)] = 337627,
  [SMALL_STATE(8273)] = 337645,
  [SMALL_STATE(8274)] = 337663,
  [SMALL_STATE(8275)] = 337681,
  [SMALL_STATE(8276)] = 337699,
  [SMALL_STATE(8277)] = 337717,
  [SMALL_STATE(8278)] = 337735,
  [SMALL_STATE(8279)] = 337753,
  [SMALL_STATE(8280)] = 337771,
  [SMALL_STATE(8281)] = 337789,
  [SMALL_STATE(8282)] = 337807,
  [SMALL_STATE(8283)] = 337825,
  [SMALL_STATE(8284)] = 337843,
  [SMALL_STATE(8285)] = 337861,
  [SMALL_STATE(8286)] = 337879,
  [SMALL_STATE(8287)] = 337897,
  [SMALL_STATE(8288)] = 337915,
  [SMALL_STATE(8289)] = 337933,
  [SMALL_STATE(8290)] = 337951,
  [SMALL_STATE(8291)] = 337969,
  [SMALL_STATE(8292)] = 337987,
  [SMALL_STATE(8293)] = 338005,
  [SMALL_STATE(8294)] = 338023,
  [SMALL_STATE(8295)] = 338041,
  [SMALL_STATE(8296)] = 338059,
  [SMALL_STATE(8297)] = 338077,
  [SMALL_STATE(8298)] = 338095,
  [SMALL_STATE(8299)] = 338113,
  [SMALL_STATE(8300)] = 338131,
  [SMALL_STATE(8301)] = 338149,
  [SMALL_STATE(8302)] = 338167,
  [SMALL_STATE(8303)] = 338185,
  [SMALL_STATE(8304)] = 338203,
  [SMALL_STATE(8305)] = 338221,
  [SMALL_STATE(8306)] = 338239,
  [SMALL_STATE(8307)] = 338257,
  [SMALL_STATE(8308)] = 338275,
  [SMALL_STATE(8309)] = 338293,
  [SMALL_STATE(8310)] = 338311,
  [SMALL_STATE(8311)] = 338329,
  [SMALL_STATE(8312)] = 338347,
  [SMALL_STATE(8313)] = 338365,
  [SMALL_STATE(8314)] = 338383,
  [SMALL_STATE(8315)] = 338401,
  [SMALL_STATE(8316)] = 338419,
  [SMALL_STATE(8317)] = 338437,
  [SMALL_STATE(8318)] = 338455,
  [SMALL_STATE(8319)] = 338473,
  [SMALL_STATE(8320)] = 338491,
  [SMALL_STATE(8321)] = 338509,
  [SMALL_STATE(8322)] = 338527,
  [SMALL_STATE(8323)] = 338545,
  [SMALL_STATE(8324)] = 338563,
  [SMALL_STATE(8325)] = 338581,
  [SMALL_STATE(8326)] = 338599,
  [SMALL_STATE(8327)] = 338617,
  [SMALL_STATE(8328)] = 338635,
  [SMALL_STATE(8329)] = 338653,
  [SMALL_STATE(8330)] = 338671,
  [SMALL_STATE(8331)] = 338689,
  [SMALL_STATE(8332)] = 338707,
  [SMALL_STATE(8333)] = 338725,
  [SMALL_STATE(8334)] = 338743,
  [SMALL_STATE(8335)] = 338761,
  [SMALL_STATE(8336)] = 338779,
  [SMALL_STATE(8337)] = 338797,
  [SMALL_STATE(8338)] = 338815,
  [SMALL_STATE(8339)] = 338833,
  [SMALL_STATE(8340)] = 338851,
  [SMALL_STATE(8341)] = 338869,
  [SMALL_STATE(8342)] = 338887,
  [SMALL_STATE(8343)] = 338905,
  [SMALL_STATE(8344)] = 338923,
  [SMALL_STATE(8345)] = 338941,
  [SMALL_STATE(8346)] = 338959,
  [SMALL_STATE(8347)] = 338977,
  [SMALL_STATE(8348)] = 338995,
  [SMALL_STATE(8349)] = 339013,
  [SMALL_STATE(8350)] = 339031,
  [SMALL_STATE(8351)] = 339049,
  [SMALL_STATE(8352)] = 339067,
  [SMALL_STATE(8353)] = 339085,
  [SMALL_STATE(8354)] = 339103,
  [SMALL_STATE(8355)] = 339121,
  [SMALL_STATE(8356)] = 339139,
  [SMALL_STATE(8357)] = 339157,
  [SMALL_STATE(8358)] = 339175,
  [SMALL_STATE(8359)] = 339193,
  [SMALL_STATE(8360)] = 339211,
  [SMALL_STATE(8361)] = 339229,
  [SMALL_STATE(8362)] = 339247,
  [SMALL_STATE(8363)] = 339251,
  [SMALL_STATE(8364)] = 339255,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3077),
  [5] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7093),
  [7] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7003),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5672),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4552),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4918),
  [17] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5294),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(655),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3617),
  [23] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5198),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5263),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5063),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4790),
  [31] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3363),
  [33] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5276),
  [35] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5055),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4890),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5005),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5157),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7097),
  [45] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4923),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5115),
  [49] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3002),
  [51] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5033),
  [53] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1182),
  [55] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5100),
  [57] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7029),
  [59] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5148),
  [61] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5272),
  [63] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4761),
  [65] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5073),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5211),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(652),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4362),
  [73] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7116),
  [75] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(224),
  [78] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5672),
  [81] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4552),
  [84] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4918),
  [87] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5294),
  [90] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(655),
  [93] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(3617),
  [96] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5198),
  [99] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5263),
  [102] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5063),
  [105] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4790),
  [108] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(3363),
  [111] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5276),
  [114] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5055),
  [117] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4890),
  [120] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5005),
  [123] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5157),
  [126] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(7097),
  [129] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4923),
  [132] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5115),
  [135] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(3002),
  [138] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5033),
  [141] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(1182),
  [144] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5100),
  [147] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(7029),
  [150] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5148),
  [153] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5272),
  [156] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4761),
  [159] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5073),
  [162] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(5211),
  [165] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(652),
  [168] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0), SHIFT_REPEAT(4362),
  [171] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 2, 0, 0),
  [173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7484),
  [175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4953),
  [177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3655),
  [179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1041),
  [181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5197),
  [183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5127),
  [185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3430),
  [187] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generate_body, 1, 0, 0),
  [189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5132),
  [191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4772),
  [193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5045),
  [195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5324),
  [197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5134),
  [199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2996),
  [201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1225),
  [203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5107),
  [205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8277),
  [207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5242),
  [209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5320),
  [211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4735),
  [213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5136),
  [215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5322),
  [217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_generate_body, 1, 0, 0),
  [221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3004),
  [223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1231),
  [225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1335),
  [227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3007),
  [229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1212),
  [231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1651),
  [233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(14),
  [237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [239] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elsif_statement, 3, 0, 0),
  [241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4911),
  [243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5078),
  [245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4912),
  [251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7494),
  [253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(663),
  [255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5011),
  [259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(633),
  [261] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 3, 0, 0),
  [263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2629),
  [269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5993),
  [271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2676),
  [275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4133),
  [277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2089),
  [279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5482),
  [281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7069),
  [283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1044),
  [285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6484),
  [287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6832),
  [291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8062),
  [293] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement_body, 1, 0, 0),
  [295] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(224),
  [298] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(5672),
  [301] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(4552),
  [304] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(655),
  [307] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1651),
  [310] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(396),
  [313] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0),
  [315] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(4911),
  [318] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(5078),
  [321] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(591),
  [324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(23),
  [327] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(4912),
  [330] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(7494),
  [333] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(663),
  [336] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(234),
  [339] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(5011),
  [342] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(633),
  [345] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(652),
  [348] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(4362),
  [351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1657),
  [353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement_alternative, 3, 0, 0),
  [355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4948),
  [357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4896),
  [359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8088),
  [361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4969),
  [365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1454),
  [367] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_statement, 1, 0, 0),
  [369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4922),
  [371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4915),
  [373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7326),
  [375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4961),
  [379] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement_alternative, 2, 0, 0),
  [381] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1657),
  [384] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(4948),
  [387] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(4896),
  [390] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(8088),
  [393] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(227),
  [396] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(4969),
  [399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4585),
  [401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5957),
  [403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(942),
  [405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6048),
  [409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(534),
  [415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5962),
  [417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1238),
  [419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2542),
  [423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4149),
  [425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2506),
  [427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6477),
  [431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1627),
  [433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6556),
  [435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1411),
  [437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7010),
  [439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6673),
  [441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5545),
  [445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [447] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0),
  [449] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(4918),
  [452] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(3617),
  [455] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5198),
  [458] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5063),
  [461] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(3363),
  [464] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5055),
  [467] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(4890),
  [470] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5005),
  [473] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5157),
  [476] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(7097),
  [479] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5115),
  [482] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5033),
  [485] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5100),
  [488] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(7029),
  [491] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5148),
  [494] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5272),
  [497] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(4761),
  [500] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5073),
  [503] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5211),
  [506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loop_body, 1, 0, 0),
  [510] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequential_block, 2, 0, 0),
  [512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequential_block, 1, 0, 0),
  [516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loop_body, 2, 0, 0),
  [520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1566),
  [522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2147),
  [524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4859),
  [528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4753),
  [532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1420),
  [534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1988),
  [536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3817),
  [538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4777),
  [540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1792),
  [542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3291),
  [546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4988),
  [548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3528),
  [550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4727),
  [552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2092),
  [554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2120),
  [556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2155),
  [558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3407),
  [560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1743),
  [562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2235),
  [566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4950),
  [568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4443),
  [570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4342),
  [572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4555),
  [574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1705),
  [576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4484),
  [578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1343),
  [580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1360),
  [582] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(1454),
  [585] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(4922),
  [588] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(4915),
  [591] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(7326),
  [594] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(230),
  [597] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 2, 0, 0), SHIFT_REPEAT(4961),
  [600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1508),
  [602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1227),
  [604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3374),
  [606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1563),
  [608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(912),
  [612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1811),
  [614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1849),
  [616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4657),
  [618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1652),
  [620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4730),
  [622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4464),
  [624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1247),
  [626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1006),
  [628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1222),
  [630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1286),
  [632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4643),
  [634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4520),
  [636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5544),
  [644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5548),
  [650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5553),
  [662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5554),
  [668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(114),
  [670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_design_unit, 1, 0, 0),
  [674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5492),
  [684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1482),
  [686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [702] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(224),
  [705] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(5672),
  [708] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(4552),
  [711] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0),
  [713] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(655),
  [716] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(3002),
  [719] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(1182),
  [722] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(652),
  [725] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(4362),
  [728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5959),
  [740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6810),
  [744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4142),
  [746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2380),
  [748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6811),
  [750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1229),
  [752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6604),
  [754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1077),
  [756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6941),
  [758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8354),
  [760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(550),
  [764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5977),
  [766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1189),
  [768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4238),
  [770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1612),
  [772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6562),
  [776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(838),
  [778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7161),
  [780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7406),
  [782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6002),
  [784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1188),
  [786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1192),
  [788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2914),
  [790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6295),
  [792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2884),
  [796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2862),
  [798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2593),
  [800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2978),
  [802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2834),
  [804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2994),
  [806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2620),
  [808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(886),
  [810] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(145),
  [813] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(534),
  [816] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(2629),
  [819] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(5962),
  [822] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(1238),
  [825] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(2676),
  [828] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(4149),
  [831] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(2506),
  [834] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(6477),
  [837] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(1627),
  [840] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(6556),
  [843] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(1411),
  [846] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(7010),
  [849] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(6673),
  [852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(9),
  [854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5985),
  [858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1414),
  [860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4268),
  [862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2530),
  [864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1615),
  [866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6168),
  [868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1336),
  [870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7197),
  [872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8037),
  [874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5987),
  [882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1560),
  [884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4006),
  [886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2559),
  [888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1675),
  [890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6242),
  [892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1475),
  [894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7221),
  [896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8340),
  [898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(600),
  [902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5992),
  [904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1024),
  [906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4123),
  [908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2366),
  [910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1236),
  [912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6398),
  [914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1081),
  [916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7269),
  [918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6823),
  [920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5963),
  [924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1910),
  [926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4167),
  [928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2756),
  [930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2422),
  [932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6039),
  [934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1975),
  [936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7049),
  [938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7357),
  [940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5986),
  [946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1417),
  [948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4298),
  [950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2496),
  [952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1642),
  [954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6205),
  [956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1353),
  [958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7209),
  [960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8201),
  [962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(555),
  [966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5988),
  [968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1422),
  [970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4021),
  [972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2554),
  [974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1680),
  [976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6265),
  [978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1499),
  [980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7233),
  [982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6654),
  [984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(613),
  [990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4144),
  [992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2452),
  [994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1257),
  [996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6537),
  [998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7305),
  [1000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6953),
  [1002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [1004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [1006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6001),
  [1008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1042),
  [1010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4154),
  [1012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2429),
  [1014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1154),
  [1016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6083),
  [1018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1116),
  [1020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7353),
  [1022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7162),
  [1024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [1026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [1028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5989),
  [1030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1177),
  [1032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4084),
  [1034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2436),
  [1036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1326),
  [1038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6312),
  [1040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1184),
  [1042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7245),
  [1044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6720),
  [1046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [1048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [1050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5991),
  [1052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1438),
  [1054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4118),
  [1056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2572),
  [1058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1709),
  [1060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6345),
  [1062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1553),
  [1064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7257),
  [1066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6769),
  [1068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [1070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5967),
  [1072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [1074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4198),
  [1076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1943),
  [1078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [1080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6076),
  [1082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(939),
  [1084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7098),
  [1086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6789),
  [1088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(150),
  [1090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(664),
  [1092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5972),
  [1094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2271),
  [1096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4209),
  [1098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2758),
  [1100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2304),
  [1102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6185),
  [1104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2082),
  [1106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7112),
  [1108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6905),
  [1110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(657),
  [1112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1776),
  [1114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7676),
  [1116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(744),
  [1118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7677),
  [1120] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(137),
  [1123] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(685),
  [1126] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(2629),
  [1129] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(5959),
  [1132] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(1003),
  [1135] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(2676),
  [1138] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(4142),
  [1141] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(2380),
  [1144] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(1229),
  [1147] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(6604),
  [1150] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(1077),
  [1153] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(6941),
  [1156] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 2, 0, 0), SHIFT_REPEAT(8354),
  [1159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [1161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [1163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5955),
  [1165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1605),
  [1167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4059),
  [1169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2610),
  [1171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1730),
  [1173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6036),
  [1175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1550),
  [1177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6870),
  [1179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6887),
  [1181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [1183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(560),
  [1185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5973),
  [1187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2279),
  [1189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4212),
  [1191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2709),
  [1193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2337),
  [1195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6260),
  [1197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2106),
  [1199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7125),
  [1201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7004),
  [1203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [1205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5974),
  [1207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2282),
  [1209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4218),
  [1211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2737),
  [1213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2354),
  [1215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6361),
  [1217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2137),
  [1219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7137),
  [1221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7148),
  [1223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [1225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [1227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5976),
  [1229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1676),
  [1231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4233),
  [1233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2648),
  [1235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2019),
  [1237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6503),
  [1239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1876),
  [1241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7149),
  [1243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7285),
  [1245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2072),
  [1247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [1249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [1251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [1253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5978),
  [1255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1945),
  [1257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4250),
  [1259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2735),
  [1261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2296),
  [1263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6063),
  [1265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2222),
  [1267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7173),
  [1269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7594),
  [1271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [1273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(579),
  [1275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5983),
  [1277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1799),
  [1279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4257),
  [1281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2683),
  [1283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2061),
  [1285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6123),
  [1287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1825),
  [1289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7185),
  [1291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7749),
  [1293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [1295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [1297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(617),
  [1299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(621),
  [1301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(625),
  [1303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [1305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [1307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [1309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5994),
  [1311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1900),
  [1313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4131),
  [1315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2651),
  [1317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2207),
  [1319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6492),
  [1321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1796),
  [1323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7281),
  [1325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6864),
  [1327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [1329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5995),
  [1331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1781),
  [1333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4137),
  [1335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2631),
  [1337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2224),
  [1339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6516),
  [1341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1838),
  [1343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7293),
  [1345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6910),
  [1347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [1349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [1351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5997),
  [1353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1669),
  [1355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4145),
  [1357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2570),
  [1359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1683),
  [1361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6583),
  [1363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1661),
  [1365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7317),
  [1367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6992),
  [1369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [1371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6000),
  [1373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1319),
  [1375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4148),
  [1377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2513),
  [1379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1467),
  [1381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6028),
  [1383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1240),
  [1385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7329),
  [1387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7035),
  [1389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [1391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4152),
  [1393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2598),
  [1395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1887),
  [1397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6062),
  [1399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7341),
  [1401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7105),
  [1403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [1405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [1407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6003),
  [1409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1322),
  [1411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4155),
  [1413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2477),
  [1415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1445),
  [1417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6102),
  [1419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1280),
  [1421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7365),
  [1423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7217),
  [1425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [1427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [1429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5966),
  [1431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1904),
  [1433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4182),
  [1435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2640),
  [1437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1927),
  [1439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6395),
  [1441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1695),
  [1443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7077),
  [1445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8210),
  [1447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [1449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [1451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5956),
  [1453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2007),
  [1455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4052),
  [1457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2739),
  [1459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2357),
  [1461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6319),
  [1463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2087),
  [1465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6706),
  [1467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7454),
  [1469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [1471] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_name, 1, 0, 0),
  [1473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5214),
  [1475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [1477] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_name, 1, 0, 0),
  [1479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3452),
  [1481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3602),
  [1483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7307),
  [1485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7284),
  [1487] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_name, 2, 0, 0),
  [1489] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_name, 2, 0, 0),
  [1491] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0),
  [1493] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5214),
  [1496] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(51),
  [1499] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0),
  [1501] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3452),
  [1504] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3602),
  [1507] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7307),
  [1510] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7284),
  [1513] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5305),
  [1516] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(67),
  [1519] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3510),
  [1522] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3608),
  [1525] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7510),
  [1528] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7467),
  [1531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5305),
  [1533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [1535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3510),
  [1537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3608),
  [1539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7510),
  [1541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7467),
  [1543] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5227),
  [1546] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(41),
  [1549] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3558),
  [1552] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3622),
  [1555] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7187),
  [1558] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7164),
  [1561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5227),
  [1563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [1565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3558),
  [1567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3622),
  [1569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7187),
  [1571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7164),
  [1573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5190),
  [1575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [1577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3499),
  [1579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3635),
  [1581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7013),
  [1583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6945),
  [1585] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5190),
  [1588] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(32),
  [1591] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3499),
  [1594] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3635),
  [1597] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7013),
  [1600] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(6945),
  [1603] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_head, 1, 0, 0),
  [1605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7127),
  [1607] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5310),
  [1610] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(34),
  [1613] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3509),
  [1616] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3597),
  [1619] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7079),
  [1622] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7052),
  [1625] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5301),
  [1628] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(66),
  [1631] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3507),
  [1634] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3600),
  [1637] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7499),
  [1640] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7456),
  [1643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5301),
  [1645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [1647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3507),
  [1649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3600),
  [1651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7499),
  [1653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7456),
  [1655] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5261),
  [1658] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(75),
  [1661] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3545),
  [1664] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3612),
  [1667] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7598),
  [1670] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7555),
  [1673] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_head, 2, 0, 0),
  [1675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5261),
  [1677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [1679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3545),
  [1681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3612),
  [1683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7598),
  [1685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7555),
  [1687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5310),
  [1689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(34),
  [1691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3509),
  [1693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3597),
  [1695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7079),
  [1697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7052),
  [1699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5289),
  [1701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [1703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3501),
  [1705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3630),
  [1707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7466),
  [1709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7423),
  [1711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5255),
  [1713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [1715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3552),
  [1717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3634),
  [1719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7615),
  [1721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7566),
  [1723] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5289),
  [1726] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(63),
  [1729] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3501),
  [1732] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3630),
  [1735] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7466),
  [1738] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7423),
  [1741] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5255),
  [1744] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(76),
  [1747] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3552),
  [1750] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3634),
  [1753] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7615),
  [1756] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7566),
  [1759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6995),
  [1761] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5273),
  [1764] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(59),
  [1767] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3495),
  [1770] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3618),
  [1773] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7422),
  [1776] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7379),
  [1779] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5265),
  [1782] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(74),
  [1785] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3540),
  [1788] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3607),
  [1791] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7587),
  [1794] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7544),
  [1797] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5277),
  [1800] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(60),
  [1803] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3496),
  [1806] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3644),
  [1809] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7433),
  [1812] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7390),
  [1815] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_head, 4, 0, 0),
  [1817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5277),
  [1819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [1821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3496),
  [1823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3644),
  [1825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7433),
  [1827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7390),
  [1829] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_head, 5, 0, 0),
  [1831] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5284),
  [1834] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(77),
  [1837] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3553),
  [1840] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3650),
  [1843] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7621),
  [1846] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7577),
  [1849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5265),
  [1851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [1853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3540),
  [1855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3607),
  [1857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7587),
  [1859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7544),
  [1861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5284),
  [1863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [1865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3553),
  [1867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3650),
  [1869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7621),
  [1871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7577),
  [1873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5308),
  [1875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [1877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3525),
  [1879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3654),
  [1881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7114),
  [1883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7101),
  [1885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5273),
  [1887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [1889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3495),
  [1891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3618),
  [1893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7422),
  [1895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7379),
  [1897] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5308),
  [1900] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(36),
  [1903] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3525),
  [1906] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3654),
  [1909] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7114),
  [1912] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7101),
  [1915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_head, 3, 0, 0),
  [1917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5304),
  [1919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [1921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3503),
  [1923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3613),
  [1925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6712),
  [1927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6713),
  [1929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5165),
  [1931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [1933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3519),
  [1935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3653),
  [1937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7554),
  [1939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7511),
  [1941] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5304),
  [1944] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(30),
  [1947] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3503),
  [1950] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3613),
  [1953] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(6712),
  [1956] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(6713),
  [1959] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5281),
  [1962] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(99),
  [1965] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3497),
  [1968] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3648),
  [1971] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7444),
  [1974] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7401),
  [1977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4956),
  [1979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3659),
  [1981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_head, 1, 0, 0),
  [1983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5144),
  [1985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3431),
  [1987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5146),
  [1989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5046),
  [1991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7738),
  [1993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5326),
  [1995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5102),
  [1997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7739),
  [1999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5135),
  [2001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8285),
  [2003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5278),
  [2005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5279),
  [2007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4824),
  [2009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5052),
  [2011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5325),
  [2013] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5285),
  [2016] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(62),
  [2019] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3498),
  [2022] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3664),
  [2025] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7455),
  [2028] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7412),
  [2031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5281),
  [2033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [2035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3497),
  [2037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3648),
  [2039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7444),
  [2041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7401),
  [2043] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5297),
  [2046] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(65),
  [2049] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3506),
  [2052] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3598),
  [2055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7488),
  [2058] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7445),
  [2061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5297),
  [2063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [2065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3506),
  [2067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3598),
  [2069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7488),
  [2071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7445),
  [2073] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5165),
  [2076] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(71),
  [2079] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3519),
  [2082] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3653),
  [2085] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7554),
  [2088] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7511),
  [2091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5285),
  [2093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(62),
  [2095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3498),
  [2097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3664),
  [2099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7455),
  [2101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7412),
  [2103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generate_head, 1, 0, 0),
  [2105] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5312),
  [2108] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(68),
  [2111] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3512),
  [2114] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3614),
  [2117] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7521),
  [2120] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7478),
  [2123] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5318),
  [2126] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(69),
  [2129] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3513),
  [2132] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3626),
  [2135] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7532),
  [2138] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7489),
  [2141] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5205),
  [2144] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(50),
  [2147] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3585),
  [2150] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3632),
  [2153] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7295),
  [2156] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7272),
  [2159] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5264),
  [2162] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(57),
  [2165] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3491),
  [2168] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3601),
  [2171] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7400),
  [2174] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7356),
  [2177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5312),
  [2179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [2181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3512),
  [2183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3614),
  [2185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7521),
  [2187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7478),
  [2189] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_architecture_head, 2, 0, 0),
  [2191] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call, 1, 0, 0),
  [2193] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call, 1, 0, 0),
  [2195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7286),
  [2197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5318),
  [2199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [2201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3513),
  [2203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3626),
  [2205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7532),
  [2207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7489),
  [2209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5264),
  [2211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [2213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3491),
  [2215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3601),
  [2217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7400),
  [2219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7356),
  [2221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4895),
  [2223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3663),
  [2225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5239),
  [2227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5059),
  [2229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3393),
  [2231] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_declaration_body, 1, 0, 0),
  [2233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5060),
  [2235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5018),
  [2237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7796),
  [2239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5328),
  [2241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7655),
  [2243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4255),
  [2245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5194),
  [2247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4993),
  [2249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5103),
  [2251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8293),
  [2253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5317),
  [2255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5314),
  [2257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4845),
  [2259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5062),
  [2261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5327),
  [2263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_head, 6, 0, 0),
  [2265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_head, 7, 0, 0),
  [2267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5274),
  [2269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [2271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3556),
  [2273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3619),
  [2275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7175),
  [2277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7152),
  [2279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_head, 8, 0, 0),
  [2281] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(4953),
  [2284] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(3655),
  [2287] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5197),
  [2290] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5127),
  [2293] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(3430),
  [2296] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5132),
  [2299] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(4772),
  [2302] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5045),
  [2305] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5324),
  [2308] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5134),
  [2311] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5107),
  [2314] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(8277),
  [2317] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5242),
  [2320] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5320),
  [2323] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(4735),
  [2326] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5136),
  [2329] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 2, 0, 0), SHIFT_REPEAT(5322),
  [2332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_architecture_head, 1, 0, 0),
  [2334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4020),
  [2336] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5274),
  [2339] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(40),
  [2342] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3556),
  [2345] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3619),
  [2348] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7175),
  [2351] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7152),
  [2354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5205),
  [2356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [2358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3585),
  [2360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3632),
  [2362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7295),
  [2364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7272),
  [2366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4179),
  [2368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4201),
  [2370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4223),
  [2372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4241),
  [2374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5187),
  [2376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [2378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3437),
  [2380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3629),
  [2382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7271),
  [2384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7248),
  [2386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__direct_name, 1, 0, 0),
  [2388] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__direct_name, 1, 0, 0),
  [2390] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesis_expression, 3, 0, 0),
  [2392] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesis_expression, 3, 0, 0),
  [2394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__external_name, 1, 0, 0),
  [2396] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__external_name, 1, 0, 0),
  [2398] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5187),
  [2401] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(48),
  [2404] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3437),
  [2407] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3629),
  [2410] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7271),
  [2413] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7248),
  [2416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5223),
  [2418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [2420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3457),
  [2422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3589),
  [2424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7319),
  [2426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7296),
  [2428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5177),
  [2430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [2432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3494),
  [2434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3658),
  [2436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7259),
  [2438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7236),
  [2440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_association_list, 4, 0, 0),
  [2442] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_association_list, 4, 0, 0),
  [2444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__literal, 1, 0, 0),
  [2446] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__literal, 1, 0, 0),
  [2448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1446),
  [2450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1458),
  [2452] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 1, 0, 0),
  [2454] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_name_repeat1, 1, 0, 0),
  [2456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__name_selector, 1, 0, 0),
  [2458] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__name_selector, 1, 0, 0),
  [2460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5163),
  [2462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [2464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3568),
  [2466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3592),
  [2468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7247),
  [2470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7224),
  [2472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7469),
  [2474] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5163),
  [2477] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(46),
  [2480] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3568),
  [2483] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3592),
  [2486] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7247),
  [2489] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7224),
  [2492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__suffix, 1, 0, 0),
  [2494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__suffix, 1, 0, 0),
  [2496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5315),
  [2498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(31),
  [2500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3433),
  [2502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3594),
  [2504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6944),
  [2506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6837),
  [2508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selection, 2, 0, 0),
  [2510] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_selection, 2, 0, 0),
  [2512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesis_group, 2, 0, 0),
  [2514] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesis_group, 2, 0, 0),
  [2516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 2, 0, 14),
  [2518] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 2, 0, 14),
  [2520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 2, 0, 0),
  [2522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 2, 0, 0),
  [2524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call, 2, 0, 0),
  [2526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call, 2, 0, 0),
  [2528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call, 3, 0, 0),
  [2530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call, 3, 0, 0),
  [2532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_head, 2, 0, 0),
  [2534] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5315),
  [2537] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(31),
  [2540] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3433),
  [2543] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3594),
  [2546] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(6944),
  [2549] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(6837),
  [2552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesis_group, 3, 0, 0),
  [2554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesis_group, 3, 0, 0),
  [2556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call, 4, 0, 0),
  [2558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call, 4, 0, 0),
  [2560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_constant_name, 6, 0, 0),
  [2562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_constant_name, 6, 0, 0),
  [2564] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_signal_name, 6, 0, 0),
  [2566] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_signal_name, 6, 0, 0),
  [2568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_variable_name, 6, 0, 0),
  [2570] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_variable_name, 6, 0, 0),
  [2572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5213),
  [2574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [2576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3551),
  [2578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3604),
  [2580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7139),
  [2582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7115),
  [2584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute_designator, 1, 0, 4),
  [2586] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute_designator, 1, 0, 4),
  [2588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_map_aspect, 3, 0, 0),
  [2590] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_map_aspect, 3, 0, 0),
  [2592] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5177),
  [2595] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(47),
  [2598] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3494),
  [2601] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3658),
  [2604] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7259),
  [2607] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7236),
  [2610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute, 1, 0, 3),
  [2612] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute, 1, 0, 3),
  [2614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 2, 0, 0),
  [2616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 2, 0, 0),
  [2618] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5223),
  [2621] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(52),
  [2624] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3457),
  [2627] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3589),
  [2630] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7319),
  [2633] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7296),
  [2636] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 3, 0, 13),
  [2638] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 3, 0, 13),
  [2640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 4, 0, 22),
  [2642] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 4, 0, 22),
  [2644] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 4, 0, 31),
  [2646] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 4, 0, 31),
  [2648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 5, 0, 40),
  [2650] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 5, 0, 40),
  [2652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 6, 0, 48),
  [2654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 6, 0, 48),
  [2656] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5213),
  [2659] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(37),
  [2662] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3551),
  [2665] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3604),
  [2668] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7139),
  [2671] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7115),
  [2674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_association_list, 3, 0, 0),
  [2676] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_association_list, 3, 0, 0),
  [2678] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__identifier, 1, 0, 0),
  [2680] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__identifier, 1, 0, 0),
  [2682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7166),
  [2684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6962),
  [2686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_based_literal, 2, 0, 0),
  [2688] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_based_literal, 2, 0, 0),
  [2690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_condition_expression, 2, 0, 0),
  [2692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2691),
  [2694] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2691),
  [2696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2619),
  [2698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2638),
  [2700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [2702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2637),
  [2704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2638),
  [2706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2678),
  [2708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_expression, 3, 0, 0),
  [2710] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_expression, 3, 0, 0),
  [2712] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_simple_expression, 3, 0, 0),
  [2714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__abstract_literal, 1, 0, 0),
  [2716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__abstract_literal, 1, 0, 0),
  [2718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_expression, 2, 0, 0),
  [2720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_simple_expression, 2, 0, 0),
  [2722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_relational_expression, 3, 0, 0),
  [2724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_relational_expression, 3, 0, 0),
  [2726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shift_expression, 3, 0, 0),
  [2728] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_shift_expression, 3, 0, 0),
  [2730] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_head, 3, 0, 0),
  [2732] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(4956),
  [2735] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(3659),
  [2738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0),
  [2740] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5144),
  [2743] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(3431),
  [2746] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5146),
  [2749] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5046),
  [2752] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5326),
  [2755] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(7097),
  [2758] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5102),
  [2761] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5033),
  [2764] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5135),
  [2767] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(8285),
  [2770] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5278),
  [2773] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5279),
  [2776] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(4824),
  [2779] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5052),
  [2782] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5325),
  [2785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7557),
  [2787] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_head, 4, 0, 0),
  [2789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1954),
  [2791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1955),
  [2793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7061),
  [2795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7458),
  [2797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2138),
  [2799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2140),
  [2801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [2803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7568),
  [2805] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat1, 2, 0, 0),
  [2807] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4811),
  [2810] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4923),
  [2813] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4761),
  [2816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_declaration_body, 2, 0, 0),
  [2818] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generate_body, 2, 0, 0),
  [2820] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_generate_body, 2, 0, 0), SHIFT(5143),
  [2823] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generate_body, 3, 0, 0),
  [2825] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_generate_body, 3, 0, 0), SHIFT(5143),
  [2828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7425),
  [2830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__range, 1, 0, 0),
  [2832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [2834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2395),
  [2836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2396),
  [2838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2373),
  [2840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2374),
  [2842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2289),
  [2844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2408),
  [2846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_declaration_body, 3, 0, 0),
  [2848] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(4895),
  [2851] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(3663),
  [2854] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5239),
  [2857] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5059),
  [2860] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(3393),
  [2863] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0),
  [2865] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5060),
  [2868] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5018),
  [2871] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5328),
  [2874] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(7655),
  [2877] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5194),
  [2880] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(4993),
  [2883] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5103),
  [2886] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(8293),
  [2889] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5317),
  [2892] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5314),
  [2895] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(4845),
  [2898] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5062),
  [2901] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5327),
  [2904] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 1, 0, 0),
  [2906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [2908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7103),
  [2910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [2912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(521),
  [2914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [2916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2459),
  [2918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2460),
  [2920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2445),
  [2922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2446),
  [2924] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_case_generate_body, 2, 0, 0), SHIFT(5089),
  [2927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_generate_body, 2, 0, 0),
  [2929] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_case_generate_body, 3, 0, 0), SHIFT(5089),
  [2932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_generate_body, 3, 0, 0),
  [2934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7381),
  [2936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7392),
  [2938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3543),
  [2940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4944),
  [2942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3651),
  [2944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [2946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5099),
  [2948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5105),
  [2950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5044),
  [2952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5321),
  [2954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2314),
  [2956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5106),
  [2958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8269),
  [2960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5191),
  [2962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5268),
  [2964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4786),
  [2966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5110),
  [2968] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__label, 1, 0, 1),
  [2970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7579),
  [2972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7546),
  [2974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [2976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(501),
  [2978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2524),
  [2980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2525),
  [2982] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_configuration_specification, 3, 0, 0),
  [2984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7066),
  [2986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2479),
  [2988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2473),
  [2990] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_generate_body, 2, 0, 0), SHIFT(5130),
  [2993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7403),
  [2995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7414),
  [2997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assertion, 2, 0, 0),
  [2999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(666),
  [3001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [3003] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_generate_body, 3, 0, 0), SHIFT(5130),
  [3006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6739),
  [3008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7447),
  [3010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generate_block, 1, 0, 0),
  [3012] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(2996),
  [3015] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(1225),
  [3018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2493),
  [3020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2494),
  [3022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2514),
  [3024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2515),
  [3026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2505),
  [3028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2528),
  [3030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7513),
  [3032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [3034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7358),
  [3036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [3038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bit_string_literal, 3, 0, 0),
  [3040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bit_string_literal, 3, 0, 0),
  [3042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__unit, 1, 0, 9),
  [3044] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__unit, 1, 0, 9),
  [3046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [3048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7154),
  [3050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2410),
  [3052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2580),
  [3054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2581),
  [3056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__unit, 1, 0, 0),
  [3058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__unit, 1, 0, 0),
  [3060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_allocator, 2, 0, 0),
  [3062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_allocator, 2, 0, 0),
  [3064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bit_string_literal, 2, 0, 0),
  [3066] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bit_string_literal, 2, 0, 0),
  [3068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__literal, 2, 0, 10),
  [3070] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__literal, 2, 0, 10),
  [3072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [3074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2595),
  [3076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2596),
  [3078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4379),
  [3080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1549),
  [3082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4954),
  [3084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4904),
  [3086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7492),
  [3088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [3090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4991),
  [3092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_expression, 2, 0, 0),
  [3094] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_waveform_element, 1, 0, 0),
  [3096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [3098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [3100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2551),
  [3102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2552),
  [3104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1653),
  [3106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4924),
  [3108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4925),
  [3110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7574),
  [3112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [3114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5021),
  [3116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2614),
  [3118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2550),
  [3120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1658),
  [3122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4934),
  [3124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4939),
  [3126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8145),
  [3128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [3130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4970),
  [3132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7274),
  [3134] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(3004),
  [3137] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(1231),
  [3140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2586),
  [3142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2589),
  [3144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2604),
  [3146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2605),
  [3148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_expression, 1, 0, 2),
  [3150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_simple_expression, 1, 0, 2),
  [3152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__primary, 1, 0, 0),
  [3154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__primary, 1, 0, 0),
  [3156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__primary, 1, 0, 2),
  [3158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__primary, 1, 0, 2),
  [3160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7480),
  [3162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7491),
  [3164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0),
  [3166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0),
  [3168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [3170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2639),
  [3172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2665),
  [3174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binding_indication, 3, 0, 0),
  [3176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signal_assignment, 1, 0, 0),
  [3178] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(3007),
  [3181] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 2, 0, 0), SHIFT_REPEAT(1212),
  [3184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [3186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [3188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [3190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binding_indication, 1, 0, 0),
  [3192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [3194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2645),
  [3196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6861),
  [3198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7238),
  [3200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_waveform_element, 3, 0, 0),
  [3202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2672),
  [3204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2682),
  [3206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2674),
  [3208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2680),
  [3210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_block, 1, 0, 0),
  [3212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binding_indication, 4, 0, 0),
  [3214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7226),
  [3216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6773),
  [3218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7250),
  [3220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binding_indication, 2, 0, 0),
  [3222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2690),
  [3224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2664),
  [3226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2686),
  [3228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2656),
  [3230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binding_indication, 5, 0, 0),
  [3232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7298),
  [3234] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_block, 2, 0, 0),
  [3236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7117),
  [3238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_statement, 6, 0, 0),
  [3240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [3242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4),
  [3244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__context_item, 1, 0, 0),
  [3246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_declaration, 7, 0, 0),
  [3248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_specification, 7, 0, 14),
  [3250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_configuration_specification, 7, 0, 0),
  [3252] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_instantiation_declaration, 7, 0, 33),
  [3254] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_group_template_declaration, 7, 0, 0),
  [3256] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_instantiation_declaration, 7, 0, 30),
  [3258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_process_statement, 7, 0, 0),
  [3260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_selected_signal_assignment, 7, 0, 0),
  [3262] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_simple_signal_assignment, 7, 0, 0),
  [3264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_conditional_signal_assignment, 7, 0, 0),
  [3266] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_instantiation_declaration, 7, 0, 8),
  [3268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(716),
  [3270] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_process_statement, 4, 0, 0),
  [3272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [3274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__block_declarative_item, 1, 0, 0),
  [3276] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_when_expression, 2, 0, 0),
  [3278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [3280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signal_declaration, 7, 0, 0),
  [3282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 7, 0, 0),
  [3284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mode_view_declaration, 7, 0, 45),
  [3286] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_range, 3, 0, 0),
  [3288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_declaration, 3, 0, 13),
  [3290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_statement, 7, 0, 0),
  [3292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_clause, 3, 0, 0),
  [3294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_generate_statement, 7, 0, 0),
  [3296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_architecture_definition, 8, 0, 50),
  [3298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_configuration_declaration, 8, 0, 51),
  [3300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_configuration_specification, 8, 0, 0),
  [3302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_instantiation_declaration, 8, 0, 33),
  [3304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_group_declaration, 8, 0, 0),
  [3306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_selected_signal_assignment, 8, 0, 0),
  [3308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_instantiation_declaration, 8, 0, 8),
  [3310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_design_unit, 2, 0, 0),
  [3312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 8, 0, 0),
  [3314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2747),
  [3316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2748),
  [3318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__library_unit, 1, 0, 0),
  [3320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_process_statement, 8, 0, 0),
  [3322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_simple_signal_assignment, 8, 0, 0),
  [3324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_conditional_signal_assignment, 8, 0, 0),
  [3326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2725),
  [3328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2729),
  [3330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_selected_signal_assignment, 9, 0, 0),
  [3332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_simple_signal_assignment, 4, 0, 0),
  [3334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_conditional_signal_assignment, 4, 0, 0),
  [3336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2755),
  [3338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2759),
  [3340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2723),
  [3342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2728),
  [3344] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_assertion_statement, 4, 0, 0),
  [3346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_procedure_call_statement, 4, 0, 0),
  [3348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat2, 1, 0, 0),
  [3350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_instantiation_statement, 4, 0, 0),
  [3352] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_generate_statement, 4, 0, 0),
  [3354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_instantiation_statement, 4, 0, 15),
  [3356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_declaration, 5, 0, 0),
  [3358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_declaration, 5, 0, 25),
  [3360] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_declaration, 5, 0, 15),
  [3362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_declaration, 5, 0, 0),
  [3364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_context_declaration, 5, 0, 0),
  [3366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_disconnection_specification, 5, 0, 0),
  [3368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_declaration, 5, 0, 17),
  [3370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_declaration, 5, 0, 0),
  [3372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat3, 1, 0, 0),
  [3374] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_configuration_specification, 5, 0, 0),
  [3376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2702),
  [3378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2703),
  [3380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [3382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__concurrent_statement, 1, 0, 0),
  [3384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(490),
  [3386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_instantiation_statement, 3, 0, 0),
  [3388] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_declaration, 5, 0, 30),
  [3390] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_process_statement, 5, 0, 0),
  [3392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [3394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_simple_signal_assignment, 5, 0, 0),
  [3396] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_conditional_signal_assignment, 5, 0, 0),
  [3398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_condition_clause, 2, 0, 0),
  [3400] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_report_expression, 2, 0, 0),
  [3402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4899),
  [3404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3665),
  [3406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5068),
  [3408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8085),
  [3410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5069),
  [3412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5047),
  [3414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5329),
  [3416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5061),
  [3418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8298),
  [3420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5159),
  [3422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5154),
  [3424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4857),
  [3426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5070),
  [3428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signal_declaration, 5, 0, 0),
  [3430] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subtype_declaration, 5, 0, 13),
  [3432] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_context_reference, 3, 0, 0),
  [3434] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_declaration, 5, 0, 13),
  [3436] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 5, 0, 0),
  [3438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_selected_signal_assignment, 5, 0, 0),
  [3440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_definition, 5, 0, 0),
  [3442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_declaration, 4, 0, 15),
  [3444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_design_file_repeat1, 1, 0, 0),
  [3446] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_statement, 5, 0, 0),
  [3448] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_assertion_statement, 2, 0, 0),
  [3450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_instantiation_statement, 5, 0, 0),
  [3452] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_instantiation_statement, 5, 0, 15),
  [3454] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_generate_statement, 5, 0, 0),
  [3456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_declaration, 6, 0, 0),
  [3458] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_declaration, 6, 0, 17),
  [3460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_declaration, 6, 0, 0),
  [3462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_library_clause, 3, 0, 0),
  [3464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_instantiation_declaration, 6, 0, 33),
  [3466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_definition, 6, 0, 39),
  [3468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_instantiation_declaration, 6, 0, 30),
  [3470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_process_statement, 6, 0, 0),
  [3472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_procedure_call_statement, 2, 0, 0),
  [3474] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_selected_signal_assignment, 6, 0, 0),
  [3476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_simple_signal_assignment, 6, 0, 0),
  [3478] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_conditional_signal_assignment, 6, 0, 0),
  [3480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_instantiation_declaration, 6, 0, 8),
  [3482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat1, 1, 0, 0),
  [3484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_assertion_statement, 3, 0, 0),
  [3486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_concurrent_procedure_call_statement, 3, 0, 0),
  [3488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2719),
  [3490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2699),
  [3492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_or_unaffected, 1, 0, 0),
  [3494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signal_declaration, 6, 0, 0),
  [3496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 6, 0, 0),
  [3498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_declaration, 2, 0, 0),
  [3500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [3502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5386),
  [3504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_definition_body, 2, 0, 0),
  [3506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [3508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7),
  [3510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [3512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequential_block_head, 2, 0, 0),
  [3514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_expression, 3, 0, 0),
  [3516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_process_head, 1, 0, 0),
  [3518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_timeout_clause, 2, 0, 0),
  [3520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3082),
  [3522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3288),
  [3524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3326),
  [3526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_severity_expression, 2, 0, 0),
  [3528] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(4944),
  [3531] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(3651),
  [3534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0),
  [3536] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5099),
  [3539] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5105),
  [3542] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5044),
  [3545] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5321),
  [3548] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(7097),
  [3551] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5106),
  [3554] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5033),
  [3557] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(8269),
  [3560] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5191),
  [3563] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5268),
  [3566] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(4786),
  [3569] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5110),
  [3572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2538),
  [3574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [3576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [3578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_head, 2, 0, 0),
  [3580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2959),
  [3582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8189),
  [3584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2047),
  [3586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [3588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_with_expression, 2, 0, 0),
  [3590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_process_head, 2, 0, 0),
  [3592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_loop, 2, 0, 0),
  [3594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_definition_body, 1, 0, 0),
  [3596] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(4899),
  [3599] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(3665),
  [3602] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5068),
  [3605] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0),
  [3607] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5069),
  [3610] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5047),
  [3613] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5329),
  [3616] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(7097),
  [3619] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5061),
  [3622] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5033),
  [3625] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(8298),
  [3628] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5159),
  [3631] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5154),
  [3634] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(4857),
  [3637] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5070),
  [3640] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(4944),
  [3643] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(3651),
  [3646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0),
  [3648] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5099),
  [3651] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5105),
  [3654] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5044),
  [3657] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5321),
  [3660] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(7097),
  [3663] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5106),
  [3666] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5033),
  [3669] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(8269),
  [3672] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5191),
  [3675] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5268),
  [3678] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(4786),
  [3681] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5110),
  [3684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_logical_name, 1, 0, 0),
  [3686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(395),
  [3688] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(4899),
  [3691] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(3665),
  [3694] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5068),
  [3697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0),
  [3699] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5069),
  [3702] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5047),
  [3705] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5329),
  [3708] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(7097),
  [3711] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5061),
  [3714] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5033),
  [3717] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(8298),
  [3720] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5159),
  [3723] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5154),
  [3726] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(4857),
  [3729] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5070),
  [3732] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequential_block_head, 1, 0, 0),
  [3734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_expression, 2, 0, 0),
  [3736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8),
  [3738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_head, 1, 0, 0),
  [3740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5546),
  [3742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5547),
  [3744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5549),
  [3746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_delay_mechanism, 3, 0, 0),
  [3748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_delay_mechanism, 1, 0, 0),
  [3750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_label_declaration, 2, 0, 0),
  [3752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_force_mode, 1, 0, 0),
  [3754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_selected_waveforms_repeat1, 3, 0, 0),
  [3756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_selected_expressions_repeat1, 3, 0, 0),
  [3758] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_adding_operator, 1, 0, 0),
  [3760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5958),
  [3762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4789),
  [3764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [3766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [3768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4128),
  [3770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4138),
  [3772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3819),
  [3774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5057),
  [3776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(381),
  [3778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3006),
  [3780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1213),
  [3782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4717),
  [3784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sign, 1, 0, 0),
  [3786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [3788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [3790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2999),
  [3792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [3794] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_operator, 1, 0, 0),
  [3796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_multiplying_operator, 1, 0, 0),
  [3798] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__direction, 1, 0, 0),
  [3800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(411),
  [3802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [3804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3011),
  [3806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1226),
  [3808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_operator, 1, 0, 0),
  [3810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shift_operator, 1, 0, 0),
  [3812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [3814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(654),
  [3816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3005),
  [3818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1208),
  [3820] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_relational_operator, 1, 0, 0),
  [3822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3317),
  [3824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5990),
  [3826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3425),
  [3828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4616),
  [3830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6796),
  [3832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3779),
  [3834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3364),
  [3836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7591),
  [3838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4390),
  [3840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5128),
  [3842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5053),
  [3844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4964),
  [3846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7119),
  [3848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5180),
  [3850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4971),
  [3852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5126),
  [3854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5179),
  [3856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5095),
  [3858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8166),
  [3860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5730),
  [3862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6808),
  [3864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6956),
  [3866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5200),
  [3868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8167),
  [3870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7610),
  [3872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6792),
  [3874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7145),
  [3876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7608),
  [3878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7534),
  [3880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7446),
  [3882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7706),
  [3884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5746),
  [3886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5219),
  [3888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5137),
  [3890] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_list, 2, 0, 0),
  [3892] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_list, 3, 0, 0),
  [3894] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_interface_list, 2, 0, 0),
  [3896] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_interface_list, 3, 0, 0),
  [3898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5224),
  [3900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [3902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3559),
  [3904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3628),
  [3906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7199),
  [3908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7176),
  [3910] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5224),
  [3913] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(42),
  [3916] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3559),
  [3919] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3628),
  [3922] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7199),
  [3925] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7176),
  [3928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5196),
  [3930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [3932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3533),
  [3934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3599),
  [3936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7283),
  [3938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7260),
  [3940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_body, 1, 0, 0),
  [3942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3508),
  [3944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1219),
  [3946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5293),
  [3948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [3950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3502),
  [3952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3590),
  [3954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7477),
  [3956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7434),
  [3958] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5293),
  [3961] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(64),
  [3964] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3502),
  [3967] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3590),
  [3970] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7477),
  [3973] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7434),
  [3976] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_body, 2, 0, 0),
  [3978] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_body_repeat1, 2, 0, 0), SHIFT_REPEAT(5672),
  [3981] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_body_repeat1, 2, 0, 0), SHIFT_REPEAT(4552),
  [3984] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_body_repeat1, 2, 0, 0), SHIFT_REPEAT(655),
  [3987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_entity_body_repeat1, 2, 0, 0),
  [3989] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_body_repeat1, 2, 0, 0), SHIFT_REPEAT(3508),
  [3992] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1219),
  [3995] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_body_repeat1, 2, 0, 0), SHIFT_REPEAT(4362),
  [3998] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5196),
  [4001] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(49),
  [4004] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3533),
  [4007] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3599),
  [4010] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7283),
  [4013] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7260),
  [4016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loop_statement, 3, 0, 0),
  [4018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 5, 0, 0),
  [4020] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loop_statement, 5, 0, 0),
  [4022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_variable_assignment, 5, 0, 0),
  [4024] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_force_assignment, 6, 0, 0),
  [4026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exit_statement, 3, 0, 0),
  [4028] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_signal_assignment, 4, 0, 0),
  [4030] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_next_statement, 4, 0, 0),
  [4032] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_null_statement, 3, 0, 0),
  [4034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_waveform_assignment, 6, 0, 0),
  [4036] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_force_assignment, 6, 0, 0),
  [4038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequential_block_statement, 6, 0, 0),
  [4040] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_wait_statement, 6, 0, 0),
  [4042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_variable_assignment, 6, 0, 0),
  [4044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_release_assignment, 6, 0, 0),
  [4046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_waveform_assignment, 6, 0, 0),
  [4048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequential_block_repeat1, 1, 0, 0),
  [4050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequential_statement, 1, 0, 0),
  [4052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assertion_statement, 3, 0, 0),
  [4054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_signal_assignment, 6, 0, 0),
  [4056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_wait_statement, 4, 0, 0),
  [4058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_force_assignment, 7, 0, 0),
  [4060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_waveform_assignment, 7, 0, 0),
  [4062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_statement, 4, 0, 0),
  [4064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_procedure_call_statement, 3, 0, 0),
  [4066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_force_assignment, 7, 0, 0),
  [4068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 2, 0, 0),
  [4070] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loop_statement, 4, 0, 0),
  [4072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_force_assignment, 8, 0, 0),
  [4074] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5203),
  [4077] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(44),
  [4080] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3564),
  [4083] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3633),
  [4086] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7223),
  [4089] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7200),
  [4092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_wait_statement, 2, 0, 0),
  [4094] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_procedure_call_statement, 2, 0, 0),
  [4096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assertion_statement, 2, 0, 0),
  [4098] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequential_block_statement, 5, 0, 0),
  [4100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_wait_statement, 5, 0, 0),
  [4102] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_wait_statement, 3, 0, 0),
  [4104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_variable_assignment, 5, 0, 0),
  [4106] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_waveform_assignment, 5, 0, 0),
  [4108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_variable_assignment, 4, 0, 0),
  [4110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_release_assignment, 4, 0, 0),
  [4112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_waveform_assignment, 4, 0, 0),
  [4114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5203),
  [4116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [4118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3564),
  [4120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3633),
  [4122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7223),
  [4124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7200),
  [4126] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_force_assignment, 5, 0, 0),
  [4128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_release_assignment, 5, 0, 0),
  [4130] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_waveform_assignment, 5, 0, 0),
  [4132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_report_statement, 2, 0, 0),
  [4134] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_signal_assignment, 5, 0, 0),
  [4136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_report_statement, 3, 0, 0),
  [4138] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_null_statement, 2, 0, 0),
  [4140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement_block, 3, 0, 0),
  [4142] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exit_statement, 5, 0, 0),
  [4144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_next_statement, 3, 0, 0),
  [4146] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequential_block_statement, 4, 0, 0),
  [4148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 3, 0, 0),
  [4150] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_next_statement, 5, 0, 0),
  [4152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exit_statement, 4, 0, 0),
  [4154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement_block, 4, 0, 0),
  [4156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exit_statement, 2, 0, 0),
  [4158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_next_statement, 2, 0, 0),
  [4160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_report_statement, 4, 0, 0),
  [4162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 4, 0, 0),
  [4164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4903),
  [4166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3704),
  [4168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2166),
  [4170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7024),
  [4172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4997),
  [4174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7834),
  [4176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4995),
  [4178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5742),
  [4180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4864),
  [4182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5258),
  [4184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [4186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3554),
  [4188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3611),
  [4190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7151),
  [4192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7128),
  [4194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_guard_condition, 3, 0, 0),
  [4196] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5258),
  [4199] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(38),
  [4202] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3554),
  [4205] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3611),
  [4208] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7151),
  [4211] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7128),
  [4214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5694),
  [4216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_clause, 5, 0, 0),
  [4218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1234),
  [4220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_clause, 4, 0, 0),
  [4222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [4224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1217),
  [4226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1230),
  [4228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_header, 1, 0, 0),
  [4230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_port_clause, 5, 0, 0),
  [4232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5283),
  [4234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [4236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3458),
  [4238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3645),
  [4240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7015),
  [4242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7044),
  [4244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_port_clause, 4, 0, 0),
  [4246] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5283),
  [4249] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(55),
  [4252] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3458),
  [4255] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3645),
  [4258] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7015),
  [4261] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7044),
  [4264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5968),
  [4266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3262),
  [4268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3278),
  [4270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7058),
  [4272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7178),
  [4274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4879),
  [4276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6128),
  [4278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6732),
  [4280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6179),
  [4282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7134),
  [4284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4495),
  [4286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6276),
  [4288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4448),
  [4290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8123),
  [4292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5927),
  [4294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7273),
  [4296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__entity_declarative_item, 1, 0, 0),
  [4298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_entity_head_repeat1, 1, 0, 0),
  [4300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7262),
  [4302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__package_declarative_item, 1, 0, 0),
  [4304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_package_declaration_body_repeat1, 1, 0, 0),
  [4306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3275),
  [4308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4151),
  [4310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7436),
  [4312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5241),
  [4314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [4316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3478),
  [4318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3643),
  [4320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7343),
  [4322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7320),
  [4324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5241),
  [4327] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(54),
  [4330] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3478),
  [4333] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3643),
  [4336] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7343),
  [4339] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7320),
  [4342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5996),
  [4344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4462),
  [4346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7837),
  [4348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4861),
  [4350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_header, 3, 0, 0),
  [4352] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4903),
  [4355] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(3704),
  [4358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_protected_type_declaration_repeat1, 2, 0, 0),
  [4360] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4997),
  [4363] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(7655),
  [4366] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4995),
  [4369] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5742),
  [4372] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_protected_type_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4864),
  [4375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4980),
  [4377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3572),
  [4379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [4381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8118),
  [4383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3573),
  [4385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2887),
  [4387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [4389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4649),
  [4391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7349),
  [4393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3541),
  [4395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5269),
  [4397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [4399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3492),
  [4401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3606),
  [4403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7411),
  [4405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7368),
  [4407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5150),
  [4409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [4411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3569),
  [4413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3656),
  [4415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7599),
  [4417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5330),
  [4419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [4421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3524),
  [4423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3605),
  [4425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7576),
  [4427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7533),
  [4429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3456),
  [4431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1220),
  [4433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7202),
  [4435] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5269),
  [4438] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(58),
  [4441] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3492),
  [4444] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3606),
  [4447] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7411),
  [4450] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7368),
  [4453] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5330),
  [4456] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(73),
  [4459] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3524),
  [4462] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3605),
  [4465] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7576),
  [4468] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7533),
  [4471] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5150),
  [4474] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(79),
  [4477] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3569),
  [4480] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3656),
  [4483] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7627),
  [4486] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7599),
  [4489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [4491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2148),
  [4493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2219),
  [4495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5964),
  [4497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3806),
  [4499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3800),
  [4501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5970),
  [4503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3450),
  [4505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3518),
  [4507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [4509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2156),
  [4511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2182),
  [4513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5975),
  [4515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3353),
  [4517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3388),
  [4519] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5210),
  [4522] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(78),
  [4525] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3557),
  [4528] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3652),
  [4531] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7624),
  [4534] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7588),
  [4537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [4539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [4541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [4543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5210),
  [4545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [4547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3557),
  [4549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3652),
  [4551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7624),
  [4553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7588),
  [4555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1224),
  [4557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [4559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2255),
  [4561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2253),
  [4563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [4565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4338),
  [4567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4380),
  [4569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [4571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4942),
  [4573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4949),
  [4575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6817),
  [4577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3072),
  [4579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [4581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7880),
  [4583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5516),
  [4585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8014),
  [4587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8015),
  [4589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5517),
  [4591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [4593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4444),
  [4595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4436),
  [4597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5259),
  [4599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [4601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3489),
  [4603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3596),
  [4605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7389),
  [4607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7344),
  [4609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [4611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4556),
  [4613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4637),
  [4615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7387),
  [4617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3034),
  [4619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7275),
  [4621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5359),
  [4623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7311),
  [4625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7380),
  [4627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5363),
  [4629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [4631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4600),
  [4633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4609),
  [4635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [4637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1697),
  [4639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1685),
  [4641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [4643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4502),
  [4645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4544),
  [4647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5250),
  [4649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [4651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3482),
  [4653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3649),
  [4655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7355),
  [4657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7332),
  [4659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [4661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2121),
  [4663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2143),
  [4665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [4667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1361),
  [4669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1355),
  [4671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [4673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1375),
  [4675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1373),
  [4677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [4679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1509),
  [4681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1502),
  [4683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [4685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1530),
  [4687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1527),
  [4689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [4691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(973),
  [4693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(988),
  [4695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [4697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1163),
  [4699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1141),
  [4701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [4703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3390),
  [4705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3387),
  [4707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [4709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1567),
  [4711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1597),
  [4713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [4715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4692),
  [4717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4758),
  [4719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [4721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1602),
  [4723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1601),
  [4725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [4727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1062),
  [4729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1061),
  [4731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [4733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1086),
  [4735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [4737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [4739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [4741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(923),
  [4743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [4745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1850),
  [4747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1841),
  [4749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [4751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1879),
  [4753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1877),
  [4755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [4757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4700),
  [4759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4690),
  [4761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [4763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4766),
  [4765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4764),
  [4767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [4769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1611),
  [4771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1580),
  [4773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [4775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4773),
  [4777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4765),
  [4779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [4781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4482),
  [4783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4445),
  [4785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [4787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1377),
  [4789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1384),
  [4791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [4793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3408),
  [4795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3426),
  [4797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [4799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1262),
  [4801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1258),
  [4803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7424),
  [4805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [4807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1020),
  [4809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1017),
  [4811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [4813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1990),
  [4815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2009),
  [4817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [4819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1197),
  [4821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1194),
  [4823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [4825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1299),
  [4827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1298),
  [4829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [4831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3976),
  [4833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3694),
  [4835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [4837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4778),
  [4839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4850),
  [4841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [4843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1793),
  [4845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1859),
  [4847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [4849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4614),
  [4851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4603),
  [4853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [4855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(933),
  [4857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [4859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [4861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3292),
  [4863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3307),
  [4865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [4867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4998),
  [4869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4975),
  [4871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [4873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3529),
  [4875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3542),
  [4877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [4879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4732),
  [4881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4844),
  [4883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [4885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2093),
  [4887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2111),
  [4889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [4891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4541),
  [4893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4540),
  [4895] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5250),
  [4898] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(17),
  [4901] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3482),
  [4904] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3649),
  [4907] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7355),
  [4910] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7332),
  [4913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6004),
  [4915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4654),
  [4917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4623),
  [4919] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5259),
  [4922] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(56),
  [4925] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3489),
  [4928] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3596),
  [4931] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7389),
  [4934] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7344),
  [4937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [4939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1788),
  [4941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1773),
  [4943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5960),
  [4945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4736),
  [4947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4865),
  [4949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4082),
  [4951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4862),
  [4953] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5149),
  [4956] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(33),
  [4959] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3504),
  [4962] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3661),
  [4965] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7051),
  [4968] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7014),
  [4971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2256),
  [4973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4012),
  [4975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3398),
  [4977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4062),
  [4979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4815),
  [4981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4094),
  [4983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2124),
  [4985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3987),
  [4987] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5161),
  [4990] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(72),
  [4993] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3521),
  [4996] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3591),
  [4999] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7565),
  [5002] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7522),
  [5005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2151),
  [5007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4199),
  [5009] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5295),
  [5012] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(45),
  [5015] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3565),
  [5018] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3638),
  [5021] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7235),
  [5024] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7212),
  [5027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4579),
  [5029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4035),
  [5031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [5033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4253),
  [5035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1614),
  [5037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4065),
  [5039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3427),
  [5041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3998),
  [5043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1021),
  [5045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4068),
  [5047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1728),
  [5049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4038),
  [5051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(875),
  [5053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4007),
  [5055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5295),
  [5057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [5059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3565),
  [5061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3638),
  [5063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7212),
  [5065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1993),
  [5067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4278),
  [5069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4513),
  [5071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4097),
  [5073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4496),
  [5075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4041),
  [5077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1274),
  [5079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4101),
  [5081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [5083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4071),
  [5085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5201),
  [5087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [5089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3517),
  [5091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3620),
  [5093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7100),
  [5095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7080),
  [5097] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sensitivity_specification, 3, 0, 0),
  [5099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3677),
  [5101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4281),
  [5103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [5105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4104),
  [5107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1570),
  [5109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4176),
  [5111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1881),
  [5113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4075),
  [5115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4780),
  [5117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4285),
  [5119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6187),
  [5121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7144),
  [5123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1376),
  [5125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4044),
  [5127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1891),
  [5129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4288),
  [5131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4793),
  [5133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4264),
  [5135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5149),
  [5137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [5139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3504),
  [5141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3661),
  [5143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7014),
  [5145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(937),
  [5147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4292),
  [5149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5313),
  [5151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [5153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3516),
  [5155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3586),
  [5157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7543),
  [5159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7500),
  [5161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6015),
  [5163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4296),
  [5165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5161),
  [5167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [5169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3521),
  [5171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3591),
  [5173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7565),
  [5175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7522),
  [5177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1794),
  [5179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4078),
  [5181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3310),
  [5183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4302),
  [5185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2186),
  [5187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3995),
  [5189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1167),
  [5191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4058),
  [5193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5010),
  [5195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4307),
  [5197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1791),
  [5199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4003),
  [5201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3544),
  [5203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3980),
  [5205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [5207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4107),
  [5209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(975),
  [5211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4220),
  [5213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6280),
  [5215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4837),
  [5217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3984),
  [5219] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5296),
  [5222] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(39),
  [5225] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3555),
  [5228] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3616),
  [5231] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7163),
  [5234] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7140),
  [5237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5296),
  [5239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [5241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3555),
  [5243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3616),
  [5245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7163),
  [5247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7140),
  [5249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4421),
  [5251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4024),
  [5253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1385),
  [5255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4048),
  [5257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4346),
  [5259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4125),
  [5261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7130),
  [5263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1532),
  [5265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4051),
  [5267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4559),
  [5269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4029),
  [5271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1303),
  [5273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4111),
  [5275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4644),
  [5277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4114),
  [5279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1424),
  [5281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4088),
  [5283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1405),
  [5285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4273),
  [5287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4489),
  [5289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4117),
  [5291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4907),
  [5293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4018),
  [5295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2159),
  [5297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3991),
  [5299] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5201),
  [5302] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(35),
  [5305] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3517),
  [5308] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3620),
  [5311] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7100),
  [5314] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7080),
  [5317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4743),
  [5319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4239),
  [5321] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5313),
  [5324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(70),
  [5327] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3516),
  [5330] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3586),
  [5333] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7543),
  [5336] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7500),
  [5339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1542),
  [5341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4055),
  [5343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6550),
  [5345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4032),
  [5347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__subprogram_declarative_item, 1, 0, 0),
  [5349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__protected_type_body_declarative_item, 1, 0, 0),
  [5351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_protected_type_body_repeat1, 1, 0, 0),
  [5353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_subprogram_head_repeat1, 1, 0, 0),
  [5355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6937),
  [5357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4905),
  [5359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5156),
  [5361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6852),
  [5363] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5232),
  [5366] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(53),
  [5369] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3466),
  [5372] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3657),
  [5375] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7331),
  [5378] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7308),
  [5381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5982),
  [5383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4548),
  [5385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4628),
  [5387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4019),
  [5389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4108),
  [5391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5998),
  [5393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4763),
  [5395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4719),
  [5397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__package_body_declarative_item, 1, 0, 0),
  [5399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_package_definition_body_repeat1, 1, 0, 0),
  [5401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6776),
  [5403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__process_declarative_item, 1, 0, 0),
  [5405] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequential_block_head_repeat1, 1, 0, 0),
  [5407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5232),
  [5409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [5411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3466),
  [5413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3657),
  [5415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7331),
  [5417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7308),
  [5419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7363),
  [5421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5965),
  [5423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4856),
  [5425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4715),
  [5427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5741),
  [5429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4546),
  [5431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4511),
  [5433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5980),
  [5435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4429),
  [5437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4434),
  [5439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5981),
  [5441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4610),
  [5443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4604),
  [5445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5984),
  [5447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4545),
  [5449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4499),
  [5451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5971),
  [5453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4836),
  [5455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4687),
  [5457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4782),
  [5459] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5316),
  [5462] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(43),
  [5465] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3561),
  [5468] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(3631),
  [5471] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7211),
  [5474] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_name_repeat1, 2, 0, 0), SHIFT_REPEAT(7188),
  [5477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5999),
  [5479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4503),
  [5481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4452),
  [5483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5969),
  [5485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5041),
  [5487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5029),
  [5489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5316),
  [5491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [5493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3561),
  [5495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3631),
  [5497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7211),
  [5499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7188),
  [5501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4670),
  [5503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5979),
  [5505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4897),
  [5507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4947),
  [5509] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subtype_indication, 1, 0, 18),
  [5511] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_resolution_indication, 1, 0, 0),
  [5513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [5515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8330),
  [5517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [5519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [5521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6894),
  [5523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5704),
  [5525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5834),
  [5527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4204),
  [5529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6836),
  [5531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5204),
  [5533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [5535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [5537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7322),
  [5539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8361),
  [5541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7370),
  [5543] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_entity_body_repeat1, 1, 0, 0),
  [5545] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__entity_statement, 1, 0, 0),
  [5547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7535),
  [5549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4487),
  [5551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8073),
  [5553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5067),
  [5555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7334),
  [5557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_header, 1, 0, 0),
  [5559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7590),
  [5561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4486),
  [5563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7675),
  [5565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7346),
  [5567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3977),
  [5569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_configuration_head, 2, 0, 0),
  [5571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5162),
  [5573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4576),
  [5575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7132),
  [5577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3776),
  [5579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [5581] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_configuration_head, 1, 0, 0),
  [5583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7038),
  [5585] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__label, 1, 0, 1),
  [5587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7025),
  [5589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3756),
  [5591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5807),
  [5593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7142),
  [5595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3309),
  [5597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4883),
  [5599] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mode, 1, 0, 0),
  [5601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7502),
  [5603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7084),
  [5605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7524),
  [5607] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_end, 1, 0, 0),
  [5609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4935),
  [5611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6733),
  [5613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7214),
  [5615] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_configuration_head_repeat1, 2, 0, 0), SHIFT_REPEAT(3977),
  [5618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_configuration_head_repeat1, 2, 0, 0),
  [5620] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_configuration_head_repeat1, 2, 0, 0), SHIFT_REPEAT(5162),
  [5623] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_configuration_head_repeat1, 2, 0, 0), SHIFT_REPEAT(4868),
  [5626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5026),
  [5628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_analysis_expression, 1, 0, 0),
  [5630] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4858),
  [5633] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4900),
  [5636] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_design_unit_repeat1, 2, 0, 0), SHIFT_REPEAT(4802),
  [5639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4147),
  [5641] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_conditional_analysis_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5026),
  [5644] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_conditional_analysis_expression_repeat1, 2, 0, 0),
  [5646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3978),
  [5648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_private_variable_declaration, 2, 0, 0),
  [5650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4175),
  [5652] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4390),
  [5655] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_type_definition_repeat1, 2, 0, 0),
  [5657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6865),
  [5659] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mode_view_body, 2, 0, 0),
  [5661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4217),
  [5663] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__protected_type_declarative_item, 1, 0, 0),
  [5665] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_protected_type_declaration_repeat1, 1, 0, 0),
  [5667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5882),
  [5669] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_analysis_expression, 2, 0, 0),
  [5671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5065),
  [5673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7310),
  [5675] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mode_view_body, 1, 0, 0),
  [5677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4858),
  [5679] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_context_declaration_body, 1, 0, 0),
  [5681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4900),
  [5683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4802),
  [5685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_header, 3, 0, 0),
  [5687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4293),
  [5689] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_mode_view_body_repeat1, 2, 0, 0), SHIFT_REPEAT(4390),
  [5692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_mode_view_body_repeat1, 2, 0, 0),
  [5694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_context_declaration_body, 2, 0, 0),
  [5696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4301),
  [5698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6344),
  [5700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2880),
  [5702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6033),
  [5704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2983),
  [5706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5006),
  [5708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6623),
  [5710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5176),
  [5712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6721),
  [5714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4008),
  [5716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_resolution_indication, 3, 0, 0),
  [5718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2866),
  [5720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2867),
  [5722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4188),
  [5724] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_physical_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4390),
  [5727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_physical_type_definition_repeat1, 2, 0, 0),
  [5729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2971),
  [5731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4015),
  [5733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2943),
  [5735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2813),
  [5737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2860),
  [5739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4221),
  [5741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [5743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2912),
  [5745] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_end, 2, 0, 0),
  [5747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2913),
  [5749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4126),
  [5751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2879),
  [5753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7190),
  [5755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2947),
  [5757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4224),
  [5759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6771),
  [5761] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6771),
  [5763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6152),
  [5765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5535),
  [5767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2936),
  [5769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(667),
  [5771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3772),
  [5773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [5775] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_analysis_relation, 4, 0, 0),
  [5777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5567),
  [5779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5569),
  [5781] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_configuration, 1, 0, 0),
  [5783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5114),
  [5785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2886),
  [5787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2922),
  [5789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5221),
  [5791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5237),
  [5793] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_entity, 1, 0, 0),
  [5795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5072),
  [5797] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_case, 2, 0, 0),
  [5799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5118),
  [5801] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_analysis_relation, 3, 0, 0),
  [5803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7067),
  [5805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2782),
  [5807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7047),
  [5809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7678),
  [5811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7060),
  [5813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4160),
  [5815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2950),
  [5817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5256),
  [5819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_procedure_specification, 2, 0, 8),
  [5821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2785),
  [5823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6847),
  [5825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7873),
  [5827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6848),
  [5829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7872),
  [5831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4210),
  [5833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5266),
  [5835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5267),
  [5837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7740),
  [5839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2775),
  [5841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6725),
  [5843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6726),
  [5845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7797),
  [5847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3019),
  [5849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5473),
  [5851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7835),
  [5853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7860),
  [5855] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_component, 1, 0, 0),
  [5857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5083),
  [5859] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subtype_indication, 2, 0, 13),
  [5861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_architecture, 1, 0, 0),
  [5863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5131),
  [5865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_context, 1, 0, 0),
  [5867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5079),
  [5869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2838),
  [5871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1707),
  [5873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5496),
  [5875] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_package_body, 1, 0, 0),
  [5877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7391),
  [5879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2843),
  [5881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_package, 1, 0, 0),
  [5883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5117),
  [5885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7057),
  [5887] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_partial_pathname_repeat1, 2, 0, 0), SHIFT_REPEAT(4390),
  [5890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7840),
  [5892] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_if, 2, 0, 0),
  [5894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4931),
  [5896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [5898] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elsif_generate, 4, 0, 0),
  [5900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6660),
  [5902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4277),
  [5904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5244),
  [5906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_block_configuration_repeat2, 2, 0, 0),
  [5908] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_configuration_repeat2, 2, 0, 0), SHIFT_REPEAT(3309),
  [5911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__conditional_analysis_identifier, 1, 0, 0),
  [5913] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__conditional_analysis_identifier, 1, 0, 0),
  [5915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_entity, 2, 0, 0),
  [5917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_view, 2, 0, 0),
  [5919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_context, 2, 0, 0),
  [5921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4359),
  [5923] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__target, 1, 0, 0),
  [5925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4122),
  [5927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_component, 2, 0, 0),
  [5929] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_process, 2, 0, 0),
  [5931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_conditional_expression_repeat1, 2, 0, 0),
  [5933] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_conditional_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(646),
  [5936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5798),
  [5938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_record, 2, 0, 0),
  [5940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_specification, 3, 0, 0),
  [5942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6013),
  [5944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2274),
  [5946] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 0),
  [5948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elsif_statement, 4, 0, 0),
  [5950] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_package_body, 3, 0, 0),
  [5952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3329),
  [5954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [5956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6126),
  [5958] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_relative_pathname_repeat1, 2, 0, 0), SHIFT_REPEAT(6852),
  [5961] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_relative_pathname_repeat1, 2, 0, 0),
  [5963] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_package_pathname_repeat1, 2, 0, 0), SHIFT_REPEAT(4390),
  [5966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_block, 2, 0, 0),
  [5968] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_generate, 2, 0, 0),
  [5970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5286),
  [5972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5195),
  [5974] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_loop, 2, 0, 0),
  [5976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3087),
  [5978] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_generate, 4, 0, 0),
  [5980] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_configuration, 2, 0, 0),
  [5982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5280),
  [5984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2163),
  [5986] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_package, 2, 0, 0),
  [5988] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_case, 3, 0, 0),
  [5990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3778),
  [5992] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_header, 4, 0, 0),
  [5994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7378),
  [5996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4228),
  [5998] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_procedure_specification, 2, 0, 8),
  [6000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4767),
  [6002] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_declaration_end, 2, 0, 0),
  [6004] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elsif_generate, 3, 0, 0),
  [6006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8296),
  [6008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_architecture, 2, 0, 0),
  [6010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_process, 3, 0, 0),
  [6012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5248),
  [6014] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_header, 3, 0, 0),
  [6016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3299),
  [6018] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_body_end, 3, 0, 0),
  [6020] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_units, 2, 0, 0),
  [6022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5777),
  [6024] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_expression, 2, 0, 0),
  [6026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_generate, 3, 0, 0),
  [6028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4816),
  [6030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4518),
  [6032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3047),
  [6034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3666),
  [6036] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_constraint, 2, 0, 0),
  [6038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1682),
  [6040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7367),
  [6042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3720),
  [6044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4728),
  [6046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2090),
  [6048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3042),
  [6050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1672),
  [6052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6642),
  [6054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3332),
  [6056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4683),
  [6058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4684),
  [6060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_secondary_unit_declaration, 4, 0, 0),
  [6062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2910),
  [6064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4685),
  [6066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_group_constituent_list_repeat1, 2, 0, 0),
  [6068] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_group_constituent_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4208),
  [6071] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_body, 1, 0, 0),
  [6073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2118),
  [6075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4367),
  [6077] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_index_constraint, 1, 0, 0),
  [6079] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__element, 1, 0, 0),
  [6081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3052),
  [6083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3255),
  [6085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3336),
  [6087] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subtype_indication, 2, 0, 18),
  [6089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2153),
  [6091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(970),
  [6093] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_body_repeat1, 2, 0, 0),
  [6095] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_body_repeat1, 2, 0, 0), SHIFT_REPEAT(252),
  [6098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2286),
  [6100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3404),
  [6102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5405),
  [6104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3257),
  [6106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4819),
  [6108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2252),
  [6110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3526),
  [6112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4141),
  [6114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4153),
  [6116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5748),
  [6118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1741),
  [6120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3339),
  [6122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2073),
  [6124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3258),
  [6126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3341),
  [6128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4631),
  [6130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7350),
  [6132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3999),
  [6134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1986),
  [6136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [6138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3016),
  [6140] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_group_constituent_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4164),
  [6143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7609),
  [6145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3722),
  [6147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2233),
  [6149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3289),
  [6151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_secondary_unit_declaration, 5, 0, 0),
  [6153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5868),
  [6155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(930),
  [6157] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_integer_incomplete_type_definition, 2, 0, 0),
  [6159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6703),
  [6161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_body, 2, 0, 0),
  [6163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_primary_unit_declaration, 2, 0, 0),
  [6165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4945),
  [6167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_configuration_head_repeat2, 2, 0, 0),
  [6169] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_configuration_head_repeat2, 2, 0, 0), SHIFT_REPEAT(7067),
  [6172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_mode_indication, 2, 0, 0),
  [6174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5423),
  [6176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8047),
  [6178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__designator, 1, 0, 0),
  [6180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__logical_name, 1, 0, 6),
  [6182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4806),
  [6184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3724),
  [6186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4400),
  [6188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_block_configuration_repeat1, 2, 0, 0),
  [6190] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_block_configuration_repeat1, 2, 0, 0), SHIFT_REPEAT(4883),
  [6193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subtype_indication, 3, 0, 13),
  [6195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4565),
  [6197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3050),
  [6199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3761),
  [6201] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sensitivity_list, 2, 0, 0),
  [6203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4208),
  [6205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1207),
  [6207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3771),
  [6209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3815),
  [6211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4553),
  [6213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2183),
  [6215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1004),
  [6217] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_element_declaration, 4, 0, 0),
  [6219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1700),
  [6221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1245),
  [6223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4478),
  [6225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3252),
  [6227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_procedure_specification, 3, 0, 8),
  [6229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1341),
  [6231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1732),
  [6233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2264),
  [6235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1358),
  [6237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [6239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4339),
  [6241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1284),
  [6243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1506),
  [6245] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_type_definition_repeat1, 1, 0, 0),
  [6247] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_mode_indication, 1, 0, 0),
  [6249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5367),
  [6251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1206),
  [6253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3765),
  [6255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3769),
  [6257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2993),
  [6259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3371),
  [6261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4721),
  [6263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4775),
  [6265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1561),
  [6267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_physical_type_definition_repeat1, 1, 0, 0),
  [6269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4673),
  [6271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [6273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2828),
  [6275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [6277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(891),
  [6279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3780),
  [6281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1323),
  [6283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sensitivity_list, 1, 0, 0),
  [6285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1083),
  [6287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3726),
  [6289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1807),
  [6291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4885),
  [6293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2142),
  [6295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5012),
  [6297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1847),
  [6299] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mode_view_element_definition, 4, 0, 0),
  [6301] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_mode_view_body_repeat1, 1, 0, 0),
  [6303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4460),
  [6305] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_or_unaffected_expression, 1, 0, 0),
  [6307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4207),
  [6309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(979),
  [6311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4230),
  [6313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3327),
  [6315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [6317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [6319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(981),
  [6321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4232),
  [6323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4812),
  [6325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4244),
  [6327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3248),
  [6329] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_conditional_waveforms_repeat1, 2, 0, 0),
  [6331] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_conditional_waveforms_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [6334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4843),
  [6336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4247),
  [6338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1094),
  [6340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4258),
  [6342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_choices, 3, 0, 0),
  [6344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3716),
  [6346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_generate_block_repeat1, 2, 0, 0),
  [6348] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_case_generate_block_repeat1, 2, 0, 0), SHIFT_REPEAT(221),
  [6351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1096),
  [6353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4260),
  [6355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4889),
  [6357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4269),
  [6359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4681),
  [6361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4271),
  [6363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_partial_pathname_repeat1, 2, 0, 0),
  [6365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1415),
  [6367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4275),
  [6369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1419),
  [6371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4276),
  [6373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_configuration_head, 3, 0, 0),
  [6375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1998),
  [6377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4279),
  [6379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2000),
  [6381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4280),
  [6383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3745),
  [6385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4283),
  [6387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3773),
  [6389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4284),
  [6391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4881),
  [6393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4286),
  [6395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4884),
  [6397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4287),
  [6399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_private_incomplete_type_definition, 1, 0, 0),
  [6401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1724),
  [6403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4289),
  [6405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1735),
  [6407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4290),
  [6409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scalar_incomplete_type_definition, 1, 0, 0),
  [6411] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__incomplete_type_definition, 1, 0, 0),
  [6413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(947),
  [6415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4294),
  [6417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [6419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4295),
  [6421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6031),
  [6423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4299),
  [6425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6095),
  [6427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4300),
  [6429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4368),
  [6431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4246),
  [6433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3311),
  [6435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4303),
  [6437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3313),
  [6439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4304),
  [6441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5020),
  [6443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4309),
  [6445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4972),
  [6447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4166),
  [6449] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_waveforms, 2, 0, 0),
  [6451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [6453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_waveform, 2, 0, 0),
  [6455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [6457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3546),
  [6459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3982),
  [6461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3548),
  [6463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3983),
  [6465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4329),
  [6467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4193),
  [6469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4872),
  [6471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3985),
  [6473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4809),
  [6475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3986),
  [6477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2128),
  [6479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3989),
  [6481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2130),
  [6483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3990),
  [6485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [6487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2170),
  [6489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3993),
  [6491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2172),
  [6493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3994),
  [6495] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_definition, 4, 0, 0),
  [6497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2190),
  [6499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3996),
  [6501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2192),
  [6503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3997),
  [6505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3354),
  [6507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4001),
  [6509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3356),
  [6511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4002),
  [6513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1828),
  [6515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4004),
  [6517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1830),
  [6519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4005),
  [6521] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_waveform, 1, 0, 0),
  [6523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unspecified_type_indication, 3, 0, 0),
  [6525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(877),
  [6527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4009),
  [6529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [6531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4010),
  [6533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2265),
  [6535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4016),
  [6537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2267),
  [6539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4017),
  [6541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4908),
  [6543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4022),
  [6545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4937),
  [6547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4023),
  [6549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4423),
  [6551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4027),
  [6553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4440),
  [6555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4028),
  [6557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4568),
  [6559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4030),
  [6561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4570),
  [6563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4031),
  [6565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6530),
  [6567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4033),
  [6569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6204),
  [6571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4034),
  [6573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4624),
  [6575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4036),
  [6577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4626),
  [6579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4037),
  [6581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1754),
  [6583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4039),
  [6585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1766),
  [6587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4040),
  [6589] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__incomplete_subtype_indication, 1, 0, 0),
  [6591] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_access_incomplete_type_definition, 2, 0, 0),
  [6593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4451),
  [6595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4042),
  [6597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4470),
  [6599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4043),
  [6601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_physical_incomplete_type_definition, 2, 0, 0),
  [6603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1378),
  [6605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4046),
  [6607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1380),
  [6609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4047),
  [6611] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_mode_indication, 3, 0, 0),
  [6613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1387),
  [6615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4049),
  [6617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1389),
  [6619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4050),
  [6621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_unit, 2, 0, 17),
  [6623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5215),
  [6625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1537),
  [6627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4053),
  [6629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1539),
  [6631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4054),
  [6633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1543),
  [6635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4056),
  [6637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1545),
  [6639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4057),
  [6641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1171),
  [6643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4060),
  [6645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1174),
  [6647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4061),
  [6649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3399),
  [6651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4063),
  [6653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3401),
  [6655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4064),
  [6657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1617),
  [6659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4066),
  [6661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1619),
  [6663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4067),
  [6665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1023),
  [6667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4069),
  [6669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1037),
  [6671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4070),
  [6673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(897),
  [6675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4073),
  [6677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(908),
  [6679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4074),
  [6681] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_name, 4, 0, 21),
  [6683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5086),
  [6685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_name, 4, 0, 7),
  [6687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1884),
  [6689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4076),
  [6691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1888),
  [6693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4077),
  [6695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1899),
  [6697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4080),
  [6699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1903),
  [6701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4081),
  [6703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4746),
  [6705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4085),
  [6707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4749),
  [6709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4086),
  [6711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1433),
  [6713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4092),
  [6715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1439),
  [6717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4093),
  [6719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4832),
  [6721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4095),
  [6723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4835),
  [6725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4096),
  [6727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4458),
  [6729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4098),
  [6731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1275),
  [6733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4102),
  [6735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1277),
  [6737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4103),
  [6739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [6741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4105),
  [6743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1030),
  [6745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4106),
  [6747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1144),
  [6749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4109),
  [6751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1148),
  [6753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4110),
  [6755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1305),
  [6757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4112),
  [6759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1307),
  [6761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4113),
  [6763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4653),
  [6765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4115),
  [6767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4599),
  [6769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4116),
  [6771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4516),
  [6773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4119),
  [6775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4538),
  [6777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4120),
  [6779] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__incomplete_type_mark, 1, 0, 0),
  [6781] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_incomplete_type_definition, 3, 0, 59),
  [6783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__incomplete_type_mark, 1, 0, 18),
  [6785] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_discrete_incomplete_type_definition, 3, 0, 0),
  [6787] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type_definition, 6, 0, 0),
  [6789] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_package_pathname_repeat1, 2, 0, 0),
  [6791] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_aspect, 2, 0, 17),
  [6793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5174),
  [6795] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_floating_incomplete_type_definition, 4, 0, 0),
  [6797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4261),
  [6799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_designator, 1, 0, 0),
  [6801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3624),
  [6803] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_or_unaffected_expression, 2, 0, 0),
  [6805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2800),
  [6807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4974),
  [6809] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_name, 3, 0, 21),
  [6811] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_name, 3, 0, 7),
  [6813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2955),
  [6815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7165),
  [6817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2051),
  [6819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interface_type_indication, 1, 0, 0),
  [6821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4263),
  [6823] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 2, 0, 32), SHIFT_REPEAT(4207),
  [6826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 2, 0, 32),
  [6828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_waveforms, 3, 0, 0),
  [6830] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_waveform_repeat1, 2, 0, 0),
  [6832] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_waveform_repeat1, 2, 0, 0), SHIFT_REPEAT(404),
  [6835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_generate_block, 1, 0, 0),
  [6837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_unit, 4, 0, 34),
  [6839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5299),
  [6841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4274),
  [6843] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generate_body, 4, 0, 0),
  [6845] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_generate_body, 4, 0, 0), SHIFT(5143),
  [6848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_header, 5, 0, 0),
  [6850] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__configuration_declarative_item, 1, 0, 0),
  [6852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_configuration_head_repeat1, 1, 0, 0),
  [6854] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_selected_name_repeat1, 2, 0, 0),
  [6856] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_name_repeat1, 2, 0, 0), SHIFT_REPEAT(5086),
  [6859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4013),
  [6861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_conditional_or_unaffected_expression_repeat1, 2, 0, 0),
  [6863] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_conditional_or_unaffected_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(646),
  [6866] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_when_element, 2, 0, 0),
  [6868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [6870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4121),
  [6872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1578),
  [6874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4183),
  [6876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3753),
  [6878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4192),
  [6880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1582),
  [6882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4185),
  [6884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2847),
  [6886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2975),
  [6888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2851),
  [6890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2188),
  [6892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4205),
  [6894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3038),
  [6896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2198),
  [6898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4211),
  [6900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2931),
  [6902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2895),
  [6904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4509),
  [6906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4099),
  [6908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [6910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1272),
  [6912] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_function_specification, 4, 0, 23),
  [6914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4596),
  [6916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4612),
  [6918] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_waveforms, 4, 0, 0),
  [6920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3247),
  [6922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3325),
  [6924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1704),
  [6926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1706),
  [6928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2017),
  [6930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2018),
  [6932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4446),
  [6934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4449),
  [6936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4704),
  [6938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_name_list, 2, 0, 0),
  [6940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4072),
  [6942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [6944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [6946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4156),
  [6948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1397),
  [6950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1398),
  [6952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5302),
  [6954] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_identifier_list, 1, 0, 27),
  [6956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3678),
  [6958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3681),
  [6960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1599),
  [6962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1600),
  [6964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2792),
  [6966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_interface_list, 1, 0, 0),
  [6968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3256),
  [6970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1606),
  [6972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1607),
  [6974] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_mode_view_indication, 2, 0, 45),
  [6976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3515),
  [6978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2280),
  [6980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1143),
  [6982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1155),
  [6984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3261),
  [6986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4705),
  [6988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3412),
  [6990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3413),
  [6992] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_index_incomplete_type_list_repeat1, 2, 0, 0), SHIFT_REPEAT(225),
  [6995] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_index_incomplete_type_list_repeat1, 2, 0, 0),
  [6997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4712),
  [6999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3265),
  [7001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1646),
  [7003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1648),
  [7005] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_name_list, 1, 0, 7),
  [7007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5013),
  [7009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [7011] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_association_or_range_list, 2, 0, 0),
  [7013] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_class_entry_list_repeat1, 2, 0, 0), SHIFT_REPEAT(3022),
  [7016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_entity_class_entry_list_repeat1, 2, 0, 0),
  [7018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4164),
  [7020] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_group_constituent_list, 2, 0, 0),
  [7022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [7024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [7026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3266),
  [7028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2281),
  [7030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(922),
  [7032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(907),
  [7034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1748),
  [7036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1749),
  [7038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6346),
  [7040] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_user_directive, 1, 0, 0),
  [7042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1893),
  [7044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1802),
  [7046] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_aggregate_repeat1, 2, 0, 0), SHIFT_REPEAT(223),
  [7049] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_aggregate_repeat1, 2, 0, 0),
  [7051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5),
  [7053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1715),
  [7055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1740),
  [7057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4089),
  [7059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3338),
  [7061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4838),
  [7063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4840),
  [7065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2030),
  [7067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [7069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [7071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1427),
  [7073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1431),
  [7075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_unit, 7, 0, 56),
  [7077] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_association_or_range_list, 1, 0, 0),
  [7079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3715),
  [7081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4855),
  [7083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4860),
  [7085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4291),
  [7087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3320),
  [7089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4459),
  [7091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4469),
  [7093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3322),
  [7095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pathname_element, 1, 0, 0),
  [7097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [7099] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_partial_pathname, 1, 0, 0),
  [7101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [7103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [7105] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_function_specification, 6, 0, 38),
  [7107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [7109] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_element_association_list, 1, 0, 0),
  [7111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3340),
  [7113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [7115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1035),
  [7117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3723),
  [7119] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_unit, 5, 0, 42),
  [7121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5821),
  [7123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5823),
  [7125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4963),
  [7127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4965),
  [7129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1169),
  [7131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [7133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [7135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1315),
  [7137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4647),
  [7139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4648),
  [7141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4498),
  [7143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4500),
  [7145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(993),
  [7147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3731),
  [7149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3566),
  [7151] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_mode_view_indication, 4, 0, 55),
  [7153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3562),
  [7155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3712),
  [7157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3742),
  [7159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3744),
  [7161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3567),
  [7163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3735),
  [7165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3736),
  [7167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_aspect, 2, 0, 16),
  [7169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [7171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1786),
  [7173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4870),
  [7175] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_case_generate_body, 4, 0, 0), SHIFT(5089),
  [7178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_generate_body, 4, 0, 0),
  [7180] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_user_directive_repeat1, 2, 0, 0), SHIFT_REPEAT(6346),
  [7183] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_user_directive_repeat1, 2, 0, 0),
  [7185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4873),
  [7187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4376),
  [7189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2951),
  [7191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5763),
  [7193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_type_declaration, 2, 0, 13),
  [7195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3488),
  [7197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2791),
  [7199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2177),
  [7201] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__entity_class, 1, 0, 0),
  [7203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3022),
  [7205] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_class_entry_list, 1, 0, 0),
  [7207] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_class_entry, 1, 0, 0),
  [7209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6290),
  [7211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5222),
  [7213] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_identifier_list, 2, 0, 11),
  [7215] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_element_association_list, 2, 0, 0),
  [7217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [7219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6518),
  [7221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2258),
  [7223] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_name_list, 2, 0, 0),
  [7225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4916),
  [7227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signal_list, 2, 0, 0),
  [7229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2262),
  [7231] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__entity_tag, 1, 0, 0),
  [7233] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_name_list, 1, 0, 0),
  [7235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2179),
  [7237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2287),
  [7239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2426),
  [7241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [7243] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_index_incomplete_type_list, 2, 0, 0),
  [7245] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_body, 2, 0, 0),
  [7247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [7249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4693),
  [7251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sensitivity_clause, 2, 0, 0),
  [7253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4697),
  [7255] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_constant_identifier_list_repeat1, 2, 0, 20), SHIFT_REPEAT(5222),
  [7258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constant_identifier_list_repeat1, 2, 0, 20),
  [7260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2193),
  [7262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2045),
  [7264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2195),
  [7266] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_index_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(385),
  [7269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_index_constraint_repeat1, 2, 0, 0),
  [7271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_selected_name_list_repeat1, 2, 0, 0),
  [7273] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_selected_name_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4916),
  [7276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5251),
  [7278] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_identifier_list, 2, 0, 0),
  [7280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4702),
  [7282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2956),
  [7284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assertion, 3, 0, 0),
  [7286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2976),
  [7288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2988),
  [7290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4703),
  [7292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list_specification, 2, 0, 0),
  [7294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4707),
  [7296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2114),
  [7298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [7300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5245),
  [7302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiation_list, 2, 0, 0),
  [7304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2205),
  [7306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2206),
  [7308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2285),
  [7310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1856),
  [7312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_access_type_definition, 2, 0, 0),
  [7314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [7316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_instantiation_definition, 2, 0, 0),
  [7318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3973),
  [7320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3974),
  [7322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3975),
  [7324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8303),
  [7326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5074),
  [7328] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_identifier_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5251),
  [7331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_identifier_list_repeat1, 2, 0, 0),
  [7333] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enumeration_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4959),
  [7336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enumeration_type_definition_repeat1, 2, 0, 0),
  [7338] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_instantiation_list_repeat1, 2, 0, 0), SHIFT_REPEAT(5245),
  [7341] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_instantiation_list_repeat1, 2, 0, 0),
  [7343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list_specification, 3, 0, 0),
  [7345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4959),
  [7347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8173),
  [7349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_aspect, 1, 0, 0),
  [7351] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_resolution_repeat1, 2, 0, 0), SHIFT_REPEAT(5124),
  [7354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_resolution_repeat1, 2, 0, 0),
  [7356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4375),
  [7358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3360),
  [7360] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_association_or_range_list_repeat1, 2, 0, 0), SHIFT_REPEAT(131),
  [7363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_association_or_range_list_repeat1, 2, 0, 0),
  [7365] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_signature_repeat1, 2, 0, 13),
  [7367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3361),
  [7369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2011),
  [7371] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_element_list, 2, 0, 0),
  [7373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_user_directive, 2, 0, 0),
  [7375] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_partial_pathname, 2, 0, 0),
  [7377] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generic_identifier_list_repeat1, 2, 0, 44), SHIFT_REPEAT(5302),
  [7380] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_identifier_list_repeat1, 2, 0, 44),
  [7382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_list_specification, 4, 0, 0),
  [7384] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_logical_name_list, 2, 0, 12),
  [7386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3345),
  [7388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1846),
  [7390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2790),
  [7392] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_list, 1, 0, 0),
  [7394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_subprogram_declaration, 1, 0, 0),
  [7396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3637),
  [7398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interface_subprogram_specification, 1, 0, 0),
  [7400] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_index_constraint, 2, 0, 0),
  [7402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1907),
  [7404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generate_block_end, 2, 0, 0),
  [7406] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_association_list_repeat1, 2, 0, 0), SHIFT_REPEAT(321),
  [7409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_association_list_repeat1, 2, 0, 0),
  [7411] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_logical_name_list_repeat1, 2, 0, 12),
  [7413] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_logical_name_list_repeat1, 2, 0, 12), SHIFT_REPEAT(5013),
  [7416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3979),
  [7418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5124),
  [7420] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_resolution, 2, 0, 0),
  [7422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1074),
  [7424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [7426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [7428] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_entity_name_list_repeat1, 2, 0, 0), SHIFT_REPEAT(4704),
  [7431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_entity_name_list_repeat1, 2, 0, 0),
  [7433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_identifier_list, 2, 0, 35),
  [7435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2793),
  [7437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [7439] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_generic_interface_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2796),
  [7442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_interface_list_repeat1, 2, 0, 0),
  [7444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2277),
  [7446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2278),
  [7448] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_index_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(382),
  [7451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1984),
  [7453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constant_identifier_list, 1, 0, 5),
  [7455] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generate_block_end, 3, 0, 0),
  [7457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generate_body, 5, 0, 0),
  [7459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3348),
  [7461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4901),
  [7463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_unit, 2, 0, 16),
  [7465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2081),
  [7467] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_verification_unit_list, 1, 0, 0),
  [7469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_name_list, 1, 0, 0),
  [7471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1921),
  [7473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4902),
  [7475] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_resolution, 1, 0, 0),
  [7477] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_block_configuration_repeat1, 1, 0, 0),
  [7479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3350),
  [7481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5786),
  [7483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4415),
  [7485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4419),
  [7487] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_procedure_specification, 3, 0, 8),
  [7489] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signal_list, 1, 0, 0),
  [7491] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_interface_list_repeat1, 2, 0, 0), SHIFT_REPEAT(2795),
  [7494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_interface_list_repeat1, 2, 0, 0),
  [7496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2980),
  [7498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_class_entry_list, 2, 0, 0),
  [7500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_group_constituent_list, 1, 0, 0),
  [7502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2979),
  [7504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1684),
  [7506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [7508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1926),
  [7510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1929),
  [7512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4129),
  [7514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_conditional_or_unaffected_expression, 3, 0, 0),
  [7516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4803),
  [7518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4617),
  [7520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3693),
  [7522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [7524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3020),
  [7526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_identifier_list, 1, 0, 0),
  [7528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4550),
  [7530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2853),
  [7532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2839),
  [7534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_aspect, 5, 0, 42),
  [7536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4382),
  [7538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4353),
  [7540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2970),
  [7542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2939),
  [7544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3752),
  [7546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4804),
  [7548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_verification_unit_list, 2, 0, 0),
  [7550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4322),
  [7552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2965),
  [7554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2855),
  [7556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2832),
  [7558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5800),
  [7560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2946),
  [7562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3018),
  [7564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3768),
  [7566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4334),
  [7568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [7570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5804),
  [7572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2848),
  [7574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2864),
  [7576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2815),
  [7578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_function_specification, 5, 0, 28),
  [7580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3116),
  [7582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1589),
  [7584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [7586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [7588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [7590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3787),
  [7592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_function_specification, 5, 0, 29),
  [7594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [7596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4349),
  [7598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3794),
  [7600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [7602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [7604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [7606] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_element_association_list_repeat1, 2, 0, 0), SHIFT_REPEAT(222),
  [7609] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_element_association_list_repeat1, 2, 0, 0),
  [7611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2852),
  [7613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [7615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2842),
  [7617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3705),
  [7619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3797),
  [7621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1591),
  [7623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2823),
  [7625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4370),
  [7627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1953),
  [7629] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_index_incomplete_type_list, 1, 0, 0),
  [7631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3030),
  [7633] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiation_list, 1, 0, 0),
  [7635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5843),
  [7637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2890),
  [7639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2891),
  [7641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5358),
  [7643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3839),
  [7645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3671),
  [7647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3037),
  [7649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5360),
  [7651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3680),
  [7653] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protect_directive, 1, 0, 0),
  [7655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2924),
  [7657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2885),
  [7659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6502),
  [7661] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protect_directive, 2, 0, 0),
  [7663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5924),
  [7665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3049),
  [7667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3697),
  [7669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6859),
  [7671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1259),
  [7673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2902),
  [7675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2892),
  [7677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2872),
  [7679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3277),
  [7681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6336),
  [7683] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_element_list, 1, 0, 0),
  [7685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3056),
  [7687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3709),
  [7689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3060),
  [7691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2896),
  [7693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2923),
  [7695] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_unit, 2, 0, 15),
  [7697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3061),
  [7699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6508),
  [7701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2894),
  [7703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3746),
  [7705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3823),
  [7707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3824),
  [7709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3825),
  [7711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3827),
  [7713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3828),
  [7715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3829),
  [7717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3830),
  [7719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3831),
  [7721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3832),
  [7723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3833),
  [7725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3834),
  [7727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3835),
  [7729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3836),
  [7731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3837),
  [7733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3838),
  [7735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3840),
  [7737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3841),
  [7739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3842),
  [7741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3316),
  [7743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3844),
  [7745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3845),
  [7747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3846),
  [7749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3847),
  [7751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3848),
  [7753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3849),
  [7755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3850),
  [7757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3851),
  [7759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3852),
  [7761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3853),
  [7763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3854),
  [7765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3855),
  [7767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3856),
  [7769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3857),
  [7771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3858),
  [7773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3859),
  [7775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3860),
  [7777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3861),
  [7779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3862),
  [7781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3863),
  [7783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3864),
  [7785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3865),
  [7787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3866),
  [7789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3867),
  [7791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3868),
  [7793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3869),
  [7795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3870),
  [7797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3871),
  [7799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3872),
  [7801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3873),
  [7803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3874),
  [7805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3875),
  [7807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3876),
  [7809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3877),
  [7811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3878),
  [7813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3879),
  [7815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3880),
  [7817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3881),
  [7819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3882),
  [7821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3883),
  [7823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3884),
  [7825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3885),
  [7827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3886),
  [7829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3887),
  [7831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3888),
  [7833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3889),
  [7835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3890),
  [7837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3891),
  [7839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3892),
  [7841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3893),
  [7843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3894),
  [7845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3895),
  [7847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3896),
  [7849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3897),
  [7851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3898),
  [7853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3899),
  [7855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3900),
  [7857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3901),
  [7859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3902),
  [7861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3903),
  [7863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3904),
  [7865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3905),
  [7867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3906),
  [7869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3907),
  [7871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3908),
  [7873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3909),
  [7875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3910),
  [7877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3911),
  [7879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3912),
  [7881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3913),
  [7883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3914),
  [7885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3915),
  [7887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3916),
  [7889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3917),
  [7891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3918),
  [7893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3919),
  [7895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3920),
  [7897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3921),
  [7899] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3922),
  [7901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3923),
  [7903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3924),
  [7905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3925),
  [7907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3926),
  [7909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3927),
  [7911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3928),
  [7913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3929),
  [7915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3930),
  [7917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3931),
  [7919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3932),
  [7921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3933),
  [7923] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3934),
  [7925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3935),
  [7927] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3936),
  [7929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3937),
  [7931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3938),
  [7933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3939),
  [7935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3940),
  [7937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3941),
  [7939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3942),
  [7941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3943),
  [7943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3944),
  [7945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3945),
  [7947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3946),
  [7949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3947),
  [7951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3948),
  [7953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3949),
  [7955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3950),
  [7957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3951),
  [7959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3952),
  [7961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3953),
  [7963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3954),
  [7965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3955),
  [7967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3956),
  [7969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3957),
  [7971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3958),
  [7973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3959),
  [7975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3960),
  [7977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3961),
  [7979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3962),
  [7981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3963),
  [7983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3964),
  [7985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3965),
  [7987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3966),
  [7989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3967),
  [7991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3968),
  [7993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3969),
  [7995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3970),
  [7997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3971),
  [7999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3972),
  [8001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_name, 1, 0, 7),
  [8003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6043),
  [8005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7442),
  [8007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6324),
  [8009] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_map_default, 5, 0, 0),
  [8011] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_map_any, 5, 0, 0),
  [8013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_configuration, 7, 0, 0),
  [8015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6644),
  [8017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8316),
  [8019] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_element_association, 3, 0, 0),
  [8021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_specification, 5, 0, 28),
  [8023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(825),
  [8025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_element_association, 1, 0, 0),
  [8027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5024),
  [8029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1477),
  [8031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7431),
  [8033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [8035] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_specification, 5, 0, 29),
  [8037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1873),
  [8039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4651),
  [8041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2427),
  [8043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5490),
  [8045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5491),
  [8047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5019),
  [8049] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__actual_part, 1, 0, 0),
  [8051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6946),
  [8053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4801),
  [8055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [8057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5009),
  [8059] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_specification, 7, 0, 46),
  [8061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3707),
  [8063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3376),
  [8065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1783),
  [8067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2298),
  [8069] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_block_configuration_repeat2, 1, 0, 0),
  [8071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8352),
  [8073] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__subprogram_specification, 1, 0, 0),
  [8075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__configuration_item, 1, 0, 0),
  [8077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [8079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1070),
  [8081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1914),
  [8083] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_generic_identifier_list_repeat1, 2, 0, 43),
  [8085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1158),
  [8087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4514),
  [8089] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constant_identifier_list_repeat1, 2, 0, 19),
  [8091] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_body_repeat1, 1, 0, 0),
  [8093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [8095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [8097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1447),
  [8099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [8101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [8103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_specification, 7, 0, 47),
  [8105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2062),
  [8107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signal_kind, 1, 0, 0),
  [8109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4808),
  [8111] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_waveform, 2, 0, 0),
  [8113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [8115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [8117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [8119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [8121] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumeration_literal, 1, 0, 0),
  [8123] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumeration_literal, 1, 0, 5),
  [8125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [8127] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_waveforms, 2, 0, 0),
  [8129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2613),
  [8131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [8133] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__association_or_range, 1, 0, 0),
  [8135] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_association_element, 3, 0, 0),
  [8137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1616),
  [8139] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_association_element, 1, 0, 0),
  [8141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [8143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [8145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [8147] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_specification, 4, 0, 23),
  [8149] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_interface_declaration, 3, 0, 0),
  [8151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2307),
  [8153] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_specification, 8, 0, 52),
  [8155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2391),
  [8157] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interface_declaration, 1, 0, 0),
  [8159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_case_generate_block_repeat1, 1, 0, 0),
  [8161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [8163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_expression_or_unaffected, 2, 0, 0),
  [8165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1649),
  [8167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_specification, 8, 0, 53),
  [8169] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_expressions, 2, 0, 0),
  [8171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2617),
  [8173] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_logical_name_list_repeat1, 2, 0, 24),
  [8175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [8177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [8179] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_element_resolution, 2, 0, 0),
  [8181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [8183] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_designator, 2, 0, 0),
  [8185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5008),
  [8187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [8189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4014),
  [8191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4968),
  [8193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__generic_interface_declaration, 1, 0, 0),
  [8195] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_initialiser, 2, 0, 0),
  [8197] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_expressions, 3, 0, 0),
  [8199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [8201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1823),
  [8203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3366),
  [8205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [8207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [8209] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_select_target, 3, 0, 0),
  [8211] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_select_target, 2, 0, 0),
  [8213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [8215] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_configuration, 4, 0, 0),
  [8217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2341),
  [8219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_specification, 6, 0, 37),
  [8221] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_procedure_specification, 4, 0, 8),
  [8223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1763),
  [8225] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_configuration, 5, 0, 0),
  [8227] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_configuration, 6, 0, 0),
  [8229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1959),
  [8231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3429),
  [8233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [8235] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_selected_waveforms, 3, 0, 0),
  [8237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8146),
  [8239] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_conditional_analysis, 1, 0, 0),
  [8241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [8243] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interface_subprogram_default, 1, 0, 0),
  [8245] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_subprogram_declaration, 3, 0, 0),
  [8247] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_declaration, 3, 0, 0),
  [8249] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__mode_indication, 1, 0, 0),
  [8251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__mode_view_indication, 1, 0, 0),
  [8253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [8255] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_class_entry, 2, 0, 0),
  [8257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [8259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_specification, 6, 0, 38),
  [8261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5536),
  [8263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_index_subtype_definition, 3, 0, 18),
  [8265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
  [8267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1330),
  [8269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [8271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4324),
  [8273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2393),
  [8275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3706),
  [8277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3362),
  [8279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [8281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8060),
  [8283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5084),
  [8285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [8287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [8289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4385),
  [8291] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_constant_declaration, 4, 0, 0),
  [8293] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_file_declaration, 4, 0, 0),
  [8295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [8297] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__alias_designator, 1, 0, 0),
  [8299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1789),
  [8301] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_user_directive_repeat1, 1, 0, 0),
  [8303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5474),
  [8305] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pathname, 6, 0, 0),
  [8307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2026),
  [8309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [8311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [8313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2355),
  [8315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [8317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [8319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4187),
  [8321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7113),
  [8323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [8325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3577),
  [8327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4173),
  [8329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [8331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_generate_alternative, 4, 0, 0),
  [8333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [8335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1930),
  [8337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1204),
  [8339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3579),
  [8341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4190),
  [8343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(833),
  [8345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [8347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [8349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3581),
  [8351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4215),
  [8353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(834),
  [8355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [8357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [8359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3583),
  [8361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4234),
  [8363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(835),
  [8365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3584),
  [8367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4248),
  [8369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [8371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3434),
  [8373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4262),
  [8375] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_signal_declaration, 4, 0, 0),
  [8377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__array_index_incomplete_type, 1, 0, 0),
  [8379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3758),
  [8381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [8383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1048),
  [8385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3032),
  [8387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3365),
  [8389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2209),
  [8391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [8393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [8395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [8397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_aggregate, 4, 0, 0),
  [8399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2023),
  [8401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3796),
  [8403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [8405] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_generate_body, 4, 0, 0), SHIFT(5130),
  [8408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [8410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [8412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2227),
  [8414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_package_declaration, 6, 0, 30),
  [8416] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_aggregate, 3, 0, 0),
  [8418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [8420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3441),
  [8422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4134),
  [8424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [8426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [8428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [8430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [8432] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_definition, 1, 0, 0),
  [8434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5119),
  [8436] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__interface_package_generic_map_aspect, 1, 0, 0),
  [8438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_type_declaration, 4, 0, 13),
  [8440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1260),
  [8442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2260),
  [8444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_mode_view_indication, 4, 0, 45),
  [8446] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_simple_mode_indication, 4, 0, 0),
  [8448] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_specification, 9, 0, 60),
  [8450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4306),
  [8452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8033),
  [8454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [8456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4308),
  [8458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8064),
  [8460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4310),
  [8462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8090),
  [8464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4311),
  [8466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8110),
  [8468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4197),
  [8470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1629),
  [8472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4312),
  [8474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8124),
  [8476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [8478] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pathname, 7, 0, 0),
  [8480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interface_variable_declaration, 4, 0, 0),
  [8482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__actual_part, 2, 0, 0),
  [8484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3242),
  [8486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3373),
  [8488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_generate_alternative, 3, 0, 0),
  [8490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [8492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1710),
  [8494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [8496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_configuration, 4, 0, 0),
  [8498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_mode_view_indication, 6, 0, 55),
  [8500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_case_generate_body, 5, 0, 0),
  [8502] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_configuration, 5, 0, 0),
  [8504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [8506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_configuration, 6, 0, 0),
  [8508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1127),
  [8510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3318),
  [8512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3370),
  [8514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2920),
  [8516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2757),
  [8518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4662),
  [8520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4584),
  [8522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4090),
  [8524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1145),
  [8526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2101),
  [8528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4590),
  [8530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4592),
  [8532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4593),
  [8534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4825),
  [8536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1149),
  [8538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4165),
  [8540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1152),
  [8542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2784),
  [8544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_entity, 2, 0, 17),
  [8546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_configuration, 3, 0, 58),
  [8548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4678),
  [8550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4433),
  [8552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2103),
  [8554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2104),
  [8556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4528),
  [8558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2105),
  [8560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [8562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4377),
  [8564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4800),
  [8566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4532),
  [8568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4536),
  [8570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4537),
  [8572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4194),
  [8574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1309),
  [8576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4814),
  [8578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1310),
  [8580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6582),
  [8582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4652),
  [8584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5098),
  [8586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2991),
  [8588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4581),
  [8590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2221),
  [8592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2579),
  [8594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4607),
  [8596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2984),
  [8598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4627),
  [8600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [8602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2578),
  [8604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2587),
  [8606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4091),
  [8608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4530),
  [8610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [8612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4539),
  [8614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4542),
  [8616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4563),
  [8618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2611),
  [8620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4831),
  [8622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3323),
  [8624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1909),
  [8626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4686),
  [8628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2127),
  [8630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4688),
  [8632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4490),
  [8634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2129),
  [8636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4689),
  [8638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2512),
  [8640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2478),
  [8642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1320),
  [8644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1321),
  [8646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2545),
  [8648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1425),
  [8650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1426),
  [8652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1190),
  [8654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2318),
  [8656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3392),
  [8658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1428),
  [8660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1045),
  [8662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(894),
  [8664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1717),
  [8666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1752),
  [8668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4817),
  [8670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1430),
  [8672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4822),
  [8674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4457),
  [8676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2487),
  [8678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1324),
  [8680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1046),
  [8682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [8684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1325),
  [8686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4601),
  [8688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4472),
  [8690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2132),
  [8692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2133),
  [8694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2134),
  [8696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4694),
  [8698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5476),
  [8700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2135),
  [8702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4695),
  [8704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7476),
  [8706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2136),
  [8708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2960),
  [8710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4083),
  [8712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4696),
  [8714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4219),
  [8716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6331),
  [8718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6332),
  [8720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2249),
  [8722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__directive, 1, 0, 0),
  [8724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2940),
  [8726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2444),
  [8728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2786),
  [8730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2942),
  [8732] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_body_end, 4, 0, 0),
  [8734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3333),
  [8736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2789),
  [8738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2787),
  [8740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2986),
  [8742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_units, 3, 0, 0),
  [8744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2448),
  [8746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2633),
  [8748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5170),
  [8750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5027),
  [8752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2250),
  [8754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_element_record_mode_view_indication, 2, 0, 45),
  [8756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_port_map_aspect, 3, 0, 0),
  [8758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6357),
  [8760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3243),
  [8762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2169),
  [8764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4243),
  [8766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [8768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2171),
  [8770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4709),
  [8772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4710),
  [8774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2174),
  [8776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2175),
  [8778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elsif_conditional_analysis, 3, 0, 0),
  [8780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2176),
  [8782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5495),
  [8784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2178),
  [8786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4711),
  [8788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2180),
  [8790] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elsif_generate, 5, 0, 0),
  [8792] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_for, 2, 0, 0),
  [8794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4713),
  [8796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8362),
  [8798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1977),
  [8800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3066),
  [8802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6589),
  [8804] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_end, 3, 0, 41),
  [8806] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_relative_pathname, 2, 0, 0),
  [8808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6793),
  [8810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2549),
  [8812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4127),
  [8814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__conditional_analysis_operator, 1, 0, 0),
  [8816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_elsif_statement, 5, 0, 0),
  [8818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6306),
  [8820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5155),
  [8822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5225),
  [8824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7684),
  [8826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7692),
  [8828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6802),
  [8830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7713),
  [8832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4214),
  [8834] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_conditional_analysis, 3, 0, 0),
  [8836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1966),
  [8838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2555),
  [8840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2532),
  [8842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6599),
  [8844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2522),
  [8846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1950),
  [8848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3418),
  [8850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4785),
  [8852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3465),
  [8854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2191),
  [8856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3473),
  [8858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7313),
  [8860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4320),
  [8862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3988),
  [8864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3422),
  [8866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3423),
  [8868] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_element_array_mode_view_indication, 4, 0, 55),
  [8870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8016),
  [8872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3424),
  [8874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4229),
  [8876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6603),
  [8878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2194),
  [8880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3068),
  [8882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2196),
  [8884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6008),
  [8886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6010),
  [8888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1964),
  [8890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5141),
  [8892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6011),
  [8894] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_body, 4, 0, 0),
  [8896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2679),
  [8898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8040),
  [8900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3334),
  [8902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4810),
  [8904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2722),
  [8906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3483),
  [8908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3073),
  [8910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2372),
  [8912] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_declaration_end, 3, 0, 0),
  [8914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3335),
  [8916] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_declaration, 4, 0, 0),
  [8918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2041),
  [8920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7342),
  [8922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2215),
  [8924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2724),
  [8926] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_record, 3, 0, 0),
  [8928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8078),
  [8930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2726),
  [8932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2376),
  [8934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6500),
  [8936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6501),
  [8938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4136),
  [8940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2475),
  [8942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1410),
  [8944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4011),
  [8946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1760),
  [8948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3028),
  [8950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3355),
  [8952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1416),
  [8954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3245),
  [8956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2788),
  [8958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2773),
  [8960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1764),
  [8962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1765),
  [8964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1767),
  [8966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5158),
  [8968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3357),
  [8970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7375),
  [8972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [8974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3358),
  [8976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5262),
  [8978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1297),
  [8980] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumeration_type_definition, 4, 0, 0),
  [8982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1304),
  [8984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6521),
  [8986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3246),
  [8988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1312),
  [8990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2660),
  [8992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5142),
  [8994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3067),
  [8996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4305),
  [8998] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_physical_type_definition, 4, 0, 0),
  [9000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7745),
  [9002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2687),
  [9004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1073),
  [9006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5499),
  [9008] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assertion, 4, 0, 0),
  [9010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2257),
  [9012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3337),
  [9014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(863),
  [9016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1328),
  [9018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1829),
  [9020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(978),
  [9022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3321),
  [9024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3249),
  [9026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [9028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [9030] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_package, 2, 0, 30),
  [9032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2590),
  [9034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [9036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2259),
  [9038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4025),
  [9040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2043),
  [9042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1853),
  [9044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2261),
  [9046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3250),
  [9048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8213),
  [9050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8214),
  [9052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8219),
  [9054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1855),
  [9056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2263),
  [9058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [9060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3675),
  [9062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3051),
  [9064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2733),
  [9066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2228),
  [9068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2974),
  [9070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5507),
  [9072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4313),
  [9074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2669),
  [9076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2012),
  [9078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2229),
  [9080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3251),
  [9082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3029),
  [9084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2624),
  [9086] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_component, 2, 0, 15),
  [9088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8252),
  [9090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4724),
  [9092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4267),
  [9094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2695),
  [9096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2244),
  [9098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1995),
  [9100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(878),
  [9102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3324),
  [9104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4829),
  [9106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2246),
  [9108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2247),
  [9110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2044),
  [9112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2248),
  [9114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3992),
  [9116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [9118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1997),
  [9120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(881),
  [9122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [9124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7563),
  [9126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3790),
  [9128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6289),
  [9130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6305),
  [9132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4976),
  [9134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2937),
  [9136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4139),
  [9138] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_loop, 2, 0, 0),
  [9140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [9142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1999),
  [9144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2938),
  [9146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2458),
  [9148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3695),
  [9150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4578),
  [9152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2002),
  [9154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2462),
  [9156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2954),
  [9158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1265),
  [9160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2003),
  [9162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6252),
  [9164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2004),
  [9166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4921),
  [9168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2957),
  [9170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3253),
  [9172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2266),
  [9174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2005),
  [9176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2046),
  [9178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3254),
  [9180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4929),
  [9182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4930),
  [9184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4933),
  [9186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2006),
  [9188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2268),
  [9190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_error_directive, 2, 0, 0),
  [9192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2269),
  [9194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2230),
  [9196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2964),
  [9198] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [9200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1967),
  [9202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2966),
  [9204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [9206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5039),
  [9208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2932),
  [9210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4430),
  [9212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3698),
  [9214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7369),
  [9216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7440),
  [9218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2577),
  [9220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_guarded_signal_specification, 3, 0, 22),
  [9222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5852),
  [9224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6148),
  [9226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5635),
  [9228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_generate, 3, 0, 0),
  [9230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2583),
  [9232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2273),
  [9234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7836),
  [9236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6760),
  [9238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2720),
  [9240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5788),
  [9242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4405),
  [9244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2212),
  [9246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6788),
  [9248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4936),
  [9250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4130),
  [9252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6158),
  [9254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6161),
  [9256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6363),
  [9258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3445),
  [9260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4079),
  [9262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4408),
  [9264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4409),
  [9266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4410),
  [9268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1968),
  [9270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2049),
  [9272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4932),
  [9274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5125),
  [9276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6330),
  [9278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2732),
  [9280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4940),
  [9282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1516),
  [9284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5077),
  [9286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1983),
  [9288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [9290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_declaration, 2, 0, 0),
  [9292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2523),
  [9294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1985),
  [9296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5087),
  [9298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3500),
  [9300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_type_definition, 2, 0, 0),
  [9302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7124),
  [9304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7155),
  [9306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6367),
  [9308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7265),
  [9310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3728),
  [9312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2778),
  [9314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2527),
  [9316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7410),
  [9318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2594),
  [9320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6130),
  [9322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6154),
  [9324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4779),
  [9326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4615),
  [9328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3755),
  [9330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4168),
  [9332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2781),
  [9334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6230),
  [9336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4426),
  [9338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4664),
  [9340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6758),
  [9342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2053),
  [9344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3505),
  [9346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4355),
  [9348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4665),
  [9350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7821),
  [9352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4356),
  [9354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8096),
  [9356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4620),
  [9358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8299),
  [9360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4357),
  [9362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6448),
  [9364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6457),
  [9366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5188),
  [9368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4621),
  [9370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4622),
  [9372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6515),
  [9374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3775),
  [9376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4401),
  [9378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3781),
  [9380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4402),
  [9382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__element_resolution, 1, 0, 0),
  [9384] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_architecture, 2, 0, 49),
  [9386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4910),
  [9388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8363),
  [9390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4378),
  [9392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3798),
  [9394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6752),
  [9396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5032),
  [9398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6812),
  [9400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3328),
  [9402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6113),
  [9404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6121),
  [9406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter_specification, 3, 0, 0),
  [9408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6147),
  [9410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3670),
  [9412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2603),
  [9414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4667),
  [9416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3676),
  [9418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1969),
  [9420] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subprogram_end, 2, 0, 33),
  [9422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8353),
  [9424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6923),
  [9426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3008),
  [9428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6211),
  [9430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6218),
  [9432] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_design_file, 1, 0, 0),
  [9434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6232),
  [9436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2607),
  [9438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4978),
  [9440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2608),
  [9442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3259),
  [9444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3260),
  [9446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4567),
  [9448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4551),
  [9450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7026),
  [9452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4569),
  [9454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6301),
  [9456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6304),
  [9458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(983),
  [9460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6326),
  [9462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [9464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_conditional_analysis, 1, 0, 0),
  [9466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7191),
  [9468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(761),
  [9470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6370),
  [9472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6371),
  [9474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4572),
  [9476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6431),
  [9478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4573),
  [9480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1915),
  [9482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4577),
  [9484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2745),
  [9486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7315),
  [9488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6513),
  [9490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6514),
  [9492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4583),
  [9494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6526),
  [9496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3263),
  [9498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4650),
  [9500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3574),
  [9502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5654),
  [9504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3264),
  [9506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7429),
  [9508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2340),
  [9510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6588),
  [9512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6012),
  [9514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6016),
  [9516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__elsif_else_generate, 1, 0, 0),
  [9518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3308),
  [9520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2754),
  [9522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4172),
  [9524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7638),
  [9526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4237),
  [9528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6087),
  [9530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6092),
  [9532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_generate, 3, 0, 0),
  [9534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6100),
  [9536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_configuration, 2, 0, 16),
  [9538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2317),
  [9540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3770),
  [9542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3684),
  [9544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4180),
  [9546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7769),
  [9548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4140),
  [9550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6141),
  [9552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6143),
  [9554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6155),
  [9556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2765),
  [9558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1872),
  [9560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2766),
  [9562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6182),
  [9564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [9566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8065),
  [9568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1906),
  [9570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6173),
  [9572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6176),
  [9574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1908),
  [9576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6183),
  [9578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1911),
  [9580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4877),
  [9582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1918),
  [9584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6511),
  [9586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8244),
  [9588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6533),
  [9590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6214),
  [9592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6217),
  [9594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6227),
  [9596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4882),
  [9598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2490),
  [9600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1919),
  [9602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_entity_specification, 3, 0, 0),
  [9604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8351),
  [9606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_open_information, 4, 0, 0),
  [9608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6246),
  [9610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6247),
  [9612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [9614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6255),
  [9616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3575),
  [9618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3826),
  [9620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2472),
  [9622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4854),
  [9624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6099),
  [9626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4387),
  [9628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6663),
  [9630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4771),
  [9632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6278),
  [9634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6282),
  [9636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4798),
  [9638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6300),
  [9640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6901),
  [9642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2225),
  [9644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4213),
  [9646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4527),
  [9648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4818),
  [9650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6730),
  [9652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6186),
  [9654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6323),
  [9656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6325),
  [9658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4625),
  [9660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6333),
  [9662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4701),
  [9664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4880),
  [9666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4533),
  [9668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4534),
  [9670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4535),
  [9672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3240),
  [9674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6785),
  [9676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3267),
  [9678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6358),
  [9680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6360),
  [9682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4635),
  [9684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6366),
  [9686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4754),
  [9688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4638),
  [9690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3268),
  [9692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3269),
  [9694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_context, 3, 0, 0),
  [9696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6835),
  [9698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4428),
  [9700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6413),
  [9702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6418),
  [9704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8286),
  [9706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6442),
  [9708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3514),
  [9710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3443),
  [9712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2074),
  [9714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1736),
  [9716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3271),
  [9718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_entity, 3, 0, 36),
  [9720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6871),
  [9722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2075),
  [9724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6494),
  [9726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6499),
  [9728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2646),
  [9730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6507),
  [9732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1348),
  [9734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_open_information, 2, 0, 0),
  [9736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4146),
  [9738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1762),
  [9740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6915),
  [9742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [9744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6523),
  [9746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6524),
  [9748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4202),
  [9750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6529),
  [9752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1871),
  [9754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [9756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_verification_unit_binding_indication, 3, 0, 0),
  [9758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1920),
  [9760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2086),
  [9762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1351),
  [9764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6958),
  [9766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2165),
  [9768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6548),
  [9770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6553),
  [9772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1352),
  [9774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6564),
  [9776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5228),
  [9778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1779),
  [9780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2623),
  [9782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1784),
  [9784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6999),
  [9786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2715),
  [9788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6602),
  [9790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6009),
  [9792] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_block, 3, 0, 0),
  [9794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6014),
  [9796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3236),
  [9798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1713),
  [9800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2977),
  [9802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(964),
  [9804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6741),
  [9806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7048),
  [9808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1726),
  [9810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6032),
  [9812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6034),
  [9814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1368),
  [9816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6047),
  [9818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3342),
  [9820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4000),
  [9822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4453),
  [9824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3343),
  [9826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1738),
  [9828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1370),
  [9830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7118),
  [9832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4045),
  [9834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6073),
  [9836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6075),
  [9838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [9840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6079),
  [9842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1372),
  [9844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1757),
  [9846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4480),
  [9848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1758),
  [9850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7180),
  [9852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4493),
  [9854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6090),
  [9856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6091),
  [9858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2770),
  [9860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6097),
  [9862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5471),
  [9864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [9866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5671),
  [9868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4161),
  [9870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3283),
  [9872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7229),
  [9874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1768),
  [9876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6111),
  [9878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6112),
  [9880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3069),
  [9882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6119),
  [9884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1490),
  [9886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4186),
  [9888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2995),
  [9890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1770),
  [9892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1922),
  [9894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1379),
  [9896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6137),
  [9898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6140),
  [9900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5420),
  [9902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6145),
  [9904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3272),
  [9906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7721),
  [9908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1494),
  [9910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1495),
  [9912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1496),
  [9914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7359),
  [9916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1381),
  [9918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6164),
  [9920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6165),
  [9922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5093),
  [9924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6170),
  [9926] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_package_body, 2, 0, 30),
  [9928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3795),
  [9930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1923),
  [9932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1382),
  [9934] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_package, 3, 0, 39),
  [9936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6174),
  [9938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6175),
  [9940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1924),
  [9942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6181),
  [9944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1925),
  [9946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1960),
  [9948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1470),
  [9950] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__iteration_scheme, 1, 0, 0),
  [9952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__external_pathname, 1, 0, 0),
  [9954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5036),
  [9956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2736),
  [9958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6194),
  [9960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6195),
  [9962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1928),
  [9964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6198),
  [9966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2251),
  [9968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1515),
  [9970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2968),
  [9972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2476),
  [9974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1388),
  [9976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6209),
  [9978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6210),
  [9980] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__process_sensitivity_list, 1, 0, 0),
  [9982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6215),
  [9984] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__elsif_else_statement, 1, 0, 0),
  [9986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1444),
  [9988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2767),
  [9990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5051),
  [9992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1521),
  [9994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6221),
  [9996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6223),
  [9998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1522),
  [10000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6229),
  [10002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1523),
  [10004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2949),
  [10006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3346),
  [10008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1390),
  [10010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5108),
  [10012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3347),
  [10014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6240),
  [10016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6241),
  [10018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6244),
  [10020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4321),
  [10022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1391),
  [10024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3010),
  [10026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(946),
  [10028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2952),
  [10030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3679),
  [10032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2712),
  [10034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6248),
  [10036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6251),
  [10038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5484),
  [10040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6254),
  [10042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(948),
  [10044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_if, 3, 0, 0),
  [10046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5485),
  [10048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4358),
  [10050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_package_body, 4, 0, 54),
  [10052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2470),
  [10054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3710),
  [10056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6257),
  [10058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6259),
  [10060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1093),
  [10062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6262),
  [10064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_loop, 3, 0, 0),
  [10066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3610),
  [10068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2987),
  [10070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4297),
  [10072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1932),
  [10074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2982),
  [10076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2693),
  [10078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6272),
  [10080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6273),
  [10082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2744),
  [10084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6279),
  [10086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3076),
  [10088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1129),
  [10090] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_design_file, 2, 0, 0),
  [10092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [10094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1538),
  [10096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6291),
  [10098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6292),
  [10100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(952),
  [10102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6302),
  [10104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2944),
  [10106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3560),
  [10108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2856),
  [10110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2948),
  [10112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3474),
  [10114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2840),
  [10116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(953),
  [10118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6313),
  [10120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6314),
  [10122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1131),
  [10124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6320),
  [10126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4381),
  [10128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2945),
  [10130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1132),
  [10132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1133),
  [10134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1577),
  [10136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3713),
  [10138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6327),
  [10140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6329),
  [10142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1540),
  [10144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6334),
  [10146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4383),
  [10148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4384),
  [10150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(954),
  [10152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1541),
  [10154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4316),
  [10156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4317),
  [10158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6341),
  [10160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6343),
  [10162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4157),
  [10164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6607),
  [10166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4982),
  [10168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_component, 3, 0, 26),
  [10170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3750),
  [10172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2934),
  [10174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3751),
  [10176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3078),
  [10178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6353),
  [10180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6354),
  [10182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6224),
  [10184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6359),
  [10186] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_case, 4, 0, 0),
  [10188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3487),
  [10190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2981),
  [10192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3083),
  [10194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2961),
  [10196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(955),
  [10198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6364),
  [10200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6365),
  [10202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6368),
  [10204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4189),
  [10206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3714),
  [10208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4874),
  [10210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3754),
  [10212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4319),
  [10214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3380),
  [10216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6391),
  [10218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6394),
  [10220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4323),
  [10222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6401),
  [10224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3981),
  [10226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1544),
  [10228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2404),
  [10230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3717),
  [10232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6423),
  [10234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6426),
  [10236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6433),
  [10238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2857),
  [10240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2808),
  [10242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3383),
  [10244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3384),
  [10246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2816),
  [10248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2972),
  [10250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6453),
  [10252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6455),
  [10254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3385),
  [10256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6472),
  [10258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2826),
  [10260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3718),
  [10262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1548),
  [10264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2992),
  [10266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3759),
  [10268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1962),
  [10270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3760),
  [10272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6483),
  [10274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6490),
  [10276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1552),
  [10278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6498),
  [10280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4325),
  [10282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1581),
  [10284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2696),
  [10286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3719),
  [10288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3762),
  [10290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4326),
  [10292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6505),
  [10294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6506),
  [10296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4955),
  [10298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4327),
  [10300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4328),
  [10302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [10304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4330),
  [10306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4962),
  [10308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6178),
  [10310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3003),
  [10312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6512),
  [10314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3024),
  [10316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3763),
  [10318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3764),
  [10320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2969),
  [10322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6519),
  [10324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5121),
  [10326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2941),
  [10328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5768),
  [10330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4225),
  [10332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3767),
  [10334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6525),
  [10336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4849),
  [10338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4332),
  [10340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6528),
  [10342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4333),
  [10344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3349),
  [10346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6531),
  [10348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4335),
  [10350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1095),
  [10352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2861),
  [10354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1593),
  [10356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2831),
  [10358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2835),
  [10360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2846),
  [10362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2706),
  [10364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1173),
  [10366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2814),
  [10368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2056),
  [10370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2869),
  [10372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1595),
  [10374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2868),
  [10376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2865),
  [10378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1596),
  [10380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1598),
  [10382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3330),
  [10384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1586),
  [10386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [10388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [10390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4960),
  [10392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1178),
  [10394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3301),
  [10396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1179),
  [10398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4174),
  [10400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3435),
  [10402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(659),
  [10404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3436),
  [10406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7231),
  [10408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [10410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3669),
  [10412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4340),
  [10414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2714),
  [10416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3438),
  [10418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3439),
  [10420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3440),
  [10422] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_component_body, 3, 0, 0),
  [10424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2013),
  [10426] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_process, 4, 0, 0),
  [10428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6081),
  [10430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4525),
  [10432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2783),
  [10434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2780),
  [10436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4178),
  [10438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3721),
  [10440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3783),
  [10442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3784),
  [10444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1115),
  [10446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3351),
  [10448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2998),
  [10450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3400),
  [10452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3303),
  [10454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3304),
  [10456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3442),
  [10458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3079),
  [10460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3080),
  [10462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3081),
  [10464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3017),
  [10466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [10468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6213),
  [10470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [10472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4184),
  [10474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [10476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5236),
  [10478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3014),
  [10480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(999),
  [10482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3305),
  [10484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3402),
  [10486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3149),
  [10488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2405),
  [10490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3403),
  [10492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7239),
  [10494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7240),
  [10496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(366),
  [10498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2989),
  [10500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2569),
  [10502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6202),
  [10504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3788),
  [10506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_absolute_pathname, 2, 0, 0),
  [10508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(904),
  [10510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6239),
  [10512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6822),
  [10514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3791),
  [10516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4347),
  [10518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1618),
  [10520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1970),
  [10522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4191),
  [10524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3446),
  [10526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [10528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3447),
  [10530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [10532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4348),
  [10534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(928),
  [10536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(911),
  [10538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(917),
  [10540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3448),
  [10542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3449),
  [10544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3451),
  [10546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4366),
  [10548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1623),
  [10550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4350),
  [10552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2774),
  [10554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2779),
  [10556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4200),
  [10558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1625),
  [10560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2967),
  [10562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2540),
  [10564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2841),
  [10566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4026),
  [10568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2032),
  [10570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2647),
  [10572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3453),
  [10574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3084),
  [10576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3085),
  [10578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3086),
  [10580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1098),
  [10582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2844),
  [10584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1827),
  [10586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4206),
  [10588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [10590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2845),
  [10592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4318),
  [10594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1036),
  [10596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1099),
  [10598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2850),
  [10600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1100),
  [10602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1832),
  [10604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1833),
  [10606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1834),
  [10608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2642),
  [10610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1038),
  [10612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2644),
  [10614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1047),
  [10616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1961),
  [10618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2242),
  [10620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4087),
  [10622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2854),
  [10624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [10626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2042),
  [10628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4216),
  [10630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3459),
  [10632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(683),
  [10634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3460),
  [10636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4979),
  [10638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2858),
  [10640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3729),
  [10642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2863),
  [10644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1858),
  [10646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3461),
  [10648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3462),
  [10650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3463),
  [10652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3312),
  [10654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [10656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3730),
  [10658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3296),
  [10660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2776),
  [10662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4222),
  [10664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1864),
  [10666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2809),
  [10668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2810),
  [10670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1865),
  [10672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1867),
  [10674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3464),
  [10676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3088),
  [10678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3089),
  [10680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3090),
  [10682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2811),
  [10684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5015),
  [10686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(914),
  [10688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5017),
  [10690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4231),
  [10692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [10694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3213),
  [10696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(896),
  [10698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5025),
  [10700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3455),
  [10702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3314),
  [10704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2421),
  [10706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3732),
  [10708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3315),
  [10710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4669),
  [10712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4235),
  [10714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3467),
  [10716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3468),
  [10718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1971),
  [10720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1952),
  [10722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3217),
  [10724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [10726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3469),
  [10728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3470),
  [10730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3734),
  [10732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2777),
  [10734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4240),
  [10736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3297),
  [10738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4675),
  [10740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4676),
  [10742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4242),
  [10744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3471),
  [10746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3091),
  [10748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3092),
  [10750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3093),
  [10752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1592),
  [10754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4245),
  [10756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [10758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4360),
  [10760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4680),
  [10762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4361),
  [10764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5802),
  [10766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2213),
  [10768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1892),
  [10770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4249),
  [10772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3475),
  [10774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3476),
  [10776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1972),
  [10778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3477),
  [10780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3479),
  [10782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4254),
  [10784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3224),
  [10786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3480),
  [10788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3094),
  [10790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3095),
  [10792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3096),
  [10794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4259),
  [10796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [10798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3799),
  [10800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1895),
  [10802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3484),
  [10804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4265),
  [10806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4266),
  [10808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3485),
  [10810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3097),
  [10812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3098),
  [10814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3099),
  [10816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4270),
  [10818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [10820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3571),
  [10822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3100),
  [10824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3101),
  [10826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3102),
  [10828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [10830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3103),
  [10832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3104),
  [10834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3105),
  [10836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3106),
  [10838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3107),
  [10840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3108),
  [10842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3109),
  [10844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3110),
  [10846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3111),
  [10848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3113),
  [10850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3114),
  [10852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3115),
  [10854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3117),
  [10856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3118),
  [10858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3119),
  [10860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3120),
  [10862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3121),
  [10864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3122),
  [10866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3123),
  [10868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3124),
  [10870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3125),
  [10872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3126),
  [10874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3127),
  [10876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3128),
  [10878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3129),
  [10880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3130),
  [10882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3026),
  [10884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3131),
  [10886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3132),
  [10888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3133),
  [10890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3134),
  [10892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3135),
  [10894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3136),
  [10896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3137),
  [10898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3138),
  [10900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3139),
  [10902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3140),
  [10904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3141),
  [10906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3142),
  [10908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3143),
  [10910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3144),
  [10912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3145),
  [10914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3146),
  [10916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3147),
  [10918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3148),
  [10920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3150),
  [10922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3151),
  [10924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3152),
  [10926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3153),
  [10928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3154),
  [10930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3155),
  [10932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3156),
  [10934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3157),
  [10936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3158),
  [10938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3159),
  [10940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3160),
  [10942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3161),
  [10944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3162),
  [10946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3163),
  [10948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3164),
  [10950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3165),
  [10952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3166),
  [10954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3167),
  [10956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3168),
  [10958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3169),
  [10960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3170),
  [10962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3171),
  [10964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3172),
  [10966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3173),
  [10968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3174),
  [10970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3175),
  [10972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3176),
  [10974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3177),
  [10976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3178),
  [10978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3179),
  [10980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3180),
  [10982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3181),
  [10984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3182),
  [10986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3183),
  [10988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3184),
  [10990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3185),
  [10992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3186),
  [10994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3187),
  [10996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3188),
  [10998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3189),
  [11000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3190),
  [11002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3191),
  [11004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3192),
  [11006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3193),
  [11008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3194),
  [11010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3195),
  [11012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3196),
  [11014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3197),
  [11016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3198),
  [11018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3199),
  [11020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3200),
  [11022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3201),
  [11024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3202),
  [11026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3203),
  [11028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3204),
  [11030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3205),
  [11032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3206),
  [11034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3207),
  [11036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3208),
  [11038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3209),
  [11040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3210),
  [11042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3211),
  [11044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3212),
  [11046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3214),
  [11048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3215),
  [11050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3216),
  [11052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3218),
  [11054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3219),
  [11056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3220),
  [11058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3221),
  [11060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3222),
  [11062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3223),
  [11064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3225),
  [11066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3226),
  [11068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3227),
  [11070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3228),
  [11072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3229),
  [11074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3230),
  [11076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3231),
  [11078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3232),
  [11080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3233),
  [11082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3234),
  [11084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3235),
  [11086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3027),
  [11088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5235),
  [11090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5527),
  [11092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3804),
  [11094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4364),
  [11096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3805),
  [11098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2812),
  [11100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2817),
  [11102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2818),
  [11104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2819),
  [11106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2820),
  [11108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2821),
  [11110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2025),
  [11112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2822),
  [11114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_physical_type_definition, 5, 0, 0),
  [11116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2827),
  [11118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3812),
  [11120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4369),
  [11122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3000),
  [11124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3419),
  [11126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4371),
  [11128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6258),
  [11130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2492),
  [11132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3811),
  [11134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5531),
  [11136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2829),
  [11138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2830),
  [11140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2833),
  [11142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1472),
  [11144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4373),
  [11146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6267),
  [11148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5533),
  [11150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2836),
  [11152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6270),
  [11154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_access_type_definition, 3, 0, 0),
  [11156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_statement, 2, 0, 0),
  [11158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6762),
  [11160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1902),
  [11162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4100),
  [11164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4416),
  [11166] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 5, 0, 0),
  [11168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1486),
  [11170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2630),
  [11172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1488),
  [11174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5133),
  [11176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1491),
  [11178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2288),
  [11180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4427),
  [11182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2997),
  [11184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2495),
  [11186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1699),
  [11188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(934),
  [11190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1780),
  [11192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3813),
  [11194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3238),
  [11196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3033),
  [11198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5319),
  [11200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4543),
  [11202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__element_mode_indication, 1, 0, 0),
  [11204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [11206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_file_type_definition, 3, 0, 22),
  [11208] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_instantiation_definition, 3, 0, 0),
  [11210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2161),
  [11212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2162),
  [11214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2671),
  [11216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3276),
  [11218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3535),
  [11220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3738),
  [11222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6813),
  [11224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4159),
  [11226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2164),
  [11228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2881),
  [11230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_body, 3, 0, 0),
  [11232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3001),
  [11234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4745),
  [11236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2888),
  [11238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4984),
  [11240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3818),
  [11242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4747),
  [11244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3739),
  [11246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4414),
  [11248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6750),
  [11250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3740),
  [11252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4751),
  [11254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2187),
  [11256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4752),
  [11258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4755),
  [11260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3843),
  [11262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4420),
  [11264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3537),
  [11266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4756),
  [11268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3672),
  [11270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3009),
  [11272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3820),
  [11274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3673),
  [11276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3035),
  [11278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3538),
  [11280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3036),
  [11282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3284),
  [11284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4760),
  [11286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4196),
  [11288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__element_mode_view_indication, 1, 0, 0),
  [11290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4823),
  [11292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_protected_type_declaration, 3, 0, 0),
  [11294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6977),
  [11296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3012),
  [11298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3821),
  [11300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3039),
  [11302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4404),
  [11304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_view, 3, 0, 55),
  [11306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3682),
  [11308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3539),
  [11310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3822),
  [11312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3741),
  [11314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2655),
  [11316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5035),
  [11318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2926),
  [11320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2928),
  [11322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_generate, 2, 0, 0),
  [11324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4473),
  [11326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2929),
  [11328] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_conditional_analysis, 2, 0, 0),
  [11330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2930),
  [11332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4967),
  [11334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1437),
  [11336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3239),
  [11338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3040),
  [11340] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_type_definition, 3, 0, 0),
  [11342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3041),
  [11344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5189),
  [11346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3685),
  [11348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2197),
  [11350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4475),
  [11352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4476),
  [11354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3043),
  [11356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3688),
  [11358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3689),
  [11360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3690),
  [11362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3691),
  [11364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3044),
  [11366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3344),
  [11368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3045),
  [11370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3046),
  [11372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4477),
  [11374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5246),
  [11376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enumeration_type_definition, 3, 0, 0),
  [11378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8147),
  [11380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3048),
  [11382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2588),
  [11384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4417),
  [11386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3696),
  [11388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1460),
  [11390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3699),
  [11392] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_relative_pathname, 1, 0, 0),
  [11394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2906),
  [11396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [11398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2915),
  [11400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2916),
  [11402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pathname_element, 4, 0, 0),
  [11404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_generate, 5, 0, 0),
  [11406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(818),
  [11408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2905),
  [11410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3683),
  [11412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3472),
  [11414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2907),
  [11416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5230),
  [11418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2878),
  [11420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2899),
  [11422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2451),
  [11424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2503),
  [11426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2200),
  [11428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2201),
  [11430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2202),
  [11432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1252),
  [11434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2203),
  [11436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3702),
  [11438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2692),
  [11440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3053),
  [11442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3054),
  [11444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3055),
  [11446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3282),
  [11448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5572),
  [11450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5575),
  [11452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4834),
  [11454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3057),
  [11456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3058),
  [11458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4418),
  [11460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3708),
  [11462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3749),
  [11464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3711),
  [11466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(8364),
  [11468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2919),
  [11470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2898),
  [11472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2873),
  [11474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2876),
  [11476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1254),
  [11478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2870),
  [11480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2904),
  [11482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2908),
  [11484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1255),
  [11486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2911),
  [11488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2917),
  [11490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2918),
  [11492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1256),
  [11494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3237),
  [11496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2531),
  [11498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3298),
  [11500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4839),
  [11502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2208),
  [11504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3733),
  [11506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3737),
  [11508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6527),
  [11510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3062),
  [11512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3063),
  [11514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3743),
  [11516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3064),
  [11518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2927),
  [11520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2901),
  [11522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2882),
  [11524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2903),
  [11526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2883),
  [11528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2889),
  [11530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2893),
  [11532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [11534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2909),
  [11536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3065),
  [11538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4439),
  [11540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4841),
  [11542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3747),
  [11544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2921),
  [11546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5113),
  [11548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3331),
  [11550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4392),
  [11552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1153),
  [11554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3578),
  [11556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2925),
  [11558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2871),
  [11560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3748),
  [11562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5139),
  [11564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5346),
  [11566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4393),
  [11568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4788),
  [11570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3580),
  [11572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2875),
  [11574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2181),
  [11576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1898),
  [11578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5054),
  [11580] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_warning_directive, 2, 0, 0),
  [11582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4394),
  [11584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3547),
  [11586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3582),
  [11588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4505),
  [11590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6783),
  [11592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3757),
  [11594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5064),
  [11596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1012),
  [11598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4395),
  [11600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4504),
  [11602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5071),
  [11604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2634),
  [11606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4396),
  [11608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4795),
  [11610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4507),
  [11612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5075),
  [11614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [11616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4397),
  [11618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1015),
  [11620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1016),
  [11622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4398),
  [11624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4796),
  [11626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4479),
  [11628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4799),
  [11630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4492),
  [11632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2771),
  [11634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5101),
  [11636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [11638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3549),
  [11640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7037),
  [11642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1176),
  [11644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3550),
  [11646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1276),
  [11648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2272),
  [11650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4132),
  [11652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1185),
  [11654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [11656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [11658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2953),
  [11660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1278),
  [11662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6133),
  [11664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1279),
  [11666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2010),
  [11668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2031),
  [11670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4424),
  [11672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_end_architecture, 3, 0, 57),
  [11674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2080),
  [11676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2958),
  [11678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4150),
  [11680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1292),
  [11682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2568),
  [11684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [11686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2973),
  [11688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4888),
  [11690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1294),
  [11692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [11694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1296),
  [11696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [11698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [11700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1032),
  [11702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2539),
  [11704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5076),
  [11706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2390),
  [11708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3777),
  [11710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6767),
  [11712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2033),
  [11714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5370),
  [11716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1957),
  [11718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5080),
  [11720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6509),
  [11722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block_comment, 3, 0, 0),
  [11724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_line_comment, 2, 0, 0),
  [11726] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tool_directive, 3, 0, 0),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token_identifier = 0,
  ts_external_token_ABS = 1,
  ts_external_token_ACCESS = 2,
  ts_external_token_AFTER = 3,
  ts_external_token_ALIAS = 4,
  ts_external_token_ALL = 5,
  ts_external_token_AND = 6,
  ts_external_token_ARCHITECTURE = 7,
  ts_external_token_ARRAY = 8,
  ts_external_token_ASSERT = 9,
  ts_external_token_ASSUME = 10,
  ts_external_token_ATTRIBUTE = 11,
  ts_external_token_BEGIN = 12,
  ts_external_token_BLOCK = 13,
  ts_external_token_BODY = 14,
  ts_external_token_BUFFER = 15,
  ts_external_token_BUS = 16,
  ts_external_token_CASE = 17,
  ts_external_token_COMPONENT = 18,
  ts_external_token_CONFIGURATION = 19,
  ts_external_token_CONSTANT = 20,
  ts_external_token_CONTEXT = 21,
  ts_external_token_COVER = 22,
  ts_external_token_DEFAULT = 23,
  ts_external_token_DISCONNECT = 24,
  ts_external_token_DOWNTO = 25,
  ts_external_token_ELSE = 26,
  ts_external_token_ELSIF = 27,
  ts_external_token_END = 28,
  ts_external_token_ENTITY = 29,
  ts_external_token_EXIT = 30,
  ts_external_token_FAIRNESS = 31,
  ts_external_token_FILE = 32,
  ts_external_token_FOR = 33,
  ts_external_token_FORCE = 34,
  ts_external_token_FUNCTION = 35,
  ts_external_token_GENERATE = 36,
  ts_external_token_GENERIC = 37,
  ts_external_token_GROUP = 38,
  ts_external_token_GUARDED = 39,
  ts_external_token_IF = 40,
  ts_external_token_IMPURE = 41,
  ts_external_token_IN = 42,
  ts_external_token_INERTIAL = 43,
  ts_external_token_INOUT = 44,
  ts_external_token_IS = 45,
  ts_external_token_LABEL = 46,
  ts_external_token_LIBRARY = 47,
  ts_external_token_LINKAGE = 48,
  ts_external_token_LITERAL = 49,
  ts_external_token_LOOP = 50,
  ts_external_token_MAP = 51,
  ts_external_token_MOD = 52,
  ts_external_token_NAND = 53,
  ts_external_token_NEW = 54,
  ts_external_token_NEXT = 55,
  ts_external_token_NOR = 56,
  ts_external_token_NOT = 57,
  ts_external_token_NULL = 58,
  ts_external_token_OF = 59,
  ts_external_token_ON = 60,
  ts_external_token_OPEN = 61,
  ts_external_token_OR = 62,
  ts_external_token_OTHERS = 63,
  ts_external_token_OUT = 64,
  ts_external_token_PACKAGE = 65,
  ts_external_token_PARAMETER = 66,
  ts_external_token_PORT = 67,
  ts_external_token_POSTPONED = 68,
  ts_external_token_PROCEDURE = 69,
  ts_external_token_PROCESS = 70,
  ts_external_token_PROPERTY = 71,
  ts_external_token_PROTECTED = 72,
  ts_external_token_PRIVATE = 73,
  ts_external_token_PURE = 74,
  ts_external_token_RANGE = 75,
  ts_external_token_RECORD = 76,
  ts_external_token_REGISTER = 77,
  ts_external_token_REJECT = 78,
  ts_external_token_RELEASE = 79,
  ts_external_token_REM = 80,
  ts_external_token_REPORT = 81,
  ts_external_token_RESTRICT = 82,
  ts_external_token_RETURN = 83,
  ts_external_token_ROL = 84,
  ts_external_token_ROR = 85,
  ts_external_token_SELECT = 86,
  ts_external_token_SEQUENCE = 87,
  ts_external_token_SEVERITY = 88,
  ts_external_token_SIGNAL = 89,
  ts_external_token_SHARED = 90,
  ts_external_token_SLA = 91,
  ts_external_token_SLL = 92,
  ts_external_token_SRA = 93,
  ts_external_token_SRL = 94,
  ts_external_token_STRONG = 95,
  ts_external_token_SUBTYPE = 96,
  ts_external_token_THEN = 97,
  ts_external_token_TO = 98,
  ts_external_token_TRANSPORT = 99,
  ts_external_token_TYPE = 100,
  ts_external_token_UNAFFECTED = 101,
  ts_external_token_UNITS = 102,
  ts_external_token_UNTIL = 103,
  ts_external_token_USE = 104,
  ts_external_token_VARIABLE = 105,
  ts_external_token_VIEW = 106,
  ts_external_token_VMODE = 107,
  ts_external_token_VPKG = 108,
  ts_external_token_VPROP = 109,
  ts_external_token_VUNIT = 110,
  ts_external_token_WAIT = 111,
  ts_external_token_WHEN = 112,
  ts_external_token_WHILE = 113,
  ts_external_token_WITH = 114,
  ts_external_token_XNOR = 115,
  ts_external_token_XOR = 116,
  ts_external_token_reserved_end_marker = 117,
  ts_external_token_directive_body = 118,
  ts_external_token_directive_constant_builtin = 119,
  ts_external_token_directive_error = 120,
  ts_external_token_directive_protect = 121,
  ts_external_token_directive_warning = 122,
  ts_external_token__directive_newline = 123,
  ts_external_token__grave_accent = 124,
  ts_external_token_box = 125,
  ts_external_token_delimiter_end_marker = 126,
  ts_external_token_decimal_integer = 127,
  ts_external_token_decimal_float = 128,
  ts_external_token_based_base = 129,
  ts_external_token_based_integer = 130,
  ts_external_token_based_float = 131,
  ts_external_token_character_literal = 132,
  ts_external_token_string_literal = 133,
  ts_external_token_string_literal_std_logic = 134,
  ts_external_token_bit_string_length = 135,
  ts_external_token_bit_string_base = 136,
  ts_external_token_bit_string_value = 137,
  ts_external_token_operator_symbol = 138,
  ts_external_token__line_comment_start = 139,
  ts_external_token__block_comment_start = 140,
  ts_external_token__block_comment_end = 141,
  ts_external_token_comment_content = 142,
  ts_external_token_token_end_marker = 143,
  ts_external_token_attribute_function = 144,
  ts_external_token_attribute_impure_function = 145,
  ts_external_token_attribute_mode_view = 146,
  ts_external_token_attribute_pure_function = 147,
  ts_external_token_attribute_range = 148,
  ts_external_token_attribute_signal = 149,
  ts_external_token_attribute_subtype = 150,
  ts_external_token_attribute_type = 151,
  ts_external_token_attribute_value = 152,
  ts_external_token_library_attribute = 153,
  ts_external_token_library_constant = 154,
  ts_external_token_library_constant_boolean = 155,
  ts_external_token_library_constant_character = 156,
  ts_external_token_library_constant_debug = 157,
  ts_external_token_library_constant_env = 158,
  ts_external_token_library_constant_standard = 159,
  ts_external_token_library_constant_std_logic = 160,
  ts_external_token_library_constant_unit = 161,
  ts_external_token_library_function = 162,
  ts_external_token_library_namespace = 163,
  ts_external_token_library_type = 164,
  ts_external_token__end_of_file = 165,
  ts_external_token_error_sentinel = 166,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token_identifier] = sym_identifier,
  [ts_external_token_ABS] = sym_ABS,
  [ts_external_token_ACCESS] = sym_ACCESS,
  [ts_external_token_AFTER] = sym_AFTER,
  [ts_external_token_ALIAS] = sym_ALIAS,
  [ts_external_token_ALL] = sym_ALL,
  [ts_external_token_AND] = sym_AND,
  [ts_external_token_ARCHITECTURE] = sym_ARCHITECTURE,
  [ts_external_token_ARRAY] = sym_ARRAY,
  [ts_external_token_ASSERT] = sym_ASSERT,
  [ts_external_token_ASSUME] = sym_ASSUME,
  [ts_external_token_ATTRIBUTE] = sym_ATTRIBUTE,
  [ts_external_token_BEGIN] = sym_BEGIN,
  [ts_external_token_BLOCK] = sym_BLOCK,
  [ts_external_token_BODY] = sym_BODY,
  [ts_external_token_BUFFER] = sym_BUFFER,
  [ts_external_token_BUS] = sym_BUS,
  [ts_external_token_CASE] = sym_CASE,
  [ts_external_token_COMPONENT] = sym_COMPONENT,
  [ts_external_token_CONFIGURATION] = sym_CONFIGURATION,
  [ts_external_token_CONSTANT] = sym_CONSTANT,
  [ts_external_token_CONTEXT] = sym_CONTEXT,
  [ts_external_token_COVER] = sym_COVER,
  [ts_external_token_DEFAULT] = sym_DEFAULT,
  [ts_external_token_DISCONNECT] = sym_DISCONNECT,
  [ts_external_token_DOWNTO] = sym_DOWNTO,
  [ts_external_token_ELSE] = sym_ELSE,
  [ts_external_token_ELSIF] = sym_ELSIF,
  [ts_external_token_END] = sym_END,
  [ts_external_token_ENTITY] = sym_ENTITY,
  [ts_external_token_EXIT] = sym_EXIT,
  [ts_external_token_FAIRNESS] = sym_FAIRNESS,
  [ts_external_token_FILE] = sym_FILE,
  [ts_external_token_FOR] = sym_FOR,
  [ts_external_token_FORCE] = sym_FORCE,
  [ts_external_token_FUNCTION] = sym_FUNCTION,
  [ts_external_token_GENERATE] = sym_GENERATE,
  [ts_external_token_GENERIC] = sym_GENERIC,
  [ts_external_token_GROUP] = sym_GROUP,
  [ts_external_token_GUARDED] = sym_GUARDED,
  [ts_external_token_IF] = sym_IF,
  [ts_external_token_IMPURE] = sym_IMPURE,
  [ts_external_token_IN] = sym_IN,
  [ts_external_token_INERTIAL] = sym_INERTIAL,
  [ts_external_token_INOUT] = sym_INOUT,
  [ts_external_token_IS] = sym_IS,
  [ts_external_token_LABEL] = sym_LABEL,
  [ts_external_token_LIBRARY] = sym_LIBRARY,
  [ts_external_token_LINKAGE] = sym_LINKAGE,
  [ts_external_token_LITERAL] = sym_LITERAL,
  [ts_external_token_LOOP] = sym_LOOP,
  [ts_external_token_MAP] = sym_MAP,
  [ts_external_token_MOD] = sym_MOD,
  [ts_external_token_NAND] = sym_NAND,
  [ts_external_token_NEW] = sym_NEW,
  [ts_external_token_NEXT] = sym_NEXT,
  [ts_external_token_NOR] = sym_NOR,
  [ts_external_token_NOT] = sym_NOT,
  [ts_external_token_NULL] = sym_NULL,
  [ts_external_token_OF] = sym_OF,
  [ts_external_token_ON] = sym_ON,
  [ts_external_token_OPEN] = sym_OPEN,
  [ts_external_token_OR] = sym_OR,
  [ts_external_token_OTHERS] = sym_OTHERS,
  [ts_external_token_OUT] = sym_OUT,
  [ts_external_token_PACKAGE] = sym_PACKAGE,
  [ts_external_token_PARAMETER] = sym_PARAMETER,
  [ts_external_token_PORT] = sym_PORT,
  [ts_external_token_POSTPONED] = sym_POSTPONED,
  [ts_external_token_PROCEDURE] = sym_PROCEDURE,
  [ts_external_token_PROCESS] = sym_PROCESS,
  [ts_external_token_PROPERTY] = sym_PROPERTY,
  [ts_external_token_PROTECTED] = sym_PROTECTED,
  [ts_external_token_PRIVATE] = sym_PRIVATE,
  [ts_external_token_PURE] = sym_PURE,
  [ts_external_token_RANGE] = sym_RANGE,
  [ts_external_token_RECORD] = sym_RECORD,
  [ts_external_token_REGISTER] = sym_REGISTER,
  [ts_external_token_REJECT] = sym_REJECT,
  [ts_external_token_RELEASE] = sym_RELEASE,
  [ts_external_token_REM] = sym_REM,
  [ts_external_token_REPORT] = sym_REPORT,
  [ts_external_token_RESTRICT] = sym_RESTRICT,
  [ts_external_token_RETURN] = sym_RETURN,
  [ts_external_token_ROL] = sym_ROL,
  [ts_external_token_ROR] = sym_ROR,
  [ts_external_token_SELECT] = sym_SELECT,
  [ts_external_token_SEQUENCE] = sym_SEQUENCE,
  [ts_external_token_SEVERITY] = sym_SEVERITY,
  [ts_external_token_SIGNAL] = sym_SIGNAL,
  [ts_external_token_SHARED] = sym_SHARED,
  [ts_external_token_SLA] = sym_SLA,
  [ts_external_token_SLL] = sym_SLL,
  [ts_external_token_SRA] = sym_SRA,
  [ts_external_token_SRL] = sym_SRL,
  [ts_external_token_STRONG] = sym_STRONG,
  [ts_external_token_SUBTYPE] = sym_SUBTYPE,
  [ts_external_token_THEN] = sym_THEN,
  [ts_external_token_TO] = sym_TO,
  [ts_external_token_TRANSPORT] = sym_TRANSPORT,
  [ts_external_token_TYPE] = sym_TYPE,
  [ts_external_token_UNAFFECTED] = sym_UNAFFECTED,
  [ts_external_token_UNITS] = sym_UNITS,
  [ts_external_token_UNTIL] = sym_UNTIL,
  [ts_external_token_USE] = sym_USE,
  [ts_external_token_VARIABLE] = sym_VARIABLE,
  [ts_external_token_VIEW] = sym_VIEW,
  [ts_external_token_VMODE] = sym_VMODE,
  [ts_external_token_VPKG] = sym_VPKG,
  [ts_external_token_VPROP] = sym_VPROP,
  [ts_external_token_VUNIT] = sym_VUNIT,
  [ts_external_token_WAIT] = sym_WAIT,
  [ts_external_token_WHEN] = sym_WHEN,
  [ts_external_token_WHILE] = sym_WHILE,
  [ts_external_token_WITH] = sym_WITH,
  [ts_external_token_XNOR] = sym_XNOR,
  [ts_external_token_XOR] = sym_XOR,
  [ts_external_token_reserved_end_marker] = sym_reserved_end_marker,
  [ts_external_token_directive_body] = sym_directive_body,
  [ts_external_token_directive_constant_builtin] = sym_directive_constant_builtin,
  [ts_external_token_directive_error] = sym_directive_error,
  [ts_external_token_directive_protect] = sym_directive_protect,
  [ts_external_token_directive_warning] = sym_directive_warning,
  [ts_external_token__directive_newline] = sym__directive_newline,
  [ts_external_token__grave_accent] = sym__grave_accent,
  [ts_external_token_box] = sym_box,
  [ts_external_token_delimiter_end_marker] = sym_delimiter_end_marker,
  [ts_external_token_decimal_integer] = sym_decimal_integer,
  [ts_external_token_decimal_float] = sym_decimal_float,
  [ts_external_token_based_base] = sym_based_base,
  [ts_external_token_based_integer] = sym_based_integer,
  [ts_external_token_based_float] = sym_based_float,
  [ts_external_token_character_literal] = sym_character_literal,
  [ts_external_token_string_literal] = sym_string_literal,
  [ts_external_token_string_literal_std_logic] = sym_string_literal_std_logic,
  [ts_external_token_bit_string_length] = sym_bit_string_length,
  [ts_external_token_bit_string_base] = sym_bit_string_base,
  [ts_external_token_bit_string_value] = sym_bit_string_value,
  [ts_external_token_operator_symbol] = sym_operator_symbol,
  [ts_external_token__line_comment_start] = sym__line_comment_start,
  [ts_external_token__block_comment_start] = sym__block_comment_start,
  [ts_external_token__block_comment_end] = sym__block_comment_end,
  [ts_external_token_comment_content] = sym_comment_content,
  [ts_external_token_token_end_marker] = sym_token_end_marker,
  [ts_external_token_attribute_function] = sym_attribute_function,
  [ts_external_token_attribute_impure_function] = sym_attribute_impure_function,
  [ts_external_token_attribute_mode_view] = sym_attribute_mode_view,
  [ts_external_token_attribute_pure_function] = sym_attribute_pure_function,
  [ts_external_token_attribute_range] = sym_attribute_range,
  [ts_external_token_attribute_signal] = sym_attribute_signal,
  [ts_external_token_attribute_subtype] = sym_attribute_subtype,
  [ts_external_token_attribute_type] = sym_attribute_type,
  [ts_external_token_attribute_value] = sym_attribute_value,
  [ts_external_token_library_attribute] = sym_library_attribute,
  [ts_external_token_library_constant] = sym_library_constant,
  [ts_external_token_library_constant_boolean] = sym_library_constant_boolean,
  [ts_external_token_library_constant_character] = sym_library_constant_character,
  [ts_external_token_library_constant_debug] = sym_library_constant_debug,
  [ts_external_token_library_constant_env] = sym_library_constant_env,
  [ts_external_token_library_constant_standard] = sym_library_constant_standard,
  [ts_external_token_library_constant_std_logic] = sym_library_constant_std_logic,
  [ts_external_token_library_constant_unit] = sym_library_constant_unit,
  [ts_external_token_library_function] = sym_library_function,
  [ts_external_token_library_namespace] = sym_library_namespace,
  [ts_external_token_library_type] = sym_library_type,
  [ts_external_token__end_of_file] = sym__end_of_file,
  [ts_external_token_error_sentinel] = sym_error_sentinel,
};

static const bool ts_external_scanner_states[291][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_ACCESS] = true,
    [ts_external_token_AFTER] = true,
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ALL] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_ARCHITECTURE] = true,
    [ts_external_token_ARRAY] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_ASSUME] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_BLOCK] = true,
    [ts_external_token_BODY] = true,
    [ts_external_token_BUFFER] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_CASE] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONFIGURATION] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_CONTEXT] = true,
    [ts_external_token_COVER] = true,
    [ts_external_token_DEFAULT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_ELSE] = true,
    [ts_external_token_ELSIF] = true,
    [ts_external_token_END] = true,
    [ts_external_token_ENTITY] = true,
    [ts_external_token_EXIT] = true,
    [ts_external_token_FAIRNESS] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FORCE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GENERATE] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_GUARDED] = true,
    [ts_external_token_IF] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_IN] = true,
    [ts_external_token_INERTIAL] = true,
    [ts_external_token_INOUT] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_LABEL] = true,
    [ts_external_token_LIBRARY] = true,
    [ts_external_token_LINKAGE] = true,
    [ts_external_token_LITERAL] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_MAP] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NEXT] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OF] = true,
    [ts_external_token_ON] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_OTHERS] = true,
    [ts_external_token_OUT] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_PROPERTY] = true,
    [ts_external_token_PROTECTED] = true,
    [ts_external_token_PRIVATE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_RECORD] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token_REJECT] = true,
    [ts_external_token_RELEASE] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_REPORT] = true,
    [ts_external_token_RESTRICT] = true,
    [ts_external_token_RETURN] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SELECT] = true,
    [ts_external_token_SEQUENCE] = true,
    [ts_external_token_SEVERITY] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_STRONG] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_THEN] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_TRANSPORT] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_UNAFFECTED] = true,
    [ts_external_token_UNITS] = true,
    [ts_external_token_UNTIL] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token_VMODE] = true,
    [ts_external_token_VPKG] = true,
    [ts_external_token_VPROP] = true,
    [ts_external_token_VUNIT] = true,
    [ts_external_token_WAIT] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_WHILE] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token_reserved_end_marker] = true,
    [ts_external_token_directive_body] = true,
    [ts_external_token_directive_constant_builtin] = true,
    [ts_external_token_directive_error] = true,
    [ts_external_token_directive_protect] = true,
    [ts_external_token_directive_warning] = true,
    [ts_external_token__directive_newline] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_box] = true,
    [ts_external_token_delimiter_end_marker] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_based_integer] = true,
    [ts_external_token_based_float] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_bit_string_value] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token__block_comment_end] = true,
    [ts_external_token_comment_content] = true,
    [ts_external_token_token_end_marker] = true,
    [ts_external_token_attribute_function] = true,
    [ts_external_token_attribute_impure_function] = true,
    [ts_external_token_attribute_mode_view] = true,
    [ts_external_token_attribute_pure_function] = true,
    [ts_external_token_attribute_range] = true,
    [ts_external_token_attribute_signal] = true,
    [ts_external_token_attribute_subtype] = true,
    [ts_external_token_attribute_type] = true,
    [ts_external_token_attribute_value] = true,
    [ts_external_token_library_attribute] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_namespace] = true,
    [ts_external_token_library_type] = true,
    [ts_external_token__end_of_file] = true,
    [ts_external_token_error_sentinel] = true,
  },
  [2] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ARCHITECTURE] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONFIGURATION] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_CONTEXT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_ENTITY] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_LIBRARY] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
    [ts_external_token__end_of_file] = true,
  },
  [3] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_ELSE] = true,
    [ts_external_token_ELSIF] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [4] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [5] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [6] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_BLOCK] = true,
    [ts_external_token_CASE] = true,
    [ts_external_token_ELSE] = true,
    [ts_external_token_ELSIF] = true,
    [ts_external_token_END] = true,
    [ts_external_token_EXIT] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_IF] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_NEXT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_REPORT] = true,
    [ts_external_token_RETURN] = true,
    [ts_external_token_WAIT] = true,
    [ts_external_token_WHILE] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [7] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_OTHERS] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [8] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_BLOCK] = true,
    [ts_external_token_CASE] = true,
    [ts_external_token_END] = true,
    [ts_external_token_EXIT] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_IF] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_NEXT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_REPORT] = true,
    [ts_external_token_RETURN] = true,
    [ts_external_token_WAIT] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_WHILE] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [9] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_BLOCK] = true,
    [ts_external_token_CASE] = true,
    [ts_external_token_END] = true,
    [ts_external_token_EXIT] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_IF] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_NEXT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_REPORT] = true,
    [ts_external_token_RETURN] = true,
    [ts_external_token_WAIT] = true,
    [ts_external_token_WHILE] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [10] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_INERTIAL] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [11] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_FORCE] = true,
    [ts_external_token_INERTIAL] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REJECT] = true,
    [ts_external_token_TRANSPORT] = true,
    [ts_external_token_UNAFFECTED] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [12] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_FORCE] = true,
    [ts_external_token_INERTIAL] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REJECT] = true,
    [ts_external_token_RELEASE] = true,
    [ts_external_token_TRANSPORT] = true,
    [ts_external_token_UNAFFECTED] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [13] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_GUARDED] = true,
    [ts_external_token_INERTIAL] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REJECT] = true,
    [ts_external_token_TRANSPORT] = true,
    [ts_external_token_UNAFFECTED] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [14] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_INERTIAL] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REJECT] = true,
    [ts_external_token_TRANSPORT] = true,
    [ts_external_token_UNAFFECTED] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [15] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_OTHERS] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [16] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_DEFAULT] = true,
    [ts_external_token_INERTIAL] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_box] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [17] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [18] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_UNAFFECTED] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [19] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_IN] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_OUT] = true,
    [ts_external_token_UNAFFECTED] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [20] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_UNAFFECTED] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [21] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_IN] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_OUT] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [22] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ABS] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token_string_literal_std_logic] = true,
    [ts_external_token_bit_string_length] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_boolean] = true,
    [ts_external_token_library_constant_character] = true,
    [ts_external_token_library_constant_debug] = true,
    [ts_external_token_library_constant_env] = true,
    [ts_external_token_library_constant_standard] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [23] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [24] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [25] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [26] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [27] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [28] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [29] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [30] = {
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [31] = {
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [32] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [33] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [34] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [35] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OF] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [36] = {
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [37] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_UNITS] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [38] = {
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [39] = {
    [ts_external_token_AFTER] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [40] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_REPORT] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SEVERITY] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [41] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [42] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERATE] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [43] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [44] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [45] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [46] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [47] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OF] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [48] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_UNITS] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [49] = {
    [ts_external_token_AND] = true,
    [ts_external_token_ELSE] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [50] = {
    [ts_external_token_AND] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [51] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [52] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SEVERITY] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [53] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [54] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_INERTIAL] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [55] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_THEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [56] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [57] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [58] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SELECT] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [59] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [60] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERATE] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [61] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [62] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [63] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [64] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [65] = {
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [66] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [67] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_ELSE] = true,
    [ts_external_token_ELSIF] = true,
    [ts_external_token_END] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [68] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [69] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [70] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [71] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [72] = {
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [73] = {
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [74] = {
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [75] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [76] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [77] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [78] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [79] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [80] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [81] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [82] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ARCHITECTURE] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONFIGURATION] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_CONTEXT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_ENTITY] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_LIBRARY] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
    [ts_external_token__end_of_file] = true,
  },
  [83] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [84] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [85] = {
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_REPORT] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SEVERITY] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [86] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OF] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [87] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AFTER] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [88] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_UNITS] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [89] = {
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [90] = {
    [ts_external_token_AND] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [91] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OF] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [92] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [93] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_UNITS] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [94] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [95] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_BLOCK] = true,
    [ts_external_token_CASE] = true,
    [ts_external_token_EXIT] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_IF] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_NEXT] = true,
    [ts_external_token_NULL] = true,
    [ts_external_token_REPORT] = true,
    [ts_external_token_RETURN] = true,
    [ts_external_token_WAIT] = true,
    [ts_external_token_WHILE] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [96] = {
    [ts_external_token_AFTER] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [97] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [98] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_REPORT] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SEVERITY] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [99] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERATE] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [100] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [101] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [102] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_ELSE] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [103] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [104] = {
    [ts_external_token_AND] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [105] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_GENERATE] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [106] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [107] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [108] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OF] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [109] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [110] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SEVERITY] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [111] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_UNITS] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [112] = {
    [ts_external_token_AND] = true,
    [ts_external_token_GENERATE] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [113] = {
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SEVERITY] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [114] = {
    [ts_external_token_AND] = true,
    [ts_external_token_ELSE] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [115] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [116] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_THEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [117] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [118] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SELECT] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [119] = {
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OF] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [120] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_INERTIAL] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [121] = {
    [ts_external_token_AND] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [122] = {
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_UNITS] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [123] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [124] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_AND] = true,
    [ts_external_token_GENERATE] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [125] = {
    [ts_external_token_AND] = true,
    [ts_external_token_INERTIAL] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [126] = {
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_THEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [127] = {
    [ts_external_token_AND] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [128] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [129] = {
    [ts_external_token_AND] = true,
    [ts_external_token_LOOP] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [130] = {
    [ts_external_token_AND] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SELECT] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [131] = {
    [ts_external_token_AND] = true,
    [ts_external_token_DOWNTO] = true,
    [ts_external_token_MOD] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_REM] = true,
    [ts_external_token_ROL] = true,
    [ts_external_token_ROR] = true,
    [ts_external_token_SLA] = true,
    [ts_external_token_SLL] = true,
    [ts_external_token_SRA] = true,
    [ts_external_token_SRL] = true,
    [ts_external_token_TO] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [132] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_BLOCK] = true,
    [ts_external_token_CASE] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONFIGURATION] = true,
    [ts_external_token_ENTITY] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_IF] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [133] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_BUFFER] = true,
    [ts_external_token_IN] = true,
    [ts_external_token_INOUT] = true,
    [ts_external_token_LINKAGE] = true,
    [ts_external_token_OUT] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [134] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [135] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [136] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [137] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [138] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [139] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [140] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BODY] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PRIVATE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [141] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [142] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_BEGIN] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [143] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token_WITH] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [144] = {
    [ts_external_token_ARCHITECTURE] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONFIGURATION] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_ENTITY] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_LABEL] = true,
    [ts_external_token_LITERAL] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PROPERTY] = true,
    [ts_external_token_SEQUENCE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_UNITS] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [145] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [146] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_DISCONNECT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_SHARED] = true,
    [ts_external_token_SUBTYPE] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [147] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [148] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_TYPE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [149] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PRIVATE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [150] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_decimal_integer] = true,
    [ts_external_token_decimal_float] = true,
    [ts_external_token_based_base] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [151] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ELSE] = true,
    [ts_external_token_ELSIF] = true,
    [ts_external_token_END] = true,
    [ts_external_token_IF] = true,
    [ts_external_token_directive_error] = true,
    [ts_external_token_directive_protect] = true,
    [ts_external_token_directive_warning] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [152] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [153] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [154] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALL] = true,
    [ts_external_token_OTHERS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [155] = {
    [ts_external_token_ACCESS] = true,
    [ts_external_token_ARRAY] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_NEW] = true,
    [ts_external_token_PROTECTED] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_RECORD] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [156] = {
    [ts_external_token_BUS] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [157] = {
    [ts_external_token_END] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [158] = {
    [ts_external_token_BUS] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [159] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [160] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_attribute_function] = true,
    [ts_external_token_attribute_impure_function] = true,
    [ts_external_token_attribute_mode_view] = true,
    [ts_external_token_attribute_pure_function] = true,
    [ts_external_token_attribute_range] = true,
    [ts_external_token_attribute_signal] = true,
    [ts_external_token_attribute_subtype] = true,
    [ts_external_token_attribute_type] = true,
    [ts_external_token_attribute_value] = true,
    [ts_external_token_library_attribute] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [161] = {
    [ts_external_token_BUS] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [162] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ASSERT] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [163] = {
    [ts_external_token_ACCESS] = true,
    [ts_external_token_ARRAY] = true,
    [ts_external_token_FILE] = true,
    [ts_external_token_PRIVATE] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_UNITS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_box] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [164] = {
    [ts_external_token_FOR] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_UNTIL] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [165] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [166] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALL] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [167] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_RETURN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [168] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [169] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_OF] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [170] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [171] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RETURN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [172] = {
    [ts_external_token_END] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [173] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_box] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [174] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [175] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [176] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [177] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_constant_unit] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [178] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_namespace] = true,
    [ts_external_token_library_type] = true,
  },
  [179] = {
    [ts_external_token_AFTER] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [180] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [181] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [182] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_BUS] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [183] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALL] = true,
    [ts_external_token_OTHERS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [184] = {
    [ts_external_token_END] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [185] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [186] = {
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [187] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [188] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_NOT] = true,
    [ts_external_token_directive_constant_builtin] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [189] = {
    [ts_external_token_ALIAS] = true,
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_END] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IMPURE] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token_PRIVATE] = true,
    [ts_external_token_PURE] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [190] = {
    [ts_external_token_ATTRIBUTE] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_GROUP] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [191] = {
    [ts_external_token_BUFFER] = true,
    [ts_external_token_IN] = true,
    [ts_external_token_INOUT] = true,
    [ts_external_token_LINKAGE] = true,
    [ts_external_token_OUT] = true,
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [192] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_constant_std_logic] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [193] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_END] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [194] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_VUNIT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_namespace] = true,
    [ts_external_token_library_type] = true,
  },
  [195] = {
    [ts_external_token_END] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [196] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_FUNCTION] = true,
    [ts_external_token_PROCEDURE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [197] = {
    [ts_external_token_AND] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_THEN] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [198] = {
    [ts_external_token_CONTEXT] = true,
    [ts_external_token_END] = true,
    [ts_external_token_LIBRARY] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [199] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_OF] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [200] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [201] = {
    [ts_external_token_AND] = true,
    [ts_external_token_NAND] = true,
    [ts_external_token_NOR] = true,
    [ts_external_token_OR] = true,
    [ts_external_token_XNOR] = true,
    [ts_external_token_XOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [202] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_namespace] = true,
    [ts_external_token_library_type] = true,
  },
  [203] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALL] = true,
    [ts_external_token_OTHERS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [204] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [205] = {
    [ts_external_token_END] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [206] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_GENERATE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [207] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [208] = {
    [ts_external_token_identifier] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [209] = {
    [ts_external_token_FOR] = true,
    [ts_external_token_ON] = true,
    [ts_external_token_UNTIL] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [210] = {
    [ts_external_token_END] = true,
    [ts_external_token_FOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [211] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_CONFIGURATION] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [212] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ENTITY] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [213] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_IS] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RETURN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [214] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [215] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_COMPONENT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [216] = {
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [217] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ARCHITECTURE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [218] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_CONTEXT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [219] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_PACKAGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [220] = {
    [ts_external_token_ELSE] = true,
    [ts_external_token_ELSIF] = true,
    [ts_external_token_END] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [221] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_BODY] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [222] = {
    [ts_external_token_IS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [223] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [224] = {
    [ts_external_token_WHEN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [225] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALL] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_library_constant] = true,
    [ts_external_token_library_function] = true,
    [ts_external_token_library_type] = true,
  },
  [226] = {
    [ts_external_token_BUS] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [227] = {
    [ts_external_token_BUS] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token_REGISTER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [228] = {
    [ts_external_token_IS] = true,
    [ts_external_token_PARAMETER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [229] = {
    [ts_external_token_BUS] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [230] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALL] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token_character_literal] = true,
    [ts_external_token_operator_symbol] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [231] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [232] = {
    [ts_external_token_FOR] = true,
    [ts_external_token_UNTIL] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [233] = {
    [ts_external_token_IS] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [234] = {
    [ts_external_token_END] = true,
    [ts_external_token_GENERIC] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [235] = {
    [ts_external_token_END] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [236] = {
    [ts_external_token_END] = true,
    [ts_external_token_WHEN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [237] = {
    [ts_external_token_CONFIGURATION] = true,
    [ts_external_token_ENTITY] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token_VUNIT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [238] = {
    [ts_external_token_GENERIC] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [239] = {
    [ts_external_token_BUS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [240] = {
    [ts_external_token_RETURN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [241] = {
    [ts_external_token_IS] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [242] = {
    [ts_external_token_ELSE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [243] = {
    [ts_external_token_FOR] = true,
    [ts_external_token_USE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [244] = {
    [ts_external_token_CONFIGURATION] = true,
    [ts_external_token_ENTITY] = true,
    [ts_external_token_OPEN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [245] = {
    [ts_external_token_PARAMETER] = true,
    [ts_external_token_RETURN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [246] = {
    [ts_external_token_PARAMETER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [247] = {
    [ts_external_token_BEGIN] = true,
    [ts_external_token_END] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [248] = {
    [ts_external_token_IN] = true,
    [ts_external_token_OUT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [249] = {
    [ts_external_token_OF] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [250] = {
    [ts_external_token_directive_body] = true,
    [ts_external_token__directive_newline] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [251] = {
    [ts_external_token_PORT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [252] = {
    [ts_external_token_CONSTANT] = true,
    [ts_external_token_SIGNAL] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [253] = {
    [ts_external_token_END] = true,
    [ts_external_token_PORT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [254] = {
    [ts_external_token__grave_accent] = true,
    [ts_external_token_box] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [255] = {
    [ts_external_token_FOR] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [256] = {
    [ts_external_token_SEVERITY] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [257] = {
    [ts_external_token_SHARED] = true,
    [ts_external_token_VARIABLE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [258] = {
    [ts_external_token_IS] = true,
    [ts_external_token_RANGE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [259] = {
    [ts_external_token_END] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [260] = {
    [ts_external_token_BEGIN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [261] = {
    [ts_external_token__grave_accent] = true,
    [ts_external_token_based_integer] = true,
    [ts_external_token_based_float] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [262] = {
    [ts_external_token_identifier] = true,
    [ts_external_token_ALL] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [263] = {
    [ts_external_token_UNITS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [264] = {
    [ts_external_token_SELECT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [265] = {
    [ts_external_token_IF] = true,
    [ts_external_token__directive_newline] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [266] = {
    [ts_external_token_LOOP] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [267] = {
    [ts_external_token_GENERATE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [268] = {
    [ts_external_token_POSTPONED] = true,
    [ts_external_token_PROCESS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [269] = {
    [ts_external_token_MAP] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [270] = {
    [ts_external_token_BLOCK] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [271] = {
    [ts_external_token__grave_accent] = true,
    [ts_external_token_bit_string_value] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [272] = {
    [ts_external_token_IN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [273] = {
    [ts_external_token__grave_accent] = true,
    [ts_external_token_bit_string_base] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [274] = {
    [ts_external_token__directive_newline] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [275] = {
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token__block_comment_end] = true,
  },
  [276] = {
    [ts_external_token__grave_accent] = true,
    [ts_external_token_string_literal] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [277] = {
    [ts_external_token_BODY] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [278] = {
    [ts_external_token_CASE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [279] = {
    [ts_external_token_AFTER] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [280] = {
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
    [ts_external_token_comment_content] = true,
  },
  [281] = {
    [ts_external_token_PROTECTED] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [282] = {
    [ts_external_token_VARIABLE] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [283] = {
    [ts_external_token_RECORD] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [284] = {
    [ts_external_token_NEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [285] = {
    [ts_external_token_VUNIT] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [286] = {
    [ts_external_token_FUNCTION] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [287] = {
    [ts_external_token_IF] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [288] = {
    [ts_external_token_PROCESS] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [289] = {
    [ts_external_token_THEN] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
  [290] = {
    [ts_external_token_VIEW] = true,
    [ts_external_token__grave_accent] = true,
    [ts_external_token__line_comment_start] = true,
    [ts_external_token__block_comment_start] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_vhdl_external_scanner_create(void);
void tree_sitter_vhdl_external_scanner_destroy(void *);
bool tree_sitter_vhdl_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_vhdl_external_scanner_serialize(void *, char *);
void tree_sitter_vhdl_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_vhdl(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_vhdl_external_scanner_create,
      tree_sitter_vhdl_external_scanner_destroy,
      tree_sitter_vhdl_external_scanner_scan,
      tree_sitter_vhdl_external_scanner_serialize,
      tree_sitter_vhdl_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "vhdl",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 1,
      .minor_version = 4,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
